(PackageInfo named: 'Morphic') postscript: 'TheWorldMainDockingBar updateInstances. "..."'!
(PackageInfo named: 'Morphic') preamble: 'SimpleBorder allSubInstancesDo: [ :each |
	(each instVarNamed: #width) ifNil: [ each instVarNamed: #width put: 0 ].
	(each instVarNamed: #color) ifNil: [ each instVarNamed: #color put: Color transparent ].
	(each instVarNamed: #baseColor) ifNil: [ each instVarNamed: #baseColor put: Color transparent ] ].'!
SystemOrganization addCategory: #'Morphic-Balloon'!
SystemOrganization addCategory: #'Morphic-Basic'!
SystemOrganization addCategory: #'Morphic-Basic-NewCurve'!
SystemOrganization addCategory: #'Morphic-Borders'!
SystemOrganization addCategory: #'Morphic-Collections-Arrayed'!
SystemOrganization addCategory: #'Morphic-Demo'!
SystemOrganization addCategory: #'Morphic-Events'!
SystemOrganization addCategory: #'Morphic-Explorer'!
SystemOrganization addCategory: #'Morphic-Kernel'!
SystemOrganization addCategory: #'Morphic-Layouts'!
SystemOrganization addCategory: #'Morphic-Menus'!
SystemOrganization addCategory: #'Morphic-Menus-DockingBar'!
SystemOrganization addCategory: #'Morphic-Pluggable Widgets'!
SystemOrganization addCategory: #'Morphic-Sound'!
SystemOrganization addCategory: #'Morphic-Sound-Synthesis'!
SystemOrganization addCategory: #'Morphic-Support'!
SystemOrganization addCategory: #'Morphic-Text Support'!
SystemOrganization addCategory: #'Morphic-TrueType'!
SystemOrganization addCategory: #'Morphic-Widgets'!
SystemOrganization addCategory: #'Morphic-Windows'!
SystemOrganization addCategory: #'Morphic-Worlds'!

!TextURL methodsFor: '*Morphic' stamp: 'rbb 2/18/2005 09:24'!
actOnClickFor: anObject
	"Do what you can with this URL.  Later a web browser."

	| response m |

	(url beginsWith: 'sqPr://') ifTrue: [
		ProjectLoading thumbnailFromUrl: (url copyFrom: 8 to: url size).
		^self		"should not get here, but what the heck"
	].
	"if it's a web browser, tell it to jump"
	anObject isWebBrowser
		ifTrue: [anObject jumpToUrl: url. ^ true]
		ifFalse: [((anObject respondsTo: #model) and: [anObject model isWebBrowser])
				ifTrue: [anObject model jumpToUrl: url. ^ true]].

		"if it's a morph, see if it is contained in a web browser"
		(anObject isKindOf: Morph) ifTrue: [
			m := anObject.
			[ m ~= nil ] whileTrue: [
				(m isWebBrowser) ifTrue: [
					m  jumpToUrl: url.
					^true ].
				(m hasProperty: #webBrowserView) ifTrue: [
					m model jumpToUrl: url.
					^true ].
				m := m owner. ]
		].

	"no browser in sight.  ask if we should start a new browser"
	((self confirm: 'open a browser to view this URL?' translated) and: [WebBrowser default notNil]) ifTrue: [
		WebBrowser default openOnUrl: url.
		^ true ].

	"couldn't display in a browser.  Offer to put up just the source"

	response := (UIManager default 
				chooseFrom: (Array with: 'View web page as source' translated
									with: 'Cancel' translated)
				title:  'Couldn''t find a web browser. View\page as source?' withCRs translated).
	response = 1 ifTrue: [HTTPSocket httpShowPage: url].
	^ true! !

!AbstractSound class methodsFor: '*Morphic-Sounds-sound library-file in/out' stamp: 'mt 4/28/2016 14:29'!
updateScorePlayers
	| soundsBeingEdited |
	"Force all ScorePlayers to update their instrument list from the sound library. This may done after loading, unloading, or replacing a sound to make all ScorePlayers feel the change."

	ScorePlayer allSubInstancesDo:
		[:p | p pause].
	SoundPlayer shutDown: true.
	soundsBeingEdited := EnvelopeEditorMorph allSubInstances collect: [:ed | ed soundBeingEdited].
	ScorePlayerMorph allSubInstancesDo:
		[:p | p updateInstrumentsFromLibraryExcept: soundsBeingEdited].
! !

!SampledSound methodsFor: '*Morphic-Sounds-sound tracks' stamp: 'nice 12/27/2009 03:10'!
sonogramMorph: height from: start to: stop nPoints: nPoints
	"FYI:  It is very cool that we can do this, but for sound tracks on a movie,
	simple volume is easier to read, easier to scale, and way faster to compute.
	Code preserved here just in case it makes a useful example."
	"In an inspector of a samplesSound...
		self currentWorld addMorph: (self sonogramMorph: 32 from: 1 to: 50000 nPoints: 256)
	"
	| fft sonogramMorph width |
	fft := FFT new: nPoints.
	width := stop-start//nPoints.
	sonogramMorph := Sonogram new
			extent: width@height
			minVal: 0.0
			maxVal: 1.0
			scrollDelta: width.
	start to: stop-nPoints by: nPoints do:
		[:i | | data |
		data := fft transformDataFrom: samples startingAt: i.
		data := data collect: [:v | v sqrt].  "square root compresses dynamic range"
		data /= 200.0.
		sonogramMorph plotColumn: data].
	^ sonogramMorph
	
! !

TextAttribute subclass: #TextAnchor
	instanceVariableNames: 'anchoredMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextAnchor commentStamp: 'md 8/10/2006 11:52' prior: 0!
TextAnchors support anchoring of images in text.  A TextAnchor exists as an attribute of text emphasis, and it gets control like a FontReference, through the emphasizeScanner: message.  Depending on whether its anchoredMorph is a Morph or a Form, it repositions the morph, or displays the form respectively.  The coordination between composition, display and selection can best be understood by browsing the various implementations of placeEmbeddedObject:.

In the morphic world, simply embed any form or morph in text.

	Workspace new
		contents: (Text withAll: 'foo') , (Text string: '*' attribute: (TextAnchor new anchoredMorph: MenuIcons confirmIcon)) , (Text withAll: 'bar');
		openLabel: 'Text with Form'.

	Workspace new
		contents: (Text withAll: 'foo') , (Text string: '*' attribute: (TextAnchor new anchoredMorph: EllipseMorph new)) , (Text withAll: 'bar');
		openLabel: 'Text with Morph'.

In this case you select a piece of the screen, and it gets anchored to a one-character text in the editor's past buffer.  If you then paste into some other text, you will see the image as an embedded image.!

!TextAnchor methodsFor: 'accessing' stamp: 'di 11/10/97 13:21'!
anchoredMorph
	^ anchoredMorph! !

!TextAnchor methodsFor: 'accessing' stamp: 'di 11/10/97 10:47'!
anchoredMorph: aMorph 
	anchoredMorph := aMorph! !

!TextAnchor methodsFor: 'accessing' stamp: 'sw 11/9/1999 17:21'!
couldDeriveFromPrettyPrinting
	^ false! !

!TextAnchor methodsFor: 'visiting' stamp: 'ar 12/17/2001 01:19'!
emphasizeScanner: scanner
	"Do nothing for emphasizing the scanner - if the anchor is valid a #embeddedObject will be encountered by the scanner and do the real thing"! !

!TextAnchor methodsFor: 'comparing' stamp: 'ar 9/9/2003 22:03'!
hash
	"#hash is re-implemented because #= is re-implemented"
	^anchoredMorph identityHash! !

!TextAnchor methodsFor: 'accessing' stamp: 'di 11/10/97 14:08'!
mayBeExtended
	"A textAnchor is designed to modify only a single character, and therefore must not be extended by the ParagraphEditor's emphasisHere facility"
	^ false! !

!TextAnchor methodsFor: 'comparing' stamp: 'di 7/1/1998 14:35'!
= other 
	^ (other class == self class) 
		and: [other anchoredMorph == anchoredMorph]! !

!SolidFillStyle methodsFor: '*Morphic-Balloon' stamp: 'topa 10/20/2015 00:49:34'!
addFillStyleMenuItems: aMenu hand: aHand from: aMorph
	^ self asColor addFillStyleMenuItems: aMenu hand: aHand from: aMorph! !

!OrderedCollection methodsFor: '*Morphic-Explorer' stamp: 'bf 8/20/1999 15:08'!
hasContentsInExplorer

	^self isEmpty not! !

!SmartRefStream methodsFor: '*Morphic-conversion' stamp: 'jm 5/21/1998 06:44'!
bookPageMorphbosfcepcbbfgcc0
	"BookPageMorph->PasteUpMorph. For reading in old BookMorphs."

	^ PasteUpMorph
! !

!SmartRefStream methodsFor: '*Morphic-conversion' stamp: 'di 5/22/1998 15:03'!
clippingMorphbosfcep0
	^ PasteUpMorph! !

!SmartRefStream methodsFor: '*Morphic-conversion' stamp: 'jm 5/21/1998 06:44'!
clippingMorphbosfcepc0
	"ClippingMorph->PasteUpMorph. For reading in old BookMorphs."

	^ PasteUpMorph! !

!SmartRefStream methodsFor: '*Morphic-conversion' stamp: 'tk 11/3/2000 18:47'!
dropShadowMorphbosfces0

	^ Morph ! !

!SmartRefStream methodsFor: '*Morphic-conversion' stamp: 'di 5/21/1998 19:24'!
layoutMorphbosfcepbbochvimol0
	^ AlignmentMorph! !

!SmartRefStream methodsFor: '*Morphic-conversion' stamp: 'tk 5/12/1998 16:18'!
layoutMorphbosfcepcbbochvimol0
	^ AlignmentMorph! !

!SmartRefStream methodsFor: '*Morphic-conversion' stamp: 'ar 10/26/2000 01:55'!
morphicEventtcbks0
	^ MorphicEvent! !

!SmartRefStream methodsFor: '*Morphic-conversion' stamp: 'ar 10/26/2000 00:48'!
morphicSoundEventtcbkss0
	^ MorphicUnknownEvent! !

!SmartRefStream methodsFor: '*Morphic-conversion' stamp: 'ar 7/8/2001 17:11'!
myMorphbosfce0

	reshaped at: #MyMorph put: #convertbosfce0:bosfce0:.
		"Be sure to define that conversion method in class Morph"
	^ Morph! !

!SmartRefStream methodsFor: '*Morphic-conversion' stamp: 'RAA 10/26/2000 09:43'!
newMorphicEventts0

	^ MorphicEvent! !

Model subclass: #AbstractHierarchicalList
	instanceVariableNames: 'currentSelection myBrowser'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Explorer'!

!AbstractHierarchicalList commentStamp: '<historical>' prior: 0!
Contributed by Bob Arning as part of the ObjectExplorer package.
!

!AbstractHierarchicalList methodsFor: 'as yet unclassified' stamp: 'RAA 6/21/1999 15:22'!
genericMenu: aMenu

	aMenu add: 'no menu yet' target: self selector: #yourself.
	^aMenu! !

!AbstractHierarchicalList methodsFor: 'as yet unclassified' stamp: 'RAA 4/7/1999 16:44'!
getCurrentSelection

	^currentSelection! !

!AbstractHierarchicalList methodsFor: 'as yet unclassified' stamp: 'RAA 4/7/1999 16:46'!
noteNewSelection: x

	currentSelection := x.
	self changed: #getCurrentSelection.
	currentSelection ifNil: [^self].
	currentSelection sendSettingMessageTo: self.
! !

!AbstractHierarchicalList methodsFor: 'as yet unclassified' stamp: 'RAA 4/7/1999 16:53'!
perform: selector orSendTo: otherTarget
	"Selector was just chosen from a menu by a user.  If can respond, then
perform it on myself. If not, send it to otherTarget, presumably the
editPane from which the menu was invoked."

	(self respondsTo: selector)
		ifTrue: [^ self perform: selector]
		ifFalse: [^ otherTarget perform: selector]! !

!AbstractHierarchicalList methodsFor: 'as yet unclassified' stamp: 'RAA 4/7/1999 16:47'!
update: aSymbol

	aSymbol == #hierarchicalList ifTrue: [
		^self changed: #getList
	].
	super update: aSymbol! !

Model subclass: #FontChooserTool
	instanceVariableNames: 'title selectedFontIndex fontList target getSelector setSelector pointSize emphasis window result offerStyleList'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Morphic-Support'!

!FontChooserTool commentStamp: 'ar 8/30/2009 14:28' prior: 0!
A ToolBuilder version of FreeTypePlus' FontChooser[Morph].!

!FontChooserTool class methodsFor: 'accessing' stamp: 'ar 8/30/2009 17:02'!
default
	"Answer the default font chooser tool"
	^Default ifNil:[self]! !

!FontChooserTool class methodsFor: 'accessing' stamp: 'ar 8/30/2009 17:02'!
default: aFontChooser
	"Answer the default font chooser tool"
	Default := aFontChooser! !

!FontChooserTool class methodsFor: 'opening' stamp: 'ar 8/30/2009 15:05'!
open
	"
		FontChooserTool open.
	"
	^self new open! !

!FontChooserTool class methodsFor: 'opening' stamp: 'ar 8/30/2009 15:20'!
openWithWindowTitle: titleString for: anObject setSelector: setSelector getSelector: getSelector
	"
		FontChooserTool 
			openWithWindowTitle: 'Choose the Menu Font' 
			for: Preferences 
			setSelector: #setMenuFontTo: 
			getSelector: #standardMenuFont.
	"
	^(self withTitle: titleString for: anObject setSelector: setSelector getSelector: getSelector) open! !

!FontChooserTool class methodsFor: 'opening' stamp: 'ar 8/30/2009 15:06'!
windowTitle: titleString for: anObject setSelector: setSelector getSelector: getSelector
	| instance |
	
	instance := self new.
	instance 
		title: titleString;
		target: anObject;
		setSelector: setSelector;
		getSelector: getSelector.
	^instance open! !

!FontChooserTool class methodsFor: 'opening' stamp: 'ar 8/30/2009 15:20'!
withTitle: titleString for: anObject setSelector: setSelector getSelector: getSelector
	"
		(FontChooserTool 
			withTitle: 'Choose the Menu Font' 
			for: Preferences 
			setSelector: #setMenuFontTo: 
			getSelector: #standardMenuFont) open.
	"
	| instance |
	instance := self new.
	instance 
		title: titleString;
		target: anObject;
		setSelector: setSelector;
		getSelector: getSelector.
	^instance! !

!FontChooserTool methodsFor: 'actions' stamp: 'ar 8/30/2009 15:03'!
accept
	self apply.
	result := self selectedFont.
	ToolBuilder default close: window.! !

!FontChooserTool methodsFor: 'actions' stamp: 'ar 8/30/2009 14:27'!
apply
	| font |
	target ifNotNil:[
		setSelector ifNotNil:[
			font := self selectedFont.
			font ifNotNil:[
				target perform: setSelector with: font]]].! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'mt 4/27/2015 16:09'!
buildButtonBarWith: builder
	"Build the button bar"
	| panelSpec buttonSpec |
	panelSpec := builder pluggablePanelSpec new.
	panelSpec
		layout: #horizontal;
		children: OrderedCollection new.

	buttonSpec := builder pluggableButtonSpec new.
	buttonSpec 
			model: self;
			label: 'Apply' translated; 
			action: #apply.
	panelSpec children addLast: buttonSpec.


	buttonSpec := builder pluggableButtonSpec new.
	buttonSpec 
			model: self;
			label: 'OK' translated; 
			action: #accept.
	panelSpec children addLast: buttonSpec.

	buttonSpec := builder pluggableButtonSpec new.
	buttonSpec 
			model: self;
			label: 'Cancel' translated; 
			action: #cancel.
	panelSpec children addLast: buttonSpec.

	^panelSpec! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'kfr 4/26/2015 20:13'!
buildFontListWith: builder
	"Build the font choosers list of font names"
	| listSpec |
	listSpec := builder pluggableListSpec new.
	listSpec 
		model: self;
		name: #fontList;
		list: #fontList; 
		getIndex: #selectedFontIndex; 
		setIndex: #selectedFontIndex:.
	^listSpec
	! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'ar 8/30/2009 14:38'!
buildPointSizeListWith: builder
	"Build the font choosers list of point sizes"
	| listSpec |
	listSpec := builder pluggableListSpec new.
	listSpec 
		model: self;
		list: #pointSizeList; 
		getIndex: #selectedPointSizeIndex; 
		setIndex: #selectedPointSizeIndex:.
	^listSpec
! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'mt 4/27/2015 15:47'!
buildPreviewPaneWith: builder
	"Build the preview panel"
	| textSpec |
	textSpec := builder pluggableTextSpec new.
	textSpec 
		name: #preview;
		model: self;
		getText: #contents;
		softLineWrap: false.
	^textSpec! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'ar 8/30/2009 16:25'!
buildStyleListWith: builder
	"Build the font choosers list of font styles"
	| listSpec |
	listSpec := builder pluggableListSpec new.
	listSpec 
		model: self;
		list: #fontStyleList; 
		getIndex: #selectedFontStyleIndex; 
		setIndex: #selectedFontStyleIndex:.
	^listSpec
! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'ar 8/30/2009 14:47'!
buildWindowWith: builder
	| windowSpec |
	windowSpec := builder pluggableWindowSpec new.
	windowSpec model: self.
	windowSpec label: #windowTitle.
	windowSpec children: OrderedCollection new.
	^windowSpec! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'nice 12/27/2009 03:10'!
buildWindowWith: builder specs: specs
	| windowSpec |
	windowSpec := self buildWindowWith: builder.
	specs do:[:assoc| | action widgetSpec rect |
		rect := assoc key.
		action := assoc value.
		widgetSpec := action value.
		widgetSpec ifNotNil:[
			widgetSpec frame: rect.
			windowSpec children add: widgetSpec]].
	^windowSpec! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'mt 4/27/2015 16:08'!
buildWith: builder
	"Create the ui for the browser"
	"ToolBuilder open: self"
	| windowSpec |
	self offerStyleList ifTrue:[
		windowSpec := self buildWindowWith: builder specs: {
			(0@0 corner: 0.4@0.4) -> [self buildFontListWith: builder].
			(0.4@0 corner: 0.8@0.4) -> [self buildStyleListWith: builder].
			(0.8@0 corner: 1.0@0.4) -> [self buildPointSizeListWith: builder].
			(LayoutFrame fractions: (0.0@0.4 corner: 1.0@1.0) offsets: (0@0 corner: 0 @ self buttonHeight negated)) -> [self buildPreviewPaneWith: builder].
			(LayoutFrame fractions: (0@1 corner: 1@1) offsets: (0@ self buttonHeight negated corner: 0@0)) -> [self buildButtonBarWith: builder].
		}.
	] ifFalse:[
		windowSpec := self buildWindowWith: builder specs: {
			(0@0 corner: 0.7@0.4) -> [self buildFontListWith: builder].
"			(0.4@0 corner: 0.8@0.4) -> [self buildStyleListWith: builder]."
			(0.7@0 corner: 1.0@0.4) -> [self buildPointSizeListWith: builder].
			(LayoutFrame fractions: (0.0@0.4 corner: 1.0@1.0) offsets: (0@0 corner: 0 @ self buttonHeight negated)) -> [self buildPreviewPaneWith: builder].
			(LayoutFrame fractions: (0@1 corner: 1@1) offsets: (0@ self buttonHeight negated corner: 0@0)) -> [self buildButtonBarWith: builder].
		}.
	].
	windowSpec extent: self initialExtent.
	window := builder build: windowSpec.

	"Now that the window has been built, notify selection again to scroll it into view."
	self changed: #selectedFontIndex.
	^window! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'mt 4/27/2015 16:05'!
buttonHeight
	^ Preferences standardButtonFont height * 3! !

!FontChooserTool methodsFor: 'actions' stamp: 'ar 8/30/2009 15:04'!
cancel
	result := nil.
	ToolBuilder default close: window.! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'topa 6/1/2016 11:55'!
contents

	^ self selectedFont
		ifNil: [Text new]
		ifNotNil: [:font | font sampleText]! !

!FontChooserTool methodsFor: 'font list' stamp: 'topa 3/9/2015 23:10'!
fontList
	"List of available font family names"
	^fontList ifNil:[fontList := TextStyle knownTextStyles]! !

!FontChooserTool methodsFor: 'style list' stamp: 'ar 8/30/2009 16:15'!
fontStyleList
	"names of simulated styles are enclosed in parenthesis"
	^#('Regular' 'Bold' 'Italic' 'Bold Italic')! !

!FontChooserTool methodsFor: 'accessing' stamp: 'ar 8/30/2009 14:27'!
getSelector
	"Answer the value of getSelector"

	^ getSelector! !

!FontChooserTool methodsFor: 'accessing' stamp: 'ar 8/30/2009 14:27'!
getSelector: aSelectorSymbolOrFont
	"Set the value of getSelector"

	getSelector := aSelectorSymbolOrFont! !

!FontChooserTool methodsFor: 'initialize' stamp: 'mt 4/27/2015 15:52'!
initialExtent
	^self offerStyleList ifTrue:[420@300] ifFalse:[320@200].! !

!FontChooserTool methodsFor: 'initialize' stamp: 'ar 9/26/2009 20:07'!
initialize
	super initialize.
	title := 'Choose A Font'.
	getSelector := TextStyle defaultFont.
	emphasis := 0.
	offerStyleList := true.! !

!FontChooserTool methodsFor: 'initialize' stamp: 'ar 9/26/2009 20:07'!
offerStyleList
	"Whether to offer a choice of styles with the font."
	^offerStyleList! !

!FontChooserTool methodsFor: 'initialize' stamp: 'ar 9/26/2009 20:08'!
offerStyleList: aBool
	"Whether to offer a choice of styles with the font."
	offerStyleList := aBool! !

!FontChooserTool methodsFor: 'toolbuilder' stamp: 'ar 8/30/2009 15:06'!
open
	^ToolBuilder open: self! !

!FontChooserTool methodsFor: 'point size' stamp: 'ar 8/30/2009 14:27'!
pointSize
	^pointSize ifNil: [pointSize := 10.0]! !

!FontChooserTool methodsFor: 'point size' stamp: 'ar 9/26/2009 20:03'!
pointSizeList
	^self selectedTextStyle pointSizes collect: [:each | each asString padded: #left to: 3 with: $ ]! !

!FontChooserTool methodsFor: 'point size' stamp: 'ar 8/30/2009 14:51'!
pointSize: aNumber
	pointSize := aNumber.
	self changed: #pointSize.
	self changed: #contents.! !

!FontChooserTool methodsFor: 'accessing' stamp: 'ar 8/30/2009 15:21'!
result
	^result! !

!FontChooserTool methodsFor: 'font list' stamp: 'ar 9/26/2009 20:03'!
selectedFont
	| font |
	font := self selectedTextStyle fontOfPointSize: pointSize.
	^font emphasized: emphasis! !

!FontChooserTool methodsFor: 'font list' stamp: 'ar 8/30/2009 16:24'!
selectedFontFamily
	^self fontList at: self selectedFontIndex ifAbsent:[nil].
	
	! !

!FontChooserTool methodsFor: 'font list' stamp: 'ar 8/30/2009 16:22'!
selectedFontIndex
	| font textStyleName family |
	selectedFontIndex ifNotNil: [^selectedFontIndex].
	selectedFontIndex := 0.
	font := (getSelector isSymbol and:[target notNil])
		ifTrue:[target perform: getSelector]
		ifFalse:[getSelector].
	font ifNotNil:[
		emphasis := font emphasis.
		pointSize := font pointSize.
		textStyleName := font textStyleName.
		family := self fontList detect:[:f | f = textStyleName] ifNone:[].
	].
	selectedFontIndex := self fontList indexOf: family ifAbsent:[0].
	self selectedFontIndex: selectedFontIndex.
	^selectedFontIndex! !

!FontChooserTool methodsFor: 'font list' stamp: 'ar 8/30/2009 16:24'!
selectedFontIndex: anIndex
	anIndex = 0 ifTrue: [^self].
	selectedFontIndex := anIndex.
	self changed: #selectedFontIndex.
	self changed: #selectedFontStyleIndex.
	self changed: #pointSizeList.
	self changed: #pointSizeIndex.
	self changed: #contents.! !

!FontChooserTool methodsFor: 'style list' stamp: 'ar 8/30/2009 16:25'!
selectedFontStyleIndex
	"This is a hack"
	^emphasis+1! !

!FontChooserTool methodsFor: 'style list' stamp: 'ar 8/30/2009 16:26'!
selectedFontStyleIndex: anIndex
	anIndex = 0 ifTrue: [^self].
	emphasis := anIndex - 1.
	self changed: #selectedFontStyleIndex.
	self changed: #contents.! !

!FontChooserTool methodsFor: 'point size' stamp: 'ar 8/30/2009 14:27'!
selectedPointSize
	^self selectedFont pointSize! !

!FontChooserTool methodsFor: 'point size' stamp: 'ar 8/30/2009 14:27'!
selectedPointSizeIndex
	^self pointSizeList indexOf: (pointSize reduce asString padded: #left to: 3 with: $ )! !

!FontChooserTool methodsFor: 'point size' stamp: 'ar 8/30/2009 16:27'!
selectedPointSizeIndex: anIndex

	anIndex = 0 ifTrue: [^self].
	pointSize := (self pointSizeList at: anIndex) withBlanksTrimmed asNumber.
	self changed: #pointSizeList.
	self changed: #contents.! !

!FontChooserTool methodsFor: 'font list' stamp: 'ar 9/26/2009 20:02'!
selectedTextStyle

	^TextStyle named: (self selectedFontFamily ifNil:[^TextStyle default]).! !

!FontChooserTool methodsFor: 'accessing' stamp: 'ar 8/30/2009 14:27'!
setSelector: anObject
	"Set the value of setSelector"

	setSelector := anObject! !

!FontChooserTool methodsFor: 'accessing' stamp: 'ar 8/30/2009 14:27'!
target
	"Answer the value of target"

	^ target! !

!FontChooserTool methodsFor: 'accessing' stamp: 'ar 8/30/2009 14:27'!
target: anObject
	"Set the value of target"

	target := anObject! !

!FontChooserTool methodsFor: 'accessing' stamp: 'ar 8/30/2009 14:27'!
title: anObject
	"Set the value of title"

	title := anObject! !

!FontChooserTool methodsFor: 'accessing' stamp: 'ar 8/30/2009 15:19'!
window
	^window! !

!FontChooserTool methodsFor: 'initialize' stamp: 'ar 8/30/2009 14:27'!
windowTitle
	^ title translated! !

Model subclass: #FontImporterTool
	instanceVariableNames: 'title allFonts emphasis currentSelection currentParent warningSeen'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!FontImporterTool commentStamp: 'topa 3/9/2015 18:56' prior: 0!
A tool to import platform (native) fonts into the image!

!FontImporterTool class methodsFor: 'class initialization' stamp: 'topa 3/19/2015 11:24'!
initialize

	self registerInOpenMenu.! !

!FontImporterTool class methodsFor: 'opening' stamp: 'topa 3/7/2015 00:50'!
open
	"
		FontChooserTool open.
	"
	^self new open! !

!FontImporterTool class methodsFor: 'class initialization' stamp: 'topa 3/19/2015 11:24'!
registerInOpenMenu
	(TheWorldMenu respondsTo: #registerOpenCommand:) ifTrue: [
		TheWorldMenu unregisterOpenCommand: 'Font Importer'.
		TheWorldMenu registerOpenCommand: {'Font Importer'. {self. #open}}].
		! !

!FontImporterTool class methodsFor: 'class initialization' stamp: 'topa 3/19/2015 11:24'!
unload

	self unregisterFromOpenMenu.! !

!FontImporterTool class methodsFor: 'class initialization' stamp: 'topa 3/19/2015 11:23'!
unregisterFromOpenMenu
	 (TheWorldMenu respondsTo: #registerOpenCommand:)
		ifTrue: [TheWorldMenu unregisterOpenCommand: 'Font Importer'].
! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 3/19/2015 10:43'!
allFonts
	^ allFonts ifNil: [ | fonts |
		fonts := Dictionary new.
		Cursor wait showWhile: [
			TTFileDescription fontPathsDo:[:path |
				TTFileDescription fontFilesIn: path do:[:font| | fontDesc filename fname |
					filename := path, FileDirectory slash, font fileName.
					fname := self textForFamily: font familyName subfamily: nil.
					fontDesc := fonts 
						at: font familyName
						ifAbsentPut: (FontImporterFontDescription new fontname: fname; yourself).
					font subfamilyName
						ifNil: [fontDesc filename: filename]
						ifNotNil: [ |subfontDesc sname | 
							sname := self textForFamily: font familyName subfamily: font subfamilyName.
							subfontDesc := FontImporterFontDescription new fontname: sname; yourself.
							subfontDesc
								parent: fontDesc;
								filename: filename.
							fontDesc addChild: subfontDesc]]]].
		allFonts := fonts values sorted.
		allFonts do: [:fontDesc | fontDesc normalize].
		allFonts].

		! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 3/19/2015 11:00'!
allFonts: anObject

	allFonts := anObject.
	self changed: #allFonts.! !

!FontImporterTool methodsFor: 'toolbuilder' stamp: 'mt 4/28/2015 13:12'!
buildButtonBarWith: builder
	"Build the button bar"
	| panelSpec buttonSpec |
	panelSpec := builder pluggablePanelSpec new.
	panelSpec
		layout: #horizontal;
		children: OrderedCollection new.

	buttonSpec := builder pluggableButtonSpec new
			model: self;
			label: 'Import' translated; 
			help: 'Include the font data in the image and provide a TextStyle for the font';
			action: #import;
			yourself.
	panelSpec children addLast: buttonSpec.


	buttonSpec := builder pluggableButtonSpec new
			model: self;
			label: 'Close' translated; 
			action: #close;
			yourself.
	panelSpec children addLast: buttonSpec.


	^panelSpec! !

!FontImporterTool methodsFor: 'toolbuilder' stamp: 'topa 3/24/2015 11:46'!
buildFontListWith: builder
	"Build the font choosers list of font names"
	
	^ builder pluggableTreeSpec new
		model: self;
		roots: #allFonts; 
		label: #labelOf: ;
		getChildren: #childrenOf: ;
		getSelected: #currentSelection;
		setSelected: #currentSelection:;
		setSelectedParent: #currentParent:;
		menu: #fontListMenu:;
		autoDeselect: false;
		yourself
! !

!FontImporterTool methodsFor: 'toolbuilder' stamp: 'mt 4/28/2015 13:24'!
buildPreviewPaneWith: builder
	"Build the preview panel"
	
	^ builder pluggablePanelSpec new
		wantsResizeHandles: true;
		children: {
			builder pluggableInputFieldSpec new
				model: self;
				getText: #filename;
				readOnly: true;
				indicateUnacceptedChanges: false;
				font: self filenameFont;
				frame: (LayoutFrame 
					fractions: (0@0 corner: 1@0)
					offsets: (0@0 corner: 0@ self filenameHeight));
				yourself.

			builder pluggableTextSpec new
				model: self;
				getText: #previewText;
				askBeforeDiscardingEdits: false;
				indicateUnacceptedChanges: false;
				softLineWrap: false;
				frame: (LayoutFrame 
					fractions: (0@0 corner: 1@0.75)
					offsets: (0@ self filenameHeight corner: 0@0));
				yourself.
				
			builder pluggableTextSpec new
				model: self;
				getText: #copyright;
				readOnly: true;
				indicateUnacceptedChanges: false;
				frame: (LayoutFrame 
					fractions: (0@0.75 corner: 1@1));
				yourself
			
		};
		yourself! !

!FontImporterTool methodsFor: 'toolbuilder' stamp: 'mt 4/28/2015 13:24'!
buildWindowWith: builder specs: specs
	| windowSpec |
	windowSpec := builder pluggableWindowSpec new.
	windowSpec
		model: self;
		label:'Choose a Font to import' translated;
		extent: self initialExtent;
		children: OrderedCollection new.
		
	specs do:[:assoc|
		| rect action widgetSpec |
		rect := assoc key.
		action := assoc value.
		widgetSpec := action value.
		widgetSpec ifNotNil:[
			widgetSpec frame: rect.
			windowSpec children add: widgetSpec]].
	^windowSpec! !

!FontImporterTool methodsFor: 'toolbuilder' stamp: 'mt 4/28/2015 13:20'!
buildWith: builder
	"Create the ui for the browser"
	"ToolBuilder open: self"
	
	| windowSpec |
	windowSpec := self buildWindowWith: builder specs: {
		(self fontListFrame) -> [self buildFontListWith: builder].
		(self previewFrame) -> [self buildPreviewPaneWith: builder].
		(self buttonsFrame) -> [self buildButtonBarWith: builder].
	}.
	^ builder build: windowSpec! !

!FontImporterTool methodsFor: 'layout' stamp: 'topa 4/26/2015 23:34'!
buttonHeight
	^ Preferences standardButtonFont height * 3! !

!FontImporterTool methodsFor: 'layout' stamp: 'topa 3/10/2015 00:09'!
buttonsFrame

	^ LayoutFrame
		fractions: (0@1 corner: 1@1)
		offsets: (0@ self buttonHeight negated corner: 0@0)
! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 3/9/2015 12:59'!
childrenOf: aFontDescription

	^ aFontDescription children! !

!FontImporterTool methodsFor: 'actions' stamp: 'mt 4/28/2015 13:00'!
close
	self changed: #close.! !

!FontImporterTool methodsFor: 'model access' stamp: 'topa 3/10/2015 00:38'!
copyright
	| f |
	f := self selectedFont ifNil:[^ ''].
	^ f isTTCFont
		ifTrue: [f ttcDescription copyright ifNil: ['']]
		ifFalse: ['']! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 3/9/2015 13:06'!
currentParent

	^ currentParent! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 3/10/2015 01:59'!
currentParent: anObject

	anObject = currentParent ifTrue: [^ self].
	currentParent := anObject.
	self changed: #currentParent.
! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 3/9/2015 19:19'!
currentSelection

	^ currentSelection! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 8/21/2015 14:12'!
currentSelection: anObject

	anObject = currentSelection ifTrue: [^ self].
	currentSelection := anObject.
	self changed: #currentSelection.
	self changed: #previewText.
	self changed: #filename.
	self changed: #copyright.! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 3/9/2015 12:51'!
emphasis

	^ emphasis! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 3/9/2015 12:51'!
emphasis: anObject

	emphasis := anObject! !

!FontImporterTool methodsFor: 'model access' stamp: 'topa 3/10/2015 00:32'!
filename

	^ self currentSelection
		ifNil: ['']
		ifNotNil: [:sel |
			String streamContents: [:stream |
				sel allFilenames
					do: [:filename | stream nextPutAll: filename]
					separatedBy: [stream nextPut: $,;nextPut: $ ]]]! !

!FontImporterTool methodsFor: 'toolbuilder' stamp: 'topa 4/26/2015 23:59'!
filenameFont
	^ Preferences standardDefaultTextFont! !

!FontImporterTool methodsFor: 'layout' stamp: 'mt 4/30/2015 10:29'!
filenameHeight
	^ self filenameFont height * 2! !

!FontImporterTool methodsFor: 'helper' stamp: 'topa 3/9/2015 21:27'!
fontFromFamily: aFamily

	| readFonts | 
	aFamily ifNil: [^ TextStyle default fonts first].
	readFonts := TTFileDescription readFontsFrom: aFamily allFilenames anyOne.
	^ (readFonts size > 1
		ifTrue: [ 
			| ftArray |
			" see TTCFontSet>>newTextStyleFromTT: "
			ftArray := readFonts collect: [:ttc | |f|
				ttc ifNil: [nil] ifNotNil: [
					f := TTCFont new.
					f ttcDescription: ttc.
					f pointSize: 11.0 .
					f]].
			TTCFontSet newFontArray: ftArray]
		ifFalse: [ |f|
			f := TTCFont new.
			f ttcDescription: readFonts anyOne.
			f pointSize: 11.0 .	
			f])! !

!FontImporterTool methodsFor: 'layout' stamp: 'mt 4/28/2015 13:04'!
fontListFrame

	^ LayoutFrame
		fractions: (0@0 corner: 0.4@1)
		offsets: (0@0 corner: 0@ self buttonHeight negated)! !

!FontImporterTool methodsFor: 'font list' stamp: 'topa 3/24/2015 11:56'!
fontListMenu: aMenu

	^ aMenu addTranslatedList: #(
		('Import Font'	import	'Include the font data in the image and provide a TextStyle for the font')
		('Link Font'		link  'Install the font as a link to its file and provide a TextStyle for the referenced font'))
	yourself! !

!FontImporterTool methodsFor: 'helper' stamp: 'topa 4/27/2015 00:09'!
font: f hasGlyphOf: aCharacter

	| font |
	font := f isFontSet ifTrue: [f fontArray first] ifFalse: [f].
	^ font isTTCFont
		ifFalse: [font hasGlyphOf: aCharacter]
		ifTrue: [
			" [(f hasGlyphOf: aCharacter) not] does not work, the fallback glyph is always found instead.
			So we fake. if aCharacter is the same form as Character null aka 0, we assume absence."
			(font characterFormAt: aCharacter) bits ~= font fallbackForm bits]
! !

!FontImporterTool methodsFor: 'actions' stamp: 'topa 3/19/2015 11:00'!
import
	| megaSize filenames fonts |
	fonts := self currentSelection.
	filenames := fonts allFilenames.
	megaSize := ((filenames inject: 0 into: [ :sum :fn |
		sum + (FileStream readOnlyFileNamed: fn do: [:file | file size])]) / (1024 * 1024)) asFloat.
	(UIManager default confirm: (
'About to import {1}{2}.\\This is at least {3} MB of space required int the image.\
Please respect the copyright and embedding restrictions of the font.\
Proceed?' 
		withCRs format: {
			self currentParent 
				ifNotNil: [:p| p fontname, ' ', self currentSelection fontname]
				ifNil: [self currentSelection fontname].
			filenames size > 1 ifTrue: [' (', filenames size, ' font files)'] ifFalse: [''].
			megaSize printShowingDecimalPlaces: 2}))
		ifTrue: [ 
			filenames do: [:filename | | readFonts |
				readFonts := TTCFontDescription addFromTTFile: filename.
				readFonts isCollection
					ifFalse: [TTCFont newTextStyleFromTT: readFonts]
					ifTrue: [self importFontFamily: readFonts]]].
	self allFonts: nil. "force redraw"
! !

!FontImporterTool methodsFor: 'helper' stamp: 'topa 3/19/2015 11:13'!
importFontFamily: readFonts

	|r rest array |
	r := readFonts detect: [:f | 
		[f isRegular] on: Error do: [false] "hack for unknown emphases"
	] ifNone: [^ TTCFont newTextStyleFromTT: readFonts first].
	rest := readFonts copyWithout: r.
	array :=TTCFont pointSizes collect: [:pt | | f | 
		f := TTCFont new ttcDescription: r; pointSize: pt; yourself.
		rest do: [:rf |
			(self isStyleNameSupported: rf subfamilyName)
				ifTrue: [f derivativeFont: (TTCFont new ttcDescription: rf; pointSize: pt; yourself)]
				ifFalse: [
					Transcript show: 'Cannot import unknown style ', rf subfamilyName, ' from Font family ', f name]]. 
		f].
	^ TTCFont reorganizeForNewFontArray: array name: array first name asSymbol.! !

!FontImporterTool methodsFor: 'initialize' stamp: 'topa 3/10/2015 00:37'!
initialExtent

	^ 600@400.! !

!FontImporterTool methodsFor: 'initialize' stamp: 'mt 4/30/2015 11:05'!
initialize

	super initialize.
	emphasis := 0.
! !

!FontImporterTool methodsFor: 'helper' stamp: 'topa 3/19/2015 10:04'!
isStyleNameSupported: subfamilyName

	^ (TextStyle decodeStyleName: subfamilyName) second isEmpty! !

!FontImporterTool methodsFor: 'model access' stamp: 'topa 3/9/2015 18:28'!
labelOf: aFontDescription

	^ aFontDescription fontname

	! !

!FontImporterTool methodsFor: 'actions' stamp: 'topa 3/24/2015 11:56'!
link
	| filenames fonts |
	fonts := self currentSelection.
	self warningSeen ifFalse: [
		(UIManager default confirm: (
'Note that linking a font instead of importing may make the
image un-portable, since the linked font must be present on
the system the next time the image is run.

This warning is only shown once per session.' ) trueChoice: 'Proceed' falseChoice: 'Cancel')
		ifFalse: [^ self].
		self warningSeen: true]..
	filenames := fonts allFilenames.
	filenames do: [:filename | | readFonts |
		readFonts := TTFileDescription readFontsFrom: filename.
		readFonts isCollection
					ifFalse: [TTCFont newTextStyleFromTT: readFonts]
					ifTrue: [self importFontFamily: readFonts]].
	self allFonts: nil. "force redraw"! !

!FontImporterTool methodsFor: 'initialize' stamp: 'topa 3/7/2015 00:50'!
open
	^ToolBuilder open: self! !

!FontImporterTool methodsFor: 'layout' stamp: 'mt 4/28/2015 13:04'!
previewFrame

	^ LayoutFrame
		fractions: (0.4@0 corner: 1@1)
		offsets: (0@0 corner: 0@ self buttonHeight negated)! !

!FontImporterTool methodsFor: 'model access' stamp: 'topa 6/1/2016 11:44'!
previewText

	^ self selectedFont
		ifNil: [Text new]
		ifNotNil: [:font | font sampleText]! !

!FontImporterTool methodsFor: 'font list' stamp: 'topa 3/9/2015 23:50'!
selectedFont
	| fontDesc font |
	fontDesc := self currentSelection.
	font := self fontFromFamily: fontDesc.
	font isFontSet ifTrue: [
		font := (self currentParent isNil or: [self currentParent = self currentSelection])
			ifTrue: [font fontArray anyOne]
			ifFalse: [ "we have selected a leaf  "
				font fontArray
					detect: [:subfont | subfont subfamilyName = fontDesc fontname]
					ifNone: [font]]].
	^font emphasized: emphasis! !

!FontImporterTool methodsFor: 'helper' stamp: 'topa 3/19/2015 10:43'!
textForFamily: familyName subfamily: subfamilyName

	subfamilyName ifNil: [
		^ (TextStyle named: familyName)
			ifNil: [familyName]
			ifNotNil: [:style | style isTTCStyle
				ifTrue: ["we are already present "
					Text string: familyName attribute: TextEmphasis underlined]
				ifFalse: [familyName]]].
		
	" frome here on it is only about subfamilies"
	
	(self isStyleNameSupported: subfamilyName)
		ifFalse: [^ Text string: subfamilyName attribute: TextColor gray].

	^ (TextStyle named: familyName)
		ifNil: ["importable" subfamilyName]
		ifNotNil: [:style |
			(style isTTCStyle and: [ | regular emph |
					regular  := style fonts anyOne.
					emph := TTCFont indexOfSubfamilyName: subfamilyName.
					" detect if this style is already imported "
					regular emphasis = emph or: [(regular emphasis: emph) ~= regular]])
				ifFalse: ["again importable" subfamilyName]
				ifTrue: [Text string: subfamilyName attribute: TextEmphasis underlined]]! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 3/19/2015 11:02'!
warningSeen

	^ warningSeen ifNil: [false]! !

!FontImporterTool methodsFor: 'accessing' stamp: 'topa 3/19/2015 11:01'!
warningSeen: anObject

	warningSeen := anObject! !

!Model methodsFor: '*morphic' stamp: 'zz 3/2/2004 23:49'!
step
	"Default for morphic models is no-op"! !

Model subclass: #SearchBar
	instanceVariableNames: 'originatingWidget searchTerm selection resultsWidget workspace scratchPad'
	classVariableNames: 'UseScratchPad UseSmartSearch'
	poolDictionaries: ''
	category: 'Morphic-Menus-DockingBar'!

!SearchBar class methodsFor: 'as yet unclassified' stamp: 'mt 4/2/2015 16:04'!
build

	^ ToolBuilder build: self new! !

!SearchBar class methodsFor: 'preferences' stamp: 'mt 11/2/2015 10:11'!
useScratchPad
	<preference: 'Ctrl+0 invokes the scratch pad'
		category: 'docking bars'
		description: 'When enabled, pressing Ctrl+0 will go straight to the Search Bar''s scratch pad instead of the Search Bar, where pressing ESCape will return keyboard focus to the originating window.'
		type: #Boolean>
	^ UseScratchPad ifNil: [ false ]! !

!SearchBar class methodsFor: 'preferences' stamp: 'cmm 5/8/2015 17:02'!
useScratchPad: aBoolean
	UseScratchPad := aBoolean! !

!SearchBar class methodsFor: 'preferences' stamp: 'mt 8/6/2016 11:03'!
useSmartSearch
	<preference: 'Smart search'
		category: 'docking bars'
		description: 'When enabled, the docking bar search tries to find globals or full class names first, then tries to browse implementors, then tries to find class names if the search term starts uppercase, and finally invoke the regular message names search. When disabled, do only the last.'
		type: #Boolean>
	^ UseSmartSearch ifNil: [ true ]! !

!SearchBar class methodsFor: 'preferences' stamp: 'mt 8/6/2016 11:01'!
useSmartSearch: aBoolean
	UseSmartSearch := aBoolean.! !

!SearchBar methodsFor: 'accessing' stamp: 'cmm 5/10/2015 20:47'!
activate: event in: morph 
	self class useScratchPad
		ifTrue:
			[ event hand keyboardFocus = self scratchPad ifFalse: [ originatingWidget := event hand keyboardFocus ].
			self scratchPad selectAll; openInWorld.
			self layoutScratchPad.
			event hand newKeyboardFocus: self scratchPad ]
		ifFalse:
			[ self selection: (1 to: self searchTerm size).
			event hand newKeyboardFocus: morph textMorph ]! !

!SearchBar methodsFor: 'accessing' stamp: 'mt 5/8/2015 11:59'!
bindingOf: x

	^ self workspace bindingOf: x! !

!SearchBar methodsFor: 'toolbuilder' stamp: 'mt 4/16/2015 22:46'!
buildWith: builder

	^ (builder build: (builder pluggableInputFieldSpec new
		model: self;
		getText: #searchTerm;
		setText: #smartSearch:in:;
		editText: #searchTermSilently:;
		menu: #menu:shifted:;
		selection: #selection;
		indicateUnacceptedChanges: false;
		help: 'Search or evaluate...' translated))
			name: #searchBar;
			wantsFrameAdornments: false;
			borderWidth: 0;
			yourself! !

!SearchBar methodsFor: 'private' stamp: 'cmm 5/8/2015 17:01'!
handleScratchPadKey: aKeyboardEvent 
	(aKeyboardEvent keyValue = Character escape asciiValue and: [ originatingWidget notNil ]) ifTrue:
		[ aKeyboardEvent hand newKeyboardFocus: originatingWidget.
		self removeScratchPad ]! !

!SearchBar methodsFor: 'private' stamp: 'cmm 5/10/2015 21:15'!
layoutScratchPad
	| pos width |
	World mainDockingBars do:
		[ : each | each searchBarMorph ifNotNil:
			[ : searchBar | pos := searchBar bottomLeft.
			width := searchBar width ] ].
	width ifNil: [ width := 250.  pos := World topRight - (width @ 5) ].
	scratchPad
		width: width ;
		position: pos ;
		startStepping: #deleteUnlessHasFocus at: Time millisecondClockValue arguments: nil stepTime: 3000! !

!SearchBar methodsFor: 'accessing' stamp: 'mt 4/2/2015 15:45'!
menu: aMenu shifted: aBoolean

	^ StringHolder codePaneMenu: aMenu shifted: aBoolean! !

!SearchBar methodsFor: 'do-its' stamp: 'mt 4/15/2015 16:59'!
printIt: code result: object

	| focusedWidget |
	focusedWidget := self currentHand keyboardFocus.

	self removeResultsWidget.
	
	ToolBuilder default in: [:builder |
		resultsWidget := (builder build: (StringHolder new
			contents: object asString;
			buildCodePaneWith: builder)).
		"This might cause the widget disappear to early. Especially if the textMorph is to small but the container needs to maintain a larger size because of its scroll bars."
		resultsWidget textMorph
			on: #mouseLeave send: #delete to: resultsWidget.
		resultsWidget
			extent: 1000@1000; fullBounds; "give text a chance to layout good"
			width: 250; fullBounds;
			height: (resultsWidget extentToFit y min: 300);
			position: (focusedWidget
				ifNotNil: [:w | w owner boundsInWorld bottomLeft]
				ifNil: [self currentHand position]);
			color: (BalloonMorph balloonColor alpha: 1.0).
			
		Preferences menuAppearance3d
			ifTrue: [resultsWidget addDropShadow].
			
		resultsWidget openInWorld].! !

!SearchBar methodsFor: 'accessing' stamp: 'mt 4/7/2015 13:40'!
removeResultsWidget

	resultsWidget ifNotNil: [:w | w delete].
	resultsWidget := nil.! !

!SearchBar methodsFor: 'accessing' stamp: 'cmm 5/8/2015 16:54'!
removeScratchPad
	scratchPad ifNotNil: [ scratchPad delete ]! !

!SearchBar methodsFor: 'accessing' stamp: 'cmm 8/4/2016 22:28'!
scratchPad
	^ scratchPad ifNil:
		[ scratchPad := TextMorphForEditView new. "we should be able to use TextMorph here; fix later"
		scratchPad
"			on: #keyboardFocusChange send: #removeScratchPad to: self ;"
			on: #mouseLeave send: #removeScratchPad to: self ;
			on: #keyStroke send: #handleScratchPadKey: to: self ;
			margins: (5@0 corner: 5@0);
			backgroundColor: (BalloonMorph balloonColor alpha: 1.0) ;
			setEditView: PluggableTextMorph new ; "dummy"
			autoFit: true ;
			wrapFlag: true ;
			newContents: '--scratch area--' ;
			font: ((UserInterfaceTheme current get: #font for: #PluggableTextMorph) ifNil: [TextStyle defaultFont]);
			textColor: ((UserInterfaceTheme current get: #textColor for: #PluggableTextMorph) ifNil: [Color black]);
			caretColor: ((UserInterfaceTheme current get: #caretColor for: #PluggableTextMorph) ifNil: [Color red]);
			selectionColor: ((UserInterfaceTheme current get: #selectionColor for: #PluggableTextMorph) ifNil: [Color blue muchDarker] ifNotNil: [ : col | col twiceLighter ]); 
			yourself.
		self layoutScratchPad.
		Preferences menuAppearance3d ifTrue: [ scratchPad addDropShadow ].
		scratchPad ]! !

!SearchBar methodsFor: 'accessing' stamp: 'mt 4/2/2015 15:41'!
searchTerm

	^ searchTerm ifNil: ['']! !

!SearchBar methodsFor: 'accessing' stamp: 'mt 4/7/2015 09:09'!
searchTermSilently: aString.
	"Do not signal it to the model."
	
	searchTerm := aString.! !

!SearchBar methodsFor: 'accessing' stamp: 'mt 4/7/2015 13:42'!
searchTerm: aString

	searchTerm := aString.
	self changed: #searchTerm.! !

!SearchBar methodsFor: 'accessing' stamp: 'mt 4/2/2015 15:56'!
selection

	^ selection ifNil: [1 to: 0]! !

!SearchBar methodsFor: 'accessing' stamp: 'mt 4/2/2015 15:48'!
selection: anInterval

	selection := anInterval.
	self changed: #selection.! !

!SearchBar methodsFor: 'searching' stamp: 'mt 8/6/2016 11:04'!
smartSearch: text in: morph
	"Take the user input and perform an appropriate search"
	| input newContents |
	self removeResultsWidget.
	input := text asString ifEmpty:[^self].
	self class useSmartSearch ifFalse: [^ ToolSet default browseMessageNames: input].

	"If it is a global or a full class name, browse that class."
	(Smalltalk bindingOf: input) ifNotNil:[:assoc| | global |
		global := assoc value.
		^ToolSet browse: (global isBehavior ifTrue:[global] ifFalse:[global class]) selector: nil].
	
	"If it is a symbol and there are implementors of it, browse those implementors."
	(Symbol lookup: input) ifNotNil: [:selector |
		(SystemNavigation new allImplementorsOf: selector) ifNotEmpty:[:list|
			^SystemNavigation new
				browseMessageList: list
				name: 'Implementors of ' , input]].

	"If it starts uppercase, browse classes if any. Otherwise, just search for messages."
	input first isUppercase
		ifTrue: [
			(UIManager default classFromPattern: input withCaption: '')
				ifNotNil:[:aClass| ^ToolSet browse: aClass selector: nil]
				ifNil: [
					newContents := input, ' -- not found.'.
					self searchTerm: newContents.
					self selection: (input size+1 to: newContents size).
					self currentHand newKeyboardFocus: morph textMorph.
					^ self]]
		ifFalse: [
			ToolSet default browseMessageNames: input].! !

!SearchBar methodsFor: 'accessing' stamp: 'mt 5/8/2015 11:58'!
workspace

	^ workspace ifNil: [workspace := Workspace new]! !

!MIDIFileReader class methodsFor: '*Morphic-Sounds' stamp: 'jm 2/1/98 19:48'!
playFileNamed: fileName

	ScorePlayerMorph
		openOn: (self scoreFromFileNamed: fileName)
		title: (FileDirectory localNameFor: fileName).
! !

!MIDIFileReader class methodsFor: '*Morphic-Sounds' stamp: 'ls 8/8/1998 03:14'!
playStream: binaryStream

	ScorePlayerMorph
		openOn: (self scoreFromStream: binaryStream)
		title: 'a MIDI stream'
! !

!MIDIFileReader class methodsFor: '*Morphic-Sounds' stamp: 'jm 5/29/1998 17:12'!
playURLNamed: urlString

	| titleString |
	titleString := urlString
		copyFrom: (urlString findLast: [:c | c=$/]) + 1
		to: urlString size.
	ScorePlayerMorph
		openOn: (self scoreFromURL: urlString)
		title: titleString.
! !

MessageSend subclass: #MorphicAlarm
	instanceVariableNames: 'scheduledTime sequenceNumber numArgs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!MorphicAlarm class methodsFor: 'instance creation' stamp: 'ar 9/11/2000 16:44'!
scheduledAt: scheduledTime receiver: aTarget selector: aSelector arguments: argArray
	^(self receiver: aTarget selector: aSelector arguments: argArray)
		scheduledTime: scheduledTime.! !

!MorphicAlarm methodsFor: 'accessing' stamp: 'ar 9/11/2000 16:44'!
scheduledTime
	"Return the time (in milliseconds) that the receiver is scheduled to be executed"
	^scheduledTime! !

!MorphicAlarm methodsFor: 'accessing' stamp: 'ar 9/11/2000 16:45'!
scheduledTime: msecs
	"Set the time (in milliseconds) that the receiver is scheduled to be executed"
	scheduledTime := msecs! !

!MorphicAlarm methodsFor: 'accessing' stamp: 'jcg 1/9/2010 12:29'!
sequenceNumber
	"Answer the sequence number of the alarm, which is used to preserve ordering for alarms scheduled for the same time."
	^sequenceNumber ifNil: [0]! !

!MorphicAlarm methodsFor: 'accessing' stamp: 'jcg 1/7/2010 10:56'!
sequenceNumber: positiveInteger
	"Set the sequence number of the alarm, which is used to preserve ordering for alarms scheduled for the same time."
	sequenceNumber := positiveInteger! !

!MorphicAlarm methodsFor: 'evaluating' stamp: 'jmv 3/2/2010 16:18'!
value: anArgument
	| nArgs |
	numArgs ifNil:[numArgs := selector numArgs].
	nArgs := arguments ifNil:[0] ifNotNil:[arguments size].
	nArgs = numArgs ifTrue:[
		"Ignore extra argument"
		^self value].
	^arguments
		ifNil: [ receiver perform: selector with: anArgument]
		ifNotNil: [ receiver perform: selector withArguments: (arguments copyWith: anArgument)]! !

MorphicAlarm subclass: #StepMessage
	instanceVariableNames: 'stepTime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!StepMessage class methodsFor: 'instance creation' stamp: 'ar 10/22/2000 15:48'!
scheduledAt: scheduledTime stepTime: stepTime receiver: aTarget selector: aSelector arguments: argArray
	^(self receiver: aTarget selector: aSelector arguments: argArray)
		scheduledTime: scheduledTime;
		stepTime: stepTime! !

!StepMessage methodsFor: 'printing' stamp: 'ar 10/22/2000 15:59'!
printOn: aStream
	super printOn: aStream.
	aStream 
		nextPut: $(;
		print: receiver;
		space;
		print: selector;
		space;
		print: scheduledTime;
		nextPut: $).! !

!StepMessage methodsFor: 'testing' stamp: 'ar 10/22/2000 16:56'!
stepTime
	"Return the step time for this message. If nil, the receiver of the message will be asked for its #stepTime."
	^stepTime! !

!StepMessage methodsFor: 'accessing' stamp: 'ar 10/22/2000 16:56'!
stepTime: aNumber
	"Set the step time for this message. If nil, the receiver of the message will be asked for its #stepTime."
	stepTime := aNumber! !

!Color methodsFor: '*Morphic' stamp: 'jrd 10/18/2010 13:45'!
asMorph

	^(RectangleMorph new)
		fillStyle: self;
		borderWidth: 0;
		yourself! !

!Color methodsFor: '*Morphic' stamp: 'ar 7/8/2006 21:00'!
iconOrThumbnailOfSize: aNumberOrPoint 
	"Answer an appropiate form to represent the receiver"
	| form |
	form := Form extent: aNumberOrPoint asPoint asPoint depth: 32.
	form fillColor: self.
	^ form! !

!Float methodsFor: '*Morphic-Explorer' stamp: 'bf 8/20/1999 12:56'!
hasContentsInExplorer

	^false! !

!OrientedFillStyle methodsFor: '*Morphic-Balloon' stamp: 'dgd 10/17/2003 22:35'!
addFillStyleMenuItems: aMenu hand: aHand from: aMorph
	"Add the items for changing the current fill style of the receiver"
	aMenu add: 'change origin' translated target: self selector: #changeOriginIn:event: argument: aMorph.
	aMenu add: 'change orientation' translated target: self selector: #changeOrientationIn:event: argument: aMorph.! !

!OrientedFillStyle methodsFor: '*Morphic-Balloon' stamp: 'ar 6/18/1999 07:41'!
changeOrientationIn: aMorph event: evt
	"Interactively change the origin of the receiver"
	| handle |
	handle := HandleMorph new forEachPointDo:[:pt|
		self direction: pt - self origin.
		self normal: nil.
		aMorph changed].
	evt hand attachMorph: handle.
	handle startStepping.! !

!OrientedFillStyle methodsFor: '*Morphic-Balloon' stamp: 'ar 6/18/1999 07:28'!
changeOriginIn: aMorph event: evt
	"Interactively change the origin of the receiver"
	| handle |
	handle := HandleMorph new forEachPointDo:[:pt|
		self origin: pt.
		aMorph changed].
	evt hand attachMorph: handle.
	handle startStepping.! !

!TextStyle class methodsFor: '*Morphic-user interface' stamp: 'tween 8/10/2006 07:16'!
emphasisMenuForFont: font target: target selector: selector highlight: currentEmphasis
	"Offer a font emphasis menu for the given style. If one is selected, pass that font to target with a call to selector. The fonts will be displayed in that font.
	Answer nil if no derivatives exist.
	"

 	| aMenu derivs |
	derivs := font derivativeFonts.
	derivs isEmpty ifTrue: [ ^nil ].
	aMenu := MenuMorph entitled: 'emphasis' translated.
	derivs := derivs asOrderedCollection.
	derivs addFirst: font.
	derivs do: [ :df | 
			aMenu 
				add: df emphasisString
				target: target 
				selector: selector
				argument: df.
                aMenu lastItem font: df.
                df emphasis == currentEmphasis ifTrue: [aMenu lastItem color: Color blue darker]].
        ^ aMenu! !

!TextStyle class methodsFor: '*Morphic-user interface' stamp: 'laza 3/25/2004 23:12'!
fontMenuForStyle: styleName target: target selector: selector
	^self fontMenuForStyle: styleName target: target selector: selector highlight: nil! !

!TextStyle class methodsFor: '*Morphic-user interface' stamp: 'nice 12/27/2009 03:19'!
fontMenuForStyle: styleName target: target selector: selector highlight: currentFont 
	"Offer a font menu for the given style. If one is selected, pass 
	that font to target with a  
	call to selector. The fonts will be displayed in that font."
	| aMenu |
	aMenu := MenuMorph entitled: styleName.
	(TextStyle named: styleName)
		ifNotNil: [:s | s isTTCStyle
				ifTrue: [aMenu
						add: 'New Size'
						target: self
						selector: #chooseTTCFontSize:
						argument: {styleName. target. selector}]].
	(self pointSizesFor: styleName)
		do: [:pointSize | 
			| font subMenu displayFont | 
			font := (self named: styleName)
						fontOfPointSize: pointSize.
			subMenu := self
						emphasisMenuForFont: font
						target: target
						selector: selector
						highlight: (currentFont
								ifNotNil: [:cf | (cf familyName = styleName
											and: [cf pointSize = font pointSize])
										ifTrue: [currentFont emphasis]]).
			subMenu
				ifNil: [aMenu
						add: pointSize asString , ' Point'
						target: target
						selector: selector
						argument: font]
				ifNotNil: [aMenu add: pointSize asString , ' Point' subMenu: subMenu].
			displayFont := font.
			(font isSymbolFont or:[(font hasDistinctGlyphsForAll: pointSize asString , ' Point') not])
				ifTrue:[
					"don't use a symbol font to display its own name!!!!"
					displayFont := self default fontOfPointSize: pointSize].
			aMenu lastItem font: displayFont.
			currentFont
				ifNotNil: [:cf | (cf familyName = styleName
							and: [cf pointSize = font pointSize])
						ifTrue: [aMenu lastItem color: Color blue darker]]].
	^ aMenu! !

!TextStyle class methodsFor: '*Morphic-user interface' stamp: 'laza 3/25/2004 23:12'!
promptForFont: aPrompt andSendTo: aTarget withSelector: aSelector
	self promptForFont: aPrompt andSendTo: aTarget withSelector: aSelector highlight: nil! !

!TextStyle class methodsFor: '*Morphic-user interface' stamp: 'nice 12/27/2009 03:11'!
promptForFont: aPrompt andSendTo: aTarget withSelector: aSelector highlight: currentFont 
	"Morphic Only!! prompt for a font and if one is provided, send it to aTarget using a 
	message with selector aSelector."
	"TextStyle promptForFont: 'Choose system font:' andSendTo: Preferences withSelector: 
	#setSystemFontTo: "
	"Derived from a method written by Robin Gibson"
	| menu currentTextStyle |
	currentTextStyle := currentFont
				ifNotNil: [currentFont textStyleName].
	menu := MenuMorph entitled: aPrompt.
	self actualTextStyles keysSortedSafely
		do: [:styleName | | subMenu | 
			subMenu := self
						fontMenuForStyle: styleName
						target: aTarget
						selector: aSelector
						highlight: currentFont.
			menu add: styleName subMenu: subMenu.
			menu lastItem
				font: ((self named: styleName)
						fontOfSize: 18).
			styleName = currentTextStyle
				ifTrue: [menu lastItem color: Color blue darker]].
	menu popUpInWorld: self currentWorld! !

!Transcripter methodsFor: '*Morphic' stamp: 'dtl 12/14/2013 16:15'!
morphicDisplayText
	para compose: self contents asText
		style: TextStyle default
		from: 1
		in: frame.
	Display
		fill: (frame insetBy: -2) fillColor: self black;
		fill: frame fillColor: self white.
	Display getCanvas
		paragraph: para
		bounds: (4@4 + frame topLeft extent: Display extent)
		color: Color black! !

!Transcripter methodsFor: '*Morphic' stamp: 'dtl 12/14/2013 16:21'!
morphicInitializeParagraph: classParagraph
	para := classParagraph new.
	para compose: self contents asText
		style: TextStyle default
		from: 1
		in: frame
! !

!PopUpMenu methodsFor: '*Morphic-Menus' stamp: 'dtl 1/30/2010 16:13'!
morphicStartUpLeftFlush
	"Build and invoke this menu with no initial selection.  By Jerry Archibald, 4/01.
	If in MVC, align menus items with the left margin.
	Answer the selection associated with the menu item chosen by the user or nil if none is chosen.  
	The mechanism for getting left-flush appearance in mvc leaves a tiny possibility for misadventure: if the user, in mvc, puts up the jump-to-project menu, then hits cmd period while it is up, then puts up a second jump-to-project menu before dismissing or proceeding through the debugger, it's possible for mvc popup-menus thereafter to appear left-aligned rather than centered; this very unlikely condition can be cleared by evaluating 'PopUpMenu alignment: 2'"

	^self startUp! !

!PopUpMenu methodsFor: '*Morphic-Menus' stamp: 'dtl 1/30/2010 15:54'!
morphicStartUpWithCaption: captionOrNil icon: aForm at: location allowKeyboard: aBoolean
	"Display the menu, with caption if supplied. Wait for the mouse button to go down, then track the selection as long as the button is pressed. When the button is released,
	Answer the index of the current selection, or zero if the mouse is not released over  any menu item. Location specifies the desired topLeft of the menu body rectangle. The final argument indicates whether the menu should seize the keyboard focus in order to allow the user to navigate it via the keyboard."

	selection := Cursor normal
				showWhile: [| menuMorph |
					menuMorph := MVCMenuMorph from: self title: nil.
					(captionOrNil notNil
							or: [aForm notNil])
						ifTrue: [menuMorph addTitle: captionOrNil icon: aForm].
					MenuIcons decorateMenu: menuMorph.
					menuMorph
						invokeAt: location
						in: ActiveWorld
						allowKeyboard: aBoolean].
	^ selection! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 7/18/2004 23:12'!
asCubic
	"Convert this point array to a Cubic object"
	self
		assert: [self size = 4].
	self
		assert: [self
				allSatisfy: [:each | each isPoint]].
	^ Cubic withAll: self! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'nice 12/26/2009 21:15'!
assertSlopesWith: knots from: start to: end
   "
	We trust everything has been checked. 
	The following assertions should hold at this point: "
	
	self assert: [ self size = knots size ] . 
	"Sizes must be consistent." 
	self assert: [ end > start]. 
	"There must be at least one slope to clamp." 
	self assert: [ 0 < start and: [start <= knots size] ]. 
	"The clamped slope may be the last one."
	self assert: [  end  <= knots size + start ] . 
	"We can wrap. There may be only one known slope."
	"xxx self assert: [ end = knots size + start ifTrue: [ (self at: start) notNil ] ] . xxx"
		"xxx If we overlap slope must be known. xxx"
	{ start . end } 
		do: [ :index |
			| slope |
			slope := (self at: index ).
			self assert: [ slope isNil 
				or: [ slope isNumber 
				or: [ slope isPoint ] ] ] ] . 
	"And a known and reasonalble value or nil." 
	^true 
	! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 1/30/2005 21:04'!
changeInSlopes: slopes 
	"A message to knots of a spline. Returns an array with the 3rd cubic coeff."
	"The last nth item is correct iff this is a closed cubic.
	Presumably that is the only time we care.
	We always return the same sized array as self."
	| n slopeChanges |
	n := self size.
	n = slopes size
		ifFalse: [^ self error: 'vertices and slopes differ in number'].
	slopeChanges := Array new: n.
	(1 to: n)
		do: [:i | slopeChanges at: i put: (self atWrap: i + 1)
					- (self at: i) * 3 - ((slopes at: i)
						* 2)
					- (slopes atWrap: i + 1)].
	
	^ slopeChanges! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'ul 7/25/2011 01:30'!
changeOfChangesInSlopes: slopes 
	"A message to knots of a spline. Returns an array with the 4rd 
	cubic coeff."
	"The last nth item is correct iff this is a closed cubic. 
	Presumably that is the only time we care. 
	We always return the same sized array as self."
	| n changes |
	n := self size.
	n = slopes size
		ifFalse: [^ self error: 'vertices and slopes differ in number'].
	changes := Array new: n.
	1 to: n do: [ :i | 
		changes at: i put: (self at: i)
			- (self atWrap: i + 1) * 2
			+ (slopes at: i)
			+ (slopes atWrap: i + 1) ].
	^ changes! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'ul 7/25/2011 11:24'!
closedCubicSlopes
	"Sent to knots returns the slopes of a closed cubic spline.
	From the same set of java sources as naturalCubic. This is a squeak  
	transliteration of the java code."
	"from java code NatCubicClosed extends NatCubic  
	solves for the set of equations for all knots: 
	b1+4*b2+b3=3*(a3-a1)
	where a1 is (knots atWrap: index + 1) etc.
	and the b's are the slopes .
	
	by decomposing the matrix into upper triangular and lower matrices  
	and then back sustitution. See Spath 'Spline Algorithms for Curves  
	and Surfaces' pp 19--21. The D[i] are the derivatives at the knots.  
	"
	
	| v w x y z n1  D F G H |
	n1 := self size.
	n1 < 3
		ifTrue: [self error: 'Less than 3 points makes a poor curve'].
	v := Array new: n1.
	w := Array new: n1.
	y := Array new: n1.
	
	D := Array new: n1.
	x := self.
	z := 1.0 / 4.0.
	v at: 2 put: z.
	w at: 2 put: z.
	y at: 1 put: z * 3.0 * ((x at: 2)
				- (x at: n1)).
	H := 4.0.
	F := 3 * ((x at: 1)
					- (x at: n1 - 1)).
	G := 1.
	2 to: n1 - 1
		do: [:k | 
			z := 1.0 / (4.0
							- (v at: k)).
			v at: k + 1 put: z.
			w at: k + 1 put: z negated
					* (w at: k).
			y at: k put: z * (3.0 * ((x at: k + 1)
							- (x at: k - 1))
						- (y at: k - 1)).
			H := H - (G
						* (w at: k)).
			F := F - (G
						* (y at: k - 1)).
			G := (v at: k) negated * G].
	H := H - (G + 1 * ((v at: n1)
						+ (w at: n1))).
	y at: n1 put: F - (G + 1
				* (y at: n1 - 1)).
	D at: n1 put: (y at: n1)
			/ H.
	D at: n1 - 1 put: (y at: n1 - 1)
			- ((v at: n1)
					+ (w at: n1)
					* (D at: n1)).
	n1 - 2 to: 1 by: -1 do: [ :k | 
		D at: k put: 
			(y at: k)
					- ((v at: k + 1)
							* (D at: k + 1)) - ((w at: k + 1)
						* (D at: n1))].
	^ D .! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 2/18/2006 12:54'!
closedCubicSlopes: clampedSlopes
	"Sent to knots returns a copy of clampedSlopes with the values of the undefined (nil)  slopes filled in.
	"
	" clampedSlopes must be the same size as knots)" 
	
	"/* Between known slopes we solve the equation for knots with end conditions:  
	4*b1+b2 = 3(a2 - a0) - b0 
	bN2+4*bN1 = 3*(aN-aN2) - bN
	and inbetween:
	b2+4*b3+b4=3*(a4-a2)
	where a2 is (knots atWrap: index + 1) etc.
	and the b's are the slopes .
	N is the last index (knots size)
	N1 is N-1.
	 
	by using row operations to convert the matrix to upper  
	triangular and then back substitution. 
	"
	| slopes tripleKnots list |
	(list := clampedSlopes closedFillinList) = { 0 to: self size } ifTrue: [ ^ self closedCubicSlopes ] .
	"Special case all unknown."
	
	tripleKnots := self * 3.0 . 
	" Premultiply and convert numbers or point coords to Floats "
	slopes := clampedSlopes copy. "slopes contents will be modified."
	
	list do: [ :r | slopes slopesWith: tripleKnots from: r first to: r last ] .
	
	^ slopes! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 2/17/2006 01:42'!
closedFillinList
	"Answers a list of ranges between which values are undertermined.
	Reciever is a list that combines known values and nil entries for
	undetermined values.
	Answer a list of ranges. Each range starts and ends with a known
	value. 
	The range inbetween the known values are nil. The ranges start and
	ends may overlap. 
	Each nil element in the list appears in exactly one range. 
	If the list starts or ends with nil the last range will wrap around to the
	next known value. There may be only one known value in the list but
	there must be atleast one know value.
	
	(self allsatisfy: [ :e | e isNil ] ) ifTrue: [ self error: 'list must contain at
	least one known value' ] 
	"
	| changes n |
	changes := self nilTransitions .
	changes isEmpty ifTrue: [ ^ { 0 to: self size } "Special case. All unknowns." ] .
	 
	changes = #(1) ifTrue: [ ^ #() "Special case. no unknowns." ] . 
	changes = { n :=  self size } ifTrue: [ ^ { n to: n + n } ] .
	"Special case. Only last element known."
	
	changes size even ifTrue: 
			[changes add: self size
							+ (changes at: 1)]
						ifFalse: [
		changes first = 1 ifFalse: [ changes add: self size + 1;
											add: self size + changes first ]
						
		].
	^ changes allButFirst pairsCollect: [ :s :e | ( s - 1 to: e ) ] .

	! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'ul 7/25/2011 01:29'!
cubicPointPolynomialAt: vIndex
	"From curve information assemble a 4-array of points representing the coefficents for curve segment between to points. Beginning point is first point in array endpoint is the pointSum of the array. Meant to be sent to newcurves idea of curve coefficents." 
	
	| result |
	result := Cubic new: 4.
	1 to: 4 do: [ :i | 
		result at: i put: ((self at: i) at: vIndex) @ ((self at: 4 + i) at: vIndex) ].
	^result! !

!SequenceableCollection methodsFor: '*Morphic-Explorer' stamp: 'ul 11/11/2009 19:49'!
explorerContents

	^Array new: self size streamContents: [ :stream |
		1 to: self size do: [ :index |
			stream nextPut: (
				ObjectExplorerWrapper
					with: (self at: index)
					name: index printString
					model: self) ] ]! !

!SequenceableCollection methodsFor: '*Morphic-fliprotate' stamp: 'wiz 4/7/2004 09:31'!
flipRotated: flipIndex 
	"Answer a copy of the receiver with element order indicated by  
	flipIndex."
	"Examples:"
	"'frog' flipRotated: 1"
	"[ :c | (1 to: c size * 2) collect:  
	[ :i | c flipRotated: i ]  
	] value: 'frog'."
	"Lsb of flipIndex indicates whether list is reversed"
	"The flipIndex // 2 gives how much to rotate by after reversing"
	"A good way to think of this is a piece of pie in a pie plate being flip  
	over its leading edge successively."
	"flipIndex > 2 * n are allowed to make it possible to store an array of  
	indexes in an integer."
	| n result src twist |
	n := self size.
	flipIndex \\ (n * 2) = 0
		ifTrue: [^ self].
	"allow for doing nothing"
	result := self species new: n.
	twist := flipIndex // 2 \\ n.
	src := 0.
	(flipIndex even
		ifTrue: [1 + twist to: n + twist]
		ifFalse: [n - 1 - twist to: twist negated by: -1])
		do: [:i | result
				at: (src := src + 1)
				put: (self atWrap: i)].
	^ result! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 2/12/2006 23:21'!
naturalCubicSlopes
	"Sent to knots returns the slopes of a natural cubic curve fit.
	This is a direct  squeak  
	transliteration of the java code."
	" public class NatCubic extends ControlCurve
	
	/* We solve the equation for knots with end conditions:  
	2*b1+b2 = 3(a1 - a0) 
	bN1+2*bN = 3*(aN-aN1)
	and inbetween:
	b2+4*b3+b4=3*(a4-a2)
	where a2 is (knots atWrap: index + 1) etc.
	and the b's are the slopes .
	N is the last index (knots size)
	N1 is N-1.
	 
	by using row operations to convert the matrix to upper  
	triangular  
	and then back sustitution. The D[i] are the derivatives at the  
	knots.  
	"
	| x gamma delta D n1 |
	n1 := self size.
	n1 < 3
		ifTrue: [self error: 'Less than 3 points makes a poor curve'].
	x := self.
	gamma := Array new: n1.
	delta := Array new: n1.
	
	D := Array new: n1.
	gamma at: 1 put: 1.0 / 2.0.
	(2 to: n1 - 1)
		do: [:i | gamma at: i put: 1.0 / (4.0
						- (gamma at: i - 1))].
	gamma at: n1 put: 1.0 / (2.0
				- (gamma at: n1 - 1)).
	delta at: 1 put: 3.0 * ((x at: 2)
				- (x at: 1))
			* (gamma at: 1).
	(2 to: n1 - 1)
		do: [:i | delta at: i put: 3.0 * ((x at: i + 1)
						- (x at: i - 1))
					- (delta at: i - 1)
					* (gamma at: i)].
	delta at: n1 put: 3.0 * ((x at: n1)
				- (x at: n1 - 1))
			- (delta at: n1 - 1)
			* (gamma at: n1).
	D
		at: n1
		put: (delta at: n1).
	(1 to: n1 - 1)
		reverseDo: [:i | D at: i put: (delta at: i)
					- ((gamma at: i)
							* (D at: i + 1))].
	^ D! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 2/18/2006 12:55'!
naturalCubicSlopes: clampedSlopes
	"Sent to knots returns a copy of clampedSlopes with the values of the undefined (nil)  slopes filled in.
	"
	" clampedSlopes must be the same size as knots)" 
	
	"/* Between known slopes we solve the equation for knots with end conditions:  
	4*b1+b2 = 3(a2 - a0) - b0 
	bN2+4*bN1 = 3*(aN-aN2) - bN
	and inbetween:
	b2+4*b3+b4=3*(a4-a2)
	where a2 is (knots atWrap: index + 1) etc.
	and the b's are the slopes .
	N is the last index (knots size)
	N1 is N-1.
	 
	by using row operations to convert the matrix to upper  
	triangular and then back substitution. 
	"
	| slopes tripleKnots |
	tripleKnots := self * 3.0 . 
	" Premultiply and convert numbers or point coords to Floats "
	slopes := clampedSlopes copy. "slopes will be modified."
	clampedSlopes naturalFillinList do: [ :r | slopes slopesWith: tripleKnots from: r first to: r last ] .
	
	^ slopes! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 2/9/2006 00:46'!
naturalFillinList
	"Return a list of fillin ranges to be used to calculate natural or clamped slopes.
	Note that this list is slightly different in mission from the closedFillinList"
	"Answers a list of ranges between which value are undertermined.
	Reciever is a list that combines known values and nil entries for
	undetermined values.
	Answer a list of ranges. Each range starts and ends with a known value. 
	With the exception of the first and last slopes on the list which may be unknown.
	If no slopes are known then the only range is the whole list.
	If all slopes are known then the fillin list is empty.
	The range inbetween the known values are nil. The ranges start and
	ends may overlap if the slope at the overlap is known.
	Each nil element in the list appears in exactly one range.  
	"
	| changes  |
	changes := self nilTransitions .
	changes isEmpty ifTrue: [ ^ { 1 to: self size } "Special case all unknown." ] .
	 
	changes = #(1) ifTrue: [ ^ #() "Special case. no unknowns." ] . 
	
	changes size even 
			ifTrue: [changes add: self size ] .  "Last slope is unknown"
	changes first = 1
			ifTrue: [ ^ changes allButFirst pairsCollect: [ :s :e | (  s - 1 to: e ) ] ] .
	 
		"Otherwise first slope is unknown."
				
						
			^ { 1 to: changes first } , 
					(changes allButFirst pairsCollect: [ :s :e | ( ( s - 1) to: e ) ]) 

	! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'nice 12/26/2009 21:21'!
nilTransitions
	"Return an OrderedCollection of transition indexes.  
	Indexes represent where the list elements transitions 
	from nil to nonNil 
		or from nonNil to nil.
	1 is an index in the list iff the first element is nonNil. "
	
	| changes nilSkip |

	changes := OrderedCollection new.
	nilSkip := true .
	
	1 to: self size
		do: [:i | (self atWrap: i) isNil == nilSkip
				ifFalse: [changes add: i.
					nilSkip := nilSkip not]].

	^ changes ! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 5/2/2004 22:43'!
segmentedSlopes
	"For a collection of floats. Returns the slopes for straight 
	segments between vertices."
	"last slope closes the polygon. Always return same size as 
	self. "
	^ self
		collectWithIndex: [:x :i | (self atWrap: i + 1)
				- x]! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 2/18/2006 12:54'!
slopesWith: tripleKnots from: start to: end 
	"Sent to modifiable list of slopes. Fills in the slope values between start
	and end. Start and end slopes can be either clamped or nil. 
	If nil the natural slope for that value will be filled in. 
	We expect that the parameters meets the assertions in
	self assertSlopesWith: knots from: start to: end."
	"
	
	/* We solve the equation for knots with end conditions either known or unknown: 
	4*b1+b2 = 3*(a2 - a0) - b0			b0 known
	Or
	2*b0+b1 = 3*(a1 - a0) .			b0 == nil
	
	bN2+4*bN1 = 3*(aN-aN2)-bN		bN known
	Or
	bN1+2*bN = 3*(aN-aN1)			bN == nil
	 .
	b0, bN are starting and ending slopes.
	
	We now handle the special closed cubic case where a0 == aN ( i.e. N = knots size )
	and b0 == bN == nil .
	
	
	
	
	and inbetween:
	b2+4*b3+b4=3*(a4-a2)
	where a2 is (knots  atWrap: index + 1) etc.
	and the b's are the slopes .
	by using row operations to convert the matrix to upper 
	triangular and then back substitution. 
	"
	| gamma delta n range isOpenRange |
	n := self size.
	gamma := Array new: n.
	delta := Array new: n.
	isOpenRange := end < (start + self size) .
	(self at: start)
		ifNotNil: [
			gamma at: start put: 0.0.
			delta
				at: start
				put: (self at: start).
			range := ( start + 1 to: end - 1 ) .
			] " clamped initial conditions"
		ifNil: [
				isOpenRange
				ifTrue:	
			[gamma at: start put: 2.0 reciprocal.
			delta
				at: start
				put:  ((tripleKnots atWrap: start + 1)
					- tripleKnots at: start ) * (gamma at: start) .
			range := ( start  to: end - 1 ) . ]  "natural initial conditions "
				ifFalse: 
			[ gamma at: start put: 4.0 reciprocal.
			delta
				at: start
				put:  ((tripleKnots atWrap: start + 1)
					- tripleKnots atWrap: start - 1 ) * (gamma at: start) .
			range := ( start + 1  to: end - 1 ) .
				]  "closed initial conditions "
				] .
	(start + 1 to: end - 1)
		do: [:i | gamma atWrap: i put: 1.0 / (4.0
						- (gamma atWrap: i - 1))].
	(start + 1 to: end - 1)
		do: [:i | delta atWrap: i put: ((tripleKnots atWrap: i + 1)
						- (tripleKnots atWrap: i - 1))
					- (delta atWrap: i - 1)
					* (gamma atWrap: i)].
	(self atWrap: end) 
		ifNil: [ isOpenRange
			ifTrue: [
			gamma atWrap: end put: 1.0 / (2.0  
										- (gamma atWrap: end - 1 )).
			delta
				atWrap: end
				put:  ((tripleKnots atWrap: end )
							- tripleKnots atWrap: end - 1 )
					 	- (delta at: end - 1 ) * (gamma atWrap: end)] "natural end conditions"
					ifFalse: [
			gamma atWrap: end put: 1.0 / (4.0  
										- (gamma atWrap: end - 1 )).
			delta
				atWrap: end
				put:  ((tripleKnots atWrap: end + 1 )
							- tripleKnots atWrap: end - 1 ) 
						- (delta at: end - 1 ) * (gamma atWrap: end)] "closed end conditions"  
					.
			self atWrap: end put: (delta atWrap: end ) .
				] 
	ifNotNil: [ 
			gamma atWrap: end put: 0.0 .
			delta
				atWrap: end
				put: (self atWrap: end)  .
		
			] "clamped end conditions" .
		
	range
		reverseDo: [:i | self atWrap: i put: 
					(delta atWrap: i)
					- ((gamma atWrap: i)
							* (self atWrap: i + 1)) ] .
	" reciever now contains the filled in slopes."
	^ self ! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 2/16/2006 20:08'!
transitions
	"Sent to a list of boolean values. 
	Return an OrderedCollection of transition indexes.  
	Indexes represent where the list elements transitions 
	from true to false
		or from false to true.
	1 is an index in the list iff the first element is false. "
	
	| changes boolSkip |

	changes := OrderedCollection new.
	boolSkip := true .
	
	self
		withIndexDo: [:truth :i | truth  == boolSkip
				ifFalse: [changes add: i.
					boolSkip := boolSkip not]].

	^ changes ! !

!SequenceableCollection methodsFor: '*Morphic-NewCurves-cubic support' stamp: 'wiz 2/16/2006 20:09'!
transitions: aSelectBlock
	"Sent to a list. Select block returns a boolean 
	Return an OrderedCollection of transition indexes.  
	Indexes represent where the list elements transitions 
	from true to false
		or from false to true.
	1 is an index in the list iff the first element tests false. "
	
	| changes boolSkip |

	changes := OrderedCollection new.
	boolSkip := true .
	
	self withIndexDo: 
		 [:e :i | (aSelectBlock value: e ) == boolSkip
				ifFalse: [changes add: i.
					boolSkip := boolSkip not]].

	^ changes ! !

!MailComposition methodsFor: '*Morphic-Support' stamp: 'mdr 4/10/2001 14:27'!
addAttachment
	| file fileResult fileName |
	textEditor
		ifNotNil: [self hasUnacceptedEdits ifTrue: [textEditor accept]].

	(fileResult := StandardFileMenu oldFile)
		ifNotNil: 
			[fileName := fileResult directory fullNameFor: fileResult name.
			file := FileStream readOnlyFileNamed: fileName.
			file ifNotNil:
				[file binary.
				self messageText:
						((MailMessage from: self messageText asString)
							addAttachmentFrom: file withName: fileResult name; text).
				file close]] ! !

!MailComposition methodsFor: '*Morphic-Support' stamp: 'dtl 2/9/2010 22:34'!
morphicOpen
	"open an interface for sending a mail message with the given initial 
	text "
	| textMorph buttonsList sendButton attachmentButton |
	morphicWindow := SystemWindow labelled: 'Mister Postman'.
	morphicWindow model: self.
	textEditor := textMorph := PluggableTextMorph
						on: self
						text: #messageText
						accept: #messageText:
						readSelection: nil
						menu: #menuGet:shifted:.
	morphicWindow addMorph: textMorph frame: (0 @ 0.1 corner: 1 @ 1).
	buttonsList := AlignmentMorph newRow.
	sendButton := PluggableButtonMorph
				on: self
				getState: nil
				action: #submit.
	sendButton
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		label: 'send message';
		setBalloonText: 'Accept any unaccepted edits and add this to the queue of messages to be sent';
		onColor: Color white offColor: Color white.
	buttonsList addMorphBack: sendButton.
	
	attachmentButton := PluggableButtonMorph
				on: self
				getState: nil
				action: #addAttachment.
	attachmentButton
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		label: 'add attachment';
		setBalloonText: 'Send a file with the message';
		onColor: Color white offColor: Color white.
	buttonsList addMorphBack: attachmentButton.
	
	morphicWindow addMorph: buttonsList frame: (0 @ 0 extent: 1 @ 0.1).
	morphicWindow openInWorld! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'kfr 10/27/2015 17:12'!
addFillStyleMenuItems: aMenu hand: aHand from: aMorph
	"Add the items for changing the current fill style of the receiver"
	self isRadialFill ifTrue:[
		aMenu add: 'linear gradient' translated target: self selector: #beLinearGradientIn: argument: aMorph.
	] ifFalse:[
		aMenu add: 'radial gradient' translated target: self selector: #beRadialGradientIn: argument: aMorph.
	].
	aMenu addLine.
	aMenu add: 'change color ramp' translated target: self selector: #changeColorRampIn:event: argument: aMorph.
	aMenu addLine.
	super addFillStyleMenuItems: aMenu hand: aHand from: aMorph.! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'ar 6/18/1999 09:49'!
addNewColorIn: aMorph event: evt
	^self inform:'not yet implemented'! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'ar 6/18/1999 07:25'!
beLinearGradientIn: aMorph
	self radial: false.
	aMorph changed.! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'ar 6/18/1999 07:25'!
beRadialGradientIn: aMorph
	self radial: true.
	aMorph changed.! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'kfr 10/21/2015 00:15'!
changeColorRampIn: aMorph event: evt
	^self changeColorSelector: #colorRamp:forMorph: hand: evt hand morph: aMorph originalColor:aMorph fillStyle colorRamp! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'kfr 10/21/2015 00:14'!
changeColorSelector: aSymbol hand: aHand morph: aMorph originalColor: originalColor 
	"Change either the firstColor or the lastColor (depending on aSymbol).  Put up a color picker to hande it.  We always use a modal picker so that the user can adjust both colors concurrently."

^(GradientEditor on: self selector: aSymbol forMorph: aMorph colorRamp: originalColor) openNear: aMorph fullBoundsInWorld. 


	"NewColorPickerMorph useIt
		ifTrue:
			[ (NewColorPickerMorph
				on: self
				originalColor: originalColor
				setColorSelector: aSymbol) openNear: aMorph fullBoundsInWorld ]
		ifFalse:
			[ ColorPickerMorph new
				 initializeModal: false ;
				 sourceHand: aHand ;
				 target: self ;
				 selector: aSymbol ;
				 argument: aMorph ;
				 originalColor: originalColor ;
				
				putUpFor: aMorph
				near: aMorph fullBoundsInWorld ]"! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'di 9/3/1999 11:34'!
changeFirstColorIn: aMorph event: evt
	^self changeColorSelector: #firstColor:forMorph:hand: hand: evt hand morph: aMorph originalColor: colorRamp first value! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'di 9/3/1999 11:34'!
changeSecondColorIn: aMorph event: evt
	^self changeColorSelector: #lastColor:forMorph:hand: hand: evt hand morph: aMorph originalColor: colorRamp last value! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'kfr 10/26/2015 15:37'!
colorRamp: aColorRamp forMorph: aMorph 
	colorRamp :=aColorRamp.
	isTranslucent := nil.
	pixelRamp := nil. 
	aMorph changed.! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'nk 7/18/2003 16:35'!
firstColor: aColor forMorph: aMorph hand: aHand
	colorRamp first value: aColor.
	isTranslucent := nil.
	pixelRamp := nil.
	aMorph changed.! !

!GradientFillStyle methodsFor: '*Morphic-Balloon' stamp: 'nk 7/18/2003 16:35'!
lastColor: aColor forMorph: aMorph hand: aHand
	colorRamp last value: aColor.
	isTranslucent := nil.
	pixelRamp := nil.
	aMorph changed.! !

Object subclass: #BorderStyle
	instanceVariableNames: ''
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Morphic-Borders'!

!BorderStyle commentStamp: 'kfr 10/27/2003 10:19' prior: 0!
See BorderedMorph

BorderedMorh new borderStyle: (BorderStyle inset width: 2); openInWorld.!

!BorderStyle class methodsFor: 'instance creation' stamp: 'sw 11/26/2001 16:05'!
borderStyleChoices
	"Answer the superset of all supported borderStyle symbols"

	^ #(simple inset raised complexAltFramed complexAltInset complexAltRaised complexFramed complexInset complexRaised)! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'yo 7/2/2004 17:21'!
borderStyleForSymbol: sym
	"Answer a border style corresponding to the given symbol"

	| aSymbol |
	aSymbol := sym == #none ifTrue: [#simple] ifFalse: [sym].
	^ self perform: aSymbol
"
	| aSymbol selector |
	aSymbol := sym == #none ifTrue: [#simple] ifFalse: [sym].
	selector := Vocabulary eToyVocabulary translationKeyFor: aSymbol.
	selector isNil ifTrue: [selector := aSymbol].
	^ self perform: selector
"
! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 8/25/2001 23:52'!
color: aColor width: aNumber
	^self width: aNumber color: aColor! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:01'!
complexAltFramed
	^ComplexBorder style: #complexAltFramed! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:01'!
complexAltInset
	^ComplexBorder style: #complexAltInset! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:00'!
complexAltRaised
	^ComplexBorder style: #complexAltRaised! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:00'!
complexFramed
	^ComplexBorder style: #complexFramed! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:00'!
complexInset
	^ComplexBorder style: #complexInset! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 15:00'!
complexRaised
	^ComplexBorder style: #complexRaised! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 8/25/2001 17:26'!
default
	^Default ifNil:[Default := self new]! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 14:59'!
inset
	^InsetBorder new! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 11/26/2001 14:59'!
raised
	^RaisedBorder new! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'sw 11/27/2001 15:22'!
simple
	"Answer a simple border style"

	^ SimpleBorder new! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'rr 6/21/2005 13:50'!
thinGray
	^ self width: 1 color: Color gray! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 8/25/2001 16:19'!
width: aNumber
	^self width: aNumber color: Color black! !

!BorderStyle class methodsFor: 'instance creation' stamp: 'ar 8/25/2001 16:19'!
width: aNumber color: aColor
	^SimpleBorder new color: aColor; width: aNumber; yourself! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 18:15'!
baseColor
	^Color transparent! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 18:15'!
baseColor: aColor
	"Ignored"! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:07'!
color
	^Color transparent! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 11/26/2001 15:22'!
colorsAtCorners
	^Array new: 4 withAll: self color! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:07'!
color: aColor
	"Ignored"! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:07'!
dotOfSize: diameter forDirection: aDirection
	| form |
	form := Form extent: diameter@diameter depth: Display depth.
	form getCanvas fillOval: form boundingBox color: self color.
	^form! !

!BorderStyle methodsFor: 'drawing' stamp: 'ar 8/25/2001 17:01'!
drawLineFrom: startPoint to: stopPoint on: aCanvas
	^aCanvas line: startPoint to: stopPoint width: self width color: self color! !

!BorderStyle methodsFor: 'drawing' stamp: 'ar 8/25/2001 16:24'!
frameOval: aRectangle on: aCanvas
	"Frame the given rectangle on aCanvas"
	aCanvas frameOval: aRectangle width: self width color: self color! !

!BorderStyle methodsFor: 'drawing' stamp: 'ar 8/25/2001 16:57'!
framePolygon: vertices on: aCanvas
	"Frame the given rectangle on aCanvas"
	self framePolyline: vertices on: aCanvas.
	self drawLineFrom: vertices last to: vertices first on: aCanvas.! !

!BorderStyle methodsFor: 'drawing' stamp: 'dgd 2/21/2003 22:59'!
framePolyline: vertices on: aCanvas 
	"Frame the given rectangle on aCanvas"

	| prev next |
	prev := vertices first.
	2 to: vertices size
		do: 
			[:i | 
			next := vertices at: i.
			self 
				drawLineFrom: prev
				to: next
				on: aCanvas.
			prev := next]! !

!BorderStyle methodsFor: 'drawing' stamp: 'ar 8/25/2001 16:24'!
frameRectangle: aRectangle on: aCanvas
	"Frame the given rectangle on aCanvas"
	aCanvas frameRectangle: aRectangle width: self width color: self color! !

!BorderStyle methodsFor: 'comparing' stamp: 'ar 8/25/2001 16:08'!
hash
	"hash is implemented because #= is implemented"
	^self species hash bitXor: (self width hash bitXor: self color hash)! !

!BorderStyle methodsFor: 'accessing' stamp: 'DS 4/22/2015 11:14:54.855'!
inset
	"Possible inset when honoring this border style. The default implementation uses #width to return a uniform inset."
	^ self width! !

!BorderStyle methodsFor: 'testing' stamp: 'ar 8/25/2001 16:08'!
isBorderStyle
	^true! !

!BorderStyle methodsFor: 'testing' stamp: 'ar 8/26/2001 19:30'!
isComplex
	^false! !

!BorderStyle methodsFor: 'initialize' stamp: 'ar 8/25/2001 16:06'!
releaseCachedState
	"Release any associated cached state"! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 17:51'!
style
	^#none! !

!BorderStyle methodsFor: 'color tracking' stamp: 'ar 8/25/2001 17:29'!
trackColorFrom: aMorph
	"If necessary, update our color to reflect a change in aMorphs color"! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:07'!
width
	^0! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:08'!
widthForRounding
	^self width! !

!BorderStyle methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:07'!
width: aNumber
	"Ignored"! !

!BorderStyle methodsFor: 'comparing' stamp: 'ar 8/25/2001 18:38'!
= aBorderStyle
	^self species = aBorderStyle species
		and:[self style == aBorderStyle style
		and:[self width = aBorderStyle width
		and:[self color = aBorderStyle color]]].! !

BorderStyle subclass: #SimpleBorder
	instanceVariableNames: 'baseColor color width'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Borders'!

!SimpleBorder commentStamp: 'kfr 10/27/2003 10:17' prior: 0!
see BorderedMorph!

SimpleBorder subclass: #ComplexBorder
	instanceVariableNames: 'style colors lineStyles'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Borders'!

!ComplexBorder commentStamp: 'nice 3/24/2010 07:36' prior: 0!
see BorderedMorph.

poly := polygon250 

baseColor := Color blue twiceLighter.
border := (ComplexBorder framed: 10) baseColor: poly color.
border frameRectangle: ((100@100 extent: 200@200) insetBy: -5) on: Display getCanvas.
baseColor := Color red twiceLighter.
border := (ComplexBorder framed: 10) baseColor: baseColor.
border drawPolygon: {100@100. 300@100. 300@300. 100@300} on: Display getCanvas.

border drawPolyPatchFrom: 100@200 via: 100@100 via: 200@100 to: 200@200 on: Display getCanvas.
border drawPolyPatchFrom: 100@100 via: 200@100 via: 200@200 to: 100@200 on: Display getCanvas.
border drawPolyPatchFrom: 200@100 via: 200@200 via: 100@200 to: 100@100 on: Display getCanvas.
border drawPolyPatchFrom: 200@200 via: 100@200 via: 100@100 to: 200@100 on: Display getCanvas.

border := (ComplexBorder raised: 10) baseColor: poly color.
border drawPolygon: poly getVertices on: Display getCanvas

360 / 16.0 22.5
points := (0 to: 15) collect:[:i| (Point r: 100 degrees: i*22.5) + 200].
Display getCanvas fillOval: (100@100 extent: 200@200) color: baseColor.
border drawPolygon: points on: Display getCanvas.

-1 to: points size + 1 do:[:i|
	border drawPolyPatchFrom: (points atWrap: i) via: (points atWrap: i+1) via: (points atWrap: i+2) to: (points atWrap: i+3) on: Display getCanvas.
].

Display getCanvas fillOval: (100@100 extent: 200@200) color: baseColor.
0 to: 36 do:[:i|
	border drawLineFrom: (Point r: 100 degrees: i*10) + 200 to: (Point r: 100 degrees: i+1*10) + 200
		on: Display getCanvas.
].
drawPolygon:
Point r: 1.0 degrees: 10
MessageTally spyOn:[
Display deferUpdates: true.
t1 := [1 to: 1000 do:[:i|
	border drawLineFrom: (100@100) to: (300@100) on: Display getCanvas.
	border drawLineFrom: (300@100) to: (300@300) on: Display getCanvas.
	border drawLineFrom: (300@300) to: (100@300) on: Display getCanvas.
	border drawLineFrom: (100@300) to: (100@100) on: Display getCanvas]] timeToRun.
Display deferUpdates: false.
].

MessageTally spyOn:[
Display deferUpdates: true.
t2 := [1 to: 1000 do:[:i|
	border drawLine2From: (100@100) to: (300@100) on: Display getCanvas.
	border drawLine2From: (300@100) to: (300@300) on: Display getCanvas.
	border drawLine2From: (300@300) to: (100@300) on: Display getCanvas.
	border drawLine2From: (100@300) to: (100@100) on: Display getCanvas]] timeToRun.
Display deferUpdates: false.
].

!

!ComplexBorder class methodsFor: 'instance creation' stamp: 'ar 8/25/2001 16:22'!
style: aSymbol
	^self new style: aSymbol! !

!ComplexBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:13'!
colors
	^colors ifNil:[colors := self computeColors].! !

!ComplexBorder methodsFor: 'private' stamp: 'aoy 2/17/2003 01:02'!
colorsForDirection: direction 
	"Return an array of colors describing the receiver in the given direction"

	| colorArray dT cc |
	cc := self colors.
	direction x * direction y <= 0 
		ifTrue: 
			["within up->right or down->left transition; no color blend needed"

			colorArray := (direction x > 0 or: [direction y < 0]) 
						ifTrue: 
							["up->right"
							cc copyFrom: 1 to: width]
						ifFalse: 
							["down->left"
							"colors are stored in reverse direction when following a line"
							(cc copyFrom: width + 1 to: cc size) reversed]]
		ifFalse: 
			["right->down or left->up transition; need color blend"

			colorArray := Array new: width.
			dT := direction x asFloat / (direction x + direction y).
			(direction x > 0 or: [direction y >= 0]) 
				ifTrue: 
					["top-right"

					1 to: width
						do: 
							[:i | 
							colorArray at: i put: ((cc at: i) mixed: dT with: (cc at: cc size - i + 1))]]
				ifFalse: 
					["bottom-left"

					1 to: width
						do: 
							[:i | 
							colorArray at: i put: ((cc at: cc size - i + 1) mixed: dT with: (cc at: i))]]].
	^colorArray! !

!ComplexBorder methodsFor: 'private' stamp: 'ar 8/25/2001 16:16'!
computeAltFramedColors
	| base light dark w hw colorArray param |
	base := self color asColor.
	light := Color white.
	dark := Color black.
	w := self width isPoint ifTrue:[self width x max: self width y] ifFalse:[self width].
	w := w asInteger.
	w = 1 ifTrue:[^{base mixed: 0.5 with: light. base mixed: 0.5 with: dark}].
	colorArray := Array new: w.
	hw := w // 2.
	"brighten"
	0 to: hw-1 do:[:i|
		param := 0.5 + (i asFloat / hw * 0.5).
		colorArray at: i+1 put: (base mixed: param with: dark). "brighten"
		colorArray at: w-i put: (base mixed: param with: light). "darken"
	].
	w odd ifTrue:[colorArray at: hw+1 put: base].
	^colorArray, colorArray! !

!ComplexBorder methodsFor: 'private' stamp: 'aoy 2/17/2003 01:03'!
computeAltInsetColors
	| base light dark w colorArray param hw |
	base := self color asColor.
	light := Color white.
	dark := Color black.
	w := self width isPoint 
				ifTrue: [self width x max: self width y]
				ifFalse: [self width].
	w := w asInteger.
	colorArray := Array new: w * 2.
	hw := 0.5 / w.
	0 to: w - 1
		do: 
			[:i | 
			param := false 
						ifTrue: 
							["whats this ???!! false ifTrue:[]"

							0.5 + (hw * i)]
						ifFalse: [0.5 + (hw * (w - i))].
			colorArray at: i + 1 put: (base mixed: param with: dark).	"darken"
			colorArray at: colorArray size - i put: (base mixed: param with: light)	"brighten"].
	^colorArray! !

!ComplexBorder methodsFor: 'private' stamp: 'aoy 2/17/2003 01:05'!
computeAltRaisedColors
	| base light dark w colorArray param hw |
	base := self color asColor.
	light := Color white.
	dark := Color black.
	w := self width isPoint 
				ifTrue: [self width x max: self width y]
				ifFalse: [self width].
	w := w asInteger.
	colorArray := Array new: w * 2.
	hw := 0.5 / w.
	0 to: w - 1
		do: 
			[:i | "again !! false ifTrue:[] ?!!"
			param := false ifTrue: [0.5 + (hw * i)] ifFalse: [0.5 + (hw * (w - i))].
			colorArray at: i + 1 put: (base mixed: param with: light).	"brighten"
			colorArray at: colorArray size - i put: (base mixed: param with: dark)	"darken"].
	^colorArray! !

!ComplexBorder methodsFor: 'private' stamp: 'ar 11/26/2001 15:00'!
computeColors
	width = 0 ifTrue:[^colors := #()].
	style == #complexFramed ifTrue:[^self computeFramedColors].
	style == #complexAltFramed ifTrue:[^self computeAltFramedColors].
	style == #complexRaised ifTrue:[^self computeRaisedColors].
	style == #complexAltRaised ifTrue:[^self computeAltRaisedColors].
	style == #complexInset ifTrue:[^self computeInsetColors].
	style == #complexAltInset ifTrue:[^self computeAltInsetColors].
	self error:'Unknown border style: ', style printString.! !

!ComplexBorder methodsFor: 'private' stamp: 'ar 8/25/2001 16:35'!
computeFramedColors
	| base light dark w hw colorArray param |
	base := self color asColor.
	light := Color white.
	dark := Color black.
	w := self width isPoint ifTrue:[self width x max: self width y] ifFalse:[self width].
	w := w asInteger.
	w = 1 ifTrue:[^{base mixed: 0.5 with: light. base mixed: 0.5 with: dark}].
	colorArray := Array new: w.
	hw := w // 2.
	"brighten"
	0 to: hw-1 do:[:i|
		param := 0.5 + (i asFloat / hw * 0.5).
		colorArray at: i+1 put: (base mixed: param with: light). "brighten"
		colorArray at: w-i put: (base mixed: param with: dark). "darken"
	].
	w odd ifTrue:[colorArray at: hw+1 put: base].
	^colorArray, colorArray! !

!ComplexBorder methodsFor: 'private' stamp: 'aoy 2/17/2003 01:06'!
computeInsetColors
	| base light dark w colorArray param hw |
	base := self color asColor.
	light := Color white.
	dark := Color black.
	w := self width isPoint 
				ifTrue: [self width x max: self width y]
				ifFalse: [self width].
	w := w asInteger.
	colorArray := Array new: w * 2.
	hw := 0.5 / w.
	0 to: w - 1
		do: 
			[:i | 
			param := true 
				ifTrue: [ 0.5 + (hw * i)]
				ifFalse: [0.5 + (hw * (w - i))].
			colorArray at: i + 1 put: (base mixed: param with: dark).	"darken"
			colorArray at: colorArray size - i put: (base mixed: param with: light)	"brighten"].
	^colorArray! !

!ComplexBorder methodsFor: 'private' stamp: 'aoy 2/17/2003 01:07'!
computeRaisedColors
	| base light dark w colorArray param hw |
	base := self color asColor.
	light := Color white.
	dark := Color black.
	w := self width isPoint 
				ifTrue: [self width x max: self width y]
				ifFalse: [self width].
	w := w asInteger.
	colorArray := Array new: w * 2.
	hw := 0.5 / w.
	0 to: w - 1
		do: 
			[:i | 
			param := true ifTrue: [0.5 + (hw * i)] ifFalse: [0.5 + (hw  * (w - i))].
			colorArray at: i + 1 put: (base mixed: param with: light).	"brighten"
			colorArray at: colorArray size - i put: (base mixed: param with: dark)	"darken"].
	^colorArray! !

!ComplexBorder methodsFor: 'drawing' stamp: 'aoy 2/17/2003 01:08'!
drawLineFrom: startPoint to: stopPoint on: aCanvas 
	"Here we're using the balloon engine since this is much faster than BitBlt w/ brushes."

	| delta length dir cos sin tfm w h w1 w2 h1 h2 fill |
	width isPoint 
		ifTrue: 
			[w := width x.
			h := width y]
		ifFalse: [w := h := width].
	w1 := w // 2.
	w2 := w - w1.
	h1 := h // 2.
	h2 := h - h1.
	"Compute the rotational transform from (0@0) -> (1@0) to startPoint -> stopPoint"
	delta := stopPoint - startPoint.
	length := delta r.
	dir := length > 1.0e-10 ifTrue: [delta / length] ifFalse: [ 1 @ 0].
	cos := dir dotProduct: 1 @ 0.
	sin := dir crossProduct: 1 @ 0.
	tfm := (MatrixTransform2x3 new)
				a11: cos;
				a12: sin;
				a21: sin negated;
				a22: cos.
	"Install the start point offset"
	tfm offset: startPoint.
	"Now get the fill style appropriate for the given direction"
	fill := self fillStyleForDirection: dir.
	"And draw..."
	aCanvas asBalloonCanvas transformBy: tfm
		during: 
			[:cc | 
			cc drawPolygon: { 
						(0 - w1) @ (0 - h1).	"top left"
						(length + w2) @ (0 - h1).	"top right"
						(length + w2) @ h2.	"bottom right"
						(0 - w1) @ h2	"bottom left"}
				fillStyle: fill]! !

!ComplexBorder methodsFor: 'drawing' stamp: 'ar 11/26/2001 15:10'!
drawPolyPatchFrom: startPoint to: stopPoint on: aCanvas usingEnds: endsArray

	| cos sin tfm fill dir fsOrigin fsDirection points x y |
	dir := (stopPoint - startPoint) normalized.
	"Compute the rotational transform from (0@0) -> (1@0) to startPoint -> stopPoint"
	cos := dir dotProduct: (1@0).
	sin := dir crossProduct: (1@0).
	"Now get the fill style appropriate for the given direction"
	fill := self fillStyleForDirection: dir.
false ifTrue:[
	"Transform the fill appropriately"
	fill := fill clone.
	"Note: Code below is inlined from tfm transformPoint:/transformDirection:"
	x := fill origin x. y := fill origin y.
	fsOrigin := ((x * cos) + (y * sin) + startPoint x) @
					((y * cos) - (x * sin) + startPoint y).
	x := fill direction x. y := fill direction y.
	fsDirection := ((x * cos) + (y * sin)) @ ((y * cos) - (x * sin)).
	fill origin: fsOrigin; 
		direction: fsDirection rounded; "NOTE: This is a bug in the balloon engine!!!!!!"
		normal: nil.
	aCanvas asBalloonCanvas drawPolygon: endsArray fillStyle: fill.
] ifFalse:[
	"Transform the points rather than the fills"
	tfm := (MatrixTransform2x3 new) a11: cos; a12: sin; a21: sin negated; a22: cos.
	"Install the start point offset"
	tfm offset: startPoint.
	points := endsArray collect:[:pt| tfm invertPoint: pt].
	aCanvas asBalloonCanvas transformBy: tfm during:[:cc|
		cc drawPolygon: points fillStyle: fill.
	].
].! !

!ComplexBorder methodsFor: 'private' stamp: 'ar 9/4/2001 19:51'!
fillStyleForDirection: direction
	"Fill the given form describing the receiver's look at a particular direction"
	| index fill dir |
	index := direction degrees truncated // 10 + 1.
	lineStyles ifNotNil:[
		fill := lineStyles at: index.
		fill ifNotNil:[^fill].
	].
	dir := Point r: 1.0 degrees: index - 1 * 10 + 5.
	fill := GradientFillStyle colors: (self colorsForDirection: dir).
	fill direction: 0 @ width asPoint y; radial: false.
	fill origin: ((width asPoint x // 2) @ (width asPoint y // 2)) negated.
	fill pixelRamp: (fill computePixelRampOfSize: 16).
	fill isTranslucent. "precompute"
	lineStyles ifNil:[lineStyles := Array new: 37].
	lineStyles at: index put: fill.
	^fill! !

!ComplexBorder methodsFor: 'drawing' stamp: 'ar 9/4/2001 19:51'!
framePolygon2: vertices on: aCanvas
	| dir1 dir2 dir3 nrm1 nrm2 nrm3 point1 point2 point3 
	 cross1 cross2 pointA pointB pointC pointD w p1 p2 p3 p4 balloon ends |
	balloon := aCanvas asBalloonCanvas.
	balloon == aCanvas ifFalse:[balloon deferred: true].
	ends := Array new: 4.
	w := width * 0.5.
	pointA := nil.
	1 to: vertices size do:[:i|
		p1 := vertices atWrap: i.
		p2 := vertices atWrap: i+1.
		p3 := vertices atWrap: i+2.
		p4 := vertices atWrap: i+3.

		dir1 := p2 - p1.
		dir2 := p3 - p2.
		dir3 := p4 - p3.

		i = 1 ifTrue:[
			"Compute the merge points of p1->p2 with p2->p3"
			cross1 := dir2 crossProduct: dir1.
			nrm1 := dir1 normalized. nrm1 := (nrm1 y * w) @ (0 - nrm1 x * w).
			nrm2 := dir2 normalized. nrm2 := (nrm2 y * w) @ (0 - nrm2 x * w).
			cross1 < 0 ifTrue:[nrm1 := nrm1 negated. nrm2 := nrm2 negated].
			point1 := (p1 x + nrm1 x) @ (p1 y + nrm1 y).
			point2 := (p2 x + nrm2 x) @ (p2 y + nrm2 y).
			pointA := self intersectFrom: point1 with: dir1 to: point2 with: dir2.
			point1 := (p1 x - nrm1 x) @ (p1 y - nrm1 y).
			point2 := (p2 x - nrm2 x) @ (p2 y - nrm2 y).
			pointB := self intersectFrom: point1 with: dir1 to: point2 with: dir2.
			pointB ifNotNil:[
				(pointB x - p2 x) abs + (pointB y - p2 y) abs > (4*w) ifTrue:[pointA := pointB := nil].
			].
		].

		"Compute the merge points of p2->p3 with p3->p4"
		cross2 := dir3 crossProduct: dir2.
		nrm2 := dir2 normalized. nrm2 := (nrm2 y * w) @ (0 - nrm2 x * w).
		nrm3 := dir3 normalized. nrm3 := (nrm3 y * w) @ (0 - nrm3 x * w).
		cross2 < 0 ifTrue:[nrm2 := nrm2 negated. nrm3 := nrm3 negated].
		point2 := (p2 x + nrm2 x) @ (p2 y + nrm2 y).
		point3 := (p3 x + nrm3 x) @ (p3 y + nrm3 y).
		pointC := self intersectFrom: point2 with: dir2 to: point3 with: dir3.
		point2 := (p2 x - nrm2 x) @ (p2 y - nrm2 y).
		point3 := (p3 x - nrm3 x) @ (p3 y - nrm3 y).
		pointD := self intersectFrom: point2 with: dir2 to: point3 with: dir3.
		pointD ifNotNil:[
			(pointD x - p3 x) abs + (pointD y - p3 y) abs > (4*w) ifTrue:[pointC := pointD := nil].
		].
		cross1 * cross2 < 0.0 ifTrue:[
			point1 := pointA.
			pointA := pointB.
			pointB := point1.
			cross1 := 0.0 - cross1].
		ends at: 1 put: pointA; at: 2 put: pointB; at: 3 put: pointD; at: 4 put: pointC.
		pointA ifNil:["degenerate and slow"
			nrm2 := dir2 normalized. nrm2 := (nrm2 y * w) @ (0 - nrm2 x * w).
			cross1 < 0 ifTrue:[nrm2 := nrm2 negated].
			point2 := (p2 x + nrm2 x) @ (p2 y + nrm2 y).
			ends at: 1 put: point2].
		pointB ifNil:["degenerate and slow"
			nrm2 := dir2 normalized. nrm2 := (nrm2 y * w) @ (0 - nrm2 x * w).
			cross1 < 0 ifTrue:[nrm2 := nrm2 negated].
			point2 := (p2 x - nrm2 x) @ (p2 y - nrm2 y).
			ends at: 2 put: point2].
		pointC ifNil:["degenerate and slow"
			nrm2 := dir2 normalized. nrm2 := (nrm2 y * w) @ (0 - nrm2 x * w).
			cross2 < 0 ifTrue:[nrm2 := nrm2 negated].
			point2 := (p3 x + nrm2 x) @ (p3 y + nrm2 y).
			ends at: 4 put: point2].
		pointD ifNil:["degenerate and slow"
			nrm2 := dir2 normalized. nrm2 := (nrm2 y * w) @ (0 - nrm2 x * w).
			cross2 < 0 ifTrue:[nrm2 := nrm2 negated].
			point2 := (p3 x - nrm2 x) @ (p3 y - nrm2 y).
			ends at: 3 put: point2].

		self drawPolyPatchFrom: p2 to: p3 on: balloon usingEnds: ends.
		pointA := pointC.
		pointB := pointD.
		cross1 := cross2.
	].
	balloon == aCanvas ifFalse:[balloon flush].! !

!ComplexBorder methodsFor: 'drawing' stamp: 'ar 9/4/2001 19:50'!
framePolygon: vertices on: aCanvas
	| dir1 dir2 dir3 nrm1 nrm2 nrm3 point1 point2 point3 
	 cross1 cross2 pointA pointB pointC pointD w p1 p2 p3 p4 balloon ends pointE pointF |
	balloon := aCanvas asBalloonCanvas.
	balloon == aCanvas ifFalse:[balloon deferred: true].
	ends := Array new: 6.
	w := width * 0.5.
	pointA := nil.
	1 to: vertices size do:[:i|
		p1 := vertices atWrap: i.
		p2 := vertices atWrap: i+1.
		p3 := vertices atWrap: i+2.
		p4 := vertices atWrap: i+3.

		dir1 := p2 - p1.
		dir2 := p3 - p2.
		dir3 := p4 - p3.

		(i = 1 | true) ifTrue:[
			"Compute the merge points of p1->p2 with p2->p3"
			cross1 := dir2 crossProduct: dir1.
			nrm1 := dir1 normalized. nrm1 := (nrm1 y * w) @ (0 - nrm1 x * w).
			nrm2 := dir2 normalized. nrm2 := (nrm2 y * w) @ (0 - nrm2 x * w).
			cross1 < 0 ifTrue:[nrm1 := nrm1 negated. nrm2 := nrm2 negated].
			point1 := (p1 x + nrm1 x) @ (p1 y + nrm1 y).
			point2 := (p2 x + nrm2 x) @ (p2 y + nrm2 y).
			pointA := self intersectFrom: point1 with: dir1 to: point2 with: dir2.
			point1 := (p1 x - nrm1 x) @ (p1 y - nrm1 y).
			point2 := (p2 x - nrm2 x) @ (p2 y - nrm2 y).
			pointB := point1 + dir1 + point2 * 0.5.
			pointB := p2 + ((pointB - p2) normalized * w).
			pointC := point2.
		].

		"Compute the merge points of p2->p3 with p3->p4"
		cross2 := dir3 crossProduct: dir2.
		nrm2 := dir2 normalized. nrm2 := (nrm2 y * w) @ (0 - nrm2 x * w).
		nrm3 := dir3 normalized. nrm3 := (nrm3 y * w) @ (0 - nrm3 x * w).
		cross2 < 0 ifTrue:[nrm2 := nrm2 negated. nrm3 := nrm3 negated].
		point2 := (p2 x + nrm2 x) @ (p2 y + nrm2 y).
		point3 := (p3 x + nrm3 x) @ (p3 y + nrm3 y).
		pointD := self intersectFrom: point2 with: dir2 to: point3 with: dir3.
		point2 := (p2 x - nrm2 x) @ (p2 y - nrm2 y).
		point3 := (p3 x - nrm3 x) @ (p3 y - nrm3 y).
		pointF := point2 + dir2.
		pointE := pointF + point3 * 0.5.
		pointE := p3 + ((pointE - p3) normalized * w).
		cross1 * cross2 < 0.0 ifTrue:[
			ends
				at: 1 put: pointA;
				at: 2 put: pointB;
				at: 3 put: pointC;
				at: 4 put: pointD;
				at: 5 put: pointE;
				at: 6 put: pointF.
		] ifFalse:[
			ends 
				at: 1 put: pointA; 
				at: 2 put: pointB;
				at: 3 put: pointC; 
				at: 4 put: pointF; 
				at: 5 put: pointE;
				at: 6 put: pointD.
		].
		self drawPolyPatchFrom: p2 to: p3 on: balloon usingEnds: ends.
		pointA := pointD.
		pointB := pointE.
		pointC := pointF.
		cross1 := cross2.
	].
	balloon == aCanvas ifFalse:[balloon flush].! !

!ComplexBorder methodsFor: 'drawing' stamp: 'ar 8/26/2001 19:01'!
frameRectangle: aRectangle on: aCanvas
	"Note: This uses BitBlt since it's roughly a factor of two faster for rectangles"
	| w h r |
	self colors ifNil:[^super frameRectangle: aRectangle on: aCanvas].
	w := self width.
	w isPoint ifTrue:[h := w y. w := w x] ifFalse:[h := w].
	1 to: h do:[:i| "top/bottom"
		r := (aRectangle topLeft + (i-1)) extent: (aRectangle width - (i-1*2))@1. "top"
		aCanvas fillRectangle: r color: (colors at: i).
		r := (aRectangle bottomLeft + (i @ (0-i))) extent: (aRectangle width - (i-1*2) - 1)@1. "bottom"
		aCanvas fillRectangle: r color: (colors at: colors size - i + 1).
	].
	1 to: w do:[:i| "left/right"
		r := (aRectangle topLeft + (i-1)) extent: 1@(aRectangle height - (i-1*2)). "left"
		aCanvas fillRectangle: r color: (colors at: i).
		r := aRectangle topRight + ((0-i)@i) extent: 1@(aRectangle height - (i-1*2) - 1). "right"
		aCanvas fillRectangle: r color: (colors at: colors size - i + 1).
	].! !

!ComplexBorder methodsFor: 'private' stamp: 'ar 8/26/2001 23:39'!
intersectFrom: startPt with: startDir to: endPt with: endDir
	"Compute the intersection of two lines. Return nil if either
		* the intersection does not exist, or
		* the intersection is 'before' startPt, or
		* the intersection is 'after' endPt
	"
	| det deltaPt alpha beta |
	det := (startDir x * endDir y) - (startDir y * endDir x).
	det = 0.0 ifTrue:[^nil]. "There's no solution for it"
	deltaPt := endPt - startPt.
	alpha := (deltaPt x * endDir y) - (deltaPt y * endDir x).
	beta := (deltaPt x * startDir y) - (deltaPt y * startDir x).
	alpha := alpha / det.
	beta := beta / det.
	alpha < 0 ifTrue:[^nil].
	beta > 1.0 ifTrue:[^nil].
	"And compute intersection"
	^(startPt x + (alpha * startDir x)) @ (startPt y + (alpha * startDir y))! !

!ComplexBorder methodsFor: 'testing' stamp: 'ar 8/26/2001 19:30'!
isComplex
	^true! !

!ComplexBorder methodsFor: 'initialize' stamp: 'ar 11/26/2001 14:43'!
releaseCachedState
	colors := nil.
	lineStyles := nil.! !

!ComplexBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:22'!
style
	^style! !

!ComplexBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:22'!
style: newStyle
	style == newStyle ifTrue:[^self].
	style := newStyle.
	self releaseCachedState.! !

!ComplexBorder methodsFor: 'color tracking' stamp: 'ar 8/25/2001 18:17'!
trackColorFrom: aMorph
	baseColor ifNil:[self color: aMorph raisedColor].! !

!ComplexBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:14'!
widthForRounding
	^0! !

SimpleBorder subclass: #InsetBorder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Borders'!

!InsetBorder commentStamp: 'kfr 10/27/2003 09:32' prior: 0!
see BorderedMorph!

!InsetBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:11'!
bottomRightColor
	^width = 1 
		ifTrue: [color twiceLighter]
		ifFalse: [color lighter]! !

!InsetBorder methodsFor: 'accessing' stamp: 'ar 11/26/2001 15:23'!
colorsAtCorners
	| c c14 c23 |
	c := self color.
	c14 := c lighter. c23 := c darker.
	^Array with: c23 with: c14 with: c14 with: c23.! !

!InsetBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 17:51'!
style
	^#inset! !

!InsetBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:11'!
topLeftColor
	^width = 1 
		ifTrue: [color twiceDarker]
		ifFalse: [color darker]! !

!InsetBorder methodsFor: 'color tracking' stamp: 'ar 8/25/2001 18:17'!
trackColorFrom: aMorph
	baseColor ifNil:[self color: aMorph insetColor].! !

SimpleBorder subclass: #RaisedBorder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Borders'!

!RaisedBorder commentStamp: 'kfr 10/27/2003 09:32' prior: 0!
see BorderedMorph!

!RaisedBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 17:34'!
bottomRightColor
	^width = 1 
		ifTrue: [color twiceDarker]
		ifFalse: [color darker]! !

!RaisedBorder methodsFor: 'accessing' stamp: 'ar 11/26/2001 15:23'!
colorsAtCorners
	| c c14 c23 |
	c := self color.
	c14 := c lighter. c23 := c darker.
	^Array with: c14 with: c23 with: c23 with: c14! !

!RaisedBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 17:51'!
style
	^#raised! !

!RaisedBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 17:34'!
topLeftColor
	^width = 1 
		ifTrue: [color twiceLighter]
		ifFalse: [color lighter]! !

!RaisedBorder methodsFor: 'color tracking' stamp: 'ar 8/25/2001 18:17'!
trackColorFrom: aMorph
	baseColor ifNil:[self color: aMorph raisedColor].! !

!SimpleBorder methodsFor: 'accessing' stamp: 'ul 4/7/2016 20:01'!
baseColor

	^baseColor! !

!SimpleBorder methodsFor: 'accessing' stamp: 'ul 4/7/2016 20:03'!
baseColor: aColor

	baseColor = aColor ifTrue:[^self].
	baseColor := aColor.
	self color: aColor "#color: will do  #releaseCachedState"! !

!SimpleBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:09'!
bottomRightColor
	^color! !

!SimpleBorder methodsFor: 'accessing' stamp: 'ul 4/7/2016 20:01'!
color

	^color! !

!SimpleBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:10'!
color: aColor
	color = aColor ifTrue:[^self].
	color := aColor.
	self releaseCachedState.! !

!SimpleBorder methodsFor: 'drawing' stamp: 'aoy 2/17/2003 01:14'!
drawLineFrom: startPoint to: stopPoint on: aCanvas 
	| lineColor |
	lineColor := (stopPoint truncated quadrantOf: startPoint truncated) > 2 
				ifTrue: [self topLeftColor]
				ifFalse: [self bottomRightColor].
	aCanvas 
		line: startPoint
		to: stopPoint 
		width: self width
		color: lineColor! !

!SimpleBorder methodsFor: 'drawing' stamp: 'ar 8/25/2001 16:27'!
frameRectangle: aRectangle on: aCanvas
	aCanvas frameAndFillRectangle: aRectangle
		fillColor: Color transparent
		borderWidth: self width
		topLeftColor: self topLeftColor
		bottomRightColor: self bottomRightColor.! !

!SimpleBorder methodsFor: 'initialize-release' stamp: 'ul 4/7/2016 20:10'!
initialize

	width := 0.
	baseColor := color := Color transparent! !

!SimpleBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 17:52'!
style
	^#simple! !

!SimpleBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:09'!
topLeftColor
	^color! !

!SimpleBorder methodsFor: 'accessing' stamp: 'ul 4/7/2016 19:57'!
width

	^width! !

!SimpleBorder methodsFor: 'accessing' stamp: 'ar 8/25/2001 16:10'!
width: aNumber
	width = aNumber ifTrue:[^self].
	width := aNumber truncated max: (width isPoint ifTrue:[0@0] ifFalse:[0]).
	self releaseCachedState.! !

Object subclass: #ComplexProgressIndicator
	instanceVariableNames: 'formerWorld targetMorph estimate prevData formerProcess translucentMorph userSuppliedMorph specificHistory historyCategory cumulativeStageTime formerProject newRatio stageCompleted start'
	classVariableNames: 'History'
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!ComplexProgressIndicator commentStamp: '<historical>' prior: 0!
Note: in an effort to remove the progress indicator if a walkback occurs, #withProgressDo: must be sent from the current uiProcess. Hopefully we can relax this restriction in the future. !

!ComplexProgressIndicator class methodsFor: 'as yet unclassified' stamp: 'nice 12/27/2009 03:15'!
historyReport
"
ComplexProgressIndicator historyReport
"
	| answer |
	History ifNil: [^Beeper beep].
	answer := String streamContents: [ :strm |
		(History keys asArray sort: [ :a :b | a asString <= b asString]) do: [ :k |
			| data |
			strm nextPutAll: k printString; cr.
			data := History at: k.
			(data keys asArray sort: [ :a :b | a asString <= b asString]) do: [ :dataKey |
				strm tab; nextPutAll: dataKey printString,'  ',
					(data at: dataKey) asArray printString; cr.
			].
			strm cr.
		].
	].
	StringHolder new
		contents: answer contents;
		openLabel: 'Progress History'! !

!ComplexProgressIndicator methodsFor: 'as yet unclassified' stamp: 'nk 8/18/2004 16:43'!
addProgressDecoration: extraParam 
	| f m |
	targetMorph ifNil: [^self].
	(extraParam isForm) 
		ifTrue: 
			[targetMorph 
				submorphsDo: [:mm | (mm isSketchMorph) ifTrue: [mm delete]].
			f := Form extent: extraParam extent depth: extraParam depth.
			extraParam displayOn: f.
			m := SketchMorph withForm: f.
			m align: m fullBounds leftCenter
				with: targetMorph fullBounds leftCenter + (2 @ 0).
			targetMorph addMorph: m.
			^self].
	(extraParam isString) 
		ifTrue: 
			[targetMorph 
				submorphsDo: [:mm | (mm isKindOf: StringMorph) ifTrue: [mm delete]].
			m := StringMorph contents: extraParam translated.
			m align: m fullBounds bottomCenter + (0 @ 8)
				with: targetMorph bounds bottomCenter.
			targetMorph addMorph: m.
			^self]! !

!ComplexProgressIndicator methodsFor: 'as yet unclassified' stamp: 'nice 12/26/2009 01:21'!
backgroundWorldDisplay

	self flag: #bob.		"really need a better way to do this"

			"World displayWorldSafely."

	"ugliness to try to track down a possible error"


	[World displayWorld] ifError: [ :a :b |
		| f |
		stageCompleted := 999.
		f := FileDirectory default fileNamed: 'bob.errors'.
		f nextPutAll: a printString,'  ',b printString; cr; cr.
		f nextPutAll: 'worlds equal ',(formerWorld == World) printString; cr; cr.
		f nextPutAll: thisContext longStack; cr; cr.
		f nextPutAll: formerProcess suspendedContext longStack; cr; cr.
		f close. Beeper beep.
	].
! !

!ComplexProgressIndicator methodsFor: 'as yet unclassified' stamp: 'nice 12/26/2009 01:22'!
forkProgressWatcher

	[
	| killTarget |
		[stageCompleted < 999 and: 
				[formerProject == Project current and: 
				[formerWorld == World and: 
				[translucentMorph world notNil and:
				[formerProcess suspendedContext notNil and: 
				[Project uiProcess == formerProcess]]]]]] whileTrue: [

			translucentMorph setProperty: #revealTimes toValue: 
					{(Time millisecondClockValue - start max: 1). (estimate * newRatio max: 1)}.
			translucentMorph changed.
			translucentMorph owner addMorphInLayer: translucentMorph.
			(Time millisecondClockValue - WorldState lastCycleTime) abs > 500 ifTrue: [
				self backgroundWorldDisplay
			].
			(Delay forMilliseconds: 100) wait.
		].
		translucentMorph removeProperty: #revealTimes.
		self loadingHistoryAt: 'total' add: (Time millisecondClockValue - start max: 1).
		killTarget := targetMorph ifNotNil: [
			targetMorph valueOfProperty: #deleteOnProgressCompletion
		].
		formerWorld == World ifTrue: [
			translucentMorph delete.
			killTarget ifNotNil: [killTarget delete].
		] ifFalse: [
			translucentMorph privateDeleteWithAbsolutelyNoSideEffects.
			killTarget ifNotNil: [killTarget privateDeleteWithAbsolutelyNoSideEffects].
		].
	] forkAt: Processor lowIOPriority.! !

!ComplexProgressIndicator methodsFor: 'as yet unclassified' stamp: 'RAA 5/23/2000 10:00'!
historyCategory: aKey

	History ifNil: [History := Dictionary new].
	specificHistory := History
		at: aKey
		ifAbsentPut: [Dictionary new].
	^specificHistory
! !

!ComplexProgressIndicator methodsFor: 'as yet unclassified' stamp: 'RAA 5/23/2000 09:55'!
loadingHistoryAt: aKey add: aNumber

	(self loadingHistoryDataForKey: aKey) add: aNumber.

! !

!ComplexProgressIndicator methodsFor: 'as yet unclassified' stamp: 'RAA 5/23/2000 10:02'!
loadingHistoryDataForKey: anObject

	| answer |
	answer := specificHistory 
		at: anObject
		ifAbsentPut: [OrderedCollection new].
	answer size > 50 ifTrue: [
		answer := answer copyFrom: 25 to: answer size.
		specificHistory at: anObject put: answer.
	].
	^answer

! !

!ComplexProgressIndicator methodsFor: 'as yet unclassified' stamp: 'RAA 6/29/2000 11:31'!
targetMorph: aMorph

	targetMorph := aMorph! !

!ComplexProgressIndicator methodsFor: 'as yet unclassified' stamp: 'nice 12/27/2009 03:11'!
withProgressDo: aBlock

	| safetyFactor totals trialRect delta targetOwner |

	Smalltalk isMorphic ifFalse: [^aBlock value].
	formerProject := Project current.
	formerWorld := World.
	formerProcess := Processor activeProcess.
	targetMorph
		ifNil: [targetMorph := ProgressTargetRequestNotification signal].
	targetMorph ifNil: [
		trialRect := Rectangle center: Sensor cursorPoint extent: 80@80.
		delta := trialRect amountToTranslateWithin: formerWorld bounds.
		trialRect := trialRect translateBy: delta.
		translucentMorph := TranslucentProgessMorph new
			opaqueBackgroundColor: Color white;
			bounds: trialRect;
			openInWorld: formerWorld.
	] ifNotNil: [
		targetOwner := targetMorph owner.
		translucentMorph := TranslucentProgessMorph new
			setProperty: #morphicLayerNumber toValue: targetMorph morphicLayerNumber - 0.1;
			bounds: targetMorph boundsInWorld;
			openInWorld: targetMorph world.
	].
	stageCompleted := 0.
	safetyFactor := 1.1.	"better to guess high than low"
	translucentMorph setProperty: #progressStageNumber toValue: 1.
	translucentMorph hide.
	targetOwner ifNotNil: [targetOwner hide].
	totals := self loadingHistoryDataForKey: 'total'.
	newRatio := 1.0.
	estimate := totals size < 2 ifTrue: [
		15000		"be a pessimist"
	] ifFalse: [
		(totals sum - totals max) / (totals size - 1 max: 1) * safetyFactor.
	].
	start := Time millisecondClockValue.
	self forkProgressWatcher.

	[
		aBlock 
			on: ProgressInitiationException
			do: [ :ex | 
				ex sendNotificationsTo: [ :min :max :curr |
					"ignore this as it is inaccurate"
				].
			].
	] on: ProgressNotification do: [ :note | | stageCompletedString |
		translucentMorph show.
		targetOwner ifNotNil: [targetOwner show].
		note extraParam ifNotNil:[self addProgressDecoration: note extraParam].
		stageCompletedString := (note messageText findTokens: ' ') first.
		stageCompleted := (stageCompletedString copyUpTo: $:) asNumber.
		cumulativeStageTime := Time millisecondClockValue - start max: 1.
		prevData := self loadingHistoryDataForKey: stageCompletedString.
		prevData isEmpty ifFalse: [
			newRatio := (cumulativeStageTime / (prevData average max: 1)) asFloat.
		].
		self 
			loadingHistoryAt: stageCompletedString 
			add: cumulativeStageTime.
		translucentMorph 
			setProperty: #progressStageNumber 
			toValue: stageCompleted + 1.
		note resume.
	].

	stageCompleted := 999.	"we may or may not get here"

! !

Object subclass: #DamageRecorder
	instanceVariableNames: 'invalidRects totalRepaint'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!DamageRecorder class methodsFor: 'instance creation' stamp: ''!
new

	^ super new reset
! !

!DamageRecorder methodsFor: 'recording' stamp: ''!
doFullRepaint
	"Record that a full redisplay is needed. No further damage rectangles will be recorded until after the next reset."

	^ totalRepaint := true.
! !

!DamageRecorder methodsFor: 'recording' stamp: ''!
invalidRectsFullBounds: aRectangle
	"Return a collection of damaged rectangles for the given canvas. If a total repaint has been requested, return the given rectangle."

	totalRepaint
		ifTrue: [^ Array with: aRectangle]
		ifFalse: [^ invalidRects copy].

! !

!DamageRecorder methodsFor: 'recording' stamp: 'nice 12/26/2009 01:28'!
recordInvalidRect: newRect
	"Record the given rectangle in my damage list, a list of rectangular areas of the display that should be redraw on the next display cycle."
	"Details: Damaged rectangles are often identical or overlap significantly. In these cases, we merge them to reduce the number of damage rectangles that must be processed when the display is updated. Moreover, above a certain threshold, we ignore the individual rectangles completely, and simply do a complete repaint on the next cycle."

	| mergeRect |
	totalRepaint ifTrue: [^ self].  "planning full repaint; don't bother collecting damage"

	invalidRects do:
		[:rect |
		| a |
		((a := (rect intersect: newRect) area) > 40
			and: ["Avoid combining a vertical and horizontal rects.
				  Can make a big diff and we only test when likely."
				  a > (newRect area // 4) or: [a > (rect area // 4)]])
			ifTrue:
			["merge rectangle in place (see note below) if there is significant overlap"
			rect setOrigin: (rect origin min: newRect origin) truncated
				corner: (rect corner max: newRect corner) truncated.
			^ self]].


	invalidRects size >= 15 ifTrue:
		["if there are too many separate areas, merge them all"
		mergeRect := Rectangle merging: invalidRects.
		self reset.
		invalidRects addLast: mergeRect].

	"add the given rectangle to the damage list"
	"Note: We make a deep copy of all rectangles added to the damage list,
		since rectangles in this list may be extended in place."
	invalidRects addLast:
		(newRect topLeft truncated corner: newRect bottomRight truncated).
! !

!DamageRecorder methodsFor: 'initialization' stamp: 'sma 6/5/2000 11:55'!
reset
	"Clear the damage list."

	invalidRects := OrderedCollection new: 15.
	totalRepaint := false
! !

!DamageRecorder methodsFor: 'testing' stamp: 'dgd 2/22/2003 14:43'!
updateIsNeeded
	"Return true if the display needs to be updated."

	^totalRepaint or: [invalidRects notEmpty]! !

Object subclass: #Editor
	instanceVariableNames: 'morph selectionShowing'
	classVariableNames: 'BlinkingCursor DestructiveBackWord DumbbellCursor KeystrokeActions SelectionsMayShrink'
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!Editor commentStamp: '<historical>' prior: 0!
New text editors.
TextEditor provides most of the functionality that used to be in TextMorphEditor.
SmalltalkEditor is has Smalltalk code specific features.
SimpleEditor provides basic functionality for single line text editing. It does not handle fonts and styles, aligning and Smalltalk utilities. It handles one single line.
CellStyleEditor allows entering alphabetic characters using only number keys, like many cell phones do.!

!Editor class methodsFor: 'preferences' stamp: 'cmm 4/19/2010 20:58'!
blinkingCursor
	<preference: 'Blinking Text Cursor'
		category: 'Morphic'
		description: 'When true, the text cursor will blink.'
		type: #Boolean>
	^ BlinkingCursor ifNil: [ true ]! !

!Editor class methodsFor: 'preferences' stamp: 'cmm 4/5/2010 15:57'!
blinkingCursor: aBoolean
	BlinkingCursor := aBoolean! !

!Editor class methodsFor: 'preferences' stamp: 'cmm 2/12/2010 15:56'!
destructiveBackWord
	<preference: 'Destructive Back-Word'
		category: 'Morphic'
		description: 'Indicates whether the back-word command deletes, or merely selects, the prior word.'
		type: #Boolean>
	^ DestructiveBackWord ifNil: [ true ]! !

!Editor class methodsFor: 'preferences' stamp: 'cmm 2/12/2010 15:57'!
destructiveBackWord: aBoolean
	DestructiveBackWord := aBoolean! !

!Editor class methodsFor: 'preferences' stamp: 'cmm 4/5/2010 16:23'!
dumbbellCursor
	<preference: 'Dumbbell-shaped Text Cursor'
		category: 'Morphic'
		description: 'When true, the text cursor assumes the shape of a dumbbell, otherwise a vertical bar..'
		type: #Boolean>
	^ DumbbellCursor ifNil: [ false ]! !

!Editor class methodsFor: 'preferences' stamp: 'cmm 4/5/2010 16:23'!
dumbbellCursor: aBoolean
	DumbbellCursor := aBoolean! !

!Editor class methodsFor: 'class initialization' stamp: 'jmv 9/7/2010 19:37'!
initialize
	"
	Editor initialize
	"
	self initializeKeystrokeActions.
	self allSubclassesDo: [ :c | c initialize ]! !

!Editor class methodsFor: 'class initialization' stamp: 'jmv 9/7/2010 23:49'!
initializeKeystrokeActions
	"Initialize the table for regular (i.e. non-command) keystroke dispatch"
	"
	self initializeKeystrokeActions
	"
	| actions |
	actions := Array new: 256 withAll: #normalCharacter:.
	0 to: 31 do: [ :i | actions at: i+1 put: #noop: ].
	actions at: 1 + 1 put: #cursorHome:.				"home key"
	actions at: 3 + 1 put: #enter:.						"enter / return key"
	actions at: 4 + 1 put: #cursorEnd:.				"end key"
	actions at: 5 + 1 put: #noop:.						"insert key"
	actions at: 8 + 1 put: #backspace:.				"macDelete winBackspace key"
	actions at: 9 + 1 put: #normalCharacter:.		"tab"
	actions at: 11 + 1 put: #cursorPageUp:.			"page up key"
	actions at: 12 + 1 put: #cursorPageDown:.		"page down key"
	actions at: 13 + 1 put: #enter:.					"enter / return key"
	actions at: 27 + 1 put: #offerMenuFromEsc:.	"escape key"
	actions at: 28 + 1 put: #cursorLeft:.				"left arrow key"
	actions at: 29 + 1 put: #cursorRight:.				"right arrow key"
	actions at: 30 + 1 put: #cursorUp:.				"up arrow key"
	actions at: 31 + 1 put: #cursorDown:.			"down arrow key"
	actions at: 127 + 1 put: #forwardDelete:.		"winDelete key"
	KeystrokeActions := actions! !

!Editor class methodsFor: 'preferences' stamp: 'ar 1/2/2010 14:45'!
selectionsMayShrink
	<preference: 'Selections may shrink'
		category: 'Morphic'
		description: 'When true, allows selection to shrink when using shift+cursor keys'
		type: #Boolean>
	^SelectionsMayShrink ifNil:[true]! !

!Editor class methodsFor: 'preferences' stamp: 'cmm 2/12/2010 15:56'!
selectionsMayShrink: aBoolean
	SelectionsMayShrink := aBoolean! !

!Editor class methodsFor: 'keyboard shortcut tables' stamp: 'jmv 11/4/2008 15:25'!
specialShiftCmdKeys

"Private - return array of key codes that represent single keys acting
as if shift-command were also being pressed"

^#(
	1	"home"
	3	"enter"
	4	"end"
	8	"backspace"
	11	"page up"
	12	"page down"
	27	"escape"
	28	"left arrow"
	29	"right arrow"
	30	"up arrow"
	31	"down arrow"
	127	"delete"
	)! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'mt 11/7/2015 17:54'!
backspace: aKeyboardEvent 
	"Backspace over the last character."

	| startIndex |
	aKeyboardEvent shiftPressed ifTrue: [^ self backWord: aKeyboardEvent].
	startIndex := self markIndex +
				(self hasCaret ifTrue: [0] ifFalse: [1]).
	startIndex := 1 max: startIndex - 1.

	^ self backTo: startIndex! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'nice 8/3/2011 20:08'!
backWord: aKeyboardEvent 
	^ self class destructiveBackWord 
		ifTrue: [ self destructiveBackWord: aKeyboardEvent ]
		ifFalse: [ self nonDestructiveBackWord: aKeyboardEvent ]! !

!Editor methodsFor: 'private' stamp: 'jmv 12/21/2010 14:46'!
beginningOfLine: position
	"Redefined in subclasses using Paragraph support"
	^ self beginningOfParagraph: position! !

!Editor methodsFor: 'private' stamp: 'jmv 9/17/2010 13:33'!
beginningOfNextParagraph: position
	| s |
	s := self string.
	^ (s
		indexOf: Character cr
		startingAt: position
		ifAbsent: [ s size ]) + 1! !

!Editor methodsFor: 'private' stamp: 'jmv 9/17/2010 13:37'!
beginningOfParagraph: position
	^ (self string
		lastIndexOf: Character cr
		startingAt: position
		ifAbsent: [ 0 ]) + 1.! !

!Editor methodsFor: 'private' stamp: 'jmv 9/17/2010 13:39'!
beginningOfText
	^1! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 3/16/2011 22:10'!
clearSelection

	self selectFrom: 1 to: 0! !

!Editor methodsFor: 'menu messages' stamp: 'jmv 11/5/2008 10:16'!
clipboardText

	^ Clipboard clipboardText! !

!Editor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 23:42'!
clipboardTextPut: text

	^ Clipboard clipboardText: text! !

!Editor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:37'!
copySelection: aKeyboardEvent
	"Copy the current text selection."

	self copySelection.
	^true! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:37'!
crlf: aKeyboardEvent
	"Append a line feed character to the stream of characters."

	self addString: String crlf.
	^false! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/7/2010 17:03'!
crWithIndent: aKeyboardEvent

	"Only for SmalltalkEditor. Regular editors don't indent"
	^ self cr: aKeyboardEvent! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/7/2010 17:01'!
cr: aKeyboardEvent
	"Append a carriage return character to the stream of characters."

	self addString: Character cr asString.
	^false! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'nice 8/3/2011 20:40'!
cursorDown: aKeyboardEvent
	"Private - Move cursor from position in current line to same position in
	next line. If next line too short, put at end. If shift key down,
	select."
	self insertAndCloseTypeIn.
	self 
		moveCursor: [:position | self
				sameColumn: position
				newLine: [:line | line + 1]
				forward: true]
		forward: true
		event: aKeyboardEvent
		specialBlock: [:dummy | dummy].
	^true! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'nice 8/3/2011 20:40'!
cursorLeft: aKeyboardEvent
	"Private - Move cursor left one character if nothing selected, otherwise 
	move cursor to beginning of selection. If the shift key is down, start 
	selecting or extending current selection. Don't allow cursor past 
	beginning of text"

	self insertAndCloseTypeIn.
	self
		moveCursor:[:position | position - 1 max: 1]
		forward: false
		event: aKeyboardEvent
		specialBlock:[:position | self previousWord: position].
	^ true! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'nice 8/3/2011 20:40'!
cursorPageDown: aKeyboardEvent

	self insertAndCloseTypeIn.
	self 
		moveCursor: [:position |
			self
				sameColumn: position
				newLine: [:lineNo | lineNo + self pageHeight]
				forward: true]
		forward: true
		event: aKeyboardEvent
		specialBlock:[:dummy | dummy].
	^true! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'nice 8/3/2011 20:40'!
cursorPageUp: aKeyboardEvent 

	self insertAndCloseTypeIn.
	self 
		moveCursor: [:position |
			self
				sameColumn: position
				newLine: [:lineNo | lineNo - self pageHeight]
				forward: false]
		forward: false
		event: aKeyboardEvent
		specialBlock:[:dummy | dummy].
	^true! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'nice 8/3/2011 20:40'!
cursorRight: aKeyboardEvent 
	"Private - Move cursor right one character if nothing selected, 
	otherwise move cursor to end of selection. If the shift key is down, 
	start selecting characters or extending already selected characters. 
	Don't allow cursor past end of text"

	self insertAndCloseTypeIn.
	self
		moveCursor: [:position | position + 1]
		forward: true
		event: aKeyboardEvent
		specialBlock:[:position | self nextWord: position].
	^ true! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:37'!
cursorTopHome: aKeyboardEvent
	"Put cursor at beginning of text -- invoked from cmd-H shortcut, useful for keyboards that have no home key."

	self selectAt: 1.
	^ true! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'nice 8/3/2011 20:40'!
cursorUp: aKeyboardEvent 
	"Private - Move cursor from position in current line to same position in
	prior line. If prior line too short, put at end"

	self insertAndCloseTypeIn.
	self
		moveCursor: [:position | self
				sameColumn: position
				newLine:[:line | line - 1]
				forward: false]
		forward: false
		event: aKeyboardEvent
		specialBlock:[:dummy | dummy].
	^true! !

!Editor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:38'!
cut: aKeyboardEvent 
	"Cut out the current text selection."

	self cut.
	^true! !

!Editor methodsFor: 'current selection' stamp: 'jmv 11/4/2008 13:20'!
deselect
	"If the text selection is visible on the screen, reverse its highlight."

	selectionShowing ifTrue: [self reverseSelection]! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'nice 8/3/2011 20:09'!
destructiveBackWord: aKeyboardEvent 
	"If the selection is not a caret, delete it and leave it in the backspace buffer.
	 Else if there is typeahead, delete it.
	 Else, delete the word before the caret."

	| startIndex |
	self hasCaret
		ifTrue: "a caret, delete at least one character"
			[startIndex := 1 max: self markIndex - 1.
			[startIndex > 1 and:
				[(self string at: startIndex - 1) tokenish]]
				whileTrue:
					[startIndex := startIndex - 1]]
		ifFalse: "a non-caret, just delete it"
			[startIndex := self markIndex].
	self backTo: startIndex.
	^false! !

!Editor methodsFor: 'private' stamp: 'jmv 12/21/2010 14:46'!
endOfLine: position
	"Redefined in subclasses using Paragraph support"
	^self endOfParagraph: position! !

!Editor methodsFor: 'private' stamp: 'jmv 9/17/2010 13:09'!
endOfParagraph: position
	| s |
	s := self string.
	^ s
		indexOf: Character cr
		startingAt: position
		ifAbsent: [ s size + 1 ].! !

!Editor methodsFor: 'private' stamp: 'jmv 9/17/2010 13:25'!
endOfText
	^self string size + 1! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/7/2010 17:37'!
enter: aKeyboardEvent
	"Enter / return key was pressed"
	"Process the various Enter / Return keystrokes"
	
	morph acceptOnCR ifTrue: [
		self closeTypeIn.
		^ true].

	aKeyboardEvent controlKeyPressed ifTrue: [
		^ self cr: aKeyboardEvent ].
	aKeyboardEvent shiftPressed ifTrue: [
		^ self lf: aKeyboardEvent ].
	aKeyboardEvent commandAltKeyPressed ifTrue: [
		^ self crlf: aKeyboardEvent ].
	^ self crWithIndent: aKeyboardEvent! !

!Editor methodsFor: 'private' stamp: 'mt 11/7/2015 16:02'!
firstWordBoundaryAfter: position
	"If the character at position is whitespace, answer the position of the first character after position which is not whitespace.
	If the character at position is not whitespace, answer the position of the first character after position which is whitespace."
	| string index atWhitespace |
	string := self string.
	index := position.
	(atWhitespace := (string at: index) isSeparator)
		ifTrue:
			[ "find next non-separator"
			[ (index <= string size) and: [ (string at: index) isSeparator ] ] whileTrue: [ index := index + 1 ] ]
		ifFalse:
			[ "find next separator"
			[ (index <= string size) and: [ (string at: index) isSeparator not ] ] whileTrue: [ index := index + 1 ] ].
	^ index! !

!Editor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 23:43'!
hasSelection
	^self hasCaret not! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:38'!
lf: aKeyboardEvent 
	"Append a line feed character to the stream of characters."

	self addString: Character lf asString.
	^false! !

!Editor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 23:42'!
lineSelectAndEmptyCheck: returnBlock
	"If the current selection is an insertion point, expand it to be the entire current line; if after that's done the selection is still empty, then evaluate the returnBlock, which will typically consist of '[^ self]' in the caller -- check senders of this method to understand this."

	self selectLine.  "if current selection is an insertion point, then first select the entire line in which occurs before proceeding"
	self hasSelection ifFalse: [morph flash.  ^ returnBlock value]! !

!Editor methodsFor: 'private' stamp: 'nice 12/25/2009 15:56'!
lines
	"Compute lines based on logical line breaks, not optical (which may change due to line wrapping of the editor).
	Subclasses using kinds of Paragraphs can instead use the service provided by it."
	| lines string lineIndex |
	string := self string.
	"Empty strings have no lines at all. Think of something."
	string isEmpty ifTrue:[^{#(1 0 0)}].
	lines := OrderedCollection new: (string size // 15).
	lineIndex := 0.
	string lineIndicesDo: [:start :endWithoutDelimiters :end |
		lines addLast: {start. (lineIndex := lineIndex + 1). end}].
	"Special workaround for last line empty."
	(string last == Character cr or: [string last == Character lf])
		ifTrue: [lines addLast: {string size + 1. lineIndex + 1. string size}].
	^lines! !

!Editor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 13:06'!
morph
	^ morph! !

!Editor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 13:06'!
morph: aMorph
	"Install a link back to the morph being edited (esp for text links)"
	morph := aMorph ! !

!Editor methodsFor: 'private' stamp: 'nice 8/3/2011 20:39'!
moveCursor: directionBlock forward: forward event: aKeyboardEvent specialBlock: specialBlock 
	"Private - Move cursor.
	directionBlock is a one argument Block that computes the new Position from a given one.
	specialBlock is a one argumentBlock that computes the new position from a given one under the alternate semantics.
	Note that directionBlock always is evaluated first."
	| indices newPosition shouldSelect |
	shouldSelect := aKeyboardEvent shiftPressed.
	indices := self setIndices: shouldSelect forward: forward.
	newPosition := directionBlock value: (indices at: #moving).
	(aKeyboardEvent commandKeyPressed or: [aKeyboardEvent controlKeyPressed])
		ifTrue: [newPosition := specialBlock value: newPosition].
	shouldSelect
		ifTrue: [self selectMark: (indices at: #fixed) point: newPosition - 1]
		ifFalse: [self selectAt: newPosition]! !

!Editor methodsFor: 'private' stamp: 'nice 8/3/2011 22:28'!
moveCursor: directionBlock forward: forward select: shouldSelect
	"Private - Move cursor.
	directionBlock is a one argument Block that computes the new Position from a given one.
	specialBlock is a one argumentBlock that computes the new position from a given one under the alternate semantics.
	Note that directionBlock always is evaluated first."
	| indices newPosition |
	indices := self setIndices: shouldSelect forward: forward.
	newPosition := directionBlock value: (indices at: #moving).
	shouldSelect
		ifTrue: [self selectMark: (indices at: #fixed) point: newPosition - 1]
		ifFalse: [self selectAt: newPosition]! !

!Editor methodsFor: 'private' stamp: 'jmv 11/4/2008 23:39'!
nextWord: position
	| string index |
	string := self string.
	index := position.
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric]]
		whileTrue: [index := index + 1].
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric not]]
		whileTrue: [index := index + 1].
	^ index! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'nice 8/3/2011 22:35'!
nonDestructiveBackWord: aKeyboardEvent 
	"Select the prior word."
	| indices newPosition |
	self closeTypeIn.
	indices := self 
		setIndices: true
		forward: false.
	newPosition := 1 max: (indices at: #moving) - 1.
	newPosition :=  self previousWord: newPosition.
	self selectMark: (indices at: #fixed) point: newPosition - 1.
	^ true! !

!Editor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:38'!
noop: aKeyboardEvent 
	"Unimplemented keyboard command; just ignore it."

	^ true! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:38'!
normalCharacter: aKeyboardEvent 
	"A nonspecial character is to be added to the stream of characters."

	self addString: aKeyboardEvent keyCharacter asString.
	^false! !

!Editor methodsFor: 'menu messages' stamp: 'jmv 11/5/2008 10:12'!
paste
	"Paste the text from the shared buffer over the current selection and 
	redisplay if necessary.  Undoer & Redoer: undoAndReselect."

	self replace: self selectionInterval with: self clipboardText and:
		[self selectAt: self pointIndex]! !

!Editor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:38'!
paste: aKeyboardEvent 
	"Replace the current text selection by the text in the shared buffer."

	self insertAndCloseTypeIn.
	self paste.
	^true! !

!Editor methodsFor: 'private' stamp: 'jmv 11/4/2008 23:02'!
previousWord: position
	| string index |
	string := self string.
	index := position.
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric not]]
		whileTrue: [index := index - 1].
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric]]
		whileTrue: [index := index - 1].
	^ index + 1! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 11/4/2008 23:48'!
selectAll

	self selectFrom: 1 to: self string size! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:38'!
selectAll: aKeyboardEvent 
	"select everything, invoked by cmd-a.  1/17/96 sw"

	self insertAndCloseTypeIn.
	self selectFrom: 1 to: self string size.
	^ true! !

!Editor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 23:10'!
selectAt: characterIndex 
	"Deselect, then place the caret before the character at characterIndex.
	 Be sure it is in view."

	self selectFrom: characterIndex to: characterIndex - 1! !

!Editor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 23:49'!
selectInterval: anInterval
	"Deselect, then select the specified characters inclusive.
	 Be sure the selection is in view."

	self selectFrom: anInterval first to: anInterval last! !

!Editor methodsFor: 'new selection' stamp: 'jmv 5/27/2011 15:25'!
selectInvisiblyFrom: start to: stop
	"Select the designated characters, inclusive.  Make no visual changes."

	self markIndex: start pointIndex: stop + 1! !

!Editor methodsFor: 'new selection' stamp: 'jmv 5/27/2011 15:25'!
selectInvisiblyMark: mark point: point
	"Select the designated characters, inclusive.  Make no visual changes."

	self markIndex: mark pointIndex: point + 1! !

!Editor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:24'!
selectionInterval
	"Answer the interval that is currently selected."

	^self startIndex to: self stopIndex - 1 ! !

!Editor methodsFor: 'new selection' stamp: 'jmv 5/30/2011 10:01'!
selectMark: mark point: point
	"Deselect, then select the specified characters inclusive.
	 Be sure the selection is in view."

	(mark =  self markIndex and: [point + 1 = self pointIndex]) ifFalse: [
		self selectInvisiblyMark: mark point: point ]! !

!Editor methodsFor: 'new selection' stamp: 'cmm 11/16/2015 19:19'!
selectWord
	"Select a word or expression, the result of pressing Command+[Space Bar] or by double-clicking."
	^self
		selectWordLeftDelimiters: '
"''|([{<'   "<--- punctuation symbols should precede the bracket symbols"
		rightDelimiters: '
"''|)]}>'! !

!Editor methodsFor: 'new selection' stamp: 'cmm 11/16/2015 15:49'!
selectWordLeftDelimiters: leftDelimiters rightDelimiters: rightDelimiters
	"Select delimited text or word--the result of double-clicking."

	| openDelimiter closeDelimiter direction match level
	string here hereChar start stop |
	string := self string.
	string size < 2 ifTrue: [^self].
	here := self pointIndex.
	"Select the whole text when clicking before first or after last character"
	(here > string size or: [here < 2]) ifTrue: [^self selectFrom: 1 to: string size].
	openDelimiter := string at: here - 1.
	closeDelimiter := string at: here.
	(match := leftDelimiters indexOf: openDelimiter) > (rightDelimiters indexOf: closeDelimiter)
		ifTrue: [
			"a more-distinct delimiter is on the left -- match to the right"
			start := here.
			direction := 1.
			here := here - 1.
			closeDelimiter := rightDelimiters at: match]
		ifFalse: [
			openDelimiter := string at: here.
			match := rightDelimiters indexOf: openDelimiter.
			match > 0
				ifTrue: [
					"delimiter is on right -- match to the left"
					stop := here - 1.
					direction := -1.
					closeDelimiter := leftDelimiters at: match]
				ifFalse: [
					"no delimiters -- select a token"
					direction := -1]].
	level := 1.
	[level > 0 and: [direction > 0
			ifTrue: [here < string size]
			ifFalse: [here > 1]]]
		whileTrue: [
			hereChar := string at: (here := here + direction).
			match = 0
				ifTrue: ["token scan goes left, then right"
					hereChar tokenish
						ifTrue: [here = 1
								ifTrue: [
									start := 1.
									"go right if hit string start"
									direction := 1]]
						ifFalse: [
							direction < 0
								ifTrue: [
									start := here + 1.
									"go right if hit non-token"
									direction := 1]
								ifFalse: [level := 0]]]
				ifFalse: ["bracket match just counts nesting level"
					hereChar = closeDelimiter
						ifTrue: [level := level - 1"leaving nest"]
						ifFalse: [hereChar = openDelimiter 
									ifTrue: [level := level + 1"entering deeper nest"]]]].

	level > 0 ifTrue: ["in case ran off string end"	here := here + direction].
	^direction > 0
		ifTrue: [self selectFrom: start to: here - 1]
		ifFalse: [self selectFrom: here + 1 to: stop]! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 12:38'!
selectWord: aKeyboardEvent

	self insertAndCloseTypeIn.
	self selectWord.
	^ true! !

!Editor methodsFor: 'private' stamp: 'ar 1/2/2010 14:45'!
setIndices: shiftPressed forward: forward
	"Little helper method that sets the moving and fixed indices according to some flags."
	| indices |
	indices := Dictionary new.
	(shiftPressed and:[self class selectionsMayShrink])
		ifTrue: [
			indices at: #moving put: self pointIndex.
			indices at: #fixed put: self markIndex
		] ifFalse: [
			forward
				ifTrue:[
					indices at: #moving put: self stopIndex.
					indices at: #fixed put: self startIndex.
				] ifFalse: [
					indices at: #moving put: self startIndex.
					indices at: #fixed put: self stopIndex.
				]
		].
	^indices! !

!Editor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 14:33'!
userHasEdited
	"Note that my text is free of user edits."

	morph hasUnacceptedEdits: true! !

!Editor methodsFor: 'menu messages' stamp: 'jmv 4/27/2011 07:50'!
wordSelectAndEmptyCheck: returnBlock
	"Ensure selecting the entire current word; if after that's done the selection is still empty, then evaluate the returnBlock, which will typically consist of '[^ self]' in the caller -- check senders of this method to understand this."

	self selectWord.  "Select exactly a whole word"
	self hasSelection ifFalse: [morph flash.  ^ returnBlock value]! !

Editor subclass: #TextEditor
	instanceVariableNames: 'model paragraph markBlock pointBlock beginTypeInIndex emphasisHere lastParenLocation otherInterval oldInterval typeAhead history'
	classVariableNames: 'AutoEnclose AutoIndent ChangeText EncloseSelection FindText'
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

TextEditor class
	instanceVariableNames: 'cmdActions shiftCmdActions yellowButtonMenu shiftedYellowButtonMenu'!

!TextEditor commentStamp: '<historical>' prior: 0!
See comment in Editor.

My instances edit Text, this is, they support multiple lines and TextAttributes.
They have no specific facilities for editing Smalltalk code. Those are found in SmalltalkEditor.!

TextEditor class
	instanceVariableNames: 'cmdActions shiftCmdActions yellowButtonMenu shiftedYellowButtonMenu'!

TextEditor subclass: #SmalltalkEditor
	instanceVariableNames: 'styler'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!SmalltalkEditor commentStamp: 'jmv 8/8/2009 15:10' prior: 0!
The editor built specifically for Smalltalk code!

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'ar 10/5/2009 22:56'!
initialize
	"SmalltalkEditor initialize"
	self initializeCmdKeyShortcuts.
	self initializeShiftCmdKeyShortcuts.
	self initializeYellowButtonMenu.
	self initializeShiftedYellowButtonMenu.
! !

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'mt 11/10/2015 09:53'!
initializeCmdKeyShortcuts
	"Initialize the (unshifted) command-key (or alt-key) shortcut table."
	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"
	"SmalltalkEditor initialize"
	| cmds |
	super initializeCmdKeyShortcuts.
	cmds := #($b #browseIt: $d #doIt: $i #inspectIt: $l #cancel: $m #implementorsOfIt: $n #sendersOfIt: $o #spawnIt: $p #printIt: $q #querySymbol: $s #save: ).
	1 to: cmds size
		by: 2
		do: [ : i | cmdActions at: (cmds at: i) asciiValue + 1 put: (cmds at: i + 1)].
	"Set up type-method argument hot keys, 1-4.."
	'1234' do:
		[ : eachKeyboardChar |
		cmdActions 
			at: eachKeyboardChar asciiValue + 1
			put: #typeMethodArgument: ]! !

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'topa 10/19/2015 19:52'!
initializeShiftCmdKeyShortcuts 
	"Initialize the shift-command-key (or control-key) shortcut table."
	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"
	"wod 11/3/1998: Fix setting of cmdMap for shifted keys to actually use the 
	capitalized versions of the letters.
	TPR 2/18/99: add the plain ascii values back in for those VMs that don't return the shifted values."

	"SmalltalkEditor initialize"

	| cmds |
	super initializeShiftCmdKeyShortcuts.
	
	cmds := #(
		$a	argAdvance:
		$b	browseItHere:
		$d	debugIt:
		$e	methodStringsContainingIt:
		$f	displayIfFalse:
		$g	fileItIn:
		$i	exploreIt:
		$n	referencesToIt:
		$s	invokePrettyPrint:
		$t	displayIfTrue:
		$v	pasteInitials:
		$w	methodNamesContainingIt:
	).
	1 to: cmds size by: 2 do: [ :i |
		shiftCmdActions at: ((cmds at: i) asciiValue + 1) put: (cmds at: i + 1).			"plain keys"
		shiftCmdActions at: ((cmds at: i) asciiValue - 32 + 1) put: (cmds at: i + 1).		"shifted keys"
		shiftCmdActions at: ((cmds at: i) asciiValue - 96 + 1) put: (cmds at: i + 1).		"ctrl keys"
	].! !

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'ar 10/5/2009 23:21'!
initializeShiftedYellowButtonMenu
	"Initialize the yellow button pop-up menu and corresponding messages."

	"SmalltalkEditor initialize"

	shiftedYellowButtonMenu := MenuMorph fromArray: StringHolder yellowButtonMenuItems.! !

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'ar 10/5/2009 23:21'!
initializeYellowButtonMenu
	"Initialize the yellow button pop-up menu and corresponding messages."

	"SmalltalkEditor initialize"

	yellowButtonMenu := MenuMorph fromArray: StringHolder yellowButtonMenuItems! !

!SmalltalkEditor methodsFor: 'parenblinking' stamp: 'Richo 10/6/2011 15:59'!
blinkPrevParen: aCharacter
	"Same as super, but tries to follow the Smalltalk syntax."

	| openDelimiter closeDelimiter level string here inside |
	string := paragraph text string.
	here := pointBlock stringIndex.
	openDelimiter := aCharacter.
	closeDelimiter := '([{' at: (')]}' indexOf: openDelimiter).
	level := 1.
	inside := nil. "Tricky."
	(here > 1 and: [ (string at: here - 1) = $$ ]) ifTrue: [ ^self ]. "Just a character literal."
	[ level > 0 and: [ here > 1 ] ] whileTrue: [
		| hereChar |
		hereChar := string at: (here := here - 1).
		inside "Are we inside a comment or string literal?"
			ifNotNil: [ "Yes."
				hereChar = inside ifTrue: [
					(here > 1 and: [ (string at: here - 1) ~= inside ])
						ifTrue: [ inside := nil ]
						ifFalse: [ here := here - 1 ] ] ]
			ifNil: [
				(here > 1 and: [ (string at: here - 1) = $$ ]) "Just a character literal."
					ifTrue: [ here := here - 1 ]
					ifFalse: [
						hereChar
							caseOf: {
								[ closeDelimiter ] -> [
									(level := level - 1) = 0 ifTrue: [
										^self blinkParenAt: here ] ].
								[ openDelimiter ] -> [  level := level + 1 ].
								[ $" ] -> [ inside := $" ].
								[ $' ] -> [ inside := $' ] }
							otherwise: [] ] ] ]! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'yo 6/29/2011 14:01'!
buttonForIt

	^ self doItButtonFromSelection ifNotNilDo: [:e | e openInHand]! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'nice 12/26/2009 21:40'!
changeEmphasis: characterStream
	"Change emphasis without styling if necessary"
	styler ifNil: [^super changeEmphasis: characterStream].
	^styler evaluateWithoutStyling: [super changeEmphasis: characterStream].! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'nice 12/16/2013 15:22'!
doItButtonFromSelection

	| button string cm |
	self lineSelectAndEmptyCheck: [^ nil].
	button := SimpleButtonMorph new.
	string := self selection.
	(model respondsTo: #doItReceiver) 
		ifTrue: [
			button setProperty: #rcvr toValue: model doItReceiver.
			button setProperty: #ctxt toValue: model doItContext].

	cm := (button valueOfProperty: #rcvr) class compilerClass new 
			compiledMethodFor: string readStream
			in: (button valueOfProperty: #ctxt)
			to: (button valueOfProperty: #rcvr)
			notifying: nil
			ifFail: [^ nil].
	button setProperty: #cm toValue: cm.
	button target: [:b |
		[(b valueOfProperty: #cm) valueWithReceiver: (b valueOfProperty: #rcvr) arguments: 
			((b valueOfProperty: #ctxt) ifNotNil: [{(b valueOfProperty: #ctxt) }] ifNil: [#()])]
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true]];
		actionSelector: #value:;
		arguments: {button}.
	button label: string.
	^ button.
! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'eem 12/5/2014 11:26'!
emphasisExtras
	^#(
		'Do it' 
		'Print it' 
		'Link to comment of class' 
		'Link to definition of class' 
		'Link to hierarchy of class' 
		'Link to method'
		'URL Link'
	).! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 21:37'!
handleEmphasisExtra: index with: aKeyboardEvent
	"Handle an extra emphasis menu item"
	| action attribute thisSel |
	action := {
		[attribute := TextDoIt new.
		thisSel := attribute analyze: self selection asString].
		[attribute := TextPrintIt new.
		thisSel := attribute analyze: self selection asString].
		[attribute := TextLink new.
		thisSel := attribute analyze: self selection asString with: 'Comment'].
		[attribute := TextLink new.
		thisSel := attribute analyze: self selection asString with: 'Definition'].
		[attribute := TextLink new.
		thisSel := attribute analyze: self selection asString with: 'Hierarchy'].
		[attribute := TextLink new.
		thisSel := attribute analyze: self selection asString].
		[attribute := TextURL new.
		thisSel := attribute analyze: self selection asString].
		["Edit hidden info"
		thisSel := self hiddenInfo.	"includes selection"
		attribute := TextEmphasis normal].
		["Copy hidden info"
		self copyHiddenInfo.
		^true].	"no other action"
	} at: index.
	action value.

	thisSel ifNil: [^true].	"Could not figure out what to link to"

	attribute ifNotNil: [
		thisSel ifEmpty:[ | oldAttributes |
			"only change emphasisHere while typing"
			oldAttributes := paragraph text attributesAt: self pointIndex.
			emphasisHere := Text addAttribute: attribute toArray: oldAttributes.
		] ifNotEmpty: [
			self replaceSelectionWith: (thisSel asText addAttribute: attribute).
		]
	].
	^true! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'cmm 3/1/2010 16:06'!
invokePrettyPrint: dummy
	self prettyPrint: false.
	^ true! !

!SmalltalkEditor methodsFor: 'private' stamp: 'cmm 4/4/2010 20:41'!
methodArgument: anInteger 
	^ (ReadStream on: self text asString) nextLine
		ifNil: [ String empty ]
		ifNotNilDo:
			[ : line | 
			line substrings
				at: 2 * anInteger
				ifAbsent: [ String empty ] ]! !

!SmalltalkEditor methodsFor: 'private' stamp: 'cmm 11/3/2015 23:32'!
nextWord: position 
	| string index boundaryCharacters |
	string := self string.
	index := position - 1.
	[ (index
		between: 1
		and: string size) and: [ (string at: index) isSeparator ] ] whileTrue: [ index := index + 1 ].
	boundaryCharacters := ')]}''"|^. 	'.
	((index
		between: 1
		and: string size) and: [ boundaryCharacters includes: (string at: index) ])
		ifTrue:
			[  index := index + 1  ]
		ifFalse:
			[ [ (index
				between: 1
				and: string size) and: [ (boundaryCharacters includes: (string at: index)) not ] ] whileTrue: [ index := index + 1 ] ].
	^ index! !

!SmalltalkEditor methodsFor: 'private' stamp: 'cmm 11/3/2015 23:31'!
previousWord: position 
	| string index boundaryCharacters |
	string := self string.
	index := position.
	"First, get out of whitespace."
	[ (index
		between: 2
		and: string size) and: [ (string at: index) isSeparator ] ] whileTrue: [ index := index - 1 ].
	boundaryCharacters := '([{''"|^. 	'.
	"Are we at a boundary character?"
	((index
		between: 2
		and: string size) and: [ boundaryCharacters includes: (string at: index) ])
		ifTrue:
			[ "yes, select it and any following whitespace of this line."
			index := index - 1 ]
		ifFalse:
			[ "no, select to the next boundary character"
			[ (index
				between: 1
				and: string size) and: [ (boundaryCharacters includes: (string at: index)) not ] ] whileTrue: [ index := index - 1 ] ].
	^ index + 1! !

!SmalltalkEditor methodsFor: 'compatibility' stamp: 'ar 8/8/2009 22:35'!
select
	"Sent by the parser when correcting variables etc. Ignored here."! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'tfel 4/7/2015 09:40'!
spyOnIt

	^ MessageTally spyOn: [self evaluateSelection]! !

!SmalltalkEditor methodsFor: 'accessing' stamp: 'ar 8/18/2009 00:11'!
styler
	"Answers the styler for this editor. Only code editors support syntax highlighting"
	^styler
! !

!SmalltalkEditor methodsFor: 'accessing' stamp: 'ar 8/18/2009 00:11'!
styler: aStyler
	"Sets the styler for this editor. Only code editors support syntax highlighting"
	^styler := aStyler! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'jmv 8/8/2009 15:36'!
tallyIt

	^ self tallySelection! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'nice 12/16/2013 15:22'!
tallySelection
	"Treat the current selection as an expression; evaluate it and return the time took for this evaluation"
	| result rcvr ctxt valueAsString v |
	self lineSelectAndEmptyCheck: [^ self].

	(model respondsTo: #doItReceiver) 
		ifTrue: [ rcvr := model doItReceiver.
				ctxt := model doItContext]
		ifFalse: [rcvr := ctxt := nil].
	result := [ | cm |
		cm := rcvr class evaluatorClass new 
			compiledMethodFor: self selectionAsStream
			in: ctxt
			to: rcvr
			notifying: self
			ifFail: [morph flash. ^ self].
		Time millisecondsToRun: 
			[v := cm valueWithReceiver: rcvr arguments: #() ].
	] 
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true].

	"We do not want to have large result displayed"
	valueAsString := v printString.
	(valueAsString size > 30) ifTrue: [valueAsString := (valueAsString copyFrom: 1 to: 30), '...'].
	PopUpMenu 
		inform: 'Time to compile and execute: ', result printString, 'ms res: ', valueAsString.
! !

!SmalltalkEditor methodsFor: 'private' stamp: 'nice 8/3/2011 20:47'!
typeMethodArgument: aKeyboardEvent 
	"Replace the current text selection with the name of the method argument represented by the keyCode."
	| keyCode |
	keyCode := ('1234' 
		indexOf: aKeyboardEvent keyCharacter
		ifAbsent: [1]).
	self  addString: (self methodArgument: keyCode).
	^ false! !

!TextEditor class methodsFor: 'class initialization' stamp: 'jmv 11/5/2008 22:21'!
abandonChangeText
	"Call this to get out of the maddening situation in which the system keeps aggressively trying to do a replacement that you no longer wish to make, every time you make choose a new method in a list."
	ChangeText := FindText

	"
	TextEditor abandonChangeText
	"! !

!TextEditor class methodsFor: 'preferences' stamp: 'mt 4/19/2016 09:08'!
autoEnclose
	<preference: 'Auto enclose brackets () {} []'
		categoryList: #('Morphic' 'editing')
		description: 'When true, typing an opening parenthesis, bracket or square-bracket will also add its corresponding closing character in front of the cursor.'
		type: #Boolean>
		
	^ AutoEnclose ifNil: [ false ]! !

!TextEditor class methodsFor: 'preferences' stamp: 'cmm 7/3/2010 16:17'!
autoEnclose: aBoolean
	AutoEnclose := aBoolean! !

!TextEditor class methodsFor: 'preferences' stamp: 'mt 4/19/2016 09:08'!
autoIndent
	<preference: 'Auto indent on new line'
		categoryList: #('Morphic' 'editing')
		description: 'When true, tabs will be inserted after pressing Enter | Return such that the new line will be indented equally with the previous line.'
		type: #Boolean>
	^ AutoIndent ifNil: [ true ]! !

!TextEditor class methodsFor: 'preferences' stamp: 'cmm 2/5/2010 11:16'!
autoIndent: aBoolean
	AutoIndent := aBoolean! !

!TextEditor class methodsFor: 'class initialization' stamp: 'mt 11/11/2015 12:54'!
cleanUp

	TextEditor allSubInstancesDo: [:editor |
		editor history ifNotNil: [:h | 
			h current ifNotNil: [editor closeTypeIn].
			h reset]].! !

!TextEditor class methodsFor: 'accessing' stamp: 'jmv 7/21/2009 10:07'!
cmdActions
	^cmdActions! !

!TextEditor class methodsFor: 'preferences' stamp: 'mt 4/19/2016 09:13'!
encloseSelection
	<preference: 'Enclose selection with brackets () {} [] '''' "" || <>'
		categoryList: #('Morphic' 'editing')
		description: 'When true, selecting text and typing an opening parenthesis, bracket, square-bracket, single quote, or double quote will add corresponding character around the selection.'
		type: #Boolean>
		
	^ EncloseSelection ifNil: [ false ]! !

!TextEditor class methodsFor: 'preferences' stamp: 'mt 4/19/2016 09:12'!
encloseSelection: boolean

	EncloseSelection := boolean.! !

!TextEditor class methodsFor: 'class initialization' stamp: 'mt 11/7/2015 18:57'!
initialize 
	"Initialize the keyboard shortcut maps and the shared buffers
	for copying text across views and managing again and undo." 
 
	"TextEditor initialize"

	FindText := ChangeText := Text new.

	self initializeCmdKeyShortcuts.
	self initializeShiftCmdKeyShortcuts.
	self initializeYellowButtonMenu.
	self initializeShiftedYellowButtonMenu! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'cmm 11/14/2015 19:20'!
initializeCmdKeyShortcuts
	"Initialize the (unshifted) command-key (or alt-key) shortcut table."

	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"

	"TextEditor initialize"

	| cmdMap cmds |
	cmdMap := Array new: 256 withAll: #noop:.		"use temp in case of a crash"
	cmdMap at: 1 + 1 put: #cursorHome:.				"home key"
	cmdMap at: 4 + 1 put: #cursorEnd:.				"end key"
	cmdMap at: 8 + 1 put: #backspace:.				"ctrl-H or delete key"
	cmdMap at: 11 + 1 put: #cursorPageUp:.			"page up key"
	cmdMap at: 12 + 1 put: #cursorPageDown:.		"page down key"
	cmdMap at: 13 + 1 put: #crWithIndent:.			"cmd-Return"
	cmdMap at: 27 + 1 put: #offerMenuFromEsc:.		"escape key"
	cmdMap at: 28 + 1 put: #cursorLeft:.				"left arrow key"
	cmdMap at: 29 + 1 put: #cursorRight:.				"right arrow key"
	cmdMap at: 30 + 1 put: #cursorUp:.				"up arrow key"
	cmdMap at: 31 + 1 put: #cursorDown:.				"down arrow key"
	cmdMap at: 32 + 1 put: #selectWord:.				"space bar key"
	cmdMap at: 127 + 1 put: #forwardDelete:.		"del key"
			
	'0123456789-=' 
		do: [:char | cmdMap at: char asciiValue + 1 put: #changeEmphasis:].
		
	'([<{|"''' do: [:char | cmdMap at: char asciiValue + 1 put: #enclose:].
	
	cmds := #($a #selectAll: $c #copySelection: $e #exchange: $f #find: $g #findAgain: $j #doAgain: $k #offerFontMenu: $u #align: $v #paste: $w #backWord: $x #cut: $y #swapChars: $z #undo:).
	1 to: cmds size
		by: 2
		do: [:i | cmdMap at: (cmds at: i) asciiValue + 1 put: (cmds at: i + 1)].
		
	cmdActions := cmdMap! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'mt 11/13/2015 10:00'!
initializeShiftCmdKeyShortcuts 
	"Initialize the shift-command-key (or control-key) shortcut table."
	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"
	"wod 11/3/1998: Fix setting of cmdMap for shifted keys to actually use the 
	capitalized versions of the letters.
	TPR 2/18/99: add the plain ascii values back in for those VMs that don't return the shifted values."

	"TextEditor initialize"
	
	| cmdMap cmds |

	"shift-command and control shortcuts"
	cmdMap := Array new: 256 withAll: #noop:.  		"use temp in case of a crash"
	cmdMap at: ( 1 + 1) put: #cursorHome:.			"home key"
	cmdMap at: ( 4 + 1) put: #cursorEnd:.				"end key"
	cmdMap at: ( 8 + 1) put: #forwardDelete:.			"ctrl-H or delete key"
	cmdMap at: (11 + 1) put: #cursorPageUp:.			"page up key"
	cmdMap at: (12 + 1) put: #cursorPageDown:.		"page down key"
	cmdMap at: (13 + 1) put: #crWithIndent:.			"ctrl-Return"
	cmdMap at: (27 + 1) put: #offerMenuFromEsc:.	"escape key"
	cmdMap at: (28 + 1) put: #cursorLeft:.			"left arrow key"
	cmdMap at: (29 + 1) put: #cursorRight:.			"right arrow key"
	cmdMap at: (30 + 1) put: #cursorUp:.				"up arrow key"
	cmdMap at: (31 + 1) put: #cursorDown:.			"down arrow key"
	cmdMap at: (32 + 1) put: #selectWord:.			"space bar key"
	cmdMap at: (45 + 1) put: #changeEmphasis:.		"cmd-sh-minus"
	cmdMap at: (61 + 1) put: #changeEmphasis:.		"cmd-sh-plus"
	cmdMap at: (127 + 1) put: #forwardDelete:.		"del key"

	"On some keyboards, these characters require a shift"
	'([<{|"''9' do: [:char | cmdMap at: char asciiValue + 1 put: #enclose:].

	"NB: sw 12/9/2001 commented out the idiosyncratic line just below, which was grabbing shift-esc in the text editor and hence which argued with the wish to have shift-esc be a universal gesture for escaping the local context and calling up the desktop menu."  
	"cmdMap at: (27 + 1) put: #shiftEnclose:." 	"ctrl-["

	"'""''(' do: [ :char | cmdMap at: (char asciiValue + 1) put: #enclose:]."

	cmds := #(
		$c	compareToClipboard:
		$h	cursorTopHome:
		$j	doAgainUpToEnd:
		$k	changeStyle:
		$m	selectCurrentTypeIn:
		$s	findAgain:
		$u	changeLfToCr:
		$x	makeLowercase:
		$y	makeUppercase:
		$z	redo: "makeCapitalized:"
	).
	1 to: cmds size by: 2 do: [ :i |
		cmdMap at: ((cmds at: i) asciiValue + 1) put: (cmds at: i + 1).			"plain keys"
		cmdMap at: ((cmds at: i) asciiValue - 32 + 1) put: (cmds at: i + 1).		"shifted keys"
		cmdMap at: ((cmds at: i) asciiValue - 96 + 1) put: (cmds at: i + 1).		"ctrl keys"
	].
	shiftCmdActions := cmdMap! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'ar 9/7/2009 11:41'!
initializeShiftedYellowButtonMenu
	"Initialize the yellow button pop-up menu and corresponding messages."

	"TextEditor initialize"
	"
	shiftedYellowButtonMenu := {
		{'set font... (k)' translated.					#offerFontMenu}.
		{'set style... (K)' translated.				#changeStyle}.
		{'set alignment...' translated.				#chooseAlignment}.
		#-.
		{'more...' translated.						#yellowButtonActivity}.
	}
	"
	shiftedYellowButtonMenu := yellowButtonMenu! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'mt 11/13/2015 09:33'!
initializeYellowButtonMenu
	"Initialize the yellow button pop-up menu and corresponding messages."

	"TextEditor initialize"

	yellowButtonMenu := MenuMorph fromArray: {
		{'find...(f)' translated.				#find}.
		{'find again (g)' translated.				#findAgain}.
		{'find and replace ...' translated.			#findReplace}.
		{'do/replace again (j)' translated.			#again}.
		#-.
		{'undo (z)' translated.				#undo}.
		{'redo (Z)' translated.				#redo}.
		#-.
		{'copy (c)' translated.				#copySelection}.
		{'cut (x)' translated.				#cut}.
		{'paste (v)' translated.				#paste}.
		{'paste...' translated.				#pasteRecent}.
		#-.
		{'set font... (k)' translated.			#offerFontMenu}.
		{'set style... (K)' translated.		#changeStyle}.
		{'set alignment...' translated.		#chooseAlignment}.
		"
		#-.
		{'more...' translated.				#shiftedTextPaneMenuRequest}.
		"
	}! !

!TextEditor class methodsFor: 'class initialization' stamp: 'mt 11/11/2015 11:44'!
resetAllHistory

	TextEditor allSubInstances do: [:editor |
		editor history reset].

	! !

!TextEditor class methodsFor: 'accessing' stamp: 'jmv 7/21/2009 10:17'!
shiftCmdActions
	^shiftCmdActions! !

!TextEditor class methodsFor: 'class initialization' stamp: 'jmv 7/21/2009 10:37'!
shiftedYellowButtonMenu
	"Answer the menu to be presented when the yellow button is pressed while the shift key is down"

	^ shiftedYellowButtonMenu! !

!TextEditor class methodsFor: 'class initialization' stamp: 'jmv 7/21/2009 10:34'!
yellowButtonMenu

	^ yellowButtonMenu! !

!TextEditor methodsFor: 'editing keys' stamp: 'ar 10/5/2009 21:39'!
abandonChangeText
	^self class abandonChangeText! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
accept
	"Save the current text of the text being edited as the current acceptable version for purposes of canceling.  Allow my morph to take appropriate action"
	morph acceptContents! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 11/7/2015 23:46'!
addString: aString

	morph readOnly ifTrue: [^ self].

	"If we modifying the text like backward or forward delete, we have to finish that operation."
	(self isTypingIn and: [self history current type notNil])
		ifTrue: [self closeTypeIn].

	self typeAhead nextPutAll: aString.! !

!TextEditor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 11:51'!
afterSelectionInsertAndSelect: aString

	self insertAndSelect: aString at: self stopIndex ! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 12/21/2015 14:16'!
again
	"Do the same find/replace command again. Looks up the editor's own command history and uses the previous command to determine find string and replace string."
	
	self history hasReplacedSomething
		ifFalse: [morph flash. ^ false].
	
	self
		setSearchFromSelectionOrHistory;
		setReplacementFromHistory.
	
	"If we have no selection, give the user one to avoid annoying surprises."
	^ self hasSelection
		ifTrue: [self findReplaceAgainNow]
		ifFalse: [self findAgainNow. false "see #againUpToEnd"]! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 11/18/2015 22:37'!
againUpToEnd
	"Find and replace until the end."
	
	| first |
	self again ifFalse: [^ self].

	first := self history previous.

	[self hasSelection] whileTrue: [
		self history previous
			isCompositeUndo: true;
			isCompositeRedo: true.
		self findReplaceAgainNow].

	first isCompositeUndo: false.
	
	self history previous isCompositeUndo: first ~~ self history previous.
	self history previous isCompositeRedo: false.! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 5/30/2011 10:03'!
align
	"Align text according to the next greater alignment value,
	cycling among leftFlush, rightFlush, center, and justified."
	self changeAlignment.
	self recomputeSelection! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:42'!
align: aKeyboardEvent
	"Triggered by Cmd-u;  cycle through alignment alternatives.  8/11/96 sw"

	self align.
	^ true! !

!TextEditor methodsFor: 'private' stamp: 'nice 8/3/2011 22:52'!
applyAttribute: aTextAttribute
	"The user selected aTextAttribute via shortcut, menu or other means.
	If there is a selection, apply the attribute to the selection.
	In any case use the attribute for the user input (emphasisHere)"
	| interval |

	emphasisHere := Text addAttribute: aTextAttribute toArray: emphasisHere.
	
	interval := self selectionInterval.
	(interval isEmpty and: [ aTextAttribute isParagraphAttribute not ])
		ifTrue: [ ^self ].
	
	self text addAttribute: aTextAttribute from: interval first to: interval last.
	paragraph recomposeFrom: interval first to: interval last delta: 0.
	self recomputeSelection.	"Needed so visible selection is updated to reflect new visual extent of selection"
	morph changed! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'cmm 3/12/2015 11:07'!
argAdvance: aKeyboardEvent 
	"Invoked by Ctrl-a or Shift+Command+A.  Useful after Ctrl-q.
	 Search forward from the end of the selection for a colon and place the caret after it.  If no colon is found, do nothing."
	| start |
	self insertAndCloseTypeIn.
	start := paragraph text
		findString: ':'
		startingAt: self stopIndex.
	start isZero ifFalse: [ self selectAt: start + 1 ].
	^ true! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 11/22/2015 14:44'!
autoEncloseFor: typedChar 
	"Answer whether typeChar was handled by auto-enclosure.  Caller should call normalCharacter if not."
	| openers closers |
	openers := '([{'.
	closers := ')]}'.
	(closers includes: typedChar) ifTrue:
		[ | pos |
		self blinkPrevParen: typedChar.
		((pos := self indexOfNextNonwhitespaceCharacter) notNil and: [ (paragraph string at: pos) = typedChar ])
			ifTrue:
				[ self
					moveCursor: [ : position | position + pos - pointBlock stringIndex + 1 ]
					forward: true
					select: false.
				^ true ]
			ifFalse: [ ^ false ] ].
	(openers includes: typedChar) ifTrue:
		[ self
			openTypeIn;
			addString: typedChar asString;
			addString: (closers at: (openers indexOf: typedChar)) asString;
			insertAndCloseTypeIn;
			
			moveCursor: [ : position | position - 1 ]
			forward: false
			select: false.
		^ true ].
	^ false! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 11/8/2015 16:28'!
backTo: startIndex
	"During typing, backspace to startIndex. If there already is a selection, just delete that selection. Otherwise, check if we did something else than backward-deletion and start a new command if so."

	morph readOnly ifTrue: [^ self].

	self hasSelection ifTrue: [
		"Add checkpoint in undo history."
		self replaceSelectionWith: self nullText.
		^ true].

	startIndex > self text size ifTrue: [^ false].

	self selectInvisiblyFrom: startIndex to: self stopIndex-1.
	
	self isTypingIn ifTrue: [
		self history current type = #backward
			ifFalse: [self closeTypeIn]
			ifTrue: [
				"Accumulate all deleted characters in current undo command."
				self history current contentsBefore replaceFrom: 1 to: 0 with: self selection.
				self history current intervalBefore in: [:i |
					self history current intervalBefore: (startIndex to: i last)]]].
		
	self openTypeInFor: #backward.
	self zapSelectionWith: self nullText.
	
	^ false! !

!TextEditor methodsFor: 'private' stamp: 'eem 11/18/2015 11:27'!
beginningOfLine: position
	"Redefined in subclasses using Paragraph support"
	^ (paragraph lines at:(paragraph lineIndexOfCharacterIndex: position)) first! !

!TextEditor methodsFor: 'binding' stamp: 'jmv 11/4/2008 11:51'!
bindingOf: aString
	^model bindingOf: aString! !

!TextEditor methodsFor: 'parenblinking' stamp: 'jmv 11/12/2010 14:56'!
blinkParen
	"Used if Shout"
	lastParenLocation ifNotNil: [
		self text string size >= lastParenLocation ifTrue: [
			self text
				addAttribute: TextEmphasis bold
				from: lastParenLocation
				to: lastParenLocation]]! !

!TextEditor methodsFor: 'parenblinking' stamp: 'jmv 11/4/2008 11:51'!
blinkParenAt: parenLocation 
	self text
		addAttribute: TextEmphasis bold
		from: parenLocation
		to: parenLocation.
	lastParenLocation := parenLocation.! !

!TextEditor methodsFor: 'parenblinking' stamp: 'nice 10/7/2011 23:36'!
blinkPrevParen: aCharacter
	"Used if not Shout"
	| openDelimiter closeDelimiter level string here hereChar |
	string := paragraph string.
	here := pointBlock stringIndex.
	openDelimiter := aCharacter.
	closeDelimiter := '([{' at: (')]}' indexOf: openDelimiter).
	level := 1.
	[level > 0 and: [here > 1]]
		whileTrue:
			[hereChar := string at: (here := here - 1).
			hereChar = closeDelimiter
				ifTrue:
					[level := level - 1.
					level = 0
						ifTrue: [^ self blinkParenAt: here]]
				ifFalse:
					[hereChar = openDelimiter
						ifTrue: [level := level + 1]]]! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
browseChangeSetsWithSelector
	"Determine which, if any, change sets have at least one change for the selected selector, independent of class"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ morph flash].
	ChangeSorter browseChangeSetsWithSelector: aSelector! !

!TextEditor methodsFor: 'menu messages' stamp: 'fbs 12/7/2013 18:32'!
browseClassFromIt
	"Launch a hierarchy browser for the class indicated by the current selection.  If multiple classes matching the selection exist, let the user choose among them."
	| aClass |
	self lineSelectAndEmptyCheck: [ ^ self ].
	aClass := UIManager default
		classFromPattern: self selection string withBlanksTrimmed
		withCaption: 'choose a class to browse...'.
	aClass ifNil: [ ^ morph flash ].
	SystemNavigation default
		spawnHierarchyForClass: aClass
		selector: nil! !

!TextEditor methodsFor: 'menu messages' stamp: 'fbs 12/6/2013 17:15'!
browseIt
	"Launch a browser for the current selection, if appropriate"

	| aSymbol anEntry brow |

	Preferences alternativeBrowseIt ifTrue: [^ self browseClassFromIt].

	self lineSelectAndEmptyCheck: [^ self].
	(aSymbol := self selectedSymbol) isNil ifTrue: [^ morph flash].

	aSymbol first isUppercase
		ifTrue:
			[anEntry := (Smalltalk
				at: aSymbol
				ifAbsent:
					[ self systemNavigation browseAllImplementorsOf: aSymbol.
					^ nil]).
			anEntry ifNil: [^ morph flash].
			(anEntry isKindOf: Class)
				ifFalse:	[anEntry := anEntry class].
			brow := SystemBrowser default new.
			brow setClass: anEntry selector: nil.
			brow class
				openBrowserView: (brow openEditString: nil)
				label: 'System Browser']
		ifFalse:
			[self systemNavigation browseAllImplementorsOf: aSymbol]! !

!TextEditor methodsFor: 'menu messages' stamp: 'fbs 3/16/2011 17:27'!
browseItHere
	"Retarget the receiver's window to look at the selected class, if appropriate.  3/1/96 sw"
	| aSymbol foundClass b |
	(((b := model) isKindOf: Browser) and: [b couldBrowseAnyClass])
		ifFalse: [^ morph flash].
	model okToChange ifFalse: [^ morph flash].
	self selectionInterval isEmpty ifTrue: [self selectWord].
	(aSymbol := self selectedSymbol) isNil ifTrue: [^ morph flash].

	foundClass := (Smalltalk at: aSymbol ifAbsent: [nil]).
		foundClass isNil ifTrue: [^ morph flash].
		(foundClass isKindOf: Class)
			ifTrue:
				[model selectSystemCategory: foundClass category.
	model classListIndex: (model classList indexOf: foundClass name)]! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 19:26'!
browseItHere: aKeyboardEvent 
	"Triggered by Cmd-shift-B; browse the thing represented by the current selection, if plausible, in the receiver's own window.  3/1/96 sw"

	self browseItHere.
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:40'!
browseIt: aKeyboardEvent
	"Triggered by Cmd-B; browse the thing represented by the current selection, if plausible.  1/18/96 sw"

	self browseIt.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
cancel
	"Cancel the changes made so far to this text"
	morph cancelEdits! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:42'!
cancel: aKeyboardEvent
	"Cancel unsubmitted changes."

	self cancel.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'ar 8/6/2009 19:11'!
changeAlignment
	| aList reply  |
	aList := #(leftFlush centered justified rightFlush).
	reply := UIManager default chooseFrom: aList values: aList.
	reply ifNil:[^self].
	self setAlignment: reply.
	paragraph composeAll.
	self recomputeSelection.
	^ true! !

!TextEditor methodsFor: 'attributes' stamp: 'nice 8/3/2011 22:41'!
changeEmphasisOrAlignment
	| aList reply  code align menuList startIndex |
	startIndex := self startIndex.
	aList := #(normal bold italic narrow underlined struckOut leftFlush centered rightFlush justified).	
	align := paragraph text alignmentAt: startIndex 
		ifAbsent: [ paragraph textStyle alignment ].
	code := paragraph text emphasisAt: startIndex.
	menuList := WriteStream on: Array new.
	menuList nextPut: (code isZero ifTrue:['<on>'] ifFalse:['<off>']), 'normal' translated.
	menuList nextPutAll: (#(bold italic underlined struckOut) collect:[:emph|
		(code anyMask: (TextEmphasis perform: emph) emphasisCode)
			ifTrue: [ '<on>', emph asString translated ]
			ifFalse: [ '<off>',emph asString translated ]]).
	((paragraph text attributesAt: startIndex)
		anySatisfy: [ :attr | attr isKern and: [attr kern < 0 ]]) 
			ifTrue: [ menuList nextPut:'<on>', 'narrow' translated ]
			ifFalse: [ menuList nextPut:'<off>', 'narrow' translated ].
	menuList nextPutAll: (#(leftFlush centered rightFlush justified) collectWithIndex: [ :type :i |
		align = (i-1)
			ifTrue: [ '<on>',type asString translated ]
			ifFalse: [ '<off>',type asString translated ]]).
	aList := #(normal bold italic underlined struckOut narrow leftFlush centered rightFlush justified).
	reply := UIManager default chooseFrom: menuList contents values: aList lines: #(1 6).
	reply notNil ifTrue: [
		(#(leftFlush centered rightFlush justified) includes: reply)
			ifTrue: [
				self setAlignment: reply.
				paragraph composeAll.
				self recomputeSelection]
			ifFalse: [
				self setEmphasis: reply.
				paragraph composeAll.
				self recomputeSelection]].
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 20:46'!
changeEmphasis: aKeyboardEvent 
	"Change the emphasis of the current selection or prepare to accept characters with the change in emphasis. Emphasis change amounts to a font change.  Keeps typeahead."

	"control 0..9 -> 0..9"

	| keyCode attribute oldAttributes index thisSel colors extras |
	keyCode := ('0123456789-=' indexOf: aKeyboardEvent keyCharacter ifAbsent: [1]) - 1.
	oldAttributes := paragraph text attributesAt: self pointIndex.
	thisSel := self selection.

	"Decipher keyCodes for Command 0-9..."
	(keyCode between: 1 and: 5) 
		ifTrue: [attribute := TextFontChange fontNumber: keyCode].

	keyCode = 6 
		ifTrue: [
			colors := #(#black #magenta #red #yellow #green #blue #cyan #white).
			extras := self emphasisExtras.
			index := UIManager default chooseFrom:colors , #('choose color...' ), extras
						lines: (Array with: colors size + 1).
			index = 0 ifTrue: [^true].
			index <= colors size 
				ifTrue: [attribute := TextColor color: (Color perform: (colors at: index))]
				ifFalse: [
					index := index - colors size - 1.	"Re-number!!!!!!"
					index = 0 
						ifTrue: [attribute := self chooseColor]
						ifFalse:[^self handleEmphasisExtra: index with: aKeyboardEvent]	"handle an extra"]].
	(keyCode between: 7 and: 11) 
		ifTrue: [
			aKeyboardEvent shiftPressed 
				ifTrue: [
					keyCode = 10 ifTrue: [attribute := TextKern kern: -1].
					keyCode = 11 ifTrue: [attribute := TextKern kern: 1]]
				ifFalse: [
					attribute := TextEmphasis 
								perform: (#(#bold #italic #narrow #underlined #struckOut) at: keyCode - 6).
					oldAttributes 
						do: [:att | (att dominates: attribute) ifTrue: [attribute turnOff]]]].
	keyCode = 0 ifTrue: [attribute := TextEmphasis normal].
	attribute ifNotNil: [
		thisSel size = 0
			ifTrue: [
				"only change emphasisHere while typing"
				self insertTypeAhead.
				emphasisHere := Text addAttribute: attribute toArray: oldAttributes ]
			ifFalse: [
				self replaceSelectionWith: (thisSel asText addAttribute: attribute) ]].
	^true! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 20:47'!
changeLfToCr: aKeyboardEvent 
	"Replace all LFs by CRs.
	Triggered by Cmd-U -- useful when getting code from FTP sites
	jmv- Modified to als change crlf by cr"
	
	| fixed |
	fixed := self selection string withSqueakLineEndings. 
	self replaceSelectionWith: (Text fromString: fixed).
	^ true! !

!TextEditor methodsFor: 'initialize-release' stamp: 'mt 11/7/2015 15:18'!
changeParagraph: aParagraph 
	"Install aParagraph as the one to be edited by the receiver."

	paragraph := aParagraph.
	self resetState! !

!TextEditor methodsFor: 'attributes' stamp: 'mt 11/10/2015 10:50'!
changeSelectionFontTo: aFont 
	| attr |
	aFont ifNil: [ ^ self ].
	attr := TextFontReference toFont: aFont.
	
	self openTypeIn.
	
	paragraph text
		addAttribute: attr
		from: self startIndex
		to:
			(self hasSelection
				ifTrue: [ self stopIndex - 1 min: paragraph text size ]
				ifFalse: [ paragraph text size ]).
	
	self closeTypeIn.
	
	paragraph composeAll.
	self recomputeSelection.
	morph changed! !

!TextEditor methodsFor: 'attributes' stamp: 'kb 2/23/2010 12:46'!
changeStyle
	"Let user change styles for the current text pane."
	| names reply style current menuList |

	current := paragraph textStyle.
	names := TextStyle knownTextStyles.
	menuList := names collect: [ :styleName |
		styleName = current name
			ifTrue: [ '<on>', styleName ]
			ifFalse: [ '<off>', styleName ]].
	reply := UIManager default chooseFrom: menuList values: names.
	reply ifNotNil: [
		(style := TextStyle named: reply) ifNil: [Beeper beep. ^ true].
		paragraph textStyle: style.
		paragraph composeAll.
		self recomputeSelection].
	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'nice 8/3/2011 19:25'!
changeStyle: aKeyboardEvent 
	"Put up the style-change menu"

	self insertAndCloseTypeIn.
	self changeStyle.
	^ true! !

!TextEditor methodsFor: 'attributes' stamp: 'cmm 12/1/2010 10:42'!
changeTextFont
	"Present a dialog which allows the user to select a font, and if one is chosen, apply it to the current selection.	If there is no selection, or the selection is empty, apply it to the whole morph."
	| curFont startIndex |
	startIndex := self startIndex.
	curFont := (paragraph text fontAt: startIndex withStyle: paragraph textStyle).
	morph openModal: (
		Cursor wait showWhile: [ 
			(FontChooserTool default
				withTitle: 'Change the selected text''s font to...' translated
				for: self 
				setSelector: #changeSelectionFontTo:
				getSelector: curFont)
			"Do not allow changing the emphasis; we don't know how to deal with
			a 'pre-emphasized' font here, so bail."
					offerStyleList: false;
					open])! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
chooseAlignment
	self changeAlignment! !

!TextEditor methodsFor: 'editing keys' stamp: 'cmm 12/4/2010 15:11'!
chooseColor
	"Make a new Text Color Attribute, let the user pick a color, and return the attribute"
	| attribute |
	attribute := TextColor color: Color black.
	NewColorPickerMorph useIt
		ifTrue:
			[ (NewColorPickerMorph on: attribute) openNear: morph fullBoundsInWorld ]
		ifFalse:
			[ ColorPickerMorph new
				 choseModalityFromPreference ;
				 sourceHand: morph activeHand ;
				 target: attribute ;
				 selector: #color: ;
				 originalColor: Color black ;
				
				putUpFor: morph
				near: morph fullBoundsInWorld ].
	^ attribute! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 8/8/2009 15:22'!
classCommentsContainingIt
	"Open a browser class comments which contain the current selection somewhere in them."

	self lineSelectAndEmptyCheck: [^ self].
	self systemNavigation browseClassCommentsWithString: self selection string! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 8/8/2009 15:22'!
classNamesContainingIt
	"Open a browser on classes whose names contain the selected string"

	self lineSelectAndEmptyCheck: [^self].
	self systemNavigation
		browseClassesWithNamesContaining: self selection string
		caseSensitive: Sensor leftShiftDown! !

!TextEditor methodsFor: 'parenblinking' stamp: 'jmv 8/16/2009 18:31'!
clearParens
	lastParenLocation ifNotNil: [
		self text string size >= lastParenLocation ifTrue: [
			self text
				removeAttribute: TextEmphasis bold
				from: lastParenLocation
				to: lastParenLocation]].
	lastParenLocation := nil! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 11/8/2015 16:52'!
closeTypeIn
	"See comment in openTypeIn.  It is important to call closeTypeIn before executing
	 any non-typing key, making a new selection, etc.  It is called automatically for
	 menu commands."

	| begin stop |
	beginTypeInIndex ifNotNil: [
		begin := beginTypeInIndex.
		stop := self stopIndex.
				
		self history current
			contentsAfter: (stop <= begin
				ifTrue: [self nullText]
				ifFalse: [paragraph text copyFrom: begin to: stop-1]);
			intervalAfter: (stop to: stop-1);
			intervalBetween: (stop < begin
				ifTrue: [stop to: stop-1]
				ifFalse: [begin to: stop-1]);
			messageToUndo: (Message selector: #undoAndReselect);
			messageToRedo: (Message selector: #redoAndReselect).
			
		self history finishRemember.

		beginTypeInIndex := nil]! !

!TextEditor methodsFor: 'menu messages' stamp: 'cmm 2/1/2015 13:05'!
compareToClipboard
	"If any text is selected, present the modifications that would be made to it if the clipboard contents were pasted over it.  If no text is selected, present the differences betwen the entire pane's contents and the clipboard text."
	| subjectText proposedText |
	subjectText := self selection string ifEmpty: [ paragraph text string ].
	proposedText := self clipboardText string.
	subjectText = proposedText ifTrue: [^ self inform: 'Exact match'].
	(StringHolder new 
		textContents:
			(TextDiffBuilder
				buildDisplayPatchFrom: subjectText 
				to: proposedText)) openLabel: 'Differences with Clipboard Text'! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:42'!
compareToClipboard: aKeyboardEvent
	"Compare the receiver to the text on the clipboard."

	self compareToClipboard.
	^ true! !

!TextEditor methodsFor: 'do-its' stamp: 'Igor.Stasenko 12/20/2009 07:27'!
compileSelectionFor: anObject in: evalContext

	| methodNode |
	methodNode := [Compiler new
		compileNoPattern: self selectionAsStream
		in: anObject class
		context: evalContext
		notifying: self
		ifFail: [^nil]]
			on: OutOfScopeNotification
			do: [:ex | ex resume: true].
	^ methodNode generateWithTempNames! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 8/11/2010 10:33'!
copyHiddenInfo
	"In TextLinks, TextDoits, TextColor, and TextURLs, there is hidden
info.  Copy that to the clipboard.  You can paste it and see what it is.
Usually enclosed in <>."

	^ self clipboardTextPut: self hiddenInfo! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 11/7/2015 18:24'!
copySelection
	"Copy the current selection and store it in the paste buffer, unless a caret."

	self lineSelectAndEmptyCheck: [^ self].
	self clipboardTextPut: self selection.! !

!TextEditor methodsFor: 'new selection' stamp: 'nice 8/3/2011 23:00'!
correctFrom: start to: stop with: aString
	"Make a correction in the model that the user has authorised from somewhere else in the system (such as from the compilier).  The user's selection is not changed, only corrected."
	| userSelection delta loc wasShowing |
	aString = '#insert period' ifTrue: [
		loc := start.
		[(loc := loc-1)>0 and: [(paragraph string at: loc) isSeparator]]
			whileTrue: [loc := loc-1].
		^ self correctFrom: loc+1 to: loc with: '.'].
	(wasShowing := selectionShowing) ifTrue: [ self reverseSelection ].
	userSelection := self selectionInterval.

	self selectInvisiblyFrom: start to: stop.
	self replaceSelectionWith: aString.

	delta := aString size - (stop - start + 1).
	self
		selectInvisiblyFrom: userSelection first + (userSelection first > start ifFalse: [ 0 ] ifTrue: [ delta ])
		to: userSelection last + (userSelection last > start ifFalse: [ 0 ] ifTrue: [ delta ]).
	wasShowing ifTrue: [ self reverseSelection ].
! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'nice 8/3/2011 19:28'!
crWithIndent: aKeyboardEvent 
	"Replace the current text selection with CR followed by as many tabs
	as on the current line (+/- bracket count) -- initiated by Shift-Return."
	self addString: (String streamContents: [:characterStream | characterStream crtab: self tabCount]).  "Now inject CR with tabCount tabs"
	^ false! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'nice 8/3/2011 20:41'!
cursorEnd: aKeyboardEvent 

	"Private - Move cursor end of current line."
	| string |
	self insertAndCloseTypeIn.
	string := paragraph text string.
	self
		moveCursor:
			[:position | Preferences wordStyleCursorMovement
				ifTrue:[| targetLine |
					targetLine := paragraph lines at:(paragraph lineIndexOfCharacterIndex: position).
					targetLine = paragraph lastLine
						ifTrue:[targetLine last + 1]
						ifFalse:[targetLine last]]
				ifFalse:[
					string
						indexOfAnyOf: CharacterSet crlf
						startingAt: position
						ifAbsent:[string size + 1]]]
		forward: true
		event: aKeyboardEvent
		specialBlock:[:dummy | string size + 1].
	^true! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'nice 8/3/2011 20:41'!
cursorHome: aKeyboardEvent 

	"Private - Move cursor from position in current line to beginning of
	current line. If control key is pressed put cursor at beginning of text"

	| string |

	string := paragraph text string.
	self
		moveCursor: [ :position | Preferences wordStyleCursorMovement
				ifTrue:[
					(paragraph lines at:(paragraph lineIndexOfCharacterIndex: position)) first]
				ifFalse:[
					(string
						lastIndexOfAnyOf: CharacterSet crlf
						startingAt: position - 1
						ifAbsent:[0]) + 1]]
		forward: false
		event: aKeyboardEvent
		specialBlock: [:dummy | 1].
	^true! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 11/8/2015 13:05'!
cut
	"Cut out the current selection and redisplay the paragraph if necessary.  Undoer & Redoer: undoCutCopy:"

	self lineSelectAndEmptyCheck: [^ self].

	self clipboardTextPut: self selection.
	self replaceSelectionWith: self nullText.! !

!TextEditor methodsFor: 'do-its' stamp: 'mt 4/11/2016 10:45'!
debugIt

	| receiver context |
	self lineSelectAndEmptyCheck: [^self].

	(model respondsTo: #debugExpression:)
		ifTrue: [^ model perform: #debugExpression: with: self selection].

	receiver := (model respondsTo: #doItReceiver) 
		ifTrue: [model doItReceiver]
		ifFalse: [nil].
		
	context := (model respondsTo: #doItContext) 
		ifTrue: [model doItContext]
		ifFalse: [nil].

	(self compileSelectionFor: receiver in: context) ifNotNil: [:method |
		ToolSet debugMethod: method forReceiver: receiver inContext: context].! !

!TextEditor methodsFor: 'editing keys' stamp: 'topa 10/19/2015 19:48'!
debugIt: aKeyboardEvent

	self debugIt.
	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'cmm 8/16/2011 22:39'!
destructiveBackWord: aKeyboardEvent
	typeAhead isEmpty
		ifTrue: [ super destructiveBackWord: aKeyboardEvent ]
		ifFalse: [ typeAhead reset ].
	^ false! !

!TextEditor methodsFor: 'typing support' stamp: 'nice 8/3/2011 20:32'!
dispatchOnEnterWith: aKeyboardEvent
	"Enter key hit.  Treat is as an 'accept', viz a synonym for cmd-s.  If cmd key is down, treat is as a synonym for print-it. "

	aKeyboardEvent commandKeyPressed
		ifTrue:
			[self printIt.]
		ifFalse: 
			[self insertAndCloseTypeIn.
			self accept].
	^ true! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 4/19/2016 09:16'!
dispatchOnKeyboardEvent: aKeyboardEvent 
	"Carry out the action associated with this character, if any. Type-ahead is passed so some routines can flush or use it."
	
	| honorCommandKeys typedChar |
	typedChar := aKeyboardEvent keyCharacter.
	
	"Handle one-line input fields."
	(typedChar == Character cr and: [morph acceptOnCR])
		ifTrue: [^ true].
	
	"Clear highlight for last opened parenthesis."
	self clearParens.
	
	"Handle line breaks and auto indent."
	typedChar == Character cr ifTrue: [
		aKeyboardEvent controlKeyPressed
			ifTrue: [^ self normalCharacter: aKeyboardEvent].
		aKeyboardEvent shiftPressed
			ifTrue: [^ self lf: aKeyboardEvent].
		aKeyboardEvent commandKeyPressed
			ifTrue: [^ self crlf: aKeyboardEvent].
		^ self crWithIndent: aKeyboardEvent].

	"Handle indent/outdent with selected text block."
	typedChar == Character tab ifTrue: [
		aKeyboardEvent shiftPressed
			ifTrue: [self outdent: aKeyboardEvent. ^ true]
			ifFalse: [self hasMultipleLinesSelected
				ifTrue: [self indent: aKeyboardEvent. ^ true]]].

	honorCommandKeys := Preferences cmdKeysInText.

	(honorCommandKeys and: [typedChar == Character enter])
		ifTrue: [^ self dispatchOnEnterWith: aKeyboardEvent].
	
	"Special keys overwrite crtl+key combinations - at least on Windows. To resolve this
	conflict, assume that keys other than cursor keys aren't used together with Crtl."
	((self class specialShiftCmdKeys includes: aKeyboardEvent keyValue)
		and: [aKeyboardEvent keyValue < 27])
			ifTrue: [^ aKeyboardEvent controlKeyPressed
				ifTrue: [self
							perform: (self class shiftCmdActions at: aKeyboardEvent keyValue + 1)
							with: aKeyboardEvent]
				ifFalse: [self
							perform: (self class cmdActions at: aKeyboardEvent keyValue + 1)
							with: aKeyboardEvent]].
			
	"backspace, and escape keys (ascii 8 and 27) are command keys"
	((honorCommandKeys and: [aKeyboardEvent commandKeyPressed])
		or: [self class specialShiftCmdKeys includes: aKeyboardEvent keyValue])
			ifTrue: [ ^ aKeyboardEvent shiftPressed
				ifTrue: [self
							perform: (self class shiftCmdActions at: aKeyboardEvent keyValue + 1)
							with: aKeyboardEvent]
				ifFalse: [self
							perform: (self class cmdActions at: aKeyboardEvent keyValue + 1)
							with: aKeyboardEvent]].

	"the control key can be used to invoke shift-cmd shortcuts"
	(honorCommandKeys and: [ aKeyboardEvent controlKeyPressed ])
		ifTrue: [^ self
					perform: (self class shiftCmdActions at: aKeyboardEvent keyValue + 1)
					with: aKeyboardEvent].

	"Enclose selection with brackets etc."
	((self class encloseSelection and: [self hasSelection]) and: [self enclose: aKeyboardEvent])
		ifTrue: [^ true].

	"Automatically enclose paired characters such as brackets."
	(self class autoEnclose and: [self autoEncloseFor: typedChar])
		ifTrue: [^ true].
					
	self normalCharacter: aKeyboardEvent.
	^ false! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:40'!
displayIfFalse: aKeyboardEvent
	"Replace the current text selection with the text 'ifFalse:'--initiated by 
	ctrl-f."

	self addString: 'ifFalse:'.
	^false! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:40'!
displayIfTrue: aKeyboardEvent
	"Replace the current text selection with the text 'ifTrue:'--initiated by 
	ctrl-t."

	self addString: 'ifTrue:'.
	^false! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'mt 11/10/2015 09:56'!
doAgainUpToEnd: aKeyboardEvent 
	"Do the previous thing again once. 1/26/96 sw"

	self insertAndCloseTypeIn.
	self againUpToEnd.
	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'mt 11/8/2015 13:33'!
doAgain: aKeyboardEvent 
	"Do the previous thing again once. 1/26/96 sw"

	self insertAndCloseTypeIn.
	self again.
	^ true! !

!TextEditor methodsFor: 'do-its' stamp: 'jmv 11/4/2008 11:51'!
doIt
	"Set the context to include pool vars of the model.  Then evaluate."
	^ self evaluateSelection! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
doIt: aKeyboardEvent
	"Called when user hits cmd-d.  Select the current line, if relevant, then evaluate and execute.  2/1/96 sw.
	2/29/96 sw: don't call selectLine; it's done by doIt now"

	self doIt.
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'ar 8/17/2009 23:27'!
emphasisExtras
	"Answer an array of extra items for the emphasis menu"
	^#()! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 5/3/2016 08:48'!
enclose: aKeyboardEvent
	"Insert or remove bracket characters around the current selection."

	| character left right startIndex stopIndex oldSelection which t |
	character := aKeyboardEvent shiftPressed
					ifTrue: ['{}|"<>' at: ('[]\'',.' indexOf: aKeyboardEvent keyCharacter) ifAbsent: [aKeyboardEvent keyCharacter]]
					ifFalse: [aKeyboardEvent keyCharacter].
	self closeTypeIn.
	startIndex := self startIndex.
	stopIndex := self stopIndex.
	oldSelection := self selection.
	which := '([<{|"''9' indexOf: character ifAbsent: [ ^ false ].
	"Allow Control key in lieu of Alt+Shift for (, {, and double-quote."
	left := ((Preferences cmdKeysInText and: [ aKeyboardEvent controlKeyPressed ])
		ifTrue: [ '({<{|""(' ]
		ifFalse: ['([<{|"''(']) at: which.
	right := ((Preferences cmdKeysInText and: [ aKeyboardEvent controlKeyPressed ])
		ifTrue: [ ')}>}|"")' ] 
		ifFalse: [')]>}|"'')']) at: which.
	t := self text.
	((startIndex > 1 and: [stopIndex <= t size])
			and: [ (t at: startIndex-1) = left and: [(t at: stopIndex) = right]])
		ifTrue:
			["already enclosed; strip off brackets"
			self selectFrom: startIndex-1 to: stopIndex.
			self replaceSelectionWith: oldSelection]
		ifFalse:
			["not enclosed; enclose by matching brackets"
			self replaceSelectionWith:
				(Text string: (String with: left), oldSelection string, (String with: right) attributes: emphasisHere).
			self selectFrom: startIndex+1 to: stopIndex].
	^true! !

!TextEditor methodsFor: 'new selection' stamp: 'nice 2/8/2010 10:07'!
encompassLine: anInterval
	"Return an interval that encompasses the entire line"
	| string left right |
	string := paragraph text string.
	left := (string lastIndexOfAnyOf: CharacterSet crlf startingAt: anInterval first - 1 ifAbsent:[0]) + 1.
	right := (string indexOfAnyOf: CharacterSet crlf startingAt: anInterval last + 1 ifAbsent: [string size + 1]) - 1.
	^left to: right! !

!TextEditor methodsFor: 'private' stamp: 'jmv 12/21/2010 14:49'!
endOfLine: position
	"Redefined in subclasses using Paragraph support"
	| targetLine |
	targetLine := paragraph lines at: (paragraph lineIndexFor: position).
	^ targetLine = paragraph lastLine
		ifFalse: [ targetLine last ]
		ifTrue: [ targetLine last + 1 ]! !

!TextEditor methodsFor: 'do-its' stamp: 'nice 3/31/2011 21:51'!
evaluateSelection
	"Treat the current selection as an expression; evaluate it and return the result"
	
	^self evaluateSelectionAndDo: [:result | result]! !

!TextEditor methodsFor: 'do-its' stamp: 'mt 5/1/2015 11:44'!
evaluateSelectionAndDo: aBlock
	"Treat the current selection as an expression; evaluate it and invoke aBlock with the result."
	| result rcvr ctxt |
	self lineSelectAndEmptyCheck: [^ nil].

	(model respondsTo: #evaluateExpression:) ifTrue: [
		^ aBlock value: (model perform: #evaluateExpression: with: self selection)].

	(model respondsTo: #doItReceiver) 
		ifTrue: [ rcvr := model doItReceiver.
				ctxt := model doItContext]
		ifFalse: [rcvr := ctxt := nil].
	result := [
		rcvr class evaluatorClass new 
			evaluate: self selectionAsStream
			in: ctxt
			to: rcvr
			notifying: self
			ifFail: [morph flash. ^ nil]
			logged: true.
	] 
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true].
		
	(model respondsTo: #expressionEvaluated:result:) ifTrue: [
		model perform: #expressionEvaluated:result: with: self selection with: result].
		
	^aBlock value: result! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 11/8/2015 17:42'!
exchange
	"See comment in exchangeWith:"

	self exchangeWith: otherInterval.! !

!TextEditor methodsFor: 'private' stamp: 'mt 11/10/2015 11:07'!
exchangeWith: prior
	"If the prior selection is non-overlapping and legal, exchange the text of
	 it with the current selection and leave the currently selected text selected
	 in the location of the prior selection (or leave a caret after a non-caret if it was
	 exchanged with a caret).  If both selections are carets, flash & do nothing.
	 Don't affect the paste buffer."

	| start stop before selection priorSelection delta altInterval |
	start := self startIndex.
	stop := self stopIndex - 1.
	
	(((prior first <= prior last) and: [start <= stop])
		and: [self isDisjointFrom: prior])
			ifFalse: [morph flash. ^ self].
		
	before := prior last < start.
	selection := self selection.
	priorSelection := paragraph text copyFrom: prior first to: prior last.
	delta := before ifTrue: [0] ifFalse: [priorSelection size - selection size].
	
	"Create first undo command."
	self replaceSelectionWith: priorSelection.
	self history previous isCompositeRedo: true.
	
	self selectInvisiblyFrom: prior first + delta to: prior last + delta.
	delta := before ifTrue: [stop - prior last] ifFalse: [start - prior first].

	"Create second undo command."
	self replaceSelectionWith: selection.
	self history previous isCompositeUndo: true.

	altInterval := prior first + delta to: prior last + delta.
				
	"If one was a caret, make it otherInterval & leave the caret after the other"
	prior first > prior last ifTrue: [self selectAt: prior last + 1].
	otherInterval := start > stop
		ifTrue: [self selectAt: altInterval last + 1. prior]
		ifFalse: [altInterval]! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 22:46'!
exchange: eKeyboardEvent
	"Exchange the current and prior selections.  Keeps typeahead."

	self insertAndCloseTypeIn.
	self exchange.
	^true! !

!TextEditor methodsFor: 'menu messages' stamp: 'nice 5/8/2011 12:11'!
explain
	"Try to shed some light on what kind of entity the current selection is. 
	The selection must be a single token or construct. Insert the answer after 
	the selection. Send private messages whose names begin with 'explain' 
	that return a string if they recognize the selection, else nil."

	
Cursor execute showWhile: 
			[ | string numbers delimitors cgVars reply selectors tiVars symbol sorry |
			sorry := '"Sorry, I can''t explain that.  Please select a single
token, construct, or special character.'.
			sorry := sorry , (morph canDiscardEdits
							ifFalse: ['  Also, please cancel or accept."']
							ifTrue: ['"']).
			(string := self selection asString) isEmpty
				ifTrue: [reply := '']
				ifFalse: [string := self explainScan: string.
					"Remove space, tab, cr"
					"Temps and Instance vars need only test strings that are all letters"
					(string allSatisfy: [:char | char isLetter or: [char isDigit]])
						ifTrue: 
							[tiVars := self explainTemp: string.
							tiVars == nil ifTrue: [tiVars := self explainInst: string]].
					(tiVars == nil and: [model respondsTo: #explainSpecial:])
						ifTrue: [tiVars := model explainSpecial: string].
					tiVars == nil
						ifTrue: [tiVars := '']
						ifFalse: [tiVars := tiVars , '\' withCRs].
					"Context, Class, Pool, and Global vars, and Selectors need 
					only test symbols"
					(Symbol hasInterned: string ifTrue: [:s | symbol := s])
						ifTrue: [cgVars := self explainCtxt: symbol.
							cgVars == nil
								ifTrue: [cgVars := self explainClass: symbol.
									cgVars == nil ifTrue: [cgVars := self explainGlobal: symbol]].
							"See if it is a Selector (sent here or not)"
							selectors := self explainMySel: symbol.
							selectors == nil
								ifTrue: 
									[selectors := self explainPartSel: string.
									selectors == nil ifTrue: [
										selectors := self explainAnySel: symbol]]]
						ifFalse: [selectors := self explainPartSel: string].
					cgVars == nil
						ifTrue: [cgVars := '']
						ifFalse: [cgVars := cgVars , '\' withCRs].
					selectors == nil
						ifTrue: [selectors := '']
						ifFalse: [selectors := selectors , '\' withCRs].
					string size = 1
						ifTrue: ["single special characters"
							delimitors := self explainChar: string]
						ifFalse: ["matched delimitors"
							delimitors := self explainDelimitor: string].
					numbers := self explainNumber: string.
					numbers == nil ifTrue: [numbers := ''].
					delimitors == nil ifTrue: [delimitors := ''].
					reply := tiVars , cgVars , selectors , delimitors , numbers].
			reply size = 0 ifTrue: [reply := sorry].
			self afterSelectionInsertAndSelect: reply]! !

!TextEditor methodsFor: 'explain' stamp: 'jmv 8/8/2009 15:26'!
explainAnySel: symbol 
	"Is this any message selector?"

	| list reply |
	list := self systemNavigation allClassesImplementing: symbol.
	list size = 0 ifTrue: [^nil].
	list size < 12
		ifTrue: [reply := ' is a message selector which is defined in these classes ' , list printString]
		ifFalse: [reply := ' is a message selector which is defined in many classes'].
	^'"' , symbol , reply , '."' , '\' withCRs, 'SystemNavigation new browseAllImplementorsOf: #' , symbol! !

!TextEditor methodsFor: 'explain' stamp: 'nice 2/8/2010 10:58'!
explainChar: string
	"Does string start with a special character?"

	| char |
	char := string at: 1.
	char = $. ifTrue: [^'"Period marks the end of a Smalltalk statement.  A period in the middle of a number means a decimal point.  (The number is an instance of class Float)."'].
	char = $' ifTrue: [^'"The characters between two single quotes are made into an instance of class String"'].
	char = $" ifTrue: [^'"Double quotes enclose a comment.  Smalltalk ignores everything between double quotes."'].
	char = $# ifTrue: [^'"The characters following a hash mark are made into an instance of class Symbol.  If parenthesis follow a hash mark, an instance of class Array is made.  It contains literal constants."'].
	(char = $( or: [char = $)]) ifTrue: [^'"Expressions enclosed in parenthesis are evaluated first"'].
	(char = $[ or: [char = $]]) ifTrue: [^'"The code inside square brackets is an unevaluated block of code.  It becomes an instance of BlockContext and is usually passed as an argument."'].
	(char = ${ or: [char = $}]) ifTrue: [^ '"A sequence of expressions separated by periods, when enclosed in curly braces, are evaluated to yield the elements of a new Array"'].
	(char = $< or: [char = $>]) ifTrue: [^'"<primitive: xx> means that this method is usually preformed directly by the virtual machine.  If this method is primitive, its Smalltalk code is executed only when the primitive fails."'].
	char = $^ ifTrue: [^'"Uparrow means return from this method.  The value returned is the expression following the ^"'].
	char = $| ifTrue: [^'"Vertical bars enclose the names of the temporary variables used in this method.  In a block, the vertical bar separates the argument names from the rest of the code."'].
	char = $_ ifTrue: [^'"Left arrow means assignment.  The value of the expression after the left arrow is stored into the variable before it."'].
	char = $; ifTrue: [^'"Semicolon means cascading.  The message after the semicolon is sent to the same object which received the message before the semicolon."'].
	char = $: ifTrue: [^'"A colon at the end of a keyword means that an argument is expected to follow.  Methods which take more than one argument have selectors with more than one keyword.  (One keyword, ending with a colon, appears before each argument).', '\\' withCRs, 'A colon before a variable name just inside a block means that the block takes an agrument.  (When the block is evaluated, the argument will be assigned to the variable whose name appears after the colon)."'].
	char = $$ ifTrue: [^'"The single character following a dollar sign is made into an instance of class Character"'].
	char = $- ifTrue: [^'"A minus sign in front of a number means a negative number."'].
	char = $e ifTrue: [^'"An e in the middle of a number means that the exponent follows."'].
	char = $r ifTrue: [^'"An r in the middle of a bunch of digits is an instance of Integer expressed in a certain radix.  The digits before the r denote the base and the digits after it express a number in that base."'].
	char = Character space ifTrue: [^'"the space Character"'].
	char = Character tab ifTrue: [^'"the tab Character"'].
	char = Character cr ifTrue: [^'"the carriage return Character"'].
	char = Character lf ifTrue: [^'"the line feed Character"'].
	^nil! !

!TextEditor methodsFor: 'explain' stamp: 'cwp 1/18/2014 10:28'!
explainClass: symbol 
	"Is symbol a class variable or a pool variable?"
	| class reply classes |
	(model respondsTo: #selectedClassOrMetaClass)
		ifFalse: [^ nil].
	(class := model selectedClassOrMetaClass) ifNil: [^ nil].
	"no class is selected"
	(class isKindOf: Metaclass)
		ifTrue: [class := class soleInstance].
	classes := class withAllSuperclasses.
	"class variables"
	reply := classes detect: [:each | each classVarNames anySatisfy: [:name | symbol = name]]
				ifNone: [].
	reply == nil ifFalse: [^ '"is a class variable, defined in class ' , reply printString , '"\' withCRs , 'SystemNavigation new browseAllCallsOn: (' , reply printString , ' classPool associationAt: #' , symbol , ').'].
	"pool variables"
	classes do: [:each | each sharedPools
			anySatisfy: [:pool | (pool includesKey: symbol)
					and: 
						[reply := pool.
						true]]].
	reply
		ifNil: [(class environment isUndeclared: symbol)
				ifTrue: [^ '"is an undeclared variable.' , '"\' withCRs , 'SystemNavigation new browseAllCallsOn: (Undeclared associationAt: #' , symbol , ').']]
		ifNotNil: 
			[classes := WriteStream on: Array new.
			self systemNavigation
				allBehaviorsDo: [:each | (each sharedPools
						detect: 
							[:pool | 
							pool == reply]
						ifNone: [])
						~~ nil ifTrue: [classes nextPut: each]].
			"Perhaps not print whole list of classes if too long. (unlikely)"
			^ '"is a pool variable from the pool ' , (Smalltalk globals keyAtIdentityValue: reply) asString , ', which is used by the following classes ' , classes contents printString , '"\' withCRs , 'SystemNavigation new browseAllCallsOn: (' , (Smalltalk globals keyAtIdentityValue: reply) asString , ' bindingOf: #' , symbol , ').'].
	^ nil! !

!TextEditor methodsFor: 'explain' stamp: 'jmv 11/4/2008 11:51'!
explainCtxt: symbol 
	"Is symbol a context variable?"

	| reply classes text cls |
	symbol = #nil ifTrue: [reply := '"is a constant.  It is the only instance of class UndefinedObject.  nil is the initial value of all variables."'].
	symbol = #true ifTrue: [reply := '"is a constant.  It is the only instance of class True and is the receiver of many control messages."'].
	symbol = #false ifTrue: [reply := '"is a constant.  It is the only instance of class False and is the receiver of many control messages."'].
	symbol = #thisContext ifTrue: [reply := '"is a context variable.  Its value is always the MethodContext which is executing this method."'].
	(model respondsTo: #selectedClassOrMetaClass) ifTrue: [
		cls := model selectedClassOrMetaClass].
	cls ifNil: [^ reply].	  "no class known"
	symbol = #self ifTrue: 
			[classes := cls withAllSubclasses.
			classes size > 12
				ifTrue: [text := cls printString , ' or a subclass']
				ifFalse: 
					[classes := classes printString.
					text := 'one of these classes' , (classes copyFrom: 4 to: classes size)].
			reply := '"is the receiver of this message; an instance of ' , text , '"'].
	symbol = #super ifTrue: [reply := '"is just like self.  Messages to super are looked up in the superclass (' , cls superclass printString , ')"'].
	^reply! !

!TextEditor methodsFor: 'explain' stamp: 'eem 3/16/2012 16:45'!
explainDelimitor: string
	"Is string enclosed in delimitors?"

	| str |
	(string at: 1) isLetter ifTrue: [^nil].  "only special chars"
	(string first = string last) ifTrue:
			[^ self explainChar: (String with: string first)]
		ifFalse:
			[(string first = $( and: [string last = $)]) ifTrue:
				[^ self explainChar: (String with: string first)].
			(string first = $[ and: [string last = $]]) ifTrue:
				[^ self explainChar: (String with: string first)].
			(string first = ${ and: [string last = $}]) ifTrue:
				[^ self explainChar: (String with: string first)].
			(string first = $< and: [string last = $>]) ifTrue:
				[^ self explainChar: (String with: string first)].
			(string first = $# and: [string last = $)]) ifTrue:
				[^'"An instance of class Array.  The Numbers, Characters, Symbols or Arrays between the parenthesis are the elements of the Array."'].
			string first = $# ifTrue:
				[^'"An instance of class Symbol."'].
			(string first = $$ and: [string size = 2]) ifTrue:
				[^'"An instance of class Character.  This one is the character ', (String with: string last), '."'].
			(string first = $:) ifTrue:
				[string = ':=' ifTrue:
					[^'"Colon equals means assignment.  The value of the expression after the := is stored into the variable before it."'].
				str := string allButFirst.
				(self explainTemp: str) ifNotNil:
					[^'"An argument to this block will be bound to the temporary variable ', str, '."']]].
	^ nil! !

!TextEditor methodsFor: 'explain' stamp: 'nice 5/8/2011 12:13'!
explainGlobal: symbol 
	"Is symbol a global variable?"
	| reply classes |
	reply := Smalltalk at: symbol ifAbsent: [^nil].
	(reply class == Dictionary or:[reply isKindOf: SharedPool class])
		ifTrue: 
			[classes := Set new.
			self systemNavigation allBehaviorsDo: [:each | (each sharedPools anySatisfy: [:pool | pool == reply])
				ifTrue: [classes add: each]].
			classes := classes printString.
			^'"is a global variable.  It is a pool which is used by the following classes ' , (classes allButFirst: 5) , '"'].
	(reply isKindOf: Behavior)
		ifTrue: [^'"is a global variable.  ' , symbol , ' is a class in category ', reply category,
			'."', '\' withCRs, 'Browser newOnClass: ' , symbol , '.'].
	symbol == #Smalltalk ifTrue: [^'"is a global.  Smalltalk is the only instance of SystemDictionary and holds all global variables."'].
	^'"is a global variable.  ' , symbol , ' is ' , reply printString , '"'! !

!TextEditor methodsFor: 'explain' stamp: 'nice 5/6/2011 00:21'!
explainInst: string 
	"Is string an instance variable of this class?"
	| classes cls |

	(model respondsTo: #selectedClassOrMetaClass) ifTrue: [
		cls := model selectedClassOrMetaClass].
	cls ifNil: [^ nil].	  "no class known"
	classes := (Array with: cls)
				, cls allSuperclasses.
	classes := classes detect: [:each | each instVarNames anySatisfy: [:name | name = string]] ifNone: [^nil].
	classes := classes printString.
	^ '"is an instance variable of the receiver; defined in class ' , classes , 
		'"\' withCRs , classes , ' systemNavigation browseAllAccessesTo: ''' , string , ''' from: ', classes, '.'! !

!TextEditor methodsFor: 'explain' stamp: 'nice 5/8/2011 12:13'!
explainMySel: symbol 
	"Is symbol the selector of this method?  Is it sent by this method?  If 
	not, then expalin will call (explainPartSel:) to see if it is a fragment of a 
	selector sent here.  If not, explain will call (explainAnySel:) to catch any 
	selector. "

	| lits classes msg |
	(model respondsTo: #selectedMessageName) ifFalse: [^ nil].
	(msg := model selectedMessageName) ifNil: [^nil].	"not in a message"
	classes := self systemNavigation allClassesImplementing: symbol.
	classes size > 12
		ifTrue: [classes := 'many classes']
		ifFalse: [classes := 'these classes ' , classes printString].
	msg = symbol
		ifTrue: [^ '"' , symbol , ' is the selector of this very method!!  It is defined in ',
			classes , '.  To see the other definitions, go to the message list pane, get the menu from the top of the scroll bar, and select ''implementors of...''."']
		ifFalse: 
			[lits := (model selectedClassOrMetaClass compiledMethodAt:
				msg) messages.
			(lits anySatisfy: [:each | each == symbol])
				ifFalse: [^nil].
			^ '"' , symbol , ' is a message selector which is defined in ', classes , '.  To see the definitions, go to the message list pane, get the menu from the top of the scroll bar, and select ''implementors of...''."'].! !

!TextEditor methodsFor: 'explain' stamp: 'jmv 11/4/2008 11:51'!
explainNumber: string 
	"Is string a Number?"

	| strm c |
	(c := string at: 1) isDigit ifFalse: [(c = $- and: [string size > 1 and: [(string at: 2) isDigit]])
			ifFalse: [^nil]].
	strm := ReadStream on: string.
	c := Number readFrom: strm.
	strm atEnd ifFalse: [^nil].
	c printString = string
		ifTrue: [^'"' , string , ' is a ' , c class name , '"']
		ifFalse: [^'"' , string , ' (= ' , c printString , ') is a ' , c class name , '"']! !

!TextEditor methodsFor: 'explain' stamp: 'nice 5/8/2011 12:14'!
explainPartSel: string 
	"Is this a fragment of a multiple-argument selector sent in this method?"
	| lits whole reply classes s msg |

	(model respondsTo: #selectedMessageName) ifFalse: [^ nil].
	(msg := model selectedMessageName) ifNil: [^ nil].  "not in a message"
	string last == $: ifFalse: [^ nil].
	"Name of this method"
	lits := Array with: msg.
	(whole := lits detect: [:each | each keywords anySatisfy: [:frag | frag = string] ]
				ifNone: []) ~~ nil
		ifTrue: [reply := ', which is the selector of this very method!!'.
			s := '.  To see the other definitions, go to the message list pane, get the menu from the top of the scroll bar, and select ''implementors of...''."']
		ifFalse: 
			["Selectors called from this method"
			lits := (model selectedClassOrMetaClass compiledMethodAt:
				msg) messages.
			(whole := lits detect: [:each | (each keywords detect: [:frag | frag = string]
							ifNone: []) ~~ nil]
						ifNone: []) ~~ nil
				ifFalse: [string = 'primitive:'
					ifTrue: [^self explainChar: '<']
					ifFalse: [^nil]].
			reply := '.'.
			s := '.  To see the definitions, go to the message list pane, get the menu from the top of the scroll bar, and select ''implementors of...''."'].
	classes := self systemNavigation allClassesImplementing: whole.
	classes size > 12
		ifTrue: [classes := 'many classes']
		ifFalse: [classes := 'these classes ' , classes printString].
	^ '"' , string , ' is one part of the message selector ' , whole, reply , '  It is defined in ' , classes , s! !

!TextEditor methodsFor: 'explain' stamp: 'jmv 11/4/2008 11:51'!
explainScan: string 
	"Remove beginning and trailing space, tab, cr.
	 1/15/96 sw: copied intact from BrowserCodeController"

	| c beg end |
	beg := 1.
	end := string size.
	
	[beg = end ifTrue: [^string copyFrom: 1 to: 1].
	"if all blank, tell about the first"
	c := string at: beg.
	c = Character space or: [c = Character tab or: [c = Character cr]]]
		whileTrue: [beg := beg + 1].
	
	[c := string at: end.
	c = Character space or: [c = Character tab or: [c = Character cr]]]
		whileTrue: [end := end - 1].
	^string copyFrom: beg to: end	"Return purely visible characters"! !

!TextEditor methodsFor: 'explain' stamp: 'nice 2/23/2011 21:50'!
explainTemp: string 
	"Is string the name of a temporary variable (method or block argument or temporary)?"

	| selectedClass methodNode tempNode |
	(model respondsTo: #selectedMessageName) ifFalse: [^ nil].
	model selectedMessageName ifNil: [^nil].	"not in a method"
	selectedClass := model selectedClassOrMetaClass.
	methodNode := selectedClass newParser parse: model selectedMessage class: selectedClass.
	tempNode := methodNode encoder tempNodes detect: [:n| n name = string] ifNone: [^nil].
	^(tempNode isArg
		ifTrue: ['"is an argument to this ']
		ifFalse: ['"is a temporary variable in this ']),
	   (tempNode isDeclaredAtMethodLevel
		ifTrue: ['method"']
		ifFalse: ['block"'])! !

!TextEditor methodsFor: 'do-its' stamp: 'mt 4/17/2015 15:46'!
exploreIt

	 self evaluateSelectionAndDo: [:result |
		(model respondsTo: #exploreIt:result:)
			ifTrue: [model
				perform: #exploreIt:result:
				with: self selection
				with: result]
			ifFalse: [result explore]].! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
exploreIt: aKeyboardEvent
	"Explore the selection -- invoked via cmd-shift-I.  If there is no current selection, use the current line."

	self exploreIt.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'nice 2/26/2013 00:14'!
fileItIn
	"Make a Stream on the text selection and fileIn it.
	 1/24/96 sw: moved here from FileController; this function can be useful from any text window that shows stuff in chunk format"

	self selectionAsStream fileIn
! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
fileItIn: aKeyboardEvent
	"File in the selection; invoked via a keyboard shortcut, -- for now, cmd-shift-G."

	self fileItIn.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 11/18/2015 22:15'!
find
	"Prompt the user for a string to search for, and search the receiver from the current selection onward for it.  1/26/96 sw"

	self setSearchFromSelectionOrHistory.

	(UIManager default request: 'Find what to select? ' initialAnswer: FindText)
		ifEmpty: [^ self]
		ifNotEmpty: [:reply |
			FindText := reply.
			self findAgainNow].! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 11/18/2015 22:25'!
findAgain

	self setSearchFromSelectionOrHistory.
	
	^ self findAgainNow! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 11/18/2015 22:10'!
findAgainNow

	| where |	
	where := self text
				findString: FindText
				startingAt: self stopIndex
				caseSensitive: Preferences caseSensitiveFinds.
	
	where = 0
		ifTrue: [self selectFrom: self stopIndex to: self stopIndex - 1]
		ifFalse: [self selectFrom: where to: where + FindText size - 1].

	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'mt 11/18/2015 22:14'!
findAgain: aKeyboardEvent 
	"Find the desired text again."

	self
		insertAndCloseTypeIn;
		findAgain.
	
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 11/18/2015 22:15'!
findReplace

	self
		setSearchFromSelectionOrHistory;
		setReplacementFromHistory.

	(UIManager default request: 'Find what to replace?' initialAnswer: FindText)
		ifEmpty: [^ self]
		ifNotEmpty: [:find |
			(UIManager default request: ('Replace ''{1}'' with?' format: {find}) initialAnswer: (ChangeText ifEmpty: [find]))
				ifEmpty: [^ self]
				ifNotEmpty: [:replace |
					FindText := find.
					ChangeText := replace.
					self findReplaceAgainNow]]! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 12/21/2015 14:16'!
findReplaceAgain
		
	^ self again! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 11/18/2015 22:13'!
findReplaceAgainNow

	self hasSelection ifTrue: [
		"Search from the beginning of the current selection. Supports a nice combination with regular find feature."
		self selectInvisiblyFrom: self startIndex to: self startIndex - 1].
	
	self findAgainNow.
	self hasSelection ifFalse: [^ false].

	self replaceSelectionWith: ChangeText.
	self findAgainNow. "Select possible next thing to replace."

	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'mt 11/18/2015 22:12'!
findReplaceAgain: aKeyboardEvent 

	self
		insertAndCloseTypeIn;
		findReplaceAgain.

	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'mt 11/18/2015 22:15'!
findReplace: aKeyboardEvent

	self
		insertAndCloseTypeIn;
		findReplace.
		
	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'mt 11/18/2015 22:15'!
find: aKeyboardEvent
	"Prompt the user for what to find, then find it, searching from the current selection onward."

	self
		insertAndCloseTypeIn;
		find.

	^ true! !

!TextEditor methodsFor: 'displaying' stamp: 'jmv 11/4/2008 11:51'!
flash
	^ morph flash! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'mt 11/8/2015 13:04'!
forwardDelete: aKeyboardEvent
	"Delete forward over the next character.
	  Make Undo work on the whole type-in, not just the one char.
	wod 11/3/1998: If there was a selection use #zapSelectionWith: rather than #backspace: which was 'one off' in deleting the selection. Handling of things like undo or typeIn area were not fully considered."

	| startIndex stopIndex |
	morph readOnly ifTrue: [^ self].
	
	self hasSelection ifTrue: [
		"Create checkpoint in history."
		self replaceSelectionWith: self nullText.
		^ true].

	startIndex := self markIndex.
	startIndex > self text size ifTrue: [^ false].

	stopIndex := startIndex.

	"Forward delete next word"
	self flag: #consistency. "mt: We might want to implemented it like #backspace: and #backWord:."
	aKeyboardEvent shiftPressed
		ifTrue: [stopIndex := (self firstWordBoundaryAfter: stopIndex) - 1].

	self selectInvisiblyFrom: startIndex to: stopIndex.
	
	self isTypingIn ifTrue: [
		self history current type = #forwardDelete
			ifFalse: [self closeTypeIn]
			ifTrue: [
				"Append next characters that will be removed."
				self history current contentsBefore append: self selection.
				self history current intervalBefore in:  [:i |
					self history current intervalBefore: (i first to: i last + (stopIndex - startIndex + 1))]]].
	
	self openTypeInFor: #forwardDelete.
	self zapSelectionWith: self nullText.
	
	^ false! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 22:46'!
handleEmphasisExtra: index with: aKeyboardEvent
	"Handle an emphasis extra choice"
	^true! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 5/27/2011 10:24'!
hasCaret
	^ markBlock = pointBlock! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 11/20/2015 14:10'!
hasMultipleLinesSelected

	^ self hasSelection and: [self startBlock top < self stopBlock top]! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 21:00'!
hiddenInfo
	"In TextLinks, TextDoits, TextColor, and TextURLs, there is hidden info.  Return the entire string that was used by Cmd-6 to create this text attribute.  Usually enclosed in < >."

	| attrList |
	attrList := self text attributesAt: (self pointIndex + self markIndex)//2.
	attrList do: [:attr |
		(attr isKindOf: TextAction) ifTrue:
			[^ self selection asString, '<', attr info, '>']].
	"If none of the above"
	attrList do: [:attr |
		attr class == TextColor ifTrue:
			[^ self selection asString, '<', attr color printString, '>']].
	^ self selection asString, '[No hidden info]'! !

!TextEditor methodsFor: 'accessing' stamp: 'mt 11/7/2015 23:26'!
history
	^ history! !

!TextEditor methodsFor: 'accessing' stamp: 'mt 11/7/2015 23:26'!
history: commandHistory
	history := commandHistory.! !

!TextEditor methodsFor: 'menu messages' stamp: 'cmm 2/18/2010 12:27'!
implementorsOfIt
	"Open an implementors browser on the selected selector"
	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ morph flash].
	model browseAllImplementorsOf: aSelector! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
implementorsOfIt: aKeyboardEvent
	"Triggered by Cmd-m; browse implementors of the selector represented by the current selection, if plausible. 2/1/96 sw"

	self implementorsOfIt.
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:44'!
indent: aKeyboardEvent
	"Add a tab at the front of every line occupied by the selection. Invoked from keyboard via cmd-shift-R.  2/29/96 sw"

	^ self inOutdent: aKeyboardEvent delta: 1! !

!TextEditor methodsFor: 'private' stamp: 'jmv 11/4/2008 11:51'!
indent: delta fromStream: inStream toStream: outStream
	"Append the contents of inStream to outStream, adding or deleting delta or -delta
	 tabs at the beginning, and after every CR except a final CR.  Do not add tabs
	 to totally empty lines, and be sure nothing but tabs are removed from lines."

	| ch skip cr tab prev atEnd |
	cr := Character cr.
	tab := Character tab.
	delta > 0
		ifTrue: "shift right"
			[prev := cr.
			 [ch := (atEnd := inStream atEnd) ifTrue: [cr] ifFalse: [inStream next].
			  (prev == cr and: [ch ~~ cr]) ifTrue:
				[delta timesRepeat: [outStream nextPut: tab]].
			  atEnd]
				whileFalse:
					[outStream nextPut: ch.
					prev := ch]]
		ifFalse: "shift left"
			[skip := delta. "a negative number"
			 [inStream atEnd] whileFalse:
				[((ch := inStream next) == tab and: [skip < 0]) ifFalse:
					[outStream nextPut: ch].
				skip := ch == cr ifTrue: [delta] ifFalse: [skip + 1]]]! !

!TextEditor methodsFor: 'private' stamp: 'cmm 5/29/2015 16:06'!
indexOfNextNonwhitespaceCharacter
	pointBlock stringIndex
		to: paragraph string size
		do:
			[ : n | | char | (char := paragraph string at: n) isSeparator ifFalse: [ ^ n ] ].
	^ nil! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 9/6/2013 02:14'!
inOutdent: aKeyboardEvent delta: delta
	"Add/remove a tab at the front of every line occupied by the selection. Flushes typeahead.  Derived from work by Larry Tesler back in December 1985.  Now triggered by Cmd-L and Cmd-R.  2/29/96 sw"

	| realStart realStop lines startLine stopLine start stop adjustStart indentation numLines oldString newString newSize |

	"Operate on entire lines, but remember the real selection for re-highlighting later"
	realStart := self startIndex.
	realStop := self stopIndex - 1.

	"Special case a caret on a line of its own, including weird case at end of paragraph"
	(realStart > realStop and:
				[realStart < 2 or: [(paragraph string at: realStart - 1) == Character cr or: [(paragraph string at: realStart - 1) == Character lf]]])
		ifTrue:
			[delta < 0
				ifTrue:
					[morph flash]
				ifFalse:
					[self replaceSelectionWith: Character tab asText.
					self selectAt: realStart + 1].
			^true].

	lines := paragraph lines.
	startLine := paragraph lineIndexOfCharacterIndex: realStart.
	"start on a real line, not a wrapped line"
	[startLine = 1 or: [CharacterSet crlf includes: (paragraph string at: (lines at: startLine-1) last)]] whileFalse: [startLine := startLine - 1].
	stopLine := paragraph lineIndexOfCharacterIndex: (realStart max: realStop).
	start := (lines at: startLine) first.
	stop := (lines at: stopLine) last.
	
	"Pin the start of highlighting unless the selection starts a line"
	adjustStart := realStart > start.

	"Find the indentation of the least-indented non-blank line; never outdent more"
	indentation := (startLine to: stopLine) inject: 1000 into:
		[:m :l |
		m min: (paragraph indentationOfLineIndex: l ifBlank: [:tabs | 1000])].
	indentation + delta <= 0 ifTrue: ["^false"].

	numLines := stopLine + 1 - startLine.
	oldString := paragraph string copyFrom: start to: stop.
	newString := oldString species new: oldString size + ((numLines * delta) max: 0).

	"Do the actual work"
	newSize := 0.
	delta > 0
		ifTrue: [| tabs |
			tabs := oldString species new: delta withAll: Character tab.
			oldString lineIndicesDo: [:startL :endWithoutDelimiters :endL |
				startL < endWithoutDelimiters ifTrue: [newString replaceFrom: 1 + newSize to: (newSize := newSize + delta) with: tabs startingAt: 1].
				newString replaceFrom: 1 + newSize to: (newSize := 1 + newSize + endL - startL) with: oldString startingAt: startL]]
		ifFalse: [| tab |
			tab := Character tab.
			oldString lineIndicesDo: [:startL :endWithoutDelimiters :endL |
				| i |
				i := 0.
				[i + delta < 0 and: [ i + startL <= endWithoutDelimiters and: [(oldString at: i + startL) == tab]]] whileTrue: [i := i + 1].
				newString replaceFrom: 1 + newSize to: (newSize := 1 + newSize + endL - (i + startL)) with: oldString startingAt: i + startL]].
	newSize < newString size ifTrue: [newString := newString copyFrom: 1 to: newSize].

	"Adjust the range that will be highlighted later"
	adjustStart ifTrue: [realStart := (realStart + delta) max: start].
	realStop := realStop + newSize - oldString size.

	"Replace selection"
	self selectInvisiblyFrom: start to: stop.
	self replaceSelectionWith: newString asText.
	self selectFrom: realStart to: realStop. 	"highlight only the original range"
	^ true! !

!TextEditor methodsFor: 'typing support' stamp: 'cmm 8/17/2011 21:13'!
insertAndCloseTypeIn
	self
		insertTypeAhead ;
		closeTypeIn! !

!TextEditor methodsFor: 'new selection' stamp: 'mt 2/5/2016 14:57'!
insertAndSelect: aString at: anInteger

	self closeTypeIn.
	
	self selectInvisiblyFrom: anInteger to: anInteger - 1.
	self openTypeIn.

	self
		replace: self selectionInterval
		with: (Text string: (' ', aString) attributes: emphasisHere)
		and: [].

	self closeTypeIn.! !

!TextEditor methodsFor: 'typing support' stamp: 'cmm 8/16/2011 22:05'!
insertTypeAhead
	self typeAhead position = 0 ifFalse:
		[self zapSelectionWith: (Text string: self typeAhead contents emphasis: emphasisHere).
		self typeAhead reset.
		self unselect]! !

!TextEditor methodsFor: 'do-its' stamp: 'mt 4/17/2015 15:46'!
inspectIt

	 self evaluateSelectionAndDo: [:result |
		(model respondsTo: #inspectIt:result:)
			ifTrue: [model
				perform: #inspectIt:result:
				with: self selection
				with: result]
			ifFalse: [result inspect]].! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
inspectIt: aKeyboardEvent
	"Inspect the selection -- invoked via cmd-i.  If there is no current selection, use the current line.  1/17/96 sw
	 2/29/96 sw: don't call selectLine; it's done by inspectIt now"

	self inspectIt.
	^ true! !

!TextEditor methodsFor: 'private' stamp: 'jmv 11/4/2008 11:51'!
isDisjointFrom: anInterval
	"Answer true if anInterval is a caret not touching or within the current
	 interval, or if anInterval is a non-caret that does not overlap the current
	 selection."

	| fudge |
	fudge := anInterval size = 0 ifTrue: [1] ifFalse: [0].
	^(anInterval last + fudge < self startIndex or:
			[anInterval first - fudge >= self stopIndex])
! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 11/7/2015 22:15'!
isTypingIn
	^ beginTypeInIndex notNil! !

!TextEditor methodsFor: 'events' stamp: 'mt 11/7/2015 22:45'!
keyStroke: anEvent
 	self resetTypeAhead; deselect.

	(self dispatchOnKeyboardEvent: anEvent) 
		ifTrue: [
			self closeTypeIn.
			self storeSelectionInParagraph.
			^self].

	self openTypeIn.
	self 
		zapSelectionWith: self typeAhead contents; 
		resetTypeAhead;
		unselect;
		storeSelectionInParagraph.! !

!TextEditor methodsFor: 'new selection' stamp: 'nice 8/3/2011 21:33'!
lineSelectAndEmptyCheck: returnBlock
	"If the current selection is an insertion point, expand it to be the entire current line; if after that's done the selection is still empty, then evaluate the returnBlock, which will typically consist of '[^ self]' in the caller -- check senders of this method to understand this."

	self selectLine.  "if current selection is an insertion point, then first select the entire line in which occurs before proceeding"
	self hasSelection ifFalse: [morph flash.  ^ returnBlock value]! !

!TextEditor methodsFor: 'editing keys' stamp: 'mt 11/13/2015 10:09'!
makeCapitalized: aKeyboardEvent
	"Force the current selection to uppercase."

	| prev |
	prev := $-.  "not a letter"
	self replaceSelectionWith: 
		(self selection string collect:
			[:c | prev := prev isLetter ifTrue: [c asLowercase] ifFalse: [c asUppercase]]).
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 8/11/2010 10:08'!
makeLowercase: aKeyboardEvent
	"Force the current selection to lowercase.  Triggered by Cmd-X."

	self replaceSelectionWith: (self selection string asLowercase).
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 8/11/2010 10:08'!
makeUppercase: aKeyboardEvent
	"Force the current selection to uppercase.  Triggered by Cmd-Y."

	self replaceSelectionWith: (self selection string asUppercase).
	^ true! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 5/27/2011 10:25'!
markIndex
	^ markBlock stringIndex! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'nice 8/3/2011 23:25'!
markIndex: anIndex pointIndex: anotherIndex
	"Called, for example, when selecting text with shift+arrow keys"
	markBlock := paragraph characterBlockForIndex: anIndex.
	pointBlock := paragraph characterBlockForIndex: anotherIndex! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 8/8/2009 15:30'!
methodNamesContainingIt
	"Open a browser on methods names containing the selected string"

	self lineSelectAndEmptyCheck: [^ self].
	Cursor wait showWhile:
		[self systemNavigation browseMethodsWhoseNamesContain: self selection string withBlanksTrimmed].
	Cursor normal show! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 22:47'!
methodNamesContainingIt: aKeyboardEvent 
	"Browse methods whose selectors containing the selection in their names"

	self methodNamesContainingIt.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 8/8/2009 15:30'!
methodSourceContainingIt
	"Open a browser on methods which contain the current selection in their source (case-sensitive full-text search of source). Slow!!"

	self lineSelectAndEmptyCheck: [^ self].
	self systemNavigation browseMethodsWithSourceString: self selection string! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 22:47'!
methodStringsContainingIt: aKeyboardEvent 
	"Invoked from cmd-E -- open a browser on all methods holding string constants containing it.  Flushes typeahead. "

	self methodStringsContainingit.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 8/8/2009 15:31'!
methodStringsContainingit
	"Open a browser on methods which contain the current selection as part of a string constant."

	self lineSelectAndEmptyCheck: [^ self].
	self systemNavigation browseMethodsWithString: self selection string! !

!TextEditor methodsFor: 'accessing' stamp: 'ul 3/30/2011 03:16'!
model

	^model! !

!TextEditor methodsFor: 'model access' stamp: 'jmv 11/25/2010 15:02'!
model: aModel
	model := aModel! !

!TextEditor methodsFor: 'events' stamp: 'mt 4/21/2015 16:55'!
mouseDown: evt 
	"An attempt to break up the old processRedButton code into threee phases"
	| clickPoint b |

	oldInterval := self selectionInterval.
	clickPoint := evt cursorPoint.
	b := paragraph characterBlockAtPoint: clickPoint.

	(paragraph clickAt: clickPoint for: model controller: self) ifTrue: [
		markBlock := b.
		pointBlock := b.
		evt hand releaseKeyboardFocus: morph.
		^ self ].
	
	evt shiftPressed
		ifFalse: [
			self closeTypeIn.
			markBlock := b.
			pointBlock := b ]
		 ifTrue: [
			self closeTypeIn.
			self mouseMove: evt ].
       self storeSelectionInParagraph! !

!TextEditor methodsFor: 'events' stamp: 'nice 8/3/2011 23:25'!
mouseMove: evt
	"Change the selection in response to mouse-down drag"

	pointBlock := paragraph characterBlockAtPoint: evt position.
	self storeSelectionInParagraph! !

!TextEditor methodsFor: 'events' stamp: 'jmv 11/4/2008 11:51'!
mouseUp: evt
	"An attempt to break up the old processRedButton code into threee phases"
	oldInterval ifNil: [^ self].  "Patched during clickAt: repair"
	(self hasCaret 
		and: [oldInterval = self selectionInterval])
		ifTrue: [self selectWord].
	self setEmphasisHere.
	(self isDisjointFrom: oldInterval) ifTrue:
		[otherInterval := oldInterval].
	self storeSelectionInParagraph! !

!TextEditor methodsFor: 'private' stamp: 'nice 8/3/2011 22:37'!
moveCursor: directionBlock forward: forward event: aKeyboardEvent specialBlock: specialBlock 
	super moveCursor: directionBlock forward: forward event: aKeyboardEvent specialBlock: specialBlock.
	self setEmphasisHere! !

!TextEditor methodsFor: 'new selection' stamp: 'jmv 9/17/2010 12:09'!
nextTokenFrom: start direction: dir
	"simple token-finder for compiler automated corrections"
	| loc str |
	loc := start + dir.
	str := paragraph string.
	[(loc between: 1 and: str size) and: [(str at: loc) isSeparator]]
		whileTrue: [loc := loc + dir].
	^ loc! !

!TextEditor methodsFor: 'new selection' stamp: 'nice 2/22/2012 02:35'!
notify: aString at: anInteger in: aStream 
	"The compilation of text failed. The syntax error is noted as the argument, 
	aString. Insert it in the text at starting character position anInteger."

	self insertAndSelect: aString at: (anInteger max: 1)! !

!TextEditor methodsFor: 'private' stamp: 'jmv 7/29/2009 22:36'!
nullText

	^Text string: '' attributes: emphasisHere! !

!TextEditor methodsFor: 'attributes' stamp: 'jmv 7/27/2009 14:20'!
offerFontMenu
	"Present a menu of available fonts, and if one is chosen, apply it to the current selection.  
	Use only names of Fonts of this paragraph  "
	
	^self changeTextFont! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:44'!
offerFontMenu: aKeyboardEvent 
	"The user typed the command key that requests a font change; Offer the font menu.  5/27/96 sw"

	self insertAndCloseTypeIn.
	self offerFontMenu.
	^ true! !

!TextEditor methodsFor: 'menu commands' stamp: 'nice 8/3/2011 21:08'!
offerMenuFromEsc: aKeyboardEvent 
	"The escape key was hit while the receiver has the keyboard focus; take action"

	ActiveEvent shiftPressed ifFalse: [
		self raiseContextMenu: aKeyboardEvent ].
	^true! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 11/7/2015 22:24'!
openTypeIn

	self openTypeInFor: nil.! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 11/8/2015 16:33'!
openTypeInFor: editType
	"Set up UndoSelection to null text (to be added to by readKeyboard and backTo:),
	 beginTypeInBlock to keep track of the leftmost backspace, and UndoParameter to tally
	 how many deleted characters were backspaced over rather than 'cut'.
	 You can't undo typing until after closeTypeIn."

	beginTypeInIndex ifNil: [
		beginTypeInIndex := self startIndex.
		self history beginRemember: (TextEditorCommand new
			type: editType;
			contentsBefore: (self hasSelection ifTrue: [self selection] ifFalse: [self nullText]);
			intervalBefore: (beginTypeInIndex to: self stopIndex-1)
			yourself)].! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:45'!
outdent: aKeyboardEvent
	"Remove a tab from the front of every line occupied by the selection.
	Invoked from keyboard via cmd-shift-L.  2/29/96 sw"

	^ self inOutdent: aKeyboardEvent delta: -1! !

!TextEditor methodsFor: 'private' stamp: 'jmv 11/4/2008 11:51'!
pageHeight
	| howManyLines visibleHeight totalHeight ratio |
	howManyLines := paragraph numberOfLines.
	visibleHeight := self visibleHeight.
	totalHeight := self totalTextHeight.
	ratio := visibleHeight / totalHeight.
	^(ratio * howManyLines) rounded - 2! !

!TextEditor methodsFor: 'accessing' stamp: 'ul 12/5/2010 15:20'!
paragraph

	^paragraph! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 11/7/2015 18:30'!
paste
	"Update command history."
	
	self openTypeIn.
	super paste.
	self closeTypeIn.! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
pasteInitials: aKeyboardEvent 
	"Replace the current text selection by an authorship name/date stamp; invoked by cmd-shift-v, easy way to put an authorship stamp in the comments of an editor."

	self insertAndCloseTypeIn.
	self replace: self selectionInterval with: (Text fromString: Utilities changeStamp) and: [self selectAt: self stopIndex].
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
pasteRecent
	"Paste an item chose from RecentClippings."

	| clipping |
	(clipping := Clipboard chooseRecentClipping) ifNil: [^ self].
	Clipboard clipboardText: clipping.
	^ self paste! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'mt 4/7/2015 10:54'!
pointBlock
	^ pointBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 5/27/2011 10:24'!
pointIndex
	^ pointBlock stringIndex! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
prettyPrint
	self prettyPrint: false! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
prettyPrintWithColor
	self prettyPrint: true! !

!TextEditor methodsFor: 'menu messages' stamp: 'nice 2/23/2011 21:42'!
prettyPrint: decorated 
	"Reformat the contents of the receiver's view (a Browser)."
	| selectedClass newText |
	selectedClass := model selectedClassOrMetaClass.
	selectedClass ifNil: [ ^ morph flash ].
	newText := selectedClass newCompiler
		format: self text
		in: selectedClass
		notifying: self
		decorated: decorated.
	newText ifNotNil:
		[ self
			deselect ;
			selectInvisiblyFrom: 1
			to: paragraph text size.
		self replaceSelectionWith: (newText asText makeSelectorBoldIn: selectedClass).
		self selectAt: self text size + 1 ]! !

!TextEditor methodsFor: 'do-its' stamp: 'mt 4/7/2015 13:17'!
printIt

	self evaluateSelectionAndDo: [:result |
		(model respondsTo: #printIt:result:)
			ifTrue: [model
				perform: #printIt:result:
				with: self selection
				with: result]
			ifFalse: [self afterSelectionInsertAndSelect: result printString]]! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
printIt: aKeyboardEvent
	"Print the results of evaluting the selection -- invoked via cmd-p.  If there is no current selection, use the current line.  1/17/96 sw
	 2/29/96 sw: don't call selectLine now, since it's called by doIt"

	self printIt.
	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'mt 8/6/2016 09:23'!
querySymbol: aKeyboardEvent
	"Invoked by Ctrl-q to query the Symbol table and display alternate symbols."

	| hintText lastOffering offering |
	self isTypingIn
		ifFalse: [
			self selectPrecedingIdentifier.
			hintText := self selection string]
		ifTrue: [
			self history current type = #query
				ifFalse: [
					self closeTypeIn.
					self selectPrecedingIdentifier.
					hintText := self selection string]
				ifTrue: [		
					self history hasPrevious
						ifFalse: [morph flash. self closeTypeIn. ^ true].
					hintText := self history previous contentsAfter string.
					hintText := hintText copyFrom: (hintText
						lastIndexOfAnyOf: Character separators, '#'
						startingAt: hintText size ifAbsent: [0])+1 to: hintText size.
					self selectInvisiblyFrom: self history current intervalBefore first to: self stopIndex-1.
					lastOffering := self selection string.
					lastOffering := (lastOffering copyReplaceAll: ':  ' with: ':') withBlanksTrimmed.]].
		
	offering := (Symbol thatStarts: hintText skipping: lastOffering) ifNil: [hintText].
	offering := offering copyReplaceAll: ':' with: ':  '.
	offering last = Character space ifTrue: [offering := offering allButLast].
	
	self openTypeInFor: #query.
	self typeAhead nextPutAll: offering.

	^ false! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'nice 8/3/2011 21:09'!
raiseContextMenu: aKeyboardEvent 
	(morph respondsTo: #editView)
		ifTrue: [morph editView yellowButtonActivity: false].
	^ true! !

!TextEditor methodsFor: 'current selection' stamp: 'jmv 5/30/2011 10:03'!
recomputeSelection
	"The same characters are selected but their coordinates may have changed."

	self markIndex: self markIndex pointIndex: self pointIndex! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 11/7/2015 23:37'!
redo

	self closeTypeIn.
	self history redoIn: self.! !

!TextEditor methodsFor: 'undo' stamp: 'mt 11/7/2015 23:38'!
redoAndReselect

	self
		replace: self history current intervalBefore
		with: self history current contentsAfter
		and: [self selectInterval: self history current intervalAfter].! !

!TextEditor methodsFor: 'editing keys' stamp: 'mt 11/7/2015 18:53'!
redo: aKeyboardEvent 
	"Redo the last edit."

	self insertAndCloseTypeIn.
	self redo.
	^true! !

!TextEditor methodsFor: 'menu messages' stamp: 'cmm 6/4/2011 16:06'!
referencesToIt
	"Open a MessageSet with the references to the selected global or variable name."
	| selection environment binding |
	self selection isEmpty ifTrue: [ self selectWord ].
	environment := (model respondsTo: #selectedClassOrMetaClass)
		ifTrue: [ model selectedClassOrMetaClass ifNil: [ Smalltalk globals ] ]
		ifFalse: [ Smalltalk globals ].
	selection := self selectedSymbol ifNil: [ self selection asString ].
	(environment isBehavior and:
		[ (environment
			instVarIndexFor: selection
			ifAbsent: [ 0 ]) ~= 0 ]) ifTrue: [ ^ self systemNavigation
			browseAllAccessesTo: selection
			from: environment ].
	selection isSymbol ifFalse: [ ^ morph flash ].
	binding := (environment bindingOf: selection) ifNil: [ ^ morph flash ].
	self systemNavigation browseAllCallsOn: binding! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
referencesToIt: aKeyboardEvent
	"Triggered by Cmd-N; browse references to the current selection"

	self referencesToIt.
	^ true! !

!TextEditor methodsFor: 'typing support' stamp: 'mt 12/21/2015 15:17'!
removePreviousInsertion
	"If the previous command was an insertion, remove it by not undoing it but replacing it with nothing."

	self history hasInsertedSomething ifFalse: [morph flash. ^ false].

	self
		selectInvisiblyFrom: self history previous intervalBefore first
		to: self history previous intervalAfter last.

	self replaceSelectionWith: self nullText.

	^ true
	! !

!TextEditor methodsFor: 'undo' stamp: 'mt 11/7/2015 18:50'!
replaceSelectionWith: aText
	"Remember the selection text in UndoSelection.
	 Deselect, and replace the selection text by aText.
	 Remember the resulting selectionInterval in UndoInterval and PriorInterval.
	 Set up undo to use UndoReplace."

	self openTypeIn.
	self zapSelectionWith: aText.
	self closeTypeIn.! !

!TextEditor methodsFor: 'undo' stamp: 'mt 11/7/2015 22:39'!
replace: interval with: newText

	self
		replace: interval
		with: newText
		and: ["Do nothing."].! !

!TextEditor methodsFor: 'undo' stamp: 'mt 11/7/2015 17:04'!
replace: xoldInterval with: newText and: selectingBlock 
	"Replace the text in oldInterval with newText and execute selectingBlock to establish the new selection.  Create an undoAndReselect:redoAndReselect: undoer to allow perfect undoing."

	| undoInterval |
	undoInterval := self selectionInterval.
	undoInterval = xoldInterval ifFalse: [self selectInterval: xoldInterval].
	
	self zapSelectionWith: newText.
	selectingBlock value.
	
	otherInterval := self selectionInterval.! !

!TextEditor methodsFor: 'initialize-release' stamp: 'mt 11/7/2015 23:41'!
resetState 
	"Establish the initial conditions for editing the paragraph: place caret 
	before first character, set the emphasis to that of the first character,
	and save the paragraph for purposes of canceling."

	pointBlock := markBlock := paragraph defaultCharacterBlock.
	beginTypeInIndex := nil.
	otherInterval := 1 to: 0.
	self setEmphasisHere.
	selectionShowing := false! !

!TextEditor methodsFor: 'private' stamp: 'cmm 8/16/2011 23:11'!
resetTypeAhead
	typeAhead := WriteStream on: (String new: 1)! !

!TextEditor methodsFor: 'current selection' stamp: 'nice 8/3/2011 22:39'!
reverseSelection
	"Reverse the valence of the current selection highlighting."
	selectionShowing := selectionShowing not.
	paragraph reverseFrom: pointBlock to: markBlock! !

!TextEditor methodsFor: 'private' stamp: 'jmv 12/8/2008 16:45'!
sameColumn: start newLine: lineBlock forward: isForward
	"Private - Compute the index in my text
	with the line number derived from lineBlock,"
	" a one argument block accepting the old line number.
	The position inside the line will be preserved as good as possible"
	"The boolean isForward is used in the border case to determine if
	we should move to the beginning or the end of the line."
	| wordStyle column currentLine offsetAtTargetLine targetEOL lines numberOfLines currentLineNumber targetLineNumber |
	wordStyle := Preferences wordStyleCursorMovement.
	wordStyle
		ifTrue: [
			lines := paragraph lines.
			numberOfLines := paragraph numberOfLines.
			currentLineNumber  := paragraph lineIndexOfCharacterIndex: start.
			currentLine := lines at: currentLineNumber]
		ifFalse: [
			lines := self lines.
			numberOfLines := lines size.
			currentLine := lines
				detect:[:lineInterval | lineInterval last >= start]
				ifNone:[lines last].
			currentLineNumber := currentLine second].
	column := start - currentLine first.
	targetLineNumber := ((lineBlock value: currentLineNumber) max: 1) min: numberOfLines.
	offsetAtTargetLine := (lines at: targetLineNumber) first.
	targetEOL := (lines at: targetLineNumber) last + (targetLineNumber = numberOfLines ifTrue:[1]ifFalse:[0]).
	targetLineNumber = currentLineNumber
	"No movement or movement failed. Move to beginning or end of line."
		ifTrue:[^isForward
			ifTrue:[targetEOL]
			ifFalse:[offsetAtTargetLine]].
	^offsetAtTargetLine + column min: targetEOL.! !

!TextEditor methodsFor: 'menu messages' stamp: 'nice 12/27/2009 03:12'!
saveContentsInFile
	"Save the receiver's contents string to a file, prompting the user for a file-name.  Suggest a reasonable file-name."

	| fileName stringToSave parentWindow labelToUse suggestedName |
	stringToSave := paragraph text string.
	stringToSave size = 0 ifTrue: [^self inform: 'nothing to save.'].
	parentWindow := model dependents 
				detect: [:dep | dep isKindOf: SystemWindow]
				ifNone: [nil].
	labelToUse := parentWindow ifNil: ['Untitled']
				ifNotNil: [parentWindow label].
	suggestedName := nil.
	#(#('Decompressed contents of: ' '.gz')) do: 
			[:leaderTrailer | | lastIndex | 
			"can add more here..."

			(labelToUse beginsWith: leaderTrailer first) 
				ifTrue: 
					[suggestedName := labelToUse copyFrom: leaderTrailer first size + 1
								to: labelToUse size.
					(labelToUse endsWith: leaderTrailer last) 
						ifTrue: 
							[suggestedName := suggestedName copyFrom: 1
										to: suggestedName size - leaderTrailer last size]
						ifFalse: 
							[lastIndex := suggestedName lastIndexOf: $. ifAbsent: [0].
							(lastIndex = 0 or: [lastIndex = 1]) 
								ifFalse: [suggestedName := suggestedName copyFrom: 1 to: lastIndex - 1]]]].
	suggestedName ifNil: [suggestedName := labelToUse , '.text'].
	fileName := UIManager default request: 'File name?'
				initialAnswer: suggestedName.
	fileName isEmptyOrNil 
		ifFalse: 
			[(FileStream newFileNamed: fileName)
				nextPutAll: stringToSave;
				close]! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
save: aKeyboardEvent
	"Submit the current text.  Equivalent to 'accept' 1/18/96 sw"

	self insertAndCloseTypeIn.
	self accept.
	^ true! !

!TextEditor methodsFor: 'scrolling' stamp: 'jmv 11/4/2008 11:51'!
scrollBy: ignore 
	"Ignore scroll requests."! !

!TextEditor methodsFor: 'current selection' stamp: 'jmv 11/4/2008 11:51'!
selectAndScroll
	"Ignore scroll requests."! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'mt 11/8/2015 16:22'!
selectCurrentTypeIn: aKeyboardEvent 
	"Select what would be replaced by an undo (e.g., the last typeIn)."

	| prior |

	self flag: #buggy.
	self insertAndCloseTypeIn.
	prior := otherInterval.
	self insertAndCloseTypeIn.
	otherInterval := prior.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
selectedSelector
	"Try to make a selector out of the current text selection"
	^self selection string findSelector! !

!TextEditor methodsFor: 'menu messages' stamp: 'nice 2/8/2010 10:48'!
selectedSymbol
	"Return the currently selected symbol, or nil if none.  Spaces, tabs and returns are ignored"

	| aString |
	self hasCaret ifTrue: [^ nil].
	aString := self selection string copyWithoutAll: CharacterSet separators.
	aString size = 0 ifTrue: [^ nil].
	Symbol hasInterned: aString  ifTrue: [:sym | ^ sym].

	^ nil! !

!TextEditor methodsFor: 'new selection' stamp: 'jmv 7/30/2009 20:17'!
selectFrom: start to: stop
	"Select the specified characters inclusive."
	self selectInvisiblyFrom: start to: stop.
	self closeTypeIn.
	self storeSelectionInParagraph.
	"Preserve current emphasis if selection is empty"
	stop > start ifTrue: [
		self setEmphasisHere ]! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
selection
	"Answer the text in the paragraph that is currently selected."

	^paragraph text copyFrom: self startIndex to: self stopIndex - 1 ! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'nice 7/22/2014 00:29'!
selectionAsStream
	"Answer a ReadStream on the text in the paragraph that is currently  selected."

	^ReadStream
		on: paragraph string
		from: self startIndex
		to: self stopIndex - 1! !

!TextEditor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 11:51'!
selectLine
	"Make the receiver's selection, if it currently consists of an insertion point only, encompass the current line."
	self hasSelection ifTrue:[^self].
	self selectInterval: (self encompassLine: self selectionInterval)! !

!TextEditor methodsFor: 'new selection' stamp: 'jmv 9/17/2010 12:10'!
selectPrecedingIdentifier
	"Invisibly select the identifier that ends at the end of the selection, if any."

	| string sep stop tok |
	tok := false.
	string := paragraph string.
	stop := self stopIndex - 1.
	[stop > 0 and: [(string at: stop) isSeparator]] whileTrue: [stop := stop - 1].
	sep := stop.
	[sep > 0 and: [(string at: sep) tokenish]] whileTrue: [tok := true. sep := sep - 1].
	tok ifTrue: [self selectInvisiblyFrom: sep + 1 to: stop]! !

!TextEditor methodsFor: 'menu messages' stamp: 'dtl 8/23/2015 21:20'!
sendContentsToPrinterWithLabel: label
	| textToPrint printer |
	textToPrint := paragraph text.
	textToPrint size = 0 ifTrue: [^self inform: 'nothing to print.'].
	printer := TextPrinter defaultTextPrinter.
	printer documentTitle: (label ifNil: ['Untitled']).
	printer printText: textToPrint! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 8/8/2009 15:32'!
sendersOfIt
	"Open a senders browser on the selected selector"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ morph flash].
	self systemNavigation browseAllCallsOn: aSelector! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
sendersOfIt: aKeyboardEvent
	"Triggered by Cmd-n; browse implementors of the selector represented by the current selection, if plausible. 2/1/96 sw"

	self sendersOfIt.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 7/29/2009 22:20'!
setAlignment: aSymbol
	| attr interval |
	attr := TextAlignment perform: aSymbol.
	interval := self encompassLine: self selectionInterval.
	paragraph 
		replaceFrom: interval first 
		to: interval last 
		with: ((paragraph text copyFrom: interval first to: interval last) addAttribute: attr)! !

!TextEditor methodsFor: 'typing support' stamp: 'jmv 8/8/2009 15:45'!
setEmphasisHere

	emphasisHere := (paragraph text attributesAt: (self pointIndex - 1 max: 1) forStyle: paragraph textStyle)
					select: [:att | att mayBeExtended]! !

!TextEditor methodsFor: 'typing support' stamp: 'jmv 9/1/2009 18:45'!
setEmphasisHereFromText

	self setEmphasisHereFromTextForward: true! !

!TextEditor methodsFor: 'typing support' stamp: 'nice 8/3/2011 21:51'!
setEmphasisHereFromTextForward: f

	| i t forward delta prevIsSeparator nextIsSeparator |
	i := self pointIndex.
	t := self text.
	"Try to set emphasisHere correctly after whitespace.
	Most important after a cr, i.e. at the start of a new line"
	prevIsSeparator :=  i > 1 and: [ (t at: i-1) isSeparator ].
	nextIsSeparator := i <= t size and: [ (t at: i) isSeparator ].
	forward := prevIsSeparator = nextIsSeparator
		ifTrue: [ f ]
		ifFalse: [ nextIsSeparator ].
	delta := forward ifTrue: [ 1 ] ifFalse: [ 0 ].
	emphasisHere := (t attributesAt: (i - delta max: 1))
					select: [:att | att mayBeExtended].! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 7/29/2009 18:37'!
setEmphasis: emphasisSymbol
	"Change the emphasis of the current selection."

	| oldAttributes attribute |
	oldAttributes := paragraph text attributesAt: self selectionInterval first.

	attribute := TextEmphasis perform: emphasisSymbol.
	(emphasisSymbol == #normal) 
		ifFalse:	[oldAttributes do:	
			[:att | (att dominates: attribute) ifTrue: [attribute turnOff]]].
	self replaceSelectionWith: (self selection addAttribute: attribute)! !

!TextEditor methodsFor: 'accessing' stamp: 'mt 11/26/2015 16:47'!
setReplacementFromHistory
	"Use history to get the previous replacement."
	
	self history hasReplacedSomething
		ifTrue: [ChangeText := self history previous contentsAfter].! !

!TextEditor methodsFor: 'accessing' stamp: 'mt 2/5/2016 15:05'!
setSearchFromSelectionOrHistory
	"Updates the current string to find with the current selection or the last change if it replaced something and thus had a prior selection."

	self hasSelection
		ifTrue: [FindText := self selection]
		ifFalse: [self history hasReplacedSomething
			ifTrue: [FindText := self history previous contentsBefore]
			ifFalse: [(self history hasInsertedSomething and: [(self history hasCursorMoved: self stopIndex) not])
				ifTrue: [
					FindText := self history previous contentsAfter.
					self removePreviousInsertion. "Undoable."]]]! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
setSearchString
	"Make the current selection, if any, be the current search string."
	self hasCaret ifTrue: [morph flash. ^ self].
	self setSearch:  self selection string! !

!TextEditor methodsFor: 'accessing' stamp: 'mt 11/18/2015 21:36'!
setSearch: aStringOrText

	FindText := aStringOrText.! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
spawn
	"Create and schedule a message browser for the code of the model's 
	selected message. Retain any edits that have not yet been accepted."
	| code |
	code := paragraph text string.
	self cancel.
	model spawn: code.! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 19:26'!
spawnIt: aKeyboardEvent
	"Triggered by Cmd-o; spawn a new code window, if it makes sense."

	self spawn.
	^ true! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 5/27/2011 10:24'!
startBlock
	^ pointBlock min: markBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
startIndex
	^ self startBlock stringIndex! !

!TextEditor methodsFor: 'typing support' stamp: 'nice 8/3/2011 21:28'!
startOfTyping
	"Compatibility during change from characterBlock to integer"
	beginTypeInIndex == nil ifTrue: [^ nil].
	beginTypeInIndex isNumber ifTrue: [^ beginTypeInIndex].
	"Last line for compatibility during change from CharacterBlock to Integer."
	^ beginTypeInIndex stringIndex! !

!TextEditor methodsFor: 'initialize-release' stamp: 'mt 11/8/2015 11:47'!
stateArray
	^ {ChangeText.
		FindText.
		history ifNil: [TextEditorCommandHistory new]. "Convert old instances"
		self selectionInterval.
		self startOfTyping.
		emphasisHere}! !

!TextEditor methodsFor: 'initialize-release' stamp: 'mt 11/7/2015 23:24'!
stateArrayPut: stateArray
	| sel |
	ChangeText := stateArray at: 1.
	FindText := stateArray at: 2.
	history := stateArray at: 3.
	sel := stateArray at: 4.
	self selectFrom: sel first to: sel last.
	beginTypeInIndex := stateArray at: 5.
	emphasisHere := stateArray at: 6! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 5/27/2011 10:24'!
stopBlock
	^ pointBlock max: markBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
stopIndex
	^ self stopBlock stringIndex! !

!TextEditor methodsFor: 'mvc compatibility' stamp: 'jmv 11/4/2008 11:51'!
storeSelectionInParagraph
	paragraph selectionStart: self startBlock selectionStop: self stopBlock! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 23:01'!
string

	^self text string! !

!TextEditor methodsFor: 'accessing' stamp: 'ar 8/17/2009 23:56'!
styler
	"Answers the styler for this editor. Only code editors support syntax highlighting"
	^nil! !

!TextEditor methodsFor: 'accessing' stamp: 'ar 8/17/2009 23:56'!
styler: aStyler
	"Sets the styler for this editor. Only code editors support syntax highlighting"
	^nil! !

!TextEditor methodsFor: 'editing keys' stamp: 'nice 8/3/2011 22:47'!
swapChars: aKeyboardEvent 
	"Triggered byCmd-Y;.  Swap two characters, either those straddling the insertion point, or the two that comprise the selection.  Suggested by Ted Kaehler.  "

	| currentSelection aString chars |
	(chars := self selection) size = 0
		ifTrue:
			[currentSelection := self pointIndex.
			self selectMark: currentSelection - 1 point: currentSelection]
		ifFalse:
			[chars size = 2
				ifFalse:
					[morph flash. ^ true]
				ifTrue:
					[currentSelection := self pointIndex - 1]].
	aString := self selection string.
	self replaceSelectionWith: (Text string: aString reversed attributes: emphasisHere).
	self selectAt: currentSelection + 1.
	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'nice 2/8/2010 10:57'!
tabCount
	^ self class autoIndent
		ifTrue:
			[ | tabCount s i char |
			s := paragraph string.
			i := self stopIndex.
			tabCount := 0.
			[(i := i-1) > 0 and: [(char := s at: i) ~= Character cr and: [char ~= Character lf]]]
				whileTrue:  "Count tabs and brackets (but not a leading bracket)"
				[(char = Character tab and: [i < s size and: [(s at: i+1) ~= $[ ]]) ifTrue: [tabCount := tabCount + 1].
				char = $[ ifTrue: [tabCount := tabCount + 1].
				char = $] ifTrue: [tabCount := tabCount - 1]].
			tabCount ]
		ifFalse: [ 0 ]! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 11:51'!
text
	"Answer the text of the paragraph being edited."

	^paragraph text! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 11:51'!
totalTextHeight

	^paragraph lines last bottom! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 11:51'!
transformFrom: owner
	^morph transformFrom: owner! !

!TextEditor methodsFor: 'private' stamp: 'cmm 8/16/2011 22:04'!
typeAhead
	^ typeAhead ifNil:
		[ self resetTypeAhead.
		typeAhead ]! !

!TextEditor methodsFor: 'private' stamp: 'nice 8/3/2011 22:52'!
unapplyAttribute: aTextAttribute
	"The user selected aTextAttribute to be removed.
	If there is a selection, unapply the attribute to the selection.
	In any case do not use the attribute for the user input (emphasisHere)"

	| interval |

	emphasisHere := emphasisHere copyWithout: aTextAttribute.

	interval := self selectionInterval.
	(interval isEmpty and: [ aTextAttribute isParagraphAttribute not ])
		ifTrue: [ ^self ].
	
	self text removeAttribute: aTextAttribute from: interval first to: interval last.
	paragraph recomposeFrom: interval first to: interval last delta: 0.
	self recomputeSelection.	"Needed so visible selection is updated to reflect new visual extent of selection"
	morph changed! !

!TextEditor methodsFor: 'menu messages' stamp: 'mt 11/10/2015 14:01'!
undo

	self closeTypeIn.
	self history undoIn: self.
	self history hasPrevious ifFalse: [morph hasUnacceptedEdits: false].! !

!TextEditor methodsFor: 'undo' stamp: 'mt 11/7/2015 23:38'!
undoAndReselect

	self
		replace: self history current intervalBetween
		with: self history current contentsBefore
		and: [self selectInterval: self history current intervalBefore].! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:45'!
undo: aKeyboardEvent 
	"Undo the last edit."

	self insertAndCloseTypeIn.
	self undo.
	^true! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'cmm 8/16/2011 22:29'!
unselect
	markBlock := pointBlock copy! !

!TextEditor methodsFor: 'scrolling' stamp: 'jmv 11/4/2008 11:51'!
updateMarker
	"Ignore scrollbar redraw requests."
! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 11:51'!
visibleHeight

	^morph owner bounds height! !

!TextEditor methodsFor: 'events' stamp: 'ar 5/5/2004 19:10'!
yellowButtonDown: event
	"Process a yellow button event. Answer true if the event was handled, false otherwise."
	(paragraph attributesAt: event cursorPoint) do:[:attr|
		attr menu ifNotNil:[
			attr menu openAt: event cursorPoint.
			^true]].
	^false! !

!TextEditor methodsFor: 'mvc compatibility' stamp: 'mt 11/7/2015 16:38'!
zapSelectionWith: replacement

	| start stop rep |
	morph readOnly ifTrue: [^ self].
	self deselect.
	start := self startIndex.
	stop := self stopIndex.
	(replacement isEmpty and: [stop > start]) ifTrue: [
		"If deleting, then set emphasisHere from 1st character of the deletion"
		emphasisHere := (self text attributesAt: start) select: [:att | att mayBeExtended]].
	(start = stop and: [ replacement isEmpty ]) ifFalse: [
		replacement isText
			ifTrue: [ rep := replacement]
			ifFalse: [ rep := Text string: replacement attributes: emphasisHere ].
		self text replaceFrom: start to: stop - 1 with: rep.
		paragraph
			recomposeFrom: start
			to:  start + rep size - 1
			delta: rep size - (stop-start).
		self markIndex: start pointIndex: start + rep size.
		otherInterval := self selectionInterval].

	self userHasEdited  " -- note text now dirty"! !

Object subclass: #EllipseMidpointTracer
	instanceVariableNames: 'rect x y a b aSquared bSquared d1 d2 inFirstRegion'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!EllipseMidpointTracer methodsFor: 'initialize' stamp: 'ar 6/28/1999 15:33'!
on: aRectangle
	rect := aRectangle.
	a := rect width // 2.
	b := rect height // 2.
	x := 0.
	y := b.
	aSquared := a * a.
	bSquared := b * b.
	d1 := bSquared - (aSquared * b) + (0.25 * aSquared).
	d2 := nil.
	inFirstRegion := true.! !

!EllipseMidpointTracer methodsFor: 'computing' stamp: 'ar 6/28/1999 15:35'!
stepInY
	"Step to the next y value"
	inFirstRegion ifTrue:[
		"In the upper region we must step until we reach the next y value"
		[(aSquared * (y-0.5)) > (bSquared * (x+1))] whileTrue:[
			d1 < 0.0
				ifTrue:[d1 := d1 + (bSquared * (2*x+3)).
						x := x + 1]
				ifFalse:[d1 := d1 + (bSquared * (2*x+3)) + (aSquared * (-2*y+2)).
						y := y - 1.
						^x := x + 1]].
		"Stepping into second region"
		d2 := (bSquared * (x + 0.5) squared) + (aSquared * (y-1) squared) - (aSquared * bSquared).
		inFirstRegion := false.
	].
	"In the lower region each step is a y-step"
	d2 < 0.0
		ifTrue:[d2 := d2 + (bSquared * (2*x+2)) + (aSquared * (-2*y+3)).
				x := x + 1]
		ifFalse:[d2 := d2 + (aSquared * (-2*y+3))].
	y := y - 1.
	^x! !

Object subclass: #EventHandler
	instanceVariableNames: 'mouseDownRecipient mouseDownSelector mouseMoveRecipient mouseMoveSelector mouseStillDownRecipient mouseStillDownSelector mouseUpRecipient mouseUpSelector mouseEnterRecipient mouseEnterSelector mouseLeaveRecipient mouseLeaveSelector mouseEnterDraggingRecipient mouseEnterDraggingSelector mouseLeaveDraggingRecipient mouseLeaveDraggingSelector mouseWheelRecipient mouseWheelSelector keyStrokeRecipient keyStrokeSelector keyUpRecipient keyUpSelector keyDownRecipient keyDownSelector valueParameter startDragRecipient startDragSelector doubleClickSelector doubleClickRecipient doubleClickTimeoutSelector doubleClickTimeoutRecipient clickSelector clickRecipient keyboardFocusChangeRecipient keyboardFocusChangeSelector wantsEveryMouseMove'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!EventHandler commentStamp: '<historical>' prior: 0!
Events in Morphic originate in a Hand, pass to a target morph, and are then dispatched by an EventHandler.  EventHandlers support redirection of mouse and keyboard activity by specifying and independent recipient object and message selector for each of the possible events.  In addition each eventHandler can supply an optional value parameter for distinguishing between, eg, events from a number of otherwise identical source morphs.

The basic protocol of an event handler is to receive a message of the form
	mouseDown: event in: targetMorph
and redirect this as one of
	mouseDownRecipient perform: mouseDownSelector0
	mouseDownRecipient perform: mouseDownSelector1 with: event
	mouseDownRecipient perform: mouseDownSelector2 with: event with: targetMorph
	mouseDownRecipient perform: mouseDownSelector3 with: event with: targetMorph with: valueParameter
depending on the arity of the mouseDownSelector.
!

!EventHandler methodsFor: 'access' stamp: 'ar 10/25/2000 17:33'!
allRecipients
	"Answer a list, without duplication, of all the objects serving as recipients to any of the events I handle.  Intended for debugging/documentation use only"
	| aList |
	aList := OrderedCollection with: mouseDownRecipient with: mouseStillDownRecipient with: mouseUpRecipient with: mouseEnterRecipient with: mouseLeaveRecipient.
	aList addAll: (OrderedCollection with:  mouseEnterDraggingRecipient with: mouseLeaveDraggingRecipient with: doubleClickRecipient with: keyStrokeRecipient).
	aList add: mouseMoveRecipient.
	^ (aList copyWithout: nil) asSet asArray! !

!EventHandler methodsFor: 'events' stamp: 'ar 10/7/2000 22:55'!
click: event fromMorph: sourceMorph 
	"This message is sent only when double clicks are handled."
	^ self
		send: clickSelector
		to: clickRecipient
		withEvent: event
		fromMorph: sourceMorph! !

!EventHandler methodsFor: 'objects from disk' stamp: 'RAA 12/20/2000 17:45'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	"20 dec 2000 - only a few (old) conversion exists"

	varDict at: 'mouseEnterLadenRecipient' ifPresent: [ :x | mouseEnterDraggingRecipient := x].
	varDict at: 'mouseEnterLadenSelector' ifPresent: [ :x | mouseEnterDraggingSelector := x].
	varDict at: 'mouseLeaveLadenRecipient' ifPresent: [ :x | mouseLeaveDraggingRecipient := x].
	varDict at: 'mouseLeaveLadenSelector' ifPresent: [ :x | mouseLeaveDraggingSelector := x].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!EventHandler methodsFor: 'events' stamp: 'jcg 9/21/2001 13:06'!
doubleClickTimeout: event fromMorph: sourceMorph 
	^ self
		send: doubleClickTimeoutSelector
		to: doubleClickTimeoutRecipient
		withEvent: event
		fromMorph: sourceMorph! !

!EventHandler methodsFor: 'events' stamp: 'LC 2/14/2000 08:38'!
doubleClick: event fromMorph: sourceMorph 
	^ self
		send: doubleClickSelector
		to: doubleClickRecipient
		withEvent: event
		fromMorph: sourceMorph! !

!EventHandler methodsFor: 'access' stamp: 'ar 10/25/2000 17:34'!
firstMouseSelector
	"Answer the selector corresponding to the first mouse-handling selector fielded.  Created in support of providing balloon-help for halo handles, triggered by the selector handled"

	mouseDownSelector ifNotNil: [^ mouseDownSelector].
	mouseMoveSelector ifNotNil:[^mouseMoveSelector].
	mouseStillDownSelector ifNotNil: [^ mouseStillDownSelector].
	mouseUpSelector ifNotNil: [^ mouseUpSelector].
	mouseEnterSelector ifNotNil: [^ mouseEnterSelector].
	mouseLeaveSelector ifNotNil: [^ mouseLeaveSelector].
	mouseEnterDraggingSelector ifNotNil: [^ mouseEnterDraggingSelector].
	mouseLeaveDraggingSelector ifNotNil: [^ mouseLeaveDraggingSelector].
	doubleClickSelector ifNotNil: [^ doubleClickSelector].
	^ nil! !

!EventHandler methodsFor: 'fixups' stamp: 'sw 3/28/2001 14:22'!
fixReversedValueMessages
	"ar 3/18/2001: Due to the change in the ordering of the value parameter old event handlers may have messages that need to be fixed up. Do this here."

	self replaceSendsIn: #( renameCharAction:sourceMorph:requestor: makeGetter:from:forPart: makeSetter:from:forPart: newMakeGetter:from:forPart: newMakeSetter:from:forPart: clickOnLine:evt:envelope: limitHandleMoveEvent:from:index: mouseUpEvent:linkMorph:formData: mouseUpEvent:linkMorph:browserAndUrl: mouseDownEvent:noteMorph:pitch: mouseMoveEvent:noteMorph:pitch: mouseUpEvent:noteMorph:pitch: dragVertex:fromHandle:vertIndex: dropVertex:fromHandle:vertIndex: newVertex:fromHandle:afterVert: prefMenu:rcvr:pref: event:arrow:upDown:
newMakeGetter:from:forMethodInterface:)
			with: #( renameCharAction:event:sourceMorph: makeGetter:event:from: makeSetter:event:from: newMakeGetter:event:from: newMakeSetter:event:from: clickOn:evt:from: limitHandleMove:event:from: mouseUpFormData:event:linkMorph: mouseUpBrowserAndUrl:event:linkMorph: mouseDownPitch:event:noteMorph: mouseMovePitch:event:noteMorph: mouseUpPitch:event:noteMorph: dragVertex:event:fromHandle: dropVertex:event:fromHandle: newVertex:event:fromHandle: prefMenu:event:rcvr: upDown:event:arrow: makeUniversalTilesGetter:event:from:).

"sw 3/28/2001 extended Andreas's original lists by one item"! !

!EventHandler methodsFor: 'testing' stamp: 'ar 10/7/2000 22:56'!
handlesClickOrDrag: evt
	clickRecipient ifNotNil:[^true].
	doubleClickRecipient ifNotNil:[^true].
	startDragRecipient ifNotNil:[^true].
	^false! !

!EventHandler methodsFor: 'testing' stamp: 'nk 2/15/2004 08:57'!
handlesGestureStart: evt
	"Does the associated morph want to handle gestures?"
	^false! !

!EventHandler methodsFor: 'testing' stamp: 'ar 10/28/2000 22:17'!
handlesKeyboard: evt
	keyStrokeRecipient ifNotNil: [^ true].
	^ false! !

!EventHandler methodsFor: 'testing' stamp: 'nk 2/15/2004 08:13'!
handlesMouseDown: evt
	mouseDownRecipient ifNotNil: [^ true].
	mouseStillDownRecipient ifNotNil: [^ true].
	mouseUpRecipient ifNotNil: [^ true].
	(self handlesClickOrDrag: evt) ifTrue:[^true].
	^self handlesGestureStart: evt! !

!EventHandler methodsFor: 'testing' stamp: 'mt 6/2/2015 09:59'!
handlesMouseMove: evt
	
	^ self wantsEveryMouseMove == true
		or: [((((mouseMoveRecipient notNil and: [mouseMoveSelector notNil])
			and: [evt hand hasSubmorphs not])
			and: [evt anyButtonPressed])
			and: [evt hand mouseFocus notNil])
			and: [evt hand mouseFocus eventHandler == self]]! !

!EventHandler methodsFor: 'testing' stamp: 'di 9/15/1998 16:35'!
handlesMouseOverDragging: evt
	mouseEnterDraggingRecipient ifNotNil: [^ true].
	mouseLeaveDraggingRecipient ifNotNil: [^ true].
	^ false! !

!EventHandler methodsFor: 'testing' stamp: ''!
handlesMouseOver: evt
	mouseEnterRecipient ifNotNil: [^ true].
	mouseLeaveRecipient ifNotNil: [^ true].
	^ false! !

!EventHandler methodsFor: 'testing' stamp: 'ar 10/22/2000 17:05'!
handlesMouseStillDown: evt
	^mouseStillDownRecipient notNil and:[mouseStillDownSelector notNil]! !

!EventHandler methodsFor: 'testing' stamp: 'mt 6/6/2016 17:04'!
handlesMouseWheel: evt
	mouseWheelRecipient ifNotNil: [^ true].
	^ false! !

!EventHandler methodsFor: 'events' stamp: 'mt 4/19/2015 12:00'!
keyboardFocusChange: aBoolean fromMorph: sourceMorph

	^ self
		send: keyboardFocusChangeSelector
		to: keyboardFocusChangeRecipient
		withEvent: aBoolean
		fromMorph: sourceMorph! !

!EventHandler methodsFor: 'events' stamp: 'mt 4/19/2015 11:52'!
keyDown: event fromMorph: sourceMorph
	^ self send: keyDownSelector to: keyDownRecipient withEvent: event fromMorph: sourceMorph! !

!EventHandler methodsFor: 'events' stamp: ''!
keyStroke: event fromMorph: sourceMorph
	^ self send: keyStrokeSelector to: keyStrokeRecipient withEvent: event fromMorph: sourceMorph! !

!EventHandler methodsFor: 'events' stamp: 'mt 4/19/2015 11:52'!
keyUp: event fromMorph: sourceMorph
	^ self send: keyUpSelector to: keyUpRecipient withEvent: event fromMorph: sourceMorph! !

!EventHandler methodsFor: 'access' stamp: ''!
messageList
	"Return a list of 'Class selector' for each message I can send. tk 
	9/13/97"
	| list |
	self flag: #mref.
	"is this still needed? I replaced the one use that I could spot with 
	#methodRefList "
	list := SortedCollection new.
	mouseDownRecipient
		ifNotNil: [list add: (mouseDownRecipient class whichClassIncludesSelector: mouseDownSelector) name , ' ' , mouseDownSelector].
	mouseMoveRecipient
		ifNotNil: [list add: (mouseMoveRecipient class whichClassIncludesSelector: mouseMoveSelector) name , ' ' , mouseMoveSelector].
	mouseStillDownRecipient
		ifNotNil: [list add: (mouseStillDownRecipient class whichClassIncludesSelector: mouseStillDownSelector) name , ' ' , mouseStillDownSelector].
	mouseUpRecipient
		ifNotNil: [list add: (mouseUpRecipient class whichClassIncludesSelector: mouseUpSelector) name , ' ' , mouseUpSelector].
	mouseEnterRecipient
		ifNotNil: [list add: (mouseEnterRecipient class whichClassIncludesSelector: mouseEnterSelector) name , ' ' , mouseEnterSelector].
	mouseLeaveRecipient
		ifNotNil: [list add: (mouseLeaveRecipient class whichClassIncludesSelector: mouseLeaveSelector) name , ' ' , mouseLeaveSelector].
	mouseEnterDraggingRecipient
		ifNotNil: [list add: (mouseEnterDraggingRecipient class whichClassIncludesSelector: mouseEnterDraggingSelector) name , ' ' , mouseEnterDraggingSelector].
	mouseLeaveDraggingRecipient
		ifNotNil: [list add: (mouseLeaveDraggingRecipient class whichClassIncludesSelector: mouseLeaveDraggingSelector) name , ' ' , mouseLeaveDraggingSelector].
	doubleClickRecipient
		ifNotNil: [list add: (doubleClickRecipient class whichClassIncludesSelector: doubleClickSelector) name , ' ' , doubleClickSelector].
	keyStrokeRecipient
		ifNotNil: [list add: (keyStrokeRecipient class whichClassIncludesSelector: keyStrokeSelector) name , ' ' , keyStrokeSelector].
	^ list! !

!EventHandler methodsFor: 'access' stamp: 'fbs 5/21/2013 22:28'!
methodRefList
	"Return a MethodReference for each message I can send. tk 9/13/97, raa 
	5/29/01 "
	| list adder |
	list := SortedCollection new.
	adder := [:recip :sel | recip
				ifNotNil: [list
						add: (MethodReference new
								class: (recip class whichClassIncludesSelector: sel)
								selector: sel)]].
	adder value: mouseDownRecipient value: mouseDownSelector.
	adder value: mouseMoveRecipient value: mouseMoveSelector.
	adder value: mouseStillDownRecipient value: mouseStillDownSelector.
	adder value: mouseUpRecipient value: mouseUpSelector.
	adder value: mouseEnterRecipient value: mouseEnterSelector.
	adder value: mouseLeaveRecipient value: mouseLeaveSelector.
	adder value: mouseEnterDraggingRecipient value: mouseEnterDraggingSelector.
	adder value: mouseLeaveDraggingRecipient value: mouseLeaveDraggingSelector.
	adder value: doubleClickRecipient value: doubleClickSelector.
	adder value: keyStrokeRecipient value: keyStrokeSelector.
	^ list! !

!EventHandler methodsFor: 'access' stamp: 'di 9/14/1998 08:32'!
mouseDownSelector
	^ mouseDownSelector! !

!EventHandler methodsFor: 'events' stamp: 'ar 10/7/2000 22:54'!
mouseDown: event fromMorph: sourceMorph 
	"Take double-clicks into account."
	((self handlesClickOrDrag: event) and:[event redButtonPressed]) ifTrue:[
		event hand waitForClicksOrDrag: sourceMorph event: event.
	].
	^self
		send: mouseDownSelector
		to: mouseDownRecipient
		withEvent: event
		fromMorph: sourceMorph.
! !

!EventHandler methodsFor: 'events' stamp: 'di 9/15/1998 16:35'!
mouseEnterDragging: event fromMorph: sourceMorph
	^ self send: mouseEnterDraggingSelector to: mouseEnterDraggingRecipient withEvent: event fromMorph: sourceMorph! !

!EventHandler methodsFor: 'events' stamp: ''!
mouseEnter: event fromMorph: sourceMorph
	^ self send: mouseEnterSelector to: mouseEnterRecipient withEvent: event fromMorph: sourceMorph! !

!EventHandler methodsFor: 'events' stamp: 'di 9/15/1998 16:35'!
mouseLeaveDragging: event fromMorph: sourceMorph
	^ self send: mouseLeaveDraggingSelector to: mouseLeaveDraggingRecipient withEvent: event fromMorph: sourceMorph! !

!EventHandler methodsFor: 'events' stamp: ''!
mouseLeave: event fromMorph: sourceMorph
	^ self send: mouseLeaveSelector to: mouseLeaveRecipient withEvent: event fromMorph: sourceMorph! !

!EventHandler methodsFor: 'events' stamp: 'ar 10/25/2000 17:32'!
mouseMove: event fromMorph: sourceMorph
	^ self send: mouseMoveSelector to: mouseMoveRecipient withEvent: event fromMorph: sourceMorph! !

!EventHandler methodsFor: 'access' stamp: 'ar 10/25/2000 18:27'!
mouseStillDownRecipient
	^mouseStillDownRecipient! !

!EventHandler methodsFor: 'access' stamp: 'ar 10/25/2000 18:27'!
mouseStillDownSelector
	^mouseStillDownSelector! !

!EventHandler methodsFor: 'events' stamp: ''!
mouseStillDown: event fromMorph: sourceMorph
	^ self send: mouseStillDownSelector to: mouseStillDownRecipient withEvent: event fromMorph: sourceMorph! !

!EventHandler methodsFor: 'access' stamp: 'di 9/14/1998 08:32'!
mouseUpSelector
	^ mouseUpSelector! !

!EventHandler methodsFor: 'events' stamp: ''!
mouseUp: event fromMorph: sourceMorph
	^ self send: mouseUpSelector to: mouseUpRecipient withEvent: event fromMorph: sourceMorph! !

!EventHandler methodsFor: 'events' stamp: 'mt 6/6/2016 17:08'!
mouseWheel: event fromMorph: sourceMorph 
	^ self
		send: mouseWheelSelector
		to: mouseWheelRecipient
		withEvent: event
		fromMorph: sourceMorph! !

!EventHandler methodsFor: 'initialization' stamp: 'nk 2/15/2004 08:59'!
onGestureSend: selector to: recipient! !

!EventHandler methodsFor: 'initialization' stamp: 'mt 6/6/2016 17:07'!
on: eventName send: selector to: recipient
	eventName == #mouseDown ifTrue:
		[mouseDownRecipient := recipient.  mouseDownSelector := selector. ^ self].
	eventName == #mouseMove ifTrue:
		[mouseMoveRecipient := recipient.  mouseMoveSelector := selector. ^ self].
	eventName == #mouseStillDown ifTrue:
		[mouseStillDownRecipient := recipient.  mouseStillDownSelector := selector. ^ self].
	eventName == #mouseUp ifTrue:
		[mouseUpRecipient := recipient.  mouseUpSelector := selector. ^ self].
	eventName == #mouseEnter ifTrue:
		[mouseEnterRecipient := recipient.  mouseEnterSelector := selector. ^ self].
	eventName == #mouseLeave ifTrue:
		[mouseLeaveRecipient := recipient.  mouseLeaveSelector := selector. ^ self].
	eventName == #mouseEnterDragging ifTrue:
		[mouseEnterDraggingRecipient := recipient.  mouseEnterDraggingSelector := selector. ^ self].
	eventName == #mouseLeaveDragging ifTrue:
		[mouseLeaveDraggingRecipient := recipient.  mouseLeaveDraggingSelector := selector. ^ self].
	eventName == #click ifTrue:
		[clickRecipient := recipient. clickSelector := selector. ^ self].
	eventName == #doubleClick ifTrue:
		[doubleClickRecipient := recipient. doubleClickSelector := selector. ^ self].
	eventName == #doubleClickTimeout ifTrue:
		[doubleClickTimeoutRecipient := recipient. doubleClickTimeoutSelector := selector. ^ self].
	eventName == #startDrag ifTrue:
		[startDragRecipient := recipient. startDragSelector := selector. ^ self].
	eventName == #mouseWheel ifTrue:
		[mouseWheelRecipient := recipient. mouseWheelSelector := selector. ^ self].
	eventName == #keyStroke ifTrue:
		[keyStrokeRecipient := recipient.  keyStrokeSelector := selector. ^ self].
	eventName == #keyUp ifTrue:
		[keyUpRecipient := recipient.  keyUpSelector := selector. ^ self].
	eventName == #keyDown ifTrue:
		[keyDownRecipient := recipient.  keyDownSelector := selector. ^ self].
	eventName == #keyboardFocusChange ifTrue:
		[keyboardFocusChangeRecipient := recipient. keyboardFocusChangeSelector := selector. ^ self].
	eventName == #gesture ifTrue:
		[ ^self onGestureSend: selector to: recipient ].
	self error: 'Event name, ' , eventName , ' is not recognizable.'
! !

!EventHandler methodsFor: 'initialization' stamp: 'wiz 8/21/2005 01:44'!
on: eventName send: selector to: recipient withValue: value
	selector numArgs = 3 ifFalse:
		[self halt: 'Warning: value parameters are passed as first of 3 arguments'].
	self on: eventName send: selector to: recipient.
	valueParameter := value
! !

!EventHandler methodsFor: 'printing' stamp: 'mt 6/6/2016 17:06'!
printOn: aStream 
	| recipients |
	super printOn: aStream.
	#('mouseDownSelector' 'mouseStillDownSelector' 'mouseUpSelector' 'mouseEnterSelector' 'mouseLeaveSelector' 'mouseEnterDraggingSelector' 'mouseLeaveDraggingSelector' 'mouseWheelSelector' 'doubleClickSelector' 'keyStrokeSelector' 'keyUpSelector' 'keyDownSelector' 'startDragSelector' 'clickSelector' 'keyboardFocusChangeSelector') 
		do: 
			[:aName | | aVal | 
			(aVal := self instVarNamed: aName) notNil 
				ifTrue: [aStream nextPutAll: '; ' , aName , '=' , aVal]].
	(recipients := self allRecipients) notEmpty 
		ifTrue: 
			[aStream nextPutAll: ' recipients: '.
			recipients printOn: aStream]! !

!EventHandler methodsFor: 'fixups' stamp: 'ar 3/18/2001 17:18'!
replaceSendsIn: array1 with: array2
	"Replace all the sends that occur in array1 with those in array2. Used for fixing old event handlers in files."
	| old index |
	1 to: self class instSize do:[:i|
		old := self instVarAt: i.
		index := array1 identityIndexOf: old.
		index > 0 ifTrue:[self instVarAt: i put: (array2 at: index)]].! !

!EventHandler methodsFor: 'events' stamp: 'ar 3/17/2001 14:34'!
send: selector to: recipient withEvent: event fromMorph: sourceMorph
	| arity |
	recipient ifNil: [^ self].
	arity := selector numArgs.
	arity = 0 ifTrue:
		[^ recipient perform: selector].
	arity = 1 ifTrue:
		[^ recipient perform: selector with: event].
	arity = 2 ifTrue:
		[^ recipient perform: selector with: event with: sourceMorph].
	arity = 3 ifTrue:
		[^ recipient perform: selector with: valueParameter with: event with: sourceMorph].
	self error: 'Event handling selectors must be Symbols and take 0-3 arguments'! !

!EventHandler methodsFor: 'events' stamp: 'mir 5/23/2000 17:43'!
startDrag: event fromMorph: sourceMorph 
	^ self
		send: startDragSelector
		to: startDragRecipient
		withEvent: event
		fromMorph: sourceMorph! !

!EventHandler methodsFor: 'copying' stamp: 'tk 1/22/2001 17:43'!
veryDeepFixupWith: deepCopier
	| old |
	"ALL inst vars were weakly copied.  If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

super veryDeepFixupWith: deepCopier.
1 to: self class instSize do:
	[:ii | old := self instVarAt: ii.
	self instVarAt: ii put: (deepCopier references at: old ifAbsent: [old])].

! !

!EventHandler methodsFor: 'copying' stamp: 'nk 2/14/2004 18:24'!
veryDeepInner: deepCopier
	"ALL fields are weakly copied!!  Can't duplicate an object by duplicating a button that activates it.  See DeepCopier."

	super veryDeepInner: deepCopier.
	"just keep old pointers to all fields"
! !

!EventHandler methodsFor: 'access' stamp: 'mt 6/2/2015 09:51'!
wantsEveryMouseMove

	^ wantsEveryMouseMove! !

!EventHandler methodsFor: 'access' stamp: 'mt 6/2/2015 09:51'!
wantsEveryMouseMove: aBoolean

	wantsEveryMouseMove := aBoolean.! !

Object subclass: #FontImporterFontDescription
	instanceVariableNames: 'fontname filename children parent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!FontImporterFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 12:37'!
addChild: aChild

	^ self children add: aChild! !

!FontImporterFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 18:52'!
allFilenames

	^ self filename
		ifNil: [
			(self children
				select: [:child | child filename notNil]
				thenCollect: [:child | child filename])
			asSet asArray]
		ifNotNil: [:f | {f}] ! !

!FontImporterFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 12:36'!
children

	^ children ifNil: [children := OrderedCollection new].! !

!FontImporterFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 12:26'!
children: anObject

	children := anObject! !

!FontImporterFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 18:41'!
filename

	^ filename! !

!FontImporterFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 18:41'!
filename: anObject

	filename := anObject! !

!FontImporterFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 12:46'!
fontname

	^ fontname! !

!FontImporterFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 12:46'!
fontname: anObject

	fontname := anObject! !

!FontImporterFontDescription methodsFor: 'testing' stamp: 'topa 3/9/2015 12:29'!
hasChildren

	^ self children notNil and: [self children notEmpty]! !

!FontImporterFontDescription methodsFor: 'actions' stamp: 'topa 3/9/2015 18:46'!
normalize

	self children size = 1 ifTrue: [ | pseudoChild |
		pseudoChild := self children removeFirst.
		(self filename notNil and: [pseudoChild filename ~=  self filename])
			ifTrue: [self error: 'Inconsistent state'].
		self filename: pseudoChild filename]! !

!FontImporterFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 12:26'!
parent

	^ parent! !

!FontImporterFontDescription methodsFor: 'accessing' stamp: 'topa 3/9/2015 12:26'!
parent: anObject

	parent := anObject! !

!FontImporterFontDescription methodsFor: 'printing' stamp: 'topa 3/9/2015 18:39'!
printOn: aStream

	self parent ifNotNil: [:p | aStream nextPutAll: p fontname; nextPut: $ ].
	aStream nextPutAll: self fontname.
	self children notEmpty ifTrue: [aStream nextPut: $ ].
	self children
		do: [:subfont | aStream nextPutAll: subfont fontname]
		separatedBy: [aStream nextPut: $/].
	aStream nextPut: $ ; nextPut: $(.
	self allFilenames
		do: [:filename | aStream nextPutAll: filename]
		separatedBy: [aStream nextPut: $,; nextPut: $ ].
	aStream nextPut: $).
! !

!FontImporterFontDescription methodsFor: 'comparing' stamp: 'topa 3/19/2015 10:45'!
<= other

	^ self fontname asString <= other fontname asString! !

Object subclass: #HaloSpec
	instanceVariableNames: 'addHandleSelector horizontalPlacement verticalPlacement color iconSymbol'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!HaloSpec commentStamp: 'kfr 10/27/2003 16:23' prior: 0!
Sets spec's for how handles are layed out in a halo.!

!HaloSpec methodsFor: 'as yet unclassified' stamp: 'sw 1/25/2000 19:54'!
addHandleSelector
	^ addHandleSelector! !

!HaloSpec methodsFor: 'as yet unclassified' stamp: 'sw 1/25/2000 18:41'!
color
	^ color! !

!HaloSpec methodsFor: 'as yet unclassified' stamp: 'sw 1/25/2000 18:41'!
horizontalPlacement
	^ horizontalPlacement! !

!HaloSpec methodsFor: 'as yet unclassified' stamp: 'sw 1/25/2000 19:54'!
horizontalPlacement: hp verticalPlacement: vp color: col iconSymbol: is addHandleSelector: sel
	horizontalPlacement := hp.
	verticalPlacement := vp.
	color:= col.
	iconSymbol := is asSymbol.
	addHandleSelector := sel! !

!HaloSpec methodsFor: 'as yet unclassified' stamp: 'sw 1/25/2000 18:41'!
iconSymbol
	^ iconSymbol! !

!HaloSpec methodsFor: 'printing' stamp: 'sw 11/15/2001 16:31'!
printOn: aStream
	"Add a textual printout representing the receiver to a stream"

	super printOn: aStream.
	aStream nextPutAll: ' (', addHandleSelector asString, ' ', iconSymbol asString, ')'! !

!HaloSpec methodsFor: 'as yet unclassified' stamp: 'sw 1/25/2000 18:41'!
verticalPlacement
	^ verticalPlacement! !

Object subclass: #KeyboardBuffer
	instanceVariableNames: 'event eventUsed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/30/97 19:53'!
commandKeyPressed
	^ event commandKeyPressed! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/30/97 19:54'!
controlKeyPressed
	^ event controlKeyPressed! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/29/97 12:34'!
flushKeyboard
	eventUsed ifFalse: [^ eventUsed := true].! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/29/97 12:34'!
keyboard
	eventUsed ifFalse: [eventUsed := true.  ^ event keyCharacter].
	^ nil! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/29/97 12:34'!
keyboardPeek
	eventUsed ifFalse: [^ event keyCharacter].
	^ nil! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'ar 7/19/2002 17:44'!
keyboardPressed
	^eventUsed not! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: 'di 9/30/97 19:54'!
leftShiftDown
	^ event shiftPressed! !

!KeyboardBuffer methodsFor: 'as yet unclassified' stamp: ''!
startingEvent: evt
	event := evt.
	eventUsed := false! !

Object subclass: #LayoutCell
	instanceVariableNames: 'target cellSize extraSpace flags nextCell'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Layouts'!

!LayoutCell commentStamp: '<historical>' prior: 0!
I am used in table layouts to hold temporary values while the layout is being computed.

Instance variables:
	target 		<Morph>		The morph contained in this cell
	cellSize 		<Point>		The size of the cell
	extraSpace 	<nil | Point>	Additional space to add after this cell
	nextCell 	<nil | LayoutCell>	The next cell in the arrangement.

Implementation note:
Both, cellSize and extraSpace contains points where
	x - represents the primary table direction
	y - represents the secondary table direction
!

!LayoutCell methodsFor: 'accessing' stamp: 'ar 11/2/2000 17:15'!
addExtraSpace: aPoint
	extraSpace 
		ifNil:[extraSpace := aPoint]
		ifNotNil:[extraSpace := extraSpace + aPoint]! !

!LayoutCell methodsFor: 'accessing' stamp: 'ar 10/28/2000 18:12'!
cellSize
	^cellSize! !

!LayoutCell methodsFor: 'accessing' stamp: 'ar 10/28/2000 18:12'!
cellSize: aPoint
	cellSize := aPoint! !

!LayoutCell methodsFor: 'collection' stamp: 'ar 10/28/2000 18:58'!
do: aBlock
	aBlock value: self.
	nextCell ifNotNil:[nextCell do: aBlock].! !

!LayoutCell methodsFor: 'accessing' stamp: 'ar 11/10/2000 17:09'!
extraSpace
	^extraSpace ifNil:[0@0]! !

!LayoutCell methodsFor: 'accessing' stamp: 'ar 10/28/2000 21:30'!
extraSpace: aPoint
	extraSpace := aPoint! !

!LayoutCell methodsFor: 'accessing' stamp: 'ls 8/5/2004 16:47'!
flags
	^flags ifNil: [ 0 ]! !

!LayoutCell methodsFor: 'accessing' stamp: 'ls 8/5/2004 16:48'!
hSpaceFill
	^self flags anyMask: 1! !

!LayoutCell methodsFor: 'accessing' stamp: 'ls 8/5/2004 16:48'!
hSpaceFill: aBool
	flags := aBool ifTrue:[self flags bitOr: 1] ifFalse:[self flags bitClear: 1].
! !

!LayoutCell methodsFor: 'collection' stamp: 'ar 10/28/2000 21:27'!
inject: thisValue into: binaryBlock 
	"Accumulate a running value associated with evaluating the argument, 
	binaryBlock, with the current value of the argument, thisValue, and the 
	receiver as block arguments. For instance, to sum the numeric elements 
	of a collection, aCollection inject: 0 into: [:subTotal :next | subTotal + 
	next]."

	| nextValue |
	nextValue := thisValue.
	self do: [:each | nextValue := binaryBlock value: nextValue value: each].
	^nextValue! !

!LayoutCell methodsFor: 'accessing' stamp: 'ar 10/28/2000 18:12'!
nextCell
	^nextCell! !

!LayoutCell methodsFor: 'accessing' stamp: 'ar 10/28/2000 18:12'!
nextCell: aCell
	nextCell := aCell! !

!LayoutCell methodsFor: 'accessing' stamp: 'dgd 2/22/2003 14:09'!
size
	| n cell |
	n := 0.
	cell := self.
	[cell isNil] whileFalse: 
			[n := n + 1.
			cell := cell nextCell].
	^n! !

!LayoutCell methodsFor: 'accessing' stamp: 'ar 10/28/2000 18:11'!
target
	^target! !

!LayoutCell methodsFor: 'accessing' stamp: 'ar 10/28/2000 18:12'!
target: newTarget
	target := newTarget! !

!LayoutCell methodsFor: 'accessing' stamp: 'ls 8/5/2004 16:47'!
vSpaceFill
	^self flags anyMask: 2! !

!LayoutCell methodsFor: 'accessing' stamp: 'ls 8/5/2004 16:48'!
vSpaceFill: aBool
	flags := aBool ifTrue:[self flags bitOr: 2] ifFalse:[self flags bitClear: 2].
! !

Object subclass: #LayoutPolicy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Layouts'!

!LayoutPolicy commentStamp: '<historical>' prior: 0!
A LayoutPolicy defines how submorphs of some morph should be arranged. Subclasses of the receiver define concrete layout policies.!

!LayoutPolicy methodsFor: 'layout' stamp: 'ar 1/27/2001 14:39'!
flushLayoutCache
	"Flush any cached information associated with the receiver"! !

!LayoutPolicy methodsFor: 'utilities' stamp: 'ar 10/29/2000 17:31'!
indexForInserting: aMorph at: aPoint in: someMorph
	"Return the insertion index based on the layout strategy defined for some morph. Used for drop insertion."
	^1 "front-most"! !

!LayoutPolicy methodsFor: 'testing' stamp: 'ar 10/29/2000 01:28'!
isProportionalLayout
	^false! !

!LayoutPolicy methodsFor: 'testing' stamp: 'ar 10/29/2000 01:28'!
isTableLayout
	^false! !

!LayoutPolicy methodsFor: 'layout' stamp: 'ar 10/28/2000 19:12'!
layout: aMorph in: newBounds
	"Compute the layout for the given morph based on the new bounds"
! !

!LayoutPolicy methodsFor: 'layout' stamp: 'ar 10/31/2000 19:59'!
minExtentOf: aMorph in: newBounds
	"Return the minimal size aMorph's children would require given the new bounds"
	^0@0! !

LayoutPolicy subclass: #ProportionalLayout
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Layouts'!

!ProportionalLayout commentStamp: '<historical>' prior: 0!
I represent a layout that places all children of some morph in their given LayoutFrame.!

!ProportionalLayout methodsFor: 'testing' stamp: 'ar 10/29/2000 01:29'!
isProportionalLayout
	^true! !

!ProportionalLayout methodsFor: 'layout' stamp: 'ar 10/29/2000 01:24'!
layout: aMorph in: newBounds
	"Compute the layout for the given morph based on the new bounds"
	aMorph submorphsDo:[:m| m layoutProportionallyIn: newBounds].! !

!ProportionalLayout methodsFor: 'layout' stamp: 'mt 4/28/2015 11:41'!
minExtentOf: aMorph in: newBounds
	"Return the minimal size aMorph's children would require given the new bounds"
	| min |
	min := 0@0.
	aMorph submorphsDo:[:m| | extent frame |
		"Map the minimal size of the child through the layout frame.
		Note: This is done here and not in the child because its specific
		for proportional layouts. Perhaps we'll generalize this for table
		layouts but I'm not sure how and when."
		extent := m minExtent.
		"frame := m layoutFrame.
		frame ifNotNil:[extent := frame minExtentFrom: extent]."
		min := min max: extent].
	^min! !

LayoutPolicy subclass: #TableLayout
	instanceVariableNames: 'properties minExtentCache'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Layouts'!

!TableLayout commentStamp: '<historical>' prior: 0!
The layout process:
For computing the new layout for the children of any morph, we start with an initial rectangle which is provided as a reference.

Step 1: The first step of layout computation is to compute the minimum extent each of our children can have. The minimum extent is mapped through both the local layout frame of the morph (for relative positioning) and the global layout frame (for insets, such as cursor indication) to obtain the minimal size required for each cell.

Step 2: Based on the cell sizes, the number of cells we can put into each row and column is computed. For equal spacing, the maximum size of the cells is taken into account here.

Step 3: Based on the row/column sizes, we compute the extra space which should be added to each row/column. For 
	#leftFlush/#topFlush - we add all extra space add the end
	#rightFlush/#bottomFlush - we add all extra space at the start
	#centering - we add 1/2 of the extra space at start and end
	#justified - we distribute the space evenly between the morphs
[NOTE: If any #spaceFill morphs are encountered during this step, #justified is implied and the space is exclusively and equally distributed between those #spaceFill morphs. This is for backward compatibility and should *never* be necessary in the new regime].

Step 4: The morphs are placed in the computed cells and the extra space is distributed as necessary. Placing the submorphs is done by mapping through the global and the local layout frame as requested.

Start point:
=> bounds: new rectangle for the morph.

Compute basic arrangement of morphs:
=> For each submorph compute minExtent
	- if global layout frame inset in global layout frame
	- if local layout frame inset in local layout frame
=> Compute number of morphs per, width and height of row/column
	- if equal spacing based on max size
=> Compute extra space per row/column
	- if centering = #justified; distribute space equally
	- if centering #leftFlush/#topFlush (-1) add start extra
	- if centering #rightFlush/#bottomFlush (1) add end extra
	- if centering #centered add 1/2 extra to start/end
	<extra space must be float and rounded accordingly!!>
=> Place morphs in appropriate cells
	- if global layout frame inset in global layout frame
	- if local layout frame inset in local layout frame
	<will likely cause #layoutChanged by submorphs>

Distribute morphs in row/column:

=> Compute the max length of each row/column
!

!TableLayout methodsFor: 'layout' stamp: 'jdl 3/28/2003 08:18'!
computeCellArrangement: cellHolder in: newBounds horizontal: aBool target: aMorph 
	"Compute number of cells we can put in each row/column. The returned array contains a list of all the cells we can put into the row/column at each level.
	Note: The arrangement is so that the 'x' value of each cell advances along the list direction and the 'y' value along the wrap direction. The returned arrangement has an extra cell at the start describing the width and height of the row."

	| cells wrap spacing output maxExtent n sum index max cell first last w cellMax maxCell hFill vFill inset |
	maxCell := cellHolder key.
	cells := cellHolder value.
	properties wrapDirection == #none 
		ifTrue: [wrap := SmallInteger maxVal]
		ifFalse: 
			[wrap := aBool ifTrue: [newBounds width] ifFalse: [newBounds height].
			wrap := wrap max: (maxCell x)].
	spacing := properties cellSpacing.
	(spacing == #globalRect or: [spacing = #globalSquare]) 
		ifTrue: 
			["Globally equal spacing is a very special case here, so get out fast and easy"

			^self 
				computeGlobalCellArrangement: cells
				in: newBounds
				horizontal: aBool
				wrap: wrap
				spacing: spacing].
	output := WriteStream on: Array new.
	inset := properties cellInset asPoint.
	aBool ifFalse: [inset := inset transposed].
	first := last := nil.
	maxExtent := 0 @ 0.
	sum := 0.
	index := 1.
	n := 0.
	hFill := vFill := false.
	[index <= cells size] whileTrue: 
			[w := sum.
			cell := cells at: index.
			cellMax := maxExtent max: cell cellSize.	"e.g., minSize"
			sum := (spacing == #localRect or: [spacing == #localSquare]) 
						ifTrue: 
							["Recompute entire size of current row"

							max := spacing == #localSquare 
										ifTrue: [cellMax x max: cellMax y]
										ifFalse: [cellMax x].
							(n + 1) * max]
						ifFalse: [sum + cell cellSize x].
			(sum + (n * inset x) > wrap and: [first notNil]) 
				ifTrue: 
					["It doesn't fit and we're not starting a new line"

					(spacing == #localSquare or: [spacing == #localRect]) 
						ifTrue: 
							[spacing == #localSquare 
								ifTrue: [maxExtent := (maxExtent x max: maxExtent y) asPoint].
							first do: [:c | c cellSize: maxExtent]].
					w := w + ((n - 1) * inset x).
					"redistribute extra space"
					first nextCell 
						ifNotNil: [first nextCell do: [:c | c addExtraSpace: inset x @ 0]].
					last := LayoutCell new.
					last cellSize: w @ maxExtent y.
					last hSpaceFill: hFill.
					last vSpaceFill: vFill.
					last nextCell: first.
					output position = 0 ifFalse: [last addExtraSpace: 0 @ inset y].
					output nextPut: last.
					first := nil.
					maxExtent := 0 @ 0.
					sum := 0.
					n := 0.
					hFill := vFill := false]
				ifFalse: 
					["It did fit; use next item from input"

					first ifNil: [first := last := cell]
						ifNotNil: 
							[last nextCell: cell.
							last := cell].
					index := index + 1.
					n := n + 1.
					maxExtent := cellMax.
					hFill := hFill or: [cell hSpaceFill].
					vFill := vFill or: [cell vSpaceFill]]].
	first ifNotNil: 
			[last := LayoutCell new.
			sum := sum + ((n - 1) * inset x).
			first nextCell 
				ifNotNil: [first nextCell do: [:c | c addExtraSpace: inset x @ 0]].
			last cellSize: sum @ maxExtent y.
			last hSpaceFill: hFill.
			last vSpaceFill: vFill.
			last nextCell: first.
			output position = 0 ifFalse: [last addExtraSpace: 0 @ inset y].
			output nextPut: last].
	output := output contents.
	properties listSpacing == #equal 
		ifTrue: 
			["Make all the heights equal"

			max := output inject: 0 into: [:size :c | size max: c cellSize y].
			output do: [:c | c cellSize: c cellSize x @ max]].
	^output! !

!TableLayout methodsFor: 'layout' stamp: 'nice 12/27/2009 03:11'!
computeCellSizes: aMorph in: newBounds horizontal: aBool
	"Step 1: Compute the minimum extent for all the children of aMorph"
	| cells block minSize maxSize maxCell |
	cells := WriteStream on: (Array new: aMorph submorphCount).
	minSize := properties minCellSize asPoint.
	maxSize := properties maxCellSize asPoint.
	aBool ifTrue:[
		minSize := minSize transposed.
		maxSize := maxSize transposed].
	maxCell := 0@0.
	block := [:m| | size cell |
		m disableTableLayout ifFalse:[
			size := m minExtent asIntegerPoint.
			cell := LayoutCell new target: m.
			aBool ifTrue:[
				cell hSpaceFill: m hResizing == #spaceFill.
				cell vSpaceFill: m vResizing == #spaceFill.
			] ifFalse:[
				cell hSpaceFill: m vResizing == #spaceFill.
				cell vSpaceFill: m hResizing == #spaceFill.
				size := size transposed.
			].
			size := (size min: maxSize) max: minSize.
			cell cellSize: size.
			maxCell := maxCell max: size.
			cells nextPut: cell]].
	properties reverseTableCells
		ifTrue:[aMorph submorphsReverseDo: block]
		ifFalse:[aMorph submorphsDo: block].
	^maxCell -> cells contents! !

!TableLayout methodsFor: 'layout' stamp: 'aoy 2/15/2003 20:49'!
computeExtraSpacing: arrangement in: newBounds horizontal: aBool target: aMorph 
	"Compute the required extra spacing for laying out the cells"

	"match newBounds extent with arrangement's orientation"

	| extent extra centering n extraPerCell cell last hFill vFill max amount allow |
	extent := newBounds extent.
	aBool ifFalse: [extent := extent transposed].

	"figure out if we have any horizontal or vertical space fillers"
	hFill := vFill := false.
	max := 0 @ 0.
	arrangement do: 
			[:c | 
			max := (max x max: c cellSize x) @ (max y + c cellSize y).
			max := max max: c cellSize.
			hFill := hFill or: [c hSpaceFill].
			vFill := vFill or: [c vSpaceFill]].

	"Take client's shrink wrap constraints into account.
	Note: these are only honored when there are no #spaceFill children,
	or when #rubberBandCells is set."
	allow := properties rubberBandCells not.
	aMorph hResizing == #shrinkWrap 
		ifTrue: 
			[aBool 
				ifTrue: [allow & hFill ifFalse: [extent := max x @ (max y max: extent y)]]
				ifFalse: [allow & vFill ifFalse: [extent := (max x max: extent x) @ max y]]].
	aMorph vResizing == #shrinkWrap 
		ifTrue: 
			[aBool 
				ifFalse: [allow & hFill ifFalse: [extent := max x @ (max y max: extent y)]]
				ifTrue: [allow & vFill ifFalse: [extent := (max x max: extent x) @ max y]]].

	"Now compute the extra v space"
	extra := extent y 
				- (arrangement inject: 0 into: [:sum :c | sum + c cellSize y]).
	extra > 0 
		ifTrue: 
			["Check if we have any #spaceFillers"

			vFill 
				ifTrue: 
					["use only #spaceFillers"

					n := arrangement inject: 0
								into: [:sum :c | c vSpaceFill ifTrue: [sum + 1] ifFalse: [sum]].
					n isZero ifFalse: [extraPerCell := extra asFloat / n asFloat].
					extra := last := 0.
					arrangement do: 
							[:c | 
							c vSpaceFill 
								ifTrue: 
									[extra := (last := extra) + extraPerCell.
									amount := 0 @ (extra truncated - last truncated).
									c do: [:cc | cc cellSize: cc cellSize + amount]]]]
				ifFalse: 
					["no #spaceFillers; distribute regularly"

					centering := properties wrapCentering.
					"centering == #topLeft ifTrue:[]."	"add all extra space to the last cell; e.g., do nothing"
					centering == #bottomRight 
						ifTrue: 
							["add all extra space to the first cell"

							arrangement first addExtraSpace: 0 @ extra].
					centering == #center 
						ifTrue: 
							["add 1/2 extra space to the first and last cell"

							arrangement first addExtraSpace: 0 @ (extra // 2)].
					centering == #justified 
						ifTrue: 
							["add extra space equally distributed to each cell"

							n := arrangement size - 1 max: 1.
							extraPerCell := extra asFloat / n asFloat.
							extra := last := 0.
							arrangement do: 
									[:c | 
									c addExtraSpace: 0 @ (extra truncated - last truncated).
									extra := (last := extra) + extraPerCell]]]].

	"Now compute the extra space for the primary direction"
	centering := properties listCentering.
	1 to: arrangement size
		do: 
			[:i | 
			cell := arrangement at: i.
			extra := extent x - cell cellSize x.
			extra > 0 
				ifTrue: 
					["Check if we have any #spaceFillers"
					cell := cell nextCell.
					cell hSpaceFill 
						ifTrue: 
							["use only #spaceFillers"

							
							n := cell inject: 0
										into: [:sum :c | c hSpaceFill ifTrue: [sum + c target spaceFillWeight] ifFalse: [sum]].
							n isZero ifFalse: [extraPerCell := extra asFloat / n asFloat].
							extra := last := 0.
							cell do: 
									[:c | 
									c hSpaceFill 
										ifTrue: 
											[extra := (last := extra) + (extraPerCell * c target spaceFillWeight).
											amount := extra truncated - last truncated.
											c cellSize: c cellSize + (amount @ 0)]]]
						ifFalse: 
							["no #spaceFiller; distribute regularly"

						
							"centering == #topLeft ifTrue:[]"	"add all extra space to the last cell; e.g., do nothing"
							centering == #bottomRight 
								ifTrue: 
									["add all extra space to the first cell"

									cell addExtraSpace: extra @ 0].
							centering == #center 
								ifTrue: 
									["add 1/2 extra space to the first and last cell"

									cell addExtraSpace: (extra // 2) @ 0].
							centering == #justified 
								ifTrue: 
									["add extra space equally distributed to each cell"

									n := cell size - 1 max: 1.
									extraPerCell := extra asFloat / n asFloat.
									extra := last := 0.
									cell do: 
											[:c | 
											c addExtraSpace: (extra truncated - last truncated) @ 0.
											extra := (last := extra) + extraPerCell]]]]]! !

!TableLayout methodsFor: 'layout' stamp: 'ar 10/31/2000 22:50'!
computeGlobalCellArrangement: cells in: newBounds horizontal: aBool wrap: wrap spacing: spacing
	"Compute number of cells we can put in each row/column. The returned array contains a list of all the cells we can put into the row/column at each level.
	Note: The arrangement is so that the 'x' value of each cell advances along the list direction and the 'y' value along the wrap direction. The returned arrangement has an extra cell at the start describing the width and height of the row."
	| output maxExtent n cell first last hFill vFill |
	output := (WriteStream on: Array new).
	first := last := nil.
	maxExtent := cells inject: 0@0 into:[:size :c| size max: c cellSize "e.g., minSize"].
	spacing == #globalSquare ifTrue:[maxExtent := (maxExtent x max: maxExtent y) asPoint].
	n := (wrap // maxExtent x) max: 1.
	hFill := vFill := false.
	1 to: cells size do:[:i|
		cell := cells at: i.
		hFill := hFill or:[cell hSpaceFill].
		vFill := vFill or:[cell vSpaceFill].
		cell cellSize: maxExtent.
		first ifNil:[first := last := cell] ifNotNil:[last nextCell: cell. last := cell].
		(i \\ n) = 0 ifTrue:[
			last := LayoutCell new.
			last cellSize: (maxExtent x * n) @ (maxExtent y).
			last hSpaceFill: hFill.
			last vSpaceFill: vFill.
			hFill := vFill := false.
			last nextCell: first.
			output nextPut: last.
			first := nil]].
	first ifNotNil:[
		last := LayoutCell new.
		last cellSize: (maxExtent x * n) @ (maxExtent y). self flag: #arNote."@@@: n is not correct!!"
		last nextCell: first.
		output nextPut: last].
	^output contents
! !

!TableLayout methodsFor: 'layout' stamp: 'ar 1/27/2001 14:40'!
flushLayoutCache
	"Flush any cached information associated with the receiver"
	minExtentCache := nil.! !

!TableLayout methodsFor: 'utilities' stamp: 'aoy 2/17/2003 01:22'!
indexForInserting: aMorph at: aPoint in: owner 
	"Return the insertion index based on the layout strategy defined for some morph. Used for drop insertion."

	| horizontal morphList index |
	owner hasSubmorphs ifFalse: [^1].
	aMorph disableTableLayout ifTrue: [^1].
	horizontal := (owner listDirection == #topToBottom 
				or: [owner listDirection == #bottomToTop]) not .
	morphList := owner submorphs.
	owner reverseTableCells ifTrue: [morphList := morphList reversed].
	index := self 
				indexForInserting: aPoint
				inList: morphList
				horizontal: horizontal
				target: owner.
	owner reverseTableCells ifTrue: [index := morphList size - index + 2].
	^index ifNil: [1]! !

!TableLayout methodsFor: 'utilities' stamp: 'nice 12/27/2009 03:11'!
indexForInserting: aPoint inList: morphList horizontal: aBool target: aMorph 
	| cmp1 cmp2 cmp3 noWrap |
	properties := aMorph layoutProperties.
	noWrap := properties wrapDirection == #none.
	aBool 
		ifTrue: 
			["horizontal"

			properties listDirection == #rightToLeft 
				ifTrue: [cmp1 := [:rect | aPoint x > rect left]]
				ifFalse: [cmp1 := [:rect | aPoint x < rect right]].
			properties wrapDirection == #bottomToTop 
				ifTrue: 
					[cmp2 := [:rect | aPoint y > rect top].
					cmp3 := [:rect | aPoint y > rect bottom]]
				ifFalse: 
					[cmp2 := [:rect | aPoint y < rect bottom].
					cmp3 := [:rect | aPoint y < rect top]]]
		ifFalse: 
			["vertical"

			properties listDirection == #bottomToTop 
				ifTrue: [cmp1 := [:rect | aPoint y > rect top]]
				ifFalse: [cmp1 := [:rect | aPoint y < rect bottom]].
			properties wrapDirection == #rightToLeft 
				ifTrue: 
					[cmp2 := [:rect | aPoint x > rect left].
					cmp3 := [:rect | aPoint x > rect right]]
				ifFalse: 
					[cmp2 := [:rect | aPoint x < rect right].
					cmp3 := [:rect | aPoint x < rect left]]]. 
	morphList keysAndValuesDo: 
			[:index :m | | box | 
			self flag: #arNote.	"it is not quite clear if we can really use #fullBounds here..."
			box := m fullBounds.
			noWrap 
				ifTrue: 
					["Only in one direction"

					(cmp1 value: box) ifTrue: [^index]]
				ifFalse: 
					["Check for inserting before current row"

					(cmp3 value: box) ifTrue: [^index].
					"Check for inserting before current cell"
					((cmp1 value: box) and: [cmp2 value: box]) ifTrue: [^index]]].
	^morphList size + 1! !

!TableLayout methodsFor: 'testing' stamp: 'ar 10/29/2000 01:29'!
isTableLayout
	^true! !

!TableLayout methodsFor: 'optimized' stamp: 'nice 12/27/2009 03:11'!
layoutLeftToRight: aMorph in: newBounds 
	"An optimized left-to-right list layout"

	| inset extent block posX posY centering extraPerCell amount minX minY maxX maxY n width extra last cell size height sum vFill first |
	size := properties minCellSize asPoint.
	minX := size x.
	minY := size y.
	size := properties maxCellSize asPoint.
	maxX := size x.
	maxY := size y.
	inset := properties cellInset asPoint x.
	extent := newBounds extent.
	n := 0.
	vFill := false.
	sum := 0.
	width := height := 0.
	first := last := nil.
	block := 
			[:m | | sizeX props sizeY | 
			props := m layoutProperties ifNil: [m].
			props disableTableLayout 
				ifFalse: 
					[n := n + 1.
					cell := LayoutCell new target: m.
					props hResizing == #spaceFill 
						ifTrue: 
							[cell hSpaceFill: true.
							extra := m spaceFillWeight.
							cell extraSpace: extra.
							sum := sum + extra]
						ifFalse: [cell hSpaceFill: false].
					props vResizing == #spaceFill ifTrue: [vFill := true].
					size := m minExtent.
					size := m minExtent.
					sizeX := size x.
					sizeY := size y.
					sizeX < minX ifTrue: [sizeX := minX] ifFalse: [sizeX := sizeX min: maxX].
					sizeY < minY ifTrue: [sizeY := minY] ifFalse: [sizeY := sizeY min: maxY].
					cell cellSize: sizeX.
					last ifNil: [first := cell] ifNotNil: [last nextCell: cell].
					last := cell.
					width := width + sizeX.
					sizeY > height ifTrue: [height := sizeY]]].
	properties reverseTableCells 
		ifTrue: [aMorph submorphsReverseDo: block]
		ifFalse: [aMorph submorphsDo: block].
	n > 1 ifTrue: [width := width + ((n - 1) * inset)].
	(properties hResizing == #shrinkWrap 
		and: [properties rubberBandCells or: [sum isZero]]) 
			ifTrue: [extent := width @ (extent y max: height)].
	(properties vResizing == #shrinkWrap 
		and: [properties rubberBandCells or: [vFill not]]) 
			ifTrue: [extent := (extent x max: width) @ height].
	posX := newBounds left.
	posY := newBounds top.

	"Compute extra vertical space"
	extra := extent y - height.
	extra := extra max: 0.
	extra > 0 
		ifTrue: 
			[vFill 
				ifTrue: [height := extent y]
				ifFalse: 
					[centering := properties wrapCentering.
					centering == #bottomRight ifTrue: [posY := posY + extra].
					centering == #center ifTrue: [posY := posY + (extra // 2)]]].


	"Compute extra horizontal space"
	extra := extent x - width.
	extra := extra max: 0.
	extraPerCell := 0.
	extra > 0 
		ifTrue: 
			[sum isZero 
				ifTrue: 
					["extra space but no #spaceFillers"

					centering := properties listCentering.
					centering == #bottomRight ifTrue: [posX := posX + extra].
					centering == #center ifTrue: [posX := posX + (extra // 2)]]
				ifFalse: [extraPerCell := extra asFloat / sum asFloat]].
	n := 0.
	extra := last := 0.
	cell := first.
	[cell isNil] whileFalse: 
			[n := n + 1.
			width := cell cellSize.
			(extraPerCell > 0 and: [cell hSpaceFill]) 
				ifTrue: 
					[extra := (last := extra) + (extraPerCell * cell extraSpace).
					amount := extra truncated - last truncated.
					width := width + amount].
			cell target layoutInBounds: (posX @ posY extent: width @ height).
			posX := posX + width + inset.
			cell := cell nextCell]! !

!TableLayout methodsFor: 'optimized' stamp: 'nice 12/27/2009 03:11'!
layoutTopToBottom: aMorph in: newBounds 
	"An optimized top-to-bottom list layout"

	| inset extent block posX posY centering extraPerCell amount minX minY maxX maxY n height extra last cell size width sum vFill first |
	size := properties minCellSize asPoint.
	minX := size x.
	minY := size y.
	size := properties maxCellSize asPoint.
	maxX := size x.
	maxY := size y.
	inset := properties cellInset asPoint y.
	extent := newBounds extent.
	n := 0.
	vFill := false.
	sum := 0.
	width := height := 0.
	first := last := nil.
	block := 
			[:m | | sizeY sizeX props | 
			props := m layoutProperties ifNil: [m].
			props disableTableLayout 
				ifFalse: 
					[n := n + 1.
					cell := LayoutCell new target: m.
					props vResizing == #spaceFill 
						ifTrue: 
							[cell vSpaceFill: true.
							extra := m spaceFillWeight.
							cell extraSpace: extra.
							sum := sum + extra]
						ifFalse: [cell vSpaceFill: false].
					props hResizing == #spaceFill ifTrue: [vFill := true].
					size := m minExtent.
					sizeX := size x.
					sizeY := size y.
					sizeX < minX ifTrue: [sizeX := minX] ifFalse: [sizeX := sizeX min: maxX].
					sizeY < minY ifTrue: [sizeY := minY] ifFalse: [sizeY := sizeY min: maxY].
					cell cellSize: sizeY.
					first ifNil: [first := cell] ifNotNil: [last nextCell: cell].
					last := cell.
					height := height + sizeY.
					sizeX > width ifTrue: [width := sizeX]]].
	properties reverseTableCells 
		ifTrue: [aMorph submorphsReverseDo: block]
		ifFalse: [aMorph submorphsDo: block].
	n > 1 ifTrue: [height := height + ((n - 1) * inset)].
	(properties vResizing == #shrinkWrap 
		and: [properties rubberBandCells or: [sum isZero]]) 
			ifTrue: [extent := (extent x max: width) @ height].
	(properties hResizing == #shrinkWrap 
		and: [properties rubberBandCells or: [vFill not]]) 
			ifTrue: [extent := width @ (extent y max: height)].
	posX := newBounds left.
	posY := newBounds top.

	"Compute extra horizontal space"
	extra := extent x - width.
	extra := extra max: 0.
	extra > 0 
		ifTrue: 
			[vFill 
				ifTrue: [width := extent x]
				ifFalse: 
					[centering := properties wrapCentering.
					centering == #bottomRight ifTrue: [posX := posX + extra].
					centering == #center ifTrue: [posX := posX + (extra // 2)]]].


	"Compute extra vertical space"
	extra := extent y - height.
	extra := extra max: 0.
	extraPerCell := 0.
	extra > 0 
		ifTrue: 
			[sum isZero 
				ifTrue: 
					["extra space but no #spaceFillers"

					centering := properties listCentering.
					centering == #bottomRight ifTrue: [posY := posY + extra].
					centering == #center ifTrue: [posY := posY + (extra // 2)]]
				ifFalse: [extraPerCell := extra asFloat / sum asFloat]].
	n := 0.
	extra := last := 0.
	cell := first.
	[cell isNil] whileFalse: 
			[n := n + 1.
			height := cell cellSize.
			(extraPerCell > 0 and: [cell vSpaceFill]) 
				ifTrue: 
					[extra := (last := extra) + (extraPerCell * cell extraSpace).
					amount := extra truncated - last truncated.
					height := height + amount].
			cell target layoutInBounds: (posX @ posY extent: width @ height).
			posY := posY + height + inset.
			cell := cell nextCell]! !

!TableLayout methodsFor: 'layout' stamp: 'aoy 2/17/2003 01:23'!
layout: aMorph in: box 
	"Compute the layout for the given morph based on the new bounds"

	| cells arrangement horizontal newBounds |
	aMorph hasSubmorphs ifFalse: [^self].
	properties := aMorph assureTableProperties.
	newBounds := box origin asIntegerPoint corner: box corner asIntegerPoint.
	(properties wrapDirection == #none and: [properties cellSpacing == #none]) 
		ifTrue: 
			["get into the fast lane"

			properties listCentering == #justified 
				ifFalse: 
					["can't deal with that"

					properties listDirection == #leftToRight 
						ifTrue: [^self layoutLeftToRight: aMorph in: newBounds].
					properties listDirection == #topToBottom 
						ifTrue: [^self layoutTopToBottom: aMorph in: newBounds]]].
	horizontal := (properties listDirection == #topToBottom 
				or: [properties listDirection == #bottomToTop]) not. 
	"Step 1: Compute the minimum extent for all the children of aMorph"
	cells := self 
				computeCellSizes: aMorph
				in: (0 @ 0 corner: newBounds extent)
				horizontal: horizontal.
	"Step 2: Compute the arrangement of the cells for each row and column"
	arrangement := self 
				computeCellArrangement: cells
				in: newBounds
				horizontal: horizontal
				target: aMorph.
	"Step 3: Compute the extra spacing for each cell"
	self 
		computeExtraSpacing: arrangement
		in: newBounds
		horizontal: horizontal
		target: aMorph.
	"Step 4: Place the children within the cells accordingly"
	self 
		placeCells: arrangement
		in: newBounds
		horizontal: horizontal
		target: aMorph! !

!TableLayout methodsFor: 'optimized' stamp: 'nice 12/27/2009 03:11'!
minExtentHorizontal: aMorph 
	"Return the minimal size aMorph's children would require given the new bounds"

	| inset minX minY maxX maxY n size width height |
	size := properties minCellSize asPoint.
	minX := size x.
	minY := size y.
	size := properties maxCellSize asPoint.
	maxX := size x.
	maxY := size y.
	inset := properties cellInset asPoint.
	n := 0.
	width := height := 0.
	aMorph submorphsDo: 
			[:m | | sizeX sizeY | 
			m disableTableLayout 
				ifFalse: 
					[n := n + 1.
					size := m minExtent.
					sizeX := size x.
					sizeY := size y.
					sizeX < minX 
						ifTrue: [sizeX := minX]
						ifFalse: [sizeX := sizeX min: maxX].
					sizeY < minY 
						ifTrue: [sizeY := minY]
						ifFalse: [sizeY := sizeY min: maxY].
					width := width + sizeX.
					sizeY > height ifTrue: [height := sizeY]]].
	n > 1 ifTrue: [width := width + ((n - 1) * inset x)].
	^minExtentCache := width @ height! !

!TableLayout methodsFor: 'layout' stamp: 'gm 2/28/2003 01:43'!
minExtentOf: aMorph in: box 
	"Return the minimal size aMorph's children would require given the new bounds"

	| cells arrangement horizontal newBounds minX minY dir |
	minExtentCache isNil ifFalse: [^minExtentCache].
	aMorph hasSubmorphs ifFalse: [^0 @ 0].
	properties := aMorph assureTableProperties.
	(properties wrapDirection == #none and: [properties cellSpacing == #none]) 
		ifTrue: 
			["Get into the fast lane"

			dir := properties listDirection.
			(dir == #leftToRight or: [dir == #rightToLeft]) 
				ifTrue: [^self minExtentHorizontal: aMorph].
			(dir == #topToBottom or: [dir == #bottomToTop]) 
				ifTrue: [^self minExtentVertical: aMorph]].
	newBounds := box origin asIntegerPoint corner: box corner asIntegerPoint.
	horizontal := (properties listDirection == #topToBottom 
				or: [properties listDirection == #bottomToTop]) not.
	"Step 1: Compute the minimum extent for all the children of aMorph"
	cells := self 
				computeCellSizes: aMorph
				in: (0 @ 0 corner: newBounds extent)
				horizontal: horizontal.
	"Step 2: Compute the arrangement of the cells for each row and column"
	arrangement := self 
				computeCellArrangement: cells
				in: newBounds
				horizontal: horizontal
				target: aMorph.
	"Step 3: Extract the minimum size out of the arrangement"
	minX := minY := 0.
	arrangement do: 
			[:cell | 
			minX := minX max: cell cellSize x + cell extraSpace x.
			minY := minY + cell cellSize y + cell extraSpace y].
	minExtentCache := horizontal ifTrue: [minX @ minY] ifFalse: [minY @ minX].
	^minExtentCache! !

!TableLayout methodsFor: 'optimized' stamp: 'nice 12/27/2009 03:11'!
minExtentVertical: aMorph 
	"Return the minimal size aMorph's children would require given the new bounds"

	| inset minX minY maxX maxY n size width height |
	size := properties minCellSize asPoint.
	minX := size x.
	minY := size y.
	size := properties maxCellSize asPoint.
	maxX := size x.
	maxY := size y.
	inset := properties cellInset asPoint.
	n := 0.
	width := height := 0.
	aMorph submorphsDo: 
			[:m | | sizeY sizeX | 
			m disableTableLayout 
				ifFalse: 
					[n := n + 1.
					size := m minExtent.
					sizeX := size x.
					sizeY := size y.
					sizeX < minX 
						ifTrue: [sizeX := minX]
						ifFalse: [sizeX := sizeX min: maxX].
					sizeY < minY 
						ifTrue: [sizeY := minY]
						ifFalse: [sizeY := sizeY min: maxY].
					height := height + sizeY.
					sizeX > width ifTrue: [width := sizeX]]].
	n > 1 ifTrue: [height := height + ((n - 1) * inset y)].
	^minExtentCache := width @ height! !

!TableLayout methodsFor: 'layout' stamp: 'dgd 2/22/2003 14:42'!
placeCells: arrangement in: newBounds horizontal: aBool target: aMorph 
	"Place the morphs within the cells accordingly"

	| xDir yDir anchor yDist place cell xDist cellRect corner inset |
	inset := properties cellInset.
	(inset isNumber and: [inset isZero]) ifTrue: [inset := nil].
	aBool 
		ifTrue: 
			["horizontal layout"

			properties listDirection == #rightToLeft 
				ifTrue: 
					[xDir := -1 @ 0.
					properties wrapDirection == #bottomToTop 
						ifTrue: 
							[yDir := 0 @ -1.
							anchor := newBounds bottomRight]
						ifFalse: 
							[yDir := 0 @ 1.
							anchor := newBounds topRight]]
				ifFalse: 
					[xDir := 1 @ 0.
					properties wrapDirection == #bottomToTop 
						ifTrue: 
							[yDir := 0 @ -1.
							anchor := newBounds bottomLeft]
						ifFalse: 
							[yDir := 0 @ 1.
							anchor := newBounds topLeft]]]
		ifFalse: 
			["vertical layout"

			properties listDirection == #bottomToTop 
				ifTrue: 
					[xDir := 0 @ -1.
					properties wrapDirection == #rightToLeft 
						ifTrue: 
							[yDir := -1 @ 0.
							anchor := newBounds bottomRight]
						ifFalse: 
							[yDir := 1 @ 0.
							anchor := newBounds bottomLeft]]
				ifFalse: 
					[xDir := 0 @ 1.
					anchor := properties wrapDirection == #rightToLeft 
								ifTrue: 
									[yDir := -1 @ 0.
									newBounds topRight]
								ifFalse: 
									[yDir := 1 @ 0.
									newBounds topLeft]]].
	1 to: arrangement size
		do: 
			[:i | 
			cell := arrangement at: i.
			cell extraSpace ifNotNil: [anchor := anchor + (cell extraSpace y * yDir)].
			yDist := cell cellSize y * yDir.	"secondary advance direction"
			place := anchor.
			cell := cell nextCell.
			[cell isNil] whileFalse: 
					[cell extraSpace ifNotNil: [place := place + (cell extraSpace x * xDir)].
					xDist := cell cellSize x * xDir.	"primary advance direction"
					corner := place + xDist + yDist.
					cellRect := Rectangle origin: (place min: corner)
								corner: (place max: corner).
					inset ifNotNil: [cellRect := cellRect insetBy: inset].
					cell target layoutInBounds: cellRect.
					place := place + xDist.
					cell := cell nextCell].
			anchor := anchor + yDist]! !

Object subclass: #LayoutProperties
	instanceVariableNames: 'hResizing vResizing disableLayout'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Layouts'!

!LayoutProperties commentStamp: '<historical>' prior: 0!
This class provides a compact bit encoding for the most commonly used layout properties.!

!LayoutProperties methodsFor: 'converting' stamp: 'ar 11/14/2000 17:52'!
asTableLayoutProperties
	^(TableLayoutProperties new)
		hResizing: self hResizing;
		vResizing: self vResizing;
		disableTableLayout: self disableTableLayout;
		yourself! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 19:53'!
cellInset
	"Default"
	^0! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:50'!
cellPositioning
	^#center! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 19:55'!
cellSpacing
	"Default"
	^#none! !

!LayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:51'!
disableTableLayout
	^disableLayout! !

!LayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:51'!
disableTableLayout: aBool
	disableLayout := aBool! !

!LayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:51'!
hResizing
	^hResizing! !

!LayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:51'!
hResizing: aSymbol
	hResizing := aSymbol! !

!LayoutProperties methodsFor: 'testing' stamp: 'ar 11/13/2000 18:34'!
includesTableProperties
	^false! !

!LayoutProperties methodsFor: 'initialize' stamp: 'alain.plantec 5/28/2009 10:00'!
initialize
	super initialize.
	hResizing := vResizing := #rigid.
	disableLayout := false.! !

!LayoutProperties methodsFor: 'initialize' stamp: 'ar 11/14/2000 17:56'!
initializeFrom: defaultProvider
	"Initialize the receiver from a default provider"
	self hResizing: defaultProvider hResizing.
	self vResizing: defaultProvider vResizing.
	self disableTableLayout: defaultProvider disableTableLayout.! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 16:38'!
layoutInset
	^0! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 19:58'!
listCentering
	"Default"
	^#topLeft! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 19:59'!
listDirection
	"Default"
	^#topToBottom! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 19:59'!
listSpacing
	"Default"
	^#none! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 20:00'!
maxCellSize
	^SmallInteger maxVal! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 20:00'!
minCellSize
	^0! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 20:00'!
reverseTableCells
	^false! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 20:01'!
rubberBandCells
	^false! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'kfr 4/25/2015 10:17'!
stringWithLayout
	| defaultValues uniqueValues |
	defaultValues := TableLayoutProperties new.
	uniqueValues := self class allInstVarNames
		select: [:title | (self instVarNamed: title)
					~= (defaultValues instVarNamed: title)].
	^ String
		streamContents: [:aStream | 
			aStream nextPutAll: 'TableLayout new; '.
			uniqueValues
				do: [:title | aStream nextPutAll: title;
						 nextPut: $:;
						 space;
						
						print: (self instVarNamed: title).
						(title = uniqueValues last)
						ifTrue:[ aStream nextPut:$.]
						ifFalse:[ aStream nextPut: $;; cr]
						]]! !

!LayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:51'!
vResizing
	^vResizing! !

!LayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:52'!
vResizing: aSymbol
	vResizing := aSymbol! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 20:02'!
wrapCentering
	^#topLeft! !

!LayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/13/2000 20:02'!
wrapDirection
	^#none! !

LayoutProperties subclass: #TableLayoutProperties
	instanceVariableNames: 'cellInset cellPositioning cellSpacing layoutInset listCentering listDirection listSpacing reverseTableCells rubberBandCells wrapCentering wrapDirection minCellSize maxCellSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Layouts'!

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:45'!
cellInset
	^cellInset! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/13/2000 17:57'!
cellInset: aNumber
	cellInset := aNumber! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:49'!
cellPositioning
	^cellPositioning! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:49'!
cellPositioning: aSymbol
	cellPositioning := aSymbol! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:46'!
cellSpacing
	^cellSpacing! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:46'!
cellSpacing: aSymbol
	cellSpacing := aSymbol.! !

!TableLayoutProperties methodsFor: 'testing' stamp: 'ar 11/13/2000 18:34'!
includesTableProperties
	^true! !

!TableLayoutProperties methodsFor: 'initialize' stamp: 'ar 11/14/2000 17:45'!
initialize
	super initialize.
	cellSpacing := listSpacing := wrapDirection := #none.
	cellPositioning := #center.
	listCentering := wrapCentering := #topLeft.
	listDirection := #topToBottom.
	reverseTableCells := rubberBandCells := false.
	layoutInset := cellInset := minCellSize := 0.
	maxCellSize := 1073741823. "SmallInteger maxVal"
! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 16:37'!
layoutInset
	^layoutInset! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 16:37'!
layoutInset: aNumber
	layoutInset := aNumber! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:46'!
listCentering
	^listCentering! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:46'!
listCentering: aSymbol
	listCentering := aSymbol! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:46'!
listDirection
	^listDirection! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:46'!
listDirection: aSymbol
	listDirection := aSymbol.! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:47'!
listSpacing
	^listSpacing! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:47'!
listSpacing: aSymbol
	listSpacing := aSymbol! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:47'!
maxCellSize
	^maxCellSize! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/13/2000 17:58'!
maxCellSize: aNumber
	maxCellSize := aNumber.! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:47'!
minCellSize
	^minCellSize! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/13/2000 17:57'!
minCellSize: aNumber
	minCellSize := aNumber.! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:47'!
reverseTableCells
	^reverseTableCells! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:47'!
reverseTableCells: aBool
	reverseTableCells := aBool! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:47'!
rubberBandCells
	^rubberBandCells! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:48'!
rubberBandCells: aBool
	rubberBandCells := aBool.! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:48'!
wrapCentering
	^wrapCentering! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:48'!
wrapCentering: aSymbol
	wrapCentering := aSymbol! !

!TableLayoutProperties methodsFor: 'table defaults' stamp: 'ar 11/14/2000 17:48'!
wrapDirection
	^wrapDirection! !

!TableLayoutProperties methodsFor: 'accessing' stamp: 'ar 11/14/2000 17:48'!
wrapDirection: aSymbol
	wrapDirection := aSymbol! !

Object subclass: #ListItemWrapper
	instanceVariableNames: 'item model'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Explorer'!

!ListItemWrapper commentStamp: '<historical>' prior: 0!
Contributed by Bob Arning as part of the ObjectExplorer package.
!

ListItemWrapper subclass: #FileDirectoryWrapper
	instanceVariableNames: 'itemName balloonText hasContents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Explorer'!

!FileDirectoryWrapper class methodsFor: 'as yet unclassified' stamp: 'RAA 6/15/2000 18:01'!
with: anObject name: aString model: aModel

	^self new 
		setItem: anObject name: aString model: aModel! !

!FileDirectoryWrapper methodsFor: 'converting' stamp: 'dgd 8/27/2004 18:45'!
asString
	 ^itemName translatedIfCorresponds! !

!FileDirectoryWrapper methodsFor: 'accessing' stamp: 'RAA 7/21/2000 11:00'!
balloonText

	^balloonText! !

!FileDirectoryWrapper methodsFor: 'as yet unclassified' stamp: 'RAA 7/21/2000 11:01'!
balloonText: aStringOrNil

	balloonText := aStringOrNil! !

!FileDirectoryWrapper methodsFor: 'accessing' stamp: 'ar 2/12/2001 16:20'!
contents

	^((model directoryNamesFor: item) sortBy: [ :a :b | a caseInsensitiveLessOrEqual: b]) collect: [ :n | 
		FileDirectoryWrapper with: (item directoryNamed: n) name: n model: self
	]
! !

!FileDirectoryWrapper methodsFor: 'as yet unclassified' stamp: 'ar 2/12/2001 16:22'!
directoryNamesFor: anItem
	^model directoryNamesFor: anItem! !

!FileDirectoryWrapper methodsFor: 'accessing' stamp: 'tpr 11/28/2003 14:02'!
hasContents
	"Return whether this directory has subfolders. The value is cached to 
	avoid a performance penalty.	Also for performance reasons, the code 
	below will just assume that the directory does indeed have contents in a 
	few of cases:  
	1. If the item is not a FileDirectory (thus avoiding the cost 
	of refreshing directories that are not local) 
	2. If it's the root directory of a given volume 
	3. If there is an error computing the FileDirectory's contents
	"
	hasContents
		ifNil: [hasContents := true. "default"
			["Best test I could think of for determining if this is a local directory "
			((item isKindOf: FileDirectory)
					and: ["test to see that it's not the root directory"
						"there has to be a better way of doing this test -tpr"
						item pathParts size > 1])
				ifTrue: [hasContents := self contents notEmpty]]
				on: Error
				do: [hasContents := true]].
	^ hasContents! !

!FileDirectoryWrapper methodsFor: 'accessing' stamp: 'dgd 9/26/2004 18:22'!
icon
	"Answer a form to be used as icon"
	^ item isRemoteDirectory
		ifTrue: [MenuIcons smallRemoteOpenIcon]
		ifFalse: [MenuIcons smallOpenIcon]! !

!FileDirectoryWrapper methodsFor: 'as yet unclassified' stamp: 'sps 12/5/2002 16:59'!
setItem: anObject name: aString model: aModel

	item := anObject.
	model := aModel.
	itemName := aString.
	hasContents := nil.
! !

!FileDirectoryWrapper methodsFor: 'as yet unclassified' stamp: 'RAA 6/16/2000 18:30'!
settingSelector

	^#setSelectedDirectoryTo:! !

FileDirectoryWrapper subclass: #SuperSwikiDirectoryWrapper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Explorer'!

!SuperSwikiDirectoryWrapper commentStamp: '<historical>' prior: 0!
The super swiki does not at present have subdirectories!

!SuperSwikiDirectoryWrapper methodsFor: 'as yet unclassified' stamp: 'RAA 2/2/2001 08:28'!
contents

	^#()		"we have no sundirectories"! !

!SuperSwikiDirectoryWrapper methodsFor: 'as yet unclassified' stamp: 'RAA 2/2/2001 08:28'!
hasContents

	^false		"we have no sundirectories"! !

!ListItemWrapper class methodsFor: 'as yet unclassified' stamp: 'RAA 3/30/1999 18:28'!
with: anObject

	^self new setItem: anObject! !

!ListItemWrapper class methodsFor: 'as yet unclassified' stamp: 'RAA 3/31/1999 16:44'!
with: anObject model: aModel

	^self new setItem: anObject model: aModel! !

!ListItemWrapper methodsFor: 'as yet unclassified' stamp: 'RAA 3/30/1999 18:13'!
acceptDroppingObject: anotherItem

	^item acceptDroppingObject: anotherItem! !

!ListItemWrapper methodsFor: 'converting' stamp: 'RAA 3/30/1999 18:17'!
asString

	^item asString! !

!ListItemWrapper methodsFor: 'accessing' stamp: 'RAA 7/21/2000 10:59'!
balloonText

	^nil! !

!ListItemWrapper methodsFor: 'as yet unclassified' stamp: 'RAA 3/31/1999 12:25'!
canBeDragged

	^true! !

!ListItemWrapper methodsFor: 'accessing' stamp: 'RAA 3/31/1999 16:32'!
contents

	^Array new! !

!ListItemWrapper methodsFor: 'as yet unclassified' stamp: 'RAA 3/31/1999 16:24'!
handlesMouseOver: evt

	^false! !

!ListItemWrapper methodsFor: 'accessing' stamp: 'RAA 4/1/1999 20:09'!
hasContents

	^self contents isEmpty not! !

!ListItemWrapper methodsFor: 'as yet unclassified' stamp: 'nice 5/8/2011 10:36'!
hasEquivalentIn: aCollection

	^aCollection anySatisfy: [ :each | 
		each withoutListWrapper = item withoutListWrapper]! !

!ListItemWrapper methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:47'!
highlightColor
	"You can override the current theme's default with a custom value."

	^ nil! !

!ListItemWrapper methodsFor: 'accessing' stamp: 'dgd 9/26/2004 18:22'!
icon
	"Answer a form to be used as icon"
	^ nil! !

!ListItemWrapper methodsFor: 'accessing' stamp: 'mt 3/6/2015 16:39'!
item
	^ item! !

!ListItemWrapper methodsFor: 'accessing' stamp: 'mt 3/13/2015 07:37'!
itemName

	^ self item asString! !

!ListItemWrapper methodsFor: 'accessing' stamp: 'mt 3/7/2015 10:16'!
model
	^ model! !

!ListItemWrapper methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:06'!
preferredColor
	"You can override the current theme's default with a custom value."
	
	^ nil! !

!ListItemWrapper methodsFor: 'as yet unclassified' stamp: 'RAA 4/4/1999 17:58'!
sendSettingMessageTo: aModel

	aModel 
		perform: (self settingSelector ifNil: [^self])
		with: self withoutListWrapper
! !

!ListItemWrapper methodsFor: 'initialization' stamp: 'mt 11/10/2015 15:09'!
setItem: anObject

	item := anObject.! !

!ListItemWrapper methodsFor: 'initialization' stamp: 'mt 3/29/2015 14:08'!
setItem: anObject model: aModel

	model := aModel.
	self setItem: anObject.! !

!ListItemWrapper methodsFor: 'as yet unclassified' stamp: 'RAA 3/31/1999 21:31'!
settingSelector

	^nil! !

!ListItemWrapper methodsFor: 'as yet unclassified' stamp: 'RAA 3/31/1999 16:32'!
wantsDroppedObject: anotherItem

	^false! !

!ListItemWrapper methodsFor: 'converting' stamp: 'RAA 3/31/1999 12:13'!
withoutListWrapper

	^item withoutListWrapper! !

ListItemWrapper subclass: #MorphListItemWrapper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!MorphListItemWrapper methodsFor: 'converting' stamp: 'dgd 9/26/2004 18:26'!
asString
	"Answer the string representation of the receiver"
	^ item externalName! !

!MorphListItemWrapper methodsFor: 'accessing' stamp: 'nice 1/1/2010 22:12'!
contents
	"Answer the receiver's contents"

	| tentative submorphs |
	tentative := item submorphs
				collect: [:each | each renderedMorph].

	submorphs := Preferences noviceMode
				ifTrue: [
					tentative
						reject: [:each |
							each isSystemWindow
								or: [each isDockingBar
								or: [(each isKindOf: HaloMorph)
								or: [(each hasProperty: #morphHierarchy)
								or: [each isFlapOrTab
								or: [each isObjectsTool]]]]]]]
				ifFalse: [
					tentative
						reject: [:each | each isKindOf: HaloMorph]].

	^ submorphs
		collect: [:each | self class with: each]! !

!MorphListItemWrapper methodsFor: 'accessing' stamp: 'dgd 7/28/2005 13:03'!
icon
	"Answer a form to be used as icon"
	^ item iconOrThumbnailOfSize: ((Preferences tinyDisplay ifTrue: [16] ifFalse: [28]))! !

ListItemWrapper subclass: #MorphWithSubmorphsWrapper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Explorer'!

!MorphWithSubmorphsWrapper commentStamp: 'ls 3/1/2004 17:32' prior: 0!
Display a morph in a SimpleHierarchicalListMorph, and arrange to recursively display the morph's submorphs.  The "item" that is wrapped is the morph to display.!

!MorphWithSubmorphsWrapper methodsFor: 'hierarchy' stamp: 'ls 3/1/2004 17:34'!
contents
	^item submorphs collect: [ :m |
		self class with: m ]! !

ListItemWrapper subclass: #ObjectExplorerWrapper
	instanceVariableNames: 'itemName parent'
	classVariableNames: 'ShowContentsInColumns'
	poolDictionaries: ''
	category: 'Morphic-Explorer'!

!ObjectExplorerWrapper commentStamp: '<historical>' prior: 0!
Contributed by Bob Arning as part of the ObjectExplorer package.
!

!ObjectExplorerWrapper class methodsFor: 'preferences' stamp: 'mt 3/6/2015 12:07'!
showContentsInColumns
	<preference: 'Use columns in object explorer'
		categoryList: #(Tools exploring)
		description: 'If enabled, the object explorer will show key in the first column and values in the second column.'
		type: #Boolean>
	^ ShowContentsInColumns ifNil: [true]! !

!ObjectExplorerWrapper class methodsFor: 'preferences' stamp: 'mt 3/6/2015 12:07'!
showContentsInColumns: aBoolean

	ShowContentsInColumns := aBoolean.! !

!ObjectExplorerWrapper class methodsFor: 'as yet unclassified' stamp: 'RAA 6/21/1999 10:50'!
with: anObject name: aString model: aModel

	^self new 
		setItem: anObject name: aString model: aModel! !

!ObjectExplorerWrapper class methodsFor: 'as yet unclassified' stamp: 'nk 7/24/2003 09:16'!
with: anObject name: aString model: aModel parent: aParent

	^self new 
		setItem: anObject name: aString model: aModel parent: aParent
! !

!ObjectExplorerWrapper methodsFor: 'converting' stamp: 'mt 3/7/2015 10:34'!
asString
	| explorerString label separator |
	explorerString := 
		[self object asExplorerString]
			on: Error 
			do: ['<error: ', self object class name, ' in asExplorerString: evaluate "' , self itemName , ' asExplorerString" to debug>'].
	(explorerString includes: Character cr)
		ifTrue: [explorerString := explorerString withSeparatorsCompacted].

	label := self itemName ifNil: [''].
	(label includes: Character cr)
		ifTrue: [label := label withSeparatorsCompacted].
	 
	separator := self class showContentsInColumns
		ifTrue: [String tab]
		ifFalse: [label ifEmpty: [''] ifNotEmpty: [': ']].

	^ '{1}{2}{3}' format: {label. separator. explorerString}! !

!ObjectExplorerWrapper methodsFor: 'accessing' stamp: 'mt 3/30/2015 10:12'!
contents

	(self object respondsTo: #explorerContents) ifTrue: [
		^ self object explorerContents
			do: [:wrapper | wrapper parent: self];
			yourself].
	"For all others, show named vars first, then indexed vars"
	^(self object class allInstVarNames asOrderedCollection withIndexCollect: [:each :index |
		self class
			with: (self object instVarAt: index)
			name: each
			model: self object
			parent: self]) ,
	((1 to: self object basicSize) collect: [:index |
		self class
			with: (self object basicAt: index)
			name: index printString
			model: self object
			parent: self])! !

!ObjectExplorerWrapper methodsFor: 'accessing' stamp: 'mt 3/7/2015 09:57'!
hasContents

	^ self object hasContentsInExplorer
	
! !

!ObjectExplorerWrapper methodsFor: 'accessing' stamp: 'mt 6/28/2016 11:33'!
icon
	"Answer a form to be used as icon"
	^ Preferences visualExplorer
		ifTrue: [([self object iconOrThumbnailOfSize: 12] on: Error do: [nil])
			ifNil: [self class showContentsInColumns
				ifTrue: [ToolIcons iconNamed: #blank] 
				ifFalse: [nil]]]
		ifFalse: [nil]! !

!ObjectExplorerWrapper methodsFor: 'accessing' stamp: 'mt 3/7/2015 09:56'!
itemName

	^ self item key! !

!ObjectExplorerWrapper methodsFor: 'accessing' stamp: 'mt 3/7/2015 09:54'!
object

	^ self item value! !

!ObjectExplorerWrapper methodsFor: 'accessing' stamp: 'mt 3/7/2015 10:01'!
object: anObject

	self item value: anObject.! !

!ObjectExplorerWrapper methodsFor: 'accessing' stamp: 'mt 3/7/2015 10:11'!
parent
	^ parent! !

!ObjectExplorerWrapper methodsFor: 'accessing' stamp: 'mt 3/30/2015 10:13'!
parent: aWrapper
	
	parent := aWrapper.! !

!ObjectExplorerWrapper methodsFor: 'monitoring' stamp: 'mt 3/7/2015 12:05'!
refresh
	"hack to refresh item given an object and a string that is either an index or an instance variable name."
	
	self parent ifNil: [^ self].
	[ | index |
		(model class allInstVarNames includes: self itemName)
			ifTrue: [ self object: (model instVarNamed: self itemName) ]
			ifFalse: [ index := self itemName asNumber.
				(index between: 1 and: model basicSize) ifTrue: [ self object: (model basicAt: index)]]
	] on: Error do: [ :ex | self object: nil ]! !

!ObjectExplorerWrapper methodsFor: 'initialization' stamp: 'mt 3/7/2015 10:12'!
setItem: anObject name: aString model: aModel

	self setItem: aString -> anObject model: aModel.! !

!ObjectExplorerWrapper methodsFor: 'initialization' stamp: 'mt 3/30/2015 10:13'!
setItem: anObject name: aString model: aModel parent: itemParent

	self parent: itemParent.
	self setItem: anObject name: aString model: aModel! !

ListItemWrapper subclass: #PluggableListItemWrapper
	instanceVariableNames: 'string getContentsSelector getStringSelector hasContentsSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Explorer'!

!PluggableListItemWrapper commentStamp: 'ar 10/14/2003 23:51' prior: 0!
luggableListItemWrapper makes it more easy for clients to use hierarchical lists. Rather than having to write a subclass of ListItemWrapper, a PluggableListItemWrapper can be used to provide the appropriate information straight from the model:
	string - an explicit string representation (contrary to the 'item' which contains any kind of object)
	getStringSelector - a message invoked to retrieve the sting representation of its item dynamically from its model (when a constant representation is undesirable)
	hasContentsSelector - a message invoked in the model to answer whether the item has any children or not.
	getContentsSelector - a message invoked in the model to retrieve the contents for its item.

All callback selectors can have zero, one or two arguments with the item and the wrapper as first and second argument.!

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 23:39'!
asString
	string ifNotNil:[^string].
	getStringSelector ifNil:[^super asString].
	^self sendToModel: getStringSelector
! !

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 21:48'!
contents
	getContentsSelector ifNil:[^#()].
	^self sendToModel: getContentsSelector.! !

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 21:49'!
getContentsSelector
	^getContentsSelector! !

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 21:50'!
getContentsSelector: aSymbol
	self validateSelector: aSymbol.
	getContentsSelector := aSymbol.! !

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 21:48'!
getStringSelector
	^getStringSelector! !

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 21:49'!
getStringSelector: aSymbol
	self validateSelector: aSymbol.
	getStringSelector := aSymbol.! !

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 21:53'!
hasContents
	hasContentsSelector ifNil:[^super hasContents].
	^self sendToModel: hasContentsSelector
! !

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 21:49'!
hasContentsSelector
	^hasContentsSelector! !

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 21:49'!
hasContentsSelector: aSymbol
	self validateSelector: aSymbol.
	hasContentsSelector := aSymbol.! !

!PluggableListItemWrapper methodsFor: 'printing' stamp: 'ar 10/11/2003 23:21'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut:$(; nextPutAll: self asString; nextPut:$).! !

!PluggableListItemWrapper methodsFor: 'private' stamp: 'ar 10/11/2003 21:47'!
sendToModel: aSelector
	aSelector numArgs = 0 
		ifTrue:[^model perform: aSelector].
	aSelector numArgs = 1 
		ifTrue:[^model perform: aSelector with: item].
	aSelector numArgs = 2 
		ifTrue:[^model perform: aSelector with: item with: self].! !

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 23:39'!
string
	^string! !

!PluggableListItemWrapper methodsFor: 'accessing' stamp: 'ar 10/11/2003 23:39'!
string: aString
	string := aString! !

!PluggableListItemWrapper methodsFor: 'private' stamp: 'ar 10/11/2003 21:50'!
validateSelector: aSymbol
	(aSymbol numArgs between: 0 and: 2) ifFalse:[^self error: 'Invalid pluggable selector'].! !

Object subclass: #MenuIcons
	instanceVariableNames: ''
	classVariableNames: 'Icons TranslatedIcons TranslationLocale'
	poolDictionaries: ''
	category: 'Morphic-Menus'!

!MenuIcons commentStamp: 'sd 11/9/2003 14:09' prior: 0!
I represent a registry for icons.  You can see the icons I contain using the following script:

| dict methods |
dict := Dictionary new. 
methods := MenuIcons class selectors select: [:each | '*Icon' match: each asString].
methods do: [:each | dict at: each put: (MenuIcons perform: each)].
GraphicalDictionaryMenu openOn: dict withLabel: 'MenuIcons'!

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/5/2004 13:37'!
backIconContents
	"Private - Method generated with the content of the file /home/dgd/back.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAADRElEQVRIie3WW4hVVRgH8N/e+5y56aCizpnR
jJzUCNFqLhqEWGlQDz0mGEJBCSL1Ur1YZA9BSIZGZg9dVBoo0B5D6qGwOzVaairZiJiKZ0hs
1NE5c2bm7N3DmXF2zQyO5lv+YbFZa/2/77+/y9prcxNpNFnoHi3/xUU4LtYiszX5WGC/ULsW
m69XMLgqo8UzEhtRlVpNJJb62Tc3TnCxnH7bBR65sra0nt8vkC9Ah9Ay7U5d2Z+jUrUaoRqx
Hr/qGp9gk0cFPsB0UFfFugUsybH1N7YfS7O7UTE4/u2vgDze02ijXUrRCLFmzwtswwRhwGO3
Wb55tcWNzQ75g7+KfNmZtqhEZoyXz2IKlutSIe+LNCnQ7A08NxRVuKHVywvXWG+lDXZ5SRs9
A7Qd53g3pSQlG1GboTY7PKojNh/hXBFK+s0oC85XocoOrASzJ6p9+wE7c69I8K7PfKrdgBK4
XYNWc+VMFgwGlkgU9Svo06OooOiksw7u3MPrhwySVmVAlTasAAummPzWg16oXWGdD+13fESe
ulxSZ5I7zTJHg1tNVxIr6FNQ1KlLuw7fOkJ9dTqHDwWarBJoA7lqPlqiflJO58gGuz78cJZn
fxya/ZQRutdQKZY1MKlifGJJzEA/A32UUs+4NDhispVcjtMRdmeU7BBag8iJM5yOmDmPYIwj
2pXn5MGy2LhQl3pJF0O/2CuwBXzfS0cnR7+j9/IoUSWcPnINYuhNcRN/DnXpRFUOoFFNwKYa
wgJ1s2mYS1TuLedOceLAsIMwIsoSZEhC4oCBgD70xpzB9ktcGExr4PHhvLVqFPsaM02NeC2L
XjKVTJtVdp7vKNfufB1bznO2RLE03lhj5P5ZqBZ3SHyFnFsyrM+SFEaabqrm6CjrY6Mf79tn
7cjOaDVfbA+mmRzyajUVqXpGE3g6NQ88IXZYKFISCUVioUAfetDjolOOKZbpo6HV3WK70SCL
F2uZ0V3eOzCFd64cm8/t8/C1hDr29dSqXuwT3AfuyjIvy66etPVT9tp2YwShWRZvYu0ouydd
MG8oVePFyOspjbxY3m4znMD9hm/90xJPOuzYmLZj4Oq/GENYZKqS1QKxPlsdNMqX4Sb+D/gb
LOgMCUjhiw4AAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'icons creation' stamp: 'dgd 9/22/2004 18:28'!
base64ContentsOfFileNamed: aString 
	"Private - convenient method"
	
	| file base64Contents |
	file := FileStream readOnlyFileNamed: aString.
	base64Contents := (Base64MimeConverter mimeEncode: file binary) contents.
	file close.
	^ base64Contents! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'KR 5/4/2006 21:37'!
blankIcon
	^self blankIconOfWidth: 16.
! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/13/2004 20:01'!
blankIconOfWidth: aNumber 
	^ Icons
		at: ('blankIcon-' , aNumber asString) asSymbol
		ifAbsentPut: [Form extent: aNumber @ 1 depth:8]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
configurationIcon
	"Private - Generated method"
	^ Icons
			at: #'configuration'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self configurationIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/5/2004 13:47'!
configurationIconContents
	"Private - Method generated with the content of the file /home/dgd/configuration.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0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'! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 10/12/2004 13:33'!
confirmIcon
	"Private - Generated method"
	^ Icons
			at: #'confirm'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self confirmIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 10/12/2004 13:33'!
confirmIconContents
	"Private - Method generated with the content of the file /home/dgd/confirm.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAACsElEQVR4nK3WvW8URxjH8c+CDxtsYg6FCJDy
IqUIWKAUiCJFWuiRIqWCKn9AynRJnS4VShEpEimSBlGkiKBAGIkUWIoUBVtC4IvBb2BsH8fd
+fD5NsWM7VO43bXPPNJoZnafme/MPL95oT8r4RCS3TbcTYNSibEynw1xusXiKyp1FvAQT5G+
LeB7H/HdCFdwcLNRStri1gK/1HiMP7G+V+ChT7hd4vwRXMB+vIxpEs+pzvB1jakIzZzp/iLa
x3w/yKWL+DECP8cpLGEULxgqcXaZ8ZR6HEtfwOQkv51h8CqGun68E6ETgoJmKadM1HkmxLWn
7cujneCDhMNfZjiOYgzlWB/mUxzO6zMXOBpjfDTHp2xbJSlN1PoGTlFJmZ3K8XmCf2O5yTRW
+gZCm2s/d3XabY9wM+YbTC5yL8N1ywpVWma8zRe/8+5BDGAef+BbVNBhvca948yfYOFZUGpP
y9uH+87yVco3+HDT+f8bLMVGHHmCOpenuY5XvTodyKKdY6jF1UFB/idj5yuCKgZwTDhQ4Q5W
6cyyJgh4d8AJmmfwPq5FWBXLoWO1MBt/xe8v0WH6dVBqT1guMDLW2nG/JzgSU7fD/TjrFOs8
iMVMYJFKmxs5P+eF9VuO9Qb/hExms1xgQrOd839T/5vAaphh/xs/pZEHnLMtpA7z9VDMXM5C
oIIZLgliaduKH3uZIZpZt2kqXAkvYr0V4rdnYCMr+nMBshW/2g4UWghMc5Z0OubLwa+6wqyg
0E7fwDyVzgkbv4X2DpezEIhGqveQl2zHb22HgikEJuGYeuMZ1hH24GSoNle3H06LewLGG/yN
Y+MJxvEaq/xQDaC/QzXf8s5SKc0ElzCIAzE1hLVb48ZT7goamimC7QT4K2Yq4VkzkjDSYThh
uEO7wk/CZfEgr59u+w8LnvWEFNR9UgAAAABJRU5ErkJggg=='! !

!MenuIcons class methodsFor: 'icons creation' stamp: 'dgd 4/20/2006 12:31'!
createIconMethodsFromDirectory: directory 
	"
	Preferences disable: #showWorldMainDockingBar. 
	MenuIcons createIconMethodsFromDirectory: '/home/dgd/'. 
	Preferences enable: #showWorldMainDockingBar.
	"
	| iconContentsSourceTemplate iconSourceTemplate normalSize smallSize |
	iconContentsSourceTemplate := '{1}IconContents
	"Private - Method generated with the content of the file {2}"
	^ ''{3}'''.
	iconSourceTemplate := '{1}Icon
	"Private - Generated method"
	^ Icons
			at: #''{1}''
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self {1}IconContents readStream) ].'.
	""
	normalSize := #('back' 'configuration' 'confirm' 'forward' 'fullScreen' 'help' 'home' 'jump' 'objectCatalog' 'objects' 'paint' 'project' 'publish' 'squeak' 'volume' 'window' 'open' 'loadProject' ).
	smallSize := #('smallExport' 'smallAuthoringTools' 'smallDebug' 'smallBack' 'smallCancel' 'smallConfiguration' 'smallCopy' 'smallCut' 'smallDelete' 'smallDoIt' 'smallExpert' 'smallFind' 'smallForward' 'smallFullScreen' 'smallHelp' 'smallHome' 'smallInspectIt' 'smallJump' 'smallLanguage' 'smallNew' 'smallObjectCatalog' 'smallObjects' 'smallOk' 'smallOpen' 'smallPaint' 'smallPaste' 'smallPrint' 'smallProject' 'smallPublish' 'smallQuit' 'smallRedo' 'smallRemoteOpen' 'smallSave' 'smallSaveAs' 'smallSelect' 'smallSqueak' 'smallUndo' 'smallUpdate' 'smallVolume' 'smallWindow' 'smallLeftFlush' 'smallCentered' 'smallJustified' 'smallRightFlush' 'smallFonts' 'smallLoadProject' ).
	normalSize , smallSize
		do: [:each | 
			| png base64 contentsSelector selector | 
			png := directory , each , '.png'.
			base64 := self base64ContentsOfFileNamed: png.
			""
			contentsSelector := (each , 'IconContents') asSymbol.
			((self respondsTo: contentsSelector)
					and: [(self perform: contentsSelector)
							= base64])
				ifFalse: [| contentsSource | 
					contentsSource := iconContentsSourceTemplate format: {each. png. base64}.
					self class compile: contentsSource classified: 'private - icons'].
			""
			selector := (each , 'Icon') asSymbol.
			(self respondsTo: selector)
				ifFalse: [| source | 
					source := iconSourceTemplate format: {each}.
					self class compile: source classified: 'private - icons']].
	""
	self initializeIcons! !

!MenuIcons class methodsFor: 'menu decoration' stamp: 'dgd 4/3/2006 14:45'!
decorateMenu: aMenu 
	"decorate aMenu with icons"

	| maxWidth |

	Preferences menuWithIcons ifFalse: [^ self].
	Preferences tinyDisplay ifTrue:[^ self].

	maxWidth := 0.

	aMenu items do: [:item | 
		item icon isNil ifTrue: [
			| icon | 
			icon := self iconForMenuItem: item.
			icon isNil ifFalse: [
				item icon: icon.
				maxWidth := maxWidth max: item icon width.
			]
		]
		ifFalse: [
			maxWidth := maxWidth max: item icon width
		].

		item hasSubMenu ifTrue: [
			self decorateMenu: item subMenu.
		].
	].

	maxWidth isZero ifFalse: [
		aMenu addBlankIconsIfNecessary: (self blankIconOfWidth: maxWidth).
	].
! !

!MenuIcons class methodsFor: 'import/export' stamp: 'sd 11/9/2003 16:16'!
exportAllIconsAsGif
	"self exportAllIconsAsGif"

	| sels | 
	sels := self class selectors select: [:each |  '*Icon' match: each asString].
	sels do: [:each | self exportIcon: (MenuIcons perform: each) asGifNamed: each asString].
! !

!MenuIcons class methodsFor: 'import/export' stamp: 'nk 2/16/2004 13:38'!
exportAllIconsAsPNG
	"self exportAllIconsAsPNG"

	| sels | 
	sels := self class selectors select: [:each |  '*Icon' match: each asString].
	sels do: [:each | self exportIcon: (MenuIcons perform: each) asPNGNamed: each asString].
! !

!MenuIcons class methodsFor: 'import/export' stamp: 'sd 11/11/2003 11:36'!
exportIcon: image asGifNamed: aString
	"self exportIcon: self newIcon asGifNamed: 'newIcon'"

	| writer |
	writer := GIFReadWriter on: (FileStream newFileNamed: aString, '.gif').
	[ writer nextPutImage: image]	
		ensure: [writer close]! !

!MenuIcons class methodsFor: 'import/export' stamp: 'nk 2/16/2004 13:38'!
exportIcon: image asPNGNamed: aString
	"self exportIcon: self newIcon asPNGNamed: 'newIcon'"

	| writer |
	writer := PNGReadWriter on: (FileStream newFileNamed: aString, '.png').
	[ writer nextPutImage: image]	
		ensure: [writer close]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/5/2004 13:34'!
forwardIconContents
	"Private - Method generated with the content of the file /home/dgd/forward.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAADXUlEQVRIie3WX2iVZRwH8M97znv2f1o2Fy4r
hxMvhhCeTewmJcuEbiLLP0GECoJWEIlYUBoR3QTVhRRIXffvJupiBF4oRCZukq5Na4LhlKnT
6XE7c/P8ebs4O8fNlh3nVdEPHnif5/d8f9/3+33e53le/usRvyv0UknzzDXgQrmQ2IzJkt4V
6BT4xVJfWKa5HFgwI7J27fIOmerQmMBOnfbeGWGrChXmoVZeSlZKjzQisEyznA4s1lhFWwMd
Z4tZIh0SNjk8vc03CZO2Yzcap32R6WLLIrYt5sgl3jvOudFiZlBki6O+vxVSsKTNK9iLurLJ
4PkFLKy38YE1Wp5b7uT1P+hJEakV2KDJiAGHpipMSmAIdeoT7GwlGzGaJT2pjeUK49l8Aflw
HZtaqI57xwv22OhD33qj+2OZXYe5OFaYF/nIUTtMmB5ot0LeAbB5kSXbn9LqIRXCUksIBZOE
Dxl2RJ8T+ktjKy2x1RpN5lh7aY/Lr+6n71qR9EvjXtLjRijnyVKt+TVOOqtNiyc84kENQnGJ
Cef7XXLKgGtG9Ruc4u4B3Q7o1mKelxue9ennCYOv76fzMoENqmXxYiDpK6wD7y9ldVPZS3i7
mKPe0PAQ6w/etDfwTIjh0qx0mounuXEdAUFAECMekqgkrCCsJFFReA7+/twYMkx9guVz+W7C
+rwVIVKlWUGK/vPlSQgC5rfSuGD6fD7H6V76SvXGBPbFBJM2aCpTHhlEEQO/E+X/mhu+TO9B
jl3lxETNyG5dTsbdb1xgK+gaY3YdFdXkagiqCWuIVxFWEYbEYhMkUUFFWEntPYWimXHOdHO2
l5Fa3koVyQ5aaJteUYCYpAtoKFtdTYwd9zF/sGBtY3OBfOgcuSz5Wbw2zHgEXVilq7B0cUSa
jKEdNWURZiL6MqyMCkrTVxhNFZTnZ7MrzWgeesSt0ulKETr1GFviXjUqkJAVE8rLysnIqZKT
0SzmR1SCfXVEIzfxV+t5e4QbEZzCY7oMTKa48+sp6TNsAetn8fi1wvY4Vssnw8Vb41c8rcuZ
W+EzIVyNH0r99gT9Oc6XvtavZWx2XHo6+Mwu4DY/iTx6y2hO5E1HfXA76Ex/MTbh50n930RW
/xMZM1VYxLZZK2/MQh2+kbuLWv/Hvyj+BJaPHpqZ0oo8AAAAAElFTkSuQmCC'! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'mt 5/10/2015 13:39'!
fugueBroomIcon

	^ Icons
			at: #broom
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self fugueBroomIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'mt 5/10/2015 13:39'!
fugueBroomIconContents
	"Fugue Icons
	(C) 2013 Yusuke Kamiyamane. All rights reserved.

	These icons are licensed under a Creative Commons
	Attribution 3.0 License.
	<http://creativecommons.org/licenses/by/3.0/>"

	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAoNJREFUeNqkU11Ik1EYfo6bm7WZW2r+MDNTZ9ZQcbYws0iN
piZEBF4UXWgXCXURXQQiBTYhgi4qRo0y+gEpNImauKVkhkzDNrxoCjrNZSSapnNLv+n2nc5H
GmFJgQcezuGc93nPc877vIRSivUMYqwkEFIIeQQQsnzA5rP1FEe0BKXZP894Bm6AQiKGPsjD
sLSEmpD/uCSX8e4JYOv9LJF+noNhb/ExrW8BBvE/yClmO824YbpaKUgwG6sjvy8gQVdwQNv4
qNkeCK6tQMHiz7H5kMs1fKe1Rw7/xARKTl8+Gq4hX5496bR/nkINz8MiKAhliGCIYohjRF2Y
LCJnW2ZelrM8I6mlxYzCQj1arX4cTBzC4ZMXyhYD11soD074E3KrglwJkyl3S2WbZBvCI8Oj
E1JjlDGqqCXvojjg4bB5ZwKaO0NRUFAKq/U5CpOGIRKJYG2638AuOxHCZLT5PDP+NF3Jvvhk
bSY3y8eOdA+IvV9nsFGlxLf+MZRle2E2NyI/vxQvnXGYc88hEABhgEifSdw2O521dfSmyyji
ZRFyxKSqIGK1Gu8bgTI5FgEfh3TFPBra+6BW70H5+dqGvBTUs2d8XKmCJVtDFqanHRd35GqK
hxz9EFMRtmakwDc1C8oFoUjcglO7PNCdKW9i8beZ/C7BMr+qwDY6+/rpTeO1h3ZFlAJJOWqM
D36CZ3IG0mg5Jj+48fjp3bdV+dTEwrtWeKt9YEnbTvy21y+q1ercIp59UHJWGsYcQ7B0tHUz
99WFSdH+O+EPIzElHa5RKnWN2pQVVce1gz1OvHn3vpcPolYiwavV8Ws50SKVgD4wNdUt98il
EALLX5tp3d243gQ/BBgABJ4L2+7frOgAAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'mt 5/10/2015 13:12'!
fugueDocumentClockIcon

	^ Icons
			at: #documentClock
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self fugueDocumentClockIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'mt 5/10/2015 13:15'!
fugueDocumentClockIconContents
	"Fugue Icons
	(C) 2013 Yusuke Kamiyamane. All rights reserved.

	These icons are licensed under a Creative Commons
	Attribution 3.0 License.
	<http://creativecommons.org/licenses/by/3.0/>"

	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAhNJREFUeNqMU11LG1EQPbvZfEgaSKKBtFX71BcFjT8hCb4I
Pjb1qc/9F4og9AcU6rPgk9KCb0Uw0D9gCsWAaSJBoSSp+Wg2rbp3d3vuDZtuWS0dmL3DvTNn
5szMaq7rQkppa2udx2P8W04cx/nq2Dbe7+yoC8N74cOTvc3N3Ts+WlIdBzbBvQSVSgV7x8ev
aWrUmhenTwBsW3NkkGUpvbu5wa/RCKZpYjgcqvNwe3uXfkW6Pw8A2ASQ0Dq/4VAIsXAYEcOA
oevQeNftdvF2fx+FhYV3X05PX3pxhq8CXQbrCgYKZCJCoFAoKDqJRAIfK5VeAEDICnwA94EI
UvR8AwC2EBMKI3Kv1+vo9/sqa5xZs3NzMCIRRUf6BnsghKLwk8Gf2XGX/KdiMSytrMCIRlE7
O4O4vUVoDKAHAAQv1XzOzxFlxqlkEh8ODtBh9+PpNBKpFK4uLhQt8QCA6sH3TgdxBs8yqNVq
4dnMDJ4yODk9jd71NcKcivBR+NMDy1IVWNwBnU5S2+02Xm1sIJvNYnVtTS7bmAJ9g1OQFPiY
YrYfrMLNZHB4dDQeMRtZrVZhDgZqDR+mwDOXy+Fbs4lGrSabpVTazUYDS8vLakr3UuDaKjvN
ClaLRZTLZXyiSpmdn0c+n1dvms/3L4BBrxf17AzLf1EqqR2Qv5J3Sioae+P3nQCMTPPy0eLi
G/yHEOTSs38LMADvCBAJ/QIAlQAAAABJRU5ErkJggg=='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'mt 5/10/2015 13:35'!
fugueUserSilhouetteQuestionIcon

	^ Icons
			at: #userSilhouetteQuestion
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self fugueUserSilhouetteQuestionIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'mt 5/10/2015 13:36'!
fugueUserSilhouetteQuestionIconContents
	"Fugue Icons
	(C) 2013 Yusuke Kamiyamane. All rights reserved.

	These icons are licensed under a Creative Commons
	Attribution 3.0 License.
	<http://creativecommons.org/licenses/by/3.0/>"

	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9i
ZSBJbWFnZVJlYWR5ccllPAAAAgBJREFUeNqkk7+LGlEQx2f9rXiLvwhYHYQENaRJ4RJFsFMQ
bLQOBCRFSBGLtJciYJc21wRsUqS7v0IRlSNFjjt/d6IJaKEXFV1v853H3rHZmKTIwIfdN+87
s/PmzUqappEkSWSyJ+AFeKSvL8FH8MUoErEHEqQ8Hs/7eDz+MBwOy+yYTCaLdrvdX61Wb7Cs
GRNY6Hd7qShKLBQKBfb7vU1VVVswGAwgYYz3zGJzgjiI+ny+o+12S8xutxP4/f4j3tM1v57D
YF/x9etsNqvN5/M7arWalslkNN5jjTHWZqrgHNy3WCxUqVREb5LJJKVSKYpEItRsNiVd88cj
PG+1Wg0EqqPRiIbDIS0WC5pOpzSbzVTeY83fErCd9nq9Gb9Eo1FKp9NUrVb56+w7/VcT2c4G
g8G13W6nUqlEnU6HxuMxV8PnPzOLbQfWr4CVF+VyWfTB7XaT7nsNPgD10C0o4DPoF4tFLZ/P
axggrV6vi/dCocDCvq5RzLfwDNP3DpMXQLfl9XpNVquVcBTCNQoB+3K53INut3sPiZ9iKt/C
/el2lC8SiUQM12dxOByiZH4ul0sR7PV6xVBtNhsxVJjQm0ajcYXYx7cVHLtcLo4np9MpEvBT
lmVRJsZZVMT9YA183PxjYw9OwA34Ab6BoT5x5zoXYAS+6xrWntz9jf9jPwUYAKt/9B8pRO/y
AAAAAElFTkSuQmCC'! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
fullScreenIcon
	"Private - Generated method"
	^ Icons
			at: #'fullScreen'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self fullScreenIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
fullScreenIconContents
	"Private - Method generated with the content of the file /home/dgd/fullScreen.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0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'! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'mt 8/2/2016 16:43'!
fullscreenWireframeIcon
	"(c) 2016, Plainicon on www.flaticon.com
	http://www.flaticon.com/free-icon/square-frame_79288
	CC 3.0 BY http://creativecommons.org/licenses/by/3.0/"
	
	^ Icons
		at: #fullscreenWireframeIcon
		ifAbsentPut: [  (Form
	extent: 16@16
	depth: 32
	fromArray: #( 1543503871 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 1543503871 2298478591 0 0 0 0 0 0 0 0 0 0 0 0 0 0 2298478591 2298478591 0 3254779903 2298478591 2281701375 285212671 0 0 0 0 301989887 2281701375 2298478591 3254779903 0 2298478591 2298478591 0 2298478591 0 0 0 0 0 0 0 0 0 0 2298478591 0 2298478591 2298478591 0 2298478591 0 0 0 0 0 0 0 0 0 0 2298478591 0 2298478591 2298478591 0 251658239 0 0 0 0 0 0 0 0 0 0 251658239 0 2298478591 2298478591 0 0 0 0 0 0 0 0 0 0 0 0 0 0 2298478591 2298478591 0 0 0 0 0 0 0 0 0 0 0 0 0 0 2298478591 2298478591 0 0 0 0 0 0 0 0 0 0 0 0 0 0 2298478591 2298478591 0 0 0 0 0 0 0 0 0 0 0 0 0 0 2298478591 2298478591 0 251658239 0 0 0 0 0 0 0 0 0 0 251658239 0 2298478591 2298478591 0 2298478591 0 0 0 0 0 0 0 0 0 0 2298478591 0 2298478591 2298478591 0 2298478591 0 0 0 0 0 0 0 0 0 0 2298478591 0 2298478591 2298478591 0 3288334335 2298478591 2281701375 285212671 0 0 0 0 301989887 2281701375 2298478591 3288334335 0 2298478591 2298478591 0 0 0 0 0 0 0 0 0 0 0 0 0 0 2298478591 1476395007 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 2583691263 1459617791)
	offset: 0@0) ]! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'mt 8/2/2016 16:09'!
fullscreenWireframeIconColorized: aColor

	^ self fullscreenWireframeIcon
		collectColors: [:c | aColor alpha: c alpha]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/5/2004 13:43'!
helpIconContents
	"Private - Method generated with the content of the file /home/dgd/help.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0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'! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
homeIcon
	"Private - Generated method"
	^ Icons
			at: #'home'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self homeIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/6/2004 17:33'!
homeIconContents
	"Private - Method generated with the content of the file /home/dgd/home.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0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'! !

!MenuIcons class methodsFor: 'menu decoration' stamp: 'sw 3/23/2006 01:07'!
iconForMenuItem: anItem
	"Answer the icon (or nil) corresponding to a given menu item."

	| aKey |
	aKey _ (anItem selector == #undoOrRedoCommand)
		ifTrue:
			['undo (z)' translated]  "Actual wording changes dynamically"
		ifFalse:
			[anItem contents asString].
	^ TranslatedIcons at: aKey asLowercase ifAbsent: [nil]! !

!MenuIcons class methodsFor: 'import/export' stamp: 'sd 11/10/2003 13:06'!
importAllIcons
	"self importAllIcons; initialize"

	| icons |
	icons := FileDirectory default fileNames select: [:each | '*Icon.gif' match: each ].
	icons do: [:icon | self importIconNamed: (icon upTo: $.)] ! !

!MenuIcons class methodsFor: 'import/export' stamp: 'nice 2/24/2013 17:53'!
importIconNamed: aString
	"self importIconNamed: 'appearanceIcon'"
	"will create a method appearanceIconContents holding the string representation of the
	icon, and a method appearanceIcon to access a form built using that string"
	
	| writer image stream |
	writer := GIFReadWriter on: (FileStream fileNamed: './icons/', aString, '.gif').
	image := [ writer nextImage]	
		ensure: [writer close].
	"store string representation"
	stream := String new writeStream.
	stream nextPutAll: aString, 'Contents' ; cr.
	stream nextPutAll: (self methodStart: aString).
	image storeOn: stream.
	stream nextPutAll: self methodEnd.
	MenuIcons class compile: stream contents classified: 'private - icons' notifying: nil.
	"create accessor method"
	stream := String new writeStream.
	stream nextPutAll: aString ; cr.
	stream nextPutAll: (self methodAccessorFor: aString).
	MenuIcons class compile: stream contents classified: 'accessing - icons' notifying: nil.
	^ stream contents! !

!MenuIcons class methodsFor: 'import/export' stamp: 'nice 2/24/2013 17:54'!
importPngIconNamed: aString
	"self importIconNamed: 'appearanceIcon'"
	"will create a method appearanceIconContents holding the string representation of the
	icon, and a method appearanceIcon to access a form built using that string"
	
	| writer image stream |
	writer := PNGReadWriter on: (FileStream fileNamed: './icons/', aString, '.png').
	image := [ writer nextImage]	
		ensure: [writer close].
	"store string representation"
	stream := String new writeStream.
	stream nextPutAll: aString, 'Contents' ; cr.
	stream nextPutAll: (self methodStart: aString).
	image storeOn: stream.
	stream nextPutAll: self methodEnd.
	MenuIcons class compile: stream contents classified: 'private - icons' notifying: nil.
	"create accessor method"
	stream := String new writeStream.
	stream nextPutAll: aString ; cr.
	stream nextPutAll: (self methodAccessorFor: aString).
	MenuIcons class compile: stream contents classified: 'accessing - icons' notifying: nil.
	^ stream contents! !

!MenuIcons class methodsFor: 'class initialization' stamp: 'dgd 9/22/2004 18:27'!
initialize
	"self initialize"
self initializeIcons.
	Smalltalk addToStartUpList: self! !

!MenuIcons class methodsFor: 'class initialization' stamp: 'dgd 9/22/2004 18:27'!
initializeIcons
	"self initialize"
	| methods |
	Icons := IdentityDictionary new.
	methods := self class selectors
				select: [:each | '*Icon' match: each asString].
	methods
		do: [:each | Icons
				at: each
				put: (MenuIcons perform: each)].
	self initializeTranslations.
! !

!MenuIcons class methodsFor: 'class initialization' stamp: 'eem 1/5/2016 17:32'!
initializeTranslations
	"Initialize the dictionary of <translated menu string>-><icon>"

	TranslationLocale := LocaleID current.
	TranslatedIcons := Dictionary new.
	self itemsIcons do:
		[:assoc|
		assoc key do:
			[:str| TranslatedIcons at: str translated asLowercase put: assoc value]]! !

!MenuIcons class methodsFor: 'menu decoration' stamp: 'ul 1/9/2016 12:02'!
itemsIcons
	"answer a collection of associations wordings -> icon to  
	decorate  
	the menus all over the image"
	| icons |
	icons := OrderedCollection new.

	"icons add: #('Test Runner' ) -> self smallTrafficIcon."

	" 
	world menu"
	"icons add: #('previous project' 'go to previous project') -> self smallProjectBackIcon."
	icons add: #('go to next project') -> self smallProjectNextIcon.
	icons add: #('select' ) -> self smallSelectIcon.
	icons add: #('jump to project...' ) -> self smallProjectJumpIcon.
	icons add: #('open...' ) -> self smallOpenIcon.
	icons add: #('appearance...' ) -> self smallConfigurationIcon.
	icons add: #('help...' ) -> self smallHelpIcon.
	"icons add: #('windows...' ) -> self smallWindowIcon."
	icons add: #('changes...' ) -> self smallDocumentClockIcon.
	icons add: #('print PS to file...' ) -> self smallPrintIcon.
	icons add: #('debug...' ) -> self smallDebugIcon.
	icons add: #('export...' ) -> self smallExportIcon.
	icons add: #('save' ) -> self smallSaveIcon.
	"icons add: #('save project on file...' ) -> self smallProjectSaveIcon."
	"icons add: #('save as...') -> self smallSaveAsIcon.
	icons add: #('save as new version') -> self smallSaveNewIcon.
	icons add: #('save and quit' ) -> self smallQuitIcon."
	icons add: #('quit') -> self smallQuitNoSaveIcon.
	"icons add: #('load project from file...' ) -> self smallProjectLoadIcon."
	""
	icons add: #('do it (d)' ) -> self smallDoItIcon.
	icons add: #('inspect it (i)' 'inspect world' 'explore world' 'inspect model' 'inspect morph' 'explore morph' 'inspect owner chain' 'explore' 'inspect' 'explore (I)' 'inspect (i)' 'basic inspect' ) -> self smallInspectItIcon.
	icons add: #('print it (p)' ) -> self smallPrintIcon.
	icons add: #('debug it (D)' ) -> self smallDebugIcon.
	icons add: #('tally it' ) -> self smallTimerIcon.
	""
	icons add: #('copy (c)' 'copy to paste buffer' 'copy text' ) -> self smallCopyIcon.
	icons add: #('paste (v)') -> self smallPasteIcon.
	icons add: #('cut (x)' ) -> self smallCutIcon.
	""
	icons add: #('accept (s)' 'yes' 'Yes' ) -> self smallOkIcon.
	icons add: #('cancel (l)' 'no' 'No' ) -> self smallCancelIcon.
	""
	icons add: #('redo (Z)' ) -> self smallRedoIcon.
	icons add: #('undo (z)' ) -> self smallUndoIcon.
	""
	icons add: #( 'find class... (f)' 'find method...' ) -> self smallSearchIcon.
	icons add: #('find...(f)') -> self smallFindIcon.
	""
	icons add: #('remove' 'remove (x)' 'remove class (x)' 'delete method from changeset (d)' 'remove method from system (x)' 'delete class from change set (d)' 'remove class from system (x)' 'destroy change set (X)' ) -> self smallDeleteIcon.
	icons add: #('add item...' 'new category...' 'new change set... (n)' ) -> self smallNewIcon.
	""
	icons add: #('objects (o)' ) -> self smallObjectCatalogIcon.
	icons add: #('authoring tools...')  -> self smallAuthoringToolsIcon.
	icons add: #('projects...')  -> self smallProjectIcon.
	""
	icons add: #('make screenshot')  -> self smallScreenshotIcon.
	
	""
	icons add: #('leftFlush' ) -> self smallLeftFlushIcon.
	icons add: #('rightFlush' ) -> self smallRightFlushIcon.
	icons add: #('centered' 'set alignment... (u)' ) -> self smallCenteredIcon.
	icons add: #('justified' ) -> self smallJustifiedIcon.
	""
	icons add: #('set font... (k)' 'list font...' 'set subtitles font' 'change font' 'system fonts...' 'change font...' 'default text font...' 'flaps font...' 'eToys font...' 'eToys title font...' 'halo label font...' 'menu font...' 'window-title font...' 'balloon-help font...' 'code font...' 'button font...') -> self smallFontsIcon.
	icons add: #('full screen on') -> self smallFullscreenOnIcon.
	icons add: #('full screen off' ) -> self smallFullscreenOffIcon.
	""
	^ icons! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
jumpIcon
	"Private - Generated method"
	^ Icons
			at: #'jump'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self jumpIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/5/2004 13:34'!
jumpIconContents
	"Private - Method generated with the content of the file /home/dgd/jump.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAACuElEQVRIie3WT2hcVRQG8N+dJtGkZiyiNpUs
0gZRqWhM1GihxU1BMLhx0UVFKiJqXAhuxJV0UxHcFCvuSjbqzo0LqxUXXWgFE5GAUmtTzWia
ltraZuwkk8kcFxOT+Zc0JSIU+sHl8c495/u++9479z5u4HpHuqbsfvdhQNIFwjRGjfnpvxXs
tzcl+9G7QsapCG8Z8+H6BAf0Jj7A7jUZ42jwilGnrl3wIbtSxhG0g2wr2zdxd1Z0tUu5v/ll
hvELzJarKwtR9qTvHVujSQzKpgGn04BID4vMgZ54v/hpTMR0XIx8vBEjIYYqY2pnpBdFGqgZ
pw3KNqPONA2WHESPTIo4+Ki333zXcOuQrTabVwKbbKwkb7lVvLZRPFtD0bPI0YANDZFH7BDe
gxi+Nw0//ZIDngMTpj1lv098bdb8cs3Mn2zOS5cwuRTt0+2oKblVV5gp21Px2FHs2LfdO55f
mtumyyEvN9qOyjuMPehswrWaYITHwWN3tj2TdrjFzTXzg+7Rrq22qFioXFuxrQnXaoKSLRAP
3GavJ0DO+aXpEV8qKC7nz12hkF8WqRL8l6saLQ2CS67Lsjqc8If7vWq3Pmf9Zay+xc5OIJbv
SysyriAYzki60/hFM0MFHzumZMFnRhur8xc4/1tNKE3UcdWh4ZGm5Btw/NzCifLvjjQTguIs
E6NE1equoEpwiasKDW0R3S6n8IKZUubHlpxCf6e8QuPKTh5nfq7W7GGqmyAyXq9vi8Y+nJLL
3KUHfZdGJ81tvclC72KTl+Y58zOT4yxUvazAF6SvqsQZie8cqqdvvpcOyqaSH9ADdt4u9nUS
uVqhikHpI5ysif4aLR70rctrE6SyeW/wuVhsxAy6KhbiDtKUyq5yTs1H6iqb99WPpzaHFe1a
NW8Z6zieqvG/HcCNwuv+xbiB6x//APor/zPsp+AoAAAAAElFTkSuQmCC'! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/22/2004 19:35'!
loadProjectIcon
	"Private - Generated method"
	^ Icons
			at: #'loadProject'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self loadProjectIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/6/2004 17:31'!
loadProjectIconContents
	"Private - Method generated with the content of the file /home/dgd/loadProject.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0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'! !

!MenuIcons class methodsFor: 'import/export' stamp: 'MAD 12/19/2010 20:51'!
methodAccessorFor: iconName
	"Template method for an icon accessor method"

	^ '
	^ Icons
		at: #{1}
		ifAbsentPut: [ self {1}Contents ]' format: {iconName}! !

!MenuIcons class methodsFor: 'import/export' stamp: 'MAD 12/19/2010 20:36'!
methodEnd

	^ ''! !

!MenuIcons class methodsFor: 'import/export' stamp: 'MAD 12/19/2010 20:36'!
methodStart: aString

	^'^ '! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
objectCatalogIcon
	"Private - Generated method"
	^ Icons
			at: #'objectCatalog'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self objectCatalogIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
objectCatalogIconContents
	"Private - Method generated with the content of the file /home/dgd/objectCatalog.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAAD5klEQVR4nN2VW2wUVRjH/+ecmdmd7c7e2qVb
uu1ulrZWoNUURYILSiKoBcRIN+gLUUKMifLiAzHEhHcVtInaxJQYY0g0Jo1CEBXaKBA3RhMp
0Ai10NpUbO32uu3u7GXO8QF6o7NLL5FE/8nJZL75n+/3zZn5zgH+S6r0s+cAVN4z4NqQfLE6
KLcDkBY6hy6D56oOSIVHDjjvC/hYI4BCACW3rznFlgh7oNTL3n5jr7ZuZ1i1MEYq/F62v6pc
fnV4nNfHE6J5iXnnyVnooH++/oI22P5psRBR/5yxdb21D0BFvgSLXdIxVSGHO25k9LLiuYtz
/NtEouN6+n0AXfkSzHzsNRE75ExRTienWVxq6euLGc19MeP8roND35xr8ganHrsd1AKA363i
GSDVG2pDgffKfN6kmTF68aoxDPhv317jQowCwMnzyURPvyEatqgFTo3tuTnE37ob9JZqdr7Y
9PkpMT4RNx2hp1/unm3fsFZpPxCxDwRLpEbNxp6tKpdaHQUkBsC6sDdcpG7GjJ4vziYP9Y8Y
pwAg3osvARAA4l8B9vYbu0zCeWHA8hp/SbrnwDlLeibanh4ajRtmxoSeUhaa1PMIHGlJeYJJ
dLshcHzinN42H8jR1tIa3dPSGjXPIpDIB7FvUlYTTupLKtyRVXUrHpKLJUrdDD80XrEDmAaS
hVadT1rYuu+x5+8/pvgtSExQ9FwbQurvNIrXSPijc0D/6/q4C6eRmvuGJrIBz/goi1AC5yBw
cswwPgaQvdPHCf/F7rEi2taD0Stjv/E0PyEovqMZ7ROtzuqfaNefiiP9VT6WJUDY94cs2shl
rVh0OXziQ9WdCFGpA8BKEz999HBVMniwUGhh676poD2svFnzUalw7bC1TMVMj6dyyj44qrq2
v2a121dQBg+heFhS5K2Sxfu1kdo4KvixO6aIuKJvKQ97QrFL8f50yDirlVqaJI1FhMxdkkZC
cZo6gj5kTdvCAbptt6LO26JWMxkPMjmAmT31ln+TpYnIvC4ukiAE6zVDXaf4pP361UzAVqtA
rVIsdkmpB8z7ULESknM/rCFygQRUT8M2WyrBySt6Z9YtNAHZy2pJlg/wDM+4nrRh4tcURs4k
khDEB5j/NOlRYeQ8Zi6LzGQW6Jyuzkj1Z51qrGhzQVHyRgaGLiih3GWM8HBiiG8EEQOqLX1i
+AImgRxtUUZZ87tW597dik2eHe/iWWybHPyp2zA2zI67HrcGeUbsAEU3Eakfxy5gJFfBuaSW
E/bzO6pzvNdRImLOleIzmye1irLfAQQXm2y28jU+8TD2khukQQDapOCnBzg/CkBfDvD/r38A
0MR5I7Mls/AAAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
objectsIcon
	"Private - Generated method"
	^ Icons
			at: #'objects'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self objectsIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
objectsIconContents
	"Private - Method generated with the content of the file /home/dgd/objects.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAADhElEQVR4nMWWW0wcVRzGv3POXHZhZdldoAud
RkzRQFsaxBpTNcUtIMZaTdoQBa+tsYRYS2mjD/riiyY2pkL6JulTaUgkaayJ8aUoohgfWlOl
GiNqTY1chr1Bmc7uzuX40E4i7A7Nsqt+yTzM+X///++cyXeSAf5n1Ycp+2gDoVcrCVUVwr4s
BXvxXyEFKO0NE5o8UxK05/zVXPdv5JO+Kr5bkBcrCZkAIBUNJgHbFcKSCX8N5+VK1nNYLtVC
oIPFYBEAqCbsq6GSwMN7RE9Ok845apdmr6vc3gxgoRAgBYA0+NZ2UXY1eQlBq+CxBKC5EJgD
FEWASjcP66o6yrwcUIoBNGxAi3J7TePXVlqjwFQxgCAE50+klnQ30xXLwPec2xsU5bZAr4Su
ynJ62Zm9WgwANM4//9k2D4Qo8TUzaYXxJ8vAXi26aFZv6tnSsr/zrsad2tWpb/9yA4Yr2ECk
WW5Sk7ag6Xx8rc2Fqwgbq6dC8nXZd+N9r998XPDEA4SqAtDumNq6+59q7e5/s6OzP5jrALU1
bMacVPj2OnEBwJbVhqykSMBWC2iWgJABfGcCFwHc+Kdn76FDJfpyaQ/n9sLYyOBZAPxWaf/B
J0o/PP1WIPjLNRN7jkdVRolOCBBdtMr0NHl17WjeRq1PH7snOj99LDE9ETRSyzt3NEieD476
KzZvFFb4/pg18eAr6rXZmL2tICAAhMrZ5P310gMfnwgyWcweZ1rAfS/Nx3741egAcClnkvJR
LGlFLk+nhx49Ek2qieyrNXR+2fhz3joN4BLgEt08lZmL2b0/zvgOtPUltNXFhlpR9HrJJue9
GEB0dr4tNbU935hOm1Fn7crvBgyTY1eTDNjY7bAE1yl5KC4lj1siTlGCl1MZjoPvJBa/uJie
ppTXtNzrkcMV9I6ZqLULwHjBsNauvp727iN3A4BSxeYb7hRioTJ6+FaZAHiopoINl/uEiLOw
QuGWF0ZlWWxyA+ipzG/qxJnHACDyXN+TDGTpwvDAOABUBdk5NW69i5t3N6eyPiljdNvUuYE6
t4bGfUcBAO3P9O+wLbvswsjAJ05NjVv73PocrSs0jzz7msKJHRkbGRzOtzd/IAdlnPYGzJmT
efeuB0gAJhH7vdHRUes/AXIC47Ozp5bWA1sXsFBlpTRjWJmuN07OuTVoeqq4wIV4rOPTb67n
/l8EAN3OFAL8G3xJPtRdAsoRAAAAAElFTkSuQmCC'! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/6/2004 17:10'!
openIconContents
	"Private - Method generated with the content of the file /home/dgd/open.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0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'! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
paintIcon
	"Private - Generated method"
	^ Icons
			at: #'paint'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self paintIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
paintIconContents
	"Private - Method generated with the content of the file /home/dgd/paint.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAAEK0lEQVR4nLXWW2xUdRoA8N85M53O2Av0Rgus
tJZCKASvUO+E3aBFN7IbWGNMdqMmPGlMdC/ZGB/U+GI0IfqiPigvGLImXtaYdReMAomu2TXB
FYGFRSUVSgstCKWdGTpzztmHKaQhiEaHLzk5+V++/M53/l9yTsqlibCHNS3kTjB8iYxKLKS1
my39c7JJX0Mm6eHFS+mF3ezYsLY7GXru2uTdh/qSK3PppJtl5zZUU5vPH3ovz65Yv34Nde2S
oMb1S2fBLVUHu+jC0w+t7aT9GtGZgiihrSUnYEnVwRRPLOnM1d7y81uJI6WxEyKhQqEMcVXB
Lhbhdw+u6WB2H2PfiGLKAvliCSarCqb4/fzZtambbr2e2kbGBsRxIBIaz5fgZNXAn5HDPf3L
ZjL3hsrkVIURJvJlAfurBmZYi8Zf9V9FXQeTp5k4WqkwCAwcGRezr2pgwAOLO3Nm9VxZmRjZ
DaKEdFA2li9FuWqB3czDL1Yvm0HTwilwVwWMA/l8Hv69h/GqgAH3Ibhz5aJKs0wMkx8Bcczg
0VMCtk7P+amv9N7u2bVauhZXRsd2nVuIksCufcclvF8V8Ao60XtTbz2N84hKjHxxbr1Yju3+
6tuxy/lXVcCQ1XDj4gZSGY7upFw4t35gaFyS2Lad8vS89I8FcUcQkCSJ/25/22XRqHmtKUFA
HLFt1yi8d4EH/dHRPP/qa3zWer+O324ycPUz/rS9x9YvSg6PTvpk38lCwuvnJ12wwn76WlmX
ITvI5q3nncNUDDe3t+vu7dXU3KwUxxb03ezz44u8smmjhDe/5tT3gneycCn/XEUqwB4eDvnN
P3jrvK3RgZ07fdLRJB1N+vqbQUE6LYoiA0fGinjhQsUEsJ7NXdz2Hx45zYcd/HUWc7NkM9Tt
5tPXWXE2qYdHc3W1G5bd2CXbkBEFjAyfduxI4tihwxNxFN31FdsuBKYRNLJ0Di37WXiKkSFe
Ha6c7ztbGOrnqrMJK0kf4vFV110hW5NRnqQcJrJtDUYPHRZH0RvfhZ0Fkw0sXUl9LS+sYktW
5QO2k+dv574t0w7/NEEYBE1zk6xykVKqApbDREdTncFD39Z9F8a0Lt3OeEDbBzz5N/5yBvdS
28Lm1dwNS6g/yfI4STaOHJ/QnE9pyodmFkKNxVAuDqH4g8DbWTHCuq08FfPnUVyGXxPW8NrN
/LHIwTDwcYqP3th7YHjGRKgtn9IydR05Oibg04uB07s03cqX/eyYw7XLpyZbsIbMOzy7sr3b
L8MOj536/OVEkJ03lnGmJlFIx947M+TgyMn/Zdl4MTCYPuhndT0vddLVgAbUIYMU/p6qiR+Z
sTxcUMgp1sSOZUt2ZE54+8SAwfzEjjQP7OfgDwbhScKPmR/SENIQVe71CTPRdpzriql0thCV
UzgcVH7lN33J3otBZ+P/SNx4XWnWkNMAAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
projectIcon
	"Private - Generated method"
	^ Icons
			at: #'project'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self projectIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/6/2004 17:21'!
projectIconContents
	"Private - Method generated with the content of the file /home/dgd/project.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0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'! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
publishIcon
	"Private - Generated method"
	^ Icons
			at: #'publish'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self publishIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/6/2004 17:26'!
publishIconContents
	"Private - Method generated with the content of the file /home/dgd/publish.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0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'! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 21:58'!
smallAuthoringToolsIcon

	^ Icons
		at: #smallAuthoringToolsIcon
		ifAbsentPut: [ self smallAuthoringToolsIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 21:58'!
smallAuthoringToolsIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 2808846277 4285241285 4285241285 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4104749094 16777215 16777215 16777215 4285241285 4288461528 3765147589 4285241285 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4156465773 4038761825 16777215 16777215 16777215 4285241285 4269385423 4287147474 4272275675 3898314177 16777215 16777215 16777215 16777215 797072384 4273576540 4294437589 4256074563 275795968 16777215 16777215 2892732357 4183986884 4286491090 4288461528 4255103704 3881930943 3064638120 3131747241 3920013734 4104492405 4293443700 4291597568 2693556224 16777215 16777215 16777215 2540345030 3580466371 4268792519 4286622160 4289118684 4270501585 4285635011 4294967295 4291342725 4292591497 4292980025 3953030672 4288257175 16777215 16777215 16777215 16777215 4083520451 3731461315 4285701320 4286227917 4289250269 4219249868 4286878622 4272851774 4292917094 4291794688 4124363449 4288257175 16777215 16777215 16777215 16777215 16777215 3951402422 4268004291 4286228432 4285636298 4272472796 4272064850 4291274851 4292526223 4071986259 4226279912 4288257175 16777215 16777215 16777215 16777215 16777215 4288257175 4259637483 4268725432 4286293452 4286030796 4273182039 4290882421 4289438090 4256807607 4260689397 4288257175 16777215 16777215 16777215 16777215 16777215 4288257175 4294901502 4259177190 4288849830 4289112743 4288981159 4288849829 4289375657 4275953373 4277795321 4288257175 16777215 16777215 16777215 16777215 16777215 4288257175 4294901502 4060020478 4294967295 4294967295 4294967295 4294967295 4294967295 3824020973 4277795321 4288257175 16777215 16777215 16777215 16777215 16777215 4288257175 4278124286 3003055870 3317805978 3874482915 4274568359 4291740331 4289906910 3102337511 3992582649 4288257175 16777215 16777215 16777215 16777215 16777215 4288257175 4278124286 3671774150 3233788567 4258252210 4292135090 4291940812 4290889933 3269254362 3791321850 4288257175 16777215 16777215 16777215 16777215 16777215 3868826775 4227726845 3923234754 2916008878 4241934256 4293847268 4290693332 4272668629 4190688481 3690790396 3868826775 16777215 16777215 16777215 16777215 16777215 3029965976 3471502058 4191934156 4192131794 4276808165 4294111470 4294177521 4292074211 4258191588 3506174971 3130694807 16777215 16777215 16777215 16777215 16777215 865835926 4070153367 3924421097 4244504061 4277926649 4294704122 4294506744 4294440951 3638484958 4019821720 1620219282 16777215 16777215 16777215 16777215 16777215 33554432 1553505174 3550059671 3952712855 4288257175 4288257175 4288257175 4288257175 3365510294 1653576333 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallBackIcon
	"Private - Generated method"
	^ Icons
			at: #'smallBack'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallBackIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/5/2004 13:37'!
smallBackIconContents
	"Private - Method generated with the content of the file /home/dgd/smallBack.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAABqklEQVQ4jb3Sv0uUARgH8M9792bHmZRJJYIt
tghFeW8HJSGBhNDY0tAQKbRWUA1tTg5lf0AtTREaERWEEE1BDRZEtQSZS3CdyqXe+eP03rfh
BClJsKAvPNPzfL/Pry//iPQfM0edsk9GwdTWJPNa5TwUSUQqcg5sVh781rVf4pa2bLNLnQx9
4Ef1O0YQowENEnMY885YXeCkUNkdiQvh2Q7DV4ZdC++p3n7PyCRhiu1pGkN2hNQSvsxXBfoC
kSweCYO+XYMn9PedN6HgsTcahLp1yunQKCMWKykbjV+Z6h1lfuV+IHITV904pPXMEQWlDXu2
aNJuD5hQMGeB0y8oLn0O0QYOZzeSi19ZmDWzWjWzuoKEVJp0yMIq1NLavMaAWiVjf4nG5npB
ucTEWxbnWK6wsrQWy1RSPFkk8bF+xC49Up673Jx1sMzOvVRK1DI8qFKMmY6ZrrIcr0+YuL7+
xi69Up4Z2J1xbJa4xngLd2cIDIp9E9iGUGJJYFKTl7/u3KVHpOhiNvG0KXE8SEQ+bWakjchr
Fxlfc2Is59zWBOoI5HXLy/8N+f/iJ/5OkGsa/+A/AAAAAElFTkSuQmCC'! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'mt 5/10/2015 13:39'!
smallBroomIcon

	^ self fugueBroomIcon! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'dgd 9/25/2004 22:59'!
smallCancelIcon
	"Private - Generated method"
	^ Icons
			at: #'smallCancel'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallCancelIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallCancelIconContents
	"Private - Method generated with the content of the file /home/dgd/smallCancel.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAACC0lEQVR4nK2RX0hTYRjGf+d85882bWZMESIC
CSHOaYu6TIksRwRCXhgUOLqWsn9IlFAQ3YRSoFTQWEFGeFUXBd4FQVC3zrrJIOgiglzTOd12
tp23myY6XBfRe/XxPt/zvs/3++B/Vgx27oXdjfQ9EI5C+8aeXjs4kDhu8zVu8dmFvi2Gdx01
WThh892F85vELohctPHKvTFZfvVUxrYHSg4cqekuREe3mdnl6UnxBnrkWgDfgRiAAaCBlfOp
+Ok5U31ZYHTmiVUaTMzqK+W4D8X+sPnm6nSySfM89A/v+eWDArUpxT44fM6i6LUhK8k7svTy
sYwEVeFmi1VYev1McjMPpdyhZMTGdyBR82kbhzjQe8xmdqJFswr3pvA1DVpb0TMZAlcucDnr
+289hubhecOfcKHvko1X7VCSn7olq+PXpdSuy7BFxYVT9feN+oYG5i6FplerlMduoGmgTKhA
WYdvDTf/eUL//RAVaUN+NCPzBpJWSCaMFCLIWZN8FA40ij7wqIlqzZw2kCGDwoRNNa2QbBjJ
R5CEyVIU3PrNh1IbzPMGcsZgzYFuB05PBvDTClluQXIRZNBg8SCE1hlo0OoY6ItF+FmE21D4
WCH+Cd4BJIsErCAp8midYehU7JirEALW1lOcVKReNCODilUHurfgMzweQB4EkRjc3ZJDFHpc
iP4F8v6GEP+1fgOfrL0AYYB91QAAAABJRU5ErkJggg=='! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:40'!
smallCenteredIcon

	^ Icons
		at: #smallCenteredIcon
		ifAbsentPut: [ self smallCenteredIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:40'!
smallCenteredIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 2862522779 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 2862522779 16777215 16777215 4288257175 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4289440683 4289440683 4289440683 4289440683 4289440683 4289440683 4289440683 4289440683 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294440950 4294440950 4294440950 4294440950 4294440950 4294440950 4294440950 4294440950 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4289309097 4289309097 4289309097 4289309097 4294572537 4294572536 4294506743 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572536 4294506743 4294440950 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572536 4294506743 4294440950 4294375157 4294309365 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4289440683 4289309097 4289309097 4289309097 4289309097 4289309097 4294309365 4294309364 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572536 4294375157 4294309365 4294309364 4294243571 4294177778 4294111985 4294243571 4294177779 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572536 4294506743 4294440950 4294375157 4294309365 4294309365 4294243572 4294177779 4294177778 4294111985 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294440950 4294375158 4294309365 4289309097 4289309097 4289309097 4289309097 4294111985 4294046193 4294046193 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294375157 4294309365 4294309364 4294046193 4294046193 4293980400 4293914607 4294046193 4293980400 4293980399 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294967295 4288454554 16777215 16777215 2862522779 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 2862522780 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 21:19'!
smallChangesIcon

	^ Icons
		at: #smallChangesIcon
		ifAbsentPut: [ self smallChangesIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 21:19'!
smallChangesIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 490044345 2671475640 3728440505 4130765241 3845814969 3157949368 1395947704 87254732 16777215 16777215 16777215 2033415863 271610031 16777215 16777215 1345746614 4130962105 4201816526 4288331486 4289908199 4289907942 4288593630 4100561352 4030167480 1479833527 16777215 1714780343 4281563063 255031227 16777215 691173044 4114250680 4288134110 4289186277 4289448678 4290631148 4291288047 4292076787 4292602100 4291288045 4115827136 3224795577 4148724669 4231297207 238518454 16777215 2604301240 4252279246 4286425299 4282351034 4281628855 4231756985 4083652551 4272736226 4292798965 4292930549 4293062134 4254381786 4291879151 4181425337 222000817 16777215 4080630456 4282745533 4265508539 4114316473 1966241975 422804152 1161066169 3443096760 4117469896 4292141808 4293324792 4293587448 4293587448 4198202553 187593913 16777215 4264916919 4214979769 2641535701 3862395320 590575030 16777215 16777215 16777215 1613920439 4281628855 4291813359 4291682544 4293390584 4198202808 171147443 16777215 4231231415 2119404740 729984219 324965577 33587328 16777215 16777215 71336127 3292101303 4204509915 4293652984 4292930549 4293587448 4214980024 154759594 16777215 222000817 16777215 16777215 16777215 16777215 16777215 16777215 2687727799 4281563063 4165830334 4149447104 4149447104 4149447360 4281563063 138444991 16777215 4180899767 4264785847 4264785847 4248008631 4231231415 4214454199 3107157943 490044345 406156725 406156725 406156725 406156725 406156725 473396918 16777215 16777215 4231231415 4293521912 4293521912 4293456120 4255630561 3761732280 271610031 16777215 16777215 16777215 16777215 445231586 964537051 4231231415 204177599 16777215 4231297207 4293587448 4291616752 4291879151 4264785847 1781757623 16777215 16777215 16777215 187593913 3912595640 2373363157 3260780998 4281563063 87254732 16777215 4248271288 4293587448 4293390584 4293456376 4291419116 4115958721 3392765112 909407928 16777215 1781757623 3980361147 4283862726 4265311417 4197940152 16777215 16777215 4248402616 4292798965 4271947742 4293258743 4293062134 4292076530 4271619037 4115892929 4214519992 4282022840 4265245624 4282417339 4283665089 2822470841 16777215 16777215 4265048504 4150630087 3560339896 4099247298 4290630890 4291879409 4291091181 4290368490 4289383143 4287871453 4287476954 4288265438 4114184888 707885494 16777215 16777215 4281563063 2469820856 16777215 1412724406 3996678584 4082995909 4287608282 4288857058 4288923108 4288002782 4235765200 4130962105 1530230966 16777215 16777215 16777215 2721216695 16777471 16777215 16777215 87254732 1395947704 3191569337 3912858808 4197742775 3829103801 2856222137 607287480 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:06'!
smallConfigurationIcon

	^ Icons
		at: #smallConfigurationIcon
		ifAbsentPut: [ self smallConfigurationIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:06'!
smallConfigurationIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4282944717 4282944717 4282944717 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1480812686 4282944717 4282944717 1480812686 4282944717 4291155690 4282944717 1480812686 4282944717 4282944717 16777215 16777215 16777215 16777215 16777215 16777215 4282944717 4291155690 4287805149 4282944717 4285374933 4289250531 4285374933 4282944717 4287805149 4291155690 4282944717 16777215 16777215 16777215 16777215 16777215 4282944717 4287805149 4289250531 4289250531 4289250531 4289250531 4289250531 4289250531 4289250531 4287805149 4282944717 16777215 16777215 16777215 16777215 16777215 1480812686 4282944717 4289250531 4287082715 4098460874 4284323793 4098460874 4287476956 4289250531 4282944717 1480812686 16777215 16777215 16777215 16777215 4282944717 4282944717 4285374933 4289250531 4098460874 2923986606 1497524108 2537911972 3998060491 4289250531 4285374933 4282944717 4282944717 16777215 16777215 16777215 4282944717 4291155690 4289250531 4289250531 4284323793 1497524108 16777215 1497524108 4284323793 4289250531 4289250531 4291155690 4282944717 16777215 16777215 16777215 4282944717 4282944717 4285374933 4289250531 4166095308 3445593287 1497524108 3225977527 4098460874 4289250531 4285374933 4282944717 4282944717 16777215 16777215 16777215 16777215 1480812686 4282944717 4291155690 4287082715 4166095308 4284323793 4098460874 4288330975 4289250531 4282944717 1480812686 16777215 16777215 16777215 16777215 16777215 4282944717 4287805149 4289250531 4289250531 4289250531 4289250531 4289250531 4289250531 4289250531 4287805149 4282944717 16777215 16777215 16777215 16777215 16777215 4282944717 4291155690 4287805149 4282944717 4285374933 4289250531 4285374933 4282944717 4287805149 4291155690 4282944717 16777215 16777215 16777215 16777215 16777215 16777215 4282944717 4282944717 1480812686 4282944717 4291155690 4282944717 1480812686 4282944717 4282944717 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4282944717 4282944717 4282944717 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 21:19'!
smallCopyIcon

	^ Icons
		at: #smallCopyIcon
		ifAbsentPut: [ self smallCopyIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 21:19'!
smallCopyIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 1335796123 2140773783 2140773783 2140773783 2140773783 2140773783 2140773783 2140773783 2140773783 2140773783 2140773783 1403036570 16777215 16777215 16777215 16777215 2107219353 2147483647 2147483647 2147483647 2147483647 2147483647 2147483647 2147483647 2147483647 2147483647 2147483647 2140773783 16777215 16777215 16777215 16777215 2107219353 2147483647 2146694131 2146694129 2146694129 2146694129 2146694129 2146694129 2146694131 2146694131 2147483647 2140773783 16777215 16777215 16777215 16777215 2107219353 2147483647 2146694131 2144588753 3500845735 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 2879299995 2107219353 2147483647 2146694131 2146694131 4254834329 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4288257175 2107219353 2147483647 2146694131 2144588753 4254768536 4294967295 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294967295 4288257175 2107219353 2147483647 2146694131 2146694129 4254834329 4294967295 4294177778 4291940816 4291940816 4291940816 4291940816 4291940816 4291940816 4294177778 4294967295 4288257175 2107219353 2147483647 2146694129 2144457169 4254768536 4294967295 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294967295 4288257175 2107219353 2147483647 2146562545 2146562545 4254834329 4294967295 4294177778 4291940816 4291940816 4291940816 4291940816 4291940816 4294177778 4294177778 4294967295 4288257175 2107219353 2147483647 2146562545 2144325583 4254768536 4294967295 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294704123 4294309365 4288257175 2107219353 2147220475 2146430959 2146430959 4254834329 4294967295 4294177778 4291940816 4291940816 4291940816 4291940816 4291940816 4294046193 4294572536 4291743694 4288257175 2107219353 2146825717 2146957301 2146825717 4254834329 4294967295 4294177778 4294177778 4294177778 4294177778 4294177778 4288520347 4288520347 4288520347 4288520347 4288257175 1352639388 2140904857 2140904857 2140904857 4254768536 4294901501 4294177778 4294177778 4294177778 4294177778 4294704123 4289243814 4294704123 4294572537 4293454312 4288257175 16777215 16777215 16777215 16777215 4221214104 4294704122 4294177778 4294177778 4294177778 4294704123 4294704123 4289243814 4294704123 4293519848 4288257175 3432619159 16777215 16777215 16777215 16777215 4221214104 4294375158 4294572536 4294506743 4294835708 4294769916 4292664540 4289243814 4293454312 4288257175 2140773783 16777215 16777215 16777215 16777215 16777215 2744950939 4288322968 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 3214580887 2140773783 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/31/2010 15:35'!
smallCutIcon

	^ Icons
		at: #smallCutIcon
		ifAbsentPut: [ self smallCutIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/31/2010 15:35'!
smallCutIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 2086296920 4270888083 4169895821 16777215 16777215 16777215 16777215 3580785006 4271151255 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4288060314 4294440951 3950673018 3178000237 16777215 16777215 16777215 4254110610 4292401881 4018045312 16777215 16777215 16777215 16777215 16777215 16777215 4288257693 4292467931 4293980656 3833232506 16777215 16777215 1851284056 4290165433 4291744465 4288060571 16777215 16777215 16777215 16777215 16777215 16777215 4271217305 4290626241 4294441207 4186343815 3178000237 16777215 4288257693 4292073430 4290297019 4169435271 16777215 16777215 16777215 16777215 16777215 16777215 1700288598 4288323486 4292928482 4294046449 4001268610 3597693552 4290823620 4292204502 4288257693 3177671015 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4220227726 4290757570 4294572537 4271151254 4289441710 4290823106 4290823363 3900144248 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1549162838 4288257693 4293257190 4294309621 4289178795 4290033847 4287994521 1549162838 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3698554484 4290954692 4293651948 4290165433 4288389279 3580785006 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 2539610207 4288060314 4291744208 4290625727 4283997890 1549162838 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 2134145207 4266566084 4283212999 4287417300 4292537599 4292537599 4266631363 2134145207 16777215 16777215 16777215 16777215 16777215 16777215 2134145207 4288667371 4292537599 4292537599 4283213770 3744684890 4285454303 4292537599 4292537599 4292537599 4283212999 2134145207 16777215 16777215 16777215 2134145207 4288667371 4292537599 3945492591 4287941348 4279934652 16777215 458557696 4292537599 4283279307 4118390190 4292537599 4283212999 2134145207 16777215 16777215 4268539848 4292537599 3125100092 16777215 4287941605 4280000187 16777215 16777215 4283212743 4272087534 16777215 3057990715 4292537599 4281899979 16777215 16777215 4280130745 4288734706 16777215 3477099345 4292537599 4281899979 16777215 16777215 4281899979 4290041584 3544143702 16777215 4292537599 4281899979 16777215 16777215 4281899979 4292537599 4288734706 4292537599 4245856928 1549473302 16777215 16777215 1549473302 4269132498 4292537599 4292537599 4292537599 4245727654 16777215 16777215 16777215 4281899979 4280195254 4280130488 1549473302 16777215 16777215 16777215 16777215 2538548780 4281899979 4281899979 4281899979 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/21/2010 22:48'!
smallDebugIcon

	^ Icons
		at: #smallDebugIcon
		ifAbsentPut: [ self smallDebugIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/21/2010 22:48'!
smallDebugIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 2134145207 16777215 16777215 4231297207 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 2134145207 16777215 16777215 4231297207 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 2134145207 16777215 16777215 4231297207 4231297207 16777215 2134145207 4231297207 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 2134145207 4231297207 4290304767 4290304767 4231297207 4231297207 4231297207 16777215 4231297207 4231297207 4231297207 16777215 16777215 16777215 16777215 16777215 16777215 4231297207 4290304767 4231297207 4290304767 4290304767 4290304767 4231297207 2134145207 16777215 16777215 16777215 16777215 16777215 4231297207 16777215 16777215 16777215 4231297207 4290304767 4290304767 4290304767 4290304767 4290304767 4231297207 16777215 16777215 16777215 16777215 16777215 16777215 4231297207 2134145207 2134145207 4231297207 4290304767 4290304767 4281694904 4290304767 4290304767 4290304767 4231297207 16777215 16777215 16777215 16777215 16777215 16777215 4231297207 4231297207 4231297207 4290304767 4290304767 4290304767 4281694904 4290304767 4290304767 4231297207 4231297207 4231297207 4281628855 16777215 16777215 16777215 16777215 16777215 16777215 4231297207 4290304767 4290304767 4290304767 4281694904 4290304767 4231297207 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4231297207 2134145207 4231297207 4290304767 4290304767 4290304767 4285966811 4231297207 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4231297207 16777215 16777215 4231297207 4231297207 4231297207 4231297207 2134145207 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4231297207 16777215 16777215 16777215 4231297207 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4231297207 16777215 16777215 16777215 4231297207 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4231297207 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:49'!
smallDeleteIcon

	^ Icons
		at: #smallDeleteIcon
		ifAbsentPut: [ self smallDeleteIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:49'!
smallDeleteIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1382511397 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 1986424869 16777215 1449488679 4287666543 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4288192890 1147498793 4284837928 4292994782 4287139466 4285493612 4285625199 4285822577 4285954164 4286151542 4286283129 4286480507 4286612094 4286809472 4286941059 4290691515 4292994782 4284837928 4284837928 4292994782 4284112991 4285888370 4286019958 4286283128 4286414715 4286612094 4286809473 4287006852 4287072902 4287270280 4287336074 4289243813 4292994782 4284837928 4284837928 4290955444 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4292994782 4290231460 4284837928 2590339368 4284837928 4288851047 4287338045 4287206716 4287075131 4286877754 4286614585 4286416951 4286022196 4285693233 4285495344 4285363760 4285035308 4284837928 2590207782 16777215 4284837928 4291812235 4290760282 4290760282 4290694233 4289904210 4289970017 4290232690 4289115217 4288325447 4287733059 4287272512 4288324968 4284837928 16777215 16777215 4284837928 4291812235 4290760282 4290760282 4290628441 4291022450 4292731857 4292731857 4292074155 4289575523 4288983124 4288391252 4288324716 4284837928 16777215 16777215 4284837928 4291943566 4290760282 4290760282 4292140202 4291088506 4290627953 4292139948 4292205742 4289509987 4289246047 4288917339 4289574266 4284837928 16777215 16777215 4284837928 4292009105 4290957409 4291220332 4292600262 4292731857 4290365032 4290101861 4290101613 4290956176 4289114972 4288719960 4289903745 4284837928 16777215 16777215 4284837928 4292337822 4291286129 4291286127 4292534726 4291285891 4290430569 4291942578 4290232689 4292468932 4288983643 4288390740 4290100612 4284837928 16777215 16777215 4284837928 4292403360 4291220334 4291154541 4291154294 4292205742 4291614119 4292600263 4292534470 4290035573 4288786006 4288259408 4289772418 4284837928 16777215 16777215 4284837928 4292206493 4291154538 4291022952 4290693989 4290365026 4290036063 4291679916 4289378136 4288983381 4288654161 4288127821 4289705854 4284837928 16777215 117440512 4284837928 4206542989 4292140187 4291943322 4291876761 4291614359 4291416725 4291284884 4291021715 4290890129 4290692752 4290363789 4204569458 4284837928 100663296 16777216 2103470115 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 4284837928 2120049952 33554432)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'mt 5/10/2015 13:13'!
smallDocumentClockIcon

	^ self fugueDocumentClockIcon! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallDoItIcon
	"Private - Generated method"
	^ Icons
			at: #'smallDoIt'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallDoItIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 4/20/2006 12:19'!
smallDoItIconContents
	"Private - Method generated with the content of the file /home/dgd/smallDoIt.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABl0
RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAAMRSURBVDiNfVNbaFRXFF3n3DNz
r3cm10nsaJxYMU5joon4QvGZYjUGtLZEBWkiRWwj+iEq4odIW6S0P/VXEIvFDzP4iB/6IRLR
ZDCjVpsJksekKsaQaXwlmceN9c6duWf7EQ0y2O6/zVqstdl7bUZE+L+qXzW3GEAFgM5QJJbK
x1m+wJ7ayl+IaCdj7HZ6TB4wvPyvJVWqO9qTyaRMuTYUif39IZ/nuW0JzhT7ThzzB5YtcG90
CYS/WqcXNW43iuo26MWJ11ZD/gQir896dEVyIdBQV6gvX2QHPyt1QzKG+AuZ0VVl457aygYC
Os0xWR+KxGzxznmmz+BnfAbH44GsJK5AuIHyOeP6BGDrJp82I+BetGKJh5+7nAy03jX3A/hN
AMBkQzn7XcMnqwt9ghV4OIgLXLgyaoXvmHYuR1hfbajbvixUv6hWOQAESzW15XZyHWPsuAAA
KVFaXu5h6jiOe1FTXmtLDjWFHx5kDP1ZZ86Piotvrts0RQ1HUnbTpZGRtq6hZgCCA4Ak+uHw
TwOJyL3XNnGBrphldT9NnJdE4Zwju3I2+/npgG0SF7BsBoDxpbOnDRNRlgPA7y09ZzIZ+qPj
wViGuILi6ZrqNyZNJ6IUAGgaX6npwkVcQU2N3/3rsbJpJVP104wxJgDg28/nba+c593d2Dir
gDjHmmq/cr8jvW13bVWh28Wl1yvW7tjx6WTiCgDAY3DQ+AXH1+w4pAZKJgmX5kIqnUVBgQtH
jlZ4e7rTX3s8AoESDZqm4OLFuNXaOpzjnLEnz9LNABwBAEQ4394+eijamSodTdjO0mVF7l3f
B/WqhVMmAkIA0mMSj4eSJyN9z05LIg7gCQeAUCSWMc3c4pY/48sfDiaPMs6zxBUMxi3s29vx
prk5niGuIFhmqH6fXuZI2UdEvUQkJ5IYisQcAL31q+Y6fb3pXGc0hZvX/3nT3T9yim5gQ9eD
ZMnwK8tq7xm6yhhTiMj56DO9S+ZCEnTk+ei/Zlv3YNM3KytuPXqemB/tf+mRRI+I6MV77kcF
JkDGBBHl/pMA4C2oHVtvUbiwmAAAAABJRU5ErkJggg=='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallExpertIcon
	"Private - Generated method"
	^ Icons
			at: #'smallExpert'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallExpertIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallExpertIconContents
	"Private - Method generated with the content of the file /home/dgd/smallExpert.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAACS0lEQVR4nJWRu09TcRTHP+1tL33QVkjB0lqN
QABRiiE6SDCEElJcYWbV4GRSw9zBP8BFSRwk6mCiI4Mp0dCQaHwwgQllACqlPEIfVOhte3vb
n4M0ImICZzq/c36f73lJnMHehJDfRiiLGW44nGRGfMj6swjU1qObfoSHNWJ2lcZ7z1B0p4W/
PsEraTR61yg7BWHFwsRKM3On5X0+Jx/uXiM3NYRYGuPTx4c0AvyvAxm4CnQDWeAxsNpgZrLX
w/eZH9ieB1GlFRZOgseQSNY214pWf6twd7oFIID3wOXDP9dDIfQCdMc7eMo5xj13PPR099Du
biefyBOdjyLbZcIvw0qlXLkFfyobjsCjyIxzG1wXXTTVNdFyvoWOng7Mo2ZiuzEKhYJl9vXs
K+AmoAIcPeMEV8BoNVJjrMEsm7Fb7HjqPbS526iz1jEwMoCl3uIDAlWoKmAAfDRAqVyiWCqS
V/NklSzxVJxoIkrqIIWiKji8DoDe4yNYARPy70fyZ5LN9CY2k428mscoGUntp4in4miSBuA8
LpAFVsjQghXWU+uYZBOlcomtvS0kvcR+fp+N9AaZRAbg80lLnGaVB1wAIQTLm8vs5fZwWBzo
9XqUosJ2bBttV8sB4Sp09IxmYJ4uOungXysAc0CW+8BkNSxVnVAopDcYDEs7iztdpa1SE4bD
7AEQB90XnXrJeWkqGAy+6+/vVyORSO4vgUAgYHe5XA19fX0LxXRxmzjFynJFsiVtCa/R+214
aPjF4ODgohBC1jRN8fv96UgkIn4BMVjYCuwVUJAAAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 4/20/2006 12:32'!
smallExportIcon
	"Private - Generated method"
	^ Icons
			at: #'smallExport'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallExportIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 4/20/2006 12:32'!
smallExportIconContents
	"Private - Method generated with the content of the file /home/dgd/smallExport.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABl0
RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAAJTSURBVDiNlZPNTxNhEMaf2d2y
29IWW6rAoiEhVRP5sBIlcqgHL0YTrScvePDEyZuJqf+BB2/ExHjxwFUTKtFEkyZgoySgpGCU
AEK4uKLUVtvdbbfd3fEgEChN0Oc4mfc3z8w7Q8yMRgrGFRnATQA5ADPFTOVbozxqBAjGlSE1
1vY0erpbbfepmExPlX4s564WM5WpAwGBC/K18EXfuO9UE9UKNkgU0NPVh6X0mvX17cYVAOsA
1ouZigsAQj1RbpdGPapA5moVuefmRz1b0RcWs1DOsexpE9NSRFglkR9t5+8BpNRwgLVatjhj
FQsvy7d/vSj3G/O1QfNzzdF/m4je6kTizmUIihjYfiPtKc/87MlqcxevUSSh6bVgXPGKIZpQ
opLoaRGhiAp0S4djuF/2OUip4WGJecBnO4cB3N0KH1KOSRFRITCAUq2E7EIWANJ7hphSw61g
XurN637ZdeUPkaAJolhCy68E48rxJlWY9g/IrXKnB/lX5uzmmD5Y7+BByLKNUNWWfbaLTsMy
wPwYAIqZykpVc0/8nLBGN8asudJy/yjFks07DsY7QlEAK5LjFgTABkAAxKoohABcSmj513Tm
3kMwj+yamUNE48x8Y7uFHgCeuh9lAIvXj4wMAZiMnT2P9o6jICLkNr9jdvoNAxiWACCh5T/V
78OOYskQAHRHT0IU/xrwB4KYm3nnOI7dsm+R/lf/AnABwND1nUC5bMJ1HQGA2/CYdotiSZlI
eM/s9nq9PgtEKJtGExFpzNx3IGAL4gcQBzCw5XoeQIaz9wt/AFR4+5g8QPOiAAAAAElFTkSu
QmCC'! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:27'!
smallFindIcon

	^ Icons
		at: #smallFindIcon
		ifAbsentPut: [ self smallFindIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:27'!
smallFindIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 3500845735 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 2879299995 16777215 16777215 16777215 16777215 4254834329 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4288257175 16777215 16777215 16777215 16777215 4254768536 4294967295 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294177778 4294967295 4288257175 16777215 16777215 16777215 16777215 4254834329 4294967295 4294177778 4291940816 4291940816 4291940816 4291940816 4291940816 4291940816 4294177778 4294967295 4288257175 16777215 16777215 16777215 16777215 4254768536 4294967295 4294177778 4294177778 4294177778 4294177778 4292796125 4287862162 4287862162 4287862162 4287862162 4287927954 16777215 16777215 16777215 16777215 4254834329 4294967295 4294177778 4291940816 4291940816 4290164917 4287862162 4290430675 4289841372 4289578718 4290037208 4287862162 1552255106 16777215 16777215 16777215 4254768536 4294967295 4294177778 4294177778 4292796381 4287862162 4290036947 4288002781 4291288045 4291682288 4289119970 4288461783 4287862162 830636669 16777215 16777215 4254834329 4294967295 4294177778 4291940816 4287862162 4289772744 4287674587 4292996085 4293718521 4293456120 4291944945 4288200674 4255366615 4287862162 16777215 16777215 4254834329 4294967295 4294177778 4294177778 4287862162 4289578459 4289711079 4292864501 4293193462 4293061877 4291419375 4289383143 4270634460 4287862162 16777215 16777215 4254768536 4294901501 4294177778 4294177778 4287862162 4290038751 4289711336 4292338931 4292864758 4292142066 4291813617 4291156462 4253726428 4287862162 16777215 16777215 4221214104 4294704122 4294177778 4294177778 4287862162 4289970895 4287936987 4290565356 4291551216 4291419376 4292273395 4289645027 4221615577 4287862162 16777215 16777215 4221214104 4294967295 4294967295 4294967295 4292401366 4287862162 4289381852 4288463072 4290367719 4290499818 4289447648 4288856541 4287862162 3800664454 16777215 16777215 2744950939 4288322968 4288257175 4288257175 4288257175 4287665037 4287862162 4290630372 4288068314 4287281116 4289052891 4287862162 4289901490 4289901490 3632300413 16777215 16777215 16777215 16777215 16777215 16777215 16777215 2306834557 4287862162 4287862162 4287862162 4287862162 4288585629 4291151300 4289901490 4289901490 4288059028 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4288585629 4291151300 4289901490 4288059028 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 33554432 4288585629 4288585629 4288059028)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'dgd 10/12/2004 20:58'!
smallFontsIcon
	"Private - Generated method"
	^ Icons
			at: #'smallFonts'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallFontsIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 10/12/2004 21:02'!
smallFontsIconContents
	"Private - Method generated with the content of the file /home/dgd/smallFonts.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAABw0lEQVR4nNWRP0xTURTGv3vP+8Pro9CqQ1Pw
34tplIGYgIZASZogFWpinEy0mGCJiX+Cg4uM3R1MHNwMMwMrcRAGFhwaEjVRQTSS1DQMbaWN
0tf77r0OakUWWDnJGb58J79zvhzg0BftFuPj03an5/XHI6erpdK6OAiA7xYVXZ8KtXW+Vu3W
44Ne8B+AOM8BgON2TOz10ulbbipzP7YX0IqQzNxJCOEnORlHiYx4JBpb3Sp+WgOAgbHJR4Lo
2U6j1nPy3MUnXV7vZScRn69sbMjWFqnFvWr52wspmrMA4IQjd/96vt94U1xfzRaW5nIqCOYs
27kSqltT/yLk85wxlmqPHiuRZS0C0Kbjjnh9yRMAEA5HKt2J/pnBzO1lTubV33HJawGGCsUx
KcR3JnFzp1YdFU3/LQOzouHYNABIKWY58esfVhYegOnnAKC15gBgAIBSavLr58LDrS8f3wHA
hdHsiAn7le123AAwo5TcNsgwvd7hs8ToGgAwzttSqbxBA+mJrNbqkuse6S5tvl8AgONnzueC
ps8CIcqxUz3Jxo/6U61UFyPq+1krvwyk8Bnnoer22iYAMADmn97v5fY+M4exfgFP/5Zri/7V
SwAAAABJRU5ErkJggg=='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallForwardIcon
	"Private - Generated method"
	^ Icons
			at: #'smallForward'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallForwardIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/5/2004 13:34'!
smallForwardIconContents
	"Private - Method generated with the content of the file /home/dgd/smallForward.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAABr0lEQVQ4jb3ST0uUcRTF8c/jPDPqlKORiqSt
JqSNRfM0LlqElkRtItAIa9HWF9DKXkCbIAQhIohoURQFtQ2iYDYZkpAEUS1CamOBjppmOvNr
MZRUi/5And2F+z33cO/ln6tXt6KBv4N365SYkwgKbivq+LElAnv1YUjQLrIkWEAdBrVkthnt
YfwFMx/nRM6YdGXDomifxIrBzcFwLjiaDQ5mgr500J8OxvIhE46FsfV7IR7vDopRkLjsuFQt
QeKiVDTSUhpyMtOvS6u02IJlT7z0yLRVaw7YZacu1x/eND9aYq161ycnIokJjalepSPqpW3X
ql7arLL3yj+tpV2z2TvPODcN52ORipUKryasVta9DlWiOuI0cYaGJjry3wxmlcnHtSJoigVv
wbs1mpdRJVQIYWNsNkeujdVl3jzn0nzNK3Y2kjiNq9/lbEixNaYjxUgjqSWyLSx+oJTl2mJZ
1YApk7UzJgqCHeo0qKqIfBZ0ilxwagv750ilub+JW/OLgkOeevzrR0pM6Y2CG7lguD5IzEgU
fg1+VcFhiRWJIPHAHm2/D2+kyEv0/Dn4v/QF8kyGaRoL4uEAAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallFullScreenIcon
	"Private - Generated method"
	^ Icons
			at: #'smallFullScreen'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallFullScreenIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallFullScreenIconContents
	"Private - Method generated with the content of the file /home/dgd/smallFullScreen.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAADIklEQVR4nH2TTWhcVRzFz//e++7Me+1MxkxS
0jR0nJZSbI0am07c2OomI7HrilB0GVy5kJDtZJNVEHQTcONKF2otlGmUEhoQSUtClTAWalJp
Pmtik/l47828efM+/m6yiFA8cDaHw1n84BAAMDNNTU2ZWutjWmsrCALDNE3HNM3q+Ph4gP8R
TU9P90oprwHIARBHHBJRi5nrQog9rfVap9N5OjEx0fzPwMzMTG6nuvVp+dmPH/fnTiad9ebO
Bxc++uVIJwbgMbMNYJ+INpj5t8nJyS0iYjk6OnpybXf12vbQ6mU+7xvVf+rd/UFuP22mIyIC
AAagiShFRD1E1EtEpxcXF7PFYrEq4ji2rIRFkR3B3nBx7t08/dS4ddULvB4AJwD0AegCoABI
Zu6N4/jNOI4vM/NZRURGvuesnX80uOqppvXEezLgd1hpqWMAICIJ4DgzHwNgA3CJKH2YaUVE
YcJIBO+9+v5TIQQeri+5sd9OLMz9MAgiYZkWeW2/menudc9deH03ne5yD9n4YRiGKgzDkIg6
RGQws3Xp5YKz8uDeqTvz91PKMAAAraaLysoy5m5/f/Hu7Tsbr7w2sjmQO+MqpTxFRB0hhB/H
sRRCSAAwrWRi6VkVXzeycCKB00kTQ6cK+OSzM+LDG5v57779Jr8wd9MYLFwtKwCNKIr2AdSZ
uYuIglS6S246LsJAISUktpvAhkO4K9O4nu7D8PAw7IbTr5RqKc/zqpZl7QA4OKSuojDg4ypE
MRWhS0vUgwjL+y10AFTaBoa0xt+7e3YGmQNRKpXazLzGzAfM/JiZ9xzb9nuojazyICMbL8kW
BlM+BgwbGdmElBLr61vPZ2dn6woAMpnMSq1WSwohiswc1WqNR1lCIYyqiEmCmZBLCJyQIZIc
od1qot5w/wDAEgDK5XI8Pz+/OTY29lgI4Wid6Cz9eq9QvPK2UtyAZBeafJjko08aqCwt88KD
Sqlet/+iFz2sVCqpPysPz9vV7a9G3iqMvDF8SXZnMxBSI/Da+PLzL36/+fP9dwDYLxw4qouA
drK4EofIsYBBgPLauHXgYQcA/gWOnHIlydMKjQAAAABJRU5ErkJggg=='! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:13'!
smallFullscreenOffIcon

	^ Icons
		at: #smallFullscreenOffIcon
		ifAbsentPut: [ self smallFullscreenOffIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:13'!
smallFullscreenOffIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 3597627500 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 3597627500 4288322713 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4288322713 4288322713 4294966267 4290033079 4290033079 4290033079 4290033079 4294045933 4294045933 4294111726 4294111726 4290033079 4290033079 4290033079 4290033079 4294966267 4288322713 4288322713 4294966267 4290033079 4290033079 4290098872 4294045933 4294045933 4294111470 4294177519 4294177519 4294177519 4290098872 4290033079 4290033079 4294966267 4288322713 4288322713 4294966267 4290033079 4290098872 4290362044 4290822852 4294111726 4294111726 4294243312 4294243312 4290559423 4290296251 4290033079 4290033079 4294966267 4288322713 4288322713 4294966267 4290033079 4294111726 4290757059 4291612368 4292270298 4294177519 4294178289 4292072919 4291480782 4290691266 4294309105 4290098872 4294966267 4288322713 4288322713 4294966267 4294177519 4294177519 4294243312 4292270298 4292730849 4293125607 4293191400 4292928228 4292270298 4294374898 4294113010 4294374898 4294966267 4288322713 4288322713 4294966267 4294243312 4294243312 4294308849 4294178289 4293059814 4294309105 4294374130 4293191400 4294440435 4294440435 4294440691 4294178803 4294966267 4288322713 4288322713 4294966267 4294440691 4294440691 4294440691 4294506228 4293257193 4294505716 4294572277 4293651951 4294571509 4294376182 4294638070 4294637302 4294966267 4288322713 4288322713 4294966267 4294506484 4294506484 4294571509 4292007126 4292862435 4293454316 4293651951 4293322986 4292533470 4294638070 4294703863 4294637302 4294966267 4288322713 4288322713 4294966267 4290033079 4294572277 4290559423 4291086024 4291941333 4294571509 4294637302 4292401884 4291349196 4290559423 4294703863 4290033079 4294966267 4288322713 4288322713 4294966267 4290033079 4290033079 4290164665 4290427837 4294376182 4294638070 4294703095 4294768888 4290362044 4290098872 4290033079 4290033079 4294966267 4288322713 4288322713 4294966267 4290033079 4290033079 4290033079 4294703863 4294637302 4294441975 4294768888 4294834681 4294834681 4290033079 4290033079 4290033079 4294966267 4288322713 4288322713 4294966267 4290033079 4290033079 4290033079 4290033079 4294769656 4294769656 4294900474 4294834681 4290033079 4290033079 4290033079 4290033079 4294966267 4288322713 4288322713 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4294966267 4288322713 3647762028 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 4288322713 3698093161)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:13'!
smallFullscreenOnIcon

	^ Icons
		at: #smallFullscreenOnIcon
		ifAbsentPut: [ self smallFullscreenOnIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:13'!
smallFullscreenOnIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 2576981400 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 2576981400 4288257175 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4288257175 4288257175 4294967295 4281563063 4281563063 4281563063 4281563063 4294046192 4294046192 4294111985 4294111985 4281563063 4281563063 4281563063 4281563063 4294967295 4288257175 4288257175 4294967295 4281563063 4281563063 4281760184 4294046192 4294046192 4294111728 4294177777 4294177777 4294177777 4281760184 4281563063 4281563063 4294967295 4288257175 4288257175 4294967295 4281563063 4281760184 4282942140 4285175492 4294111985 4294111985 4294243571 4294243571 4283861695 4282679483 4281957047 4281563063 4294967295 4288257175 4288257175 4294967295 4281628855 4294111985 4284912835 4287672272 4289643482 4294177777 4294243827 4289052375 4287080654 4284584386 4294309364 4281825720 4294967295 4288257175 4288257175 4294967295 4294177778 4294177778 4294243570 4289511898 4291023329 4291943399 4292206568 4291286500 4289446362 4294375157 4294309621 4294375157 4294967295 4288257175 4288257175 4294967295 4294243570 4294243571 4294309107 4294243827 4291877862 4294309363 4294374901 4292272104 4294440693 4294440693 4294440949 4294375414 4294967295 4288257175 4288257175 4294967295 4294440949 4294440949 4294440950 4294506486 4292141033 4294506487 4294572536 4292863983 4294572537 4294572793 4294638329 4294638073 4294967295 4288257175 4288257175 4294967295 4294506743 4294506742 4294572280 4288461014 4290892259 4292535020 4292995311 4292009706 4289840862 4294638329 4294704122 4294638330 4294967295 4288257175 4288257175 4294967295 4281628855 4294572536 4283664831 4285701320 4288198101 4294572537 4294638330 4289381084 4286292684 4283599039 4294704122 4281563063 4294967295 4288257175 4288257175 4294967295 4281563063 4281759927 4282219705 4283204797 4294572793 4294638329 4294704123 4294769659 4282942140 4281825720 4281563063 4281563063 4294967295 4288257175 4288257175 4294967295 4281563063 4281563063 4281563063 4294704122 4294638330 4294638586 4294769916 4294835452 4294835452 4281563063 4281563063 4281563063 4294967295 4288257175 4288257175 4294967295 4281563063 4281563063 4281563063 4281563063 4294769915 4294769915 4294901245 4294835709 4281563063 4281563063 4281563063 4281563063 4294967295 4288257175 4288257175 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4288257175 2643827092 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 2694091921)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 21:19'!
smallHelpIcon

	^ Icons
		at: #smallHelpIcon
		ifAbsentPut: [ self smallHelpIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 21:19'!
smallHelpIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 672815513 2232900251 3306642333 4061617052 4061617052 3306642333 2232900251 672815513 16777215 16777215 16777215 16777215 16777215 16777215 33587328 2031573916 4113327778 4254773454 4291285987 4293849847 4293849847 4291285987 4254708174 4130104994 2014796699 33587328 16777215 16777215 16777215 33587328 3289865116 4133981369 4293455348 4290102746 4285699261 4282479274 4282479274 4285699261 4290168539 4293520883 4134047161 3289865116 33587328 16777215 16777215 2031573916 4134441147 4293652469 4284844728 4283530415 4291549157 4294178297 4293586676 4289971418 4283925171 4285633726 4293718518 4150890169 2031573916 16777215 672815513 4113459107 4293521140 4284779191 4279720860 4286159552 4294835710 4292732142 4294112760 4294967295 4291286244 4282151596 4286094019 4293586933 4130236322 672815513 2232900251 4271748304 4290102746 4279720860 4279720860 4283793842 4283925171 4281626024 4284845241 4294967295 4294770173 4283334836 4283071922 4290892002 4255234002 2232900251 3306642333 4291417573 4285699261 4279720860 4280574624 4281165988 4281822889 4282348973 4288065743 4294967295 4293718262 4283729335 4283860408 4287869136 4291812328 3307101598 4061617052 4293981175 4282085031 4280311967 4281165988 4281823145 4282414509 4286225605 4294638845 4294901502 4287606223 4284386493 4284518078 4286095046 4294112761 4061748381 4061617052 4293981175 4282216616 4280903075 4281757096 4282348973 4283006130 4293718518 4294967295 4289512155 4284846529 4285109698 4285241284 4286752203 4294178298 4061944989 3306642333 4291417573 4286027967 4281363109 4282151595 4282874544 4284188857 4291878122 4292009707 4285372612 4285438406 4285701320 4285898697 4289183708 4292141036 3307167391 2232900251 4254971088 4290365661 4281757352 4282480557 4283269044 4284254907 4288460756 4288723928 4285569735 4285964490 4286358733 4286555855 4291944429 4255957466 2233031837 672815513 4130236323 4293586933 4285962689 4282874544 4283597493 4285437634 4294967295 4294967295 4286292940 4286424783 4286884818 4289250528 4293981433 4130433443 672815513 16777215 2014796699 4134572988 4293784054 4286554310 4283794871 4285635011 4294967295 4294967295 4286621647 4286884818 4289447393 4294178811 4151876034 2015125149 16777215 16777215 33587328 3289865116 4151415997 4293784054 4291089381 4288197844 4286423754 4286949839 4289381599 4292141550 4294112762 4152270533 3290652832 33587328 16777215 16777215 16777215 33587328 2031573916 4130367396 4272471510 4292075243 4294178554 4294244347 4292403949 4256351709 4130630309 2015256221 33587328 16777215 16777215 16777215 16777215 16777215 16777215 672815513 2232900251 3306970782 4061879709 4061879709 3307233441 2233031837 672815513 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:52'!
smallHomeIcon

	^ Icons
		at: #smallHomeIcon
		ifAbsentPut: [ self smallHomeIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:52'!
smallHomeIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 1291782669 4290904064 4290904064 2180975120 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1291782669 4290904064 4294904335 4294904335 4290904064 1291782669 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1291782669 4290904064 4294904335 4287978586 4288768873 4294904335 4290904064 1291782669 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1291782669 4290904064 4294904335 4289883499 4288059284 4289440682 4292057497 4294904335 4290904064 1291782669 16777215 16777215 16777215 16777215 16777215 1291782669 4290904064 4294904335 4289955460 4288717214 4291085765 4291677644 4291217093 4292128687 4294904335 4290904064 1291782669 16777215 16777215 16777215 1291782669 4290904064 4294904335 4290158236 4289704110 4291546058 4293190884 4293190884 4293190884 4292861919 4292264385 4294904335 4290904064 1291782669 16777215 1291782669 4290904064 4294904335 4290228141 4289901233 4292993505 4293848814 4293848814 4293848814 4293783021 4293783021 4293717228 4293253336 4294904335 4290904064 1291782669 4290904064 4294904335 4290822334 4289835696 4292927712 4293651435 4293717228 4293717228 4293717228 4293651435 4293519849 4293519849 4293519849 4293716457 4294904335 4290904064 1740701696 4294904335 4293388776 4284967526 4284967526 4284967526 4284967526 4293717228 4287269773 4287138960 4287137928 4287138960 4287138185 4294440951 4294904335 1455489024 16777215 4284967526 4294243572 4284967526 4287665037 4287730830 4287796367 4293717228 4287138959 4286949841 4287137928 4286818512 4287073167 4294375158 4284967526 16777215 16777215 4284967526 4294177779 4284967526 4288059539 4288059539 4288059539 4293717228 4287073166 4285110471 4287137928 4283533503 4286941325 4294309365 4284967526 16777215 16777215 4284967526 4294111986 4284967526 4288191125 4288191125 4288453494 4293717228 4286811028 4281563063 4287137928 4281563063 4286613908 4294309365 4284967526 16777215 16777215 4284967526 4294111986 4284967526 4288191382 4288191382 4288322448 4293717228 4290691267 4291086541 4291217352 4291152333 4289967027 4294309365 4284967526 16777215 16777215 4284967526 4294046193 4284967526 4288257175 4288257175 4288257175 4293717228 4289835441 4289901234 4289901234 4289901234 4289901234 4294309365 4284967526 16777215 16777215 4284967526 4294440951 4286085750 4289178277 4289243814 4289309350 4294309365 4294440951 4294440951 4294440951 4294440951 4294440951 4294769916 4284967526 16777215 16777215 2892392806 4284967526 4284967526 4284967526 4284967526 4284967526 4284967526 4284967526 4284967526 4284967526 4284967526 4284967526 4284967526 2540137061 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallInspectItIcon
	"Private - Generated method"
	^ Icons
			at: #'smallInspectIt'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallInspectItIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallInspectItIconContents
	"Private - Method generated with the content of the file /home/dgd/smallInspectIt.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAACX0lEQVR4nKWST08TURTFT2f6nyntlJbWAq2h
BYxUSAoblBg0rsTECiEkfgANSgyuXejaj2CMicaFhhhdINGYKAYQYjUYTG1EBIZph1JaSodO
W9uZqSvISAYT41299847v3fuzQP+szQHD3wNtf2+o9Yhu03frNUSBp6vrJZF6cnUNPtCDaBV
rG2dQef9sZFQX9/p5oLB6lmRqwa+xDMdk5OxS5IoR1bZ7eF4vJhQS0C1+unx188H+yr0ufdr
5LBBeYkqzZXq+YehkbFX62+mmB4A4p5GAgBNG++8fBo+L3kGp1lywHww5qZA10wu6G03wnnv
x4VkY3JTmNjTCADobHd2N3rdCY64YFHrsyLJBJu1GCVrx/LxFnuvUiMAWNpa6LqCxsuqmQFA
KPzSAcDiRqPJ5TI7AeiUAKfnCAVJY5IOA2R3BCMAVGQ9HLRZC8CqBKzMRzi+prrmUDOn0rwp
9oN1AUDAniww8VwKQPqPGURjmSW5lAlZq9Hd/b5FiYgtJ+xvZ7/6RVEm9ZDELvMXOs7ll5QP
kACQ48vR+chGz/WBYluy4E7PLPJNc5+WfCyXpiVJJkx6qXzv2ShVM/7ZzpxsIKKxbEQoivF9
AICdrUwpOheJnxnqSXV1eLitOqpYDjhyhXDw2+5F7yxtybnNuHqC6KV4K6MznF1niu+EUjl1
8CvbAs22260BOuRxU26djtRv7xQTCS4fJQnNsYkHV05RmTRZZRYrg4/XuZ+s0H7Y4PfSEYq9
sdVPP/o+c3NXmr8s3r3VnQoG611/A6iW02kaDff7P/iaaq/9s1mtfgOeOPMUwSrvxgAAAABJ
RU5ErkJggg=='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallJumpIcon
	"Private - Generated method"
	^ Icons
			at: #'smallJump'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallJumpIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/5/2004 13:34'!
smallJumpIconContents
	"Private - Method generated with the content of the file /home/dgd/smallJump.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAABdklEQVQ4jcWSPUuCcRTFf/d51LSIgmfphSBo
iobCBx16oSUaApemgqCGQGhq6wNEn6ChrakIWlskiKgpqCdqqKkXSlJCibA3M/U2aFZaEDR4
xv8593DP/R+oNuRXxqYDxS6qHBzO/2Zg4xeXLJPV7jLlgSphHA6/Pptlw8MCO572RmtsYda0
Jnq49L8h6UclmmkVmKaFPeI/bRPEkiC3xlzT82p+W1/1TRd1Q306qjwNKEuo2KjYxAhifYwZ
JYMsIbxm3dT8nG9cBsmRZ41dXsiAYUIA6BWAZrKEKgwMIUBn/dOke4gkKby4aaC2QD4kAdBu
/dQW4fp+KDG8eOhiBh81XJMAVbiLFXitTF7aIK/sc5KqW8lskSBVHM7DxQE83hX8nS/ail9o
Iyo5wufRM/dzX6NJ+h6ujiGVKPC7IBEA4riY5YYXKO+BzbAIESwR7VdoAe5BjgROFSCnMILD
Zil1RSgbv9TLOg/a8f0+Pxfp31WuPt4BdGiFIfxoncEAAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:41'!
smallJustifiedIcon

	^ Icons
		at: #smallJustifiedIcon
		ifAbsentPut: [ self smallJustifiedIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:41'!
smallJustifiedIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 2862522779 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 2862522779 16777215 16777215 4288257175 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4289440683 4289440683 4289440683 4289440683 4289440683 4289440683 4289440683 4289440683 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294440950 4294440950 4294440950 4294440950 4294440950 4294440950 4294440950 4294440950 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4294506743 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572536 4294506743 4294440950 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572536 4294506743 4294440950 4294375157 4294309365 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4294309364 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294375157 4294375157 4294309365 4294309364 4294243571 4294177778 4294111985 4294046193 4294177779 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572536 4294506743 4294440950 4294375157 4294309365 4294309365 4294243572 4294177779 4294177778 4294111985 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294440950 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4294046193 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294375157 4294177778 4294111985 4294046193 4294046193 4293980400 4293914607 4293848814 4293783021 4293980399 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294967295 4288454554 16777215 16777215 2862522779 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 2862522780 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallLanguageIcon
	"Private - Generated method"
	^ Icons
			at: #'smallLanguage'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallLanguageIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallLanguageIconContents
	"Private - Method generated with the content of the file /home/dgd/smallLanguage.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAABlElEQVR4nMWRv0sbARTHP+bulOCQwaQnauoF
rYGiDpa2Wtupg7q5Fsng/+CmODqUigouDuLg4L/goCBxiFkEtRXj0CVNuWhE6cn9MOZeB1NI
S07qlC+84T34fHnf96DRaqozawE+Aa8ikchLTdOaS6XSN+AY2ASsxwxTuq7/XFz8LPn8d/F9
V0Q8Mc28rK+vSSJh3AAzgXBvb49/eVkQEa9u2faNDA+/FWCu3tr24WE2EP5TppkXTdMEeA6g
Vg3eGUZ3uF9R+PVlCTXehZowqBQvENtG6ezAy2RRYlHaPrxndHSEvb30R2AjVDVwXNflLneO
1vcCpV2n+c1rymc5lK5OvEwWymWaWlupFAo4jgvg1EZQAHN3Z1tEPPHvbh9Wrjh/974r52cn
EgqFHKDt3zsM6fqz61zua2D+YvGHJJN9LjAe9AkjHA5nVleXxbKuxPddsawrOTjYl/n52fto
NLoF9AfBAKiqmj09PZJ0elcGBwcqwA4wCUQeBauKT0yM+anUlAesAPH/gWo1DSwAsaeCjdNv
pz0BrfQjq7kAAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:38'!
smallLeftFlushIcon

	^ Icons
		at: #smallLeftFlushIcon
		ifAbsentPut: [ self smallLeftFlushIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:38'!
smallLeftFlushIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 2862522779 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 2862522779 16777215 16777215 4288257175 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4294572537 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4289309097 4289309097 4289309097 4289309097 4289309097 4294572537 4294572537 4294572536 4294506743 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572536 4294506743 4294440950 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294440950 4294440950 4294440950 4294440950 4294375157 4294309365 4294309364 4294375157 4294309365 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4289440683 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4294309365 4294309364 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572536 4294572536 4294506743 4294440950 4294375157 4294309365 4294309364 4294243571 4294177779 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572536 4294309365 4294309364 4294243571 4294177778 4294309365 4294243572 4294177779 4294177778 4294111985 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294440950 4289309097 4289309097 4289309097 4289309097 4294243571 4294177778 4294111985 4294046193 4294046193 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294375157 4294309365 4294309364 4294243571 4294177779 4294111986 4294046193 4294046193 4293980400 4293980399 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294967295 4288454554 16777215 16777215 2862522779 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 2862522780 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/22/2004 19:35'!
smallLoadProjectIcon
	"Private - Generated method"
	^ Icons
			at: #'smallLoadProject'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallLoadProjectIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/6/2004 17:32'!
smallLoadProjectIconContents
	"Private - Method generated with the content of the file /home/dgd/smallLoadProject.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAADA0lEQVQ4jY2SW0zbZRjGf9+/358eKIUW6AE7
SlvcxiImc8PIlnmKkmAyt5ktWaYx8cAu1At3ockWZ2I27iDemHjjMTEsgnE6ncQsJrhMM2RM
kQnLQKRoS4GW0lJKD/T/eWe8EOJ7++b3vHne5xH8j2l56MjjhmF6QkBIYYwZQl2YunpxEsC0
JXnsmOke966+XU3Os/aO4w3r7SeahMvrZ+7O87X+7So5Nzm8pcBO1843nzpQ9/K7r3jNDTua
Xc/UjFf9WH/QW7r/iKN889vttf7wLW0LXjjsFafOnqgXCUuA8nKaVtsM+8wR0M0m/fgZuzC0
V+Wm1x8+HPDVWUV8zU2hsooLyd2YSgYRw0l7dYxFm9t7S4jWTQVM6InpeDH1UzbgOOoaw+w8
wBvZw/zQ8g5kkqTWFB0Wrbiphd+GBrLlYjFyfvKu+cvxe/l51c16WZIvCC7F7+N0v1peXS/1
b/UDBMaLhc96ymfGGmMFw0RX7QhmY5We79KR769P/1m2r3eLrr10un3B3lp/yOkNNEtzZaWm
61JomgCBSC8l1ODlQUatj2Y8TmtFKbfmmM1MDUdDE/VMr+2XUuqd574cbREqDarAYux3ZkY+
wtCqaXvkKLl8lmvRiJqc/XRgWtpkydZxkj23r9IVaOXQRKes99/dmFqJk1gpgTKwCI1CbgV7
jRldK/LYh29x4+mUoDF4snRlKc9zA130VgU0iwsDhHR5/N50ZpmNsg9QFFWUBw+dQlAklVnC
7HWit4cpsQHNixY+dxYIaTL8l5cpjT9kPpvJBX02ssk73P7gEpVtYYQnxLMfv41UGsFcgbkF
Mw2eMCPBLL7dbumPVDHR/fV17PwqkwuRGIZivG+QB3q+YnjvNrb1v8R8fIk2TxNWn5XZ3pny
L/tmTcb54aEo9EcVCzj4hiHyMp2Yj61mNmh74SA33DUE9+xAaE6uvNYNqgBaNXXj76Wvnbt4
GsEnjJL7d9RSCfpef3J/SJXLJqXUPwtNSqHrZk1WWEQuu/IFN3n/v7ryN/fUMc5/02mQAAAA
AElFTkSuQmCC'! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallNewIcon
	"Private - Generated method"
	^ Icons
			at: #'smallNew'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallNewIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallNewIconContents
	"Private - Method generated with the content of the file /home/dgd/smallNew.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAACb0lEQVR4nI2RzWsUaRCHn3r7ne6e7p5MPjTu
mMRE8bIOe9iFsKwiePAgQcGFBW+aYy7e/APyXwQPe9uL5CJIbnszgrCw7IQISr7Nl05kY9Rm
ZpJ53/LQKIgfWJeqQz1Vv1+V8I24dH2y98iVhowJhsXrsBcdBjmuav569GDmMYAtGqfGvNM/
EH6Kk2j05OCxU0kS1yppEvdkKZVKRpYlVLKUl81X3Lv/98/ABQD57erUkBVduHljov/X8XHS
ai/tQ8g7nrztyDuOI6c4p3Q9XK5H/H7rTuvhL7WM6WlvrLizfb2V/omJK7hSD/u5J+84nPNY
K1TigJ5yQDW1VJOAUlSmr1opn/9v53RhQSQ7MTiAKqRxgPeKU3Be8b7IH+quV7pOGRup8Xox
rwMrRpUDAZLQUA4N5ciQhIY0MiRRkbM4II0NQ/0hWRwwOlJD8fVCgdVm3moTlQxOQb3iFbwW
W70qcWioJpbQCgBjIzWAc8WAQ7uX5y1KgRBZQVUKWAVrhCQy2EA+ee/o8A8AdQAzPzfz+s3b
vCOilKxQskJoTXG8JPgMBhg7VQP4kelpYwA96naf/79/gDWCDYQ4NJTs5+CHGOirkibl+NK/
zTMGQGF9e7dJYArZ5uvsJyqcuroBEJWN7Rd7iIB8BwwwUhsEtBighvXt3VffR1J85uBtjogJ
LYBR1nde7H2xud3usLKxw/LaJkurmyyvbbG6sU27c9gQCf60AD5w/yw+W+0uPl2x7/IWK+tb
LK1usby2ydZOs6WqT0AXvNAQrw31dmF+7u4+wEfHF69NTSJ6G9hFpCHogvHaOB7vL83Ozrqv
2XkPgdkFGpcRD6AAAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallObjectCatalogIcon
	"Private - Generated method"
	^ Icons
			at: #'smallObjectCatalog'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallObjectCatalogIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallObjectCatalogIconContents
	"Private - Method generated with the content of the file /home/dgd/smallObjectCatalog.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAAB9UlEQVR4nGNgIAxYGBgYuIlQhx2YqrNFR7hx
bcAlz4xHL6OiDKuZkQZbUUMynyM/N6MTEyNj0IMXf1cQZbOICINkuAvXpR+HpP//Py7z/9Q8
sY9OJuye6OpYGBgYGMzDC90EBHidYIJ//zIw7Tl7unvlnu1mJhqs5xJ9uDXff/7HdPvx/8dY
DZCWFvVZ0JSbCxO8dOshw57YS8sYGBgu7D714+/eUz/KmJmZ5MSF/hs+fslwBcMAfGDXqZ82
DAwMH3HJMxEyAJ9muAvevvt0o3ry0o0wwU9ffrAzMP9F1RjKwMz7gtPCwFk2+8e/X/qnDz3Q
ZzjA8IcRh8EcDAwMP5AFzBtVXnOKcIs8f/jxJ6vQX+an+94Gvt/1bQuKF5zZ2ELS2LhOVLDz
XvZg5VjAwMAAt+A/4//Hl3bf2nfz1AOeL/e+bWNXZS5kYEBNSMzBrFwLJ3EKmrqwcgipM7No
nfj748ar//+vCbow8P/48t+XRZxRk+kDM9M/1n9ZXCpssh/Yvnchu0BGg4lZHsbRY2ZlE2Fg
cWRgYGD4+4PN6zfLHxdGDiZJxn9/bzFzMj3+cu7nKr4PaHkkio1r2zt+qf9/BWT+t3HyvRVn
YFBiYGBgYDBmYOWz5YjhteO0YgjFm/wZ+GxZ2Ds9WdhXq7CyWuBTSDUAANeVnXOvNaMbAAAA
AElFTkSuQmCC'! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallObjectsIcon
	"Private - Generated method"
	^ Icons
			at: #'smallObjects'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallObjectsIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallObjectsIconContents
	"Private - Method generated with the content of the file /home/dgd/smallObjects.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAAB/0lEQVR4nJWQXUgUURTH/zN3ZnMGbXcM+9AN
spaelETtQyLbsiWofAjspYwegggfpJ6KoqjH6KEEeyjDoBfFCvbBqChZiGJD7EMa66koQzPX
zf1o292ZuXN6aWSM2KYDl8s55///nXsuw+/wAfW1jO3LEsEGZvA/sVGSe64qgXm9YgXdVLTM
dubr92xeDmnrdTWQokCQnBNVl+VCjLV78YvVIsJ7JMXvLu6Uy1RNEHd5AiQEGn/LDcNdfG+b
dpZs/S96bUOtvGkRYIrz4cvFH09GedEiADo37bP5THzVga7pyMETDW5xS53v+Pq18i0AFU5N
cO56Ue5SBWzOEY3rttUDwAp3ngxJHI0kGyMjt69Rxw5lqO+M1tZ5YX6MCZT8PGu/8bImGuuq
hi8e0+bmHlYTxYNE8SDdOK1NhWqkFk8AAFq4SYl9uLfScgC7tyy5CwDMI6CwdF2EKoVPrYwJ
wkzSpi/frMLoO7NPWpA0tavgzLfIZso/MXHHiBzqbnv5YGhsIJc3++/nz3/PcL3KLx4GUOZ8
Iho6uqN7tzU3O3kqm2OPnr++tKay/IXAjcnHg70Jvx+r02l8dM9YeIHmL8epo/trnHzyawLP
XulBblM2Ntg7DQB/mgFALLW4ZSMfG7gyUUpTEkBE6VJ9wLXCbDIVbT1yLuVuZgrFp/8C/AKq
oMu3BdqijAAAAABJRU5ErkJggg=='! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:35'!
smallOkIcon

	^ Icons
		at: #smallOkIcon
		ifAbsentPut: [ self smallOkIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:35'!
smallOkIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 248039624 79675327 33554431 33554431 33554431 33554431 33554431 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 16777215 16777215 16777215 16777215 2438055777 2135606620 16777215 16777215 33554431 33554431 33554431 33554431 33554431 33554431 50331647 50331647 50331647 50331647 61516458 2471741027 4285252467 4284989551 2034219344 16777215 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 2437793121 4285186674 4287618449 4286961290 4284332135 2066328132 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 2437923681 4285055344 4287487120 4287356046 4284397928 4281637445 134225920 33554431 33554431 1147794552 1718086006 33554431 33554431 33554431 33554431 2437792609 4284989551 4287290254 4287159181 4284332135 4231042883 134225920 16777215 196721081 1047000951 3865102958 4286107261 1885067371 33554431 33554431 2437792091 4284923758 4287224460 4287027851 4284266341 4230977347 152844572 61516458 93952409 1198060663 3831679084 4287027080 4287552399 4233345377 1884605796 2336866654 4284857964 4286962059 4286896264 4284069475 4230911554 152844572 33554431 33554431 33554431 169502259 4284334440 4285910139 4287684243 4286961546 4284990064 4285252723 4286764935 4286634375 4284003682 4247426367 152844572 33554431 33554431 33554431 33554431 33554431 186080558 4283676768 4285252979 4287159181 4287027594 4286568581 4286568325 4283806817 4264203071 171134259 50331647 33554431 33554431 33554431 33554431 33554431 33554431 186080535 4283018583 4284595563 4286634118 4286371460 4283741024 4264071742 152844572 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 186074647 4282229325 4284003683 4283544158 4263940413 152844572 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 186074647 4281243715 4213674302 136331296 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 186074647 136331296 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 33554431 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 21:19'!
smallOpenIcon

	^ Icons
		at: #smallOpenIcon
		ifAbsentPut: [ self smallOpenIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 21:19'!
smallOpenIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 4167658343 4218055784 4234767207 4234833000 4234833000 4251741802 4251807595 3597825392 359492973 16777215 16777215 16777215 2676854157 3968633996 3968633996 3968633996 4285098855 4294572537 4294638330 4294704123 4294769916 4294769916 4294835709 4289835697 3782111083 242052461 16777215 16777215 4035677067 4292072403 4291940817 4291809231 4284967526 4292993505 4292598747 4292598747 4292598747 4292401368 4292269782 4294901502 4272202915 3563942253 158429553 16777215 4018570886 4291875024 4290690750 4290427578 4284967526 4293125091 4289506732 4289506732 4289506732 4289506732 4292730332 4294901502 4294835709 4272335015 2590863724 16777215 4001464705 4291611851 4290427578 4290295992 4285098855 4293454056 4293256677 4293190884 4293190884 4293190884 4293322470 4293783021 4293585642 4292467161 4050678382 16777215 3967647101 4291414473 4290295992 4290032820 4285098855 4293651435 4289704110 4289506731 4289506731 4289506731 4293190884 4293519849 4293519849 4292467417 4084101228 16777215 3950540920 4291085508 4290098613 4289835441 4285164648 4293914607 4293848814 4293848814 4293783021 4293783021 4293783021 4293783021 4293717228 4292927712 4050546540 16777215 3916591730 4290822336 4282744759 4281957047 4281957047 4281826232 4281759928 4281759928 4281628856 4281628855 4281628855 4281628855 4281563063 4281563063 4281563063 3828906425 3916394351 4290559164 4281628856 4291354096 4291354096 4291354096 4291354096 4291354096 4291354096 4291419632 4291419888 4291419888 4291419888 4291419888 4291025902 4264785848 3882445161 4290295992 4281628855 4291550960 4288463588 4288463588 4288463588 4288463588 4288463588 4288463588 4288463588 4288463588 4288463588 4288660965 4290894317 4247943351 3848561764 4290032819 4281759927 4291682288 4288857574 4288857574 4288857574 4288857574 4288857574 4288857574 4288857574 4288857574 4288857574 4288660709 4290893805 4214519992 3831389790 4289835441 4281825720 4291288047 4289054438 4288988902 4288988902 4288988902 4288988902 4288988902 4288988902 4288726245 4288331748 4288003810 4290434027 4197742776 3797572186 4289440683 4281957047 4289711337 4288200675 4287938018 4287609569 4287477985 4287346912 4287149791 4286821086 4286689758 4286689758 4286952670 4288725988 4164188600 3764083547 4289309097 4282088119 4288660196 4287412448 4287083999 4287083999 4287083999 4287083999 4287083999 4287083999 4287083999 4287083999 4287215583 4287609312 4147542200 3764017754 4287601320 4282088119 4284193486 4284259279 4284259279 4284259279 4284259279 4284259279 4284259279 4284259279 4284259279 4284259279 4284259279 4283404487 3644225976 2656459876 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 3996481463 1194293175)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallPaintIcon
	"Private - Generated method"
	^ Icons
			at: #'smallPaint'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallPaintIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallPaintIconContents
	"Private - Method generated with the content of the file /home/dgd/smallPaint.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAAB60lEQVR4nJXQT0iTcRzH8ffze/5srT0zNvqr
KM42JTskoRJISVARCf0RCTp4KCMhELpJx4gQomNBl+jgtX8QnQorqIyg/5atpZmOlk62PXPP
9uCz5+kQ7bAObd/rh8+LD1+o8cJwIgzDgKi1S0Qw/Gi0wx0f3O62wAi1KG0Q6tu1/nLnnr3o
uh8JumoCXFmMjZze56eYxgYcyFQNNENTT7v/VKi1B4w5Vl0AFqoGBAwdP9IlIWuQX8SyS4hq
gV5QhMTJju6dsPwJcFk2TFyYrAr4AQd2tKzd4tkQhdQUpVVIpsylbxCvCpBgf19vM5hLkEtQ
tFzeTqfu/M2V/wKStNsQIV6+eMNWbN7FinxdyI3/AwyAvAbOBED9CFcfgw2gB4NqtP88iqZx
/d5dpiZuGrPwvAwcgnA9DM7Bg04YC4DPgM3AaBgiqrAablw6h+INkU1bzvRs9ixQKgNe8Ot/
kGPfYTIGFySYAVCFGDrYHq4reSQsJcur1x+exaE8H0Dcgvef4WIaMl7YdhSeKnA4At224yS0
lE1jRqHekPFpmlT5IwXAgTYVGppgXStIebjWHNy0uNGr1+krwowWPL77TpKcaV2pBGSAOHxp
hPwvMGMQL8D8z8KKFdACqt+W1du5+YmHyUT/DDypBH4Dwguh75kYJgEAAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/31/2010 15:26'!
smallPasteIcon

	^ Icons
		at: #smallPasteIcon
		ifAbsentPut: [ self smallPasteIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/31/2010 15:26'!
smallPasteIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 33554432 3983042920 4286743170 4286743170 4286743170 4286743170 3983042920 33554432 16777215 16777215 16777215 16777215 16777215 1278376377 4281628855 4281628855 4281628855 4286677633 4289377716 4289378230 4289378230 4289377716 4286677633 4281628855 4281628855 4281628855 1278376377 16777215 138420224 4281628855 4289252074 4292537599 4287207318 4286874756 4288652194 4288717730 4288717730 4288586401 4286874756 4287206547 4292537599 4289252074 4281628855 71319552 524958216 4281628855 4292537599 4287207061 4294375158 4293651435 4291940817 4291940817 4291940817 4291940817 4293651435 4294440951 4287206290 4292537599 4281628855 273694720 524958216 4281628855 4292537599 4287270541 4294967295 4294309365 4294309365 4294309365 4294309365 4294309365 4294309365 4294967295 4287270798 4292537599 4281628855 273694720 524958216 4281628855 4292537599 4287270541 4294967295 4294309365 4291546572 4291546572 4291546572 4291546572 4294243828 4294967295 4287270798 4292537599 4281628855 273694720 524958216 4281628855 4292537599 4287270541 4294967295 4294309365 4294309365 4294309365 4294309365 4294243828 4294112499 4294967295 4287270798 4292537599 4281628855 273694720 524958216 4281628855 4292537599 4287270541 4294967295 4294309365 4291546572 4291546572 4291480779 4291480522 4293980656 4294967295 4287270798 4292537599 4281628855 273694720 524958216 4281628855 4292537599 4287270541 4294967295 4294309365 4294243828 4294178035 4294046449 4293980656 4293322983 4294967295 4287270798 4292537599 4281628855 273694720 524958216 4281628855 4292537599 4287270541 4294967295 4294243572 4294112242 4294046449 4293980656 4293257190 4292665053 4294967295 4287270798 4292537599 4281628855 273694720 524958216 4281628855 4292537599 4287270541 4294967295 4294046706 4293980656 4293915120 4292730846 4291744208 4291678672 4294967295 4287270798 4292537599 4281628855 273694720 524958216 4281628855 4292537599 4287270541 4294967295 4293980656 4293849327 4293191397 4291743951 4294967295 4294967295 4294967295 4287270798 4292537599 4281628855 273694720 524958216 4281628855 4292537599 4287270541 4294967295 4293783791 4293125861 4292138709 4291612622 4294967295 4294967295 4287270798 4292537599 4292537599 4281628855 273694720 524958216 4281628855 4292537599 4287076246 4294178035 4294901502 4294901502 4294901502 4294243828 4294835965 4287270798 4292537599 4292537599 4292537599 4281628855 273694720 138420224 4281628855 4289252074 4292537599 4287142553 4287468177 4287468177 4287468177 4287468177 4287468177 4292537599 4292537599 4292537599 4289252074 4281628855 71319552 16777215 1278376377 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 1278376377 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 1/5/2011 20:43'!
smallPinIcon

	^ Icons
		at: #smallPinIcon
		ifAbsentPut: [ self smallPinIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 1/5/2011 20:43'!
smallPinIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1560237654 2550092884 872371541 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1560237654 4294923605 4294923605 4294923605 1560237654 16777215 16777215 16777215 16777215 4294923605 16777215 16777215 16777215 16777215 16777215 1560237654 4294923605 4294923605 4294936712 4294936712 4294923605 1560237654 16777215 16777215 16777215 4294923605 4294923605 16777215 16777215 16777215 1560237654 4294923605 4294923605 4294936712 4294936712 4294923605 4294923605 4294923605 872371541 16777215 16777215 1560237654 4294923605 4294923605 1560226346 2113885782 4294923605 4294923605 4294936712 4294936712 4294923605 4294923605 4294923605 4293283149 2550092884 16777215 16777215 16777215 2113885782 4294923605 4294923605 4294923605 4294923605 4294936712 4294936712 4294923605 4294923605 4294923605 4293283149 4291576900 2107979575 16777215 16777215 16777215 16777215 2902410822 4294923605 4294923605 4294936712 4294936712 4294923605 4294923605 4294923605 4293283149 4291576900 2107979575 16777215 16777215 16777215 16777215 16777215 872371541 3187614499 4294923605 4294923605 4294923605 4294923605 4294923605 4293283149 4291576900 2107979575 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1560237654 3405717793 4294923605 4294923605 4294923605 4293283149 4291576900 2107979575 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3212409209 4286151033 3405717793 4294923605 4294923605 4291773765 2545302845 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3212409209 4286151033 3212409209 1560237654 3187614499 4294923605 4294923605 1560226346 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3212409209 4286151033 3212409209 16777215 16777215 872371541 2902410822 4294923605 4294923605 16777215 16777215 16777215 16777215 16777215 16777215 2675472504 4286151033 3212409209 16777215 16777215 16777215 16777215 16777215 2113885782 4294923605 4294923605 16777215 16777215 16777215 16777215 2138601592 4286151033 2675472504 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1560237654 4294923605 4294923605 16777215 16777215 16777215 4286151033 2138601592 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:45'!
smallPrintIcon

	^ Icons
		at: #smallPrintIcon
		ifAbsentPut: [ self smallPrintIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:45'!
smallPrintIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 2176367288 4290296247 4290296247 4290296247 4290296247 4290296247 4290296247 4290296247 4290296247 4290296247 2176367288 16777215 16777215 16777215 16777215 16777215 4290296247 4294638330 4294638330 4294638330 4294638330 4294638330 4294638330 4294638330 4294638330 4294572537 4290296247 16777215 16777215 16777215 16777215 16777215 4290296247 4294638330 4291809231 4291809231 4291809231 4291809231 4291809231 4291809231 4291809231 4294572537 4290296247 16777215 16777215 16777215 16777215 16777215 4290296247 4294638330 4294243572 4294243572 4294243572 4294309365 4294309365 4294309365 4294309365 4294572537 4290296247 16777215 16777215 16777215 16777215 16777215 4290296247 4294572537 4291809231 4291809231 4291875024 4291875024 4291875024 4291940817 4291940817 4294572537 4290296247 16777215 16777215 16777215 16777215 16777215 4290296247 4294572537 4293322470 4293454056 4293519849 4293519849 4293585642 4293651435 4293717228 4294572537 4290296247 16777215 16777215 16777215 16777215 16777215 4290296247 4294506744 4289835441 4289835441 4289835441 4289967027 4293388263 4293388263 4293388263 4294572537 4290296247 16777215 16777215 16777215 2173603982 4287598991 4288125078 4290361785 4290295992 4290361785 4290361785 4290361785 4290361785 4290361785 4290295992 4290295992 4288125078 4287598991 2173603982 16777215 4287598991 4294835709 4294638330 4294375158 4294375158 4294375158 4294375158 4294375158 4294375158 4294375158 4294375158 4294375158 4293651435 4293783021 4287598991 16777215 4287598991 4294835709 4292664540 4294046193 4290559164 4294309365 4293256677 4293519849 4293388263 4293454056 4293322470 4293125091 4289703855 4292335317 4287598991 16777215 4287598991 4294177779 4292269525 4293190884 4293190884 4293190884 4293190884 4293190884 4293125091 4293190884 4293190884 4292927712 4288585116 4291875024 4287598991 16777215 4287598991 4294309365 4291480266 4288190358 4287466635 4287466635 4287203463 4287334792 4287203205 4286940290 4287071877 4287334792 4290756543 4291875024 4287598991 16777215 4287598991 4294309365 4291282887 4291151301 4291151301 4291151301 4291151301 4291151300 4291019715 4290822336 4290953922 4291085508 4291151301 4291940817 4287598991 16777215 4287598991 4293980400 4292861919 4292927712 4292927712 4292861919 4292861919 4292861919 4292861919 4292861919 4292730333 4292861919 4292796126 4292664540 4287598991 16777215 2811269520 4287598991 4287598991 4287598991 4287598991 4287598991 4287598991 4287598991 4287598991 4287598991 4287598991 4287598991 4287598991 4287598991 2811269520)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/31/2010 16:58'!
smallProjectBackIcon

	^ Icons
		at: #smallProjectBackIcon
		ifAbsentPut: [ self smallProjectBackIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/31/2010 16:58'!
smallProjectBackIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 2083747767 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 2083747767 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292537599 4292537599 4293176763 4293176763 4292537599 4292537599 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292537599 4292985039 4293751676 4293751676 4292985039 4292537599 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292793830 4293496216 4294901760 4294901760 4293496216 4292793830 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292985039 4294901760 4294901760 4294901760 4294901760 4292985039 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4290694609 4287727293 4287727293 4290694609 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4287595701 4281014002 4281014002 4287661239 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4287463852 4280948465 4280948465 4287464110 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4290628297 4287595700 4287595700 4290628298 4292537599 4281628855 16777215 16777215 2083747767 4291090666 4284271650 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 2083747767 16777215 16777215 3169845236 4278231040 4278231040 3053453311 3053453311 3053453311 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3053453311 4284271650 4289589680 4278231040 4278231040 4278231040 4278231040 3053453311 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281118976 4283491340 4278248192 4278247936 4278246144 4278244352 4278231040 3053453311 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3053453311 4281118976 4285526614 4278231040 4278231040 4278231040 4278231040 3053453311 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3169910772 4278231040 4278231040 3053453311 3053453311 3053453311 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/31/2010 16:56'!
smallProjectIcon

	^ Icons
		at: #smallProjectIcon
		ifAbsentPut: [ self smallProjectIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/31/2010 16:56'!
smallProjectIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 2083747767 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 2083747767 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292537599 4292537599 4293176763 4293176763 4292537599 4292537599 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292537599 4292985039 4293751676 4293751676 4292985039 4292537599 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292793830 4293496216 4294901760 4294901760 4293496216 4292793830 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292985039 4294901760 4294901760 4294901760 4294901760 4292985039 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4290694609 4287727293 4287727293 4290694609 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4287595701 4281014002 4281014002 4287661239 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4287463852 4280948465 4280948465 4287464110 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4290628297 4287595700 4287595700 4290628298 4292537599 4281628855 16777215 16777215 2083747767 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 2083747767 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/31/2010 16:54'!
smallProjectJumpIcon

	^ Icons
		at: #smallProjectJumpIcon
		ifAbsentPut: [ self smallProjectJumpIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/31/2010 16:54'!
smallProjectJumpIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 2083747767 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 2083747767 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292537599 4292537599 4293176763 4293176763 4292537599 4292537599 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292537599 4292985039 4293751676 4293751676 4292985039 4292537599 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292793830 4293496216 4294901760 4294901760 4293496216 4292793830 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292985039 4294901760 4294901760 4294901760 4294901760 4292985039 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4290694609 4287727293 4287727293 4290694609 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4287595701 4281014002 4281014002 4287661239 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4287463852 4280948465 4280948465 4287464110 4292537599 4281628855 16777215 16777215 4281628855 4292537599 4278239744 4278239744 4278239744 4278239744 4292537599 4292537599 4290628297 4287595700 4287595700 4290628298 4292537599 4281628855 16777215 16777215 2083747767 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4284271650 4291090666 2083747767 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3053453311 3053453311 3053453311 4278231040 4278231040 3500729258 16777215 16777215 1275109632 4278231040 1275109632 4278231040 1275109632 4278231040 1275109632 3053453311 4278231040 4278231040 4278231040 4278231040 4278241792 4278231040 3417498546 16777215 4278231040 4278246144 4278231040 4278246144 4278231040 4278246144 4278231040 3053453311 4278231040 4278246144 4278246144 4278246144 4278246144 4278241792 4278231040 16777215 1275109632 4278231040 1275109632 4278231040 1275109632 4278231040 1275109632 3053453311 4278231040 4278231040 4278231040 4278231040 4278241792 4278231040 3417498546 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3053453311 3053453311 3053453311 4278231040 4278231040 3500729258 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/31/2010 15:23'!
smallProjectLoadIcon

	^ Icons
		at: #smallProjectLoadIcon
		ifAbsentPut: [ self smallProjectLoadIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/31/2010 15:23'!
smallProjectLoadIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 1278376377 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 1278376377 16777215 16777215 16777215 16777215 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4294928966 4294928966 4292537599 4292537599 4292537599 4281628855 16777215 16777215 16777215 16777215 4281628855 4281628855 4281628855 4292537599 4292537599 4294949256 4294901760 4294901760 4294949256 4292537599 4292537599 4281628855 16777215 16777215 1412725174 4281628855 4281628855 4287412967 4281628855 4288923367 4292537599 4294933841 4294901760 4294901760 4294933841 4292537599 4292537599 4281628855 16777215 1412725174 4281628855 4290304767 4290304767 4290304767 4287412967 4281628855 4278247936 4278247936 4292537599 4292537599 4290947299 4290947299 4292537599 4281628855 16777215 4281628855 4290304767 4287412967 4281628855 4287412967 4281628855 4279353404 4278247936 4278247936 4292537599 4290947299 4282515711 4282515711 4290947299 4281628855 16777215 4281628855 4290304767 4281628855 4281628855 4281628855 4281628855 4278247936 4278247936 4278247936 4292537599 4290947299 4282515711 4282515711 4290947299 4281628855 16777215 4281628855 4281628855 4281628855 16777215 16777215 1278376377 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 1278376377 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4294770683 4294770683 4294770683 4294770683 4294770683 4294770683 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4294770683 4294770683 4294770683 4294770683 4294770683 4294770683 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4294770683 4294770683 4294770683 4294770683 4294770683 4294770683 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4294770683 4294770683 4294770683 4294770683 4294770683 4294770683 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/31/2010 14:29'!
smallProjectNextIcon

	^ Icons
		at: #smallProjectNextIcon
		ifAbsentPut: [ self smallProjectNextIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/31/2010 14:29'!
smallProjectNextIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 3178078063 4282619024 4282619024 4282619024 4282619024 4282619024 4282619024 4282619024 4282619024 4282619024 4282619024 4282619024 4282619024 3178078063 16777215 16777215 4282619024 4292666586 4292666586 4292666586 4292666586 4292666586 4294159759 4294159503 4292666586 4292666586 4292666586 4292666586 4292600791 4282619024 16777215 16777215 4282619024 4292666587 4292337348 4292337348 4292337348 4293637757 4294139970 4294139970 4293571966 4292271812 4292271812 4292337350 4292600791 4282619024 16777215 16777215 4282619025 4292666589 4292337349 4292337349 4292337349 4294079056 4294901760 4294901760 4294079056 4292337349 4292337349 4292271812 4292600790 4282619024 16777215 16777215 4282619025 4292732381 4292337608 4292337608 4292337608 4294901760 4294901760 4294901760 4294901760 4292337350 4292271810 4292271553 4292600791 4282619024 16777215 16777215 4282619025 4292666586 4292337609 4285977428 4285977170 4285911632 4285911631 4292271554 4291084195 4287727293 4287727293 4291084196 4292534996 4282619024 16777215 16777215 4282619025 4292469200 4292140219 4278239744 4278239744 4278239744 4285713988 4292074168 4287595701 4281014002 4281014002 4287661239 4292337610 4282619024 16777215 16777215 4282619025 4292140221 4291942577 4278239744 4278239744 4278239744 4285582397 4291942577 4287463852 4280948465 4280948465 4287464110 4292140221 4282619024 16777215 16777215 4282619026 4292074170 4292074167 4278239744 4278239744 4278239744 4285713987 4292074167 4290886292 4287595700 4287595700 4290886293 4292073912 4282619024 16777215 16777215 4282619026 4291415968 4291415968 4291415968 4291415968 4291415968 4291415968 4291415968 4291415968 4291415968 4284271650 4293915107 4291153049 4282619024 16777215 16777215 3178078063 4282619024 4282619024 4282619024 4282619024 4282619024 4282619024 4291354847 4291354847 4291354847 4278231040 4278231040 4290962394 3178078063 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3053453311 4278231040 4278231040 4278231040 4278231040 4289589680 4284271650 3053453311 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3053453311 4278231040 4278244352 4278246144 4278247936 4278248192 4283491340 4281118976 3053453311 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3053453311 4278231040 4278231040 4278231040 4278231040 4285526614 4281118976 3053453311 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 3053453311 3053453311 3053453311 4278231040 4278231040 3169910772 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281118976 3053453311 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/31/2010 15:18'!
smallProjectSaveIcon

	^ Icons
		at: #smallProjectSaveIcon
		ifAbsentPut: [ self smallProjectSaveIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/31/2010 15:18'!
smallProjectSaveIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 1278376377 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 1278376377 16777215 16777215 16777215 16777215 16777215 16777215 4281628855 4292537599 4292537599 4292537599 4294928966 4294928966 4292537599 4292537599 4292537599 4281628855 16777215 16777215 16777215 16777215 16777215 16777215 4281628855 4292537599 4292537599 4294949256 4294901760 4294901760 4294949256 4292537599 4292537599 4281628855 16777215 16777215 16777215 16777215 16777215 16777215 4281628855 4292537599 4292537599 4294933841 4294901760 4294901760 4294933841 4292537599 4292537599 4281628855 16777215 16777215 16777215 16777215 16777215 16777215 4281628855 4278247936 4278247936 4278247936 4292537599 4292537599 4290947299 4290947299 4292537599 4281628855 16777215 4281628855 4281628855 1412725174 16777215 16777215 4281628855 4278247936 4278247936 4278247936 4292537599 4290947299 4282515711 4282515711 4290947299 4281628855 16777215 4281628855 4290304767 4281628855 1412725174 16777215 4281628855 4278247936 4278247936 4278247936 4292537599 4290947299 4282515711 4282515711 4290947299 4281628855 16777215 4281628855 4287412967 4290304767 4281628855 16777215 1278376377 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 4281628855 1278376377 16777215 16777215 4281628855 4290304767 4281628855 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281628855 4281628855 4290304767 4281628855 4281628855 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4286151288 4286151288 4281628855 4287412967 4290304767 4287412967 4281628855 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4294770683 4294770683 4290434020 4281628855 4287412967 4281628855 4284643980 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4294770683 4294770683 4294770683 4290434020 4281628855 4290434020 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4294770683 4294770683 4294770683 4294770683 4294770683 4294770683 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4294770683 4294770683 4294770683 4294770683 4294770683 4294770683 4286151288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288 4286151288)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
smallPublishIcon
	"Private - Generated method"
	^ Icons
			at: #'smallPublish'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallPublishIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/6/2004 17:26'!
smallPublishIconContents
	"Private - Method generated with the content of the file /home/dgd/smallPublish.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAADG0lEQVQ4jY2Sb2jUdQDGP9/fn7vN29+7tv2c
t5tenulsUd1miWlEFCjZzRdRVPiiEnohBBaLIiRIg9heDAuiFwVJONgcWrBilajlKlKU5cEa
zrlluzvnbfd/9/f3/fYiKgoOel4/z4cHnkfwP7Tl4X2PSanvEeBXyCkp1PC1705NA4h/Obfi
oMU8R6fLQzi5H/9Tl+6+rT7b3Nn4RK63Lxap73LZ878sVc6NNqhycWjmwtgx/e9wkG4cvM2T
vr5nB171zPiSL3gmbgeeeyjwzIcHLWf7XRvdzzddrf+hZa9V3ravwb48scnjvTOsE8R0Pej5
eus7oYH0675guUcn7IzTGLDE8tpE99D6dmFaXUTjgu1101wvdjAj12rC11WSV852GMChbe89
/ehZbRy+SkuUlCrkMlYSNxC/RVh0P0Kdqmd4+T70smRBNrO9McLSmlYrLES3galtDnsXYSxt
M5TowGsexrr8MhP2zc7JHkfh8M62QGUKZ/NO3sr2MbnlA0gvk8gpHq/RShoV+X08lgS30pFU
iJf7eS3h52jaVyM9M0em10XHY/dwJdNK3jYoFAVfxO7njRG1ksmXRzTyjKjha1FWAEEjk2T4
mRt/TiRfKp4ctN+c8kWKUueA5yJOmWHwTGrh/E+zN+26/FFxoIfdjpaOY7hb3Tu6g9Lpcmmm
oQtNEyCEmJu/pY5fiIqcozXV1lzrKK/mGlK50qlVQxyc+3Y0ZRiGufv98amNQiVAFliKzDJ3
8ThSr6d3VwhTt7HuOK8Ojcbay6vZ9tl15jzjJ+2/1jdavAFfIrlEPFkCVaFGGBTzaeoaHZh6
BWSB2lpd/f7jaB64/t+XGu42r5XKJKnYLYCipBbZFXoFoQokUnFqTYXDIWS1mxuFbHp1g+Ui
uzLHrx+fxtXrR7StZ/+nQxhK0NHkYgf3qmoAbfnWQgTg6okvCQ5+TnRgDGkXiMbiWKzBUIpS
SaveIBWPRjKpEr0vhrjU2syG4CaE3sQ3/e+CLIDWwEdHPslVBSjBif69D/iVbetK/dNUMwxh
mk7NcNSI1WzydDXAH4NwUmsvAPcyAAAAAElFTkSuQmCC'! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/22/2010 00:16'!
smallQuitIcon

	^ Icons
		at: #smallQuitIcon
		ifAbsentPut: [ self smallQuitIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/22/2010 00:16'!
smallQuitIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 4284901482 4290098613 4288124823 4287401100 4286677377 4286085240 4285558896 4285295724 4294046194 4294112244 4294309366 4294440952 4294440952 4293980658 4287598995 679904902 4284901482 4290493371 4288585374 4287730065 4287137928 4286545791 4286019447 4285756275 4294046194 4294309365 4294572537 4294704124 4294769916 4294638330 4289243563 1066044042 4284901482 4291217094 4289111718 4288387995 4287664272 4287137928 4286677377 4286282619 4294046194 4294309366 4294638330 4288727272 4282685399 4294704123 4289769906 1183484554 4284901482 4291282887 4289572269 4289045925 4288387995 4287730065 4287203721 4286808963 4294046451 4294112502 4285771742 4284261853 4278219466 4292536564 4288260277 1736008038 4284901482 4291414473 4289769648 4289243304 4288716960 4288190616 4287664272 4287203721 4293324017 4282553556 4285444580 4289977081 4281306583 4280321491 4279467214 4278544569 4284901482 4291480266 4289835441 4289374890 4288848546 4288322202 4287795858 4286351764 4279335885 4285707496 4288007672 4288401400 4288861177 4288926713 4289189626 4278348735 4284901482 4291480266 4289967027 4289506476 4288980132 4288519581 4285894306 4278218436 4285182955 4285906421 4283345650 4283542770 4283674098 4283805170 4288335864 4278348992 4284901482 4291480266 4290164406 4289638062 4289177511 4288585374 4278217663 4281832927 4284527603 4280522222 4278946541 4278224107 4278224106 4278224364 4286300406 4278348992 4284901482 4291480266 4290295992 4289835441 4289243304 4288716960 4286944417 4278217920 4280782815 4281638640 4278223849 4278223849 4278223849 4278224106 4284133619 4278348735 4284901482 4291480266 4290295992 4289901234 4289374890 4288914339 4288387995 4287337370 4280057289 4279600088 4281835503 4282689009 4282097645 4282294251 4282425579 4278348735 4284901482 4291480266 4290295992 4290032820 4289572269 4289111718 4288716960 4287927444 4285625720 4283604181 4278220498 4281572332 4278218436 4278482381 4278218435 4079241345 4284901482 4291480266 4290295992 4290295992 4289835441 4289309097 4286743170 4288322460 4292467419 4293454314 4286230746 4278219466 4278218951 4293520107 4288190874 814056837 4284901482 4291480266 4290361785 4290624957 4288124824 4288454561 4291875283 4292533470 4292730592 4292796385 4292862178 4288002776 4283209424 4292665056 4286875272 495223940 4284901482 4291743438 4289177769 4288783012 4291283661 4291612369 4291743954 4291743955 4291809747 4291809747 4291809747 4291809747 4291809747 4291612626 4285690998 176193664 3309849418 4286282878 4284967275 4285033068 4285098861 4285098861 4285098861 4285098861 4285164654 4285164654 4285164654 4285164654 4285164654 4285098861 3326889805 41975936 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 21:23'!
smallQuitNoSaveIcon

	^ Icons
		at: #smallQuitNoSaveIcon
		ifAbsentPut: [ self smallQuitNoSaveIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 21:23'!
smallQuitNoSaveIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 4284967526 4290098613 4288124823 4287401100 4286677377 4286085240 4285558896 4285295724 4294111985 4294243570 4294375157 4294506743 4294506743 4293980912 4287730575 369098751 4284967526 4290493371 4288585374 4287730065 4287137928 4286545791 4286019447 4285756275 4294111985 4294309365 4294572537 4294769915 4294769916 4294638330 4289375144 587202559 4284967526 4291217094 4289111718 4288387995 4287664272 4287137928 4286677377 4286282619 4294111985 4294375157 4294638330 4293435552 4292297797 4294704123 4289835696 654311423 4284967526 4291282887 4289572269 4289045925 4288387995 4287730065 4287203721 4286808963 4294177777 4294374642 4292768627 4292697180 4291428352 4294237146 4290091673 988390307 4284967526 4291414473 4289769648 4289243304 4288716960 4288190616 4287664272 4287203721 4294043878 4292100674 4293160558 4294554547 4292292400 4292026402 4291695379 4173529088 4284967526 4291480266 4289835441 4289374890 4288848546 4288322202 4287795858 4287921276 4291629331 4293423730 4294481302 4294482843 4294550178 4294550435 4294616999 4240637952 4284967526 4291480266 4289967027 4289506476 4288980132 4288519581 4288836982 4291035136 4293618282 4294276470 4294069839 4294070609 4294071124 4294071637 4294482586 4240703488 4284967526 4291480266 4290164406 4289638062 4289177511 4288585374 4290707456 4292818744 4294140000 4293796644 4293724943 4293591040 4293525504 4293656576 4294343547 4240703488 4284967526 4291480266 4290295992 4289835441 4289243304 4288716960 4288775557 4290772992 4292814633 4293932085 4293459968 4293459968 4293459968 4293525504 4294138458 4240637952 4284967526 4291480266 4290295992 4289901234 4289374890 4288914339 4288387995 4288318347 4291370268 4292351255 4293867320 4294001733 4293737276 4293606975 4293607489 4257415168 4284967526 4291480266 4290295992 4290032820 4289572269 4289111718 4288716960 4287927444 4286083697 4292170578 4291952640 4293669684 4291035136 4291625991 4290969600 3368550400 4284967526 4291480266 4290295992 4290295992 4289835441 4289309097 4286743170 4288388250 4292467673 4293520104 4292508538 4291428352 4291231744 4293585897 4288256664 436207615 4284967526 4291480266 4290361785 4290624957 4288125079 4288586140 4291941328 4292665050 4292796637 4292862430 4292928223 4292384661 4291841612 4292731100 4287006852 268435455 4284967526 4291743438 4289243559 4288914593 4291415495 4291678668 4291875534 4291875790 4291875791 4291875791 4291875791 4291875791 4291875791 4291809996 4285822578 100663295 2238541675 4286348923 4285098855 4285164648 4285230441 4285230441 4285230441 4285230441 4285296234 4285296234 4285296234 4285296234 4285296234 4285230441 2255647856 33554431 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 21:20'!
smallRedoIcon

	^ Icons
		at: #smallRedoIcon
		ifAbsentPut: [ self smallRedoIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 21:20'!
smallRedoIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4282691328 860270336 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4282691328 4282691328 1128640256 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4282691328 4289526392 4282691328 1094955264 16777215 16777215 16777215 16777215 16777215 16777215 16777215 122271232 1178971904 2655301120 3477384960 4282691328 4282691328 4289526392 4289526392 4282691328 1094955264 16777215 16777215 16777215 16777215 16777215 910339840 2974068224 3298668032 3719096654 4289591414 4289591671 4289919871 4289001324 4285128704 4289526392 4282691328 1094955264 16777215 16777215 16777215 1246081024 3359944448 4256228425 4289722227 4288737880 4288866048 4288866048 4288866048 4288866048 4285128704 4285128704 4289526392 4282691328 998676224 16777215 725725696 3125128960 4289586251 4289853296 4289128704 4288080156 4288866048 4288866048 4290044672 4285128704 4285128704 4285128704 4285260032 4289526392 4282691328 16777215 2655301120 3936735531 4289852778 4289194502 4290044672 4290044672 4291877888 4285586944 4285128704 4285128704 4285128704 4285128704 4286901554 4282691328 1094955264 16777215 3662721792 4289917550 4290701860 4292009728 4288537088 4288931328 4288931328 4288931328 4288931328 4292864000 4285128704 4286901554 4282691328 1094955264 16777215 16777215 4048204032 4289983599 4292141312 4272009984 4282691328 4282691328 4282691328 4282691328 4282691328 4292075264 4286901554 4282691328 1094955264 16777215 16777215 16777215 4282691328 4290964552 4289116160 4047810304 1933946624 809873152 16777215 16777215 4282691328 4286901554 4282691328 1094955264 16777215 16777215 16777215 16777215 4282691328 4256752690 4282691328 1212395008 16777215 16777215 16777215 16777215 4282691328 4282691328 860270336 16777215 16777215 16777215 16777215 16777215 4282691328 4122863670 4265914112 16777215 16777215 16777215 16777215 16777215 4282691328 860270336 16777215 16777215 16777215 16777215 16777215 16777215 3242438144 3333936384 3466506240 1212395008 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1581559808 2253699584 2529681430 2359929600 205564416 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 541634304 1782820864 1531162112 491171840 79675136 61516288 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 23:00'!
smallRemoteOpenIcon
	"Private - Generated method"
	^ Icons
			at: #'smallRemoteOpen'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallRemoteOpenIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/6/2004 17:12'!
smallRemoteOpenIconContents
	"Private - Method generated with the content of the file /home/dgd/smallRemoteOpen.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABIAAAASCAYAAABWzo5XAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAAD2klEQVQ4ja3T+0/VdRzH8efn+/18OTcOnJtc
zhFCg+EVD56BiYU4qQw1XdocttSt6VrL1fyh1vpBN7dq9mPZlv1Wm1OsDRZqU/DSkKkFelKS
wEsonANM1AOHw/d7Lt/TD9lWW2219f4DHnu/fnjC/3RidzkWIx+r1MkkLcxVBDcO9ZD6r5A0
8jiTY7XWyTzNrG9aq4TPnU/sDEXbEHwjspw81EPiX320K0T7gfbWtWORcc61tFBSWUnZwoX0
dnTQ09k5E4/FjitZvlYNjn/WR/yfIDUUYGFDrfpMyawEoRe3YySh5cAH7HpvG2t2vq2VLapa
kM2am8fHRvYsKcjU1BQjQ36GeqIYf4GWFuGrCC56uTjvEUz0UlBRzVRCYWqgk1JLmCJvhlBD
Hat3vCGL5lTMM/SZlx5Eo3uW+rO1NX76f4wwCqAuC5B2eArfXFw1G0QOZE0cpTV8d/QEoWXl
qGYCpm4jH1yixC956oXnWbl1l2pzuipvhsM7gn6zpXeECXVdhIfj7vQ79WtWaD/fy9AVjhET
RXiLA5w6doaxsQROtxun2w3pBEzexDJ1lYrqJVh88+T17ou23ijfyn1gNg/Hr2zeN1SXVG3k
BOZgvznKSH+E5cGNNK4OcPnCKUZb+ygpyadpQxCRTYE+wRPz5xM3RDnkeuXip7e6b9mM4Meb
illWW8prt5p5zjXBkChjuO0TPj38Awf3b0M4Z3O96zxHDrfSvH0lCAuqlAAapOcraWVm/871
flv9AgsXp5/kfVcb613XyAhJcMtb9EcFHae7IdLFouXLyS9dwODtSVCsKFKigAbZHMVplWub
an1i2l7EsbtzqbRH0UhimJCjKpStWMWFn6ZAOiA+QsOWV+i5fAe0PKQqEAIVMGU6K9xfjDey
t+AsxZ4Mtbfe5dlZD9jt78KuqkzmjjE0LDBMK5Z0EnuuxO4pQk/noKgqQkGCMJT4tNF3YsKj
p5IKm51XMYXCptyL+Geu4IqdJXZviEKvHc2WD5oD9Bh1GzYS7hlAVSWCx1AmlTo62XZwbN2d
1/loaAUAg7qLG/dtNHc3cvpKjODiEhRL7u/zEPgKvDg8hSjyD0hJqg8jg5dcroIN8TsD6fvl
q1xznLrY62sl/MtE9qtD7Y/kve5f49c6s+N3hx0zehprngdHfh4Ojw9D1+k40vIoPCq+FI9T
sZZWNXxoc3k3qtKSa5OmLZVMDURu931+/27/CWCkrkyr8lozTXY1Wx8I+JZWVi+ZZZom506e
/f7odeur4k/dCSCgaZovlUpJIAZE4e+Kz3dXFU5Xq6oouzGuDupp/fJvhgd0/Din8aMAAAAA
SUVORK5CYII='! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:39'!
smallRightFlushIcon

	^ Icons
		at: #smallRightFlushIcon
		ifAbsentPut: [ self smallRightFlushIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:39'!
smallRightFlushIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 2862522779 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 4288257175 2862522779 16777215 16777215 4288257175 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4289638062 4289440683 4289440683 4289440683 4289440683 4289440683 4289440683 4294440950 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294440950 4294440950 4294440950 4294440950 4294440950 4294440950 4294440950 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294572537 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4289638062 4289440683 4289440683 4289440683 4289440683 4294309365 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294440950 4294440950 4294440950 4294375157 4294309365 4294440950 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4294572537 4294572537 4294572537 4294572536 4294506743 4294440950 4294375157 4294309365 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572537 4289440683 4289309097 4289309097 4289309097 4289309097 4289309097 4289309097 4294309364 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572537 4294572536 4294375157 4294309365 4294309364 4294243571 4294177778 4294111985 4294046193 4294177779 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294572536 4294506743 4294440950 4294375157 4294309365 4294309365 4294243572 4294177779 4294177778 4294111985 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294440950 4294375158 4294309365 4294309365 4294309364 4289440683 4289309097 4289309097 4289309097 4293848814 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294375157 4294309365 4294309364 4294243571 4294177779 4293980400 4293914607 4293848814 4293783021 4293980399 4294901502 4288454554 16777215 16777215 4288257175 4294967295 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294901502 4294967295 4288454554 16777215 16777215 2862522779 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 4288454554 2862522780 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/20/2010 18:48'!
smallSaveAsIcon

	^ Icons
		at: #smallSaveAsIcon
		ifAbsentPut: [ self smallSaveAsIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/20/2010 18:48'!
smallSaveAsIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 2088343690 2119535780 1866038431 2352512158 3560536991 4281957279 4165041566 3847390612 2237959310 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1581744805 2120722634 1569307355 3451508718 4292273651 4291617521 4288990696 4252212658 4249387676 461084520 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4287143846 4285694857 4283662745 4282615725 4282813362 4289252322 4291355120 4283799744 4284120721 4286677886 4286677886 4286677886 4286677886 4287796623 16777215 16777215 4287534739 4294967295 4293126109 4291944164 4286032581 4281957279 4289975786 4288464094 4284256430 4292601323 4294967295 4294967295 4294967295 4287862160 16777215 16777215 4287994512 4294046193 4294046193 4293257444 4286818752 4281957279 4286756567 4288266713 4282416802 4290234575 4294046193 4293980657 4293980657 4287862160 16777215 16777215 4287862160 4293914864 4281957279 4281957279 4281957279 4281957279 4286756567 4284195788 4281957279 4281957279 4281957279 4281957279 4293980400 4287862160 16777215 16777215 4287862160 4293783022 4292073678 4281957279 4288793573 4286230477 4286230477 4286230477 4286361550 4288661474 4281957279 4288854206 4293848558 4287862160 16777215 16777215 4287862160 4294704123 4293519849 4291810762 4281957279 4288793573 4286361550 4286361550 4290040551 4281957279 4288656827 4293519849 4294375158 4287862160 16777215 16777215 4287862160 4294046193 4294704123 4293519849 4291810762 4281957279 4290106858 4288793573 4282154659 4288656827 4293519849 4294704123 4294046193 4287862160 16777215 16777215 4287862160 4293519849 4294046193 4294704123 4294704123 4292271570 4282613922 4281957279 4289182914 4294704123 4294835709 4294046193 4293519849 4287862160 16777215 2158208927 4289901488 4289835952 4289901488 4289901488 4289901488 4289901488 4289901488 4289901488 4289901488 4289901488 4289901488 4289901488 4289901488 4289901488 2158208927 4286677886 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4278190080 4294967295 4294967295 4286677886 4286677886 4294967295 4290625211 4294967295 4290625212 4294967295 4290625211 4294967295 4290625211 4294967295 4291085763 4294967295 4278190080 4294967295 4294967295 4286677886 4286677886 4294967295 4291085763 4294967295 4291085763 4294967295 4291085763 4294967295 4291085763 4294967295 4291085763 4294967295 4278190080 4294967295 4294967295 4286677886 4286677886 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4278190080 4294967295 4294967295 4286677886 2155971454 4286677886 4286677886 4286677886 4286677886 4286677886 4286677886 4286677886 4286677886 4286677886 4286677886 4286677886 4286677886 4286677886 4286677886 2155971454)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/21/2010 20:30'!
smallSaveIcon

	^ Icons
		at: #smallSaveIcon
		ifAbsentPut: [ self smallSaveIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/21/2010 20:30'!
smallSaveIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 2089200540 2119802550 1865322161 2351730865 3559886514 4281306802 4164849841 3847723431 2238554017 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1581749686 2120857815 1569835238 3451969523 4292602614 4292012277 4289452015 4252872386 4249392559 461679227 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4287999926 4286748316 4284126380 4282227390 4282424771 4289779178 4291749876 4283738575 4284780964 4287862160 4287862160 4287862160 4287862160 4288849568 16777215 16777215 4288587684 4294967295 4293455075 4292404714 4286495186 4281306802 4290436848 4288991207 4284654271 4292930288 4294967295 4294967295 4294967295 4288980641 16777215 16777215 4289047457 4294243572 4294243572 4293586409 4287477965 4281306802 4287218657 4288859618 4282159540 4290892761 4294243572 4294178036 4294178036 4288980641 16777215 16777215 4288980641 4294112243 4281306802 4281306802 4281306802 4281306802 4287218657 4284068825 4281306802 4281306802 4281306802 4281306802 4294177779 4288980641 16777215 16777215 4288980641 4294046193 4292534487 4281306802 4289254892 4286693081 4286693081 4286693081 4286823898 4289188586 4281306802 4289578443 4294111729 4288980641 16777215 16777215 4288980641 4294769916 4293783021 4292337107 4281306802 4289254892 4286823898 4286823898 4290567150 4281306802 4289446856 4293783021 4294506744 4288980641 16777215 16777215 4288980641 4294243572 4294769916 4293783021 4292337107 4281306802 4290567664 4289254892 4281635253 4289446856 4293783021 4294769916 4294243572 4288980641 16777215 16777215 4288980641 4293783021 4294243572 4294769916 4294769916 4292732122 4282422196 4281306802 4289907150 4294769916 4294835709 4294243572 4293783021 4288980641 16777215 16777215 4288980641 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294506744 4294506744 4294111986 4294111986 4293783021 4288980641 16777215 16777215 4288980641 4288980641 4288980641 4288980641 4288980641 4288980641 4288980641 4288980641 4288980641 4288980641 4288980641 4288980641 4288980641 4288980641 16777215 16777215 4288980641 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4288980641 16777215 16777215 4288980641 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4288980641 16777215 16777215 4288980641 4291151301 4291151301 4291151301 4291151301 4291151301 4291151301 4291151301 4291151301 4291151301 4291151301 4291151301 4291151301 4288980641 16777215 16777215 4288849568 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4288849568 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/20/2010 18:48'!
smallSaveNewIcon

	^ Icons
		at: #smallSaveNewIcon
		ifAbsentPut: [ self smallSaveNewIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/20/2010 18:48'!
smallSaveNewIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 2089200540 2119802550 1865322161 2351730865 3559886514 4281306802 4164849841 3847723431 2238554017 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1581749686 2120857815 1569835238 3451969523 4292602614 4292012277 4289452015 4252872386 4249392559 461679227 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4287999926 4286748316 4284126380 4282227390 4282424771 4289779178 4291749876 4283738575 4284780964 4287862160 4287862160 4287862160 4287862160 4288849568 16777215 16777215 4288587684 4294967295 4293455075 4292404714 4286495186 4281306802 4290436848 4288991207 4284654271 4292930288 4294967295 4294967295 4294967295 4288980641 16777215 16777215 4289047457 4294243572 4294243572 4293586409 4287477965 4281306802 4287218657 4288859618 4282159540 4290892761 4294243572 4294178036 4294178036 4288980641 16777215 16777215 4288980641 4294112243 4281306802 4281306802 4281306802 4281306802 4287218657 4284068825 4281306802 4281306802 4281306802 4281306802 4294177779 4288980641 16777215 16777215 4288980641 4294046193 4292534487 4281306802 4289254892 4286693081 4286693081 4286693081 4286823898 4289188586 4281306802 4289578443 4294111729 4288980641 16777215 16777215 4288980641 4294769916 4293783021 4292337107 4281306802 4289254892 4286823898 4286823898 4290567150 4281306802 4289446856 4293783021 4294506744 4288980641 16777215 16777215 4288980641 4294243572 4294769916 4293783021 4292337107 4281306802 4290567664 4289254892 4281635253 4289446856 4293717228 4294704123 4294177779 4288980641 16777215 16777215 4288980641 4293783021 4294243572 4294769916 4294769916 4292732122 4282422196 4281306802 4289907150 4294769916 4294901502 4285890304 4294638330 4288980641 16777215 16777215 4288980641 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294440951 4294835709 4285890304 4294508288 4285890304 4293190883 16777216 16777215 4288980641 4292598747 4292598747 4292598747 4292598747 4292467161 4292598747 4292598747 4294243572 4285890304 4289311744 4294508288 4289311744 4285890304 3003121663 16777215 4288980641 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4285890304 4294508288 4294508288 4294508288 4294508288 4294508288 4285890304 16777215 4288980641 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4291085252 4293783021 4285890304 4289311744 4294508288 4289311744 4285890304 3003121663 16777215 4288980641 4291151301 4291151301 4291151301 4291151301 4291151301 4291151301 4291151301 4291151301 4293783021 4285890304 4294508288 4285890304 4293190883 16777215 16777215 4288849568 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4287862160 4292796381 4285890304 4292796381 4288849568 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/21/2010 22:00'!
smallScreenshotIcon

	^ Icons
		at: #smallScreenshotIcon
		ifAbsentPut: [ self smallScreenshotIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/21/2010 22:00'!
smallScreenshotIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 1982540587 4283392626 4284446346 4284446346 4284446346 4265561434 1462117926 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4282009685 4294967295 4294967295 4294967295 4294967295 4278190079 4282075478 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4282010199 4292209663 4294967295 4294967295 4294967295 4289649663 4282075478 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281483596 4285105307 4288332514 4288859118 4289649151 4285237151 4282207836 16777215 16777215 16777215 16777215 2738567995 4283524727 4283129455 4282339422 4282471266 4282997611 4287410126 16777215 16777215 16777215 4288595688 4281088066 4282207836 4282339422 4283393141 2738567995 4284315017 4291028223 4290371327 4294967295 4294967295 4289649663 4287871704 4288793068 4289123060 4289583358 4289649151 4291553535 4289649151 4291028223 4291815679 4262992929 4284907927 4289649663 4284578703 4289386490 4289649663 4288595688 4291422207 4292472319 4291815679 4289649663 4288464102 4288464102 4287673812 4285500581 4290633983 4262992929 4284512397 4289452028 4284578703 4285830059 4288661482 4291028223 4290043391 4288661482 4288201184 4288201184 4287805142 4289386490 4285895853 4283524727 4289649663 4262992929 4284248967 4289188854 4284578703 4286093489 4289452028 4288135134 4288595688 4284512397 4281351496 4283656571 4287673812 4286159027 4288398052 4284117381 4289649663 4262992929 4284248967 4289188854 4284578703 4286093489 4288595688 4285500581 4281614926 4280429620 4279440147 4279440147 4281483596 4285105307 4285500581 4285171101 4289452028 4262992929 4284248967 4289188854 4284578703 4286685886 4285632423 4282997611 4282405472 4294967295 4289649663 4283393141 4281219910 4284973977 4285171101 4286356407 4289452028 4262795547 4284248967 4289188854 4284578703 4287476176 4285171101 4281746770 4283393141 4289649663 4294047999 4291422207 4284248967 4286159027 4285237151 4285830059 4289386490 4262795547 4284248967 4289188854 4284644497 4286685372 4285171101 4283063661 4279440147 4285961903 4292997631 4292866559 4286817728 4287673812 4285237151 4285961903 4288661482 4262269971 4283854207 4287937498 4288332771 4288003291 4288332514 4284842133 4280693050 4283524727 4284248967 4287410126 4290239999 4288201184 4288464102 4287015107 4286357176 4198056779 2738567995 4283854207 4283392626 4283392626 4262335254 4285171101 4285434531 4283393141 4284315017 4286093489 4284512397 4284051331 4283524727 4283392626 4283392369 2738567995 16777215 16777215 16777215 16777215 1261844022 2218408506 3578481483 4250166386 4283985793 4250166386 3595324751 1849178168 858927666 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 22:28'!
smallSearchIcon

	^ Icons
		at: #smallSearchIcon
		ifAbsentPut: [ self smallSearchIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 22:28'!
smallSearchIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 2844298120 4118971263 4270032000 4286677886 4102062717 2542571658 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1049398408 4169171584 4275361492 4292993763 4275887841 4292468193 4292796385 4275360979 4152328574 982356365 16777215 16777215 16777215 16777215 16777215 898667920 4269966464 4292993762 4188256462 4051150803 4202607067 4286296026 4050822352 4188584911 4293190884 4253123454 832018322 16777215 16777215 16777215 16777215 4118774141 4292993505 4204180946 4269518811 3989823728 3957780472 4242861558 3736589034 4252741594 4187666386 4293059298 4085153916 16777215 16777215 16777215 1989120396 4206278581 4188782032 4269518810 3890014453 3958108664 4042257914 4243058680 3705071349 3265316323 4269453018 4189701845 4155683760 1871811470 16777215 16777215 4286348921 4292993504 4285571534 4189508073 4225821173 4226149879 4209504247 4293258742 4040549618 3616191967 3850022619 4287476438 4292927711 4286348921 16777215 16777215 4269900670 4292862434 4286230234 3418150637 3453149937 3336103923 3235703028 3856394484 1854058716 2054926555 2340269274 4286230234 4293454056 4269900670 16777215 16777215 4202725756 4293256677 4269846745 3013262819 3049314029 2999902193 2580077294 2762853096 1517990106 1820176348 2189209564 4235700950 4293059298 4185883004 16777215 16777215 3348469650 4224700367 4237407702 3312690387 2340269274 1970974426 1585098972 1920839387 1333702363 1702801116 3211500488 4254447831 4174631890 3231094675 16777215 16777215 1654103191 3801388691 4292403431 4286296026 2558111192 2004528861 1652207325 2172366553 1434169051 2071374035 4286296026 4292468967 3933895800 2122877572 16777215 16777215 16777215 3883761531 3971463351 4291024100 4286296026 3496384451 1987685334 2742594518 3445920446 4286361562 4291811035 4272203427 4120418711 4068179065 426141286 16777215 16777215 781949846 4152328572 3819086754 4292995310 4288790749 4269977817 4286296026 4288856541 4292666346 4252860027 4291151300 4288125334 4253518213 3917052791 813727872 16777215 16777215 781949846 3833430139 3700527759 3802177952 4073573837 4293256677 3634076827 3782703477 1668247407 4286611838 4291085506 4288322970 4258386385 4286546045 16777215 16777215 16777215 16777215 1637391254 3063520407 4118774140 4286546300 3063520407 1637391254 16777215 2256436862 4286546045 4292203988 4292796126 4286546045 16777215 16777215 16777215 16777215 33554432 67108864 83886080 100663296 100663296 117440512 100663296 100663296 1685419381 4269834365 4286546045 1451262077 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 23:00'!
smallSelectIcon
	"Private - Generated method"
	^ Icons
			at: #'smallSelect'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallSelectIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 23:00'!
smallSelectIconContents
	"Private - Method generated with the content of the file /home/dgd/smallSelect.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAABK0lEQVR4nKXOO0oEQRAG4L+6q2cfjojgBmJi
YmroEQwNBBPxFF7AzBt4DMHcyFwUE2MRPIDruuxMd9VvsKCbrOJOQVNNPT5KcHK7gTjdA2UH
MQzgnoCQIJ6AEAE3QFqQBSFk0DMcBaot1uxOobPjm/ODy/WhxrqvWvei1r35v1+FUIzM5mwz
PZt7a2Qx98eX9/Hp1cOFJpV0uD8aYUmkKJJixLBCXKxv1qmCSz/0VKtly79FEJlnjdBVgG+o
CljpgoVLJP499guQVFInIDtKJ6DJ3nQCPps86wTkgoYAVwYAycXYARAW8/8DKQZBgCgczdPr
x3h3azBwEtmc5qQZWZx0gsXJUpzFOe85+fw2mQCIgrPrbVg6AtFAQgHpAOznyTzTHRCDLNRt
ev8FfDedurn+q90AAAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 10/7/2004 13:42'!
smallSqueakIcon
	"Private - Generated method"
	^ Icons
			at: #'smallSqueak'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallSqueakIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/5/2004 13:49'!
smallSqueakIconContents
	"Private - Method generated with the content of the file /home/dgd/smallSqueak.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAACWklEQVQ4jcWSW0hTARzGv52zeXY25ub0eEFd
BBsztbYcC8PWhZEsctm6jFlEWkQRBT4U0ioiirSHMAgzijDqLeqlop60jGhGiEoaydKtXXWX
pmdubpN5eukhTKM3v8f/x//38v2A1Q75j05UpVNfy5NLbdFwzANgCgCZV5h3qLiMaZ2JsjSA
sZUA5Jb62uFae3Mju1FZM+cLH6eynENdvfbe4dPWMw1NJj3KGWPSG6KXBejqNHeNF47uejo1
zWP7R5AqZ0hJcr7x7PlmlUgioi6fuA6/qpQWpRfy+csBKtUbjoiHEqj7HoCsQIVMGYEhZ0Cq
3WNAgUCAm/YupL66QcokxQQAHgkcLAH/EgAZAFRVa4TvevvQcqAJ2dgcnHMp7LfVo292Ft1+
PwiSACWikQzFUoQExK1OMHceg7m6GULHTsv2jkg0QnBiGi6XC8FgEJkgiwabCYowi8EbT+Af
/wG5VsnFIzND/CoI1+8DU7QIHixEVilpaWwrUinw6dE0KIqCx+MBXVOCjJdEca4YW5UK8E9a
4B33JOIT/nYiBzn5I+AjCAoDQoLU1WkgpQRob7XhQ+Ab1pmNsO7ehp+RBE7Z7CHn6EQ0PDYZ
i/YPdwIY4FMQHOuAp4uleSrVOaucTWfIZDyB9HwabRetWOA4OD5OYtTnRTKZvvK858XD3/6k
AID3pziMVvlerKuoUGyqFNDy3Jzc0kL0vv3M7fAtcu5w8OXgs9d7ly7GW3owm82v9Hq9UV6Q
H7/dc/+Ly+3uzoZn3wBIrCDd3xYaDAa7yWR6AGDN/z6tXn4B2ljYZPClNw8AAAAASUVORK5C
YII='! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/31/2010 17:09'!
smallTimerIcon

	^ Icons
		at: #smallTimerIcon
		ifAbsentPut: [ self smallTimerIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/31/2010 17:09'!
smallTimerIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 4288526525 4116212343 4116212343 4288526525 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4284515726 4284515726 4284515726 4284515726 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4284515726 4284515726 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 2639020888 4116212343 4284449675 4284515726 4284909712 3377418592 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1414084431 4149767289 4288526525 4291550689 4292208362 4292274155 4291419360 4288525497 4284515981 1414084431 16777215 16777215 16777215 16777215 16777215 1414084431 4267474568 4291090909 4291286488 4291875796 4284534498 4284534498 4293322728 4291877082 4291090909 4284450703 1414084431 16777215 16777215 16777215 16777215 4099434870 4291090909 4290693837 4292798956 4284534498 4284534498 4284534498 4294046451 4293849072 4290693837 4291090909 4099434870 16777215 16777215 16777215 2639020888 4288526525 4291089366 4292730334 4278190080 4284534498 4284534498 4284534498 4294178037 4293914865 4292730334 4291089366 4288526525 2639020888 16777215 16777215 3931463791 4291550689 4290561222 4294046451 4293522164 4278190080 4285054413 4284534498 4294178037 4293980401 4294046451 4290561222 4291550689 4285369491 75530240 16777215 4284450446 4292208362 4291415245 4293980401 4294243573 4294310137 4278190080 4280778896 4293190885 4293914865 4293914608 4294440951 4292076519 4284582032 16777215 16777215 3931463791 4291550689 4290363586 4294243573 4293717229 4293783279 4284966759 4278190080 4293651693 4293454314 4294506745 4294046194 4291550689 4284842635 16777215 16777215 2639020888 4288526525 4290497486 4293651436 4294046193 4293717229 4294506745 4293585900 4293454057 4294111987 4294506744 4292272611 4288526525 4066077045 16777215 16777215 16777215 4116146551 4291090909 4289838785 4293980401 4294309365 4293717229 4293191142 4293783279 4294309366 4293323499 4291090909 4284712333 890442520 16777215 16777215 16777215 1731932739 4284251783 4291090909 4290562765 4291678674 4294309366 4294177780 4294375159 4292600805 4291090909 4284581518 3493280325 1325400064 67108864 16777215 16777215 385875968 2385323570 4216680057 4288526525 4291550689 4292208362 4292274155 4291550689 4288526525 4284514697 3593219383 2147483648 1090519040 16777215 16777215 16777215 16777215 520093696 1509949440 3493476934 4233325429 4284383883 4284450703 4284185988 4064366169 2650800128 2097152000 939524096 33554432 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/21/2010 23:21'!
smallTrafficIcon

	^ Icons
		at: #smallTrafficIcon
		ifAbsentPut: [ self smallTrafficIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/21/2010 23:21'!
smallTrafficIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 422804152 4281694648 4281694648 4281694648 4281694648 4281694648 4281694648 422804152 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4289448421 4292255918 4293546071 4293546071 4292255918 4289448421 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4291288047 4293546071 4294836738 4294836738 4293546071 4291288047 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4291288047 4293546071 4294836738 4294836738 4293546071 4291288047 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4291288047 4292255918 4293546071 4293546071 4292255918 4291288047 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4291288047 4292265390 4293569111 4293569111 4292265390 4291288047 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4291288047 4293569111 4294872578 4294872578 4293569111 4291288047 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4291288047 4293569111 4294872578 4294872578 4293569111 4291288047 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4291288047 4292265390 4293569111 4293569111 4292265390 4291288047 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4291288047 4287747246 4282960727 4282960727 4287747246 4291288047 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4291288047 4282960727 4278370818 4278370818 4282960727 4291288047 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4291288047 4282960727 4278370818 4278370818 4282960727 4291288047 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4281694648 4289448421 4287747246 4282960727 4282960727 4287747246 4289448421 4281694648 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 808811194 4281694648 4281694648 4281694648 4281694648 4281694648 4281694648 808811194 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 21:20'!
smallUndoIcon

	^ Icons
		at: #smallUndoIcon
		ifAbsentPut: [ self smallUndoIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 21:20'!
smallUndoIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 16777215 16777215 16777215 16777215 16777215 16777215 4292129024 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 348959488 4292129024 4292129024 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 295077632 4292129024 4294834103 4292129024 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 295077632 4292129024 4294834103 4294962176 4292129024 4292129024 3889082880 3486232576 2764878080 193694464 16777215 16777215 16777215 16777215 16777215 295077632 4292129024 4294834103 4294633738 4294896384 4294832763 4294832503 4294635126 4177194867 3570184704 3587224576 479367680 16777215 16777215 16777215 177838336 4292129024 4294834103 4294633472 4294699535 4294370304 4294633472 4294370304 4293646336 4294107683 4294635120 4276869437 3838882816 848269056 16777215 16777215 4292129024 4294834103 4294633472 4294765071 4294699535 4294699535 4294633472 4294370304 4293909760 4293909760 4293909760 4294107683 4293515323 3402478080 177838336 16777215 295077632 4292129024 4294832770 4294765071 4294567680 4292988160 4292988160 4292988160 4293909760 4293909760 4293909760 4293448704 4294438002 4174625536 3419123968 16777215 16777215 295077632 4292129024 4294832766 4293185536 4294832766 4294766704 4294766704 4294766704 4294766704 4294764032 4293448704 4293054208 4294042744 4107187200 16777215 16777215 16777215 295077632 4292129024 4294832766 4294832766 4292129024 4292129024 4292129024 4292129024 4292000000 4294175095 4293448960 4294372208 4174361600 16777215 16777215 16777215 16777215 295077632 4292129024 4294832766 4292129024 16777215 16777215 819310336 1355393280 3386159616 4294569333 4294240640 4291867648 16777215 16777215 16777215 16777215 16777215 16777215 4292129024 4292129024 16777215 16777215 16777215 16777215 449753344 4292262912 4277595539 4291932928 16777215 16777215 16777215 16777215 16777215 16777215 16777215 4292129024 16777215 16777215 16777215 16777215 16777215 4275681792 4277134179 4023367168 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 584233984 4242919490 4276277570 2144319232 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1574154752 3002197102 2933044736 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 16777215 1775351296 1775482368 416658688 16777215 16777215)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 23:00'!
smallUpdateIcon
	"Private - Generated method"
	^ Icons
			at: #'smallUpdate'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallUpdateIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 23:00'!
smallUpdateIconContents
	"Private - Method generated with the content of the file /home/dgd/smallUpdate.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAACz0lEQVR4nI3SW2hTBwDG8f85iSdpEnIxR+01
0TlmbawOWy8PdXZYFRXLVHwYLt5ARMUXH0Rh0DxIq1jQl3irw6KCSMecKDiYVZEZ6+jEFsrs
apektsY2tWnNSU/S0+T45IMgxO/5+35Pn8CXRyqysRmROoOIVReJSDl+MeZbOcx85XaYL2/Z
tnV1bd16saKyGovVgZJ8z/ZNNQvzAYt27tja/dOe3UJfLEXwfhjloQqiAX+1i5JS75J8gDh/
/jzhbqifJxNejHIFVWUzmcxo9IR7sNrs5nzAlDqp6mFVEYQCAZ9nJovnyuR0nbe9r3n+Ovom
HzCazkzp5aUOwW1zM8tupvvvByhDL1lTs5Shwcg1sbzUsq92xcK++rqVI99VL+gvdnLLbWHZ
R6C9/VFkTbWXRF+IcPtF0rF/WVtTyc0r5yfC72gRdmypGzjbcrPsRU8vnuLZSKLGrbartN24
8bb/VeRoMsd/P9av+8vhchlNZhMDkQGms9ncLM/immAw+FS0O53uI40XOPnnMAdau/Cf/oNh
oZizF84VHjy0r1VVGej8p6szq88gqYgsqKhiND7aVVRUdALAKCLqU7qRVb5isjmdjl4DnUkd
57OX2O02Ach0vxpe2xe97isw4RlTuAcotRtzjwHEWGww7il0kZvWMBkNfF3oACClTDIxkcwC
aUBRNZ6NKbQBSiAQcEqSpAIYX3SGWv0HjwXu3G7BOnseGYsXr7uQ5XPLaAn+PgJMAjQ1NS0T
BEFOp9OqLMtHotHoJQABkDasXvLcv3evr6M7yrSWwfNNJb45Art37W86fDTwsyzLv42Pj49o
mpaSJMmUSCTuNTc33/kIUAUW/6njqY5QCKvNSmwwRkpNx1d+X7+npKTkQCKRaG9oaDjzuaMY
AGKgxYf+r9hU/0O5JBWIY2PvctGheOjbquWueDz+a2NjY2uew30SJ2D70vIHGPAd/9jj14AA
AAAASUVORK5CYII='! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'mt 5/10/2015 13:36'!
smallUserQuestionIcon

	^ self fugueUserSilhouetteQuestionIcon! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 23:00'!
smallVolumeIcon
	"Private - Generated method"
	^ Icons
			at: #'smallVolume'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self smallVolumeIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 23:00'!
smallVolumeIconContents
	"Private - Method generated with the content of the file /home/dgd/smallVolume.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAQAAAC1+jfqAAAAAmJLR0QA/4ePzL8AAAE3SURB
VHjajdE/SJRxHMfxF/f8Hp/n7vLhlAKhyKRCOpccA6EbFEW0BsciyKWphiBoCRuiQXEJnIyG
ampyEppydy8wBOFEyPzHA0X1wNGggV0Ivbbv8h3eH44rKWkT/XUNqdhxgrjz5kBTA90SiVlD
h0+PdI6PvZhKwDmPVL13G4KaOqcuTL+6k3z8iis+Sd0175pxy0H91ss066g+SKo+Q68uq4Zd
9M5DyyUdE6dHzg7XIrmdM7XHPriq8NoNaxJdgRypA7/Ezo8e9NjDpgzfhMC+kp92FXZb++u+
KKPbd5S1Ak2xVJ+WH3nzqQE5Jq3IpPJgY2GOrDe+NxhCYUvZkkGXvHHfkuJPhkr/4pNiZlsD
dbPq+ryVHS9Zqc1c39JAJhF5Zqw9dsXzw7yIXBb/u0fUNt7/+A0ZzEdfNHX2/wAAAABJRU5E
rkJggg=='! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'MAD 12/19/2010 21:20'!
smallWindowIcon

	^ Icons
		at: #smallWindowIcon
		ifAbsentPut: [ self smallWindowIconContents ]! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'MAD 12/19/2010 21:20'!
smallWindowIconContents
^ (Form
	extent: 16@16
	depth: 32
	fromArray: #( 1556466370 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 1556466370 16777215 16777215 16777215 16777215 4291217859 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4291217859 16777215 16777215 16777215 16777215 4291217859 4294901502 4279720860 4279720860 4279720860 4279720860 4279720860 4279720860 4279720860 4279720860 4294835709 4291217859 16777215 16777215 16777215 16777215 4291217859 4294835708 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4294769916 4291217859 16777215 16777215 16777215 16777215 4291217859 4294704379 4294638330 4294638330 4293388774 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 1556466370 4291217859 4294572537 4294506744 4294441208 4291217859 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4294967295 4291217859 4291217859 4294572537 4294374901 4294309365 4291217859 4294901502 4279720860 4279720860 4279720860 4279720860 4279720860 4279720860 4279720860 4279720860 4294835709 4291217859 4291217859 4294309365 4294046193 4294111986 4291217859 4294835708 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4281563063 4294769916 4291217859 4291217859 4294111987 4293914350 4293848814 4291217859 4294704379 4294638330 4294638330 4294638330 4294572793 4294572537 4294572537 4294572537 4294572537 4294572793 4291217859 4291217859 4294046449 4293585643 4293585642 4291217859 4294572537 4294506744 4294441208 4294506488 4294506744 4294506487 4294440951 4294506487 4294506487 4294572537 4291217859 4291217859 4294374901 4294112242 4294177522 4291217859 4294572537 4294374901 4294309365 4294309365 4294309365 4294243829 4294309365 4294309365 4294374901 4294506744 4291217859 1589888702 4291217859 4291217859 4291217859 4291217859 4294309365 4294046193 4294111986 4294111986 4294111986 4294111986 4294111986 4294111986 4294046193 4294243828 4291217859 16777215 16777215 16777215 16777215 4291217859 4294111987 4293914350 4293848814 4293848814 4293848814 4293783021 4293848814 4293848557 4293848558 4294111986 4291217859 16777215 16777215 16777215 16777215 4291217859 4294046449 4293585643 4293585642 4293585898 4293585642 4293585642 4293585642 4293585642 4293585642 4294046193 4291217859 16777215 16777215 16777215 16777215 4291217859 4294374901 4294112242 4294177522 4294111986 4294112243 4294111986 4294111986 4294177522 4294111986 4294111986 4291217859 16777215 16777215 16777215 16777215 1589888702 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 4291217859 1556466370)
	offset: 0@0)! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
squeakIcon
	"Private - Generated method"
	^ Icons
			at: #'squeak'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self squeakIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 11/5/2004 13:49'!
squeakIconContents
	"Private - Method generated with the content of the file /home/dgd/squeak.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0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'! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'mt 7/10/2016 12:54'!
squeakLogoIcon
^ Icons
		at: #squeakLogoIcon ifAbsentPut: [(Form
	extent: 24@16
	depth: 32
	fromArray: #( 0 0 67108865 2063597569 1778384897 889192449 0 0 0 0 0 0 0 0 0 0 0 0 1610612737 1811939329 1610612737 0 0 0 0 0 905969665 889192449 0 1442840577 788529153 0 0 0 0 0 0 0 0 0 0 1711276033 469762049 0 1728053249 0 0 0 0 0 1006632961 704643073 0 0 1895825409 234881025 0 0 0 0 0 0 0 0 1207959553 838860801 0 0 1728053249 0 0 0 0 0 738197505 989855745 0 0 167772161 1862270977 0 0 0 0 0 0 0 83886081 1879048193 0 0 0 1744830465 0 0 0 0 0 469762049 1258291201 0 0 0 1543503873 520093697 0 0 0 0 0 0 1543503873 335544321 0 0 0 1761607681 0 0 0 0 0 16777217 1644167169 0 0 0 33554433 67108865 0 0 0 0 0 0 33554433 0 0 0 0 1795162113 0 0 0 0 0 0 1711276033 0 0 0 0 0 117440513 0 0 0 0 0 0 0 0 0 100663297 1677721601 0 0 0 0 0 0 1744830465 0 0 0 0 3019898881 4009754625 1056964609 0 0 1090519041 3204448257 1325400065 0 0 0 889192449 889192449 0 0 0 0 0 0 1392508929 352321537 0 0 637534209 4278190081 4278190081 2650800129 0 0 3372220417 4278190081 3758096385 0 0 0 1560281089 117440513 0 0 0 0 0 0 0 0 0 0 134217729 4060086273 4278190081 1929379841 0 0 3187671041 4278190081 3590324225 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 704643073 1577058305 83886081 0 0 771751937 2667577345 956301313 0 0 0 0 0 0 0 0 0 0 369098753 973078529 1577058305 1728053249 1711276033 1325400065 805306369 218103809 0 855638017 922746881 16777217 0 268435457 788529153 1308622849 1593835521 1056964609 436207617 16777217 0 0 1442840577 1761607681 1409286145 788529153 301989889 872415233 1124073473 1627389953 1929379841 973078529 2030043137 4278190081 4278190081 2164260865 1040187393 1946157057 1577058305 939524097 452984833 687865857 1325400065 1778384897 1795162113 1392508929 0 0 855638017 1795162113 1677721601 855638017 620756993 436207617 486539265 285212673 2030043137 4278190081 4278190081 2164260865 486539265 1023410177 922746881 1157627905 1392508929 1845493761 1744830465 486539265 0 0 201326593 2013265921 1006632961 0 553648129 1694498817 1778384897 1694498817 1728053249 738197505 0 855638017 922746881 16777217 738197505 1728053249 1744830465 1778384897 1476395009 251658241 117440513 1476395009 1879048193 150994945 0 16777217 0 486539265 1426063361 16777217 0 0 0 0 0 0 0 0 0 0 0 0 234881025 1728053249 486539265 0 167772161 50331649)
	offset: 0@0)]! !

!MenuIcons class methodsFor: 'accessing - icons' stamp: 'mt 7/10/2016 12:57'!
squeakLogoIconColorized: aColor

	^ self squeakLogoIcon
		collectColors: [:c | aColor alpha: c alpha]! !

!MenuIcons class methodsFor: 'class initialization' stamp: 'eem 1/5/2016 17:33'!
startUp: resuming 
	resuming ifFalse: [^self].
	TranslationLocale = LocaleID current ifTrue:
		[^self].
	self initializeTranslations! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
volumeIcon
	"Private - Generated method"
	^ Icons
			at: #'volume'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self volumeIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
volumeIconContents
	"Private - Method generated with the content of the file /home/dgd/volume.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAQAAABKfvVzAAAABGdBTUEAALGOfPtRkwAAACBj
SFJNAAB6JQAAgIMAAPn/AACA6QAAdTAAAOpgAAA6mAAAF2+SX8VGAAAAAmJLR0QA/4ePzL8A
AAAJcEhZcwAACxMAAAsTAQCanBgAAAHrSURBVDjLldTLahRBFMbx/xu02hEvm1mZIC4qovaA
LhoE142+QIMvILoRgjASdKEilSBxO5CFIhKbmUAQNWmD6KCQxCAMEi+JGGKISuETfC665uIM
DqRqd6hfn9Nd5zRidxsBV/jOH5rc5QhjJBDwnyUQ43xmix9ssMo8GRiygaBBg0dMM887FsCE
jvyfUwEZaRdY5CrDnOUyt8GE7pbaIKEEgEGYNnjpowmYIXdfeQfErJMAUCliArFMAJwEM+Sm
9V5r6iopIC+eXdDiK1lycux+90Qb2tavAlQxFxgBwwoAl8haIH+sTKN6oZ/6rW1ttjJYzEMO
QZUEKPnTgnxHmzqjHW3pm77ooxApEFC9QwoxVV+U8eCTmirrq9bU1KqWVNZBEewDe55JwGfM
iT1Y0hud0gct661ea1HndEzEQOUiUxzoBwt6quNq6JVyPdOcTuuwCE6AtUywpx/Mqq5RPdec
ZlVTTTjSBxBgp7gGCRYA136HG7qpYU2qrppqqou8VtxAcI8RyDBACddzD6GzbTAOFYIxjkLs
C6pQbYHWMqGbUF11kcNeH2OdEhDgiHsBmNBZzXRaI2XFN4YtWr4LREREYEJ3vQNKfpRSXNGh
PaBMBIb+ebA+Tz/wZPDE0ZMjgiAaBHb919jd/gvLk2YJrn7DBgAAAABJRU5ErkJggg=='! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
windowIcon
	"Private - Generated method"
	^ Icons
			at: #'window'
			ifAbsentPut:[ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self windowIconContents readStream) ].! !

!MenuIcons class methodsFor: 'private - icons' stamp: 'dgd 9/25/2004 22:59'!
windowIconContents
	"Private - Method generated with the content of the file /home/dgd/window.png"
	^ 'iVBORw0KGgoAAAANSUhEUgAAABwAAAAcCAYAAAByDd+UAAAABHNCSVQICAgIfAhkiAAAABF0
RVh0U29mdHdhcmUAU29kaXBvZGmU4xfQAAAB5UlEQVR4nO2WP2gTURzHP/fuEqMpZ1vUKBqq
FHFTB/+0XVSKRSxCBUVHQXHoUHeh4thFVwcnFwWdNFBErWgVVBIXBRGs1pZEjamtucvl7nKX
O4fkoPXPkHqCw32m977vz4f3ePB7EBEyUtDoGzrbLRDpMDeX63XjceZqDvCDTAHoP3HuwsHd
uy5uVNukuXQXruSR8ORlSaqyS6IuYwqX79MFfKG8mLx9pWeJcNv2HSOnx69Lq18+Z+zRJHmj
xNEt+5Al0ZJs3tZ4kM/SrW7ivZant7+Xzx9n9n79NHDkbfbeXcBRAN69eq3cHzzJ+sPHWGsn
SMopSvlCy6fz8NmqpOiwYgglxdybaWanpjArCzuBBeCpAmCZVZ4Uiwzs30NXpdRcXm5ZCLAO
AJ2U6/Iwl8N1nGCoA5pXGospjI0Oo7YllyX5E4cO9HD8zHlmF2UCYE1nuwhbBhCPx9ic3rAk
a+1VhEAkjISR8D8RGlXT9zzvnwjKuvGrsKxXStdujmPZtdBEtZrDrTsTfJgpUDMqepBLAH2D
p4aFiF+WJGlFaMYGvmVoz7ITNy4BDpAJvhirVibVoaTaqYZps23L1ue/BHXOATJKs2OahvbN
NDQrTOFPFGHRJwqQaRRJ6bfT/w6XRsWPCJ8feQeiSATBXJcAAAAASUVORK5CYII='! !

Object subclass: #MenuUpdater
	instanceVariableNames: 'updater updateSelector arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Menus'!

!MenuUpdater methodsFor: 'as yet unclassified' stamp: 'kb 2/22/2010 14:54'!
updater: anObject updateSelector: aSelector

	self updater: anObject updateSelector: aSelector arguments: nil! !

!MenuUpdater methodsFor: 'as yet unclassified' stamp: 'kb 2/22/2010 14:53'!
updater: anObject updateSelector: aSelector arguments: anArray

	updater := anObject.
	updateSelector := aSelector.
	arguments := anArray! !

!MenuUpdater methodsFor: 'as yet unclassified' stamp: 'kb 2/22/2010 14:58'!
update: aMenuMorph
	"Reconstitute the menu by first removing the contents and then building it afresh"

	aMenuMorph removeAllMorphs.
	arguments 
		ifNil: [ updater perform: updateSelector with: aMenuMorph ]
		ifNotNil: [ 
			updater 
				perform: updateSelector 
				withArguments: (arguments copyWith: aMenuMorph) ].
	aMenuMorph changed! !

Object subclass: #Morph
	instanceVariableNames: 'bounds owner submorphs fullBounds color extension'
	classVariableNames: 'IndicateKeyboardFocus PreferredCornerRadius UseSoftDropShadow'
	poolDictionaries: ''
	category: 'Morphic-Kernel'!

!Morph commentStamp: 'efc 2/26/2003 20:01' prior: 0!
A Morph (from the Greek "shape" or "form") is an interactive graphical object. General information on the Morphic system can be found at http://minnow.cc.gatech.edu/squeak/30. 

Morphs exist in a tree, rooted at a World (generally a PasteUpMorph). The morphs owned by a morph are its submorphs. Morphs are drawn recursively; if a Morph has no owner it never gets drawn. To hide a Morph and its submorphs, set its #visible property to false using the #visible: method. 

The World (screen) coordinate system is used for most coordinates, but can be changed if there is a TransformMorph somewhere in the owner chain. 

My instance variables have accessor methods (e.g., #bounds, #bounds:). Most users should use the accessor methods instead of using the instance variables directly.

Structure:
instance var 	Type 			Description 
bounds 			Rectangle 		A Rectangle indicating my position and a size that will enclose 									me. 
owner 			Morph		 	My parent Morph, or nil for the top-level Morph, which is a
 				or nil			world, typically a PasteUpMorph.
submorphs 		Array 			My child Morphs. 
fullBounds 		Rectangle 		A Rectangle minimally enclosing me and my submorphs. 
color 			Color 			My primary color. Subclasses can use this in different ways. 
extension 		MorphExtension Allows extra properties to be stored without adding a
				or nil  				storage burden to all morphs. 

By default, Morphs do not position their submorphs. Morphs may position their submorphs directly or use a LayoutPolicy to automatically control their submorph positioning.

Although Morph has some support for BorderStyle, most users should use BorderedMorph if they want borders.!

Morph subclass: #AbstractResizerMorph
	instanceVariableNames: 'dotColor handleColor lastMouse'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!AbstractResizerMorph commentStamp: 'jmv 1/29/2006 17:15' prior: 0!
I am the superclass of a hierarchy of morph specialized in allowing the user to resize or rearrange windows and panes.!

!AbstractResizerMorph class methodsFor: 'as yet unclassified' stamp: 'mt 4/24/2015 11:29'!
gripThickness
	"A number in pixels that encodes the area were the user can target splitters or edge grips."

	^ 4! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'jrp 7/5/2005 21:36'!
dotColor

	^ dotColor ifNil: [self setDefaultColors. dotColor]! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'jrp 7/5/2005 21:35'!
handleColor

	^ handleColor ifNil: [self setDefaultColors. handleColor]! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'jrp 7/5/2005 21:37'!
handlesMouseDown: anEvent

	^ true! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'jrp 7/5/2005 21:37'!
handlesMouseOver: anEvent

	^ true
	! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'md 2/24/2006 23:01'!
initialize

	super initialize.
	self color: Color transparent! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'jrp 7/5/2005 21:40'!
isCursorOverHandle

	^ true! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'jrp 7/5/2005 21:42'!
mouseDown: anEvent

	lastMouse := anEvent cursorPoint! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'md 2/24/2006 23:01'!
mouseEnter: anEvent

	self isCursorOverHandle ifTrue:
		[self setInverseColors.
		self changed.
		anEvent hand showTemporaryCursor: self resizeCursor]! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'jrp 7/5/2005 21:37'!
mouseLeave: anEvent

	anEvent hand showTemporaryCursor: nil.
	self setDefaultColors.
	self changed! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'mt 4/14/2015 08:39'!
referencePoint: aPoint

	lastMouse := aPoint.! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'jrp 7/5/2005 21:37'!
resizeCursor

	self subclassResponsibility! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'jrp 7/29/2005 13:25'!
setDefaultColors

	handleColor := Color lightGray lighter lighter.
	dotColor := Color gray lighter! !

!AbstractResizerMorph methodsFor: 'as yet unclassified' stamp: 'jrp 7/30/2005 21:30'!
setInverseColors

	handleColor := Color lightGray.
	dotColor := Color white! !

AbstractResizerMorph subclass: #CornerGripMorph
	instanceVariableNames: 'target'
	classVariableNames: 'ActiveForm DrawCornerResizeHandles PassiveForm'
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!CornerGripMorph commentStamp: 'jmv 1/29/2006 17:15' prior: 0!
I am the superclass of a hierarchy of morph specialized in allowing the user to resize windows.!

CornerGripMorph subclass: #BorderGripMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!BorderGripMorph methodsFor: 'as yet unclassified' stamp: 'spfa 4/20/2010 13:33'!
drawOn: aCanvas

	"aCanvas fillRectangle: self bounds color: Color red" "for debugging"
! !

!BorderGripMorph methodsFor: 'as yet unclassified' stamp: 'spfa 4/19/2010 22:22'!
setDefaultColors! !

!BorderGripMorph methodsFor: 'as yet unclassified' stamp: 'spfa 4/19/2010 22:21'!
setInverseColors! !

BorderGripMorph subclass: #BottomGripMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!BottomGripMorph methodsFor: 'target resize' stamp: 'ael 6/3/2011 00:39'!
apply: delta 
	| oldBounds |
	oldBounds := target bounds.
	target
		bounds: (oldBounds origin corner: oldBounds corner + (0 @ delta y))! !

!BottomGripMorph methodsFor: 'initialize' stamp: 'mt 4/24/2015 11:31'!
defaultHeight

	^ self class gripThickness! !

!BottomGripMorph methodsFor: 'accessing' stamp: 'mt 4/28/2015 14:28'!
gripLayoutFrame
	^ LayoutFrame
		fractions: (0 @ 1 corner: 1 @ 1)
		offsets: (0 @ 0 negated corner: 0@ self defaultHeight)! !

!BottomGripMorph methodsFor: 'initialize' stamp: 'ar 6/23/2010 17:01'!
initialize
	super initialize.
	self hResizing: #spaceFill.! !

!BottomGripMorph methodsFor: 'accessing' stamp: 'spfa 8/15/2009 11:37'!
ptName
	^#bottom! !

!BottomGripMorph methodsFor: 'accessing' stamp: 'spfa 8/15/2009 11:37'!
resizeCursor

	^ Cursor resizeForEdge: #top! !

BorderGripMorph subclass: #LeftGripMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!LeftGripMorph methodsFor: 'target resize' stamp: 'ael 6/3/2011 00:38'!
apply: delta 
	| oldBounds |
	oldBounds := target bounds.
	target
		bounds: (oldBounds origin + (delta x @ 0) corner: oldBounds corner)! !

!LeftGripMorph methodsFor: 'initialize' stamp: 'mt 4/24/2015 11:31'!
defaultWidth

	^ self class gripThickness! !

!LeftGripMorph methodsFor: 'accessing' stamp: 'mt 4/28/2015 14:26'!
gripLayoutFrame
	^ LayoutFrame
		fractions: (0 @ 0 corner: 0 @ 1)
		offsets: (self defaultWidth negated @ 0 corner: 0 @ 0)! !

!LeftGripMorph methodsFor: 'initialize' stamp: 'ar 6/23/2010 17:01'!
initialize
	super initialize.
	self vResizing: #spaceFill.! !

!LeftGripMorph methodsFor: 'accessing' stamp: 'spfa 8/15/2009 11:41'!
ptName
	^#left! !

!LeftGripMorph methodsFor: 'accessing' stamp: 'spfa 8/15/2009 11:41'!
resizeCursor

	^ Cursor resizeForEdge: #left! !

BorderGripMorph subclass: #RightGripMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!RightGripMorph methodsFor: 'target resize' stamp: 'ael 6/3/2011 00:37'!
apply: delta 
	| oldBounds |
	oldBounds := target bounds.
	target
		bounds: (oldBounds origin corner: oldBounds corner + (delta x @ 0))! !

!RightGripMorph methodsFor: 'initialize' stamp: 'mt 4/24/2015 11:31'!
defaultWidth

	^ self class gripThickness! !

!RightGripMorph methodsFor: 'accessing' stamp: 'mt 4/28/2015 14:26'!
gripLayoutFrame
	^ LayoutFrame
		fractions: (1 @ 0 corner: 1 @ 1)
		offsets: (0 @ self defaultHeight negated corner: self defaultWidth @ 0)! !

!RightGripMorph methodsFor: 'initialize' stamp: 'ar 6/23/2010 17:01'!
initialize
	super initialize.
	self vResizing: #spaceFill.! !

!RightGripMorph methodsFor: 'accessing' stamp: 'spfa 8/15/2009 11:40'!
ptName
	^#right! !

!RightGripMorph methodsFor: 'accessing' stamp: 'spfa 8/15/2009 11:40'!
resizeCursor

	^ Cursor resizeForEdge: #right! !

BorderGripMorph subclass: #TopGripMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!TopGripMorph methodsFor: 'target resize' stamp: 'ael 6/3/2011 00:38'!
apply: delta 
	| oldBounds |
	oldBounds := target bounds.
	target
		bounds: (oldBounds origin + (0 @ delta y) corner: oldBounds corner)! !

!TopGripMorph methodsFor: 'initialize' stamp: 'mt 4/24/2015 11:31'!
defaultHeight

	^ self class gripThickness! !

!TopGripMorph methodsFor: 'accessing' stamp: 'mt 4/28/2015 14:27'!
gripLayoutFrame
	^ LayoutFrame
		fractions: (0 @ 0 corner: 1 @ 0)
		offsets: (0 @ (-40 - self defaultHeight)  corner: 0@ 0)! !

!TopGripMorph methodsFor: 'initialize' stamp: 'ar 6/23/2010 17:01'!
initialize
	super initialize.
	self hResizing: #spaceFill.! !

!TopGripMorph methodsFor: 'accessing' stamp: 'spfa 8/15/2009 11:23'!
ptName
	^#top! !

!TopGripMorph methodsFor: 'accessing' stamp: 'spfa 8/15/2009 11:23'!
resizeCursor

	^ Cursor resizeForEdge: #top! !

CornerGripMorph subclass: #BottomLeftGripMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!BottomLeftGripMorph commentStamp: 'jmv 1/29/2006 17:17' prior: 0!
I am the handle in the left bottom of windows used for resizing them.!

!BottomLeftGripMorph methodsFor: 'target resize' stamp: 'jmv 1/29/2006 18:06'!
apply: delta 
	| oldBounds |
	oldBounds := target bounds.
	target
		bounds: (oldBounds origin + (delta x @ 0) corner: oldBounds corner + (0 @ delta y))! !

!BottomLeftGripMorph methodsFor: 'private' stamp: 'hsj 12/10/2009 16:44'!
borderOffset
	|width|
	width :=SystemWindow borderWidth +1.
	 ^self handleOrigin + (width @ width negated)! !

!BottomLeftGripMorph methodsFor: 'accessing' stamp: 'mt 4/28/2015 14:41'!
gripLayoutFrame
	^ LayoutFrame
		fractions: (0 @ 1 corner: 0 @ 1)
		offsets: (self class gripThickness negated @ (self defaultHeight negated + self class gripThickness)
			corner: 0 @ 0)! !

!BottomLeftGripMorph methodsFor: 'private' stamp: 'hsj 12/10/2009 16:04'!
handleOrigin
^25@0! !

!BottomLeftGripMorph methodsFor: 'accessing' stamp: 'md 2/24/2006 22:43'!
ptName
	^#bottomLeft! !

!BottomLeftGripMorph methodsFor: 'accessing' stamp: 'jmv 1/29/2006 17:52'!
resizeCursor

	^ Cursor resizeForEdge: #bottomLeft! !

CornerGripMorph subclass: #BottomRightGripMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!BottomRightGripMorph commentStamp: 'jmv 1/29/2006 17:18' prior: 0!
I am the handle in the right bottom of windows used for resizing them.!

!BottomRightGripMorph methodsFor: 'target resize' stamp: 'jmv 1/29/2006 17:59'!
apply: delta 
	| oldBounds |
	oldBounds := target bounds.
	target
		bounds: (oldBounds origin corner: oldBounds corner + delta)! !

!BottomRightGripMorph methodsFor: 'private' stamp: 'hsj 12/10/2009 16:45'!
borderOffset
	|width|
	width :=SystemWindow borderWidth +1.
	 ^self handleOrigin - (width asPoint)! !

!BottomRightGripMorph methodsFor: 'accessing' stamp: 'mt 4/28/2015 14:43'!
gripLayoutFrame
	^ LayoutFrame
		fractions: (1 @ 1 corner: 1 @ 1)
		offsets: ((self defaultWidth negated + self class gripThickness) @ (self defaultHeight negated + self class gripThickness)
			corner: 0@0)! !

!BottomRightGripMorph methodsFor: 'private' stamp: 'hsj 12/10/2009 16:07'!
handleOrigin
	^0@0! !

!BottomRightGripMorph methodsFor: 'accessing' stamp: 'md 2/24/2006 22:43'!
ptName
	^#bottomRight! !

!BottomRightGripMorph methodsFor: 'accessing' stamp: 'jmv 1/29/2006 17:51'!
resizeCursor

	^ Cursor resizeForEdge: #bottomRight! !

!CornerGripMorph class methodsFor: 'handle settings' stamp: 'mt 7/2/2016 13:28'!
activeColor
	<preference: 'Corner Grip highlight color'
		category: 'colors'
		description: 'The highlight-color of window corners'
		type: #Color>
	^(self activeForm colorAt: 24@24) alpha:  1! !

!CornerGripMorph class methodsFor: 'handle settings' stamp: 'HenrikSperreJohansen 2/25/2010 01:41'!
activeColor: aColor
	|canvas|
	canvas := self initializeActiveForm getCanvas.
	canvas 
		privatePort fillPattern: aColor;
		combinationRule: Form rgbMul;
		fillRect: (self activeForm boundingBox) offset: 0@0.

		! !

!CornerGripMorph class methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/10/2009 21:06'!
activeForm
	^ActiveForm ifNil: [self initializeActiveForm]! !

!CornerGripMorph class methodsFor: 'class initialization' stamp: 'HenrikSperreJohansen 12/10/2009 21:03'!
defaultForm

	^(Form
		extent: 48@48
		depth: 32
		fromArray: 
			#( 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 50331647 100663295 134217727 167772159 184549375 201326591 218103807 218103807 201326591 184549375 167772159 134217727 100663295 50331647 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 50331647 117440511 167772159 218103807 301989887 352321535 385875967 436207615 469762047 469762047 469762047 469762047 452984831 402653183 369098751 318767103 251658239 167772159 117440511 50331647 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 50331647 117440511 201326591 301989887 385875967 452984831 536870911 587202559 637534207 654311423 687865855 687865855 704643071 687865855 671088639 637534207 603979775 553648127 486539263 402653183 335544319 218103807 117440511 50331647 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 33554431 100663295 201326591 301989887 402653183 520093695 637534207 687865855 754974719 838860799 889192447 922746879 956301311 973078527 973078527 956301311 939524095 905969663 855638015 788529151 704643071 654311423 570425343 436207615 352321535 218103807 100663295 33554431 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 67108863 167772159 268435455 385875967 520093695 654311423 721420287 838860799 922746879 989855743 1056964607 1107296255 1140850687 1174405119 1191182335 1191182335 1191182335 1157627903 1124073471 1073741823 1006632959 939524095 872415231 771751935 671088639 570425343 436207615 318767103 167772159 67108863 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 83886079 201326591 335544319 469762047 620756991 738197503 855638015 956301311 1040187391 1124073471 1224736767 1258291199 1325400063 1358954495 1375731711 1375731711 1375731711 1375731711 1358954495 1342177279 1291845631 1224736767 1157627903 1073741823 989855743 889192447 788529151 654311423 520093695 385875967 234881023 83886079 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 83886079 218103807 369098751 520093695 654311423 805306367 939524095 1056964607 1157627903 1241513983 1325400063 1409286143 1476395007 1509949439 1543503871 1577058303 1610612735 1610612735 1593835519 1560281087 1509949439 1493172223 1426063359 1342177279 1275068415 1191182335 1090519039 973078527 855638015 704643071 570425343 419430399 251658239 83886079 0 0 0 0 0 0 0 0 0 0 0 0 0 83886079 218103807 385875967 536870911 687865855 855638015 1006632959 1124073471 1224736767 1342177279 1442840575 1526726655 1593835519 1660944383 1711276031 1761607679 1778384895 1778384895 1778384895 1778384895 1778384895 1728053247 1677721599 1627389951 1543503871 1476395007 1375731711 1275068415 1157627903 1040187391 905969663 754974719 587202559 419430399 268435455 83886079 0 0 0 0 0 0 0 0 0 0 0 83886079 218103807 369098751 536870911 704643071 889192447 1040187391 1191182335 1308622847 1409286143 1543503871 1627389951 1694498815 1778384895 1845493759 1895825407 1946157055 1962934271 1979711487 1979711487 1979711487 1962934271 1912602623 1862270975 1795162111 1728053247 1660944383 1577058303 1459617791 1342177279 1224736767 1107296255 939524095 771751935 587202559 419430399 251658239 83886079 0 0 0 0 0 0 0 0 0 67108863 184549375 352321535 536870911 704643071 889192447 1056964607 1207959551 1342177279 1476395007 1610612735 1694498815 1811939327 1879048191 1962934271 2030043135 2080374783 2113929215 2147483647 2147483647 2147483647 2147483647 2130706431 2097151999 2046820351 1979711487 1912602623 1828716543 1744830463 1644167167 1526726655 1375731711 1258291199 1124073471 939524095 771751935 587202559 419430399 218103807 67108863 0 0 0 0 0 0 0 33554431 134217727 285212671 486539263 671088639 872415231 1056964607 1207959551 1358954495 1509949439 1644167167 1761607679 1862270975 1962934271 2046820351 2130706431 2181038079 2231369727 2264924159 2298478591 2315255807 2315255807 2315255807 2281701375 2248146943 2197815295 2147483647 2080374783 1996488703 1895825407 1795162111 1677721599 1560281087 1409286143 1258291199 1124073471 922746879 738197503 553648127 369098751 134217727 33554431 0 0 0 0 0 0 83886079 234881023 419430399 637534207 822083583 1023410175 1207959551 1358954495 1509949439 1660944383 1795162111 1895825407 2013265919 2097151999 2197815295 2281701375 2348810239 2382364671 2432696319 2466250751 2466250751 2466250751 2466250751 2432696319 2399141887 2365587455 2298478591 2231369727 2130706431 2030043135 1946157055 1828716543 1711276031 1560281087 1409286143 1258291199 1090519039 889192447 704643071 503316479 285212671 83886079 0 0 0 0 0 50331647 167772159 335544319 553648127 754974719 973078527 1140850687 1308622847 1476395007 1660944383 1795162111 1929379839 2030043135 2147483647 2264924159 2348810239 2415919103 2499805183 2533359615 2583691263 2600468479 2634022911 2634022911 2617245695 2583691263 2550136831 2516582399 2449473535 2365587455 2298478591 2181038079 2080374783 1962934271 1845493759 1711276031 1543503871 1358954495 1207959551 1023410175 838860799 620756991 419430399 184549375 50331647 0 0 0 0 83886079 234881023 452984831 687865855 889192447 1090519039 1258291199 1442840575 1627389951 1761607679 1912602623 2046820351 2181038079 2281701375 2382364671 2466250751 2550136831 2617245695 2684354559 2717908991 2751463423 2768240639 2785017855 2751463423 2734686207 2701131775 2650800127 2583691263 2499805183 2415919103 2332033023 2214592511 2097151999 1962934271 1811939327 1677721599 1509949439 1325400063 1140850687 956301311 754974719 536870911 301989887 83886079 0 0 0 50331647 167772159 335544319 553648127 771751935 989855743 1191182335 1375731711 1560281087 1728053247 1879048191 2030043135 2164260863 2281701375 2399141887 2499805183 2600468479 2684354559 2751463423 2801795071 2852126719 2885681151 2902458367 2902458367 2885681151 2868903935 2818572287 2768240639 2701131775 2617245695 2533359615 2432696319 2332033023 2197815295 2063597567 1929379839 1778384895 1610612735 1442840575 1258291199 1056964607 855638015 620756991 419430399 184549375 50331647 0 0 83886079 234881023 436207615 671088639 872415231 1090519039 1291845631 1476395007 1660944383 1828716543 1979711487 2113929215 2264924159 2399141887 2516582399 2617245695 2717908991 2801795071 2868903935 2919235583 2986344447 3003121663 3003121663 3019898879 3003121663 2986344447 2952790015 2885681151 2818572287 2751463423 2650800127 2550136831 2432696319 2315255807 2181038079 2030043135 1862270975 1728053247 1543503871 1358954495 1157627903 939524095 738197503 536870911 285212671 83886079 0 0 100663295 268435455 503316479 738197503 973078527 1174405119 1375731711 1560281087 1744830463 1929379839 2080374783 2214592511 2365587455 2483027967 2617245695 2717908991 2818572287 2902458367 2969567231 3036676095 3087007743 3103784959 3137339391 3137339391 3120562175 3103784959 3053453311 2986344447 2919235583 2835349503 2751463423 2650800127 2533359615 2399141887 2264924159 2130706431 1979711487 1811939327 1610612735 1442840575 1241513983 1040187391 822083583 587202559 352321535 100663295 0 50331647 167772159 369098751 603979775 822083583 1040187391 1258291199 1442840575 1627389951 1811939327 1996488703 2164260863 2298478591 2449473535 2583691263 2701131775 2818572287 2902458367 3003121663 3087007743 3137339391 3187671039 3221225471 3254779903 3254779903 3238002687 3204448255 3154116607 3103784959 3019898879 2936012799 2835349503 2751463423 2617245695 2483027967 2348810239 2214592511 2046820351 1879048191 1694498815 1509949439 1308622847 1124073471 889192447 671088639 436207615 201326591 50331647 67108863 201326591 402653183 654311423 889192447 1107296255 1308622847 1526726655 1711276031 1879048191 2063597567 2214592511 2365587455 2516582399 2650800127 2768240639 2885681151 2986344447 3087007743 3154116607 3238002687 3271557119 3321888767 3338665983 3338665983 3321888767 3305111551 3238002687 3187671039 3120562175 3019898879 2902458367 2801795071 2684354559 2566914047 2432696319 2264924159 2113929215 1946157055 1761607679 1593835519 1392508927 1174405119 956301311 721420287 486539263 251658239 67108863 83886079 234881023 452984831 704643071 956301311 1157627903 1375731711 1560281087 1761607679 1946157055 2113929215 2281701375 2432696319 2583691263 2701131775 2818572287 2936012799 3053453311 3154116607 3238002687 3305111551 3355443199 3405774847 3439329279 3456106495 3422552063 3388997631 3321888767 3271557119 3187671039 3087007743 2969567231 2868903935 2751463423 2617245695 2483027967 2332033023 2181038079 2013265919 1811939327 1627389951 1442840575 1241513983 1023410175 788529151 536870911 301989887 83886079 100663295 268435455 503316479 738197503 989855743 1207959551 1409286143 1610612735 1811939327 1979711487 2164260863 2315255807 2466250751 2617245695 2751463423 2885681151 3003121663 3103784959 3204448255 3288334335 3372220415 3439329279 3506438143 3523215359 3523215359 3506438143 3472883711 3388997631 3321888767 3238002687 3137339391 3036676095 2936012799 2801795071 2667577343 2516582399 2365587455 2214592511 2046820351 1862270975 1694498815 1459617791 1275068415 1056964607 838860799 570425343 352321535 100663295 100663295 285212671 553648127 771751935 1023410175 1241513983 1442840575 1627389951 1845493759 2030043135 2197815295 2332033023 2516582399 2650800127 2785017855 2919235583 3019898879 3137339391 3254779903 3338665983 3422552063 3506438143 3556769791 3607101439 3607101439 3590324223 3523215359 3456106495 3372220415 3271557119 3187671039 3070230527 2952790015 2835349503 2684354559 2550136831 2399141887 2231369727 2080374783 1895825407 1711276031 1509949439 1308622847 1090519039 872415231 620756991 385875967 100663295 117440511 301989887 553648127 788529151 1040187391 1258291199 1459617791 1677721599 1845493759 2030043135 2197815295 2365587455 2516582399 2667577343 2801795071 2936012799 3053453311 3170893823 3288334335 3372220415 3456106495 3539992575 3623878655 3674210303 3674210303 3640655871 3573547007 3489660927 3405774847 3305111551 3204448255 3103784959 2969567231 2868903935 2717908991 2566914047 2432696319 2264924159 2080374783 1929379839 1728053247 1543503871 1308622847 1124073471 889192447 620756991 402653183 117440511 117440511 301989887 553648127 805306367 1040187391 1258291199 1459617791 1677721599 1845493759 2046820351 2197815295 2365587455 2533359615 2684354559 2818572287 2936012799 3070230527 3170893823 3288334335 3372220415 3489660927 3556769791 3640655871 3724541951 3741319167 3674210303 3590324223 3506438143 3422552063 3305111551 3221225471 3103784959 2986344447 2868903935 2734686207 2583691263 2432696319 2264924159 2080374783 1929379839 1728053247 1543503871 1308622847 1124073471 889192447 637534207 402653183 117440511 117440511 301989887 553648127 788529151 1040187391 1258291199 1459617791 1677721599 1845493759 2030043135 2197815295 2365587455 2516582399 2667577343 2801795071 2936012799 3053453311 3170893823 3288334335 3372220415 3456106495 3539992575 3623878655 3674210303 3674210303 3640655871 3573547007 3489660927 3405774847 3305111551 3204448255 3103784959 2969567231 2868903935 2717908991 2566914047 2432696319 2264924159 2080374783 1929379839 1728053247 1543503871 1308622847 1124073471 889192447 620756991 402653183 117440511 100663295 285212671 553648127 771751935 1023410175 1241513983 1442840575 1627389951 1845493759 2030043135 2197815295 2332033023 2516582399 2650800127 2785017855 2919235583 3019898879 3137339391 3254779903 3338665983 3422552063 3506438143 3556769791 3607101439 3607101439 3590324223 3523215359 3456106495 3372220415 3271557119 3187671039 3070230527 2952790015 2835349503 2684354559 2550136831 2399141887 2231369727 2080374783 1895825407 1711276031 1509949439 1308622847 1090519039 872415231 620756991 385875967 100663295 100663295 268435455 503316479 738197503 989855743 1207959551 1409286143 1610612735 1811939327 1979711487 2164260863 2315255807 2466250751 2617245695 2751463423 2885681151 3003121663 3103784959 3204448255 3288334335 3372220415 3439329279 3506438143 3523215359 3523215359 3506438143 3472883711 3388997631 3321888767 3238002687 3137339391 3036676095 2936012799 2801795071 2667577343 2516582399 2365587455 2214592511 2046820351 1862270975 1694498815 1459617791 1275068415 1056964607 838860799 570425343 352321535 100663295 83886079 234881023 452984831 704643071 956301311 1157627903 1375731711 1560281087 1761607679 1946157055 2113929215 2281701375 2432696319 2583691263 2701131775 2818572287 2936012799 3053453311 3154116607 3238002687 3305111551 3355443199 3405774847 3439329279 3456106495 3422552063 3388997631 3321888767 3271557119 3187671039 3087007743 2969567231 2868903935 2751463423 2617245695 2483027967 2332033023 2181038079 2013265919 1811939327 1627389951 1442840575 1241513983 1023410175 788529151 536870911 301989887 83886079 67108863 201326591 402653183 654311423 889192447 1107296255 1308622847 1526726655 1711276031 1879048191 2063597567 2214592511 2365587455 2516582399 2650800127 2768240639 2885681151 2986344447 3087007743 3154116607 3238002687 3271557119 3321888767 3338665983 3338665983 3321888767 3305111551 3238002687 3187671039 3120562175 3019898879 2902458367 2801795071 2684354559 2566914047 2432696319 2264924159 2113929215 1946157055 1761607679 1593835519 1392508927 1174405119 956301311 721420287 486539263 251658239 67108863 50331647 167772159 369098751 603979775 822083583 1040187391 1258291199 1442840575 1627389951 1811939327 1996488703 2164260863 2298478591 2449473535 2583691263 2701131775 2818572287 2902458367 3003121663 3087007743 3137339391 3187671039 3221225471 3254779903 3254779903 3238002687 3204448255 3154116607 3103784959 3019898879 2936012799 2835349503 2751463423 2617245695 2483027967 2348810239 2214592511 2046820351 1879048191 1694498815 1509949439 1308622847 1124073471 889192447 671088639 436207615 201326591 50331647 0 100663295 268435455 503316479 738197503 973078527 1174405119 1375731711 1560281087 1744830463 1929379839 2080374783 2214592511 2365587455 2483027967 2617245695 2717908991 2818572287 2902458367 2969567231 3036676095 3087007743 3103784959 3137339391 3137339391 3120562175 3103784959 3053453311 2986344447 2919235583 2835349503 2751463423 2650800127 2533359615 2399141887 2264924159 2130706431 1979711487 1811939327 1610612735 1442840575 1241513983 1040187391 822083583 587202559 352321535 100663295 0 0 83886079 234881023 436207615 671088639 872415231 1090519039 1291845631 1476395007 1660944383 1828716543 1979711487 2113929215 2264924159 2399141887 2516582399 2617245695 2717908991 2801795071 2868903935 2919235583 2986344447 3003121663 3003121663 3019898879 3003121663 2986344447 2952790015 2885681151 2818572287 2751463423 2650800127 2550136831 2432696319 2315255807 2181038079 2030043135 1862270975 1728053247 1543503871 1358954495 1157627903 939524095 738197503 536870911 285212671 83886079 0 0 50331647 167772159 335544319 553648127 771751935 989855743 1191182335 1375731711 1560281087 1728053247 1879048191 2030043135 2164260863 2281701375 2399141887 2499805183 2600468479 2684354559 2751463423 2801795071 2852126719 2885681151 2902458367 2902458367 2885681151 2868903935 2818572287 2768240639 2701131775 2617245695 2533359615 2432696319 2332033023 2197815295 2063597567 1929379839 1778384895 1610612735 1442840575 1258291199 1056964607 855638015 620756991 419430399 184549375 50331647 0 0 0 83886079 234881023 452984831 687865855 889192447 1090519039 1258291199 1442840575 1627389951 1761607679 1912602623 2046820351 2181038079 2281701375 2382364671 2466250751 2550136831 2617245695 2684354559 2717908991 2751463423 2768240639 2785017855 2751463423 2734686207 2701131775 2650800127 2583691263 2499805183 2415919103 2332033023 2214592511 2097151999 1962934271 1811939327 1677721599 1509949439 1325400063 1140850687 956301311 754974719 536870911 301989887 83886079 0 0 0 0 50331647 167772159 335544319 553648127 754974719 973078527 1140850687 1308622847 1476395007 1660944383 1795162111 1929379839 2030043135 2147483647 2264924159 2348810239 2415919103 2499805183 2533359615 2583691263 2600468479 2634022911 2634022911 2617245695 2583691263 2550136831 2516582399 2449473535 2365587455 2298478591 2181038079 2080374783 1962934271 1845493759 1711276031 1543503871 1358954495 1207959551 1023410175 838860799 620756991 419430399 184549375 50331647 0 0 0 0 0 83886079 234881023 419430399 637534207 822083583 1023410175 1207959551 1358954495 1509949439 1660944383 1795162111 1895825407 2013265919 2097151999 2197815295 2281701375 2348810239 2382364671 2432696319 2466250751 2466250751 2466250751 2466250751 2432696319 2399141887 2365587455 2298478591 2231369727 2130706431 2030043135 1946157055 1828716543 1711276031 1560281087 1409286143 1258291199 1090519039 889192447 704643071 503316479 285212671 83886079 0 0 0 0 0 0 33554431 134217727 285212671 486539263 671088639 872415231 1056964607 1207959551 1358954495 1509949439 1644167167 1761607679 1862270975 1962934271 2046820351 2130706431 2181038079 2231369727 2264924159 2298478591 2315255807 2315255807 2315255807 2281701375 2248146943 2197815295 2147483647 2080374783 1996488703 1895825407 1795162111 1677721599 1560281087 1409286143 1258291199 1124073471 922746879 738197503 553648127 369098751 134217727 33554431 0 0 0 0 0 0 0 67108863 184549375 352321535 536870911 704643071 889192447 1056964607 1207959551 1342177279 1476395007 1610612735 1694498815 1811939327 1879048191 1962934271 2030043135 2080374783 2113929215 2147483647 2147483647 2147483647 2147483647 2130706431 2097151999 2046820351 1979711487 1912602623 1828716543 1744830463 1644167167 1526726655 1375731711 1258291199 1124073471 939524095 771751935 587202559 419430399 218103807 67108863 0 0 0 0 0 0 0 0 0 83886079 218103807 369098751 536870911 704643071 889192447 1040187391 1191182335 1308622847 1409286143 1543503871 1627389951 1694498815 1778384895 1845493759 1895825407 1946157055 1962934271 1979711487 1979711487 1979711487 1962934271 1912602623 1862270975 1795162111 1728053247 1660944383 1577058303 1459617791 1342177279 1224736767 1107296255 939524095 771751935 587202559 419430399 251658239 83886079 0 0 0 0 0 0 0 0 0 0 0 83886079 218103807 385875967 536870911 687865855 855638015 1006632959 1124073471 1224736767 1342177279 1442840575 1526726655 1593835519 1660944383 1711276031 1761607679 1778384895 1778384895 1778384895 1778384895 1778384895 1728053247 1677721599 1627389951 1543503871 1476395007 1375731711 1275068415 1157627903 1040187391 905969663 754974719 587202559 419430399 268435455 83886079 0 0 0 0 0 0 0 0 0 0 0 0 0 83886079 218103807 369098751 520093695 654311423 805306367 939524095 1056964607 1157627903 1241513983 1325400063 1409286143 1476395007 1509949439 1543503871 1577058303 1610612735 1610612735 1593835519 1560281087 1509949439 1493172223 1426063359 1342177279 1275068415 1191182335 1090519039 973078527 855638015 704643071 570425343 419430399 251658239 83886079 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 83886079 201326591 335544319 469762047 620756991 738197503 855638015 956301311 1040187391 1124073471 1224736767 1258291199 1325400063 1358954495 1375731711 1375731711 1375731711 1375731711 1358954495 1342177279 1291845631 1224736767 1157627903 1073741823 989855743 889192447 788529151 654311423 520093695 385875967 234881023 83886079 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 67108863 167772159 268435455 385875967 520093695 654311423 721420287 838860799 922746879 989855743 1056964607 1107296255 1140850687 1174405119 1191182335 1191182335 1191182335 1157627903 1124073471 1073741823 1006632959 939524095 872415231 771751935 671088639 570425343 436207615 318767103 167772159 67108863 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 33554431 100663295 201326591 301989887 402653183 520093695 637534207 687865855 754974719 838860799 889192447 922746879 956301311 973078527 973078527 956301311 939524095 905969663 855638015 788529151 704643071 654311423 570425343 436207615 352321535 218103807 100663295 33554431 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 50331647 117440511 201326591 301989887 385875967 452984831 536870911 587202559 637534207 654311423 687865855 687865855 704643071 687865855 671088639 637534207 603979775 553648127 486539263 402653183 335544319 218103807 117440511 50331647 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 50331647 117440511 167772159 218103807 301989887 352321535 385875967 436207615 469762047 469762047 469762047 469762047 452984831 402653183 369098751 318767103 251658239 167772159 117440511 50331647 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 50331647 100663295 134217727 167772159 184549375 201326591 218103807 218103807 201326591 184549375 167772159 134217727 100663295 50331647 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0)
	offset: 0@0)! !

!CornerGripMorph class methodsFor: 'handle settings' stamp: 'HenrikSperreJohansen 12/10/2009 21:04'!
defaultFormFromFileNamed: aString
	"If you dislike the alpha scale of the default handle, use this method to install a new one.
	File should be in a readable image format, and contain a 48x48 32bit radial gradient with color white.
	Use passiveColor: / activeColor: to change them after file is loaded, see initialize for an example"

	|sourceStream|
	sourceStream := WriteStream on: String new. 
	sourceStream nextPutAll: 'defaultForm';
		nextPut: Character cr;
		nextPut: Character cr;
		nextPut: $^;
		nextPut: $(;
		nextPutAll: 	(ImageReadWriter formFromFileNamed: aString) storeString;
			nextPut: $).
	self class compile: sourceStream contents.! !

!CornerGripMorph class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 13:42'!
drawCornerResizeHandles
	<preference: 'Draw Corner Resize Handles'
		category: 'windows'
		description: 'Governs the resize handles on windows should be drawn. This does not disable them'
		type: #Boolean>
	^ DrawCornerResizeHandles ifNil: [ false ]! !

!CornerGripMorph class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 14:04'!
drawCornerResizeHandles: aBoolean
	
	DrawCornerResizeHandles := aBoolean.
	World invalidRect: World bounds from: World.! !

!CornerGripMorph class methodsFor: 'class initialization' stamp: 'HenrikSperreJohansen 2/25/2010 02:23'!
initialize
	"CornerGripMorph initialize"
	
	super initialize.
	self initializeActiveForm.
	self initializePassiveForm.
	self activeColor: Color orange.! !

!CornerGripMorph class methodsFor: 'class initialization' stamp: 'HenrikSperreJohansen 12/10/2009 21:05'!
initializeActiveForm

	^ActiveForm := self defaultForm! !

!CornerGripMorph class methodsFor: 'class initialization' stamp: 'HenrikSperreJohansen 12/10/2009 21:05'!
initializePassiveForm

	^PassiveForm := self defaultForm! !

!CornerGripMorph class methodsFor: 'handle settings' stamp: 'mt 7/2/2016 13:28'!
passiveColor
	<preference: 'Corner Grip color'
		category: 'colors'
		description: 'The default color of window corners'
		type: #Color>
	^(self passiveForm colorAt: 24@24) alpha:  1! !

!CornerGripMorph class methodsFor: 'handle settings' stamp: 'HenrikSperreJohansen 2/25/2010 01:53'!
passiveColor: aColor 
	| canvas |
	canvas := self initializePassiveForm getCanvas.
	canvas privatePort fillPattern: aColor;
		 combinationRule: Form rgbMul;
		 fillRect: self passiveForm boundingBox offset: 0 @ 0.
	self
		allSubInstancesDo: [:each | each setDefaultColors; changed]! !

!CornerGripMorph class methodsFor: 'accessing' stamp: 'HenrikSperreJohansen 12/10/2009 21:06'!
passiveForm
	^PassiveForm ifNil: [self initializePassiveForm]! !

!CornerGripMorph methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/10/2009 21:08'!
activeForm
	^self clipForm: self class activeForm! !

!CornerGripMorph methodsFor: 'drawing' stamp: 'HenrikSperreJohansen 12/10/2009 21:36'!
alphaHandle

	handleColor ifNil: [handleColor := self passiveForm].
	"The following line is only needed on first load, so existing windows don't blow up from the new handles. 
	Can safely be deleted along with this comment in a later update"
	(handleColor class == Form) 
		ifFalse: [handleColor := self passiveForm].
	^handleColor

	
	! !

!CornerGripMorph methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/10/2009 20:42'!
borderOffset
	"The offset from my corner to where the border starts"
	^self subclassResponsibility! !

!CornerGripMorph methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/10/2009 21:13'!
clipForm: aHandle
	|cutArea|
	"This doesn't really needs to be done every draw, but only if border width changes.
	In that case,, we'd have to install a newly initialized one anyways, so the current method will fail.
	Good enough for now though."
	cutArea := self transparentRectangle.
	aHandle getCanvas image: cutArea 
		at: self borderOffset 
		sourceRect: cutArea boundingBox 
		rule: Form and. 
	^aHandle
	
	! !

!CornerGripMorph methodsFor: 'accessing' stamp: 'mt 4/24/2015 11:32'!
defaultHeight

	^ self class gripThickness * 6! !

!CornerGripMorph methodsFor: 'accessing' stamp: 'mt 4/24/2015 11:32'!
defaultWidth

	^ self class gripThickness * 6! !

!CornerGripMorph methodsFor: 'drawing' stamp: 'tfel 2/28/2010 13:43'!
drawOn: aCanvas

	self class drawCornerResizeHandles
		ifTrue: [
			bounds := self bounds.
			aCanvas 
				translucentImage: (self alphaHandle) 
				at: (bounds origin ) 
				sourceRect: (self handleOrigin extent: bounds extent)]! !

!CornerGripMorph methodsFor: 'private' stamp: 'hsj 12/10/2009 16:03'!
handleOrigin
	"The handles origin is the offset into the alphaForm"
^self subclassResponsibility! !

!CornerGripMorph methodsFor: 'as yet unclassified' stamp: 'jmv 2/19/2006 14:30'!
handlesMouseDown: anEvent
	^ true! !

!CornerGripMorph methodsFor: 'as yet unclassified' stamp: 'jmv 2/19/2006 14:30'!
handlesMouseOver: anEvent
	^true! !

!CornerGripMorph methodsFor: 'as yet unclassified' stamp: 'jmv 2/2/2006 14:24'!
initialize
	super initialize.
	self extent: self defaultWidth+2 @ (self defaultHeight+2).
	self layoutFrame: self gripLayoutFrame! !

!CornerGripMorph methodsFor: 'as yet unclassified' stamp: 'mt 5/27/2016 14:33'!
mouseDown: anEvent 
	"Disable drop shadow to improve performance."

	super mouseDown: anEvent.

	target ifNil: [^ self].
	target fastFramingOn ifFalse: [
		self setProperty: #targetHadDropShadow toValue: target hasDropShadow.
		target hasDropShadow: false].! !

!CornerGripMorph methodsFor: 'as yet unclassified' stamp: 'mt 5/27/2016 14:06'!
mouseMove: anEvent 
	| delta |
	target ifNil: [^ self].
	target fastFramingOn 
		ifTrue: [delta := target doFastWindowReframe: self ptName] 
		ifFalse: [
			delta := lastMouse ifNil: [0@0] ifNotNil: [anEvent cursorPoint - lastMouse].
			lastMouse := anEvent cursorPoint.
			self apply: delta.
			self bounds: (self bounds origin + delta extent: self bounds extent)].! !

!CornerGripMorph methodsFor: 'as yet unclassified' stamp: 'mt 5/27/2016 14:33'!
mouseUp: anEvent 

	target ifNil: [^ self].
	target fastFramingOn ifFalse: [
		(self valueOfProperty: #targetHadDropShadow ifAbsent: [false]) ifTrue: [target hasDropShadow: true].
		self removeProperty: #targetHadDropShadow].! !

!CornerGripMorph methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/10/2009 21:11'!
passiveForm
	^self clipForm: self class passiveForm! !

!CornerGripMorph methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/10/2009 21:11'!
setDefaultColors

	handleColor := self passiveForm.! !

!CornerGripMorph methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/10/2009 21:11'!
setInverseColors
	handleColor := self activeForm.! !

!CornerGripMorph methodsFor: 'as yet unclassified' stamp: 'bvs 3/24/2004 14:38'!
target: aMorph

	target := aMorph! !

!CornerGripMorph methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/10/2009 20:40'!
transparentRectangle
	"This could be a class var, provided either bounds of grips does not change, or one ensures a new one is installed when such an event occurs"
	^Form extent: self bounds extent depth: 32! !

CornerGripMorph subclass: #TopLeftGripMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!TopLeftGripMorph commentStamp: 'jmv 1/29/2006 17:18' prior: 0!
I am the handle in the left top of windows used for resizing them.!

!TopLeftGripMorph methodsFor: 'target resize' stamp: 'jmv 1/29/2006 18:02'!
apply: delta 
	| oldBounds |
	oldBounds := target bounds.
	target
		bounds: (oldBounds origin + delta corner: oldBounds corner)! !

!TopLeftGripMorph methodsFor: 'private' stamp: 'hsj 12/10/2009 16:44'!
borderOffset
	|width|
	width :=SystemWindow borderWidth +1.
	 ^self handleOrigin + width asPoint! !

!TopLeftGripMorph methodsFor: 'accessing' stamp: 'mt 4/28/2015 14:31'!
gripLayoutFrame
	^ LayoutFrame
		fractions: (0 @ 0 corner: 0 @ 0)
		offsets: (0 @ 0 corner: 0 @ 0)! !

!TopLeftGripMorph methodsFor: 'private' stamp: 'hsj 12/10/2009 16:05'!
handleOrigin
^25@25! !

!TopLeftGripMorph methodsFor: 'layout' stamp: 'Igor.Stasenko 9/4/2009 16:04'!
layoutProportionallyIn: newBounds
	| b |
	b := owner bounds.
	self bounds: (b topLeft extent: self extent)
! !

!TopLeftGripMorph methodsFor: 'target resize' stamp: 'md 2/24/2006 22:44'!
ptName
	^#topLeft! !

!TopLeftGripMorph methodsFor: 'accessing' stamp: 'jmv 1/29/2006 17:52'!
resizeCursor

	^ Cursor resizeForEdge: #topLeft! !

CornerGripMorph subclass: #TopRightGripMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!TopRightGripMorph commentStamp: 'jmv 1/29/2006 17:18' prior: 0!
I am the handle in the right top of windows used for resizing them.!

!TopRightGripMorph methodsFor: 'target resize' stamp: 'jmv 1/29/2006 18:05'!
apply: delta 
	| oldBounds |
	oldBounds := target bounds.
	target
		bounds: (oldBounds origin + (0@delta y) corner: oldBounds corner + (delta x @ 0))! !

!TopRightGripMorph methodsFor: 'private' stamp: 'hsj 12/10/2009 16:43'!
borderOffset
	|width|
	width :=SystemWindow borderWidth +1 .
	 ^self handleOrigin + ( width negated @ (width) )! !

!TopRightGripMorph methodsFor: 'accessing' stamp: 'mt 4/28/2015 14:38'!
gripLayoutFrame
	^ LayoutFrame
		fractions: (1 @ 0 corner: 1 @ 0)
		offsets:  (0 @ 0 corner: 0 @ 0)! !

!TopRightGripMorph methodsFor: 'private' stamp: 'hsj 12/10/2009 16:04'!
handleOrigin
^0@25! !

!TopRightGripMorph methodsFor: 'layout' stamp: 'Igor.Stasenko 9/4/2009 16:02'!
layoutProportionallyIn: newBounds
	| b |
	b := owner bounds.
	
	self bounds: (b right - self width @ b top extent: self extent)
! !

!TopRightGripMorph methodsFor: 'accessing' stamp: 'md 2/24/2006 22:44'!
ptName
	^#topRight! !

!TopRightGripMorph methodsFor: 'accessing' stamp: 'jmv 1/29/2006 17:52'!
resizeCursor

	^ Cursor resizeForEdge: #topRight! !

AbstractResizerMorph subclass: #ProportionalSplitterMorph
	instanceVariableNames: 'leftOrTop rightOrBottom splitsTopAndBottom oldColor traceMorph handle movements'
	classVariableNames: 'SmartHorizontalSplitters SmartVerticalSplitters'
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!ProportionalSplitterMorph commentStamp: 'jmv 1/29/2006 17:16' prior: 0!
I am the morph the user grabs to adjust pane splitters.!

!ProportionalSplitterMorph class methodsFor: 'preferences' stamp: 'md 2/24/2006 22:47'!
fastSplitterResize
	
	^ Preferences fastDragWindowForMorphic.! !

!ProportionalSplitterMorph class methodsFor: 'private' stamp: 'cmm 9/10/2013 14:06'!
preferenceChanged: aBoolean 
	"Take immediate effect for all in a  World."
	self allInstances do:
		[ : each | (each isInWorld and: [ aBoolean and: [ each wantsSteps ] ])
			ifTrue: [ each startStepping ]
			ifFalse: [ each stopStepping ] ]! !

!ProportionalSplitterMorph class methodsFor: 'preferences' stamp: 'jrp 7/9/2005 17:43'!
showSplitterHandles

	^ Preferences valueOfPreference: #showSplitterHandles ifAbsent: [true]! !

!ProportionalSplitterMorph class methodsFor: 'preferences' stamp: 'cmm 8/24/2013 23:27'!
smartHorizontalSplitters
	<preference: 'Smart Horizontal Splitters'
		category: 'Morphic'
		description: 'When true, horizontal splitter bars will automatically reposition themselves to increase the quantity of exposed information, if possible..'
		type: #Boolean>
	^ SmartHorizontalSplitters ifNil: [ false ]! !

!ProportionalSplitterMorph class methodsFor: 'preferences' stamp: 'cmm 8/24/2013 23:22'!
smartHorizontalSplitters: aBoolean 
	SmartHorizontalSplitters := aBoolean.
	self preferenceChanged: aBoolean! !

!ProportionalSplitterMorph class methodsFor: 'preferences' stamp: 'cmm 8/24/2013 23:27'!
smartVerticalSplitters
	<preference: 'Smart Vertical Splitters'
		category: 'Morphic'
		description: 'When true, vertical bars between lists will automatically reposition themselves to balance the number of characters occluded on either side of the bar.'
		type: #Boolean>
	^ SmartVerticalSplitters ifNil: [ false ]! !

!ProportionalSplitterMorph class methodsFor: 'preferences' stamp: 'cmm 8/24/2013 23:25'!
smartVerticalSplitters: aBoolean 
	SmartVerticalSplitters := aBoolean.
	self preferenceChanged: aBoolean! !

!ProportionalSplitterMorph methodsFor: 'controlled morphs' stamp: 'bvs 3/24/2004 16:57'!
addLeftOrTop: aMorph

	leftOrTop add: aMorph! !

!ProportionalSplitterMorph methodsFor: 'controlled morphs' stamp: 'bvs 3/24/2004 16:55'!
addRightOrBottom: aMorph

	rightOrBottom add: aMorph.
	
	! !

!ProportionalSplitterMorph methodsFor: 'direction' stamp: 'mt 4/24/2015 11:35'!
beSplitsLeftAndRight

	splitsTopAndBottom := false.
	self
		hResizing: #rigid;
		vResizing: #spaceFill.
	self width: self class gripThickness.! !

!ProportionalSplitterMorph methodsFor: 'direction' stamp: 'mt 4/24/2015 11:34'!
beSplitsTopAndBottom

	splitsTopAndBottom := true.
	self
		vResizing: #rigid;
		hResizing: #spaceFill.
	self height: self class gripThickness.! !

!ProportionalSplitterMorph methodsFor: 'controlled morphs' stamp: 'ml 9/25/2009 08:31'!
bordersOn: aMorph
	"Answer true if the aMorph is one of my neighbours."

	^ (leftOrTop includes: aMorph) or: [rightOrBottom includes: aMorph]! !

!ProportionalSplitterMorph methodsFor: 'boundaries' stamp: 'spfa 4/20/2010 22:38'!
bottomBoundary
	"Answert the bottommost x position the receiver could be moved."

	| splitter morphs |
	splitter := self splitterBelow.
	morphs := self commonNeighbours: rightOrBottom with: splitter.
	^ (splitter
		ifNil: [owner isSystemWindow ifTrue: [owner panelRect bottom]
				ifFalse: [owner innerBounds bottom]]
		ifNotNil: [splitter top])
		- (self minimumHeightOf: morphs)! !

!ProportionalSplitterMorph methodsFor: 'layout' stamp: 'cmm 8/25/2013 13:09'!
canEncroachWhiteSpaceOf: morphs 
	^ morphs allSatisfy: [ : each | each canBeEncroached ]! !

!ProportionalSplitterMorph methodsFor: 'layout' stamp: 'cmm 4/2/2015 16:40'!
charactersOccludedIn: aCollection
	^ aCollection
		inject: 0
		into:
			[ : sum : each | sum +
				(each isMorphicModel
					ifTrue: [ each charactersOccluded ]
					ifFalse: [ 0 ]) ]! !

!ProportionalSplitterMorph methodsFor: 'controlled morphs' stamp: 'ml 9/25/2009 08:29'!
commonNeighbours: morphs with: aProportionalSplitterMorphOrNil
	"Answer the subset of morphs which is also confined by aProportionalSplitterMorphOrNil."

	^ aProportionalSplitterMorphOrNil isNil
		ifTrue: [morphs]
		ifFalse: [morphs select: [ :which |
				aProportionalSplitterMorphOrNil bordersOn: which]]! !

!ProportionalSplitterMorph methodsFor: 'displaying' stamp: 'apl 7/8/2005 13:38'!
getOldColor
	^ oldColor ifNil: [Color transparent]! !

!ProportionalSplitterMorph methodsFor: 'displaying' stamp: 'jrp 7/4/2005 10:50'!
handleRect

	^ Rectangle
		center: self bounds center 
		extent: (self splitsTopAndBottom
			ifTrue: [self handleSize transposed] 
			ifFalse: [self handleSize])! !

!ProportionalSplitterMorph methodsFor: 'displaying' stamp: 'mt 4/24/2015 11:45'!
handleSize

	^ self class gripThickness @ 30! !

!ProportionalSplitterMorph methodsFor: 'initialization' stamp: 'mt 4/24/2015 11:35'!
initialize

	super initialize.
	
	self beSplitsLeftAndRight.
	
	leftOrTop := OrderedCollection new.
	rightOrBottom := OrderedCollection new.
	
	Preferences showSplitterHandles 
		ifTrue: [
			handle := CircleMorph new
					borderWidth: 0;
					extent: 4@4;
					yourself.
			handle fillStyle: ((GradientFillStyle 
						ramp: {0.0 -> Preferences defaultWindowColor muchLighter. 
							1.0 -> Preferences defaultWindowColor darker})
						origin: handle topLeft;
						direction: 0 @ handle bounds extent y;
						normal: handle bounds extent x @ 0;
						radial: false;
						yourself).
			self addMorphCentered: handle].
	self initializeMovements! !

!ProportionalSplitterMorph methodsFor: 'initialization' stamp: 'cmm 4/17/2015 23:18'!
initializeMovements
	movements := OrderedCollection new: 3 withAll: 0! !

!ProportionalSplitterMorph methodsFor: 'displaying' stamp: 'jrp 7/9/2005 17:44'!
isCursorOverHandle
	^ self class showSplitterHandles not or: [self handleRect containsPoint: ActiveHand cursorPoint]! !

!ProportionalSplitterMorph methodsFor: 'layout' stamp: 'tfel 2/27/2010 17:45'!
layoutChanged	
	
	super layoutChanged.
	handle ifNotNil: [handle position: self bounds center - (2@2)]! !

!ProportionalSplitterMorph methodsFor: 'boundaries' stamp: 'mt 4/21/2010 09:30:03.495'!
leftBoundary
	"Answer the leftmost y position the receiver could be moved."

	| splitter morphs |
	splitter := self splitterLeft.
	morphs := self commonNeighbours: leftOrTop with: splitter.
	^ (splitter
		ifNil: [owner isSystemWindow ifTrue: [owner panelRect left]
				ifFalse: [owner innerBounds left]]
		ifNotNil: [splitter right])
		 + (self minimumWidthOf: morphs)! !

!ProportionalSplitterMorph methodsFor: 'layout' stamp: 'cmm 3/23/2015 10:52'!
leftRightImbalance
	"First check if I find myself out of range due to user having reduced size of parent."
	^ self left < self leftBoundary "too far left"
		ifTrue: [ self leftBoundary-self left ]
		ifFalse:
			[ self right > self rightBoundary "too far right"
				ifTrue: [ self rightBoundary-self right ]
				ifFalse: [ self occlusionDifference ] ]! !

!ProportionalSplitterMorph methodsFor: 'boundaries' stamp: 'mt 4/28/2015 11:41'!
minimumHeightOf: aCollection
	"Answer the minimum height needed to display any of the morphs in aCollection.
	See ProportionalLayout >> #minExtentOf:in:."

	^ aCollection inject: 0 into: [ :height :morph |
		| minHeight |
		minHeight := morph minHeight.
		"morph layoutFrame ifNotNil: [:frame |
			minHeight := frame minHeightFrom: minHeight]."
		minHeight + self height max: height]! !

!ProportionalSplitterMorph methodsFor: 'boundaries' stamp: 'mt 4/28/2015 11:19'!
minimumWidthOf: aCollection
	"Answer the minimum width needed to display any of the morphs in aCollection.
	See ProportionalLayout >> #minExtentOf:in:."
	
	^ aCollection inject: 0 into: [ :width :morph |
		| minWidth |
		minWidth := morph minWidth.
		"morph layoutFrame ifNotNil: [:frame |
			minWidth := frame minWidthFrom: minWidth]."
		minWidth + self width max: width]! !

!ProportionalSplitterMorph methodsFor: 'events' stamp: 'mt 5/19/2016 08:05'!
mouseDown: anEvent 
	"If the user manually drags me, don't override him with auto positioning."

	anEvent redButtonChanged
		ifTrue: [ self withSiblingSplittersDo: [ : each | each stopStepping ] ]
		ifFalse:
			[ anEvent shiftPressed
				ifTrue: [ self startStepping ]
				ifFalse:
					[ self startStepping.
					self withSiblingSplittersDo: [ : each | each startStepping ] ] ].
	(self class showSplitterHandles not and: [ self bounds containsPoint: anEvent cursorPoint ]) ifTrue:
		[ oldColor := self color.
		self color: Color black ].
	^ super mouseDown: anEvent! !

!ProportionalSplitterMorph methodsFor: 'events' stamp: 'tfel 2/27/2010 17:50'!
mouseMove: anEvent 
	anEvent hand temporaryCursor
		ifNil: [^ self].
	self class fastSplitterResize
		ifFalse:  [self updateFromEvent: anEvent]
		ifTrue: [traceMorph
				ifNil: [traceMorph := Morph newBounds: self bounds.
					traceMorph color: (Color gray alpha: 0.5).
					traceMorph borderWidth: 0.
					self owner addMorph: traceMorph].
			splitsTopAndBottom
				ifTrue: [traceMorph position: traceMorph position x @ (self normalizedY: anEvent cursorPoint y)]
				ifFalse: [traceMorph position: (self normalizedX: anEvent cursorPoint x) @ traceMorph position y]]! !

!ProportionalSplitterMorph methodsFor: 'events' stamp: 'jrp 7/8/2005 21:42'!
mouseUp: anEvent 
	(self bounds containsPoint: anEvent cursorPoint)
		ifFalse: [anEvent hand showTemporaryCursor: nil].
	self class fastSplitterResize
		ifTrue: [self updateFromEvent: anEvent].
	traceMorph ifNotNil: [traceMorph delete. traceMorph := nil].
	self color: self getOldColor! !

!ProportionalSplitterMorph methodsFor: 'layout' stamp: 'cmm 9/10/2013 14:02'!
movements
	"Used to track my pattern of movement for the last 3 steps to fix the twitching."
"Lazy-init for now for smooth transition -- want to convert this back to direct-var access after a few months."
	^ movements ifNil: [ self initializeMovements. movements ]! !

!ProportionalSplitterMorph methodsFor: 'boundaries' stamp: 'jrp 3/21/2006 22:45'!
normalizedX: x

	^ (x max: self leftBoundary) min: self rightBoundary! !

!ProportionalSplitterMorph methodsFor: 'boundaries' stamp: 'jrp 3/21/2006 23:12'!
normalizedY: y

	^ (y max: self topBoundary) min: self bottomBoundary! !

!ProportionalSplitterMorph methodsFor: 'events' stamp: 'cmm 8/25/2013 18:15'!
occlusionDifference
	^ (self charactersOccludedIn: leftOrTop) - (self charactersOccludedIn: rightOrBottom)! !

!ProportionalSplitterMorph methodsFor: 'layout' stamp: 'mt 5/19/2016 11:21'!
proposedCorrectionWouldCauseFocusChange: correction 
	^ Preferences mouseOverForKeyboardFocus and:
		[ | edge | splitsTopAndBottom
			ifTrue:
				[ edge := correction positive
					ifTrue: [ self bottom + 3 ]
					ifFalse: [ self top - 3 ].
				self activeHand position y
					inRangeOf: edge
					and: edge + correction ]
			ifFalse:
				[ edge := correction positive
					ifTrue: [ self right ]
					ifFalse: [ self left ].
				self activeHand position x
					inRangeOf: edge
					and: edge + correction ] ]! !

!ProportionalSplitterMorph methodsFor: 'layout' stamp: 'cmm 4/17/2015 23:16'!
reduceLeftRightImbalance
	| correction |
	correction := self leftRightImbalance.
	correction abs > 1
		ifTrue:
			[ (self proposedCorrectionWouldCauseFocusChange: correction) ifFalse:
				[ self repositionBy:
					(correction abs > 4
						ifTrue: [ correction sign * 2 @ 0 ]
						ifFalse: [ correction sign @ 0 ]) ] ]
		ifFalse:
			[ self class smartVerticalSplitters ifFalse: [ self stopStepping ] ].
	^ correction! !

!ProportionalSplitterMorph methodsFor: 'layout' stamp: 'cmm 4/19/2015 21:15'!
reduceTopBottomImbalance
	| correction |
	(correction := self topBottomCorrection) isZero
		ifTrue:
			[ self class smartHorizontalSplitters ifFalse: [ self stopStepping ] ]
		ifFalse:
			[ (self proposedCorrectionWouldCauseFocusChange: correction) ifFalse: [ self repositionBy: 0 @ correction ] ].
	^ correction! !

!ProportionalSplitterMorph methodsFor: 'layout' stamp: 'mt 4/28/2015 10:03'!
repositionBy: delta
	| selfTop selfBottom selfLeft selfRight |
	leftOrTop do:
		[ : each | | firstRight firstBottom firstLeft firstTop |
		firstRight := each layoutFrame rightOffset ifNil: [ 0 ].
		firstBottom := each layoutFrame bottomOffset ifNil: [ 0 ].
		each layoutFrame rightOffset: firstRight + delta x.
		each layoutFrame bottomOffset: firstBottom + delta y.
		each layoutFrame hasFixedHeight ifTrue: [
			firstTop := each layoutFrame topOffset ifNil: [ 0 ].
			each layoutFrame topOffset: firstTop + delta y ].
		each layoutFrame hasFixedWidth ifTrue: [
			firstLeft := each layoutFrame leftOffset ifNil: [ 0 ].
			each layoutFrame leftOffset: firstLeft + delta x. ] ].
	rightOrBottom do:
		[ : each | | secondLeft secondTop secondRight secondBottom |
		secondLeft := each layoutFrame leftOffset ifNil: [ 0 ].
		secondTop := each layoutFrame topOffset ifNil: [ 0 ].
		each layoutFrame leftOffset: secondLeft + delta x.
		each layoutFrame topOffset: secondTop + delta y.
		each layoutFrame hasFixedHeight ifTrue: [
			secondBottom := each layoutFrame bottomOffset ifNil: [ 0 ].
			each layoutFrame bottomOffset: secondBottom + delta y. ].
		each layoutFrame hasFixedWidth ifTrue: [
			secondRight := each layoutFrame rightOffset ifNil: [ 0 ].
			each layoutFrame rightOffset: secondRight + delta x. ] ].

	selfTop := self layoutFrame topOffset ifNil: [ 0 ].
	selfBottom := self layoutFrame bottomOffset ifNil: [ 0 ].
	selfLeft := self layoutFrame leftOffset ifNil: [ 0 ].
	selfRight := self layoutFrame rightOffset ifNil: [ 0 ].
	self layoutFrame
		 topOffset: selfTop + delta y ;
		 bottomOffset: selfBottom + delta y ;
		 leftOffset: selfLeft + delta x ;
		 rightOffset: selfRight + delta x.
	self owner layoutChanged.
	self movements removeFirst; add: (splitsTopAndBottom ifTrue: [ delta y sign ] ifFalse: [ delta x sign ])! !

!ProportionalSplitterMorph methodsFor: 'displaying' stamp: 'bvs 3/24/2004 16:39'!
resizeCursor

	^ Cursor resizeForEdge: (splitsTopAndBottom
		ifTrue: [#top]
		ifFalse: [#left])
		! !

!ProportionalSplitterMorph methodsFor: 'boundaries' stamp: 'mt 4/29/2015 17:01'!
rightBoundary
	"Answer the rightmost x position the receiver could be moved to."

	| splitter morphs |
	splitter := self splitterRight.
	morphs := self commonNeighbours: rightOrBottom with: splitter.
	
	^ (splitter
		ifNil: [owner isSystemWindow ifTrue: [owner panelRect right]
				ifFalse: [owner innerBounds right]]
		ifNotNil: [splitter left])
		- (self minimumWidthOf: morphs)! !

!ProportionalSplitterMorph methodsFor: 'adjacent splitters' stamp: 'cmm 5/4/2015 20:10'!
siblingSplittersDo: aBlock 
	owner ifNotNil:
		[ owner submorphsDo:
			[ : each | ((each isKindOf: self class) and:
				[ self splitsTopAndBottom = each splitsTopAndBottom and: [ each ~= self ] ]) ifTrue: [ aBlock value: each ] ] ]! !

!ProportionalSplitterMorph methodsFor: 'direction' stamp: 'mt 4/28/2015 08:54'!
splitsLeftAndRight

	^ self splitsTopAndBottom not! !

!ProportionalSplitterMorph methodsFor: 'direction' stamp: 'bvs 3/24/2004 17:25'!
splitsTopAndBottom

	^ splitsTopAndBottom! !

!ProportionalSplitterMorph methodsFor: 'adjacent splitters' stamp: 'cmm 4/18/2015 14:19'!
splitterAbove
	| immediatelyAbove |
	immediatelyAbove := nil.
	self siblingSplittersDo:
		[ : each | "Splitter y's map 0 at the bottom."
		each y > self y ifTrue:
			[ immediatelyAbove
				ifNil: [ immediatelyAbove := each ]
				ifNotNil:
					[ each y < immediatelyAbove y ifTrue: [ immediatelyAbove := each ] ] ] ].
	^ immediatelyAbove! !

!ProportionalSplitterMorph methodsFor: 'adjacent splitters' stamp: 'cmm 4/18/2015 14:19'!
splitterBelow
	| immediatelyBelow |
	immediatelyBelow := nil.
	self siblingSplittersDo:
		[ : each | each y < self y ifTrue:
			[ immediatelyBelow
				ifNil: [ immediatelyBelow := each ]
				ifNotNil:
					[ each y > immediatelyBelow y ifTrue: [ immediatelyBelow := each ] ] ] ].
	^ immediatelyBelow! !

!ProportionalSplitterMorph methodsFor: 'adjacent splitters' stamp: 'cmm 4/18/2015 14:14'!
splitterLeft
	| immediateLeft |
	immediateLeft := nil.
	self siblingSplittersDo:
		[ : each | each x < self x ifTrue:
			[ immediateLeft
				ifNil: [ immediateLeft := each ]
				ifNotNil:
					[ each x > immediateLeft x ifTrue: [ immediateLeft := each ] ] ] ].
	^ immediateLeft! !

!ProportionalSplitterMorph methodsFor: 'adjacent splitters' stamp: 'cmm 4/18/2015 14:16'!
splitterRight
	| immediateRight |
	immediateRight := nil.
	self siblingSplittersDo:
		[ : each | each x > self x ifTrue:
			[ immediateRight
				ifNil: [ immediateRight := each ]
				ifNotNil:
					[ each x < immediateRight x ifTrue: [ immediateRight := each ] ] ] ].
	^ immediateRight! !

!ProportionalSplitterMorph methodsFor: 'adjacent splitters' stamp: 'cmm 4/18/2015 13:40'!
splittersAboveDo: aBlock 
	self splitterAbove ifNotNil:
		[ : splitter | aBlock value: splitter.
		splitter splittersAboveDo: aBlock ]! !

!ProportionalSplitterMorph methodsFor: 'adjacent splitters' stamp: 'cmm 4/18/2015 13:41'!
splittersBelowDo: aBlock 
	self splitterBelow ifNotNil:
		[ : splitter | aBlock value: splitter.
		splitter splittersBelowDo: aBlock ]! !

!ProportionalSplitterMorph methodsFor: 'adjacent splitters' stamp: 'cmm 4/18/2015 13:41'!
splittersLeftDo: aBlock 
	self splitterLeft ifNotNil:
		[ : splitter | aBlock value: splitter.
		splitter splittersLeftDo: aBlock ]! !

!ProportionalSplitterMorph methodsFor: 'adjacent splitters' stamp: 'cmm 4/18/2015 13:42'!
splittersRightDo: aBlock 
	self splitterRight ifNotNil:
		[ : splitter | aBlock value: splitter.
		splitter splittersRightDo: aBlock ]! !

!ProportionalSplitterMorph methodsFor: 'events' stamp: 'cmm 4/19/2015 21:20'!
step
	splitsTopAndBottom
		ifTrue: [ self reduceTopBottomImbalance ]
		ifFalse:
			[ self reduceLeftRightImbalance abs > 1 ifTrue:
				[ self splittersLeftDo:
					[ : splitter | splitter reduceLeftRightImbalance ].
				self splittersRightDo:
					[ : splitter | splitter reduceLeftRightImbalance ] ] ]! !

!ProportionalSplitterMorph methodsFor: 'events' stamp: 'cmm 2/22/2014 15:23'!
stepTime
	"When a splitter finds itself in the right place, let it rest for about 3 seconds to avoid performance impacts of constant, rapid stepping."
	| pause |
	pause := 3000. "Frozen image when atRandom failed due to lock on its Mutex."
	^ ({#(1 -1 1 ).  #(-1 1 -1 )} includes: self movements asArray)
		ifTrue: [ pause "don't twitch" ]
		ifFalse:
			[ splitsTopAndBottom
				ifTrue:
					[ self topBottomCorrection isZero
						ifTrue: [ pause ]
						ifFalse: [ 0 ] ]
				ifFalse:
					[ self leftRightImbalance abs > 1
						ifTrue: [ ">1 rather than 0 to discourage one-off twitching"
							0 ]
						ifFalse: [ pause ] ] ]! !

!ProportionalSplitterMorph methodsFor: 'layout' stamp: 'cmm 4/27/2015 14:56'!
topBottomCorrection
	^ self top < self topBoundary
		ifTrue: [ self topBoundary - self top ]
		ifFalse:
			[ self bottom > (self bottomBoundary)
				ifTrue: [ self bottomBoundary - self bottom ]
				ifFalse:
					[ | wsAbove wsBelow |
					wsAbove := self canEncroachWhiteSpaceOf: leftOrTop.
					wsBelow := self canEncroachWhiteSpaceOf: rightOrBottom.
					wsAbove
						ifTrue:
							[ wsBelow
								ifTrue:
									[ self splitterBelow
										ifNil: [0]
										ifNotNil: [ : below | below reduceTopBottomImbalance min: 0 ] ]
								ifFalse: [ (self top > self topBoundary) ifTrue: [-2] ifFalse: [0] ] ]
						ifFalse:
							[ wsBelow
								ifTrue: [ (self bottom < self bottomBoundary) ifTrue: [2] ifFalse: [0] ]
								ifFalse:
									[ self splitterBelow
										ifNil: [0]
										ifNotNil: [ : below | below reduceTopBottomImbalance max: 0 ] ] ] ] ]! !

!ProportionalSplitterMorph methodsFor: 'boundaries' stamp: 'mt 4/29/2015 17:01'!
topBoundary
	"Answer the topmost x position the receiver could be moved to."

	| splitter morphs |
	splitter := self splitterAbove.
	morphs := self commonNeighbours: leftOrTop with: splitter.
	
	^ (splitter
		ifNil: [owner isSystemWindow ifTrue: [owner panelRect top]
				ifFalse: [owner innerBounds top]]
		ifNotNil: [splitter bottom])
		+ (self minimumHeightOf: morphs)! !

!ProportionalSplitterMorph methodsFor: 'events' stamp: 'cmm 10/9/2015 14:04'!
updateFromEvent: anEvent 
	| delta |
	lastMouse ifNil: [ lastMouse := anEvent position ].
	delta := splitsTopAndBottom
		ifTrue: [ 0 @ ((self normalizedY: anEvent cursorPoint y) - lastMouse y) ]
		ifFalse: [ (self normalizedX: anEvent cursorPoint x) - lastMouse x @ 0 ].
	lastMouse := splitsTopAndBottom
		ifTrue: [ lastMouse x @ (self normalizedY: anEvent cursorPoint y) ]
		ifFalse: [ (self normalizedX: anEvent cursorPoint x) @ lastMouse y ].
	self repositionBy: delta! !

!ProportionalSplitterMorph methodsFor: 'events' stamp: 'cmm 8/24/2013 23:25'!
wantsSteps
	^ splitsTopAndBottom
		ifTrue: [ self class smartHorizontalSplitters ]
		ifFalse: [ self class smartVerticalSplitters ]! !

!ProportionalSplitterMorph methodsFor: 'adjacent splitters' stamp: 'cmm 5/4/2015 20:10'!
withSiblingSplittersDo: aBlock 
	aBlock value: self.
	self siblingSplittersDo: aBlock! !

Morph subclass: #BorderedMorph
	instanceVariableNames: 'borderWidth borderColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Kernel'!

!BorderedMorph commentStamp: 'kfr 10/27/2003 11:17' prior: 0!
BorderedMorph introduce borders to morph. Borders have the instanceVariables borderWidth and borderColor.
 
BorderedMorph new borderColor: Color red; borderWidth: 10; openInWorld.

BorderedMorph also have a varaity of border styles: simple, inset, raised, complexAltFramed, complexAltInset, complexAltRaised, complexFramed, complexInset, complexRaised.
These styles are set using the classes BorderStyle, SimpleBorder, RaisedBorder, InsetBorder and ComplexBorder.

BorderedMorph new borderStyle: (SimpleBorder width: 1 color: Color white); openInWorld.
BorderedMorph new borderStyle: (BorderStyle inset width: 2); openInWorld.


!

!BorderedMorph methodsFor: 'geometry' stamp: 'ul 11/15/2010 11:17'!
acquireBorderWidth: aBorderWidth
	"Gracefully acquire the new border width, keeping the interior area intact and not seeming to shift"

	| delta |
	(delta := aBorderWidth- self borderWidth) = 0 ifTrue: [^ self].
	self bounds: ((self bounds origin - (delta @ delta)) corner: (self bounds corner + (delta @ delta))).
	self borderWidth: aBorderWidth.
	self layoutChanged! !

!BorderedMorph methodsFor: 'menu' stamp: 'dgd 9/18/2004 19:16'!
addBorderStyleMenuItems: aMenu hand: aHandMorph
	"Add border-style menu items"

	| subMenu |
	subMenu := MenuMorph new defaultTarget: self.
	"subMenu addTitle: 'border' translated."
	subMenu addStayUpItemSpecial.
	subMenu addList: 
		{{'border color...' translated. #changeBorderColor:}.
		{'border width...' translated. #changeBorderWidth:}}.
	subMenu addLine.
	BorderStyle borderStyleChoices do:
		[:sym | (self borderStyleForSymbol: sym)
			ifNotNil:
				[subMenu add: sym translated target: self selector: #setBorderStyle: argument: sym]].
	aMenu add: 'border style' translated subMenu: subMenu
! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'kfr 11/5/2006 21:36'!
addCornerGrips
	self
		addMorphBack: (TopLeftGripMorph new target: self; position: self position).
	self
		addMorphBack: (TopRightGripMorph new target: self; position: self position).
	self
		addMorphBack: (BottomLeftGripMorph new target: self;position: self position).
	self
		addMorphBack: (BottomRightGripMorph new target: self;position: self position)! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'ar 6/23/2010 17:09'!
addEdgeGrips
	"Add resizers along the four edges of the receiver"

	self
		addMorphBack: (TopGripMorph new target: self;position: self position).
	self
		addMorphBack: (BottomGripMorph new target: self;position: self position).
	self
		addMorphBack: (RightGripMorph new target: self;position: self position).
	self
		addMorphBack: (LeftGripMorph new target: self;position: self position).! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 4/29/2015 17:24'!
addMorph: aMorph fullFrame: aLayoutFrame
	"Add aMorph according to aLayoutFrame."

	self cellInset > 0 ifTrue: [
		| left right bottom top spacing |

		spacing := self cellInset.
		left := aLayoutFrame leftOffset ifNil: [0].
		right := aLayoutFrame rightOffset ifNil: [0].

		bottom := aLayoutFrame bottomOffset ifNil: [0].
		top := aLayoutFrame topOffset ifNil: [0].
		
		"Add a spacing to the frame if it is not top or leftmost."
		aLayoutFrame leftFraction = 0
			ifFalse: [left := left + spacing]
			ifTrue: [
				"Expect another, resizable widget besides me if I have a fixed width."
				aLayoutFrame hasFixedWidth ifTrue: [right := right - spacing]].
		aLayoutFrame topFraction = 0
			ifFalse: [top := top + spacing]
			ifTrue: [
				"Expect another, resizable widget besides me if I have a fixed height."
				aLayoutFrame hasFixedHeight ifTrue: [bottom := bottom - spacing]].
			
		aLayoutFrame
			topOffset: top;
			bottomOffset: bottom;
			leftOffset: left;
			rightOffset: right].
	
	super addMorph: aMorph fullFrame: aLayoutFrame.
	
	self wantsPaneSplitters ifTrue: [self addPaneSplitters].! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 4/28/2015 13:35'!
addPaneHSplitterBetween: topMorph and: bottomMorphs

	| targetY minX maxX splitter |
	targetY := topMorph layoutFrame bottomFraction.

	minX := (bottomMorphs detectMin: [:each | each layoutFrame leftFraction]) layoutFrame leftFraction.
	maxX := (bottomMorphs detectMax: [:each | each layoutFrame rightFraction]) layoutFrame rightFraction.
	splitter := ProportionalSplitterMorph new beSplitsTopAndBottom; yourself.
	splitter layoutFrame: (LayoutFrame
		fractions: (minX @ targetY corner: maxX @ targetY)
		offsets: (((topMorph layoutFrame leftOffset ifNil: [0]) @ 0 
			corner: (topMorph layoutFrame rightOffset ifNil: [0]) @ AbstractResizerMorph gripThickness) 
			translateBy: 0 @ (topMorph layoutFrame bottomOffset ifNil: [0]))).

	self addMorphBack: (splitter position: self position).! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 4/28/2015 13:29'!
addPaneHSplitters

	| remaining targetY sameY |
	remaining := self paneMorphs reject: [:each |
		each layoutFrame bottomFraction = 1
			or: [each layoutFrame hasFixedHeight]].
	[remaining notEmpty] whileTrue:
		[targetY := remaining first layoutFrame bottomFraction.
		sameY := self paneMorphs select: [:each | each layoutFrame bottomFraction = targetY].
		self addPaneHSplitterBetween: remaining first and: sameY.
		remaining := remaining copyWithoutAll: sameY]! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 4/28/2015 13:31'!
addPaneSplitters

	self removePaneSplitters.
	
	self addPaneVSplitters.
	self addPaneHSplitters.	

	self linkSubmorphsToSplitters! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 4/24/2015 11:30'!
addPaneVSplitterBetween: leftMorph and: rightMorphs 

	| targetX minY maxY splitter |
	targetX := leftMorph layoutFrame rightFraction.
	minY := (rightMorphs detectMin: [:each | each layoutFrame topFraction]) layoutFrame topFraction.
	maxY := (rightMorphs detectMax: [:each | each layoutFrame bottomFraction]) layoutFrame bottomFraction.
	
	splitter := ProportionalSplitterMorph new.
	splitter layoutFrame: (LayoutFrame
		fractions: (targetX @ minY corner: targetX @ maxY)
		offsets: ((0 @ (leftMorph layoutFrame topOffset ifNil: [0]) corner: (AbstractResizerMorph gripThickness@ (leftMorph layoutFrame bottomOffset ifNil: [0]))) translateBy: (leftMorph layoutFrame rightOffset ifNil: [0]) @ 0)).

	self addMorphBack: (splitter position: self position).! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 4/28/2015 13:29'!
addPaneVSplitters

	| remaining targetX sameX |
	remaining := self paneMorphs reject: [:each |
		each layoutFrame rightFraction = 1
			or: [each layoutFrame hasFixedWidth]].
	[remaining notEmpty] whileTrue:
		[targetX := remaining first layoutFrame rightFraction.
		sameX := self paneMorphs select: [:each | each layoutFrame rightFraction = targetX].
		self addPaneVSplitterBetween: remaining first and: sameX.
		remaining := remaining copyWithoutAll: sameX]! !

!BorderedMorph methodsFor: 'drawing' stamp: 'ar 8/25/2009 11:37'!
areasRemainingToFill: aRectangle
	"Fixed here to test the fillStyle rather than color for translucency.
	Since can have a translucent fillStyle while the (calculated) color is not."
	
	self fillStyle isTranslucent
		ifTrue: [^ Array with: aRectangle].
	self wantsRoundedCorners
		ifTrue: [(self borderWidth > 0
					and: [self borderColor isColor
							and: [self borderColor isTranslucent]])
				ifTrue: [^ aRectangle
						areasOutside: (self innerBounds intersect: self boundsWithinCorners)]
				ifFalse: [^ aRectangle areasOutside: self boundsWithinCorners]]
		ifFalse: [(self borderWidth > 0
					and: [self borderColor isColor
							and: [self borderColor isTranslucent]])
				ifTrue: [^ aRectangle areasOutside: self innerBounds]
				ifFalse: [^ aRectangle areasOutside: self bounds]]
! !

!BorderedMorph methodsFor: 'accessing' stamp: 'sw 8/6/97 14:34'!
borderColor
	^ borderColor! !

!BorderedMorph methodsFor: 'accessing' stamp: 'ar 8/17/2001 16:52'!
borderColor: colorOrSymbolOrNil
	self doesBevels ifFalse:[
		colorOrSymbolOrNil isColor ifFalse:[^self]].
	borderColor = colorOrSymbolOrNil ifFalse: [
		borderColor := colorOrSymbolOrNil.
		self changed].
! !

!BorderedMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 15:53'!
borderInitialize
	"initialize the receiver state related to border"
	borderColor:= self defaultBorderColor.
	borderWidth := self defaultBorderWidth! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 6/20/97 11:24'!
borderInset
	self borderColor: #inset! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 6/20/97 11:25'!
borderRaised
	self borderColor: #raised! !

!BorderedMorph methodsFor: 'accessing' stamp: 'aoy 2/17/2003 01:19'!
borderStyle
	"Work around the borderWidth/borderColor pair"

	| style |
	borderColor ifNil: [^BorderStyle default].
	borderWidth isZero ifTrue: [^BorderStyle default].
	style := self valueOfProperty: #borderStyle ifAbsent: [BorderStyle default].
	(borderWidth = style width and: 
			["Hah!! Try understanding this..."

			borderColor == style style or: 
					["#raised/#inset etc"

					#simple == style style and: [borderColor = style color]]]) 
		ifFalse: 
			[style := borderColor isColor 
				ifTrue: [BorderStyle width: borderWidth color: borderColor]
				ifFalse: [(BorderStyle perform: borderColor) width: borderWidth	"argh."].
			self setProperty: #borderStyle toValue: style].
	^style trackColorFrom: self! !

!BorderedMorph methodsFor: 'accessing' stamp: 'dgd 2/21/2003 22:42'!
borderStyle: aBorderStyle 
	"Work around the borderWidth/borderColor pair"

	aBorderStyle = self borderStyle ifTrue: [^self].
	"secure against invalid border styles"
	(self canDrawBorder: aBorderStyle) 
		ifFalse: 
			["Replace the suggested border with a simple one"

			^self borderStyle: (BorderStyle width: aBorderStyle width
						color: (aBorderStyle trackColorFrom: self) color)].
	aBorderStyle width = self borderStyle width ifFalse: [self changed].
	(aBorderStyle isNil or: [aBorderStyle == BorderStyle default]) 
		ifTrue: 
			[self removeProperty: #borderStyle.
			borderWidth := 0.
			^self changed].
	self setProperty: #borderStyle toValue: aBorderStyle.
	borderWidth := aBorderStyle width.
	borderColor := aBorderStyle style == #simple 
				ifTrue: [aBorderStyle color]
				ifFalse: [aBorderStyle style].
	self changed! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 6/20/97 11:09'!
borderWidth
	^ borderWidth! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 6/4/1999 09:42'!
borderWidth: anInteger
	borderColor ifNil: [borderColor := Color black].
	borderWidth := anInteger max: 0.
	self changed! !

!BorderedMorph methodsFor: 'menu' stamp: 'ar 10/5/2000 18:50'!
changeBorderColor: evt
	| aHand |
	aHand := evt ifNotNil: [evt hand] ifNil: [self primaryHand].
	self changeColorTarget: self selector: #borderColor: originalColor: self borderColor hand: aHand! !

!BorderedMorph methodsFor: 'menu' stamp: 'ul 12/12/2009 14:11'!
changeBorderWidth: evt
	| handle origin aHand newWidth oldWidth |
	aHand := evt ifNil: [self primaryHand] ifNotNil: [evt hand].
	origin := aHand position.
	oldWidth := borderWidth.
	(handle := HandleMorph new)
		forEachPointDo:
			[:newPoint | handle removeAllMorphs.
			handle addMorph:
				(LineMorph from: origin to: newPoint color: Color black width: 1).
			newWidth := (newPoint - origin) r asInteger // 5.
			self borderWidth: newWidth]
		lastPointDo:
			[:newPoint | handle deleteBalloon.
			self halo ifNotNil: [:halo | halo addHandles].
			self rememberCommand:
				(Command new cmdWording: 'border change' translated;
					undoTarget: self selector: #borderWidth: argument: oldWidth;
					redoTarget: self selector: #borderWidth: argument: newWidth)].
	aHand attachMorph: handle.
	handle setProperty: #helpAtCenter toValue: true.
	handle showBalloon:
'Move cursor farther from
this point to increase border width.
Click when done.' translated hand: evt hand.
	handle startStepping! !

!BorderedMorph methodsFor: 'geometry' stamp: 'nk 4/5/2001 14:24'!
closestPointTo: aPoint
	"account for round corners. Still has a couple of glitches at upper left and right corners"
	| pt |
	pt := self bounds pointNearestTo: aPoint.
	self wantsRoundedCorners ifFalse: [ ^pt ].
	self bounds corners with: (self bounds insetBy: 6) corners do: [ :out :in |
		(pt - out) abs < (6@6)
			ifTrue: [ ^(in + (Point r: 5.0 degrees: (pt - in) degrees)) asIntegerPoint ].
	].
	^pt.! !

!BorderedMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:33'!
defaultBorderColor
	"answer the default border color/fill style for the receiver"
	^ Color black! !

!BorderedMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:36'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 2! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 6/20/97 11:19'!
doesBevels
	"To return true means that this object can show bevelled borders, and
	therefore can accept, eg, #raised or #inset as valid borderColors.
	Must be overridden by subclasses that do not support bevelled borders."

	^ true! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 7/19/2016 10:39'!
doFastWindowReframe: ptName

	| newBounds |
	"For fast display, only higlight the rectangle during loop"
	newBounds := self bounds newRectButtonPressedDo: [:f | 
		f 
			withSideOrCorner: ptName
			setToPoint: (self pointFromWorld: Sensor cursorPoint)
			minExtent: self minimumExtent].
	self bounds: newBounds.
	^newBounds.! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 7/14/2016 16:28'!
fastFramingOn

	^ Preferences fastDragWindowForMorphic! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 5/19/2016 07:39'!
grips

	^ self submorphsSatisfying: [:each | each isKindOf: CornerGripMorph]! !

!BorderedMorph methodsFor: 'accessing' stamp: 'di 1/3/1999 12:24'!
hasTranslucentColor
	"Answer true if this any of this morph is translucent but not transparent."

	(color isColor and: [color isTranslucentColor]) ifTrue: [^ true].
	(borderColor isColor and: [borderColor isTranslucentColor]) ifTrue: [^ true].
	^ false
! !

!BorderedMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:07'!
initialize
	"initialize the state of the receiver"
	super initialize.
""
	self borderInitialize! !

!BorderedMorph methodsFor: 'geometry' stamp: 'nk 4/5/2001 14:23'!
intersectionWithLineSegmentFromCenterTo: aPoint
	"account for round corners. Still has a couple of glitches at upper left and right corners"
	| pt |
	pt := super intersectionWithLineSegmentFromCenterTo: aPoint.
	self wantsRoundedCorners ifFalse: [ ^pt ].
	self bounds corners with: (self bounds insetBy: 6) corners do: [ :out :in |
		(pt - out) abs < (6@6)
			ifTrue: [ ^(in + (Point r: 5.0 degrees: (pt - in) degrees)) asIntegerPoint ].
	].
	^pt.! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 4/29/2015 16:58'!
linkSubmorphsToSplitters

	self splitters do: [:splitter |
		splitter splitsTopAndBottom ifTrue: [
			self submorphsDo: [:morph |
					((morph ~= splitter
						and: [morph layoutFrame bottomFraction = splitter layoutFrame topFraction])
						and: [morph layoutFrame hasFixedHeight not
							or: [(morph layoutFrame topOffset ifNil: [0]) < 0]])
								ifTrue: [splitter addLeftOrTop: morph].
					((morph ~= splitter
						and: [morph layoutFrame topFraction = splitter layoutFrame bottomFraction])
						and: [morph layoutFrame hasFixedHeight not
							or: [(morph layoutFrame bottomOffset ifNil: [0]) > 0]])
								ifTrue: [splitter addRightOrBottom: morph]]].
		splitter splitsLeftAndRight ifTrue: [
			self submorphsDo: [:morph |
					((morph ~= splitter
						and: [morph layoutFrame rightFraction = splitter layoutFrame leftFraction])
						and: [morph layoutFrame hasFixedWidth not
							or: [(morph layoutFrame leftOffset ifNil: [0]) < 0]])
								ifTrue: [splitter addLeftOrTop: morph].
					((morph ~= splitter 
						and: [morph layoutFrame leftFraction = splitter layoutFrame rightFraction])
						and: [morph layoutFrame hasFixedWidth not
							or: [(morph layoutFrame rightOffset ifNil: [0]) > 0]])
								ifTrue: [splitter addRightOrBottom: morph]]]]! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 4/28/2015 14:06'!
paneMorphs

	^ self submorphs copyWithoutAll: self splitters! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'jrp 7/23/2005 00:03'!
removeCornerGrips

	| corners |
	corners := self submorphsSatisfying: [:each | each isKindOf: CornerGripMorph].
	corners do: [:each | each delete]! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'jrp 7/23/2005 22:28'!
removePaneSplitters

	self splitters do: [:each | each delete]! !

!BorderedMorph methodsFor: 'private' stamp: 'di 6/20/97 11:21'!
setBorderWidth: w borderColor: bc
	self borderWidth: w.
	self borderColor: bc.! !

!BorderedMorph methodsFor: 'private' stamp: 'di 6/20/97 11:22'!
setColor: c borderWidth: w borderColor: bc
	self color: c.
	self borderWidth: w.
	self borderColor: bc.! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'jrp 7/23/2005 22:16'!
splitters

	^ self submorphsSatisfying: [:each | each isKindOf: ProportionalSplitterMorph]! !

!BorderedMorph methodsFor: 'accessing' stamp: 'sw 11/29/1999 17:36'!
useRoundedCorners
	self cornerStyle: #rounded! !

!BorderedMorph methodsFor: 'accessing' stamp: 'sw 11/29/1999 17:37'!
useSquareCorners
	self cornerStyle: #square! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 4/28/2015 14:11'!
wantsPaneSplitters

	^ self valueOfProperty: #allowPaneSplitters ifAbsent: [false]! !

!BorderedMorph methodsFor: 'resize handling' stamp: 'mt 4/28/2015 14:52'!
wantsPaneSplitters: aBoolean

	self setProperty: #allowPaneSplitters toValue: aBoolean.! !

BorderedMorph subclass: #BorderedSubpaneDividerMorph
	instanceVariableNames: 'resizingEdge'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!BorderedSubpaneDividerMorph class methodsFor: 'as yet unclassified' stamp: 'ar 8/15/2001 23:24'!
forBottomEdge
	^self new horizontal resizingEdge: #bottom! !

!BorderedSubpaneDividerMorph class methodsFor: 'as yet unclassified' stamp: 'ar 8/15/2001 23:25'!
forTopEdge
	^self new horizontal resizingEdge: #top! !

!BorderedSubpaneDividerMorph class methodsFor: 'as yet unclassified' stamp: 'ar 8/15/2001 23:25'!
horizontal
	^self new horizontal! !

!BorderedSubpaneDividerMorph class methodsFor: 'as yet unclassified' stamp: 'ar 8/15/2001 23:25'!
vertical
	^self new vertical! !

!BorderedSubpaneDividerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:09'!
defaultBorderWidth
"answer the default border width for the receiver"
	^ 0! !

!BorderedSubpaneDividerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:09'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color black! !

!BorderedSubpaneDividerMorph methodsFor: 'as yet unclassified' stamp: 'ar 8/15/2001 23:24'!
firstEnter: evt
	"The first time this divider is activated, find its window and redirect further interaction there."
	| window |

	window := self firstOwnerSuchThat: [:m | m respondsTo: #secondaryPaneTransition:divider:].
	window ifNil: [ self suspendEventHandler. ^ self ]. "not working out"
	window secondaryPaneTransition: evt divider: self.
	self on: #mouseEnter send: #secondaryPaneTransition:divider: to: window.
! !

!BorderedSubpaneDividerMorph methodsFor: 'as yet unclassified' stamp: 'ar 8/15/2001 23:24'!
horizontal

	self hResizing: #spaceFill.! !

!BorderedSubpaneDividerMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:09'!
initialize
	"initialize the state of the receiver"
	super initialize.
""
	self extent: 1 @ 1! !

!BorderedSubpaneDividerMorph methodsFor: 'as yet unclassified' stamp: 'ar 8/15/2001 23:24'!
resizingEdge

	^resizingEdge
! !

!BorderedSubpaneDividerMorph methodsFor: 'as yet unclassified' stamp: 'ar 8/15/2001 23:24'!
resizingEdge: edgeSymbol

	(#(top bottom) includes: edgeSymbol) ifFalse:
		[ self error: 'resizingEdge must be #top or #bottom' ].
	resizingEdge := edgeSymbol.
	self on: #mouseEnter send: #firstEnter: to: self.
! !

!BorderedSubpaneDividerMorph methodsFor: 'as yet unclassified' stamp: 'ar 8/15/2001 23:24'!
vertical

	self vResizing: #spaceFill.! !

BorderedMorph subclass: #DoCommandOnceMorph
	instanceVariableNames: 'target command actionBlock innerArea'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!DoCommandOnceMorph commentStamp: '<historical>' prior: 0!
I am used to execute a once-only command. My first use was in loading/saving the current project. In such cases it is necessary to be in another project to do the actual work. So an instance of me is added to a new world/project and that project is entered. I do my stuff (save/load followed by a re-enter of the previous project) and everyone is happy.!

!DoCommandOnceMorph methodsFor: 'as yet unclassified' stamp: 'RAA 7/7/2000 11:49'!
actionBlock: aBlock

	actionBlock := aBlock! !

!DoCommandOnceMorph methodsFor: 'as yet unclassified' stamp: 'RAA 7/29/2000 17:30'!
addText: aString

	| t |
	t := TextMorph new 
		beAllFont: (TextStyle default fontOfSize: 26);
		contents: aString.
	self extent: t extent * 3.
	innerArea := Morph new 
		color: Color white; 
		extent: self extent - (16@16);
		position: self position + (8@8);
		lock.
	self addMorph: innerArea. 
	self addMorph: (t position: self position + t extent; lock).! !

!DoCommandOnceMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:34'!
defaultBorderColor
	"answer the default border color/fill style for the receiver"
	^ Color blue! !

!DoCommandOnceMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:36'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 8! !

!DoCommandOnceMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:08'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	
	self useRoundedCorners! !

!DoCommandOnceMorph methodsFor: 'initialization' stamp: 'RAA 7/7/2000 12:19'!
openInWorld: aWorld

	self position: aWorld topLeft + (aWorld extent - self extent // 2).
	super openInWorld: aWorld! !

!DoCommandOnceMorph methodsFor: 'stepping and presenter' stamp: 'ul 8/2/2011 19:59'!
step

	| goForIt |

	actionBlock ifNil: [^self stopStepping].
	goForIt := actionBlock.
	actionBlock := nil.
	goForIt
		on: ProgressTargetRequestNotification
		do: [ :ex | ex resume: innerArea].		"in case a save/load progress display needs a home"
! !

!DoCommandOnceMorph methodsFor: 'testing' stamp: 'RAA 7/7/2000 11:46'!
stepTime

	^1
! !

!DoCommandOnceMorph methodsFor: 'testing' stamp: 'RAA 7/7/2000 11:50'!
wantsSteps

	^actionBlock notNil
! !

BorderedMorph subclass: #EllipseMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!EllipseMorph commentStamp: 'kfr 10/27/2003 10:32' prior: 0!
A round BorderedMorph. Supports borderWidth and borderColor. 
Only simple borderStyle is implemented.

EllipseMorph new borderWidth:10; borderColor: Color green; openInWorld.
EllipseMorph new borderStyle:(SimpleBorder width: 5 color: Color blue); openInWorld.!

EllipseMorph subclass: #CircleMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!CircleMorph commentStamp: '<historical>' prior: 0!
I am a specialization of EllipseMorph that knows enough to remain circular.
!

!CircleMorph class methodsFor: 'as yet unclassified' stamp: 'nk 8/23/2004 11:45'!
newPin
	"Construct a pin for embedded attachment"
	"CircleMorph newPin openInHand"
	^self new
		removeAllMorphs;
		extent: 18@18;
		hResizing: #rigid;
		vResizing: #rigid;
		layoutPolicy: nil;
		color: Color orange lighter;
		borderColor: Color orange darker;
		borderWidth: 2;
		wantsConnectionWhenEmbedded: true;
		name: 'Pin'! !

!CircleMorph class methodsFor: 'as yet unclassified' stamp: 'tak 7/25/2007 00:04'!
supplementaryPartsDescriptions
	"Extra items for parts bins"

	^ {DescriptionForPartsBin
		formalName: 'Circle' translatedNoop
		categoryList: {'Graphics' translatedNoop}
		documentation: 'A circular shape' translatedNoop
		globalReceiverSymbol: #CircleMorph 
		nativitySelector: #newStandAlone.

	DescriptionForPartsBin
		formalName: 'Pin' translatedNoop
		categoryList: {'Connectors' translatedNoop}
		documentation: 'An attachment point for Connectors that you can embed in another Morph.' translatedNoop
		globalReceiverSymbol: #NCPinMorph 
		nativitySelector: #newPin.
}! !

!CircleMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/1/2002 07:04'!
addFlexShellIfNecessary
	"When scaling or rotating from a halo, I can do this without a flex shell"

	^ self
! !

!CircleMorph methodsFor: 'geometry' stamp: 'nk 7/1/2002 07:01'!
bounds: aRectangle
	| size |
	size := aRectangle width min: aRectangle height.
	super bounds: (Rectangle origin: aRectangle origin extent: size @ size).! !

!CircleMorph methodsFor: 'geometry' stamp: 'nk 7/1/2002 16:39'!
extent: aPoint
	| size oldRotationCenter |
	oldRotationCenter := self rotationCenter.
	size := aPoint x min: aPoint y.
	super extent: size @ size.
	self rotationCenter: oldRotationCenter.! !

!CircleMorph methodsFor: 'geometry eToy' stamp: 'nk 7/1/2002 07:12'!
heading: newHeading
	"Set the receiver's heading (in eToy terms).
	Note that circles never use flex shells."
	self rotationDegrees: newHeading.! !

!CircleMorph methodsFor: 'parts bin' stamp: 'nice 5/6/2011 00:12'!
initialize
	super initialize.
	self extent: 40@40;
		color: Color green lighter! !

!CircleMorph methodsFor: 'parts bin' stamp: 'nk 7/1/2002 16:42'!
initializeToStandAlone
	^super initializeToStandAlone
		extent: 40@40;
		color: Color green lighter;
		yourself! !

!CircleMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/1/2002 16:29'!
privateMoveBy: delta
	self setProperty: #referencePosition toValue: self referencePosition + delta.
	self setProperty: #originalCenter toValue: (self valueOfProperty: #originalCenter ifAbsent: [ self center ]) + delta.
	super privateMoveBy: delta.
! !

!CircleMorph methodsFor: 'geometry eToy' stamp: 'nk 7/1/2002 07:31'!
referencePosition 
	"Return the current reference position of the receiver"
	^ self valueOfProperty: #referencePosition ifAbsent: [ self center ]
! !

!CircleMorph methodsFor: 'geometry eToy' stamp: 'nk 7/1/2002 11:16'!
rotationCenter
	"Return the rotation center of the receiver. The rotation center defines the relative offset inside the receiver's bounds for locating the reference position."
	| refPos |
	refPos := self referencePosition.
	^ (refPos - self bounds origin) / self bounds extent asFloatPoint! !

!CircleMorph methodsFor: 'geometry eToy' stamp: 'nk 7/1/2002 13:48'!
rotationCenter: aPointOrNil
	"Set the new rotation center of the receiver. The rotation center defines the relative offset inside the receiver's bounds for locating the reference position."
	| newRef box |
	aPointOrNil isNil
		ifTrue: [self removeProperty: #referencePosition.
			self removeProperty: #originalCenter.
			self removeProperty: #originalAngle. ]
		ifFalse: [ box := self bounds.
				newRef := box origin + (aPointOrNil * box extent).
				self setRotationCenterFrom: newRef ].

! !

!CircleMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/1/2002 07:28'!
rotationDegrees

	^ self forwardDirection! !

!CircleMorph methodsFor: 'rotate scale and flex' stamp: 'nk 7/1/2002 15:52'!
rotationDegrees: degrees
	| ref newPos flex origAngle origCenter |
	ref := self referencePosition.
	origAngle := self valueOfProperty: #originalAngle ifAbsentPut: [ self heading ].
	origCenter := self valueOfProperty: #originalCenter ifAbsentPut: [ self center ].
	flex := (MorphicTransform offset: ref negated)
			withAngle: (degrees - origAngle) degreesToRadians.
	newPos := (flex transform: origCenter) - flex offset.
	self position: (self position + newPos - self center) asIntegerPoint.
	self setProperty: #referencePosition toValue: ref.
	self setProperty: #originalAngle toValue: origAngle.
	self setProperty: #originalCenter toValue: origCenter.
	self forwardDirection: degrees.
	self changed.
! !

!CircleMorph methodsFor: 'menus' stamp: 'nk 7/1/2002 11:30'!
setRotationCenterFrom: aPoint
	"Called by halo rotation code.
	Circles store their referencePosition."
	self setProperty: #referencePosition toValue: aPoint.
	self setProperty: #originalCenter toValue: self center.
	self setProperty: #originalAngle toValue: self heading.! !

!CircleMorph methodsFor: 'geometry' stamp: 'nk 7/1/2002 08:49'!
transformedBy: aTransform
	aTransform isIdentity ifTrue:[^self].
	^self center: (aTransform localPointToGlobal: self center).
! !

!EllipseMorph class methodsFor: 'parts bin' stamp: 'tak 7/25/2007 00:04'!
descriptionForPartsBin
	^ self partName:	'Ellipse' translatedNoop
		categories:		{'Graphics' translatedNoop. 'Basic' translatedNoop}
		documentation:	'An elliptical or circular shape' translatedNoop! !

!EllipseMorph methodsFor: 'drawing' stamp: 'di 6/24/1998 14:27'!
areasRemainingToFill: aRectangle
	"Could be improved by quick check of inner rectangle"

	^ Array with: aRectangle! !

!EllipseMorph methodsFor: 'geometry' stamp: 'nk 8/31/2004 14:25'!
bottomLeftCorner
	^self intersectionWithLineSegmentFromCenterTo: bounds bottomLeft 
! !

!EllipseMorph methodsFor: 'geometry' stamp: 'nk 8/31/2004 14:23'!
bottomRightCorner
	^self intersectionWithLineSegmentFromCenterTo: bounds bottomRight 
! !

!EllipseMorph methodsFor: 'testing' stamp: 'ar 8/25/2001 19:08'!
canDrawBorder: aBorderStyle
	^aBorderStyle style == #simple! !

!EllipseMorph methodsFor: 'visual properties' stamp: 'ar 6/25/1999 11:14'!
canHaveFillStyles
	"Return true if the receiver can have general fill styles; not just colors.
	This method is for gradually converting old morphs."
	^true! !

!EllipseMorph methodsFor: 'geometry' stamp: 'nk 2/15/2001 16:08'!
closestPointTo: aPoint
	^self intersectionWithLineSegmentFromCenterTo: aPoint! !

!EllipseMorph methodsFor: 'geometry testing' stamp: 'di 11/14/97 13:50'!
containsPoint: aPoint

	| radius other delta xOverY |
	(bounds containsPoint: aPoint) ifFalse: [^ false].  "quick elimination"
	(bounds width = 1 or: [bounds height = 1])
		ifTrue: [^ true].  "Degenerate case -- code below fails by a bit"

	radius := bounds height asFloat / 2.
	other := bounds width asFloat / 2.
	delta := aPoint - bounds topLeft - (other@radius).
	xOverY := bounds width asFloat / bounds height asFloat.
	^ (delta x asFloat / xOverY) squared + delta y squared <= radius squared! !

!EllipseMorph methodsFor: 'rounding' stamp: 'ka 12/4/2005 00:52'!
cornerStyle: aSymbol
	"Set the receiver's corner style.  But, in this case, do *not*"

	self removeProperty: #cornerStyle.
	self changed! !

!EllipseMorph methodsFor: 'accessing' stamp: 'sw 11/24/1999 14:59'!
couldHaveRoundedCorners
	^ false! !

!EllipseMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:37'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 1! !

!EllipseMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:26'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color yellow! !

!EllipseMorph methodsFor: 'accessing' stamp: 'di 6/20/97 11:29'!
doesBevels
	^ false! !

!EllipseMorph methodsFor: 'drawing' stamp: 'di 5/25/2001 01:37'!
drawOn: aCanvas 

	aCanvas isShadowDrawing
		ifTrue: [^ aCanvas fillOval: bounds fillStyle: self fillStyle borderWidth: 0 borderColor: nil].
	aCanvas fillOval: bounds fillStyle: self fillStyle borderWidth: borderWidth borderColor: borderColor.
! !

!EllipseMorph methodsFor: 'geometry' stamp: 'nk 2/13/2001 18:16'!
intersectionWithLineSegmentFromCenterTo: aPoint 
	| dx aSquared bSquared m mSquared xSquared x y dy |
	(self containsPoint: aPoint)
		ifTrue: [ ^aPoint ].
	dx := aPoint x - self center x.
	dy := aPoint y - self center y.
	dx = 0
		ifTrue: [ ^self bounds pointNearestTo: aPoint ].
	m := dy / dx.
	mSquared := m squared.
	aSquared := (self bounds width / 2) squared.
	bSquared := (self bounds height / 2) squared.
	xSquared := 1 / ((1 / aSquared) + (mSquared / bSquared)).
	x := xSquared sqrt.
	dx < 0 ifTrue: [ x := x negated ].
	y := m * x.
	^ self center + (x @ y) asIntegerPoint.
! !

!EllipseMorph methodsFor: 'geometry' stamp: 'nk 8/31/2004 14:23'!
topLeftCorner
	^self intersectionWithLineSegmentFromCenterTo: bounds topLeft 
! !

!EllipseMorph methodsFor: 'geometry' stamp: 'nk 8/31/2004 14:26'!
topRightCorner
	^self intersectionWithLineSegmentFromCenterTo: bounds topRight
! !

EllipseMorph subclass: #HandleMorph
	instanceVariableNames: 'pointBlock lastPointBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!HandleMorph commentStamp: '<historical>' prior: 0!
A HandleMorph provides mouse-up control behavior.!

!HandleMorph methodsFor: 'initialize' stamp: 'di 11/3/97 16:34'!
forEachPointDo: aBlock
	pointBlock := aBlock! !

!HandleMorph methodsFor: 'initialize' stamp: 'di 8/30/2000 21:48'!
forEachPointDo: aBlock lastPointDo: otherBlock
	pointBlock := aBlock.
	lastPointBlock := otherBlock! !

!HandleMorph methodsFor: 'initialization' stamp: 'kfr 10/13/2007 13:57'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	self extent: 16 @ 16.
	! !

!HandleMorph methodsFor: 'dropping/grabbing' stamp: 'dgd 9/10/2004 13:40'!
justDroppedInto: aMorph event: anEvent
	"So that when the hand drops me (into the world) I go away"
	self removeHalo.
	lastPointBlock ifNotNil: [lastPointBlock value: self center].
	self flag: #arNote. "Probably unnecessary"
	anEvent hand releaseKeyboardFocus: self.
	self changed.
	self delete.
! !

!HandleMorph methodsFor: 'event handling' stamp: 'ar 9/15/2000 23:30'!
keyStroke: evt
	"Check for cursor keys"
	| keyValue |
	owner isHandMorph ifFalse:[^self].
	keyValue := evt keyValue.
	keyValue = 28 ifTrue:[^self position: self position - (1@0)].
	keyValue = 29 ifTrue:[^self position: self position + (1@0)].
	keyValue = 30 ifTrue:[^self position: self position - (0@1)].
	keyValue = 31 ifTrue:[^self position: self position + (0@1)].
	"Special case for return"
	keyValue = 13 ifTrue:[
		"Drop the receiver and be done"
	self flag: #arNote. "Probably unnecessary"
		owner releaseKeyboardFocus: self.
		self delete].
! !

!HandleMorph methodsFor: 'stepping and presenter' stamp: 'ar 9/15/2000 23:24'!
startStepping
	"Make the receiver the keyboard focus for editing"
	super startStepping.
	"owner isHandMorph ifTrue:[owner newKeyboardFocus: self]."
self flag: #arNote. "make me #handleKeyboard:"! !

!HandleMorph methodsFor: 'stepping and presenter' stamp: 'di 11/3/97 16:34'!
step
	pointBlock value: self center! !

!HandleMorph methodsFor: 'testing' stamp: 'JMM 10/21/2003 18:15'!
stepTime
	"Update every hundredth of a second."
	^ 10
! !

HandleMorph subclass: #NewHandleMorph
	instanceVariableNames: 'hand offset waitingForClickInside'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!NewHandleMorph class methodsFor: 'new-morph participation' stamp: 'di 5/3/1998 10:08'!
includeInNewMorphMenu
	^ false! !

!NewHandleMorph methodsFor: 'submorphs-add/remove' stamp: 'ar 8/16/2001 15:38'!
delete
	hand ifNotNil:[
		hand showTemporaryCursor: nil.
	].
	super delete.! !

!NewHandleMorph methodsFor: 'all' stamp: 'di 5/18/1998 15:27'!
followHand: aHand forEachPointDo: block1 lastPointDo: block2
	hand := aHand.
	pointBlock := block1.
	lastPointBlock := block2.
	self position: hand lastEvent cursorPoint - (self extent // 2)! !

!NewHandleMorph methodsFor: 'all' stamp: 'ar 8/16/2001 15:48'!
followHand: aHand forEachPointDo: block1 lastPointDo: block2 withCursor: aCursor
	hand := aHand.
	hand showTemporaryCursor: aCursor "hotSpotOffset: aCursor offset negated".
	borderWidth := 0.
	color := Color transparent.
	pointBlock := block1.
	lastPointBlock := block2.
	self position: hand lastEvent cursorPoint - (self extent // 2)! !

!NewHandleMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:29'!
initialize
"initialize the state of the receiver"

	super initialize.
""
	waitingForClickInside := true.
	Preferences noviceMode
		ifTrue: [self setBalloonText: 'stretch']! !

!NewHandleMorph methodsFor: 'dropping/grabbing' stamp: 'di 4/30/1999 14:06'!
justDroppedInto: aMorph event: anEvent
	"No dropping behavior because stepping will delete me.
	Moreover it needs to be done that way to evaluate lastPointBlock"
! !

!NewHandleMorph methodsFor: 'WiW support' stamp: 'RAA 1/10/2001 10:15'!
morphicLayerNumber

	^1		"handles are very front-like - e.g. the spawn reframe logic actually asks if the first submorph of the world is one of us before deciding to create one"! !

!NewHandleMorph methodsFor: 'all' stamp: 'RAA 4/19/2001 11:36'!
sensorMode

	"If our client is still addressing the Sensor directly, we need to do so as well"
	^self valueOfProperty: #sensorMode ifAbsent: [false].
! !

!NewHandleMorph methodsFor: 'all' stamp: 'RAA 4/19/2001 11:36'!
sensorMode: aBoolean

	"If our client is still addressing the Sensor directly, we need to do so as well"
	self setProperty: #sensorMode toValue: aBoolean.
! !

!NewHandleMorph methodsFor: 'stepping and presenter' stamp: 'RAA 4/19/2001 11:37'!
step
	| eventSource |

	eventSource := self sensorMode ifTrue: [
		Sensor
	] ifFalse: [
		hand lastEvent
	].
	eventSource anyButtonPressed
		ifTrue: [waitingForClickInside := false.
				self position: eventSource cursorPoint - (self extent // 2).
				pointBlock value: self center]
		ifFalse: [waitingForClickInside
					ifTrue: [(self containsPoint: eventSource cursorPoint)
								ifFalse: ["mouse wandered out before clicked"
										^ self delete]]
					ifFalse: [lastPointBlock value: self center.
							^ self delete]]! !

!NewHandleMorph methodsFor: 'dropping/grabbing' stamp: 'ar 10/5/2000 18:16'!
undoGrabCommand
	^nil! !

BorderedMorph subclass: #MorphicModel
	instanceVariableNames: 'model slotName open'
	classVariableNames: 'TimeOfError'
	poolDictionaries: ''
	category: 'Morphic-Kernel'!

MorphicModel class
	instanceVariableNames: 'prototype'!

!MorphicModel commentStamp: '<historical>' prior: 0!
MorphicModels are used to represent structures with state and behavior as well as graphical structure.  A morphicModel is usually the root of a morphic tree depicting its appearance.  The tree is constructed concretely by adding its consituent morphs to a world.

When a part is named in a world, it is given a new slot in the model.  When a part is sensitized, it is named, and a set of mouse-driven methods is also generated in the model.  These may be edited to induce particular behavior.  When a variable is added through the morphic world, it is given a slot in the model, along with a set of access methods.

In addition for public variables (and this is the default for now), methods are generated and called in any outer model in which this model gets embedded, thus propagating variable changes outward.!

MorphicModel class
	instanceVariableNames: 'prototype'!

MorphicModel subclass: #ColorPresenterMorph
	instanceVariableNames: 'contentMorph labelMorph solidLabelMorph getColorSelector'
	classVariableNames: 'HatchForm'
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!ColorPresenterMorph commentStamp: 'gvc 5/18/2007 13:38' prior: 0!
Displays a colour with alpha against a white, hatched and black background.!

!ColorPresenterMorph class methodsFor: 'graphics constants' stamp: 'gvc 9/18/2006 16:05'!
hatchForm
	"Answer a form showing a grid hatch pattern."

	^HatchForm ifNil: [HatchForm := self newHatchForm]! !

!ColorPresenterMorph class methodsFor: 'graphics constants' stamp: 'gvc 9/18/2006 16:24'!
newHatchForm
	"Answer a new hatch form."
	
	^(Form
	extent: 8@8
	depth: 1
	fromArray: #( 4026531840 4026531840 4026531840 4026531840 251658240 251658240 251658240 251658240)
	offset: 0@0)! !

!ColorPresenterMorph class methodsFor: 'instance creation' stamp: 'StephaneDucasse 3/3/2010 16:12'!
on: anObject color: getSel
	"Answer a new instance of the receiver on the given model using
	the given selectors as the interface."
	
	"(ColorPresenterMorph on: (BorderedMorph new) color: #color) openInWorld"
	
	^self new
		on: anObject 
		color: getSel! !

!ColorPresenterMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 10:46'!
contentMorph
	"The outer, containing Morph."
	^ contentMorph! !

!ColorPresenterMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 10:47'!
contentMorph: aMorph
	"The outer, containing Morph."
	contentMorph := aMorph! !

!ColorPresenterMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 10:32'!
getColorSelector
	"The selector symbol used to retrieve the color from my model."
	^ getColorSelector! !

!ColorPresenterMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 10:33'!
getColorSelector: aSymbol
	"The selector symbol used to retrieve the color from my model."
	getColorSelector := aSymbol! !

!ColorPresenterMorph methodsFor: 'initializing' stamp: 'gvc 9/22/2006 09:34'!
initialize
	"Initialize the receiver."

	super initialize.
	self
		borderWidth: 0;
		changeTableLayout;
		labelMorph: self newLabelMorph;
		solidLabelMorph: self newLabelMorph;
		contentMorph: self newContentMorph;
		addMorphBack: self contentMorph! !

!ColorPresenterMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 11:26'!
labelMorph
	"The morph that renders the actual color being presented."
	^ labelMorph! !

!ColorPresenterMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 11:27'!
labelMorph: aMorph
	"The morph that renders the actual color being presented."
	labelMorph := aMorph! !

!ColorPresenterMorph methodsFor: 'initializing' stamp: 'gvc 9/22/2006 09:23'!
newContentMorph
	"Answer a new content morph"

	^Morph new
		color: Color transparent;
		changeTableLayout;
		borderStyle: (BorderStyle inset width: 1);
		vResizing: #spaceFill;
		hResizing: #spaceFill;
		addMorph: self newHatchMorph;
		yourself! !

!ColorPresenterMorph methodsFor: 'initializing' stamp: 'cmm 12/3/2010 14:39'!
newHatchMorph
	"Answer a new morph showing the three backgrounds; white, hatch pattern, and black, against which my labelMorph is displayed."
	^ Morph new
		 color: Color transparent ;
		 changeProportionalLayout ;
		 vResizing: #spaceFill ;
		 hResizing: #spaceFill ;
		 minWidth: 48 ;
		 minHeight: 12 ;
		
		addMorph: (Morph new color: Color white)
		fullFrame: (LayoutFrame fractions: (0 @ 0 corner: 0.3 @ 1)) ;
		
		addMorph: (Morph new fillStyle: (InfiniteForm with: self class hatchForm))
		fullFrame: (LayoutFrame fractions: (0.3 @ 0 corner: 0.7 @ 1)) ;
		
		addMorph: self solidLabelMorph
		fullFrame: (LayoutFrame fractions: (0.7 @ 0 corner: 1 @ 1)) ;
		
		addMorph: self labelMorph
		fullFrame: (LayoutFrame fractions: (0 @ 0 corner: 1 @ 1))! !

!ColorPresenterMorph methodsFor: 'initializing' stamp: 'gvc 9/18/2006 16:20'!
newLabelMorph
	"Answer a new label morph"

	^Morph new! !

!ColorPresenterMorph methodsFor: 'initializing' stamp: 'gvc 9/22/2006 09:45'!
on: anObject color: getColSel
	"Set the receiver to the given model parameterized by the given message selectors."

	self
		model: anObject;
		getColorSelector: getColSel;
		updateColor! !

!ColorPresenterMorph methodsFor: 'initializing' stamp: 'gvc 6/3/2009 18:14'!
setColor: aColor
	"Update the colour of the labels."

	self labelMorph color: aColor.
	self solidLabelMorph color: aColor asNontranslucentColor! !

!ColorPresenterMorph methodsFor: 'accessing' stamp: 'gvc 9/18/2006 16:17'!
solidLabelMorph
	"Answer the value of solidLabelMorph"

	^ solidLabelMorph! !

!ColorPresenterMorph methodsFor: 'accessing' stamp: 'gvc 9/18/2006 16:17'!
solidLabelMorph: anObject
	"Set the value of solidLabelMorph"

	solidLabelMorph := anObject! !

!ColorPresenterMorph methodsFor: 'initializing' stamp: 'gvc 7/30/2009 18:36'!
updateColor
	"Update the color state."

	|col|
	self getColorSelector ifNotNil: [
		col := (self model perform: self getColorSelector) ifNil: [Color transparent].
		self setColor: col]! !

!ColorPresenterMorph methodsFor: 'initializing' stamp: 'gvc 9/22/2006 09:34'!
update: aSymbol 
	"Refer to the comment in View|update:."

	aSymbol == self getColorSelector ifTrue: 
		[self updateColor.
		^ self]! !

!MorphicModel class methodsFor: 'compiling' stamp: 'sw 5/13/1998 14:33'!
acceptsLoggingOfCompilation
	"Dont log sources for my automatically-generated subclasses.  Can easily switch this back when it comes to deal with Versions, etc."

	^ self == MorphicModel or: [(name last isDigit) not]! !

!MorphicModel class methodsFor: 'compilation' stamp: 'tk 3/10/98 18:03'!
categoryForSubclasses
	^ 'Morphic-Models'! !

!MorphicModel class methodsFor: 'compilation' stamp: 'ar 8/6/2009 20:43'!
chooseNewName
	"Choose a new name for the receiver, persisting until an acceptable name is provided or until the existing name is resubmitted"

	| oldName newName |
	oldName := self name.
		[newName := (UIManager default request: 'Please give this Model a name'
					initialAnswer: oldName) asSymbol.
		newName = oldName ifTrue: [^ self].
		Smalltalk includesKey: newName]
		whileTrue:
		[self inform: 'Sorry, that name is already in use.'].
	self rename: newName.! !

!MorphicModel class methodsFor: 'compilation' stamp: ''!
compileAccessorsFor: varName
	self compile: (
'&var
	"Return the value of &var"
	^ &var'
			copyReplaceAll: '&var' with: varName)
		classified: 'public access' notifying: nil.
	self compile: (
'&varPut: newValue
	"Assign newValue to &var.
	Add code below to update related graphics appropriately..."

	&var := newValue.'
			copyReplaceAll: '&var' with: varName)
		classified: 'public access' notifying: nil.
	self compile: (
'&var: newValue
	"Assigns newValue to &var and updates owner"
	&var := newValue.
	self propagate: &var as: ''&var:'''
			copyReplaceAll: '&var' with: varName)
		classified: 'private - propagation' notifying: nil.
! !

!MorphicModel class methodsFor: 'compilation' stamp: ''!
compilePropagationForVarName: varName slotName: slotName
	self compile: ((
'&slot&var: newValue
	"The value of &var in &slot has changed to newValue.
	This value can be read elsewhere in code with
		&slot &var
	and it can be stored into with
		&slot &varPut: someValue"

	"Add code for appropriate response here..."'
			copyReplaceAll: '&var' with: varName)
			copyReplaceAll: '&slot' with: slotName)
		classified: 'input events' notifying: nil.
! !

!MorphicModel class methodsFor: 'queries' stamp: ''!
hasPrototype
	"Return true if there is a prototype for this morph."

	^ prototype ~~ nil
! !

!MorphicModel class methodsFor: 'new-morph participation' stamp: 'di 2/21/98 11:01'!
includeInNewMorphMenu
	"Only include Models that are appropriate"
	^ false! !

!MorphicModel class methodsFor: 'instance creation' stamp: 'tk 8/13/1998 12:58'!
new
	"Return a copy of the prototype, if there is one.
	Otherwise create a new instance normally."

	self hasPrototype ifTrue: [^ prototype veryDeepCopy].
	^ super new
! !

!MorphicModel class methodsFor: 'instance creation' stamp: 'di 6/22/97 09:27'!
newBounds: bounds model: thang slotName: nameOfThisPart
	^ (super new model: thang slotName: nameOfThisPart)
		newBounds: bounds! !

!MorphicModel class methodsFor: 'testing' stamp: 'tk 3/15/98 20:13'!
officialClass
	"We want to make a new instance of the receiver, which is a subclass of MorphicModel.  Answer who to make a new subclass of.  Also used to tell if a given class is a UniClass, existing only for its single instance."

	^ self name last isDigit ifTrue: [MorphicModel] ifFalse: [self]
		"MorphicModel7 can not have subclasses, but Slider and SystemWindow may"! !

!MorphicModel class methodsFor: 'prototype access' stamp: ''!
prototype
	"Return the prototype for this morph."

	^ prototype
! !

!MorphicModel class methodsFor: 'prototype access' stamp: 'gm 2/22/2003 19:13'!
prototype: aMorph
	"Store a copy of the given morph as a prototype to be copied to make new instances."

	aMorph ifNil: [prototype := nil. ^ self].

	prototype := aMorph veryDeepCopy.
	(prototype isMorphicModel) ifTrue: 
		[prototype model: nil slotName: nil].
! !

!MorphicModel class methodsFor: 'housekeeping' stamp: 'nice 12/27/2009 03:11'!
removeUninstantiatedModels
	"With the user's permission, remove the classes of any models that have neither instances nor subclasses."
	"MorphicModel removeUninstantiatedModels"

	| candidatesForRemoval |
	Smalltalk garbageCollect.
	candidatesForRemoval :=
		MorphicModel subclasses select: [:c |
			(c instanceCount = 0) and: [c subclasses size = 0]].
	candidatesForRemoval do: [:c | | ok |
		ok := self confirm: 'Are you certain that you
want to delete the class ', c name, '?'.
		ok ifTrue: [c removeFromSystem]].
! !

!MorphicModel class methodsFor: 'compiling' stamp: 'sw 8/4/97 17:16'!
wantsChangeSetLogging
	"Log changes for MorphicModel itself and for things like PlayWithMe2, but not for automatically-created subclasses like MorphicModel1, MorphicModel2, etc."

	^ self == MorphicModel or:
		[(self class name beginsWith: 'Morphic') not]! !

!MorphicModel methodsFor: 'menu' stamp: 'dgd 8/30/2003 21:53'!
addCustomMenuItems: aCustomMenu hand: aHandMorph

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	model ifNotNil: [model addModelMenuItemsTo: aCustomMenu forMorph: self hand: aHandMorph].
	self isOpen ifTrue: [aCustomMenu add: 'close editing' translated action: #closeToEdits]
			ifFalse: [aCustomMenu add: 'open editing' translated action: #openToEdits].
! !

!MorphicModel methodsFor: 'compilation' stamp: 'ar 8/6/2009 20:42'!
addPartNameLike: className withValue: aMorph
	| otherNames i default partName stem |
	stem := className first asLowercase asString , className allButFirst.
	otherNames := self class allInstVarNames.
	i := 1.
	[otherNames includes: (default := stem, i printString)]
		whileTrue: [i := i + 1].
	partName := UIManager default
		request: 'Please give this part a name'
		initialAnswer: default.
	(otherNames includes: partName)
		ifTrue: [self inform: 'Sorry, that name is already used'. ^ nil].
	self class addInstVarName: partName.
	self instVarAt: self class instSize put: aMorph.  "Assumes added as last field"
	^ partName! !

!MorphicModel methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 18:51'!
allKnownNames
	"Return a list of all known names based on the scope of the receiver.  If the receiver is a member of a uniclass, incorporate the original 1997 logic that queries the known names of the values of all the instance variables."

	| superNames |
	superNames := super allKnownNames.	"gather them from submorph tree"
	^self belongsToUniClass 
		ifTrue: 
			[superNames , (self instanceVariableValues 
						select: [:e | e notNil and: [e knownName notNil]]
						thenCollect: [:e | e knownName])]
		ifFalse: [superNames]! !

!MorphicModel methodsFor: 'drag and drop' stamp: 'di 6/22/97 23:17'!
allowSubmorphExtraction
	^ self isOpen
! !

!MorphicModel methodsFor: 'geometry' stamp: 'cmm 3/27/2015 14:55'!
charactersOccluded
	"Subclasses override as necessary to keep smart-splitters balanced."
	^ 0! !

!MorphicModel methodsFor: 'naming' stamp: 'dgd 2/21/2003 23:00'!
choosePartName
	"When I am renamed, get a slot, make default methods, move any existing methods.  ** Does not clean up old inst var name or methods**  "

	| old |
	old := slotName.
	super choosePartName.
	slotName ifNil: [^self].	"user chose bad slot name"
	self model: self world model slotName: slotName.
	old isNil
		ifTrue: [self compilePropagationMethods]
		ifFalse: [self copySlotMethodsFrom: old]
	"old ones not erased!!"! !

!MorphicModel methodsFor: 'menu' stamp: 'di 6/20/97 15:36'!
closeToEdits
	"Disable this morph's ability to add and remove morphs via drag-n-drop."

	open := false
! !

!MorphicModel methodsFor: 'compilation' stamp: 'tk 4/18/97'!
compileAccessForSlot: aSlotName
	"Write the method to get at this inst var.  "
	"Instead call the right thing to make this happen?"

	| s  |
	s := WriteStream on: (String new: 2000).
	s nextPutAll: aSlotName; cr; tab; nextPutAll: '^', aSlotName.
	self class
		compile: s contents
		classified: 'public access'
		notifying: nil.
! !

!MorphicModel methodsFor: 'compilation' stamp: 'nice 12/27/2009 20:13'!
compilePropagationMethods
	
	(self class organization listAtCategoryNamed: 'private - propagation' asSymbol)
		do: [:sel | | varName |
			varName := sel allButLast.
			model class compilePropagationForVarName: varName slotName: slotName]! !

!MorphicModel methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:35'!
defaultBorderColor
	"answer the default border color/fill style for the receiver"
	^ Color yellow! !

!MorphicModel methodsFor: 'initialization' stamp: 'dgd 3/7/2003 15:07'!
defaultBounds
"answer the default bounds for the receiver"
	^ 0 @ 0 corner: 200 @ 100! !

!MorphicModel methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:28'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color transparent! !

!MorphicModel methodsFor: 'submorphs-add/remove' stamp: 'ar 8/6/2009 18:29'!
delete
	(model isMorphicModel) ifFalse: [^super delete].
	slotName ifNotNil: 
			[(UIManager default confirm: 'Shall I remove the slot ' , slotName 
						, '
along with all associated methods?') 
				ifTrue: 
					[(model class selectors select: [:s | s beginsWith: slotName]) 
						do: [:s | model class removeSelector: s].
					(model class instVarNames includes: slotName) 
						ifTrue: [model class removeInstVarName: slotName]]
				ifFalse: 
					[(UIManager default 
						confirm: '...but should I at least dismiss this morph?
[choose no to leave everything unchanged]') 
							ifFalse: [^self]]].
	super delete! !

!MorphicModel methodsFor: 'initialization' stamp: 'tk 4/15/97'!
duplicate: newGuy from: oldGuy
	"oldGuy has just been duplicated and will stay in this world.  Make sure all the MorphicModel requirements are carried out for the copy.  Ask user to rename it.  "

	newGuy installModelIn: oldGuy world.
	newGuy copySlotMethodsFrom: oldGuy slotName.! !

!MorphicModel methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:47'!
initialize
	"initialize the state of the receiver"
	super initialize.
""
	open := false! !

!MorphicModel methodsFor: 'printing' stamp: ''!
initString

	^ String streamContents:
		[:s | s nextPutAll: self class name;
			nextPutAll: ' newBounds: (';
			print: bounds;
			nextPutAll: ') model: self slotName: ';
			print: slotName]! !

!MorphicModel methodsFor: 'debug and other' stamp: '6/7/97 10:43 di'!
installModelIn: aWorld

	self wantsSlot ifFalse: [^ self].  "No real need to install"
	slotName := aWorld model addPartNameLike: self class name withValue: self.
	slotName ifNil: [^ self].  "user chose bad slot name"
	self model: aWorld model slotName: slotName.
	self compilePropagationMethods.
	aWorld model compileAccessForSlot: slotName.
! !

!MorphicModel methodsFor: 'classification' stamp: 'ar 10/5/2000 16:40'!
isMorphicModel
	^true! !

!MorphicModel methodsFor: 'drag and drop' stamp: 'di 6/22/97 23:16'!
isOpen
	"Support drag/drop and other edits."
	^ open! !

!MorphicModel methodsFor: 'accessing' stamp: ''!
model 
	^ model! !

!MorphicModel methodsFor: 'accessing' stamp: 'sw 10/23/1999 22:36'!
modelOrNil
	^ model! !

!MorphicModel methodsFor: 'initialization' stamp: 'jm 8/20/1998 09:08'!
model: anObject
	"Set my model and make me me a dependent of the given object."

	model ifNotNil: [model removeDependent: self].
	anObject ifNotNil: [anObject addDependent: self].
	model := anObject.
! !

!MorphicModel methodsFor: 'initialization' stamp: 'di 6/21/97 13:25'!
model: thang slotName: nameOfThisPart
	model := thang.
	slotName := nameOfThisPart.
	open := false.! !

!MorphicModel methodsFor: 'compilation' stamp: ''!
nameFor: aMorph
	"Return the name of the slot containing the given morph or nil if that morph has not been named."

	| allNames start |
	allNames := self class allInstVarNames.
	start := MorphicModel allInstVarNames size + 1.
	start to: allNames size do: [:i |
		(self instVarAt: i) == aMorph ifTrue: [^ allNames at: i]].
	^ nil
! !

!MorphicModel methodsFor: 'geometry' stamp: ''!
newBounds: newBounds
	self bounds: newBounds! !

!MorphicModel methodsFor: 'menu' stamp: 'di 6/20/97 15:36'!
openToEdits
	"Enable this morph's ability to add and remove morphs via drag-n-drop."

	open := true
! !

!MorphicModel methodsFor: 'compilation' stamp: ''!
propagate: value as: partStoreSelector
	model ifNil: [^ self].
"
	Later we can cache this for more speed as follows...
	(partName == cachedPartName and: [slotName == cachedSlotName])
		ifFalse: [cachedPartName := partName.
				cachedSlotName := slotName.
				cachedStoreSelector := (slotName , partStoreSelector) asSymbol].
	model perform: cachedStoreSelector with: value].
"
	model perform: (self slotSelectorFor: partStoreSelector) with: value! !

!MorphicModel methodsFor: 'geometry' stamp: ''!
recomputeBounds

	| bnds |
	bnds := submorphs first bounds.
	bounds := bnds origin corner: bnds corner. "copy it!!"
	fullBounds := nil.
	bounds := self fullBounds.
! !

!MorphicModel methodsFor: 'caching' stamp: 'sw 3/6/2001 11:22'!
releaseCachedState
	"Release cached state of the receiver"

	(model ~~ self and: [model respondsTo: #releaseCachedState]) ifTrue:
		[model releaseCachedState].
	super releaseCachedState! !

!MorphicModel methodsFor: 'compilation' stamp: 'tk 10/31/97 12:33'!
removeAll
	"Clear out all script methods and subpart instance variables in me.  Start over."
	"self removeAll"
	"MorphicModel2 removeAll"

self class == MorphicModel ifTrue: [^ self].	"Must be a subclass!!"
self class removeCategory: 'scripts'.
self class instVarNames do: [:nn | self class removeInstVarName: nn].! !

!MorphicModel methodsFor: 'accessing' stamp: ''!
slotName
	^ slotName! !

!MorphicModel methodsFor: 'compilation' stamp: ''!
slotSelectorFor: selectorBody
	| selector |
	model ifNil: [^ nil].
	"Make up selector from slotname if any"
	selector := (slotName ifNil: [selectorBody]
					ifNotNil: [slotName , selectorBody]) asSymbol.
	(model canUnderstand: selector) ifFalse:
		[self halt: 'Compiling a null response for ' , model class name , '>>' , selector].
	^ selector! !

!MorphicModel methodsFor: 'compilation' stamp: ''!
use: cachedSelector orMakeModelSelectorFor: selectorBody in: selectorBlock
	| selector |
	model ifNil: [^ nil].
	cachedSelector ifNil:
			["Make up selector from slotname if any"
			selector := (slotName ifNil: [selectorBody]
								ifNotNil: [slotName , selectorBody]) asSymbol.
			(model class canUnderstand: selector) ifFalse:
				[(self confirm: 'Shall I compile a null response for'
							, Character cr asString
							, model class name , '>>' , selector)
						ifFalse: [self halt].
				model class compile: (String streamContents:
								[:s | selector keywords doWithIndex:
										[:k :i | s nextPutAll: k , ' arg' , i printString].
								s cr; nextPutAll: '"Automatically generated null response."'.
								s cr; nextPutAll: '"Add code below for appropriate behavior..."'.])
							classified: 'input events'
							notifying: nil]]
		ifNotNil:
			[selector := cachedSelector].
	^ selectorBlock value: selector! !

!MorphicModel methodsFor: 'accessing' stamp: '6/7/97 10:40 di'!
wantsSlot
	"Override this default for models that want to be installed in theri model"
	^ false! !

MorphicModel subclass: #ScrollPane
	instanceVariableNames: 'scrollBar scroller retractableScrollBar scrollBarOnLeft getMenuSelector getMenuTitleSelector hasFocus hScrollBar hScrollBarPolicy vScrollBarPolicy scrollBarThickness'
	classVariableNames: 'UseRetractableScrollBars'
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!ScrollPane commentStamp: 'mk 8/9/2005 10:34' prior: 0!
The scroller (a transform) of a scrollPane is driven by the scrollBar.  The scroll values vary from 0.0, meaning zero offset to 1.0 meaning sufficient offset such that the bottom of the scrollable material appears 3/4 of the way down the pane.  The total distance to achieve this range is called the totalScrollRange.

Basic clue about utilization of the ScrollPane class is given in:
	ScrollPane example1.
	ScrollPane example2.!

ScrollPane subclass: #PluggableListMorph
	instanceVariableNames: 'list getListSelector getListSizeSelector getListElementSelector getIndexSelector setIndexSelector keystrokeActionSelector autoDeselect lastKeystrokeTime lastKeystrokes lastClickTime doubleClickSelector handlesBasicKeys potentialDropRow hoverRow listMorph hScrollRangeCache keystrokePreviewSelector priorSelection getIconSelector getHelpSelector'
	classVariableNames: 'ClearFilterAutomatically FilterableLists HighlightHoveredRow MenuRequestUpdatesSelection'
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!PluggableListMorph commentStamp: 'cmm 8/21/2011 23:37' prior: 0!
When a PluggableListMorph is in focus, type in a letter (or several letters quickly) to go to the next item that begins with that letter (if FilterableLists is false).

Special keys (up, down, home, etc.) are also supported.!

PluggableListMorph subclass: #AlternatePluggableListMorphOfMany
	instanceVariableNames: 'getSelectionListSelector setSelectionListSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!AlternatePluggableListMorphOfMany commentStamp: 'cmm 3/2/2010 14:39' prior: 0!
This is a multi-select list that is more conventional in its behavior than PluggableListMorphOfMany.  It utilizes a shift+click mechanism for selecting ranges, and control+click for toggling individual selections.  This list also allows fast mouse swipes without missing any message selections.!

!AlternatePluggableListMorphOfMany class methodsFor: 'instance creation' stamp: 'cmm 1/18/2002 16:08'!
on: anObject list: listSel primarySelection: getSelectionSel changePrimarySelection: setSelectionSel listSelection: getListSel changeListSelection: setListSel menu: getMenuSel
	^ self new
		on: anObject
		list: listSel
		primarySelection: getSelectionSel
		changePrimarySelection: setSelectionSel
		listSelection: getListSel
		changeListSelection: setListSel
		menu: getMenuSel
		keystroke: #arrowKey:from:		"default"! !

!AlternatePluggableListMorphOfMany class methodsFor: 'instance creation' stamp: 'cmm 1/18/2002 16:08'!
on: anObject list: listSel primarySelection: getSelectionSel changePrimarySelection: setSelectionSel listSelection: getListSel changeListSelection: setListSel menu: getMenuSel keystroke: keyActionSel 
	^ self new
		on: anObject
		list: listSel
		primarySelection: getSelectionSel
		changePrimarySelection: setSelectionSel
		listSelection: getListSel
		changeListSelection: setListSel
		menu: getMenuSel
		keystroke: keyActionSel! !

!AlternatePluggableListMorphOfMany methodsFor: 'model access' stamp: 'cmm 1/18/2002 17:49'!
changeModelSelection: anInteger
	"Change the model's selected item index to be anInteger."

	^self
		changeModelSelection: anInteger
		shifted: Sensor shiftPressed
		controlled: Sensor controlKeyPressed
! !

!AlternatePluggableListMorphOfMany methodsFor: 'model access' stamp: 'cmm 1/18/2002 16:50'!
changeModelSelection: anInteger shifted: shiftedBoolean controlled: controlledBoolean
	"Change the model's selected item index to be anInteger."

	setIndexSelector ifNotNil:
		[ model 
			perform: setIndexSelector 
			with: anInteger
			with: shiftedBoolean
			with: controlledBoolean ]
! !

!AlternatePluggableListMorphOfMany methodsFor: 'event handling' stamp: 'cmm 3/2/2004 23:14'!
handlesMouseDown: evt
	^ true! !

!AlternatePluggableListMorphOfMany methodsFor: 'model access' stamp: 'bf 7/19/2012 16:18'!
itemSelectedAmongMultiple: index
	^self listSelectionAt: (self modelIndexFor: index)! !

!AlternatePluggableListMorphOfMany methodsFor: 'drawing' stamp: 'cmm 1/18/2002 16:08'!
listSelectionAt: index
	getSelectionListSelector ifNil:[^false].
	^model perform: getSelectionListSelector with: index! !

!AlternatePluggableListMorphOfMany methodsFor: 'drawing' stamp: 'cmm 1/18/2002 16:08'!
listSelectionAt: index put: value
	setSelectionListSelector ifNil:[^false].
	^model perform: setSelectionListSelector with: index with: value! !

!AlternatePluggableListMorphOfMany methodsFor: 'initialization' stamp: 'tpr 12/31/2013 16:42'!
list: listOfStrings
self isThisEverCalled .
	scroller removeAllMorphs.
	list := listOfStrings ifNil: [Array new].
	list isEmpty ifTrue: [^ self selectedMorph: nil].
	super list: listOfStrings.

	"At this point first morph is sensitized, and all morphs share same handler."
	scroller firstSubmorph on: #mouseEnterDragging
						send: #mouseEnterDragging:onItem:
						to: self.
	scroller firstSubmorph on: #mouseUp
						send: #mouseUp:onItem:
						to: self.
	"This should add this behavior to the shared event handler thus affecting all items"! !

!AlternatePluggableListMorphOfMany methodsFor: 'event handling' stamp: 'bf 7/19/2012 16:39'!
mouseDown: event
	| row |

	event yellowButtonPressed ifTrue: [^ self yellowButtonActivity: event shiftPressed].

	row := self rowAtLocation: event position.
	
	row = 0 ifTrue: [^super mouseDown: event].

	model okToChange ifFalse: [^ self].  "No change if model is locked"

	"Inform model of selected item and let it toggle."
	self
		changeModelSelection: (self modelIndexFor: row)
		shifted: event shiftPressed
		controlled: event controlKeyPressed.


"
	event hand releaseMouseFocus: aMorph.
	submorphs do: [ :each | each changed ]
"! !

!AlternatePluggableListMorphOfMany methodsFor: 'event handling' stamp: 'cmm 4/18/2016 18:19'!
mouseMove: event 
	"The mouse has moved, as characterized by the event provided.  Adjust the scrollbar, and alter the selection as appropriate"

	| oldIndex oldVal row |
	event position y < self top 
		ifTrue: 
			[scrollBar scrollUp: 1.
			row := self rowAtLocation: scroller topLeft + (1 @ 1)]
		ifFalse: 
			[row := event position y > self bottom 
				ifTrue: 
					[scrollBar scrollDown: 1.
					self rowAtLocation: scroller bottomLeft + (1 @ -1)]
				ifFalse: [ self rowAtLocation: event position]].
	row = 0 ifTrue: [^super mouseDown: event].

	model okToChange ifFalse: [^self].	"No change if model is locked"

	"Set meaning for subsequent dragging of selection"
	oldIndex := self getCurrentSelectionIndex.
	oldIndex ~= 0 ifTrue: [oldVal := self listSelectionAt: oldIndex].
	"Need to restore the old one, due to how model works, and set new one."
	oldIndex ~= 0 ifTrue: [self listSelectionAt: oldIndex put: oldVal].

	"Inform model of selected item and let it toggle."
	self 
		changeModelSelection: (self modelIndexFor: row)
		shifted: true
		controlled: true.
	submorphs do: [:each | each changed]! !

!AlternatePluggableListMorphOfMany methodsFor: 'event handling' stamp: 'mt 4/9/2015 13:42'!
mouseUp: event
	
	event hand newKeyboardFocus: self. 
	hasFocus := true.! !

!AlternatePluggableListMorphOfMany methodsFor: 'initialization' stamp: 'cmm 12/24/2002 15:47'!
on: anObject list: listSel primarySelection: getSelectionSel changePrimarySelection: setSelectionSel listSelection: getListSel changeListSelection: setListSel menu: getMenuSel keystroke: keyActionSel 
	"setup a whole load of pluggability options"

	getSelectionListSelector := getListSel.
	setSelectionListSelector := setListSel.
	self 
		on: anObject
		list: listSel
		selected: getSelectionSel
		changeSelected: setSelectionSel
		menu: getMenuSel
		keystroke: keyActionSel! !

!AlternatePluggableListMorphOfMany methodsFor: 'event handling' stamp: 'cmm 1/18/2002 16:08'!
update: aSymbol 
	aSymbol == #allSelections ifTrue:
		[self selectionIndex: self getCurrentSelectionIndex.
		^ self changed].
	^ super update: aSymbol! !

PluggableListMorph subclass: #PluggableListMorphByItem
	instanceVariableNames: 'itemList'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!PluggableListMorphByItem methodsFor: 'model access' stamp: 'ls 8/19/2001 15:58'!
changeModelSelection: anInteger
	"Change the model's selected item to be the one at the given index."

	| item |
	setIndexSelector ifNotNil: [
		item := (anInteger = 0 ifTrue: [nil] ifFalse: [itemList at: anInteger]).
		model perform: setIndexSelector with: item].
	self update: getIndexSelector.
! !

!PluggableListMorphByItem methodsFor: 'model access' stamp: 'ls 8/19/2001 14:51'!
getCurrentSelectionIndex
	"Answer the index of the current selection."
	| item |
	getIndexSelector == nil ifTrue: [^ 0].
	item := model perform: getIndexSelector.
	^ list findFirst: [ :x | x = item]
! !

!PluggableListMorphByItem methodsFor: 'as yet unclassified' stamp: 'ls 8/19/2001 15:57'!
getList
	"cache the raw items in itemList"
	itemList := getListSelector ifNil: [ #() ] ifNotNil: [ model perform: getListSelector ].
	^super getList! !

!PluggableListMorphByItem methodsFor: 'initialization' stamp: 'ls 8/19/2001 14:52'!
list: arrayOfStrings
	"Set the receivers items to be the given list of strings."
	"Note: the instance variable 'items' holds the original list.
	 The instance variable 'list' is a paragraph constructed from
	 this list."
"NOTE: this is no longer true; list is a real list, and itemList is no longer used.  And this method shouldn't be called, incidentally."
self isThisEverCalled .
	itemList := arrayOfStrings.
	^ super list: arrayOfStrings! !

PluggableListMorph subclass: #PluggableListMorphOfMany
	instanceVariableNames: 'dragOnOrOff getSelectionListSelector setSelectionListSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!PluggableListMorphOfMany commentStamp: 'hpt 4/5/2004 11:21' prior: 0!
A variant of its superclass that allows multiple items to be selected simultaneously.  There is still a distinguished element which is selected, but each other element in the list may be flagged on or off.
!

!PluggableListMorphOfMany class methodsFor: 'instance creation' stamp: 'tpr 10/4/2001 21:54'!
on: anObject list: listSel primarySelection: getSelectionSel changePrimarySelection: setSelectionSel listSelection: getListSel changeListSelection: setListSel menu: getMenuSel
	^ self new
		on: anObject
		list: listSel
		primarySelection: getSelectionSel
		changePrimarySelection: setSelectionSel
		listSelection: getListSel
		changeListSelection: setListSel
		menu: getMenuSel
		keystroke: #arrowKey:from:		"default"! !

!PluggableListMorphOfMany class methodsFor: 'instance creation' stamp: 'tpr 10/4/2001 21:52'!
on: anObject list: listSel primarySelection: getSelectionSel changePrimarySelection: setSelectionSel listSelection: getListSel changeListSelection: setListSel menu: getMenuSel keystroke: keyActionSel 
	^ self new
		on: anObject
		list: listSel
		primarySelection: getSelectionSel
		changePrimarySelection: setSelectionSel
		listSelection: getListSel
		changeListSelection: setListSel
		menu: getMenuSel
		keystroke: keyActionSel! !

!PluggableListMorphOfMany methodsFor: 'model access' stamp: 'bf 4/13/2012 16:27'!
itemSelectedAmongMultiple: index
	^self listSelectionAt: (self modelIndexFor: index)! !

!PluggableListMorphOfMany methodsFor: 'drawing' stamp: 'tpr 10/4/2001 21:26'!
listSelectionAt: index
	getSelectionListSelector ifNil:[^false].
	^model perform: getSelectionListSelector with: index! !

!PluggableListMorphOfMany methodsFor: 'drawing' stamp: 'tpr 10/4/2001 21:27'!
listSelectionAt: index put: value
	setSelectionListSelector ifNil:[^false].
	^model perform: setSelectionListSelector with: index with: value! !

!PluggableListMorphOfMany methodsFor: 'initialization' stamp: 'ar 3/17/2001 17:07'!
list: listOfStrings
	scroller removeAllMorphs.
	list := listOfStrings ifNil: [Array new].
	list isEmpty ifTrue: [^ self selectedMorph: nil].
	super list: listOfStrings.

	"At this point first morph is sensitized, and all morphs share same handler."
	scroller firstSubmorph on: #mouseEnterDragging
						send: #mouseEnterDragging:onItem:
						to: self.
	scroller firstSubmorph on: #mouseUp
						send: #mouseUp:onItem:
						to: self.
	"This should add this behavior to the shared event handler thus affecting all items"! !

!PluggableListMorphOfMany methodsFor: 'event handling' stamp: 'bf 4/13/2012 16:23'!
mouseDown: event
	| oldIndex oldVal row index |
	event yellowButtonPressed ifTrue: [^ self yellowButtonActivity: event shiftPressed].
	row := self rowAtLocation: event position.

	row = 0 ifTrue: [^super mouseDown: event].
	index := self modelIndexFor: row.

	model okToChange ifFalse: [^ self].  "No change if model is locked"

	"Set meaning for subsequent dragging of selection"
	dragOnOrOff := (self listSelectionAt: index) not.
	oldIndex := self getCurrentSelectionIndex.
	oldIndex ~= 0 ifTrue: [oldVal := self listSelectionAt: oldIndex].

	"Set or clear new primary selection (listIndex)"
	dragOnOrOff
		ifTrue: [self changeModelSelection: index]
		ifFalse: [self changeModelSelection: 0].

	"Need to restore the old one, due to how model works, and set new one."
	oldIndex ~= 0 ifTrue: [self listSelectionAt: oldIndex put: oldVal].
	self listSelectionAt: index put: dragOnOrOff.
	"event hand releaseMouseFocus: aMorph."
! !

!PluggableListMorphOfMany methodsFor: 'event handling' stamp: 'bf 4/13/2012 16:22'!
mouseMove: event 
	"The mouse has moved, as characterized by the event provided.  Adjust the scrollbar, and alter the selection as appropriate"

	| oldIndex oldVal row index |
	event position y < self top 
		ifTrue: 
			[scrollBar scrollUp: 1.
			row := self rowAtLocation: scroller topLeft + (1 @ 1)]
		ifFalse: 
			[row := event position y > self bottom 
				ifTrue: 
					[scrollBar scrollDown: 1.
					self rowAtLocation: scroller bottomLeft + (1 @ -1)]
				ifFalse: [ self rowAtLocation: event position]].
	row = 0 ifTrue: [^super mouseDown: event].
	index := self modelIndexFor: row.

	model okToChange ifFalse: [^self].	"No change if model is locked"

	dragOnOrOff ifNil: 
			["Was not set at mouse down, which means the mouse must have gone down in an area where there was no list item"
			dragOnOrOff := (self listSelectionAt: index) not].

	"Set meaning for subsequent dragging of selection"
	oldIndex := self getCurrentSelectionIndex.
	oldIndex ~= 0 ifTrue: [oldVal := self listSelectionAt: oldIndex].

	"Set or clear new primary selection (listIndex)"
	dragOnOrOff 
		ifTrue: [self changeModelSelection: index]
		ifFalse: [self changeModelSelection: 0].

	"Need to restore the old one, due to how model works, and set new one."
	oldIndex ~= 0 ifTrue: [self listSelectionAt: oldIndex put: oldVal].
	self listSelectionAt: index put: dragOnOrOff.
! !

!PluggableListMorphOfMany methodsFor: 'event handling' stamp: 'mt 4/9/2015 13:44'!
mouseUp: event

	dragOnOrOff := nil.  "So improperly started drags will have not effect"
	event hand newKeyboardFocus: self. 
	hasFocus := true.! !

!PluggableListMorphOfMany methodsFor: 'initialization' stamp: 'tpr 10/4/2001 21:24'!
on: anObject list: listSel primarySelection: getSelectionSel changePrimarySelection: setSelectionSel listSelection: getListSel changeListSelection: setListSel menu: getMenuSel keystroke: keyActionSel
	"setup a whole load of pluggability options"
	getSelectionListSelector := getListSel.
	setSelectionListSelector := setListSel.
	super on: anObject list: listSel selected: getSelectionSel changeSelected: setSelectionSel menu: getMenuSel keystroke: keyActionSel
! !

!PluggableListMorphOfMany methodsFor: 'updating' stamp: 'di 11/10/1998 14:44'!
update: aSymbol 
	aSymbol == #allSelections ifTrue:
		[self selectionIndex: self getCurrentSelectionIndex.
		^ self changed].
	^ super update: aSymbol! !

!PluggableListMorph class methodsFor: 'preferences' stamp: 'cmm 10/4/2011 22:09'!
clearFilterAutomatically
	<preference: 'Filterable Lists Clear Automatically'
		category: 'scrolling'
		description: 'When using the Filterable Lists option, if this option is also selected, then the filter will be cleared as soon as the list loses keyboard focus.'
		type: #Boolean>
	^ ClearFilterAutomatically ifNil: [ true ]! !

!PluggableListMorph class methodsFor: 'preferences' stamp: 'cmm 10/18/2011 21:05'!
clearFilterAutomatically: aBoolean
	ClearFilterAutomatically := aBoolean! !

!PluggableListMorph class methodsFor: 'preferences' stamp: 'cmm 10/18/2011 21:07'!
filterableLists
	<preference: 'Filterable Lists'
		category: 'scrolling'
		description: 'When true, using the keyboard on a list will filter it rather than scroll.  Pressing enter clears the filter and keeps the filtered selection.  Backspace clears the filter and returns to the prior selection.'
		type: #Boolean>
	^ FilterableLists ifNil: [ true ]! !

!PluggableListMorph class methodsFor: 'preferences' stamp: 'cmm 10/18/2011 21:08'!
filterableLists: aBoolean
	"When true, using the keyboard on a list will filter it rather than scroll.  Pressing enter clears the filter and keeps the filtered selection.  Backspace clears the filter and returns to the prior selection."
	FilterableLists := aBoolean! !

!PluggableListMorph class methodsFor: 'preferences' stamp: 'mt 3/31/2015 17:16'!
highlightHoveredRow

	<preference: 'Highlight Hovered Row in Lists'
		category: #Morphic
		description: 'Indicate, which row will be affected before any further iteraction takes place.'
		type: #Boolean>
	^ HighlightHoveredRow ifNil: [true]! !

!PluggableListMorph class methodsFor: 'preferences' stamp: 'mt 3/31/2015 17:15'!
highlightHoveredRow: aBoolean

	HighlightHoveredRow := aBoolean.! !

!PluggableListMorph class methodsFor: 'preferences' stamp: 'mt 4/3/2015 09:09'!
menuRequestUpdatesSelection

	<preference: 'Menu request updates list/tree selection'
		category: #Morphic
		description: 'When invoking a menu, this will for the current selection by default. If the mouse cursor it at a different location, however, this preference will update the selection before showing the menu if enabled.'
		type: #Boolean>
	^ MenuRequestUpdatesSelection ifNil: [ false ]! !

!PluggableListMorph class methodsFor: 'preferences' stamp: 'mt 4/3/2015 09:08'!
menuRequestUpdatesSelection: aBoolean

	MenuRequestUpdatesSelection := aBoolean.! !

!PluggableListMorph class methodsFor: 'instance creation' stamp: 'md 7/13/2005 16:33'!
on: anObject list: getListSel selected: getSelectionSel changeSelected: setSelectionSel
	"Create a 'pluggable' list view on the given model parameterized by the given message selectors."

	^ self new
		on: anObject
		list: getListSel
		selected: getSelectionSel
		changeSelected: setSelectionSel
		menu: nil
		keystroke: #arrowKey:from:		"default"! !

!PluggableListMorph class methodsFor: 'instance creation' stamp: 'md 7/13/2005 16:33'!
on: anObject list: getListSel selected: getSelectionSel changeSelected: setSelectionSel menu: getMenuSel
	"Create a 'pluggable' list view on the given model parameterized by the given message selectors."

	^ self new
		on: anObject
		list: getListSel
		selected: getSelectionSel
		changeSelected: setSelectionSel
		menu: getMenuSel
		keystroke: #arrowKey:from:		"default"
! !

!PluggableListMorph class methodsFor: 'instance creation' stamp: 'md 7/13/2005 16:33'!
on: anObject list: getListSel selected: getSelectionSel changeSelected: setSelectionSel menu: getMenuSel keystroke: keyActionSel
	"Create a 'pluggable' list view on the given model parameterized by the given message selectors."

	^ self new
		on: anObject
		list: getListSel
		selected: getSelectionSel
		changeSelected: setSelectionSel
		menu: getMenuSel
		keystroke: keyActionSel
! !

!PluggableListMorph class methodsFor: 'preferences' stamp: 'mt 7/30/2016 08:27'!
themeProperties

	^ super themeProperties, {
		{ #font. 'Fonts'. 'Font of the list items.' }.
		{ #textColor. 'Colors'. 'Color of the list items.' }.
		{ #selectionColor. 'Colors'. 'Color used for items when hovering or selecting them.' }.
		{ #multiSelectionColor. 'Colors'. 'Colors used for items that are selected among others.'}.
		{ #selectionTextColor. 'Colors'. 'Color used for label when hovering or selecting them.' }.
		{ #filterColor. 'Colors'. 'Color used for items to indicate the matching filter.' }.
		{ #filterTextColor. 'Colors'. 'Color used for items to indicate the matching filter.' }.
		
		{ #preSelectionModifier. 'Colors'. 'How to derive the pre-selection color from the selection color.'}.
		{ #hoverSelectionModifier. 'Colors'. 'How to derive the hover color from the selection color.'}.
	}! !

!PluggableListMorph methodsFor: 'drag and drop' stamp: 'ls 6/22/2001 23:56'!
acceptDroppingMorph: aMorph event: evt 
	"This message is sent when a morph is dropped onto a morph that has     
	agreed to accept the dropped morph by responding 'true' to the     
	wantsDroppedMorph:Event: message. The default implementation just     
	adds the given morph to the receiver."
	"Here we let the model do its work."

	self model
		acceptDroppingMorph: aMorph
		event: evt
		inMorph: self.
	self resetPotentialDropRow.
	evt hand releaseMouseFocus: self.
	Cursor normal show.
! !

!PluggableListMorph methodsFor: 'menus' stamp: 'dgd 8/30/2003 21:56'!
addCustomMenuItems:  aMenu hand: aHandMorph
	"Add halo menu items to be handled by the invoking hand. The halo menu is invoked by clicking on the menu-handle of the receiver's halo."

	super addCustomMenuItems: aMenu hand: aHandMorph.
	aMenu addLine.
	aMenu add: 'list font...' translated target: self action: #setListFont.
	aMenu add: 'copy list to clipboard' translated target: self action: #copyListToClipboard.
	aMenu add: 'copy selection to clipboard' translated target: self action: #copySelectionToClipboard! !

!PluggableListMorph methodsFor: 'filtering' stamp: 'mt 7/19/2016 11:15'!
allowEmptyFilterResult
	^ self valueOfProperty: #allowEmptyFilterResult ifAbsent: [false]! !

!PluggableListMorph methodsFor: 'filtering' stamp: 'mt 7/19/2016 11:15'!
allowEmptyFilterResult: aBoolean
	
	self
		setProperty: #allowEmptyFilterResult
		toValue: aBoolean.! !

!PluggableListMorph methodsFor: 'submorphs-accessing' stamp: 'di 11/14/2001 13:57'!
allSubmorphNamesDo: nameBlock
	"Assume list morphs do not have named parts -- saves MUCH time"

	^ self! !

!PluggableListMorph methodsFor: 'updating' stamp: 'mt 7/24/2016 10:23'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'di 4/10/98 16:20'!
autoDeselect: trueOrFalse
	"Enable/disable autoDeselect (see class comment)"
	autoDeselect := trueOrFalse.! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'mt 11/6/2015 11:27'!
balloonText
	"Overridden to send selector to model and not self. Do not use #perform:orSendTo: because super does more than just the send.."
	
	self getHelpSelector ifNotNil: [:selector |
		((self model respondsTo: selector) and: [self hoverRow > 0]) ifTrue: [
			^ self model perform: selector with: (self modelIndexFor: self hoverRow)]].
	
	^ super balloonText! !

!PluggableListMorph methodsFor: 'model access' stamp: 'ul 10/22/2011 05:17'!
basicKeyPressed: aChar 
	| milliseconds slowKeyStroke listSize newSelectionIndex oldSelectionIndex startIndex |
	oldSelectionIndex := newSelectionIndex := self getCurrentSelectionIndex.
	listSize := self getListSize.
	listSize = 0 ifTrue: [ ^self flash ].
	milliseconds := Time millisecondClockValue.
	slowKeyStroke := (Time
		milliseconds: milliseconds
		since: lastKeystrokeTime) > (self class filterableLists ifTrue: [500] ifFalse: [ 300 ]).
	lastKeystrokeTime := milliseconds.
	slowKeyStroke
		ifTrue:
			[ self class filterableLists ifTrue: [ self hasFilter ifFalse: [ priorSelection := self modelIndexFor: self selectionIndex] ].
			"forget previous keystrokes and search in following elements"
			lastKeystrokes := aChar asLowercase asString.
			newSelectionIndex := newSelectionIndex \\ listSize + 1.
			self class filterableLists ifTrue: [ list := self getFullList ] ]
		ifFalse: [ "append quick keystrokes but don't move selection if it still matches"
			lastKeystrokes := lastKeystrokes , aChar asLowercase asString.
			newSelectionIndex := newSelectionIndex max: 1 ].
	"No change if model is locked"
	model okToChange ifFalse: [ ^ self ].
	self class filterableLists
		ifTrue:
			[ self
				 filterList ;
				 updateList.
			newSelectionIndex := self modelIndexFor: 1 ]
		ifFalse:
			[ startIndex := newSelectionIndex.
			listSize := self getListSize.
			[ (self getListItem: newSelectionIndex) asString withBlanksTrimmed asLowercase beginsWith: lastKeystrokes ] whileFalse:
				[ (newSelectionIndex := newSelectionIndex \\ listSize + 1) = startIndex ifTrue: [ ^ self flash"Not in list." ] ].
			newSelectionIndex = oldSelectionIndex ifTrue: [ ^ self flash ] ].
	(self hasFilter and: [(self getCurrentSelectionIndex = newSelectionIndex) not]) ifTrue:
		[self changeModelSelection: newSelectionIndex]! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'cmm 8/20/2013 16:23'!
bottomVisibleRowIndex
	^ self rowAtLocation: self bottomLeft+(3@3 negated)! !

!PluggableListMorph methodsFor: 'testing' stamp: 'cmm 8/26/2013 20:42'!
canBeEncroached
	"Answer whether my bottom edge can be encroached by horizontal smart-splitter.  If my list is larger than my outermost containing window, go ahead and report true since moving a splitter will never allow my entire list to be displayed.  In that case go ahead and be encroachable to allow lower truncated text-panes to be exposed, but leave a reasonable height (70) to ensure at least few items are displayed."
	^ self height > 24 and:
		[ | outermostContainer |
		outermostContainer := self outermostMorphThat:
			[ : e | e owner = World ].
		listMorph height + 8 < self height or:
			[ outermostContainer notNil and: [ listMorph height > (outermostContainer height / 1.2) and: [ self height > 70 ] ] ] ]! !

!PluggableListMorph methodsFor: 'model access' stamp: 'mt 4/19/2016 12:08'!
changeModelSelection: anInteger
	" Change the model's selected item index to be anInteger. Enable the pre selection highlight. Step the World forward to let the pre selection highlight take effect. "

	self rowAboutToBecomeSelected: (self uiIndexFor: anInteger).
	self refreshWorld.
	setIndexSelector ifNotNil: [
		model perform: setIndexSelector with: anInteger ].! !

!PluggableListMorph methodsFor: 'geometry' stamp: 'cmm 4/13/2016 22:22'!
charactersOccluded
	"Answer the number of characters occluded in my #visibleList by my right edge."
	| listIndex | listIndex:=0.
	^ self visibleList
		inject: 0
		into:
			[ : sum : each | | eachString totalWidth indexOfLastVisible iconWidth |
			totalWidth:=0.
			eachString := each asString "withBlanksTrimmed".
			iconWidth := (self iconAt: (listIndex := listIndex+1)) ifNil:[0] ifNotNil: [ : icon | icon width+2 ].
			indexOfLastVisible := ((1 to: eachString size)
				detect:
					[ : stringIndex | (totalWidth:=totalWidth+(self font widthOf: (eachString at: stringIndex))) >
						(self width -
							(scrollBar
								ifNil: [ 0 ]
								ifNotNil: [ scrollBar width ]) - iconWidth) ]
				ifNone: [ eachString size + 1 ]) - 1.
			sum + ((eachString size - indexOfLastVisible) min: 10) ]! !

!PluggableListMorph methodsFor: 'model access' stamp: 'sw 12/4/2001 20:51'!
commandKeyTypedIntoMenu: evt
	"The user typed a command-key into a menu which has me as its command-key handler"

	^ self modifierKeyPressed: evt keyCharacter! !

!PluggableListMorph methodsFor: 'menus' stamp: 'dao 6/6/2006 15:58'!
copyListToClipboard
	"Copy my items to the clipboard as a multi-line string"

	| stream |
	stream := WriteStream on: (String new: self getList size * 40).
	list do: [:ea | stream nextPutAll: ea asString] separatedBy: [stream nextPut: Character cr].
	Clipboard clipboardText: stream contents! !

!PluggableListMorph methodsFor: 'menus' stamp: 'sw 3/31/2002 02:38'!
copySelectionToClipboard
	"Copy my selected item to the clipboard as a string"

	self selection
		ifNotNil:
			[Clipboard clipboardText: self selection asString]
		ifNil:
			[self flash]! !

!PluggableListMorph methodsFor: 'scroll cache' stamp: 'sps 4/3/2005 15:29'!
deriveHScrollRange

	|  unadjustedRange totalRange |
	(list isNil or: [list isEmpty]) 
		ifTrue:[hScrollRangeCache := Array with: 0 with: 0 with: 0 with: 0 with: 0 ]
		ifFalse:[ 
			unadjustedRange := self listMorph hUnadjustedScrollRange.
			totalRange := unadjustedRange + self hExtraScrollRange + self hMargin.
			hScrollRangeCache := Array 
										with: totalRange 
										with: unadjustedRange 
										with: list size 
										with: list first 
										with: list last .
		].
! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'sw 1/12/2000 16:22'!
doubleClickSelector: aSymbol
	doubleClickSelector := aSymbol! !

!PluggableListMorph methodsFor: 'events' stamp: 'jmv 3/2/2010 16:21'!
doubleClick: event
	| index |
	doubleClickSelector ifNil: [^super doubleClick: event].
	index := self rowAtLocation: event position.
	index = 0 ifTrue: [^super doubleClick: event].
	"selectedMorph ifNil: [self setSelectedMorph: aMorph]."
	^ self model perform: doubleClickSelector! !

!PluggableListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:19'!
doubleClick: event onItem: aMorph
	self removeObsoleteEventHandlers.! !

!PluggableListMorph methodsFor: 'geometry' stamp: 'mt 4/15/2015 15:51'!
extent: newExtent
	super extent: newExtent.
	
	"Change listMorph's bounds to the new width. It is either the size
	of the widest list item, or the size of self, whatever is bigger"
	self listMorph width: (self width max: listMorph hUnadjustedScrollRange). 
! !

!PluggableListMorph methodsFor: 'filtering' stamp: 'mt 7/19/2016 11:17'!
filterList
	self hasFilter
		ifTrue:
			[ | frontMatching substringMatching newList |
			self indicateFiltered.
			frontMatching := OrderedCollection new.
			substringMatching := OrderedCollection new.
			list withIndexDo:
				[ : each : n | | foundPos |
				foundPos := each asString
					findString: lastKeystrokes
					startingAt: 1
					caseSensitive: false.
				foundPos = 1
					ifTrue: [ frontMatching add: each ]
					ifFalse:
						[ foundPos = 0 ifFalse: [ substringMatching add: each ] ] ].
			newList := frontMatching , substringMatching.
			(newList isEmpty not or: [ self allowEmptyFilterResult ])
				ifTrue: [ list := newList ]
				ifFalse: 
					[ lastKeystrokes := lastKeystrokes allButLast: 1.
					self
						 flash ;
						 filterList ] ]
		ifFalse: [ self indicateUnfiltered ]! !

!PluggableListMorph methodsFor: 'filtering' stamp: 'mt 7/19/2016 11:18'!
filterList: aString
	"Manually set the list filter."

	lastKeystrokes := aString.
	self filterList.
	self updateList.
	self changeModelSelection: (list ifEmpty: [0] ifNotEmpty: [self modelIndexFor: 1]).! !

!PluggableListMorph methodsFor: 'filtering' stamp: 'mt 4/12/2015 13:37'!
filterTerm
	^ lastKeystrokes ifNil: ['']! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'ls 5/15/2001 22:21'!
font

	^ self listMorph font
! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'mt 4/24/2015 10:45'!
font: aFontOrNil

	self listMorph font: aFontOrNil.
	
	self minimumWidth: (self font widthOf: $m) * 5.
! !

!PluggableListMorph methodsFor: 'model access' stamp: 'jmv 3/2/2010 16:21'!
getCurrentSelectionIndex
	"Answer the index of the current selection."

	getIndexSelector ifNil: [^0].
	^model perform: getIndexSelector! !

!PluggableListMorph methodsFor: 'model access' stamp: 'cmm 8/20/2011 16:00'!
getFullList
	"The full, unfiltered list."
	^ model perform: getListSelector! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'mt 11/6/2015 11:29'!
getHelpSelector
	
	^ getHelpSelector! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'mt 11/6/2015 11:29'!
getHelpSelector: aSelector
	"Get help for list entries."
	
	getHelpSelector := aSelector.! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'cwp 12/6/2011 14:04'!
getIconSelector: aSymbol
	getIconSelector := aSymbol! !

!PluggableListMorph methodsFor: 'model access' stamp: 'cmm 8/22/2011 09:17'!
getList
	"Answer the list to be displayed.  Caches the returned list in the 'list' ivar"
	getListSelector == nil ifTrue: [ ^ Array empty ].
	list := self getFullList.
	self class filterableLists ifTrue: [ self filterList ].
	^ list ifNil: [ Array empty ]! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'ar 1/23/2010 14:31'!
getListElementSelector: aSymbol
	"specify a selector that can be used to obtain a single element in the underlying list"
	getListElementSelector == aSymbol ifTrue:[^self].
	getListElementSelector := aSymbol.
	list := nil.  "this cache will not be updated if getListElementSelector has been specified, so go ahead and remove it"! !

!PluggableListMorph methodsFor: 'model access' stamp: 'ls 7/1/2001 10:39'!
getListItem: index
	"get the index-th item in the displayed list"
	getListElementSelector ifNotNil: [ ^(model perform: getListElementSelector with: index) asStringOrText ].
	list ifNotNil: [ ^list at: index ].
	^self getList at: index! !

!PluggableListMorph methodsFor: 'selection' stamp: 'di 6/21/1998 22:19'!
getListSelector
	^ getListSelector! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'cmm 3/2/2011 10:49'!
getListSelector: sel 
	"Set the receiver's getListSelector as indicated, and trigger a recomputation of the list"
	self
		 setGetListSelector: sel ;
		 changed ;
		 updateList! !

!PluggableListMorph methodsFor: 'model access' stamp: 'ls 5/17/2001 22:04'!
getListSize
	"return the current number of items in the displayed list"
	getListSizeSelector ifNotNil: [ ^model perform: getListSizeSelector ].
	^self getList size! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'ls 5/22/2001 18:21'!
getListSizeSelector: aSymbol
	"specify a selector that can be used to specify the list's size"
	getListSizeSelector := aSymbol! !

!PluggableListMorph methodsFor: 'menu' stamp: 'tk 12/10/2001 20:33'!
getMenu: shiftKeyState
	"Answer the menu for this text view, supplying an empty menu to be filled in. If the menu selector takes an extra argument, pass in the current state of the shift key."

	| aMenu |
	aMenu := super getMenu: shiftKeyState.
	aMenu ifNotNil: [aMenu commandKeyHandler: self].
	^ aMenu! !

!PluggableListMorph methodsFor: 'events' stamp: 'ls 10/14/2001 13:08'!
handleBasicKeys: aBoolean
	"set whether the list morph should handle basic keys like arrow keys, or whether everything should be passed to the model"
	handlesBasicKeys := aBoolean! !

!PluggableListMorph methodsFor: 'events-processing' stamp: 'mt 3/31/2015 16:46'!
handleMouseMove: anEvent
	"Reimplemented because we really want #mouseMove when a morph is dragged around"
	anEvent wasHandled ifTrue:[^self]. "not interested"
	self hoverRow: (self rowAtLocation: anEvent position).
	(anEvent anyButtonPressed and:[anEvent hand mouseFocus == self]) ifFalse:[^self].
	anEvent wasHandled: true.
	self mouseMove: anEvent.
	(self handlesMouseStillDown: anEvent) ifTrue:[
		"Step at the new location"
		self startStepping: #handleMouseStillDown: 
			at: Time millisecondClockValue
			arguments: {anEvent copy resetHandlerFields}
			stepTime: 1].
! !

!PluggableListMorph methodsFor: 'events' stamp: 'ls 10/14/2001 13:09'!
handlesBasicKeys
	" if ya don't want the list to automatically handle non-modifier key 
	(excluding shift key) input, return false"
	^ handlesBasicKeys ifNil: [ true ]! !

!PluggableListMorph methodsFor: 'event handling' stamp: 'ar 3/17/2001 16:05'!
handlesMouseOverDragging: evt
	^self dropEnabled! !

!PluggableListMorph methodsFor: 'filtering' stamp: 'cmm 8/21/2011 19:56'!
hasFilter
	^ lastKeystrokes isEmptyOrNil not! !

!PluggableListMorph methodsFor: 'scrolling' stamp: 'sps 12/24/2002 18:31'!
hExtraScrollRange
	"Return the amount of extra blank space to include to the right of the scroll content."
	^5 
! !

!PluggableListMorph methodsFor: 'drawing' stamp: 'ls 5/17/2001 20:53'!
highlightSelection! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'ar 1/31/2001 19:21'!
highlightSelector
	^self valueOfProperty: #highlightSelector! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'ls 5/15/2001 22:31'!
highlightSelector: aSelector
	self setProperty: #highlightSelector toValue: aSelector.
	self updateList! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'mt 3/31/2015 16:26'!
hoverItem

	^ self hoverRow = 0
		ifTrue: [nil]
		ifFalse: [self getListItem: self hoverRow]! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'mt 3/31/2015 16:25'!
hoverRow

	^ hoverRow ifNil: [0]! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'mt 11/6/2015 11:26'!
hoverRow: anInteger

	hoverRow = anInteger ifTrue: [^ self].
	
	hoverRow ifNotNil: [self listMorph rowChanged: hoverRow].
	hoverRow := anInteger.
	hoverRow ifNotNil: [self listMorph rowChanged: hoverRow].
	
	self wantsBalloon ifTrue: [
		self activeHand
			removePendingBalloonFor: self;
			triggerBalloonFor: self after: self balloonHelpDelayTime].! !

!PluggableListMorph methodsFor: 'scroll cache' stamp: 'sps 4/3/2005 15:29'!
hTotalScrollRange
	"Return the entire scrolling range."

	 self resetHScrollRangeIfNecessary.

	^hScrollRangeCache first
! !

!PluggableListMorph methodsFor: 'scroll cache' stamp: 'sps 4/3/2005 15:29'!
hUnadjustedScrollRange
	"Return the entire scrolling range."

	 self resetHScrollRangeIfNecessary.

	^hScrollRangeCache second
! !

!PluggableListMorph methodsFor: 'model access' stamp: 'cmm 3/27/2015 14:43'!
iconAt: anInteger

	| index |
	index := (self hasFilter and: [list notNil])
		ifTrue: [self getFullList indexOf: (list at: anInteger ifAbsent: [^nil])]
		ifFalse: [anInteger].
	^ getIconSelector ifNotNil: [model perform: getIconSelector with: index]! !

!PluggableListMorph methodsFor: 'filtering' stamp: 'cmm 4/18/2015 11:16'!
indicateFiltered
	"Don't need this pink background anymore, since now the current filter is highlighted in every matching substring entry in the list."
	"self color: Color red muchLighter muchLighter"! !

!PluggableListMorph methodsFor: 'filtering' stamp: 'mt 7/23/2016 17:19'!
indicateUnfiltered
	! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'di 10/11/1999 08:45'!
initForKeystrokes
	lastKeystrokeTime := 0.
	lastKeystrokes := ''! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'mt 4/24/2015 21:38'!
initialize
	super initialize.
	
	self minimumWidth: (self font widthOf: $m) * 5.! !

!PluggableListMorph methodsFor: 'debug and other' stamp: 'di 5/6/1998 21:19'!
installModelIn: aWorld
	"No special inits for new components"
	^ self! !

!PluggableListMorph methodsFor: 'drawing' stamp: 'cmm 8/19/2013 22:12'!
isPluggableListMorph
	^ true! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'ar 3/17/2001 15:32'!
itemFromPoint: aPoint
	"Return the list element (morph) at the given point or nil if outside"
	| ptY |
	scroller hasSubmorphs ifFalse:[^nil].
	(scroller fullBounds containsPoint: aPoint) ifFalse:[^nil].
	ptY := (scroller firstSubmorph point: aPoint from: self) y.
	"note: following assumes that submorphs are vertical, non-overlapping, and ordered"
	scroller firstSubmorph top > ptY ifTrue:[^nil].
	scroller lastSubmorph bottom < ptY ifTrue:[^nil].
	"now use binary search"
	^scroller 
		findSubmorphBinary:[:item|
			(item top <= ptY and:[item bottom >= ptY])
				ifTrue:[0] "found"
				ifFalse:[ (item top + item bottom // 2) > ptY ifTrue:[-1] ifFalse:[1]]]! !

!PluggableListMorph methodsFor: 'model access' stamp: 'ls 6/10/2001 12:26'!
itemSelectedAmongMultiple: index
	"return whether the index-th row is selected.  Always false in PluggableListMorph, but sometimes true in PluggableListMorphOfMany"
	^false! !

!PluggableListMorph methodsFor: 'event handling' stamp: 'mt 4/10/2015 16:28'!
keyboardFocusChange: aBoolean 
	"The message is sent to a morph when its keyboard focus changes.
	The given argument indicates that the receiver is gaining (versus losing) the keyboard focus.
	In this case, all we need to do is to redraw border feedback"
	aBoolean ifFalse: [
		self hoverRow: nil.
		self class clearFilterAutomatically ifTrue:
			[ self hasFilter ifTrue:
				[ self
					 removeFilter ;
					 updateList ] ] ].
		
	super keyboardFocusChange: aBoolean.! !

!PluggableListMorph methodsFor: 'event handling' stamp: 'mt 6/6/2016 17:58'!
keyStroke: event 
	"Process keys 
	The model is allowed to preview all keystrokes. If it's not interested:
	specialKeys are things like up, down, etc. ALWAYS HANDLED 
	modifierKeys are regular characters either 1) accompanied with ctrl, 
	cmd or 2) any character if the list doesn't want to handle basic 
	keys (handlesBasicKeys returns false) 
	basicKeys are any characters"
	
	| aChar aSpecialKey |
	(self previewKeystroke: event)
		ifTrue: [^ self].
	
	aChar := event keyCharacter.
	
	(aSpecialKey := aChar asciiValue) < 32
		ifTrue: [^ self specialKeyPressed: aSpecialKey].
	
	(event anyModifierKeyPressed or: [self handlesBasicKeys not])
		ifTrue: [^ self modifierKeyPressed: aChar].
		
	^ self basicKeyPressed: aChar! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'sw 1/18/2001 13:08'!
keystrokeActionSelector: keyActionSel
	"Set the keystroke action selector as specified"

	keystrokeActionSelector := keyActionSel! !

!PluggableListMorph methodsFor: 'selection' stamp: 'MAD 3/13/2010 14:53'!
keystrokePreviewSelector
	^ keystrokePreviewSelector! !

!PluggableListMorph methodsFor: 'selection' stamp: 'MAD 3/13/2010 14:53'!
keystrokePreviewSelector: sel
	"The method on the model that will be given first view of any keystroke events.  For access via scripting"

	keystrokePreviewSelector := sel! !

!PluggableListMorph methodsFor: 'as yet unclassified' stamp: 'ls 2/5/2004 18:01'!
listMorph
	listMorph ifNil: [
		"crate this lazily, in case the morph is legacy"
		listMorph := self listMorphClass new.
		listMorph listSource: self.
		listMorph width: self scroller width.
		listMorph color: self textColor ].

	listMorph owner ~~ self scroller ifTrue: [
		"list morph needs to be installed.  Again, it's done this way to accomodate legacy PluggableListMorphs"
		self scroller removeAllMorphs.
		self scroller addMorph: listMorph ].

	^listMorph! !

!PluggableListMorph methodsFor: 'as yet unclassified' stamp: 'ls 5/17/2001 09:04'!
listMorphClass
	^LazyListMorph! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'cmm 2/27/2011 15:48'!
list: listOfStrings  
	"lex doesn't think this is used any longer, but is not yet brave enough to remove it.  It should be removed eventually"
	
	
	"Set the receiver's list as specified"

	| morphList h index converter aSelector textColor font loc |
self isThisEverCalled.
	scroller removeAllMorphs.
	list := listOfStrings ifNil: [Array new].
	list isEmpty ifTrue: [self setScrollDeltas.  ^ self selectedMorph: nil].
	"NOTE: we will want a quick StringMorph init message, possibly even
		combined with event install and positioning"
	font ifNil: [font := Preferences standardListFont].
	converter := self valueOfProperty: #itemConversionMethod.
	converter ifNil: [converter := #asStringOrText].
	textColor := self valueOfProperty: #textColor.
	morphList := list collect: [:each | | stringMorph item |
		item := each.
		item := item perform: converter.
		stringMorph := item isText
			ifTrue: [StringMorph contents: item font: font emphasis: (item emphasisAt: 1)]
			ifFalse: [StringMorph contents: item font: font].
		textColor ifNotNil: [ stringMorph color: textColor ].
		stringMorph
	].
	
	(aSelector := self valueOfProperty: #balloonTextSelectorForSubMorphs)
		ifNotNil:
			[morphList do: [:m | m balloonTextSelector: aSelector]].

	self highlightSelector ifNotNil:
		[model perform: self highlightSelector with: list with: morphList].

	"Lay items out vertically and install them in the scroller"
	h := morphList first height "self listItemHeight".
	loc := 0@0.
	morphList do: [:m | m bounds: (loc extent: 9999@h).  loc := loc + (0@h)].
	scroller addAllMorphs: morphList.

	index := self getCurrentSelectionIndex.
	self selectedMorph: ((index = 0 or: [index > morphList size]) ifTrue: [nil] ifFalse: [morphList at: index]).
	self setScrollDeltas.
	scrollBar setValue: 0.0! !

!PluggableListMorph methodsFor: 'selection' stamp: 'ls 5/17/2001 23:06'!
maximumSelection
	^ self getListSize! !

!PluggableListMorph methodsFor: 'selection' stamp: 'di 5/22/1998 00:20'!
minimumSelection
	^ 1! !

!PluggableListMorph methodsFor: 'selection' stamp: 'cmm 8/21/2011 23:49'!
modelIndexFor: selectionIndex 
	"The model does not know anything about the receiver's filtering, so if I am filtered, we must determine the correct index by scanning the full list in the model."
	^ (selectionIndex > 0 and: [ self hasFilter ])
		ifTrue:
			[ list
				ifEmpty: [ 0 ]
				ifNotEmpty: [ self getFullList indexOf: (self getListItem: selectionIndex) ] ]
		ifFalse: [ selectionIndex ]! !

!PluggableListMorph methodsFor: 'model access' stamp: 'dgd 2/21/2003 23:05'!
modifierKeyPressed: aChar 
	| args |
	keystrokeActionSelector isNil ifTrue: [^nil].
	args := keystrokeActionSelector numArgs.
	args = 1 ifTrue: [^model perform: keystrokeActionSelector with: aChar].
	args = 2 
		ifTrue: 
			[^model 
				perform: keystrokeActionSelector
				with: aChar
				with: self].
	^self error: 'keystrokeActionSelector must be a 1- or 2-keyword symbol'! !

!PluggableListMorph methodsFor: 'events' stamp: 'mt 4/7/2015 16:45'!
mouseDown: evt
	| selectors row |
	row := self rowAtLocation: evt position.

	evt yellowButtonPressed  "First check for option (menu) click"
		ifTrue: [
			(self class menuRequestUpdatesSelection and: [model okToChange]) ifTrue: [
				"Models depend on the correct selection:"
				self selectionIndex = (self modelIndexFor: row)
					ifFalse: [self changeModelSelection: (self modelIndexFor: row)]].
			
			^ self yellowButtonActivity: evt shiftPressed].
	row = 0  ifTrue: [^super mouseDown: evt].
	"self dragEnabled ifTrue: [aMorph highlightForMouseDown]."
	selectors := Array 
		with: #click:
		with: (doubleClickSelector ifNotNil:[#doubleClick:])
		with: nil
		with: (self dragEnabled ifTrue:[#startDrag:] ifFalse:[nil]).
	evt hand waitForClicksOrDrag: self event: evt selectors: selectors threshold: HandMorph dragThreshold "pixels".! !

!PluggableListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:19'!
mouseDown: event onItem: aMorph
	self removeObsoleteEventHandlers.! !

!PluggableListMorph methodsFor: 'events' stamp: 'ls 6/22/2001 23:58'!
mouseEnterDragging: evt

	(evt hand hasSubmorphs and:[self dropEnabled]) ifFalse: ["no d&d"
		^super mouseEnterDragging: evt].

	(self wantsDroppedMorph: evt hand firstSubmorph event: evt )
		ifTrue:[
			potentialDropRow := self rowAtLocation: evt position.
			evt hand newMouseFocus: self.
			self changed.
			"above is ugly but necessary for now"
		].
! !

!PluggableListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:20'!
mouseEnterDragging: anEvent onItem: aMorph 
	self removeObsoleteEventHandlers.! !

!PluggableListMorph methodsFor: 'events' stamp: 'mt 5/19/2016 11:15'!
mouseEnter: event

	super mouseEnter: event.
	Preferences mouseOverForKeyboardFocus
		ifTrue: [event hand newKeyboardFocus: self].! !

!PluggableListMorph methodsFor: 'events' stamp: 'mt 3/31/2015 16:47'!
mouseLeaveDragging: anEvent

	self hoverRow: nil.
	(self dropEnabled and:[anEvent hand hasSubmorphs]) ifFalse: ["no d&d"
		^ super mouseLeaveDragging: anEvent].
	self resetPotentialDropRow.
	anEvent hand releaseMouseFocus: self.
	"above is ugly but necessary for now"
! !

!PluggableListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:20'!
mouseLeaveDragging: anEvent onItem: aMorph 
	self removeObsoleteEventHandlers.! !

!PluggableListMorph methodsFor: 'events' stamp: 'mt 5/19/2016 11:16'!
mouseLeave: event 

	super mouseLeave: event.
	self hoverRow: nil.

	Preferences mouseOverForKeyboardFocus
		ifTrue: [event hand releaseKeyboardFocus: self].! !

!PluggableListMorph methodsFor: 'events' stamp: 'ls 6/22/2001 23:55'!
mouseMove: evt

	(self dropEnabled and:[evt hand hasSubmorphs]) 
		ifFalse:[^super mouseMove: evt].
	potentialDropRow ifNotNil:[
		potentialDropRow = (self rowAtLocation: evt position)
			ifTrue:[^self].
	].
	self mouseLeaveDragging: evt.
	(self containsPoint: evt position) 
		ifTrue:[self mouseEnterDragging: evt].! !

!PluggableListMorph methodsFor: 'events' stamp: 'kfr 6/22/2014 16:13'!
mouseUp: event 
	"The mouse came up within the list; take appropriate action"
	| row |
	row := self rowAtLocation: event position.
	"aMorph ifNotNil: [aMorph highlightForMouseDown: false]."
	model okToChange
		ifFalse: [^ self].
	"No change if model is locked"
	row = self selectionIndex
		ifTrue: [(autoDeselect ifNil: [true]) ifTrue:[row = 0 ifFalse: [self changeModelSelection: 0] ]]
		ifFalse: [self changeModelSelection: (self modelIndexFor: row)].
	event hand newKeyboardFocus: self. 
	hasFocus := true.
	Cursor normal show! !

!PluggableListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:20'!
mouseUp: event onItem: aMorph 
	self removeObsoleteEventHandlers.! !

!PluggableListMorph methodsFor: 'scrolling' stamp: 'ge 9/6/2006 19:02'!
numSelectionsInView
	"Answer the scroller's height based on the average number of submorphs."
	
	(scroller submorphCount > 0) ifFalse:[ ^0 ].
	
	"ugly hack, due to code smell.
	PluggableListMorph added another level of indirection, 
	There is always only one submorph - a LazyListMorph which holds the actual list,
	but TransformMorph doesn't know that and we are left with a breach of interface.
	
	see vUnadjustedScrollRange for another bad example."
		
	^scroller numberOfItemsPotentiallyInViewWith: (scroller 
												submorphs last getListSize).! !

!PluggableListMorph methodsFor: 'scrolling' stamp: 'mt 5/19/2015 16:19'!
offsetToShow: aRectangle
	"Due to the approximation of the horizontal scroll range and the primary interaction along the vertical axis, we snap back to 0 on the horizontal axis."
	
	^ 0 @ (super offsetToShow: aRectangle) y! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'mt 7/26/2016 11:28'!
on: anObject list: getListSel selected: getSelectionSel changeSelected: setSelectionSel menu: getMenuSel keystroke: keyActionSel 
	self model: anObject.
	getListSelector := getListSel.
	getIndexSelector := getSelectionSel.
	setIndexSelector := setSelectionSel.
	getMenuSelector := getMenuSel.
	keystrokeActionSelector := keyActionSel.
	autoDeselect := true.
	self updateList.
	self selectionIndex: self getCurrentSelectionIndex.
	self initForKeystrokes! !

!PluggableListMorph methodsFor: 'drag and drop' stamp: 'ls 6/23/2001 00:01'!
potentialDropItem
	"return the item that the most recent drop hovered over, or nil if there is no potential drop target"
	self potentialDropRow = 0 ifTrue: [ ^self ].
	^self getListItem: self potentialDropRow! !

!PluggableListMorph methodsFor: 'drag and drop' stamp: 'ls 6/23/2001 00:10'!
potentialDropRow
	"return the row of the item that the most recent drop hovered over, or 0 if there is no potential drop target"
	^potentialDropRow ifNil: [ 0 ].
! !

!PluggableListMorph methodsFor: 'model access' stamp: 'MAD 3/13/2010 14:44'!
previewKeystroke: event
	"Let the model decide if it's going to handle the event for us"

	^ keystrokePreviewSelector 
		ifNil: [ false ]
		ifNotNil: [ model perform: keystrokePreviewSelector with: event ]! !

!PluggableListMorph methodsFor: 'filtering' stamp: 'cmm 8/21/2011 23:25'!
removeFilter
	lastKeystrokes := String empty.
	list := nil! !

!PluggableListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:20'!
removeObsoleteEventHandlers
	scroller submorphs do:[:m|
		m eventHandler: nil; highlightForMouseDown: false; resetExtension].! !

!PluggableListMorph methodsFor: 'scroll cache' stamp: 'sps 4/3/2005 15:29'!
resetHScrollRange

	hScrollRangeCache := nil.
	self deriveHScrollRange.
! !

!PluggableListMorph methodsFor: 'scroll cache' stamp: 'sps 4/3/2005 15:29'!
resetHScrollRangeIfNecessary

	hScrollRangeCache ifNil: [ ^self deriveHScrollRange ].

	(list isNil or: [list isEmpty]) 
		ifTrue:[^hScrollRangeCache := Array with: 0 with: 0 with: 0 with: 0 with: 0].

"Make a guess as to whether the scroll ranges need updating based on whether the size, first item, or last item of the list has changed"
	(
		(hScrollRangeCache third == list size) and: [
		(hScrollRangeCache fourth == list first) and: [
		(hScrollRangeCache fifth == list last)
	]])
		ifFalse:[self deriveHScrollRange].

! !

!PluggableListMorph methodsFor: 'drag and drop' stamp: 'ls 6/23/2001 00:01'!
resetPotentialDropRow
	potentialDropRow ifNotNil: [
	potentialDropRow ~= 0 ifTrue: [
		potentialDropRow := 0.
		self changed. ] ]! !

!PluggableListMorph methodsFor: 'selection' stamp: 'kb 5/22/2013 16:24'!
rowAboutToBecomeSelected: anInteger

	self listMorph preSelectedRow: anInteger! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'ls 5/17/2001 20:31'!
rowAtLocation: aPoint
	"Return the row at the given point or 0 if outside"
	| pointInListMorphCoords |
	pointInListMorphCoords := (self scroller transformFrom: self) transform: aPoint.
	^self listMorph rowAtLocation: pointInListMorphCoords.! !

!PluggableListMorph methodsFor: 'geometry' stamp: 'ls 5/17/2001 21:01'!
scrollDeltaHeight
	"Return the increment in pixels which this pane should be scrolled."
	^ self font height! !

!PluggableListMorph methodsFor: 'geometry' stamp: 'sps 3/9/2004 17:31'!
scrollDeltaWidth
"A guess -- assume that the width of a char is approx 1/2 the height of the font"
	^ self scrollDeltaHeight // 2

! !

!PluggableListMorph methodsFor: 'selection' stamp: 'cmm 8/21/2011 23:44'!
scrollSelectionIntoView
	"make sure that the current selection is visible"
	| row |
	row := self uiIndexFor: self getCurrentSelectionIndex.
	row = 0 ifTrue: [ ^ self ].
	self scrollToShow: (self listMorph drawBoundsForRow: row)! !

!PluggableListMorph methodsFor: 'selection' stamp: 'ls 8/19/2001 14:20'!
selectedMorph
	"this doesn't work with the LargeLists patch!!  Use #selectionIndex and #selection instead."
	^self scroller submorphs at: self selectionIndex! !

!PluggableListMorph methodsFor: 'selection' stamp: 'tpr 12/31/2013 15:57'!
selectedMorph: aMorph 
	"this shouldn't be used any longer. Only sent by implemetations of #list: such as PluggableListMorph, SimpleHierarchicalListMorph etc"

	self isThisEverCalled .
	true ifTrue: [^self]! !

!PluggableListMorph methodsFor: 'selection' stamp: 'ls 8/19/2001 14:29'!
selection 
	self selectionIndex = 0 ifTrue: [ ^nil ].
	list ifNotNil: [ ^list at: self selectionIndex ].
	^ self getListItem: self selectionIndex! !

!PluggableListMorph methodsFor: 'selection' stamp: 'ls 6/22/2001 22:49'!
selectionIndex
	"return the index we have currently selected, or 0 if none"
	^self listMorph selectedRow ifNil: [ 0 ]! !

!PluggableListMorph methodsFor: 'selection' stamp: 'ls 6/22/2001 22:50'!
selectionIndex: index
	"Called internally to select the index-th item."
	| row |
	self unhighlightSelection.
	row := index ifNil: [ 0 ].
	row := row min: self getListSize.  "make sure we don't select past the end"
	self listMorph selectedRow: row.
	self highlightSelection.
	self scrollSelectionIntoView.! !

!PluggableListMorph methodsFor: 'selection' stamp: 'ls 6/22/2001 22:49'!
selection: item
	"Called from outside to request setting a new selection."

	self selectionIndex: (self getList indexOf: item)! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'mt 7/25/2016 14:12'!
setDefaultParameters

	super setDefaultParameters.
	
	self
		font: (self userInterfaceTheme font ifNil: [TextStyle defaultFont]);
		textColor: (self userInterfaceTheme textColor ifNil: [Color black]).
		
	self setListParameters.! !

!PluggableListMorph methodsFor: 'selection' stamp: 'sw 10/30/2000 11:16'!
setGetListSelector: sel
	"Set the the receiver's getListSelector as indicated.  For access via scripting"

	getListSelector := sel! !

!PluggableListMorph methodsFor: 'menus' stamp: 'tween 9/8/2007 14:56'!
setListFont
	"set the font for the list"

	Preferences 
		chooseFontWithPrompt: 'Choose the font for this list' translated 
		andSendTo: self 
		withSelector: #font: 
		highlightSelector: #font! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'mt 7/30/2016 08:28'!
setListParameters

	self listMorph
		selectionColor: (self userInterfaceTheme selectionColor ifNil: [Color r: 0.72 g: 0.72 b: 0.9]);
		multiSelectionColor: (self userInterfaceTheme multiSelectionColor ifNil: [(Color r: 0.72 g: 0.72 b: 0.9) lighter]);
		setProperty: #selectionTextColor
			toValue: (self userInterfaceTheme selectionTextColor ifNil: [Color black]);
		filterColor: (self userInterfaceTheme filterColor ifNil: [Color yellow paler]);
		setProperty: #filterTextColor
			toValue: (self userInterfaceTheme filterTextColor ifNil: [Color black]);
		
		setProperty: #preSelectionColor
			toValue: ((self userInterfaceTheme preSelectionModifier ifNil: [ [:c | Color gray: 0.9] ]) value: self listMorph selectionColor asColor);
		setProperty: #hoverColor
			toValue: ((self userInterfaceTheme hoverSelectionModifier ifNil: [ [:c | c darker alpha: 0.3] ]) value: self listMorph selectionColor asColor)! !

!PluggableListMorph methodsFor: 'selection' stamp: 'cmm 8/21/2011 22:04'!
setSelectedMorph: aMorph 
	self changeModelSelection: (self modelIndexFor: (scroller submorphs indexOf: aMorph))! !

!PluggableListMorph methodsFor: 'model access' stamp: 'mt 4/19/2015 07:01'!
specialKeyPressed: asciiValue
	"A special key with the given ascii-value was pressed; dispatch it"
	| oldSelection nextSelection max howManyItemsShowing |
	(#(8 13) includes: asciiValue) ifTrue:
		[ "backspace key - clear the filter, restore the list with the selection" 
		model okToChange ifFalse: [^ self].
		self removeFilter.
		priorSelection ifNotNil:
			[ | prior |
			prior := priorSelection.
			priorSelection := self getCurrentSelectionIndex.
			asciiValue = 8 ifTrue: [ self changeModelSelection: prior ] ].
		^ self updateList ].
	asciiValue = 27 ifTrue: 
		[" escape key"
		^ ActiveEvent shiftPressed
			ifTrue:
				[ActiveWorld putUpWorldMenuFromEscapeKey]
			ifFalse:
				[self yellowButtonActivity: false]].

	max := self maximumSelection.
	max > 0 ifFalse: [^ self].
	nextSelection := oldSelection := self selectionIndex.
	asciiValue = 31 ifTrue: 
		[" down arrow"
		nextSelection := oldSelection + 1.
		nextSelection > max ifTrue: [nextSelection := 1]].
	asciiValue = 30 ifTrue: 
		[" up arrow"
		nextSelection := oldSelection - 1.
		nextSelection < 1 ifTrue: [nextSelection := max]].
	asciiValue = 1 ifTrue:
		[" home"
		nextSelection := 1].
	asciiValue = 4 ifTrue:
		[" end"
		nextSelection := max].
	howManyItemsShowing := self numSelectionsInView.
	asciiValue = 11 ifTrue:
		[" page up"
		nextSelection := 1 max: oldSelection - howManyItemsShowing].
	asciiValue = 12 ifTrue:
		[" page down"
		nextSelection := oldSelection + howManyItemsShowing min: max].
	model okToChange ifFalse: [^ self].
	"No change if model is locked"
	oldSelection = nextSelection ifTrue: [^ self flash].
	^ self changeModelSelection: (self modelIndexFor: nextSelection)! !

!PluggableListMorph methodsFor: 'drag and drop' stamp: 'mt 11/5/2015 08:58'!
startDrag: evt 

	| item itemMorph |
	evt hand hasSubmorphs ifTrue: [^ self].
	self model okToChange ifFalse: [^ self].

	"Ensure selection to save additional click."
	(self modelIndexFor: (self rowAtLocation: evt position)) in: [:evtIndex |
		self selectionIndex = evtIndex ifFalse: [self changeModelSelection: evtIndex]].

	item := self selection ifNil: [^ self].
	itemMorph := StringMorph contents: item asStringOrText.
	
	[ "Initiate drag."
		(self model dragPassengerFor: itemMorph inMorph: self) ifNotNil: [:passenger | | ddm |
			ddm := (self valueOfProperty: #dragTransferClass ifAbsent: [TransferMorph]) withPassenger: passenger from: self.
			ddm dragTransferType: (self model dragTransferTypeForMorph: self).
			ddm updateFromUserInputEvent: evt.
			self model dragStartedFor: itemMorph transferMorph: ddm.
			evt hand grabMorph: ddm]
	] ensure: [
		Cursor normal show.
		evt hand releaseMouseFocus: self].! !

!PluggableListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:20'!
startDrag: evt onItem: itemMorph 
	self removeObsoleteEventHandlers.! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'nk 5/16/2003 14:41'!
textColor
	"Answer my default text color."
	^self valueOfProperty: #textColor ifAbsent: [ Color black ]
! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'ls 2/5/2004 18:02'!
textColor: aColor
	"Set my default text color."
	self setProperty: #textColor toValue: aColor.
	self listMorph color: aColor.! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'MAD 10/6/2009 19:54'!
textHighlightColor
	"Answer my default text highlight color."
	^self valueOfProperty: #textHighlightColor ifAbsent: [ self textColor negated ].
! !

!PluggableListMorph methodsFor: 'initialization' stamp: 'nk 5/16/2003 14:37'!
textHighlightColor: aColor
	"Set my default text highlight color."
	self setProperty: #textHighlightColor toValue: aColor.
! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'cmm 8/20/2013 16:23'!
topVisibleRowIndex
	^ self rowAtLocation: self topLeft+(3@3)! !

!PluggableListMorph methodsFor: 'selection' stamp: 'cmm 8/21/2011 22:40'!
uiIndexFor: modelIndex 
	"The model does not know anything about the receiver's filtering.  Answer the index in my (possibly filtered) list for modelIndex."
	(modelIndex > 0 and: [ self hasFilter ])
		ifTrue:
			[ | selectedItem |
			selectedItem := self getFullList at: modelIndex.
			(list ifNil: [ self getList ]) withIndexDo:
				[ : eachMatchingItem : n | eachMatchingItem = selectedItem ifTrue: [ ^ n ] ].
			^ 0 ]
		ifFalse: [ ^ modelIndex ]! !

!PluggableListMorph methodsFor: 'drawing' stamp: 'ls 5/17/2001 20:53'!
unhighlightSelection
! !

!PluggableListMorph methodsFor: 'updating' stamp: 'mt 4/8/2015 20:16'!
updateList
	| index |
	"the list has changed -- update from the model"
	self listMorph listChanged.
	self setScrollDeltas.
	index := self getCurrentSelectionIndex.
	self resetPotentialDropRow.
	self selectionIndex: (self uiIndexFor: index).
! !

!PluggableListMorph methodsFor: 'updating' stamp: 'mt 4/19/2015 13:24'!
update: aSymbol 
	"Refer to the comment in View|update:."
	aSymbol == getListSelector ifTrue:
		[ self updateList.
		^ self ].
	aSymbol == getIndexSelector ifTrue:
		[ | uiIndex modelIndex |
		uiIndex := self uiIndexFor: (modelIndex := self getCurrentSelectionIndex).
		self selectionIndex:
			(uiIndex = 0
				ifTrue:
					[ "The filter is preventing us from selecting the item we want - remove it."
					(modelIndex > 0 and: [list notNil and: [list size > 0]]) ifTrue: [ self removeFilter; updateList ].
					modelIndex ]
				ifFalse: [ uiIndex ]).
		^ self ]! !

!PluggableListMorph methodsFor: 'updating' stamp: 'mt 6/1/2016 16:37'!
update: aSymbol with: arg1

	(aSymbol == #inputRequested and: [getIndexSelector == arg1 or: [setIndexSelector == arg1]])
		ifTrue: [self activeHand newKeyboardFocus: self].
		
	super update: aSymbol with: arg1.! !

!PluggableListMorph methodsFor: 'debug and other' stamp: 'bf 6/25/2011 13:36'!
userString
	"Do I have a text string to be searched on?"

	^ String streamContents: [:strm |
		1 to: self getListSize do: [:i |
			"must use asStringOrText because that's what the drawing uses, too"
			strm nextPutAll: (self getListItem: i) asStringOrText; cr]]! !

!PluggableListMorph methodsFor: 'updating' stamp: 'cmm 8/18/2013 13:23'!
verifyContents
	"Verify the contents of the receiver, reconstituting if necessary.  Called whenever window is reactivated, to react to possible structural changes.  Also called periodically in morphic if the smartUpdating preference is true"
	| newList existingSelection anIndex oldList |
	oldList := list ifNil: [ #() ].
	newList := self getList.
	oldList = newList ifTrue: [ ^ self ].
	existingSelection :=  oldList at: self selectionIndex ifAbsent: [ nil ].
	self updateList.
	(existingSelection notNil and: [(anIndex := self getFullList indexOf: existingSelection asStringOrText ifAbsent: [nil]) notNil])
		ifTrue:
			[model noteSelectionIndex: anIndex for: getListSelector.
			self selectionIndex: anIndex]
		ifFalse:
			[self changeModelSelection: 0]! !

!PluggableListMorph methodsFor: 'accessing' stamp: 'cmm 9/9/2013 21:14'!
visibleList
	^ list isEmptyOrNil
		ifTrue: [ Array empty ]
		ifFalse:
			[ list
				copyFrom: self topVisibleRowIndex
				to: (self bottomVisibleRowIndex min: list size) ]! !

!PluggableListMorph methodsFor: 'scrolling' stamp: 'sps 12/26/2002 13:36'!
vUnadjustedScrollRange
	"Return the height extent of the receiver's submorphs."
	(scroller submorphs size > 0) ifFalse:[ ^0 ].
	^(scroller submorphs last fullBounds bottom)
! !

!PluggableListMorph methodsFor: 'drag and drop' stamp: 'panda 4/25/2000 18:26'!
wantsDroppedMorph: aMorph event: anEvent 
	^ self model wantsDroppedMorph: aMorph event: anEvent inMorph: self! !

PluggableListMorph subclass: #PluggableMultiColumnListMorph
	instanceVariableNames: 'lists'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!PluggableMultiColumnListMorph commentStamp: '<historical>' prior: 0!
This morph can be used to show a list having multiple columns,  The columns are self width sized to make the largest entry in each list fit.  In some cases the pane may then be too narrow.

Use it like a regular PluggableListMorph except pass in an array of lists instead of a single list.

There are base assumptions made here that each list in the array of lists is the same size.

Also, the highlight color for the selection is easy to modify in the #highlightSelection method.  I used blue
when testing just to see it work.!

PluggableMultiColumnListMorph subclass: #PluggableMultiColumnListMorphByItem
	instanceVariableNames: 'itemList'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!PluggableMultiColumnListMorphByItem methodsFor: 'model access' stamp: 'ls 8/19/2001 14:57'!
changeModelSelection: anInteger 
	"Change the model's selected item to be the one at the given index."
	| item |
	setIndexSelector
		ifNotNil: [item := anInteger = 0
						ifFalse: [list first at: anInteger].
			model perform: setIndexSelector with: item].
	self update: getIndexSelector! !

!PluggableMultiColumnListMorphByItem methodsFor: 'model access' stamp: 'ls 8/19/2001 15:11'!
getCurrentSelectionIndex
	"Answer the index of the current selection."
	| item |
	getIndexSelector == nil
		ifTrue: [^ 0].
	item := model perform: getIndexSelector.

	^ list first
		findFirst: [:x | x  = item]! !

!PluggableMultiColumnListMorphByItem methodsFor: 'initialization' stamp: 'ls 8/19/2001 14:55'!
list: arrayOfStrings 
	"Set the receivers items to be the given list of strings."
	"Note: the instance variable 'items' holds the original list.  
	The instance variable 'list' is a paragraph constructed from  
	this list."
"NO LONGER TRUE.  list is a real list, and listItems is obsolete."
self isThisEverCalled .
	itemList := arrayOfStrings first.
	^ super list: arrayOfStrings! !

!PluggableMultiColumnListMorph methodsFor: 'model access' stamp: 'ls 11/14/2002 13:13'!
basicKeyPressed: aChar
	"net supported for multi-column lists; which column should be used?!!  The issue is that the base class implementation uses getList expecting a single collectino to come back instead of several of them"
	^self! !

!PluggableMultiColumnListMorph methodsFor: 'initialization' stamp: 'sbw 12/2/2000 08:37'!
calculateColumnOffsetsFrom: maxWidths
	| offsets previous current |
	offsets := Array new: maxWidths size.
	1
		to: offsets size
		do: [:indx | offsets at: indx put: (maxWidths at: indx)
					+ 10].
	2
		to: offsets size
		do: [:indx | 
			previous := offsets at: indx - 1.
			current := offsets at: indx.
			current := previous + current.
			offsets at: indx put: current].
	^offsets
! !

!PluggableMultiColumnListMorph methodsFor: 'initialization' stamp: 'sbw 12/2/2000 08:36'!
calculateColumnWidthsFrom: arrayOfMorphs 
	| maxWidths |
	maxWidths := Array new: arrayOfMorphs size - 1.
	1
		to: maxWidths size
		do: [:idx | maxWidths at: idx put: 0].
	1
		to: maxWidths size
		do: [:idx | (arrayOfMorphs at: idx)
				do: [:mitem | mitem width
							> (maxWidths at: idx)
						ifTrue: [maxWidths at: idx put: mitem width]]].
	^maxWidths! !

!PluggableMultiColumnListMorph methodsFor: 'initialization' stamp: 'ls 5/15/2001 22:22'!
createMorphicListsFrom: arrayOfLists 
	| array |

	array := Array new: arrayOfLists size.
	1 to: arrayOfLists size do: [:arrayIndex |
		array at: arrayIndex put: (
			(arrayOfLists at: arrayIndex) collect: [:item | item isText
						ifTrue: [StringMorph
								contents: item
								font: self font
								emphasis: (item emphasisAt: 1)]
						ifFalse: [StringMorph contents: item font: self font]])
		].
	^array! !

!PluggableMultiColumnListMorph methodsFor: 'model access' stamp: 'ls 7/12/2001 23:24'!
getList
	"fetch and answer the lists to be displayed"
	getListSelector == nil ifTrue: [^ #()].
	list := model perform: getListSelector.
	list == nil ifTrue: [^ #()].
	list := list collect: [ :column | column collect: [ :item | item asStringOrText ] ].
	^ list! !

!PluggableMultiColumnListMorph methodsFor: 'accessing' stamp: 'ls 5/18/2001 10:32'!
getListRow: row
	"return the strings that should appear in the requested row"
	getListElementSelector ifNotNil: [ ^model perform: getListElementSelector with: row ].
	^self getList collect: [ :l | l at: row ]! !

!PluggableMultiColumnListMorph methodsFor: 'accessing' stamp: 'ls 5/17/2001 23:03'!
getListSize
	| l |
	getListSizeSelector ifNotNil: [ ^model perform: getListSizeSelector ].

	l := self getList.
	l isEmpty ifTrue: [ ^ 0 ].
	^l first size! !

!PluggableMultiColumnListMorph methodsFor: 'selection' stamp: 'ls 5/16/2001 22:24'!
highlightSelection
^self! !

!PluggableMultiColumnListMorph methodsFor: 'accessing' stamp: 'nk 4/5/2001 23:18'!
itemFromPoint: aPoint
	"Return the list element (morph) at the given point or nil if outside"
	| ptY |
	scroller hasSubmorphs ifFalse:[^nil].
	(scroller fullBounds containsPoint: aPoint) ifFalse:[^nil].
	ptY := (scroller firstSubmorph point: aPoint from: self) y.
	"note: following assumes that submorphs are vertical, non-overlapping, and ordered"
	scroller firstSubmorph top > ptY ifTrue:[^nil].
	scroller lastSubmorph bottom < ptY ifTrue:[^nil].
	"now use binary search"
	^scroller submorphThat: [ :item | item top <= ptY and:[item bottom >= ptY] ] ifNone: [].
! !

!PluggableMultiColumnListMorph methodsFor: 'initialization' stamp: 'sbw 12/2/2000 08:38'!
layoutMorphicLists: arrayOfMorphs 
	| maxWidths offsets locs h |
	maxWidths := self calculateColumnWidthsFrom: arrayOfMorphs.
	offsets := self calculateColumnOffsetsFrom: maxWidths.
	locs := Array new: arrayOfMorphs size.
	locs at: 1 put: 0 @ 0.
	2
		to: locs size
		do: [:indx | locs at: indx put: (offsets at: indx - 1)
					@ 0].
	h := arrayOfMorphs first first height.
	1
		to: arrayOfMorphs size
		do: [:indx | (arrayOfMorphs at: indx)
				do: [:morphItem | 
					morphItem
						bounds: ((locs at: indx)
								extent: 9999 @ h).
					locs at: indx put: (locs at: indx)
							+ (0 @ h)]]! !

!PluggableMultiColumnListMorph methodsFor: 'accessing' stamp: 'ls 5/17/2001 20:01'!
listMorphClass
	^MulticolumnLazyListMorph! !

!PluggableMultiColumnListMorph methodsFor: 'initialization' stamp: 'ls 5/17/2001 21:16'!
list: arrayOfLists 
	| listOfStrings |
	lists := arrayOfLists.
	scroller removeAllMorphs.
	listOfStrings := arrayOfLists == nil
				ifTrue: [Array new]
				ifFalse: [
					arrayOfLists isEmpty ifFalse: [
					arrayOfLists at: 1]].
	list := listOfStrings
				ifNil: [Array new].
	self listMorph listChanged..

	self setScrollDeltas.
	scrollBar setValue: 0.0! !

!PluggableMultiColumnListMorph methodsFor: 'selection' stamp: 'ls 5/16/2001 22:23'!
unhighlightSelection
^self! !

ScrollPane subclass: #PluggableScrollPane
	instanceVariableNames: 'morph morphClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!PluggableScrollPane class methodsFor: 'instance creation' stamp: 'mt 6/15/2014 07:07:16.698'!
onClass: morphClass

	^ self new
		morphClass: morphClass;
		updateMorph;
		yourself! !

!PluggableScrollPane class methodsFor: 'instance creation' stamp: 'mt 6/15/2014 07:00:05.055'!
on: morph

	^ self new
		morph: morph;
		yourself! !

!PluggableScrollPane methodsFor: 'accessing' stamp: 'mt 6/15/2014 06:57:04.763'!
morph

	^ morph ifNil: [
		self morph: self morphClass new.
		morph]! !

!PluggableScrollPane methodsFor: 'accessing' stamp: 'mt 6/15/2014 06:51:48.148'!
morphClass

	^ morphClass ifNil: [Morph]! !

!PluggableScrollPane methodsFor: 'accessing' stamp: 'mt 6/15/2014 06:52:13.517'!
morphClass: aMorphClass

	morphClass := aMorphClass.! !

!PluggableScrollPane methodsFor: 'accessing' stamp: 'mt 6/5/2016 13:20'!
morph: morphToScroll

	morph ifNotNil: [:m | m removeDependent: self].

	morphToScroll topLeft: 0@0.
	morph := morphToScroll.
	morphClass := morphToScroll class.

	morph addDependent: self.

	self scroller
		removeAllMorphs;
		addMorph: morph.
	
	self updateMorph.! !

!PluggableScrollPane methodsFor: 'updating' stamp: 'mt 6/5/2016 13:19'!
updateMorph

	self morph fullBounds.
	self setScrollDeltas.! !

!PluggableScrollPane methodsFor: 'updating' stamp: 'mt 6/5/2016 13:19'!
update: aspect

	aspect == #submorphs
		ifTrue: [self updateMorph].

	super update: aspect.! !

ScrollPane subclass: #PluggableTextMorph
	instanceVariableNames: 'textMorph getTextSelector setTextSelector getSelectionSelector hasUnacceptedEdits askBeforeDiscardingEdits selectionInterval hasEditingConflicts editTextSelector wantsWrapBorder'
	classVariableNames: 'AdornmentCache SimpleFrameAdornments SoftLineWrap VisualWrapBorder VisualWrapBorderLimit'
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

PluggableTextMorph subclass: #AcceptableCleanTextMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!AcceptableCleanTextMorph methodsFor: 'menu commands' stamp: 'dgd 2/21/2003 22:50'!
accept
	"Overridden to allow accept of clean text"

	| textToAccept ok |
	textToAccept := textMorph asText.
	ok := setTextSelector isNil or: 
					[setTextSelector numArgs = 2 
						ifTrue: 
							[model 
								perform: setTextSelector
								with: textToAccept
								with: self]
						ifFalse: [model perform: setTextSelector with: textToAccept]].
	ok 
		ifTrue: 
			[self setText: self getText.
			self hasUnacceptedEdits: false]! !

PluggableTextMorph subclass: #PluggableTextMorphWithModel
	instanceVariableNames: 'myContents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!PluggableTextMorphWithModel class methodsFor: 'scripting' stamp: 'sw 10/30/2000 11:14'!
authoringPrototype
	"Answer an instance of the receiver suitable for placing in a parts bin"

	| proto |
	proto := super authoringPrototype.
	proto color: (Color r: 0.972 g: 0.972 b: 0.662).
	^ proto! !

!PluggableTextMorphWithModel methodsFor: 'submorphs-add/remove' stamp: 'sw 10/26/2000 14:39'!
delete
	"Delete the receiver.  Since I have myself as a dependent, I need to remove it. which is odd in itself.  Also, the release of dependents will seemingly not be done if the *container* of the receiver is deleted rather than the receiver itself, a further problem"

	self removeDependent: self.
	super delete! !

!PluggableTextMorphWithModel methodsFor: 'contents' stamp: 'mjg 12/3/1999 11:57'!
getMyText
	^myContents! !

!PluggableTextMorphWithModel methodsFor: 'initialization' stamp: 'dgd 2/14/2003 18:25'!
initialize
	"initialize the state of the receiver"
	super initialize.
	self
		on: self
		text: #getMyText
		accept: #setMyText:
		readSelection: nil
		menu: nil! !

!PluggableTextMorphWithModel methodsFor: 'contents' stamp: 'ar 1/3/2010 19:52'!
newTextContents: stringOrText
	"Accept new text contents."

	| newText myText |
	"Just underway; trying to make this work like TextMorph does, but not quite there yet."

	newText := stringOrText asText.
	(myText := textMorph text) = newText ifTrue: [^ self].  "No substantive change"
	
	self world ifNotNil:
		[self world startSteppingSubmorphsOf: self ].
! !

!PluggableTextMorphWithModel methodsFor: 'contents' stamp: 'mjg 12/3/1999 11:59'!
setMyText: someText
	myContents := someText.
	^true.! !

!PluggableTextMorph class methodsFor: 'frame adornments' stamp: 'ar 2/26/2010 16:23'!
adornmentCache
	"Cache for frame adornments"

	^AdornmentCache ifNil:[AdornmentCache := Dictionary new].! !

!PluggableTextMorph class methodsFor: 'frame adornments' stamp: 'mt 7/24/2016 14:37'!
adornmentWithColor: aColor
	"Create and return a frame adornment with the given color"

	| size box form fillStyle |
	^self adornmentCache at: aColor ifAbsentPut:[
		size := 20. 
		box := 0@0 extent: size asPoint.
		form := Form extent: size@size depth: 32.
		fillStyle := MenuMorph gradientMenu ifFalse: [SolidFillStyle color: aColor] ifTrue: [
			(GradientFillStyle ramp: {
				0.0->(aColor alpha: 0.01).
				0.8->aColor.
				1.0->aColor})
				origin: box topRight - (size@0);
				direction: (size @ size negated) // 4;
				radial: false].
		form getCanvas drawPolygon:  {
			box topRight. 
			box topRight + (0@size). 
			box topRight - (size@0)
		} fillStyle: fillStyle.
		form].
! !

!PluggableTextMorph class methodsFor: 'frame adornments' stamp: 'ar 2/26/2010 16:23'!
flushAdornmentCache
	"Cache for frame adornments"

	AdornmentCache := nil! !

!PluggableTextMorph class methodsFor: 'as yet unclassified' stamp: 'di 4/7/98 16:03'!
on: anObject text: getTextSel accept: setTextSel

	^ self on: anObject
		text: getTextSel
		accept: setTextSel
		readSelection: nil
		menu: nil! !

!PluggableTextMorph class methodsFor: 'as yet unclassified' stamp: 'di 4/7/98 16:03'!
on: anObject text: getTextSel accept: setTextSel readSelection: getSelectionSel menu: getMenuSel

	^ self new on: anObject
		text: getTextSel
		accept: setTextSel
		readSelection: getSelectionSel
		menu: getMenuSel! !

!PluggableTextMorph class methodsFor: 'frame adornments' stamp: 'ar 3/1/2010 19:43'!
simpleFrameAdornments
	<preference: 'Simple Frame Adornments'
		category: 'Morphic'
		description: 'When true, use a simple rectangular feedback for indicating unsaved changes in text editors'
		type: #Boolean>
	^SimpleFrameAdornments ifNil:[false]! !

!PluggableTextMorph class methodsFor: 'frame adornments' stamp: 'ar 3/1/2010 19:40'!
simpleFrameAdornments: aBool
	"Sets the simpleFrameAdornment preference
		PluggableTextMorph simpleFrameAdornments: true.
		PluggableTextMorph simpleFrameAdornments: false.
	"
	SimpleFrameAdornments := aBool
! !

!PluggableTextMorph class methodsFor: 'preferences' stamp: 'mt 4/27/2015 15:25'!
softLineWrap
	<preference: 'Use soft line wrap'
		categoryList: #(scrolling editing)
		description: 'Wrap text lines to avoid horizontal scrolling.'
		type: #Boolean>
	^ SoftLineWrap ifNil: [true]! !

!PluggableTextMorph class methodsFor: 'preferences' stamp: 'mt 4/27/2015 15:28'!
softLineWrap: aBoolean

	aBoolean == SoftLineWrap ifTrue: [^ self].
	SoftLineWrap := aBoolean.
	PluggableTextMorph allSubInstancesDo: [:m |
		m text lineCount > 1 ifTrue: [m wrapFlag: aBoolean]].! !

!PluggableTextMorph class methodsFor: 'preferences' stamp: 'mt 8/5/2016 08:42'!
themeProperties

	^ super themeProperties, {
		{ #font. 'Fonts'. 'Font for text if not styled.' }.
		{ #textColor. 'Colors'. 'Color for text if not styled.' }.		
		{ #caretColor. 'Colors'. 'The color of the text cursor.' }.
		{ #selectionColor. 'Colors'. 'The color of the text selection.' }.
		{ #unfocusedSelectionModifier. 'Colors'. 'How to derive the text selection color if not focused.' }.
		
		{ #adornmentReadOnly. 'Color'. 'How to indicate read-only contents.' }.
		{ #adornmentRefuse. 'Color'. 'How to indicate that the model refuses to accept.' }.
		{ #adornmentConflict. 'Color'. 'How to indicate that there are editing conflicts.' }.
		{ #adornmentDiff. 'Color'. 'How to indicate that the model wants diff feedback.' }.
		{ #adornmentNormalEdit. 'Color'. 'How to indicate that there are unaccepted edits.' }.
		{ #adornmentDiffEdit. 'Color'. 'How to indicate that there are unaccepted edits in a diff view.' }.
		
		{ #wrapBorderColorModifier. 'Color'. 'How to indicate a specific wrap border.' }.
		{ #frameAdornmentWidth. 'Geometry'. 'Width of simple frame adornments.' }.
	}! !

!PluggableTextMorph class methodsFor: 'preferences' stamp: 'mt 4/27/2015 17:12'!
visualWrapBorder
	<preference: 'Show wrap border in code panes.'
		categoryList: #(editing visuals performance)
		description: 'Show a visual border after a specific amout of characters. Makes sense for monospaced fonts.'
		type: #Boolean>
	^ VisualWrapBorder ifNil: [false]! !

!PluggableTextMorph class methodsFor: 'preferences' stamp: 'mt 4/27/2015 16:53'!
visualWrapBorderLimit
	<preference: 'Wrap border limit'
		categoryList: #(editing visuals performance)
		description: 'Amount of characters after the border should be drawn.'
		type: #Number>
	^ VisualWrapBorderLimit ifNil: [80]! !

!PluggableTextMorph class methodsFor: 'preferences' stamp: 'mt 4/27/2015 16:53'!
visualWrapBorderLimit: aNumber

	VisualWrapBorderLimit := aNumber asInteger.! !

!PluggableTextMorph class methodsFor: 'preferences' stamp: 'mt 4/27/2015 16:52'!
visualWrapBorder: aBoolean

	VisualWrapBorder := aBoolean.! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'mt 5/13/2016 15:01'!
accept 
	"Inform the model of text to be accepted, and return true if OK."

	| priorSelection priorScrollerOffset |

	(self canDiscardEdits and: [(self hasProperty: #alwaysAccept) not])
		ifTrue: [^ self flash].

	self hasEditingConflicts ifTrue: [
		(self confirm: 'Caution!! This method may have been\changed elsewhere since you started\editing it here.  Accept anyway?' withCRs translated) ifFalse: [^ self flash]].

	priorSelection := self selectionInterval copy.
	priorScrollerOffset := scroller offset copy.
	
	self acceptTextInModel == true
		ifFalse: [^ self "something went wrong"].
		
	self setText: self getText.
	self hasUnacceptedEdits: false.

	(model dependents
		detect: [:dep | (dep isKindOf: PluggableTextMorph) and: [dep getTextSelector == #annotation]]
		ifNone: [nil])
			ifNotNil: [:aPane | model changed: #annotation].

	"Update the model's internal caches. Note that this is specific to CodeHolder and the stepping it uses for updating. We have to trigger this here manually to avoid that the next step message destroys our selection and scrolling offset."
	(model respondsTo: #updateCodePaneIfNeeded)
		ifTrue: [model updateCodePaneIfNeeded].
	
	"Restore prior selection:"
	scroller offset: priorScrollerOffset.
	selectionInterval := priorSelection.
	self selectFrom: priorSelection first to: priorSelection last.! !

!PluggableTextMorph methodsFor: 'layout' stamp: 'jcg 7/7/2000 11:08'!
acceptDroppingMorph: aMorph event: evt 
	"This message is sent when a morph is dropped onto a morph that has     
	agreed to accept the dropped morph by responding 'true' to the     
	wantsDroppedMorph:Event: message. The default implementation just     
	adds the given morph to the receiver."
	"Here we let the model do its work."

	self model
		acceptDroppingMorph: aMorph
		event: evt
		inMorph: self.

! !

!PluggableTextMorph methodsFor: 'initialization' stamp: 'di 9/11/1998 15:46'!
acceptOnCR: trueOrFalse
	textMorph acceptOnCR: trueOrFalse! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'mt 5/3/2015 10:53'!
acceptTextInModel
	"Inform the model that the receiver's textMorph's text should be accepted.
	Answer true if the model accepted ok, false otherwise"
	| textToAccept |

	textToAccept := textMorph contents copy.
	^setTextSelector isNil or:
		[setTextSelector numArgs = 2
			ifTrue: [model perform: setTextSelector with: textToAccept with: self]
			ifFalse: [model perform: setTextSelector with: textToAccept]]
! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 7/25/2016 11:31'!
adoptPaneColor: aColor

	super adoptPaneColor: aColor.
	
	self wrapBorderColor: ((self userInterfaceTheme wrapBorderColorModifier ifNil: [ [:c | c muchLighter alpha: 0.3] ])
								value: self borderColor).! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 5/10/1998 12:41'!
again
	self handleEdit: [textMorph editor again]! !

!PluggableTextMorph methodsFor: 'transcript' stamp: 'di 5/13/1998 14:29'!
appendEntry
	"Append the text in the model's writeStream to the editable text. "
	textMorph asText size > model characterLimit ifTrue:
		["Knock off first half of text"
		self selectInvisiblyFrom: 1 to: textMorph asText size // 2.
		self replaceSelectionWith: Text new].
	self selectInvisiblyFrom: textMorph asText size + 1 to: textMorph asText size.
	self replaceSelectionWith: model contents asText.
	self selectInvisiblyFrom: textMorph asText size + 1 to: textMorph asText size! !

!PluggableTextMorph methodsFor: 'updating' stamp: 'mt 7/24/2016 14:17'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	self textMorph releaseParagraph; paragraph.! !

!PluggableTextMorph methodsFor: 'unaccepted edits' stamp: 'di 4/20/1998 18:53'!
askBeforeDiscardingEdits: aBoolean
	"Set the flag that determines whether the user should be asked before discarding unaccepted edits."

	askBeforeDiscardingEdits := aBoolean! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'sw 4/24/2001 12:24'!
browseChangeSetsWithSelector
	"Help the user track down which change sets mention a particular selector"

	self handleEdit: [textMorph editor browseChangeSetsWithSelector]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'tk 7/13/2000 20:04'!
browseIt
	self handleEdit: [textMorph editor browseIt]! !

!PluggableTextMorph methodsFor: 'transcript' stamp: 'sma 3/15/2000 21:40'!
bsText
	self changeText: (self text copyFrom: 1 to: (self text size - 1 max: 0))! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'yo 6/23/2011 00:33'!
buttonForIt
	self handleEdit: [textMorph editor buttonForIt]! !

!PluggableTextMorph methodsFor: 'testing' stamp: 'cmm 8/25/2013 13:09'!
canBeEncroached
	"Fixed-height always report true, since they cannot be encroached."
	self layoutFrame ifNotNil: [ : frame | frame topFraction = frame bottomFraction ifTrue: [ ^ true ] ].
	^ (textMorph height+10) < self height! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'ul 12/12/2009 14:08'!
cancel
	self setText: self getText.
	self setSelection: self getSelection.
	getTextSelector == #annotation ifFalse:
		[(model dependents detect: [:dep | (dep isKindOf: PluggableTextMorph) and: [dep getTextSelector == #annotation]] ifNone: [nil]) ifNotNil:
			[:aPane | model changed: #annotation]]! !

!PluggableTextMorph methodsFor: 'dependents access' stamp: 'di 4/20/1998 18:52'!
canDiscardEdits
	"Return true if this view either has no text changes or does not care."

	^ (hasUnacceptedEdits & askBeforeDiscardingEdits) not
! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 5/10/1998 12:41'!
changeStyle
	self handleEdit: [textMorph editor changeStyle]! !

!PluggableTextMorph methodsFor: 'transcript' stamp: 'di 5/8/1998 21:22'!
changeText: aText
	"The paragraph to be edited is changed to aText."
	self setText: aText! !

!PluggableTextMorph methodsFor: 'geometry' stamp: 'cmm 4/18/2015 14:09'!
charactersOccluded
	"Let the receiver suggest to its neighboring vertical Splitter(s) that they try to optimize their position such that none of the receivers visible lines need to wrap."
	| numberOfPhysicalLines lastPhysicalLine logicalLines physicalLines |
	numberOfPhysicalLines := 0.
	lastPhysicalLine := (physicalLines := textMorph paragraph lines)
		detect:
			[ : each | numberOfPhysicalLines := numberOfPhysicalLines + 1.
			each bottom > self height ]
		ifNone: [ physicalLines last ].
	logicalLines := (self text
		copyFrom: 1
		to: lastPhysicalLine last) lineCount.
	^ numberOfPhysicalLines - logicalLines! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'sw 9/27/1999 11:57'!
chooseAlignment
	self handleEdit: [textMorph editor changeAlignment]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'tk 5/1/2001 21:37'!
classCommentsContainingIt
	self handleEdit: [textMorph editor classCommentsContainingIt]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'sw 7/31/2002 01:48'!
classNamesContainingIt
	self handleEdit: [textMorph editor classNamesContainingIt]! !

!PluggableTextMorph methodsFor: 'html' stamp: 'dgd 9/2/2005 12:41'!
copyHtml
	"put the html representation of the receiver's text into the clipboard"
	Clipboard clipboardText: self text printHtmlString! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'tk 7/13/2000 20:04'!
copySelection
	self handleEdit: [textMorph editor copySelection]! !

!PluggableTextMorph methodsFor: 'interactive error protocol' stamp: 'di 6/22/1998 15:15'!
correctFrom: start to: stop with: aString
	^ self handleEdit: [textMorph editor correctFrom: start to: stop with: aString]! !

!PluggableTextMorph methodsFor: 'interactive error protocol' stamp: 'jcg 11/5/2000 22:25'!
correctSelectionWithString: aString
	| result newPosition |

	"I can't tell if this is a hack or if it's the right thing to do."
	self setSelection: selectionInterval. 

	result := self correctFrom: selectionInterval first to: selectionInterval last with: aString.
	newPosition := selectionInterval first + aString size.
	self setSelection: (newPosition to: newPosition - 1).
	^ result! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 5/10/1998 12:41'!
cut
	self handleEdit: [textMorph editor cut]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'vb 7/29/2001 12:45'!
debugIt
	self handleEdit: [textMorph editor debugIt]! !

!PluggableTextMorph methodsFor: 'interactive error protocol' stamp: 'di 5/6/1998 15:16'!
deselect
	^ textMorph editor deselect! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'eem 6/12/2013 17:25'!
doIt
	^self handleEdit: [textMorph editor evaluateSelection]! !

!PluggableTextMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 14:27'!
drawFrameAdornmentsOn: aCanvas 
	"Include a thin red inset border for unaccepted edits, or, if the unaccepted edits are known to conflict with a change made somewhere else to the same method (typically), put a thick red frame"

	self wantsFrameAdornments ifFalse: [^ self].
	
	self readOnly ifTrue: [^ self drawFrameAdornment: (self valueOfProperty: #adornmentReadOnly ifAbsent: [Color black]) on: aCanvas].
	
	(model notNil and: [model refusesToAcceptCode])
		ifTrue: [
			"Put up feedback showing that code cannot be submitted in this state"
			^ self drawFrameAdornment: (self valueOfProperty: #adornmentRefuse ifAbsent: [Color tan]) on: aCanvas].
	
	self hasEditingConflicts
		ifTrue: [^ self drawFrameAdornment: (self valueOfProperty: #adornmentConflict ifAbsent: [Color red]) on: aCanvas].
		 
	self hasUnacceptedEdits ifTrue: [
		model wantsDiffFeedback
			ifTrue: [self drawFrameAdornment: (self valueOfProperty: #adornmentDiffEdit ifAbsent: [Color yellow]) on: aCanvas]
			ifFalse: [self drawFrameAdornment: (self valueOfProperty: #adornmentNormalEdit ifAbsent: [Color orange]) on: aCanvas].
		^ self].

	model wantsDiffFeedback
		ifTrue: [self drawFrameAdornment: (self valueOfProperty: #adornmentDiff ifAbsent: [Color green]) on: aCanvas].! !

!PluggableTextMorph methodsFor: 'drawing' stamp: 'mt 8/5/2016 08:46'!
drawFrameAdornment: aColor on: aCanvas 
	"Indicate edit status for the text editor"
	self class simpleFrameAdornments
		ifTrue:
			[ aCanvas
				frameRectangle: self innerBounds
				width: (self valueOfProperty: #frameAdornmentWidth ifAbsent: [1])
				color: aColor.
			aCanvas
				frameRectangle: (self innerBounds insetBy: 1)
				width: (self valueOfProperty: #frameAdornmentWidth ifAbsent: [1])
				color: (aColor alpha: aColor alpha / 3.0) ]
		ifFalse:
			[ | form |
			"Class-side adornment cache is currently using pre-multiplied alpha, so we need to use rule 34 which works for < 32bpp, too."
			form := self class adornmentWithColor: aColor.
			aCanvas
				image: form
				at: self innerBounds topRight - (form width @ 0)
				sourceRect: form boundingBox
				rule: 34 ]! !

!PluggableTextMorph methodsFor: 'drawing' stamp: 'mt 4/28/2015 07:55'!
drawOverlayOn: aCanvas 
	"Draw frame adornments on top of everything otherwise they will partially overlap with text selection which looks ugly."
	
	self drawWrapBorderOn: aCanvas.
	super drawOverlayOn: aCanvas. 
	self drawFrameAdornmentsOn: aCanvas.
! !

!PluggableTextMorph methodsFor: 'drawing' stamp: 'mt 7/25/2016 11:32'!
drawWrapBorderOn: aCanvas

	| offset rect |
	self wantsWrapBorder ifFalse: [^ self].
	textMorph ifNil: [^ self].
	
	offset := textMorph margins isRectangle
		ifTrue: [textMorph margins left]
		ifFalse: [textMorph margins isPoint
			ifTrue: [textMorph margins x]
			ifFalse: [textMorph margins]].
	offset := offset + ((textMorph textStyle defaultFont widthOf: $x) * self class visualWrapBorderLimit).
	offset > self width ifTrue: [^ self].
	
	rect := scroller topLeft + (offset @ 0) corner: scroller bottomRight.

	aCanvas
		fillRectangle: rect
		color: self wrapBorderColor.
	aCanvas
		line: rect topLeft
		to: rect bottomLeft
		width: self borderStyle width
		color: (self wrapBorderColor muchDarker alpha: 0.5).! !

!PluggableTextMorph methodsFor: 'initialization' stamp: 'di 5/4/1998 15:55'!
editString: aString 
	"Jam some text in.  This is treated as clean text by default."

	self setText: aString asText! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 4/7/2015 08:55'!
editTextSelector

	^ editTextSelector! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 4/7/2015 08:55'!
editTextSelector: aSymbol

	editTextSelector := aSymbol.! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 5/10/1998 12:41'!
explain
	self handleEdit: [textMorph editor explain]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'nice 3/31/2011 21:47'!
exploreIt

	
	self handleEdit:
		[textMorph editor evaluateSelectionAndDo: [:result | result explore]].! !

!PluggableTextMorph methodsFor: 'geometry' stamp: 'mt 4/15/2015 16:47'!
extent: aPoint

	super extent: aPoint.
	
	"Update vertical scroll bars because if text is wrapped."
	textMorph ifNotNil: [:tm |
		tm isAutoFit ifTrue: [self vSetScrollDelta]].! !

!PluggableTextMorph methodsFor: 'geometry' stamp: 'JW 2/21/2001 22:15'!
extraScrollRange
	^ self height // 4! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 11/2/1998 16:02'!
fileItIn
	self handleEdit: [textMorph editor fileItIn]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 5/10/1998 12:42'!
find
	self handleEdit: [textMorph editor find]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 5/10/1998 12:42'!
findAgain
	self handleEdit: [textMorph editor findAgain]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'mt 11/8/2015 12:51'!
findReplace
	self handleEdit: [textMorph editor findReplace]! !

!PluggableTextMorph methodsFor: 'initialization' stamp: 'bolot 11/2/1999 03:18'!
font: aFont
	textMorph beAllFont: aFont! !

!PluggableTextMorph methodsFor: 'model access' stamp: 'jmv 3/2/2010 16:21'!
getSelection
	"Answer the model's selection interval."

	getSelectionSelector ifNil: [^1 to: 0].	"null selection"
	^model perform: getSelectionSelector! !

!PluggableTextMorph methodsFor: 'model access' stamp: 'jmv 3/2/2010 16:22'!
getText
	"Retrieve the current model text"

	| newText |
	getTextSelector ifNil: [^Text new].
	newText := model perform: getTextSelector.
	newText ifNil: [^Text new].
	^newText shallowCopy! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'sr 4/25/2000 07:21'!
getTextSelector
	^getTextSelector! !

!PluggableTextMorph methodsFor: 'editor access' stamp: 'nice 1/10/2010 19:26'!
handleEdit: editBlock
	| result |
	textMorph editor selectFrom: selectionInterval first to: selectionInterval last;
						model: model.  "For, eg, evaluateSelection"
	result := textMorph handleEdit: editBlock.   "Update selection after edit"
	self scrollSelectionIntoView.
	^ result! !

!PluggableTextMorph methodsFor: 'unaccepted edits' stamp: 'sw 10/10/1999 22:55'!
hasEditingConflicts
	"Return true if a conflicting edit to the same code (typically) is known to have occurred after the current contents started getting edited"

	^ hasEditingConflicts == true! !

!PluggableTextMorph methodsFor: 'unaccepted edits' stamp: 'sw 10/10/1999 22:55'!
hasEditingConflicts: aBoolean

	hasEditingConflicts := aBoolean! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 4/15/2015 07:53'!
hasFocus

	^ super hasFocus or: [textMorph notNil and: [textMorph hasKeyboardFocus]]! !

!PluggableTextMorph methodsFor: 'dependents access' stamp: 'di 4/20/1998 18:56'!
hasUnacceptedEdits
	"Return true if this view has unaccepted edits."

	^ hasUnacceptedEdits! !

!PluggableTextMorph methodsFor: 'unaccepted edits' stamp: 'sw 10/10/1999 23:06'!
hasUnacceptedEdits: aBoolean
	"Set the hasUnacceptedEdits flag to the given value. "
	aBoolean == hasUnacceptedEdits ifFalse:
		[hasUnacceptedEdits := aBoolean.
		self changed].
	aBoolean ifFalse: [hasEditingConflicts := false]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 11/2/1998 16:02'!
implementorsOfIt
	self handleEdit: [textMorph editor implementorsOfIt]! !

!PluggableTextMorph methodsFor: 'initialization' stamp: 'mt 7/24/2016 14:00'!
initialize

	self initializeTextMorph.
	super initialize.

	hasUnacceptedEdits := false.
	hasEditingConflicts := false.
	askBeforeDiscardingEdits := true.

	self minimumWidth: (TextStyle defaultFont widthOf: $m) * 10.
		
	scroller addMorph: textMorph.
	
	"Reset minExtent because only now we can anser #isAutoFit correctly."
	self minimumExtent: 0@0; updateMinimumExtent.! !

!PluggableTextMorph methodsFor: 'initialization' stamp: 'mt 7/24/2016 14:01'!
initializeTextMorph

	textMorph := self textMorphClass new
		margins: (3@0 corner: 0@0);
		setEditView: self;
		autoFit: true;
		setProperty: #indicateKeyboardFocus toValue: #never;
		yourself.! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'nice 3/31/2011 21:47'!
inspectIt
	
	self handleEdit:
		[textMorph editor evaluateSelectionAndDo: [:result | result inspect]]! !

!PluggableTextMorph methodsFor: 'debug and other' stamp: 'di 5/3/1998 12:45'!
installModelIn: aWorld
	"No special inits for new components"
	^ self! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 4/27/2015 14:49'!
isAutoFit
	"Whether I adjust the contents of my scroller to save a scroll bar."
	
	^ self wrapFlag! !

!PluggableTextMorph methodsFor: 'testing' stamp: 'dtl 2/10/2010 17:28'!
isTextView
	"True if the reciever is a view on a text model, such as a view on a TranscriptStream"
	^true! !

!PluggableTextMorph methodsFor: 'testing' stamp: 'mt 4/27/2015 14:47'!
isWrapped

	^ self wrapFlag! !

!PluggableTextMorph methodsFor: 'event handling' stamp: 'mt 4/10/2015 12:20'!
keyboardFocusDelegate

	^ textMorph ifNil: [self]! !

!PluggableTextMorph methodsFor: 'event handling' stamp: 'nice 8/2/2011 20:50'!
keyStroke: evt
	"A keystroke was hit while the receiver had keyboard focus.  Pass the keystroke on to my textMorph, and and also, if I have an event handler, pass it on to that handler"

	textMorph keyStroke: evt.
	self eventHandler ifNotNil:
		[self eventHandler keyStroke: evt fromMorph: self].
! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 11/2/1998 16:01'!
methodNamesContainingIt
	self handleEdit: [textMorph editor methodNamesContainingIt]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 11/2/1998 16:01'!
methodSourceContainingIt
	self handleEdit: [textMorph editor methodSourceContainingIt]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 11/2/1998 16:01'!
methodStringsContainingit
	self handleEdit: [textMorph editor methodStringsContainingit]! !

!PluggableTextMorph methodsFor: 'event handling' stamp: 'mt 5/19/2016 11:18'!
mouseEnter: event
	"Restore the selection in the text morph if there was a selection."

	super mouseEnter: event.
	
	selectionInterval ifNotNil: [:interval |
		textMorph editor
			selectInterval: selectionInterval;
			setEmphasisHere].
		
	Preferences mouseOverForKeyboardFocus
		ifTrue:[event hand newKeyboardFocus: self]! !

!PluggableTextMorph methodsFor: 'event handling' stamp: 'mt 5/19/2016 11:20'!
mouseLeave: event
	"Save the selection interval for later."

	selectionInterval := textMorph editor selectionInterval.

	super mouseLeave: event.

	Preferences mouseOverForKeyboardFocus
		ifTrue: [event hand releaseKeyboardFocus: self]! !

!PluggableTextMorph methodsFor: 'interactive error protocol' stamp: 'di 5/6/1998 15:26'!
nextTokenFrom: start direction: dir
	^ textMorph editor nextTokenFrom: start direction: dir! !

!PluggableTextMorph methodsFor: 'interactive error protocol' stamp: 'di 5/20/1998 08:32'!
notify: aString at: anInteger in: aStream
	^ textMorph editor notify: aString at: anInteger in: aStream! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'ar 12/17/2001 13:00'!
offerFontMenu
	self handleEdit: [textMorph editor changeTextFont]! !

!PluggableTextMorph methodsFor: 'event handling' stamp: 'dvf 7/28/2003 14:50'!
onKeyStrokeSend: sel to: recipient
	textMorph on: #keyStroke send: sel to: recipient.! !

!PluggableTextMorph methodsFor: 'initialization' stamp: 'mt 7/26/2016 11:15'!
on: anObject text: getTextSel accept: setTextSel readSelection: getSelectionSel menu: getMenuSel

	self model: anObject.
	getTextSelector := getTextSel.
	setTextSelector := setTextSel.
	getSelectionSelector := getSelectionSel.
	getMenuSelector := getMenuSel.
	self setText: self getText.
	self setSelection: self getSelection.! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 5/10/1998 12:44'!
paste
	self handleEdit: [textMorph editor paste]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'ar 1/15/2001 18:36'!
pasteRecent
	"Paste an item chosen from RecentClippings."

	| clipping |
	(clipping := Clipboard chooseRecentClipping) ifNil: [^ self].
	Clipboard clipboardText: clipping.
	^ self handleEdit: [textMorph editor paste]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 5/10/1998 12:44'!
presentSpecialMenu
	self handleEdit: [textMorph editor presentSpecialMenu]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'sw 11/7/1999 00:01'!
prettyPrint
	self handleEdit: [textMorph editor prettyPrint]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'sw 11/7/1999 00:01'!
prettyPrintWithColor
	self handleEdit: [textMorph editor prettyPrintWithColor]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'dew 3/7/2000 21:10'!
printerSetup
	self handleEdit: [textMorph editor printerSetup]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'nice 3/31/2011 21:48'!
printIt
	| oldEditor |
	textMorph editor selectFrom: selectionInterval first to: selectionInterval last;
						model: model.  "For, eg, evaluateSelection"
	textMorph handleEdit: [(oldEditor := textMorph editor) evaluateSelectionAndDo:
		[:result |
		selectionInterval := oldEditor selectionInterval.
		textMorph installEditorToReplace: oldEditor.
		textMorph handleEdit: [oldEditor afterSelectionInsertAndSelect: result printString].
		selectionInterval := oldEditor selectionInterval.
	
		textMorph editor selectFrom: selectionInterval first to: selectionInterval last.
		self scrollSelectionIntoView]]! !

!PluggableTextMorph methodsFor: 'unaccepted edits' stamp: 'mt 5/1/2015 20:46'!
promptForCancel
	"Ask if it is OK to cancel changes to text"
	(self confirm:
'Changes have not been saved.
Is it OK to cancel those changes?' translated)
		ifTrue: [model changed: #clearUserEdits].
! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 4/19/2015 13:38'!
readOnly

	^ textMorph readOnly! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 4/19/2015 13:38'!
readOnly: aBoolean

	textMorph readOnly: aBoolean.! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'mt 11/8/2015 11:52'!
redo
	self handleEdit: [textMorph editor redo]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 11/2/1998 15:59'!
referencesToIt
	self handleEdit: [textMorph editor referencesToIt]! !

!PluggableTextMorph methodsFor: 'transcript' stamp: 'di 5/9/1998 21:40'!
replaceSelectionWith: aText
	^ textMorph editor replaceSelectionWith: aText! !

!PluggableTextMorph methodsFor: 'geometry' stamp: 'mt 4/27/2015 14:48'!
resizeScroller
	"Also needs to resize the text morph"

	super resizeScroller.

	textMorph ifNotNil: [:tm |
		tm isWrapped ifTrue: [textMorph extent: self scroller extent]].! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'dew 3/7/2000 21:15'!
saveContentsInFile
	self handleEdit: [textMorph editor saveContentsInFile]! !

!PluggableTextMorph methodsFor: 'scroll bar events' stamp: 'rr 3/10/2004 09:30'!
scrollBarMenuButtonPressed: event
	"The menu button in the scrollbar was pressed; put up the menu"

	| menu |
	(menu := self getMenu: event shiftPressed) ifNotNil:
		["Set up to use perform:orSendTo: for model/view dispatch"
		menu setInvokingView: self.
		menu invokeModal]! !

!PluggableTextMorph methodsFor: 'geometry' stamp: 'mt 4/9/2015 17:21'!
scrollDeltaHeight
	"Return the increment in pixels which this pane should be scrolled."
	^ textMorph ifNil: [super scrollDeltaHeight] ifNotNil: [:tm | tm defaultLineHeight]
! !

!PluggableTextMorph methodsFor: 'editor access' stamp: 'di 5/22/1998 12:35'!
scrollSelectionIntoView
	"Scroll my text into view if necessary and return true, else return false"
	^ self scrollSelectionIntoView: nil! !

!PluggableTextMorph methodsFor: 'editor access' stamp: 'mt 11/20/2015 12:00'!
scrollSelectionIntoView: event 
	"Scroll my text into view. Due to line composition mechanism, we must never use the right of a character block because the lines last character block right value always comes from a global container and is *not* line specific."
	
	selectionInterval := textMorph editor selectionInterval.
	
	textMorph editor hasSelection
		ifFalse: [self scrollToShow: (textMorph editor startBlock withWidth: 1)]
		ifTrue: [
			self scrollToShow: (textMorph editor startBlock topLeft corner: textMorph editor stopBlock bottomLeft).
			self scrollToShow: (textMorph editor pointBlock withWidth: 1). "Ensure text cursor visibility."].
		
	^ true! !

!PluggableTextMorph methodsFor: 'interactive error protocol' stamp: 'di 5/6/1998 14:59'!
select
	^ textMorph editor select! !

!PluggableTextMorph methodsFor: 'editor access' stamp: 'mt 5/3/2015 19:37'!
selectAll
	"Tell my textMorph to select all"

	textMorph selectAll.! !

!PluggableTextMorph methodsFor: 'interactive error protocol' stamp: 'di 5/6/1998 14:58'!
selectFrom: start to: stop
	^ textMorph editor selectFrom: start to: stop! !

!PluggableTextMorph methodsFor: 'interactive error protocol' stamp: 'di 5/6/1998 15:18'!
selectInvisiblyFrom: start to: stop
	^ textMorph editor selectInvisiblyFrom: start to: stop! !

!PluggableTextMorph methodsFor: 'interactive error protocol' stamp: 'di 5/6/1998 14:56'!
selectionInterval
	^ textMorph editor selectionInterval! !

!PluggableTextMorph methodsFor: 'model access' stamp: 'di 6/22/1998 01:32'!
selectionInterval: sel
	selectionInterval := sel! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'dtl 8/23/2015 21:15'!
sendContentsToPrinter
	self handleEdit: [textMorph editor sendContentsToPrinterWithLabel: owner knownName]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 11/2/1998 15:31'!
sendersOfIt
	self handleEdit: [textMorph editor sendersOfIt]! !

!PluggableTextMorph methodsFor: 'initialization' stamp: 'mt 8/5/2016 08:45'!
setDefaultParameters

	super setDefaultParameters.
	
	self
		font: (self userInterfaceTheme font ifNil: [TextStyle defaultFont]);
		setTextColor: (self userInterfaceTheme textColor ifNil: [Color black]).
	
	self wrapBorderColor: ((self userInterfaceTheme wrapBorderColorModifier ifNil: [ [:c | c muchLighter alpha: 0.3] ])
								value: self borderColor).
	
	self
		setProperty: #adornmentReadOnly
		toValue: (self userInterfaceTheme adornmentReadOnly ifNil: [Color black]);
		setProperty: #adornmentRefuse
		toValue: (self userInterfaceTheme adornmentRefuse ifNil: [Color tan]);
		setProperty: #adornmentConflict
		toValue: (self userInterfaceTheme adornmentConflict ifNil: [Color red]);
		setProperty: #adornmentDiff
		toValue: (self userInterfaceTheme adornmentDiff ifNil: [Color green]);
		setProperty: #adornmentNormalEdit
		toValue: (self userInterfaceTheme adornmentNormalEdit ifNil: [Color orange]);
		setProperty: #adornmentDiffEdit
		toValue: (self userInterfaceTheme adornmentDiffEdit ifNil: [Color yellow]).
		
	self
		setProperty: #frameAdornmentWidth
		toValue: (self userInterfaceTheme frameAdornmentWidth ifNil: [1]).
	
	textMorph
		caretColor: (self userInterfaceTheme caretColor ifNil: [Color red]);
		selectionColor: (self userInterfaceTheme selectionColor ifNil: [TranslucentColor r: 0.0 g: 0.0 b: 0.8 alpha: 0.2]);
		unfocusedSelectionColor: ((self userInterfaceTheme unfocusedSelectionModifier ifNil: [ [:c | Color gray: 0.9] ])
			value: textMorph selectionColor).! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'tk 7/13/2000 20:09'!
setSearchString
	self handleEdit: [textMorph editor setSearchString]! !

!PluggableTextMorph methodsFor: 'model access' stamp: 'mt 11/20/2015 11:47'!
setSelection: sel
	selectionInterval := sel.
	textMorph editor selectFrom: sel first to: sel last.
	self scrollSelectionIntoView.! !

!PluggableTextMorph methodsFor: 'model access' stamp: 'yo 9/28/2006 14:20'!
setTextColor: aColor
	"Set the color of my text to the given color"

	textMorph textColor: aColor! !

!PluggableTextMorph methodsFor: 'editor access' stamp: 'sw 7/24/2001 02:24'!
setTextMorphToSelectAllOnMouseEnter
	"Tell my textMorph's editor to select all when the mouse enters"

	textMorph on: #mouseEnter send: #selectAll to: textMorph! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 6/1/2016 16:34'!
setTextSelector
	^setTextSelector! !

!PluggableTextMorph methodsFor: 'model access' stamp: 'mt 7/24/2016 13:57'!
setText: aText

	textMorph newContents: aText.
	self hasUnacceptedEdits: false.
	self setScrollDeltas.	
	self changed. "Redraw the whole area. For example, it might not be necssary to draw the help text anymore."! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'di 5/20/1998 23:11'!
spawn
	self handleEdit: [textMorph editor spawn].
	self cancel! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'tfel 4/7/2015 09:38'!
spyOnIt
	self handleEdit: [textMorph editor spyOnIt]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'ab 3/23/2005 16:50'!
tallyIt
	self handleEdit: [textMorph editor tallyIt]! !

!PluggableTextMorph methodsFor: 'model access' stamp: 'di 4/20/1998 07:59'!
text
	^ textMorph contents! !

!PluggableTextMorph methodsFor: 'editor access' stamp: 'mt 4/7/2015 09:00'!
textEdited: someText
	"Tell the model about some edits in the text if interested. This is not #accept, which means that it will be send on every keystroke."
	
	self editTextSelector ifNotNil: [:selector |
		model perform: selector with: someText].! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'rr 3/21/2004 12:29'!
textMorph
	^ textMorph! !

!PluggableTextMorph methodsFor: 'private' stamp: 'tween 8/29/2004 20:42'!
textMorphClass
	"Answer the class used to create the receiver's textMorph"
	
	^TextMorphForEditView! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'nice 3/31/2011 21:48'!
tileForIt
	"Return a tile referring to the object resulting form evaluating my current selection.  Not currently threaded in, but useful in earlier demos and possibly still of value."

	
	self handleEdit:
		[textMorph editor evaluateSelectionAndDo: [:result | self currentHand attachMorph: result tileToRefer]]! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'nice 12/27/2009 03:10'!
toggleAnnotationPaneSize

	| handle origin aHand siblings |

	self flag: #bob.		"CRUDE HACK to enable changing the size of the annotations pane"

	owner ifNil: [^self].
	siblings := owner submorphs.
	siblings size > 3 ifTrue: [^self].
	siblings size < 2 ifTrue: [^self].

	aHand := self primaryHand.
	origin := aHand position.
	(handle := HandleMorph new)
		forEachPointDo: [:newPoint | | lf ht prevBottom newHeight m |
			handle removeAllMorphs.
			newHeight := (newPoint - origin) y asInteger min: owner height - 50 max: 16.
			lf := siblings last layoutFrame.
			lf bottomOffset: newHeight.
			prevBottom := newHeight.
			siblings size - 1 to: 1 by: -1 do: [ :index |
				m := siblings at: index.
				lf := m layoutFrame.
				ht := lf bottomOffset - lf topOffset.
				lf topOffset: prevBottom.
				lf bottomOffset = 0 ifFalse: [
					lf bottomOffset: (prevBottom + ht).
				].
				prevBottom := prevBottom + ht.
			].
			owner layoutChanged.

		]
		lastPointDo:
			[:newPoint | handle deleteBalloon.
			self halo ifNotNil: [:halo | halo addHandles].
		].
	aHand attachMorph: handle.
	handle setProperty: #helpAtCenter toValue: true.
	handle showBalloon:
'Move cursor farther from
this point to increase pane.
Click when done.' hand: aHand.
	handle startStepping

! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'tk 7/13/2000 20:13'!
undo
	self handleEdit: [textMorph editor undo]! !

!PluggableTextMorph methodsFor: 'updating' stamp: 'mt 8/2/2016 09:43'!
update: aSymbol 
	aSymbol ifNil: [^self].
	aSymbol == #flash ifTrue: [^self flash].

	aSymbol == getTextSelector
		ifTrue: [
			self setText: self getText.
			getSelectionSelector
				ifNotNil: [self setSelection: self getSelection].
			^ self].
	aSymbol == getSelectionSelector 
		ifTrue: [^self setSelection: self getSelection].

	aSymbol == #acceptChanges ifTrue: [^ self accept].
	aSymbol == #revertChanges ifTrue: [^ self cancel].

	(aSymbol == #autoSelect and: [getSelectionSelector notNil]) 
		ifTrue: 
			[self handleEdit: 
					[(textMorph editor)
						abandonChangeText; "no replacement!!"
						setSearch: model autoSelectString;
						findAgain]].
	aSymbol == #clearUserEdits ifTrue: [^self hasUnacceptedEdits: false].
	aSymbol == #wantToChange 
		ifTrue: 
			[self canDiscardEdits ifFalse: [^self promptForCancel].
			^self].
	aSymbol == #appendEntry 
		ifTrue: 
			[self handleEdit: [self appendEntry].
			^self refreshWorld].
	aSymbol == #appendEntryLater
		ifTrue: [self handleEdit: [self appendEntry]].
	aSymbol == #clearText 
		ifTrue: 
			[self handleEdit: [self changeText: Text new].
			^self refreshWorld].
	aSymbol == #bs 
		ifTrue: 
			[self handleEdit: [self bsText].
			^self refreshWorld].
	aSymbol == #codeChangedElsewhere 
		ifTrue: 
			[self hasEditingConflicts: true.
			^self changed].
	aSymbol == #saveContents
		ifTrue:
			[^self saveContentsInFile].
	aSymbol == #showContents
		ifTrue:
			[^ self scrollToTop].
! !

!PluggableTextMorph methodsFor: 'updating' stamp: 'mt 6/1/2016 16:35'!
update: aSymbol with: arg1

	aSymbol == #editString ifTrue:[
		self editString: arg1.
		self hasUnacceptedEdits: true.
	].

	(aSymbol == #inputRequested and: [self getTextSelector == arg1 or: [self setTextSelector == arg1]])
		ifTrue: [self activeHand newKeyboardFocus: self].

	^super update: aSymbol with: arg1! !

!PluggableTextMorph methodsFor: 'halos and balloon help' stamp: 'mt 4/2/2015 09:15'!
wantsBalloon

	^ self textMorph contents notEmpty and: [super wantsBalloon]! !

!PluggableTextMorph methodsFor: 'dropping/grabbing' stamp: 'jcg 7/7/2000 11:13'!
wantsDroppedMorph: aMorph event: anEvent 
	^ self model wantsDroppedMorph: aMorph event: anEvent inMorph: self! !

!PluggableTextMorph methodsFor: 'drawing' stamp: 'dgd 10/1/2004 12:32'!
wantsFrameAdornments
	"Answer whether the receiver wishes to have red borders, etc.,  
	used to show editing state"
	"A 'long-term temporary workaround': a nonmodular,  
	unsavory, but expedient way to get the desired effect, sorry.  
	Clean up someday."
	^ self
		valueOfProperty: #wantsFrameAdornments
		ifAbsent: [(#(#annotation #searchString #infoViewContents ) includes: getTextSelector) not]! !

!PluggableTextMorph methodsFor: 'drawing' stamp: 'dgd 10/1/2004 12:32'!
wantsFrameAdornments: aBoolean 
	self setProperty: #wantsFrameAdornments toValue: aBoolean! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 4/27/2015 16:57'!
wantsWrapBorder

	^ wantsWrapBorder ifNil: [false]! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 4/27/2015 16:57'!
wantsWrapBorder: aBoolean

	wantsWrapBorder := aBoolean.! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 7/25/2016 11:27'!
wrapBorderColor

	^ self valueOfProperty: #wrapBorderColor ifAbsent: [Color gray alpha: 0.3]! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 7/25/2016 11:27'!
wrapBorderColor: aColor

	self setProperty: #wrapBorderColor toValue: aColor.
	self changed.! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 4/27/2015 14:46'!
wrapFlag

	^ textMorph ifNil: [false] ifNotNil: [:tm | tm isWrapped]! !

!PluggableTextMorph methodsFor: 'accessing' stamp: 'mt 4/27/2015 15:14'!
wrapFlag: aBoolean

	textMorph ifNil: [self setText: ''].
	textMorph
		wrapFlag: aBoolean;
		autoFit: true.
	
	"Text navigation will be tedious if there is no horizontal scroll bar w/o wrapping."
	aBoolean ifFalse: [self showHScrollBarOnlyWhenNeeded].

	self
		resizeScrollBars;
		resizeScroller;
		hideOrShowScrollBars;
		setScrollDeltas.! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'dgd 10/1/2004 13:27'!
yellowButtonActivity
	"Called when the shifted-menu's 'more' item is chosen"
	self yellowButtonActivity: false! !

!PluggableTextMorph methodsFor: 'menu commands' stamp: 'dgd 10/1/2004 13:32'!
yellowButtonActivity: shiftKeyState 
	"Called when the shifted-menu's 'more' item is chosen"
	| menu |
	(menu := self getMenu: shiftKeyState)
		ifNotNil: [""
			menu setInvokingView: self.
			menu invokeModal]! !

!ScrollPane class methodsFor: 'examples' stamp: 'mk 8/9/2005 10:28'!
example1
	| window scrollPane pasteUpMorph |
	window := SystemWindow new.
	scrollPane := ScrollPane new.
	pasteUpMorph := PasteUpMorph new.
	pasteUpMorph extent: 1000@1000.
	scrollPane scroller addMorph: pasteUpMorph.
	window addMorph: scrollPane frame: (0@0 corner: 1@1).
	window openInWorld.! !

!ScrollPane class methodsFor: 'examples' stamp: 'mk 8/9/2005 10:29'!
example2
	| window scrollPane pasteUpMorph point textMorph |
	window := SystemWindow new.
	scrollPane := ScrollPane new.
	pasteUpMorph := PasteUpMorph new.
	pasteUpMorph extent: 1000@1000.
	scrollPane scroller addMorph: pasteUpMorph.
	window addMorph: scrollPane frame: (0@0 corner: 1@1).
	0 to: 1000 by: 100 do: 
		[:x | 0 to: 1000 by: 100 do:
			[:y |
				point :=  x@y.
				textMorph := TextMorph new contents: point asString.
				textMorph position: point.
				pasteUpMorph addMorph: textMorph
			]
		].
	window openInWorld.! !

!ScrollPane class methodsFor: 'new-morph participation' stamp: 'di 2/21/98 11:02'!
includeInNewMorphMenu
	"OK to instantiate"
	^ true! !

!ScrollPane class methodsFor: 'defaults' stamp: 'mt 4/25/2015 09:10'!
scrollBarThickness

	^ Preferences scrollBarsNarrow
		ifTrue: [10]
		ifFalse: [14]! !

!ScrollPane class methodsFor: 'preferences' stamp: 'mt 7/23/2016 11:33'!
themeProperties

	^ super themeProperties, {
		{ #borderColor. 'Colors'. 'Color of the pane''s border.' }.
		{ #borderWidth. 'Borders'. 'Width of the pane''s border.' }.
		{ #borderStyle. 'Borders'. 'Whether to use a plain border, inset, or outset.' }.
		{ #color. 'Colors'. 'Background color of the pane.' }.
	}! !

!ScrollPane class methodsFor: 'preferences' stamp: 'mt 4/14/2015 19:47'!
useRetractableScrollBars
	
	<preference: 'Use retractable scrollbars'
		category: #scrolling
		description: 'When enabled, scrollbars are hidden when scroll pane is not focused.'
		type: #Boolean>
	^ UseRetractableScrollBars ifNil: [false]! !

!ScrollPane class methodsFor: 'preferences' stamp: 'mt 4/14/2015 19:50'!
useRetractableScrollBars: aBoolean
	
	UseRetractableScrollBars = aBoolean ifTrue: [^ self].
	UseRetractableScrollBars := aBoolean.
	ScrollPane allSubInstances do: [:pane | 
		pane retractable: aBoolean.
		pane setScrollDeltas].! !

!ScrollPane methodsFor: 'menu' stamp: 'dgd 8/30/2003 22:06'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	retractableScrollBar
		ifTrue: [aCustomMenu add: 'make scrollbar inboard' translated action: #retractableOrNot]
		ifFalse: [aCustomMenu add: 'make scrollbar retractable' translated action: #retractableOrNot].
	scrollBarOnLeft
		ifTrue: [aCustomMenu add: 'scroll bar on right' translated action: #leftOrRight]
		ifFalse: [aCustomMenu add: 'scroll bar on left' translated action: #leftOrRight]! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/10/2015 17:54'!
adoptPaneColor: paneColor
	super adoptPaneColor: paneColor.
	
	"May not be in the hierarchy at the moment."
	scrollBar adoptPaneColor: paneColor.
	hScrollBar adoptPaneColor: paneColor.

	paneColor ifNotNil: [:c | self borderColor: (c adjustBrightness: -0.3)].! !

!ScrollPane methodsFor: 'accessing options' stamp: 'mt 4/24/2015 21:53'!
alwaysShowHScrollBar

	self hScrollBarPolicy: #always.
	self hHideOrShowScrollBar.
! !

!ScrollPane methodsFor: 'accessing options' stamp: 'mt 4/19/2015 07:55'!
alwaysShowScrollBars	
	
	self 
		alwaysShowHScrollBar;
		alwaysShowVScrollBar.
! !

!ScrollPane methodsFor: 'accessing options' stamp: 'mt 4/24/2015 21:53'!
alwaysShowVScrollBar

	self vScrollBarPolicy: #always.
	self vHideOrShowScrollBar.
! !

!ScrollPane methodsFor: 'updating' stamp: 'mt 7/23/2016 17:15'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	self setDefaultParameters.! !

!ScrollPane methodsFor: 'accessing' stamp: 'dgd 10/1/2004 12:53'!
borderStyle: aBorderStyle 
	super borderStyle: aBorderStyle.
	self setScrollDeltas! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 7/26/2016 15:43'!
borderWidth: aNumber 
	super borderWidth: aNumber.
	self resizeScroller; setScrollDeltas! !

!ScrollPane methodsFor: 'support' stamp: 'cmm 8/25/2013 13:08'!
canBeEncroached
	"For support of the smartHorizontalSplitters preference."
	^ scrollBar isInWorld not! !

!ScrollPane methodsFor: 'geometry testing' stamp: 'mt 4/25/2015 08:32'!
containsPoint: aPoint

	(super containsPoint: aPoint) ifTrue: [^ true].
	
	"Also include v scrollbar when it is extended..."
	((retractableScrollBar and: [self vIsScrollbarShowing]) and:
		[scrollBar containsPoint: aPoint])
			ifTrue:[ ^true ].
		
	"Also include hScrollbar when it is extended..."
	((retractableScrollBar and: [self hIsScrollbarShowing]) and:
		[hScrollBar containsPoint: aPoint])
			ifTrue: [ ^true ].
	
	^ false
! !

!ScrollPane methodsFor: 'initialization' stamp: 'sps 4/3/2005 15:21'!
defaultExtent
	^150@120
! !

!ScrollPane methodsFor: 'geometry' stamp: 'mt 4/24/2015 21:58'!
extentToFit
	"Resize scroll pane to exactly fit its contents."
	
	| offset |
	offset := 0@0.
	(retractableScrollBar not and: [self vScrollBarPolicy == #always])
		ifTrue: [offset := (self scrollBarThickness - self borderWidth) @ offset y].
	(retractableScrollBar not and: [self hScrollBarPolicy == #always])
		ifTrue: [offset := offset x @ (self scrollBarThickness - self borderWidth)].
		
	^ scroller submorphBounds extent + offset + (2* (self borderWidth @ self borderWidth))! !

!ScrollPane methodsFor: 'geometry' stamp: 'mt 8/27/2015 11:57'!
extent: aPoint
	
	self handleResizeAction: [
		(bounds extent closeTo: aPoint)
			ifTrue: [false]
			ifFalse: [
				super extent: aPoint.
				owner ifNotNil: [owner layoutChanged].
				true]].! !

!ScrollPane methodsFor: 'event filtering' stamp: 'mt 6/12/2016 08:14'!
filterEvent: aKeyboardEvent for: morphOrNil
	"See #initialize. This filter should be installed as keyboard event filter during the capture phase."

	aKeyboardEvent isKeystroke
		ifFalse: [^ aKeyboardEvent].

	^ aKeyboardEvent
		wasIgnored: (self scrollByKeyboard: aKeyboardEvent);
		yourself! !

!ScrollPane methodsFor: 'geometry' stamp: 'mt 4/15/2015 15:44'!
fit
	"Resize scroll pane to exactly fit its contents."
	
	self extent: self extentToFit.! !

!ScrollPane methodsFor: 'accessing' stamp: 'md 2/24/2006 16:25'!
flatColoredScrollBarLook
	"Currently only show the flat (not rounded) + colored-to-match-window scrollbar look when inboard."
	^ retractableScrollBar not or: [ScrollBar alwaysShowFlatScrollbarForAlternativeLook]
! !

!ScrollPane methodsFor: 'menu' stamp: 'yo 10/15/2006 01:14'!
getMenu: shiftKeyState
	"Answer the menu for this text view, supplying an empty menu to be filled in. If the menu selector takes an extra argument, pass in the current state of the shift key."
	| menu aMenu aTitle |
	getMenuSelector == nil ifTrue: [^ nil].
	(self valueOfProperty: #withMenuButton) == false ifTrue: [^ nil].
	menu _ MenuMorph new defaultTarget: model.
	aTitle _ getMenuTitleSelector ifNotNil: [model perform: getMenuTitleSelector].
	getMenuSelector numArgs = 1 ifTrue:
		[aMenu _ model perform: getMenuSelector with: menu.
		aTitle ifNotNil:  [aMenu addTitle: aTitle].
		^ aMenu].
	getMenuSelector numArgs = 2 ifTrue:
		[aMenu _ model perform: getMenuSelector with: menu with: shiftKeyState.
		aTitle ifNotNil:  [aMenu addTitle: aTitle].
		^ aMenu].
	^ self error: 'The getMenuSelector must be a 1- or 2-keyword symbol'! !

!ScrollPane methodsFor: 'geometry' stamp: 'mt 5/19/2015 15:57'!
handleResizeAction: aBlock
	"Ensure layout properties after resizing takes place."
	
	| oldExtent |	
	oldExtent := self extent.
	
	aBlock value ifFalse: [^ self].
	
	"Now reset widget sizes"
	self extent ~= oldExtent ifTrue: [
		self
			resizeScrollBars;
			resizeScroller;
			setScrollDeltas].! !

!ScrollPane methodsFor: 'event handling' stamp: 'mt 4/10/2015 14:45'!
handlesKeyboard: evt
	^ true! !

!ScrollPane methodsFor: 'event handling' stamp: 'sps 3/10/2004 10:23'!
handlesMouseDown: evt
	^ true
! !

!ScrollPane methodsFor: 'event handling' stamp: 'ar 9/18/2000 22:11'!
handlesMouseOver: evt
	"Could just ^ true, but this ensures that scroll bars won't flop out
	if you mouse-over appendages such as connecting pins."
	self flag: #arNote. "I have no idea how the code below could've ever worked. If the receiver does not handle mouse over events then it should not receive any #mouseLeave if the mouse leaves the receiver for real. This is because 'evt cursorPoint' describes the *end* point of the movement and considering that the code would return false if the move ends outside the receiver the scroll bars should never pop back in again. Which is exactly what happens with the new event logic if you don't just ^true. I'm leaving the code in for reference - perhaps somebody can make sense from it; I sure cannot."
	^true
"
	| cp |
	cp := evt cursorPoint.
	(bounds containsPoint: cp)
		ifTrue: [^ true]			
		ifFalse: [self submorphsDo:
					[:m | (m containsPoint: cp) ifTrue:
							[m == scrollBar
								ifTrue: [^ true]
								ifFalse: [^ false]]].
				^ false]
"! !

!ScrollPane methodsFor: 'event handling' stamp: 'mt 6/6/2016 18:05'!
handlesMouseWheel: evt
	^ true! !

!ScrollPane methodsFor: 'accessing' stamp: 'dew 10/17/1999 19:40'!
hasFocus
	"hasFocus is currently set by mouse enter/leave events.
	This inst var should probably be moved up to a higher superclass."

	^ hasFocus ifNil: [false]! !

!ScrollPane methodsFor: 'geometry ranges' stamp: 'mt 4/8/2015 19:23'!
hExtraScrollRange
	"Return the amount of extra blank space to include below the bottom of the scroll content."
	^ 0
! !

!ScrollPane methodsFor: 'scrolling' stamp: 'sps 12/25/2002 01:10'!
hHideOrShowScrollBar
	"Hide or show the scrollbar depending on if the pane is scrolled/scrollable."

	self hIsScrollbarNeeded
		ifTrue:[ self hShowScrollBar ]
		ifFalse: [ self hHideScrollBar ].
! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 4/7/2015 11:16'!
hHideScrollBar

	self hIsScrollbarShowing ifFalse: [^self].
	self removeMorph: hScrollBar.
	retractableScrollBar ifFalse: [self resetExtent].

! !

!ScrollPane methodsFor: 'geometry testing' stamp: 'mt 4/8/2015 19:40'!
hIsScrollable
	"If the contents of the pane are too small to scroll, return false."
	
	^ self hLeftoverScrollRange > 0
! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 4/24/2015 21:56'!
hIsScrollbarNeeded
"Return whether the horz scrollbar is needed"

	"Don't do anything with the retractable scrollbar unless we have focus"
	retractableScrollBar & self hasFocus not ifTrue: [^false].
	
	"Don't show it if we were told not to."
	self hScrollBarPolicy == #never ifTrue: [^false].

	"Always show it if we were told to"
	self hScrollBarPolicy == #always ifTrue: [^true].

	^self hIsScrollable
! !

!ScrollPane methodsFor: 'geometry testing' stamp: 'mt 4/27/2015 17:09'!
hIsScrollbarShowing
	"Return true if a horz scroll bar is currently showing"

	^ hScrollBar owner notNil! !

!ScrollPane methodsFor: 'geometry testing' stamp: 'sps 3/10/2004 10:22'!
hIsScrolled
	"If the scroller is not set to x = 0, then the pane has been h-scrolled."
	^scroller offset x > 0
! !

!ScrollPane methodsFor: 'accessing options' stamp: 'mt 4/24/2015 21:53'!
hideHScrollBarIndefinitely
	"Get rid of scroll bar for short panes that don't want it shown."

	self hScrollBarPolicy: #never.
	self hHideOrShowScrollBar.
! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 4/15/2015 16:29'!
hideOrShowScrollBars

	self 
		vHideOrShowScrollBar; 
		hHideOrShowScrollBar; 
		resizeScrollBars.! !

!ScrollPane methodsFor: 'scrolling' stamp: 'sps 12/25/2002 16:30'!
hideScrollBars
	self
		vHideScrollBar;
		hHideScrollBar
! !

!ScrollPane methodsFor: 'accessing options' stamp: 'mt 4/19/2015 07:56'!
hideScrollBarsIndefinitely

	self
		hideVScrollBarIndefinitely;
		hideHScrollBarIndefinitely.! !

!ScrollPane methodsFor: 'accessing options' stamp: 'mt 4/24/2015 21:53'!
hideVScrollBarIndefinitely
	"Get rid of scroll bar for short panes that don't want it shown."

	self vScrollBarPolicy: #never.
	self vHideOrShowScrollBar.
! !

!ScrollPane methodsFor: 'geometry ranges' stamp: 'mt 4/8/2015 19:30'!
hLeftoverScrollRange
	"Return the entire scrolling range minus the currently viewed area."

	^ scroller hasSubmorphs
		ifFalse: [0]
		ifTrue: [self hTotalScrollRange - scroller width max: 0]
! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/1/2015 16:57'!
hMargin
"pixels of whitespace at to the left of the scroller when the hScrollBar offset is 0"
	^0
! !

!ScrollPane methodsFor: 'geometry' stamp: 'mt 4/25/2015 08:57'!
hResizeScrollBar

	| topLeft h border offset |
	self hScrollBarPolicy == #never ifTrue: [^self].
	bounds ifNil: [ self fullBounds ].
	
	h := self scrollBarThickness.
	border := borderWidth.
	offset := (scrollBarOnLeft and: [self vIsScrollbarShowing and: [retractableScrollBar not]])
		ifTrue: [h]
		ifFalse: [0].
	
	topLeft := retractableScrollBar
				ifTrue: [bounds bottomLeft + (offset @ border negated)]
				ifFalse: [bounds bottomLeft + (offset @ h negated)].

	hScrollBar bounds: (topLeft extent: self hScrollBarWidth@ h)! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/8/2015 19:35'!
hScrollBar
	^ hScrollBar! !

!ScrollPane methodsFor: 'scroll bar events' stamp: 'sps 12/27/2002 00:13'!
hScrollBarMenuButtonPressed: event
	^ self scrollBarMenuButtonPressed: event
! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/24/2015 21:47'!
hScrollBarPolicy

	^ hScrollBarPolicy! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/24/2015 21:48'!
hScrollBarPolicy: aSymbol
	"#always, #never, #whenNeeded"

	hScrollBarPolicy := aSymbol.! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 5/13/2016 14:33'!
hScrollBarValue: scrollValue

	scroller hasSubmorphs ifFalse: [^ self].
	scroller offset: (scrollValue max: 0) @ scroller offset y.! !

!ScrollPane methodsFor: 'geometry' stamp: 'hmm 9/19/2009 11:35'!
hScrollBarWidth
"Return the width of the horizontal scrollbar"


	| w |
	
	w := bounds width "- (2 * borderWidth)".
	
	(retractableScrollBar not and: [self vIsScrollbarNeeded])
		ifTrue: [w := w - self scrollBarThickness ].
		
	^w 
! !

!ScrollPane methodsFor: 'geometry' stamp: 'mt 4/8/2015 19:53'!
hSetScrollDelta
	"Set the ScrollBar deltas, value and interval, based on the current scroll pane size, offset and range."

	| delta |	
	delta := self scrollDeltaWidth.

	hScrollBar
			truncate: true;
			scrollDelta: delta 
			pageDelta: 10*delta;
			maximumValue: self hLeftoverScrollRange;
			interval: (self hTotalScrollRange = 0
				ifTrue: [1.0]
				ifFalse: [scroller width / self hTotalScrollRange]);
			setValue: scroller offset x.! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 4/14/2015 20:02'!
hShowScrollBar

	self hIsScrollbarShowing ifTrue: [^self].
	self hResizeScrollBar.
	self privateAddMorph: hScrollBar atIndex: 1.
	retractableScrollBar
		ifTrue: [self comeToFront]
		ifFalse: [self resetExtent].
! !

!ScrollPane methodsFor: 'geometry ranges' stamp: 'sps 3/10/2004 11:51'!
hTotalScrollRange
	"Return the entire scrolling range."
	^ self hUnadjustedScrollRange + self hExtraScrollRange + self hMargin
! !

!ScrollPane methodsFor: 'geometry ranges' stamp: 'sps 12/24/2002 16:07'!
hUnadjustedScrollRange
	"Return the width extent of the receiver's submorphs."

	| submorphBounds |
	submorphBounds := scroller localSubmorphBounds ifNil: [^ 0].
	^ submorphBounds right
! !

!ScrollPane methodsFor: 'initialization' stamp: 'mt 7/23/2016 11:35'!
initialize
	
	"initialize the state of the receiver"
	super initialize.
	""
	self initializePreferences.
	hasFocus := false.
	self initializeScrollBars.
	""
	
	self extent: self defaultExtent.
	self
		resizeScrollBars;
		resizeScroller;
		hideOrShowScrollBars;
		updateMinimumExtent.
	
	self setDefaultParameters.
	self addKeyboardCaptureFilter: self.! !

!ScrollPane methodsFor: 'initialization' stamp: 'mt 4/25/2015 09:13'!
initializePreferences
	"initialize the receiver's Preferences"
	
	retractableScrollBar := self class useRetractableScrollBars.
	scrollBarOnLeft := (Preferences valueOfFlag: #scrollBarsOnRight) not.
	scrollBarThickness := self class scrollBarThickness.
	
	Preferences alwaysShowVScrollbar
		ifTrue: [ self vScrollBarPolicy: #always ]
		ifFalse: [ self vScrollBarPolicy: #whenNeeded ].
		
	Preferences alwaysHideHScrollbar
		ifTrue:[ self hScrollBarPolicy: #never ]
		ifFalse: [ Preferences alwaysShowHScrollbar
			ifTrue: [ self hScrollBarPolicy: #always ]
			ifFalse: [ self hScrollBarPolicy: #whenNeeded ]].! !

!ScrollPane methodsFor: 'initialization' stamp: 'mt 7/23/2016 13:41'!
initializeScrollBars
	"Initialize vertical and horizontal scroll bars."

	(scrollBar := ScrollBar on: self getValue: nil setValue: #vScrollBarValue:)
			menuSelector: #vScrollBarMenuButtonPressed:;
			orientation: #vertical;
			extent: 1@1.
	(hScrollBar := ScrollBar on: self getValue: nil setValue: #hScrollBarValue:)
			menuSelector: #hScrollBarMenuButtonPressed:;
			orientation: #horizontal;
			extent: 1@1.

	""
	scroller := TransformMorph new color: Color transparent.
	scroller offset: 0 @ 0.
	self addMorph: scroller.
	""
	retractableScrollBar ifFalse: 
			[self 
				addMorph: scrollBar;
				addMorph: hScrollBar].
	self updateMinimumExtent.! !

!ScrollPane methodsFor: 'geometry' stamp: 'mt 4/25/2015 09:40'!
innerBounds

	| inner bottomOffset leftOffset rightOffset |
	(retractableScrollBar or: [self vIsScrollbarShowing not])
		ifTrue: [leftOffset := rightOffset := 0]
		ifFalse: [
			scrollBarOnLeft
				ifTrue: [
					leftOffset := self scrollBarThickness - self borderWidth.
					rightOffset := 0.]
				ifFalse: [
					leftOffset := 0.
					rightOffset := self scrollBarThickness - self borderWidth]].
			
	(retractableScrollBar or: [self hIsScrollbarShowing not])
		ifTrue: [bottomOffset := 0]
		ifFalse: [bottomOffset := self scrollBarThickness - self borderWidth].
	
	inner := super innerBounds.
	^ (inner left + leftOffset) @ (inner top "+ topOffset")
		corner: (inner right - rightOffset) @ (inner bottom - bottomOffset)! !

!ScrollPane methodsFor: 'geometry testing' stamp: 'mt 4/25/2015 08:40'!
isAnyScrollbarShowing

	^ self hIsScrollbarShowing or: [self vIsScrollbarShowing]
! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/19/2015 11:22'!
isAutoFit
	"Does this scroll pane modifies the extent of its scrollable content when it resizes itself to avoid, for example, a horizontal or vertical scrollbar?"

	^ false! !

!ScrollPane methodsFor: 'support' stamp: 'mt 4/25/2015 08:42'!
isScrolledFromTop
	"Have the contents of the pane been scrolled, so that the top of the contents are not visible?"
	^ self vIsScrolled
! !

!ScrollPane methodsFor: 'event handling' stamp: 'mt 6/6/2016 17:59'!
keyStroke: evt
	"If pane is not empty, pass the event to the last submorph,
	assuming it is the most appropriate recipient (!!)"

	scroller submorphs last keyStroke: evt! !

!ScrollPane methodsFor: 'layout' stamp: 'mt 4/22/2015 09:53'!
layoutChanged
	"Do not tell owner. We might keep our bounds."
	
	fullBounds := nil.! !

!ScrollPane methodsFor: 'menu' stamp: 'di 11/14/97 09:09'!
leftOrRight  "Change scroll bar location"
	scrollBarOnLeft := scrollBarOnLeft not.
	self extent: self extent! !

!ScrollPane methodsFor: 'menu' stamp: 'sw 8/18/1998 12:38'!
menuTitleSelector: aSelector
	getMenuTitleSelector := aSelector! !

!ScrollPane methodsFor: 'layout' stamp: 'mt 4/25/2015 17:27'!
minScrollbarHeight
	"Answer the minimum extent occupied by the receiver..
	It is assumed the if the receiver is sized to its minimum both scrollbars will be used (and visible) unless they have been turned off explicitly.
	This makes the behaviour also more predictable."
	
	^ (self vScrollBarPolicy == #never 
		ifTrue: [0]
		ifFalse: [scrollBar minHeight])
			+ (((self hScrollBarPolicy == #never or: [retractableScrollBar]) or: [self isAutoFit])
				ifTrue: [0]
				ifFalse: [self scrollBarThickness])! !

!ScrollPane methodsFor: 'layout' stamp: 'mt 4/25/2015 17:24'!
minScrollbarWidth
	"Answer the minimum extent occupied by the receiver..
	It is assumed the if the receiver is sized to its minimum both scrollbars will be used (and visible) unless they have been turned off explicitly.
	This makes the behaviour also more predictable."
	
	^ ((self hScrollBarPolicy == #never or: [self isAutoFit])
		ifTrue: [0]
		ifFalse: [hScrollBar minWidth])
			+ ((retractableScrollBar or: [self vScrollBarPolicy == #never])
				ifTrue: [0]
				ifFalse: [self scrollBarThickness])! !

!ScrollPane methodsFor: 'event handling' stamp: 'md 2/13/2006 11:53'!
mouseDown: evt
	evt yellowButtonPressed  "First check for option (menu) click"
		ifTrue: [^ self yellowButtonActivity: evt shiftPressed].
	"If pane is not empty, pass the event to the last submorph,
	assuming it is the most appropriate recipient (!!)"
	scroller hasSubmorphs ifTrue:
		[scroller submorphs last mouseDown: (evt transformedBy: (scroller transformFrom: self))]! !

!ScrollPane methodsFor: 'event handling' stamp: 'mt 8/3/2016 14:09'!
mouseEnter: event
	hasFocus := true.
	owner isSystemWindow ifTrue: [owner paneTransition: event].
	retractableScrollBar ifTrue: [self hideOrShowScrollBars].! !

!ScrollPane methodsFor: 'event handling' stamp: 'mt 8/3/2016 14:09'!
mouseLeave: event
	hasFocus := false.
	retractableScrollBar ifTrue: [self hideScrollBars].
	owner isSystemWindow ifTrue: [owner paneTransition: event].! !

!ScrollPane methodsFor: 'event handling' stamp: 'md 2/13/2006 11:53'!
mouseMove: evt
	"If pane is not empty, pass the event to the last submorph,
	assuming it is the most appropriate recipient (!!)."
	scroller hasSubmorphs ifTrue:
		[scroller submorphs last mouseMove: (evt transformedBy: (scroller transformFrom: self))]! !

!ScrollPane methodsFor: 'event handling' stamp: 'md 2/13/2006 11:53'!
mouseUp: evt
	"If pane is not empty, pass the event to the last submorph,
	assuming it is the most appropriate recipient (!!)"
	scroller hasSubmorphs ifTrue:
		[scroller submorphs last mouseUp: (evt transformedBy: (scroller transformFrom: self))]! !

!ScrollPane methodsFor: 'event handling' stamp: 'mt 6/6/2016 18:06'!
mouseWheel: evt

	evt isWheelUp ifTrue: [scrollBar scrollUp: 3].
	evt isWheelDown ifTrue: [scrollBar scrollDown: 3].! !

!ScrollPane methodsFor: 'accessing' stamp: 'nk 3/8/2004 11:07'!
numSelectionsInView
	"Answer the scroller's height based on the average number of submorphs."
	
	^scroller numberOfItemsPotentiallyInView! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 5/19/2015 15:37'!
offsetToShow: aRectangle
	"Calculate the offset necessary to show the rectangle."
	
	| offset scrollRange target |
	offset := scroller offset.
	scrollRange := self hTotalScrollRange @ self vTotalScrollRange.
	
	"Normalize the incoming rectangle."
	target := 
			(scroller width < aRectangle width
				ifTrue: [offset x < aRectangle left "Coming from left?"
					ifTrue: [aRectangle right - scroller width]
					ifFalse: [aRectangle left]]
				ifFalse: [aRectangle left])
		@
			(scroller height < aRectangle height
				ifTrue: [offset y < aRectangle top "Coming from top?"
					ifTrue: [aRectangle bottom - scroller height]
					ifFalse: [aRectangle top]]
				ifFalse: [aRectangle top])
		corner: 
			(scroller width < aRectangle width
				ifTrue: [offset x + scroller width > aRectangle right "Coming from right?"
					ifTrue: [aRectangle left + scroller width]
					ifFalse: [aRectangle right]]
				ifFalse: [aRectangle right])
		@
			(scroller height < aRectangle height
				ifTrue: [offset y + scroller height > aRectangle bottom "Coming from bottom?"
					ifTrue: [aRectangle top + scroller height]
					ifFalse: [aRectangle bottom]]
				ifFalse: [aRectangle bottom]).

	"Vertical Scrolling"
	target top < offset y
		ifTrue: [offset := offset x @ target top].	
	target bottom > (offset y + scroller height)
		ifTrue: [offset := offset x @ (target bottom - scroller height)].
	
	"Horizontal Scrolling"
	target left < offset x
		ifTrue: [offset := target left @ offset y].
	target right > (offset x + scroller width)
		ifTrue: [offset := (target right - scroller width) @ offset y].

	^ (offset min: scrollRange - scroller extent) max: 0@0! !

!ScrollPane methodsFor: 'geometry' stamp: 'dew 10/17/1999 19:41'!
resetExtent
	"Reset the extent. (may be overridden by subclasses which need to do more than this)"
	self resizeScroller! !

!ScrollPane methodsFor: 'geometry' stamp: 'sps 12/25/2002 16:34'!
resizeScrollBars
	self vResizeScrollBar; hResizeScrollBar
! !

!ScrollPane methodsFor: 'geometry' stamp: 'di 11/11/1998 09:48'!
resizeScroller

	scroller bounds: self innerBounds! !

!ScrollPane methodsFor: 'accessing retractable' stamp: 'mt 4/26/2015 21:59'!
retractableOrNot
	"Change scroll bar operation"

	retractableScrollBar := retractableScrollBar not.
	retractableScrollBar
		ifTrue: [
			self removeMorph: scrollBar; removeMorph: hScrollBar]
		ifFalse: [(submorphs includes: scrollBar) 
					ifFalse: 
						[self privateAddMorph: scrollBar atIndex: 1.
						self privateAddMorph: hScrollBar atIndex: 1]].
	self
		resizeScrollBars;
		resizeScroller;
		hideOrShowScrollBars;
		updateMinimumExtent.! !

!ScrollPane methodsFor: 'accessing retractable' stamp: 'LC 6/12/2000 09:28'!
retractableScrollBar
	^ retractableScrollBar! !

!ScrollPane methodsFor: 'accessing retractable' stamp: 'sw 11/5/1998 14:14'!
retractable: aBoolean
	retractableScrollBar == aBoolean ifFalse: [self retractableOrNot "toggles it"]! !

!ScrollPane methodsFor: 'support' stamp: 'sps 3/10/2004 10:22'!
scrollBarFills: aRectangle
	"Return true if a flop-out scrollbar fills the rectangle"

	retractableScrollBar ifFalse:[^false].
	
	((submorphs includes: scrollBar) and: [scrollBar bounds containsRect: aRectangle])
				ifTrue:[ ^true ].
	^((submorphs includes: hScrollBar) and: [hScrollBar bounds containsRect: aRectangle])
! !

!ScrollPane methodsFor: 'scroll bar events' stamp: 'di 6/26/1998 13:31'!
scrollBarMenuButtonPressed: event
	^ self yellowButtonActivity: event shiftPressed! !

!ScrollPane methodsFor: 'accessing' stamp: 'LC 6/12/2000 09:28'!
scrollBarOnLeft
	^ scrollBarOnLeft! !

!ScrollPane methodsFor: 'accessing' stamp: 'sw 1/13/98 21:27'!
scrollBarOnLeft: aBoolean
	scrollBarOnLeft := aBoolean.
	self extent: self extent! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/25/2015 10:30'!
scrollBarThickness
	"Includes border"
	
	^ scrollBarThickness! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/25/2015 09:54'!
scrollBarThickness: anInteger
	
	scrollBarThickness := anInteger.
	self updateMinimumExtent.! !

!ScrollPane methodsFor: 'event handling' stamp: 'mt 6/6/2016 18:05'!
scrollByKeyboard: event 
	"If event is ctrl+up/down then scroll and answer true. Backwards compatibility."
	
	(event controlKeyPressed or:[event commandKeyPressed "??? key decode table in event sensor does not change CTRL+up/down !!!!!!"]) ifFalse: [^ false].
	event keyCharacter = Character arrowUp
		ifTrue: 
			[scrollBar scrollUp: 3.
			^ true].
	event keyCharacter = Character arrowDown
		ifTrue: 
			[scrollBar scrollDown: 3.
			^ true].
	"event keyCharacter = Character arrowRight
		ifTrue: 
			[hScrollBar scrollDown: 3.
			^ true].
	event keyCharacter = Character arrowLeft
		ifTrue: 
			[hScrollBar scrollUp: 3.
			^ true]."
	^ false! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 4/8/2015 19:49'!
scrollBy: delta
	"Move the contents in the direction delta."

	self flag: #negated. "mt: Who uses this and why is does the expected behavor negate the delta?"
	self vScrollBar scrollBy: delta y negated.
	self hScrollBar scrollBy: delta x negated.! !

!ScrollPane methodsFor: 'geometry' stamp: 'sps 3/10/2004 13:26'!
scrollDeltaHeight
	"Return the increment in pixels which this pane should be scrolled (normally a subclass responsibility)."
	^ 10
! !

!ScrollPane methodsFor: 'geometry' stamp: 'sps 3/9/2004 17:29'!
scrollDeltaWidth
	"Return the increment in pixels which this pane should be scrolled (normally a subclass responsibility)."
	
	^10
! !

!ScrollPane methodsFor: 'accessing' stamp: ''!
scroller
	^ scroller! !

!ScrollPane methodsFor: 'accessing' stamp: 'ar 5/19/1999 18:06'!
scroller: aTransformMorph
	scroller ifNotNil:[scroller delete].
	scroller := aTransformMorph.
	self addMorph: scroller.
	self resizeScroller.! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 5/2/2016 17:57'!
scrollToBottom

	self vScrollBar scrollToMaximum..! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 5/2/2016 17:58'!
scrollToLeft

	self hScrollBar scrollToMinimum..! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 5/2/2016 17:58'!
scrollToRight

	self hScrollBar scrollToMaximum.! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 5/13/2016 14:34'!
scrollToShow: aRectangle

	scroller offset: (self offsetToShow: aRectangle).
	self setScrollDeltas.
! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 5/2/2016 17:57'!
scrollToTop

	self vScrollBar scrollToMinimum.! !

!ScrollPane methodsFor: 'initialization' stamp: 'mt 7/23/2016 17:20'!
setDefaultParameters
	"change the receiver's appareance parameters"

	self
		color: (self userInterfaceTheme color ifNil: [Color white]);
		borderStyle: (self userInterfaceTheme borderStyle ifNil: [BorderStyle default]);
		borderColor: (self userInterfaceTheme borderColor ifNil: [Color gray: 0.6]);
		borderWidth: (self userInterfaceTheme borderWidth ifNil: [1]).! !

!ScrollPane methodsFor: 'geometry' stamp: 'mt 4/8/2015 20:06'!
setScrollDeltas
	"Set the ScrollBar deltas, value and interval, based on the current scroll pane size, offset and range."

	scroller hasSubmorphs ifFalse: [^ self].
	
"NOTE: fullbounds commented out now -- trying to find a case where this expensive step is necessary -- perhaps there is a less expensive way to handle that case."
	"scroller fullBounds." "force recompute so that leftoverScrollRange will be up-to-date"
	self hideOrShowScrollBars.
	self vSetScrollDelta.
	self hSetScrollDelta..
! !

!ScrollPane methodsFor: 'scroll bar events' stamp: 'sw 3/22/2001 12:03'!
shiftedTextPaneMenuRequest
	"The more... button was hit from the text-pane menu"

	^ self yellowButtonActivity: true! !

!ScrollPane methodsFor: 'scroll bar events' stamp: 'di 6/26/1998 13:31'!
shiftedYellowButtonActivity
	^ self yellowButtonActivity: true! !

!ScrollPane methodsFor: 'accessing options' stamp: 'mt 4/24/2015 21:54'!
showHScrollBarOnlyWhenNeeded

	self hScrollBarPolicy: #whenNeeded.
	self hHideOrShowScrollBar.
! !

!ScrollPane methodsFor: 'scrolling' stamp: 'sps 12/25/2002 16:35'!
showScrollBars
	self  vShowScrollBar; hShowScrollBar
! !

!ScrollPane methodsFor: 'accessing options' stamp: 'mt 4/19/2015 07:52'!
showScrollBarsOnlyWhenNeeded

	self
		showHScrollBarOnlyWhenNeeded;
		showVScrollBarOnlyWhenNeeded.
! !

!ScrollPane methodsFor: 'accessing options' stamp: 'mt 4/24/2015 21:54'!
showVScrollBarOnlyWhenNeeded
	"Get rid of scroll bar for short panes that don't want it shown."

	self vScrollBarPolicy: #whenNeeded.
	self vHideOrShowScrollBar.
! !

!ScrollPane methodsFor: 'scroll bar events' stamp: 'di 6/26/1998 13:32'!
unshiftedYellowButtonActivity
	^ self yellowButtonActivity: false! !

!ScrollPane methodsFor: 'layout' stamp: 'mt 4/25/2015 09:47'!
updateMinimumExtent
	"This grows only. For shrinking scroll bars, this might have to be changed."
	
	self minimumWidth: (self minimumWidth max: self minScrollbarWidth).
	self minimumHeight: (self minimumHeight max: self minScrollbarHeight).! !

!ScrollPane methodsFor: 'geometry ranges' stamp: 'mt 4/8/2015 19:23'!
vExtraScrollRange
	"Return the amount of extra blank space to include below the bottom of the scroll content."
	^ 0
! !

!ScrollPane methodsFor: 'scrolling' stamp: 'sps 12/25/2002 16:31'!
vHideOrShowScrollBar

	self vIsScrollbarNeeded
		ifTrue:[ self vShowScrollBar ]
		ifFalse:[ self vHideScrollBar ].
! !

!ScrollPane methodsFor: 'scrolling' stamp: 'sps 3/10/2004 12:07'!
vHideScrollBar
	self vIsScrollbarShowing ifFalse: [^self].
	self removeMorph: scrollBar.
	retractableScrollBar ifFalse: [self resetExtent].
	
! !

!ScrollPane methodsFor: 'geometry testing' stamp: 'mt 4/8/2015 19:40'!
vIsScrollable
	"Return whether the verticle scrollbar is scrollable. If the contents of the pane are too small to scroll, return false."
	
	^ self vLeftoverScrollRange > 0! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 4/24/2015 21:59'!
vIsScrollbarNeeded
"Return whether the verticle scrollbar is needed"

	"Don't do anything with the retractable scrollbar unless we have focus"
	retractableScrollBar & self hasFocus not ifTrue: [^false].
	
	"Don't show it if we were told not to."
	self vScrollBarPolicy == #never ifTrue: [^false].

	"Always show it if we were told to"
	self vScrollBarPolicy == #always ifTrue: [^true].
	
	^self vIsScrollable
! !

!ScrollPane methodsFor: 'geometry testing' stamp: 'mt 4/27/2015 17:09'!
vIsScrollbarShowing
	"Return true if a retractable scroll bar is currently showing"

	^ scrollBar owner notNil
! !

!ScrollPane methodsFor: 'geometry testing' stamp: 'sps 3/10/2004 10:23'!
vIsScrolled
	"If the scroller is not set to y = 0, then the pane has been scrolled."
	^scroller offset y > 0
! !

!ScrollPane methodsFor: 'geometry ranges' stamp: 'mt 4/8/2015 19:31'!
vLeftoverScrollRange
	"Return the entire scrolling range minus the currently viewed area."

	^ scroller hasSubmorphs
		ifFalse: [0]
		ifTrue: [self vTotalScrollRange - scroller height max: 0]
! !

!ScrollPane methodsFor: 'geometry' stamp: 'mt 4/14/2015 19:32'!
vResizeScrollBar
	| w topLeft border |
	w := self scrollBarThickness.
	border := self borderWidth.
	topLeft := scrollBarOnLeft 
		ifTrue: [retractableScrollBar 
			ifTrue: [bounds topLeft - ((w - border) @ 0)]
			ifFalse: [bounds topLeft]]
		ifFalse: [retractableScrollBar 
			ifTrue: [bounds topRight - (border @ 0)]
			ifFalse: [bounds topRight - (w @ 0)]].
			
	scrollBar bounds: (topLeft extent: w @ self vScrollBarHeight)
	
! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/8/2015 19:36'!
vScrollBar
	^ scrollBar! !

!ScrollPane methodsFor: 'geometry' stamp: 'hmm 9/19/2009 11:32'!
vScrollBarHeight
	| h |

	h := bounds height "- (2 * borderWidth)".
	(retractableScrollBar not and: [self hIsScrollbarNeeded]) 
		ifTrue:[ h := h - self scrollBarThickness. ].
	
	^h
! !

!ScrollPane methodsFor: 'scroll bar events' stamp: 'sps 12/27/2002 00:13'!
vScrollBarMenuButtonPressed: event
	^ self scrollBarMenuButtonPressed: event
! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/24/2015 21:47'!
vScrollBarPolicy

	^ vScrollBarPolicy! !

!ScrollPane methodsFor: 'accessing' stamp: 'mt 4/24/2015 21:47'!
vScrollBarPolicy: aSymbol
	"#always, #never, #whenNeeded"

	vScrollBarPolicy := aSymbol.! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 5/13/2016 14:33'!
vScrollBarValue: scrollValue

	scroller hasSubmorphs ifFalse: [^ self].
	scroller offset: scroller offset x @ (scrollValue max: 0).! !

!ScrollPane methodsFor: 'geometry' stamp: 'mt 4/8/2015 19:55'!
vSetScrollDelta
	"Set the ScrollBar deltas, value and interval, based on the current scroll pane size, offset and range."

	| delta |	
	delta := self scrollDeltaHeight.

	scrollBar
			truncate: true;
			scrollDelta: delta 
			pageDelta: 10*delta;
			maximumValue: self vLeftoverScrollRange;
			interval: (self vTotalScrollRange = 0
				ifTrue: [1.0]
				ifFalse: [scroller height / self vTotalScrollRange]);
			setValue: scroller offset y.! !

!ScrollPane methodsFor: 'scrolling' stamp: 'mt 4/14/2015 20:02'!
vShowScrollBar

	self vIsScrollbarShowing ifTrue: [^ self].
	self vResizeScrollBar.
	self privateAddMorph: scrollBar atIndex: 1.
	retractableScrollBar
		ifTrue: [self comeToFront]
		ifFalse: [self resetExtent]
! !

!ScrollPane methodsFor: 'geometry ranges' stamp: 'sps 3/10/2004 11:06'!
vTotalScrollRange
	"Return the entire scrolling range."
	^ self vUnadjustedScrollRange + self vExtraScrollRange
! !

!ScrollPane methodsFor: 'geometry ranges' stamp: 'sps 12/25/2002 14:43'!
vUnadjustedScrollRange
	"Return the height extent of the receiver's submorphs."
	| submorphBounds |
	submorphBounds := scroller localSubmorphBounds ifNil: [^ 0].
	^ submorphBounds bottom
! !

!ScrollPane methodsFor: 'event handling' stamp: 'mt 4/10/2015 13:15'!
wantsKeyboardFocus

	^ true! !

!ScrollPane methodsFor: 'accessing' stamp: '6/7/97 10:42 di'!
wantsSlot
	"For now do it the old way, until we sort this out"
	^ true! !

!ScrollPane methodsFor: 'menu' stamp: 'dgd 9/18/2004 18:29'!
wantsYellowButtonMenu
	"Answer true if the receiver wants a yellow button menu"
	^ getMenuSelector notNil! !

!ScrollPane methodsFor: 'scroll bar events' stamp: 'RAA 6/12/2000 09:02'!
yellowButtonActivity: shiftKeyState
	| menu |
	(menu := self getMenu: shiftKeyState) ifNotNil:
		[menu setInvokingView: self.
		menu popUpEvent: self activeHand lastEvent in: self world]! !

ScrollPane subclass: #SimpleHierarchicalListMorph
	instanceVariableNames: 'selectedMorph hoveredMorph getListSelector keystrokeActionSelector autoDeselect columns columnsCache sortingSelector getSelectionSelector setSelectionSelector potentialDropMorph lineColor font textColor'
	classVariableNames: 'WrappedNavigation'
	poolDictionaries: ''
	category: 'Morphic-Explorer'!

SimpleHierarchicalListMorph class
	instanceVariableNames: 'expandedForm notExpandedForm'!

!SimpleHierarchicalListMorph commentStamp: 'ls 3/1/2004 12:15' prior: 0!
Display a hierarchical list of items.  Each item should be wrapped with a ListItemWrapper.

For a simple example, look at submorphsExample.  For beefier examples, look at ObjectExplorer or FileList2.!

SimpleHierarchicalListMorph class
	instanceVariableNames: 'expandedForm notExpandedForm'!

SimpleHierarchicalListMorph subclass: #MorphHierarchyListMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!MorphHierarchyListMorph methodsFor: 'private' stamp: 'kfr 4/19/2015 11:54'!
createContainer
	"Private - Create a container"
	| container |
	container := BorderedMorph new.
	container extent: (World extent * (1 / 4 @ (2 / 3))) rounded.
	container layoutPolicy: TableLayout new.
	container hResizing: #rigid.
	container vResizing: #rigid.
	container
		setColor: MenuMorph menuColor
		borderWidth: MenuMorph menuBorderWidth
		borderColor: MenuMorph menuBorderColor.
	container layoutInset: 0.
	"container useRoundedCorners."
	""
	container setProperty: #morphHierarchy toValue: true.
	container setNameTo: 'Objects Hierarchy' translated.
	""
	^ container! !

!MorphHierarchyListMorph methodsFor: 'private' stamp: 'dgd 9/26/2004 18:27'!
inAContainer
	"Answer the receiver contained in a proper container"
	| container |
	container := self createContainer.
	container addMorphBack: self.
	" 
	nasty hack to force the scroolbar recreation"
	self extent: container extent - container borderWidth.
	""
	^ container! !

!MorphHierarchyListMorph methodsFor: 'initialization' stamp: 'dgd 9/26/2004 18:18'!
on: anObject list: getListSel selected: getSelectionSel changeSelected: setSelectionSel menu: getMenuSel keystroke: keyActionSel 
	super
		on: anObject
		list: getListSel
		selected: getSelectionSel
		changeSelected: setSelectionSel
		menu: getMenuSel
		keystroke: keyActionSel.
	""
	self borderWidth: 0.
	self autoDeselect: false.
	self enableDrag: false.
	self enableDrop: true.
	self hResizing: #spaceFill.
	self vResizing: #spaceFill.
self expandRoots! !

!MorphHierarchyListMorph methodsFor: 'selection' stamp: 'dgd 9/25/2004 21:28'!
setSelectedMorph: aMorph 
	super setSelectedMorph: aMorph.
self owner isNil ifFalse:[self owner delete]! !

!SimpleHierarchicalListMorph class methodsFor: 'instance creation' stamp: 'kfr 2/18/2016 17:31'!
expandedForm
	expandedForm
		ifNotNil: [expandedForm depth ~= Display depth
				ifTrue: [expandedForm := nil]].
	^ expandedForm
		ifNil: [expandedForm := ((Form
						extent: 10 @ 9
						depth: 8
						fromArray: #(4294967295 4294967295 4294901760 4294967295 4294967295 4294901760 4278255873 16843009 16842752 4294902089 1229539657 33488896 4294967041 1229539585 4294901760 4294967295 21561855 4294901760 4294967295 4278321151 4294901760 4294967295 4294967295 4294901760 4294967295 4294967295 4294901760 )
						offset: 0 @ 0)
						asFormOfDepth: Display depth) replaceColor: Color white withColor: Color transparent;
						 yourself]! !

!SimpleHierarchicalListMorph class methodsFor: 'instance creation' stamp: 'mt 2/18/2016 19:11'!
notExpandedForm
	notExpandedForm
		ifNotNil: [notExpandedForm depth ~= Display depth
				ifTrue: [notExpandedForm := nil]].
	^ notExpandedForm
		ifNil: [notExpandedForm := ((Form
						extent: 10 @ 9
						depth: 8
						fromArray: #(4294967041 4294967295 4294901760 4294967041 33554431 4294901760 4294967041 1224867839 4294901760 4294967041 1229521407 4294901760 4294967041 1229539585 4294901760 4294967041 1229521407 4294901760 4294967041 1224867839 4294901760 4294967041 33554431 4294901760 4294967041 4294967295 4294901760 )
						offset: 0 @ 0)
							asFormOfDepth: Display depth)
								replaceColor: Color white withColor: Color transparent;
								yourself]! !

!SimpleHierarchicalListMorph class methodsFor: 'instance creation' stamp: 'md 7/13/2005 16:34'!
on: anObject list: getListSel selected: getSelectionSel changeSelected: setSelectionSel
	"Create a 'pluggable' list view on the given model parameterized by the given message selectors."

	^ self new
		on: anObject
		list: getListSel
		selected: getSelectionSel
		changeSelected: setSelectionSel
		menu: nil
		keystroke: #arrowKey:from:		"default"! !

!SimpleHierarchicalListMorph class methodsFor: 'instance creation' stamp: 'md 7/13/2005 16:34'!
on: anObject list: getListSel selected: getSelectionSel changeSelected: setSelectionSel menu: getMenuSel
	"Create a 'pluggable' list view on the given model parameterized by the given message selectors."

	^ self new
		on: anObject
		list: getListSel
		selected: getSelectionSel
		changeSelected: setSelectionSel
		menu: getMenuSel
		keystroke: #arrowKey:from:		"default"
! !

!SimpleHierarchicalListMorph class methodsFor: 'instance creation' stamp: 'md 7/13/2005 16:34'!
on: anObject list: getListSel selected: getSelectionSel changeSelected: setSelectionSel menu: getMenuSel keystroke: keyActionSel
	"Create a 'pluggable' list view on the given model parameterized by the given message selectors."

	^ self new
		on: anObject
		list: getListSel
		selected: getSelectionSel
		changeSelected: setSelectionSel
		menu: getMenuSel
		keystroke: keyActionSel
! !

!SimpleHierarchicalListMorph class methodsFor: 'examples' stamp: 'ls 3/1/2004 17:32'!
submorphsExample
	"display a hierarchical list of the World plus its submorphs plus its submorphs' submorphs etc."
	"[SimpleHierarchicalListMorph submorphsExample]"
	| morph |
	morph :=
		SimpleHierarchicalListMorph
			on: [ Array with:  (MorphWithSubmorphsWrapper with: World)  ]
			list: #value
			selected: nil
			changeSelected: nil
			menu: nil
			keystroke: nil.

	morph openInWindow! !

!SimpleHierarchicalListMorph class methodsFor: 'preferences' stamp: 'mt 7/24/2016 10:49'!
themeProperties

	^ super themeProperties, {
		{ #font. 'Fonts'. 'Font of the list items.' }.
		{ #textColor. 'Colors'. 'Color of the list items.' }.
		{ #selectionColor. 'Colors'. 'Color used for items when hovering or selecting them.' }.
		{ #selectionTextColor. 'Colors'. 'Color used for label when hovering or selecting them.' }.
		
		{ #filterColor. 'Colors'. 'Color used for items to indicate the matching filter.' }.
		{ #filterTextColor. 'Colors'. 'Color used for items to indicate the matching filter.' }.
		{ #highlightTextColor. 'Colors'. 'Color used for highlighted items.' }.
		
		{ #hoverSelectionModifier. 'Colors'. 'How to derive the hover color from the selection color.'}.

		{ #lineColor. 'Colors'. 'How the lines are drawn.' }.
	}! !

!SimpleHierarchicalListMorph class methodsFor: 'preferences' stamp: 'cmm 4/7/2016 15:51'!
wrappedNavigation
	<preference: 'Wrapped Tree Navigation'
		category: 'Morphic'
		description: 'When enabled, use of the arrow keys at the top or bottom of a hierarchical list will wrap to the opposite side of the list.'
		type: #Boolean>
	^ WrappedNavigation ifNil: [ false ]! !

!SimpleHierarchicalListMorph class methodsFor: 'preferences' stamp: 'cmm 4/7/2016 15:23'!
wrappedNavigation: aBoolean 
	WrappedNavigation := aBoolean! !

!SimpleHierarchicalListMorph methodsFor: 'drag and drop' stamp: 'nk 6/15/2003 11:49'!
acceptDroppingMorph: aMorph event: evt

	self model
		acceptDroppingMorph: aMorph
		event: evt
		inMorph: self.
	self resetPotentialDropMorph.
	evt hand releaseMouseFocus: self.
	Cursor normal show.
! !

!SimpleHierarchicalListMorph methodsFor: 'private' stamp: 'mt 7/24/2016 15:06'!
addMorphsTo: morphList from: aCollection allowSorting: sortBoolean withExpandedItems: expandedItems atLevel: newIndent

	| priorMorph newCollection firstAddition |
	priorMorph := nil.
	newCollection := (sortBoolean and: [sortingSelector notNil]) ifTrue: [
		(aCollection asSortedCollection: [ :a :b | 
			(a perform: sortingSelector) <= (b perform: sortingSelector)]) asOrderedCollection
	] ifFalse: [
		aCollection
	].
	firstAddition := nil.
	newCollection do: [:item | 
		priorMorph := self indentingItemClass basicNew 
			initWithContents: item 
			prior: priorMorph 
			forList: self
			indentLevel: newIndent.
		priorMorph
			color: self textColor;
			font: self font;
			selectionColor: self selectionColor;
			selectionTextColor: self selectionTextColor;
			hoverColor: self hoverColor;
			highlightTextColor: self highlightTextColor;
			filterColor: self filterColor;
			filterTextColor: self filterTextColor.
		firstAddition ifNil: [firstAddition := priorMorph].
		morphList add: priorMorph.
		((item hasEquivalentIn: expandedItems) or: [priorMorph isExpanded]) ifTrue: [
			priorMorph isExpanded: true.
			priorMorph 
				addChildrenForList: self 
				addingTo: morphList
				withExpandedItems: expandedItems.
		].
	].
	^firstAddition
	
! !

!SimpleHierarchicalListMorph methodsFor: 'private' stamp: 'mt 7/24/2016 15:06'!
addSubmorphsAfter: parentMorph fromCollection: aCollection allowSorting: sortBoolean

	| priorMorph morphList newCollection |
	priorMorph := nil.
	newCollection := (sortBoolean and: [sortingSelector notNil]) ifTrue: [
		(aCollection asSortedCollection: [ :a :b | 
			(a perform: sortingSelector) <= (b perform: sortingSelector)]) asOrderedCollection
	] ifFalse: [
		aCollection
	].
	morphList := OrderedCollection new.
	newCollection do: [:item | 
		priorMorph := self indentingItemClass basicNew 
			initWithContents: item 
			prior: priorMorph 
			forList: self
			indentLevel: parentMorph indentLevel + 1.
		priorMorph
			color: self textColor;
			font: self font;
			selectionColor: self selectionColor;
			selectionTextColor: self selectionTextColor;
			hoverColor: self hoverColor;
			highlightTextColor: self highlightTextColor;
			filterColor: self filterColor;
			filterTextColor: self filterTextColor.
		morphList add: priorMorph.
	].
	scroller addAllMorphs: morphList after: parentMorph.
	^morphList
	
! !

!SimpleHierarchicalListMorph methodsFor: 'private' stamp: 'mt 4/22/2016 10:17'!
adjustSubmorphPositions

	| p |
	p := 0@0.
	scroller submorphsDo: [ :each | | h |
		each visible ifTrue: [
			h := each height.
			each privateBounds: (p extent: self preferredSubmorphWidth@h).
			p := p + (0@h) ]].
	self 
		clearColumnsCache;
		changed;
		layoutChanged;
		setScrollDeltas.
! !

!SimpleHierarchicalListMorph methodsFor: 'updating' stamp: 'mt 7/29/2016 15:04'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.! !

!SimpleHierarchicalListMorph methodsFor: 'keyboard navigation' stamp: 'cmm 4/7/2016 15:46'!
arrowKey: asciiValue
	"Handle a keyboard navigation character. Answer true if handled, false if not."
	| keyEvent max oldSelection nextSelection howManyItemsShowing keyHandled |
	keyHandled := false.
	keyEvent := asciiValue.
	max := self maximumSelection.
	nextSelection := oldSelection := self getSelectionIndex.
     keyEvent = 31 ifTrue:["down"
		keyHandled := true.
		nextSelection :=oldSelection + 1.
		nextSelection > max ifTrue: [nextSelection := (self class wrappedNavigation ifTrue: [1] ifFalse: [max])]].
     keyEvent = 30 ifTrue:["up"
		keyHandled := true.
		nextSelection := oldSelection - 1.
		nextSelection < 1 ifTrue: [nextSelection := self class wrappedNavigation ifTrue: [max] ifFalse: [1]]].
     keyEvent = 1  ifTrue: ["home"
		keyHandled := true.
		nextSelection := 1].
     keyEvent = 4  ifTrue: ["end"
		keyHandled := true.
		nextSelection := max].
	howManyItemsShowing := self numSelectionsInView.
      keyEvent = 11 ifTrue: ["page up"
		keyHandled := true.
		nextSelection := 1 max: oldSelection - howManyItemsShowing].
     keyEvent = 12  ifTrue: ["page down"
		keyHandled := true.
		nextSelection := oldSelection + howManyItemsShowing min: max].

	(nextSelection ~= oldSelection or: [ keyHandled and: [ self class wrappedNavigation not ]]) ifTrue: [
		self setSelectionIndex: nextSelection.
		^ true].
	
	keyEvent = 29 ifTrue:["right"
		selectedMorph ifNotNil:[
			(selectedMorph canExpand and:[selectedMorph isExpanded not])
				ifTrue:[self toggleExpandedState: selectedMorph]
				ifFalse:[self setSelectionIndex: self getSelectionIndex+1].
		].
		^true].
	keyEvent = 28 ifTrue:["left"
		selectedMorph ifNotNil:[
			(selectedMorph isExpanded)
				ifTrue:[self toggleExpandedState: selectedMorph]
				ifFalse:[self setSelectionIndex: (self getSelectionIndex-1 max: 1)].
		].
		^true].
	^false! !

!SimpleHierarchicalListMorph methodsFor: 'initialization' stamp: 'mt 3/6/2015 09:54'!
autoDeselect: trueOrFalse
	"When selecting a selected item, it will be deselected. See #mouseUp:."
	
	autoDeselect := trueOrFalse.! !

!SimpleHierarchicalListMorph methodsFor: 'private - caching' stamp: 'mt 4/22/2016 10:16'!
clearColumnsCache

	columnsCache := self columns
		ifNil: [Array empty]
		ifNotNil: [:c | Array new: c size].! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'RAA 7/24/1998 22:52'!
columns

	^columns! !

!SimpleHierarchicalListMorph methodsFor: 'private - caching' stamp: 'mt 4/22/2016 10:20'!
columnsCache
	columnsCache ifNil: [self clearColumnsCache].
	^ columnsCache! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 6/28/2016 11:34'!
columns: anArray
	"You can specify columns or specs for columns. This depends on the actual morph that is used as item. ObjectExplorers produce items that support exact numbers, blocks, and nil for max width."
	
	columns := anArray.
	self clearColumnsCache.! !

!SimpleHierarchicalListMorph methodsFor: 'initialization' stamp: 'RAA 3/30/1999 22:29'!
currentlyExpanded

	^(scroller submorphs select: [ :each | each isExpanded]) collect: [ :each |
		each complexContents
	].
	! !

!SimpleHierarchicalListMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 10:37'!
drawHoverOn: aCanvas 

	self hoveredMorph ifNil: [^ self].
	PluggableListMorph highlightHoveredRow ifFalse: [^ self].
	
	aCanvas
		transformBy: scroller transform
		clippingTo: scroller innerBounds
		during: [:c | self hoveredMorph drawHoverOn: c].! !

!SimpleHierarchicalListMorph methodsFor: 'drawing' stamp: 'mt 4/22/2016 10:52'!
drawLinesOn: aCanvas

	| lColor |
	lColor := self lineColor.
	aCanvas 
		transformBy: scroller transform
		clippingTo: scroller innerBounds
		during:[:clippedCanvas |
			scroller submorphsDo: [ :submorph |
				(submorph visible and: [(submorph isExpanded
					or: [clippedCanvas isVisible: submorph fullBounds] )
					or: [	submorph nextSibling notNil and: [clippedCanvas isVisible: submorph nextSibling fullBounds]]]) 
						 ifTrue: [submorph drawLinesOn: clippedCanvas lineColor: lColor indentThreshold: 0] ] ]
		smoothing: scroller smoothing.
! !

!SimpleHierarchicalListMorph methodsFor: 'drawing' stamp: 'mt 11/5/2015 09:39'!
drawOn: aCanvas 
	
	super drawOn: aCanvas.
	
	self drawHoverOn: aCanvas.
	self drawSelectionOn: aCanvas.
	self drawLinesOn: aCanvas.! !

!SimpleHierarchicalListMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 10:32'!
drawSelectionOn: aCanvas 

	self selectedMorph ifNotNil: [:m |
		aCanvas
			transformBy: scroller transform
			clippingTo: scroller innerBounds
			during: [:c | m drawSelectionOn: c] ].! !

!SimpleHierarchicalListMorph methodsFor: 'events' stamp: ''!
expandAll
	(selectedMorph isNil
		or: [selectedMorph isExpanded])
		ifTrue: [^self].
	self expandAll: selectedMorph.
	self adjustSubmorphPositions! !

!SimpleHierarchicalListMorph methodsFor: 'events' stamp: ''!
expandAll: aMorph
	| allChildren |
	aMorph toggleExpandedState.
	allChildren := OrderedCollection new: 10.
	aMorph recursiveAddTo: allChildren.
	allChildren do: [:each | 
		(each canExpand and: [each isExpanded not])
			ifTrue: [self expandAll: each]].
! !

!SimpleHierarchicalListMorph methodsFor: 'events' stamp: 'ar 3/17/2001 17:39'!
expandAll: aMorph except: aBlock
	| allChildren |
	(aBlock value: aMorph complexContents)
		ifFalse: [^self].
	aMorph toggleExpandedState.
	allChildren := OrderedCollection new: 10.
	aMorph recursiveAddTo: allChildren.
	allChildren do: [:each | 
		(each canExpand
			and: [each isExpanded not])
			ifTrue: [self expandAll: each except: aBlock]].! !

!SimpleHierarchicalListMorph methodsFor: 'drawing' stamp: 'nk 3/8/2004 10:05'!
expandedForm

	^self class expandedForm! !

!SimpleHierarchicalListMorph methodsFor: 'events' stamp: 'dgd 9/26/2004 18:24'!
expandRoots
	"Expand all the receiver's roots"
	self roots
		do: [:each |
			(each canExpand and: [each isExpanded not])
				ifTrue: [each toggleExpandedState]].
	self adjustSubmorphPositions! !

!SimpleHierarchicalListMorph methodsFor: 'events' stamp: 'ar 3/17/2001 17:39'!
expand: aMorph to: level
	| allChildren |
	aMorph toggleExpandedState.
	allChildren := OrderedCollection new: 10.
	aMorph recursiveAddTo: allChildren.
	allChildren do: [:each | 
		((each canExpand
			and: [each isExpanded not])
			and: [level > 0])
			ifTrue: [self expand: each to: level-1]].! !

!SimpleHierarchicalListMorph methodsFor: 'geometry' stamp: 'nk 7/11/2004 20:07'!
extent: newExtent
	bounds extent = newExtent ifTrue: [^ self].
	super extent: newExtent.
	self setScrollDeltas ! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:17'!
filterColor

	^ self valueOfProperty: #filterColor ifAbsent: [Color yellow]! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/29/2016 13:13'!
filterColor: aColor

	self setProperty: #filterColor toValue: aColor.
	scroller submorphsDo: [:ea | ea filterColor: aColor].! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:16'!
filterTextColor

	^ self valueOfProperty: #filterTextColor ifAbsent: [Color black]! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/29/2016 13:13'!
filterTextColor: aColor

	self setProperty: #filterTextColor toValue: aColor.
	scroller submorphsDo: [:ea | ea filterTextColor: aColor].! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 09:54'!
font
	^ font ifNil: [TextStyle defaultFont]! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 09:52'!
font: aFont
	font := aFont.
	scroller submorphsDo: [:ea | ea font: font].! !

!SimpleHierarchicalListMorph methodsFor: 'selection' stamp: 'RAA 8/1/1998 00:19'!
getCurrentSelectionItem

	^model perform: (getSelectionSelector ifNil: [^nil])
	! !

!SimpleHierarchicalListMorph methodsFor: 'model access' stamp: 'RAA 8/1/1998 00:10'!
getList 
	"Answer the list to be displayed."

	^(model perform: (getListSelector ifNil: [^#()])) ifNil: [#()]

! !

!SimpleHierarchicalListMorph methodsFor: 'keyboard navigation' stamp: 'ar 10/14/2003 23:38'!
getSelectionIndex
	^scroller submorphs indexOf: selectedMorph! !

!SimpleHierarchicalListMorph methodsFor: 'events-processing' stamp: 'mt 11/5/2015 09:44'!
handleMouseMove: anEvent
	"Reimplemented because we really want #mouseMove when a morph is dragged around"
	anEvent wasHandled ifTrue:[^self]. "not interested"
	self hoveredMorph: (self itemFromPoint: anEvent position).
	(anEvent anyButtonPressed and:[anEvent hand mouseFocus == self]) ifFalse:[^self].
	anEvent wasHandled: true.
	self mouseMove: anEvent.
	(self handlesMouseStillDown: anEvent) ifTrue:[
		"Step at the new location"
		self startStepping: #handleMouseStillDown: 
			at: Time millisecondClockValue
			arguments: {anEvent copy resetHandlerFields}
			stepTime: 1].
! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'ar 3/17/2001 17:27'!
handlesMouseOverDragging: evt
	^self dropEnabled! !

!SimpleHierarchicalListMorph methodsFor: 'scrolling' stamp: 'sps 12/24/2002 18:31'!
hExtraScrollRange
	"Return the amount of extra blank space to include below the bottom of the scroll content."
	^5
! !

!SimpleHierarchicalListMorph methodsFor: 'drawing' stamp: 'RAA 8/3/1999 09:44'!
highlightSelection

	selectedMorph ifNotNil: [selectedMorph highlight]! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:46'!
highlightTextColor

	^ self valueOfProperty: #highlightTextColor ifAbsent: [Color red]! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/29/2016 13:13'!
highlightTextColor: aColor

	self setProperty: #highlightTextColor toValue: aColor.
	scroller submorphsDo: [:ea | ea highlightTextColor: aColor].! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:19'!
hoverColor

	^ self valueOfProperty: #hoverColor ifAbsent: [Color veryLightGray]! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:33'!
hoverColor: aColor

	self setProperty: #hoverColor toValue: aColor.
	scroller submorphsDo: [:ea | ea setProperty: #hoverColor toValue: aColor].! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 11/5/2015 09:27'!
hoveredMorph

	^ hoveredMorph! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 11/5/2015 09:42'!
hoveredMorph: aMorph

	hoveredMorph == aMorph ifTrue: [^ self].
	
	hoveredMorph ifNotNil: [:m | m changed].
	hoveredMorph := aMorph.
	hoveredMorph ifNotNil: [:m | m changed].	! !

!SimpleHierarchicalListMorph methodsFor: 'scrolling' stamp: 'nice 12/27/2009 03:10'!
hUnadjustedScrollRange
"Return the width of the widest item in the list"

	| max count |

	max := 0.
	count := 0.
	scroller submorphsDo: [ :each | | stringW right |
		stringW := each font widthOfStringOrText: each contents.
		right := (each toggleRectangle right + stringW + 10).
		max := max max: right.
		
"NOTE: need to optimize this method by caching list item morph widths (can init that cache most efficiently in the #list: method before the item widths are reset to 9999).  For now, just punt on really long lists"
		((count := count + 1) > 200) ifTrue:[ ^max * 3].
	].

	^max 
! !

!SimpleHierarchicalListMorph methodsFor: 'initialization' stamp: 'RAA 7/29/2000 22:15'!
indentingItemClass
	
	^IndentingListItemMorph! !

!SimpleHierarchicalListMorph methodsFor: 'initialization' stamp: 'yo 9/11/2009 10:25'!
initialize
	"initialize the state of the receiver"
	super initialize.
	self setProperty: #autoExpand toValue: false.
	self
		on: #mouseMove
		send: #mouseStillDown:onItem:
		to: self! !

!SimpleHierarchicalListMorph methodsFor: 'private' stamp: 'RAA 4/2/1999 15:33'!
insertNewMorphs: morphList

	scroller addAllMorphs: morphList.
	self adjustSubmorphPositions.
	self selection: self getCurrentSelectionItem.
	self setScrollDeltas.
! !

!SimpleHierarchicalListMorph methodsFor: 'debug and other' stamp: 'di 5/6/1998 21:19'!
installModelIn: aWorld
	"No special inits for new components"
	^ self! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'mt 3/8/2015 18:15'!
itemFromPoint: aPoint
	"Return the list element (morph) at the given point or nil if outside"
	| ptY visibleRows |
	scroller hasSubmorphs ifFalse:[^nil].
	(scroller fullBounds containsPoint: aPoint) ifFalse:[^nil].
	
	visibleRows := scroller submorphs select: [:m | m visible].
	
	ptY := (visibleRows first point: aPoint from: self) y.
	"note: following assumes that submorphs are vertical, non-overlapping, and ordered"
	visibleRows first top > ptY ifTrue:[^nil].
	visibleRows last bottom < ptY ifTrue:[^nil].
	"now use binary search"
	^visibleRows
		findBinary:[:item|
			(item top <= ptY and:[item bottom >= ptY])
				ifTrue:[0] "found"
				ifFalse:[ (item top + item bottom // 2) > ptY ifTrue:[-1] ifFalse:[1]]]
		ifNone: [nil]! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'mt 3/13/2015 09:42'!
keyStrokeAction: event 

	| numArgs |
	keystrokeActionSelector ifNil: [^false].
	
	numArgs := keystrokeActionSelector numArgs.
	
	numArgs = 1 ifTrue: [
		^ model
			perform: keystrokeActionSelector
			with: event keyCharacter].
	numArgs = 2 ifTrue: [
		^ model 
			perform: keystrokeActionSelector
			with: event keyCharacter
			with: self].
	numArgs = 3 ifTrue: [
		^ model 
			perform: keystrokeActionSelector
			with: event keyCharacter
			with: self
			with: event].
	^self error: 'The keystrokeActionSelector must be a 1-, 2-, or 3-keyword symbol'! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'mt 6/6/2016 17:58'!
keyStroke: event 
	"Process potential command keys"
	
	event keyCharacter asciiValue < 32 ifTrue: [
		^ self specialKeyPressed: event keyCharacter asciiValue].
	
	" A lot of actions do not return a Boolean here (#messageListKey:from:, 
	#classListKey:from:, ...); even most #arrowKey:from: implementors do not.
	Hence, we explicitely check for false and else assume truthieness"
	^ (self keyStrokeAction: event) ~~ false
! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'nk 3/8/2004 09:55'!
lineColor
	"Answer a good color to use for drawing the lines that connect members of the hierarchy view.
	Used the cached color, or derive it if necessary by finding the first owner (up to my root) that is not transparent, then picking a contrasting color.
	Fall back to veryLightGray if all my owners are transparent."

	| coloredOwner targetLuminance ownerColor darken |
	lineColor ifNotNil: [ ^lineColor ].
	coloredOwner := self firstOwnerSuchThat: [ :o | o isWorldOrHandMorph not and: [ o color isTransparent not ]].
	coloredOwner ifNil: [ ^Color veryLightGray ].
	ownerColor := coloredOwner color.
	darken := ownerColor luminance > 0.5.
	targetLuminance := ownerColor luminance + (darken ifTrue: [ -0.2 ] ifFalse: [ 0.2 ]).
	^darken
		ifTrue: [ ownerColor atMostAsLuminentAs: targetLuminance ]
		ifFalse: [ ownerColor atLeastAsLuminentAs: targetLuminance ]
	
! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'sps 12/28/2002 02:27'!
lineColor: aColor
	^lineColor := aColor
! !

!SimpleHierarchicalListMorph methodsFor: 'initialization' stamp: 'di 5/22/1998 00:32'!
listItemHeight
	"This should be cleaned up.  The list should get spaced by this parameter."
	^ 12! !

!SimpleHierarchicalListMorph methodsFor: 'initialization' stamp: 'ar 3/17/2001 17:39'!
list: aCollection

	| wereExpanded morphList |
	wereExpanded := self currentlyExpanded.
	scroller removeAllMorphs.
	(aCollection isNil or: [aCollection isEmpty]) ifTrue: [^ self selectedMorph: nil].
	morphList := OrderedCollection new.
	self 
		addMorphsTo: morphList
		from: aCollection 
		allowSorting: false
		withExpandedItems: wereExpanded
		atLevel: 0.
	self insertNewMorphs: morphList.! !

!SimpleHierarchicalListMorph methodsFor: 'selection' stamp: 'RAA 7/31/1998 00:25'!
maximumSelection

	^ scroller submorphs size
! !

!SimpleHierarchicalListMorph methodsFor: 'selection' stamp: 'di 5/22/1998 00:20'!
minimumSelection
	^ 1! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'mt 5/27/2015 13:36'!
mouseDown: evt
	| aMorph selectors |
	aMorph := self itemFromPoint: evt position.
	evt yellowButtonPressed  "First check for option (menu) click"
		ifTrue: [
			(PluggableListMorph menuRequestUpdatesSelection and: [model okToChange]) ifTrue: [
				aMorph == selectedMorph 
					ifFalse: [self setSelectedMorph: aMorph]].
			^ self yellowButtonActivity: evt shiftPressed].
	(aMorph notNil and:[aMorph inToggleArea: (aMorph point: evt position from: self)])
		ifTrue:[^self toggleExpandedState: aMorph event: evt]. 
	aMorph ifNil:[^super mouseDown: evt].
	
	aMorph highlightForMouseDown.
	self setProperty: #highlightedMorph toValue: aMorph.
	
	selectors := Array 
		with: #click:
		with: nil
		with: nil
		with: (self dragEnabled ifTrue:[#startDrag:] ifFalse:[nil]).
	evt hand waitForClicksOrDrag: self event: evt selectors: selectors threshold: HandMorph dragThreshold "pixels".! !

!SimpleHierarchicalListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:40'!
mouseDown: event onItem: aMorph
	self removeObsoleteEventHandlers.
! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'ar 3/17/2001 17:26'!
mouseEnterDragging: evt
	| aMorph |
	(evt hand hasSubmorphs and:[self dropEnabled]) ifFalse: ["no d&d"
		^super mouseEnterDragging: evt].
	(self wantsDroppedMorph: evt hand firstSubmorph event: evt )
		ifTrue:[
			aMorph := self itemFromPoint: evt position.
			aMorph ifNotNil:[self potentialDropMorph: aMorph].
			evt hand newMouseFocus: self.
			"above is ugly but necessary for now"
		].! !

!SimpleHierarchicalListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:40'!
mouseEnterDragging: anEvent onItem: aMorph 
	self removeObsoleteEventHandlers.! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'mt 5/19/2016 11:21'!
mouseEnter: event

	super mouseEnter: event.
	
	Preferences mouseOverForKeyboardFocus
		ifTrue: [event hand newKeyboardFocus: self].! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'mt 11/8/2015 09:27'!
mouseLeaveDragging: anEvent

	self hoveredMorph: nil.
	(self dropEnabled and:[anEvent hand hasSubmorphs]) ifFalse: ["no d&d"
		^ super mouseLeaveDragging: anEvent].
	self resetPotentialDropMorph.
	anEvent hand releaseMouseFocus: self.
	"above is ugly but necessary for now"
! !

!SimpleHierarchicalListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:39'!
mouseLeaveDragging: anEvent onItem: aMorph 
	self removeObsoleteEventHandlers.! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'mt 5/19/2016 11:21'!
mouseLeave: aMouseEvent 

	super mouseLeave: aMouseEvent.
	self hoveredMorph: nil.
	
	Preferences mouseOverForKeyboardFocus
		ifTrue: [aMouseEvent hand releaseKeyboardFocus: self].! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'ar 3/17/2001 17:27'!
mouseMove: evt

	(self dropEnabled and:[evt hand hasSubmorphs]) 
		ifFalse:[^super mouseMove: evt].
	potentialDropMorph ifNotNil:[
		(potentialDropMorph containsPoint: (potentialDropMorph point: evt position from: self))
			ifTrue:[^self].
	].
	self mouseLeaveDragging: evt.
	(self containsPoint: evt position) 
		ifTrue:[self mouseEnterDragging: evt].! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'mt 5/27/2015 14:10'!
mouseUp: event
 
	| clickedMorph highlightedMorph |
	clickedMorph := self itemFromPoint: event position.
	highlightedMorph := self valueOfProperty: #highlightedMorph ifAbsent: [].
	self removeProperty: #highlightedMorph.

	((model okToChange not
		or: [clickedMorph isNil])
		or: [clickedMorph ~~ highlightedMorph]) ifTrue: [
			highlightedMorph ifNotNil: [:m | m highlightForMouseDown: false].
			^ self].
	
	((autoDeselect isNil or: [autoDeselect]) and: [clickedMorph == selectedMorph]) 
		ifTrue: [self setSelectedMorph: nil]
		ifFalse: [self setSelectedMorph: clickedMorph].

	highlightedMorph highlightForMouseDown: false.	
	event hand newKeyboardFocus: self.
	Cursor normal show.! !

!SimpleHierarchicalListMorph methodsFor: 'drawing' stamp: 'nk 3/8/2004 10:06'!
notExpandedForm

	^self class notExpandedForm! !

!SimpleHierarchicalListMorph methodsFor: 'private' stamp: 'RAA 8/2/1999 12:42'!
noteRemovalOfAll: aCollection

	scroller removeAllMorphsIn: aCollection.
	(aCollection includes: selectedMorph) ifTrue: [self setSelectedMorph: nil].! !

!SimpleHierarchicalListMorph methodsFor: 'scrolling' stamp: 'mt 5/21/2015 16:19'!
offsetToShow: aRectangle
	"Focus rectangles are too wide and the view hence jumps to the right. Snap it back to 0 on the horizontal axis."
	
	^ 0 @ (super offsetToShow: aRectangle) y! !

!SimpleHierarchicalListMorph methodsFor: 'initialization' stamp: 'RAA 8/1/1998 00:19'!
on: anObject list: getListSel selected: getSelectionSel changeSelected: setSelectionSel menu: getMenuSel keystroke: keyActionSel

	self model: anObject.
	getListSelector := getListSel.
	getSelectionSelector := getSelectionSel.
	setSelectionSelector := setSelectionSel.
	getMenuSelector := getMenuSel.
	keystrokeActionSelector := keyActionSel.
	autoDeselect := true.
	self borderWidth: 1.
	self list: self getList.! !

!SimpleHierarchicalListMorph methodsFor: 'drag and drop' stamp: 'panda 4/25/2000 17:39'!
potentialDropMorph
	^potentialDropMorph! !

!SimpleHierarchicalListMorph methodsFor: 'drag and drop' stamp: 'mir 5/8/2000 15:37'!
potentialDropMorph: aMorph
	potentialDropMorph := aMorph.
	aMorph highlightForDrop! !

!SimpleHierarchicalListMorph methodsFor: 'private' stamp: 'mt 2/1/2016 10:16'!
preferredSubmorphWidth

	^ 9999! !

!SimpleHierarchicalListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:38'!
removeObsoleteEventHandlers
	scroller submorphs do:[:m|
		m eventHandler: nil; highlightForMouseDown: false; resetExtension].! !

!SimpleHierarchicalListMorph methodsFor: 'drag and drop' stamp: 'mir 5/8/2000 15:38'!
resetPotentialDropMorph
	potentialDropMorph ifNotNil: [
		potentialDropMorph resetHighlightForDrop.
		potentialDropMorph := nil]
! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'dgd 9/26/2004 18:23'!
roots
	"Answer the receiver's roots"
	^ scroller submorphs
		select: [:each | each indentLevel isZero]! !

!SimpleHierarchicalListMorph methodsFor: 'geometry' stamp: 'mt 4/9/2015 17:24'!
scrollDeltaHeight
	^ scroller hasSubmorphs
		ifTrue: [scroller firstSubmorph height]
		ifFalse: [super scrollDeltaHeight]! !

!SimpleHierarchicalListMorph methodsFor: 'geometry' stamp: 'sps 3/9/2004 17:31'!
scrollDeltaWidth
"A guess -- assume that the width of a char is approx 1/2 the height of the font"
	^ self scrollDeltaHeight // 2


! !

!SimpleHierarchicalListMorph methodsFor: 'selection' stamp: 'panda 4/25/2000 18:56'!
selectedMorph
	^selectedMorph! !

!SimpleHierarchicalListMorph methodsFor: 'selection' stamp: 'RAA 8/1/1998 00:13'!
selectedMorph: aMorph

	self unhighlightSelection.
	selectedMorph := aMorph.
	self highlightSelection! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:13'!
selectionColor

	^ self valueOfProperty: #selectionColor ifAbsent: [Color blue]! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/29/2016 14:32'!
selectionColor: aColor

	self setProperty: #selectionColor toValue: aColor.
	scroller submorphsDo: [:ea | ea selectionColor: aColor].! !

!SimpleHierarchicalListMorph methodsFor: 'selection' stamp: 'mt 3/9/2015 07:50'!
selectionIndex: idx
	"Called internally to select the index-th item."
	| theMorph index |
	idx ifNil: [^ self].
	index := idx min: scroller submorphs size max: 0.
	(theMorph := index = 0 ifTrue: [nil] ifFalse: [scroller submorphs at: index])
		ifNotNil: [self scrollToShow: theMorph bounds].
	self selectedMorph: theMorph! !

!SimpleHierarchicalListMorph methodsFor: 'selection' stamp: 'nice 12/27/2009 03:10'!
selectionOneOf: aListOfItems
	"Set the selection to the first item in the list which is represented by one of my submorphs"

	
	aListOfItems do: [ :item | | index |
		index := scroller submorphs findFirst: [:m | 
			m withoutListWrapper = item withoutListWrapper
		].
		index > 0 ifTrue: [^self selectionIndex: index].
	].
	self selectionIndex: 0.! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:17'!
selectionTextColor

	^ self valueOfProperty: #selectionTextColor ifAbsent: [Color white]! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/29/2016 13:12'!
selectionTextColor: aColor

	self setProperty: #selectionTextColor toValue: aColor.
	scroller submorphsDo: [:ea | ea selectionTextColor: aColor].! !

!SimpleHierarchicalListMorph methodsFor: 'selection' stamp: 'RAA 8/31/1999 08:36'!
selection: item
	"Called from outside to request setting a new selection.
	Assumes scroller submorphs is exactly our list.
	Note: MAY NOT work right if list includes repeated items"

	| i |
	item ifNil: [^self selectionIndex: 0].
	i := scroller submorphs findFirst: [:m | m complexContents == item].
	i > 0 ifTrue: [^self selectionIndex: i].
	i := scroller submorphs findFirst: [:m | m withoutListWrapper = item withoutListWrapper].
	self selectionIndex: i! !

!SimpleHierarchicalListMorph methodsFor: 'initialization' stamp: 'mt 7/24/2016 15:05'!
setDefaultParameters

	super setDefaultParameters.
	
	self
		font: (self userInterfaceTheme font ifNil: [TextStyle defaultFont]);
		textColor: (self userInterfaceTheme textColor ifNil: [Color black]).
		
	self
		selectionColor: (self userInterfaceTheme selectionColor ifNil: [Color r: 0.72 g: 0.72 b: 0.9]);
		selectionTextColor: (self userInterfaceTheme selectionTextColor ifNil: [Color black]);
		filterColor: (self userInterfaceTheme filterColor ifNil: [Color yellow paler]);
		filterTextColor: (self userInterfaceTheme filterTextColor ifNil: [Color black]);
		lineColor: (self userInterfaceTheme lineColor ifNil: [Color veryLightGray]);
		hoverColor: ((self userInterfaceTheme hoverSelectionModifier ifNil: [ [:c | c darker alpha: 0.3] ]) value: self selectionColor);
		highlightTextColor: (self userInterfaceTheme highlightTextColor ifNil: [Color red]).! !

!SimpleHierarchicalListMorph methodsFor: 'selection' stamp: 'mt 11/4/2015 13:58'!
setSelectedMorph: aMorph

	"Avoid unnecessary model callbacks."
	self selectedMorph == aMorph ifTrue: [^ self].

	model 
		perform: (setSelectionSelector ifNil: [^self]) 
		with: aMorph complexContents	"leave last wrapper in place"

 ! !

!SimpleHierarchicalListMorph methodsFor: 'keyboard navigation' stamp: 'mt 3/8/2015 18:08'!
setSelectionIndex: idx
	"Called internally to select the index-th item."
	| theMorph index max currentIndex |
	idx ifNil: [^ self].
	max := scroller submorphs size.
	currentIndex := self getSelectionIndex.
	
	index := idx min: max max: 0.
	theMorph := index = 0 ifTrue: [nil] ifFalse: [scroller submorphs at: index].

	"Skip invisible rows."
	[theMorph notNil and: [theMorph visible not]] whileTrue: [
		currentIndex < index
			ifTrue: [index := index + 1]
			ifFalse: [index := index - 1].		
		(index < 1 or: [index > max]) ifTrue: [^ self].
		theMorph := scroller submorphs at: index.
	].
	self setSelectedMorph: theMorph.! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'RAA 7/20/1998 12:09'!
sortingSelector: s

	sortingSelector := s! !

!SimpleHierarchicalListMorph methodsFor: 'event handling' stamp: 'mt 3/8/2015 16:30'!
specialKeyPressed: asciiValue

	^ self arrowKey: asciiValue! !

!SimpleHierarchicalListMorph methodsFor: 'drag and drop' stamp: 'mt 11/4/2015 18:41'!
startDrag: evt 
	
	| itemMorph |
	evt hand hasSubmorphs ifTrue: [^ self].
	self model okToChange ifFalse: [^ self].
	
	itemMorph := scroller submorphs
		detect: [:any | any highlightedForMouseDown]
		ifNone: [^ self].

	"Prepare visuals."
	itemMorph highlightForMouseDown: false.
	self setSelectedMorph: itemMorph.

	[ "Initiate drag."
		(self model dragPassengerFor: itemMorph inMorph: self) ifNotNil: [:passenger | | ddm |
			ddm := (self valueOfProperty: #dragTransferClass ifAbsent: [TransferMorph]) withPassenger: passenger from: self.
			ddm dragTransferType: (self model dragTransferTypeForMorph: self).
			ddm updateFromUserInputEvent: evt.
			self model dragStartedFor: itemMorph transferMorph: ddm.
			evt hand grabMorph: ddm].
	] ensure: [
		Cursor normal show.
		evt hand releaseMouseFocus: self].! !

!SimpleHierarchicalListMorph methodsFor: 'obsolete' stamp: 'ar 3/17/2001 17:39'!
startDrag: evt onItem: itemMorph 
	self removeObsoleteEventHandlers.! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 09:54'!
textColor
	^ textColor ifNil: [Color black]! !

!SimpleHierarchicalListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 09:53'!
textColor: aColor
	textColor := aColor.
	scroller submorphsDo: [:ea | ea color: textColor].! !

!SimpleHierarchicalListMorph methodsFor: 'keyboard navigation' stamp: 'ar 10/14/2003 23:39'!
toggleExpandedState: aMorph
	aMorph toggleExpandedState.
	self adjustSubmorphPositions.
! !

!SimpleHierarchicalListMorph methodsFor: 'events' stamp: 'ar 3/17/2001 17:39'!
toggleExpandedState: aMorph event: event
	| oldState |
	"self setSelectedMorph: aMorph."
	event yellowButtonPressed ifTrue: [
		oldState := aMorph isExpanded.
		scroller submorphs copy do: [ :each |
			(each canExpand and: [each isExpanded = oldState]) ifTrue: [
				each toggleExpandedState.
			].
		].
	] ifFalse: [
		aMorph toggleExpandedState.
	].
	self adjustSubmorphPositions.
	! !

!SimpleHierarchicalListMorph methodsFor: 'drawing' stamp: 'RAA 8/3/1999 09:44'!
unhighlightSelection
	selectedMorph ifNotNil: [selectedMorph unhighlight]! !

!SimpleHierarchicalListMorph methodsFor: 'updating' stamp: 'dgd 2/21/2003 23:16'!
update: aSymbol 
	aSymbol == getSelectionSelector 
		ifTrue: 
			[self selection: self getCurrentSelectionItem.
			^self].
	aSymbol == getListSelector 
		ifTrue: 
			[self list: self getList.
			^self].
	((aSymbol isKindOf: Array) 
		and: [aSymbol notEmpty and: [aSymbol first == #openPath]]) 
			ifTrue: 
				[^(scroller submorphs at: 1 ifAbsent: [^self]) 
					openPath: aSymbol allButFirst]! !

!SimpleHierarchicalListMorph methodsFor: 'scrolling' stamp: 'sps 12/26/2002 13:37'!
vUnadjustedScrollRange
"Return the width of the widest item in the list"

	(scroller submorphs size > 0) ifFalse:[ ^0 ].
	^scroller submorphs last fullBounds bottom
! !

!SimpleHierarchicalListMorph methodsFor: 'drag and drop' stamp: 'panda 4/25/2000 17:38'!
wantsDroppedMorph: aMorph event: anEvent 
	^ self model wantsDroppedMorph: aMorph event: anEvent inMorph: self! !

!SimpleHierarchicalListMorph methodsFor: 'private - caching' stamp: 'mt 5/1/2016 11:27'!
widthOfColumn: index 

	self columns ifNil: [^ nil].
	
	"Try columns cache."
	(self columnsCache at: index)
		ifNotNil: [:cachedWidth | ^ cachedWidth].
		
	^ (self columns at: index)
		ifNil: [nil]
		ifNotNil: [:widthOrSpec |
			"Fill columns cache."
			self columnsCache
				at: index
				put: (widthOrSpec isNumber
					ifTrue: [widthOrSpec]
					ifFalse: [widthOrSpec isBlock
						ifTrue: [widthOrSpec cull: self]
						ifFalse: [50 "fall back for bad spec"]])]! !

MorphicModel subclass: #Slider
	instanceVariableNames: 'slider value setValueSelector getValueSelector sliderShadow sliderColor descending minimumValue maximumValue quantum getMinimumValueSelector getMaximumValueSelector orientation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!Slider commentStamp: 'mt 2/12/2016 10:21' prior: 0!
I am a widget that helps users enter a numerically bounded value. I have a minimum and a maximum value. My thumb indicates my current value. My orientation can be vertical or horizontal, which is usually determined by the ratio of my extent. Nevertheless, you can force me to use a specific orientation,

Note that I can be pluggable and get data from a model.!

Slider subclass: #BracketSliderMorph
	instanceVariableNames: 'getEnabledSelector enabled'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!BracketSliderMorph commentStamp: 'gvc 5/18/2007 13:39' prior: 0!
Abstract superclass for morphs that are used to select a component (R, G, B or A) of a colour.!

BracketSliderMorph subclass: #AColorSelectorMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!AColorSelectorMorph commentStamp: 'gvc 5/18/2007 13:52' prior: 0!
ColorComponentSelector showing an alpha gradient over a hatched background.!

!AColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/3/2009 13:43'!
color: aColor
	"Set the gradient colors."
	
	super color: aColor asNontranslucentColor.
	self fillStyle: self defaultFillStyle! !

!AColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'mt 4/16/2015 23:30'!
defaultFillStyle
	"Answer the hue gradient."

	^(GradientFillStyle colors: {self color alpha: 0. self color})
		origin: self topLeft;
		direction: (self orientation == #horizontal
					ifTrue: [self width@0]
					ifFalse: [0@self height])! !

!AColorSelectorMorph methodsFor: 'drawing' stamp: 'cmm 12/3/2010 11:28'!
drawOn: aCanvas 
	"Draw a hatch pattern first."
	aCanvas
		fillRectangle: self innerBounds
		fillStyle: (InfiniteForm with: ColorPresenterMorph hatchForm).
	super drawOn: aCanvas! !

!AColorSelectorMorph methodsFor: 'visual properties' stamp: 'gvc 9/19/2006 15:46'!
fillStyle: fillStyle
	"If it is a color then override with gradient."
	
	fillStyle isColor
		ifTrue: [self color: fillStyle]
		ifFalse: [super fillStyle: fillStyle]! !

!AColorSelectorMorph methodsFor: 'initialization' stamp: 'gvc 9/26/2006 11:54'!
initialize
	"Initialize the receiver."

	super initialize.
	self
		value: 1.0;
		color: Color black! !

!BracketSliderMorph methodsFor: 'accessing - ui' stamp: 'mt 4/8/2015 17:46'!
adoptPaneColor: paneColor 
	"Pass on to the border too."
	super adoptPaneColor: paneColor.
	paneColor ifNil: [ ^ self ].
	self
		 fillStyle: self fillStyleToUse ;
		 borderStyle:
			(BorderStyle inset
				 width: 1 ;
				 baseColor: self color twiceDarker) ;
		 sliderColor:
			(self enabled
				ifTrue: [ paneColor twiceDarker ]
				ifFalse: [ self paneColor twiceDarker paler ])! !

!BracketSliderMorph methodsFor: 'accessing - ui' stamp: 'mt 4/8/2015 17:47'!
defaultColor
	"Answer the default color/fill style for the receiver."
	
	^Color white! !

!BracketSliderMorph methodsFor: 'accessing - ui' stamp: 'gvc 9/3/2009 13:40'!
defaultFillStyle
	"Answer the defauolt fill style."

	^Color gray! !

!BracketSliderMorph methodsFor: 'as yet unclassified' stamp: 'mt 4/8/2015 17:46'!
disable
	"Disable the receiver."
	
	self enabled: false! !

!BracketSliderMorph methodsFor: 'as yet unclassified' stamp: 'mt 4/8/2015 17:46'!
enable
	"Enable the receiver."
	
	self enabled: true! !

!BracketSliderMorph methodsFor: 'accessing' stamp: 'mt 4/8/2015 17:44'!
enabled
	"Answer the value of enabled"

	^ enabled! !

!BracketSliderMorph methodsFor: 'accessing' stamp: 'mt 4/8/2015 17:45'!
enabled: anObject
	"Set the value of enabled"

	enabled = anObject ifTrue: [^self].
	enabled := anObject.
	self changed: #enabled.
	self
		adoptPaneColor: self color;
		changed! !

!BracketSliderMorph methodsFor: 'geometry' stamp: 'gvc 9/3/2009 13:40'!
extent: aPoint
	"Update the gradient directions."

	super extent: aPoint.
	self updateFillStyle! !

!BracketSliderMorph methodsFor: 'accessing - ui' stamp: 'gvc 8/7/2007 16:05'!
fillStyleToUse
	"Answer the fillStyle that should be used for the receiver."
	
	^self fillStyle! !

!BracketSliderMorph methodsFor: 'accessing' stamp: 'mt 4/8/2015 17:45'!
getEnabledSelector
	"Answer the value of getEnabledSelector"

	^ getEnabledSelector! !

!BracketSliderMorph methodsFor: 'accessing' stamp: 'mt 4/8/2015 17:45'!
getEnabledSelector: aSymbol
	"Set the value of getEnabledSelector"

	getEnabledSelector := aSymbol.
	self updateEnabled! !

!BracketSliderMorph methodsFor: 'accessing - ui' stamp: 'gvc 9/19/2006 14:06'!
gradient
	"Answer the gradient."

	self subclassResponsibility! !

!BracketSliderMorph methodsFor: 'event handling' stamp: 'mt 4/8/2015 17:47'!
handlesMouseDown: evt
	"Answer true." 

	^true! !

!BracketSliderMorph methodsFor: 'initialization' stamp: 'mt 4/8/2015 17:48'!
initialize
	"Initialize the receiver."

	super initialize.
	self
		enabled: true;
		fillStyle: self defaultFillStyle;
		borderStyle: (BorderStyle inset baseColor: self color; width: 1);
		sliderColor: Color black;
		clipSubmorphs: true! !

!BracketSliderMorph methodsFor: 'initialization' stamp: 'mt 4/16/2015 23:30'!
initializeSlider
	"Make the slider raised."
	
	slider :=( BracketMorph newBounds: self totalSliderArea)
		horizontal: self orientation == #horizontal;
		color: self thumbColor;
		borderStyle: (BorderStyle raised baseColor: Color white; width: 1).
	sliderShadow := (BracketMorph newBounds: self totalSliderArea)
		horizontal: self orientation == #horizontal;
		color: self pagingArea color;
		borderStyle: (BorderStyle inset baseColor: (Color white alpha: 0.6); width: 1).
	slider on: #mouseMove send: #scrollAbsolute: to: self.
	slider on: #mouseDown send: #mouseDownInSlider: to: self.
	slider on: #mouseUp send: #mouseUpInSlider: to: self.
	"(the shadow must have the pagingArea as its owner to highlight properly)"
	self pagingArea addMorph: sliderShadow.
	sliderShadow hide.
	self addMorph: slider.
	self computeSlider.
! !

!BracketSliderMorph methodsFor: 'layout' stamp: 'mt 4/8/2015 17:50'!
minHeight
	"Answer the receiver's minimum height.
	Give it a bit of a chance..."
	
	^8 max: super minHeight! !

!BracketSliderMorph methodsFor: 'other events' stamp: 'mt 4/8/2015 17:50'!
mouseDownInSlider: event
	"Ignore if disabled."
	
	self enabled ifFalse: [^self].
	^super mouseDownInSlider: event! !

!BracketSliderMorph methodsFor: 'event handling' stamp: 'mt 4/8/2015 17:47'!
mouseDown: anEvent
	"Set the value directly."
	
	self enabled ifTrue: [
		self
			scrollPoint: anEvent;
			computeSlider].
	super mouseDown: anEvent.
	self enabled ifFalse: [^self].
	anEvent hand newMouseFocus: slider event: anEvent.
	slider
		mouseEnter: anEvent copy;
		mouseDown: anEvent copy
! !

!BracketSliderMorph methodsFor: 'geometry' stamp: 'mt 4/16/2015 23:30'!
roomToMove
	"Allow to run off the edges a bit."
	
	^self orientation == #horizontal
		ifTrue: [self totalSliderArea insetBy: ((self sliderThickness // 2@0) negated corner: (self sliderThickness // 2 + 1)@0)]
		ifFalse: [self totalSliderArea insetBy: (0@(self sliderThickness // 2) negated corner: 0@(self sliderThickness // 2 - (self sliderThickness \\ 2) + 1))]! !

!BracketSliderMorph methodsFor: 'scrolling' stamp: 'mt 4/8/2015 17:50'!
scrollAbsolute: event
	"Ignore if disabled."
	
	self enabled ifFalse: [^self].
	^super scrollAbsolute: event! !

!BracketSliderMorph methodsFor: 'event handling' stamp: 'mt 4/16/2015 23:31'!
scrollPoint: event
	"Scroll to the event position."
	
	| r p |
	r := self roomToMove.
	self orientation == #horizontal
		ifTrue: [r width = 0 ifTrue: [^ self]]
		ifFalse: [r height = 0 ifTrue: [^ self]].
	p := event position - (self sliderThickness // 2) adhereTo: r.
	self descending
		ifFalse:
			[self setValue: (self orientation == #horizontal 
				ifTrue: [(p x - r left) asFloat / r width]
				ifFalse: [(p y - r top) asFloat / r height])]
		ifTrue:
			[self setValue: (self orientation == #horizontal
				ifTrue: [(r right - p x) asFloat / r width]
				ifFalse:	[(r bottom - p y) asFloat / r height])]! !

!BracketSliderMorph methodsFor: 'accessing - ui' stamp: 'gvc 9/8/2009 13:29'!
sliderColor: newColor
	"Set the slider colour."
	
	super sliderColor: (self enabled ifTrue: [Color black] ifFalse: [self sliderShadowColor]).
	slider ifNotNil: [slider borderStyle baseColor: Color white]! !

!BracketSliderMorph methodsFor: 'accessing - ui' stamp: 'gvc 9/19/2006 15:43'!
sliderShadowColor
	"Answer the color for the slider shadow."
	
	^Color black alpha: 0.6! !

!BracketSliderMorph methodsFor: 'geometry' stamp: 'mt 4/16/2015 23:31'!
sliderThickness
	"Answer the thickness of the slider."
	
	^((self orientation == #horizontal
		ifTrue: [self height]
		ifFalse: [self width]) // 2 max: 8) // 2 * 2 + 1! !

!BracketSliderMorph methodsFor: 'testing' stamp: 'mt 4/8/2015 17:47'!
updateEnabled
	"Update the enablement state."

	self model ifNotNil: [
		self getEnabledSelector ifNotNil: [
			self enabled: (self model perform: self getEnabledSelector)]]! !

!BracketSliderMorph methodsFor: 'initialization' stamp: 'mt 4/16/2015 23:31'!
updateFillStyle
	"Update the fill style directions."

	|b fs|
	fs := self fillStyle.
	fs isOrientedFill ifTrue: [
		b := self innerBounds.
		fs origin: b topLeft.
		fs direction: (self orientation == #horizontal
			ifTrue: [b width@0]
			ifFalse: [0@b height])]! !

!BracketSliderMorph methodsFor: 'initialization' stamp: 'mt 4/16/2015 23:31'!
updateSlider

	super updateSlider.
	
	slider horizontal: self orientation == #horizontal.
	sliderShadow horizontal: self orientation == #horizontal.! !

!BracketSliderMorph methodsFor: 'updating' stamp: 'mt 4/8/2015 17:50'!
update: aSymbol
	"Update the value."
	
	super update: aSymbol.
	aSymbol == self getEnabledSelector ifTrue: [
		^self updateEnabled].! !

BracketSliderMorph subclass: #HColorSelectorMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!HColorSelectorMorph commentStamp: 'gvc 5/18/2007 12:58' prior: 0!
ColorComponentSelector showing a hue rainbow palette.!

!HColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/19/2006 14:13'!
color: aColor
	"Ignore to preserve fill style."
	! !

!HColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'mt 4/16/2015 23:31'!
defaultFillStyle
	"Answer the hue gradient."

	^(GradientFillStyle colors: ((0.0 to: 359.9 by: 0.1) collect: [:a | Color h: a s: 1.0 v: 1.0]))
		origin: self topLeft;
		direction: (self orientation == #horizontal
					ifTrue: [self width@0]
					ifFalse: [0@self height])! !

Slider subclass: #ScrollBar
	instanceVariableNames: 'menuButton upButton downButton pagingArea scrollDelta pageDelta interval menuSelector timeOfMouseDown timeOfLastScroll nextPageDirection currentScrollDelay'
	classVariableNames: 'ArrowImagesCache BoxesImagesCache GradientScrollBar RoundedScrollBarLook ScrollBarsWithoutArrowButtons ScrollBarsWithoutMenuButton UpArrow UpArrow8Bit'
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!ScrollBar commentStamp: '<historical>' prior: 0!
Inspired by an oiginal design of Hans-Martin Mosner, this ScrollBar is intended to exercise the handling of input events in Morphic.  With sufficient flexibility in this area, all particular behavior can be concentrated in this single class with no need to specialize any other morphs to achieve button, slider and menu-button behavior.

Once we have this working, put in logic for horizontal operation as well.

CachedImages was added to reduce the number of forms created and thrown away. This will be helpful for Nebraska and others as well.!

!ScrollBar class methodsFor: 'as yet unclassified' stamp: 'dew 3/23/2002 01:30'!
alwaysShowFlatScrollbarForAlternativeLook
	"Set this value to true, if you want to see the flat scrollbar look in flop-out mode as well as inboard.  Otherwise the flop-out scrollbar will be rounded and inboard will be flat."
	^ false! !

!ScrollBar class methodsFor: 'images' stamp: 'dgd 3/28/2003 10:22'!
arrowOfDirection: aSymbol size: finalSizeInteger color: aColor 
	"answer a form with an arrow based on the parameters"
	^ ArrowImagesCache at: {aSymbol. finalSizeInteger. aColor}! !

!ScrollBar class methodsFor: 'images - samples' stamp: 'sd 11/8/2003 16:02'!
arrowSamples
	"create a set of arrow with different sizes, colors and directions"
	" 
	ScrollBar arrowSamples.  
	"
	| column |
	column := AlignmentMorph newColumn vResizing: #shrinkWrap;
				 hResizing: #shrinkWrap;
				 layoutInset: 1;
				 borderColor: Color black;
				 borderWidth: 0;
				 wrapCentering: #center;
				 cellPositioning: #center;
				 color: Color white;
				 yourself.
	
	self sampleSizes
		do: [:size | 
			| row | 
			row := AlignmentMorph newRow color: Color transparent;
						 vResizing: #shrinkWrap;
						 cellInset: 2 @ 0 yourself.
			
			self sampleColors
				do: [:color | 
					#(#top #right #bottom #left )
						do: [:direction | 
							row addMorphBack: (ScrollBar
									arrowOfDirection: direction
									size: size
									color: color) asMorph]].
			
			column addMorphBack: row].
	
	column openInHand! !

!ScrollBar class methodsFor: 'images' stamp: 'dgd 3/28/2003 10:22'!
boxOfSize: finalSizeInteger color: aColor 
	"answer a form with an box based on the parameters"
	^ BoxesImagesCache at: {finalSizeInteger. aColor}! !

!ScrollBar class methodsFor: 'images - samples' stamp: 'dgd 3/28/2003 10:18'!
boxSamples
	"create a set of box with different sizes and colors"
	" 
	ScrollBar boxSamples.  
	"
	| column |
	column := AlignmentMorph newColumn vResizing: #shrinkWrap;
				 hResizing: #shrinkWrap;
				 layoutInset: 1;
				 borderColor: Color black;
				 borderWidth: 0;
				 wrapCentering: #center;
				 cellPositioning: #center;
				 color: Color white;
				 yourself.
	""
	self sampleSizes
		do: [:size | 
			| row | 
			row := AlignmentMorph newRow color: Color transparent;
						 vResizing: #shrinkWrap;
						 cellInset: 2 @ 0 yourself.
			""
			self sampleColors
				do: [:color | 
					row addMorphBack: (ScrollBar boxOfSize: size color: color) asMorph].
			""
			column addMorphBack: row].
	""
	""
	column openInHand! !

!ScrollBar class methodsFor: 'images' stamp: 'dgd 3/28/2003 20:54'!
changesInPreferences
	"the related preferences changed"
	self initializeImagesCache
	" ScrollBar allInstances do: [:each | each removeAllMorphs; initializeSlider] "! !

!ScrollBar class methodsFor: 'class initialization' stamp: 'ar 2/26/2010 23:05'!
cleanUp
	"Re-initialize the image cache"

	self initializeImagesCache! !

!ScrollBar class methodsFor: 'class initialization' stamp: 'dgd 3/28/2003 10:13'!
createArrowImagesCache
	"creates the cache to store the arrow forms"
	^ LRUCache
		size: 40
		factory: [:key | ""
			self
				createArrowOfDirection: key first
				size: key second
				color: key third]! !

!ScrollBar class methodsFor: 'images' stamp: 'mt 7/23/2016 17:05'!
createArrowOfDirection: aSymbolDirection size: size color: aColor 
	"PRIVATE - create an arrow with aSymbolDirectionDirection, finalSizeInteger and aColor  
	aSymbolDirectionDirection = #top, #bottom. #left or #right  
	 ScrollBar initializeImagesCache.
	Try with: 
	(ScrollBar createArrowOfDirection: #right size: 14 color: Color lightGreen) asMorph openInHand."
	
	| form canvas vertices margin |
	form := Form extent: size asPoint depth: 32.
	canvas := form getCanvas.
	margin := size < 14 ifTrue: [2] ifFalse: [3].
	vertices := {
		size // 2 @ margin.
		size asPoint - margin asPoint.
		margin @ (size-margin).
		size // 2 @ margin}.
	
	"Preferences gradientScrollBars
		ifTrue: [
			fillStyle := GradientFillStyle ramp: {
					0.0 -> (aColor adjustBrightness: 0.5).
					0.1-> (aColor adjustBrightness: 0.05).
					0.6 -> (aColor darker)}.
			fillStyle origin: size // 2 @ margin.
			fillStyle direction: 0 @ size]
		ifFalse: [
			fillStyle := SolidFillStyle color: aColor]. "
	
	canvas
		drawPolygon: vertices
		fillStyle: (SolidFillStyle color: aColor).
	
	^ (form rotateBy: (aSymbolDirection caseOf: {
		[#top] -> [0].
		[#bottom] -> [180].
		[#left] -> [270].
		[#right] -> [90]})) clippedToSize: size asPoint! !

!ScrollBar class methodsFor: 'class initialization' stamp: 'dgd 3/28/2003 10:13'!
createBoxImagesCache
	"creates the cache to store the arrow forms"
	^ LRUCache
		size: 20
		factory: [:key | self createBoxOfSize: key first color: key second]! !

!ScrollBar class methodsFor: 'images' stamp: 'mt 7/23/2016 17:05'!
createBoxOfSize: anInteger color: aColor 
	"PRIVATE - create a box with anInteger and aColor  
	 
	Try with:  
	(ScrollBar createBoxOfSize: 32 color: Color gray) asMorph  
	openInHand.  
	"
	
	| form canvas margin |
	form := Form extent: anInteger asPoint depth: 32.
	canvas := form getCanvas.
	margin := anInteger < 14 ifTrue: [3] ifFalse: [4].

	"Preferences gradientScrollBars
		ifTrue: [
			fillStyle := GradientFillStyle ramp: {
					0.0 -> (aColor adjustBrightness: 0.5).
					0.1-> (aColor adjustBrightness: 0.05).
					0.6 -> (aColor darker)}.
			fillStyle origin: margin asPoint.
			fillStyle direction: anInteger @ 0]
		ifFalse: [
			fillStyle := SolidFillStyle color: aColor]. "
	
	canvas
		fillRectangle: ((0@0 extent: anInteger asPoint) insetBy: margin)
		fillStyle: (SolidFillStyle color: aColor).
		
	^ form! !

!ScrollBar class methodsFor: 'preferences' stamp: 'mt 7/28/2016 11:49'!
gradientScrollBar

	<preference: 'gradientScrollBar'
		categoryList: #(windows scrolling)
		description: 'If true, scroll bars will have a gradient look.'
		type: #Boolean>
	^ GradientScrollBar ifNil: [false]! !

!ScrollBar class methodsFor: 'preferences' stamp: 'mt 7/28/2016 11:51'!
gradientScrollBar: aBoolean

	GradientScrollBar = aBoolean ifTrue: [^ self].
	GradientScrollBar := aBoolean.

	ScrollBar allSubInstancesDo: [:ea | ea updateSliderColor].! !

!ScrollBar class methodsFor: 'class initialization' stamp: 'ar 9/18/2009 22:18'!
initialize
	"ScrollBar initialize"
	UpArrow := Form
				extent: 6 @ 3
				fromArray: #(805306368 2013265920 4227858432 )
				offset: 0 @ 0.
	""
	self initializeImagesCache.

	"Adjustments for FaceLift"
	Preferences disable: #scrollBarsNarrow.
	SystemWindow allSubInstancesDo:[:w| w updatePaneColors].
	ScrollPane allSubInstancesDo:[:pane| pane hideOrShowScrollBars].
! !

!ScrollBar class methodsFor: 'class initialization' stamp: 'md 3/23/2006 15:55'!
initializeImagesCache
	"initialize the receiver's ImagesCache. 
	 
	normally this method is not evaluated more than in the class 
	initializazion. "

	" 
	ScrollBar initializeImagesCache.
	"

	ArrowImagesCache := self createArrowImagesCache.
	BoxesImagesCache := self createBoxImagesCache! !

!ScrollBar class methodsFor: 'class initialization' stamp: 'mt 4/9/2015 09:51'!
refreshAllScrollBars

	ScrollBar allSubInstances do: [:s |
		s updateSlider].! !

!ScrollBar class methodsFor: 'preferences' stamp: 'mt 4/1/2015 11:01'!
roundedScrollBarLook

	<preference: 'roundedScrollBarLook'
		category: #scrolling
		description: 'If true, morphic scrollbars will look rounded.'
		type: #Boolean>
	^ RoundedScrollBarLook ifNil: [false]! !

!ScrollBar class methodsFor: 'preferences' stamp: 'mt 7/25/2016 13:51'!
roundedScrollBarLook: aBoolean

	RoundedScrollBarLook = aBoolean ifTrue: [^ self].
	RoundedScrollBarLook := aBoolean.
	
	ScrollBar allSubInstancesDo: [:ea | ea updateSliderCornerStyle].! !

!ScrollBar class methodsFor: 'images - samples' stamp: 'dgd 3/28/2003 10:18'!
sampleColors
	"private"
	^ (Color lightCyan wheel: 5)! !

!ScrollBar class methodsFor: 'images - samples' stamp: 'dgd 3/28/2003 10:17'!
sampleSizes
	
"private"
	^ #(10 12 14 16 18 32 64 )! !

!ScrollBar class methodsFor: 'preferences' stamp: 'mt 4/1/2015 11:08'!
scrollBarsWithoutArrowButtons

	<preference: 'scrollBarsWithoutArrowButtons'
		category: #scrolling
		description: 'If true, morphic scrollbars will not include arrow buttons but only the slider.'
		type: #Boolean>
	^ ScrollBarsWithoutArrowButtons ifNil: [false]! !

!ScrollBar class methodsFor: 'preferences' stamp: 'mt 4/9/2015 09:52'!
scrollBarsWithoutArrowButtons: aBoolean

	ScrollBarsWithoutArrowButtons = aBoolean ifTrue: [^ self].
	ScrollBarsWithoutArrowButtons := aBoolean.
	self refreshAllScrollBars.! !

!ScrollBar class methodsFor: 'preferences' stamp: 'mt 4/9/2015 09:46'!
scrollBarsWithoutMenuButton

	<preference: 'scrollBarsWithoutMenuButton'
		category: #scrolling
		description: 'If true, morphic scrollbars will not include a menu button.'
		type: #Boolean>
	^ ScrollBarsWithoutMenuButton ifNil: [false]! !

!ScrollBar class methodsFor: 'preferences' stamp: 'mt 4/9/2015 09:52'!
scrollBarsWithoutMenuButton: aBoolean

	ScrollBarsWithoutMenuButton = aBoolean ifTrue: [^ self].
	ScrollBarsWithoutMenuButton := aBoolean.
	self refreshAllScrollBars.! !

!ScrollBar class methodsFor: 'preferences' stamp: 'mt 7/23/2016 16:35'!
themeProperties

	^ super themeProperties, {		
		{ #pagingAreaColorModifier. 'Colors'. 'How to modify the paging area color when adopting a pane color.' }.
		{ #thumbColorModifier. 'Colors'. 'How to modify the thumb color when adopting a pane color.' }.
		{ #borderColorModifier. 'Colors'. 'How to modify the border color when adopting a pane color.' }.
	}! !

!ScrollBar methodsFor: 'updating' stamp: 'md 2/24/2006 16:27'!
adoptPaneColor: aColor
	"Adopt the given pane color"
	aColor ifNil:[^self].
	self sliderColor: aColor.! !

!ScrollBar methodsFor: 'updating' stamp: 'mt 7/25/2016 15:00'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 4/9/2015 09:16'!
boundsForDownButton
	
	^ self innerBounds bottomRight - self buttonExtent 
		extent: self buttonExtent! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 4/9/2015 09:16'!
boundsForMenuButton

	^ self innerBounds topLeft extent: self buttonExtent! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 4/16/2015 23:31'!
boundsForUpButton

	^ (self menuButton visible
		ifFalse: [self innerBounds topLeft]
		ifTrue: [self orientation == #horizontal
			ifTrue: [self menuButton bounds topRight - (1@0)]
			ifFalse: [self menuButton bounds bottomLeft - (0@1)]])
		extent: self buttonExtent! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 4/16/2015 23:32'!
buttonExtent
	^ self orientation == #horizontal
		ifTrue: [self innerBounds height asPoint]
		ifFalse: [self innerBounds width asPoint]! !

!ScrollBar methodsFor: 'scrolling' stamp: 'mt 4/8/2015 17:57'!
doScrollByPage
	"Scroll automatically while mouse is down"
	(self waitForDelay1: 300 delay2: 100) ifFalse: [^ self].
	nextPageDirection
		ifTrue: [self setValue: value + pageDelta]
		ifFalse: [self setValue: value - pageDelta]
! !

!ScrollBar methodsFor: 'scrolling' stamp: 'mt 4/8/2015 17:58'!
doScrollDown
	"Scroll automatically while mouse is down"
	(self waitForDelay1: 200 delay2: 40) ifFalse: [^ self].
	self setValue: value + scrollDelta.! !

!ScrollBar methodsFor: 'scrolling' stamp: 'mt 4/8/2015 17:58'!
doScrollUp
	"Scroll automatically while mouse is down"
	(self waitForDelay1: 200 delay2: 40) ifFalse: [^ self].
	self setValue: value - scrollDelta.! !

!ScrollBar methodsFor: 'accessing' stamp: 'mt 7/23/2016 13:04'!
downButton
	^ downButton! !

!ScrollBar methodsFor: 'images' stamp: 'mt 7/23/2016 16:40'!
downImage
	"answer a form to be used in the down button"
	^ self class
		arrowOfDirection: (self orientation == #horizontal
				ifTrue: [#right]
				ifFalse: [#bottom])
		size: (self buttonExtent x min: self buttonExtent y)
		color: self imageColor! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 4/16/2015 23:32'!
expandSlider
	"Compute the new size of the slider (use the old sliderThickness as a minimum)."
	| r |
	r := self totalSliderArea.
	slider extent: (self orientation == #horizontal
		ifTrue: [(((r width * self interval) asInteger max: self minThumbThickness) min: r width) @ slider height]
		ifFalse: [slider width @ (((r height * self interval) asInteger max: self minThumbThickness) min: r height)])! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 7/27/2016 10:05'!
extent: p 

	(bounds extent closeTo: p) ifTrue: [^ self].
	
	p x > p y
		ifTrue: [super
				extent: (p max: 8 @ 4)]
		ifFalse: [super
				extent: (p max: 4 @ 8)].
			
	self updateSliderColor.
	! !

!ScrollBar methodsFor: 'scrolling' stamp: 'mt 7/23/2016 13:10'!
finishedScrolling
	self stopStepping.
	self scrollBarAction: nil.
! !

!ScrollBar methodsFor: 'testing' stamp: 'mt 4/9/2015 11:31'!
hasButtons

	^ (self menuButton visible or: [upButton visible]) or: [downButton visible]! !

!ScrollBar methodsFor: 'accessing' stamp: 'mt 7/31/2016 13:03'!
imageColor

	^ slider ifNil: [Color black] ifNotNil: [:s | 
		(s color = s borderColor or: [s borderColor isTransparent])
			ifTrue: [s color adjustBrightness: (s color brightness < 0.5 ifTrue: [0.2] ifFalse: [-0.2])]
			ifFalse: [s borderColor]]! !

!ScrollBar methodsFor: 'initialize' stamp: 'mt 7/23/2016 12:21'!
initialize

	interval := 0.2.
	scrollDelta := 0.02.
	pageDelta := 0.2.
	
	super initialize.! !

!ScrollBar methodsFor: 'initialize' stamp: 'mt 7/23/2016 16:46'!
initializeDownButton
	"initialize the receiver's downButton"

	downButton := RectangleMorph newBounds: self boundsForDownButton.
	downButton 
		on: #mouseDown
		send: #scrollDownInit
		to: self.
	downButton 
		on: #mouseUp
		send: #finishedScrolling
		to: self.
	self updateDownButtonImage.	
	self addMorphFront: downButton.
	downButton visible: self class scrollBarsWithoutArrowButtons not.! !

!ScrollBar methodsFor: 'initialize' stamp: 'mt 7/23/2016 16:46'!
initializeMenuButton
"initialize the receiver's menuButton"
	"Preferences disable: #scrollBarsWithoutMenuButton"
	"Preferences enable: #scrollBarsWithoutMenuButton"
	menuButton := RectangleMorph newBounds: self boundsForMenuButton.
	menuButton
		on: #mouseEnter
		send: #menuButtonMouseEnter:
		to: self.
	menuButton
		on: #mouseDown
		send: #menuButtonMouseDown:
		to: self.
	menuButton
		on: #mouseLeave
		send: #menuButtonMouseLeave:
		to: self.
	"menuButton 
	addMorphCentered: (RectangleMorph 
	newBounds: (0 @ 0 extent: 4 @ 2) 
	color: Color black)."
	self updateMenuButtonImage.
	self addMorphFront: menuButton.
	menuButton visible: (self class scrollBarsWithoutMenuButton or: [self orientation == #horizontal]) not.! !

!ScrollBar methodsFor: 'initialize' stamp: 'mt 7/23/2016 12:48'!
initializePagingArea
	
	"Appearance"
	pagingArea := RectangleMorph newBounds: self totalSliderArea.
	self addMorphBack: pagingArea.
			
	"Interactions"
	pagingArea
		on: #mouseDown
		send: #scrollPageInit:
		to: self.
	pagingArea
		on: #mouseUp
		send: #finishedScrolling
		to: self.
	
! !

!ScrollBar methodsFor: 'initialize' stamp: 'mt 7/23/2016 12:22'!
initializeSlider

	self
		initializeMenuButton;
		initializeUpButton;
		initializeDownButton;
		initializePagingArea.
		
	super initializeSlider.
	
	self expandSlider.! !

!ScrollBar methodsFor: 'initialize' stamp: 'mt 7/23/2016 13:07'!
initializeUpButton
"initialize the receiver's upButton"
	upButton := RectangleMorph newBounds: self boundsForUpButton.
	upButton
		on: #mouseDown
		send: #scrollUpInit
		to: self.
	upButton
		on: #mouseUp
		send: #finishedScrolling
		to: self.
	self updateUpButtonImage.
	self addMorph: upButton.	
	upButton visible: self class scrollBarsWithoutArrowButtons not.! !

!ScrollBar methodsFor: 'accessing' stamp: 'mt 4/9/2015 11:45'!
interval
	
	^ interval ifNil: [interval := 0.2]! !

!ScrollBar methodsFor: 'accessing' stamp: 'dew 2/21/1999 03:08'!
interval: d
	"Supply an optional floating fraction so slider can expand to indicate range"
	interval := d min: 1.0.
	self expandSlider.
	self computeSlider.! !

!ScrollBar methodsFor: 'accessing' stamp: 'mt 4/9/2015 11:31'!
menuButton

	^ menuButton ifNil: [menuButton := RectangleMorph new]! !

!ScrollBar methodsFor: 'event handling' stamp: ''!
menuButtonMouseDown: event
	event hand showTemporaryCursor: nil.
	self use: menuSelector orMakeModelSelectorFor: 'MenuButtonPressed:'
		in: [:sel | menuSelector := sel.  model perform: sel with: event]! !

!ScrollBar methodsFor: 'images' stamp: 'mt 7/23/2016 16:40'!
menuImage
	"answer a form to be used in the menu button"
	^ self class
		boxOfSize: (self buttonExtent x min: self buttonExtent y)
		color: self imageColor! !

!ScrollBar methodsFor: 'accessing' stamp: 'mt 4/9/2015 09:55'!
menuSelector
	^ menuSelector! !

!ScrollBar methodsFor: 'accessing' stamp: 'mt 4/9/2015 09:55'!
menuSelector: aSymbol
	menuSelector := aSymbol.! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 4/22/2015 17:23'!
minHeight
	"The minimum extent is that of 2 or 3 buttons in a row or column,
	the 'up' and 'down' button and optionally the 'menu' button."

	| buttonCount refExtent refBorder |
	refExtent := upButton minExtent.
	refBorder := upButton borderWidth.
	buttonCount := {
		upButton visible.
		downButton visible.
		self menuButton visible. } count: [:ea | ea].
	^ self orientation == #horizontal
		ifTrue: [super minHeight max: 5]
		ifFalse: [super minHeight max: ((buttonCount * refExtent y) - (buttonCount-1 * refBorder))]! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 4/16/2015 22:01'!
minThumbThickness

	^ Preferences standardDefaultTextFont height! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 4/22/2015 17:23'!
minWidth
	"The minimum extent is that of 2 or 3 buttons in a row or column,
	the 'up' and 'down' button and optionally the 'menu' button."

	| buttonCount refExtent refBorder |
	refExtent := upButton minExtent.
	refBorder := upButton borderWidth.
	buttonCount := {
		upButton visible.
		downButton visible.
		self menuButton visible. } count: [:ea | ea].
	^ self orientation == #horizontal
		ifTrue: [super minWidth max: ((buttonCount * refExtent x) - (buttonCount-1 * refBorder))]
		ifFalse: [super minWidth max: 5]! !

!ScrollBar methodsFor: 'event handling' stamp: 'mt 7/23/2016 12:11'!
mouseDownInSlider: event

	"make the entire scrollable area visible if a full scrollbar is clicked on"
	self interval = self maximumValue ifTrue: [
		self setValue: 0.
		self model hideOrShowScrollBars].
	
	self showSliderShadow.! !

!ScrollBar methodsFor: 'accessing' stamp: 'dew 2/15/1999 18:25'!
pagingArea
	^pagingArea! !

!ScrollBar methodsFor: 'scroll timing' stamp: 'di 8/17/1998 09:22'!
resetTimer
	timeOfMouseDown := Time millisecondClockValue.
	timeOfLastScroll := timeOfMouseDown - 1000 max: 0.
	nextPageDirection := nil.
	currentScrollDelay := nil! !

!ScrollBar methodsFor: 'scrolling' stamp: 'ar 10/7/2000 14:56'!
scrollBarAction
	^self valueOfProperty: #scrollBarAction! !

!ScrollBar methodsFor: 'scrolling' stamp: 'ar 10/7/2000 14:56'!
scrollBarAction: aSymbol
	self setProperty: #scrollBarAction toValue: aSymbol! !

!ScrollBar methodsFor: 'accessing' stamp: ''!
scrollDelta
	^ scrollDelta! !

!ScrollBar methodsFor: 'accessing' stamp: ''!
scrollDelta: d1 pageDelta: d2
	"Supply optional increments for better scrolling of, eg, text"
	scrollDelta := d1.
	pageDelta := d2.! !

!ScrollBar methodsFor: 'scrolling' stamp: 'mt 7/23/2016 13:13'!
scrollDownInit
	self resetTimer.
	self scrollBarAction: #doScrollDown.
	self startStepping.! !

!ScrollBar methodsFor: 'scrolling' stamp: 'mt 4/8/2015 18:02'!
scrollDown: count
	self setValue: value + (scrollDelta * count).! !

!ScrollBar methodsFor: 'scrolling' stamp: 'RAA 12/29/2000 11:56'!
scrollPageInit: evt
	self resetTimer.
	self setNextDirectionFromEvent: evt.
	self scrollBarAction: #doScrollByPage.
	self startStepping.! !

!ScrollBar methodsFor: 'scrolling' stamp: 'mt 7/23/2016 13:13'!
scrollUpInit
	self resetTimer.
	self scrollBarAction: #doScrollUp.
	self startStepping.! !

!ScrollBar methodsFor: 'scrolling' stamp: 'mt 4/8/2015 18:03'!
scrollUp: count
	self setValue: value - (scrollDelta * count).! !

!ScrollBar methodsFor: 'initialize' stamp: 'mt 7/25/2016 13:51'!
setDefaultParameters

	"Compared to generic sliders, I am not my own paging area. Thus, make me transparent."
	self
		color: Color transparent;
		borderWidth: 0.

	pagingArea
		color: (self userInterfaceTheme color ifNil: [Color veryVeryLightGray darker alpha: 0.35]);
		borderWidth: 0. "no border for the paging area"

	slider
		color: (self userInterfaceTheme thumbColor ifNil: [Color veryVeryLightGray]);
		borderColor: (self userInterfaceTheme thumbBorderColor ifNil: [Color gray: 0.6]);
		borderWidth: (self userInterfaceTheme thumbBorderWidth ifNil: [1]).
	
	self updateSliderCornerStyle.
	
	sliderShadow
		cornerStyle: slider cornerStyle;
		borderWidth: slider borderWidth;
		borderColor: Color transparent.
	
	sliderColor := slider color.
	self updateSliderColor: slider color.! !

!ScrollBar methodsFor: 'scrolling' stamp: 'mt 4/16/2015 23:33'!
setNextDirectionFromEvent: event

	nextPageDirection := self orientation == #horizontal
		ifTrue: [event cursorPoint x >= slider center x]
		ifFalse: [event cursorPoint y >= slider center y].
! !

!ScrollBar methodsFor: 'accessing' stamp: 'mt 7/23/2016 13:44'!
sliderColor: aColor 
	"Change the color of the scrollbar to go with aColor."

	super sliderColor: aColor.
	self updateSliderColor: aColor.! !

!ScrollBar methodsFor: 'geometry' stamp: 'dew 2/21/1999 03:08'!
sliderExtent
	"The sliderExtent is now stored in the slider itself, not hardcoded as it is in the superclass."
	^slider extent! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 4/17/2015 13:59'!
sliderThickness
	^ self orientation == #horizontal
		ifTrue: [self height]
		ifFalse: [self width]! !

!ScrollBar methodsFor: 'stepping and presenter' stamp: 'ar 10/7/2000 15:02'!
step
	| action |
	action := self scrollBarAction.
	action ifNotNil:[self perform: action].! !

!ScrollBar methodsFor: 'testing' stamp: 'di 4/22/2001 18:30'!
stepTime
	^ currentScrollDelay ifNil: [300]! !

!ScrollBar methodsFor: 'geometry' stamp: 'mt 4/16/2015 23:33'!
totalSliderArea
	| upperReferenceBounds lowerReferenceBounds |
	upperReferenceBounds := (upButton visible ifFalse: [self menuButton visible ifTrue: [self menuButton] ifFalse: [nil]] ifTrue: [upButton])
		ifNil: [self topLeft corner: (self orientation == #horizontal ifTrue: [self bottomLeft + (1@0)] ifFalse: [self topRight + (0@1)])]
		ifNotNil: [:button | button bounds].
	lowerReferenceBounds := downButton visible
		ifFalse: [(self orientation == #horizontal ifTrue: [self topRight - (1@0)] ifFalse: [self bottomLeft - (0@1)]) corner: self bottomRight]
		ifTrue: [downButton bounds].
	^ self orientation == #horizontal
		ifTrue: [upperReferenceBounds topRight - (1@0) corner: lowerReferenceBounds bottomLeft + (1@0)]
		ifFalse:[upperReferenceBounds bottomLeft - (0@1) corner: lowerReferenceBounds topRight + (0@1)].
! !

!ScrollBar methodsFor: 'images' stamp: 'RAA 7/28/2000 10:12'!
upArrow8Bit

	"convert to 8-bit and convert white to transparent to avoid gratuitous conversion every time we put one in an ImageMorph"

	^UpArrow8Bit ifNil: [
		UpArrow8Bit := (ColorForm mappingWhiteToTransparentFrom: UpArrow) asFormOfDepth: 8
	]! !

!ScrollBar methodsFor: 'accessing' stamp: 'mt 7/23/2016 13:04'!
upButton
	^ upButton! !

!ScrollBar methodsFor: 'updating' stamp: 'mt 7/23/2016 16:54'!
updateDownButtonImage
	"update the receiver's downButton.  put a new image inside"
	downButton removeAllMorphs.
	downButton addMorphCentered: (ImageMorph new image: self downImage).! !

!ScrollBar methodsFor: 'updating' stamp: 'mt 7/23/2016 16:54'!
updateMenuButtonImage
	"update the receiver's menuButton. put a new image inside"

	self menuButton removeAllMorphs.
	self menuButton addMorphCentered: (ImageMorph new image: self menuImage).! !

!ScrollBar methodsFor: 'updating' stamp: 'mt 7/23/2016 12:07'!
updateSlider

	| imagesNeedUpdate |
	imagesNeedUpdate := upButton width ~= (self orientation == #horizontal ifTrue: [self height] ifFalse: [self width]).
	
	self menuButton
		visible: (self orientation == #horizontal or: [self class scrollBarsWithoutMenuButton]) not;
		bounds: self boundsForMenuButton.
	upButton
		visible: self class scrollBarsWithoutArrowButtons not;
		bounds: self boundsForUpButton.
	downButton
		visible: self class scrollBarsWithoutArrowButtons not;
		bounds: self boundsForDownButton.

	super updateSlider.

	pagingArea bounds: self totalSliderArea.
	self expandSlider.

	imagesNeedUpdate ifTrue: [
		self menuButton visible ifTrue: [self updateMenuButtonImage].
		upButton visible ifTrue: [self updateUpButtonImage].
		downButton visible ifTrue: [self updateDownButtonImage]].! !

!ScrollBar methodsFor: 'updating' stamp: 'mt 7/28/2016 11:50'!
updateSliderColor

	self updateSliderColor: self sliderColor.! !

!ScrollBar methodsFor: 'updating' stamp: 'mt 7/24/2016 14:48'!
updateSliderColorGradient: aColor

	| gradient cc |
	pagingArea color: ((self userInterfaceTheme pagingAreaColorModifier
		ifNil: [ [:c | c darker alpha: 0.35] ]) value: aColor).
	slider borderColor: ((self userInterfaceTheme borderColorModifier
		ifNil: [ [:c | c adjustBrightness: -0.3] ]) value: aColor).
	
	cc := ((self userInterfaceTheme thumbColorModifier
		ifNil: [ [:c | c] ]) value: aColor).
	
	gradient := GradientFillStyle ramp: {
		0.0 -> Color white.
		0.1-> (cc adjustBrightness: 0.05).
		0.6 -> (cc darker)}.
	
	"Fill the thumb"
	gradient origin: slider topLeft.
	gradient direction: (self orientation == #horizontal
		ifTrue:[0@slider height]
		ifFalse:[slider width@0]).
	slider fillStyle: gradient.
	
	"Fill the buttons"
	{self menuButton. self upButton. self downButton} do: [:ea |
		gradient := gradient copy.
		gradient origin: ea topLeft.
		gradient direction: (self orientation == #horizontal
			ifTrue:[0@ea height]
			ifFalse:[ea width@0]).
		ea
			fillStyle: gradient;
			borderWidth: slider borderWidth;
			borderColor: slider borderColor].! !

!ScrollBar methodsFor: 'updating' stamp: 'mt 7/23/2016 17:11'!
updateSliderColorSolid: aColor

	sliderShadow color: self sliderShadowColor.

	slider color: ((self userInterfaceTheme thumbColorModifier
		ifNil: [ [:c | c] ]) value: aColor).

	slider borderColor: ((self userInterfaceTheme borderColorModifier
		ifNil: [ [:c | c adjustBrightness: -0.3] ]) value: aColor).

	pagingArea color: ((self userInterfaceTheme pagingAreaColorModifier
		ifNil: [ [:c | c darker alpha: 0.35] ]) value: aColor).
		
	{self menuButton. self upButton. self downButton} do: [:ea |
		ea
			color: slider color;
			borderWidth: slider borderWidth;
			borderColor: slider borderColor].! !

!ScrollBar methodsFor: 'updating' stamp: 'mt 7/25/2016 13:41'!
updateSliderColor: aColor

	self class gradientScrollBar
		ifFalse: [self updateSliderColorSolid: aColor]
		ifTrue: [self updateSliderColorGradient: aColor].
		
	self flag: #performance. "mt: This is slow..."
	self updateMenuButtonImage.
	self updateUpButtonImage.
	self updateDownButtonImage.! !

!ScrollBar methodsFor: 'updating' stamp: 'mt 7/25/2016 13:50'!
updateSliderCornerStyle
	"Right now, only the slider's corner style is affected. Buttons and paging area remain square. Looks better somehow... :-)"
	
	slider cornerStyle: (self class roundedScrollBarLook ifTrue: [#rounded] ifFalse: [#square]).! !

!ScrollBar methodsFor: 'updating' stamp: 'mt 7/23/2016 16:54'!
updateUpButtonImage
"update the receiver's upButton. put a new image inside"
	upButton removeAllMorphs.
	upButton addMorphCentered: (ImageMorph new image: self upImage).! !

!ScrollBar methodsFor: 'images' stamp: 'mt 7/23/2016 16:40'!
upImage
	"answer a form to be used in the up button"
	^ self class
		arrowOfDirection: (self orientation == #horizontal
				ifTrue: [#left]
				ifFalse: [#top])
		size: (self buttonExtent x min: self buttonExtent y)
		color: self imageColor! !

!ScrollBar methodsFor: 'scroll timing' stamp: 'jmv 3/2/2010 16:29'!
waitForDelay1: delay1 delay2: delay2 
	"Return true if an appropriate delay has passed since the last scroll operation.
	The delay decreases exponentially from delay1 to delay2."

	| now scrollDelay |
	timeOfLastScroll ifNil: [self resetTimer].	"Only needed for old instances"
	now := Time millisecondClockValue.
	(scrollDelay := currentScrollDelay) isNil 
		ifTrue: [scrollDelay := delay1	"initial delay"].
	currentScrollDelay := scrollDelay * 9 // 10 max: delay2.	"decrease the delay"
	timeOfLastScroll := now.
	^true! !

!ScrollBar methodsFor: 'testing' stamp: 'ar 10/7/2000 15:02'!
wantsSteps
	^self scrollBarAction notNil! !

!Slider class methodsFor: 'new-morph participation' stamp: 'di 2/21/98 11:03'!
includeInNewMorphMenu
	"OK to instantiate"
	^ true! !

!Slider class methodsFor: 'instance creation' stamp: 'mt 4/8/2015 16:51'!
on: anObject getValue: getSel setValue: setSel
	"Answer a new instance of the receiver with
	the given selectors as the interface."

	^self new
		on: anObject
		getValue: getSel
		setValue: setSel! !

!Slider class methodsFor: 'instance creation' stamp: 'mt 4/8/2015 17:16'!
on: anObject getValue: getSel setValue: setSel min: min max: max quantum: quantum
	"Answer a new instance of the receiver with
	the given selectors as the interface."

	| instance |
	instance := self new
		quantum: quantum;
		on: anObject
		getValue: getSel
		setValue: setSel.
	min isSymbol
		ifTrue: [instance getMinimumValueSelector: min]
		ifFalse: [instance minimumValue: min].
	max isSymbol
		ifTrue: [instance getMaximumValueSelector: max]
		ifFalse: [instance maximumValue: max].
	^ instance! !

!Slider class methodsFor: 'preferences' stamp: 'mt 7/23/2016 16:18'!
themeProperties

	^ super themeProperties, {
		{ #borderColor. 'Colors'. 'Color of the slider''s border.' }.
		{ #borderWidth. 'Borders'. 'Width of the slider''s border.' }.
		{ #color. 'Colors'. 'Background color of the slider.' }.

		{ #thumbBorderColor. 'Colors'. 'Color of the slider thumb''s border.' }.
		{ #thumbBorderWidth. 'Borders'. 'Width of the slider thumb''s border.' }.
		{ #thumbColor. 'Colors'. 'Background color of the slider thumb.' }.
		
		{ #thumbShadowModifier. 'Colors'. 'How to modify the thumb color to get the shadow color.' }.
	}! !

!Slider methodsFor: 'accessing - ui' stamp: 'mt 7/23/2016 16:28'!
adoptPaneColor: paneColor

	super adoptPaneColor: paneColor.
	paneColor ifNotNil: [:c | self sliderColor: c].! !

!Slider methodsFor: 'updating' stamp: 'mt 7/23/2016 16:21'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	
	self setDefaultParameters.! !

!Slider methodsFor: 'geometry' stamp: 'mt 4/16/2015 23:34'!
computeSlider
	| r v |
	r := self roomToMove.
	v := self maximumValue = self minimumValue
		ifTrue: [0]
		ifFalse: [(value - self minimumValue) / (self maximumValue - self minimumValue)].
	self descending
		ifFalse:
			[slider position: (self orientation == #horizontal
				ifTrue: [r topLeft + ((r width * v) asInteger @ 0)]
				ifFalse: [r topLeft + (0 @ (r height * v)  asInteger)])]
		ifTrue:
			[slider position: (self orientation == #horizontal
				ifTrue:	[r bottomRight - ((r width * v) asInteger @ 0)]
				ifFalse:	[r bottomRight - ((0 @ (r height * v) asInteger))])].
	slider extent: self sliderExtent! !

!Slider methodsFor: 'initialization' stamp: 'dgd 3/7/2003 15:07'!
defaultBounds
"answer the default bounds for the receiver"
	^ 0 @ 0 corner: 16 @ 100! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 17:08'!
descending

	^ descending! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 17:07'!
descending: aBoolean

	descending := aBoolean.
	self computeSlider.! !

!Slider methodsFor: 'geometry' stamp: 'mt 4/17/2015 14:07'!
extent: newExtent

	(bounds extent closeTo: newExtent) ifTrue: [^ self].

	super extent: newExtent.

	self updateSlider.! !

!Slider methodsFor: 'model access' stamp: 'mt 4/8/2015 17:17'!
getMaximumValue
	
	self getMaximumValueSelector ifNotNil: [:selector |
		self maximumValue: (model perform: selector)].
	^ self maximumValue! !

!Slider methodsFor: 'accessing - model' stamp: 'mt 4/8/2015 17:12'!
getMaximumValueSelector

	^ getMaximumValueSelector! !

!Slider methodsFor: 'accessing - model' stamp: 'mt 4/8/2015 17:12'!
getMaximumValueSelector: aSymbol

	getMaximumValueSelector := aSymbol.! !

!Slider methodsFor: 'model access' stamp: 'mt 4/8/2015 17:18'!
getMinimumValue
	
	self getMinimumValueSelector ifNotNil: [:selector |
		self minimumValue: (model perform: selector)].
	^ self minimumValue! !

!Slider methodsFor: 'accessing - model' stamp: 'mt 4/8/2015 17:13'!
getMinimumValueSelector

	^ getMinimumValueSelector! !

!Slider methodsFor: 'accessing - model' stamp: 'mt 4/8/2015 17:13'!
getMinimumValueSelector: aSymbol

	getMinimumValueSelector := aSymbol.! !

!Slider methodsFor: 'model access' stamp: 'mt 4/8/2015 17:04'!
getValue
	"Updates internal value with model data if possible. Returns the updated value or the current one."
	
	self getValueSelector ifNotNil: [:selector |
		self value: (model perform: selector)].
	^ self value! !

!Slider methodsFor: 'accessing - model' stamp: 'mt 4/8/2015 15:29'!
getValueSelector

	^ getValueSelector! !

!Slider methodsFor: 'accessing - model' stamp: 'mt 4/8/2015 15:29'!
getValueSelector: aSymbol

	getValueSelector := aSymbol.! !

!Slider methodsFor: 'other events' stamp: 'mt 7/23/2016 12:10'!
hideSliderShadow

	sliderShadow hide.! !

!Slider methodsFor: 'as yet unclassified' stamp: 'mt 2/12/2016 09:39'!
hResizing
	"Due to possible automatic orientation change on extent changes, we have to calculate the resizing property based on the current orientation. Otherwise, layout will break.There is, however, support for using the resizing property in layout properties if set."

	self layoutProperties ifNotNil: [:props |
		^ props hResizing].
	
	^ self orientation == #horizontal
		ifTrue: [#spaceFill]
		ifFalse: [#rigid]! !

!Slider methodsFor: 'initialization' stamp: 'mt 7/23/2016 12:19'!
initialize

	super initialize.

	value := 0.0.
	descending := false.
	
	self initializeSlider.
	self setDefaultParameters.! !

!Slider methodsFor: 'initialization' stamp: 'mt 7/23/2016 12:24'!
initializeSlider

	slider := RectangleMorph
		newBounds: self totalSliderArea
		color: self thumbColor.
	
	sliderShadow := RectangleMorph
		newBounds: self totalSliderArea
		color: self pagingArea color.
	
	slider on: #mouseMove send: #scrollAbsolute: to: self.
	slider on: #mouseDown send: #mouseDownInSlider: to: self.
	slider on: #mouseUp send: #mouseUpInSlider: to: self.

	self addMorphFront: sliderShadow.
	self addMorphFront: slider.

	self computeSlider.
	self hideSliderShadow.! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 10:29'!
maximumValue
	
	^ maximumValue ifNil: [1.0]! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 18:40'!
maximumValue: aNumber
	
	maximumValue := aNumber.
	self setValue: self value.! !

!Slider methodsFor: 'geometry' stamp: 'mt 4/22/2015 17:22'!
minHeight

	^ self orientation == #horizontal
		ifTrue: [super minHeight max: self borderWidth + 1]
		ifFalse: [super minHeight max: self sliderThickness * 2]! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 10:28'!
minimumValue
	^ minimumValue ifNil: [0.0]! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 18:40'!
minimumValue: aNumber

	minimumValue := aNumber.
	self setValue: self value.! !

!Slider methodsFor: 'geometry' stamp: 'mt 4/22/2015 17:21'!
minWidth

	^ self orientation == #horizontal
		ifTrue: [super minWidth max: self sliderThickness * 2]
		ifFalse: [super minWidth max: self borderWidth + 1]! !

!Slider methodsFor: 'other events' stamp: 'mt 7/23/2016 12:10'!
mouseDownInSlider: event 

	slider borderStyle style == #raised
		ifTrue: [slider borderColor: #inset].
	
	self showSliderShadow.! !

!Slider methodsFor: 'other events' stamp: 'mt 7/23/2016 12:10'!
mouseUpInSlider: event 

	slider borderStyle style == #inset
		ifTrue: [slider borderColor: #raised].
	
	self hideSliderShadow.! !

!Slider methodsFor: 'initialization' stamp: 'mt 4/8/2015 16:51'!
on: anObject getValue: getSel setValue: setSel

	self
		model: anObject;
		getValueSelector: getSel;
		setValueSelector: setSel;
		getValue.! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/16/2015 23:28'!
orientation

	^ orientation ifNil: [bounds isWide ifTrue: [#horizontal] ifFalse: [#vertical]]! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/16/2015 23:28'!
orientation: aSymbol

	orientation == aSymbol ifTrue: [^ self].
	orientation := aSymbol.
	self updateSlider.! !

!Slider methodsFor: 'accessing - ui' stamp: 'dew 2/15/1999 18:24'!
pagingArea
	^self! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 16:17'!
quantum

	^ quantum! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 18:40'!
quantum: aNumber

	quantum := aNumber.
	self setValue: self value.! !

!Slider methodsFor: 'geometry' stamp: ''!
roomToMove
	^ self totalSliderArea insetBy: (0@0 extent: self sliderExtent)! !

!Slider methodsFor: 'scrolling' stamp: 'mt 4/16/2015 23:35'!
scrollAbsolute: event
	| r p |
	r := self roomToMove.
	self orientation == #horizontal
		ifTrue: [r width = 0 ifTrue: [^ self]]
		ifFalse: [r height = 0 ifTrue: [^ self]].
	p := event targetPoint adhereTo: r.
	self descending
		ifFalse:
			[self setValueFraction: (self orientation == #horizontal 
				ifTrue: [(p x - r left) asFloat / r width]
				ifFalse: [(p y - r top) asFloat / r height])]
		ifTrue:
			[self setValueFraction: (self orientation == #horizontal
				ifTrue: [(r right - p x) asFloat / r width]
				ifFalse:	[(r bottom - p y) asFloat / r height])]! !

!Slider methodsFor: 'scrolling' stamp: 'mt 4/8/2015 19:47'!
scrollBy: delta

	self setValue: self value + delta.! !

!Slider methodsFor: 'scrolling' stamp: 'mt 5/2/2016 17:57'!
scrollToMaximum

	self setValue: self maximumValue.! !

!Slider methodsFor: 'scrolling' stamp: 'mt 5/2/2016 17:57'!
scrollToMinimum

	self setValue: self minimumValue.! !

!Slider methodsFor: 'initialization' stamp: 'mt 7/23/2016 16:48'!
setDefaultParameters
	"change the receiver's appareance parameters"

	self
		color: (self userInterfaceTheme color ifNil: [Color lightGray]);
		borderColor: (self userInterfaceTheme borderColor ifNil: [Color gray]);
		borderWidth: (self userInterfaceTheme borderWidth ifNil: [1]).
	
	slider
		color: (self userInterfaceTheme thumbColor ifNil: [Color veryVeryLightGray]);
		borderColor: (self userInterfaceTheme thumbBorderColor ifNil: [Color gray: 0.6]);
		borderWidth: (self userInterfaceTheme thumbBorderWidth ifNil: [0]).

	sliderShadow
		borderWidth: slider borderWidth;
		borderColor: Color transparent.

	sliderColor := slider color.
	self updateSliderColor: slider color.! !

!Slider methodsFor: 'support' stamp: 'mt 4/8/2015 10:46'!
setValueFraction: newValueFraction

	self setValue: (newValueFraction * (self maximumValue - self minimumValue)) + self minimumValue.! !

!Slider methodsFor: 'accessing - model' stamp: 'mt 4/8/2015 15:05'!
setValueSelector

	^ setValueSelector! !

!Slider methodsFor: 'accessing - model' stamp: 'mt 4/8/2015 15:05'!
setValueSelector: aSymbol

	setValueSelector := aSymbol.! !

!Slider methodsFor: 'model access' stamp: 'mt 4/8/2015 19:00'!
setValue: newValue
	"Either changes the value directly or tries to go the loop through the model. See #update:."

	self setValueSelector ifNotNil: [:selector |
		| trimmedValue |
		trimmedValue := self trimmedValue: newValue.
		"Only notify about changed values."
		trimmedValue ~= self value ifTrue: [
			model perform: selector with: trimmedValue]].

	(self setValueSelector isNil or: [self getValueSelector isNil])
		ifTrue: [self value: newValue].! !

!Slider methodsFor: 'other events' stamp: 'mt 7/23/2016 12:10'!
showSliderShadow

	sliderShadow color: self sliderShadowColor.
	sliderShadow cornerStyle: slider cornerStyle.
	sliderShadow bounds: slider bounds.
	sliderShadow show.! !

!Slider methodsFor: 'accessing - ui' stamp: 'mt 7/23/2016 16:24'!
sliderColor
	"color scheme for the whole slider widget"
	
	^ sliderColor ifNil: [self userInterfaceTheme thumbColor ifNil: [Color veryVeryLightGray]]! !

!Slider methodsFor: 'accessing - ui' stamp: 'mt 7/23/2016 16:14'!
sliderColor: newColor

	sliderColor := newColor.
	self updateSliderColor: sliderColor.! !

!Slider methodsFor: 'geometry' stamp: 'mt 4/16/2015 23:35'!
sliderExtent
	^ self orientation == #horizontal
		ifTrue: [self sliderThickness @ self innerBounds height]
		ifFalse: [self innerBounds width @ self sliderThickness]! !

!Slider methodsFor: 'accessing - ui' stamp: 'mt 7/23/2016 16:23'!
sliderShadowColor

	^ ((self userInterfaceTheme thumbShadowModifier
			ifNil: [ [:c | c alpha: 0.7 ] ]) value: self sliderColor)! !

!Slider methodsFor: 'geometry' stamp: 'yo 1/5/2007 16:04'!
sliderThickness
	"^ 7"

	| w |
	w _ bounds isWide
		ifTrue: [super height]
		ifFalse: [super width].

	^ (w // 32) max: 16.
! !

!Slider methodsFor: 'accessing - ui' stamp: 'mt 4/8/2015 17:09'!
thumb

	^ slider! !

!Slider methodsFor: 'accessing - ui' stamp: 'dew 1/21/2002 01:31'!
thumbColor
	"Color of the draggable 'thumb'"
	^ self sliderColor! !

!Slider methodsFor: 'geometry' stamp: ''!
totalSliderArea
	^ self innerBounds! !

!Slider methodsFor: 'support' stamp: 'mt 4/8/2015 17:06'!
trimmedValue: aValue

	| trimmedValue |
	trimmedValue := aValue min: self maximumValue max: self minimumValue.
	self quantum ifNotNil: [:q | trimmedValue := trimmedValue roundTo: q].
	^ trimmedValue
! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 16:21'!
truncate

	^ self quantum == 1! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 16:21'!
truncate: aBoolean

	self quantum: (aBoolean ifTrue: [1] ifFalse: [nil]).! !

!Slider methodsFor: 'updating' stamp: 'mt 4/9/2015 08:59'!
updateSlider
	"Updates layout properties of the slider."
	
	slider bounds: self totalSliderArea.
	sliderShadow bounds: slider bounds.
	
	self computeSlider.
! !

!Slider methodsFor: 'updating' stamp: 'mt 7/23/2016 16:27'!
updateSliderColor: aColor

	slider color: aColor.
	sliderShadow color: self sliderShadowColor.! !

!Slider methodsFor: 'updating' stamp: 'mt 4/8/2015 17:17'!
update: aSymbol
	"Update the value."
	
	super update: aSymbol.
	
	aSymbol = self getValueSelector ifTrue: [self getValue. ^ self].
	aSymbol = self getMinimumValueSelector ifTrue: [self getMinimumValue. ^ self].
	aSymbol = self getMaximumValueSelector ifTrue: [self getMaximumValue. ^ self].! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 17:19'!
value

	^ value! !

!Slider methodsFor: 'accessing' stamp: 'mt 4/8/2015 18:02'!
value: newValue

	| t |
	t := self trimmedValue: newValue.
	t = value ifTrue: [^ self].
	
	value := t.
	self computeSlider.! !

!Slider methodsFor: 'as yet unclassified' stamp: 'mt 2/12/2016 09:39'!
vResizing
	"Due to possible automatic orientation change on extent changes, we have to calculate the resizing property based on the current orientation. Otherwise, layout will break. There is, however, support for using the resizing property in layout properties if set."

	self layoutProperties ifNotNil: [:props |
		^ props vResizing].

	^ self orientation == #vertical
		ifTrue: [#spaceFill]
		ifFalse: [#rigid]! !

!Slider methodsFor: 'testing' stamp: '6/7/97 10:42 di'!
wantsSlot
	"For now do it the old way, until we sort this out"
	^ true! !

MorphicModel subclass: #SystemWindow
	instanceVariableNames: 'labelString stripes label closeBox collapseBox paneMorphs paneRects collapsedFrame fullFrame isCollapsed isActive isLookingFocused menuBox mustNotClose labelWidgetAllowance updatablePanes allowReframeHandles labelArea expandBox'
	classVariableNames: 'ClickOnLabelToEdit CloseBoxFrame CloseBoxImageFlat CloseBoxImageGradient CollapseBoxImageFlat CollapseBoxImageGradient DoubleClickOnLabelToExpand ExpandBoxFrame ExpandBoxImageFlat ExpandBoxImageGradient FocusFollowsMouse GradientWindow HideExpandButton MenuBoxFrame MenuBoxImageFlat MenuBoxImageGradient ResizeAlongEdges ReuseWindows RoundedWindowCorners TopWindow WindowsRaiseOnClick WindowTitleActiveOnFirstClick'
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!SystemWindow commentStamp: '<historical>' prior: 0!
SystemWindow is the Morphic equivalent of StandardSystemView -- a labelled container for rectangular views, with iconic facilities for close, collapse/expand, and resizing.

The attribute onlyActiveOnTop, if set to true (and any call to activate will set this), determines that only the top member of a collection of such windows on the screen shall be active.  To be not active means that a mouse click in any region will only result in bringing the window to the top and then making it active.!

SystemWindow subclass: #CollapsedMorph
	instanceVariableNames: 'uncollapsedMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!CollapsedMorph methodsFor: 'collapse/expand' stamp: 'sw 5/9/2000 00:18'!
beReplacementFor: aMorph

	| itsWorld priorPosition |
	(itsWorld := aMorph world) ifNil: [^self].
	uncollapsedMorph := aMorph.
			
	self setLabel: aMorph externalName.
	aMorph delete.
	itsWorld addMorphFront: self.
	self collapseOrExpand.
	(priorPosition := aMorph valueOfProperty: #collapsedPosition ifAbsent: [nil])
	ifNotNil:
		[self position: priorPosition].
! !

!CollapsedMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 16:41'!
buildWindowMenu
	"Answer the menu to be put up in response to the user's clicking on the window-menu control in the window title.  Specialized for CollapsedMorphs."

	| aMenu |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu add: 'change name...' translated action: #relabel.
	aMenu addLine.
	aMenu add: 'send to back' translated action: #sendToBack.
	aMenu add: 'make next-to-topmost' translated action: #makeSecondTopmost.
	aMenu addLine.
	self mustNotClose
		ifFalse:
			[aMenu add: 'make unclosable' translated action: #makeUnclosable]
		ifTrue:
			[aMenu add: 'make closable' translated action: #makeClosable].
	aMenu
		add: (self isSticky ifTrue: ['make draggable'] ifFalse: ['make undraggable']) translated 
		action: #toggleStickiness.
	^aMenu! !

!CollapsedMorph methodsFor: 'resize/collapse' stamp: 'sw 9/1/2000 11:07'!
collapseOrExpand
	"Toggle the expand/collapsd state of the receiver.  If expanding, copy the window title back to the name of the expanded morph"

	| aWorld |
	isCollapsed
		ifTrue: 
			[uncollapsedMorph setProperty: #collapsedPosition toValue: self position.
			labelString ifNotNil: [uncollapsedMorph setNameTo: labelString].
			mustNotClose := false.	"We're not closing but expanding"
			self delete.
			(aWorld := self currentWorld) addMorphFront: uncollapsedMorph.
			aWorld startSteppingSubmorphsOf: uncollapsedMorph]
		ifFalse:
			[super collapseOrExpand]! !

!CollapsedMorph methodsFor: 'collapse/expand' stamp: 'sw 4/9/2001 14:23'!
uncollapseToHand
	"Hand the uncollapsedMorph to the user, placing it in her hand, after remembering appropriate state for possible future use"

	| nakedMorph |
	nakedMorph := uncollapsedMorph.
	uncollapsedMorph := nil.
	nakedMorph setProperty: #collapsedPosition toValue: self position.
	mustNotClose := false.  "so the delete will succeed"
	self delete.
	ActiveHand attachMorph: nakedMorph! !

!CollapsedMorph methodsFor: 'resize/collapse' stamp: 'sw 6/5/2001 22:55'!
wantsExpandBox
	"Answer whether I'd like an expand box"

	^ false! !

SystemWindow subclass: #SystemWindowWithButton
	instanceVariableNames: 'buttonInTitle'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!SystemWindowWithButton commentStamp: '<historical>' prior: 0!
A SystemWindow with a single extra button in its title bar.!

!SystemWindowWithButton class methodsFor: 'preferences' stamp: 'tfel 3/27/2010 12:36'!
hideExpandButton

	^ true! !

!SystemWindowWithButton methodsFor: 'geometry' stamp: 'tfel 3/27/2010 13:22'!
adjustExtraButton
	| leftMargin |
	buttonInTitle ifNil: [^ self].
	leftMargin := self boxExtent x + 3.
	self class moveMenuButtonRight ifTrue: [leftMargin := leftMargin * 2].
	buttonInTitle align: buttonInTitle topRight with: self innerBounds topRight - (leftMargin@0)! !

!SystemWindowWithButton methodsFor: 'label' stamp: 'sw 2/15/1999 22:41'!
buttonInTitle: aButton
	buttonInTitle := aButton.
	self addMorphFront: aButton! !

!SystemWindowWithButton methodsFor: 'geometry' stamp: 'sw 9/29/1999 07:26'!
extent: newExtent
	super extent: (newExtent max: 120 @ 50).
	self adjustExtraButton! !

!SystemWindowWithButton methodsFor: 'label' stamp: 'sw 9/29/1999 07:27'!
setLabelWidgetAllowance
	^ labelWidgetAllowance := 115! !

!SystemWindowWithButton methodsFor: 'resize/collapse' stamp: 'sw 5/30/2001 11:11'!
wantsExpandBox
	"Answer whether I'd like an expand box"

	^ false! !

!SystemWindow class methodsFor: 'initializing' stamp: 'mt 4/24/2015 11:30'!
borderWidth

	^ AbstractResizerMorph gripThickness! !

!SystemWindow class methodsFor: 'preferences' stamp: 'tfel 3/25/2010 23:05'!
boxExtent
	"answer the extent to use in all the buttons"
	
	^ (Preferences alternativeWindowBoxesLook
		ifTrue: [18 @ 18]
		ifFalse: [16 @ 16])! !

!SystemWindow class methodsFor: 'top window' stamp: 'cmm 6/2/2016 15:09'!
bringWindowUnderHandToFront
"This only works when All Windows Active is enabled."
	(self windowsIn: World) do: [ : each | each isLookingFocused ifTrue: [ ^ each beKeyWindow ]]! !

!SystemWindow class methodsFor: 'initializing' stamp: 'RAA 12/21/2000 12:01'!
classVersion
	"Changed to 1 for SystemWindow Dec 2000 - see if this helps loading old ones"
	^ 1! !

!SystemWindow class methodsFor: 'top window' stamp: 'RAA 7/7/2000 09:34'!
clearTopWindow

	TopWindow := nil.	"if leaving morphic to export from mvc, this ref could cause problems"! !

!SystemWindow class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 16:15'!
clickOnLabelToEdit

	<preference: 'Click On Label To Edit'
		category: 'windows'
		description: 'If true, a click on the label of a system window lets you edit it'
		type: #Boolean>
	^ ClickOnLabelToEdit ifNil: [false].
! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 7/28/2016 13:05'!
clickOnLabelToEdit: aBoolean

	ClickOnLabelToEdit := aBoolean.
	self refreshAllWindows.! !

!SystemWindow class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 14:11'!
closeBoxFrame

	^ CloseBoxFrame ifNil: [
		CloseBoxFrame := (LayoutFrame new
								leftFraction: 0;
								leftOffset: 2;
								topFraction: 0;
								topOffset: 0;
								yourself)]! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/8/2015 23:24'!
closeBoxImage

	^ self gradientWindow
		ifTrue: [self closeBoxImageGradient]
		ifFalse: [self closeBoxImageFlat].
! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/8/2015 23:50'!
closeBoxImageFlat

	^ CloseBoxImageFlat ifNil: [CloseBoxImageFlat :=  (Form
	extent: 12@12
	depth: 32
	fromArray: #( 0 0 552294233 2683000665 4025177945 4293613401 4293613401 4025177945 2683000665 552294233 0 0 0 1089165145 4025177945 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4025177945 820729689 0 552294233 4025177945 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4025177945 283858777 2683000665 4293613401 4293613401 4294492353 4294492353 4293613401 4293613401 4294492353 4294492353 4293613401 4293613401 2683000665 4025177945 4293613401 4293613401 4294492353 4294967295 4294492353 4294492353 4294967295 4294492353 4293613401 4293613401 4025177945 4293613401 4293613401 4293613401 4293613401 4294492353 4294967295 4294967295 4294492353 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4294492353 4294967295 4294967295 4294492353 4293613401 4293613401 4293613401 4293613401 4025177945 4293613401 4293613401 4294492353 4294967295 4294492353 4294492353 4294967295 4294492353 4293613401 4293613401 3756742489 2683000665 4293613401 4293613401 4294492353 4294492353 4293613401 4293613401 4294492353 4294492353 4293613401 4293613401 2162906969 552294233 4025177945 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4025177945 283858777 0 820729689 4025177945 4293613401 4293613401 4293613401 4293613401 4293613401 4293613401 4025177945 820729689 0 0 0 283858777 2683000665 4025177945 4293613401 4293613401 3756742489 2162906969 283858777 0 0)
	offset: 0@0)]! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/8/2015 23:23'!
closeBoxImageGradient

	^ CloseBoxImageGradient ifNil: [CloseBoxImageGradient := (Form
	extent: 14@14
	depth: 32
	fromArray: #(0 0 6032910 288687365 1550655006 2844865845 3450028610 3450028610 2828088372 1550655006 288687365 5967374 0 0 1377027 15091784 844830230 3063889984 4158092147 4294153621 4294681250 4294681250 4294153364 4158091376 3047046718 844764694 14763847 1376770 7742243 861541910 3585164870 4294282123 4294943908 4294943908 4294943651 4294943651 4294943908 4294943908 4294216330 3585099848 844764694 7676450 288098055 3047109684 4294211447 4294939539 4293757578 4293363848 4294939025 4294939282 4293954186 4293560456 4294939539 4294212732 3047110455 288031748 1584209695 4158081610 4294932343 4293753466 4291733929 4291802549 4293164156 4293884281 4291666594 4291670192 4293294457 4294933885 4158018130 1567432222 2861706541 4294136145 4294924376 4292761945 4292198592 4294704894 4292064950 4291403679 4294375158 4293057499 4291846497 4294925147 4294136659 2844863789 3483644980 4294660947 4294923605 4294857555 4292566879 4292658886 4294770173 4294704380 4293452000 4291981681 4294595154 4294923605 4294660947 3466867764 3500422196 4294726483 4294923605 4294923605 4293742415 4291402908 4294836223 4294967295 4292198592 4292760660 4294923348 4294923605 4294660947 3483644980 2878484014 4294136145 4294923605 4293939794 4291728018 4294375158 4293452257 4292793042 4294704637 4292128944 4293154390 4294923605 4294136145 2878484014 1617830177 4174923591 4294923862 4292957011 4292000442 4293386721 4291915888 4292238685 4292659400 4292595406 4292303450 4294923605 4158146375 1617829920 305072650 3097572147 4294267217 4294923348 4292631901 4291781989 4294595154 4294857555 4292435036 4292238942 4294857812 4294267217 3097506611 305007114 8070437 912005658 3635624249 4294333010 4294923605 4294857555 4294923605 4294923605 4294923348 4294923348 4294333010 3635624249 895228442 8004901 1639428 16735838 912071451 3131126579 4191766343 4294267217 4294726740 4294726740 4294201681 4191766343 3131126579 895293978 16735067 1573892 0 0 6888733 322506254 1651515681 2945723950 3550885173 3550885173 2945723950 1651515681 322440718 6822940 0 0)
	offset: 0@0)]! !

!SystemWindow class methodsFor: 'top window' stamp: 'sw 12/6/2000 20:13'!
closeTopWindow
	"Try to close the top window.  It may of course decline"

	TopWindow ifNotNil:
		[TopWindow delete]! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/8/2015 23:56'!
collapseBoxImage

	^ self gradientWindow
		ifTrue: [self collapseBoxImageGradient]
		ifFalse: [self collapseBoxImageFlat].
! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/8/2015 23:56'!
collapseBoxImageFlat

	^ CollapseBoxImageFlat ifNil: [CollapseBoxImageFlat :=  (Form
	extent: 12@12
	depth: 32
	fromArray: #( 0 0 552699410 2683405842 4025583122 4294018578 4294018578 4025583122 2683405842 552699410 0 0 0 1089570322 4025583122 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4025583122 821134866 0 552699410 4025583122 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4025583122 284263954 2683405842 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 2683405842 4025583122 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4025583122 4294018578 4294221118 4294762948 4294762948 4294762948 4294762948 4294762948 4294762948 4294762948 4294762948 4294221118 4294018578 4294018578 4294221118 4294762948 4294762948 4294762948 4294762948 4294762948 4294762948 4294762948 4294762948 4294221118 4294018578 4025583122 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 3757147666 2683405842 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 2163312146 552699410 4025583122 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4025583122 284263954 0 821134866 4025583122 4294018578 4294018578 4294018578 4294018578 4294018578 4294018578 4025583122 821134866 0 0 0 284263954 2683405842 4025583122 4294018578 4294018578 3757147666 2163312146 284263954 0 0)
	offset: 0@0)]! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/8/2015 23:51'!
collapseBoxImageGradient

	^ CollapseBoxImageGradient ifNil: [CollapseBoxImageGradient := (Form
	extent: 14@14
	depth: 32
	fromArray: #(0 768 5853184 288569344 1550537988 2844748052 3449976096 3449976096 2827970835 1550537731 288569600 5787392 768 0 1256960 15043082 844713216 3063837470 4158037328 4294162549 4294689156 4294689156 4294162292 4158036813 3046994460 844713216 14714890 1256704 7690756 861424896 3585112609 4294226279 4294952069 4294952069 4294951812 4294951812 4294952069 4294952068 4294226278 3585047331 844647680 7625220 287914752 3047058190 4294222924 4294949230 4294948716 4294948716 4294948716 4294948716 4294948716 4294948717 4294949231 4294223955 3047058707 287914240 1584092675 4158030361 4294944327 4294945619 4294945877 4294946134 4294946134 4294945877 4294945619 4294945102 4294944587 4294945360 4158031652 1567315202 2861589770 4294150937 4294938911 4294217252 4293955628 4294021937 4294021938 4294021421 4294020390 4293954079 4294084893 4294873892 4294151196 2844812298 3483593741 4294741530 4294478360 4291271017 4292461743 4292461485 4292461485 4292461485 4292461485 4292527535 4291340680 4293625630 4294807066 3466816525 3500370957 4294741530 4294347032 4291669904 4294112759 4294112244 4294112244 4294112244 4294112244 4294243831 4292200893 4293428769 4294807066 3483593741 2878432779 4294150937 4294806810 4292511536 4292382532 4292382531 4292382531 4292382531 4292382531 4292382532 4292249912 4294347548 4294216473 2878432523 1617712902 4174872597 4294938651 4294938393 4294938136 4294938136 4294938136 4294938136 4294938136 4294938136 4294938393 4294938651 4158095381 1617712902 304955136 3097520909 4294282265 4294938395 4294938395 4294938395 4294938395 4294938395 4294938395 4294938395 4294938395 4294282009 3097455373 304889600 7953671 911888388 3635573264 4294347801 4294938651 4294938395 4294938395 4294938395 4294938395 4294938651 4294347801 3635573264 895111172 7888135 1520128 16753431 911954180 3131075341 4191781141 4294216473 4294741530 4294741530 4294216473 4191781141 3131075341 895176964 16752406 1454592 0 3072 6708483 322388481 1651398919 2945607436 3550833934 3550833934 2945607436 1651398919 322322945 6642947 2816 0)
	offset: 0@0)]! !

!SystemWindow class methodsFor: 'top window' stamp: 'cmm 6/7/2016 18:25'!
deleteTopWindow
	TopWindow ifNotNil: [ TopWindow delete ]! !

!SystemWindow class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 16:21'!
doubleClickOnLabelToExpand

	<preference: 'Double-Click On Label To Expand'
		category: 'windows'
		description: 'Activates expansion through double-clicking on the window label area. This mimics the behavior in many current operating systems'
		type: #Boolean>
	^ DoubleClickOnLabelToExpand ifNil: [true].
! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 7/28/2016 13:05'!
doubleClickOnLabelToExpand: aBoolean

	DoubleClickOnLabelToExpand := aBoolean.
	self refreshAllWindows.
! !

!SystemWindow class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 14:10'!
expandBoxFrame

	^ ExpandBoxFrame ifNil: [
		ExpandBoxFrame := (LayoutFrame new
								leftFraction: 1;
								leftOffset: (self boxExtent x * 2 + 3) negated;
								topFraction: 0;
								topOffset: 0;
								yourself)]! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/9/2015 00:00'!
expandBoxImage

	^ self gradientWindow
		ifTrue: [self expandBoxImageGradient]
		ifFalse: [self expandBoxImageFlat].
! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/9/2015 00:00'!
expandBoxImageFlat

	^ ExpandBoxImageFlat ifNil: [ExpandBoxImageFlat :=  (Form
	extent: 12@12
	depth: 32
	fromArray: #( 0 0 542547506 2673253938 4015431218 4283866674 4283866674 4015431218 2673253938 542547506 0 0 0 1079418418 4015431218 4283866674 4283866674 4285968472 4285968472 4283866674 4283866674 4015431218 810982962 0 542547506 4015431218 4283866674 4283866674 4283866674 4292208588 4292208588 4283866674 4283866674 4283866674 4015431218 274112050 2673253938 4283866674 4283866674 4283866674 4283866674 4292208588 4292208588 4283866674 4283866674 4283866674 4283866674 2673253938 4015431218 4283866674 4283866674 4283866674 4283866674 4292208588 4292208588 4283866674 4283866674 4283866674 4283866674 4015431218 4283866674 4285968472 4292208588 4292208588 4292208588 4294244850 4294244850 4292208588 4292208588 4292208588 4285968472 4283866674 4283866674 4285968472 4292208588 4292208588 4292208588 4294244850 4294244850 4292208588 4292208588 4292208588 4285968472 4283866674 4015431218 4283866674 4283866674 4283866674 4283866674 4292208588 4292208588 4283866674 4283866674 4283866674 4283866674 3746995762 2673253938 4283866674 4283866674 4283866674 4283866674 4292208588 4292208588 4283866674 4283866674 4283866674 4283866674 2153160242 542547506 4015431218 4283866674 4283866674 4283866674 4292208588 4292208588 4283866674 4283866674 4283866674 4015431218 274112050 0 810982962 4015431218 4283866674 4283866674 4285968472 4285968472 4283866674 4283866674 4015431218 810982962 0 0 0 274112050 2673253938 4015431218 4283866674 4283866674 3746995762 2153160242 274112050 0 0)
	offset: 0@0)]! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/8/2015 23:59'!
expandBoxImageGradient

	^ ExpandBoxImageGradient ifNil: [ExpandBoxImageGradient := (Form
	extent: 14@14
	depth: 32
	fromArray: #(0 0 1058562 285544960 1545616142 2838976800 3443812652 3443812652 2822199328 1545550349 285610496 1058562 0 0 197889 5015843 840445449 3057804586 4151680347 4288134782 4288989580 4288989580 4288134525 4151549016 3040961576 840379656 4949796 132352 2441745 857222409 3578360366 4287543665 4289187213 4289121422 4288857993 4288792456 4289121421 4289121677 4287478128 3578491696 840445193 2375953 285740802 3040304923 4286295641 4288070777 4288004983 4287543410 4288524177 4289312671 4287344755 4288004983 4288136313 4286624095 3040502048 285543680 1579236111 4149118761 4286297173 4286822495 4286888545 4286426462 4291679432 4293717227 4286752873 4286560090 4286428504 4286691421 4149578546 1562393102 2855294231 4283865130 4284326448 4284325172 4284718907 4284716865 4291416516 4293651691 4285043786 4284061998 4283996461 4284523316 4283996461 2838517015 3476513307 4284063275 4283865386 4287341947 4290759608 4290693559 4293717483 4294572537 4291153599 4290759607 4288129929 4283864876 4283997739 3459736090 3493290523 4284063275 4283799593 4289312927 4294243316 4294046193 4294769916 4294901502 4294177779 4294243316 4290364338 4283864621 4284063275 3476513306 2872071703 4283865130 4284063275 4284453693 4285176395 4285109069 4291811019 4293849070 4285962336 4285110602 4284585024 4283997227 4283799593 2872071447 1612856592 4165698852 4284195372 4284063786 4284064041 4283864620 4291285440 4293651690 4284782402 4283998248 4284063786 4284195372 4148921636 1612856336 302649860 3090571034 4283865642 4284129580 4284129580 4283930157 4289378465 4291218626 4284519486 4284064042 4284129580 4283865642 3090505498 302649604 2573587 907751437 3627904029 4283865642 4284195372 4284063275 4284060978 4284126260 4283996972 4284195372 4283865642 3627838237 890974221 2573330 263937 6533168 907817229 3124125722 4182476324 4283865386 4284063531 4284063531 4283865386 4182476324 3124125722 891040013 6335279 263937 0 0 1979406 319691015 1646476817 2939246103 3543687963 3543687963 2939246103 1646476817 319691015 1979406 0 0)
	offset: 0@0) ]! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 2/24/2015 08:33'!
gradientWindow

	<preference: 'gradientWindow'
		category: 'windows'
		description: 'If true, windows will have a gradient look.'
		type: #Boolean>
	^ GradientWindow ifNil: [true]
! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 4/4/2015 13:03'!
gradientWindow: aBoolean

	aBoolean = GradientWindow ifTrue: [^ self].
	GradientWindow := aBoolean.
	self refreshAllWindows.! !

!SystemWindow class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 14:27'!
hideExpandButton

	<preference: 'Hide Expand Button'
		category: 'windows'
		description: 'Hides the expand button in all windows'
		type: #Boolean>
	^ HideExpandButton ifNil: [ false ]
! !

!SystemWindow class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 14:44'!
hideExpandButton: aBoolean

	HideExpandButton := aBoolean.
	"Have the menu button frame rebuilt accordingly"
	self moveMenuButtonRight: self moveMenuButtonRight.
! !

!SystemWindow class methodsFor: 'new-morph participation' stamp: 'di 2/3/98 11:54'!
includeInNewMorphMenu
	"Include my subclasses but not me"
	^ self ~~ SystemWindow! !

!SystemWindow class methodsFor: 'top window' stamp: 'mt 5/19/2016 07:32'!
keyWindow
	"Multiple windows may appear to be on top, especially on large screens. Only one window, however, can be the key window, which gets the programmers primary attention."
	
	^ TopWindow! !

!SystemWindow class methodsFor: 'instance creation' stamp: 'mt 4/3/2015 17:35'!
labelled: labelString

	^ self basicNew
		initializeWithLabel: labelString;
		yourself! !

!SystemWindow class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 14:13'!
menuBoxFrame

	^ MenuBoxFrame ifNil: [
		MenuBoxFrame := (LayoutFrame new
								leftFraction: 0;
								leftOffset: self boxExtent x + 3;
								topFraction: 0;
								topOffset: 0;
								yourself)]! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/9/2015 00:07'!
menuBoxImage

	^ self gradientWindow
		ifTrue: [self menuBoxImageGradient]
		ifFalse: [self menuBoxImageFlat].
! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/9/2015 00:06'!
menuBoxImageFlat

	^ MenuBoxImageFlat ifNil: [MenuBoxImageFlat :=  (Form
	extent: 12@12
	depth: 32
	fromArray: #( 0 0 540831669 2671538101 4013715381 4282150837 4282150837 4013715381 2671538101 540831669 0 0 0 1077702581 4013715381 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 4013715381 809267125 0 540831669 4013715381 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 4013715381 272396213 2671538101 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 2671538101 4013715381 4282150837 4284582595 4288591834 4288591834 4288591834 4288591834 4288591834 4288591834 4283728318 4282150837 4013715381 4282150837 4282150837 4282939578 4292535537 4294967295 4294967295 4294967295 4294967295 4292535537 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 4284582595 4294967295 4294967295 4294967295 4294178554 4283728318 4282150837 4282150837 4282150837 4013715381 4282150837 4282150837 4282150837 4287737557 4294967295 4294967295 4286948817 4282150837 4282150837 4282150837 3745279925 2671538101 4282150837 4282150837 4282150837 4282150837 4291746797 4290958056 4282150837 4282150837 4282150837 4282150837 2151444405 540831669 4013715381 4282150837 4282150837 4282150837 4282939578 4282939578 4282150837 4282150837 4282150837 4013715381 272396213 0 809267125 4013715381 4282150837 4282150837 4282150837 4282150837 4282150837 4282150837 4013715381 809267125 0 0 0 272396213 2671538101 4013715381 4282150837 4282150837 3745279925 2151444405 272396213 0 0)
	offset: 0@0)]! !

!SystemWindow class methodsFor: 'initializing' stamp: 'topa 3/9/2015 00:01'!
menuBoxImageGradient

	^ MenuBoxImageGradient ifNil: [MenuBoxImageGradient := (Form
	extent: 14@14
	depth: 32
	fromArray: #(0 0 7774 285216318 1544039783 2837267842 3442168720 3442168720 2820490625 1544039782 285216574 7774 0 0 2087 1532860 839001178 3056161165 4150166716 4286817494 4287738079 4287803615 4286751958 4150035131 3039252364 839001178 1467065 1831 668526 855712858 3576584601 4286029012 4287935459 4287935458 4287869666 4287869666 4287935458 4287869923 4285963476 3576715929 838935642 602733 285216827 3038463881 4284517582 4286621149 4286489564 4286489564 4286489564 4286489564 4286489564 4286489820 4286621149 4284911823 3038726794 285216058 1577594471 4147012271 4284387029 4284978391 4285043927 4285109719 4285109719 4285044183 4284978135 4284649685 4284518357 4284912599 4147603633 1560751463 2853453696 4281560257 4281888189 4283200687 4284186808 4284515256 4284515257 4284252600 4283858102 4283529653 4282740653 4282151359 4281757378 2836676480 3474606220 4281757894 4281559740 4285761703 4292993507 4293454315 4293388522 4293388522 4293454315 4292862177 4285432998 4281494205 4281692102 3457829004 3491383436 4281757638 4281758153 4281953722 4288589251 4294835708 4294967295 4294967295 4294769659 4288194752 4281887931 4281823689 4281692102 3474606220 2870296704 4281560257 4281823689 4281758153 4282215861 4289838026 4294967294 4294967038 4289377991 4282084534 4281758153 4281823689 4281560257 2870296704 1611280232 4163592111 4281823946 4281823688 4281692360 4282740914 4291152852 4290758353 4282544051 4281692361 4281823688 4281823946 4146749358 1611214696 301994814 3088664202 4281626050 4281823689 4281823688 4281692103 4282937773 4282806447 4281692360 4281823688 4281823689 4281626050 3088664202 301994814 734577 906307420 3625930390 4281626050 4281823946 4281823689 4281692360 4281692360 4281823689 4281823946 4281626050 3625930390 889530204 734576 2347 2392036 906373213 3122218634 4180369583 4281560257 4281757638 4281757638 4281560257 4180369583 3122218634 889595997 2325471 2346 0 3 76907 318773060 1644834921 2937405569 3541780621 3541780621 2937405569 1644834921 318773059 76906 2 0)
	offset: 0@0)]! !

!SystemWindow class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 14:21'!
moveMenuButtonRight

	<preference: 'Move Menu Button Right'
		category: 'windows'
		description: 'Moves the menu to the right side of the window label area, while traditionally it used to be on the left side'
		type: #Boolean>
	^ self menuBoxFrame leftOffset negative
! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 7/28/2016 13:05'!
moveMenuButtonRight: aBoolean

	| absLeftOffset |
	absLeftOffset := ((self hideExpandButton and: [aBoolean])
		ifTrue: [absLeftOffset := self boxExtent x * 2]
		ifFalse: [absLeftOffset := self boxExtent x]) + 3.
	self menuBoxFrame leftOffset: (aBoolean 
										ifTrue: [absLeftOffset negated]
										ifFalse: [absLeftOffset]).
	self refreshAllWindows.! !

!SystemWindow class methodsFor: 'top window' stamp: 'mt 5/17/2016 16:04'!
noteTopWindowIn: aWorld
	"Look for a new top window in the given world. We have to reset the former top window because this is global state shared between all worlds."

	| newTopWindow |
	TopWindow := nil.
	newTopWindow := nil.

	aWorld ifNil: [^ self].

	aWorld submorphsDo: [:m |
		(m isSystemWindow and: [newTopWindow isNil])
			ifTrue: [newTopWindow := m].
		
		self flag: #refactor. "This really belongs in a special ProjWindow class"
		(m model isKindOf: Project)
			ifTrue: [m label ~= m model name ifTrue: [m setLabel: m model name]]].

	newTopWindow ifNotNil: [newTopWindow beKeyWindow].! !

!SystemWindow class methodsFor: 'private' stamp: 'mt 5/20/2016 11:21'!
reconfigureWindowsForFocus

	self allSubInstancesDo: [:window |
		window
			passivate;
			activate;
			unlockWindowDecorations;
			passivateIfNeeded].
	self topWindow activate.! !

!SystemWindow class methodsFor: 'initializing' stamp: 'mt 7/28/2016 13:11'!
refreshAllWindows
	"If there is some prominent UI change, use this method to update all open windows."
	
	SystemWindow allSubInstances do: [:w |
		w
			replaceBoxes;
			setDefaultParameters;
			refreshWindowColor].! !

!SystemWindow class methodsFor: 'preferences' stamp: 'ar 6/23/2010 17:12'!
resizeAlongEdges
	<preference: 'Resize Windows along edges'
		category: 'Morphic'
		description: 'When true, windows can be resized along their edges as well as the corners'
		type: #Boolean>
	^ResizeAlongEdges ifNil:[true]! !

!SystemWindow class methodsFor: 'preferences' stamp: 'ar 6/23/2010 17:13'!
resizeAlongEdges: aBool
	"Preference setter"
	ResizeAlongEdges := aBool! !

!SystemWindow class methodsFor: 'preferences' stamp: 'kb 2/15/2010 11:10'!
reuseWindows

	<preference: 'Reuse Windows'
		category: 'browsing'
		description: 'When enabled, before opening a new window check if there is any open window like it, and if there is, reuse it.'
		type: #Boolean>
	^ReuseWindows ifNil: [ false ]
! !

!SystemWindow class methodsFor: 'preferences' stamp: 'kb 2/15/2010 11:11'!
reuseWindows: aBoolean

	ReuseWindows := aBoolean! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 7/28/2016 10:34'!
roundedWindowCorners

	<preference: 'Rounded Window Corners'
		category: 'windows'
		description: 'If true, windows will have a rounded corners.'
		type: #Boolean>
	^ RoundedWindowCorners ifNil: [false]! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 7/28/2016 10:34'!
roundedWindowCorners: aBoolean

	RoundedWindowCorners = aBoolean ifTrue: [^ self].
	RoundedWindowCorners := aBoolean.
	self refreshAllWindows.! !

!SystemWindow class methodsFor: 'top window' stamp: 'sw 12/6/2000 19:43'!
sendTopWindowToBack
	"Send the top window of the world to the back, activating the one just beneath it"

	TopWindow ifNotNil:
		[TopWindow sendToBack]! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 7/25/2016 15:48'!
themeProperties

	^ super themeProperties, {	
		{ #borderColorModifier. 'Colors'. 'How to derive the borderColor from the window color.' }.
		{ #borderWidth. 'Borders'. 'Width of the menu''s border.' }.
		{ #titleFont. 'Fonts'. 'Font of the window title.' }.
		
		{ #unfocusedWindowColorModifier. 'Colors'. 'A block with one argument that modifies a color to look unfocused.' }.
		{ #unfocusedLabelColor. 'Colors'. 'Window title color when window has no focus.'}.
		{ #focusedLabelColor. 'Colors'. 'Window title color when window has focus.'}.
	}! !

!SystemWindow class methodsFor: 'top window' stamp: 'rr 3/10/2006 14:19'!
topWindow
	^ TopWindow! !

!SystemWindow class methodsFor: 'top window' stamp: 'sw 1/4/2000 15:22'!
wakeUpTopWindowUponStartup
	TopWindow ifNotNil:
		[TopWindow isCollapsed ifFalse:
			[TopWindow model ifNotNil:
				[TopWindow model modelWakeUpIn: TopWindow]]]! !

!SystemWindow class methodsFor: 'top window' stamp: 'mt 5/14/2016 14:35'!
windowsIn: aWorld

	^ self windowsIn: aWorld satisfying: [:window | true].! !

!SystemWindow class methodsFor: 'top window' stamp: 'nice 12/27/2009 03:11'!
windowsIn: aWorld satisfying: windowBlock
	| windows |

	windows := OrderedCollection new.
	aWorld ifNil: [^windows].	"opening MVC in Morphic - WOW!!"
	aWorld submorphs do:
		[:m | | s |
		((m isSystemWindow) and: [windowBlock value: m])
			ifTrue: [windows addLast: m]
			ifFalse: [((m isKindOf: TransformationMorph) and: [m submorphs size = 1])
					ifTrue: [s := m firstSubmorph.
							((s isSystemWindow) and: [windowBlock value: s])
								ifTrue: [windows addLast: s]]]].
	^ windows! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 5/26/2016 17:58'!
windowsRaiseOnClick
	<preference: 'Windows Raise On Any Click'
		category: 'windows'
		description: 'If false, windows only raise when clicking on window decorations.'
		type: #Boolean>
	^ WindowsRaiseOnClick ifNil: [ true ]! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 5/26/2016 17:03'!
windowsRaiseOnClick: boolean
	WindowsRaiseOnClick := boolean.! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 5/27/2016 14:38'!
windowTitleActiveOnFirstClick
	<preference: 'Windows'' Controls Are Always Active'
		category: 'windows'
		description: '... except for grips and splitters. Those remain always active and an option to bring the window to the front.'
		type: #Boolean>
	^ WindowTitleActiveOnFirstClick ifNil: [ true ]! !

!SystemWindow class methodsFor: 'preferences' stamp: 'mt 5/27/2016 14:38'!
windowTitleActiveOnFirstClick: boolean
	WindowTitleActiveOnFirstClick := boolean.
	self reconfigureWindowsForFocus.! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 5/29/2016 12:02'!
activate
	"Bring the receiver to the top.  If I am modal, bring along my modal owning window and my model child as well."

	self isActive ifTrue: [self lookFocused. ^ self].
	self topRendererOrSelf owner ifNil: [^ self "avoid spurious activate when drop in trash"].
	
	self isActive: true.
			
	"Special handling for expanded windows."
	self isCollapsed ifFalse: [
		model modelWakeUpIn: self.
		self positionSubmorphs].

	self submorphsDo: [:each | each unlock].

	self lookFocused.! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 7/28/2016 13:07'!
addCloseBox
	"If I have a labelArea, add a close box to it"

	labelArea ifNil: [^ self].
	mustNotClose == true ifTrue: [^ self].
	closeBox ifNotNil: [closeBox delete].
	closeBox := self createCloseBox.
	closeBox layoutFrame: self class closeBoxFrame.
	labelArea addMorphFront: closeBox! !

!SystemWindow methodsFor: 'initialization' stamp: 'ar 6/23/2010 17:13'!
addCornerGrips
	"When enabled via preference, also add edge grips"
	super addCornerGrips.
	self class resizeAlongEdges ifTrue:[self addEdgeGrips].! !

!SystemWindow methodsFor: 'menu' stamp: 'md 11/14/2003 17:30'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
"template..."
	aCustomMenu addLine.
	aCustomMenu add: 'edit label...' translated action: #relabel.
! !

!SystemWindow methodsFor: 'initialization' stamp: 'tfel 2/28/2010 14:51'!
addExpandBox
	"If I have a labelArea, add a close box to it"
	
	labelArea ifNil: [^ self].
	self class hideExpandButton ifTrue: [^ self].
	expandBox ifNotNil: [expandBox delete].
	expandBox := self createExpandBox.
	expandBox layoutFrame: self class expandBoxFrame.
	labelArea addMorphBack: expandBox! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 5/31/2016 12:07'!
addKeyboardShortcuts
	"Install an event capture filter to add shortcuts for window control operations."
	
	self addKeyboardCaptureFilter: self.! !

!SystemWindow methodsFor: 'label' stamp: 'tfel 2/28/2010 16:24'!
addLabel

	(labelArea isNil or: [label isNil]) ifTrue: [^ self].
	labelArea 
		addMorphBack: (Morph new extent: self class borderWidth @ 0);
		addMorphBack: label.
	label on: #startDrag send: #startDragFromLabel: to: self.
	self class clickOnLabelToEdit
		ifTrue: [label on: #mouseUp send: #relabel to: self].
	self class doubleClickOnLabelToExpand
		ifTrue: [label on: #doubleClick send: #expandBoxHit to: self].! !

!SystemWindow methodsFor: 'initialization' stamp: 'tfel 3/25/2010 23:00'!
addLabelArea

	labelArea ifNotNil: [labelArea abandon].
	labelArea := (AlignmentMorph newSpacer: Color transparent)
			vResizing: #spaceFill;
			layoutPolicy: ProportionalLayout new.
	self addMorph: labelArea.! !

!SystemWindow methodsFor: 'initialization' stamp: 'tfel 2/28/2010 14:14'!
addMenuControl
	"If I have a label area, add a menu control to it."
	
	labelArea ifNil: [^ self].
	"No menu if no label area"
	menuBox ifNotNil: [menuBox delete].
	menuBox := self createMenuBox.
	menuBox layoutFrame: self class menuBoxFrame.
	labelArea addMorphBack: menuBox! !

!SystemWindow methodsFor: 'panes' stamp: 'RAA 1/8/2001 20:37'!
addMorph: aMorph frame: relFrame
	| frame |
	frame := LayoutFrame new.
	frame 
		leftFraction: relFrame left; 
		rightFraction: relFrame right; 
		topFraction: relFrame top; 
		bottomFraction: relFrame bottom.
	self addMorph: aMorph fullFrame: frame.

! !

!SystemWindow methodsFor: 'panes' stamp: 'mt 8/25/2015 11:49'!
addMorph: aMorph fullFrame: aLayoutFrame
	"Add aMorph according to aLayoutFrame."

	super addMorph: aMorph fullFrame: aLayoutFrame.

	paneMorphs := paneMorphs copyReplaceFrom: 1 to: 0 with: (Array with: aMorph).
	aMorph isImageMorph ifFalse: [aMorph adoptPaneColor: self paneColor].! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 4/28/2015 13:31'!
addPaneSplitters

	self removeCornerGrips.

	super addPaneSplitters.
	
	self addCornerGrips.! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'dns 2/2/2000 14:20'!
allowReframeHandles

	^ allowReframeHandles! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'dns 2/2/2000 14:20'!
allowReframeHandles: aBoolean

	allowReframeHandles := aBoolean! !

!SystemWindow methodsFor: 'stepping' stamp: 'sw 10/19/1999 09:30'!
amendSteppingStatus
	"Circumstances having changed, find out whether stepping is wanted and assure that the new policy is carried out"

	self wantsSteps
		ifTrue:
			[self arrangeToStartStepping]
		ifFalse:
			[self stopStepping]! !

!SystemWindow methodsFor: 'open/close' stamp: 'kb 2/15/2010 11:11'!
anyOpenWindowLikeMe
	
	self class reuseWindows ifFalse: [ ^Array empty ].
	^ SystemWindow
		windowsIn: World 
		satisfying: 
			[ : each |
			each model class = self model class
				and: [ (each model respondsTo: #representsSameBrowseeAs:) 
				and: [ each model representsSameBrowseeAs: self model ] ] ]
! !

!SystemWindow methodsFor: 'initialization' stamp: 'dgd 10/26/2004 19:44'!
applyModelExtent
	| initialExtent |
	initialExtent := Preferences bigDisplay
				ifTrue: [(model initialExtent * 1.5) rounded]
				ifFalse: [model initialExtent].
	self extent: initialExtent ! !

!SystemWindow methodsFor: 'user interface' stamp: 'mt 7/29/2016 14:47'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	
	self
		setDefaultParameters;
		refreshWindowColor.

	self isLookingFocused
		ifTrue: [self lookUnfocused; lookFocused]
		ifFalse: [self lookFocused; lookUnfocused].
		
	self isCollapsed ifTrue: [self setProperty: #applyTheme toValue: true].! !

!SystemWindow methodsFor: 'drawing' stamp: 'di 8/16/1998 01:14'!
areasRemainingToFill: aRectangle
	| areas |
	(areas := super areasRemainingToFill: aRectangle) isEmpty
		ifTrue: [^ areas "good news -- complete occlusion"].
	"Check for special case that this is scrollbar damage"
	((bounds topLeft - (14@0) corner: bounds bottomRight) containsRect: aRectangle) ifTrue:
		[paneMorphs do: [:p | ((p isKindOf: ScrollPane) and: [p scrollBarFills: aRectangle])
							ifTrue: [^ Array new]]].
	^ areas! !

!SystemWindow methodsFor: 'layout' stamp: 'mt 5/17/2016 13:53'!
assureLabelAreaVisible
	"Label should be visible to interact with."

	(RealEstateAgent maximumUsableAreaInWorld: self world) in: [:rect |
		 self innerBounds top < rect top ifTrue: [self top: rect top]].! !

!SystemWindow methodsFor: 'top window' stamp: 'mt 5/31/2016 11:21'!
beKeyWindow
	"Let me be the most important window on the screen. I am at the top and I can have a shadow to get more attention by the user. I am the window that is responsible for window keyboard shortcuts."

	| oldKeyWindow |
	self isKeyWindow ifTrue: [^ self].

	oldKeyWindow := TopWindow.
	TopWindow := self.

	self
		unlockWindowDecorations; "here, because all windows might be active anyway"
		activate; "if not already active, activate now"
		comeToFront. "key windows are on top"

	"Change appearance to get noticed."
	self hasDropShadow: Preferences menuAppearance3d.
	(self valueOfProperty: #borderWidthWhenActive)
		ifNotNil: [:bw | self acquireBorderWidth: bw].

	oldKeyWindow ifNotNil: [:wnd |
		wnd passivateIfNeeded.
		
		"Change appearance to not look prettier than the new key window."
		wnd hasDropShadow: false.
		(wnd valueOfProperty: #borderWidthWhenInactive)
			ifNotNil: [:bw | wnd acquireBorderWidth: bw]].

	"Synchronize focus look with position of current hand because any call could have made this window the new key window."
	self updateFocusLookAtHand.! !

!SystemWindow methodsFor: 'top window' stamp: 'mt 6/9/2016 18:03'!
beKeyWindowIfNeeded: evt
	"Make me the new key window if needed. Ensure that the focus look matches the keyboard focus."

	(self isKeyWindow not
		and: [self class windowsRaiseOnClick
			or: [self windowDecorations anySatisfy: [:morph | morph bounds containsPoint: evt position]] ])
				ifFalse: [
					"Just ensure the correct look after the event."
					self updateFocusLookForKeyboardFocus]
				ifTrue: [
					"We are the new key window at the front etc."
					self beKeyWindow.
					"There may have been a pop-up menu spawned. We do not want to occlude that."
					evt hand mouseFocus ifNotNil: [:morph |
						morph owner = self world ifTrue: [morph comeToFront]].
					evt hand balloonHelpList do: [:morph |
						morph comeToFront]]! !

!SystemWindow methodsFor: 'initialization' stamp: 'tfel 2/28/2010 14:00'!
boxExtent
	"the label height is used to be proportional to the fonts preferences"
	
	^ self class boxExtent
		max: label height @ label height ! !

!SystemWindow methodsFor: 'menu' stamp: 'laza 2/26/2010 22:06'!
buildWindowMenu
	| aMenu |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu add: 'change title...' translated action: #relabel.
	aMenu addLine.
	aMenu add: 'send to back' translated action: #sendToBack.
	aMenu add: 'make next-to-topmost' translated action: #makeSecondTopmost.
	aMenu addLine.
	self mustNotClose
		ifFalse:
			[aMenu add: 'make unclosable' translated action: #makeUnclosable]
		ifTrue:
			[aMenu add: 'make closable' translated action: #makeClosable].
	aMenu
		add: (self isSticky ifTrue: ['make draggable'] ifFalse: ['make undraggable']) translated 
		action: #toggleStickiness.
	aMenu addLine.
	self unexpandedFrame 
		ifNil: [aMenu add: 'full screen' translated action: #expandBoxHit]
		ifNotNil: [aMenu add: 'original size' translated action: #expandBoxHit].
	self isCollapsed ifFalse: [aMenu add: 'window color...' translated action: #setWindowColor].
	^aMenu! !

!SystemWindow methodsFor: 'menu' stamp: 'cmm 12/4/2010 15:11'!
changeColor
	"Change the color of the receiver -- triggered, e.g. from a menu.  This variant allows the recolor triggered from the window's halo recolor handle to have the same result as choosing change-window-color from the window-title menu"
	NewColorPickerMorph useIt
		ifTrue:
			[ (NewColorPickerMorph
				on: self
				originalColor: self color
				setColorSelector: #setWindowColor:) openNear: self fullBoundsInWorld ]
		ifFalse:
			[ ColorPickerMorph new
				 choseModalityFromPreference ;
				 sourceHand: self activeHand ;
				 target: self ;
				 selector: #setWindowColor: ;
				 originalColor: self color ;
				
				putUpFor: self
				near: self fullBoundsInWorld ]! !

!SystemWindow methodsFor: 'open/close' stamp: 'ar 10/5/2000 17:26'!
closeBoxHit
	"The user clicked on the close-box control in the window title.  For Mac users only, the Mac convention of option-click-on-close-box is obeyed if the mac option key is down."

	Preferences dismissAllOnOptionClose ifTrue:
		[Sensor rawMacOptionKeyPressed ifTrue:
			[^ self world closeUnchangedWindows]].
	self delete
! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'ar 2/10/1999 04:19'!
collapse
	self isCollapsed ifFalse:[self collapseOrExpand]! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'di 5/20/1998 08:25'!
collapsedFrame
	^ collapsedFrame! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'mt 7/29/2016 14:48'!
collapseOrExpand
	"Collapse or expand the window, depending on existing state"
	| cf |
	isCollapsed
		ifTrue: 
			["Expand -- restore panes to morphics structure"
			isCollapsed := false.
			self beKeyWindow.  "Bring to frint first"
			Preferences collapseWindowsInPlace
				ifTrue: 
					[fullFrame := fullFrame align: fullFrame topLeft with: self getBoundsWithFlex topLeft]
				ifFalse:
					[collapsedFrame := self getBoundsWithFlex].
			collapseBox ifNotNil: [collapseBox setBalloonText: 'collapse this window' translated].
			self setBoundsWithFlex: fullFrame.
			paneMorphs reverseDo: 
					[:m |  self addMorph: m unlock.
					self world startSteppingSubmorphsOf: m].
			self addPaneSplitters.
			(self hasProperty: #applyTheme) ifTrue: [
				self removeProperty: #applyTheme.
				self userInterfaceTheme applyTo: self allMorphs]]
		ifFalse: 
			["Collapse -- remove panes from morphics structure"
			isCollapsed := true.
			fullFrame := self getBoundsWithFlex.
			"First save latest fullFrame"
			paneMorphs do: [:m | m delete; releaseCachedState].
			self removePaneSplitters.
			self removeCornerGrips.
			model modelSleep.
			cf := self getCollapsedFrame.
			(collapsedFrame isNil and: [Preferences collapseWindowsInPlace not]) ifTrue:
				[collapsedFrame := cf].
			self setBoundsWithFlex: cf.
			collapseBox ifNotNil: [collapseBox setBalloonText: 'expand this window' translated ].
			expandBox ifNotNil: [expandBox setBalloonText: 'expand this window' translated ].
			self sendToBack].
	self layoutChanged! !

!SystemWindow methodsFor: 'drawing' stamp: 'ar 8/15/2001 21:55'!
colorForInsets
	^self paneColor colorForInsets! !

!SystemWindow methodsFor: 'top window' stamp: 'mt 6/4/2016 19:18'!
comeToFront

	"Modal windows: Walk along the modal owner chain, the bring all to top."
	self modalOwner ifNotNil: [:mo | mo isSystemWindow ifTrue: [
		^ mo modalOwner
			ifNil: [mo comeToFrontModally]
			ifNotNil: [:omo | omo comeToFront]]].
	self modalChild ifNotNil: [^ self comeToFrontModally].
	
	"Now show me."
	super comeToFront.
		
	"Label should be visible to interact with."
	self assureLabelAreaVisible.
	
	"Handle historic companions. May be removed in the future."
	self extantSketchEditor ifNotNil: [:sketchEditor | 
		sketchEditor comeToFront.
		(self world findA: PaintBoxMorph) ifNotNil: [:pal | pal comeToFront]].! !

!SystemWindow methodsFor: 'top window' stamp: 'mt 6/4/2016 19:16'!
comeToFrontModally
	"If I am modal, bring along my modal owning window and my model child as well."

	super comeToFront.
	
	"Bring up my modal children in front of me."
	self modalChild ifNotNil: [:wnd | 
		wnd comeToFrontModally.
		wnd modalChild ifNil: [wnd flash]].! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'cmm 6/21/2011 20:13'!
contractToOriginalSize
	self bounds: self unexpandedFrame.
	self unexpandedFrame: nil.
	expandBox ifNotNil: [expandBox setBalloonText: 'expand this window' translated].! !

!SystemWindow methodsFor: 'layout' stamp: 'nice 1/10/2010 17:48'!
convertAlignment
	"Primarily Jesse Welton's code to convert old system windows to ones with modern layout scheme"

	self layoutPolicy: ProportionalLayout new.
	(paneMorphs isNil 
		or: [paneRects isNil or: [paneMorphs size ~= paneRects size]]) 
			ifFalse: 
				[self addLabelArea.
				self putLabelItemsInLabelArea.
				self setFramesForLabelArea.
				paneMorphs with: paneRects
					do: 
						[:m :r | 
						| frame |
						frame := LayoutFrame new.
						frame
							leftFraction: r left;
							rightFraction: r right;
							topFraction: r top;
							bottomFraction: r bottom.
						m layoutFrame: frame.
						m
							hResizing: #spaceFill;
							vResizing: #spaceFill]].
	labelArea isNil 
		ifTrue: 
			[self addLabelArea.
			self putLabelItemsInLabelArea.
			self setFramesForLabelArea.
			paneMorphs ifNotNil: 
					[paneMorphs do: 
							[:m | 
							| frame |
							frame := m layoutFrame ifNil: [LayoutFrame new].
							frame topOffset: (frame topOffset ifNil: [0]) - self labelHeight.
							frame bottomFraction ~= 1.0 
								ifTrue: 
									[frame bottomOffset: (frame bottomOffset ifNil: [0]) - self labelHeight]]]].
	label ifNotNil: 
			[| frame |
			frame := LayoutFrame new.
			frame
				leftFraction: 0.5;
				topFraction: 0;
				leftOffset: label width negated // 2.
			label layoutFrame: frame].
	collapseBox ifNotNil: 
			[| frame |
			frame := LayoutFrame new.
			frame
				rightFraction: 1;
				topFraction: 0;
				rightOffset: -1;
				topOffset: 1.
			collapseBox layoutFrame: frame].
	stripes ifNotNil: 
			[| frame |
			frame := LayoutFrame new.
			frame
				leftFraction: 0;
				topFraction: 0;
				rightFraction: 1;
				leftOffset: 1;
				topOffset: 1;
				rightOffset: -1.
			stripes first layoutFrame: frame.
			stripes first height: self labelHeight - 2.
			stripes first hResizing: #spaceFill.
			frame := LayoutFrame new.
			frame
				leftFraction: 0;
				topFraction: 0;
				rightFraction: 1;
				leftOffset: 3;
				topOffset: 3;
				rightOffset: -3.
			stripes last layoutFrame: frame.
			stripes last height: self labelHeight - 6.
			stripes last hResizing: #spaceFill].
	menuBox ifNotNil: 
			[| frame |
			frame := LayoutFrame new.
			frame
				leftFraction: 0;
				leftOffset: 19;
				topFraction: 0;
				topOffset: 1.
			menuBox layoutFrame: frame].
	closeBox ifNotNil: 
			[| frame |
			frame := LayoutFrame new.
			frame
				leftFraction: 0;
				leftOffset: 4;
				topFraction: 0;
				topOffset: 1.
			closeBox layoutFrame: frame]! !

!SystemWindow methodsFor: 'object fileIn' stamp: 'JW 1/31/2001 08:57'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	allowReframeHandles ifNil: [allowReframeHandles := true].
	self layoutPolicy ifNil: [self convertAlignment].
	labelArea ifNil: [self convertAlignment].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!SystemWindow methodsFor: 'initialization' stamp: 'ar 3/24/2010 20:10'!
createBox: aForm
	"create a button with a form to be used in the label area"
	
	| box |
	box := SystemWindowButton new.
	box color: Color transparent;
		 target: self;
		 useSquareCorners;
		 borderWidth: 0;
		 labelGraphic: aForm;
		 extent: self boxExtent.
	^ box! !

!SystemWindow methodsFor: 'initialization' stamp: 'tfel 2/27/2010 19:09'!
createCloseBox
	^ (self createBox: self class closeBoxImage)
		actionSelector: #closeBoxHit;
		setBalloonText: 'close this window' translated! !

!SystemWindow methodsFor: 'initialization' stamp: 'tfel 2/27/2010 19:09'!
createCollapseBox
	^ (self createBox: self class collapseBoxImage)
		actionSelector: #collapseOrExpand;
		setBalloonText: 'collapse this window' translated.
! !

!SystemWindow methodsFor: 'initialization' stamp: 'cmm 6/21/2011 20:13'!
createExpandBox
	^ (self createBox: self class expandBoxImage)
		actionSelector: #expandBoxHit;
		setBalloonText: 'expand this window' translated! !

!SystemWindow methodsFor: 'initialization' stamp: 'tfel 2/27/2010 22:31'!
createMenuBox
	^ (self createBox: self class menuBoxImage)
		actionSelector: #offerWindowMenu;
		setBalloonText: 'window menu' translated! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 4/3/2015 17:24'!
defaultBounds

	^ 0@0 corner: 300@200! !

!SystemWindow methodsFor: 'open/close' stamp: 'mt 5/17/2016 17:10'!
delete
	| thisWorld sketchEditor aPaintBox |
	self mustNotClose ifTrue: [^self].
	model okToClose ifFalse: [^self].
	thisWorld := self world.
	sketchEditor := self extantSketchEditor.
	
	self activeHand removeKeyboardListener: self.
	
	self isFlexed
		ifTrue: [owner delete]
		ifFalse: [super delete].
	model windowIsClosing; release.
	model := nil.
	sketchEditor ifNotNil:
		[sketchEditor deleteSelfAndSubordinates.
		 (thisWorld notNil 
		  and: [(aPaintBox := thisWorld paintBoxOrNil) notNil]) ifTrue:
			[aPaintBox delete]].
		
	SystemWindow noteTopWindowIn: thisWorld! !

!SystemWindow methodsFor: 'menu' stamp: 'sw 8/7/2000 17:35'!
deleteCloseBox
	closeBox ifNotNil:
		[closeBox delete.
		closeBox := nil]! !

!SystemWindow methodsFor: 'top window' stamp: 'mt 5/14/2016 14:26'!
dimWindowButtons
	{closeBox. collapseBox. menuBox. expandBox}
		do: [:b | b ifNotNil: [b dim]]! !

!SystemWindow methodsFor: 'events' stamp: 'bmk 3/19/2002 02:09'!
doFastFrameDrag: grabPoint
	"Do fast frame dragging from the given point"

	| offset newBounds outerWorldBounds |
	outerWorldBounds := self boundsIn: nil.
	offset := outerWorldBounds origin - grabPoint.
	newBounds := outerWorldBounds newRectFrom: [:f | 
		Sensor cursorPoint + offset extent: outerWorldBounds extent].
	self position: (self globalPointToLocal: newBounds topLeft); comeToFront! !

!SystemWindow methodsFor: 'colors' stamp: 'md 2/24/2006 15:47'!
existingPaneColor
	"Answer the existing pane color for the window, obtaining it from the first paneMorph if any, and fall back on using the second stripe color if necessary."

	| aColor |
	aColor := self valueOfProperty: #paneColor.
	aColor ifNil: [self setProperty: #paneColor toValue: (aColor := self paneColor)].
	^aColor.! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'ar 2/10/1999 04:20'!
expand
	self isCollapsed ifTrue:[self collapseOrExpand]! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'laza 2/26/2010 21:58'!
expandBoxHit
	isCollapsed
		ifTrue: [self	hide;
					collapseOrExpand;
					expandToFullScreen;
					show]
		ifFalse: [self unexpandedFrame 
					ifNil: [self expandToFullScreen]
					ifNotNil: [self contractToOriginalSize]]! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'tfel 2/27/2010 20:21'!
expandToFullScreen
	self unexpandedFrame ifNil: [ self unexpandedFrame: fullFrame ].
	self fullScreen.
	expandBox ifNotNil: [expandBox setBalloonText: 'contract to original size' translated]! !

!SystemWindow methodsFor: 'top window' stamp: 'LC 9/28/1999 19:04'!
extantSketchEditor
	"If my world has an extant SketchEditorMorph associated with anything  
	in this window, return that SketchEditor, else return nil"
	| w sketchEditor pasteUp |
	(w := self world) isNil ifTrue: [^ nil].
	(sketchEditor := w sketchEditorOrNil) isNil ifTrue: [^ nil].
	(pasteUp := sketchEditor enclosingPasteUpMorph) isNil ifTrue: [^ nil].
	self findDeepSubmorphThat: [:m | m = pasteUp]
		ifAbsent: [^ nil].
	^ sketchEditor! !

!SystemWindow methodsFor: 'geometry' stamp: 'bvs 3/16/2004 11:12'!
extent: aPoint 
	"Set the receiver's extent to value provided. Respect my minimumExtent."

	| newExtent |
	newExtent := self isCollapsed
		ifTrue: [aPoint]
		ifFalse: [aPoint max: self minimumExtent].
	newExtent = self extent ifTrue: [^ self].

	isCollapsed
		ifTrue: [super extent: newExtent x @ (self labelHeight + 2)]
		ifFalse: [super extent: newExtent].
	isCollapsed
		ifTrue: [collapsedFrame := self bounds]
		ifFalse: [fullFrame := self bounds]! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'di 10/28/1999 13:21'!
fastFramingOn

	^ Preferences fastDragWindowForMorphic and: [self isFlexed not]! !

!SystemWindow methodsFor: 'events' stamp: 'cmm 6/11/2016 14:19'!
filterEvent: aKeyboardEvent for: anObject
	"Provide keyboard shortcuts."
		
	aKeyboardEvent isKeystroke
		ifFalse: [^ aKeyboardEvent].
	
	aKeyboardEvent commandKeyPressed ifTrue: [
		aKeyboardEvent keyCharacter caseOf: {
			[$\] -> [self class sendTopWindowToBack].
			[Character escape] -> [self class deleteTopWindow].
			[$/] -> [self class bringWindowUnderHandToFront].
		} otherwise: [^ aKeyboardEvent "no hit"].
		^ aKeyboardEvent ignore "hit!!"].
	
	aKeyboardEvent controlKeyPressed ifTrue: [
		aKeyboardEvent keyCharacter caseOf: {
			[Character escape] -> [self world findWindow: aKeyboardEvent].
		} otherwise: [^ aKeyboardEvent "no hit"].
		^ aKeyboardEvent ignore "hit!!"].

	^ aKeyboardEvent "no hit"! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'di 5/20/1998 08:25'!
fullFrame
	^ fullFrame! !

!SystemWindow methodsFor: 'menu' stamp: 'dhn 11/13/2014 15:39'!
fullScreen
	"Zoom Window to full owner size, allowing for scroll bars and desk margins"
	
	| left right possibleBounds |
	left := right := 0.
	self paneMorphs
		do: [:pane | ((pane isKindOf: ScrollPane)
					and: [pane retractableScrollBar])
				ifTrue: [pane scrollBarOnLeft
						ifTrue: [left := left max: pane scrollBarThickness]
						ifFalse: [right := right max: pane scrollBarThickness]]].
	possibleBounds := (RealEstateAgent maximumUsableAreaInWorld: self owner)
				insetBy: (left @ 0 corner: right @ 0).
	Preferences fullScreenLeavesDeskMargins
		ifTrue: [possibleBounds := possibleBounds insetBy: 22].
	self bounds: possibleBounds! !

!SystemWindow methodsFor: 'menu' stamp: 'dao 10/1/2004 12:57'!
fullScreenMaximumExtent
	"Zoom Window to Full World size with possible DeskMargins
	obey the maximum extent rules"
	
	| left right possibleBounds |
	left := right := 0.
	self paneMorphs
		do: [:pane | ((pane isKindOf: ScrollPane)
					and: [pane retractableScrollBar])
				ifTrue: [pane scrollBarOnLeft
						ifTrue: [left := left max: pane scrollBarThickness]
						ifFalse: [right := right max: pane scrollBarThickness]]].
	possibleBounds := self worldBounds
				insetBy: (left @ 0 corner: right @ 0).

	self maximumExtent ifNotNil:
		[possibleBounds := possibleBounds origin extent: ( self maximumExtent min: ( possibleBounds extent ))].
	((Flaps sharedFlapsAllowed
				and: [Project current flapsSuppressed not])
			or: [Preferences fullScreenLeavesDeskMargins])
		ifTrue: [possibleBounds := possibleBounds insetBy: 22].
	self bounds: possibleBounds! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'di 10/28/1999 14:14'!
getBoundsWithFlex
	"Return the lastest bounds rectangle with origin forced to global coordinates"

	self isFlexed
		ifTrue: [^ ((owner transform localPointToGlobal: bounds topLeft)
										extent: bounds extent)]
		ifFalse: [^ self bounds].
! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'svp 8/21/2001 11:13'!
getCollapsedFrame

	| tmp |
	^Preferences collapseWindowsInPlace 
		ifTrue:
			[tmp := self getBoundsWithFlex.
			tmp origin corner: (tmp corner x @ 18)]
		ifFalse:
			[RealEstateAgent assignCollapseFrameFor: self]! !

!SystemWindow methodsFor: 'label' stamp: 'bvs 4/2/2004 16:55'!
getRawLabel
	| contentsFit |
	contentsFit := label duplicate fitContents.
	contentsFit extent: (label extent x min: contentsFit extent x) @ contentsFit extent y.
	
	^ contentsFit
! !

!SystemWindow methodsFor: 'colors' stamp: 'mt 7/26/2016 08:04'!
gradientWithColor: aColor

	| gradient cc |
	cc := aColor adjustSaturation: -0.08 brightness: 0.4.
	gradient := GradientFillStyle ramp: {
		0.0 -> (self isKeyWindow ifTrue: [cc muchLighter] ifFalse: [cc]). 
		0.25 -> (aColor mixed: 0.5 with: cc). 
		1.0 -> aColor.
	}.
	gradient origin: self topLeft.
	gradient direction: 0 @ self labelHeight.
	^gradient! !

!SystemWindow methodsFor: 'events' stamp: 'mt 5/31/2016 11:20'!
handleListenEvent: aUserInputEvent 
	"See #mouseEnterDragging:. Watch for finished drag-and-drop action and lock contents accordingly."

	(aUserInputEvent isMouse and: [ aUserInputEvent hand hasSubmorphs not ]) ifTrue: 
		[ self isKeyWindow ifFalse: [ self passivateIfNeeded ].
		aUserInputEvent hand removeMouseListener: self ].! !

!SystemWindow methodsFor: 'events' stamp: 'mt 6/4/2016 16:18'!
handleMouseDown: evt 

	self beKeyWindowIfNeeded: evt.
	^ super handleMouseDown: evt! !

!SystemWindow methodsFor: 'events' stamp: 'mt 6/4/2016 16:21'!
handleMouseUp: evt 
	"Some widgets grab the keyboard focus only on mouse-up. Hence, we need to update our looks."
	
	self updateFocusLookForKeyboardFocus.
	^ super handleMouseUp: evt! !

!SystemWindow methodsFor: 'events' stamp: 'mt 4/5/2015 11:48'!
handlesMouseDown: evt 

	^ true! !

!SystemWindow methodsFor: 'events' stamp: 'ar 1/31/2001 21:02'!
handlesMouseOverDragging: evt
	^true! !

!SystemWindow methodsFor: 'events' stamp: 'cmm 9/9/2015 13:51'!
handlesMouseOver: anEvent
	^ true! !

!SystemWindow methodsFor: 'panes' stamp: 'ul 11/15/2010 12:06'!
holdsTranscript
	"ugh"
	| plug |
	^ paneMorphs size = 1 and: [((plug := paneMorphs first) isKindOf: PluggableTextMorph) and: [plug model isKindOf: TranscriptStream]]! !

!SystemWindow methodsFor: 'thumbnail' stamp: 'dgd 9/22/2004 19:29'!
icon
	"Answer a form with an icon to represent the receiver"
	^ MenuIcons windowIcon! !

!SystemWindow methodsFor: 'open/close' stamp: 'mt 4/19/2015 18:21'!
initialExtent
	"Ask the model for it's initial extent."
	
	^ model initialExtent! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 5/31/2016 11:21'!
initialize
	"Initialize a system window. Add label, stripes, etc., if desired"

	super initialize.

	self layoutPolicy: ProportionalLayout new.

	self wantsPaneSplitters: true.
	self layoutInset: ProportionalSplitterMorph gripThickness.
	self cellInset: ProportionalSplitterMorph gripThickness.

	self initializeLabelArea.				
	self addCornerGrips.
	self setDefaultParameters.

	allowReframeHandles := true.
	isCollapsed := false.
	paneMorphs := Array new.
	mustNotClose := false.
	updatablePanes := Array new.

	self initializeKeyboardShortcuts.! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 5/31/2016 12:07'!
initializeKeyboardShortcuts
	"Install an event capture filter to add shortcuts for window control operations."
	
	PasteUpMorph globalCommandKeysEnabled
		ifTrue: [self addKeyboardShortcuts].! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 7/25/2016 15:50'!
initializeLabelArea
	"Initialize the label area (titlebar) for the window."
	
	labelString ifNil: [labelString := 'Untitled Window'].
	label := StringMorph new contents: labelString.
			"Add collapse box so #labelHeight will work"
			collapseBox := self createCollapseBox.
			stripes := Array
						with: (RectangleMorph newBounds: bounds)
						with: (RectangleMorph newBounds: bounds).
			"see extent:"
			self addLabelArea.
			self setLabelWidgetAllowance.
			self addCloseBox.
			self class moveMenuButtonRight 
				ifTrue: [self addLabel. self addMenuControl]
				ifFalse: [self addMenuControl. self addLabel].
			self addExpandBox.
			labelArea addMorphBack: collapseBox.
			self setFramesForLabelArea.
			Preferences noviceMode
				ifTrue: [closeBox
						ifNotNil: [closeBox setBalloonText: 'close window'].
					menuBox
						ifNotNil: [menuBox setBalloonText: 'window menu'].
					collapseBox
						ifNotNil: [collapseBox setBalloonText: 'collapse/expand window']].
! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 4/3/2015 17:31'!
initializeWithLabel: aString

	labelString := aString.	
	self initialize.! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 5/17/2016 16:12'!
isActive
	
	^ isActive ifNil: [false]! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 5/17/2016 15:04'!
isActive: aBoolean
	
	isActive := aBoolean.! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'di 5/19/1998 09:34'!
isCollapsed
	^ isCollapsed! !

!SystemWindow methodsFor: 'top window' stamp: 'mt 5/19/2016 07:33'!
isKeyWindow
	"The key window is the window with the user's main attention. It usually has a shadow or other visual hints to separate it from other windows. Usually, the key window is at the top of the screen. On large screens, however, many windows may appear at the top."

	^ self == self class keyWindow! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 6/1/2016 14:46'!
isLookingFocused
	
	^ isLookingFocused ifNil: [false]! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 6/1/2016 14:45'!
isLookingFocused: aBoolean
	
	isLookingFocused := aBoolean.! !

!SystemWindow methodsFor: 'testing' stamp: 'jam 3/9/2003 15:13'!
isSystemWindow
"answer whatever the receiver is a SystemWindow"
	^ true! !

!SystemWindow methodsFor: 'testing' stamp: 'ar 8/15/2009 15:05'!
isWindowForModel: aModel
	"Return true if the receiver acts as the window for the given model"
	^aModel == self model! !

!SystemWindow methodsFor: 'geometry' stamp: 'mt 5/27/2016 14:46'!
justDroppedInto: aMorph event: anEvent

	isCollapsed
		ifTrue: [self position: ((self position max: 0@0) grid: 8@8).
				collapsedFrame := self bounds]
		ifFalse: [fullFrame := self bounds].

	self beKeyWindow.
	self hasDropShadow: true. "See #startDragFromLabel:."
			
	aMorph == self world ifTrue: [self assureLabelAreaVisible].

	(ToolBuilder openToolsAttachedToMouseCursor and: (self hasProperty: #initialDrop))
		ifTrue: [
			self removeProperty: #initialDrop.
			(self submorphs detect: [:m | m isKindOf: BottomRightGripMorph] ifNone: [])
				ifNotNil: [:grip | 
					grip
						referencePoint: anEvent position;
						setProperty: #targetHadDropShadow toValue: true "See MorphicToolBuilder >> #open:".
					self hasDropShadow: false.
					anEvent hand newMouseFocus: grip]].
			
	^super justDroppedInto: aMorph event: anEvent! !

!SystemWindow methodsFor: 'label' stamp: 'mt 4/2/2015 11:23'!
knownName

	^ self label! !

!SystemWindow methodsFor: 'label' stamp: 'di 5/4/1998 23:42'!
label
	^ labelString! !

!SystemWindow methodsFor: 'label' stamp: 'cmm 4/21/2015 15:25'!
labelArea
	^ labelArea! !

!SystemWindow methodsFor: 'label' stamp: 'jrp 8/6/2005 23:49'!
labelHeight
	"Answer the height for the window label.  The standard behavior is at bottom; a hook is provided so that models can stipulate other heights, in support of various less-window-looking demos."

	| aHeight |
	(model notNil and: [model respondsTo: #desiredWindowLabelHeightIn:]) ifTrue:
		[(aHeight := model desiredWindowLabelHeightIn: self) ifNotNil: [^ aHeight]].

	^ label ifNil: [0] ifNotNil:
		 [(label height + (self class borderWidth * 2)) max:
			(collapseBox ifNotNil: [collapseBox height] ifNil: [10])]! !

!SystemWindow methodsFor: 'geometry' stamp: 'di 6/16/1998 07:56'!
labelRect
	^ self innerBounds withHeight: self labelHeight.
! !

!SystemWindow methodsFor: 'label' stamp: 'sw 9/29/1999 07:22'!
labelWidgetAllowance
	^ labelWidgetAllowance ifNil: [self setLabelWidgetAllowance]! !

!SystemWindow methodsFor: 'layout' stamp: 'JW 1/30/2001 22:45'!
layoutBounds
	"Bounds of pane area only."
	| box |

	box := super layoutBounds.
	^box withTop: box top + self labelHeight! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 5/27/2016 14:14'!
lockWindowDecorations
	"Lock all window decrations, that is grips, splitters, and title bar."
	
	self windowDecorations do: [:m | m lock].! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 5/27/2016 14:41'!
lockWindowTitle
	
	labelArea ifNotNil: [:m | m lock].! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 7/9/2016 11:42'!
lookFocused

	"Optimize performance."
	self isLookingFocused ifTrue: [^ self].
	self isLookingFocused: true.

	label ifNotNil: [ label color: (self userInterfaceTheme focusedLabelColor ifNil: [Color black]) ].
	
	(self isKeyWindow or: [self class windowTitleActiveOnFirstClick])
		ifTrue: [self undimWindowButtons].
		
	self paneColorToUse in: [ : col |
		self
			 setStripeColorsFrom: col ;
			 adoptPaneColor: col].! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 7/9/2016 11:43'!
lookUnfocused

	"Optimize performance."
	self isLookingFocused ifFalse: [^ self].
	self isLookingFocused: false.

	label ifNotNil: [ label color: (self userInterfaceTheme unfocusedLabelColor ifNil: [Color darkGray]) ].
	self dimWindowButtons.
	self paneColorToUseWhenNotActive in: [ : col |
		self
			 setStripeColorsFrom: col ;
			 adoptPaneColor: col ]! !

!SystemWindow methodsFor: 'menu' stamp: 'laza 2/26/2010 20:41'!
makeClosable
	mustNotClose := false.
	closeBox ifNil: [self addCloseBox]! !

!SystemWindow methodsFor: 'drawing' stamp: 'RAA 6/12/2000 18:16'!
makeMeVisible 

	self world extent > (0@0) ifFalse: [^ self].

	((self world bounds insetBy: (0@0 corner: self labelHeight asPoint))
		containsPoint: self position) ifTrue: [^ self "OK -- at least my top left is visible"].

	"window not on screen (probably due to reframe) -- move it now"
	self isCollapsed
		ifTrue: [self position: (RealEstateAgent assignCollapsePointFor: self)]
		ifFalse: [self position: (RealEstateAgent initialFrameFor: self initialExtent: self extent world: self world) topLeft].

! !

!SystemWindow methodsFor: 'menu' stamp: 'mt 5/17/2016 16:02'!
makeSecondTopmost

	| aWorld nextWindow |
	aWorld := self world.
	nextWindow := aWorld submorphs 
				detect: [:m | (m isSystemWindow) and: [m ~~ self]]
				ifNone: [^self].

	self comeToFront.
	nextWindow beKeyWindow.! !

!SystemWindow methodsFor: 'menu' stamp: 'sw 8/7/2000 17:35'!
makeUnclosable
	mustNotClose := true.
	self deleteCloseBox! !

!SystemWindow methodsFor: 'initialization' stamp: 'jlb 5/29/2001 23:24'!
maximumExtent
	"This returns the maximum extent that the morph may be expanded to.
	Return nil if this property has not been set."

	^ self valueOfProperty: #maximumExtent! !

!SystemWindow methodsFor: 'initialization' stamp: 'jlb 5/29/2001 23:24'!
maximumExtent: aPoint
	"This returns the maximum extent that the morph may be expanded to.
	Return nil if this property has not been set."

	^ self setProperty: #maximumExtent toValue: aPoint! !

!SystemWindow methodsFor: 'polymorph' stamp: 'gvc 1/16/2007 12:24'!
modalChild
	"Answer the modal child of the receiver, if any."

	^self valueOfProperty: #modalChild! !

!SystemWindow methodsFor: 'polymorph' stamp: 'gvc 1/16/2007 12:27'!
modalLockTo: aSystemWindow
	"Lock the receiver as a modal owner of the given window."

	aSystemWindow
		setProperty: #modalOwner toValue: self.
	self setProperty: #modalChild toValue: aSystemWindow! !

!SystemWindow methodsFor: 'polymorph' stamp: 'gvc 1/16/2007 12:24'!
modalOwner
	"Answer the modal owner of the receiver, if any."

	^self valueOfProperty: #modalOwner! !

!SystemWindow methodsFor: 'polymorph' stamp: 'mt 5/17/2016 16:02'!
modalUnlockFrom: aSystemWindow
	"Unlock the receiver as a modal owner of the given window."

	aSystemWindow removeProperty: #modalOwner.
	self removeProperty: #modalChild.
	self beKeyWindow.! !

!SystemWindow methodsFor: 'initialization' stamp: 'ar 8/15/2001 22:16'!
model: anObject
	super model: anObject.
	self paneColor: nil.! !

!SystemWindow methodsFor: 'events' stamp: 'mt 5/26/2016 17:23'!
mouseDown: evt

	evt hand 
		waitForClicksOrDrag: self 
		event: evt 
		selectors: { nil. nil. nil. #startDragFromLabel: }
		threshold: HandMorph dragThreshold.! !

!SystemWindow methodsFor: 'events' stamp: 'mt 5/26/2016 17:10'!
mouseEnterDragging: evt
	"unlock children for drop operations"
	
	self flag: #performance. "mt: There may be no need to change appearance if no widget wants the drop."
	self isActive ifTrue: [self lookFocused].
		
	(self isActive not and: [evt hand hasSubmorphs]) ifTrue: [
		self activate. "unlock contents for drop"
		evt hand addMouseListener: self. "for drop completion on submorph"
	].! !

!SystemWindow methodsFor: 'events' stamp: 'mt 6/1/2016 13:29'!
mouseEnter: anEvent 
	"Handle a mouseEnter event, meaning the mouse just entered my bounds with no button pressed. The default response is to let my eventHandler, if any, handle it."
	super mouseEnter: anEvent.

	(self isActive and: [Preferences mouseOverForKeyboardFocus])
		ifTrue: [self lookFocused].! !

!SystemWindow methodsFor: 'events' stamp: 'mt 6/1/2016 13:48'!
mouseLeaveDragging: evt
	"Passivate after drop operations if needed. Unfortunately, we get a leave-dragging event for the system window we want to drag. Watch out for it."
	
	((model windowActiveOnFirstClick and: [(evt hand submorphs includes: self) not])
		or: [self isKeyWindow not])
			ifTrue: [self lookUnfocused].
	
	(self isKeyWindow not and: [evt hand hasSubmorphs]) ifTrue:[
		self passivateIfNeeded.
		evt hand removeMouseListener: self. "no more drop completion possible on submorph"
	].! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'jm 6/17/1998 11:55'!
mouseLeaveEvent: event fromPane: pane
	"For backward compatibility only.  Not used by any newly created window"
	(pane isKindOf: ScrollPane) ifTrue: [pane mouseLeave: event].
! !

!SystemWindow methodsFor: 'events' stamp: 'mt 6/1/2016 14:31'!
mouseLeave: anEvent 
	"Handle a mouseEnter event, meaning the mouse just entered my bounds with no button pressed. The default response is to let my eventHandler, if any, handle it."
	super mouseLeave: anEvent.
	
	(model windowActiveOnFirstClick and: [Preferences mouseOverForKeyboardFocus
		or: [anEvent hand keyboardFocus notNil and: [anEvent hand keyboardFocus containingWindow ~= self]]])
			ifTrue: [self lookUnfocused].! !

!SystemWindow methodsFor: 'open/close' stamp: 'sw 9/28/1999 13:32'!
mustNotClose
	^ mustNotClose == true! !

!SystemWindow methodsFor: 'menu' stamp: 'RAA 6/12/2000 09:01'!
offerWindowMenu
	| aMenu |
	aMenu := self buildWindowMenu.
	model ifNotNil:
		[model addModelItemsToWindowMenu: aMenu].
	aMenu popUpEvent: self currentEvent in: self world! !

!SystemWindow methodsFor: 'open/close' stamp: 'ar 5/11/2001 23:46'!
openAsIs
	^self openAsIsIn: self currentWorld
! !

!SystemWindow methodsFor: 'open/close' stamp: 'mt 6/1/2016 14:25'!
openAsIsIn: aWorld
	"This msg and its callees result in the window being activeOnlyOnTop"
	aWorld addMorph: self.
	self beKeyWindow.
	aWorld startSteppingSubmorphsOf: self.! !

!SystemWindow methodsFor: 'open/close' stamp: 'kfr 4/19/2015 08:19'!
openInWorldExtent: extent
	"This msg and its callees result in the window being activeOnlyOnTop"
	^ self openInWorld: self currentWorld extent: extent! !

!SystemWindow methodsFor: 'open/close' stamp: 'mt 6/1/2016 14:25'!
openInWorld: aWorld
	"This msg and its callees result in the window being activeOnlyOnTop"
	^ self anyOpenWindowLikeMe
		ifEmpty: 
			[ self 
				bounds: (RealEstateAgent initialFrameFor: self world: aWorld) ;
				openAsIsIn: aWorld ]
		ifNotEmptyDo:
			[ : windows | 
			windows anyOne
				expand ;
				beKeyWindow ; 
				postAcceptBrowseFor: self ].! !

!SystemWindow methodsFor: 'open/close' stamp: 'mt 6/1/2016 14:25'!
openInWorld: aWorld extent: extent
	"This msg and its callees result in the window being activeOnlyOnTop"
	^ self anyOpenWindowLikeMe
		ifEmpty:
			[ self 
				position: (RealEstateAgent initialFrameFor: self initialExtent: extent world: aWorld) topLeft ;
				extent: extent.
			self openAsIsIn: aWorld ]
		ifNotEmptyDo:
			[ : windows | 
			windows anyOne
				expand ;
				beKeyWindow ; 
				postAcceptBrowseFor: self ].! !

!SystemWindow methodsFor: 'colors' stamp: 'mt 8/5/2016 16:08'!
paneColor
	| cc |
	(cc := self valueOfProperty: #paneColor) ifNotNil: [^cc].
	(model respondsTo: #windowColorToUse) ifTrue: [cc := model windowColorToUse].
	cc ifNil: [cc := paneMorphs isEmptyOrNil ifFalse: [paneMorphs first color]].
	cc ifNil: [cc := self defaultColor].
	self paneColor: cc.
	^cc! !

!SystemWindow methodsFor: 'colors' stamp: 'ar 8/15/2001 22:14'!
paneColorToUse
	^ Display depth <= 2
		ifTrue:
			[Color white]
		ifFalse:
			[self paneColor]! !

!SystemWindow methodsFor: 'colors' stamp: 'mt 7/9/2016 11:41'!
paneColorToUseWhenNotActive

	^ (self userInterfaceTheme unfocusedWindowColorModifier ifNil: [ [:color | color darker] ])
		value: self paneColorToUse! !

!SystemWindow methodsFor: 'colors' stamp: 'bvs 3/17/2004 09:15'!
paneColor: aColor
	self setProperty: #paneColor toValue: aColor.

	self adoptPaneColor: aColor.! !

!SystemWindow methodsFor: 'geometry' stamp: 'sma 2/5/2000 14:09'!
panelRect
	"Answer the area below the title bar which is devoted to panes."

	^ self innerBounds insetBy: (0 @ self labelHeight corner: 0 @ 0)! !

!SystemWindow methodsFor: 'panes' stamp: 'ul 8/2/2011 20:11'!
paneMorphSatisfying: aBlock

	^paneMorphs detect: aBlock ifNone: [ nil ]! !

!SystemWindow methodsFor: 'geometry' stamp: 'mt 5/19/2016 07:38'!
paneMorphs
	"Nominally private but a need for obtaining this from the outside arose"
	^ paneMorphs! !

!SystemWindow methodsFor: 'events' stamp: 'jmv 2/19/2006 14:47'!
paneTransition: event
	"Mouse has entered or left a pane"! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'nice 12/27/2009 20:19'!
paneWithLongestSide: sideBlock near: aPoint 
	| thePane theSide theLen |
	theLen := 0.
	paneMorphs do:
		[:pane | | box |
		box := pane bounds.
		box forPoint: aPoint closestSideDistLen:
			[:side :dist :len |
			(dist <= 5 and: [len > theLen]) ifTrue:
				[thePane := pane.
				theSide := side.
				theLen := len]]].
	sideBlock value: theSide.
	^ thePane! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 6/1/2016 15:25'!
passivate
	"Reconfigure my focus according to preferences."

	self isActive ifFalse: [^ self].
	self isActive: false.
	
	self isCollapsed ifFalse: [model modelSleep].

	self submorphsDo: [:each | each lock].

	self activeHand keyboardFocus ifNotNil: [:morph |
		morph containingWindow == self
			ifTrue: [self activeHand releaseKeyboardFocus]].

	self lookUnfocused.! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 5/27/2016 14:52'!
passivateIfNeeded

	model windowActiveOnFirstClick
		ifFalse: [self passivate]
		ifTrue: [self lookUnfocused].
	
	self unlockWindowDecorations.
	
	self class windowTitleActiveOnFirstClick
		ifFalse: [self lockWindowTitle]
		ifTrue: [self unlockWindowTitle].! !

!SystemWindow methodsFor: 'geometry' stamp: 'di 5/22/1998 13:24'!
position: newPos
	super position: newPos.
	isCollapsed
		ifTrue: [collapsedFrame := self bounds]
		ifFalse: [fullFrame := self bounds].
! !

!SystemWindow methodsFor: 'open/close' stamp: 'cmm 2/13/2010 20:07'!
postAcceptBrowseFor: anotherSystemWindow
	"If I am taking over browsing for anotherSystemWindow, sucblasses may override to, for example, position me to the object to be focused on."
	self model postAcceptBrowseFor: anotherSystemWindow model! !

!SystemWindow methodsFor: 'layout' stamp: 'sw 5/18/2001 16:09'!
putLabelItemsInLabelArea
	"Put label items into the label area, if there is one"

	labelArea ifNotNil:
		[stripes ifNotNil: [stripes do: [:stripe | labelArea addMorph: stripe]].
		closeBox ifNotNil: [labelArea addMorph: closeBox].
		menuBox ifNotNil: [labelArea addMorph: menuBox].
		collapseBox ifNotNil: [labelArea addMorph: collapseBox].
		label ifNotNil: [labelArea addMorph: label]]

! !

!SystemWindow methodsFor: 'colors' stamp: 'ar 8/16/2001 12:47'!
raisedColor
	^self paneColor raisedColor! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'di 10/22/1998 22:55'!
reframePanesAdjoining: growingPane along: side to: aDisplayBox 
	| delta newRect minDim theMin horiz |
	growingPane ifNil: [^ self].  "As from click outside"
	newRect := aDisplayBox.
	horiz := #(left right) includes: side.
	theMin := horiz ifTrue: [40] ifFalse: [20].

	"First check that this won't make any pane smaller than theMin screen dots"
	minDim := (((paneMorphs select: [:pane | pane bounds bordersOn: growingPane bounds along: side])
		collect: [:pane | pane bounds adjustTo: newRect along: side]) copyWith: aDisplayBox)
			inject: 999 into:
				[:was :rect | was min: (horiz ifTrue: [rect width] ifFalse: [rect height])].
	"If so, amend newRect as required"
	minDim > theMin ifFalse:
		[delta := minDim - theMin.
		newRect := newRect withSide: side setTo: 
				((newRect perform: side) > (growingPane bounds perform: side)
					ifTrue: [(newRect perform: side) + delta]
					ifFalse: [(newRect perform: side) - delta])].

	"Now adjust all adjoining panes for real"
	paneMorphs do:
		[:pane | (pane bounds bordersOn: growingPane bounds along: side) ifTrue:
			[pane bounds: (pane bounds adjustTo: newRect along: side)]].
	"And adjust the growing pane itself"
	growingPane bounds: newRect.

	"Finally force a recomposition of the whole window"
	self setPaneRectsFromBounds.
	self extent: self extent! !

!SystemWindow methodsFor: 'colors' stamp: 'mt 7/28/2016 13:14'!
refreshWindowColor
	"For changing the underlying model's default window color"
	self setProperty: #paneColor toValue: nil.
	self setWindowColor: self paneColor.

	self isLookingFocused
		ifTrue: [self lookUnfocused; lookFocused]
		ifFalse: [self lookFocused; lookUnfocused].! !

!SystemWindow methodsFor: 'label' stamp: 'rbb 3/1/2005 11:17'!
relabel
	| newLabel |
	newLabel := UIManager default 
		request: 'New title for this window' translated
		initialAnswer: labelString.
	newLabel isEmpty ifTrue: [^self].
	(model windowReqNewLabel: newLabel)
		ifTrue: [self setLabel: newLabel]! !

!SystemWindow methodsFor: 'label' stamp: 'di 12/1/2001 23:19'!
relabelEvent: evt
	"No longer used, but may be referred to by old eventHandlers"

	^ Preferences clickOnLabelToEdit
		ifFalse:	[self mouseDown: evt]
		ifTrue:	[self relabel]! !

!SystemWindow methodsFor: 'polymorph' stamp: 'gvc 9/4/2007 13:42'!
rememberedKeyboardFocus
	"Answer the remembered keyboard focus for the receiver."
	
	^self valueOfProperty: #rememberedFocus! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 5/31/2016 12:07'!
removeKeyboardShortcuts
	
	self removeKeyboardCaptureFilter: self.! !

!SystemWindow methodsFor: 'geometry' stamp: 'sw 9/28/1999 11:02'!
removeMenuBox
	menuBox ifNotNil:
		[menuBox delete.
		menuBox := nil].
! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 5/19/2016 08:03'!
replaceBoxes
	"Rebuild the various boxes."
	self setLabelWidgetAllowance.
	label ifNotNil: [label delete].
	labelArea ifNotNil: [labelArea delete].
	self initializeLabelArea.
	self setFramesForLabelArea.
	self setWindowColor: self paneColor.
	self isActive
		ifTrue: [self passivate; activate]
		ifFalse: [self activate; passivate].! !

!SystemWindow methodsFor: 'panes' stamp: 'md 2/24/2006 15:46'!
replacePane: oldPane with: newPane
	"Make newPane exactly occupy the position and extent of oldPane"

	| aLayoutFrame hadDep |
	hadDep := model dependents includes: oldPane.
	oldPane owner replaceSubmorph: oldPane by: newPane.
	newPane
		position: oldPane position;
		extent: oldPane extent.
	aLayoutFrame := oldPane layoutFrame.
	paneMorphs := paneMorphs collect:
		[:each |
		each == oldPane ifTrue: [newPane] ifFalse: [each]].
	aLayoutFrame ifNotNil: [newPane layoutFrame: aLayoutFrame].
	newPane color: Color transparent.
	hadDep ifTrue: [model removeDependent: oldPane. model addDependent: newPane].

	self changed

! !

!SystemWindow methodsFor: 'colors' stamp: 'md 2/24/2006 15:46'!
restoreDefaultPaneColor
	"Useful when changing from monochrome to color display"

	self setStripeColorsFrom: self paneColor.! !

!SystemWindow methodsFor: 'events' stamp: 'jmv 2/19/2006 14:54'!
secondaryPaneTransition: event divider: aMorph
	"Mouse has entered or left a pane"! !

!SystemWindow methodsFor: 'menu' stamp: 'mt 5/17/2016 16:03'!
sendToBack
	| aWorld nextWindow |
	aWorld := self world.
	nextWindow := aWorld submorphs 
				detect: [:m | (m isSystemWindow) and: [m ~~ self]]
				ifNone: [^self].
	nextWindow beKeyWindow.
	aWorld addMorphNearBack: self! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'di 10/28/1999 14:15'!
setBoundsWithFlex: newFrame
	"Set bounds from newFrame with origin preserved from global coordinates"

	self isFlexed
		ifTrue: [super bounds: ((owner transform globalPointToLocal: newFrame topLeft)
										extent: newFrame extent)]
		ifFalse: [super bounds: newFrame].! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 7/25/2016 15:50'!
setDefaultParameters

	Preferences menuAppearance3d ifTrue: [
		self addDropShadow.
		self hasDropShadow: self isKeyWindow. "maybe turn off again"].
	
	self borderWidth: (self userInterfaceTheme borderWidth ifNil: [1]).
	label font: (self userInterfaceTheme titleFont ifNil: [TextStyle defaultFont]).! !

!SystemWindow methodsFor: 'initialization' stamp: 'mt 4/28/2015 14:46'!
setFramesForLabelArea
	"an aid to converting old instances, but then I found  
	convertAlignment (jesse welton's note)"
	| frame |
	labelArea ifNil: [^ self].	
	labelArea
		layoutPolicy: TableLayout new;
		listDirection: #leftToRight;
		hResizing: #spaceFill;
		layoutInset: 0.
	label hResizing: #spaceFill.
	labelArea
		ifNotNil: [frame := LayoutFrame new.
			frame leftFraction: 0;
				topFraction: 0;
				rightFraction: 1;
				bottomFraction: 0;
				topOffset: self labelHeight negated.
			labelArea layoutFrame: frame]! !

!SystemWindow methodsFor: 'label' stamp: 'dgd 3/21/2003 13:03'!
setLabelWidgetAllowance
	^ labelWidgetAllowance :=  (self boxExtent x * 4) + 19! !

!SystemWindow methodsFor: 'label' stamp: 'mt 4/2/2015 11:41'!
setLabel: aString
	| frame |
	labelString := aString.
	label ifNil: [^ self].
	label contents: (aString ifNil: ['']).
	self labelWidgetAllowance.  "Sets it if not already"
	self isCollapsed
		ifTrue: [self extent: (label width + labelWidgetAllowance) @ (self labelHeight + 2)]
		ifFalse: [label fitContents; setWidth: (label width min: bounds width - labelWidgetAllowance).
				label align: label bounds topCenter with: bounds topCenter + (0@borderWidth).
				collapsedFrame ifNotNil:
					[collapsedFrame := collapsedFrame withWidth: label width + labelWidgetAllowance]].
	frame := LayoutFrame new.
	frame leftFraction: 0.5;
		 topFraction: 0.5;
		 leftOffset: label width negated // 2;
		 topOffset: label height negated // 2.
	label layoutFrame: frame.
! !

!SystemWindow methodsFor: 'geometry' stamp: 'nice 12/27/2009 03:10'!
setPaneRectsFromBounds
	"Reset proportional specs from actual bounds, eg, after reframing panes"
	| layoutBounds |
	layoutBounds := self layoutBounds.
	paneMorphs do:[:m| | box left bottom top frame right |
		frame := m layoutFrame.
		box := m bounds.
		frame ifNotNil:[
			left := box left - layoutBounds left - (frame leftOffset ifNil:[0]).
			right := box right - layoutBounds left - (frame rightOffset ifNil:[0]).
			top := box top - layoutBounds top - (frame topOffset ifNil:[0]).
			bottom := box bottom - layoutBounds top - (frame bottomOffset ifNil:[0]).
			frame leftFraction: (left / layoutBounds width asFloat).
			frame rightFraction: (right / layoutBounds width asFloat).
			frame topFraction: (top / layoutBounds height asFloat).
			frame bottomFraction: (bottom / layoutBounds height asFloat).
		].
	].! !

!SystemWindow methodsFor: 'colors' stamp: 'mt 7/25/2016 15:45'!
setStripeColorsFrom: paneColor 
	"Set the stripe color based on the given paneColor"

	labelArea ifNotNil: [labelArea color: Color transparent].
	self updateBoxesColor: paneColor.
	stripes ifNil: [^self].

	self borderColor: ((self userInterfaceTheme borderColorModifier ifNil: [ [:c | c adjustBrightness: -0.3] ])
							value: paneColor).

	self class gradientWindow
		ifTrue: [self fillStyle: (self gradientWithColor: paneColor)]
		ifFalse: [self color: paneColor].! !

!SystemWindow methodsFor: 'panes' stamp: 'nice 12/27/2009 03:10'!
setUpdatablePanesFrom: getSelectors
	| aList possibles |
	"Set my updatablePanes inst var to the list of panes which are list panes with the given get-list selectors.  Order is important here!!  Note that the method is robust in the face of panes not found, but a warning is printed in the transcript in each such case"

	aList := OrderedCollection new.
	possibles := OrderedCollection new.
	self allMorphsDo: [ :pane | 
		(pane isKindOf: PluggableListMorph) ifTrue: [
			possibles add: pane.
		].
	].

	getSelectors do: [:sel | | aPane | 
		aPane := possibles detect: [ :pane | pane getListSelector == sel] ifNone: [nil].
		aPane
			ifNotNil:
				[aList add: aPane]
			ifNil:
				[Transcript cr; show: 'Warning: pane ', sel, ' not found.']].
	updatablePanes := aList asArray! !

!SystemWindow methodsFor: 'menu' stamp: 'cmm 12/4/2010 15:11'!
setWindowColor
	"Allow the user to select a new basic color for the window"
	NewColorPickerMorph useIt
		ifTrue:
			[ (NewColorPickerMorph
				on: self
				originalColor: self paneColorToUse
				setColorSelector: #setWindowColor:) openNear: self fullBounds ]
		ifFalse:
			[ ColorPickerMorph new
				 choseModalityFromPreference ;
				 sourceHand: self activeHand ;
				 target: self ;
				 selector: #setWindowColor: ;
				 originalColor: self paneColorToUse ;
				
				putUpFor: self
				near: self fullBounds ]! !

!SystemWindow methodsFor: 'colors' stamp: 'mt 4/6/2015 12:47'!
setWindowColor: incomingColor
	| existingColor aColor |
	incomingColor ifNil: [^ self].  "it happens"
	aColor := incomingColor.
	(aColor = ColorPickerMorph perniciousBorderColor 
		or: [aColor = Color black]) ifTrue: [^ self].
	existingColor := self paneColorToUse.
	existingColor ifNil: [^ Beeper beep].
	self paneColor: aColor.
	self setStripeColorsFrom: aColor.
	self changed.! !

!SystemWindow methodsFor: 'testing' stamp: 'ar 12/2/2001 21:43'!
shouldDropOnMouseUp
	"Return true for consistency with fastdrag"
	^true! !

!SystemWindow methodsFor: 'events' stamp: 'mt 4/5/2015 20:26'!
startDragFromLabel: evt
	"When label events are active, we need to pass dragging to the window explicitely
	 The window only recognizes a drag with an offset of more than 3 pixels"
	
	self isSticky ifTrue: [^ self].
	self fastFramingOn 
		ifTrue: [self doFastFrameDrag: evt cursorPoint]
		ifFalse: [
			self hasDropShadow: false.
			evt hand grabMorph: self topRendererOrSelf]
! !

!SystemWindow methodsFor: 'stepping' stamp: 'di 4/9/2001 17:04'!
stepAt: millisecondClockValue
	"If the receiver is not collapsed, step it, after first stepping the model."

	(isCollapsed not or: [self wantsStepsWhenCollapsed]) ifTrue:
		[model ifNotNil: [model stepAt: millisecondClockValue in: self].
		super stepAt: millisecondClockValue "let player, if any, step"]

"Since this method ends up calling step, the model-stepping logic should not be duplicated there."! !

!SystemWindow methodsFor: 'stepping' stamp: 'sw 10/19/1999 08:22'!
stepTime
	^ model
		ifNotNil:
			[model stepTimeIn: self]
		ifNil:
			[200] "milliseconds"! !

!SystemWindow methodsFor: 'menu' stamp: 'ul 11/15/2010 12:06'!
takeOutOfWindow
	"Take the receiver's pane morph out the window and place it, naked, where once the window was"
	| aMorph |
	paneMorphs size = 1 ifFalse: [^ Beeper beep].
	aMorph := paneMorphs first.
	owner addMorphFront: aMorph.
	self delete! !

!SystemWindow methodsFor: 'panes' stamp: 'ul 11/15/2010 12:06'!
titleAndPaneText
	"If the receiver represents a workspace, return an Association between the title and that text, else return nil"
	(paneMorphs size ~= 1 or: [(paneMorphs first isKindOf: PluggableTextMorph) not])
		ifTrue: [^ nil].
	^ labelString -> paneMorphs first text

! !

!SystemWindow methodsFor: 'label' stamp: 'fc 12/15/2004 13:33'!
tryToRenameTo: newLabel
	"Triggered eg by typing a new name in the halo"

	newLabel isEmpty ifTrue: [^self].
	(model windowReqNewLabel: newLabel)
		ifTrue: [self setLabel: newLabel]
	! !

!SystemWindow methodsFor: 'top window' stamp: 'bf 3/26/2010 10:34'!
undimWindowButtons
	{closeBox. collapseBox. menuBox. expandBox}
		do: [:b | b ifNotNil: [b undim]]! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'jlb 5/29/2001 23:06'!
unexpandedFrame
	"Return the frame size of an unexpanded window"

	^ self valueOfProperty: #unexpandedFrame! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'jlb 5/29/2001 23:07'!
unexpandedFrame: aRectangle
	"Set the frame size of an unexpanded window"

	^ self setProperty: #unexpandedFrame toValue: aRectangle! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 5/27/2016 14:15'!
unlockWindowDecorations
	"Unlock all window decrations, that is grips, splitters, and title bar."
	
	self windowDecorations do: [:m | m unlock].
	
	"Migrate old window instances. Can be removed in the future."
	labelArea ifNotNil: [:m | m submorphsDo: [:sm | sm unlock]].! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 5/27/2016 14:41'!
unlockWindowTitle
	
	labelArea ifNotNil: [:m | m unlock].! !

!SystemWindow methodsFor: 'panes' stamp: 'sw 10/19/1999 09:53'!
updatablePanes
	"Answer the list of panes, in order, which should be sent the #verifyContents message"
	^ updatablePanes ifNil: [updatablePanes := #()]! !

!SystemWindow methodsFor: 'panes' stamp: 'md 2/24/2006 15:36'!
updateBoxesColor: aColor 
	| opaqueColor |
	aColor isNil
		ifTrue: [^ self].
	Preferences alternativeWindowBoxesLook ifFalse:[^ self].
	
	opaqueColor := aColor asNontranslucentColor.
	
	self
		updateBox: closeBox
		color: (opaqueColor alphaMixed: 0.5 with: Color red).
	self updateBox: menuBox color: opaqueColor.
	self updateBox: expandBox color: opaqueColor.
	self updateBox: collapseBox color: opaqueColor! !

!SystemWindow methodsFor: 'panes' stamp: 'mt 2/24/2015 08:52'!
updateBox: anIconMorph color: aColor 

	| fill |
	anIconMorph isNil
		ifTrue: [^ self].
	anIconMorph
		extent: self boxExtent;
		useRoundedCorners.

	SystemWindow gradientWindow
		ifFalse: [anIconMorph color: aColor]
		ifTrue: [	
			fill := GradientFillStyle ramp: {
				0.0 -> aColor muchLighter muchLighter.
				1.0 -> aColor twiceDarker}.
			fill origin: anIconMorph topLeft + (5 @ 5).
			fill direction: anIconMorph extent.
			anIconMorph fillStyle: fill].

	anIconMorph
		borderWidth: (Preferences alternativeWindowBoxesLook
			ifTrue: [1]
			ifFalse: [0]);
		borderColor: aColor darker.! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 6/1/2016 15:19'!
updateFocusLookAtHand
	"If there is more than one active window, look for the mouse cursor and update the window focus look accordingly. This method is not on the class-side because we need our world and some active hand."
	
	(model windowActiveOnFirstClick and: [Preferences mouseOverForKeyboardFocus])
		ifFalse: [self updateFocusLookForKeyboardFocus]
		 ifTrue: [
			((self class windowsIn: self world)
				do: [:window | window lookUnfocused];
				select: [:window | window bounds containsPoint: self activeHand position])
					ifNotEmpty: [:windowsPointed | windowsPointed first lookFocused "only to foremost window"]].! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 6/9/2016 17:46'!
updateFocusLookForKeyboardFocus

	| f w |
	(((f := self activeHand keyboardFocus) notNil and: [(w := f containingWindow) notNil])
		and: [w isActive])
			ifTrue: [
				(self class windowsIn: self world) do: [:window |
					w ~~ window ifTrue: [window lookUnfocused]].
				w lookFocused]! !

!SystemWindow methodsFor: 'colors' stamp: 'md 2/24/2006 15:36'!
updatePaneColors
	"Useful when changing from monochrome to color display"

	self setStripeColorsFrom: self paneColorToUse.! !

!SystemWindow methodsFor: 'top window' stamp: 'nk 8/6/2003 10:53'!
updatePanesFromSubmorphs
	"Having removed some submorphs, make sure this is reflected in my paneMorphs."
	paneMorphs := paneMorphs select: [ :pane | submorphs includes: pane ].! !

!SystemWindow methodsFor: 'label' stamp: 'BG 8/27/2003 19:07'!
update: aSymbol
	aSymbol = #relabel
		ifTrue: [^ model ifNotNil: [ self setLabel: model labelString ] ].
	aSymbol = #close
		ifTrue: [self delete]! !

!SystemWindow methodsFor: 'resize/collapse' stamp: 'tfel 2/27/2010 20:20'!
wantsExpandBox
	"Answer whether I'd like an expand box"

	^ Preferences alwaysHideExpandButton not! !

!SystemWindow methodsFor: 'events' stamp: 'sw 12/22/1999 18:31'!
wantsHalo
	^ false! !

!SystemWindow methodsFor: 'drawing' stamp: 'mt 7/28/2016 11:23'!
wantsRoundedCorners
	^ self class roundedWindowCorners or: [super wantsRoundedCorners]! !

!SystemWindow methodsFor: 'stepping' stamp: 'jcg 9/6/2000 23:01'!
wantsSteps
	"Return true if the model wants its view to be stepped.  For an open system window, we give the model to offer an opinion"

	self isPartsDonor ifTrue: [^ false].
	self player wantsSteps ifTrue: [^ true].
	^ isCollapsed not and: [model wantsStepsIn: self]! !

!SystemWindow methodsFor: 'stepping' stamp: 'di 4/9/2001 16:45'!
wantsStepsWhenCollapsed
	"Default is not to bother updating collapsed windows"

	^ false! !

!SystemWindow methodsFor: 'testing' stamp: 'ar 6/23/2001 16:06'!
wantsToBeCachedByHand
	"Return true if the receiver wants to be cached by the hand when it is dragged around."
	self hasTranslucentColor ifTrue:[^false].
	self bounds = self fullBounds ifTrue:[^true].
	self submorphsDo:[:m|
		(self bounds containsRect: m fullBounds) ifFalse:[
			m wantsToBeCachedByHand ifFalse:[^false].
		].
	].
	^true! !

!SystemWindow methodsFor: 'events' stamp: 'ar 9/18/2000 18:34'!
wantsToBeDroppedInto: aMorph
	"Return true if it's okay to drop the receiver into aMorph"
	^aMorph isWorldMorph or:[Preferences systemWindowEmbedOK]! !

!SystemWindow methodsFor: 'menu' stamp: 'dgd 9/18/2004 18:27'!
wantsYellowButtonMenu
	"Answer true if the receiver wants a yellow button menu"
	^ false! !

!SystemWindow methodsFor: 'label' stamp: 'jmv 8/6/2009 08:57'!
widthOfFullLabelText
	^Preferences windowTitleFont widthOfString: labelString! !

!SystemWindow methodsFor: 'focus' stamp: 'mt 5/27/2016 14:14'!
windowDecorations
	
	^ self submorphs copyWithoutAll: self paneMorphs! !

BorderedMorph subclass: #PasteUpMorph
	instanceVariableNames: 'presenter model cursor padding backgroundMorph turtleTrailsForm turtlePen lastTurtlePositions isPartsBin indicateCursor wantsMouseOverHalos worldState griddingOn'
	classVariableNames: 'DisableDeferredUpdates GlobalCommandKeysEnabled MinCycleLapse StillAlive WindowEventHandler'
	poolDictionaries: ''
	category: 'Morphic-Worlds'!

!PasteUpMorph commentStamp: '<historical>' prior: 0!
A morph whose submorphs comprise a paste-up of rectangular subparts which "show through".  Anything called a 'Playfield' is a PasteUpMorph.

Facilities commonly needed on pages of graphical presentations and on simulation playfields, such as the painting of new objects, turtle trails, gradient fills, background paintings, parts-bin behavior, collision-detection, etc., are (or will be) provided.

A World, the entire Smalltalk screen, is a PasteUpMorph.  A World responds true to isWorld.  Morph subclasses that have specialized menus (BookMorph) build them in the message addBookMenuItemsTo:hand:.  A PasteUpMorph that is a world, builds its menu in HandMorph buildWorldMenu.

presenter	A Presenter in charge of stopButton stepButton and goButton, 
			mouseOverHalosEnabled soundsEnabled fenceEnabled coloredTilesEnabled.
model		<not used>
cursor		??
padding		??
backgroundMorph		A Form that covers the background.
turtleTrailsForm			Moving submorphs may leave trails on this form.
turtlePen				Draws the trails.
lastTurtlePositions		A Dictionary of (aPlayer -> aPoint) so turtle trails can be drawn 
						only once each step cycle.  The point is the start of the current stroke.
isPartsBin		If true, every object dragged out is copied.
autoLineLayout		??
indicateCursor		??
resizeToFit		??
wantsMouseOverHalos		If true, simply moving the cursor over a submorph brings up its halo.
worldState		If I am also a World, keeps the hands, damageRecorder, stepList etc.
griddingOn		If true, submorphs are on a grid

!

!PasteUpMorph class methodsFor: 'scripting' stamp: 'yo 5/5/2007 10:38'!
authoringPrototype
	"Answer an instance of the receiver suitable for placing in a parts bin for authors"
	
	| proto |
	proto _ self new markAsPartsDonor.
	proto color: Color green muchLighter;  extent: 100 @ 80; borderColor: (Color r: 0.645 g: 0.935 b: 0.161).
	proto extent: 300 @ 240.
	proto wantsMouseOverHalos: false.
	proto beSticky.
	^ proto! !

!PasteUpMorph class methodsFor: 'printing' stamp: 'KR 5/23/2008 01:29'!
defaultNameStemForInstances
	"Answer a basis for names of default instances of the receiver"
	^ 'playfield' translatedNoop! !

!PasteUpMorph class methodsFor: 'parts bin' stamp: 'sw 8/2/2001 16:48'!
descriptionForPartsBin
	^ self partName:	'Playfield'
		categories:		#('Presentation')
		documentation:	'A place for assembling parts or for staging animations'! !

!PasteUpMorph class methodsFor: 'project' stamp: 'RAA 5/25/2000 15:26'!
disableDeferredUpdates

	^DisableDeferredUpdates ifNil: [DisableDeferredUpdates := false]
! !

!PasteUpMorph class methodsFor: 'project' stamp: 'di 7/15/1999 09:51'!
disableDeferredUpdates: aBoolean
	"If the argument is true, disable deferred screen updating."
	"Details: When deferred updating is used, Morphic performs double-buffered screen updates by telling the VM to de-couple the Display from the hardware display buffer, drawing directly into the Display, and then forcing the changed regions of the Display to be copied to the screen. This saves both time (an extra BitBlt is avoided) and space (an extra display buffer is avoided). However, on platforms on which the Display points directly to the hardware screen buffer, deferred updating can't be used (you'd see ugly flashing as the layers of the drawing were assembled). In this case, the drawing is composited into an offscreen FormCanvas  and then copied to the hardware display buffer."

	DisableDeferredUpdates := aBoolean.
! !

!PasteUpMorph class methodsFor: 'preferences' stamp: 'cmm 10/17/2015 18:54'!
globalCommandKeysEnabled
	<preference: 'Enable Global Command Keys'
		category: 'Morphic'
		description: 'When true, the global command keys are enabled.  Command-key help is available from the help menu.'
		type: #Boolean>
	^ GlobalCommandKeysEnabled ifNil: [ true ]! !

!PasteUpMorph class methodsFor: 'preferences' stamp: 'mt 5/31/2016 12:09'!
globalCommandKeysEnabled: aBoolean

	GlobalCommandKeysEnabled = aBoolean ifTrue: [^ self].
	GlobalCommandKeysEnabled := aBoolean.
	
	SystemWindow allSubInstancesDo: [:ea |
		aBoolean
			ifTrue: [ea addKeyboardShortcuts]
			ifFalse: [ea removeKeyboardShortcuts]].

	PasteUpMorph allSubInstancesDo: [:ea |
		aBoolean
			ifTrue: [ea addKeyboardShortcuts]
			ifFalse: [ea removeKeyboardShortcuts]].! !

!PasteUpMorph class methodsFor: 'project' stamp: 'di 7/15/1999 09:51'!
MinCycleLapse: milliseconds
	"set the minimum amount of time that may transpire between two calls to doOneCycle"
	MinCycleLapse := milliseconds ifNotNil: [ milliseconds rounded ].! !

!PasteUpMorph class methodsFor: 'project' stamp: 'RAA 5/25/2000 15:15'!
newWorldForProject: projectOrNil 
	"Return a new pasteUpMorph configured as a world (ie project notNil).
	projectOrNil is no longer used."

	^ self new initForProject: WorldState new! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 9/28/2000 19:25'!
abandonAllHalos
	self flag: #arNote. "Remove the method"
	^self deleteAllHalos! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 9/9/1998 10:45'!
abandonCostumeHistory
	self allMorphsDo:
		[:m | m player ifNotNil: [m player forgetOtherCostumes]]! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'kfr 3/16/2016 20:26'!
acceptDroppingMorph: dropped event: evt
	"The supplied morph, known to be acceptable to the receiver, is now to be assimilated; the precipitating event is supplied"

	| aMorph |
	aMorph := self morphToDropFrom: dropped.
	self isWorldMorph
		ifTrue:["Add the given morph to this world and start stepping it if it wants to be."
				self addMorphFront: aMorph.
				(aMorph fullBounds intersects: self viewBox) ifFalse:
					[Beeper beep.  aMorph position: self bounds center]]
		ifFalse:[super acceptDroppingMorph: aMorph event: evt].

	aMorph submorphsDo: [:m | (m isKindOf: HaloMorph) ifTrue: [m delete]].
	aMorph allMorphsDo:  "Establish any penDown morphs in new world"
		[:m | | tfm mm |
		m player ifNotNil:
			[m player getPenDown ifTrue:
				[((mm := m player costume) notNil and: [(tfm := mm owner transformFrom: self) notNil])
					ifTrue: [self noteNewLocation: (tfm localPointToGlobal: mm referencePosition)
									forPlayer: m player]]]].

	self isPartsBin
		ifTrue:
			[aMorph isPartsDonor: true.
			aMorph stopSteppingSelfAndSubmorphs.
			aMorph suspendEventHandler]
		ifFalse:
			[self world startSteppingSubmorphsOf: aMorph].

"	self presenter morph: aMorph droppedIntoPasteUpMorph: self."
	self griddingOn ifTrue: [aMorph position: (self gridPoint: aMorph position)].
	self showingListView ifTrue:
		[self sortSubmorphsBy: (self valueOfProperty: #sortOrder).
		self currentWorld abandonAllHalos].

	self bringTopmostsToFront.
! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'sw 2/15/1999 20:37'!
accommodateFlap: aFlapTab
	"Shift submorphs over, if appropriate"
	| offset |
	aFlapTab slidesOtherObjects ifTrue:
		[offset := self offsetForAccommodating: aFlapTab referent extent onEdge: aFlapTab edgeToAdhereTo.
		self shiftSubmorphsBy: offset]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 8/12/2001 17:47'!
activateObjectsTool
	"Offer the user a parts bin of morphs -- if one already exists, bring it to the front and flash its border beckoningly; if none exists yet, create a new one and place it in the center of the screen"

	| anObjectTool |
	submorphs do:
		[:aMorph | (aMorph renderedMorph isKindOf: ObjectsTool)
			ifTrue:
				[aMorph comeToFront.
				aMorph flash.
				^ self]].
	"None found, so create one"

	anObjectTool := ObjectsTool newStandAlone.
	self addMorphFront: anObjectTool.
	anObjectTool fullBounds.
	anObjectTool center: self center

	"ActiveWorld activateObjectsTool"! !

!PasteUpMorph methodsFor: 'structure' stamp: 'mt 5/1/2016 18:11'!
activeHand

	^ worldState
		ifNotNil: [:ws | ws activeHand ifNil: [ws hands first]]
		ifNil: [super activeHand]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 5/24/2000 22:29'!
activeHand: aHandMorph
	"temporarily retained for old main event loops"

	worldState activeHand: aHandMorph.

! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'ar 3/17/2001 20:12'!
adaptedToWorld: aWorld
	"If I refer to a world or a hand, return the corresponding items in the new world."
	self isWorldMorph ifTrue:[^aWorld].! !

!PasteUpMorph methodsFor: 'alarms-scheduler' stamp: 'ar 9/11/2000 16:40'!
addAlarm: aSelector withArguments: argArray for: aTarget at: scheduledTime
	"Add a new alarm with the given set of parameters"
	worldState addAlarm: aSelector withArguments: argArray for: aTarget at: scheduledTime.! !

!PasteUpMorph methodsFor: 'submorphs-add/remove' stamp: 'di 7/15/1999 09:51'!
addAllMorphs: array

	super addAllMorphs: array.
	self isWorldMorph
		ifTrue: [array do: [:m | self startSteppingSubmorphsOf: m]].
! !

!PasteUpMorph methodsFor: 'layout' stamp: 'sw 3/24/1999 14:11'!
addCenteredAtBottom: aMorph offset: anOffset
	"Add aMorph beneath all other morphs currently in the receiver, centered horizontally, with the vertical offset from the bottom of the previous morph given by anOffset"
	| curBot |
	curBot := 0.
	submorphs do: [:m | curBot := curBot max: m bottom].
	self addMorphBack: aMorph.
	aMorph position: ((self center x - (aMorph width // 2)) @ (curBot + anOffset))! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'tbn 1/4/2010 01:02'!
addCustomMenuItems: menu hand: aHandMorph 
	"Add morph-specific menu itemns to the menu for the hand"
	super addCustomMenuItems: menu hand: aHandMorph.

	menu addLine.
	Preferences noviceMode
		ifFalse: [
			self addStackMenuItems: menu hand: aHandMorph.
			self addPenMenuItems: menu hand: aHandMorph.
			self addPlayfieldMenuItems: menu hand: aHandMorph].

	self isWorldMorph
		ifTrue: [
			menu addLine.
			Preferences noviceMode
				ifFalse: [(owner isKindOf: BOBTransformationMorph)
						ifTrue: [self addScalingMenuItems: menu hand: aHandMorph]].
			menu addUpdating: #showWorldMainDockingBarString action: #toggleShowWorldMainDockingBar.

			Flaps sharedFlapsAllowed ifTrue: [
				menu
					addUpdating: #suppressFlapsString
					target: Project current
					action: #toggleFlapsSuppressed.
			].
			 
			Preferences noviceMode ifFalse: [| twm |
				menu addLine.

				twm := TheWorldMenu new.
				twm world: self project: Project current hand: aHandMorph.

				menu add: 'old desktop menu... (W)' translated subMenu: twm buildWorldMenu.
			].
		].
! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'dao 10/1/2004 12:56'!
addGlobalFlaps 
	"Must make global flaps adapt to world.  Do this even if not shown, so the old world will not be pointed at by the flaps."

	| use thisWorld |
	use := Flaps sharedFlapsAllowed.
	Project current flapsSuppressed ifTrue: [use := false].
	"Smalltalk isMorphic ifFalse: [use := false]."
	thisWorld := use 
		ifTrue: [self]
		ifFalse: [PasteUpMorph new initForProject:  "fake to be flap owner"
						WorldState new;
					bounds: (0@0 extent: 4000@4000);
					viewBox: (0@0 extent: 4000@4000)].
	
	Flaps globalFlapTabsIfAny do: [:aFlapTab |
		(Project current isFlapEnabled: aFlapTab) ifTrue:
			[(aFlapTab world == thisWorld) ifFalse:
				[thisWorld addMorphFront: aFlapTab.
				aFlapTab adaptToWorld: thisWorld].	"always do"
			use ifTrue:
				[aFlapTab spanWorld.
				aFlapTab adjustPositionAfterHidingFlap.
				aFlapTab flapShowing ifTrue: [aFlapTab showFlap]]]]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 10/26/2000 14:52'!
addHand: aHandMorph
	"Add the given hand to the list of hands for this world."

	aHandMorph owner ifNotNil:[aHandMorph owner removeHand: aHandMorph].
	worldState addHand: aHandMorph.
	aHandMorph privateOwner: self.
! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'mt 5/31/2016 12:06'!
addKeyboardShortcuts
	"Install an event capture filter to add shortcuts for global operations like opening a tool."
	
	self addKeyboardCaptureFilter: self.! !

!PasteUpMorph methodsFor: 'submorphs-add/remove' stamp: 'RAA 12/16/2000 18:37'!
addMorphFront: aMorph

	^self addMorphInFrontOfLayer: aMorph
! !

!PasteUpMorph methodsFor: 'WiW support' stamp: 'dgd 8/31/2004 16:25'!
addMorphInLayer: aMorph
	super addMorphInLayer: aMorph.
	aMorph wantsToBeTopmost ifFalse:[self bringTopmostsToFront].! !

!PasteUpMorph methodsFor: 'world state' stamp: 'dgd 2/22/2003 14:09'!
addMorphsAndModel: aMorphOrList 
	"Dump in submorphs, model, and stepList from aMorphOrList.  Used to bring a world, paste-up, or other morph in from an object file."

	aMorphOrList isMorph 
		ifTrue: 
			[aMorphOrList isWorldMorph 
				ifFalse: 
					["one morph, put on hand"

					"aMorphOrList installModelIn: self.  	a chance to install model pointers"

					aMorphOrList privateOwner: nil.
					self firstHand attachMorph: aMorphOrList.
					self startSteppingSubmorphsOf: aMorphOrList]
				ifTrue: 
					[model isNil 
						ifTrue: [self setModel: aMorphOrList modelOrNil]
						ifFalse: 
							[aMorphOrList modelOrNil ifNotNil: 
									[aMorphOrList modelOrNil privateOwner: nil.
									self addMorph: aMorphOrList modelOrNil]].
					aMorphOrList privateSubmorphs reverseDo: 
							[:m | 
							m privateOwner: nil.
							self addMorph: m.
							m changed].
					(aMorphOrList instVarNamed: 'stepList') 
						do: [:entry | entry first startSteppingIn: self]]]
		ifFalse: 
			["list, add them all"

			aMorphOrList reverseDo: 
					[:m | 
					m privateOwner: nil.
					self addMorph: m.
					self startSteppingSubmorphsOf: m.	"It may not want this!!"
					m changed]]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'di 7/15/1999 09:51'!
addMorph: aMorph centeredNear: aPoint
	"Add the given morph to this world, attempting to keep its center as close to the given point possible while also keeping the it entirely within the bounds of this world."

	| trialRect delta |
	trialRect := Rectangle center: aPoint extent: aMorph fullBounds extent.
	delta := trialRect amountToTranslateWithin: bounds.
	aMorph position: trialRect origin + delta.
	self addMorph: aMorph.
! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'sw 10/2/2007 21:37'!
addPenMenuItems: menu hand: aHandMorph
	"Add a pen-trails-within submenu to the given menu"

	menu add: 'pen trails...' translated target: self selector: #putUpPenTrailsSubmenu.
	menu balloonTextForLastItem: 'its governing pen trails drawn within' translated! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'sw 10/2/2007 21:39'!
addPenTrailsMenuItemsTo: aMenu
	"Add items relating to pen trails to aMenu"

	| oldTarget |
	oldTarget _ aMenu defaultTarget.
	aMenu defaultTarget: self.
	aMenu add: 'clear pen trails' translated action: #clearTurtleTrails.
	aMenu addLine.
	aMenu add: 'all pens up' translated action: #liftAllPens.
	aMenu add: 'all pens down' translated action: #lowerAllPens.
	aMenu addLine.
	aMenu add: 'all pens show lines' translated action: #linesForAllPens.
	aMenu add: 'all pens show arrowheads' translated action: #arrowsForAllPens.
	aMenu add: 'all pens show arrows' translated action: #linesAndArrowsForAllPens.
	aMenu add: 'all pens show dots' translated action: #dotsForAllPens.
	aMenu  addLine.
	aMenu addUpdating:  #batchPenTrailsString  action: #toggleBatchPenTrails.
	aMenu balloonTextForLastItem: 'if true, detailed movement of pens between display updates is ignored.  Thus multiple line segments drawn within a script may not be seen individually.' translated.

	aMenu defaultTarget: oldTarget! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'sw 3/3/2004 17:52'!
addPlayfieldMenuItems: menu hand: aHandMorph
	"Add playfield-related items to the menu"

	menu add: 'playfield options...' translated target: self action: #presentPlayfieldMenu.
	(self hasProperty: #donorTextMorph) ifTrue:
		[menu add: 'send contents back to donor' translated action: #sendTextContentsBackToDonor]! !

!PasteUpMorph methodsFor: 'Nebraska' stamp: 'ar 10/26/2000 14:19'!
addRemoteClient: aClient
	self addHand: aClient hand.
	worldState addRemoteCanvas: aClient canvas.
	aClient canvas fullDrawMorph: self.
	self changed.  "force a redraw"
! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 8/30/2003 21:18'!
addScalingMenuItems: menu hand: aHandMorph

	| subMenu |

	(subMenu := MenuMorph new)
		defaultTarget: self;
		add: 'show application view' translated action: #showApplicationView;
		add: 'show factory view' translated action: #showFactoryView;
		add: 'show whole world view' translated action: #showFullView;
		add: 'expand' translated action: #showExpandedView;
		add: 'reduce' translated action: #showReducedView;
		addLine;
		add: 'define application view' translated action: #defineApplicationView;
		add: 'define factory view' translated action: #defineFactoryView.
	menu
		add: 'world scale and clip...' translated
		subMenu: subMenu! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'sw 10/23/2000 14:41'!
addStackMenuItems: menu hand: aHandMorph
	"Add appropriate stack-related items to the given menu"

	self isStackBackground
		ifTrue:
			[menu add: 'card & stack...' target: self action: #presentCardAndStackMenu]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/5/2000 18:34'!
addUndoItemsTo: aWorldMenu
	"Add undo-related items to the given menu.  Will add zero, one or two items, depending on the settings of the #useUndo and #infiniteUndo preferences"

	Preferences useUndo ifFalse: [^ self].
	Preferences infiniteUndo
		ifFalse:
			[aWorldMenu addUpdating: #undoOrRedoMenuWording target: self commandHistory action: #undoOrRedoCommand]
		ifTrue:
			[aWorldMenu addUpdating: #undoMenuWording target: self commandHistory  action: #undoLastCommand.
			aWorldMenu addUpdating: #redoMenuWording target: self commandHistory action: #redoNextCommand.
			self flag: #deferred.  "The following feature to be unblocked in due course"
			"aWorldMenu add: 'undo to...' target: self commandHistory action: #undoTo"].
	aWorldMenu addLine! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'ul 11/15/2010 11:56'!
addWorldHaloMenuItemsTo: aMenu hand: aHandMorph
	"Add standard halo items to the menu, given that the receiver is a World"

	| unlockables |
	self addFillStyleMenuItems: aMenu hand: aHandMorph.
	self addLayoutMenuItems: aMenu hand: aHandMorph.

	aMenu addLine.
	self addWorldToggleItemsToHaloMenu: aMenu.
	aMenu addLine.
	self addCopyItemsTo: aMenu.
	self addPlayerItemsTo: aMenu.
	self addExportMenuItems: aMenu hand: aHandMorph.
	self addStackItemsTo: aMenu.
	self addMiscExtrasTo: aMenu.

	Preferences noviceMode ifFalse:
		[self addDebuggingItemsTo: aMenu hand: aHandMorph].

	aMenu addLine.
	aMenu defaultTarget: self.

	aMenu addLine.

	unlockables := self submorphs select:
		[:m | m isLocked].
	unlockables size = 1 ifTrue:
		[aMenu add: ('unlock "{1}"' translated format:{unlockables first externalName})action: #unlockContents].
	unlockables size > 1 ifTrue:
		[aMenu add: 'unlock all contents' translated action: #unlockContents.
		aMenu add: 'unlock...' translated action: #unlockOneSubpart].

	aMenu defaultTarget: aHandMorph.
! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'sw 7/6/2009 15:47'!
addWorldToggleItemsToHaloMenu: aMenu
	"Add toggle items for the world to the halo menu ....  July 2009:  no longer in world halo menu"

	"aMenu addUpdating: #showTabsString
				target: CurrentProjectRefactoring 
				action: #currentToggleFlapsSuppressed  "! !

!PasteUpMorph methodsFor: 'submorphs-accessing' stamp: 'RAA 5/24/2000 12:09'!
allMorphsDo: aBlock
	"Enumerate all morphs in the world, including those held in hands."

	super allMorphsDo: aBlock.
	self isWorldMorph
		ifTrue: [worldState handsReverseDo: [:h | h allMorphsDo: aBlock]].
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'dgd 8/31/2004 16:23'!
allNonFlapRelatedSubmorphs
	"Answer all non-window submorphs that are not flap-related"

	^submorphs 
		select: [:m | (m isSystemWindow) not and: [m wantsToBeTopmost not]]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'ar 1/3/2010 19:01'!
allScriptEditors
	^ self allMorphs select:
		[:s | s isScriptEditorMorph]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 10/24/2000 13:51'!
allScriptors
	"Answer a list of all active scriptors running on behalf of the receiver.  This is a hook used in past demos and with a future life which however presently is vacuous"

	^ #()
"
	^ self allMorphs select: [:m | m isKindOf: Scriptor]"! !

!PasteUpMorph methodsFor: 'scripting' stamp: 'tk 1/31/2001 15:58'!
allTileScriptingElements
	"Answer a list of all the morphs that pertain to tile-scripting.  A sledge-hammer"

	| all morphs |
	morphs := IdentitySet new: 400.
	self allMorphsAndBookPagesInto: morphs.
	all := morphs select: [:s | s isTileScriptingElement].
"	self closedViewerFlapTabs do:
		[:aTab | all addAll: aTab referent allTileScriptingElements].
"
	^ all asOrderedCollection! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 7/6/1998 14:19'!
alwaysShowThumbnail
	^ self hasProperty: #alwaysShowThumbnail! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 5/24/2000 10:45'!
assuredCanvas
	
	^worldState assuredCanvas! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'sw 6/17/1999 16:02'!
assureFlapTabsFitOnScreen
	self flapTabs do:
		[:m | m fitOnScreen]! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'ar 6/30/2001 13:21'!
assureFlapWidth: requestedWidth
	| tab |
	self width: requestedWidth.
	tab := self flapTab ifNil:[^self].
	tab flapShowing ifTrue:[tab hideFlap; showFlap].! !

!PasteUpMorph methodsFor: 'world state' stamp: 'tak 3/15/2005 17:31'!
assureNotPaintingElse: aBlock
	"If painting is already underway in the receiver, put up an informer to that effect and evalute aBlock"
	self removeModalWindow.
	self sketchEditorOrNil ifNotNil:
		[self inform: 'Sorry, you can only paint
one object at a time' translated.
		Cursor normal show.
		^ aBlock value]
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 12/18/2000 01:16'!
assureNotPaintingEvent: evt
	"If painting is already underway
	in the receiver, put up an informer to that effect and evalute aBlock"
	| editor |
	(editor := self sketchEditorOrNil) ifNotNil:[
		editor save: evt.
		Cursor normal show.
	].! !

!PasteUpMorph methodsFor: 'options' stamp: 'ar 11/9/2000 12:48'!
autoLineLayout
	| layout |
	layout := self layoutPolicy ifNil:[^false].
	layout isTableLayout ifFalse:[^false].
	self listDirection == #leftToRight ifFalse:[^false].
	self wrapDirection == #topToBottom ifFalse:[^false].
	^true! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 9/6/2003 17:54'!
autoLineLayoutString
	"Answer the string to be shown in a menu to represent the  
	auto-line-layout status"
	^ (self autoLineLayout
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'auto-line-layout' translated! !

!PasteUpMorph methodsFor: 'options' stamp: 'ar 11/9/2000 15:07'!
autoLineLayout: aBoolean
	"Make the receiver be viewed with auto-line-layout, which means that its submorphs will be laid out left-to-right and then top-to-bottom in the manner of a word processor, or (if aBoolean is false,) cease applying auto-line-layout"

	aBoolean ifTrue:
		[self viewingNormally ifTrue: [self saveBoundsOfSubmorphs]].
	aBoolean ifTrue:[
		self layoutPolicy: TableLayout new.
		self layoutInset: 8; cellInset: 4.
		self listDirection: #leftToRight; wrapDirection: #topToBottom.
	] ifFalse:[
		self layoutPolicy: nil.
		self layoutInset: 0; cellInset: 0.
	].
! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'sw 6/18/1998 09:10'!
automaticPhraseExpansion
	^ self hasProperty: #automaticPhraseExpansion! !

!PasteUpMorph methodsFor: 'e-toy support' stamp: 'sw 7/6/1999 08:39'!
automaticViewing
	^ self hasProperty: #automaticViewing! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'ul 8/2/2011 20:07'!
becomeActiveDuring: aBlock
	"Make the receiver the ActiveWorld during the evaluation of aBlock.
	Note that this method does deliberately *not* use #ensure: to prevent
	re-installation of the world on project switches."
	| priorWorld priorHand priorEvent |
	priorWorld := ActiveWorld.
	priorHand := ActiveHand.
	priorEvent := ActiveEvent.
	ActiveWorld := self.
	ActiveHand := self hands first. "default"
	ActiveEvent := nil. "not in event cycle"
	aBlock
		on: Error
		do: [:ex | 
			ActiveWorld := priorWorld.
			ActiveEvent := priorEvent.
			ActiveHand := priorHand.
			ex pass]! !

!PasteUpMorph methodsFor: 'options' stamp: 'tk 10/30/2001 18:40'!
behaveLikeHolder
 
	self vResizeToFit: true; autoLineLayout: true; indicateCursor: true! !

!PasteUpMorph methodsFor: 'options' stamp: 'sw 9/27/2007 04:24'!
behaveLikeHolder: aBoolean
 	"Change the receiver's viewing properties such that they conform to what we commonly call a Holder, viz: resize-to-fit, do auto-line-layout, and indicate the 'cursor'"

	self vResizeToFit: aBoolean; autoLineLayout: aBoolean; indicateCursor: aBoolean.
	self changed "redraw"
	! !

!PasteUpMorph methodsFor: 'options' stamp: 'sw 10/17/2000 12:04'!
behavingLikeAHolder
	"Answer whether the receiver is currently behaving like a Holder"

	^ self resizeToFit and: [self indicateCursor and: [self autoLineLayout]]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 6/19/2000 07:44'!
beWorldForProject: aProject

	self privateOwner: nil.
	worldState := WorldState new.
	self addHand: HandMorph new.
	self setProperty: #automaticPhraseExpansion toValue: true.
	self setProperty: #optimumExtentFromAuthor toValue: Display extent.
	self startSteppingSubmorphsOf: self! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'dgd 4/4/2006 13:58'!
bringTopmostsToFront
	submorphs
		select:[:m| m wantsToBeTopmost]
		thenDo:[:m| self addMorphInLayer: m].! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 1/30/2001 20:37'!
bringWindowsFullOnscreen
	"Make ever SystemWindow on the desktop be totally on-screen, whenever possible."
	
	(SystemWindow windowsIn: self satisfying: [:w | true]) do:
		[:aWindow | 
			aWindow right: (aWindow right min: bounds right).
			aWindow bottom: (aWindow bottom min: bounds bottom).
			aWindow left: (aWindow left max: bounds left).
			aWindow top: (aWindow top max: bounds top)]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 6/4/2004 15:00'!
browseAllScriptsTextually
	"Put up a browser showing all scripts in the project textually"

	self presenter browseAllScriptsTextually

"ActiveWorld browseAllScriptsTextually"! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 8/30/2003 20:42'!
buildDebugMenu: aHandMorph
	| aMenu |
	aMenu := super buildDebugMenu: aHandMorph.
	aMenu add:  'abandon costume history' translated target: self action: #abandonCostumeHistory.
	^ aMenu! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/5/2000 18:27'!
buildWorldMenu: evt
	^(TheWorldMenu new
		world: self
		project: (self project ifNil: [Project current])       "mvc??"
		hand: evt hand) buildWorldMenu.! !

!PasteUpMorph methodsFor: 'misc' stamp: 'di 12/23/1998 14:44'!
cachedOrNewThumbnailFrom: newThumbnail
	"If I have a cached thumbnail, and it is of the desired extent, then ruturn it.
	Otherwise produce one in newThumbnail and return it (after caching).
	This code parallels what happens in page: to match resultant extent."
	| cachedThumbnail scale ext |
	scale := newThumbnail height / self fullBounds height.
	ext := (self fullBounds extent * scale) truncated.
	(cachedThumbnail := self valueOfProperty: #cachedThumbnail) ifNotNil:
		[cachedThumbnail extent = ext ifTrue: [^ cachedThumbnail]].
	self setProperty: #cachedThumbnail toValue: (newThumbnail page: self).
	^ newThumbnail! !

!PasteUpMorph methodsFor: 'visual properties' stamp: 'bf 5/4/2000 15:27'!
canHaveFillStyles
	"Return true if the receiver can have general fill styles; not just colors.
	This method is for gradually converting old morphs."
	^ true! !

!PasteUpMorph methodsFor: 'project state' stamp: 'di 7/27/1999 10:46'!
canvas

	^ worldState canvas! !

!PasteUpMorph methodsFor: 'project state' stamp: 'ar 6/17/2010 20:00'!
canvas: aCanvas
	"Set this world's canvas"

	worldState canvas: aCanvas.
! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 7/6/1999 13:30'!
cartesianOrigin
	^ self originAtCenter
		ifFalse:
			[self bottomLeft]
		ifTrue:
			[self center]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 8/6/2009 18:30'!
checkCurrentHandForObjectToPaste

	| response |
	self primaryHand pasteBuffer ifNil: [^self].
	response := UIManager default chooseFrom: #('Delete' 'Keep')
		title: 'Hand is holding a Morph in its paste buffer:\' withCRs,
			self primaryHand pasteBuffer printString.
	response = 1 ifTrue: [self primaryHand pasteBuffer: nil].
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 3/17/2001 23:57'!
checkCurrentHandForObjectToPaste2

	self primaryHand pasteBuffer ifNil: [^self].
	self inform: 'Hand is holding a Morph in its paste buffer:\' withCRs,
		self primaryHand pasteBuffer printString.

! !

!PasteUpMorph methodsFor: 'world state' stamp: 'bf 5/16/2011 23:45'!
chooseClickTarget
	Cursor crossHair showWhile:
		[Sensor waitButton].
	Cursor down showWhile:
		[Sensor anyButtonPressed].
	^ (self morphsAt: Sensor cursorPoint) first topRendererOrSelf! !

!PasteUpMorph methodsFor: 'stepping' stamp: 'RAA 5/24/2000 11:10'!
cleanseStepList
	"Remove morphs from the step list that are not in this World.  Often were in a flap that has moved on to another world."

	worldState cleanseStepListForWorld: self! !

!PasteUpMorph methodsFor: 'undo' stamp: 'RAA 9/21/2000 20:07'!
clearCommandHistory

	worldState ifNotNil: [worldState clearCommandHistory]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 9/8/2000 18:52'!
closedViewerFlapTabs
	"Answer all the viewer flap tabs in receiver that are closed"

	^ self submorphs select:
		[:m | (m isKindOf: ViewerFlapTab) and: [m flapShowing not]]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 8/6/2009 19:01'!
closeUnchangedWindows
	"Present a menu of window titles for all windows with changes,
	and activate the one that gets chosen."
	(UIManager default confirm:
'Do you really want to close all windows
except those with unaccepted edits?' translated)
		ifFalse: [^ self].

	(SystemWindow windowsIn: self satisfying: [:w | w model canDiscardEdits])
		do: [:w | w delete]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/5/2000 17:27'!
collapseAll
	"Collapse all windows"
	(SystemWindow windowsIn: self satisfying: [:w | w isCollapsed not])
		reverseDo: [:w | w collapseOrExpand.  self displayWorld].
	self collapseNonWindows! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/5/2000 17:27'!
collapseNonWindows
	self allNonFlapRelatedSubmorphs do:
		[:m | m collapse]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'di 7/15/1999 09:51'!
colorAt: aPoint belowMorph: aMorph
	"Return the color of the pixel immediately behind the given morph at the given point.
	NOTE: due to some bounds wobble in flexing, we take the middle of 3x3 rect."
	^ (self patchAt: (aPoint-1 extent: 3) without: aMorph andNothingAbove: true)
		colorAt: 1@1
! !

!PasteUpMorph methodsFor: 'undo' stamp: 'ar 8/31/2000 23:16'!
commandHistory
	"Return the command history for the receiver"
	^self isWorldMorph
		ifTrue:[worldState commandHistory]
		ifFalse:[super commandHistory]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 3/13/2003 13:52'!
commandKeySelectors
	"Answer my command-key table"

	| aDict |
	aDict := self valueOfProperty: #commandKeySelectors ifAbsentPut: [self initializeDesktopCommandKeySelectors].
	^ aDict! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/24/2000 14:03'!
connectRemoteUser
	
	^self
		connectRemoteUserWithName: nil 
		picture: nil 
		andIPAddress: nil
! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'fbs 12/3/2010 23:33'!
connectRemoteUserWithName: nameStringOrNil picture: aFormOrNil andIPAddress: aStringOrNil
	"Prompt for the initials to be used to identify the cursor of a remote user, then create a cursor for that user and wait for a connection."

	| initials addr h |
	initials := nameStringOrNil.
	initials isEmptyOrNil ifTrue: [
		initials := UIManager default request: 'Enter initials for remote user''s cursor?'.
	].
	initials isEmpty ifTrue: [^ self].  "abort"
	addr := 0.
	aStringOrNil isEmptyOrNil ifFalse: [
		addr := NetNameResolver addressForName: aStringOrNil timeout: 30
	].
	addr = 0 ifTrue: [
		addr := NetNameResolver promptUserForHostAddress.
	].
	addr = 0 ifTrue: [^ self].  "abort"

	(RemoteHandMorph ensureNetworkConnected) ifFalse: [^ self]. "abort"

	h := RemoteHandMorph new userInitials: initials andPicture: aFormOrNil.
	self addHand: h.
	h changed.
	h startListening.
	h startTransmittingEventsTo: addr.
! !

!PasteUpMorph methodsFor: 'layout' stamp: 'cmm 11/18/2010 17:29'!
convertAlignment
	self 
		clipSubmorphs: true ;
		layoutPolicy: nil ;
		layoutInset: 0 ;
		cellInset: 0 ;
		vResizing: #rigid! !

!PasteUpMorph methodsFor: 'Nebraska' stamp: 'RAA 11/8/2000 15:04'!
convertRemoteClientToBuffered: aClient

	worldState removeRemoteCanvas: aClient canvas.
	aClient convertToBuffered.
	worldState addRemoteCanvas: aClient canvas.
	self changed.  "force a redraw"
! !

!PasteUpMorph methodsFor: 'objects from disk' stamp: 'RAA 12/20/2000 17:48'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	"transition from project to worldState (8/16/1999)"
	worldState ifNil: [varDict at: 'project' ifPresent: [ :x | worldState := x]].

	"elimination of specific gradient stuff (5/6/2000)"
	varDict at: 'fillColor2' ifPresent: [ :color2 |
		(color isColor and: [color2 isColor and: [color ~= color2]]) ifTrue: [
			self useGradientFill.
			self fillStyle
				colorRamp: {0.0 -> color. 1.0 -> color2};
				radial: false;
				origin: self position;
				direction: ((varDict at: 'gradientDirection') == #vertical 
					ifTrue:[0@self height] 
					ifFalse:[self width@0]).
		]
	].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.
! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'sw 10/5/2006 04:19'!
correspondingFlapTab
	"If there is a flap tab whose referent is me, return it, else return nil.  Will also work for flaps on the edge of embedded subareas such as within scripting-areas, but more slowly."

	self currentWorld flapTabs do:
		[:aTab | aTab referent == self ifTrue: [^ aTab]].

	"Catch guys in embedded worldlets"
	ActiveWorld allMorphs do:
		[:aTab | ((aTab isKindOf: FlapTab) and: [aTab referent == self]) ifTrue: [^ aTab]].

	^ nil! !

!PasteUpMorph methodsFor: 'model' stamp: 'dgd 2/22/2003 14:09'!
createCustomModel
	"Create a model object for this world if it does not yet have one. A model object is an initially empty subclass of MorphicModel. As the user names parts and adds behavior, instance variables and methods are added to this class."

	model isNil ifFalse: [^self].
	model := MorphicModel newSubclass new! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'sw 7/9/2005 18:19'!
currentlyUsingVectorVocabulary
	"Answer whether this world is currently set up to use the vector vocabulary"

	^ (self valueOfProperty: #currentVocabularySymbol) == #Vector! !

!PasteUpMorph methodsFor: 'e-toy support' stamp: 'di 7/27/1999 10:24'!
cursor 
	^ cursor
! !

!PasteUpMorph methodsFor: 'cursor' stamp: 'tak 11/7/2004 18:33'!
cursorWrapped: aNumber 
	"Set the cursor to the given number, modulo the number of items I
	contain. Fractional cursor values are allowed."
	| oldRect newRect offset |
	cursor = aNumber
		ifTrue: [^ self].
	self hasSubmorphs
		ifFalse: [cursor := 1.
			^ self].
	oldRect := self selectedRect.
	offset := (self asNumber: aNumber) - 1 \\ submorphs size.
	cursor := offset + 1.
	newRect := self selectedRect.
	self indicateCursor
		ifTrue: [self invalidRect: oldRect;
				 invalidRect: newRect]! !

!PasteUpMorph methodsFor: 'e-toy support' stamp: 'sw 9/7/2000 11:43'!
cursor: aNumber
	"for backward compatibility"

	self cursorWrapped: aNumber! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 9/15/2000 23:28'!
deEmphasizeViewMVC: asTwoTone
	self flag: #arNote. "Probably unnecessary"
	worldState handsDo:          "free dependents links if any"
		[:h | h releaseKeyboardFocus].
	worldState canvas: nil.		"free model's canvas to save space"
	self fullReleaseCachedState.
	asTwoTone ifTrue: [
		"draw deEmphasized as a two-tone (monochrome) form"
		self displayWorldAsTwoTone].
! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:35'!
defaultBorderColor
	"answer the default border color/fill style for the receiver"
	^ Color
		r: 0.861
		g: 1.0
		b: 0.722! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:38'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 1! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'tfel 2/27/2010 19:56'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Preferences defaultWorldColor muchLighter.! !

!PasteUpMorph methodsFor: 'viewer' stamp: 'KR 5/23/2008 01:29'!
defaultNameStemForInstances
	"Answer a basis for names of default instances of the receiver"
	^ self isWorldMorph
		ifFalse:
			[super defaultNameStemForInstances]
		ifTrue:
			['world' translatedNoop]! !

!PasteUpMorph methodsFor: 'halos and balloon help' stamp: 'sw 1/10/2000 16:44'!
defersHaloOnClickTo: aSubMorph
	"If a cmd-click on aSubMorph would make it a preferred recipient of the halo, answer true"
	^ true
	! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'RAA 6/26/2000 11:31'!
defineApplicationView

	| r |
	r := Rectangle fromUser.
	self 
		setProperty: #applicationViewBounds 
		toValue: ((self transformFromOutermostWorld) globalBoundsToLocal: r) truncated ! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'RAA 6/26/2000 11:29'!
defineFactoryView

	| r |
	r := Rectangle fromUser.
	self 
		setProperty: #factoryViewBounds 
		toValue: ((self transformFromOutermostWorld) globalBoundsToLocal: r) truncated ! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'dgd 9/11/2004 20:45'!
delayedInvokeWorldMenu: evt 
	self
		addAlarm: #invokeWorldMenu:
		with: evt
		after: 200! !

!PasteUpMorph methodsFor: 'world state' stamp: 'cmm 12/4/2010 20:55'!
deleteAllHalos
	self haloMorphs do:
		[ : m | m target isSelectionMorph ifTrue: [ m target delete ] ].
	self hands do:
		[ : each | each removeHalo ]! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'ar 10/3/2000 17:02'!
deleteBalloonTarget: aMorph
	"Delete the balloon help targeting the given morph"
	self handsDo:[:h| h deleteBalloonTarget: aMorph].! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'sw 5/5/2001 00:27'!
deleteGlobalFlapArtifacts
	"Delete all flap-related detritus from the world"

	| localFlaps |
	localFlaps := self localFlapTabs collect: [:m | m referent].
	self submorphs do:
		[:m | 
			((m isFlapTab) and: [m isGlobalFlap]) ifTrue: [m delete].
			m isFlap ifTrue:[(localFlaps includes: m) ifFalse: [m delete]]]

"ActiveWorld deleteGlobalFlapArtifacts"

! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 8/6/2009 19:01'!
deleteNonWindows
	(UIManager default confirm:
'Do you really want to discard all objects
that are not in windows?' translated)
		ifFalse: [^ self].

	self allNonFlapRelatedSubmorphs do:
		[:m | m delete]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/5/2000 17:29'!
detachableScriptingSpace
	ScriptingSystem newScriptingSpace openInWorld: self! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/24/2000 14:07'!
disconnectAllRemoteUsers
	"Disconnect all remote hands and stop transmitting events."
	self world handsDo: [:h |
		(h isKindOf: RemoteHandMorph) 
			ifTrue: [h withdrawFromWorld]].! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 8/6/2009 20:43'!
disconnectRemoteUser
	"Prompt for the initials of the remote user, then remove the remote hand with those initials, breaking its connection."

	"select hand to remove"
	| initials handToRemove |
	initials := UIManager default request: 'Enter initials for remote user''s cursor?'.
	initials isEmpty ifTrue: [^ self].  "abort"
	handToRemove := nil.
	self handsDo: [:h |
		h userInitials = initials ifTrue: [handToRemove := h]].
	handToRemove ifNil: [^ self].  "no hand with those initials"
	handToRemove withdrawFromWorld.
! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 3/13/2003 12:19'!
dispatchCommandKeyInWorld: aChar event: evt
	"Dispatch the desktop command key if possible.  Answer whether handled"

	| aMessageSend |
	aMessageSend := self commandKeySelectors at: aChar ifAbsent: [^ false].
	aMessageSend selector numArgs = 0
		ifTrue:
			[aMessageSend value]
		ifFalse:
			[aMessageSend valueWithArguments: (Array with: evt)].
	^ true
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'mt 7/28/2016 15:13'!
displayWorld

	(self valueOfProperty: #shouldDisplayWorld ifAbsent: [true])
		ifTrue: [worldState displayWorld: self submorphs: submorphs].! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 5/24/2000 13:20'!
displayWorldAsTwoTone
	"Display the world in living black-and-white. (This is typically done to save space.)"

	worldState displayWorldAsTwoTone: self submorphs: submorphs color: color
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'dgd 2/22/2003 14:10'!
displayWorldNonIncrementally
	"Display the morph world non-incrementally. Used for testing."

	(worldState canvas isNil or: 
			[worldState canvas extent ~= self viewBox extent 
				or: [worldState canvas form depth ~= Display depth]]) 
		ifTrue: 
			["allocate a new offscreen canvas the size of the window"

			worldState 
				canvas: (Display defaultCanvasClass extent: self viewBox extent)].
	worldState canvas fillColor: color.
	submorphs reverseDo: [:m | worldState canvas fullDrawMorph: m].
	worldState handsReverseDo: [:h | worldState canvas fullDrawMorph: h].
	worldState canvas form displayOn: Display at: self viewBox origin.
	self fullRepaintNeeded.	"don't collect damage"
	Display forceDisplayUpdate! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 5/24/2000 12:23'!
displayWorldSafely

	worldState displayWorldSafely: self.
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ls 5/6/2003 16:51'!
doOneCycle
	"see the comment in doOneCycleFor:"

	worldState doOneCycleFor: self! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 5/24/2000 10:14'!
doOneCycleInBackground
	
	worldState doOneCycleInBackground
! !

!PasteUpMorph methodsFor: 'interaction loop' stamp: 'ls 5/6/2003 16:51'!
doOneCycleNow
	"see the comment in doOneCycleNowFor:"
	worldState doOneCycleNowFor: self.
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 5/24/2000 11:59'!
doOneSubCycle
	"Like doOneCycle, but preserves activeHand."

	worldState doOneSubCycleFor: self! !

!PasteUpMorph methodsFor: 'world state' stamp: 'di 9/19/2000 22:17'!
dragThroughOnDesktop: evt
	"Draw out a selection rectangle"
	| selection |
	selection := SelectionMorph newBounds: (evt cursorPoint extent: 8@8).
	self addMorph: selection.
	^ selection extendByHand: evt hand
! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/5/2000 19:11'!
drawingClass

	^ SketchMorph! !

!PasteUpMorph methodsFor: 'drawing' stamp: 'ar 1/3/2010 16:31'!
drawOn: aCanvas 
	"Draw in order:
	- background color
	- grid, if any
	- background sketch, if any
	- Update and draw the turtleTrails form. See the comment in updateTrailsForm.
	- cursor box if any

	Later (in drawSubmorphsOn:) I will skip drawing the background sketch."

	"draw background fill"
	super drawOn: aCanvas.

	"draw grid"
	(self griddingOn and: [self gridVisible]) 
		ifTrue: 
			[aCanvas fillRectangle: self bounds
				fillStyle: (self 
						gridFormOrigin: self gridOrigin
						grid: self gridModulus
						background: nil
						line: Color lightGray)].

	"draw background sketch."
	backgroundMorph ifNotNil: [
		self clipSubmorphs ifTrue: [
			aCanvas clipBy: self clippingBounds
				during: [ :canvas | canvas fullDrawMorph: backgroundMorph ]]
			ifFalse: [ aCanvas fullDrawMorph: backgroundMorph ]].

	"draw turtle trails"
	(lastTurtlePositions isNil or: [lastTurtlePositions isEmpty]) ifFalse:[
		self updateTrailsForm.
	].
	turtleTrailsForm 
		ifNotNil: [aCanvas paintImage: turtleTrailsForm at: self position].

	"draw cursor"
	(submorphs notEmpty and: [self indicateCursor]) 
		ifTrue: 
			[aCanvas 
				frameRectangle: self selectedRect
				width: 2
				color: Color black]! !

!PasteUpMorph methodsFor: 'painting' stamp: 'nk 7/4/2003 15:59'!
drawSubmorphsOn: aCanvas 
	"Display submorphs back to front, but skip my background sketch."

	| drawBlock |
	submorphs isEmpty ifTrue: [^self].
	drawBlock := [:canvas | submorphs reverseDo: [:m | m ~~ backgroundMorph ifTrue: [ canvas fullDrawMorph: m ]]].
	self clipSubmorphs 
		ifTrue: [aCanvas clipBy: self clippingBounds during: drawBlock]
		ifFalse: [drawBlock value: aCanvas]! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'sw 2/4/2001 00:54'!
dropEnabled
	"Get this morph's ability to add and remove morphs via drag-n-drop."

	^ (self valueOfProperty: #dropEnabled) ~~ false
! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'topa 11/2/2015 10:04'!
dropFiles: anEvent
	"Handle a number of dropped files from the OS.
	TODO:
		- use a more general mechanism for figuring out what to do with the file (perhaps even offering a choice from a menu)
		- remember the resource location or (when in browser) even the actual file handle
	"
	| numFiles |
	numFiles := anEvent contents.
	1 to: numFiles do: [ :i |
		(FileDirectory requestDropDirectory: i) 
			ifNotNil: [:directory | self handleDroppedItem: directory event: anEvent]
			ifNil: [(FileStream requestDropStream: i) ifNotNil: [:stream |
				[self handleDroppedItem: stream event: anEvent] ensure: [stream close]]]].
				! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'ar 1/3/2010 18:18'!
dumpPresenter
	"Dump my current presenter"
	presenter := nil.! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 6/11/2000 17:45'!
embeddedProjectDisplayMode

	"#naked - the embedded project/world is just a pasteup in the outer p/w
	#window - the embedded p/w is in a system window in the outer p/w
	#frame - the embedded p/w is in a green frame and clipped
	#scaled - the embedded p/w is in a green frame and scaled to fit"

	^#scaled
! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'sw 4/17/2001 11:23'!
enableGlobalFlaps 
	"Restore saved global flaps, or obtain brand-new system defaults if necessary"

	Flaps globalFlapTabs. 		 "If nil, creates new ones"
	self addGlobalFlaps 			 "put them on screen"! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 12/19/2000 19:23'!
endDrawing: evt
	"If painting is already underway
	in the receiver, finish and save it."
	| editor |
	(editor := self sketchEditorOrNil) ifNotNil:[
		editor save: evt.
		Cursor normal show.
	].! !

!PasteUpMorph methodsFor: 'world state' stamp: 'dao 10/1/2004 13:14'!
exit

	Project current exit
! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/5/2000 17:33'!
expandAll
	"Expand all windows"
	(SystemWindow windowsIn: self satisfying: [:w | w isCollapsed])
		reverseDo: [:w | w collapseOrExpand.  self displayWorld]! !

!PasteUpMorph methodsFor: 'geometry' stamp: 'RAA 6/20/2000 12:42'!
extent: aPoint

	super extent: aPoint.
	worldState ifNotNil: [
		worldState viewBox ifNotNil: [
			worldState canvas: nil.
			worldState viewBox: bounds
		].
	].! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'yo 12/12/2007 21:03'!
extractScreenRegion: poly andPutSketchInHand: hand
	"The user has specified a polygonal area of the Display.
	Now capture the pixels from that region, and put in the hand as a Sketch."
	| screenForm outline topLeft innerForm exterior |
	outline _ poly shadowForm.
	topLeft _ outline offset.
	exterior _ (outline offset: 0@0) anyShapeFill reverse.
	screenForm _ Form fromDisplay: (topLeft extent: outline extent).
	screenForm eraseShape: exterior.
	innerForm _ screenForm trimBordersOfColor: Color transparent.
	ActiveHand showTemporaryCursor: nil.
	innerForm isAllWhite ifFalse:
		[hand attachMorph: (self drawingClass withForm: innerForm)]! !

!PasteUpMorph methodsFor: 'events-processing' stamp: 'mt 5/31/2016 12:10'!
filterEvent: aKeyboardEvent for: anObject
	"Provide keyboard shortcuts."
	
	"Delegate keyboard shortcuts to my docking bars."
	self submorphsDo: [:ea | ea isDockingBar ifTrue: [
		ea filterEvent: aKeyboardEvent for: anObject. "No support for converting events here!!"
		aKeyboardEvent wasIgnored ifTrue: [^ aKeyboardEvent "early out"]]].
	
	aKeyboardEvent isKeystroke
		ifFalse: [^ aKeyboardEvent].
	
	aKeyboardEvent commandKeyPressed ifTrue: [
		aKeyboardEvent keyCharacter caseOf: {
			[$R] -> [Utilities browseRecentSubmissions].
			[$L] -> [self findAFileList: aKeyboardEvent].
			[$O] -> [self findAMonticelloBrowser].
			[$P] -> [self findAPreferencesPanel: aKeyboardEvent].
			"[$Z] -> [ChangeList browseRecentLog]."
			[$]] -> [Smalltalk snapshot: true andQuit: false].
		} otherwise: [^ aKeyboardEvent "no hit"].
		^ aKeyboardEvent ignore "hit!!"].

	^ aKeyboardEvent "no hit"! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 8/7/2009 22:22'!
findAChangeSorter: evt
	"Locate a change sorter, open it, and bring it to the front.  Create one if necessary"

	self findAWindowSatisfying:
		[:aWindow | (aWindow model isMemberOf: ChangeSorter) or:
				[aWindow model isKindOf: DualChangeSorter]] orMakeOneUsing: [DualChangeSorter open]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'cmm 6/3/2012 21:52'!
findAFileList: evt 
	"Bring a file list to the foreground, reusing an existing one if possible."
	self
		findAWindowSatisfying: [ : aWindow | (aWindow model class = FileList) and: [ aWindow model hasUnacceptedEdits not ] ]
		orMakeOneUsing: [ FileList prototypicalToolWindow openInWorld ]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 8/7/2009 22:12'!
findAMessageNamesWindow: evt
	"Locate a MessageNames tool, open it, and bring it to the front.  Create one if necessary"

	self findAWindowSatisfying:
		[:aWindow | aWindow model isKindOf: MessageNames] orMakeOneUsing: [MessageNames openMessageNames]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 1/2/2010 15:42'!
findAPreferencesPanel: evt
	"Locate a Preferences Panel, open it, and bring it to the front.  Create one if necessary"
	Smalltalk at: #PreferenceBrowser ifPresent:[:pb|
		self findAWindowSatisfying:[:aWindow | aWindow model isKindOf: pb]
			orMakeOneUsing:[pb open]
	].! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 8/7/2009 22:28'!
findATranscript: evt
	"Locate a transcript, open it, and bring it to the front.  Create one if necessary"

	self findAWindowSatisfying:
		[:aWindow | aWindow model == Transcript] orMakeOneUsing: [Transcript openLabel: 'Transcript']! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'mt 5/17/2016 15:59'!
findAWindowSatisfying: qualifyingBlock orMakeOneUsing: makeBlock 
	"Locate a window satisfying a block, open it, and bring it to the front.  Create one if necessary, by using the makeBlock"

	submorphs do:
		[:aMorph | | aWindow | 
		(((aWindow := aMorph renderedMorph) isSystemWindow) 
		and: [qualifyingBlock value: aWindow]) ifTrue:
			[aWindow isCollapsed ifTrue: [aWindow expand].
			self addMorphFront: aWindow.
			aWindow beKeyWindow.
			^self]].
	"None found, so create one"
	makeBlock value! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'mt 5/17/2016 15:59'!
findDirtyBrowsers: evt 
	"Present a menu of window titles for browsers with changes,
	and activate the one that gets chosen."

	| menu |
	menu := MenuMorph new.
	(SystemWindow windowsIn: self
		satisfying: [:w | (w model isKindOf: Browser) and: [w model canDiscardEdits not]]) 
			do: 
				[:w | 
				menu 
					add: w label
					target: w
					action: #beKeyWindow].
	menu submorphs notEmpty ifTrue: [menu popUpEvent: evt in: self]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'mt 5/17/2016 16:00'!
findDirtyWindows: evt 
	"Present a menu of window titles for all windows with changes,
	and activate the one that gets chosen."

	| menu |
	menu := MenuMorph new.
	(SystemWindow windowsIn: self
		satisfying: [:w | w model canDiscardEdits not]) do: 
				[:w | 
				menu 
					add: w label
					target: w
					action: #beKeyWindow].
	menu submorphs notEmpty ifTrue: [menu popUpEvent: evt in: self]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'mt 5/17/2016 16:00'!
findWindow: evt
	"Present a menu names of windows and naked morphs, and activate the one that gets chosen.  Collapsed windows appear below line, expand if chosen; naked morphs appear below second line; if any of them has been given an explicit name, that is what's shown, else the class-name of the morph shows; if a naked morph is chosen, bring it to front and have it don a halo."
	| menu expanded collapsed nakedMorphs |
	menu := MenuMorph new.
	expanded := SystemWindow windowsIn: self satisfying: [:w | w isCollapsed not].
	collapsed := SystemWindow windowsIn: self satisfying: [:w | w isCollapsed].
	nakedMorphs := self submorphsSatisfying:
		[:m | (m isSystemWindow not and: [(m isStickySketchMorph) not]) and:
			[(m isFlapTab) not]].
	(expanded isEmpty & (collapsed isEmpty & nakedMorphs isEmpty)) ifTrue: [^ Beeper beep].
	(expanded asSortedCollection: [:w1 :w2 | w1 label caseInsensitiveLessOrEqual: w2 label]) do:
		[:w | menu add: w label target: w action: #beKeyWindow.
			w model canDiscardEdits ifFalse: [menu lastItem color: Color red]].
	(expanded isEmpty | (collapsed isEmpty & nakedMorphs isEmpty)) ifFalse: [menu addLine].
	(collapsed asSortedCollection: [:w1 :w2 | w1 label caseInsensitiveLessOrEqual: w2 label]) do: 
		[:w | menu add: w label target: w action: #collapseOrExpand.
		w model canDiscardEdits ifFalse: [menu lastItem color: Color red]].
	nakedMorphs isEmpty ifFalse: [menu addLine].
	(nakedMorphs asSortedCollection: [:w1 :w2 | w1 nameForFindWindowFeature caseInsensitiveLessOrEqual: w2 nameForFindWindowFeature]) do:
		[:w | menu add: w nameForFindWindowFeature target: w action: #comeToFrontAndAddHalo].
	menu addTitle: 'find window' translated.
	
	menu popUpEvent: evt in: self.! !

!PasteUpMorph methodsFor: 'project state' stamp: 'di 7/27/1999 10:46'!
firstHand

	^ worldState hands first! !

!PasteUpMorph methodsFor: 'objects from disk' stamp: 'ul 12/12/2009 14:08'!
fixUponLoad: aProject seg: anImageSegment
	"We are in an old project that is being loaded from disk.
Fix up conventions that have changed."

	self isWorldMorph ifTrue: [
			(self valueOfProperty: #soundAdditions) ifNotNil:
				[:additions | SampledSound
assimilateSoundsFrom: additions]].

	^ super fixUponLoad: aProject seg: anImageSegment! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'sw 9/19/2006 14:08'!
flapTab
	"Answer the tab affilitated with the receiver.  Normally every flap tab is expected to have a PasteUpMorph which serves as its 'referent.'"

	| ww |
	self isFlap ifFalse:[^nil].
	ww _ self presenter associatedMorph ifNil: [ActiveWorld].
	^ ww ifNotNil: [ww flapTabs detect:[:any| any referent == self] ifNone: [nil]]! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'ar 9/28/2000 13:56'!
flapTabs
	^ self submorphs select:[:m| m isFlapTab]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'tpr 3/15/2013 12:36'!
flashRects: rectangleList color: aColor
	"For testing. Flashes the given list of rectangles on the Display so you can watch incremental redisplay at work."
	"Details: Uses two reverses so that the display is restored to its original state. This is necessary when in deferred update mode."

	| blt |
	blt := (BitBlt toForm: Display)
		sourceForm: nil;
		sourceOrigin: 0@0;
		clipRect: self viewBox;
		combinationRule: Form reverse.
	rectangleList do: [:r | | screenRect |
		screenRect := r translateBy: self viewBox origin.
		blt destRect: screenRect; copyBits.
		Display forceToScreen: screenRect; forceDisplayUpdate.
		(Delay forMilliseconds: 15) wait.
		blt destRect: screenRect; copyBits.
		Display forceToScreen: screenRect; forceDisplayUpdate].
! !

!PasteUpMorph methodsFor: 'geometry testing' stamp: 'RAA 6/2/2000 10:22'!
fullContainsPoint: pt
	"The world clips its children"

	worldState ifNil: [^super fullContainsPoint: pt].
	^bounds containsPoint: pt

! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 5/25/2000 15:12'!
fullRepaintNeeded

	worldState doFullRepaint.
	SystemWindow windowsIn: self
		satisfying: [:w | w makeMeVisible. false].

! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/5/2000 19:09'!
getWorldMenu: aSymbol
	^(TheWorldMenu new
		world: self
		project: (self project ifNil: [Project current])       "mvc??"
		hand: self primaryHand) perform: aSymbol! !

!PasteUpMorph methodsFor: 'world state' stamp: 'di 11/27/1999 10:11'!
goBack

	Project returnToPreviousProject.
! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'di 10/18/2001 03:33'!
grabDrawingFromScreen: evt
	"Allow the user to specify a rectangular area of the Display, capture the pixels from that area, and use them to create a new drawing morph. Attach the result to the hand."
	| m |
	m := self drawingClass new form: Form fromUser.
	evt hand position: Sensor cursorPoint.  "update hand pos after Sensor loop in fromUser"
	evt hand attachMorph: m.! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'nice 1/19/2010 01:57'!
grabFloodFromScreen: evt
	"Allow the user to plant a flood seed on the Display, and create a new drawing morph from the resulting region. Attach the result to the hand."
	| screenForm exterior p1 box |
	p1 := Cursor crossHair showWhile: [Sensor waitButton].
	box := Display floodFill: Color transparent at: p1.
	exterior := ((Display copy: box) makeBWForm: Color transparent) reverse.
	self world invalidRect: box; displayWorldSafely.
	(box area > (Display boundingBox area // 2))
		ifTrue: [^ UIManager default notify: 'Sorry, the region was too big'].
	(exterior deepCopy reverse anyShapeFill reverse)  "save interior bits"
		displayOn: exterior at: 0@0 rule: Form and.
	screenForm := Form fromDisplay: box.
	screenForm eraseShape: exterior.
	screenForm isAllWhite ifFalse:
		[evt hand attachMorph: (self drawingClass withForm: screenForm)]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'di 10/18/2001 02:58'!
grabLassoFromScreen: evt
	"Allow the user to specify a polygonal area of the Display, capture the pixels from that area, and use them to create a new drawing morph. Attach the result to the hand."

	self extractScreenRegion: (PolygonMorph fromHandFreehand: evt hand)
		andPutSketchInHand: evt hand
! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'di 10/18/2001 01:13'!
grabRubberBandFromScreen: evt
	"Allow the user to specify a polygonal area of the Display, capture the pixels from that area, and use them to create a new drawing morph. Attach the result to the hand."

	self extractScreenRegion: (PolygonMorph fromHand: evt hand)
		andPutSketchInHand: evt hand! !

!PasteUpMorph methodsFor: 'display' stamp: 'ar 9/7/2002 15:24'!
gradientFillColor: aColor
	"For backwards compatibility with GradientFillMorph"

	self flag: #fixThis.
	self useGradientFill.
	self fillStyle colorRamp: {0.0 -> self fillStyle colorRamp first value. 1.0 -> aColor}.
	self changed! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:15'!
griddingOn

	^ griddingOn ifNil: [false]! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 16:48'!
griddingOnOff

	griddingOn := self griddingOn not.
	self changed! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'dgd 12/13/2003 19:30'!
griddingString
	"Answer a string to use in a menu offering the user the 
	opportunity to start or stop using gridding"
	^ (self griddingOn
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'use gridding' translated! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:29'!
gridModulus

	^ self gridSpec extent! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 16:47'!
gridModulus: newModulus

	self gridSpecPut: (self gridOrigin extent: newModulus).
	self changed! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:28'!
gridOrigin

	^ self gridSpec origin! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:29'!
gridOrigin: newOrigin

	^ self gridSpecPut: (newOrigin extent: self gridModulus)! !

!PasteUpMorph methodsFor: 'geometry' stamp: 'di 8/28/2000 23:13'!
gridPoint: ungriddedPoint

	self griddingOn ifFalse: [^ ungriddedPoint].
	^ (ungriddedPoint - self position - self gridOrigin grid: self gridModulus)
					+ self position + self gridOrigin! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:26'!
gridSpec
	"Gridding rectangle provides origin and modulus"

	^ self valueOfProperty: #gridSpec ifAbsent: [0@0 extent: 8@8]! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 13:28'!
gridSpecPut: newSpec
	"Gridding rectangle provides origin and modulus"

	^ self setProperty: #gridSpec toValue: newSpec! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 16:11'!
gridVisible

	^ self hasProperty: #gridVisible! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'di 8/24/2000 16:47'!
gridVisibleOnOff

	self setProperty: #gridVisible toValue: self gridVisible not.
	self changed! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'sw 9/27/2007 04:38'!
gridVisibleString
	"Answer a string to be used in a menu offering the opportunity 
	to show or hide the grid"
	^ (self gridVisible
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'grid visible when gridding' translated! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 9/28/2000 18:00'!
haloMorphs
	^ self hands collect:[:h| h halo] thenSelect:[:halo| halo notNil]! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'topa 11/2/2015 10:02'!
handleDroppedItem: anItem event: anEvent
	
	(ExternalDropHandler lookupExternalDropHandler: anItem)
		ifNotNil: [:handler | handler handle: anItem in: self dropEvent: anEvent].! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'dgd 8/28/2004 18:44'!
handlesKeyboard: evt
	^self isWorldMorph or:[evt keyCharacter == Character tab and:[self tabAmongFields]]! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'ar 10/3/2000 22:46'!
handlesMouseDown: evt
	^true! !

!PasteUpMorph methodsFor: 'project state' stamp: 'di 7/27/1999 10:46'!
hands

	^ worldState hands! !

!PasteUpMorph methodsFor: 'project state' stamp: 'nk 7/4/2003 16:47'!
handsDo: aBlock

	^ worldState ifNotNil: [ worldState handsDo: aBlock ]! !

!PasteUpMorph methodsFor: 'project state' stamp: 'nk 7/4/2003 16:46'!
handsReverseDo: aBlock

	^ worldState ifNotNil: [ worldState handsReverseDo: aBlock ]! !

!PasteUpMorph methodsFor: 'Nebraska' stamp: 'ar 10/26/2000 15:03'!
hasRemoteServer
	^self remoteServer notNil! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'cmm 3/11/2010 22:42'!
hasTransferMorphConverter
	^ self transferMorphConverter ~= #yourself! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 10/8/1998 16:50'!
heightForThumbnails
	^ self valueOfProperty: #heightForThumbnails ifAbsent: [50]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 11/18/2001 18:36'!
hideFlapsOtherThan: aFlapTab ifClingingTo: anEdgeSymbol
	"Hide flaps on the given edge unless they are the given one"

	self flapTabs do:
		[:aTab | (aTab edgeToAdhereTo == anEdgeSymbol)
			ifTrue:
				[aTab  == aFlapTab
					ifFalse:
						[aTab hideFlap]]]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'ar 2/12/2001 23:17'!
hideViewerFlaps
	self flapTabs do:[:aTab |
		(aTab isKindOf: ViewerFlapTab) ifTrue:[aTab hideFlap]]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 6/25/1999 22:44'!
hideViewerFlapsOtherThanFor: aPlayer
	self flapTabs do:
		[:aTab | (aTab isKindOf: ViewerFlapTab)
			ifTrue:
				[aTab scriptedPlayer == aPlayer
					ifFalse:
						[aTab hideFlap]]]! !

!PasteUpMorph methodsFor: 'thumbnail' stamp: 'dgd 9/22/2004 19:37'!
icon
	"Answer a form with an icon to represent the receiver"
	^ self isWorldMorph
		ifTrue: [MenuIcons homeIcon]
		ifFalse: [MenuIcons projectIcon]! !

!PasteUpMorph methodsFor: 'options' stamp: 'sw 4/1/98 16:58'!
indicateCursor
	^ indicateCursor == true! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 9/6/2003 17:56'!
indicateCursorString
	"Answer the string to be shown in a menu to represent the  
	whether-to-indicate-cursor status"
	^ (self indicateCursor
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'indicate cursor' translated! !

!PasteUpMorph methodsFor: 'options' stamp: 'sw 4/1/98 16:59'!
indicateCursor: aBoolean
	indicateCursor := aBoolean! !

!PasteUpMorph methodsFor: 'world state' stamp: 'dgd 9/5/2004 16:19'!
initForProject: aWorldState

	worldState := aWorldState.
	bounds := Display boundingBox.
	self color: Preferences defaultWorldColor.
	self addHand: HandMorph new.
	self setProperty: #automaticPhraseExpansion toValue: true.
	self setProperty: #optimumExtentFromAuthor toValue: Display extent.
	self wantsMouseOverHalos: Preferences mouseOverHalos.
	self borderWidth: 0.
	model := nil.
! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'mt 5/31/2016 12:01'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	cursor := 1.
	padding := 3.
	self enableDragNDrop.
	self isWorldMorph
		ifTrue: [self setProperty: #automaticPhraseExpansion toValue: true].
	self clipSubmorphs: true.
	self initializeKeyboardShortcuts.! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'nice 12/27/2009 03:11'!
initializeDesktopCommandKeySelectors
	"Provide the starting settings for desktop command key selectors.  Answer the dictionary."

	"ActiveWorld initializeDesktopCommandKeySelectors"
	| dict |
	dict := IdentityDictionary new.
	self defaultDesktopCommandKeyTriplets do:
		[:trip | | messageSend |
			messageSend := MessageSend receiver: trip second selector: trip third.
			dict at: trip first put: messageSend].
	self setProperty: #commandKeySelectors toValue: dict.
	^ dict

! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'mt 5/31/2016 12:07'!
initializeKeyboardShortcuts
	"Install an event capture filter to add shortcuts for global operations like opening a tool."
	
	PasteUpMorph globalCommandKeysEnabled
		ifTrue: [self addKeyboardShortcuts].! !

!PasteUpMorph methodsFor: 'parts bin' stamp: 'sw 8/2/2001 17:50'!
initializeToStandAlone
	"Answer an instance of the receiver suitable for placing in a parts bin for authors"
	
	self initialize.
	self color: Color green muchLighter;  extent: 100 @ 80; borderColor: (Color r: 0.645 g: 0.935 b: 0.161).
	self extent: 300 @ 240.
	self beSticky! !

!PasteUpMorph methodsFor: 'misc' stamp: 'dgd 8/30/2003 15:52'!
innocuousName
	^ (self isFlap)
		ifTrue:
			['flap' translated]
		ifFalse:
			[super innocuousName]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'mt 4/17/2016 13:08'!
install
	owner := nil.	"since we may have been inside another world previously"
	ActiveWorld := self.
	ActiveHand := self hands first.	"default"
	ActiveEvent := nil.
	submorphs do: [:ss | ss owner isNil ifTrue: [ss privateOwner: self]].
	"Transcript that was in outPointers and then got deleted."
	self viewBox: Display boundingBox.
	Sensor flushAllButDandDEvents.
	worldState handsDo: [:h | h initForEvents].
	self installFlaps.
	self borderWidth: 0.	"default"
	(Preferences showSecurityStatus 
		and: [SecurityManager default isInRestrictedMode]) 
			ifTrue: 
				[self
					borderWidth: 2;
					borderColor: Color red].
	self presenter allExtantPlayers do: [:player | player prepareToBeRunning].
	SystemWindow noteTopWindowIn: self.! !

!PasteUpMorph methodsFor: 'world state' stamp: 'yo 7/12/2007 15:03'!
installFlaps
	"Get flaps installed within the bounds of the receiver"

	| localFlapTabs |
	Project current assureFlapIntegrity.
	self addGlobalFlaps.
	localFlapTabs := self localFlapTabs.
	localFlapTabs do: [:each | each visible: false].

	Preferences eToyFriendly ifTrue: [
		ProgressInitiationException display: 'Building Viewers...' translated
			during: [:bar |
				localFlapTabs keysAndValuesDo: [:i :each |
					each adaptToWorld.
					each visible: true.
					self displayWorld.
					bar value: i / self localFlapTabs size]].
	] ifFalse: [
		localFlapTabs keysAndValuesDo: [:i :each |
			each adaptToWorld.
			each visible: true.
			self displayWorld]].

	self assureFlapTabsFitOnScreen.
	self bringTopmostsToFront! !

!PasteUpMorph methodsFor: 'change reporting' stamp: 'ar 6/17/2010 17:28'!
invalidRect: damageRect from: aMorph
	"Clip damage reports to my bounds, since drawing is clipped to my bounds."
	self isWorldMorph
		ifTrue: [worldState recordDamagedRect: damageRect].
	^super invalidRect: damageRect from: aMorph! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 4/30/2001 21:03'!
invokeWorldMenu: evt
	"Put up the world menu, triggered by the passed-in event.  But don't do it if the eToyFriendly preference is set to true."

	Preferences eToyFriendly ifFalse:
		[self putUpWorldMenu: evt]! !

!PasteUpMorph methodsFor: 'testing' stamp: 'wiz 2/25/2006 00:09'!
isEasySelecting
"This is to isolate easySelection predicate. 
Selectors in holders make no sense so we are limiting easy selection to the worldMorph.
It would also make sense in playfield so feel free to adjust this predicate.  Selection can always be forced by using the shift before mouse down."

^ self isWorldMorph and: [  Preferences easySelection ]! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 9/6/2003 17:58'!
isOpenForDragNDropString
	"Answer the string to be shown in a menu to represent the  
	open-to-drag-n-drop status"
	^ (self dragNDropEnabled
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'open to drag & drop' translated! !

!PasteUpMorph methodsFor: 'parts bin' stamp: 'sw 4/13/1998 18:15'!
isPartsBin
	^ isPartsBin == true! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 9/6/2003 17:57'!
isPartsBinString
	"Answer the string to be shown in a menu to represent the 
	parts-bin status"
	^ (self isPartsBin
		ifTrue: ['<on>']
		ifFalse: ['<off>']), 'parts bin' translated! !

!PasteUpMorph methodsFor: 'options' stamp: 'sw 4/23/1998 16:49'!
isPartsBin: aBoolean
	isPartsBin := aBoolean! !

!PasteUpMorph methodsFor: 'classification' stamp: 'sw 1/29/98 21:50'!
isPlayfieldLike
	^ true! !

!PasteUpMorph methodsFor: 'project state' stamp: 'sw 10/9/1999 22:51'!
isStepping: aMorph
	^ worldState isStepping: aMorph! !

!PasteUpMorph methodsFor: 'project state' stamp: 'ar 10/22/2000 16:43'!
isStepping: aMorph selector: aSelector
	^ worldState isStepping: aMorph selector: aSelector! !

!PasteUpMorph methodsFor: 'classification' stamp: 'di 7/27/1999 10:46'!
isWorldMorph

	^ worldState notNil! !

!PasteUpMorph methodsFor: 'world state' stamp: 'dtl 11/23/2009 23:38'!
jumpToProject

	Project current jumpToProject.
! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'sw 9/1/2000 05:37'!
justDroppedInto: aMorph event: anEvent
	"This message is sent to a dropped morph after it has been dropped on--and been accepted by--a drop-sensitive morph"

	super justDroppedInto: aMorph event: anEvent.
	self isPartsBin ifTrue: [self setPartsBinStatusTo: true]  "gets some things right about the subtle case of dropping a parts bin"
! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 5/20/2003 15:04'!
keyboardNavigationHandler
	"Answer the receiver's existing keyboardNavigationHandler, or nil if none."

	| aHandler |
	aHandler := self valueOfProperty: #keyboardNavigationHandler ifAbsent: [^ nil].
	(aHandler hasProperty: #moribund) ifTrue:  "got clobbered in another project"
		[self removeProperty: #keyboardNavigationHander.
		^ nil].
	^ aHandler! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 3/18/2003 23:10'!
keyboardNavigationHandler: aHandler
	"Set the receiver's keyboard navigation handler as indicated.  A nil argument means to remove the handler"

	aHandler
		ifNil:
			[self removeProperty: #keyboardNavigationHandler]
		ifNotNil:
			[self setProperty: #keyboardNavigationHandler toValue: aHandler]! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'dgd 4/4/2006 14:42'!
keyStroke: anEvent
	"A keystroke has been made.  Service event handlers and, if it's a keystroke presented to the world, dispatch it to #unfocusedKeystroke:"

	| selected |
	super keyStroke: anEvent.  "Give event handlers a chance"

	selected := self selectedObject.
	selected isNil
		ifFalse:[ selected moveOrResizeFromKeystroke: anEvent ].

	(anEvent keyCharacter == Character tab) ifTrue:
		[self tabAmongFields
			ifTrue:[^ self tabHitWithEvent: anEvent]].
	self isWorldMorph ifTrue:
		[self keystrokeInWorld: anEvent]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'cmm 10/19/2015 16:23:45'!
keystrokeInWorld: evt
	"A keystroke was hit when no keyboard focus was set, so it is sent here to the world instead."

	|  aChar isCmd ascii |
	aChar := evt keyCharacter.
	(ascii := aChar asciiValue) = Character escape asciiValue ifTrue:
		[evt commandKeyPressed ifFalse: [^ self putUpWorldMenuFromEscapeKey]].
	(evt controlKeyPressed not
		and: [(#(1 4 8 28 29 30 31 32) includes: ascii)  "home, end, backspace, arrow keys, space"
			and: [self keyboardNavigationHandler notNil]])
				ifTrue: [self keyboardNavigationHandler navigateFromKeystroke: aChar].

	isCmd := evt commandKeyPressed and: [Preferences cmdKeysInText].
	(evt commandKeyPressed and: [Preferences eToyFriendly])
			ifTrue:
				[(aChar == $W) ifTrue: [^ self putUpWorldMenu: evt]].
	(isCmd and: [Preferences honorDesktopCmdKeys]) ifTrue:
		[^ self dispatchCommandKeyInWorld: aChar event: evt].

	"It was unhandled. Remember the keystroke."
	self lastKeystroke: evt keyString.
	self triggerEvent: #keyStroke! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'nk 10/13/2004 11:26'!
lastKeystroke
	"Answer the last keystroke fielded by the receiver"

	^ self valueOfProperty: #lastKeystroke ifAbsent: ['']! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'nk 10/13/2004 11:27'!
lastKeystroke: aString
	"Remember the last keystroke fielded by the receiver"

	^ self setProperty: #lastKeystroke toValue: aString! !

!PasteUpMorph methodsFor: 'layout' stamp: 'ar 11/9/2000 13:43'!
layoutChanged
	"The receiver's layout changed; inform above and below"
	super layoutChanged.
	(self valueOfProperty: #SqueakPage) ifNotNil: [
		self setProperty: #pageDirty toValue: true].
		"I am the morph of a SqueakPage, I have changed and 
		need to be written out again"
! !

!PasteUpMorph methodsFor: 'layout' stamp: 'sw 8/3/1998 13:43'!
laySubpartsOutInOneRow
	| aPosition |
	aPosition := 0 @ padding.
	submorphs do:
	[:aMorph |
		aMorph position: (aPosition + (padding @ 0)).
		aPosition := aMorph topRight]! !

!PasteUpMorph methodsFor: 'project state' stamp: 'sw 9/5/2000 06:45'!
listOfSteppingMorphs
	^ worldState listOfSteppingMorphs

"self currentWorld listOfSteppingMorphs"! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'nice 12/27/2009 20:14'!
localFlapTabs
	"Answer a list of local flap tabs in the current project"

	| globalList aList |
	globalList := Flaps globalFlapTabsIfAny.
	aList := OrderedCollection new.
	submorphs do:
		[:m | | aFlapTab |
		((m isFlapTab) and: [(globalList includes: m) not])
			ifTrue:
				[aList add: m]
			ifFalse:
				[((m isFlap) and:
					[(aFlapTab := m submorphs detect: [:n | n isFlapTab] ifNone: [nil]) notNil])
						ifTrue:
							[aList add: aFlapTab]]].
	^ aList! !

!PasteUpMorph methodsFor: 'painting' stamp: 'sw 3/24/2001 23:58'!
makeNewDrawingWithin
	"Start a painting session in my interior which will result in a new SketchMorph being created as one of my submorphs"

	| evt |
	evt := MouseEvent new setType: nil position: self center buttons: 0 hand: self world activeHand.
	self makeNewDrawing: evt! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 2/8/2001 19:26'!
makeNewDrawing: evt
	^self makeNewDrawing: evt at: evt position! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 1/3/2010 16:37'!
makeNewDrawing: evt at: aPoint
	"make a new drawing, triggered by the given event, with the painting area centered around the given point"

	| w newSketch newPlayer sketchEditor aPalette rect aPaintBox aPaintTab aWorld |
	w := self world.
	w assureNotPaintingElse: [^ self].
	rect := self paintingBoundsAround: aPoint.
	aPalette := self standardPalette.
	aPalette ifNotNil: [aPalette showNoPalette; layoutChanged].
	w prepareToPaint.

	newSketch := self drawingClass new.
	Smalltalk at: #UnscriptedPlayer ifPresent:[:aClass|
		newSketch player: (newPlayer := aClass newUserInstance).
		newPlayer costume: newSketch.
	].
	newSketch nominalForm: (Form extent: rect extent depth: w assuredCanvas depth).
	newSketch bounds: rect.
	sketchEditor := SketchEditorMorph new.
	w addMorphFront: sketchEditor.
	sketchEditor initializeFor: newSketch inBounds: rect pasteUpMorph: self.
	sketchEditor
		afterNewPicDo: [:aForm :aRect | | tfx ownerBeforeHack whereToPresent |
			whereToPresent := self presenter.
			newSketch form: aForm.
			tfx := self transformFrom: w.
			newSketch position: (tfx globalPointToLocal: aRect origin).
			newSketch rotationStyle: sketchEditor rotationStyle.
			newSketch forwardDirection: sketchEditor forwardDirection.

			ownerBeforeHack := newSketch owner.	"about to break the invariant!!!!"
			newSketch privateOwner: self. "temp for halo access"
			newPlayer ifNotNil:[newPlayer setHeading: sketchEditor forwardDirection].
			(aPaintTab := (aWorld := self world) paintingFlapTab)
				ifNotNil:[aPaintTab hideFlap]
				ifNil:[(aPaintBox := aWorld paintBox) ifNotNil:[aPaintBox delete]].

			"Includes  newSketch rotationDegrees: sketchEditor forwardDirection."
			newSketch privateOwner: ownerBeforeHack. "probably nil, but let's be certain"

			self addMorphFront: (newPlayer ifNil:[newSketch] ifNotNil:[newPlayer costume]).
			w startSteppingSubmorphsOf: newSketch.
			whereToPresent drawingJustCompleted: newSketch]
		 ifNoBits:[
			(aPaintTab := (aWorld := self world) paintingFlapTab)
				ifNotNil:[aPaintTab hideFlap]
				ifNil:[(aPaintBox := aWorld paintBox) ifNotNil:[aPaintBox delete]].
			aPalette ifNotNil: [aPalette showNoPalette].]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 10/8/1998 16:50'!
maxHeightToAvoidThumbnailing
	^ self valueOfProperty: #maxHeightToAvoidThumbnailing ifAbsent: [80]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 11/13/1998 10:06'!
maximumThumbnailWidth
	^ self valueOfProperty: #maximumThumbnailWidth ifAbsent: [200 min: (self width - 10)]! !

!PasteUpMorph methodsFor: 'polymorph' stamp: 'gvc 1/12/2007 16:57'!
modalLockTo: aSystemWindow
	"Don't lock the world!! Lock the submorphs.
	The modal window gets opened afterwards so is OK."
	
	|lockStates|
	lockStates := IdentityDictionary new.
	self submorphsDo: [:m |
		lockStates at: m put: m isLocked.
		m lock].
	self
		setProperty: #submorphLockStates
		toValue: lockStates! !

!PasteUpMorph methodsFor: 'polymorph' stamp: 'gvc 1/12/2007 17:01'!
modalUnlockFrom: aSystemWindow
	"Don't unlock the world!! Unlock the submorphs
	that were not originally locked."
	
	|lockStates|
	lockStates := self
		valueOfProperty: #submorphLockStates
		ifAbsent: [^self].
	self removeProperty: #submorphLockStates.
	lockStates keysAndValuesDo: [:m :locked |
		locked ifFalse: [m unlock]]! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'ul 12/12/2009 14:08'!
modalWindow: aMorph 
	(self valueOfProperty: #modalWindow)
		ifNotNil: [:morph | morph doCancel].
	self setProperty: #modalWindow toValue: aMorph.
	aMorph
		ifNotNil: [self
				when: #aboutToLeaveWorld
				send: #removeModalWindow
				to: self]! !

!PasteUpMorph methodsFor: 'model' stamp: 'sw 4/1/98 21:18'!
model
	"Return the model object for this world. If the world has no model, then create one."

	self createCustomModel.
	^ model! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'sw 4/1/98 21:18'!
modelOrNil
	"Return the model object for this world, or nil if it doesn't have one."

	^ model
! !

!PasteUpMorph methodsFor: 'user interface' stamp: 'dgd 2/22/2003 14:11'!
modelWakeUp
	"I am the model of a SystemWindow, that has just been activated"

	| aWindow |
	owner isNil ifTrue: [^self].	"Not in Morphic world"
	(owner isKindOf: TransformMorph) ifTrue: [^self viewBox: self fullBounds].
	(aWindow := self containingWindow) ifNotNil: 
			[self viewBox = aWindow panelRect 
				ifFalse: [self viewBox: aWindow panelRect]]! !

!PasteUpMorph methodsFor: 'WiW support' stamp: 'RAA 10/3/2000 09:24'!
morphicLayerNumber

	self isFlap ifTrue:[^26]. 	"As navigators"
	^super morphicLayerNumber.! !

!PasteUpMorph methodsFor: 'submorphs-accessing' stamp: 'nk 7/4/2003 16:49'!
morphsInFrontOf: someMorph overlapping: aRectangle do: aBlock
	"Include hands if the receiver is the World"
	self handsDo:[:m|
		m == someMorph ifTrue:["Try getting out quickly"
			owner ifNil:[^self].
			^owner morphsInFrontOf: self overlapping: aRectangle do: aBlock].
		"The hand only overlaps if it's not the hardware cursor"
		m needsToBeDrawn ifTrue:[
			(m fullBoundsInWorld intersects: aRectangle)
				ifTrue:[aBlock value: m]]].
	^super morphsInFrontOf: someMorph overlapping: aRectangle do: aBlock! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'cmm 3/11/2010 22:41'!
morphToDropForTransferMorph: aTransferMorph 
	"aTransferMorph has been dragged directly onto the desktop.  TransferMorphs, by nature, are about transferring a 'logical object' from another source. Answer the Morph that should actually be dropped on to the desktop."
	^ self
		perform: self transferMorphConverter
		with: aTransferMorph! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'cmm 3/11/2010 21:40'!
morphToDropFrom: aMorph 
	"Given a morph being carried by the hand, which the hand is about to drop, answer the actual morph to be deposited.  Normally this would be just the morph itself, but several unusual cases arise, which this method is designed to service."

	| aNail representee handy posBlock |
	handy := self primaryHand.
	posBlock := 
			[:z | | tempPos | 
			tempPos := handy position 
						- ((handy targetOffset - aMorph formerPosition) 
								* (z extent / aMorph extent)) rounded.
			self pointFromWorld: tempPos].
	self alwaysShowThumbnail 
		ifTrue: 
			[aNail := aMorph 
						representativeNoTallerThan: self maxHeightToAvoidThumbnailing
						norWiderThan: self maximumThumbnailWidth
						thumbnailHeight: self heightForThumbnails.
			aNail == aMorph 
				ifFalse: 
					[aMorph formerPosition: aMorph position.
					aNail position: (posBlock value: aNail)].
			^aNail].
	((aMorph isKindOf: MorphThumbnail) 
		and: [(representee := aMorph morphRepresented) owner isNil]) 
			ifTrue: 
				[representee position: (posBlock value: representee).
				^representee].
	self showingListView 
		ifTrue: 
			[^aMorph 
				listViewLineForFieldList: (self valueOfProperty: #fieldListSelectors)].
	(aMorph hasProperty: #newPermanentScript) 
		ifTrue: [^aMorph asEmptyPermanentScriptor].
	((aMorph isPhraseTileMorph) or: [aMorph isSyntaxMorph]) 
		ifFalse: [^aMorph morphToDropInPasteUp: self].
	aMorph userScriptSelector isEmptyOrNil 
		ifTrue: 
			["non-user"

			self automaticPhraseExpansion ifFalse: [^aMorph]].
	^aMorph morphToDropInPasteUp: self! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'ar 2/23/2001 16:44'!
morphToGrab: event
	"Return the morph to grab from a mouse down event. If none, return nil."
	self submorphsDo:[:m|
		((m rejectsEvent: event) not and:[m fullContainsPoint: event cursorPoint]) ifTrue:[^m].
	].
	^nil! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'mtf 2/20/2011 20:40'!
mouseDown: evt
	"Handle a mouse down event."
	| grabbedMorph handHadHalos |

	(Preferences generalizedYellowButtonMenu
			and: [evt yellowButtonPressed])
		ifTrue: [^ self yellowButtonActivity: evt shiftPressed].

	grabbedMorph := self morphToGrab: evt.
	grabbedMorph ifNotNil:[
		grabbedMorph isSticky ifTrue:[^self].
		self isPartsBin ifFalse:[^evt hand grabMorph: grabbedMorph].
		grabbedMorph := grabbedMorph partRepresented duplicate.
		grabbedMorph restoreSuspendedEventHandler.
		(grabbedMorph fullBounds containsPoint: evt position) 
			ifFalse:[grabbedMorph position: evt position].
		"Note: grabbedMorph is ownerless after duplicate so use #grabMorph:from: instead"
		^ evt hand grabMorph: grabbedMorph from: self].

	(super handlesMouseDown: evt)
		ifTrue:[^super mouseDown: evt].

	handHadHalos := evt hand halo notNil.

	evt hand removeHalo. "shake off halos"
	evt hand releaseKeyboardFocus. "shake of keyboard foci"

	self submorphs
		select:[:each | each hasProperty: #morphHierarchy]
		thenDo:[:each | each delete].

	Preferences noviceMode
		ifTrue:[
			self submorphs
				select:[:each | (each isKindOf: MenuMorph) and:[each stayUp not]]
				thenDo:[:each | each delete].
		].

	(evt shiftPressed not
			and:[ self isWorldMorph not 
			and:[ self wantsEasySelection not ]])
	ifTrue:[
		"explicitly ignore the event if we're not the world and we'll not select,
		so that we could be picked up if need be"
		evt wasHandled: false.
		^ self.
	].

	( evt shiftPressed or: [ self wantsEasySelection ] ) ifTrue:[
		"We'll select on drag, let's decide what to do on click"
		| clickSelector |

		clickSelector := nil.

		evt shiftPressed ifTrue:[
			clickSelector := #findWindow:.
		]
		ifFalse:[
			self isWorldMorph ifTrue:[
				clickSelector := handHadHalos
										ifTrue: [ #delayedInvokeWorldMenu: ]
										ifFalse: [ #invokeWorldMenu: ]
			]
		].

		evt hand 
				waitForClicksOrDrag: self 
				event: evt 
				selectors: { clickSelector. nil. nil. #dragThroughOnDesktop: }
				threshold: HandMorph dragThreshold.
	]
	ifFalse:[
		"We wont select, just bring world menu if I'm the world"
		self isWorldMorph ifTrue:[
			handHadHalos
				ifTrue: [ self delayedInvokeWorldMenu: evt ]
				ifFalse: [ self invokeWorldMenu: evt ]
		]
	].
! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 9/6/2003 17:58'!
mouseOverHalosString
	"Answer the string to be shown in a menu to represent the  
	mouse-over-halos status"
	^ (self wantsMouseOverHalos
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'mouse-over halos' translated! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'ar 10/6/2000 00:04'!
mouseUp: evt
	self isWorldMorph ifTrue:[self removeAlarm: #invokeWorldMenu:].
	super mouseUp: evt.! !

!PasteUpMorph methodsFor: 'misc' stamp: 'yo 2/17/2005 16:58'!
mouseX
	"Answer the x-coordinate of the mouse, in my coordinate system"

	^ self isInWorld
		ifTrue:
			[((self pointFromWorld: self cursorPoint) x) - self cartesianOrigin x]
		ifFalse:
			[0]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'yo 2/17/2005 16:58'!
mouseY
	"Answer the y-coordinate of the mouse, in my coordinate system"

	^ self isInWorld
		ifTrue:
			[self cartesianOrigin y - ((self pointFromWorld: self cursorPoint) y)]
		ifFalse:
			[0]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 8/16/2000 17:42'!
nameForCopyIfAlreadyNamed: aMorph
	"Answer a name to set for a copy of aMorph if aMorph itself is named, else nil"

	| aName usedNames |
	^ (aName := aMorph knownName) ifNotNil:
		[usedNames := self allKnownNames.
		Utilities keyLike: aName satisfying: [:f | (usedNames includes: f) not]]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/5/2000 17:40'!
newDrawingFromMenu: evt
	self assureNotPaintingElse: [^ self].
	evt hand attachMorph: PaintInvokingMorph new markAsPartsDonor! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'ar 3/3/2001 15:30'!
newResourceLoaded
	"Some resource has just been loaded. Notify all morphs in case somebody wants to update accordingly."
	self allMorphsDo:[:m| m resourceJustLoaded ].
	self fullRepaintNeeded.! !

!PasteUpMorph methodsFor: 'world state' stamp: 'nb 6/17/2003 12:25'!
nextPage
	"backstop for smart next-page buttons that look up the containment hierarchy until they find somone who is willing to field this command.  If we get here, the 'next' button was not embedded in a book, so we can do nothing useful"

	Beeper beep! !

!PasteUpMorph methodsFor: 'world state' stamp: 'mt 7/28/2016 15:02'!
noDisplayDuring: block
	"Some morphs insist on updating the display like the system progress bar does. However, some code might be in an inconsistent state. For this, you can use this method to be safe."

	self
		setProperty: #shouldDisplayWorld
		toValue: false.
		
	block ensure: [self removeProperty: #shouldDisplayWorld].! !

!PasteUpMorph methodsFor: 'cursor' stamp: 'sw 9/8/2000 16:41'!
numberAtCursor
	"Answer the number represented by the object at my current cursor position"

	| chosenMorph |
	submorphs isEmpty ifTrue: [^ 0].
	chosenMorph := submorphs at: ((cursor truncated max: 1) min: submorphs size).
	^ chosenMorph getNumericValue
! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'sw 2/11/1999 10:53'!
offsetForAccommodating: anExtent onEdge: edgeSymbol
	"Answer a delta to be applied to my submorphs in order tfor anExtent to be slid inboard on the indicated edge"
	edgeSymbol == #left ifTrue: [^ anExtent x @ 0].
	edgeSymbol == #right ifTrue: [^ anExtent x negated @ 0].
	edgeSymbol == #top ifTrue: [^ 0 @ anExtent y].
	edgeSymbol == #bottom ifTrue: [^ 0 @ anExtent y negated].! !

!PasteUpMorph methodsFor: 'undo' stamp: 'fbs 12/31/2013 12:08'!
onceAgainDismiss: aMorph
	"Occasioned by a redo of a dismiss-via-halo"

	aMorph dismissMorph.
	TrashCanMorph preserveTrash ifTrue: 
		[TrashCanMorph slideDismissalsToTrash
			ifTrue:[aMorph slideToTrash: nil]
			ifFalse:[TrashCanMorph moveToTrash: aMorph]]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'fbs 11/25/2013 08:41'!
openScrapsBook: evt
	"Open up the Scraps book in the center of the screen"

	evt hand world addMorphCentered: ScrapBook default scrapBook! !

!PasteUpMorph methodsFor: 'world state' stamp: 'nice 12/27/2009 03:11'!
optimumExtentFromAuthor

	
	^self 
		valueOfProperty: #optimumExtentFromAuthor 
		ifAbsent: [ | opt |
			opt := bounds extent.
			self setProperty: #optimumExtentFromAuthor toValue: opt.
			^opt
		]

! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'sw 7/6/1999 13:26'!
originAtCenter
	^ self hasProperty: #originAtCenter! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 9/6/2003 18:00'!
originAtCenterString
	"Answer the string to be shown in a menu to represent the 
	origin-at-center status"
	^ ((self hasProperty: #originAtCenter)
		ifTrue: ['<on>']
		ifFalse: ['<off>']), 'origin-at-center' translated! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 4/23/1998 18:50'!
padding: aNumber
	padding := aNumber! !

!PasteUpMorph methodsFor: 'world state' stamp: 'wiz 12/4/2006 00:32'!
paintArea
	"What rectangle should the user be allowed to create a new painting in??
	An area beside the paintBox. Allow playArea to override with its own
	bounds!! "
	| playfield paintBoxBounds |
	playfield := self
				submorphNamed: 'playfield'
				ifNone: [].
	playfield
		ifNotNil: [^ playfield bounds].
	paintBoxBounds := self paintBox bounds.
	self firstHand targetPoint x < paintBoxBounds center x
		ifTrue: [^ bounds topLeft corner: paintBoxBounds left @ bounds bottom"paint on left side"]
		ifFalse: [^ paintBoxBounds right @ bounds top corner: bounds bottomRight]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'dgd 2/22/2003 19:01'!
paintAreaFor: aSketchMorph 
	"Answer the area to comprise the onion-skinned canvas for painting/repainting aSketchMorph"

	| itsOwner |
	((itsOwner := aSketchMorph owner) notNil and: [itsOwner isPlayfieldLike]) 
		ifTrue: [^itsOwner bounds].	"handles every plausible situation"
	^self paintArea! !

!PasteUpMorph methodsFor: 'world state' stamp: 'tk 8/21/2000 15:36'!
paintBox
	"Return the painting controls widget (PaintBoxMorph) to be used for painting in this world. If there is not already a PaintBox morph, or if it has been deleted from this world, create a new one."

	| newPaintBox refPoint aPalette |
	self allMorphsDo: [:m | (m isKindOf: PaintBoxMorph) ifTrue: [^ m]].
	refPoint := (aPalette := self standardPalette)
		ifNotNil:
			[aPalette showNoPalette.
			aPalette topRight + (0 @ 12)]
		ifNil:
			[self topRight].
	newPaintBox := PaintBoxMorph new.
	newPaintBox position: (refPoint - (newPaintBox width @ 0)). 
	self addMorph: newPaintBox.
	^ newPaintBox
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'sw 9/2/1999 12:01'!
paintBoxOrNil
	"Return the painting controls widget (PaintBoxMorph) to be used for painting in this world. If there is not already a PaintBox morph return nil"

	self allMorphsDo: [:m | (m isKindOf: PaintBoxMorph) ifTrue: [^ m]].
	^ nil
! !

!PasteUpMorph methodsFor: 'painting' stamp: 'sw 9/22/1998 12:26'!
paintingBoundsAround: aPoint
	"Return a rectangle for painting centered on the given point. Both the argument point and the result rectangle are in world coordinates."

	| paintExtent maxPaintArea myBnds |
	paintExtent := self reasonablePaintingExtent.
	maxPaintArea := paintExtent x * paintExtent y.
	myBnds := self boundsInWorld.
	(myBnds area <= maxPaintArea) ifTrue: [^ myBnds].
	^ (aPoint - (paintExtent // 2) extent: paintExtent) intersect: myBnds
! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'sw 6/25/1999 21:35'!
paintingFlapTab
	"If the receiver has a flap which has a paintbox, return it, else return nil"
	self flapTabs do:
		[:aTab | aTab referent submorphsDo:
			[:aMorph | (aMorph isKindOf: PaintBoxMorph) ifTrue: [^ aTab]]].
	^ nil! !

!PasteUpMorph methodsFor: 'world state' stamp: 'nk 7/7/2003 11:15'!
patchAt: patchRect without: stopMorph andNothingAbove: stopThere
	"Return a complete rendering of this patch of the display screen
	without stopMorph, and possibly without anything above it."

	| c |
	c := ColorPatchCanvas
		extent: patchRect extent
		depth: Display depth
		origin: patchRect topLeft negated
		clipRect: (0@0 extent: patchRect extent).
	c stopMorph: stopMorph.
	c doStop: stopThere.

	(self bounds containsRect: patchRect) ifFalse:
		["Need to fill area outside bounds with black."
		c form fillColor: Color black].
	(self bounds intersects: patchRect) ifFalse:
		["Nothing within bounds to show."
		^ c form].
	self fullDrawOn: c.
	stopThere ifFalse: [ self world handsReverseDo: [:h | h drawSubmorphsOn: c]].
	^c form
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'nice 12/27/2009 03:11'!
pauseEventRecorder
	"Suspend any event recorder, and return it if found"

	
	worldState handsDo: [:h | | er | (er := h pauseEventRecorderIn: self) ifNotNil: [^ er]].
	^ nil! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'sw 5/2/1998 11:31'!
positionNear: aPoint forExtent: anExtent adjustmentSuggestion: adjustmentPoint
	"Compute a plausible positioning for adding a subpart of size anExtent, somewhere near aPoint, using adjustmentPoint as the unit of adjustment"

	| adjustedPosition |
	adjustedPosition := aPoint.
	[((self morphsAt: (adjustedPosition + (anExtent // 2))) size > 1) and:  "that 1 is self here"
		[bounds containsPoint: adjustedPosition]]
	whileTrue:
		[adjustedPosition := adjustedPosition + adjustmentPoint].

	^ adjustedPosition! !

!PasteUpMorph methodsFor: 'geometry' stamp: 'RAA 6/1/2000 10:28'!
position: aPoint
	"Prevent moving a world (e.g. via HandMorph>>specialGesture:)"

	"for now, let's allow it and see what happens"

	self isWorldMorph ifFalse: [^super position: aPoint].
	super position: aPoint.
	self viewBox ifNotNil: [self viewBox: (aPoint extent: self viewBox extent)].

! !

!PasteUpMorph methodsFor: 'misc' stamp: 'ar 1/11/2010 19:55'!
prepareToBeSaved
	"Prepare for export via the ReferenceStream mechanism"

	| exportDict soundKeyList players |
	super prepareToBeSaved.
	turtlePen := nil.
	self isWorldMorph
		ifTrue:
			[self removeProperty: #scriptsToResume.
			soundKeyList := Set new.
			(players := self presenter allExtantPlayers)
				do: [:aPlayer | aPlayer slotInfo
						associationsDo: [:assoc | assoc value type == #Sound
								ifTrue: [soundKeyList
										add: (aPlayer instVarNamed: assoc key)]]].
			players
				do: [:p | p allScriptEditors
						do: [:e | (e allMorphs
								select: [:m | m isSoundTile])
								do: [:aTile | soundKeyList add: aTile literal]]].
			(self allMorphs
				select: [:m | m isSoundTile])
				do: [:aTile | soundKeyList add: aTile literal].
			soundKeyList removeAllFoundIn: SampledSound universalSoundKeys.
			soundKeyList
				removeAllSuchThat: [:aKey | (SampledSound soundLibrary includesKey: aKey) not].
			soundKeyList isEmpty
				ifFalse: [exportDict := Dictionary new.
					soundKeyList
						do: [:aKey | exportDict
								add: (SampledSound soundLibrary associationAt: aKey)].
					self setProperty: #soundAdditions toValue: exportDict]]! !

!PasteUpMorph methodsFor: 'painting' stamp: 'sw 10/4/2004 05:06'!
prepareToPaint
	"We're about to start painting. Do a few preparations that make the system more responsive."

	^ self prepareToPaint: Preferences keepTickingWhilePainting not! !

!PasteUpMorph methodsFor: 'painting' stamp: 'sw 5/1/2005 00:32'!
prepareToPaint: stopRunningScripts
	"We're about to start painting. Do a few preparations that make the system more responsive."

	self hideViewerFlaps. "make room"
	stopRunningScripts ifTrue:
		[self setProperty: #scriptsToResume toValue: self presenter allCurrentlyTickingScriptInstantiations.  "We'll restart these when painting is done"
		self stopRunningAll]. "stop scripts"
	self abandonAllHalos. "no more halos"! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'kfr 5/17/2011 17:16'!
presentCardAndStackMenu
	"Put up a menu holding card/stack-related options."

	| aMenu |
	aMenu _ MenuMorph new defaultTarget: self.		
	aMenu addStayUpItem.
	aMenu addTitle: 'card and stack' translated.
	aMenu add: 'add new card' translated action: #insertCard.
	aMenu add: 'delete this card' translated action: #deleteCard.
	aMenu add: 'go to next card' translated action: #goToNextCardInStack.
	aMenu add: 'go to previous card' translated action: #goToPreviousCardInStack.
	aMenu addLine.
	aMenu add: 'show foreground objects' translated action: #showForegroundObjects.
	aMenu add: 'show background objects' translated action: #showBackgroundObjects.
	aMenu add: 'show designations' translated action: #showDesignationsOfObjects.
	aMenu add: 'explain designations'  translated action: #explainDesignations.
	aMenu popUpInWorld: (self world ifNil: [self currentWorld])! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'ar 1/3/2010 18:11'!
presenter
	"Normally only the world will have a presenter, but the architecture supports individual localized presenters as well"

	^ presenter ifNil:
		[self isWorldMorph
			ifTrue: [presenter := Presenter defaultPresenterClass new associatedMorph: self]
			ifFalse: [super presenter]]! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'RAA 6/12/2000 09:14'!
presentPlayfieldMenu

	self playfieldOptionsMenu popUpForHand: self activeHand in: self world! !

!PasteUpMorph methodsFor: 'world state' stamp: 'nb 6/17/2003 12:25'!
previousPage
	"backstop for smartprev-page buttons that look up the containment hierarchy until they find somone who is willing to field this command.  If we get here, the button was not embedded in a book, so we can do nothing useful"

	Beeper beep! !

!PasteUpMorph methodsFor: 'printing' stamp: 'sw 10/18/2000 10:54'!
printOn: aStream
	"Reimplemented to add a tag showing that the receiver is currently functioning as a 'world', if it is"

	super printOn: aStream.
	self isWorldMorph ifTrue: [aStream nextPutAll: ' [world]']! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 1/12/2001 20:02'!
printScriptSummary
	"Put up a window with summaries of all scripts in the world"

	self presenter reportPlayersAndScripts

"self currentWorld printScriptSummary"! !

!PasteUpMorph methodsFor: 'private' stamp: 'ar 3/14/2000 23:20'!
privateFullMoveBy: delta
	"Private. Overridden to prevent drawing turtle trails when a playfield is moved"
	self setProperty: #turtleTrailsDelta toValue: delta.
	super privateFullMoveBy: delta.
	self removeProperty: #turtleTrailsDelta.
! !

!PasteUpMorph methodsFor: 'private' stamp: 'RAA 6/1/2000 14:23'!
privateMoveBy: delta

	super privateMoveBy: delta.
	worldState ifNotNil: [
		worldState viewBox ifNotNil: [
			worldState viewBox: bounds
		].
	].! !

!PasteUpMorph methodsFor: 'private' stamp: 'nk 7/8/2003 09:18'!
privateRemoveMorph: aMorph
	backgroundMorph == aMorph ifTrue: [ backgroundMorph := nil ].
	^super privateRemoveMorph: aMorph.
! !

!PasteUpMorph methodsFor: 'events-processing' stamp: 'mt 5/21/2016 09:29'!
processEvent: anEvent using: defaultDispatcher
	"Reimplemented to install the receiver as the new ActiveWorld if it is one"
	| priorWorld result |
	self isWorldMorph ifFalse:[^super processEvent: anEvent using: defaultDispatcher].
	priorWorld := ActiveWorld.
	ActiveWorld := self.
	[result := super processEvent: anEvent using: defaultDispatcher]
		ensure: [ActiveWorld := priorWorld].
	^result
! !

!PasteUpMorph methodsFor: 'project' stamp: 'tk 9/3/1999 12:07'!
project
	"Find the project that owns me.  Not efficient to call this."

	^ Project ofWorld: self! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 3/13/2003 11:51'!
putUpNewMorphMenu
	"Put up the New Morph menu in the world"

	TheWorldMenu new adaptToWorld: self; newMorph! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'sw 3/3/2004 01:17'!
putUpPenTrailsSubmenu
	"Put up the pen trails menu"

	| aMenu |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu title: 'pen trails' translated.
	aMenu addStayUpItem.
	self addPenTrailsMenuItemsTo: aMenu.
	aMenu popUpInWorld: ActiveWorld! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'dgd 9/20/2004 20:17'!
putUpWorldMenuFromEscapeKey
	Preferences noviceMode
		ifFalse: [self putUpWorldMenu: ActiveEvent]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'dgd 8/31/2004 16:25'!
putUpWorldMenu: evt
	"Put up a menu in response to a click on the desktop, triggered by evt."

	| menu |
	self bringTopmostsToFront.
	evt isMouse ifTrue:
		[evt yellowButtonPressed
			ifTrue: [^ self yellowButtonClickOnDesktopWithEvent: evt].
		evt shiftPressed ifTrue:[^ self findWindow: evt]].
	"put up screen menu"
	menu := self buildWorldMenu: evt.
	menu addTitle: Preferences desktopMenuTitle translated.
	menu popUpEvent: evt in: self.
	^ menu! !

!PasteUpMorph methodsFor: 'painting' stamp: 'sw 9/29/1998 07:35'!
reasonablePaintingExtent
	^ Preferences unlimitedPaintArea
		ifTrue:
			[3000 @ 3000]
		ifFalse:
			[Preferences defaultPaintingExtent]! !

!PasteUpMorph methodsFor: 'cursor' stamp: 'sw 5/12/1998 10:55'!
rectifyCursor
	cursor := ((cursor truncated max: 1) min: submorphs size)
! !

!PasteUpMorph methodsFor: 'e-toy support' stamp: 'nk 9/8/2003 17:17'!
referencePlayfield
	"Answer a pasteup morph to be used as the reference for cartesian coordinates.
	Do not get fooled by other morphs (like viewers) that happen to be named 'playfield'."

	^self isWorldMorph
		ifTrue: [ self submorphThat: [ :s | (s knownName = 'playfield') and: [ s isPlayfieldLike] ] ifNone: [self]]
		ifFalse: [ super referencePlayfield ]! !

!PasteUpMorph methodsFor: 'objects from disk' stamp: 'tfel 8/2/2016 14:11'!
referencePool
	^ self 
		valueOfProperty: #References 
		ifAbsentPut: [WeakValueDictionary new]
! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'sw 4/23/2001 12:33'!
reformulateUpdatingMenus
	"Give any updating menu morphs in the receiver a fresh kiss of life"

	(self submorphs select: [:m | m isKindOf: UpdatingMenuMorph]) do:
		[:m | m updateMenu] 

	"NB: to do the perfect job here one might well want to extend across allMorphs here, but the expense upon project entry is seemingly too high a price to pay at this point"! !

!PasteUpMorph methodsFor: 'undo' stamp: 'ul 12/12/2009 14:08'!
reintroduceIntoWorld: aMorph
	"The given morph is being raised from the dead.  Bring it back to life."

	(aMorph valueOfProperty: #lastPosition) ifNotNil:
		[:pos | aMorph position: pos].
	aMorph openInWorld; goHome

	! !

!PasteUpMorph methodsFor: 'caching' stamp: 'tak 3/15/2005 17:33'!
releaseCachedState
	super releaseCachedState.
	self removeModalWindow.
	presenter ifNotNil:[presenter flushPlayerListCache].
	self isWorldMorph ifTrue:[self cleanseStepList].! !

!PasteUpMorph methodsFor: 'Nebraska' stamp: 'ar 10/26/2000 14:44'!
releaseRemoteServer
	"My server has been transferred to some other world. Release pending references"
	^worldState releaseRemoteServer.! !

!PasteUpMorph methodsFor: 'project' stamp: 'nice 12/27/2009 20:16'!
releaseSqueakPages
	
	"If this world has a book with SqueakPages, then clear the SqueakPageCache"

	submorphs do: [:sub | | uu |
		(sub isKindOf: BookMorph) ifTrue: [
		uu := sub valueOfProperty: #url ifAbsent: [nil].
		uu ifNotNil: [(SqueakPageCache pageCache includesKey: uu) ifTrue: [
				SqueakPageCache initialize]]]].	"wipe the cache"! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'RAA 1/9/2001 06:59'!
releaseViewers
	"In preparation for saving, make the flapTabs release their viewers."

	self flapTabs do: [:ft | 
		(ft respondsTo: #hibernate) ifTrue: [ft hibernate]]! !

!PasteUpMorph methodsFor: 'Nebraska' stamp: 'ar 10/26/2000 14:08'!
remoteServer
	^worldState remoteServer.! !

!PasteUpMorph methodsFor: 'Nebraska' stamp: 'RAA 11/6/2000 17:24'!
remoteServer: aNebraskaServer

	| h |

	worldState remoteServer: aNebraskaServer.
	h := self primaryHand.
	aNebraskaServer ifNil:[
		(h hasProperty: #representingTheServer) ifTrue: [
			h removeProperty: #representingTheServer.
			h userInitials: '' andPicture: nil.
		]
	] ifNotNil:[
		(h hasProperty: #representingTheServer) ifFalse: [
			h setProperty: #representingTheServer toValue: true.
			h userInitials: Utilities authorName andPicture: nil.
		]
	].! !

!PasteUpMorph methodsFor: 'flaps' stamp: 'sw 2/15/1999 20:36'!
removeAccommodationForFlap: aFlapTab
	"Shift submorphs over, if appropriate"
	| offset |
	aFlapTab slidesOtherObjects ifTrue:
		[offset := self offsetForAccommodating: aFlapTab referent extent onEdge: aFlapTab edgeToAdhereTo.
		self shiftSubmorphsBy: offset negated]! !

!PasteUpMorph methodsFor: 'alarms-scheduler' stamp: 'ar 9/11/2000 16:39'!
removeAlarm: aSelector for: aTarget
	"Remove the alarm with the given selector"
	worldState removeAlarm: aSelector for: aTarget! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 5/3/2001 20:56'!
removeAllViewers
	"Delete all the viewers lined up along my right margin."

	(self submorphs select: [:m | m isKindOf: ViewerFlapTab]) do:
		[:m |
			m referent ifNotNil: [m referent delete].
			m delete.]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 10/5/2000 16:23'!
removeHand: aHandMorph
	"Remove the given hand from the list of hands for this world."

	(worldState hands includes: aHandMorph) ifFalse: [^self].
	aHandMorph dropMorphs.
	self invalidRect: aHandMorph fullBounds.
	worldState removeHand: aHandMorph.
! !

!PasteUpMorph methodsFor: 'initialization' stamp: 'mt 5/31/2016 12:06'!
removeKeyboardShortcuts
	
	self removeKeyboardCaptureFilter: self.! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'tak 3/15/2005 17:31'!
removeModalWindow
	self modalWindow: nil! !

!PasteUpMorph methodsFor: 'Nebraska' stamp: 'ar 10/26/2000 14:19'!
removeRemoteClient: aClient
	self removeHand: aClient hand.
	worldState removeRemoteCanvas: aClient canvas.
	self changed.  "force a redraw"
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'nice 12/27/2009 03:11'!
repairEmbeddedWorlds

	| toDoList |

	toDoList := OrderedCollection new.
	self allMorphsDo: [ :each | | transform eWorld |
		(each isKindOf: EmbeddedWorldBorderMorph) ifTrue: [
			transform := each submorphs at: 1 ifAbsent: [nil].
			transform ifNotNil: [
				eWorld := transform submorphs at: 1 ifAbsent: [nil].
				eWorld ifNotNil: [
					toDoList add: {transform. eWorld}.
				].
			].
			"Smalltalk at: #Q put: {self. each. transform. eWorld}."
		].
	].
	toDoList do: [ :each |
		each first addMorph: each second.
	].! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:22'!
repelsMorph: aMorph event: ev
	(aMorph wantsToBeDroppedInto: self) ifFalse: [^ false].
	self dropEnabled ifFalse: [^ true].
	(self wantsDroppedMorph: aMorph event: ev) ifFalse: [^ true].
	^ super repelsMorph: aMorph event: ev "consults #repelling flag"! !

!PasteUpMorph methodsFor: 'options' stamp: 'nice 12/27/2009 03:11'!
replaceTallSubmorphsByThumbnails
	"Any submorphs that seem to tall get replaced by thumbnails; their balloon text is copied over to the thumbnail"

	| heightForThumbnails maxHeightToAvoidThumbnailing maxWidthForThumbnails |
	heightForThumbnails := self heightForThumbnails.
	maxHeightToAvoidThumbnailing := self maxHeightToAvoidThumbnailing.
	maxWidthForThumbnails := self maximumThumbnailWidth.
	self submorphs do:
		[:aMorph | | existingHelp itsThumbnail |
			itsThumbnail := aMorph representativeNoTallerThan: maxHeightToAvoidThumbnailing norWiderThan: maxWidthForThumbnails thumbnailHeight: heightForThumbnails.
			(aMorph == itsThumbnail)
				ifFalse:
					[existingHelp := aMorph balloonText.
					self replaceSubmorph: aMorph by: itsThumbnail.
					existingHelp ifNotNil:
						[itsThumbnail setBalloonText: existingHelp]]]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'ar 10/24/2000 14:40'!
reportLocalAddress
	"Report the local host address of this computer."

	| addrString m s |
	Socket initializeNetwork.
	addrString := NetNameResolver localAddressString.
	m := RectangleMorph new
		color: (Color r: 0.6 g: 0.8 b: 0.6);
		extent: 118@36;
		borderWidth: 1.
	s := StringMorph contents: 'Local Host Address:'.
	s position: m position + (5@4).
	m addMorph: s.
	s := StringMorph contents: addrString.
	s position: m position + (5@19).
	m addMorph: s.
	self primaryHand attachMorph: m.
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'sw 2/7/2002 16:22'!
repositionFlapsAfterScreenSizeChange
	"Reposition flaps after screen size change"

	(Flaps globalFlapTabsIfAny, ActiveWorld localFlapTabs) do:
		[:aFlapTab |
			aFlapTab applyEdgeFractionWithin: self bounds].
	Flaps doAutomaticLayoutOfFlapsIfAppropriate! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'cmm 3/11/2010 22:50'!
resetTransferMorphConverter
	^ self transferMorphConverter: #yourself! !

!PasteUpMorph methodsFor: 'parts bin' stamp: 'di 6/8/1999 14:22'!
residesInPartsBin
	"Answer true if the receiver is, or has some ancestor owner who is, a parts bin"

	self isWorldMorph
		ifTrue: [^ self isPartsBin]
		ifFalse: [^ self isPartsBin or: [super residesInPartsBin]]! !

!PasteUpMorph methodsFor: 'options' stamp: 'ar 11/9/2000 12:49'!
resizeToFit
	^self vResizing == #shrinkWrap! !

!PasteUpMorph methodsFor: 'options' stamp: 'dgd 12/13/2003 19:30'!
resizeToFitString
	"Answer a string, to be used in a self-updating menu, to 
	represent whether the receiver is currently using resize-to-fit 
	or not"
	^ (self resizeToFit
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'resize to fit' translated! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 3/13/2003 13:56'!
respondToCommand: aCharacter bySending: aSelector to: aReceiver
	"Respond to the command-key use of the given character by sending the given selector to the given receiver.  If the selector is nil, retract any prior such setting"

	aSelector
		ifNil:
			[self commandKeySelectors removeKey: aCharacter]
		ifNotNil:
			[self commandKeySelectors at: aCharacter put: (MessageSend receiver: aReceiver selector: aSelector)]! !

!PasteUpMorph methodsFor: 'WiW support' stamp: 'dtl 7/13/2013 17:09'!
restartWorldCycleWithEvent: evt

	"RAA 27 Nov 99 - redispatch that click picked up from our inner world"
	evt ifNotNil: [
		self primaryHand handleEvent: (evt setHand: self primaryHand).
	].
	Project current spawnNewProcessAndTerminateOld: true
! !

!PasteUpMorph methodsFor: 'viewing' stamp: 'sw 10/5/2000 06:42'!
restoreBoundsOfSubmorphs
	"restores the saved xy-positions and extents"

	submorphs do:
		[:aSubmorph |
			aSubmorph valueOfProperty: #savedExtent ifPresentDo:
				[:anExtent | aSubmorph extent: anExtent].
			aSubmorph valueOfProperty: #savedPosition ifPresentDo:
				[:aPosition | aSubmorph position: aPosition]]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'md 2/12/2006 20:08'!
restoreFlapsDisplay
	"Restore the display of flaps"

	(Flaps sharedFlapsAllowed and: [Project current flapsSuppressed not]) ifTrue:
		[Flaps globalFlapTabs do:
			[:aFlapTab | aFlapTab adaptToWorld]].
	self localFlapTabs do:
			[:aFlapTab | aFlapTab adaptToWorld].
	self assureFlapTabsFitOnScreen.
	self bringTopmostsToFront.! !

!PasteUpMorph methodsFor: 'world state' stamp: 'dgd 9/5/2004 19:46'!
restoreMainDockingBarDisplay
	"Restore the display of docking bars"
	self dockingBars
		do: [:each | each updateBounds]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'mt 7/28/2016 15:17'!
restoreMorphicDisplay

	self removeProperty: #shouldDisplayWorld.

	ThumbnailMorph recursionReset.
	
	self
		extent: Display extent;
		viewBox: Display boundingBox;
		handsDo: [:h | h visible: true; showTemporaryCursor: nil];
		restoreFlapsDisplay;
		restoreMainDockingBarDisplay;
		fullRepaintNeeded.
		
	WorldState
		addDeferredUIMessage: [Cursor normal show].
! !

!PasteUpMorph methodsFor: 'painting' stamp: 'sw 5/1/2005 00:33'!
resumeScriptsPausedByPainting
	"If there were any scripts running when painting was initiated, resume them now"

	| aList |
	(aList := self valueOfProperty: #scriptsToResume) isEmptyOrNil ifFalse:
		[aList do:
			[:aScriptInstantiation |
				aScriptInstantiation player costume isInWorld ifTrue:
					[aScriptInstantiation startRunningIfPaused]]].
	self removeProperty: #scriptsToResume! !

!PasteUpMorph methodsFor: 'stepping' stamp: 'RAA 6/7/2000 10:12'!
runLocalStepMethods

	worldState runLocalStepMethodsIn: self
! !

!PasteUpMorph methodsFor: 'stepping' stamp: 'RAA 5/24/2000 10:27'!
runStepMethods

	worldState runStepMethodsIn: self
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'ar 8/6/2009 20:43'!
saveAsWorld
	| worldName s |
	worldName := UIManager default
		request: 'Please give this world a name'
		initialAnswer: 'test'.
	((self class class includesSelector: worldName asSymbol) and:
		[(UIManager default confirm: 'OK to overwrite ' , worldName , '?') not])
		ifTrue: [^ self].

	s := WriteStream on: (String new: 1000).
	s	nextPutAll: worldName; cr; tab;
		nextPutAll: '"' , self class name , ' ' , worldName, ' open"'; cr; cr; tab;
		nextPutAll: '^ '.
	self printConstructorOn: s indent: 0.
	s cr.

	self class class
		compile: s contents
		classified: 'examples'
		notifying: nil.! !

!PasteUpMorph methodsFor: 'viewing' stamp: 'sw 10/18/2000 10:59'!
saveBoundsOfSubmorphs
	"store the current xy-positions and extents of submorphs for future use"

	submorphs do:
		[:aSubmorph |
			aSubmorph setProperty: #savedExtent toValue: aSubmorph extent.
			aSubmorph setProperty: #savedPosition toValue: aSubmorph position]! !

!PasteUpMorph methodsFor: 'objects from disk' stamp: 'ar 8/6/2009 20:43'!
saveOnFile
	"Ask the user for a filename and save myself on a SmartReferenceStream file.  Writes out the version and class structure.  The file is fileIn-able.  UniClasses will be filed out."

	| aFileName fileStream ok |

	self flag: #bob0302.
	self isWorldMorph ifTrue: [^self project saveAs].

	aFileName := ('my {1}' translated format: {self class name}) asFileName.	"do better?"
	aFileName := UIManager default request: 'File name? (".project" will be added to end)' translated 
			initialAnswer: aFileName.
	aFileName isEmpty ifTrue: [^ Beeper beep].
	self allMorphsDo: [:m | m prepareToBeSaved].

	ok := aFileName endsWith: '.project'.	"don't double them"
	ok := ok | (aFileName endsWith: '.sp').
	ok ifFalse: [aFileName := aFileName,'.project'].
	fileStream := FileStream newFileNamed: aFileName asFileName.
	fileStream fileOutClass: nil andObject: self.	"Puts UniClass definitions out anyway"! !

!PasteUpMorph methodsFor: 'cursor' stamp: 'jdl 3/28/2003 08:17'!
selectedRect
	"Return a rectangle enclosing the morph at the current cursor. Note that the cursor may be a float and may be out of range, so pick the nearest morph. Assume there is at least one submorph."

	| p |
	p := cursor asInteger.
	p := p min: submorphs size.
	p := p max: 1.
	^(submorphs at: p) fullBounds expandBy: 2! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'ul 12/12/2009 14:08'!
sendTextContentsBackToDonor
	"Send my string contents back to the Text Morph from whence I came"

	(self valueOfProperty: #donorTextMorph) ifNotNil:
		[:aDonor | aDonor setCharacters: self assuredPlayer getStringContents]! !

!PasteUpMorph methodsFor: 'display' stamp: 'ar 10/5/2000 18:52'!
setGradientColor: evt
	"For backwards compatibility with GradientFillMorph"

	self flag: #fixThis.
	self changeColorTarget: self selector: #gradientFillColor:
		originalColor: (self fillStyle isGradientFill
			ifTrue: [self fillStyle colorRamp last value]
			ifFalse: [color])
		hand: evt hand.! !

!PasteUpMorph methodsFor: 'gridding' stamp: 'ar 8/6/2009 20:43'!
setGridSpec
	"Gridding rectangle provides origin and modulus"
	| response result |
	response := UIManager default
			request: 'New grid origin (usually 0@0):' translated
			initialAnswer: self gridOrigin printString.
	response isEmpty ifTrue: [^ self].
	result := [Compiler evaluate: response] ifError: [^ self].
	(result isPoint and: [(result >= (0@0))])
		ifTrue: [self gridOrigin: result]
		ifFalse: [self inform: ('Must be a Point with coordinates (for example 10@10)' translated )].

	response := UIManager default
			request: 'New grid spacing:' translated
			initialAnswer: self gridModulus printString.
	response isEmpty ifTrue: [^ self].
	result := [Compiler evaluate: response] ifError: [^ self].
	(result isPoint and: [(result > (0@0)) ])
		ifTrue: [self gridModulus: result]
		ifFalse: [self inform: ('Must be a Point with coordinates (for example 10@10)' translated )].

! !

!PasteUpMorph methodsFor: 'model' stamp: 'sw 4/1/98 21:17'!
setModel: aModelMorph
	"Set the model for this world. Methods for sensitized morphs will be compiled into the class for this model."

	model := aModelMorph
! !

!PasteUpMorph methodsFor: 'options' stamp: 'panda 4/25/2000 15:42'!
setPartsBinStatusTo: aBoolean
	isPartsBin := aBoolean.
	aBoolean ifFalse: [self enableDragNDrop].
		"but note that we no longer reset openToDragNDrop to false upon making it a parts bin again"
	isPartsBin
		ifTrue:
			[submorphs do:
				[:m | m isPartsDonor: true.
					m stopStepping.
					m suspendEventHandler]]
		ifFalse:
			[submorphs do:
				[:m | m isPartsDonor: false.
					m restoreSuspendedEventHandler].
			self world ifNotNil: [self world startSteppingSubmorphsOf: self]]! !

!PasteUpMorph methodsFor: 'WiW support' stamp: 'RAA 8/14/2000 12:10'!
shouldGetStepsFrom: aWorld

	(self isWorldMorph and: [owner notNil]) ifTrue: [
		^self outermostWorldMorph == aWorld
	].
	^super shouldGetStepsFrom: aWorld! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'RAA 6/26/2000 11:33'!
showApplicationView

	self transformToShow: (self valueOfProperty: #applicationViewBounds ifAbsent: [bounds])
		! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'RAA 6/26/2000 11:47'!
showExpandedView

	owner	"the transform"
		owner	"the green border"
			bounds: Display boundingBox! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'RAA 6/26/2000 11:33'!
showFactoryView

	self transformToShow: (self valueOfProperty: #factoryViewBounds ifAbsent: [bounds])
		! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'RAA 7/13/2000 10:17'!
showFullView

	self transformToShow: bounds
		! !

!PasteUpMorph methodsFor: 'viewing' stamp: 'sw 10/23/2000 19:01'!
showingListView
	"Answer whether the receiver is currently showing a list view"

	^ self hasProperty: #showingListView
! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'RAA 6/26/2000 11:46'!
showReducedView

	| r |
	r := Display extent // 4 extent: Display extent // 2.
	owner	"the transform"
		owner	"the green border"
			bounds: r! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 1/14/2001 03:04'!
showStatusOfAllScripts
	"Put up a window that shows, and allows you to change, the status of all scripts"

	self presenter toolToViewScriptInstantiations! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 9/6/2003 18:00'!
showThumbnailString
	"Answer the string to be shown in a menu to represent the 
	show-thumbnails status"
	^ ((self hasProperty: #alwaysShowThumbnail)
		ifTrue: ['<on>']
		ifFalse: ['<off>']), 'show thumbnails' translated! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 9/20/2004 19:27'!
showWorldMainDockingBarString
	^ self project showWorldMainDockingBarString! !

!PasteUpMorph methodsFor: 'world state' stamp: 'sw 9/2/1999 13:09'!
sketchEditorOrNil
	"Return a SketchEditorMorph found in the world, if any, else nil"

	^ self findA: SketchEditorMorph
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 7/16/2000 00:36'!
sleep

	self flag: #bob.		"Alan wanted this"

	worldState canvas ifNil: [^ self  "already called (clean this up)"].
	Cursor normal show.	"restore the normal cursor"

">>>> Alan wanted this out
	(turtleTrailsForm ~~ nil and: [self confirm: 'May I clear the pen trails
in this worldState to save space?']) ifTrue: [self clearTurtleTrails].
<<<<<"

	worldState canvas: nil.		"free my canvas to save space"
	self fullReleaseCachedState.
! !

!PasteUpMorph methodsFor: 'misc' stamp: 'di 12/23/1998 14:45'!
smallThumbnailForPageSorter

	^ self cachedOrNewThumbnailFrom: BookPageThumbnailMorph new smaller! !

!PasteUpMorph methodsFor: 'world state' stamp: 'dgd 2/22/2003 14:11'!
someHalo
	"Return some halo that's currently visible in the world"

	| m |
	^(m := self haloMorphs) notEmpty ifTrue: [m first] ifFalse: [nil]! !

!PasteUpMorph methodsFor: 'viewing' stamp: 'ar 11/12/2000 22:37'!
sortSubmorphsBy: sortOrderSymbol
	"Sort the receiver's submorphs by the criterion indicated in the provided symbol"
	self invalidRect: self fullBounds.
	submorphs := submorphs sortBy:[:a :b | (a perform: sortOrderSymbol) <= (b perform: sortOrderSymbol)].
	self layoutChanged.! !

!PasteUpMorph methodsFor: 'world state' stamp: 'di 7/15/1999 09:51'!
specialNameInModelFor: aMorph
	^ model ifNotNil: [model nameFor: aMorph] ifNil: [nil]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'nb 6/17/2003 12:25'!
standardPlayerHit

	self playSoundNamed: 'peaks'.
! !

!PasteUpMorph methodsFor: 'update cycle' stamp: 'ul 6/17/2011 12:38'!
startBackgroundProcess
	"Start a process to update this world in the background. Return the process created."

	| p |
	p := [ [
		self doOneCycleInBackground.
		(Delay forMilliseconds: 20) wait] repeat ] newProcess.
	p resume.
	^ p
! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 11/30/2006 18:35'!
startRunningAll
	"Start running all scripted morphs.  Triggered by user hitting GO button"

	self presenter flushPlayerListCache.  "Inefficient, but makes sure things come right whenever GO hit"
	self presenter allExtantPlayers do: [:aPlayer | aPlayer costume residesInPartsBin ifFalse: [aPlayer startRunning]].

	self world updateStatusForAllScriptEditors! !

!PasteUpMorph methodsFor: 'world state' stamp: 'RAA 6/1/2000 19:01'!
startSteppingSubmorphsOf: aMorph

	"Ensure that all submorphs of the given morph that want to be stepped are added to the step list.   Typically used after adding a morph to the world."

	aMorph allMorphsDo: [:m |
		m wantsSteps ifTrue: [m arrangeToStartSteppingIn: m world].
	]

! !

!PasteUpMorph methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:39'!
startStepping: aMorph
	"Add the given morph to the step list. Do nothing if it is already being stepped."
	^self startStepping: aMorph at: Time millisecondClockValue selector: #stepAt: arguments: nil stepTime: nil! !

!PasteUpMorph methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:36'!
startStepping: aMorph at: scheduledTime selector: aSelector arguments: args stepTime: stepTime
	worldState startStepping: aMorph at: scheduledTime selector: aSelector arguments: args stepTime: stepTime.! !

!PasteUpMorph methodsFor: 'stepping and presenter' stamp: 'RAA 8/14/2000 11:50'!
step

	(self isWorldMorph and: [owner notNil]) ifTrue: [
		^self runLocalStepMethods
	].
	super step! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 11/30/2006 18:35'!
stepAll
	"tick all the paused player scripts in the receiver"

	self presenter allExtantPlayers do:
		[:aPlayer | 
			aPlayer startRunning; step; stopRunning]! !

!PasteUpMorph methodsFor: 'project state' stamp: 'sw 9/5/2000 09:56'!
stepListSize
	^ worldState stepListSize

"Transcript cr; show: self currentWorld stepListSize printString, ' items on steplist as of ', Date dateAndTimeNow printString"! !

!PasteUpMorph methodsFor: 'project state' stamp: 'sw 9/5/2000 09:56'!
stepListSummary
	^ worldState stepListSummary

"Transcript cr show: self currentWorld stepListSummary"! !

!PasteUpMorph methodsFor: 'project state' stamp: 'sw 9/5/2000 09:59'!
steppingMorphsNotInWorld
	| all |
	all := self allMorphs.
	^ self listOfSteppingMorphs select: [:m | (all includes: m) not]

	"self currentWorld steppingMorphsNotInWorld do: [:m | m delete]"! !

!PasteUpMorph methodsFor: 'testing' stamp: 'RAA 8/14/2000 11:50'!
stepTime

	(self isWorldMorph and: [owner notNil]) ifTrue: [
		^1
	].
	^super stepTime! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 11/30/2006 18:35'!
stopRunningAll
	"Reset all ticking scripts to be paused.  Triggered by user hitting STOP button"

	self presenter allExtantPlayers do:
		[:aPlayer |
			aPlayer stopSound.
			aPlayer stopRunning].

	self world updateStatusForAllScriptEditors! !

!PasteUpMorph methodsFor: 'stepping' stamp: 'RAA 5/24/2000 11:08'!
stopStepping: aMorph
	"Remove the given morph from the step list."

	worldState stopStepping: aMorph
! !

!PasteUpMorph methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:40'!
stopStepping: aMorph selector: aSelector
	"Remove the given morph from the step list."

	worldState stopStepping: aMorph selector: aSelector
! !

!PasteUpMorph methodsFor: 'project' stamp: 'gm 2/16/2003 20:35'!
storeProjectsAsSegments
	"Force my sub-projects out to disk"

	submorphs do: 
			[:sub | 
			(sub isSystemWindow) 
				ifTrue: [(sub model isKindOf: Project) ifTrue: [sub model storeSegment]]]	"OK if was already out"! !

!PasteUpMorph methodsFor: 'misc' stamp: 'di 12/23/1998 14:44'!
thumbnailForPageSorter

	^ self cachedOrNewThumbnailFrom: BookPageThumbnailMorph new! !

!PasteUpMorph methodsFor: 'options' stamp: 'ar 11/8/2000 22:37'!
toggleAutoLineLayout
	"Toggle the auto-line-layout setting"

	self autoLineLayout: self autoLineLayout not.
	self autoLineLayout ifFalse: [self restoreBoundsOfSubmorphs].! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'mt 8/26/2015 17:41'!
toggleClassicNavigatorIfAppropriate
	"If appropriate, toggle the presence of classic navigator"

	Preferences classicNavigatorEnabled ifTrue: [^ Preferences toggle: #showProjectNavigator]! !

!PasteUpMorph methodsFor: 'options' stamp: 'sw 9/30/1998 17:24'!
toggleIsPartsBin
	"Not entirely happy with the openToDragNDrop not being directly manipulable etc, but still living with it for now."
	self setPartsBinStatusTo: self isPartsBin not! !

!PasteUpMorph methodsFor: 'options' stamp: 'sw 1/27/2000 14:51'!
toggleMouseOverHalos
	wantsMouseOverHalos := self wantsMouseOverHalos not! !

!PasteUpMorph methodsFor: 'options' stamp: 'tk 10/30/2001 18:41'!
toggleResizeToFit
	"Toggle whether the receiver is set to resize-to-fit"

	self vResizeToFit: self resizeToFit not! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'dgd 9/20/2004 19:27'!
toggleShowWorldMainDockingBar
	self project toggleShowWorldMainDockingBar! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'cmm 3/11/2010 22:43'!
transferMorphConverter
	^self
		valueOfProperty: #transferMorphConverter
		ifAbsent: [ #yourself ]! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'cmm 3/11/2010 22:07'!
transferMorphConverter: aMessageSend 
	self
		setProperty: #transferMorphConverter
		toValue: aMessageSend! !

!PasteUpMorph methodsFor: 'Nebraska' stamp: 'ar 10/26/2000 15:19'!
transferRemoteServerFrom: aWorld
	"Transfer the remote server which was associated with aWorld (if any) to the receiver"
	| server |
	(aWorld notNil and:[aWorld isMorph and:[aWorld isWorldMorph]]) ifFalse:[^self].
	server := aWorld remoteServer.
	server ifNotNil:[
		self remoteServer: server.
		server clients do:[:each| self addRemoteClient: each].
		self primaryHand
			userInitials: (aWorld primaryHand userInitials)
			andPicture: (aWorld primaryHand userPicture).
		aWorld primaryHand userInitials: '' andPicture: nil].
	aWorld releaseRemoteServer.! !

!PasteUpMorph methodsFor: 'menu & halo' stamp: 'RAA 6/26/2000 19:10'!
transformToShow: aRectangle

	owner changeWorldBoundsToShow: aRectangle
! !

!PasteUpMorph methodsFor: 'world state' stamp: 'kfr 11/24/2011 17:33'!
triggerClosingScripts
	"If the receiver has any scripts set to run on closing, run them now"
	| aPlayer |
	self allMorphsDo:[ :m|
	(aPlayer := m player) ifNotNil:
		[aPlayer runAllClosingScripts]]! !

!PasteUpMorph methodsFor: 'world state' stamp: 'kfr 11/24/2011 17:26'!
triggerOpeningScripts
	"If the receiver has any scripts set to run on opening, run them now"
	| aPlayer |
	self allMorphsDo:[ :m|
	(aPlayer := m player) ifNotNil:
		[aPlayer runAllOpeningScripts]]! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'sw 3/13/2003 12:25'!
undoOrRedoCommand
	"Undo or redo the last command recorded in the world"

	^ self commandHistory undoOrRedoCommand! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 1/1/1999 16:04'!
unhideHiddenObjects
	self allMorphsDo:
		[:m | m show]! !

!PasteUpMorph methodsFor: 'name' stamp: 'sw 6/17/2004 01:46'!
unusedMorphNameLike: stem
	"Answer a suitable name for a morph in this world, based on the stem provided"

	| names |
	names := self allKnownNames.
	^ Utilities keyLike: stem asString satisfying:
		[:aName | (names includes: aName) not]! !

!PasteUpMorph methodsFor: 'misc' stamp: 'sw 6/24/1998 19:47'!
updateStatusForAllScriptEditors
	self allScriptEditors do: [:anEditor | anEditor updateStatus]! !

!PasteUpMorph methodsFor: 'accessing' stamp: 'ar 4/25/2001 17:15'!
useRoundedCorners
	"Somewhat special cased because we do have to fill Display for this"
	super useRoundedCorners.
	self == World ifTrue:[Display bits primFill: 0]. "done so that we *don't* get a flash"! !

!PasteUpMorph methodsFor: 'WiW support' stamp: 'RAA 11/14/1999 15:07'!
validateMouseEvent: evt

	! !

!PasteUpMorph methodsFor: 'cursor' stamp: 'bf 9/30/2002 23:37'!
valueAtCursor
	"Answer the submorph of mine indexed by the value of my 'cursor' slot"

	submorphs isEmpty ifTrue: [^ self presenter standardPlayer costume].
	^ (submorphs at: ((cursor truncated max: 1) min: submorphs size)) morphRepresented! !

!PasteUpMorph methodsFor: 'cursor' stamp: 'sw 5/12/1998 10:55'!
valueAtCursor: aMorph
	submorphs isEmpty ifTrue: [^ self].
	self rectifyCursor.
	self replaceSubmorph: self valueAtCursor by: aMorph! !

!PasteUpMorph methodsFor: 'copying' stamp: 'tk 7/30/2001 09:26'!
veryDeepCopyWith: deepCopier
	"See storeDataOn:"

	^ self isWorldMorph
		ifTrue: [self]	"never copy the World"
		ifFalse: [super veryDeepCopyWith: deepCopier]! !

!PasteUpMorph methodsFor: 'project state' stamp: 'dgd 9/27/2004 11:45'!
viewBox
	"This tortured workaround arises from a situation encountered 
	in which a PasteUpMorph was directliy lodged as a submorph 
	of another PasteUpMorph of identical size, with the former 
	bearing flaps but the latter being the world"
	^ worldState
		ifNil: [super viewBox]
		ifNotNil: [worldState viewBox]! !

!PasteUpMorph methodsFor: 'project state' stamp: 'dgd 2/22/2003 14:12'!
viewBox: newViewBox 
	"I am now displayed within newViewBox; react."

	self isWorldMorph 
		ifTrue: 
			[(self viewBox isNil or: [self viewBox extent ~= newViewBox extent]) 
				ifTrue: [worldState canvas: nil].
			worldState viewBox: newViewBox].
	super position: newViewBox topLeft.
	fullBounds := bounds := newViewBox.

	"Paragraph problem workaround; clear selections to avoid screen
droppings."
	self flag: #arNote.	"Probably unnecessary"
	self isWorldMorph 
		ifTrue: 
			[worldState handsDo: [:hand | hand releaseKeyboardFocus].
			self fullRepaintNeeded]! !

!PasteUpMorph methodsFor: 'viewing' stamp: 'ar 11/8/2000 18:13'!
viewingNormally
	"Answer whether the receiver is being viewed normally, viz not in list-view or auto-line-layout"

	^ (self showingListView or: [self autoLineLayout == true]) not
! !

!PasteUpMorph methodsFor: 'halos and balloon help' stamp: 'di 9/26/2000 21:39'!
wantsDirectionHandles

	^ super wantsDirectionHandles and: [self isWorldMorph not]! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'ar 1/10/2001 21:29'!
wantsDropFiles: anEvent
	^self isWorldMorph! !

!PasteUpMorph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:22'!
wantsDroppedMorph: aMorph event: evt
	self isWorldMorph ifTrue:[^true]. "always"
	self visible ifFalse: [^ false].  "will be a call to #hidden again very soon"
	self dropEnabled ifFalse: [^ false].
	^ true! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'dgd 9/27/2004 13:55'!
wantsEasySelection
	"Answer if the receiver want easy selection mode"
	^ Preferences easySelection! !

!PasteUpMorph methodsFor: 'halos and balloon help' stamp: 'sw 10/10/2006 11:51'!
wantsHaloFor: aSubMorph
	"Answer whether the receiver wishes for a mouse-over halo to be produced for aSubMorph"

	^ wantsMouseOverHalos == true and:
		 [self visible and:
			[isPartsBin ~~ true and:
				[self dropEnabled and:
					[aSubMorph renderedMorph isLikelyRecipientForMouseOverHalos]]]]! !

!PasteUpMorph methodsFor: 'halos and balloon help' stamp: 'yo 2/17/2005 14:45'!
wantsHaloFromClick
	(owner isSystemWindow) ifTrue: [^ false].
	self paintBoxOrNil ifNotNil: [^ false].
	^ true.
! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'sw 5/6/1998 17:07'!
wantsKeyboardFocusFor: aSubmorph
	aSubmorph inPartsBin ifTrue: [^ false].
	aSubmorph wouldAcceptKeyboardFocus ifTrue: [ ^ true].
	^ super wantsKeyboardFocusFor: aSubmorph! !

!PasteUpMorph methodsFor: 'options' stamp: 'sw 6/5/1998 18:13'!
wantsMouseOverHalos
	^ wantsMouseOverHalos == true! !

!PasteUpMorph methodsFor: 'options' stamp: 'sw 6/5/1998 18:13'!
wantsMouseOverHalos: aBoolean
	wantsMouseOverHalos := aBoolean! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'bf 8/30/2007 17:59'!
wantsWindowEvent: anEvent
	^self isWorldMorph or: [self windowEventHandler notNil]! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'bf 8/30/2007 18:18'!
windowEventHandler
	"This is a class variable so it is global to all projects and does not get saved"
	^WindowEventHandler! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'bf 8/30/2007 18:18'!
windowEventHandler: anObject
	"This is a class variable so it is global to all projects and does not get saved"
	WindowEventHandler := anObject
! !

!PasteUpMorph methodsFor: 'event handling' stamp: 'bf 8/30/2007 20:37'!
windowEvent: anEvent
	self windowEventHandler
		ifNotNil: [^self windowEventHandler windowEvent: anEvent].

	anEvent type == #windowClose
		ifTrue: [
			^Preferences eToyFriendly 
				ifTrue: [ProjectNavigationMorph basicNew quitSqueak]
				ifFalse: [TheWorldMenu basicNew quitSession]].
! !

!PasteUpMorph methodsFor: 'structure' stamp: 'jmv 3/2/2010 16:10'!
world
	worldState ifNil: [^super world].
	^self! !

!PasteUpMorph methodsFor: 'world menu' stamp: 'dgd 9/19/2003 11:22'!
yellowButtonClickOnDesktopWithEvent: evt 
	"Put up either the personalized menu or the world menu when 
	the user clicks on the morphic desktop with the yellow button. 
	The preference 'personalizedWorldMenu' governs which one 
	is used"
	| aMenu |
	Preferences personalizedWorldMenu
		ifTrue: [aMenu := MenuMorph new defaultTarget: self.
			Preferences personalizeUserMenu: aMenu.
			aMenu addLine.
			aMenu
				add: 'personalize...' translated
				target: Preferences
				action: #letUserPersonalizeMenu]
		ifFalse: [aMenu := self buildWorldMenu: evt.
			aMenu addTitle: 'World' translated].
	aMenu popUpEvent: evt in: self! !

BorderedMorph subclass: #PolygonMorph
	instanceVariableNames: 'vertices closed filledForm arrows arrowForms smoothCurve curveState borderDashSpec handles borderForm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!PolygonMorph commentStamp: 'md 2/24/2006 20:34' prior: 0!
This class implements a morph which can behave as four different objects depending on the the following two facts:
- is it OPEN or CLOSED?
- is it SEGMENTED or SMOOTHED.

1. The OPEN and SEGMENTED variant looks like polyline.

2. The OPEN and SMOOTHED variant looks like spline (kind of curve)

3. The CLOSED and SEGMENTED variant looks like polygon. This is actually what you get when you do
	PolygonMorph new openInWorld
You get a triangle. See below how to manipulate these objects...

4. The CLOSED and SMOOTHED variant looks like blob (???)

Prototypes of this morph can also be found in "Object Catalog". Several (different variants) of this object are among "Basic" morphs.

Explore the assiciated morph-menu. It enables you
- to toggle showing of "handles". They make it possible to
	- reposition already existing vertices (by moving yellow handles)
	- create new vertices (by moving green handles)
	- delete already existing vertices (by dragging and dropping one yellow handle closely
	  nearby the adjacent yellow handle
  Handles can be made visible/hidden by shift+leftclicking the morph. This way it is possible
  to quickly show handles, adjust vertices and then again hide handles.
- making closed polygon open, i.e. converting it to a curve (and vice versa)
- toggle smoothed/segmented line/outline
- set up custom dashing (for line, curves or borders of closed polygons
- set up custom arrow-heads (for lines resp. curves)

------------------------------------------------------------------------------------------
Implementation notes:

This class combines the old Polygon and Curve classes.

The 1-bit fillForm to make display and containment tests reasonably fast.  However, this functionality is in the process of being supplanted by balloon capabilities, which should eventually provide anti-aliasing as well.

wiz 7/18/2004 21:26
s have made some changes to this class to

1) correct some bugs associated with one vertex polygons.

2) prepare for some enhancements with new curves.

3) add shaping items to menu.!

PolygonMorph subclass: #BalloonMorph
	instanceVariableNames: 'target offsetFromTarget balloonOwner'
	classVariableNames: 'BalloonColor BalloonFont'
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!BalloonMorph commentStamp: '<historical>' prior: 0!
A balloon with text used for the display of explanatory information.

Balloon help is integrated into Morphic as follows:
If a Morph has the property #balloonText, then it will respond to #showBalloon by adding a text balloon to the world, and to #deleteBalloon by removing the balloon.

Moreover, if mouseOverEnabled is true (see class msg), then the Hand will arrange to cause display of the balloon after the mouse has lingered over the morph for a while, and removal of the balloon when the mouse leaves the bounds of that morph.  In any case, the Hand will attempt to remove any such balloons before handling mouseDown events, or displaying other balloons.

Balloons should not be duplicated with veryDeepCopy unless their target is also duplicated at the same time.!

!BalloonMorph class methodsFor: 'preferences' stamp: 'mt 8/5/2016 09:16'!
balloonColor

	self flag: #remove. "mt: We should remove this additional getter  in the future and use UI themes instead:"
	^ (UserInterfaceTheme current get: #color for: #BalloonMorph) ifNil: [(TranslucentColor r: 0.92 g: 0.92 b: 0.706 alpha: 0.749)]! !

!BalloonMorph class methodsFor: 'preferences' stamp: 'mt 8/5/2016 09:16'!
balloonColor: aColor

	self flag: #remove. "mt: We should remove this additional setter in the future and use UI themes instead:"
	UserInterfaceTheme current set: #color for: #BalloonMorph to: aColor.! !

!BalloonMorph class methodsFor: 'utility' stamp: 'mt 8/5/2016 09:16'!
balloonFont

	self flag: #remove. "mt: We should remove this additional getter  in the future and use UI themes instead:"
	^ (UserInterfaceTheme current get: #font for: #BalloonMorph) ifNil: [TextStyle defaultFont]! !

!BalloonMorph class methodsFor: 'preferences' stamp: 'mt 8/5/2016 09:17'!
balloonTextColor

	self flag: #remove. "mt: We should remove this additional getter  in the future and use UI themes instead:"
	^ (UserInterfaceTheme current get: #textColor for: #BalloonMorph) ifNil: [Color black]! !

!BalloonMorph class methodsFor: 'preferences' stamp: 'mt 8/5/2016 09:17'!
balloonTextColor: aColor

	self flag: #remove. "mt: We should remove this additional setter in the future and use UI themes instead:"
	UserInterfaceTheme current set: #textColor for: #BalloonMorph to: aColor.! !

!BalloonMorph class methodsFor: 'utility' stamp: 'tween 8/7/2007 01:59'!
chooseBalloonFont
	"BalloonMorph chooseBalloonFont"

	Preferences 
		chooseFontWithPrompt:  'Ballon Help font...' translated
		andSendTo: self 
		withSelector: #setBalloonFontTo: 
		highlightSelector: #balloonFont! !

!BalloonMorph class methodsFor: 'private' stamp: 'wiz 1/24/2005 00:32'!
getBestLocation: vertices for: morph corner: cornerName
	"Try four rel locations of the balloon for greatest unclipped area.   12/99 sma"

	| rect maxArea verts rectCorner morphPoint mbc a mp dir bestVerts result usableArea |
	"wiz 1/8/2005 Choose rect independantly of vertice order or size. Would be nice it this took into account curveBounds but it does not." 
	rect := Rectangle encompassing: vertices.  
	maxArea := -1.
	verts := vertices.
	usableArea := (morph world ifNil: [self currentWorld]) viewBox.
	1 to: 4 do: [:i |
		dir := #(vertical horizontal) atWrap: i.
		verts := verts collect: [:p | p flipBy: dir centerAt: rect center].
		rectCorner := #(bottomLeft bottomRight topRight topLeft) at: i.
		morphPoint := #(topCenter topCenter bottomCenter bottomCenter) at: i.
		a := ((rect
			align: (rect perform: rectCorner)
			with: (mbc := morph boundsForBalloon perform: morphPoint))
				intersect: usableArea) area.
		(a > maxArea or: [a = rect area and: [rectCorner = cornerName]]) ifTrue:
			[maxArea := a.
			bestVerts := verts.
			mp := mbc]].
	result := bestVerts collect: [:p | p + (mp - bestVerts first)] "Inlined align:with:".
	^ result! !

!BalloonMorph class methodsFor: 'private' stamp: 'mt 8/5/2016 09:16'!
getTextMorph: aStringOrMorph for: balloonOwner
	"Construct text morph."
	| m text |
	aStringOrMorph isMorph
		ifTrue: [m := aStringOrMorph]
		ifFalse: [BalloonFont
				ifNil: [text := aStringOrMorph]
				ifNotNil: [text := aStringOrMorph asText
									addAttribute: (TextFontReference toFont: balloonOwner balloonFont);
									addAttribute: (TextColor color: self balloonTextColor)].
			m := (TextMorph new contents: text) centered].
	m setToAdhereToEdge: #adjustedCenter.
	^ m! !

!BalloonMorph class methodsFor: 'private' stamp: 'wiz 1/8/2005 18:05'!
getVertices: bounds
	"Construct vertices for a balloon up and to left of anchor"

	| corners |
	corners := bounds corners atAll: #(1 4 3 2).
	^ (Array
		with: corners first + (0 - bounds width // 2 @ 0)
		with: corners first + (0 - bounds width // 4 @ (bounds height // 2))) , corners! !

!BalloonMorph class methodsFor: 'utility' stamp: 'mt 8/5/2016 09:17'!
setBalloonFontTo: aFont

	self flag: #remove. "mt: We should remove this additional setter in the future and use UI themes instead:"
	UserInterfaceTheme current set: #font for: #BalloonMorph to: aFont.! !

!BalloonMorph class methodsFor: 'instance creation' stamp: 'sma 12/23/1999 20:05'!
string: str for: morph
	^ self string: str for: morph corner: #bottomLeft! !

!BalloonMorph class methodsFor: 'instance creation' stamp: 'cmm 8/4/2016 16:59'!
string: str for: morph corner: cornerName 
	"Make up and return a balloon for morph. Find the quadrant that 
	clips the text the least, using cornerName as a tie-breaker. tk 9/12/97"
	| text vertices balloon |
	balloon := self new.
	text := self
		getTextMorph: (str asString withNoLineLongerThan: Preferences maxBalloonHelpLineLength)
		for: morph.
	balloon userInterfaceTheme textColor ifNotNil:
		[:col | text color: col].
	balloon userInterfaceTheme font ifNotNil: [ : font | text font: font ].
	vertices := self getVertices: text bounds.
	vertices := self
		getBestLocation: vertices
		for: morph
		corner: cornerName.
	^ balloon
		 color: morph balloonColor ;
		 setVertices: vertices ;
		 addMorph: text ;
		 setTarget: morph! !

!BalloonMorph class methodsFor: 'preferences' stamp: 'mt 8/5/2016 08:58'!
themeProperties

	^ super themeProperties, {
		{ #borderColor. 'Colors'. 'Color of the balloon''s border.' }.
		{ #borderWidth. 'Borders'. 'Width of the balloon''s border.' }.
		{ #color. 'Colors', 'Color for the balloon background.' }.
		{ #font. 'Fonts'. 'Font for balloon text if not overridden by text attributes.' }.
		{ #textColor. 'Colors'. 'Color for the balloon text if not overridden by text attributes.' }.
	}! !

!BalloonMorph methodsFor: 'menus' stamp: 'wiz 12/30/2004 17:14'!
adjustedCenter
	"Return the center of the original textMorph box within the balloon."

	^ (self vertices last: 4) average rounded  ! !

!BalloonMorph methodsFor: 'updating' stamp: 'mt 8/5/2016 09:02'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	self setDefaultParameters.! !

!BalloonMorph methodsFor: 'accessing' stamp: 'ar 10/3/2000 17:19'!
balloonOwner
	^balloonOwner! !

!BalloonMorph methodsFor: 'initialization' stamp: 'mt 8/5/2016 09:00'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	self beSmoothCurve.

	offsetFromTarget := 0 @ 0.
	
	self setDefaultParameters.! !

!BalloonMorph methodsFor: 'WiW support' stamp: 'RAA 6/27/2000 18:07'!
morphicLayerNumber

	"helpful for insuring some morphs always appear in front of or behind others.
	smaller numbers are in front"

	^5		"Balloons are very front-like things"! !

!BalloonMorph methodsFor: 'initialization' stamp: 'cmm 8/4/2016 23:42'!
popUp
	| w worldBounds |
	target ifNil: [^ self].
	target isInWorld ifFalse: [^ self].
	w := target world.
	self prepareToOpen.
	"So that if the translation below makes it overlap the receiver, it won't interfere with the rootMorphsAt: logic and hence cause flashing.  Without this, flashing happens, believe me!!"
	((worldBounds := w bounds) containsRect: self bounds) ifFalse:
		[self bounds: (self bounds translatedToBeWithin: worldBounds)].
	self openInWorld.
	w activeHand addBalloonHelp: self.
	! !

!BalloonMorph methodsFor: 'initialization' stamp: 'mt 4/1/2016 11:20'!
popUpAt: point

	self position: point.
	self popUp.! !

!BalloonMorph methodsFor: 'initialization' stamp: 'mt 3/31/2015 11:46'!
popUpAt: aPoint forHand: aHand

	self popUpForHand: aHand.! !

!BalloonMorph methodsFor: 'initialization' stamp: 'cmm 8/4/2016 16:28'!
popUpForHand: aHand
	"Pop up the receiver as balloon help for the given hand"
	| worldBounds |
	self prepareToOpen.
	aHand world addMorphFront: self.
	"So that if the translation below makes it overlap the receiver, it won't
	interfere with the rootMorphsAt: logic and hence cause flashing.  Without
	this, flashing happens, believe me!!"
	((worldBounds := aHand world bounds) containsRect: self bounds) ifFalse:
		[self bounds: (self bounds translatedToBeWithin: worldBounds)].
	aHand resetBalloonHelp: self.
! !

!BalloonMorph methodsFor: 'initialization' stamp: 'ar 10/4/2000 10:13'!
popUpFor: aMorph hand: aHand
	"Pop up the receiver as balloon help for the given hand"
	balloonOwner := aMorph.
	self popUpForHand: aHand.! !

!BalloonMorph methodsFor: 'private' stamp: 'cmm 8/4/2016 16:29'!
prepareToOpen
	"Override the color if not already set."
	self userInterfaceTheme color ifNotNil: [ : col | self color: col].
	self
		lock ;
		fullBounds ;
		setProperty: #morphicLayerNumber
		toValue: self morphicLayerNumber! !

!BalloonMorph methodsFor: 'initialization' stamp: 'mt 8/5/2016 09:00'!
setDefaultParameters

	self
		borderWidth: (self userInterfaceTheme borderWidth ifNil: [0]);
		borderColor: (self userInterfaceTheme borderColor ifNil: [Color black]);
		color: (self userInterfaceTheme color ifNil: [TranslucentColor r: 0.92 g: 0.92 b: 0.706 alpha: 0.749]).! !

!BalloonMorph methodsFor: 'private' stamp: 'sma 12/23/1999 14:06'!
setTarget: aMorph
	(target := aMorph) ifNotNil: [offsetFromTarget := self position - target position]! !

!BalloonMorph methodsFor: 'stepping and presenter' stamp: 'sma 12/23/1999 14:05'!
step
	"Move with target."

	target ifNotNil: [self position: target position + offsetFromTarget].
! !

!BalloonMorph methodsFor: 'testing' stamp: 'di 9/18/97 10:10'!
stepTime
	^ 0  "every cycle"! !

PolygonMorph subclass: #CurveMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!CurveMorph commentStamp: '<historical>' prior: 0!
This is really only a shell for creating Shapes with smooth outlines.!

!CurveMorph class methodsFor: 'instance creation' stamp: 'tk 11/14/2001 17:47'!
arrowPrototype

	| aa |
	aa := PolygonMorph vertices: (Array with: 5@40 with: 5@8 with: 35@8 with: 35@40) 
		color: Color black 
		borderWidth: 2 
		borderColor: Color black.
	aa beSmoothCurve; makeOpen; makeForwardArrow.		"is already open"
	aa dashedBorder: {10. 10. Color red}.
		"A dash spec is a 3- or 5-element array with
		{ length of normal border color.
		length of alternate border color.
		alternate border color}"
	aa computeBounds.
	^ aa! !

!CurveMorph class methodsFor: 'parts bin' stamp: 'nk 8/23/2004 18:11'!
descriptionForPartsBin
	^ self partName:	'Curve'
		categories:		#('Graphics' 'Basic')
		documentation:	'A smooth wiggly curve, or a curved solid.  Shift-click to get handles and move the points.'! !

!CurveMorph methodsFor: 'initialization' stamp: 'di 9/10/2000 14:28'!
initialize

	super initialize.
	self beSmoothCurve.
! !

!CurveMorph methodsFor: 'parts bin' stamp: 'sw 6/28/2001 11:32'!
initializeToStandAlone

	super initializeToStandAlone.
	self beSmoothCurve.
! !

!CurveMorph methodsFor: 'testing' stamp: 'wiz 1/7/2005 20:02'!
isCurvier
	"Test used by smoothing routines.  If true use true closed curve splines for closed curves. If not mimic old stodgy curveMorph curves with one sharp bend. Curve overrides this test for backward compatability.."
	^ (false)! !

PolygonMorph subclass: #CurvierMorph
	instanceVariableNames: ''
	classVariableNames: 'DrawCurvier SlopeConstantsCache'
	poolDictionaries: ''
	category: 'Morphic-Basic-NewCurve'!

!CurvierMorph commentStamp: '<historical>' prior: 0!
I want to be merged into PolygonMorph.
I implement Closed Cubic Curves and restructured routines to ease maintence and development.



New way to calculate curves.

cVariables
SlopeConstantsCache anArray size 2  indexed by nVerts \\2 .
		Each element is an array of integers. The integers represent the constants for
 		calculating slopes for closed cubic curves from the vertices. 




Class Variable SlopeConstantsCache holds a pair of arrays for even and odd number of vertices( aka knots).
Each array holds a series of constants in Integer form.
This allows slopes to be calculated directly from the array of knots.
Wonderfully it turns out that only two arrays are needed.
By matching up the knots equidistant from the point in question;
Summing the weighted differences of the pairs the unscaled slope can be arrived at.
The scale needed to get the slopes needed is trice the reciprical of the next integer in the series.
We leave the division til last to get the full benifit of the integer arithmetic.

Rounding the vertices before calculation is recommended.


Instead of calculating the number of curve subsegments in lineSegDo we add a ninth array to curve state to allow the number to be precalculated.
Getting better looking curves motivates finding a better way of guessing n. So this provides a framework for trying.

For the first pass we just used the constant 12 for every thing.
In the second pass we judge the number of segments by starting with two and doubling the number until the distance of the curve no longer increases.
Then we hone in using a binary search to find the smallest number of segments with that same curve length.


We have changed some assumptions. Previously curves were figured by solving for the second derivative  first and using the results to determine the slope and the third derivative. So lineSegDo counted on the last second deriv being a number it could use in its calculation of the number of subsegments.

Currently we just solve for slopes and the second and third derivs are derived from that. 
Also the derivation for the second and third derivs only assume C(1) (first derivitive continuity). The calculations for the slopes are the only calcs using C(2) continuity. Therefore the slopes can alternately be chosen to fit some other chriteria  and the resulting curves will still be smooth to the first degree.
A useful variant of closed slopes is to scale them by a constant.


Also the last of each element of curvestate always reflects a closing segment. And we don't add an extra row for closed curves anymore. 
That is now lineSegDo's responsibility to be aware of as it was already doing with segmented curves. So the last n does not track its old value.

Preferences:
A Preference has been added to toggle between the old (ugly) closed curves based on natural cubic slopes and the new smooth algorythim. This doesn't make much difference while newcurves are a subclass of polygons but the ambition is for newcurves to supercede polygons. This will allow backwards  compatibility.

Shapes: With closed curves a smooth oval results from rectagular or diamond vertices. So two menuitems have been added (to PolygonMorph) that allow the vertices to be set to these shapes using the current bounds of the polygon. The former state of vertices will be lost but it seems useful to lose a complicated shape and start fresh with a simple symmetrical one. 

Furthur on: Modify curveState to only contain slope and higher deriv information. Let the information about the knots only be held only in the vertices of the polygon. Once that is done curvestate will not have to be recalcutaled each time the polygon is moved but only when its shape changes.

There is also some possible speed up to be had by refining or experimenting with other number of segment calculating schemes but not as much as preserving curvestate over a move.

Furthur furthur on: Figure out how to combine straight and curved segments into a single shape in a pleasing way.

 







!

!CurvierMorph class methodsFor: 'parts bin' stamp: 'wiz 11/6/2004 23:24'!
descriptionForPartsBin
	"We are very much like curve only better looking."
	^ self
		partName: 'Curvier'
		categories: #('Graphics' 'Basic' )
		documentation: 'A smooth wiggly curve, or a smooth curved solid without bends.  Shift-click to get handles and move the points.'! !

!CurvierMorph class methodsFor: 'class initialization' stamp: 'eem 7/1/2015 15:11'!
drawCurvier
	<preference: 'Curvier'
		category: 'morphic'
		description: 'If true, closed CurvierMorphs will be smoother and more symmetrical all about. If false they will mimic the old curve shapes with the one sharp bend.'
		type: #Boolean>
	^DrawCurvier ifNil: [true]! !

!CurvierMorph class methodsFor: 'class initialization' stamp: 'fbs 8/10/2013 17:03'!
drawCurvier: aBoolean
	DrawCurvier := aBoolean.! !

!CurvierMorph class methodsFor: 'class initialization' stamp: 'eem 7/1/2015 15:12'!
initialize
	"CurvierMorph initialize"
	self registerInFlapsRegistry! !

!CurvierMorph class methodsFor: 'class initialization' stamp: 'wiz 11/6/2004 23:17'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | 
			cl registerQuad: #(#CurvierMorph #authoringPrototype 'Curvier' 'A curve' ) forFlapNamed: 'PlugIn Supplies'.
			cl registerQuad: #(#CurvierMorph #authoringPrototype 'Curvier' 'A curve' ) forFlapNamed: 'Supplies']! !

!CurvierMorph methodsFor: 'initialization' stamp: 'wiz 11/16/2004 21:35'!
initialize
	"We use an oval shape because we wear it well."
	super initialize.
	self beSmoothCurve.
	self diamondOval! !

PolygonMorph subclass: #LineMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!LineMorph commentStamp: '<historical>' prior: 0!
This is really only a shell for creating single-segment straight-line Shapes.!

!LineMorph class methodsFor: 'parts bin' stamp: 'tak 7/25/2007 00:06'!
descriptionForPartsBin
	"Answer a description for the parts bin."

	^ self partName:	'Line' translatedNoop
		categories:		{'Graphics' translatedNoop}
		documentation:	'A straight line.  Shift-click to get handles and move the ends.' translatedNoop! !

!LineMorph class methodsFor: 'instance creation' stamp: 'di 8/20/2000 12:18'!
from: startPoint to: endPoint color: lineColor width: lineWidth

	^ PolygonMorph vertices: {startPoint. endPoint}
			color: Color black borderWidth: lineWidth borderColor: lineColor! !

!LineMorph class methodsFor: 'instance creation' stamp: 'di 8/20/2000 12:16'!
new
	^ self from: 0@0 to: 50@50 color: Color black width: 2! !

!LineMorph class methodsFor: 'new-morph participation' stamp: 'sw 11/13/2001 14:37'!
newStandAlone
	"Answer a suitable instance for use in a parts bin, for example"

	^ self new setNameTo: 'Line'! !

PolygonMorph subclass: #MixedCurveMorph
	instanceVariableNames: 'slopeClamps'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic-NewCurve'!

!MixedCurveMorph commentStamp: '<historical>' prior: 0!
A MixedCurveMorph is Curve that can be broken up into separately curved segments. It allows for the creation of matching edges( e. g. for jigsaw puzzle pieces).

Instance Variables
	slopeClamps:		<Array>

slopeClamps
	- elements of array are either 0 or nil. Indicating whether slope for the corresponding vertex is 0@0 or unknown and therefore to be calculated. There is one element for each vertex.
	
	
!

!MixedCurveMorph class methodsFor: 'as yet unclassified' stamp: 'wiz 2/12/2006 17:12'!
descriptionForPartsBin
	^ self partName:	'Mixed'
		categories:		#('Graphics' 'Basic')
		documentation:	'A Curve with optional bends and segments. Shift click to get handles.
			Click handles to change bends. Move handles to move the points.'! !

!MixedCurveMorph methodsFor: 'access' stamp: 'wiz 2/8/2006 18:59'!
clamps
" Return a collection of clamps the same size as vertices.
	If necessary default to unclamped slopes.
"

slopeClamps 
	ifNil:   [ ^ slopeClamps := Array new: vertices size  ] .
slopeClamps size = vertices size
	ifFalse: [ ^ slopeClamps := Array new: vertices size  ] . 
	^ slopeClamps           ! !

!MixedCurveMorph methodsFor: 'editing' stamp: 'wiz 2/8/2006 17:50'!
clickVertex: ix event: evt fromHandle: handle
" Toggle the state of the clamp. "
"Note: self clamps assures slopeClamps will be same size as vertices"

(self clamps at: ix) 
	ifNil:	 [ slopeClamps  at: ix put: 0 ]
	ifNotNil: [ slopeClamps  at: ix put: nil ] .
	self setVertices: vertices .
	
! !

!MixedCurveMorph methodsFor: 'editing' stamp: 'wiz 2/8/2006 18:01'!
deleteVertexAt: anIndex
			(slopeClamps :=
						slopeClamps
						copyReplaceFrom: anIndex
						to: anIndex
						with: Array new) .
			self
				setVertices: (vertices
						copyReplaceFrom: anIndex
						to: anIndex
						with: Array new).
						! !

!MixedCurveMorph methodsFor: 'access' stamp: 'wiz 2/8/2006 17:44'!
handleColorAt: vertIndex
      " clamped handles are cyan     and 
	unclamped handles are yellow."

(self clamps at: vertIndex ) ifNil: [ ^ Color yellow ] .
^ Color cyan 
! !

!MixedCurveMorph methodsFor: 'initialization' stamp: 'wiz 2/12/2006 05:59'!
initialize
"initialize the state of the receiver"
	super initialize.
	self extent: 32@20 .

	self rectOval.
	self clamps . "This initializes slopeClamps."
	slopeClamps at: 1 put: 0 .
	slopeClamps at: 4 put: 0 .
	
	closed := true.
	smoothCurve := true.
	arrows := #none.
	self computeBounds! !

!MixedCurveMorph methodsFor: 'editing' stamp: 'wiz 2/8/2006 18:29'!
insertVertexAt: anIndex put: aValue
	"New vertexs are unclamped."
	"Note: order is important. 
	The clamps array must match vertex size before setVertices: is performed."
	slopeClamps := slopeClamps 
		copyReplaceFrom: anIndex + 1 to: anIndex with: (Array with: nil).
	self setVertices: (vertices copyReplaceFrom: anIndex + 1 to: anIndex 
									with: (Array with: aValue)).! !

!MixedCurveMorph methodsFor: 'smoothing' stamp: 'wiz 2/18/2006 12:53'!
slopes: knots 
	"Choose slopes according to state of polygon and preferences"
	self isCurvy
		ifFalse: [^ knots segmentedSlopes].
	^ (closed
			and: [self isCurvier])
		ifTrue: [ knots closedCubicSlopes: self clamps ]
		ifFalse: [knots naturalCubicSlopes: self clamps ]! !

!PolygonMorph class methodsFor: 'instance creation' stamp: 'wiz 7/13/2005 00:43'!
arrowPrototype
	"Answer an instance of the receiver that will serve as a prototypical arrow"

	| aa |
	aa := self new. 
	aa vertices: (Array with: 0@0 with: 40@40) 
		color: Color black 
		borderWidth: 2 
		borderColor: Color black.
	"aa setProperty: #noNewVertices toValue: true."
	"Revert to expected behavior. Remove vestigial code."
	aa makeForwardArrow.		"is already open"
	aa computeBounds.
	^ aa! !

!PolygonMorph class methodsFor: 'parts bin' stamp: 'tak 7/25/2007 00:10'!
descriptionForPartsBin
	^ self partName:	'Polygon' translatedNoop
		categories:		{'Graphics' translatedNoop. 'Basic' translatedNoop}
		documentation:	'A series of connected line segments, which may be a closed solid, or a zig-zag line.  Shift-click to get handles and move the points.' translatedNoop! !

!PolygonMorph class methodsFor: 'instance creation' stamp: 'di 10/18/2001 04:42'!
fromHandFreehand: hand
	"Let the user draw a polygon, holding the mouse down, and ending
		by clicking within 5 of the first point..."
	| p1 poly pN opposite |
	Cursor crossHair showWhile:
		[[Sensor anyButtonPressed] whileFalse:
			[self currentWorld displayWorldSafely; runStepMethods].
		p1 := Sensor cursorPoint].
	opposite := (Display colorAt: p1) negated.
	opposite = Color transparent ifTrue: [opposite := Color red].
	(poly := LineMorph from: p1 to: p1 color: opposite width: 2) openInWorld.
	self currentWorld displayWorldSafely; runStepMethods.
	[Sensor anyButtonPressed] whileTrue:
			[pN := Sensor cursorPoint.
			(pN dist: poly vertices last) > 3 ifTrue:
				[poly setVertices: (poly vertices copyWith: pN).
				self currentWorld displayWorldSafely; runStepMethods]].
	hand position: Sensor cursorPoint.  "Done -- update hand pos"
	^ (poly setVertices: (poly vertices copyWith: p1)) delete! !

!PolygonMorph class methodsFor: 'instance creation' stamp: 'ul 6/17/2011 12:40'!
fromHand: hand
	"Let the user draw a polygon, clicking at each vertex, and ending
		by clicking within 5 of the first point..."
	| p1 poly oldVerts pN opposite |
	Cursor crossHair showWhile:
		[[Sensor anyButtonPressed] whileFalse:
			[self currentWorld displayWorldSafely; runStepMethods].
		p1 := Sensor cursorPoint].
	opposite := (Display colorAt: p1) negated.
	opposite = Color transparent ifTrue: [opposite := Color red].
	(poly := LineMorph from: p1 to: p1 color: opposite width: 2) openInWorld.
	oldVerts := {p1}.
	self currentWorld displayWorldSafely; runStepMethods.
	
	[[Sensor anyButtonPressed] whileTrue:
		[pN := Sensor cursorPoint.
		poly setVertices: (oldVerts copyWith: pN).
		self currentWorld displayWorldSafely; runStepMethods].
	(oldVerts size > 1 and: [(pN dist: p1) < 5]) ifTrue:
		[hand position: Sensor cursorPoint.  "Done -- update hand pos"
		^ (poly setVertices: (poly vertices copyWith: p1)) delete].
	oldVerts := poly vertices.
	[Sensor anyButtonPressed] whileFalse:
		[pN := Sensor cursorPoint.
		poly setVertices: (oldVerts copyWith: pN).
		self currentWorld displayWorldSafely; runStepMethods]] repeat! !

!PolygonMorph class methodsFor: 'instance creation' stamp: 'di 9/9/2000 11:41'!
shapeFromPen: penBlock color: c borderWidth: bw borderColor: bc
	"World addMorph: (PolygonMorph
		shapeFromPen: [:p | p hilbert: 4 side: 5. p go: 5.
						p hilbert: 4 side: 5. p go: 5]
		color: Color red borderWidth: 1 borderColor: Color black)"

	| pen |
	penBlock value: (pen := PenPointRecorder new).
	^ (self vertices: pen points asArray color: c borderWidth: bw borderColor: bc)
		quickFill: false! !

!PolygonMorph class methodsFor: 'instance creation' stamp: 'di 9/7/2000 17:05'!
vertices: verts color: c borderWidth: bw borderColor: bc
	^ self basicNew beStraightSegments vertices: verts color: c borderWidth: bw borderColor: bc! !

!PolygonMorph methodsFor: 'menu' stamp: 'sw 10/3/2007 02:07'!
addCustomMenuItems: aMenu hand: aHandMorph
	"Add morph-specific items to the given menu which was invoked by the given hand.  This method provides is invoked both from the halo-menu and from the control-menu regimes."

	super addCustomMenuItems: aMenu hand: aHandMorph.
	aMenu addUpdating: #handlesShowingPhrase target: self action: #showOrHideHandles.
	vertices size > 2 ifTrue:
		[aMenu addUpdating: #openOrClosePhrase target: self action: #makeOpenOrClosed].

	aMenu addUpdating: #smoothPhrase target: self action: #toggleSmoothing.
	aMenu addLine.
	aMenu add: 'specify dashed line' translated action:  #specifyDashedLine.

	self isOpen ifTrue:
		[aMenu addLine.
		aMenu addWithLabel: '---' enablement: [self isOpen and: [arrows ~~ #none]] action:  #makeNoArrows.
		aMenu addWithLabel: '-->' enablement: [self isOpen and: [arrows ~~ #forward]] action:  #makeForwardArrow.
		aMenu addWithLabel: '<--' enablement: [self isOpen and: [arrows ~~ #back]] action:  #makeBackArrow.
		aMenu addWithLabel: '<->' enablement: [self isOpen and: [arrows ~~ #both]] action:  #makeBothArrows.
		aMenu add: 'customize arrows' translated action: #customizeArrows:.
		(self hasProperty: #arrowSpec)
			ifTrue: [aMenu add: 'standard arrows' translated action: #standardArrows]].! !

!PolygonMorph methodsFor: 'rotate scale and flex' stamp: 'di 11/28/2001 18:23'!
addFlexShellIfNecessary
	"When scaling or rotating from a halo, I can do this without a flex shell"

	^ self
! !

!PolygonMorph methodsFor: 'editing' stamp: 'nice 1/1/2010 22:06'!
addHandles
	"Put moving handles at the vertices. Put adding handles at
	edge midpoints.
	Moving over adjacent vertex and dropping will delete a
	vertex. "
	| tri |
	self removeHandles.
	handles := OrderedCollection new.
	tri := Array
				with: 0 @ -4
				with: 4 @ 3
				with: -3 @ 3.
	vertices
		withIndexDo: [:vertPt :vertIndex | 
			| handle |
			handle := EllipseMorph
						newBounds: (Rectangle center: vertPt extent: 8 @ 8)
						color: (self handleColorAt: vertIndex) .
			handle
				on: #mouseMove
				send: #dragVertex:event:fromHandle:
				to: self
				withValue: vertIndex.
			handle
				on: #mouseUp
				send: #dropVertex:event:fromHandle:
				to: self
				withValue: vertIndex.
				handle
				on: #click
				send: #clickVertex:event:fromHandle:
				to: self
				withValue: vertIndex.
			self addMorph: handle.
			handles addLast: handle.
			(closed
					or: [1 = vertices size
						"Give a small polygon a chance to grow. 
						-wiz"
					or: [vertIndex < vertices size]])
				ifTrue: [| newVert |
					newVert := PolygonMorph
								vertices: (tri
										collect: [:p | p + (vertPt
													+ (vertices atWrap: vertIndex + 1) // 2)])
								color: Color green
								borderWidth: 1
								borderColor: Color black.
					newVert
						on: #mouseDown
						send: #newVertex:event:fromHandle:
						to: self
						withValue: vertIndex.
					self addMorph: newVert.
					handles addLast: newVert]].
	self isCurvy
		ifTrue: [self updateHandles; layoutChanged].
	self changed! !

!PolygonMorph methodsFor: 'menu' stamp: 'wiz 12/29/2004 13:53'!
addPolyArrowMenuItems: aMenu hand: aHandMorph 
aMenu addLine.
			aMenu
				addWithLabel: '---'
				enablement: [self isOpen
						and: [arrows ~~ #none]]
				action: #makeNoArrows.
			aMenu
				addWithLabel: '-->'
				enablement: [self isOpen
						and: [arrows ~~ #forward]]
				action: #makeForwardArrow.
			aMenu
				addWithLabel: '<--'
				enablement: [self isOpen
						and: [arrows ~~ #back]]
				action: #makeBackArrow.
			aMenu
				addWithLabel: '<->'
				enablement: [self isOpen
						and: [arrows ~~ #both]]
				action: #makeBothArrows.
			aMenu add: 'customize arrows' translated action: #customizeArrows:.
			(self hasProperty: #arrowSpec)
				ifTrue: [aMenu add: 'standard arrows' translated action: #standardArrows]! !

!PolygonMorph methodsFor: 'menu' stamp: 'wiz 4/1/2006 19:05'!
addPolyLIneCurveMenuItems: aMenu hand: aHandMorph 

	aMenu addLine;
				addUpdating: #openOrClosePhrase
				target: self
				action: #makeOpenOrClosed.
			
			aMenu
				addUpdating: #smoothOrSegmentedPhrase
				target: self
				action: #toggleSmoothing.! !

!PolygonMorph methodsFor: 'menu' stamp: 'wiz 12/29/2004 13:50'!
addPolyShapingMenuItems: aMenu hand: aHandMorph 
	aMenu addLine.
			aMenu
				addWithLabel: 'make inscribed diamondOval'
				enablement: [self isClosed ]
				action: #diamondOval.
			aMenu
				addWithLabel: 'make enclosing rectangleOval'
				enablement: [self isClosed ]
					action: #rectOval.
					! !

!PolygonMorph methodsFor: 'drawing' stamp: 'di 6/24/1998 14:36'!
areasRemainingToFill: aRectangle
	"Could be improved by quick check of inner rectangle"

	^ Array with: aRectangle! !

!PolygonMorph methodsFor: 'private' stamp: 'di 10/3/2000 09:02'!
arrowBoundsAt: endPoint from: priorPoint 
	"Answer a triangle oriented along the line from priorPoint to endPoint."
	| d v angle wingBase arrowSpec length width |
	v := endPoint - priorPoint.
	angle := v degrees.
	d := borderWidth max: 1.
	arrowSpec := self valueOfProperty: #arrowSpec ifAbsent: [5@4].
	length := arrowSpec x abs.  width := arrowSpec y abs.
	wingBase := endPoint + (Point r: d * length degrees: angle + 180.0).
	arrowSpec x >= 0
		ifTrue: [^ {	endPoint.
					wingBase + (Point r: d * width degrees: angle + 125.0).
					wingBase + (Point r: d * width degrees: angle - 125.0) }]
		ifFalse: ["Negative length means concave base."
				^ {	endPoint.
					wingBase + (Point r: d * width degrees: angle + 125.0).
					wingBase.
					wingBase + (Point r: d * width degrees: angle - 125.0) }]! !

!PolygonMorph methodsFor: 'private' stamp: 'wiz 6/22/2004 15:54'!
arrowForms
	"ArrowForms are computed only upon demand"
	arrowForms
		ifNotNil: [^ arrowForms].
	arrowForms := Array new.
	self hasArrows
		ifFalse: [^ arrowForms].
	(arrows == #forward
			or: [arrows == #both])
		ifTrue: [arrowForms := arrowForms
						copyWith: (self computeArrowFormAt: vertices last from: self nextToLastPoint)].
	(arrows == #back
			or: [arrows == #both])
		ifTrue: [arrowForms := arrowForms
						copyWith: (self computeArrowFormAt: vertices first from: self nextToFirstPoint)].
	^ arrowForms! !

!PolygonMorph methodsFor: 'menu' stamp: 'sw 9/24/2002 19:23'!
arrowLength: aLength
	"Assumes that I have exactly two vertices"

	| theta horizontalOffset verticalOffset newTip delta |
	delta := vertices second - vertices first.
	theta := delta theta.
	horizontalOffset := aLength * (theta cos).
	verticalOffset := aLength * (theta sin).
	newTip := vertices first + (horizontalOffset @ verticalOffset).
	self verticesAt: 2 put: newTip! !

!PolygonMorph methodsFor: 'menu' stamp: 'di 10/3/2000 09:09'!
arrowSpec: specPt
	"Specify a custom arrow for this line.
	specPt x abs gives the length of the arrow (point to base) in terms of borderWidth.
	If specPt x is negative, then the base of the arrow will be concave.
	specPt y abs gives the width of the arrow.
	The standard arrow is equivalent to arrowSpec: 5@4.
	See arrowBoundsAt:From: for details."

	self setProperty: #arrowSpec toValue: specPt.
	self computeBounds! !

!PolygonMorph methodsFor: 'menu' stamp: 'nk 2/26/2001 20:11'!
arrows
	^arrows! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 2/15/2001 09:09'!
arrowsContainPoint: aPoint
	"Answer an Array of two Booleans that indicate whether the given point is inside either arrow"

	| retval f |

	retval := { false . false }.
	(super containsPoint: aPoint) ifFalse: [^ retval ].
	(closed or: [arrows == #none or: [vertices size < 2]]) ifTrue: [^ retval].

	(arrows == #forward or: [arrows == #both]) ifTrue: [	"arrowForms first has end form"
		f := self arrowForms first.
		retval at: 2 put: ((f pixelValueAt: aPoint - f offset) > 0)
	].
	(arrows == #back or: [arrows == #both]) ifTrue: [ "arrowForms last has start form"
		f := self arrowForms last.
		retval at: 1 put: ((f pixelValueAt: aPoint - f offset) > 0)
	].
	^retval.! !

!PolygonMorph methodsFor: 'initialization' stamp: 'di 9/8/2000 09:44'!
beSmoothCurve

	smoothCurve == true ifFalse:
		[smoothCurve := true.
		self computeBounds]! !

!PolygonMorph methodsFor: 'initialization' stamp: 'di 9/8/2000 09:45'!
beStraightSegments

	smoothCurve == false ifFalse:
		[smoothCurve := false.
		self computeBounds]! !

!PolygonMorph methodsFor: 'access' stamp: 'aoy 2/15/2003 20:51'!
borderColor: aColor 

	super borderColor: aColor.
	(borderColor isColor and: [borderColor isTranslucentColor]) 
		== (aColor isColor and: [aColor isTranslucentColor]) 
			ifFalse: 
				["Need to recompute fillForm and borderForm
					if translucency of border changes."

				self releaseCachedState]! !

!PolygonMorph methodsFor: 'dashes' stamp: 'dgd 2/22/2003 18:55'!
borderDashOffset
	borderDashSpec size < 4 ifTrue: [^0.0].
	^(borderDashSpec fourth) asFloat! !

!PolygonMorph methodsFor: 'private' stamp: 'ar 5/25/2000 18:04'!
borderForm
	"A form must be created for drawing the border whenever the borderColor is translucent."

	| borderCanvas |
	borderForm ifNotNil: [^ borderForm].
	borderCanvas := (Display defaultCanvasClass extent: bounds extent depth: 1)
		shadowColor: Color black.
	borderCanvas translateBy: bounds topLeft negated
		during:[:tempCanvas| self drawBorderOn: tempCanvas].
	borderForm := borderCanvas form.
	self arrowForms do:
		[:f |  "Eliminate overlap between line and arrowheads if transparent."
		borderForm copy: f boundingBox from: f to: f offset - self position rule: Form erase].
	^ borderForm! !

!PolygonMorph methodsFor: 'accessing' stamp: 'nk 9/4/2004 17:23'!
borderWidth: anInteger

	borderColor ifNil: [borderColor := Color black].
	borderWidth := anInteger max: 0.
	self computeBounds! !

!PolygonMorph methodsFor: 'attachments' stamp: 'nk 8/14/2004 13:58'!
boundsSignatureHash
	^(vertices - (self positionInWorld))  hash
! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nice 2/16/2008 02:30'!
bounds: newBounds
	"This method has to be reimplemented since self extent: will also change self bounds origin,
	super bounds would leave me in wrong position when container is growing.
	Always change extent first then position"
	
	self extent: newBounds extent; position: newBounds topLeft
! !

!PolygonMorph methodsFor: 'visual properties' stamp: 'ar 6/25/1999 11:18'!
canHaveFillStyles
	"Return true if the receiver can have general fill styles; not just colors.
	This method is for gradually converting old morphs."
	^true! !

!PolygonMorph methodsFor: 'editing' stamp: 'wiz 2/7/2006 23:35'!
clickVertex: ix event: evt fromHandle: handle
	"Backstop for MixedCurveMorph"! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nice 12/27/2009 03:11'!
closestPointTo: aPoint 
	| closestPoint minDist |
	closestPoint := minDist := nil.
	self lineSegmentsDo: 
			[:p1 :p2 | | dist curvePoint | 
			curvePoint := aPoint nearestPointOnLineFrom: p1 to: p2.
			dist := curvePoint dist: aPoint.
			(closestPoint isNil or: [dist < minDist]) 
				ifTrue: 
					[closestPoint := curvePoint.
					minDist := dist]].
	^closestPoint! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nice 12/27/2009 03:11'!
closestSegmentTo: aPoint
	"Answer the starting index of my (big) segment nearest to aPoint"
	| closestPoint minDist vertexIndex closestVertexIndex |
	vertexIndex := 0.
	closestVertexIndex := 0.
	closestPoint := minDist := nil.
	self lineSegmentsDo:
		[:p1 :p2 | | dist curvePoint | 
		(p1 = (self vertices at: vertexIndex + 1))
			ifTrue: [ vertexIndex := vertexIndex + 1 ].
		curvePoint := aPoint nearestPointOnLineFrom: p1 to: p2.
		dist := curvePoint dist: aPoint.
		(closestPoint == nil or: [dist < minDist])
			ifTrue: [closestPoint := curvePoint.
					minDist := dist.
					closestVertexIndex := vertexIndex. ]].
	^ closestVertexIndex! !

!PolygonMorph methodsFor: 'smoothing' stamp: 'nice 2/8/2010 08:34'!
coefficients
	"Compute an array for the coefficients."
	| verts vertXs vertYs slopeXs slopeYs coefficients |
	curveState
		ifNotNil: [^ curveState at: 1].
	verts := self vertices.
	verts size < 1
		ifTrue: [^ self].
	"Less than three points handled as segments by our 
	lineSegmentsDo:"
	(self isCurvier)
		ifFalse: [closed
				ifTrue: [verts := verts , verts first asOrderedCollection]].
	coefficients := {
		vertXs := verts collect: [:p | p x asFloat].
		slopeXs := self slopes: vertXs.
		vertXs changeInSlopes: slopeXs.
		vertXs changeOfChangesInSlopes: slopeXs.
		vertYs := verts collect: [:p | p y asFloat].
		slopeYs := self slopes: vertYs.
		vertYs changeInSlopes: slopeYs.
		vertYs changeOfChangesInSlopes: slopeYs.
		Array new: verts size withAll: 12}.
	coefficients
		at: 9
		put: ((1 to: verts size)
				collect: [:i | (coefficients cubicPointPolynomialAt: i) bestSegments]).
	(self isCurvier)
		ifFalse: [closed
				ifTrue: [coefficients := coefficients
								collect: [:each | each allButLast]]].
	curveState := {coefficients. nil. nil}.
	self computeNextToEndPoints.
	^ coefficients! !

!PolygonMorph methodsFor: 'private' stamp: 'tpr 3/15/2013 12:36'!
computeArrowFormAt: endPoint from: priorPoint 
	"Compute a triangle oriented along the line from priorPoint to  
	endPoint. Then draw those lines in a form and return that  
	form, with appropriate offset"

	| p1 pts box arrowForm bb origin |
	pts := self arrowBoundsAt: endPoint from: priorPoint.
	box := ((pts first rect: pts last) encompass: (pts second)) expandBy: 1.
	arrowForm := Form extent: box extent asIntegerPoint.
	bb := (BitBlt toForm: arrowForm)
				sourceForm: nil;
				fillColor: Color black;
				combinationRule: Form over;
				width: 1;
				height: 1.
	origin := box topLeft.
	p1 := pts last - origin.
	pts do: 
			[:p | 
			bb drawFrom: p1 to: p - origin.
			p1 := p - origin].
	arrowForm convexShapeFill: Color black.
	^arrowForm offset: box topLeft! !

!PolygonMorph methodsFor: 'private' stamp: 'nice 12/27/2009 22:39'!
computeBounds
	| oldBounds delta excludeHandles |
	vertices ifNil: [^ self].

	self changed.
	oldBounds := bounds.
	self releaseCachedState.
	bounds := self curveBounds expanded copy.
	self arrowForms do:
		[:f | bounds swallow: (f offset extent: f extent)].
	handles ifNotNil: [self updateHandles].

	"since we are directly updating bounds, see if any ordinary submorphs exist and move them accordingly"
	(oldBounds notNil and: [(delta := bounds origin - oldBounds origin) ~= (0@0)]) ifTrue: [
		excludeHandles := IdentitySet new.
		handles ifNotNil: [excludeHandles addAll: handles].
		self submorphsDo: [ :each |
			(excludeHandles includes: each) ifFalse: [
				each position: each position + delta
			].
		].
	].
	self layoutChanged.
	self changed.
! !

!PolygonMorph methodsFor: 'smoothing' stamp: 'jmv 3/2/2010 16:22'!
computeNextToEndPoints
	| pointAfterFirst pointBeforeLast |
	pointAfterFirst := nil.
	self lineSegmentsDo: 
			[:p1 :p2 | 
			pointAfterFirst ifNil: [pointAfterFirst := p2 asIntegerPoint].
			pointBeforeLast := p1 asIntegerPoint].
	curveState at: 2 put: pointAfterFirst.
	curveState at: 3 put: pointBeforeLast! !

!PolygonMorph methodsFor: 'geometry testing' stamp: 'di 8/20/2000 14:33'!
containsPoint: aPoint
	(super containsPoint: aPoint) ifFalse: [^ false].

	closed & color isTransparent not ifTrue:
		[^ (self filledForm pixelValueAt: aPoint - bounds topLeft + 1) > 0].

	self lineSegmentsDo:
		[:p1 :p2 |
		(aPoint onLineFrom: p1 to: p2 within: (3 max: borderWidth+1//2) asFloat)
				ifTrue: [^ true]].

	self arrowForms do:
		[:f | (f pixelValueAt: aPoint - f offset) > 0 ifTrue: [^ true]].

	^ false! !

!PolygonMorph methodsFor: 'objects from disk' stamp: 'RAA 12/20/2000 17:48'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	smoothCurve ifNil: [smoothCurve := false].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!PolygonMorph methodsFor: 'rounding' stamp: 'ka 12/4/2005 00:56'!
cornerStyle: aSymbol 
	"Set the receiver's corner style.  But, in this case, do *not*"

	self removeProperty: #cornerStyle.
	self changed! !

!PolygonMorph methodsFor: 'accessing' stamp: 'sw 11/24/1999 14:57'!
couldHaveRoundedCorners
	^ false! !

!PolygonMorph methodsFor: 'private' stamp: 'wiz 2/12/2006 02:58'!
curveBounds
	"Compute the bounds from actual curve traversal, with 
	leeway for borderWidth. 
	Also note the next-to-first and next-to-last points for arrow 
	directions."
	"wiz - to avoid roundoff errors we return unrounded curvebounds."
	"we expect our receiver to take responsibility for approriate rounding adjustment."
	"hint: this is most likely 'self curveBounds expanded' "
	| pointAfterFirst pointBeforeLast  oX oY cX cY |
	self isCurvy
		ifFalse: [^ (Rectangle encompassing: vertices)
				expandBy: borderWidth * 0.5 ].
	curveState := nil.
	"Force recomputation"
	"curveBounds := vertices first corner: vertices last."
	pointAfterFirst := nil.
	self
		lineSegmentsDo: [:p1 :p2 | 
			pointAfterFirst isNil
				ifTrue: [pointAfterFirst := p2 floor .
					oX := cX := p1 x.
					oY := cY := p1 y. ].
			"curveBounds := curveBounds encompass: p2 ."
			oX:= oX min: p2 x.
			cX := cX max: p2 x.
			oY := oY min: p2 y.
			cY := cY max: p2 y.
			pointBeforeLast := p1 floor ].
	curveState at: 2 put: pointAfterFirst.
	curveState at: 3 put: pointBeforeLast.
	^ ( oX @ oY corner: cX @ cY )  expandBy: borderWidth * 0.5 ! !

!PolygonMorph methodsFor: 'menu' stamp: 'ul 12/12/2009 14:07'!
customizeArrows: evt
	| handle origin aHand |
	aHand := evt ifNil: [self primaryHand] ifNotNil: [evt hand].
	origin := aHand position.
	(handle := HandleMorph new)
		forEachPointDo:
			[:newPoint | handle removeAllMorphs.
			handle addMorph:
				(LineMorph from: origin to: newPoint color: Color black width: 1).
			self arrowSpec: (newPoint - origin) / 5.0]
		lastPointDo:
			[:newPoint | handle deleteBalloon.
			self halo ifNotNil: [:halo | halo addHandles].].
	aHand attachMorph: handle.
	handle setProperty: #helpAtCenter toValue: true.
	handle showBalloon:
'Move cursor left and right
to change arrow length and style.
Move it up and down to change width.
Click when done.' hand: evt hand.
	handle startStepping! !

!PolygonMorph methodsFor: 'dashes' stamp: 'nk 2/27/2001 12:11'!
dashedBorder
	^borderDashSpec
	"A dash spec is a 3- or 5-element array with
		{ length of normal border color.
		length of alternate border color.
		alternate border color.
		starting offset.
		amount to add to offset at each step }
	Starting offset is usually = 0, but changing it moves the dashes along the curve."
! !

!PolygonMorph methodsFor: 'dashes' stamp: 'di 9/9/2000 09:20'!
dashedBorder: dashSpec
	"A dash spec is a 3- or 5-element array with
		{ length of normal border color.
		length of alternate border color.
		alternate border color.
		starting offset.
		amount to add to offset at each step }
	Starting offset is usually = 0, but changing it moves the dashes along the curve."

	borderDashSpec := dashSpec.
	self changed! !

!PolygonMorph methodsFor: 'attachments' stamp: 'nk 2/25/2001 17:21'!
defaultAttachmentPointSpecs
	^{ 
		{ #firstVertex } .
		{ #midpoint  } .
		{ #lastVertex }
	}! !

!PolygonMorph methodsFor: 'initialization' stamp: 'sw 9/10/2007 01:23'!
defaultBorderColor
	"answer the default border color/fill style for the receiver"

	^ Color black

"Until September 2007, this had long been...
	^ Color
		r: 0.0
		g: 0.419
		b: 0.935"! !

!PolygonMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:30'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color orange! !

!PolygonMorph methodsFor: 'editing' stamp: 'wiz 2/8/2006 18:20'!
deleteVertexAt: anIndex
	"This acts as a backstop for MixedCurveMorph."
			self
				setVertices: (vertices
						copyReplaceFrom: anIndex
						to: anIndex
						with: Array new).
						! !

!PolygonMorph methodsFor: 'smoothing' stamp: 'dgd 2/22/2003 14:16'!
derivs: a first: point1 second: point2 third: point3 
	"Compute the first, second and third derivitives (in coeffs) from
	the Points in this Path (coeffs at: 1 and coeffs at: 5)."

	| len v anArray |
	len := a size.
	len < 2 ifTrue: [^self].
	len > 2 
		ifTrue: 
			[v := Array new: len.
			v at: 1 put: 4.0.
			anArray := Array new: len.
			anArray at: 1 put: 6.0 * (a first - (a second * 2.0) + (a third)).
			2 to: len - 2
				do: 
					[:i | 
					v at: i put: 4.0 - (1.0 / (v at: i - 1)).
					anArray at: i
						put: 6.0 * ((a at: i) - ((a at: i + 1) * 2.0) + (a at: i + 2)) 
								- ((anArray at: i - 1) / (v at: i - 1))].
			point2 at: len - 1 put: (anArray at: len - 2) / (v at: len - 2).
			len - 2 to: 2
				by: 0 - 1
				do: 
					[:i | 
					point2 at: i
						put: ((anArray at: i - 1) - (point2 at: i + 1)) / (v at: i - 1)]].
	point2 at: 1 put: (point2 at: len put: 0.0).
	1 to: len - 1
		do: 
			[:i | 
			point1 at: i
				put: (a at: i + 1) - (a at: i) 
						- (((point2 at: i) * 2.0 + (point2 at: i + 1)) / 6.0).
			point3 at: i put: (point2 at: i + 1) - (point2 at: i)]! !

!PolygonMorph methodsFor: 'shaping' stamp: 'wiz 1/8/2005 19:27'!
diamondOval
	"Set my vertices to an array of edge midpoint vertices. 
	Order of vertices is in the tradion of warpblt quads."
	| b r |
	b := self bounds.
	r := {b leftCenter. b bottomCenter. b rightCenter. b topCenter}.
	self setVertices: r! !

!PolygonMorph methodsFor: 'editing' stamp: 'ar 3/17/2001 14:30'!
dragVertex: ix event: evt fromHandle: handle
	| p |
	p := self isCurve
		ifTrue: [evt cursorPoint]
		ifFalse: [self griddedPoint: evt cursorPoint].
	handle position: p - (handle extent//2).
	self verticesAt: ix put: p.
! !

!PolygonMorph methodsFor: 'editing' stamp: 'ar 3/18/2001 17:28'!
dragVertex: arg1 fromHandle: arg2 vertIndex: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self dragVertex: arg1 event: arg2 fromHandle: arg3! !

!PolygonMorph methodsFor: 'drawing' stamp: 'dgd 2/22/2003 18:56'!
drawArrowOn: aCanvas at: endPoint from: priorPoint 
	"Draw a triangle oriented along the line from priorPoint to  
	endPoint. Answer the wingBase."

	| pts spec wingBase |
	pts := self arrowBoundsAt: endPoint from: priorPoint.
	wingBase := pts size = 4 
				ifTrue: [pts third]
				ifFalse: [(pts copyFrom: 2 to: 3) average].
	spec := self valueOfProperty: #arrowSpec ifAbsent: [5 @ 4].
	spec x sign = spec y sign 
		ifTrue: [aCanvas drawPolygon: pts fillStyle: borderColor]
		ifFalse: 
			[aCanvas 
				drawPolygon: pts
				fillStyle: Color transparent
				borderWidth: (borderWidth + 1) // 2
				borderColor: borderColor].
	^wingBase! !

!PolygonMorph methodsFor: 'drawing' stamp: 'wiz 2/23/2006 19:19'!
drawArrowsOn: aCanvas 
	"Answer (possibly modified) endpoints for border drawing"
	"ArrowForms are computed only upon demand"
	| array |

	self hasArrows
		ifFalse: [^ #() ].
	"Nothing to do"

	array := Array with: vertices first with: vertices last.

	"Prevent crashes for #raised or #inset borders"
	borderColor isColor
		ifFalse: [ ^array ].

	(arrows == #forward or: [arrows == #both])
		ifTrue: [ array at: 2 put: (self
				drawArrowOn: aCanvas
				at: vertices last
				from: self nextToLastPoint) ].

	(arrows == #back or: [arrows == #both])
		ifTrue: [ array at: 1 put: (self
				drawArrowOn: aCanvas
				at: vertices first
				from: self nextToFirstPoint) ].

	^array! !

!PolygonMorph methodsFor: 'drawing' stamp: 'ar 11/26/2001 23:15'!
drawBorderOn: aCanvas 
	self
		drawClippedBorderOn: aCanvas
		usingEnds: (Array with: vertices first with: vertices last)! !

!PolygonMorph methodsFor: 'drawing' stamp: 'nice 12/27/2009 03:11'!
drawBorderOn: aCanvas usingEnds: anArray 
	"Display my border on the canvas."
	"NOTE: Much of this code is also copied in  
	drawDashedBorderOn:  
	(should be factored)"
	| bigClipRect style |
	borderDashSpec
		ifNotNil: [^ self drawDashedBorderOn: aCanvas usingEnds: anArray].
	style := self borderStyle.
	bigClipRect := aCanvas clipRect expandBy: self borderWidth + 1 // 2.
	self
		lineSegmentsDo: [:p1 :p2 | | p2i p1i | 
			p1i := p1 asIntegerPoint.
			p2i := p2 asIntegerPoint.
			self hasArrows
				ifTrue: ["Shorten line ends so as not to interfere with tip  
					of arrow."
					((arrows == #back
								or: [arrows == #both])
							and: [p1 = vertices first])
						ifTrue: [p1i := anArray first asIntegerPoint].
					((arrows == #forward
								or: [arrows == #both])
							and: [p2 = vertices last])
						ifTrue: [p2i := anArray last asIntegerPoint]].
			(closed
					or: ["bigClipRect intersects: (p1i rect: p2i)  
						optimized:"
						((p1i min: p2i)
							max: bigClipRect origin)
							<= ((p1i max: p2i)
									min: bigClipRect corner)])
				ifTrue: [style
						drawLineFrom: p1i
						to: p2i
						on: aCanvas]]! !

!PolygonMorph methodsFor: 'drawing' stamp: 'ar 11/26/2001 23:15'!
drawClippedBorderOn: aCanvas usingEnds: anArray 
	aCanvas clipBy: self bounds during:[:cc| self drawBorderOn: cc usingEnds: anArray].! !

!PolygonMorph methodsFor: 'drawing' stamp: 'nk 10/4/2000 12:23'!
drawDashedBorderOn: aCanvas 
	self
		drawDashedBorderOn: aCanvas
		usingEnds: (Array with: vertices first with: vertices last)! !

!PolygonMorph methodsFor: 'drawing' stamp: 'nice 12/27/2009 03:11'!
drawDashedBorderOn: aCanvas usingEnds: anArray 
	"Display my border on the canvas. NOTE: mostly copied from  
	drawBorderOn:"
	| bevel topLeftColor bottomRightColor bigClipRect lineColor segmentOffset |
	(borderColor isNil
			or: [borderColor isColor
					and: [borderColor isTransparent]])
		ifTrue: [^ self].
	lineColor := borderColor.
	bevel := false.
	"Border colors for bevelled effects depend on CW ordering of  
	vertices"
	borderColor == #raised
		ifTrue: [topLeftColor := color lighter.
			bottomRightColor := color darker.
			bevel := true].
	borderColor == #inset
		ifTrue: [topLeftColor := owner colorForInsets darker.
			bottomRightColor := owner colorForInsets lighter.
			bevel := true].
	bigClipRect := aCanvas clipRect expandBy: self borderWidth + 1 // 2.
	segmentOffset := self borderDashOffset.
	self
		lineSegmentsDo: [:p1 :p2 | | p1i p2i | 
			p1i := p1 asIntegerPoint.
			p2i := p2 asIntegerPoint.
			self hasArrows
				ifTrue: ["Shorten line ends so as not to interfere with tip  
					of arrow."
					((arrows == #back
								or: [arrows == #both])
							and: [p1 = vertices first])
						ifTrue: [p1i := anArray first asIntegerPoint].
					((arrows == #forward
								or: [arrows == #both])
							and: [p2 = vertices last])
						ifTrue: [p2i := anArray last asIntegerPoint]].
			(closed
					or: ["bigClipRect intersects: (p1i rect: p2i)  
						optimized:"
						((p1i min: p2i)
							max: bigClipRect origin)
							<= ((p1i max: p2i)
									min: bigClipRect corner)])
				ifTrue: [bevel
						ifTrue: [lineColor := (p1i quadrantOf: p2i)
											> 2
										ifTrue: [topLeftColor]
										ifFalse: [bottomRightColor]].
					segmentOffset := aCanvas
								line: p1i
								to: p2i
								width: borderWidth
								color: lineColor
								dashLength: borderDashSpec first
								secondColor: borderDashSpec third
								secondDashLength: borderDashSpec second
								startingOffset: segmentOffset]]! !

!PolygonMorph methodsFor: 'drawing' stamp: ''!
drawOnFormCanvas: aCanvas 
	"Display the receiver, a spline curve, approximated by straight line segments."

	| |
	vertices size < 1 ifTrue: [self error: 'a polygon must have at least one point'].
	closed & color isTransparent not
		ifTrue: [aCanvas stencil: self filledForm at: bounds topLeft - 1 color: color].
	(borderColor isColor and: [borderColor isTranslucentColor])
		ifTrue: [aCanvas stencil: self borderForm at: bounds topLeft
						color: borderColor]
		ifFalse: [self drawBorderOn: aCanvas].
	self arrowForms do:
		[:f | aCanvas stencil: f at: f offset
			color: (borderColor isColor ifTrue: [borderColor] ifFalse: [color])]! !

!PolygonMorph methodsFor: 'drawing' stamp: 'ar 11/26/2001 23:15'!
drawOn: aCanvas 
	"Display the receiver, a spline curve, approximated by straight 
	line segments."
	| array |
	vertices size < 1
		ifTrue: [self error: 'a polygon must have at least one point'].
	closed ifTrue:
		[aCanvas drawPolygon: self getVertices fillStyle: self fillStyle.
		aCanvas isShadowDrawing ifTrue: [^ self]].
	array := self drawArrowsOn: aCanvas.
	self drawClippedBorderOn: aCanvas usingEnds: array.
! !

!PolygonMorph methodsFor: 'editing' stamp: 'wiz 2/8/2006 18:35'!
dropVertex: ix event: evt fromHandle: handle
	"Leave vertex in new position. If dropped ontop another vertex delete this one.
	Check for too few vertices before deleting. The alternative 
				is not pretty -wiz"
	| p |
	p := vertices at: ix.
	(vertices size >= 2
			and: ["check for too few vertices before deleting. The alternative 
				is not pretty -wiz"
				((vertices atWrap: ix - 1)
						dist: p)
						< 3
					or: [((vertices atWrap: ix + 1)
							dist: p)
							< 3]])
		ifTrue: ["Drag a vertex onto its neighbor means delete"
				self deleteVertexAt: ix .].
	evt shiftPressed
		ifTrue: [self removeHandles]
		ifFalse: [self addHandles
			"remove then add to recreate"]! !

!PolygonMorph methodsFor: 'editing' stamp: 'ar 3/18/2001 17:28'!
dropVertex: arg1 fromHandle: arg2 vertIndex: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self dropVertex: arg1 event: arg2 fromHandle: arg3! !

!PolygonMorph methodsFor: 'attachments' stamp: 'nk 4/18/2001 11:43'!
endShapeColor: aColor
	self borderColor: aColor.
	self isClosed ifTrue: [ self color: aColor ].! !

!PolygonMorph methodsFor: 'attachments' stamp: 'nk 5/31/2003 10:56'!
endShapeWidth: aWidth
	| originalWidth originalVertices transform |
	originalWidth := self valueOfProperty: #originalWidth ifAbsentPut: [ self borderWidth isZero ifFalse: [ self borderWidth ] ifTrue: [ 2 ] ].
	self borderWidth: aWidth.
	originalVertices := self valueOfProperty: #originalVertices ifAbsentPut: [
		self vertices collect: [ :ea | (ea - (self referencePosition)) rotateBy: self heading degreesToRadians about: 0@0 ]
	].
	transform := MorphicTransform offset: 0@0 angle: self heading degreesToRadians scale: originalWidth / aWidth.
	self setVertices: (originalVertices collect: [ :ea |
		((transform transform: ea) + self referencePosition) asIntegerPoint
	]).
	self computeBounds.! !

!PolygonMorph methodsFor: 'geometry' stamp: 'di 9/24/2000 08:44'!
extent: newExtent 
	"Not really advisable, but we can preserve most of the geometry if we don't
	shrink things too small."
	| safeExtent center |
	center := self referencePosition.
	safeExtent := newExtent max: 20@20.
	self setVertices: (vertices collect:
		[:p | p - center * (safeExtent asFloatPoint / (bounds extent max: 1@1)) + center])! !

!PolygonMorph methodsFor: 'private' stamp: 'tpr 3/15/2013 12:36'!
filledForm
	"Note: The filled form is actually 2 pixels bigger than bounds, and the point corresponding to this morphs' position is at 1@1 in the form.  This is due to the details of the fillig routines, at least one of which requires an extra 1-pixel margin around the outside.  Computation of the filled form is done only on demand."
	| bb origin |
	closed ifFalse: [^ filledForm := nil].
	filledForm ifNotNil: [^ filledForm].
	filledForm := Form extent: bounds extent+2.

	"Draw the border..."
	bb := (BitBlt toForm: filledForm) sourceForm: nil; fillColor: Color black;
			combinationRule: Form over; width: 1; height: 1.
	origin := bounds topLeft asIntegerPoint-1.
	self lineSegmentsDo: [:p1 :p2 | bb drawFrom: p1 asIntegerPoint-origin
										to: p2 asIntegerPoint-origin].

	"Fill it in..."
	filledForm convexShapeFill: Color black.

	(borderColor isColor and: [borderColor isTranslucentColor]) ifTrue:
		["If border is stored as a form, then erase any overlap now."
		filledForm copy: self borderForm boundingBox from: self borderForm
			to: 1@1 rule: Form erase].

	^ filledForm! !

!PolygonMorph methodsFor: 'visual properties' stamp: 'sw 3/26/2008 15:14'!
fillStyle
	"Answer the receiver's fillStyle.  For an *open* polygon, we return the borderColor, provided it's a true color rather than something strange like the symbol #raised."

	| aColor |
	self isOpen
		ifTrue:
			[(aColor := self borderColor) isColor ifTrue: [^ aColor]].   "easy access to line color from halo -- di's old note"
	
	^ super fillStyle! !

!PolygonMorph methodsFor: 'visual properties' stamp: 'wiz 1/7/2005 20:39'!
fillStyle: newColor

	self isOpen
		ifTrue: [^ self borderColor: newColor asColor "easy access to line color from halo"]
		ifFalse: [^ super fillStyle: newColor]! !

!PolygonMorph methodsFor: 'attachments' stamp: 'nk 2/25/2001 17:19'!
firstVertex
	^vertices first! !

!PolygonMorph methodsFor: 'geometry' stamp: 'edc 3/20/2002 14:24'!
flipHAroundX: centerX
	"Flip me horizontally around the center.  If centerX is nil, compute my center of gravity."

	| cent |
	cent := centerX 
		ifNil: [bounds center x
			"cent := 0.
			vertices do: [:each | cent := cent + each x].
			cent asFloat / vertices size"]		"average is the center"
		ifNotNil: [centerX].
	self setVertices: (vertices collect: [:vv |
			((vv x - cent) * -1 + cent) @ vv y]) reversed.! !

!PolygonMorph methodsFor: 'geometry' stamp: 'sw 9/14/97 18:22'!
flipVAroundY: centerY
	"Flip me vertically around the center.  If centerY is nil, compute my center of gravity."

	| cent |
	cent := centerY 
		ifNil: [bounds center y
			"cent := 0.
			vertices do: [:each | cent := cent + each y].
			cent asFloat / vertices size"]		"average is the center"
		ifNotNil: [centerY].
	self setVertices: (vertices collect: [:vv |
			vv x @ ((vv y - cent) * -1 + cent)]) reversed.! !

!PolygonMorph methodsFor: 'private' stamp: 'di 9/7/2000 16:17'!
getVertices

	smoothCurve ifFalse: [^ vertices].

	"For curves, enumerate the full set of interpolated points"
	^ Array streamContents:
		[:s | self lineSegmentsDo: [:pt1 :pt2 | s nextPut: pt1]]! !

!PolygonMorph methodsFor: 'editing' stamp: 'wiz 2/8/2006 18:37'!
handleColorAt: vertIndex
      "This is a backstop for MixedCurveMorph"

^ Color yellow
! !

!PolygonMorph methodsFor: 'event handling' stamp: 'di 8/20/2000 14:29'!
handlesMouseDown: evt

	^ (super handlesMouseDown: evt) or: [evt shiftPressed]! !

!PolygonMorph methodsFor: 'menu' stamp: 'sw 8/17/2007 01:12'!
handlesShowingPhrase
	"Answer a phrase characterizing whether handles are showing or not."

	^ (self showingHandles ifTrue: ['<yes>'] ifFalse: ['<no>']), ('show handles' translated)! !

!PolygonMorph methodsFor: 'testing' stamp: 'wiz 7/18/2004 23:00'!
hasArrows
	"Are all the conditions meet for having arrows?"
	^ (closed
		or: [arrows == #none
				or: [vertices size < 2]]) not! !

!PolygonMorph methodsFor: 'private' stamp: 'di 8/31/2000 13:46'!
includesHandle: aMorph

	handles ifNil: [^ false].
	^ handles includes: aMorph! !

!PolygonMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:10'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	vertices := Array
				with: 5 @ 0
				with: 20 @ 10
				with: 0 @ 20.
	closed := true.
	smoothCurve := false.
	arrows := #none.
	self computeBounds! !

!PolygonMorph methodsFor: 'editing' stamp: 'wiz 2/8/2006 18:19'!
insertVertexAt: anIndex put: aValue
	"This serves as a hook and a backstop for MixedCurveMorph."
	self setVertices: (vertices copyReplaceFrom: anIndex + 1 to: anIndex 
									with: (Array with: aValue)).! !

!PolygonMorph methodsFor: 'debug and other' stamp: 'di 9/26/97 10:33'!
installModelIn: aWorld
	aWorld isWorldMorph ifTrue: [self addHandles]! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 3/30/2002 12:29'!
intersectionsWith: aRectangle
	"Answer a Set of points where the given Rectangle intersects with me.
	Ignores arrowForms."

	| retval |
	retval := IdentitySet new: 4.
	(self bounds intersects: aRectangle) ifFalse: [^ retval].

	self lineSegmentsDo: [ :lp1 :lp2 | | polySeg |
		polySeg := LineSegment from: lp1 to: lp2.
		aRectangle lineSegmentsDo: [ :rp1 :rp2 | | rectSeg int |
			rectSeg := LineSegment from: rp1 to: rp2.
			int := polySeg intersectionWith: rectSeg.
			int ifNotNil: [ retval add: int ].
		].
	].

	^retval
! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 2/15/2001 15:45'!
intersectionWithLineSegmentFromCenterTo: aPoint 
	^self closestPointTo: aPoint! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 4/27/2003 16:15'!
intersects: aRectangle 
	"Answer whether any of my segments intersects aRectangle, which is in World coordinates."
	| rect |
	(super intersects: aRectangle) ifFalse: [ ^false ].
	rect := self bounds: aRectangle in: self world.
	self
		lineSegmentsDo: [:p1 :p2 | (rect intersectsLineFrom: p1 to: p2)
				ifTrue: [^ true]].
	^ false! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 7/18/2003 17:38'!
isBordered
	^false! !

!PolygonMorph methodsFor: 'access' stamp: 'sw 8/25/2000 22:37'!
isClosed
	^ closed! !

!PolygonMorph methodsFor: 'access' stamp: 'di 9/7/2000 16:18'!
isCurve
	^ smoothCurve! !

!PolygonMorph methodsFor: 'testing' stamp: 'fbs 8/10/2013 20:50'!
isCurvier
	"Test used by smoothing routines.  If true use true closed curve splines for closed curves. If not mimic old stodgy curveMorph curves with one sharp bend.. Override this routine in classes where backward compatability is still needed."
	^ CurvierMorph drawCurvier! !

!PolygonMorph methodsFor: 'testing' stamp: 'wiz 5/2/2004 22:03'!
isCurvy
	"Test for significant curves.  
	Small smoothcurves in practice are straight."
	^ smoothCurve
		and: [vertices size > 2]! !

!PolygonMorph methodsFor: 'testing' stamp: 'nk 10/13/2003 18:36'!
isLineMorph
	^closed not! !

!PolygonMorph methodsFor: 'access' stamp: 'jm 11/19/97 18:55'!
isOpen
	^ closed not! !

!PolygonMorph methodsFor: 'dropping/grabbing' stamp: 'di 9/8/2000 09:56'!
justDroppedInto: newOwner event: evt

	| delta |
	(newOwner isKindOf: PasteUpMorph) ifTrue:
		["Compensate for border width so that gridded drop
			is consistent with gridded drag of handles."
		delta := borderWidth+1//2.
		self position: (newOwner gridPoint: self position + delta) - delta].
	^ super justDroppedInto: newOwner event: evt! !

!PolygonMorph methodsFor: 'attachments' stamp: 'nk 2/25/2001 17:19'!
lastVertex
	^vertices last! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 7/18/2003 17:06'!
lineBorderColor
	^self borderColor! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 7/18/2003 17:06'!
lineBorderColor: aColor
	self borderColor: aColor! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 7/18/2003 17:07'!
lineBorderWidth

	^self borderWidth! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 7/18/2003 17:46'!
lineBorderWidth: anInteger

	self borderWidth: anInteger! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 7/18/2003 16:48'!
lineColor
	^self borderColor! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 7/18/2003 16:48'!
lineColor: aColor
	self borderColor: aColor! !

!PolygonMorph methodsFor: 'private' stamp: 'di 11/21/97 21:29'!
lineSegments
	| lineSegments |
	lineSegments := OrderedCollection new.
	self lineSegmentsDo: [:p1 :p2 | lineSegments addLast: (Array with: p1 with: p2)].
	^ lineSegments! !

!PolygonMorph methodsFor: 'smoothing' stamp: 'nice 12/27/2009 03:11'!
lineSegmentsDo: endPointsBlock 
	"Emit a sequence of segment endpoints into endPointsBlock."
	"Unlike the method this one replaces we expect the curve 
	coefficents not the dirivatives"
	"Also unlike the replaced method the smooth closed curve
	does 
	not need an extra vertex. 
	We take care of the extra endpoint here. Just like for 
	segmented curves."
	| cs x y beginPoint |
	vertices size < 1
		ifTrue: [^ self].
	"test too few vertices first"
	self isCurvy
		ifFalse: [beginPoint := nil.
			"smoothCurve 
			ifTrue: [cs := self coefficients]."
			"some things still depend on smoothCurves having 
			curveState"
			vertices
				do: [:vert | 
					beginPoint
						ifNotNil: [endPointsBlock value: beginPoint value: vert].
					beginPoint := vert].
			(closed
					or: [vertices size = 1])
				ifTrue: [endPointsBlock value: beginPoint value: vertices first].
			^ self].
	"For curves we include all the interpolated sub segments."
	"self assert: [(vertices size > 2 )].	"
	cs := self coefficients.
	beginPoint := (x := cs first first) @ (y := cs fifth first).
	(closed
		ifTrue: [1 to: cs first size]
		ifFalse: [1 to: cs first size - 1])
		do: [:i | | x1 y1 endPoint n y2 t x3 y3 x2 | 
			"taylor series coefficients"
			x1 := cs second at: i.
			y1 := cs sixth at: i.
			x2 := cs third at: i.
			y2 := cs seventh at: i.
			x3 := cs fourth at: i.
			y3 := cs eighth at: i.
			n := cs ninth at: i.
			"guess n 
			n := 5 max: (x2 abs + y2 abs * 2.0 + (cs third atWrap:
			i 
			+ 1) abs + (cs seventh atWrap: i + 1) abs / 100.0) 
			rounded."
			1
				to: n - 1
				do: [:j | 
					t := j asFloat / n asFloat.
					endPoint := x3 * t + x2 * t + x1 * t + x @ (y3 * t + y2 * t + y1 * t + y).
					endPointsBlock value: beginPoint value: endPoint.
					beginPoint := endPoint].
			endPoint := (x := cs first atWrap: i + 1) @ (y := cs fifth atWrap: i + 1).
			endPointsBlock value: beginPoint value: endPoint.
			beginPoint := endPoint]! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 7/18/2003 16:47'!
lineWidth

	^self borderWidth! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 7/18/2003 17:48'!
lineWidth: anInteger

	self borderWidth: (anInteger rounded max: 1)! !

!PolygonMorph methodsFor: 'caching' stamp: 'di 11/13/97 15:16'!
loadCachedState
	"Prepare for fast response -- next page of a book?"
	self filledForm.
	self arrowForms! !

!PolygonMorph methodsFor: 'menu' stamp: '6/9/97 21:32 di'!
makeBackArrow
	arrows := #back.
	self computeBounds! !

!PolygonMorph methodsFor: 'menu' stamp: '6/9/97 21:32 di'!
makeBothArrows
	arrows := #both.
	self computeBounds! !

!PolygonMorph methodsFor: 'menu' stamp: 'di 8/20/2000 14:27'!
makeClosed
	closed := true.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!PolygonMorph methodsFor: 'menu' stamp: '6/9/97 21:32 di'!
makeForwardArrow
	arrows := #forward.
	self computeBounds! !

!PolygonMorph methodsFor: 'menu' stamp: '6/9/97 21:32 di'!
makeNoArrows
	arrows := #none.
	self computeBounds! !

!PolygonMorph methodsFor: 'menu' stamp: 'di 8/20/2000 14:27'!
makeOpen
	closed := false.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!PolygonMorph methodsFor: 'access' stamp: 'sw 8/23/2000 16:16'!
makeOpenOrClosed
	"toggle the open/closed status of the receiver"
	closed ifTrue: [self makeOpen] ifFalse: [self makeClosed]! !

!PolygonMorph methodsFor: 'geometry' stamp: 'dgd 2/22/2003 18:57'!
mergeDropThird: mv in: hv from: shared 
	"We are merging two polygons.  In this case, they have at least three identical shared vertices.  Make sure they are sequential in each, and drop the middle one from vertex lists mv, hv, and shared.  First vertices on lists are identical already."

	"know (mv first = hv first)"

	| mdrop vv |
	(shared includes: (mv at: mv size - 2)) 
		ifTrue: [(shared includes: mv last) ifTrue: [mdrop := mv last]]
		ifFalse: 
			[(shared includes: mv last) 
				ifTrue: [(shared includes: mv second) ifTrue: [mdrop := mv first]]].
	(shared includes: (mv third)) 
		ifTrue: [(shared includes: mv second) ifTrue: [mdrop := mv second]].
	mdrop ifNil: [^nil].
	mv remove: mdrop.
	hv remove: mdrop.
	shared remove: mdrop.
	[shared includes: mv first] whileFalse: 
			["rotate them"

			vv := mv removeFirst.
			mv addLast: vv].
	[mv first = hv first] whileFalse: 
			["rotate him until same shared vertex is first"

			vv := hv removeFirst.
			hv addLast: vv]! !

!PolygonMorph methodsFor: 'geometry' stamp: 'dgd 2/22/2003 18:56'!
merge: aPolygon 
	"Expand myself to enclose the other polygon.  (Later merge overlapping or disjoint in a smart way.)  For now, the two polygons must share at least two vertices.  Shared vertices must come one after the other in each polygon.  Polygons must not overlap."

	| shared mv vv hv xx |
	shared := vertices select: [:mine | aPolygon vertices includes: mine].
	shared size < 2 ifTrue: [^nil].	"not sharing a segment"
	mv := vertices asOrderedCollection.
	[shared includes: mv first] whileFalse: 
			["rotate them"

			vv := mv removeFirst.
			mv addLast: vv].
	hv := aPolygon vertices asOrderedCollection.
	[mv first = hv first] whileFalse: 
			["rotate him until same shared vertex is first"

			vv := hv removeFirst.
			hv addLast: vv].
	[shared size > 2] whileTrue: 
			[shared := shared asOrderedCollection.
			(self 
				mergeDropThird: mv
				in: hv
				from: shared) ifNil: [^nil]].
	"works by side effect on the lists"
	(mv second) = hv last 
		ifTrue: 
			[mv
				removeFirst;
				removeFirst.
			^self setVertices: (hv , mv) asArray].
	(hv second) = mv last 
		ifTrue: 
			[hv
				removeFirst;
				removeFirst.
			^self setVertices: (mv , hv) asArray].
	(mv second) = (hv second) 
		ifTrue: 
			[hv removeFirst.
			mv remove: (mv second).
			xx := mv removeFirst.
			^self setVertices: (hv , (Array with: xx) , mv reversed) asArray].
	mv last = hv last 
		ifTrue: 
			[mv removeLast.
			hv removeFirst.
			^self setVertices: (mv , hv reversed) asArray].
	^nil! !

!PolygonMorph methodsFor: 'attachments' stamp: 'nice 1/10/2010 19:20'!
midpoint
	"Answer the midpoint along my segments"
	| middle |
	middle := self totalLength.
	middle < 2 ifTrue: [ ^ self center ].
	middle := middle / 2.
	self lineSegmentsDo: [ :a :b | | dist |
		dist := (a dist: b).
		middle < dist
			ifTrue: [ ^(a + ((b - a) * (middle / dist))) asIntegerPoint ].
		middle := middle - dist.
	].
	self error: 'can''t happen'! !

!PolygonMorph methodsFor: 'access' stamp: 'wiz 6/24/2004 22:50'!
midVertices
	"Return and array of midpoints for this line or closed curve"
	| midPts nextVertIx tweens |
	vertices size < 2
		ifTrue: [^ vertices].
	midPts := OrderedCollection new.
	nextVertIx := 2.
	tweens := OrderedCollection new.
	tweens add: vertices first asIntegerPoint.
	"guarantee at least two points."
	self
		lineSegmentsDo: [:p1 :p2 | 
			tweens addLast: p2 asIntegerPoint.
			p2
					= (vertices atWrap: nextVertIx)
				ifTrue: ["Found endPoint."
					midPts addLast: (tweens atWrap: tweens size + 1 // 2)
							+ (tweens at: tweens size // 2 + 1) // 2.
					"wiz 6/19/2004 20:11 adjusted to handle  
					one segment properly"
					tweens := OrderedCollection new.
					tweens add: p2 asIntegerPoint.
					"guarantee at least two points."
					nextVertIx := nextVertIx + 1]].
	^ midPts asArray! !

!PolygonMorph methodsFor: 'event handling' stamp: 'sw 8/21/2008 22:21'!
mouseDown: evt
	"Handle a mouse-down event."

	^ (evt shiftPressed and: [(self hasProperty: #activateOnShift) not])
		ifTrue: [((owner isKindOf: PolygonMorph) and: [owner includesHandle: self])
					ifTrue: ["Prevent insertion handles from getting edited"
							^ super mouseDown: evt].
				self toggleHandles.
				handles ifNil: [^ self].
				vertices withIndexDo:  "Check for click-to-drag at handle site"
					[:vertPt :vertIndex |
					((handles at: vertIndex*2-1 ifAbsent: [ ^self ]) containsPoint: evt cursorPoint) ifTrue:
						["If clicked near a vertex, jump into drag-vertex action"
						evt hand newMouseFocus: (handles at: vertIndex*2-1)]]]
		ifFalse: [super mouseDown: evt]! !

!PolygonMorph methodsFor: 'editing' stamp: 'wiz 2/8/2006 18:25'!
newVertex: ix event: evt fromHandle: handle
	"Insert a new vertex and fix everything up!! Install the drag-handle of the new vertex as recipient of further mouse events."

	| pt |
	"(self hasProperty: #noNewVertices) ifFalse:
		[pt := evt cursorPoint.
		self setVertices: (vertices copyReplaceFrom: ix + 1 to: ix with: (Array with: pt)).
		evt hand newMouseFocus: (handles at: ((ix + 1) * 2) - 1)]"
	"modified to remove now vestigial test. see PolygonMorph class>>arrowprototype"
	pt := evt cursorPoint.
	self  insertVertexAt: ix put:  pt .
	evt hand newMouseFocus: (handles at: ((ix + 1) * 2) - 1)! !

!PolygonMorph methodsFor: 'editing' stamp: 'ar 3/18/2001 17:28'!
newVertex: arg1 fromHandle: arg2 afterVert: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self newVertex: arg1 event: arg2 fromHandle: arg3! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 4/27/2003 15:39'!
nextDuplicateVertexIndex
	vertices
		doWithIndex: [:vert :index | ((index between: 2 and: vertices size - 1)
					and: [| epsilon v1 v2 | 
						v1 := vertices at: index - 1.
						v2 := vertices at: index + 1.
						epsilon := ((v1 x - v2 x) abs max: (v1 y - v2 y) abs)
									// 32 max: 1.
						vert
							onLineFrom: v1
							to: v2
							within: epsilon])
				ifTrue: [^ index]].
	^ 0! !

!PolygonMorph methodsFor: 'smoothing' stamp: 'wiz 11/16/2004 19:54'!
nextToFirstPoint
	"For arrow direction"
	self isCurvy
		ifTrue: [curveState
				ifNil: [self coefficients].
			^ curveState second]
		ifFalse: [^ vertices second]! !

!PolygonMorph methodsFor: 'smoothing' stamp: 'wiz 11/16/2004 19:51'!
nextToLastPoint
	"For arrow direction"
	self isCurvy
		ifTrue: [curveState
				ifNil: [self coefficients].
			^ curveState third]
		ifFalse: [^ vertices at: vertices size - 1]! !

!PolygonMorph methodsFor: 'attachments' stamp: 'nice 2/14/2011 21:49'!
nudgeForLabel: aRectangle
	"Try to move the label off me. Prefer labels on the top and right."

	| i flags nudge |
	(self bounds intersects: aRectangle) ifFalse: [^ 0 @ 0 ].
	flags := 0.
	nudge := 0 @ 0.
	i := 1.
	aRectangle lineSegmentsDo: [ :rp1 :rp2 | | rectSeg |
		rectSeg := LineSegment from: rp1 to: rp2.
		self straightLineSegmentsDo: [ :lp1 :lp2 | | polySeg int |
			polySeg := LineSegment from: lp1 to: lp2.
			int := polySeg intersectionWith: rectSeg.
			int ifNotNil: [ flags := flags bitOr: i ].
		].
		i := i * 2.
	].
	"Now flags has bitflags for which sides"
	nudge := flags caseOf: {
"no intersection"
		[ 2r0000 ] -> [ 0 @ 0 ].
"2 adjacent sides only" 
		[ 2r1001 ] -> [ 1 @ 1 ].
		[ 2r0011 ] -> [ -1 @ 1 ].
		[ 2r1100 ] -> [ 1 @ -1 ].
		[ 2r0110 ] -> [ -1 @ -1 ].
"2 opposite sides only" 
		[ 2r1010 ] -> [ 0 @ -1 ].
		[ 2r0101 ] -> [ 1 @ 0 ].
"only 1 side" 
		[ 2r1000 ] -> [ -1 @ 0 ].
		[ 2r0001 ] -> [ 0 @ -1 ].
		[ 2r0010 ] -> [ 1 @ 0 ].
		[ 2r0100 ] -> [ 0 @ 1 ].
"3 sides" 
		[ 2r1011 ] -> [ 0 @ 1 ].
		[ 2r1101 ] -> [ 1 @ 0 ].
		[ 2r1110 ] -> [ 0 @ -1 ].
		[ 2r0111 ] -> [ -1 @ 0 ].
 "all sides" 
		[ 2r1111 ] -> [ 1 @ -1 "move up and to the right" ].
	}.
	^nudge! !

!PolygonMorph methodsFor: 'access' stamp: 'sw 10/3/2007 02:02'!
openOrClosePhrase
	"Answer a string indicating whether the receiver is open or closed."

	^ (closed ifTrue: ['<yes>'] ifFalse: ['<no>']), 'closed' translated! !

!PolygonMorph methodsFor: 'private' stamp: 'ul 12/12/2009 14:07'!
privateMoveBy: delta
	super privateMoveBy: delta.
	vertices := vertices collect: [:p | p + delta].
	self arrowForms do: [:f | f offset: f offset + delta].
	curveState := nil.  "Force recomputation"
	(self valueOfProperty: #referencePosition) ifNotNil:
		[:oldPos | self setProperty: #referencePosition toValue: oldPos + delta]! !

!PolygonMorph methodsFor: 'menu' stamp: 'di 9/7/2000 13:10'!
quickFill: ignored! !

!PolygonMorph methodsFor: 'shaping' stamp: 'wiz 1/8/2005 19:20'!
rectOval
	"Set my vertices to an array of corner vertices.
	Order of vertices is in the tradion of warpblt quads."

	self setVertices: self bounds corners.! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 4/18/2002 16:58'!
reduceVertices
	"Reduces the vertices size, when 3 vertices are on the same line with a 
	little epsilon. Based on code by Steffen Mueller"
	| dup |
	[ (dup := self nextDuplicateVertexIndex) > 0 ] whileTrue: [
		self setVertices: (vertices copyWithoutIndex: dup)
	].
	^vertices size.! !

!PolygonMorph methodsFor: 'geometry eToy' stamp: 'di 9/24/2000 08:38'!
referencePosition 
	"Return the current reference position of the receiver"
	^ self valueOfProperty: #referencePosition ifAbsent: [super referencePosition]
! !

!PolygonMorph methodsFor: 'caching' stamp: 'di 9/4/2000 13:36'!
releaseCachedState
	super releaseCachedState.
	filledForm := nil.
	arrowForms := nil.
	borderForm := nil.
	curveState := nil.
	(self hasProperty: #flex) ifTrue:
		[self removeProperty: #unflexedVertices;
			removeProperty: #flex].
! !

!PolygonMorph methodsFor: 'menu' stamp: 'tk 9/2/97 16:04'!
removeHandles
	"tk 9/2/97 allow it to be called twice (when nil already)"

	handles ifNotNil: [
		handles do: [:h | h delete].
		handles := nil].! !

!PolygonMorph methodsFor: 'dashes' stamp: 'nk 4/5/2001 16:02'!
removeVertex: aVert
	"Make sure that I am not left with less than two vertices"
	| newVertices |
	vertices size < 2 ifTrue: [ ^self ].
	newVertices := vertices copyWithout: aVert.
	newVertices size caseOf: {
		[1] -> [ newVertices := { newVertices first . newVertices first } ].
		[0] -> [ newVertices := { aVert . aVert } ]
	} otherwise: [].
	self setVertices: newVertices 
! !

!PolygonMorph methodsFor: 'debug and other' stamp: 'wiz 5/1/2004 00:21'!
rotateTestFlip: aBool 
	"Return one copy of me for each vertex using each vertex as  
	the  
	starting point.  
	Vary to border color to destinguish the copies.  
	This tests closed curves for their consistency.  
	The flip boolean tests the reversed rotations."
	| len colors verts flip |
	verts := self vertices.
	flip := aBool == true
				ifTrue: [1]
				ifFalse: [0].
	len := verts size.
	colors := Color wheel: len*2 .
	(1 to: len)
		do: [:i | | j | (self copy
				borderColor: (colors at: (j:=i * 2 - flip));
				 yourself)
				setVertices: (verts flipRotated: j);
				 openInWorld]! !

!PolygonMorph methodsFor: 'geometry eToy' stamp: 'di 9/24/2000 09:21'!
rotationCenter
	"Return the rotation center of the receiver. The rotation center defines the relative offset inside the receiver's bounds for locating the reference position."
	| refPos |
	refPos := self valueOfProperty: #referencePosition
		ifAbsent: [^ 0.5@0.5].
	^ (refPos - self bounds origin) / self bounds extent asFloatPoint! !

!PolygonMorph methodsFor: 'geometry eToy' stamp: 'di 9/24/2000 09:31'!
rotationCenter: aPointOrNil
	"Set the new rotation center of the receiver. The rotation center defines the relative offset inside the receiver's bounds for locating the reference position."
	| box |
	aPointOrNil isNil
		ifTrue: [self removeProperty: #referencePosition]
		ifFalse: [box := self bounds.
				self setProperty: #referencePosition
					toValue: box origin + (aPointOrNil * box extent)]
! !

!PolygonMorph methodsFor: 'rotate scale and flex' stamp: 'di 9/24/2000 08:42'!
rotationDegrees

	^ self forwardDirection! !

!PolygonMorph methodsFor: 'halo control' stamp: 'di 9/24/2000 09:42'!
rotationDegrees: degrees 
	| flex center |
	(center := self valueOfProperty: #referencePosition) ifNil:
		[self setProperty: #referencePosition toValue: (center := self bounds center)].
	flex := (MorphicTransform offset: center negated)
			withAngle: (degrees - self forwardDirection) degreesToRadians.
	self setVertices: (vertices collect: [:v | (flex transform: v) - flex offset]).
	self forwardDirection: degrees.

! !

!PolygonMorph methodsFor: 'menu' stamp: 'di 9/24/2000 09:25'!
setRotationCenterFrom: aPoint
	"Polygons store their referencePosition."
	self setProperty: #referencePosition toValue: aPoint! !

!PolygonMorph methodsFor: 'private' stamp: 'di 9/8/2000 10:36'!
setVertices: newVertices
	vertices := newVertices.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!PolygonMorph methodsFor: 'menu' stamp: 'sw 8/19/2000 15:17'!
showingHandles
	^ handles notNil! !

!PolygonMorph methodsFor: 'menu' stamp: 'sw 8/19/2000 15:16'!
showOrHideHandles
	self showingHandles
		ifTrue:	[self removeHandles]
		ifFalse:	[self addHandles]! !

!PolygonMorph methodsFor: 'smoothing' stamp: 'wiz 1/7/2005 19:53'!
slopes: knots 
	"Choose slopes according to state of polygon and preferences"
	self isCurvy
		ifFalse: [^ knots segmentedSlopes].
	^ (closed
			and: [self isCurvier])
		ifTrue: [knots closedCubicSlopes]
		ifFalse: [knots naturalCubicSlopes]! !

!PolygonMorph methodsFor: 'access' stamp: 'wiz 6/6/2004 21:56'!
smoothOrSegmentedPhrase
				| lineName |
	lineName := (closed
						ifTrue: ['outline']
						ifFalse: ['line']) translated.

			^ self isCurve
				ifTrue: ['make segmented {1}' translated format: {lineName}]
				ifFalse: ['make smooth {1}' translated format: {lineName}].! !

!PolygonMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:43'!
specifyDashedLine

	| executableSpec newSpec |
	executableSpec := UIManager default
		request:
'Enter a dash specification as
{ major dash length. minor dash length. minor dash color }
The major dash will have the normal border color.
A blank response will remove the dash specification.
[Note: You may give 5 items as, eg, {10. 5. Color white. 0. 3}
where the 4th ityem is zero, and the 5th is the number of pixels
by which the dashes will move in each step of animation]' translated
		initialAnswer: '{ 10. 5. Color red }'.
	executableSpec isEmpty ifTrue:
		[^ self stopStepping; dashedBorder: nil].
	newSpec := [Compiler evaluate: executableSpec] ifError:
		[^ self stopStepping; dashedBorder: nil].
	newSpec first isNumber & newSpec second isNumber & newSpec third isColor ifFalse:
		[^ self stopStepping; dashedBorder: nil].
	newSpec size = 3 ifTrue:
		[^ self stopStepping; dashedBorder: newSpec].
	(newSpec size = 5 and: [newSpec fourth isNumber & newSpec fifth isNumber]) ifTrue:
		[^ self dashedBorder: newSpec; startStepping].
! !

!PolygonMorph methodsFor: 'menu' stamp: 'di 10/3/2000 07:12'!
standardArrows

	self removeProperty: #arrowSpec.
	self computeBounds! !

!PolygonMorph methodsFor: 'stepping and presenter' stamp: 'dgd 2/22/2003 18:58'!
step
	borderDashSpec ifNil: [^super step].
	borderDashSpec size < 5 ifTrue: [^super step].

	"Only for dashed lines with creep"
	borderDashSpec at: 4 put: (borderDashSpec fourth) + borderDashSpec fifth.
	self changed.
	^super step! !

!PolygonMorph methodsFor: 'testing' stamp: 'sw 7/28/2008 13:24'!
stepTime
	"Answer the desired time between steps in milliseconds."

	^ self topRendererOrSelf player ifNotNil: [10] ifNil: [100]

	"NB:  in all currently known cases, polygons are not actually wrapped  in TransformationMorphs, so the #topRendererOrSelf call above is probably redundant, but is retained for safety."! !

!PolygonMorph methodsFor: 'geometry' stamp: 'nk 3/6/2001 16:36'!
straighten
	self setVertices: { vertices first . vertices last }! !

!PolygonMorph methodsFor: 'smoothing' stamp: 'nk 4/23/2002 15:48'!
straightLineSegmentsDo: endPointsBlock
	"Emit a sequence of segment endpoints into endPointsBlock.
	Work the same way regardless of whether I'm curved."
	| beginPoint |
	beginPoint := nil.
		vertices do:
			[:vert | beginPoint ifNotNil:
				[endPointsBlock value: beginPoint
								value: vert].
			beginPoint := vert].
		(closed or: [vertices size = 1])
			ifTrue: [endPointsBlock value: beginPoint
									value: vertices first].! !

!PolygonMorph methodsFor: 'menu' stamp: 'di 8/20/2000 14:31'!
toggleHandles

	handles ifNil: [self addHandles] ifNotNil: [self removeHandles].

! !

!PolygonMorph methodsFor: 'menu' stamp: 'di 9/7/2000 15:43'!
toggleSmoothing

	smoothCurve := smoothCurve not.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!PolygonMorph methodsFor: 'attachments' stamp: 'nk 7/3/2003 14:42'!
totalLength
	"Answer the full length of my segments. Can take a long time if I'm curved."
	| length |
	length := 0.
	self lineSegmentsDo: [ :a :b | length := length + (a dist: b) ].
	^length.! !

!PolygonMorph methodsFor: 'geometry' stamp: 'ar 10/6/2000 15:40'!
transformedBy: aTransform
	self setVertices: (self vertices collect:[:v| aTransform localPointToGlobal: v])! !

!PolygonMorph methodsFor: 'private' stamp: 'nk 3/27/2001 21:23'!
transformVerticesFrom: oldOwner to: newOwner
	| oldTransform newTransform world newVertices |
	world := self world.
	oldTransform := oldOwner
		ifNil: [ IdentityTransform new ]
		ifNotNil: [ oldOwner transformFrom: world ].
	newTransform := newOwner
		ifNil: [ IdentityTransform new ]
		ifNotNil: [ newOwner transformFrom: world ].
	newVertices := vertices collect: [ :ea | newTransform globalPointToLocal:
		(oldTransform localPointToGlobal: ea) ].
	self setVertices: newVertices.
! !

!PolygonMorph methodsFor: 'menu' stamp: 'ul 11/15/2010 11:59'!
unrotatedLength
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"

	vertices size = 2 ifTrue:
		[^ (vertices second - vertices first) r].

	^ ((PolygonMorph new setVertices: vertices) rotationDegrees: self rotationDegrees negated) height! !

!PolygonMorph methodsFor: 'menu' stamp: 'ul 11/15/2010 11:59'!
unrotatedLength: aLength
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"

	vertices size = 2 ifTrue: [^ self arrowLength: aLength].

	self setVertices: ((((PolygonMorph new setVertices: vertices) rotationDegrees: self rotationDegrees negated) height: aLength) rotationDegrees: 0) vertices! !

!PolygonMorph methodsFor: 'menu' stamp: 'ul 11/15/2010 11:59'!
unrotatedWidth
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"
	
	vertices size = 2 ifTrue: [^ self borderWidth].
	^ ((PolygonMorph new setVertices: vertices) rotationDegrees: self rotationDegrees negated) width! !

!PolygonMorph methodsFor: 'menu' stamp: 'sw 9/24/2002 18:18'!
unrotatedWidth: aWidth
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"

	self borderWidth: aWidth! !

!PolygonMorph methodsFor: 'editing' stamp: 'nice 2/8/2010 08:45'!
updateHandles
	self isCurvy
		ifTrue: [handles first center: vertices first.
			handles last center: vertices last.
			self midVertices
				withIndexDo: [:midPt :vertIndex | (closed
							or: [vertIndex < vertices size])
						ifTrue: [| newVert |
							newVert := handles atWrap: vertIndex * 2.
							newVert position: midPt - (newVert extent // 2)]]]
		ifFalse: [vertices
				withIndexDo: [:vertPt :vertIndex |
					| oldVert | 
					oldVert := handles at: vertIndex * 2 - 1.
					oldVert position: vertPt - (oldVert extent // 2).
					(closed
							or: [vertIndex < vertices size])
						ifTrue: [| newVert |
							newVert := handles at: vertIndex * 2.
							newVert position: vertPt
									+ (vertices atWrap: vertIndex + 1) - newVert extent // 2 + (1 @ -1)]]]! !

!PolygonMorph methodsFor: 'dashes' stamp: 'nk 2/25/2001 17:05'!
vertexAt: n
	^vertices at: (n min: vertices size).! !

!PolygonMorph methodsFor: 'access' stamp: 'sw 9/14/97 18:22'!
vertices
	^ vertices! !

!PolygonMorph methodsFor: 'editing' stamp: 'tetha 12/1/2006 19:25'!
verticesAt: anInteger put: aPoint

	self vertices at: anInteger put: aPoint asFloatPoint.
	self computeBounds! !

!PolygonMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:09'!
vertices: verts color: aColor borderWidth: borderWidthInteger borderColor: anotherColor 
	super initialize.
""
	vertices := verts.
	color := aColor.
	borderWidth := borderWidthInteger.
	borderColor := anotherColor.
	closed := vertices size > 2.
	arrows := #none.
	self computeBounds! !

!PolygonMorph methodsFor: 'testing' stamp: 'dgd 2/22/2003 18:58'!
wantsSteps
	super wantsSteps ifTrue: [^true].

	"For crawling ants effect of dashed line."
	borderDashSpec ifNil: [^false].
	^borderDashSpec size = 5 and: [(borderDashSpec fifth) > 0]! !

BorderedMorph subclass: #RectangleMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!RectangleMorph commentStamp: 'kfr 10/27/2003 11:12' prior: 0!
A subclass of BorderedMorph that supports different fillStyles.

RectangleMorph diagonalPrototype openInWorld.
RectangleMorph gradientPrototype openInWorld.!

RectangleMorph subclass: #AlignmentMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!AlignmentMorph commentStamp: 'kfr 10/27/2003 10:25' prior: 0!
Used for layout.
Since all morphs now support layoutPolicy the main use of this class is no longer needed.
Kept around for compability. 
Supports a few methods not found elsewhere that can be convenient, eg. newRow
!

!AlignmentMorph class methodsFor: 'instance creation' stamp: 'dgd 9/20/2003 19:05'!
columnPrototype
	"Answer a prototypical column"

	| sampleMorphs aColumn |
	sampleMorphs := #(red yellow green) collect:
		[:aColor | Morph new extent: 130 @ 38; color: (Color perform: aColor); setNameTo: aColor asString; yourself].
	aColumn := self inAColumn: sampleMorphs.
	aColumn setNameTo: 'Column'.
	aColumn color: Color veryVeryLightGray.
	aColumn cellInset: 4; layoutInset: 6.
	aColumn enableDragNDrop.
	aColumn setBalloonText: 'Things dropped into here will automatically be organized into a column. Once you have added your own items here, you will want to remove the sample colored rectangles that this started with, and you will want to change this balloon help message to one of your own!!' translated.
	^ aColumn! !

!AlignmentMorph class methodsFor: 'scripting' stamp: 'sw 11/16/2004 00:44'!
defaultNameStemForInstances
	"The code just below, now commented out, resulted in every instance of every sublcass of AlignmentMorph being given a default name of the form 'Alignment1', rather than the desired 'MoviePlayer1', 'ScriptEditor2', etc."

	"^ 'Alignment'"

	^ super defaultNameStemForInstances! !

!AlignmentMorph class methodsFor: 'instance creation' stamp: 'sw 11/2/2001 04:45'!
inAColumn: aCollectionOfMorphs
	"Answer a columnar AlignmentMorph holding the given collection"

	| col |
	col := self newColumn
		color: Color transparent;
		vResizing: #shrinkWrap;
		hResizing: #shrinkWrap;
		layoutInset: 1;
		borderColor: Color black;
		borderWidth: 1;
		wrapCentering: #center;
		cellPositioning: #topCenter.
	aCollectionOfMorphs do: [:each | col addMorphBack: each].
	^ col! !

!AlignmentMorph class methodsFor: 'instance creation' stamp: 'sw 11/5/2001 15:11'!
inARow: aCollectionOfMorphs
	"Answer a row-oriented AlignmentMorph holding the given collection"

	| aRow |
	aRow := self newRow
		color: Color transparent;
		vResizing: #shrinkWrap;
		hResizing: #shrinkWrap;
		layoutInset: 1;
		borderColor: Color black;
		borderWidth: 1;
		wrapCentering: #center;
		cellPositioning: #topCenter.
	aCollectionOfMorphs do: [ :each | aRow addMorphBack: each].
	^ aRow! !

!AlignmentMorph class methodsFor: 'instance creation' stamp: 'ar 11/9/2000 20:51'!
newColumn

	^ self new
		listDirection: #topToBottom;
		hResizing: #spaceFill;
		extent: 1@1;
		vResizing: #spaceFill
! !

!AlignmentMorph class methodsFor: 'instance creation' stamp: 'ar 11/9/2000 20:50'!
newRow

	^ self new
		listDirection: #leftToRight;
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		extent: 1@1;
		borderWidth: 0
! !

!AlignmentMorph class methodsFor: 'instance creation' stamp: 'ar 11/9/2000 20:37'!
newSpacer: aColor
	"Answer a space-filling instance of me of the given color."

	^ self new
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		layoutInset: 0;
		borderWidth: 0;
		extent: 1@1;
		color: aColor.
! !

!AlignmentMorph class methodsFor: 'instance creation' stamp: 'ar 11/9/2000 20:37'!
newVariableTransparentSpacer
	"Answer a space-filling instance of me of the given color."

	^ self new
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		layoutInset: 0;
		borderWidth: 0;
		extent: 1@1;
		color: Color transparent
! !

!AlignmentMorph class methodsFor: 'instance creation' stamp: 'dgd 9/20/2003 19:05'!
rowPrototype
	"Answer a prototypical row"

	| sampleMorphs aRow |
	sampleMorphs := (1 to: (2 + 3 atRandom)) collect:
		[:integer | EllipseMorph new extent: ((60 + (20 atRandom)) @ (80 + ((20 atRandom)))); color: Color random; setNameTo: ('egg',  integer asString); yourself].
	aRow := self inARow: sampleMorphs.
	aRow setNameTo: 'Row'.
	aRow enableDragNDrop.
	aRow cellInset: 6.
	aRow layoutInset: 8.
	aRow setBalloonText: 'Things dropped into here will automatically be organized into a row. Once you have added your own items here, you will want to remove the sample colored eggs that this started with, and you will want to change this balloon help message to one of your own!!' translated.
	aRow color: Color veryVeryLightGray.
	^ aRow

			"AlignmentMorph rowPrototype openInHand"! !

!AlignmentMorph methodsFor: 'visual properties' stamp: 'sw 11/5/2001 15:11'!
canHaveFillStyles
	"Return true if the receiver can have general fill styles; not just colors.
	This method is for gradually converting old morphs."

	^ self class == AlignmentMorph "no subclasses"! !

!AlignmentMorph methodsFor: 'object fileIn' stamp: 'gm 2/22/2003 13:12'!
convertOldAlignmentsNov2000: varDict using: smartRefStrm 
	"major change - much of AlignmentMorph is now implemented more generally in Morph"

	"These are going away 
	#('orientation' 'centering' 'hResizing' 'vResizing' 
	'inset' 'minCellSize' 'layoutNeeded' 'priorFullBounds')"

	| orientation centering hResizing vResizing inset minCellSize inAlignment |
	orientation := varDict at: 'orientation'.
	centering := varDict at: 'centering'.
	hResizing := varDict at: 'hResizing'.
	vResizing := varDict at: 'vResizing'.
	inset := varDict at: 'inset'.
	minCellSize := varDict at: 'minCellSize'.
	(orientation == #horizontal or: [orientation == #vertical]) 
		ifTrue: [self layoutPolicy: TableLayout new].
	self cellPositioning: #topLeft.
	self rubberBandCells: true.
	orientation == #horizontal ifTrue: [self listDirection: #leftToRight].
	orientation == #vertical ifTrue: [self listDirection: #topToBottom].
	centering == #topLeft ifTrue: [self wrapCentering: #topLeft].
	centering == #bottomRight ifTrue: [self wrapCentering: #bottomRight].
	centering == #center 
		ifTrue: 
			[self wrapCentering: #center.
			orientation == #horizontal 
				ifTrue: [self cellPositioning: #leftCenter]
				ifFalse: [self cellPositioning: #topCenter]].
	(inset isNumber or: [inset isPoint]) ifTrue: [self layoutInset: inset].
	(minCellSize isNumber or: [minCellSize isPoint]) 
		ifTrue: [self minCellSize: minCellSize].
	(self hasProperty: #clipToOwnerWidth) ifTrue: [self clipSubmorphs: true].

	"now figure out if our owner was an AlignmentMorph, even if it is reshaped..."
	inAlignment := false.
	owner isMorph 
		ifTrue: [(owner isAlignmentMorph) ifTrue: [inAlignment := true]]
		ifFalse: 
			["e.g., owner may be reshaped"

			(owner class instanceVariablesString 
				findString: 'orientation centering hResizing vResizing') > 0 
				ifTrue: 
					["this was an alignment morph being reshaped"

					inAlignment := true]].
	"And check for containment in system windows"
	owner isSystemWindow ifTrue: [inAlignment := true].
	(hResizing == #spaceFill and: [inAlignment not]) 
		ifTrue: [self hResizing: #shrinkWrap]
		ifFalse: [self hResizing: hResizing].
	(vResizing == #spaceFill and: [inAlignment not]) 
		ifTrue: [self vResizing: #shrinkWrap]
		ifFalse: [self vResizing: vResizing]! !

!AlignmentMorph methodsFor: 'objects from disk' stamp: 'tk 11/26/2004 05:51'!
convertToCurrentVersion: varDict refStream: smartRefStrm

	| newish |
	newish := super convertToCurrentVersion: varDict refStream:
smartRefStrm.

	"major change - much of AlignmentMorph is now implemented
more generally in Morph"
	varDict at: 'hResizing' ifPresent: [ :x |
		^ newish convertOldAlignmentsNov2000: varDict using:
smartRefStrm].
	^ newish
! !

!AlignmentMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:36'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 0! !

!AlignmentMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:24'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color
		r: 0.8
		g: 1.0
		b: 0.8! !

!AlignmentMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 19:19'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	self layoutPolicy: TableLayout new;
	 listDirection: #leftToRight;
	 wrapCentering: #topLeft;
	 hResizing: #spaceFill;
	 vResizing: #spaceFill;
	 layoutInset: 2;
	 rubberBandCells: true! !

!AlignmentMorph methodsFor: 'classification' stamp: 'di 5/7/1998 01:20'!
isAlignmentMorph

	^ true
! !

!AlignmentMorph methodsFor: 'initialization' stamp: 'ar 11/9/2000 20:34'!
openInWindowLabeled: aString inWorld: aWorld

	self layoutInset: 0.
	^super openInWindowLabeled: aString inWorld: aWorld.! !

!AlignmentMorph methodsFor: 'event handling' stamp: 'sw 5/6/1998 15:58'!
wantsKeyboardFocusFor: aSubmorph
	aSubmorph wouldAcceptKeyboardFocus ifTrue: [^ true].
	^ super wantsKeyboardFocusFor: aSubmorph! !

AlignmentMorph subclass: #ScorePlayerMorph
	instanceVariableNames: 'scorePlayer trackInstNames instrumentSelector scrollSlider'
	classVariableNames: 'LastMIDIPort'
	poolDictionaries: ''
	category: 'Morphic-Sound'!

!ScorePlayerMorph commentStamp: '<historical>' prior: 0!
A ScorePlayerMorph mediates between a score such as a MIDIScore, a PianoRollScoreMorph, and the actual SoundPlayer synthesizer.

It provides control over volume, tempo, instrumentation, and location in the score.!

!ScorePlayerMorph class methodsFor: 'parts bin' stamp: 'sw 10/20/2001 17:17'!
descriptionForPartsBin
	^ self partName: 	'ScorePlayer'
		categories:		#('Multimedia')
		documentation:	' Mediates between a score such as a MIDIScore, a PianoRollScoreMorph, and the actual SoundPlayer synthesizer'! !

!ScorePlayerMorph class methodsFor: 'fileIn/Out' stamp: 'sd 2/6/2002 21:37'!
fileReaderServicesForFile: fullName suffix: suffix

	^(suffix = 'mid') | (suffix = '*') 
		ifTrue: [ self services]
		ifFalse: [#()]
! !

!ScorePlayerMorph class methodsFor: 'class initialization' stamp: 'ar 9/5/2010 09:25'!
initialize

	FileServices registerFileReader: self! !

!ScorePlayerMorph class methodsFor: 'system hookup' stamp: 'jm 3/3/98 11:48'!
onMIDIFileNamed: fileName
	"Return a ScorePlayerMorph on the score from the MIDI file of the given name."

	| score player |
	score := MIDIFileReader scoreFromFileNamed: fileName	.
	player := ScorePlayer onScore: score.
	^ self new onScorePlayer: player title: fileName
! !

!ScorePlayerMorph class methodsFor: 'system hookup' stamp: 'jm 10/12/1998 16:29'!
openOn: aScore title: aString

	| player |
	player := ScorePlayer onScore: aScore.
	(self new onScorePlayer: player title: aString) openInWorld.
! !

!ScorePlayerMorph class methodsFor: 'class initialization' stamp: 'nice 12/26/2009 21:11'!
playMidiFile: fullName
	"Play a MIDI file."
 
	Smalltalk at: #MIDIFileReader ifPresent: [:midiReader |
			| f score |
			f := (FileStream oldFileNamed: fullName) binary.
			score := (midiReader new readMIDIFrom: f) asScore.
			f close.
			self openOn: score title: (FileDirectory localNameFor: fullName)]
! !

!ScorePlayerMorph class methodsFor: 'class initialization' stamp: 'sw 2/17/2002 02:45'!
servicePlayMidiFile
	"Answer a service for opening player on a midi file"

	^ SimpleServiceEntry 
		provider: self 
		label: 'open in midi player'
		selector: #playMidiFile:
		description: 'open the midi-player tool on this file'
		buttonLabel: 'open'! !

!ScorePlayerMorph class methodsFor: 'fileIn/Out' stamp: 'sd 2/1/2002 22:21'!
services

	^ Array with: self servicePlayMidiFile

	! !

!ScorePlayerMorph class methodsFor: 'initialize-release' stamp: 'ar 9/5/2010 09:25'!
unload

	FileServices unregisterFileReader: self ! !

!ScorePlayerMorph methodsFor: 'controls' stamp: 'gm 2/28/2003 00:28'!
atTrack: trackIndex from: aPopUpChoice selectInstrument: selection 
	| oldSnd name snd |
	oldSnd := scorePlayer instrumentForTrack: trackIndex.
	(selection beginsWith: 'edit ') 
		ifTrue: 
			[name := selection copyFrom: 6 to: selection size.
			aPopUpChoice contentsClipped: name.
			(oldSnd isKindOf: FMSound) | (oldSnd isKindOf: LoopedSampledSound) 
				ifTrue: [EnvelopeEditorMorph openOn: oldSnd title: name].
			(oldSnd isKindOf: SampledInstrument) 
				ifTrue: [EnvelopeEditorMorph openOn: oldSnd allNotes first title: name].
			^self].
	snd := nil.
	1 to: instrumentSelector size
		do: 
			[:i | 
			(trackIndex ~= i and: [selection = (instrumentSelector at: i) contents]) 
				ifTrue: [snd := scorePlayer instrumentForTrack: i]].	"use existing instrument prototype"
	snd ifNil: 
			[snd := (selection = 'clink' 
				ifTrue: 
					[(SampledSound samples: SampledSound coffeeCupClink
								samplingRate: 11025) ]
				ifFalse: [(AbstractSound soundNamed: selection)]) copy].
	scorePlayer instrumentForTrack: trackIndex put: snd.
	(instrumentSelector at: trackIndex) contentsClipped: selection! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'jm 10/12/1998 17:14'!
closeMIDIPort

	scorePlayer closeMIDIPort.
	LastMIDIPort := nil.
! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:39'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 2! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:30'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color veryLightGray! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 21:25'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	self listDirection: #topToBottom;
		 wrapCentering: #center;
		 cellPositioning: #topCenter;
		 hResizing: #shrinkWrap;
		 vResizing: #shrinkWrap;
		 layoutInset: 3;
		 onScorePlayer: ScorePlayer new initialize title: ' ';
		 extent: 20 @ 20 ! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'nice 12/26/2009 21:11'!
instrumentChoicesForTrack: trackIndex
	| names |
	names := AbstractSound soundNames asOrderedCollection.
	names := names collect: [:n |
		| inst |
		inst := AbstractSound soundNamed: n.
		(inst isKindOf: UnloadedSound)
			ifTrue: [n, '(out)']
			ifFalse: [n]].
	names add: 'clink'.
	names add: 'edit ', (instrumentSelector at: trackIndex) contents.
	^ names asArray
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'dgd 9/21/2003 14:58'!
invokeMenu
	"Invoke a menu of additonal functions for this ScorePlayer."

	| aMenu |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu add: 'open a MIDI file' translated action: #openMIDIFile.
	aMenu addList: {
		#-.
		{'save as AIFF file' translated.	#saveAsAIFF}.
		{'save as WAV file' translated.		#saveAsWAV}.
		{'save as Sun AU file' translated.	#saveAsSunAudio}.
		#-}.
	aMenu add: 'reload instruments' translated target: AbstractSound selector: #updateScorePlayers.
	aMenu addLine.
	scorePlayer midiPort
		ifNil: [
			aMenu add: 'play via MIDI' translated action: #openMIDIPort]
		ifNotNil: [
			aMenu add: 'play via built in synth' translated action: #closeMIDIPort.
			aMenu add: 'new MIDI controller' translated action: #makeMIDIController:].
	aMenu addLine.
	aMenu add: 'make a pause marker' translated action: #makeAPauseEvent:.

	aMenu popUpInWorld: self world.
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'dgd 9/21/2003 17:20'!
makeAPauseEvent: evt

	| newWidget |

	newWidget := AlignmentMorph newRow.
	newWidget 
		color: Color orange; 
		borderWidth: 0; 
		layoutInset: 0;
		hResizing: #shrinkWrap; 
		vResizing: #shrinkWrap; 
		extent: 5@5;
		addMorph: (StringMorph contents: '[pause]' translated) lock;
		addMouseUpActionWith: (
			MessageSend receiver: self selector: #showResumeButtonInTheWorld
		).

	evt hand attachMorph: newWidget.! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'dgd 9/19/2003 12:50'!
makeControls

	| bb r reverbSwitch repeatSwitch |
	r := AlignmentMorph newRow.
	r color: color; borderWidth: 0; layoutInset: 0.
	r hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	bb := SimpleButtonMorph new target: self; borderColor: #raised;
			borderWidth: 2; color: color.
	r addMorphBack: (bb label: '<>'; actWhen: #buttonDown;
												actionSelector: #invokeMenu).
	bb := SimpleButtonMorph new target: self; borderColor: #raised;
			borderWidth: 2; color: color.
	r addMorphBack: (bb label: 'Piano Roll' translated;		actionSelector: #makePianoRoll).
	bb := SimpleButtonMorph new target: self; borderColor: #raised;
			borderWidth: 2; color: color.
	r addMorphBack: (bb label: 'Rewind' translated;		actionSelector: #rewind).
	bb := SimpleButtonMorph new target: scorePlayer; borderColor: #raised;
			borderWidth: 2; color: color.
	r addMorphBack: (bb label: 'Play' translated;			actionSelector: #resumePlaying).
	bb := SimpleButtonMorph new target: scorePlayer; borderColor: #raised;
			borderWidth: 2; color: color.
	r addMorphBack: (bb label: 'Pause' translated;			actionSelector: #pause).
	reverbSwitch := SimpleSwitchMorph new
		offColor: color;
		onColor: (Color r: 1.0 g: 0.6 b: 0.6);
		borderWidth: 2;
		label: 'Reverb Disable' translated;
		actionSelector: #disableReverb:;
		target: scorePlayer;
		setSwitchState: SoundPlayer isReverbOn not.
	r addMorphBack: reverbSwitch.
	scorePlayer ifNotNil:
		[repeatSwitch := SimpleSwitchMorph new
			offColor: color;
			onColor: (Color r: 1.0 g: 0.6 b: 0.6);
			borderWidth: 2;
			label: 'Repeat' translated;
			actionSelector: #repeat:;
			target: scorePlayer;
			setSwitchState: scorePlayer repeat.
		r addMorphBack: repeatSwitch].
	^ r
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'jm 9/28/1998 23:05'!
makeMIDIController: evt

	self world activeHand attachMorph:
		(MIDIControllerMorph new midiPort: scorePlayer midiPort).
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'panda 4/25/2000 15:42'!
makePianoRoll
	"Create a piano roll viewer for this score player."

	| pianoRoll hand |
	pianoRoll := PianoRollScoreMorph new on: scorePlayer.
	hand := self world activeHand.
	hand ifNil: [self world addMorph: pianoRoll]
		ifNotNil: [hand attachMorph: pianoRoll.
				hand lastEvent shiftPressed ifTrue:
					["Special case for NOBM demo"
					pianoRoll contractTime; contractTime; enableDragNDrop]].
	pianoRoll startStepping.
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'ar 11/9/2000 21:23'!
makeRow

	^ AlignmentMorph newRow
		color: color;
		layoutInset: 0;
		wrapCentering: #center; cellPositioning: #leftCenter;
		hResizing: #spaceFill;
		vResizing: #shrinkWrap
! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'tk 2/19/2001 18:46'!
onScorePlayer: aScorePlayer title: scoreName
	| divider col r |
	scorePlayer := aScorePlayer.
	scorePlayer ifNotNil:
		[scorePlayer  reset.
		instrumentSelector := Array new: scorePlayer score tracks size].

	self removeAllMorphs.
	self addMorphBack: self makeControls.
	scorePlayer ifNil: [^ self].

	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap.
	r addMorphBack: self rateControl;
		addMorphBack: (Morph newBounds: (0@0 extent: 20@0) color: Color transparent);
		addMorphBack: self volumeControl.
	self addMorphBack: r.
	self addMorphBack: self scrollControl.

	col := AlignmentMorph newColumn color: color; layoutInset: 0.
	self addMorphBack: col.
	1 to: scorePlayer trackCount do: [:trackIndex |
		divider := AlignmentMorph new
			extent: 10@1;
			borderWidth: 1;
			layoutInset: 0;
			borderColor: #raised;
			color: color;
			hResizing: #spaceFill;
			vResizing: #rigid.
		col addMorphBack: divider.
		col addMorphBack: (self trackControlsFor: trackIndex)].

	LastMIDIPort ifNotNil: [
		"use the most recently set MIDI port"
		scorePlayer openMIDIPort: LastMIDIPort].
! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'dgd 9/19/2003 12:51'!
openMIDIFile
	"Open a MIDI score and re-init controls..."
	| score fileName f player |
	fileName := Utilities chooseFileWithSuffixFromList: #('.mid' '.midi')
					withCaption: 'Choose a MIDI file to open' translated.
	(fileName isNil or: [ fileName == #none ])
		ifTrue: [^ self inform: 'No .mid/.midi files found in the Squeak directory' translated].
	f := FileStream readOnlyFileNamed: fileName.
	score := (MIDIFileReader new readMIDIFrom: f binary) asScore.
	f close.
	player := ScorePlayer onScore: score.
	self onScorePlayer: player title: fileName! !

!ScorePlayerMorph methodsFor: 'initialization' stamp: 'jm 10/12/1998 18:00'!
openMIDIPort

	| portNum |
	portNum := SimpleMIDIPort outputPortNumFromUser.
	portNum ifNil: [^ self].
	scorePlayer openMIDIPort: portNum.
	LastMIDIPort := portNum.
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'dgd 9/19/2003 12:53'!
panAndVolControlsFor: trackIndex

	| volSlider panSlider c r middleLine pianoRollColor |
	pianoRollColor := (Color wheel: scorePlayer score tracks size) at: trackIndex.
	volSlider := SimpleSliderMorph new
		color: color;
		sliderColor: pianoRollColor;
		extent: 101@2;
		target: scorePlayer;
		arguments: (Array with: trackIndex);
		actionSelector: #volumeForTrack:put:;
		minVal: 0.0;
		maxVal: 1.0;
		adjustToValue: (scorePlayer volumeForTrack: trackIndex).
	panSlider := SimpleSliderMorph new
		color: color;
		sliderColor: pianoRollColor;
		extent: 101@2;
		target: scorePlayer;
		arguments: (Array with: trackIndex);
		actionSelector: #panForTrack:put:;
		minVal: 0.0;
		maxVal: 1.0;		
		adjustToValue: (scorePlayer panForTrack: trackIndex).
	c := AlignmentMorph newColumn
		color: color;
		layoutInset: 0;
		wrapCentering: #center; cellPositioning: #topCenter;
		hResizing: #spaceFill;
		vResizing: #shrinkWrap.
	middleLine := Morph new  "center indicator for pan slider"
		color: (Color r: 0.4 g: 0.4 b: 0.4);
		extent: 1@(panSlider height - 4);
		position: panSlider center x@(panSlider top + 2).
	panSlider addMorphBack: middleLine.
	r := self makeRow.
	r addMorphBack: (StringMorph contents: '0').
	r addMorphBack: volSlider.
	r addMorphBack: (StringMorph contents: '10').
	c addMorphBack: r.
	r := self makeRow.
	r addMorphBack: (StringMorph contents: 'L' translated).
	r addMorphBack: panSlider.
	r addMorphBack: (StringMorph contents: 'R' translated).
	c addMorphBack: r.
	^ c
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'dgd 9/19/2003 12:53'!
rateControl

	| rateSlider middleLine r |
	rateSlider := SimpleSliderMorph new
		color: color;
		sliderColor: Color gray;
		extent: 180@2;
		target: self;
		actionSelector: #setLogRate:;
		minVal: -1.0;
		maxVal: 1.0;
		adjustToValue: 0.0.
	middleLine := Morph new  "center indicator for pan slider"
		color: (Color r: 0.4 g: 0.4 b: 0.4);
		extent: 1@(rateSlider height - 4);
		position: rateSlider center x@(rateSlider top + 2).
	rateSlider addMorphBack: middleLine.
	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #rigid;
		height: 24.
	r addMorphBack: (StringMorph contents: 'slow ' translated).
	r addMorphBack: rateSlider.
	r addMorphBack: (StringMorph contents: ' fast' translated).
	^ r
! !

!ScorePlayerMorph methodsFor: 'controls' stamp: 'jm 10/12/1998 17:18'!
rewind

	scorePlayer pause; reset.
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:43'!
saveAsAIFF
	"Create a stereo AIFF audio file with the result of performing my score."

	| fileName |
	fileName := UIManager default request: 'New file name?' translated.
	fileName isEmpty ifTrue: [^ self].
	(fileName asLowercase endsWith: '.aif') ifFalse: [
		fileName := fileName, '.aif'].

	scorePlayer storeAIFFOnFileNamed: fileName.
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:44'!
saveAsSunAudio
	"Create a stereo Sun audio file with the result of performing my score."

	| fileName |
	fileName := UIManager default request: 'New file name?' translated.
	fileName isEmpty ifTrue: [^ self].
	(fileName asLowercase endsWith: '.au') ifFalse: [
		fileName := fileName, '.au'].

	scorePlayer storeSunAudioOnFileNamed: fileName.
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:44'!
saveAsWAV
	"Create a stereo WAV audio file with the result of performing my score."

	| fileName |
	fileName := UIManager default request: 'New file name?' translated.
	fileName isEmpty ifTrue: [^ self].
	(fileName asLowercase endsWith: '.wav') ifFalse: [
		fileName := fileName, '.wav'].

	scorePlayer storeWAVOnFileNamed: fileName.
! !

!ScorePlayerMorph methodsFor: 'accessing' stamp: 'jm 6/1/1998 09:10'!
scorePlayer

	^ scorePlayer
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'dgd 9/19/2003 12:53'!
scrollControl

	| r |
	scrollSlider := SimpleSliderMorph new
		color: color;
		sliderColor: Color gray;
		extent: 360@2;
		target: scorePlayer;
		actionSelector: #positionInScore:;
		adjustToValue: scorePlayer positionInScore.
	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #rigid;
		height: 24.
	r addMorphBack: (StringMorph contents: 'start ' translated).
	r addMorphBack: scrollSlider.
	r addMorphBack: (StringMorph contents: ' end' translated).
	^ r
! !

!ScorePlayerMorph methodsFor: 'controls' stamp: 'jm 1/30/98 15:23'!
setLogRate: logOfRate

	scorePlayer rate: (3.5 raisedTo: logOfRate).
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'nice 12/26/2009 21:11'!
showResumeButtonInTheWorld
	WorldState addDeferredUIMessage: [
		| w |
		w := self world.
		w ifNotNil: [
			w addMorphFront:
				(self standaloneResumeButton position: (w right - 100) @ (w top + 10)).
			scorePlayer pause.
			].
	]
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'dgd 9/19/2003 12:53'!
standaloneResumeButton

	| r |

	r := AlignmentMorph newRow.
	r color: Color red; borderWidth: 0; layoutInset: 6; useRoundedCorners.
	r hResizing: #shrinkWrap; vResizing: #shrinkWrap; extent: 5@5.
	r addMorphBack: (
		SimpleButtonMorph new
			target: [
				scorePlayer resumePlaying.
				r delete
			];
			borderColor: #raised;
			borderWidth: 2;
			color: Color green;
			label: 'Continue' translated;
			actionSelector: #value
	).
	r setBalloonText: 'Continue playing a paused presentation' translated.
	^r


! !

!ScorePlayerMorph methodsFor: 'stepping and presenter' stamp: 'di 6/15/1999 11:48'!
step

	scrollSlider adjustToValue: scorePlayer positionInScore.

! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'jm 1/30/98 14:52'!
trackControlsFor: trackIndex

	| r |
	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap.
	r addMorphBack: (self trackNumAndMuteButtonFor: trackIndex).
	r addMorphBack: (Morph new extent: 10@5; color: color).  "spacer"
	r addMorphBack: (self panAndVolControlsFor: trackIndex).
	^ r
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'dgd 9/19/2003 12:54'!
trackNumAndMuteButtonFor: trackIndex

	| muteButton instSelector pianoRollColor r |
	muteButton := SimpleSwitchMorph new
		onColor: (Color r: 1.0 g: 0.6 b: 0.6);
		offColor: color;
		color: color;
		label: 'Mute' translated;
		target: scorePlayer;
		actionSelector: #mutedForTrack:put:;
		arguments: (Array with: trackIndex).
	instSelector := PopUpChoiceMorph new
		extent: 95@14;
		contentsClipped: 'oboe1';
		target: self;
		actionSelector: #atTrack:from:selectInstrument:;
		getItemsSelector: #instrumentChoicesForTrack:;
		getItemsArgs: (Array with: trackIndex).
	instSelector arguments:
		(Array with: trackIndex with: instSelector).
	instrumentSelector at: trackIndex put: instSelector.

	"select track color using same color list as PianoRollScoreMorph"
	pianoRollColor := (Color wheel: scorePlayer score tracks size) at: trackIndex.

	r := self makeRow
		hResizing: #rigid;
		vResizing: #spaceFill;
		extent: 70@10.
	r addMorphBack:
		((StringMorph
			contents: trackIndex printString
			font: (TextStyle default fontOfSize: 24)) color: pianoRollColor).
	trackIndex < 10
		ifTrue: [r addMorphBack: (Morph new color: color; extent: 19@8)]  "spacer"
		ifFalse: [r addMorphBack: (Morph new color: color; extent: 8@8)].  "spacer"
	r addMorphBack:
		(StringMorph new
			extent: 140@14;
			contentsClipped: (scorePlayer infoForTrack: trackIndex)).
	r addMorphBack: (Morph new color: color; extent: 8@8).  "spacer"
	r addMorphBack: instSelector.
	r addMorphBack: (AlignmentMorph newRow color: color).  "spacer"
	r addMorphBack: muteButton.
	^ r
! !

!ScorePlayerMorph methodsFor: 'menu' stamp: 'di 11/7/2000 12:42'!
updateInstrumentsFromLibraryExcept: soundsBeingEdited
	"The instrument library has been modified. Update my instruments with the new versions from the library. Use a single instrument prototype for all parts with the same name; this allows the envelope editor to edit all the parts by changing a single sound prototype."

	"soundsBeingEdited is a collection of sounds being edited (by an EnvelopeEditor).  If any of my instruments share one of these, then they will be left alone so as not to disturb that dynamic linkage."

	| unloadPostfix myInstruments name displaysAsUnloaded isUnloaded |
	unloadPostfix := '(out)'.
	myInstruments := Dictionary new.
	1 to: instrumentSelector size do: [:i |
		name := (instrumentSelector at: i) contents.
		displaysAsUnloaded := name endsWith: unloadPostfix.
		displaysAsUnloaded ifTrue: [
			name := name copyFrom: 1 to: name size - unloadPostfix size].
		(myInstruments includesKey: name) ifFalse: [
			myInstruments at: name put:
				(name = 'clink'
					ifTrue: [
						(SampledSound
							samples: SampledSound coffeeCupClink
							samplingRate: 11025) copy]
					ifFalse: [
						(AbstractSound
							soundNamed: name
							ifAbsent: [
								(instrumentSelector at: i) contentsClipped: 'default'.
								FMSound default]) copy])].
		(soundsBeingEdited includes: (scorePlayer instrumentForTrack: i)) ifFalse:
			["Do not update any instrument that is currently being edited"
			scorePlayer instrumentForTrack: i put: (myInstruments at: name)].

		"update loaded/unloaded status in instrumentSelector if necessary"
		isUnloaded := (myInstruments at: name) isKindOf: UnloadedSound.
		(displaysAsUnloaded and: [isUnloaded not])
			ifTrue: [(instrumentSelector at: i) contentsClipped: name].
		(displaysAsUnloaded not and: [isUnloaded])
			ifTrue: [(instrumentSelector at: i) contentsClipped: name, unloadPostfix]].
! !

!ScorePlayerMorph methodsFor: 'layout' stamp: 'dgd 9/19/2003 12:54'!
volumeControl

	| volumeSlider r |
	volumeSlider := SimpleSliderMorph new
		color: color;
		sliderColor: Color gray;
		extent: 80@2;
		target: scorePlayer;
		actionSelector: #overallVolume:;
		adjustToValue: scorePlayer overallVolume.
	r := self makeRow
		hResizing: #shrinkWrap;
		vResizing: #rigid;
		height: 24.
	r addMorphBack: (StringMorph contents: 'soft  ' translated).
	r addMorphBack: volumeSlider.
	r addMorphBack: (StringMorph contents: ' loud' translated).
	^ r
! !

!ScorePlayerMorph methodsFor: 'rounding' stamp: 'mt 7/28/2016 10:39'!
wantsRoundedCorners
	^ SystemWindow roundedWindowCorners or: [super wantsRoundedCorners]! !

RectangleMorph subclass: #ClickExerciser
	instanceVariableNames: 'buttons'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Demo'!

!ClickExerciser class methodsFor: 'parts bin' stamp: 'mt 3/30/2015 11:17'!
descriptionForPartsBin
	"ClickExerciser descriptionForPartsBin"
	^ self partName:	'Exercise Click'
		categories:		#('Demo')
		documentation:	'An exerciser for click, double-click, and drag-click in morphic'! !

!ClickExerciser methodsFor: 'accessing' stamp: 'wiz 8/23/2005 19:31'!
allSelectors
^ #( #click: #doubleClick: #doubleClickTimeout: #startDrag:) ! !

!ClickExerciser methodsFor: 'accessing' stamp: 'wiz 8/23/2005 19:31'!
alternateBorderColor
"answer the alternate color/fill style for the receiver"
	^ Color yellow! !

!ClickExerciser methodsFor: 'accessing' stamp: 'wiz 8/25/2005 22:38'!
alternateColor
"answer the alternate color/fill style for the receiver"
	^ Color cyan! !

!ClickExerciser methodsFor: 'initialization' stamp: 'wiz 8/23/2005 19:31'!
alternateSide
"initial extent is square. We return an alternate height/width for drag to manipulate."
^ (self defaultSide asFloat * 1.618) rounded .
! !

!ClickExerciser methodsFor: 'accessing' stamp: 'wiz 8/23/2005 19:31'!
balloonText
	^ 'Double-click on me to change my color; 
single-click on me to change border color;
hold mouse down within me and then move it to grow 
 or shrink. When I time out my border changes width.
Choose which of the above will work
by selecting the boxes in the middle.
See the boxes balloons.' translated! !

!ClickExerciser methodsFor: 'event handling' stamp: 'wiz 8/23/2005 19:31'!
click: evt
	self showBalloon: 'click' hand: evt hand.
	self borderColor: (self borderColor = self defaultBorderColor ifTrue: [self alternateBorderColor] ifFalse: [self defaultBorderColor])
! !

!ClickExerciser methodsFor: 'accessing' stamp: 'wiz 8/23/2005 19:31'!
defaultBorderColor
"answer the alternate color/fill style for the receiver"
	^ Color black! !

!ClickExerciser methodsFor: 'accessing' stamp: 'wiz 8/23/2005 19:31'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color red! !

!ClickExerciser methodsFor: 'initialization' stamp: 'wiz 8/23/2005 19:31'!
defaultSide
"initial extent is square. We return the default height/width."
^ 100 .
! !

!ClickExerciser methodsFor: 'event handling' stamp: 'wiz 8/23/2005 19:31'!
doubleClickTimeout: evt 
	self showBalloon: 'ClickTimeout' hand: evt hand.
	self borderWidth: self borderWidth \\ 11 + 2! !

!ClickExerciser methodsFor: 'event handling' stamp: 'wiz 8/23/2005 19:31'!
doubleClick: evt
	self showBalloon: 'doubleClick' hand: evt hand.
	self color: ((color = self alternateColor ) ifTrue: [self defaultColor] ifFalse: [self alternateColor])
! !

!ClickExerciser methodsFor: 'event handling' stamp: 'wiz 8/23/2005 19:31'!
handlesMouseDown: evt
	^ true! !

!ClickExerciser methodsFor: 'initialization' stamp: 'mt 3/30/2015 09:16'!
initButtons
	| aButton positions |
	aButton := ThreePhaseButtonMorph checkBox. 
	positions := ((0@0) rect: aButton extent negated) corners + self center .
	buttons := positions collect: [ :p | ThreePhaseButtonMorph checkBox position: p; state: #on ] .
	buttons with: self allSelectors do: [ :b :s | b balloonText:  s asString ] .
	self removeAllMorphs .
	self addAllMorphs:  buttons .
! !

!ClickExerciser methodsFor: 'initialization' stamp: 'wiz 8/23/2005 19:31'!
initialize
super initialize.
self extent: self defaultSide asPoint.
self initButtons .! !

!ClickExerciser methodsFor: 'event handling' stamp: 'mtf 2/20/2011 20:41'!
mouseDown: evt 
	"Do nothing upon mouse-down except inform the hand to watch for a  
	double-click; wait until an ensuing click:, doubleClick:, or drag:  
	message gets dispatched"
	Preferences disable: #NewClickTest .
	evt hand
		waitForClicksOrDrag: self
		event: evt
		selectors: self selectors
		threshold: HandMorph dragThreshold! !

!ClickExerciser methodsFor: 'accessing' stamp: 'wiz 8/23/2005 19:31'!
selectors
^  self allSelectors with: buttons collect: [ :s :b | b isOn ifTrue: [ s ] ifFalse: [nil ] ] .! !

!ClickExerciser methodsFor: 'event handling' stamp: 'wiz 8/23/2005 19:31'!
startDrag: evt
	"We'll get a mouseDown first, some mouseMoves, and a mouseUp event last"
	| height width both  |
	"evt isMouseDown ifTrue:
		[self showBalloon: 'drag (mouse down)' hand: evt hand.
		self world displayWorld.
		(Delay forMilliseconds: 750) wait].
	evt isMouseUp ifTrue:
		[self showBalloon: 'drag (mouse up)' hand: evt hand].
	(evt isMouseUp or: [evt isMouseDown]) ifFalse:
		[self showBalloon: 'drag (mouse still down)' hand: evt hand].
	(self containsPoint: evt cursorPoint)
		ifFalse: [^ self]."
	self showBalloon: 'drag (mouse down)' hand: evt hand.

	width  := (self defaultSide max: self extent x) min: self alternateSide.
	height := (self defaultSide max: self extent y) min: self alternateSide.
	both  := self defaultSide + self alternateSide. 

self extent:	
 (((color = (self defaultColor)
		ifTrue:
			[ (both - height) @ width  ]
		ifFalse:
			[ height @ (both - width) ]) 
	max: self defaultSide asPoint) 
	min: self alternateSide asPoint )
	
	! !

RectangleMorph subclass: #GradientDisplayMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!GradientDisplayMorph methodsFor: 'as yet unclassified' stamp: 'kfr 3/20/2015 11:29'!
colorRamp
	^self fillStyle colorRamp! !

!GradientDisplayMorph methodsFor: 'as yet unclassified' stamp: 'kfr 3/18/2015 18:50'!
colorRamp: aColorRamp
	self fillStyle colorRamp: aColorRamp! !

!GradientDisplayMorph methodsFor: 'as yet unclassified' stamp: 'kfr 10/27/2015 16:59'!
drawOn: aCanvas 
	"Draw a hatch pattern first."
	aCanvas
		fillRectangle: self innerBounds
		fillStyle: (InfiniteForm with: ColorPresenterMorph hatchForm).
	super drawOn: aCanvas! !

!GradientDisplayMorph methodsFor: 'as yet unclassified' stamp: 'kfr 3/20/2015 16:00'!
initialize
	| fill colorRamp |
	super initialize.
	"self hResizing: #spaceFill. "
	colorRamp := {0.0 -> Color green. 0.3 -> Color red. 0.7 -> Color black. 1.0 -> Color blue}.
	fill := GradientFillStyle ramp: colorRamp.
	fill origin: 0@0.
	fill direction: self bounds extent x @ 0.
	fill radial: false.
	self fillStyle: fill! !

RectangleMorph subclass: #GradientEditor
	instanceVariableNames: 'gradientDisplay rampMorphs selectedSketch row text target selector morph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!GradientEditor class methodsFor: 'as yet unclassified' stamp: 'kfr 10/21/2015 00:20'!
on: aTarget selector: aSelector forMorph: aMorph colorRamp: aColorRamp
	^self new 
		setTarget: aTarget
		selector: aSelector
		forMorph: aMorph
		colorRamp: aColorRamp.
		
! !

!GradientEditor methodsFor: 'initialization' stamp: 'kfr 10/29/2015 12:48'!
addButtonRow
	| button button2 buttonRow button3 |
	buttonRow := RectangleMorph new borderWidth: 0; 
		color: Color transparent;
		layoutPolicy: TableLayout new;
		 hResizing: #spaceFil;
		 vResizing: #spaceFill;
		 cellPositioning: #center;
		 listCentering: #topLeft;
		 listDirection: #LeftToRight;
		 reverseTableCells: true;
		 cellInset: 4.

	button := PluggableButtonMorph on: self
				getState: nil
				action: #addHandle
				label: #addColorButtonLabel.
	button hResizing: #spaceFill;
			vResizing: #spaceFill.
	buttonRow  addMorph: button.

	button2 := PluggableButtonMorph on: self
				getState: nil
				action: #deleteHandle
				label: #removeColorButtonLabel.
	button2 hResizing: #spaceFill;
			vResizing: #spaceFill.
	buttonRow addMorph: button2.

	button3 := PluggableButtonMorph on: self
				getState: nil
				action: #delete
				label: #closeButtonLabel.
	button3 hResizing: #spaceFill;
			vResizing: #spaceFill.
	buttonRow addMorph: button3.
	
	self addMorph: buttonRow! !

!GradientEditor methodsFor: 'initialization' stamp: 'kfr 10/28/2015 13:36'!
addColorButtonLabel
	^ 'Add color' translated! !

!GradientEditor methodsFor: 'change reporting' stamp: 'kfr 10/27/2015 22:46'!
addHandle
	self addHandleForColor: Color random position: 0.5.
	self updateColorRamp! !

!GradientEditor methodsFor: 'change reporting' stamp: 'kfr 10/27/2015 22:49'!
addHandleForColor: aColor position: aPosition
	| handleInstance colorIcon delta |
	handleInstance := self handle.
	colorIcon := SketchMorph
						withForm: ((aColor iconOrThumbnailOfSize: 20) borderWidth: 1 color: Color black)..
	self eventHandler: colorIcon target: colorIcon.
	self eventHandler: handleInstance target: self.
	row addMorph: handleInstance.
	delta := gradientDisplay left - 10 + (gradientDisplay width * aPosition).
	handleInstance position:  delta @ (gradientDisplay top - 18).
	colorIcon position: delta @ (gradientDisplay bottom + 5).
	handleInstance addMorph: colorIcon.
	rampMorphs addLast: handleInstance.
	
	! !

!GradientEditor methodsFor: 'change reporting' stamp: 'kfr 11/1/2015 21:06'!
addHandles
	| handle colorRamp |
	rampMorphs := OrderedCollection new.
	colorRamp := gradientDisplay colorRamp asOrderedCollection.
	handle := self handle.
	colorRamp
		do: [:i | 
			self addHandleForColor: i value position: i key
			].
	
	self changed.! !

!GradientEditor methodsFor: 'change reporting' stamp: 'kfr 3/23/2015 12:22'!
changeColor: aSketchMorph event: evt target: aMorph
	| newColor |
	newColor := aSketchMorph rotatedForm colorAt: aSketchMorph rotatedForm center.
	selectedSketch := aSketchMorph.
 	self changeColorTarget: self selector: #updateColor: originalColor: newColor value hand: evt hand.
	! !

!GradientEditor methodsFor: 'initialization' stamp: 'kfr 10/28/2015 13:35'!
closeButtonLabel
	^ 'Close' translated! !

!GradientEditor methodsFor: 'accessing' stamp: 'kfr 11/1/2015 21:05'!
colorRamp
	^gradientDisplay fillStyle colorRamp asArray printString.
	! !

!GradientEditor methodsFor: 'initialization' stamp: 'kfr 11/1/2015 21:16'!
colorRampExpressionMorph
	| inputField builder |
	builder := ToolBuilder default.
	inputField := (builder build: (builder pluggableInputFieldSpec new
		model: self;
		getText: #colorRamp;
		setText: #colorRampExpression:;
		softLineWrap: true)).
	inputField
		 hResizing: #spaceFill ;
		 vResizing: #spaceFill ;
		 
		 height: (Preferences standardDefaultTextFont height * 3/2).
	^ inputField! !

!GradientEditor methodsFor: 'accessing' stamp: 'kfr 11/1/2015 21:10'!
colorRampExpression: aString 
	"Set my color by evaluating aString, a Smalltalk expression which results in a Color instance."
	| col |
	
	{aString.}
		detect:
			[ : each | ([ col := Compiler evaluate: each ]
				on: Error
				do:
					[ : err | nil ]) notNil ]
		ifNone: [ nil ]. 
	col ifNotNil: [ self colorRamp: col.
	rampMorphs do:[ :each| each delete].
	self addHandles; updateColorRamp. ]! !

!GradientEditor methodsFor: 'accessing' stamp: 'kfr 10/31/2015 17:52'!
colorRamp: aColorRamp
	gradientDisplay colorRamp: aColorRamp.
	self changed: #colorRamp.
	self changed
	
! !

!GradientEditor methodsFor: 'change reporting' stamp: 'kfr 10/28/2015 12:14'!
deleteHandle
	rampMorphs do:[ :i | (i color == Color black) ifTrue:[ rampMorphs remove: i. row removeMorph: i]].
	self updateColorRamp
	
! !

!GradientEditor methodsFor: 'event handling' stamp: 'kfr 11/1/2015 21:07'!
eventHandler: anInstance target: aTarget 
	(anInstance isKindOf: SketchMorph)
		ifTrue: [anInstance
				on: #mouseUp
				send: #changeColor:event:target:
				to: self
				withValue: aTarget]
		ifFalse: [anInstance
				on: #mouseDown
				send: #limitHandleMove:event:from:
				to: self
				withValue: aTarget.
			anInstance
				on: #mouseMove
				send: #limitHandleMove:event:from:
				to: self
				withValue: aTarget]! !

!GradientEditor methodsFor: 'initialization' stamp: 'kfr 11/1/2015 21:07'!
handle
	| handle |
	handle := PolygonMorph
				vertices: (Array
						with: 0 @ 0
						with: 16 @ 0
						with: 8 @ 16)
				color: Color white darker
				borderWidth: 1
				borderColor: Color black.
	^ handle
		addMorph: ((RectangleMorph
				newBounds: (8 @ 18 extent: 1 @ (gradientDisplay height - 2))
				color: Color orange)
				borderWidth: 0)! !

!GradientEditor methodsFor: 'initialization' stamp: 'kfr 11/1/2015 21:07'!
initialize
	super initialize.
	self myLayout.
	self extent: 500 @ 200.
	row := RectangleMorph new extent: self width @ 100;
				 color: Color transparent;
				 borderColor: #inset.
	row addMorph: (gradientDisplay := GradientDisplayMorph new position: 20 @ 20;
					 extent: self width - 40 @ 40).
	gradientDisplay fillStyle direction: gradientDisplay width @ 0.
	self addMorph: row.
	self addButtonRow.
	self addMorph: self colorRampExpressionMorph! !

!GradientEditor methodsFor: 'change reporting' stamp: 'kfr 10/28/2015 12:24'!
limitHandleMove: association event: evt from: handle
	| p  newBounds |
	rampMorphs do:[ : i | i color: (Color white darker)].
	newBounds := gradientDisplay bounds.
	newBounds := (newBounds left: (newBounds left - 10)).
	newBounds := (newBounds right: (newBounds right - 10)). 
	p := (evt cursorPoint - (10@0)) adhereTo: newBounds.
	handle position: (p x )@ (handle position y).
	handle color: Color black.
	self updateColorRamp! !

!GradientEditor methodsFor: 'accessing' stamp: 'kfr 10/21/2015 00:18'!
morph: aMorph
	^morph := aMorph! !

!GradientEditor methodsFor: 'initialization' stamp: 'kfr 10/29/2015 12:48'!
myLayout
	self color: (Color white darker) ;
		 cornerStyle: #rounded ;
		 borderWidth: 0;
		 layoutPolicy: TableLayout new;
		 hResizing: #shrinkWrap;
		 vResizing: #shrinkWrap;
		 cellPositioning: #center;
		 listCentering: #topLeft;
		 layoutInset: 4;
		 listDirection: #topToBottom;
		 reverseTableCells: true;
		 wrapCentering: #topLeft;
		 cellInset: 4! !

!GradientEditor methodsFor: 'initialization' stamp: 'kfr 10/28/2015 13:36'!
removeColorButtonLabel
	^ 'Remove color' translated! !

!GradientEditor methodsFor: 'accessing' stamp: 'kfr 10/19/2015 21:44'!
selector: aSelector	
	^selector := aSelector! !

!GradientEditor methodsFor: 'initialization' stamp: 'kfr 10/31/2015 17:48'!
setTarget: aTarget selector: aSelector forMorph:aMorph colorRamp: aColorRamp

	self target: aTarget; selector: aSelector; morph: aMorph; colorRamp: aColorRamp; addHandles
	
		! !

!GradientEditor methodsFor: 'accessing' stamp: 'kfr 10/19/2015 21:44'!
target: aTarget
	^target := aTarget! !

!GradientEditor methodsFor: 'change reporting' stamp: 'kfr 10/31/2015 11:40'!
updateColorRamp
	| newAssociation newKey newColor sketch colorRamp |
	
	self updateRampMorphs.
	colorRamp := OrderedCollection new.
	rampMorphs
		do: [:i | 
			newKey := ((i position x - gradientDisplay left / gradientDisplay width) asFloat roundUpTo: 0.01)
						min: 1.0
						max: 0.0.
			sketch := i findA: SketchMorph.
			newColor := sketch rotatedForm colorAt: sketch rotatedForm center.
			newAssociation := newKey -> newColor.
			colorRamp addLast: newAssociation].
	colorRamp := colorRamp sorted.
	gradientDisplay colorRamp: colorRamp.
	gradientDisplay fillStyle direction: gradientDisplay extent x @ 0.
	self changed.
	target ifNotNil:[
			target perform: selector
			with: colorRamp 
			with: morph].
	self changed: #colorRamp! !

!GradientEditor methodsFor: 'change reporting' stamp: 'kfr 3/23/2015 12:17'!
updateColor: aColor
	selectedSketch rotatedForm  floodFill: aColor at: selectedSketch rotatedForm center.
	self updateColorRamp
! !

!GradientEditor methodsFor: 'change reporting' stamp: 'kfr 3/27/2015 14:21'!
updateRampMorphs
	rampMorphs do:[ :i | i isInWorld ifFalse:[ rampMorphs remove: i]]! !

!RectangleMorph class methodsFor: 'parts bin' stamp: 'tak 7/25/2007 00:10'!
descriptionForPartsBin
	^ self partName:	'Rectangle' translatedNoop
		categories:		{'Graphics' translatedNoop. 'Basic' translatedNoop}
		documentation:	'A rectangular shape, with border and fill style' translatedNoop! !

!RectangleMorph class methodsFor: 'parts bin' stamp: 'tk 11/14/2001 20:09'!
diagonalPrototype

	| rr |
	rr := self authoringPrototype.
	rr useGradientFill; borderWidth: 0.
	rr fillStyle direction: rr extent.
	^ rr! !

!RectangleMorph class methodsFor: 'parts bin' stamp: 'tk 11/14/2001 20:09'!
gradientPrototype

	| rr |
	rr := self authoringPrototype.
	rr useGradientFill; borderWidth: 0.
	^ rr! !

!RectangleMorph class methodsFor: 'as yet unclassified' stamp: 'sw 9/14/2006 10:38'!
roundRectPrototype
	"Answer a prototypical RoundRect object for a parts bin."

	^ self authoringPrototype useRoundedCorners 
		color: (Color r: 1.0 g: 0.3 b: 0.6); 
		borderWidth: 1;
		setNameTo: 'RoundRect'! !

!RectangleMorph methodsFor: 'visual properties' stamp: 'ar 6/25/1999 11:13'!
canHaveFillStyles
	"Return true if the receiver can have general fill styles; not just colors.
	This method is for gradually converting old morphs."
	^self class == RectangleMorph "no subclasses"! !

!RectangleMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:30'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color
		r: 0.613
		g: 0.903
		b: 1.0! !

!RectangleMorph methodsFor: 'accessing' stamp: 'ar 6/23/2001 16:06'!
wantsToBeCachedByHand
	"Return true if the receiver wants to be cached by the hand when it is dragged around."
	self hasTranslucentColor ifTrue:[^false].
	self bounds = self fullBounds ifTrue:[^true].
	self submorphsDo:[:m|
		(self bounds containsRect: m fullBounds) ifFalse:[
			m wantsToBeCachedByHand ifFalse:[^false].
		].
	].
	^true! !

RectangleMorph subclass: #SimpleButtonMorph
	instanceVariableNames: 'target actionSelector arguments actWhen oldColor mouseDownTime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!SimpleButtonMorph commentStamp: 'ul 7/22/2010 04:30' prior: 0!
I am labeled, rectangular morph which allows the user to click me. I can be configured to send my "target" the message "actionSelector" with "arguments" when I am clicked. I may have a label, implemented as a StringMorph.

Example:

	SimpleButtonMorph new
		target: Beeper;
		label: 'Beep!!';
		actionSelector: #beep; 
		openInWorld

Structure:
instance var 	Type		Description 
target 			Object 		The Object to notify upon a click 
actionSelector 	Symbol 		The message to send to Target (#messageName) 
arguments 		Array 		Arguments to send with #actionSelection (optional) 
actWhen 		Symbol 		When to take action: may be #buttonUp (default), #buttonDown,
								#whilePressed, or #startDrag 
oldColor 		Color 		Used to restore color after click 

Another example: a button which quits the image without saving it.

	SimpleButtonMorph new
		target: Smalltalk;
		label: 'quit';
		actionSelector: #snapshot:andQuit:;
		arguments: (Array with: false with: true); 
		openInWorld

!

SimpleButtonMorph subclass: #IconicButton
	instanceVariableNames: 'darkenedForm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!IconicButton commentStamp: '<historical>' prior: 0!
A "Simple Button" in which the appearance is provided by a Form.!

!IconicButton methodsFor: 'menu' stamp: 'sw 9/28/1999 20:42'!
addLabelItemsTo: aCustomMenu hand: aHandMorph
	"don't do the inherited behavior, since there is no textual label in this case"! !

!IconicButton methodsFor: 'accessing' stamp: 'ar 12/12/2001 01:38'!
borderInset
	self borderStyle: (BorderStyle inset width: 2).! !

!IconicButton methodsFor: 'initialization' stamp: 'ar 12/12/2001 01:38'!
borderNormal
	self borderStyle: (BorderStyle width: 2 color: Color transparent).! !

!IconicButton methodsFor: 'accessing' stamp: 'ar 12/12/2001 01:41'!
borderRaised
	self borderStyle: (BorderStyle raised width: 2).! !

!IconicButton methodsFor: 'initialization' stamp: 'ar 12/17/2001 21:17'!
borderThick
	self borderStyle: (BorderStyle width: 2 color: self raisedColor twiceDarker).! !

!IconicButton methodsFor: 'initialization' stamp: 'ar 12/15/2001 14:43'!
buttonSetup
	self actWhen: #buttonUp.
	self cornerStyle: #rounded.
	self borderNormal.
	self on: #mouseEnter send: #borderRaised to: self.
	self on: #mouseLeave send: #borderNormal to: self.
	self on: #mouseLeaveDragging send: #borderNormal to: self.
	self on: #mouseDown send: #borderInset to: self.
	self on: #mouseUp send: #borderRaised to: self.! !

!IconicButton methodsFor: 'as yet unclassified' stamp: 'mha 3/26/2010 00:00'!
darken

	self firstSubmorph form: self darkenedForm! !

!IconicButton methodsFor: 'as yet unclassified' stamp: 'bf 3/26/2010 09:01'!
darkenedForm
	^ darkenedForm ifNil: [ darkenedForm := self firstSubmorph baseGraphic darker ]! !

!IconicButton methodsFor: 'button' stamp: 'tfel 3/18/2010 21:38'!
doButtonAction

	super doButtonAction.
	self restoreImage.! !

!IconicButton methodsFor: 'initialization' stamp: 'sw 11/29/1999 20:52'!
initialize
	super initialize.
	self useSquareCorners! !

!IconicButton methodsFor: 'initialization' stamp: 'nk 9/7/2004 11:43'!
initializeWithThumbnail: aThumbnail withLabel: aLabel andColor: aColor andSend: aSelector to: aReceiver 	
	"Initialize the receiver to show aThumbnail on its face, giving it the label supplied and arranging for it, when the button goes down on it, to obtain a new morph by sending the supplied selector to the supplied receiver"

	| labeledItem nonTranslucent |
	nonTranslucent := aColor asNontranslucentColor.
	labeledItem := AlignmentMorph newColumn.
	labeledItem color: nonTranslucent.
	labeledItem borderWidth: 0.
	labeledItem
		layoutInset: 4@0;
		cellPositioning: #center.
	labeledItem addMorph: aThumbnail.
	labeledItem addMorphBack: (Morph new extent: (4@4)) beTransparent.
	labeledItem addMorphBack: (TextMorph new
		backgroundColor: nonTranslucent;
		contentsAsIs: aLabel;
		beAllFont: Preferences standardEToysFont;
		centered).

	self
		beTransparent;
		labelGraphic: ((labeledItem imageForm: 32 backgroundColor: nonTranslucent forRectangle: labeledItem fullBounds) replaceColor: nonTranslucent withColor: Color transparent);
		borderWidth: 0;
		target: aReceiver;
		actionSelector: #launchPartVia:label:;
		arguments: {aSelector. aLabel};
		actWhen: #buttonDown.

	self stationarySetup.! !

!IconicButton methodsFor: 'initialization' stamp: 'nk 8/6/2004 11:34'!
initializeWithThumbnail: aThumbnail withLabel: aLabel andSend: aSelector to: aReceiver 	
	"Initialize the receiver to show aThumbnail on its face, giving it the label supplied and arranging for it, when the button goes down on it, to obtain a new morph by sending the supplied selector to the supplied receiver"

	^self initializeWithThumbnail: aThumbnail withLabel: aLabel andColor: Color transparent   andSend: aSelector to: aReceiver 	! !

!IconicButton methodsFor: 'as yet unclassified' stamp: 'sw 9/28/1999 14:11'!
labelFromString: aString
	"Make an iconic label from aString"

	self labelGraphic: (StringMorph contents: aString) imageForm
! !

!IconicButton methodsFor: 'as yet unclassified' stamp: 'tfel 3/18/2010 21:11'!
labelGraphic: aForm
	| oldLabel graphicalMorph |
	(oldLabel := self findA: SketchMorph)
		ifNotNil: [oldLabel delete].
	graphicalMorph := SketchMorph withForm: aForm.
	self extent: graphicalMorph extent + (borderWidth + 6).
	graphicalMorph position: self center - (graphicalMorph extent // 2).
	self addMorph: graphicalMorph.
	graphicalMorph 
		baseGraphic;
		lock.
! !

!IconicButton methodsFor: 'events' stamp: 'jrp 7/5/2005 23:15'!
mouseEnter: evt

	self borderStyle: BorderStyle thinGray! !

!IconicButton methodsFor: 'events' stamp: 'jrp 7/5/2005 22:50'!
mouseLeave: evt

	self borderNormal! !

!IconicButton methodsFor: 'as yet unclassified' stamp: 'tfel 3/18/2010 21:05'!
restoreImage

	self firstSubmorph restoreBaseGraphic.! !

!IconicButton methodsFor: 'initialization' stamp: 'sw 9/28/1999 18:38'!
setDefaultLabel
	self labelGraphic: (ScriptingSystem formAtKey: 'squeakyMouse')! !

!IconicButton methodsFor: 'as yet unclassified' stamp: 'sw 11/29/1999 20:56'!
shedSelvedge
	self extent: (self extent - (6@6))! !

!IconicButton methodsFor: 'initialization' stamp: 'sw 6/22/2007 17:04'!
stationarySetup
	"Set up event handlers for mouse actions.  Should be spelled stationery..."

	self actWhen: #startDrag.
	self cornerStyle: #rounded.
	self borderNormal.
	self on: #mouseEnter send: #borderThick to: self.
	self on: #mouseDown send: nil to: nil.
	self on: #mouseLeave send: #borderNormal to: self.
	self on: #mouseLeaveDragging send: #borderNormal to: self.
	self on: #mouseUp send: #borderThick to: self.

	self on: #click send: #launchPartFromClick to: self! !

IconicButton subclass: #SystemWindowButton
	instanceVariableNames: 'dimmed dimmedForm highlightedForm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!SystemWindowButton methodsFor: 'visual properties' stamp: 'bf 3/26/2010 09:03'!
dim
	dimmed := true.
	self restoreImage.! !

!SystemWindowButton methodsFor: 'visual properties' stamp: 'bf 3/26/2010 09:01'!
dimmedForm
	^ dimmedForm ifNil: [ dimmedForm := self firstSubmorph baseGraphic dimmed ]! !

!SystemWindowButton methodsFor: 'visual properties' stamp: 'ar 3/24/2010 20:08'!
handlesMouseOver: evt

	^ true
! !

!SystemWindowButton methodsFor: 'visual properties' stamp: 'mha 3/25/2010 23:35'!
highlight

	self firstSubmorph form: self highlightedForm
! !

!SystemWindowButton methodsFor: 'visual properties' stamp: 'bf 3/26/2010 09:01'!
highlightedForm
	^ highlightedForm ifNil: [ highlightedForm := self firstSubmorph baseGraphic lighter ]! !

!SystemWindowButton methodsFor: 'accessing' stamp: 'bf 3/26/2010 10:24'!
lock
	self firstSubmorph form: self dimmedForm.
	super lock! !

!SystemWindowButton methodsFor: 'visual properties' stamp: 'ar 3/24/2010 20:07'!
mouseEnter: evt

	self highlight.
! !

!SystemWindowButton methodsFor: 'visual properties' stamp: 'ar 3/24/2010 20:07'!
mouseLeave: evt

	self restoreImage.
! !

!SystemWindowButton methodsFor: 'visual properties' stamp: 'bf 3/26/2010 10:23'!
restoreImage
	dimmed == true
		ifTrue: [self firstSubmorph form: self dimmedForm]
		ifFalse: [super restoreImage]
! !

!SystemWindowButton methodsFor: 'visual properties' stamp: 'bf 3/26/2010 09:14'!
undim
	dimmed := false..
	self isLocked ifFalse: [self restoreImage].! !

!SystemWindowButton methodsFor: 'accessing' stamp: 'laza 3/25/2010 14:19'!
unlock
	self restoreImage.
	super unlock! !

!SystemWindowButton methodsFor: 'visual properties' stamp: 'ar 3/24/2010 20:06'!
updateVisualState: evt

	(self containsPoint: evt cursorPoint)
		ifTrue: [self darken]
		ifFalse: [self restoreImage].
! !

SimpleButtonMorph subclass: #SimpleButtonDelayedMenuMorph
	instanceVariableNames: 'didMenu'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!SimpleButtonDelayedMenuMorph methodsFor: 'event handling' stamp: 'RAA 10/26/2000 12:19'!
handlesMouseStillDown: evt

	^true! !

!SimpleButtonDelayedMenuMorph methodsFor: 'event handling' stamp: 'nk 1/11/2004 12:35'!
mouseDown: evt

	didMenu := nil.
	super mouseDown: evt.
! !

!SimpleButtonDelayedMenuMorph methodsFor: 'event handling' stamp: 'ar 10/25/2000 18:08'!
mouseStillDown: evt
	(mouseDownTime isNil or: [(Time millisecondClockValue - mouseDownTime) abs < 1000]) ifTrue: [
		^super mouseStillDown: evt
	].
	didMenu ifNotNil: [^super mouseStillDown: evt].
	self color: oldColor.		"in case menu never returns"
	didMenu := target showMenuFor: actionSelector event: evt.
! !

!SimpleButtonDelayedMenuMorph methodsFor: 'event handling' stamp: 'RAA 7/5/2000 16:50'!
mouseUp: evt

	didMenu == true ifFalse: [^super mouseUp: evt].
	oldColor ifNotNil: [
		self color: oldColor.
		oldColor := nil
	].! !

!SimpleButtonMorph class methodsFor: 'scripting' stamp: 'sw 5/6/1998 14:07'!
authoringPrototype
	^ super authoringPrototype label: 'Button'! !

!SimpleButtonMorph class methodsFor: 'printing' stamp: 'KR 5/23/2008 01:30'!
defaultNameStemForInstances
	^ self = SimpleButtonMorph
		ifTrue: ['Button' translatedNoop]
		ifFalse: [^ super defaultNameStemForInstances]! !

!SimpleButtonMorph class methodsFor: 'as yet unclassified' stamp: 'di 6/5/2000 08:42'!
newWithLabel: labelString

	^ self basicNew initializeWithLabel: labelString
! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: ''!
actionSelector

	^ actionSelector
! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: ''!
actionSelector: aSymbolOrString

	(nil = aSymbolOrString or:
	 ['nil' = aSymbolOrString or:
	 [aSymbolOrString isEmpty]])
		ifTrue: [^ actionSelector := nil].

	actionSelector := aSymbolOrString asSymbol.
! !

!SimpleButtonMorph methodsFor: 'submorphs-add/remove' stamp: 'sw 10/8/2000 08:14'!
actWhen
	"acceptable symbols:  #buttonDown, #buttonUp, and #whilePressed"

	^ actWhen! !

!SimpleButtonMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 6/13/2004 13:46'!
actWhen: condition
	"Accepts symbols:  #buttonDown, #buttonUp, and #whilePressed, #startDrag"
	actWhen := condition.
	actWhen == #startDrag
		ifFalse: [self on: #startDrag send: nil to: nil ]
		ifTrue:[self on: #startDrag send: #doButtonAction to: self].! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'bf 8/14/2010 19:37'!
addCustomMenuItems: aCustomMenu hand: aHandMorph

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	self addLabelItemsTo: aCustomMenu hand: aHandMorph.
	(target isKindOf: BookMorph)
		ifTrue:
			[aCustomMenu add: 'set page sound' translated action: #setPageSound:.
			aCustomMenu add: 'set page visual' translated action: #setPageVisual:]
		ifFalse:
			[
			aCustomMenu add: 'change action selector' translated action: #setActionSelector.
			aCustomMenu add: 'change arguments' translated action: #setArguments.
			aCustomMenu add: 'change when to act' translated action: #setActWhen.
			aCustomMenu add: 'set target' translated action: #sightTargets:.
			target ifNotNil: [aCustomMenu add: 'clear target' translated action: #clearTarget]].
! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 22:06'!
addLabelItemsTo: aCustomMenu hand: aHandMorph 
	aCustomMenu add: 'change label' translated action: #setLabel! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'wiz 1/16/2006 20:10'!
addTargetingMenuItems: aCustomMenu hand: aHandMorph 
	"Add targeting menu items"
	aCustomMenu addLine.

	aCustomMenu add: 'set target' translated action: #targetWith:.
	aCustomMenu add: 'sight target' translated action: #sightTargets:.
	target
		ifNotNil: [aCustomMenu add: 'clear target' translated action: #clearTarget]! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: ''!
arguments

	^ arguments
! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: ''!
arguments: aCollection

	arguments := aCollection asArray copy.
! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'wiz 1/16/2006 20:05'!
clearTarget

	target := nil.
! !

!SimpleButtonMorph methodsFor: 'button' stamp: 'yo 10/7/2006 15:45'!
doButtonAction
	"Perform the action of this button. Subclasses may override this method. The default behavior is to send the button's actionSelector to its target object with its arguments."

	(target notNil and: [actionSelector notNil]) 
		ifTrue: 
			[target perform: actionSelector withArguments: arguments].
	actWhen == #startDrag ifTrue: [oldColor ifNotNil: [self color: oldColor]]! !

!SimpleButtonMorph methodsFor: 'geometry' stamp: '6/7/97 10:53 di'!
extent: newExtent
	| label |
	super extent: newExtent.
	submorphs size = 1 ifTrue:
		["keep the label centered"
		"NOTE: may want to test more that it IS a label..."
		label := self firstSubmorph.
		label position: self center - (label extent // 2)]! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: 'dgd 2/21/2003 22:57'!
fitContents
	| aMorph aCenter |
	aCenter := self center.
	submorphs isEmpty ifTrue: [^self].
	aMorph := submorphs first.
	self extent: aMorph extent + (borderWidth + 6).
	self center: aCenter.
	aMorph position: aCenter - (aMorph extent // 2)! !

!SimpleButtonMorph methodsFor: 'event handling' stamp: 'sw 8/16/97 22:10'!
handlesMouseDown: evt
	^  self isPartsDonor not
! !

!SimpleButtonMorph methodsFor: 'event handling' stamp: 'ar 10/25/2000 18:15'!
handlesMouseStillDown: evt
	^actWhen == #whilePressed! !

!SimpleButtonMorph methodsFor: 'initialization' stamp: 'di 6/5/2000 09:08'!
initialize

	self initializeAllButLabel; setDefaultLabel! !

!SimpleButtonMorph methodsFor: 'initialization' stamp: 'cmm 5/8/2014 10:59'!
initializeAllButLabel
	super initialize.
	self
		 borderWidth: 1 ;
		 color: (Color r: 0.4 g: 0.8 b: 0.6) ;
		 borderColor: self color darker ;
		 borderStyle: BorderStyle thinGray.
	target := nil.
	actionSelector := #flash.
	arguments := Array empty.
	actWhen := #buttonUp! !

!SimpleButtonMorph methodsFor: 'initialization' stamp: 'di 6/5/2000 09:09'!
initializeWithLabel: labelString

	self initializeAllButLabel; label: labelString
! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: ''!
label

	| s |
	s := ''.
	self allMorphsDo: [:m | (m isKindOf: StringMorph) ifTrue: [s := m contents]].
	^ s! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: 'sw 6/11/1999 18:40'!
labelString: aString

	| existingLabel |
	(existingLabel := self findA: StringMorph)
		ifNil:
			[self label: aString]
		ifNotNil:
			[existingLabel contents: aString.
			self fitContents]
! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: 'sw 12/7/1999 18:11'!
label: aString

	| oldLabel m |
	(oldLabel := self findA: StringMorph)
		ifNotNil: [oldLabel delete].
	m := StringMorph contents: aString font: TextStyle defaultFont.
	self extent: m extent + (borderWidth + 6).
	m position: self center - (m extent // 2).
	self addMorph: m.
	m lock! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: 'sw 12/10/1999 09:06'!
label: aString font: aFont

	| oldLabel m |
	(oldLabel := self findA: StringMorph)
		ifNotNil: [oldLabel delete].
	m := StringMorph contents: aString font: (aFont ifNil: [Preferences standardButtonFont]).
	self extent: (m width + 6) @ (m height + 6).
	m position: self center - (m extent // 2).
	self addMorph: m.
	m lock
! !

!SimpleButtonMorph methodsFor: 'event handling' stamp: 'wiz 5/18/2006 21:36'!
mouseDown: evt

	super mouseDown: evt.
	evt yellowButtonPressed ifTrue: [ ^self ] .
	mouseDownTime := Time millisecondClockValue.
	oldColor := self fillStyle. 
	actWhen == #buttonDown
		ifTrue: [ self doButtonAction]
		ifFalse: [ self updateVisualState: evt ].
	self mouseStillDown: evt.! !

!SimpleButtonMorph methodsFor: 'event handling' stamp: 'ar 10/25/2000 18:10'!
mouseMove: evt
	actWhen == #buttonDown ifTrue: [^ self].
	self updateVisualState: evt.! !

!SimpleButtonMorph methodsFor: 'events-processing' stamp: 'nk 1/11/2004 13:25'!
mouseStillDownStepRate
	"Answer how often I want the #handleMouseStillDown: stepped"
	^200! !

!SimpleButtonMorph methodsFor: 'event handling' stamp: 'ar 10/25/2000 18:15'!
mouseStillDown: evt
	actWhen == #whilePressed ifFalse:[^self].
	(self containsPoint: evt cursorPoint) ifTrue:[self doButtonAction].! !

!SimpleButtonMorph methodsFor: 'event handling' stamp: 'wiz 5/18/2006 03:33'!
mouseUp: evt
	super mouseUp: evt.
	oldColor ifNotNil:
		["if oldColor nil, it signals that mouse had not gone DOWN
		inside me, e.g. because of a cmd-drag; in this case we want
		to avoid triggering the action!!"
		self color: oldColor.
		oldColor := nil.
		(self containsPoint: evt cursorPoint) 
				ifTrue: [ actWhen == #buttonUp 
							ifTrue: [self doButtonAction]  ]
				ifFalse: [ self mouseLeave: evt "This is a balk. Note that we have left." ]]
		
! !

!SimpleButtonMorph methodsFor: 'objects from disk' stamp: 'tk 3/2/2007 10:51'!
objectForDataStream: refStrm

	^ super objectForDataStream: refStrm


	"I am about to be written on an object file.  If I send a message to a BookMorph, it would be bad to write that object out.  Create and write out a URLMorph instead.
	Feb 2007: It seems unlikely that Squeak Pages will be used in the OLPC image.  Don't use this code.  Consider removing all code that supports SqueakPages."
"
	| bb thatPage um stem ind sqPg |
	(actionSelector == #goToPageMorph:fromBookmark:) | 
		(actionSelector == #goToPageMorph:) ifFalse: [
			^ super objectForDataStream: refStrm].	'normal case'.

	target url ifNil: ['Later force target book to get a url.'.
		bb _ SimpleButtonMorph new.	'write out a dummy'.
		bb label: self label.
		bb bounds: bounds.
		refStrm replace: self with: bb.
		^ bb].

	(thatPage _ arguments first) url ifNil: [
			'Need to assign a url to a page that will be written later.
			It might have bookmarks too.  Don't want to recurse deeply.  
			Have that page write out a dummy morph to save its url on the server.'.
		stem _ target getStemUrl.	'know it has one'.
		ind _ target pages identityIndexOf: thatPage.
		thatPage reserveUrl: stem,(ind printString),'.sp'].
	um _ URLMorph newForURL: thatPage url.
	sqPg _ thatPage sqkPage clone.
	sqPg contentsMorph: nil.
	um setURL: thatPage url page: sqPg.
	(SqueakPage stemUrl: target url) = (SqueakPage stemUrl: thatPage url) 
		ifTrue: [um book: true]
		ifFalse: [um book: target url].  	'remember which book'.
	um privateOwner: owner.
	um bounds: bounds.
	um isBookmark: true; label: self label.
	um borderWidth: borderWidth; borderColor: borderColor.
	um color: color.
	refStrm replace: self with: um.
	^ um
"! !

!SimpleButtonMorph methodsFor: 'copying' stamp: 'sw 2/15/98 03:49'!
recolor: c
	self color: c.
	oldColor := c! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:44'!
setActionSelector

	| newSel |
	newSel := UIManager default
		request:
'Please type the selector to be sent to
the target when this button is pressed' translated
		initialAnswer: actionSelector.
	newSel isEmpty ifFalse: [self actionSelector: newSel].
! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 19:09'!
setActWhen

	| selections |
	selections := #(buttonDown buttonUp whilePressed startDrag).
	actWhen := UIManager default 
		chooseFrom: (selections collect: [:t | t translated]) 
		values: selections
		title: 'Choose one of the following conditions' translated.! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'nice 12/16/2013 15:46'!
setArguments

	| s newArgs newArgsArray |
	s := WriteStream on: ''.
	arguments do: [:arg | arg printOn: s. s nextPutAll: '. '].
	newArgs := UIManager default
		request:
'Please type the arguments to be sent to the target
when this button is pressed separated by periods' translated
		initialAnswer: s contents.
	newArgs isEmpty ifFalse: [
		newArgsArray := Compiler evaluate: '{', newArgs, '}' for: self.
		self arguments: newArgsArray].
! !

!SimpleButtonMorph methodsFor: 'initialization' stamp: 'sw 9/28/1999 14:05'!
setDefaultLabel
	self label: 'Flash'.
! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:44'!
setLabel

	| newLabel |
	newLabel := UIManager default
		request: 'Please enter a new label for this button'
		initialAnswer: self label.
	newLabel isEmpty ifFalse: [self labelString: newLabel].
! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'di 12/20/1998 16:55'!
setPageSound: event

	^ target menuPageSoundFor: self event: event! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'di 12/20/1998 16:55'!
setPageVisual: event

	^ target menuPageVisualFor: self event: event! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:44'!
setTarget
	
	| newLabel |
	newLabel := UIManager default request: 'Enter an expression that create the target' translated initialAnswer: 'World'.
	newLabel isEmpty
		ifFalse: [self target: (Compiler evaluate: newLabel)]! !

!SimpleButtonMorph methodsFor: 'menu' stamp: 'wiz 12/4/2006 00:32'!
setTarget: evt 
	| rootMorphs |
	rootMorphs := self world rootMorphsAt: evt hand targetPoint.
	target := rootMorphs size > 1
				ifTrue: [rootMorphs second]! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: ''!
target

	^ target
! !

!SimpleButtonMorph methodsFor: 'accessing' stamp: ''!
target: anObject

	target := anObject
! !

!SimpleButtonMorph methodsFor: 'visual properties' stamp: 'yo 8/27/2008 23:58'!
updateVisualState: evt
	
	oldColor ifNotNil: [
		 self color: 
			((self containsPoint: evt cursorPoint)
				ifTrue: [oldColor mixed: 0.5 with: Color white]
				ifFalse: [oldColor])]
! !

!SimpleButtonMorph methodsFor: 'copying' stamp: 'tk 1/6/1999 17:55'!
veryDeepFixupWith: deepCopier
	"If target and arguments fields were weakly copied, fix them here.  If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

super veryDeepFixupWith: deepCopier.
target := deepCopier references at: target ifAbsent: [target].
arguments := arguments collect: [:each |
	deepCopier references at: each ifAbsent: [each]].
! !

!SimpleButtonMorph methodsFor: 'copying' stamp: 'nk 1/23/2004 17:14'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all, but shared.  	Warning!!!!  Every instance variable defined in this class must be handled.  We must also implement veryDeepFixupWith:.  See DeepCopier class comment."

super veryDeepInner: deepCopier.
"target := target.		Weakly copied"
"actionSelector := actionSelector.		a Symbol"
"arguments := arguments.		All weakly copied"
actWhen := actWhen veryDeepCopyWith: deepCopier.
oldColor := oldColor veryDeepCopyWith: deepCopier.
mouseDownTime := nil.! !

SimpleButtonMorph subclass: #UpdatingSimpleButtonMorph
	instanceVariableNames: 'wordingProvider wordingSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!UpdatingSimpleButtonMorph commentStamp: '<historical>' prior: 0!
Adds to SimpleButtonMorph the ability to keep its own wording up to date by send a given message (indicated by its wordingSelector) to a given object (indicated by its wordingTarget, and normally the same as its target.)!

!UpdatingSimpleButtonMorph methodsFor: 'stepping and presenter' stamp: 'sw 10/30/2000 08:56'!
step
	"If appropriate update the receiver's label"

	| newString |
	super step.
	wordingProvider ifNotNil:
		[newString := wordingProvider perform: wordingSelector.
		newString = self label ifFalse: [self labelString: newString; changed]]! !

!UpdatingSimpleButtonMorph methodsFor: 'testing' stamp: 'sw 10/30/2000 08:57'!
stepTime
	"Answer the desired time between steps in milliseconds.  If the receiver has a wordingProvider that may dynamically provide changed wording for the label, step once every 1.5 seconds"

	^ wordingProvider ifNotNil: [1500] ifNil: [super stepTime]! !

!UpdatingSimpleButtonMorph methodsFor: 'testing' stamp: 'sw 10/30/2000 08:55'!
wantsSteps
	"Answer whether the receiver wishes to be sent the #step message.  In the current case, this decision depends on whether there is a wordingProvider which can dynamically provide fresh wording for the button's label"

	^ wordingProvider notNil! !

!UpdatingSimpleButtonMorph methodsFor: 'as yet unclassified' stamp: 'sw 6/11/1999 18:30'!
wordingSelector: aSelector
	wordingSelector := aSelector.
	wordingProvider ifNil: [wordingProvider := target]! !

RectangleMorph subclass: #TextFieldMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextFieldMorph commentStamp: '<historical>' prior: 0!
Act as a field in a HyperCard-like setting.  Has both properties of a Rectangle, and exposes some proteries of the TextMorph it owns.

!

!TextFieldMorph class methodsFor: 'scripting' stamp: 'md 11/14/2003 17:32'!
authoringPrototype 
	"Answer an instance of the receiver that can serve as a prototype for authoring"

	| proto |
	proto := super authoringPrototype.
	proto setProperty: #shared toValue: true.
	proto extent: 170 @ 30.
	proto color: Color veryLightGray lighter.
	proto contents: 'on a clear day you can...'.
	^ proto
! !

!TextFieldMorph class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 11:56'!
initialize

	self registerInFlapsRegistry.	! !

!TextFieldMorph class methodsFor: 'class initialization' stamp: 'KR 8/12/2007 11:16'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | cl registerQuad: {#TextFieldMorph. #exampleBackgroundField. 'Scrolling Field' translatedNoop. 'A scrolling data field which will have a different value on every card of the background' translatedNoop}
						forFlapNamed: 'Scripting'.]! !

!TextFieldMorph class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 12:41'!
unload
	"Unload the receiver from global registries"

	self environment at: #Flaps ifPresent: [:cl |
	cl unregisterQuadsWithReceiver: self] ! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'tk 9/6/2000 11:03'!
append: stringOrText
	"add to my text"
	| tm |

	(tm := self findA: TextMorph) ifNil: [^ nil].
	tm contents append: stringOrText.
	tm releaseParagraph; paragraph.


	! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'tk 8/30/2000 14:22'!
contents
	| tm |
	"talk to my text"

	(tm := self findA: TextMorph) ifNil: [^ nil].
	^ tm contents! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'ar 4/10/2005 18:54'!
contents: textOrString
	"talk to my text"
	| tm newText atts |

	(tm := self findA: TextMorph) ifNil: [^ nil].
	textOrString isString ifTrue: [
		tm contents ifNotNil: ["Keep previous properties of the field"
			newText := textOrString asText.
			atts := tm contents attributesAt: 1.
			atts do: [:each | newText addAttribute: each].
			^ tm contents: newText]].

	^ tm contents: textOrString! !

!TextFieldMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:54'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color veryLightGray lighter! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'tk 9/4/2000 16:28'!
fit
	"tell my text to recompute its looks"
	| tm |

	(tm := self findA: TextMorph) ifNil: [^ nil].
	tm releaseParagraph; paragraph.! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'tk 8/30/2000 14:24'!
fontName: fontName size: fontSize
	| tm |
	"talk to my text"

	(tm := self findA: TextMorph) ifNil: [^ nil].
	^ tm fontName: fontName size: fontSize
! !

!TextFieldMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:54'!
initialize
	"initialize the state of the receiver"
	| tm |
	super initialize.
	""
	
	self addMorph: (tm := TextMorph new).
	tm fillingOnOff! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'tk 9/6/2000 12:33'!
lineCount
	| tm |
	"how many lines in my text"

	(tm := self findA: TextMorph) ifNil: [^ nil].
	^ tm contents string lineCount! !

!TextFieldMorph methodsFor: 'just like textMorph' stamp: 'ar 12/27/2001 00:03'!
prepend: stringOrText
	"add to my text"
	| tm |

	(tm := self findA: TextMorph) ifNil: [^ nil].
	tm contents prepend: stringOrText.
	tm releaseParagraph; paragraph.


	! !

RectangleMorph subclass: #TextMorph
	instanceVariableNames: 'textStyle text wrapFlag paragraph editor container predecessor successor backgroundColor margins editHistory readOnly autoFit'
	classVariableNames: 'CaretForm DefaultEditorClass'
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!TextMorph commentStamp: 'nice 3/24/2010 07:40' prior: 0!
TextMorphs support display of text with emphasis.  They also support reasonable text-editing capabilities, as well as embedded hot links, and the ability to embed submorphs in the text.

Late in life, TextMorph was made a subclass of BorderedMorph to provide border and background color if desired.  In order to keep things compatible, protocols have been redirected so that color (preferably textColor) relates to the text, and backgroundColor relates to the inner fill color.

Text display is clipped to the innerBounds of the rectangle, and text composition is normally performed within a rectangle which is innerBounds inset by the margins parameter.

If text has been embedded in another object, one can elect to fill the owner's shape, in which case the text will be laid out in the shape of the owner's shadow image (including any submorphs other than the text).  One can also elect to have the text avoid occlusions, in which case it will avoid the bounds of any sibling morphs that appear in front of it.  It may be necessary to update bounds in order for the text runaround to notice the presence of a new occluding shape.

The optional autoFitContents property enables the following feature:  if the text contents changes, then the bounds of the morph will be adjusted to fit the minimum rectangle that encloses the text (plus any margins specified).  Similarly, any attempt to change the size of the morph will be resisted if this parameter is set.  Except...

If the wrapFlag parameter is true, then text will be wrapped at word boundaries based on the composition width (innerBounds insetBy: margins) width.  Thus an attempt to resize the morph in autofit mode, if it changes the width, will cause the text to be recomposed with the new width, and then the bounds will be reset to the minimum enclosing rectangle.  Similarly, if the text contents are changed with the wrapFlag set to true, word wrap will be performed based on the current compostion width, after which the bounds will be set (or not), based on the autoFitcontents property.

Note that fonts can only be applied to the TextMorph as a whole.  While you can change the size, color, and emphasis of a subsection of the text and have it apply to only that subsection, changing the font changes the font for the entire contents of the TextMorph. 

Still a TextMorph can be composed of several texts of different fonts
| font1 font2 t1 t2 tMorph|
tMorph := TextMorph new.
font1 := (TextFontReference toFont: (StrikeFont familyName: 'Atlanta' size: 22)).
font2 := (TextFontReference toFont: (StrikeFont familyName: 'Atlanta' size: 11)).
t1 := 'this is font1' asText addAttribute: font1.
t2 := ' and this is font2' asText addAttribute: font2.
tMorph contents: (t1,t2).
tMorph openInHand.


Yet to do:
Make a comprehensive control for the eyedropper, with border width and color, inner color and text color, and margin widths.!

TextMorph subclass: #StringMorphEditor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!StringMorphEditor commentStamp: '<historical>' prior: 0!
I am a textMorph used as a pop-up editor for StringMorphs.  I present a yellow background and I go away when a CR is typed or when the user clicks elsewhere.!

!StringMorphEditor class methodsFor: 'new-morph participation' stamp: 'kfr 5/1/2000 13:41'!
includeInNewMorphMenu
	"Not to be instantiated from the menu"
	^ false! !

!StringMorphEditor methodsFor: 'drawing' stamp: 'sw 9/7/1999 16:22'!
drawOn: aCanvas

	aCanvas fillRectangle: self bounds color: Color yellow muchLighter.
	^ super drawOn: aCanvas! !

!StringMorphEditor methodsFor: 'display' stamp: 'yo 9/29/2006 07:14'!
initialize
	"Initialize the receiver.  Give it a white background"

	super initialize.
	self backgroundColor: Color white.
	self textColor: Color red.! !

!StringMorphEditor methodsFor: 'event handling' stamp: 'nk 1/23/2004 13:18'!
keyboardFocusChange: aBoolean
	| hadFocus |
	owner ifNil: [ ^self ].
	hadFocus := owner hasFocus.
	super keyboardFocusChange: aBoolean.
	aBoolean ifFalse:
		[hadFocus ifTrue:
			[owner lostFocusWithoutAccepting; doneWithEdits].
		^ self delete]! !

!StringMorphEditor methodsFor: 'event handling' stamp: 'bf 4/20/2016 17:58'!
keyStroke: evt
	"This is hugely inefficient, but it seems to work, and it's unlikely it will ever need
	to be any more efficient -- it's only intended to edit single-line strings."

	| char priorEditor newSel |
	(((char := evt keyCharacter) = Character enter) or: [(char = Character cr)
			or: [char = $s and: [evt commandKeyPressed]]])
				ifTrue: [owner doneWithEdits; acceptContents.
	self flag: #arNote. "Probably unnecessary"
						evt hand releaseKeyboardFocus.
						^ self delete].
	
	(char = $l and: [evt commandKeyPressed]) ifTrue:   "cancel"
		[owner cancelEdits.
		evt hand releaseKeyboardFocus.
		^ self delete].

	super keyStroke: evt.
	
	owner ifNil: [^self].
	owner interimContents: self contents asString.
	newSel := self editor selectionInterval.

	priorEditor := self editor.  "Save editor state"
	self releaseParagraph.  "Release paragraph so it will grow with selection."
	self paragraph.      "Re-instantiate to set new bounds"
	self installEditorToReplace: priorEditor.  "restore editor state"
	self editor selectFrom: newSel first to: newSel last.
! !

TextMorph subclass: #TextMorphForEditView
	instanceVariableNames: 'editView acceptOnCR'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextMorphForEditView class methodsFor: 'new-morph participation' stamp: 'kfr 5/1/2000 13:41'!
includeInNewMorphMenu
	"Not to be instantiated from the menu"
	^ false! !

!TextMorphForEditView methodsFor: 'editing' stamp: 'di 4/22/1998 10:57'!
acceptContents
	"The message is sent when the user hits enter or Cmd-S.
	Accept the current contents and end editing."
	self updateFromParagraph.
	editView accept.! !

!TextMorphForEditView methodsFor: 'editing' stamp: 'sw 8/12/2002 00:02'!
acceptOnCR
	"Answer whether the receiver wants to accept when the Return key is hit"

	^ acceptOnCR == true! !

!TextMorphForEditView methodsFor: 'accept/cancel' stamp: 'di 9/11/1998 15:42'!
acceptOnCR: trueOrFalse
	acceptOnCR := trueOrFalse! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'ar 11/8/2000 15:50'!
autoScrollView: evt
	"This is kind of a hack because the PluggableTextMorph expects me to first expand the selection before auto scrolling will work."
	| localEvt |
	localEvt := evt transformedBy: (self transformedFrom: editView).
	super mouseMove: localEvt.
	editView scrollSelectionIntoView: localEvt.! !

!TextMorphForEditView methodsFor: 'editing' stamp: 'di 4/22/1998 11:03'!
cancelEdits
	"The message is sent when the user hits enter or Cmd-L.
	Cancel the current contents and end editing."
	self releaseParagraph.
	editView cancel! !

!TextMorphForEditView methodsFor: 'accessing' stamp: 'mt 7/24/2016 14:05'!
caretColor
	^ self valueOfProperty: #caretColor ifAbsent: [Color red]! !

!TextMorphForEditView methodsFor: 'accessing' stamp: 'mt 8/3/2016 09:54'!
caretColor: aColor
	self
		setProperty: #caretColor
		toValue: aColor.! !

!TextMorphForEditView methodsFor: 'objects from disk' stamp: 'RAA 12/20/2000 17:50'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	acceptOnCR ifNil: [acceptOnCR := false].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!TextMorphForEditView methodsFor: 'private' stamp: 'mt 7/24/2016 14:13'!
createParagraph

	super createParagraph.
	
	paragraph
		caretColor: self caretColor;
		selectionColor: self selectionColor;
		unfocusedSelectionColor: self unfocusedSelectionColor.
		
	^ paragraph! !

!TextMorphForEditView methodsFor: 'drawing' stamp: 'di 6/22/1998 10:44'!
drawNullTextOn: aCanvas
	"Just run the normal code to show selection in a window"
	aCanvas paragraph: self paragraph bounds: bounds color: color
! !

!TextMorphForEditView methodsFor: 'edit view' stamp: 'di 6/22/1998 01:31'!
editView
	^ editView! !

!TextMorphForEditView methodsFor: 'macpal' stamp: 'di 11/10/1998 10:13'!
flash
	^ editView flash! !

!TextMorphForEditView methodsFor: 'editing' stamp: 'di 10/5/1998 14:03'!
handleInteraction: interActionBlock fromEvent: evt
	"Overridden to pass along a model to the editor for, eg, link resolution, doits, etc"

	self editor model: editView model.  "For evaluateSelection, etc"
	^ super handleInteraction: interActionBlock fromEvent: evt! !

!TextMorphForEditView methodsFor: 'editing' stamp: 'di 4/21/1998 13:23'!
hasUnacceptedEdits: aBoolean
	"Set the hasUnacceptedEdits flag in my view."

	editView hasUnacceptedEdits: aBoolean! !

!TextMorphForEditView methodsFor: 'initialization' stamp: 'di 9/11/1998 15:43'!
initialize
	super initialize.
	acceptOnCR := false! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'mt 6/5/2016 09:19'!
keyboardFocusChange: weHaveFocus 
	"Overwritten because we do not want to release the editor if we do not have the focus. Editors and their paragraphs store the text selection, which we want to reuse after getting keyboard focus again. This supports keyboard navigation."

	paragraph ifNotNil: [:para | para focused: weHaveFocus].
	
	weHaveFocus
		ifFalse: [self stopBlinking]
		ifTrue: [
			self hasFocus
				ifFalse: [self editor "forces install"].
			Editor blinkingCursor
				ifTrue: [self startBlinking]
				ifFalse: [self resetBlinkCursor "ensure caret visible"]].
			
	self changed.

	"Tell my edit-view about this because I am his delegate."
	self editView keyboardFocusChange: weHaveFocus.
! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'mt 6/6/2016 17:53'!
keyStroke: evt
	| view |
	editView deleteBalloon.
	self editor model: editView model.  "For evaluateSelection"
	view := editView.  "Copy into temp for case of a self-mutating doit"
	(acceptOnCR and: [evt keyCharacter = Character cr])
		ifTrue: [^ self editor accept].
	super keyStroke: evt.
	view scrollSelectionIntoView.
	
	"Make a cheap check and guess editing. (Alternative would be to copy the old contents and then compare them against the new ones. Maybe add a better hook in the TextEditor."
	(self readOnly not and: [evt keyCharacter isAlphaNumeric or: [evt keyCharacter isSeparator]])
		ifTrue: [view textEdited: self contents].! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'kfr 7/3/2014 13:23'!
mouseDown: event

	event yellowButtonPressed ifTrue: [
		(self editor yellowButtonDown: event) ifTrue:[^self].
		^ editView yellowButtonActivity: event shiftPressed].
	^ super mouseDown: event
! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'ar 9/26/2001 22:28'!
mouseMove: evt
	| editEvt |
	super mouseMove: evt.
	evt redButtonPressed ifFalse: [^ self].
	editEvt := evt transformedBy: (self transformedFrom: editView) inverseTransformation.
	(editEvt position y between: editView top and: editView bottom) ifFalse:[
		"Start auto-scrolling"
		self startStepping: #autoScrollView:
			at: Time millisecondClockValue
			arguments: (Array with: editEvt)
			stepTime: 100. "fast enough"
	] ifTrue:[
		self stopSteppingSelector: #autoScrollView:.
	].! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'yo 11/7/2002 18:44'!
mouseUp: evt
	super mouseUp: evt.
	self stopSteppingSelector: #autoScrollView:.
	editView scrollSelectionIntoView: evt.

	self setCompositionWindow.
! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'tpr 6/6/2016 15:04'!
preferredKeyboardPosition

	| pos |
	pos := super preferredKeyboardPosition.
	^ pos + (self boundsInWorld) topLeft.
! !

!TextMorphForEditView methodsFor: 'accessing' stamp: 'mt 7/24/2016 14:06'!
selectionColor
	^ self valueOfProperty: #selectionColor ifAbsent: [Color blue muchLighter]! !

!TextMorphForEditView methodsFor: 'accessing' stamp: 'mt 8/3/2016 09:54'!
selectionColor: aColor

	self
		setProperty: #selectionColor
		toValue: aColor.! !

!TextMorphForEditView methodsFor: 'edit view' stamp: 'di 4/21/1998 13:09'!
setEditView: editPane
	editView := editPane! !

!TextMorphForEditView methodsFor: 'debug and other' stamp: 'sw 11/2/1998 15:51'!
tempCommand
	"Smalltalk browseAllImplementorsOf: #tempCommand"
	"Place your definition for tempCommand for this class here"! !

!TextMorphForEditView methodsFor: 'accessing' stamp: 'mt 7/24/2016 14:06'!
unfocusedSelectionColor
	^ self valueOfProperty: #unfocusedSelectionColor ifAbsent: [Color blue muchLighter]! !

!TextMorphForEditView methodsFor: 'accessing' stamp: 'mt 8/3/2016 09:55'!
unfocusedSelectionColor: aColor

	self
		setProperty: #unfocusedSelectionColor
		toValue: aColor.! !

!TextMorphForEditView methodsFor: 'private' stamp: 'dew 2/21/1999 03:09'!
updateFromParagraph  
	super updateFromParagraph.
	editView setScrollDeltas.! !

!TextMorphForEditView methodsFor: 'event handling' stamp: 'sw 8/29/2000 15:06'!
wouldAcceptKeyboardFocusUponTab
	"Answer whether the receiver would be a happy inheritor of keyboard focus if tab were hit in an enclosing playfield under propitious circumstances.  Does not make sense for this kind of morph, which is encased in a window"

	^ false! !

!TextMorph class methodsFor: 'scripting' stamp: 'yo 7/2/2004 21:27'!
authoringPrototype
	| t |
	t := super authoringPrototype.
	t contents: 'abc' translated asText.
	t wrapFlag: true. 

"Strangeness here in order to avoid two offset copies of the default contents when operating in an mvc project before cursor enters the morphic window"
	t paragraph.
	^ t! !

!TextMorph class methodsFor: 'connectorstext-parts bin' stamp: 'yo 1/20/2005 12:44'!
boldAuthoringPrototype
	"TextMorph boldAuthoringPrototype openInHand"
	| text |
	text := Text string: 'Text' translated attributes: { TextEmphasis bold. }.
	^self new
		contentsWrapped: text;
		fontName: 'BitstreamVeraSans' pointSize: 24;
		paragraph;
		extent: 79@36;
		margins: 4@0;
		fit;
		yourself
! !

!TextMorph class methodsFor: 'parts bin' stamp: 'tk 7/2/2007 15:01'!
borderedPrototype

	| t |
	t _ self authoringPrototype.
	t fontName: 'BitstreamVeraSans' pointSize: 24.
	t autoFit: false; extent: 250@100.
	t borderWidth: 1; margins: 4@0; backgroundColor: Color white.

"Strangeness here in order to avoid two offset copies of the default contents when operating in an mvc project before cursor enters the morphic window"
	t paragraph.
	^ t! !

!TextMorph class methodsFor: 'class initialization' stamp: 'mt 11/11/2015 12:55'!
cleanUp

	TextMorph allSubInstancesDo: [:m | m releaseEditor].! !

!TextMorph class methodsFor: 'class initialization' stamp: 'ar 8/6/2009 16:03'!
defaultEditorClass
	"Answers the default editor class for TextMorph"
	^DefaultEditorClass! !

!TextMorph class methodsFor: 'class initialization' stamp: 'ar 8/6/2009 16:03'!
defaultEditorClass: aTextEditorClass
	"Sets the default editor class for TextMorph"
	"
		TextMorph defaultEditorClass: TextMorphEditor.
		TextMorph defaultEditorClass: TextEditor.
	"
	DefaultEditorClass := aTextEditorClass! !

!TextMorph class methodsFor: 'scripting' stamp: 'KR 5/23/2008 01:31'!
defaultNameStemForInstances
	^ 'Text' translatedNoop! !

!TextMorph class methodsFor: 'parts bin' stamp: 'sw 6/13/2001 21:58'!
exampleBackgroundLabel
	"Answer a background label for a parts bin"

	| aTextMorph |
	aTextMorph := self authoringPrototype.
	aTextMorph contents: 'background
label' asText.  
	aTextMorph beAllFont: (StrikeFont familyName: #NewYork size: 18).
	aTextMorph color: Color brown.
	aTextMorph setProperty: #shared toValue: true.
	^ aTextMorph
! !

!TextMorph class methodsFor: 'parts bin' stamp: 'yo 9/27/2006 22:34'!
fancyPrototype

	| t |
	t _ self authoringPrototype.
	t autoFit: false; extent: 150@75.
	t borderWidth: 2; margins: 4@0; useRoundedCorners.	"Why not rounded?"
	"fancy font, shadow, rounded"
	t fontName: Preferences standardEToysFont familyName size: 18; textColor: Color blue; fillStyle: Color lightBrown.
	t addDropShadow.

"Strangeness here in order to avoid two offset copies of the default contents when operating in an mvc project before cursor enters the morphic window"
	t paragraph.
	^ t! !

!TextMorph class methodsFor: 'new-morph participation' stamp: 'kfr 5/1/2000 13:42'!
includeInNewMorphMenu
	^ true! !

!TextMorph class methodsFor: 'class initialization' stamp: 'ar 1/2/2010 14:47'!
initialize	"TextMorph initialize"

	"Initialize the default text editor class to use"
	DefaultEditorClass := SmalltalkEditor.

	"Initialize constants shared by classes associated with text display."

	CaretForm := (ColorForm extent: 16@5
					fromArray: #(2r001100e26 2r001100e26 2r011110e26 2r111111e26 2r110011e26)
					offset: -2@0)
					colors: (Array with: Color transparent with: Preferences textHighlightColor).

	self registerInFlapsRegistry.
! !

!TextMorph class methodsFor: 'class initialization' stamp: 'kfr 5/18/2011 22:17'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | cl registerQuad: {#TextMorph. #authoringPrototype. 'Text' translatedNoop. 'Text that you can edit into anything you desire.' translatedNoop}
						forFlapNamed: 'PlugIn Supplies'.
						cl registerQuad: {#TextMorph	. #exampleBackgroundLabel. 'Background Label' translatedNoop. 'A piece of text that will occur on every card of the background' translatedNoop}
						forFlapNamed: 'Scripting'.
						cl registerQuad: {#TextMorph	. #exampleBackgroundField. 'Background Field' translatedNoop. 'A data field which will have a different value on every card of the background' translatedNoop}
						forFlapNamed: 'Scripting'.
						cl registerQuad: {#TextMorph	. #authoringPrototype. 'Simple Text'	translatedNoop. 'Text that you can edit into anything you wish' translatedNoop}
						forFlapNamed: 'Stack Tools'.
						cl registerQuad: {#TextMorph	. #fancyPrototype. 'Fancy Text' 	translatedNoop. 'A text field with a rounded shadowed border, with a fancy font.' translatedNoop}
						forFlapNamed: 'Stack Tools'.
						cl registerQuad: {#TextMorph	. #authoringPrototype. 'Text' translatedNoop. 'Text that you can edit into anything you desire.' translatedNoop}
						forFlapNamed: 'Supplies'.]! !

!TextMorph class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 12:41'!
unload
	"Unload the receiver from global registries"

	self environment at: #Flaps ifPresent: [:cl |
	cl unregisterQuadsWithReceiver: self] ! !

!TextMorph methodsFor: 'editing' stamp: 'di 4/22/1998 10:57'!
acceptContents
	"The message is sent when the user hits enter or Cmd-S.
	Accept the current contents and end editing.
	This default implementation does nothing."
	self updateFromParagraph! !

!TextMorph methodsFor: 'layout' stamp: 'tk 6/30/1998 17:06'!
acceptDroppingMorph: aMorph event: evt
	"This message is sent when a morph is dropped onto me."

	self addMorphFront: aMorph fromWorldPosition: aMorph position.
		"Make a TextAnchor and install it in a run."! !

!TextMorph methodsFor: 'editing' stamp: 'sw 8/12/2002 01:10'!
acceptOnCR
	"Answer whether the receiver wants to accept when the Return key is hit.  Generic TextMorph has no such feature, but subclasses may."

	^ false! !

!TextMorph methodsFor: 'menu' stamp: 'fbs 7/3/2013 21:39'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	| outer |
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu add: 'text properties...' translated action: #changeTextColor.
	aCustomMenu addUpdating: #autoFitString target: self action: #autoFitOnOff.
	aCustomMenu addUpdating: #wrapString target: self action: #wrapOnOff.
	aCustomMenu add: 'text margins...' translated action: #changeMargins:.
	aCustomMenu add: 'add predecessor' translated action: #addPredecessor:.
	aCustomMenu add: 'add successor' translated action: #addSuccessor:.
	
	outer := self owner.
	outer ifNotNil: [
	outer isLineMorph ifTrue:
		[container isNil
			ifTrue: [Smalltalk at: #TextOnCurveContainer ifPresent: [:ignored | aCustomMenu add: 'follow owner''s curve' translated action: #followCurve]]
			ifFalse: [aCustomMenu add: 'reverse direction' translated action: #reverseCurveDirection.
					aCustomMenu add: 'set baseline' translated action: #setCurveBaseline:]]
		ifFalse:
		[self fillsOwner
			ifFalse: [aCustomMenu add: 'fill owner''s shape' translated action: #fillingOnOff]
			ifTrue: [aCustomMenu add: 'rectangular bounds' translated action: #fillingOnOff].
		self avoidsOcclusions
			ifFalse: [aCustomMenu add: 'avoid occlusions' translated action: #occlusionsOnOff]
			ifTrue: [aCustomMenu add: 'ignore occlusions' translated action: #occlusionsOnOff]]].
	aCustomMenu addLine.
	aCustomMenu add: 'holder for characters' translated action: #holderForCharacters
! !

!TextMorph methodsFor: 'submorphs-add/remove' stamp: 'ar 12/17/2001 13:21'!
addMorphFront: aMorph fromWorldPosition: wp 
	"Overridden for more specific re-layout and positioning"
	aMorph textAnchorType == #document 
		ifFalse:[^self anchorMorph: aMorph at: wp type: aMorph textAnchorType].
	self addMorphFront: aMorph.
! !

!TextMorph methodsFor: 'linked frames' stamp: 'di 7/28/2001 10:34'!
addPredecessor: evt
	| newMorph |
	newMorph := self copy predecessor: predecessor successor: self.
	newMorph extent: self width @ 100.
	predecessor ifNotNil: [predecessor setSuccessor: newMorph].
	self setPredecessor: newMorph.
	predecessor recomposeChain.
	evt hand attachMorph: newMorph! !

!TextMorph methodsFor: 'linked frames' stamp: 'di 7/28/2001 10:35'!
addSuccessor: evt
	| newMorph |
	newMorph := self copy predecessor: self successor: successor.
	newMorph extent: self width @ 100.
	successor ifNotNil: [successor setPredecessor: newMorph].
	self setSuccessor: newMorph.
	successor recomposeChain.
	evt hand attachMorph: newMorph! !

!TextMorph methodsFor: 'private' stamp: 'di 11/8/97 16:02'!
adjustLineIndicesBy: delta
	paragraph ifNotNil: [paragraph adjustLineIndicesBy: delta]! !

!TextMorph methodsFor: 'anchors' stamp: 'ar 12/17/2001 13:38'!
adjustTextAnchor: aMorph
	"Later compute the new relative position of aMorph if it is #paragraph anchored."! !

!TextMorph methodsFor: 'anchors' stamp: 'ar 8/10/2003 18:19'!
anchorMorph: aMorph at: aPoint type: anchorType
	| relPt index newText block |
	aMorph owner == self ifTrue:[self removeMorph: aMorph].
	aMorph textAnchorType: nil.
	aMorph relativeTextAnchorPosition: nil.
	self addMorphFront: aMorph.
	aMorph textAnchorType: anchorType.
	aMorph relativeTextAnchorPosition: nil.
	anchorType == #document ifTrue:[^self].
	relPt := self transformFromWorld globalPointToLocal: aPoint.
	index := (self paragraph characterBlockAtPoint: relPt) stringIndex.
	newText := Text string: (String value: 1) attribute: (TextAnchor new anchoredMorph: aMorph).
	anchorType == #inline ifTrue:[
		self paragraph replaceFrom: index to: index-1 with: newText displaying: false.
	] ifFalse:[
		index := index min: paragraph text size.
		index := paragraph text string lastIndexOf: Character cr startingAt: index ifAbsent:[0].
		block := paragraph characterBlockForIndex: index+1.
		aMorph relativeTextAnchorPosition: (relPt x - bounds left) @ (relPt y - block top ).
		self paragraph replaceFrom: index+1 to: index with: newText displaying: false.
	].
	self fit.! !

!TextMorph methodsFor: 'drawing' stamp: 'yo 9/27/2006 22:40'!
areasRemainingToFill: aRectangle
	"Overridden from BorderedMorph to test backgroundColor instead of (text) color."
	(self backgroundColor isNil or: [self backgroundColor asColor isTranslucent])
		ifTrue: [^ Array with: aRectangle].
	self wantsRoundedCorners
	ifTrue: [(borderWidth > 0 and: [borderColor isColor and: [borderColor isTranslucent]])
				ifTrue: [^ aRectangle areasOutside: (self innerBounds intersect: self boundsWithinCorners)]
				ifFalse: [^ aRectangle areasOutside: self boundsWithinCorners]]
	ifFalse: [(borderWidth > 0 and: [borderColor isColor and: [borderColor isTranslucent]])
				ifTrue: [^ aRectangle areasOutside: self innerBounds]
				ifFalse: [^ aRectangle areasOutside: self bounds]]! !

!TextMorph methodsFor: 'accessing' stamp: 'sw 1/12/98 23:40'!
asText
	^ text! !

!TextMorph methodsFor: 'menu' stamp: 'mt 4/27/2015 15:19'!
autoFitOnOff
	autoFit := self isAutoFit not.
	self isAutoFit ifTrue: [self fit]! !

!TextMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 22:06'!
autoFitString
	"Answer the string to put in a menu that will invite the user to 
	switch autoFit mode"
	^ (self isAutoFit
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'text auto fit' translated! !

!TextMorph methodsFor: 'accessing' stamp: 'spd 1/4/2011 17:36'!
autoFit: trueOrFalse
	"Whether I automatically adjust my size to fit text as it changes"
	
	self isAutoFit = trueOrFalse ifTrue: [^ self].
	self autoFitOnOff.! !

!TextMorph methodsFor: 'containment' stamp: 'nk 8/18/2004 11:23'!
avoidsOcclusions
	^container notNil and: [ container avoidsOcclusions ]
! !

!TextMorph methodsFor: 'accessing' stamp: 'yo 9/27/2006 22:25'!
backgroundColor
	^ self fillStyle.
! !

!TextMorph methodsFor: 'accessing' stamp: 'yo 9/27/2006 22:25'!
backgroundColor: newColor
	self fillStyle: newColor.
! !

!TextMorph methodsFor: 'initialization' stamp: 'bf 8/26/2008 17:25'!
beAllFont: aFont

	textStyle _ TextStyle fontArray: (Array with: aFont).
	text ifNotNil: [text addAttribute: (TextFontReference toFont: aFont)].
	self releaseCachedState; changed! !

!TextMorph methodsFor: 'blinking' stamp: 'ar 12/15/2008 23:00'!
blinkStart
	"Reset time for blink cursor after which blinking should actually start"
	^self valueOfProperty: #blinkStart ifAbsent:[Time millisecondClockValue]
! !

!TextMorph methodsFor: 'blinking' stamp: 'ar 12/15/2008 23:00'!
blinkStart: msecs
	"Reset time for blink cursor after which blinking should actually start"
	^self setProperty: #blinkStart toValue: msecs
! !

!TextMorph methodsFor: 'accessing' stamp: 'di 7/19/2001 11:08'!
borderWidth: newWidth
	super borderWidth: newWidth.
	paragraph ifNotNil: [self composeToBounds].! !

!TextMorph methodsFor: 'geometry' stamp: 'di 10/8/1998 23:46'!
bounds
	container ifNil: [^ bounds].
	^ container bounds ifNil: [bounds]! !

!TextMorph methodsFor: 'editing' stamp: 'di 4/22/1998 11:00'!
cancelEdits
	"The message is sent when the user hits enter or Cmd-L.
	Cancel the current contents and end editing.
	This default implementation does nothing."
	self releaseParagraph! !

!TextMorph methodsFor: 'alignment' stamp: 'di 10/25/97 19:19'!
centered 
	self paragraph centered.
	self updateFromParagraph ! !

!TextMorph methodsFor: 'menu' stamp: 'ul 12/12/2009 14:05'!
changeMargins: evt
	| handle origin aHand oldMargin newMargin |
	aHand := evt ifNil: [self primaryHand] ifNotNil: [evt hand].
	origin := aHand position.
	oldMargin := margins.
	(handle := HandleMorph new)
		forEachPointDo:
			[:newPoint | handle removeAllMorphs.
			handle addMorph:
				(LineMorph from: origin to: newPoint color: Color black width: 1).
			newMargin := (newPoint - origin max: 0@0) // 5.
			self margins: newMargin]
		lastPointDo:
			[:newPoint | handle deleteBalloon.
			self halo ifNotNil: [ :halo | halo addHandles].
			self rememberCommand:
				(Command new cmdWording: ('margin change for ' translated,self nameForUndoWording);
					undoTarget: self selector: #margins: argument: oldMargin;
					redoTarget: self selector: #margins: argument: newMargin;
					yourself)].
	aHand attachMorph: handle.
	handle setProperty: #helpAtCenter toValue: true.
	handle showBalloon:
'Move cursor down and to the right
to increase margin inset.
Click when done.' hand: evt hand.
	handle startStepping! !

!TextMorph methodsFor: 'menu' stamp: 'RAA 8/21/2001 11:08'!
changeTextColor
	"Change the color of the receiver -- triggered, e.g. from a menu"

	self openATextPropertySheet.
">>>>>
	ColorPickerMorph new
		choseModalityFromPreference;
		sourceHand: self activeHand;
		target: self;
		selector: #textColor:;
		originalColor: self textColor;
		putUpFor: self near: self fullBoundsInWorld
<<<<"! !

!TextMorph methodsFor: 'editing' stamp: 'di 10/5/1998 13:55'!
chooseAlignment
	self editor changeAlignment.
	self updateFromParagraph! !

!TextMorph methodsFor: 'editing' stamp: 'mt 5/3/2015 19:13'!
chooseEmphasis
	self editor changeEmphasisOrAlignment.
	self updateFromParagraph! !

!TextMorph methodsFor: 'editing' stamp: 'sw 9/27/1999 12:13'!
chooseEmphasisOrAlignment
	self editor changeEmphasisOrAlignment.
	self updateFromParagraph! !

!TextMorph methodsFor: 'editing' stamp: 'ar 12/17/2001 13:09'!
chooseFont
	self editor changeTextFont.
	self updateFromParagraph.! !

!TextMorph methodsFor: 'editing' stamp: 'vj 9/14/2003 20:53'!
chooseStyle
	self editor changeStyle.
	self updateFromParagraph.! !

!TextMorph methodsFor: 'private' stamp: 'di 6/22/2001 09:10'!
clippingRectangle
	^ self innerBounds! !

!TextMorph methodsFor: 'private' stamp: 'di 8/4/2000 16:06'!
composeToBounds
	"Compose my text to fit my bounds.
	If any text lies outside my bounds, it will be clipped, or
	if I have successors, it will be shown in the successors."
	| |
	self releaseParagraph; paragraph.
	container ifNotNil:
		[self privateBounds: container bounds truncated].
	self paragraph positionWhenComposed: self position.
	successor ifNotNil:
		[successor predecessorChanged].

! !

!TextMorph methodsFor: 'private' stamp: 'di 7/20/2001 22:18'!
compositionRectangle
	| compRect |
	compRect := self innerBounds.
	margins ifNotNil: [compRect := compRect insetBy: margins].
	compRect width < 9 ifTrue: [compRect := compRect withWidth: 9].
	compRect height < 16 ifTrue: [compRect := compRect withHeight: 16].
	^ compRect! !

!TextMorph methodsFor: 'geometry' stamp: 'mt 11/20/2015 11:35'!
container
	"Return the container for composing this text.  There are four cases:
	1.  container is specified as, eg, an arbitrary shape,
	2.  container is specified as the bound rectangle, because
		this morph is linked to others,
	3.  container is nil, and wrap is true -- grow downward as necessary,
	4.  container is nil, and wrap is false -- grow in 2D as necessary."

	container ifNil:
		[successor ifNotNil: [^ self compositionRectangle].
		wrapFlag ifTrue: [^ self compositionRectangle withHeight: self maximumContainerExtent y].
		^ self compositionRectangle topLeft extent: self maximumContainerExtent].
	^ container! !

!TextMorph methodsFor: 'geometry testing' stamp: 'di 7/12/2001 22:15'!
containsPoint: aPoint
	(super containsPoint: aPoint) ifFalse: [^ false].  "Not in my bounds"
	container ifNil: [^ true].  "In bounds of simple text"
	self startingIndex > text size ifTrue:
		["make null text frame visible"
		^ super containsPoint: aPoint].
	"In complex text (non-rect container), test by line bounds"
	^ self paragraph containsPoint: aPoint
! !

!TextMorph methodsFor: 'accessing' stamp: ''!
contents

	^ text! !

!TextMorph methodsFor: 'accessing' stamp: 'tk 8/31/2000 14:59'!
contentsAsIs: stringOrText
	"Accept new text contents with line breaks only as in the text.
	Fit my width and height to the result."
	wrapFlag := false.
	container ifNotNil: [container fillsOwner ifTrue: [wrapFlag := true]].
	self newContents: stringOrText! !

!TextMorph methodsFor: 'accessing' stamp: 'di 9/30/97 09:51'!
contentsWrapped: stringOrText
	"Accept new text contents.  Lay it out, wrapping within my current width.
	Then fit my height to the result."
	wrapFlag := true.
	self newContents: stringOrText! !

!TextMorph methodsFor: 'accessing' stamp: 'di 9/30/97 15:48'!
contents: stringOrText
	^ self contentsAsIs: stringOrText! !

!TextMorph methodsFor: 'accessing' stamp: 'di 4/14/98 08:33'!
contents: stringOrText wrappedTo: width
	"Accept new text contents.  Lay it out, wrapping to width.
	Then fit my height to the result."
	self newContents: ''.
	wrapFlag := true.
	super extent: width truncated@self height.
	self newContents: stringOrText! !

!TextMorph methodsFor: 'objects from disk' stamp: 'di 7/30/2001 14:19'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	borderWidth ifNil:
		[borderWidth := 0.
		self removeProperty: #fillStyle].
	^ super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!TextMorph methodsFor: 'copying' stamp: 'di 11/12/97 09:31'!
copy
	^ super copy text: text copy textStyle: textStyle copy 
		wrap: wrapFlag color: color
		predecessor: nil successor: nil! !

!TextMorph methodsFor: 'accessing' stamp: 'ar 8/23/2001 21:23'!
crAction
	"Return the action to perform when encountering a CR in the input"
	^self valueOfProperty: #crAction! !

!TextMorph methodsFor: 'accessing' stamp: 'ar 8/23/2001 21:23'!
crAction: aMessageSend
	"Return the action to perform when encountering a CR in the input"
	^self setProperty: #crAction toValue: aMessageSend! !

!TextMorph methodsFor: 'private' stamp: 'mt 7/24/2016 14:11'!
createParagraph

	self setProperty: #CreatingParagraph toValue: true.

	[
		self setDefaultContentsIfNil.

		"...Code here to recreate the paragraph..."
		paragraph := (self paragraphClass new textOwner: self owner).
		paragraph wantsColumnBreaks: successor notNil.
		paragraph
			compose: text
			style: textStyle copy
			from: self startingIndex
			in: self container.
		wrapFlag ifFalse:
			["Was given huge container at first... now adjust"
			paragraph adjustRightX].
		paragraph focused: (self currentHand keyboardFocus == self).
	
		self fit.
	] ensure: [self removeProperty: #CreatingParagraph].

	^ paragraph! !

!TextMorph methodsFor: 'accessing' stamp: 'sw 5/22/2003 02:39'!
cursor
	"Answer the receiver's logical cursor position"

	| loc |
	loc := self valueOfProperty: #textCursorLocation  ifAbsentPut: [1].
	loc := loc min: text string size.
	^ loc rounded
	! !

!TextMorph methodsFor: 'accessing' stamp: 'sw 2/17/2003 18:20'!
cursorWrapped: aNumber
	"Set the cursor as indicated"

	self setProperty: #textCursorLocation toValue: (((aNumber rounded - 1) \\  text string size) + 1)

	! !

!TextMorph methodsFor: 'drawing' stamp: 'di 7/12/2001 10:45'!
debugDrawLineRectsOn: aCanvas
	"Shows where text line rectangles are"
	self paragraph lines do:
		[:line | aCanvas frameRectangle: line rectangle color: Color brown]
! !

!TextMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:31'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color black! !

!TextMorph methodsFor: 'geometry' stamp: 'kfr 12/2/2007 13:47'!
defaultLineHeight
	^ ( textStyle fontAt: textStyle defaultFontIndex) pointSize! !

!TextMorph methodsFor: 'submorphs-add/remove' stamp: 'di 11/7/97 10:00'!
delete
	predecessor ifNotNil: [predecessor setSuccessor: successor].
	successor ifNotNil: [successor setPredecessor: predecessor.
						successor recomposeChain].
	super delete! !

!TextMorph methodsFor: 'drawing' stamp: 'nk 1/1/2004 21:10'!
drawNullTextOn: aCanvas
	"make null text frame visible"

	aCanvas isPostscriptCanvas ifFalse: [
	aCanvas fillRectangle: bounds color: 
		((Color black) alpha: 0.1).
		]! !

!TextMorph methodsFor: 'drawing' stamp: 'sw 4/25/2002 00:52'!
drawOn: aCanvas
	"Draw the receiver on a canvas"

	| fauxBounds |
	self setDefaultContentsIfNil.
	super drawOn: aCanvas.  "Border and background if any"
	false ifTrue: [self debugDrawLineRectsOn: aCanvas].  "show line rects for debugging"
	(self startingIndex > text size)
		ifTrue: [self drawNullTextOn: aCanvas].
	"Hack here:  The canvas expects bounds to carry the location of the text, but we also need to communicate clipping."
	fauxBounds := self bounds topLeft corner: self innerBounds bottomRight.
	aCanvas paragraph: self paragraph bounds: fauxBounds color: color! !

!TextMorph methodsFor: 'multi level undo' stamp: 'mt 11/15/2015 13:38'!
editHistory
	^ self editor history! !

!TextMorph methodsFor: 'accessing' stamp: 'di 10/5/1998 13:56'!
editor
	"Return my current editor, or install a new one."
	editor ifNotNil: [^ editor].
	^ self installEditorToReplace: nil! !

!TextMorph methodsFor: 'private' stamp: 'ar 8/6/2009 16:05'!
editorClass
	"Answer the class used to create the receiver's editor"
	^DefaultEditorClass! !

!TextMorph methodsFor: 'accessing' stamp: 'sw 2/18/2003 02:58'!
elementCount
	"Answer how many sub-objects are within me"

	^ self text string size ! !

!TextMorph methodsFor: 'editing' stamp: 'ar 9/26/2001 22:45'!
enterClickableRegion: evt
	| index isLink |
	evt hand hasSubmorphs ifTrue:[^self].
	evt hand temporaryCursor ifNotNil:[^self].
	paragraph ifNotNil:[
		index := (paragraph characterBlockAtPoint: evt position) stringIndex.
		isLink := (paragraph text attributesAt: index forStyle: paragraph textStyle) 
					anySatisfy:[:attr| attr mayActOnClick].
		isLink ifTrue:[Cursor webLink show] ifFalse:[Cursor normal show].
	].
! !

!TextMorph methodsFor: 'geometry' stamp: 'nk 7/11/2004 20:07'!
extent: aPoint 
	| newExtent priorEditor |
	bounds extent = aPoint ifTrue: [^ self].
	priorEditor := editor.
	self isAutoFit
		ifTrue: [wrapFlag ifFalse: [^ self].  "full autofit can't change"
				newExtent := aPoint truncated max: self minimumExtent.
				newExtent x = self extent x ifTrue: [^ self].  "No change of wrap width"
				self releaseParagraphReally.  "invalidate the paragraph cache"
				super extent: newExtent.
				priorEditor
					ifNil: [self fit]  "since the width has changed..."
					ifNotNil: [self installEditorToReplace: priorEditor]]
		ifFalse: [super extent: (aPoint truncated max: self minimumExtent).
				wrapFlag ifFalse: [^ self].  "no effect on composition"
				self composeToBounds]
! !

!TextMorph methodsFor: 'containment' stamp: 'nk 8/18/2004 11:37'!
fillingOnOff
	"Establish a container for this text, with opposite filling status"
	self fillsOwner: (self fillsOwner not)! !

!TextMorph methodsFor: 'visual properties' stamp: 'dgd 2/16/2003 20:03'!
fillStyle
	"Return the current fillStyle of the receiver."
	^ self
		valueOfProperty: #fillStyle
		ifAbsent: [backgroundColor
				ifNil: [Color transparent]]! !

!TextMorph methodsFor: 'visual properties' stamp: 'di 6/22/2001 09:52'!
fillStyle: aFillStyle
	"Set the current fillStyle of the receiver."
	self setProperty: #fillStyle toValue: aFillStyle.
	"Workaround for Morphs not yet converted"
	backgroundColor := aFillStyle asColor.
	self changed.! !

!TextMorph methodsFor: 'containment' stamp: 'nk 8/18/2004 11:21'!
fillsOwner
	"Answer true if I fill my owner's shape."
	^container notNil and: [container fillsOwner]! !

!TextMorph methodsFor: 'containment' stamp: 'nk 8/18/2004 11:38'!
fillsOwner: aBoolean 
	self fillsOwner == aBoolean
		ifTrue: [^ self].
	self
		setContainer: (aBoolean
				ifTrue: [wrapFlag := true.
					container
						ifNil: [TextContainer new for: self minWidth: textStyle lineGrid * 2]
						ifNotNil: [container fillsOwner: true]]
				ifFalse: [self avoidsOcclusions
						ifFalse: [ nil ]
						ifTrue: [container fillsOwner: false]])! !

!TextMorph methodsFor: 'linked frames' stamp: 'di 11/8/97 15:51'!
firstCharacterIndex
	^ self paragraph firstCharacterIndex! !

!TextMorph methodsFor: 'linked frames' stamp: 'dgd 2/21/2003 22:26'!
firstInChain
	"Return the first morph in a chain of textMorphs"

	| first |
	first := self.
	[first predecessor isNil] whileFalse: [first := first predecessor].
	^first! !

!TextMorph methodsFor: 'private' stamp: 'yo 1/30/2008 16:27'!
fit
	"Adjust my bounds to fit the text.  Should be a no-op if autoFit is not specified.
	Required after the text changes,
	or if wrapFlag is true and the user attempts to change the extent."

	| newExtent para cBounds lastOfLines heightOfLast wid |
	self isAutoFit 
		ifTrue: 
			[wid _ (text notNil and: [text size > 2]) ifTrue: [5] ifFalse: [40].
			newExtent := (self paragraph extent max: wid @ ( self defaultLineHeight)) + (0 @ 2).
			newExtent := newExtent + (2 * borderWidth).
			margins 
				ifNotNil: [newExtent := ((0 @ 0 extent: newExtent) expandBy: margins) extent].
			newExtent ~= bounds extent 
				ifTrue: 
					[(container isNil and: [successor isNil]) 
						ifTrue: 
							[para := paragraph.	"Save para (layoutChanged smashes it)"
							super extent: newExtent.
							paragraph := para]].
			container notNil & successor isNil 
				ifTrue: 
					[cBounds := container bounds truncated.
					"23 sept 2000 - try to allow vertical growth"
					lastOfLines := self paragraph lines last.
					heightOfLast := lastOfLines bottom - lastOfLines top.
					(lastOfLines last < text size 
						and: [lastOfLines bottom + heightOfLast >= self bottom]) 
							ifTrue: 
								[container releaseCachedState.
								cBounds := cBounds origin corner: cBounds corner + (0 @ heightOfLast)].
					self privateBounds: cBounds]].

	"These statements should be pushed back into senders"
	self paragraph positionWhenComposed: self position.
	successor ifNotNil: [successor predecessorChanged].
	self changed	"Too conservative: only paragraph composition
					should cause invalidation."! !

!TextMorph methodsFor: 'objects from disk' stamp: 'pre 7/21/2016 07:50'!
fixUponLoad: aProject seg: anImageSegment
	"We are in an old project that is being loaded from disk. Fix up conventions that have changed."

	| substituteFont |
	substituteFont := aProject projectParameters at: #substitutedFont ifAbsent: [#none].
	(substituteFont ~~ #none and: [self textStyle fontArray includes: substituteFont])
			ifTrue: [ self fit ].

	^ super fixUponLoad: aProject seg: anImageSegment! !

!TextMorph methodsFor: 'menu' stamp: 'di 10/8/1998 23:40'!
followCurve
	self setContainer: (TextOnCurveContainer new baseline: 0; textDirection: 1).
	self changed! !

!TextMorph methodsFor: 'accessing' stamp: 'nk 8/30/2004 05:43'!
fontName: fontName pointSize: fontSize
	| newTextStyle |
	newTextStyle := ((TextStyle named: fontName asSymbol) ifNil: [ TextStyle default ]) copy.
	newTextStyle ifNil: [self error: 'font ', fontName, ' not found.'].

	textStyle := newTextStyle.
	text addAttribute: (TextFontChange fontNumber: (newTextStyle fontIndexOfPointSize: fontSize)).
	paragraph ifNotNil: [paragraph textStyle: newTextStyle]! !

!TextMorph methodsFor: 'accessing' stamp: 'nk 7/12/2003 08:39'!
fontName: fontName size: fontSize
	| newTextStyle |
	newTextStyle := ((TextStyle named: fontName asSymbol) ifNil: [ TextStyle default ]) copy.
	textStyle := newTextStyle.
	text addAttribute: (TextFontChange fontNumber: (newTextStyle fontIndexOfSize: fontSize)).
	paragraph ifNotNil: [paragraph textStyle: newTextStyle]! !

!TextMorph methodsFor: 'accessing' stamp: 'nk 10/16/2003 16:42'!
font: aFont
	| newTextStyle |
	newTextStyle := aFont textStyle copy ifNil: [ TextStyle fontArray: { aFont } ].
	textStyle := newTextStyle.
	text addAttribute: (TextFontChange fontNumber: (newTextStyle fontIndexOf: aFont)).
	paragraph ifNotNil: [paragraph textStyle: newTextStyle]! !

!TextMorph methodsFor: 'scripting access' stamp: 'pre 7/21/2016 07:51'!
getAllButFirstCharacter
	"Obtain all but the first character from the receiver; if that would be empty, return a black dot"

	| aString |
	^ (aString := text string) size > 1 
		ifTrue: [aString copyFrom: 2 to: aString size] 
		ifFalse: ['·']! !

!TextMorph methodsFor: 'accessing' stamp: 'kfr 9/21/2003 21:47'!
getFirstCharacter
	"obtain the first character from the receiver if it is empty, return a  
	black dot"
	| aString |
	^ (aString := text string) isEmpty
		ifTrue: ['·']
		ifFalse: [aString first asString] ! !

!TextMorph methodsFor: 'accessing' stamp: 'pre 7/21/2016 07:51'!
getLastCharacter
	"obtain the last character from the receiver if it is empty, return a black dot"

	| aString |
	^ (aString := text string) size > 0 
		ifTrue: [aString last asString] 
		ifFalse: ['·']! !

!TextMorph methodsFor: 'event handling' stamp: 'pre 7/21/2016 07:51'!
getMenu: shiftKeyState 
	^ (shiftKeyState not or: [Preferences noviceMode])
		ifTrue: [TextEditor yellowButtonMenu]
		ifFalse: [TextEditor shiftedYellowButtonMenu]! !

!TextMorph methodsFor: 'submorphs-add/remove' stamp: 'di 11/16/97 16:52'!
goBehind
	"We need to save the container, as it knows about fill and run-around"
	| cont |
	container ifNil: [^ super goBehind].
	self releaseParagraph.  "Cause recomposition"
	cont := container.  "Save the container"
	super goBehind.  "This will change owner, nilling the container"
	container := cont.  "Restore the container"
	self changed! !

!TextMorph methodsFor: 'editing' stamp: 'nice 1/10/2010 19:25'!
handleEdit: editBlock
	"Ensure that changed areas get suitably redrawn"
	| result |
	self selectionChanged.  "Note old selection"
	result := editBlock value.
	self selectionChanged.  "Note new selection"
	self updateFromParagraph.  "Propagate changes as necessary"
	^result! !

!TextMorph methodsFor: 'editing' stamp: 'mt 4/1/2015 16:29'!
handleInteraction: interactionBlock fromEvent: evt
	"Perform the changes in interactionBlock, noting any change in selection
	and possibly a change in the size of the paragraph (ar 9/22/2001 - added for TextPrintIts)"
	| oldEditor oldParagraph oldText |
	oldEditor := editor.
	oldParagraph := paragraph.
	oldText := oldParagraph text copy.

	self selectionChanged.  "Note old selection"

		interactionBlock value.

	(oldParagraph == paragraph) ifTrue:[
		"this will not work if the paragraph changed"
		editor := oldEditor.     "since it may have been changed while in block"
	].
	self selectionChanged.  "Note new selection"
	(oldText = paragraph text and: [ oldText runs = paragraph text runs ])
		ifFalse:[ 
			self paragraph composeAll.
			self updateFromParagraph ].
	self setCompositionWindow.! !

!TextMorph methodsFor: 'events-processing' stamp: 'mt 6/4/2016 16:46'!
handleKeystroke: anEvent
	"Overwritten to support tab-among-fields preference."

	| pasteUp |
	anEvent wasHandled ifTrue:[^self].
	(self handlesKeyboard: anEvent) ifFalse: [^ self].

	anEvent keyCharacter = Character tab ifTrue: [
		"Allow passing through text morph inside pasteups"
		(self wouldAcceptKeyboardFocusUponTab
			and: [(pasteUp := self pasteUpMorphHandlingTabAmongFields) notNil])
				ifTrue: [
					anEvent wasHandled: true.
					^ pasteUp tabHitWithEvent: anEvent]].
	
	^ super handleKeystroke: anEvent! !

!TextMorph methodsFor: 'events-processing' stamp: 'ar 9/26/2001 22:21'!
handleMouseMove: anEvent
	"Re-implemented to allow for mouse-up move events"
	anEvent wasHandled ifTrue:[^self]. "not interested"
	(anEvent hand hasSubmorphs) ifTrue:[^self].
	anEvent wasHandled: true.
	self mouseMove: anEvent.
	(anEvent anyButtonPressed and:[anEvent hand mouseFocus == self]) ifFalse:[^self].
	(self handlesMouseStillDown: anEvent) ifTrue:[
		"Step at the new location"
		self startStepping: #handleMouseStillDown: 
			at: Time millisecondClockValue
			arguments: {anEvent copy resetHandlerFields}
			stepTime: 1].
! !

!TextMorph methodsFor: 'event handling' stamp: 'ar 10/4/2000 19:09'!
handlesKeyboard: evt
	^true! !

!TextMorph methodsFor: 'event handling' stamp: 'di 7/21/2001 09:44'!
handlesMouseDown: evt
	self isPartsDonor ifTrue: [^ false].
	^ self innerBounds containsPoint: evt cursorPoint! !

!TextMorph methodsFor: 'event handling' stamp: 'dgd 2/22/2003 19:04'!
hasFocus
	^editor notNil! !

!TextMorph methodsFor: 'accessing' stamp: 'pre 7/21/2016 07:53'!
hasTranslucentColor
	"Overridden from BorderedMorph to test backgroundColor instead of (text) color."

	^ backgroundColor isNil
		or: [backgroundColor isColor and: [backgroundColor isTranslucentColor]]
		or: [borderColor isColor and: [borderColor isTranslucentColor]]! !

!TextMorph methodsFor: 'editing' stamp: 'di 4/21/1998 13:22'!
hasUnacceptedEdits: aBoolean
	"Ignored here, but noted in TextMorphForEditView"
! !

!TextMorph methodsFor: 'menu' stamp: 'sw 2/18/2003 03:20'!
holderForCharacters
	"Hand the user a Holder that is populated with individual text morphs representing my characters"

	| aHolder |
	aHolder := ScriptingSystem prototypicalHolder.
	aHolder setNameTo: 'H', self externalName.
	text string do:
		[:aChar |
			aHolder addMorphBack: (TextMorph new contents: aChar asText)].
	aHolder setProperty: #donorTextMorph toValue: self.
	aHolder fullBounds.
	aHolder openInHand! !

!TextMorph methodsFor: 'initialization' stamp: 'di 7/27/2001 11:56'!
initialize
	super initialize.
	borderWidth := 0.
	textStyle := TextStyle default copy.
	wrapFlag := true.
! !

!TextMorph methodsFor: 'scripting access' stamp: 'bf 5/16/2011 18:35'!
insertCharacters: aString
	"Insert the characters from the given source at my current cursor position"

	| aLoc aText attributes |
	aLoc := self cursor max: 1.
	aText := aLoc > text size
		ifTrue: [aString asText]
		ifFalse: [
			attributes := (text attributesAt: aLoc)
				select: [:attr | attr mayBeExtended].
			Text string: aString attributes: attributes].
	paragraph replaceFrom: aLoc to: (aLoc - 1) with: aText displaying: true.
	self updateFromParagraph  ! !

!TextMorph methodsFor: 'scripting access' stamp: 'sw 2/18/2003 02:46'!
insertContentsOf: aPlayer
	"Insert the characters from the given player at my current cursor position"

	| aLoc |
	aLoc := self cursor.
	paragraph replaceFrom: aLoc to: (aLoc - 1) with: aPlayer getStringContents displaying: true.
	self updateFromParagraph  ! !

!TextMorph methodsFor: 'private' stamp: 'mt 11/7/2015 23:29'!
installEditorToReplace: priorEditor
	"Install an editor for my paragraph.  This constitutes 'hasFocus'.
	If priorEditor is not nil, then initialize the new editor from its state.
	We may want to rework this so it actually uses the prior editor."

	| stateArray |
	priorEditor ifNotNil: [stateArray := priorEditor stateArray].
	editor := self editorClass new morph: self.
	editor changeParagraph: self paragraph.
	priorEditor
		ifNil: [editor history: TextEditorCommandHistory new]
		ifNotNil: [editor stateArrayPut: stateArray].
	self selectionChanged.
	^ editor! !

!TextMorph methodsFor: 'accessing' stamp: 'mt 4/27/2015 15:32'!
isAutoFit
	"Migrating old instances. The #isNil check may be removed in the future."
	^ autoFit isNil or: [autoFit]! !

!TextMorph methodsFor: 'linked frames' stamp: 'di 11/16/97 15:15'!
isLinkedTo: aMorph
	self firstInChain withSuccessorsDo:
		[:m | m == aMorph ifTrue: [^ true]].
	^ false! !

!TextMorph methodsFor: 'classification' stamp: 'ar 12/17/2001 12:46'!
isTextMorph
	^true! !

!TextMorph methodsFor: 'accessing' stamp: 'RAA 8/21/2001 11:18'!
isWrapped
	
	^wrapFlag! !

!TextMorph methodsFor: 'alignment' stamp: 'di 10/25/97 19:20'!
justified 
	self paragraph justified.
	self updateFromParagraph! !

!TextMorph methodsFor: 'event handling' stamp: 'mt 6/5/2016 09:26'!
keyboardFocusChange: weHaveFocus 

	paragraph ifNotNil: [:para | para focused: weHaveFocus].

	weHaveFocus
		ifTrue: [ "A hand is wanting to send us characters..."
			self hasFocus ifFalse: [self editor	"Forces install"].
			Editor blinkingCursor
				ifTrue: [self startBlinking]
				ifFalse: [self resetBlinkCursor "ensure caret visible"]]
		ifFalse: [ "A hand has clicked elsewhere..."
			self world ifNotNil: [:w |
				self flag: #suspicious. "mt: Why should this be called with 'false' if a hand holds on the focus? Do hands not talk to each other before sending this call? They should..."
				w handsDo: [:h | h keyboardFocus == self ifTrue: [^self]].
				"Release control unless some hand is still holding on"
				self releaseEditor].
			self stopBlinking].
! !

!TextMorph methodsFor: 'event handling' stamp: 'nice 8/5/2011 19:42'!
keyStroke: evt
	"Handle a keystroke event."
	
	self resetBlinkCursor. "don't blink during type-in"
	ToolSet 
		codeCompletionAround: [
			evt keyValue = 13 ifTrue: [ "CR - check for special action"
				self crAction ifNotNil: [ :action |
					"Note: Code below assumes that this was some
					input field reacting on CR. Break the keyboard
					focus so that the receiver can be safely deleted."
					evt hand newKeyboardFocus: nil.
					^action valueWithEnoughArguments: { evt } ] ].
			self handleInteraction: [ self editor keyStroke: evt ] fromEvent: evt.
			"self updateFromParagraph."
			super keyStroke: evt  "sends to keyStroke event handler, if any" ]
		textMorph: self
		keyStroke: evt! !

!TextMorph methodsFor: 'linked frames' stamp: 'jm 10/28/97 18:31'!
lastCharacterIndex
	^ self paragraph lastCharacterIndex! !

!TextMorph methodsFor: 'alignment' stamp: 'di 10/25/97 19:20'!
leftFlush 
	self paragraph leftFlush.
	self updateFromParagraph! !

!TextMorph methodsFor: 'caching' stamp: 'di 11/13/97 15:17'!
loadCachedState
	"Prepare for fast response -- next page of a book?"
	self paragraph! !

!TextMorph methodsFor: 'accessing' stamp: 'RAA 8/21/2001 11:41'!
margins

	^margins! !

!TextMorph methodsFor: 'accessing' stamp: 'di 7/20/2001 22:44'!
margins: newMargins
	"newMargins can be a number, point or rectangle, as allowed by, eg, insetBy:."

	margins := newMargins.
	self composeToBounds! !

!TextMorph methodsFor: 'geometry' stamp: 'mt 11/20/2015 11:34'!
maximumContainerExtent
	"For text composition. Returns the maximum area for text to be composed."
	
	^ 9999999@9999999! !

!TextMorph methodsFor: 'layout' stamp: 'mt 4/24/2015 09:36'!
minHeight

	| result |
	textStyle ifNil: [^ 16].
	borderWidth ifNil: [^ 16].

	result := (textStyle lineGrid + 2) + (borderWidth*2).
	margins ifNil: [^ result].
	
	^ margins isRectangle
		ifTrue: [result + margins top + margins bottom]
		ifFalse: [margins isPoint
			ifTrue: [result + margins y + margins y]
			ifFalse: [result + (2*margins)]]! !

!TextMorph methodsFor: 'layout' stamp: 'mt 4/24/2015 09:37'!
minWidth

	| result |
	textStyle ifNil: [^ 9].
	borderWidth ifNil: [^ 9].

	result := 9 + (borderWidth*2).
	margins ifNil: [^ result].
	
	^ margins isRectangle
		ifTrue: [result + margins left + margins right]
		ifFalse: [margins isPoint
			ifTrue: [result + margins x + margins x]
			ifFalse: [result + (2*margins)]]! !

!TextMorph methodsFor: 'event handling' stamp: 'mt 8/3/2016 14:50'!
mouseDown: evt 
	"Make this TextMorph be the keyboard input focus, if it isn't  
	already, and repond to the text selection gesture."
	
	evt yellowButtonPressed ifTrue: [
		"First check for option (menu) click"
		^ self yellowButtonActivity: evt shiftPressed].

	"Show the caret immediately on mouse down to give user feedback."
	self resetBlinkCursor.

	"If focus does not follow the mouse cursor and we click below everything, just grab the focus to not destroy the selection."
	((self hasKeyboardFocus: evt hand) not and: [(self bounds containsPoint: evt position) not])
		ifTrue: [evt hand newKeyboardFocus: self]
		ifFalse: [
			evt hand newKeyboardFocus: self.
			self
				handleInteraction: [editor mouseDown: evt]
				fromEvent: evt].! !

!TextMorph methodsFor: 'event handling' stamp: 'nice 8/3/2011 18:26'!
mouseMove: evt
	evt redButtonPressed ifFalse: [^ self enterClickableRegion: evt].
	self handleInteraction: [self editor mouseMove: evt] fromEvent: evt! !

!TextMorph methodsFor: 'event handling' stamp: 'di 12/3/97 20:04'!
mouseUp: evt
	self handleInteraction: [editor mouseUp: evt] fromEvent: evt! !

!TextMorph methodsFor: 'accessing' stamp: 'pre 7/21/2016 07:44'!
newContents: stringOrText 
	"Accept new text contents."
	| newText embeddedMorphs oldSelection |
	newText := stringOrText isString 
		ifTrue: [Text fromString: stringOrText copy ]
		ifFalse: [ stringOrText copy asText.	"should be veryDeepCopy?" ].

	(text = newText and: [text runs = newText runs]) ifTrue: [^ self].	"No substantive change"
	text ifNotNil: [(embeddedMorphs := text embeddedMorphs)
			ifNotNil: 
				[self removeAllMorphsIn: embeddedMorphs.
				embeddedMorphs do: [:m | m delete]]].

	oldSelection := editor ifNotNil: [:ed | ed selectionInterval].
	text := newText.

	"add all morphs off the visible region; they'll be moved into the right 
	place when they become visible. (this can make the scrollable area too 
	large, though)"
	newText embeddedMorphs do: 
		[:m | 
		self addMorph: m.
		m position: -1000 @ 0].
	self releaseParagraph.
	"update the paragraph cache"
	self paragraph.
	oldSelection ifNotNil: [:sel | self selectFrom: sel first to: sel last].
	"re-instantiate to set bounds"
	self world ifNotNil: [self world startSteppingSubmorphsOf: self]! !

!TextMorph methodsFor: 'containment' stamp: 'di 11/4/97 15:37'!
occlusionsOnOff
	"Establish a container for this text, with opposite occlusion avoidance status"
	self setContainer:
	(container
	ifNil: [(TextContainer new for: self minWidth: textStyle lineGrid*2)
							fillsOwner: false; avoidsOcclusions: true]
	ifNotNil: [(container avoidsOcclusions and: [container fillsOwner not])
			ifTrue: [nil  "Return to simple rectangular bounds"]
			ifFalse: [container avoidsOcclusions: container avoidsOcclusions not]])! !

!TextMorph methodsFor: 'blinking' stamp: 'mt 6/5/2016 09:22'!
onBlinkCursor

	| para |
	para := paragraph ifNil: [^nil].
	Time millisecondClockValue < self blinkStart 
		ifTrue: [para showCaret: para focused]
		ifFalse: [para showCaret: para showCaret not].
	para caretRect ifNotNil: [ :r | self invalidRect: r].! !

!TextMorph methodsFor: 'change reporting' stamp: 'di 10/18/2004 13:50'!
ownerChanged
	| priorEditor |
	super ownerChanged.
	container ifNotNil: 
			[editor isNil 
				ifTrue:
					[self releaseParagraph.
					(container isKindOf: TextContainer) ifTrue:
						["May need to recompose due to changes in owner"
						self installEditorToReplace: nil.
						self releaseParagraph]]
				ifFalse: 
					[priorEditor := editor.
					self releaseParagraph.
					self installEditorToReplace: priorEditor]]! !

!TextMorph methodsFor: 'private' stamp: 'mt 7/24/2016 14:12'!
paragraph
	"Paragraph instantiation is lazy -- create it only when needed"

	^ paragraph ifNil: [self createParagraph]! !

!TextMorph methodsFor: 'private' stamp: 'jm 11/19/97 19:30'!
paragraphClass
	container ifNil: [^ NewParagraph].
	^ container paragraphClass! !

!TextMorph methodsFor: 'editing' stamp: 'dgd 2/21/2003 22:29'!
passKeyboardFocusTo: otherMorph 
	| w |
	self flag: #arNote.	"Do we need this?!!"
	(w := self world) isNil 
		ifFalse: 
			[w 
				handsDo: [:h | h keyboardFocus == self ifTrue: [h newKeyboardFocus: otherMorph]]]! !

!TextMorph methodsFor: 'linked frames' stamp: 'jm 10/28/97 18:31'!
predecessor
	^ predecessor! !

!TextMorph methodsFor: 'private' stamp: 'pre 7/21/2016 07:55'!
predecessorChanged
	| newStart oldStart |
	(self hasProperty: #CreatingParagraph) ifTrue: [^self].
	newStart := predecessor isNil 
				ifTrue: [1]
				ifFalse: [predecessor lastCharacterIndex + 1].
	(self paragraph adjustedFirstCharacterIndex ~= newStart 
		or: [newStart >= text size]) 
			ifTrue: 
				[paragraph composeAllStartingAt: newStart.
				self fit]
			ifFalse: 
				["If the offset to end of text has not changed, just slide"
				oldStart := self firstCharacterIndex.
				self withSuccessorsDo: [:m | m adjustLineIndicesBy: newStart - oldStart]]! !

!TextMorph methodsFor: 'private' stamp: 'di 7/28/2001 10:33'!
predecessor: pred successor: succ
	"Private -- for use only in morphic duplication"
	predecessor := pred.
	successor := succ.
! !

!TextMorph methodsFor: 'editing' stamp: 'tpr 6/6/2016 17:42'!
preferredKeyboardPosition
	| default rects |
	default := (self bounds: self bounds in: World) topLeft.
	paragraph
		ifNil: [^ default].
	rects := paragraph selectionRects.
	rects size = 0
		ifTrue: [^ default].
	^ rects first bottomLeft! !

!TextMorph methodsFor: 'geometry' stamp: 'jmv 3/2/2010 16:42'!
privateMoveBy: delta 
	super privateMoveBy: delta.
	editor 
		ifNil: [ paragraph ifNotNil: [paragraph moveBy: delta]]
		ifNotNil: [ 
			"When moving text with an active editor, save and restore all state."
			paragraph moveBy: delta.
			self installEditorToReplace: editor]! !

!TextMorph methodsFor: 'private' stamp: 'tk 9/28/1999 16:50'!
privateOwner: newOwner
	"Nil the container when text gets extracted"
	super privateOwner: newOwner.
	container ifNotNil: [
		newOwner ifNotNil: [
			newOwner isWorldOrHandMorph ifTrue: [self setContainer: nil]]]! !

!TextMorph methodsFor: 'accessing' stamp: 'mt 4/19/2015 13:37'!
readOnly

	^ readOnly ifNil: [false]! !

!TextMorph methodsFor: 'accessing' stamp: 'mt 4/19/2015 13:37'!
readOnly: aBoolean

	readOnly := aBoolean.! !

!TextMorph methodsFor: 'containment' stamp: 'pre 7/21/2016 07:55'!
recognizerArena
	"Answer the rectangular area, in world coordinates, that the character recognizer should regard as its tablet"

	| outer |
	^ (outer := self ownerThatIsA: PluggableTextMorph)
		ifNotNil: [outer boundsInWorld]
		ifNil: [self boundsInWorld]! !

!TextMorph methodsFor: 'linked frames' stamp: 'di 11/12/97 09:10'!
recomposeChain
	"Recompose this textMorph and all that follow it."
	self withSuccessorsDo:
		[:m |  m text: text textStyle: textStyle;  "Propagate new style if any"
				releaseParagraph;  "Force recomposition"
				fit  "and propagate the change"]! !

!TextMorph methodsFor: 'caching' stamp: 'BG 11/20/2004 12:36'!
releaseCachedState

	super releaseCachedState.
	self releaseParagraph; paragraph.
! !

!TextMorph methodsFor: 'private' stamp: 'di 10/5/1998 16:39'!
releaseEditor 
	"Release the editor for my paragraph.  This morph no longer 'hasFocus'."
	editor ifNotNil:
		[self selectionChanged.
		self paragraph selectionStart: nil selectionStop: nil.
		editor := nil].! !

!TextMorph methodsFor: 'private' stamp: 'RAA 12/5/2001 11:20'!
releaseParagraph

	"a slight kludge so subclasses can have a bit more control over whether the paragraph really 
	gets released. important for GeeMail since the selection needs to be accessible even if the 
	hand is outside me"

	self releaseParagraphReally.
! !

!TextMorph methodsFor: 'private' stamp: 'sw 9/14/2007 11:01'!
releaseParagraphReally

	"a slight kludge so subclasses can have a bit more control over whether the paragraph really 
	gets released. important for GeeMail since the selection needs to be accessible even if the 
	hand is outside me"

	"Paragraph instantiation is lazy -- it will be created only when needed"
	self releaseEditor.
	paragraph ifNotNil:
		[paragraph _ nil].
	container ifNotNil:
		[container isMorph ifTrue: [container releaseCachedState]]! !

!TextMorph methodsFor: 'private' stamp: 'ar 8/10/2003 18:12'!
removedMorph: aMorph
	| range |
	range := text find: (TextAnchor new anchoredMorph: aMorph).
	range ifNotNil:
		[self paragraph replaceFrom: range first to: range last
				with: Text new displaying: false.
		self fit].
	aMorph textAnchorType: nil.
	aMorph relativeTextAnchorPosition: nil.
	super removedMorph: aMorph.! !

!TextMorph methodsFor: 'blinking' stamp: 'pre 7/21/2016 07:56'!
resetBlinkCursor
	"Reset the blinking cursor"
	| para |
	self blinkStart: Time millisecondClockValue + 500.
	para := paragraph ifNil: [^self].
	para showCaret = para focused ifFalse:[
		para caretRect ifNotNil: [ :r | self invalidRect: r].
		para showCaret: para focused.].
! !

!TextMorph methodsFor: 'menu' stamp: 'di 12/3/97 09:40'!
reverseCurveDirection
	container textDirection: container textDirection negated.
	self paragraph composeAll! !

!TextMorph methodsFor: 'alignment' stamp: 'di 10/25/97 19:20'!
rightFlush 
	self paragraph rightFlush.
	self updateFromParagraph! !

!TextMorph methodsFor: 'accessing' stamp: 'mt 5/3/2015 19:43'!
selectAll

	self selectionChanged.
	self editor selectAll.
	self selectionChanged.! !

!TextMorph methodsFor: 'accessing' stamp: 'nk 2/26/2001 12:45'!
selectFrom: a to: b
	self editor selectFrom: a to: b! !

!TextMorph methodsFor: 'accessing' stamp: 'nk 2/26/2001 12:42'!
selection
	^editor ifNotNil: [ editor selection ]! !

!TextMorph methodsFor: 'private' stamp: 'nk 8/29/2004 21:40'!
selectionChanged
	"Invalidate all the selection rectangles. 
	Make sure that any drop shadow is accounted for too."
	self paragraph selectionRects
		do: [:r | self
				invalidRect: (self expandFullBoundsForDropShadow: (r intersect: self fullBounds))]! !

!TextMorph methodsFor: 'scripting access' stamp: 'tk 4/23/2007 10:51'!
setAllButFirstCharacter: source 
	"Set all but the first char of the receiver to the source"
	| chars |
	(chars _ self getCharacters) isEmpty
		ifTrue: [self newContents: '·' , source asString]
		ifFalse: [self newContents: (String
								streamContents: [:aStream | 
									aStream nextPut: chars first.
									aStream nextPutAll: source])]! !

!TextMorph methodsFor: 'accessing' stamp: 'sw 9/1/2000 10:43'!
setCharacters: chars
	"obtain a string value from the receiver"

	(self getCharacters = chars) ifFalse:
		[self newContents: chars]! !

!TextMorph methodsFor: 'editing' stamp: 'yo 11/7/2002 19:11'!
setCompositionWindow

	| hand |
	hand := self primaryHand.
	hand ifNotNil: [hand compositionWindowManager keyboardFocusForAMorph: self].
! !

!TextMorph methodsFor: 'containment' stamp: 'di 11/12/97 09:06'!
setContainer: newContainer
	"Adopt (or abandon) container shape"
	self changed.
	container := newContainer.
	self releaseParagraph! !

!TextMorph methodsFor: 'menu' stamp: 'di 12/3/97 10:25'!
setCurveBaseline: evt
	| handle origin |
	origin := evt cursorPoint.
	handle := HandleMorph new forEachPointDo:
		[:newPoint | handle removeAllMorphs.
		handle addMorph:
			(PolygonMorph vertices: (Array with: origin with: newPoint)
				color: Color black borderWidth: 1 borderColor: Color black).
		container baseline: (newPoint - origin) y negated asInteger // 5.
		self paragraph composeAll].
	evt hand attachMorph: handle.
	handle startStepping	! !

!TextMorph methodsFor: 'private' stamp: 'tk 11/13/2001 01:57'!
setDefaultContentsIfNil
	"Set the default contents"

	| toUse |
	text ifNil:
		[toUse := self valueOfProperty: #defaultContents.
		toUse ifNil: [toUse :='abc' asText "allBold"].	"try it plain for a while"
		text := toUse]! !

!TextMorph methodsFor: 'accessing' stamp: 'dgd 2/21/2003 22:32'!
setFirstCharacter: source 
	"Set the first character of the receiver as indicated"
	| aChar chars |
	aChar := source asCharacter.
	(chars := self getCharacters) isEmpty
		ifTrue: [self
				newContents: (String with: aChar)]
		ifFalse: [chars first = aChar
				ifFalse: [self
						newContents: (String
								streamContents: [:aStream | 
									aStream nextPut: aChar.
									aStream
										nextPutAll: (chars copyFrom: 2 to: chars size)])]] ! !

!TextMorph methodsFor: 'accessing' stamp: 'sw 8/10/2004 00:56'!
setLastCharacter: source
	"Set the last character of the receiver as indicated"

	| aChar chars |
	aChar := source asCharacter.
	(chars := self getCharacters) size > 0 
		ifFalse:
			[self newContents: (String with: aChar)]
		ifTrue:
			[(chars last = aChar) ifFalse:
				[self newContents: (String streamContents:
					[:aStream |
						aStream nextPutAll: (chars copyFrom: 1 to: (chars size - 1)).
						aStream nextPut: aChar])]]! !

!TextMorph methodsFor: 'private' stamp: 'di 10/25/97 17:11'!
setPredecessor: newPredecessor
	predecessor := newPredecessor! !

!TextMorph methodsFor: 'private' stamp: 'RAA 5/6/2001 15:12'!
setSuccessor: newSuccessor

	successor := newSuccessor.
	paragraph ifNotNil: [paragraph wantsColumnBreaks: successor notNil].
! !

!TextMorph methodsFor: 'initialization' stamp: 'di 11/17/2001 15:34'!
setTextStyle: aTextStyle

	textStyle := aTextStyle.
	self releaseCachedState; changed! !

!TextMorph methodsFor: 'blinking' stamp: 'ar 12/15/2008 22:49'!
startBlinking
	self startStepping: #onBlinkCursor 
		at: Time millisecondClockValue 
		arguments: nil stepTime: 500.
	self resetBlinkCursor.
! !

!TextMorph methodsFor: 'linked frames' stamp: 'dgd 2/21/2003 22:32'!
startingIndex
	predecessor isNil
		ifTrue: [^ 1].
	^ predecessor lastCharacterIndex + 1 ! !

!TextMorph methodsFor: 'blinking' stamp: 'mt 6/5/2016 09:14'!
stopBlinking
	self stopSteppingSelector: #onBlinkCursor.
	self resetBlinkCursor. "Ensure caret visible."! !

!TextMorph methodsFor: 'initialization' stamp: 'mir 8/2/1999 10:34'!
string: aString fontName: aName size: aSize

	self string: aString fontName: aName size: aSize wrap: true! !

!TextMorph methodsFor: 'initialization' stamp: 'mir 8/2/1999 10:35'!
string: aString fontName: aName size: aSize wrap: shouldWrap

	shouldWrap
		ifTrue: [self contentsWrapped: aString]
		ifFalse: [self contents: aString].
	self fontName: aName size: aSize! !

!TextMorph methodsFor: 'linked frames' stamp: 'jm 10/28/97 18:31'!
successor
	^ successor! !

!TextMorph methodsFor: 'accessing' stamp: 'tk 1/10/2001 13:52'!
text
	^ text! !

!TextMorph methodsFor: 'accessing' stamp: 'nk 7/3/2003 18:33'!
textAlignment
	"Answer 1..4, representing #leftFlush, #rightFlush, #centered, or #justified"
	^self editor textAlignment! !

!TextMorph methodsFor: 'accessing' stamp: 'nk 6/18/2003 14:28'!
textAlignmentSymbol
	"Answer one of #leftFlush, #rightFlush, #centered, or #justified"
	^self editor textAlignmentSymbol! !

!TextMorph methodsFor: 'geometry' stamp: 'di 3/1/98 11:40'!
textBounds
	^ bounds! !

!TextMorph methodsFor: 'accessing' stamp: 'di 7/21/2001 10:37'!
textColor

	^ color! !

!TextMorph methodsFor: 'accessing' stamp: 'yo 9/28/2006 14:20'!
textColor: aColor

	self editor selectFrom: 1 to: 0.
	self selectionColor: aColor.
! !

!TextMorph methodsFor: 'accessing' stamp: 'tk 9/1/2000 13:50'!
textStyle
	^textStyle! !

!TextMorph methodsFor: 'private' stamp: 'di 10/24/97 11:35'!
text: t textStyle: s
	"Private -- for use only in morphic duplication"
	text := t.
	textStyle := s.
	paragraph ifNotNil: [paragraph textStyle: s]! !

!TextMorph methodsFor: 'private' stamp: 'di 7/28/2001 10:34'!
text: t textStyle: s wrap: wrap color: c
	predecessor: pred successor: succ
	"Private -- for use only in morphic duplication"
	text := t.
	textStyle := s.
	wrapFlag := wrap.
	color := c.
	paragraph := editor := container := nil.
	self predecessor: pred successor: succ! !

!TextMorph methodsFor: 'private' stamp: 'mt 4/27/2015 15:33'!
updateFromParagraph
	"A change has taken place in my paragraph, as a result of editing and I must be updated.  If a line break causes recomposition of the current paragraph, or it the selection has entered a different paragraph, then the current editor will be released, and must be reinstalled with the resulting new paragraph, while retaining any editor state, such as selection, undo state, and current typing emphasis."

	| newStyle sel oldLast oldEditor back |
	paragraph ifNil: [^self].
	wrapFlag ifNil: [wrapFlag := true].
	editor ifNotNil: 
			[oldEditor := editor.
			sel := editor selectionInterval.
			editor storeSelectionInParagraph].
	text := paragraph text.
	paragraph textStyle = textStyle 
		ifTrue: [self fit]
		ifFalse: 
			["Broadcast style changes to all morphs"

			newStyle := paragraph textStyle.
			(self firstInChain text: text textStyle: newStyle) recomposeChain.
			editor ifNotNil: [self installEditorToReplace: editor]].
	
	(self isAutoFit and: [self isWrapped not])
		ifTrue: [self extent: self paragraph extent; composeToBounds]
		ifFalse: [super layoutChanged].
	sel ifNil: [^self].

	"If selection is in top line, then recompose predecessor for possible ripple-back"
	predecessor ifNotNil: 
			[sel first <= (self paragraph lines first last + 1) 
				ifTrue: 
					[oldLast := predecessor lastCharacterIndex.
					predecessor paragraph 
						recomposeFrom: oldLast
						to: text size
						delta: 0.
					oldLast = predecessor lastCharacterIndex 
						ifFalse: 
							[predecessor changed.	"really only last line"
							self predecessorChanged]]].
	((back := predecessor notNil 
				and: [sel first <= self paragraph firstCharacterIndex]) or: 
				[successor notNil 
					and: [sel first > (self paragraph lastCharacterIndex + 1)]]) 
		ifTrue: 
			["The selection is no longer inside this paragraph.
		Pass focus to the paragraph that should be in control."

			back ifTrue: [predecessor recomposeChain] ifFalse: [self recomposeChain].
			self firstInChain withSuccessorsDo: 
					[:m | 
					(sel first between: m firstCharacterIndex and: m lastCharacterIndex + 1) 
						ifTrue: 
							[m installEditorToReplace: oldEditor.
							^self passKeyboardFocusTo: m]].
			self error: 'Inconsistency in text editor'	"Must be somewhere in the successor chain"].
	editor ifNil: 
			["Reinstate selection after, eg, style change"

			self installEditorToReplace: oldEditor].
	"self setCompositionWindow."
! !

!TextMorph methodsFor: 'accessing' stamp: 'tk 12/16/1998 11:58'!
userString
	"Do I have a text string to be searched on?"

	^ text string! !

!TextMorph methodsFor: 'copying' stamp: 'tk 2/20/2001 18:55'!
veryDeepFixupWith: deepCopier 
	"If target and arguments fields were weakly copied, fix them here.  If 
	they were in the tree being copied, fix them up, otherwise point to the 
	originals!!"

	super veryDeepFixupWith: deepCopier.
	"It makes no sense to share pointers to an existing predecessor and successor"
	predecessor := deepCopier references at: predecessor ifAbsent: [nil].
	successor := deepCopier references at: successor ifAbsent: [nil]! !

!TextMorph methodsFor: 'copying' stamp: 'mt 11/15/2015 13:38'!
veryDeepInner: deepCopier 
	"Copy all of my instance variables. Some need to be not copied at all, but shared.
	Warning!!!! Every instance variable defined in this class must be handled.
	We must also implement veryDeepFixupWith:.  See DeepCopier class comment."

	super veryDeepInner: deepCopier.
	textStyle := textStyle veryDeepCopyWith: deepCopier.
	text := text veryDeepCopyWith: deepCopier.
	wrapFlag := wrapFlag veryDeepCopyWith: deepCopier.
	paragraph := paragraph veryDeepCopyWith: deepCopier.
	editor := editor veryDeepCopyWith: deepCopier.
	container := container veryDeepCopyWith: deepCopier.
	predecessor := predecessor.
	successor := successor.
	backgroundColor := backgroundColor veryDeepCopyWith: deepCopier.
	margins := margins veryDeepCopyWith: deepCopier.! !

!TextMorph methodsFor: 'event handling' stamp: 'mt 6/4/2016 16:42'!
wantsKeyboardFocus

	^ true! !

!TextMorph methodsFor: 'linked frames' stamp: 'dgd 2/21/2003 22:32'!
withSuccessorsDo: aBlock 
	"Evaluate aBlock for each morph in my successor chain"

	| each |
	each := self.
	[each isNil] whileFalse: 
			[aBlock value: each.
			each := each successor]! !

!TextMorph methodsFor: 'event handling' stamp: 'dgd 8/28/2004 13:54'!
wouldAcceptKeyboardFocusUponTab
	"Answer whether the receiver might accept keyboard focus if 
	tab were hit in some container playfield"
	^ self inPartsBin not! !

!TextMorph methodsFor: 'accessing' stamp: 'spd 1/4/2011 17:33'!
wrapFlag: aBoolean
	"Whether contained text will adjust its bounds as I change shape:
		|	wrapFlag 	|		TextMorph grows			|		TextMorph shrinks 	|
		|		true 		| wrapped lines fill new space	|	long lines wrap to fit 		|
		|		false 		|   wrapped lines stay same 		|	long lines are cut off 		|"

	aBoolean == wrapFlag ifTrue: [^ self].
	wrapFlag := aBoolean.
	self composeToBounds! !

!TextMorph methodsFor: 'menu' stamp: 'di 7/27/2001 13:20'!
wrapOnOff
	self wrapFlag: wrapFlag not! !

!TextMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 22:18'!
wrapString
	"Answer the string to put in a menu that will invite the user to 
	switch autoFit mode"
	^ (wrapFlag
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'text wrap to bounds' translated! !

!TextMorph methodsFor: 'editing' stamp: ''!
xeqLinkText: sourceString withParameter: param
	self confirm: 'xeqLinkText:
' asText allBold , sourceString asText! !

!TextMorph methodsFor: 'event handling' stamp: 'dgd 10/1/2004 13:15'!
yellowButtonActivity: shiftKeyState 
	"Invoke the text-editing menu"
	| menu |
	(menu := self getMenu: shiftKeyState)
		ifNotNil: [""
			menu setInvokingView: self editor.
			menu invokeModal. self changed]! !

BorderedMorph subclass: #SelectionMorph
	instanceVariableNames: 'selectedItems slippage dupLoc dupDelta itemsAlreadySelected otherSelection undoProperties'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!SelectionMorph commentStamp: '<historical>' prior: 0!
A selectionMorph supports the selection of multiple objects in a morphic world or pasteUp.

Structure:
	selectedItems	an OrderedCollection of Morphs
					These are the morphs that have been selected
	slippage		a Point
					Keeps track of actual movement between the 
					steps of gridded movement
	dupLoc		a Point
					Notes the position when first duplicate request occurs from halo
	dupDelta	a Point
					Holds the final delta of the first duplicate plus subsequent moves.
!

!SelectionMorph class methodsFor: 'scripting' stamp: 'KR 5/23/2008 01:30'!
defaultNameStemForInstances
	^ 'Selection' translatedNoop! !

!SelectionMorph methodsFor: 'dropping/grabbing' stamp: 'ar 10/10/2000 15:02'!
aboutToBeGrabbedBy: aHand
	slippage := 0@0.
	^ super aboutToBeGrabbedBy: aHand
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'bf 8/24/2010 14:48'!
addCustomMenuItems: aMenu hand: aHandMorph
	"Add custom menu items to the menu"

	super addCustomMenuItems: aMenu hand: aHandMorph.
	aMenu addList: {
		#-.
		{'place into a row' translated. #organizeIntoRow}.
		{'place into a column' translated. #organizeIntoColumn}.
		#-.
		{'align left edges' translated. #alignLeftEdges}.
		{'align top edges' translated. #alignTopEdges}.
		{'align right edges' translated. #alignRightEdges}.
		{'align bottom edges' translated. #alignBottomEdges}.
		#-.
		{'align centers vertically' translated. #alignCentersVertically}.
		{'align centers horizontally' translated. #alignCentersHorizontally}.
		#-.
		{'distribute vertically' translated. #distributeVertically}.
		{'distribute horizontally' translated. #distributeHorizontally}.
		}


! !

!SelectionMorph methodsFor: 'halos and balloon help' stamp: 'di 8/31/2000 22:29'!
addHandlesTo: aHaloMorph box: box
	| onlyThese |
	aHaloMorph haloBox: box.
	onlyThese := #(addDismissHandle: addMenuHandle: addGrabHandle: addDragHandle: addDupHandle: addHelpHandle: addGrowHandle: addFontSizeHandle: addFontStyleHandle: addFontEmphHandle: addRecolorHandle:).
	Preferences haloSpecifications do:
		[:aSpec | (onlyThese includes: aSpec addHandleSelector) ifTrue:
				[aHaloMorph perform: aSpec addHandleSelector with: aSpec]].
	aHaloMorph innerTarget addOptionalHandlesTo: aHaloMorph box: box! !

!SelectionMorph methodsFor: 'halos and balloon help' stamp: 'di 8/31/2000 22:59'!
addOptionalHandlesTo: aHalo box: box
	aHalo addHandleAt: box leftCenter color: Color blue icon: nil
		on: #mouseUp send: #addOrRemoveItems: to: self.! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'dgd 10/26/2004 14:53'!
addOrRemoveItems: handOrEvent 
	"Make a new selection extending the current one."

	| hand |
	hand := (handOrEvent isMorphicEvent) 
				ifFalse: [handOrEvent]
				ifTrue: [handOrEvent hand].
	hand 
		addMorphBack: ((self class 
				newBounds: (hand lastEvent cursorPoint extent: 16 @ 16)) 
					setOtherSelection: self).
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'dgd 9/20/2004 13:40'!
alignBottomEdges
	"Make the bottom coordinate of all my elements be the same"

	| maxBottom |
	maxBottom := (selectedItems collect: [:itm | itm bottom]) max.
	selectedItems do:
		[:itm | itm bottom: maxBottom].

	self changed
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'dgd 9/20/2004 13:41'!
alignCentersHorizontally
	"Make every morph in the selection have the same vertical center as the topmost item."

	| minLeft leftMost |
	selectedItems size > 1 ifFalse: [^ self].
	minLeft := (selectedItems collect: [:itm | itm left]) min.
	leftMost := selectedItems detect: [:m | m left = minLeft].
	selectedItems do:
		[:itm | itm center: (itm center x @ leftMost center y)].

	self changed
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'dgd 9/20/2004 13:41'!
alignCentersVertically
	"Make every morph in the selection have the same horizontal center as the topmost item."

	| minTop topMost |
	selectedItems size > 1 ifFalse: [^ self].
	minTop := (selectedItems collect: [:itm | itm top]) min.
	topMost := selectedItems detect: [:m | m top = minTop].
	selectedItems do:
		[:itm | itm center: (topMost center x @ itm center y)].

	self changed
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'dgd 9/20/2004 13:41'!
alignLeftEdges
	"Make the left coordinate of all my elements be the same"

	| minLeft |
	minLeft := (selectedItems collect: [:itm | itm left]) min.
	selectedItems do:
		[:itm | itm left: minLeft].

	self changed
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'dgd 9/20/2004 13:41'!
alignRightEdges
	"Make the right coordinate of all my elements be the same"

	| maxRight |
	maxRight := (selectedItems collect: [:itm | itm right]) max.
	selectedItems do:
		[:itm | itm right: maxRight].

	self changed
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'dgd 9/20/2004 13:50'!
alignTopEdges
	"Make the top coordinate of all my elements be the same"

	| minTop |
	minTop := (selectedItems collect: [:itm | itm top]) min.
	selectedItems do:
		[:itm | itm top: minTop].

	self changed
! !

!SelectionMorph methodsFor: 'halos and balloon help' stamp: 'di 8/31/2000 22:59'!
balloonHelpTextForHandle: aHandle
	aHandle eventHandler firstMouseSelector == #addOrRemoveItems:
		ifTrue: [^'Add items to, or remove them from, this selection.'].
	^ super balloonHelpTextForHandle: aHandle! !

!SelectionMorph methodsFor: 'undo' stamp: 'di 8/31/2000 00:24'!
borderColorForItems: colorCollection

	(selectedItems select: [:m | m isKindOf: BorderedMorph])
		with: colorCollection
		do: [:m :c | m borderColor: c]! !

!SelectionMorph methodsFor: 'accessing' stamp: 'nk 9/4/2004 17:33'!
borderColor: aColor

	| bordered |
	bordered := selectedItems.
	undoProperties ifNil: [undoProperties := bordered collect: [:m | m borderColor]].
	bordered do: [:m | m borderColor: aColor]! !

!SelectionMorph methodsFor: 'undo' stamp: 'di 8/31/2000 00:26'!
borderWidthForItems: widthCollection

	(selectedItems select: [:m | m isKindOf: BorderedMorph])
		with: widthCollection
		do: [:m :c | m borderWidth: c]! !

!SelectionMorph methodsFor: 'accessing' stamp: 'di 9/19/2000 18:25'!
borderWidth: aWidth

	| bordered |
	bordered := selectedItems select: [:m | m isKindOf: BorderedMorph].
	undoProperties ifNil: [undoProperties := bordered collect: [:m | m borderWidth]].
	bordered do: [:m | m borderWidth: aWidth]! !

!SelectionMorph methodsFor: 'geometry' stamp: 'jcg 2/14/2001 08:58'!
bounds: newBounds
	"Make sure position: gets called before extent:; Andreas' optimization for growing/shrinking in ChangeSet 3119 screwed up selection of morphs from underlying pasteup."

	selectedItems := OrderedCollection new.  "Avoid repostioning items during super position:"
	self position: newBounds topLeft; extent: newBounds extent
! !

!SelectionMorph methodsFor: 'testing' stamp: 'sw 3/30/2005 04:17'!
couldMakeSibling
	"Answer whether it is appropriate to ask the receiver to make a sibling"

	^ false! !

!SelectionMorph methodsFor: 'initialization' stamp: 'kfr 4/19/2015 12:11'!
defaultBorderColor
	"answer the default border color/fill style for the receiver"
	^ (	MenuMorph menuSelectionColor ifNil: [Color blue]) twiceDarker alpha: 0.75! !

!SelectionMorph methodsFor: 'initialization' stamp: 'dgd 8/28/2004 19:00'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 2! !

!SelectionMorph methodsFor: 'initialization' stamp: 'kfr 4/19/2015 12:11'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ (MenuMorph menuSelectionColor ifNil: [Color blue]) alpha: 0.08
! !

!SelectionMorph methodsFor: 'submorphs-add/remove' stamp: 'dgd 9/11/2004 18:27'!
delete
	self setProperty: #deleting toValue: true.
	super delete.
	! !

!SelectionMorph methodsFor: 'submorphs-add/remove' stamp: 'kfr 2/22/2013 21:49'!
dismissViaHalo
	selectedItems do: [:m | m dismissViaHalo].
	super dismissViaHalo.
	! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'dgd 9/20/2004 13:50'!
distributeHorizontally
	"Distribute the empty vertical space in a democratic way."
	| minLeft maxRight totalWidth currentLeft space |

	self selectedItems size > 2
		ifFalse: [^ self].

	minLeft := self selectedItems anyOne left.
	maxRight := self selectedItems anyOne right.
	totalWidth := 0.
	self selectedItems
		do: [:each | 
			minLeft := minLeft min: each left.
			maxRight := maxRight max: each right.
			totalWidth := totalWidth + each width].

	currentLeft := minLeft.
	space := (maxRight - minLeft - totalWidth / (self selectedItems size - 1)) rounded.
	(self selectedItems
		asSortedCollection: [:x :y | x left <= y left])
		do: [:each | 
			each left: currentLeft.
			currentLeft := currentLeft + each width + space].

	self changed
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'dgd 9/20/2004 13:51'!
distributeVertically
	"Distribute the empty vertical space in a democratic way."
	| minTop maxBottom totalHeight currentTop space |
	self selectedItems size > 2
		ifFalse: [^ self].

	minTop := self selectedItems anyOne top.
	maxBottom := self selectedItems anyOne bottom.
	totalHeight := 0.
	self selectedItems
		do: [:each | 
			minTop := minTop min: each top.
			maxBottom := maxBottom max: each bottom.
			totalHeight := totalHeight + each height].

	currentTop := minTop.
	space := (maxBottom - minTop - totalHeight / (self selectedItems size - 1)) rounded.
	(self selectedItems asSortedCollection:[:x :y | x top <= y top])
		do: [:each | 
			each top: currentTop.
			currentTop := currentTop + each height + space].

	self changed
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'md 11/14/2003 17:18'!
doDup: evt fromHalo: halo handle: dupHandle

	selectedItems := self duplicateMorphCollection: selectedItems.
	selectedItems do: [:m | self owner addMorph: m].
	dupDelta isNil
		ifTrue: ["First duplicate operation -- note starting location"
				dupLoc := self position.
				evt hand grabMorph: self.
				halo removeAllHandlesBut: dupHandle]
		ifFalse: ["Subsequent duplicate does not grab, but only moves me and my morphs"
				dupLoc := nil.
				self position: self position + dupDelta]
! !

!SelectionMorph methodsFor: 'private' stamp: 'dgd 9/20/2004 14:01'!
doneExtending

	otherSelection ifNotNil:
		[selectedItems := otherSelection selectedItems , selectedItems.
		otherSelection delete.
		self setOtherSelection: nil].
	self changed.
	self layoutChanged.
	super privateBounds:
		((Rectangle merging: (selectedItems collect: [:m | m fullBounds]))
			expandBy: 8).
	self changed.
	self addHalo.! !

!SelectionMorph methodsFor: 'drawing' stamp: 'nice 12/27/2009 22:39'!
drawOn: aCanvas

	| canvas form1 form2 box |
	super drawOn: aCanvas.
	box := self bounds copy.
	selectedItems do: [:m | box swallow: m fullBounds].
	box := box expandBy: 1.
	canvas := Display defaultCanvasClass extent: box extent depth: 8.
	canvas translateBy: box topLeft negated
		during: [:tempCanvas | selectedItems do: [:m | tempCanvas fullDrawMorph: m]].
	form1 := (Form extent: box extent) copyBits: (0@0 extent: box extent) from: canvas form at: 0@0 colorMap: (Color maskingMap: 8).
	form2 := Form extent: box extent.
	(0@0) fourNeighbors do: [:d | form1 displayOn: form2 at: d rule: Form under].
	form1 displayOn: form2 at: 0@0 rule: Form erase.
	aCanvas stencil: form2
		at: box topLeft
		sourceRect: form2 boundingBox
		color: self borderColor
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'sw 4/19/2005 18:14'!
duplicate
	"Make a duplicate of the receiver and havbe the hand grab it"

	selectedItems := self duplicateMorphCollection: selectedItems.
	selectedItems reverseDo: [:m | (owner ifNil: [ActiveWorld]) addMorph: m].
	dupLoc := self position.
	ActiveHand grabMorph: self.
	ActiveWorld presenter flushPlayerListCache! !

!SelectionMorph methodsFor: 'initialization' stamp: 'dgd 9/11/2004 21:07'!
extendByHand: aHand
	"Assumes selection has just been created and added to some pasteUp or world"
	| startPoint handle |

	startPoint := self position.

	handle := NewHandleMorph new followHand: aHand
		forEachPointDo: [:newPoint |
					| localPt |
					localPt := (self transformFrom: self world) globalPointToLocal: newPoint.
					self bounds: (startPoint rect: localPt)
				]
		lastPointDo: [:newPoint |
					selectedItems isEmpty
						ifTrue: [self delete]
						ifFalse: [
							selectedItems size = 1
								ifTrue:[self delete.  selectedItems anyOne addHalo]
								ifFalse:[self doneExtending]
						]
				].

	aHand attachMorph: handle.
	handle startStepping.! !

!SelectionMorph methodsFor: 'geometry' stamp: 'sw 5/17/2007 12:17'!
extent: newExtent
	"Set the receiver's extent   Extend or contract the receiver's selection to encompass morphs within the new extent."

	super extent: newExtent.
	self selectSubmorphsOf: (self pasteUpMorph ifNil: [^ self])! !

!SelectionMorph methodsFor: 'viewer' stamp: 'dgd 8/29/2004 12:34'!
externalName
	^ 'Selected {1} objects' translated format:{self selectedItems size}! !

!SelectionMorph methodsFor: 'undo' stamp: 'di 9/19/2000 21:53'!
fillStyleForItems: fillStyleCollection

	selectedItems with: fillStyleCollection do: [:m :c | m fillStyle: c]! !

!SelectionMorph methodsFor: 'visual properties' stamp: 'di 9/19/2000 21:53'!
fillStyle: aColor
	undoProperties ifNil: [undoProperties := selectedItems collect: [:m | m fillStyle]].
	selectedItems do: [:m | m fillStyle: aColor]! !

!SelectionMorph methodsFor: 'halos and balloon help' stamp: 'dgd 9/11/2004 18:29'!
hasHalo: aBool 
	super hasHalo: aBool.
	aBool
		ifFalse: [ (self hasProperty: #deleting) ifFalse: [self delete] ]
! !

!SelectionMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:23'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	
	selectedItems := OrderedCollection new.
	itemsAlreadySelected := OrderedCollection new.
	slippage := 0 @ 0! !

!SelectionMorph methodsFor: 'testing' stamp: 'nk 9/4/2004 17:38'!
isSelectionMorph
	^true! !

!SelectionMorph methodsFor: 'dropping/grabbing' stamp: 'kfr 6/6/2009 09:50'!
justDroppedInto: newOwner event: evt
	"This message is sent to a dropped morph after it has been dropped on -- and been accepted by -- a drop-sensitive morph"

	selectedItems isEmpty ifTrue:
		["Hand just clicked down to draw out a new selection"
		^ self extendByHand: evt hand].
	dupLoc ifNotNil: [dupDelta _ self position - dupLoc].
	selectedItems reverseDo: [:m | 
		WorldState addDeferredUIMessage:
			[m referencePosition: (newOwner localPointToGlobal: m referencePosition).
			newOwner handleDropMorph:
				(DropEvent new setPosition: evt cursorPoint contents: m hand: evt hand)] fixTemps].
	selectedItems _ nil.
	self removeHalo. 
	self halo ifNotNil: [self halo visible: false]. 
	self delete.
	evt wasHandled: true! !

!SelectionMorph methodsFor: 'menus' stamp: 'di 8/31/2000 20:50'!
maybeAddCollapseItemTo: aMenu
	"... don't "! !

!SelectionMorph methodsFor: 'WiW support' stamp: 'dgd 9/11/2004 21:58'!
morphicLayerNumber
	"helpful for insuring some morphs always appear in front of or  
	behind others. smaller numbers are in front"
	^ 8! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'tb 8/10/2003 14:47'!
organizeIntoColumn
	"Place my objects in a column-enforcing container"

	((AlignmentMorph inAColumn: (selectedItems asSortedCollection: [:x :y | x top < y top])) setNameTo: 'Column'; color: Color orange muchLighter; enableDragNDrop: true; yourself) openInHand
! !

!SelectionMorph methodsFor: 'halo commands' stamp: 'tb 8/10/2003 14:48'!
organizeIntoRow
	"Place my objects in a row-enforcing container"

	((AlignmentMorph inARow: (selectedItems asSortedCollection: [:x :y | x left < y left])) setNameTo: 'Row'; color: Color orange muchLighter; enableDragNDrop: true; yourself) openInHand
! !

!SelectionMorph methodsFor: 'testing' stamp: 'sw 3/30/2005 04:30'!
preferredDuplicationHandleSelector
	"Answer the selector, to be offered as the default in a halo open on me"

	^ #addDupHandle:! !

!SelectionMorph methodsFor: 'private' stamp: 'nk 8/14/2003 08:12'!
privateFullMoveBy: delta

	| griddedDelta griddingMorph |
	selectedItems isEmpty ifTrue: [^ super privateFullMoveBy: delta].
	griddingMorph := self pasteUpMorph.
	griddingMorph ifNil: [^ super privateFullMoveBy: delta].
	griddedDelta := (griddingMorph gridPoint: self position + delta + slippage) -
					(griddingMorph gridPoint: self position).
	slippage := slippage + (delta - griddedDelta).  "keep track of how we lag the true movement."
	griddedDelta = (0@0) ifTrue: [^ self].
	super privateFullMoveBy: griddedDelta.
	selectedItems do:
		[:m | m position: (m position + griddedDelta) ]
! !

!SelectionMorph methodsFor: 'undo' stamp: 'di 9/19/2000 21:52'!
refineUndoTarget: target selector: selector arguments: arguments in: refineBlock
	"Any morph can override this method to refine its undo specification"
	selector == #fillStyle: ifTrue:
		[refineBlock value: target value: #fillStyleForItems: value: {undoProperties}.
		^ undoProperties := nil].
	selector == #borderColor: ifTrue:
		[refineBlock value: target value: #borderColorForItems: value: {undoProperties}.
		^ undoProperties := nil].
	selector == #borderWidth: ifTrue:
		[refineBlock value: target value: #borderWidthForItems: value: {undoProperties}.
		^ undoProperties := nil].
	selector == #undoMove:redo:owner:bounds:predecessor: ifTrue:
		["This is the biggy.  Need to gather parameters for all selected items"
		refineBlock value: target
			value: #undoMove:redo:owner:bounds:predecessor:
			value: {arguments first.
					arguments second.
					selectedItems collect: [:m | m owner].
					selectedItems collect: [:m | m bounds].
					selectedItems collect: [:m | m owner morphPreceding: m]}].
	refineBlock value: target value: selector value: arguments! !

!SelectionMorph methodsFor: 'private' stamp: 'di 8/31/2000 21:36'!
selectedItems

	^ selectedItems! !

!SelectionMorph methodsFor: 'private' stamp: 'sw 5/17/2007 12:06'!
selectSubmorphsOf: aMorph
	"Given the receiver's current bounds, select submorphs of the indicated pasteup morph that fall entirely within those bounds.  If nobody is within the bounds, delete the receiver."

	| newItems removals |
	newItems _ aMorph submorphs select:
		[:m | (bounds containsRect: m fullBounds) 
					and: [m~~self
					and: [(m isKindOf: HaloMorph) not]]].
	otherSelection ifNil: [^ selectedItems _ newItems].

	removals _ newItems intersection: itemsAlreadySelected.
	otherSelection setSelectedItems: (itemsAlreadySelected copyWithoutAll: removals).
	selectedItems _ (newItems copyWithoutAll: removals).
	selectedItems ifEmpty: [self delete]
! !

!SelectionMorph methodsFor: 'private' stamp: 'dgd 2/21/2003 23:18'!
setOtherSelection: otherOrNil 
	otherSelection := otherOrNil.
	otherOrNil isNil 
		ifTrue: [super borderColor: Color blue]
		ifFalse: 
			[itemsAlreadySelected := otherSelection selectedItems.
			super borderColor: Color green]! !

!SelectionMorph methodsFor: 'private' stamp: 'di 8/31/2000 21:45'!
setSelectedItems: items

	selectedItems := items.
	self changed! !

!SelectionMorph methodsFor: 'dropping/grabbing' stamp: 'kfr 2/22/2013 21:41'!
slideToTrash: evt
	self delete.
	"selectedItems do: [:m | m slideToTrash: evt]"! !

!SelectionMorph methodsFor: 'undo' stamp: 'mt 5/17/2016 16:00'!
undoMove: cmd redo: redo owner: oldOwners bounds: oldBoundses predecessor: oldPredecessors 
	"Handle undo and redo of move commands in morphic"

	| item |
	redo 
		ifFalse: 
			["undo sets up the redo state first"

			cmd 
				redoTarget: self
				selector: #undoMove:redo:owner:bounds:predecessor:
				arguments: { 
						cmd.
						true.
						selectedItems collect: [:m | m owner].
						selectedItems collect: [:m | m bounds].
						selectedItems collect: [:m | m owner morphPreceding: m]}].
	1 to: selectedItems size do: 
				[:i | 
				item := selectedItems at: i.
				(oldOwners at: i) ifNotNil: 
						[(oldPredecessors at: i) ifNil: [(oldOwners at: i) addMorphFront: item]
							ifNotNil: [(oldOwners at: i) addMorph: item after: (oldPredecessors at: i)]].
				item bounds: (oldBoundses at: i).
				item isSystemWindow ifTrue: [item beKeyWindow]]! !

!SelectionMorph methodsFor: 'accessing' stamp: 'dgd 9/11/2004 21:57'!
wantsToBeTopmost
	"Answer if the receiver want to be one of the topmost objects in 
	its owner"
	^ true! !

BorderedMorph subclass: #TTSampleFontMorph
	instanceVariableNames: 'font transform smoothing'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-TrueType'!

!TTSampleFontMorph commentStamp: '<historical>' prior: 0!
An example for using TrueType fonts.!

!TTSampleFontMorph class methodsFor: 'connectors' stamp: 'nk 8/17/2003 11:34'!
fontWithoutString: aTTFontDescription
	^self new fontWithoutString: aTTFontDescription! !

!TTSampleFontMorph class methodsFor: 'instance creation' stamp: 'ar 11/14/1998 23:06'!
font: aTTFontDescription
	^self new font: aTTFontDescription! !

!TTSampleFontMorph methodsFor: 'menus' stamp: 'ar 6/16/1999 07:21'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu addUpdating: #getSmoothingLevel action: #nextSmoothingLevel.! !

!TTSampleFontMorph methodsFor: 'halos and balloon help' stamp: 'sw 1/27/2000 15:43'!
addOptionalHandlesTo: aHalo box: box
	aHalo addHandleAt: box center color: Color magenta icon: nil on: #mouseDown send: #createSample to: self.! !

!TTSampleFontMorph methodsFor: 'drawing' stamp: 'ar 11/14/1998 23:12'!
areasRemainingToFill: aRectangle
	^ Array with: aRectangle! !

!TTSampleFontMorph methodsFor: 'halos and balloon help' stamp: 'ar 11/14/1998 23:39'!
balloonHelpTextForHandle: aHandle
	aHandle eventHandler firstMouseSelector == #createSample
		ifTrue:[^'Create a sample string'].
	^super balloonHelpTextForHandle: aHandle! !

!TTSampleFontMorph methodsFor: 'testing' stamp: 'ar 8/25/2001 19:09'!
canDrawBorder: aBorderStyle
	^aBorderStyle style == #simple! !

!TTSampleFontMorph methodsFor: 'updating' stamp: 'sma 1/1/2000 17:59'!
changed
	self invalidRect: (self fullBounds expandBy: 1)! !

!TTSampleFontMorph methodsFor: 'private' stamp: 'ar 11/15/1998 22:48'!
computeTransform
	| fullExtent scale |
	fullExtent := font bounds extent * 16.
	scale := self extent asFloatPoint / fullExtent asFloatPoint.
	transform := MatrixTransform2x3 withScale: scale.
	transform := transform composedWithGlobal: (MatrixTransform2x3 withOffset: self position).
	^transform! !

!TTSampleFontMorph methodsFor: 'menu' stamp: 'ar 11/14/1998 23:46'!
createSample
	self world primaryHand attachMorph: (TTSampleStringMorph font: font)! !

!TTSampleFontMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:40'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 1! !

!TTSampleFontMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:31'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color black! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'ar 11/14/1998 22:52'!
doesBevels
	^false! !

!TTSampleFontMorph methodsFor: 'drawing' stamp: 'nice 12/29/2010 15:02'!
drawCharactersOn: aCanvas
	| glyph origin r offset cy m |
	0 to: 255 do: [:i |
		glyph := font at: i.
		origin := font bounds extent * ((i \\ 16) @ (i // 16)).
		r := origin extent: font bounds extent.
		offset := r center - glyph bounds center.
		cy := glyph bounds center y.
		m := MatrixTransform2x3 withOffset: 0@cy.
		m := m composedWithLocal: (MatrixTransform2x3 withScale: 1 @ -1).
		m := m composedWithLocal: (MatrixTransform2x3 withOffset: 0 @ cy negated).
		m := m composedWithGlobal: (MatrixTransform2x3 withOffset: offset).
		aCanvas asBalloonCanvas preserveStateDuring: [:balloonCanvas |
			balloonCanvas transformBy: m.
			balloonCanvas drawGeneralBezierShape: glyph contours
					color: color
					borderWidth: 0
					borderColor: Color black.
		].
	].! !

!TTSampleFontMorph methodsFor: 'drawing' stamp: 'ar 12/30/1998 10:49'!
drawOn: aCanvas
	| origin extent offset |
	(font isNil) 
		ifTrue:[^aCanvas frameRectangle: bounds color: Color black].
	origin := self position asIntegerPoint.
	extent := self extent asIntegerPoint.
	0 to: 16 do:[:i|
		offset := (extent x * i // 16) @ (extent y * i // 16).
		aCanvas line: origin x @ (origin y + offset y) 
				to: (origin x + extent x) @ (origin y + offset y)
				width: borderWidth color: borderColor.
		aCanvas line: (origin x + offset x) @ origin y 
				to: (origin x + offset x) @ (origin y + extent y)
				width: borderWidth color: borderColor.
	].
	aCanvas asBalloonCanvas preserveStateDuring:[:balloonCanvas|
		balloonCanvas transformBy: self transform.
		balloonCanvas aaLevel: self smoothing.
		self drawCharactersOn: balloonCanvas.
	].! !

!TTSampleFontMorph methodsFor: 'geometry' stamp: 'ar 11/14/1998 22:53'!
extent: extentPoint
	super extent: extentPoint.
	transform := nil.! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'sma 1/1/2000 18:08'!
font
	^ font! !

!TTSampleFontMorph methodsFor: 'connectors' stamp: 'nk 8/17/2003 11:34'!
fontWithoutString: aTTFontDescription
	font := aTTFontDescription.
! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'ar 11/15/1998 00:31'!
font: aTTFontDescription
	| morph |
	font := aTTFontDescription.
	morph := (TTSampleStringMorph font: font).
	morph extent: morph extent * 2.
	morph color: Color magenta.
	self addMorphCentered: morph.
	morph position: morph position x @ (self bounds bottom + 10).
	self privateFullMoveBy: self fullBounds origin negated.! !

!TTSampleFontMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 21:42'!
getSmoothingLevel
	"Menu support"
	smoothing = 1
		ifTrue: [^ 'turn on smoothing' translated].
	smoothing = 2
		ifTrue: [^ 'more smoothing' translated].
	smoothing = 4
		ifTrue: [^ 'turn off smoothing' translated]! !

!TTSampleFontMorph methodsFor: 'connectors' stamp: 'nk 8/17/2003 11:43'!
glyphAt: position
	^font at: (self glyphIndexAt: position).! !

!TTSampleFontMorph methodsFor: 'connectors' stamp: 'nk 8/17/2003 11:54'!
glyphIndexAt: position
	| offset |
	offset := (position adhereTo: (bounds insetBy: 1)) - bounds origin.
	offset := (offset asFloatPoint / bounds extent) * 16.
	offset := offset truncated.
	^offset y * 16 + offset x! !

!TTSampleFontMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 18:43'!
initialize
	"initialize the state of the receiver"
	super initialize.
	smoothing := 4.
	self extent: 300 @ 300! !

!TTSampleFontMorph methodsFor: 'rotate scale and flex' stamp: 'ar 11/15/1998 22:42'!
newTransformationMorph
	^MatrixTransformMorph new! !

!TTSampleFontMorph methodsFor: 'menu' stamp: 'sma 1/1/2000 17:51'!
nextSmoothingLevel
	smoothing = 1
		ifTrue: [smoothing := 2]
		ifFalse: [smoothing = 2
			ifTrue: [smoothing := 4]
			ifFalse: [smoothing = 4
				ifTrue: [smoothing := 1]]].
	self changed! !

!TTSampleFontMorph methodsFor: 'initialize' stamp: 'ar 8/15/2009 14:54'!
open
	^self openInWorld! !

!TTSampleFontMorph methodsFor: 'initialization' stamp: 'ar 8/15/2009 14:54'!
openInWorld
	HandMorph attach: self! !

!TTSampleFontMorph methodsFor: 'geometry' stamp: 'ar 11/14/1998 22:53'!
position: pos
	super position: pos.
	transform := nil.! !

!TTSampleFontMorph methodsFor: 'connectors' stamp: 'nk 7/15/2003 15:12'!
printOn: aStream
	aStream nextPutAll: 'TTSampleFont(';
		nextPutAll: font familyName;
		nextPut: $)! !

!TTSampleFontMorph methodsFor: 'private' stamp: 'ar 11/14/1998 22:55'!
privateMoveBy: delta
	super privateMoveBy: delta.
	transform := nil.! !

!TTSampleFontMorph methodsFor: 'connectors' stamp: 'nk 8/17/2003 12:12'!
selectGlyph
	| retval done |
	"Modal glyph selector"
	done := false.
	self on: #mouseDown send: #selectGlyphBlock:event:from: to: self withValue: [ :glyph | retval := glyph. done := true. ].
	self on: #keyStroke send: #value to: [ done := true ].
	[ done ] whileFalse: [ self world doOneCycle ].
	self on: #mouseDown send: nil to: nil.
	self on: #keyStroke send: nil to: nil.
	^retval! !

!TTSampleFontMorph methodsFor: 'connectors' stamp: 'nk 8/17/2003 11:58'!
selectGlyphAndSendTo: aBlock
	self on: #mouseDown send: #selectGlyphBlock:event:from: to: self withValue: aBlock.! !

!TTSampleFontMorph methodsFor: 'connectors' stamp: 'nk 8/17/2003 12:17'!
selectGlyphBlock: aBlock event: evt from: me
	aBlock value: (self glyphAt: evt position).
! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'sma 1/1/2000 17:53'!
smoothing
	^ smoothing! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'bf 10/18/1999 16:19'!
smoothing: aNumber
	smoothing := aNumber.
	self changed! !

!TTSampleFontMorph methodsFor: 'accessing' stamp: 'ar 11/14/1998 22:53'!
transform
	^transform ifNil:[self computeTransform].! !

!TTSampleFontMorph methodsFor: 'copying' stamp: 'sma 2/26/2000 19:20'!
veryDeepFixupWith: deepCopier
	"If fields were weakly copied, fix them here. If they were in the 
	tree being copied, fix them up, otherwise point to the originals!!!!"

	super veryDeepFixupWith: deepCopier.
	font := deepCopier references at: font ifAbsent: [font]! !

!TTSampleFontMorph methodsFor: 'copying' stamp: 'sma 2/26/2000 19:18'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all,
	but shared. Warning!!!! Every instance variable defined in this class
	must be handled.  We must also implement veryDeepFixupWith:.
	See DeepCopier class comment."

	super veryDeepInner: deepCopier.
	"font := font"
	transform := transform veryDeepCopyWith: deepCopier.
	smoothing := smoothing veryDeepCopyWith: deepCopier! !

TTSampleFontMorph subclass: #TTSampleStringMorph
	instanceVariableNames: 'string ttBounds'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-TrueType'!

!TTSampleStringMorph commentStamp: '<historical>' prior: 0!
I allow the display of a string in a TrueType font as a stand-alone morph.

Morph's color changes the inside of the characters.
Morph's borderColor changes the outline.

Many free fonts are stored at www.FontGuy.com.  
Use a normal web browser (not our Scamper) and go there.  
Choose 'categories' and browse to a font you like.  
Hold the mouse down on the example text in that font.  
When the menu comes up, choose "Copy this link location".  
Come back into Squeak, choose "load font from web..."
from my menu, and paste in the url.!

!TTSampleStringMorph class methodsFor: 'parts bin' stamp: 'tak 7/25/2007 00:12'!
descriptionForPartsBin
	^ self partName:	'TrueType banner' translatedNoop
		categories:		#()
		documentation:	'A short text in a beautiful font.  Use the resize handle to change size.' translatedNoop! !

!TTSampleStringMorph methodsFor: 'menus' stamp: 'dgd 8/30/2003 22:17'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu add: 'edit contents...' translated action: #edit.
	aCustomMenu add: 'how to find more fonts...' translated action: #howTo.
	aCustomMenu add: 'load font from web...' translated action: #loadFromURL.! !

!TTSampleStringMorph methodsFor: 'halos and balloon help' stamp: 'ar 11/14/1998 23:44'!
addOptionalHandlesTo: aHalo box: box! !

!TTSampleStringMorph methodsFor: 'private' stamp: 'nice 12/29/2010 15:02'!
computeTransform
	| cy |
	cy := bounds origin y + bounds corner y * 0.5.
	transform := MatrixTransform2x3 
			transformFromLocal: (ttBounds insetBy: borderWidth negated)
			toGlobal: bounds.
	transform := transform composedWithGlobal:(MatrixTransform2x3 withOffset: 0 @ cy negated).
	transform := transform composedWithGlobal:(MatrixTransform2x3 withScale: 1.0 @ -1.0).
	transform := transform composedWithGlobal:(MatrixTransform2x3 withOffset: 0 @ cy).
	^transform! !

!TTSampleStringMorph methodsFor: 'geometry testing' stamp: 'dgd 2/22/2003 14:42'!
containsPoint: aPoint 
	"^ super containsPoint: aPoint"

	"so much faster..."

	| picker |
	(self bounds containsPoint: aPoint) ifFalse: [^false].
	picker := BalloonCanvas on: (Form extent: 1 @ 1 depth: 32).
	picker transformBy: (MatrixTransform2x3 withOffset: aPoint negated).
	self drawOn: picker.
	^(picker form bits first) ~= 0! !

!TTSampleStringMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:40'!
defaultBorderWidth
	"answer the default border width for the receiver"
	^ 0! !

!TTSampleStringMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:31'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ {Color magenta. Color yellow. Color orange. Color lightGray} atRandom! !

!TTSampleStringMorph methodsFor: 'drawing' stamp: 'nice 12/27/2009 03:10'!
drawOn: aCanvas
	| xStart |
	(font isNil or:[string isNil or:[string isEmpty]]) 
		ifTrue:[^aCanvas frameRectangle: bounds color: Color black].
	xStart := 0.
	aCanvas asBalloonCanvas preserveStateDuring:[:balloonCanvas|
		balloonCanvas transformBy: self transform.
		balloonCanvas aaLevel: self smoothing.
		string do:[:char| | glyph |
			glyph := font at: char.
			balloonCanvas preserveStateDuring:[:subCanvas|
				subCanvas transformBy: (MatrixTransform2x3 withOffset: xStart@0).
				subCanvas 
					drawGeneralBezierShape: glyph contours
					color: color 
					borderWidth: borderWidth 
					borderColor: borderColor].
			xStart := xStart + glyph advanceWidth.
		].
	].! !

!TTSampleStringMorph methodsFor: 'menus' stamp: 'tk 12/7/2001 11:24'!
edit
	"Allow the user to change the text in a crude way"

	| str |
	str := FillInTheBlankMorph request: 'Type in new text for this TrueType displayer.'
				 initialAnswer: 'some text'.
	str isEmpty ifTrue: [^ self].
	self string: str.
! !

!TTSampleStringMorph methodsFor: 'accessing' stamp: 'tk 12/10/2001 16:21'!
font: aTTFontDescription
	font := aTTFontDescription.
	string ifNil: [self string: aTTFontDescription fullName]
		ifNotNil: [self initializeString].! !

!TTSampleStringMorph methodsFor: 'menus' stamp: 'tk 12/10/2001 16:02'!
howTo

	self inform: 'Many free fonts are stored at www.FontGuy.com.  
Use a normal web browser (not our Scamper) and go there.  
Choose ''categories'' and browse to a font you like.  
Hold the mouse down on the example text in that font.  
When the menu comes up, choose "Copy this link location".  
Come back into Squeak, choose "load font from web..."
from this menu, and paste in the url.'! !

!TTSampleStringMorph methodsFor: 'initialize' stamp: 'sma 1/1/2000 18:08'!
initializeString
	| xStart char glyph |
	(font isNil or: [string isNil]) ifTrue: [^ self].
	xStart := 0.
	ttBounds := 0@0 corner: 0@0.
	1 to: string size do:
		[:i |
		char := string at: i.
		glyph := font at: char.
		ttBounds := ttBounds quickMerge: (glyph bounds translateBy: xStart@0).
		xStart := xStart + glyph advanceWidth.
	].
	self extent: ttBounds extent // 40.
	borderWidth := ttBounds height // 40! !

!TTSampleStringMorph methodsFor: 'parts bin' stamp: 'tk 12/10/2001 17:36'!
initializeToStandAlone
	"Make me into an example"

	| dd |
	dd := TTFontDescription default.
	dd ifNil: [^ RectangleMorph initializeToStandAlone].	"not available"

	super initializeToStandAlone.
	self font: dd; color: (TranslucentColor r: 1.0 g: 0.097 b: 1.0 alpha: 0.6).
	self string: 'TrueType fonts are beautiful'.
! !

!TTSampleStringMorph methodsFor: 'menus' stamp: 'tk 12/10/2001 16:03'!
loadFromURL
	"Allow the user to change the text in a crude way"

	| url |
	url := FillInTheBlankMorph request: ' Type in the url for a TrueType font on the web. '
				 initialAnswer: 'http://www.fontguy.com/download.asp?fontid=1494'.
	url isEmpty ifTrue: [^ self].
	self loadFromURL: url.
! !

!TTSampleStringMorph methodsFor: 'menus' stamp: 'tk 12/10/2001 16:03'!
loadFromURL: urlString
	"Fetch the file, unarchive, unzip, and use as my font."

	| rawStrm |
	rawStrm := HTTPSocket httpGet: urlString. 	"Later use an HttpURL?"
	self font: (TTFontReader readFrom: rawStrm asUnZippedStream).
! !

!TTSampleStringMorph methodsFor: 'printing' stamp: 'nk 7/15/2003 15:12'!
printOn: aStream
	aStream nextPutAll: 'TTSampleString(';
		nextPutAll: font familyName;
		nextPut: $)! !

!TTSampleStringMorph methodsFor: 'accessing' stamp: 'sma 1/1/2000 18:08'!
string
	^ string! !

!TTSampleStringMorph methodsFor: 'accessing' stamp: 'ar 11/14/1998 23:53'!
string: aString
	string := aString.
	self initializeString.! !

Morph subclass: #BracketMorph
	instanceVariableNames: 'orientation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!BracketMorph commentStamp: 'gvc 5/18/2007 13:48' prior: 0!
Morph displaying opposing arrows.!

!BracketMorph methodsFor: 'drawing' stamp: 'gvc 9/21/2006 16:16'!
drawOn: aCanvas
	"Draw triangles at the edges."
	
	|r|
	r := self horizontal
		ifTrue: [self bounds insetBy: (2@1 corner: 2@1)]
		ifFalse: [self bounds insetBy: (1@2 corner: 1@2)].
	aCanvas
		drawPolygon: (self leftOrTopVertices: self bounds)
		fillStyle: self borderColor;
		drawPolygon: (self leftOrTopVertices: r)
		fillStyle: self fillStyle;
		drawPolygon: (self rightOrBottomVertices: self bounds)
		fillStyle: self borderColor;
		drawPolygon: (self rightOrBottomVertices: r)
		fillStyle: self fillStyle! !

!BracketMorph methodsFor: 'accessing' stamp: 'gvc 9/21/2006 15:48'!
horizontal
	"Answer whether horizontal or vertical."
	
	^self orientation == #horizontal! !

!BracketMorph methodsFor: 'accessing' stamp: 'gvc 9/19/2006 15:54'!
horizontal: aBoolean
	"Set whether horizontal or vertical."
	
	^self orientation: (aBoolean ifTrue: [#horizontal] ifFalse: [#vertical])! !

!BracketMorph methodsFor: 'initialization' stamp: 'gvc 9/19/2006 15:52'!
initialize
	"Initialize the receiver."

	super initialize.
	self
		orientation: #horizontal! !

!BracketMorph methodsFor: 'geometry' stamp: 'gvc 9/21/2006 15:45'!
leftOrTopVertices: r
	"Answer the vertices for a left or top bracket in the given rectangle."
	
	^self orientation == #vertical
		ifTrue: [{r topLeft - (0@1). r left + (r height // 2 + (r height \\ 2))@(r center y - (r height + 1 \\ 2)).
				r left + (r height // 2 + (r height \\ 2))@(r center y). r bottomLeft}]
		ifFalse: [{r topLeft. (r center x - (r width + 1 \\ 2))@(r top + (r width // 2 + (r width \\ 2))).
				r center x@(r top + (r width // 2 + (r width \\ 2))). r topRight}]! !

!BracketMorph methodsFor: 'accessing' stamp: 'gvc 9/19/2006 15:51'!
orientation
	"Answer the value of orientation"

	^ orientation! !

!BracketMorph methodsFor: 'accessing' stamp: 'gvc 9/19/2006 15:52'!
orientation: anObject
	"Set the value of orientation"

	orientation := anObject.
	self changed! !

!BracketMorph methodsFor: 'geometry' stamp: 'gvc 9/21/2006 16:18'!
rightOrBottomVertices: r
	"Answer the vertices for a right or bottom bracket in the given rectangle."
	
	^self orientation == #vertical
		ifTrue: [{r topRight - (0@1). r right - (r height // 2 + (r height \\ 2))@(r center y - (r height + 1 \\ 2)).
				r right - (r height // 2 + (r height \\ 2))@(r center y). r bottomRight}]
		ifFalse: [{(r center x)@(r bottom - 1 - (r width // 2 + (r width \\ 2))).
				r center x @(r bottom - 1 - (r width // 2 + (r width \\ 2))). r bottomRight. r bottomLeft - (1@0)}]! !

Morph subclass: #DialogWindow
	instanceVariableNames: 'titleMorph messageMorph paneMorph buttonRow result selectedButton cancelButton timeout preferredPosition keyMap exclusive filter filterEnabled filterMorph autoCancel'
	classVariableNames: 'GradientDialog RoundedDialogCorners UseWiggleAnimation'
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!DialogWindow commentStamp: '<historical>' prior: 0!
A DialogBoxMorph is Morph used in simple yes/no/confirm dialogs. Strongly modal.!

!DialogWindow class methodsFor: 'preferences' stamp: 'mt 7/25/2016 14:35'!
gradientDialog

	<preference: 'gradientDialog'
		category: 'dialogs'
		description: 'If true, dialogs will have a gradient look.'
		type: #Boolean>
	^ GradientDialog ifNil: [true]
! !

!DialogWindow class methodsFor: 'preferences' stamp: 'mt 7/25/2016 14:35'!
gradientDialog: aBoolean

	aBoolean = GradientDialog ifTrue: [^ self].
	GradientDialog := aBoolean.
	self refreshAllDialogs.! !

!DialogWindow class methodsFor: 'preferences' stamp: 'mt 7/25/2016 14:34'!
refreshAllDialogs
	
	self allSubInstances do: [:instance | instance setDefaultParameters; setTitleParameters].! !

!DialogWindow class methodsFor: 'preferences' stamp: 'mt 7/25/2016 14:27'!
roundedDialogCorners
	<preference: 'Rounded Dialog Corners'
		categoryList: #(windows dialogs)
		description: 'Governs whether dialog windows should have rounded corners'
		type: #Boolean>
	^ RoundedDialogCorners ifNil: [ true ]! !

!DialogWindow class methodsFor: 'preferences' stamp: 'mt 7/25/2016 14:34'!
roundedDialogCorners: aBoolean
	
	RoundedDialogCorners = aBoolean ifTrue: [^ self].
	RoundedDialogCorners := aBoolean.
	self refreshAllDialogs.! !

!DialogWindow class methodsFor: 'preferences' stamp: 'mt 7/13/2016 19:24'!
themeProperties

	^ super themeProperties, {
		{ #borderColor. 'Colors'. 'Color of the dialogs''s border.' }.
		{ #borderWidth. 'Borders'. 'Width of the dialogs''s border.' }.
		{ #borderStyle. 'Borders'. 'Whether to use a plain border, inset, or outset.' }.
		{ #color. 'Colors'. 'Background color of the dialog.' }.
		{ #font. 'Fonts'. 'Font for dialog messages.' }.
		{ #textColor. 'Colors'. 'Color for dialog messages.' }.
		
		{ #titleBorderColor. 'Colors'. 'Color of the dialogs title border.' }.
		{ #titleBorderWidth. 'Geometry'. 'Width of the dialog title border.' }.
		{ #titleBorderStyle. 'Borders'. 'Whether to use a plain border, inset, or outset for the title.' }.
		{ #titleColor. 'Colors'. 'Background color of the dialogs'' title.' }.
		{ #titleFont. 'Fonts'. 'Font for dialog title.' }.
		{ #titleTextColor. 'Colors'. 'Color for the dialog title label.' }.
		
		{ #okColor. 'Colors'. 'Color for the OK button.' }.
		{ #cancelColor. 'Colors'. 'Color for the Cancel button.' }.
		{ #buttonColor. 'Colors'. 'Color for a normal button.' }.
		{ #selectionModifier. 'Colors'. 'How to convert the color of a selected button?' }.
	}! !

!DialogWindow class methodsFor: 'preferences' stamp: 'mt 7/19/2016 13:49'!
useWiggleAnimation
	
	<preference: 'Use Wiggle Animation in Modal Dialogs'
		categoryList: #(Morphic windows dialogs)
		description: 'In order to indicate that a modal dialog is waiting for a user''s input, wiggle instead of flash.'
		type: #Boolean>
	^ UseWiggleAnimation ifNil: [true]! !

!DialogWindow class methodsFor: 'preferences' stamp: 'mt 7/19/2016 13:49'!
useWiggleAnimation: aBoolean
	
	UseWiggleAnimation := aBoolean.! !

!DialogWindow methodsFor: 'constructing' stamp: 'mt 7/13/2016 19:41'!
addPaneMorph: aMorph

	self paneMorph addMorphBack: aMorph.! !

!DialogWindow methodsFor: 'updating' stamp: 'mt 7/26/2016 16:56'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	
	self
		setDefaultParameters;
		setTitleParameters;
		setMessageParameters.
		
	"Update all buttons."
	selectedButton in: [:sb |
		self buttons do: [:ea |
			ea setProperty: #normalColor toValue: self defaultButtonColor.
			self deselectButton: ea].
		sb ifNotNil: [self selectButton: sb]].! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 8/5/2016 09:41'!
autoCancel
	"Whether to automatically cancel and close this dialog if the user clicks outside it. Like dismissing a pop-up menu."
	
	^ autoCancel! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 8/5/2016 09:39'!
autoCancel: aBoolean

	autoCancel := aBoolean.! !

!DialogWindow methodsFor: 'accessing - ui' stamp: 'mt 7/14/2016 09:44'!
buttonRowMorph
	^ buttonRow! !

!DialogWindow methodsFor: 'accessing - ui' stamp: 'mt 7/13/2016 12:21'!
buttons

	^buttonRow submorphs! !

!DialogWindow methodsFor: 'accessing - ui' stamp: 'mt 7/13/2016 17:40'!
cancelButton
	^ cancelButton! !

!DialogWindow methodsFor: 'running' stamp: 'mt 7/13/2016 17:40'!
cancelDialog

	self cancelButton
		ifNil: [self closeDialog: nil]
		ifNotNil: [:btn | btn performAction].	! !

!DialogWindow methodsFor: 'running' stamp: 'mt 7/13/2016 18:01'!
closeDialog

	self selectedButton
		ifNil: [self closeDialog: nil]
		ifNotNil: [:btn | btn performAction].	! !

!DialogWindow methodsFor: 'running' stamp: 'mt 7/13/2016 17:37'!
closeDialog: returnValue
	result := returnValue.
	self delete.! !

!DialogWindow methodsFor: 'constructing' stamp: 'mt 7/22/2016 15:52'!
createAcceptButton

	^ self
		createButton: 'Accept' translated
		value: true
		color: (self userInterfaceTheme okColor ifNil: [Color r: 0.49 g: 0.749 b: 0.49])! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 8/1/2016 14:30'!
createBody

	| body |
	body := Morph new
		changeTableLayout;
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap;
		listDirection: #topToBottom;
		cellPositioning: #leftCenter;
		layoutInset: (10@5 corner: 10@10);
		cellInset: 5;
		color: Color transparent;
		yourself.
	body addAllMorphs: {self createMessage: ''. self createPane. self createButtonRow. self createFilter}.
	self addMorphBack: body.! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 7/21/2016 10:00'!
createButtonRow

	^ buttonRow := Morph new
		color: Color transparent;
		changeTableLayout;
		vResizing: #shrinkWrap;
		hResizing: #spaceFill;
		listDirection: #leftToRight;
		listCentering: #center;
		cellInset: 5;
		yourself! !

!DialogWindow methodsFor: 'constructing' stamp: 'mt 7/13/2016 17:59'!
createButton: buttonLabel

	^ self
		createButton: buttonLabel
		value: self buttonRow submorphs size + 1
		color: self defaultButtonColor! !

!DialogWindow methodsFor: 'constructing' stamp: 'mt 7/13/2016 17:54'!
createButton: buttonLabel value: buttonValue

	^ self
		createButton: buttonLabel
		value: buttonValue
		color: self defaultButtonColor! !

!DialogWindow methodsFor: 'constructing' stamp: 'mt 7/21/2016 09:16'!
createButton: buttonLabel value: buttonValue color: buttonColor

	| button |
	button := PluggableButtonMorphPlus new
		label: buttonLabel ;
		action: [ self closeDialog: buttonValue ] ;
		setProperty: #normalColor toValue: buttonColor ;
		setProperty: #normalLabel toValue: buttonLabel ;
		hResizing: #rigid;
		vResizing: #rigid;
		yourself.
	
	self deselectButton: button.
	buttonRow addMorphBack: button.
	self updateButtonExtent.
	
	^ button! !

!DialogWindow methodsFor: 'constructing' stamp: 'mt 7/22/2016 15:53'!
createCancelButton

	^ cancelButton := self
		createButton: 'Cancel' translated
		value: false
		color: (self userInterfaceTheme cancelColor ifNil: [Color r: 1 g: 0.6 b: 0.588])! !

!DialogWindow methodsFor: 'constructing' stamp: 'mt 7/13/2016 20:14'!
createCancelButton: label value: result

	^ cancelButton := self
		createButton: label
		value: result! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 8/1/2016 14:30'!
createFilter
	
	filterMorph := '' asText asMorph lock.
	filterMorph
		visible: false;
		disableTableLayout: true.
	^ filterMorph! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 7/22/2016 10:53'!
createMessage: aString 
	
	messageMorph := aString asText asMorph lock.
	self setMessageParameters.	
	^ messageMorph! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 7/14/2016 08:38'!
createPane

	^ paneMorph := BorderedMorph new
		changeProportionalLayout;
		hResizing: #rigid;
		vResizing: #rigid;
		layoutInset: 0;
		color: Color transparent;
		borderWidth: 0;
		yourself.! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 7/25/2016 10:04'!
createTitle: aString 
	"Mimick behavior of MenuMorph title creation."
	
	| box closeButton menuButton |
	box := Morph new
		name: #title;
		changeTableLayout;
		listDirection: #leftToRight;
		listCentering: #justified;
		yourself.
	
	titleMorph := aString asText asMorph lock.

	closeButton := SystemWindowButton new
		color: Color transparent;
		target: self;
		extent: 12@12;
		actionSelector: #cancelDialog;
		balloonText: 'Cancel this dialog' translated;
		borderWidth: 0;
		labelGraphic: SystemWindow closeBoxImage;
		extent: SystemWindow closeBoxImage extent;
		yourself.

	menuButton := SystemWindowButton new
		color: Color transparent;
		target: self;
		actionSelector: #offerDialogMenu;
		balloonText: 'Dialog menu' translated;
		borderWidth: 0;
		labelGraphic: SystemWindow menuBoxImage;
		extent: SystemWindow menuBoxImage extent;
		yourself.
	
	box addAllMorphs: {closeButton. titleMorph. menuButton}.
	
	self addMorphBack: box.
	self setTitleParameters.
! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/13/2016 17:51'!
defaultButtonColor

	^  self userInterfaceTheme buttonColor ifNil: [(Color r: 0.658 g: 0.678 b: 0.78) twiceLighter]! !

!DialogWindow methodsFor: 'selection' stamp: 'mt 7/22/2016 15:44'!
deselectButton: aButton

	aButton ifNil: [^ self].
	aButton offColor: (aButton valueOfProperty: #normalColor).
	aButton == selectedButton ifTrue: [selectedButton := nil].! !

!DialogWindow methodsFor: 'drawing' stamp: 'mt 8/1/2016 14:31'!
drawOverlayOn: aCanvas

	| title inset |
	super drawOverlayOn: aCanvas.

	title := self submorphs first.

	self wantsRoundedCorners ifTrue: [
		inset := (self class roundedDialogCorners and: [self class gradientDialog])
			"This check compensates a bug in balloon."
			ifTrue: [0@0 corner: 0@ -1] ifFalse: [self borderWidth @ 0].

		"Overdraw lower part of title bar to hide bottom corners."		
		aCanvas
			fillRectangle:( (title bottomLeft - (0 @ self submorphs first cornerRadius) corner: title bottomRight) insetBy: inset)
			color: self color].
		
	"Draw a line between the title and the contents."
	self borderWidth > 0 ifTrue: [			
		"Redraw the border all around. Needed because rounded borders do not align very well."
		self wantsRoundedCorners
			ifTrue: [	aCanvas frameRoundRect: self bounds radius: self cornerRadius width: self borderStyle width color: self borderStyle color]
			ifFalse: [aCanvas frameRectangle: self bounds width: self borderStyle width color: self borderStyle color]].! !

!DialogWindow methodsFor: 'selection' stamp: 'mt 8/1/2016 14:37'!
ensureSelectedButton

	self buttons ifEmpty: [^ self].
	self selectedButton ifNil: [self selectButton: self buttons first].
	self selectedButton enabled ifFalse: [self selectNextButton].! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/19/2016 13:55'!
exclusive

	^ exclusive! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/25/2016 09:44'!
exclusive: aBoolean

	exclusive := aBoolean.
	
	exclusive
		ifTrue: [self activeHand newMouseFocus: self]
		ifFalse: [self activeHand releaseMouseFocus: self].! !

!DialogWindow methodsFor: 'running' stamp: 'mt 7/25/2016 09:59'!
exploreInvocation

	| result context |
	self exclusive: false. "We want to explore."

	result := OrderedCollection new.
	context := thisContext.
	
	[context method selector = #getUserResponse]
		whileFalse: [context := context sender].
	
	[context sender] whileNotNil: [
		result add: context method.
		context := context sender].
	result add: context method.

	result explore.! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 8/1/2016 14:10'!
filter
	^ filter ifNil: ['']! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 8/1/2016 14:11'!
filterEnabled
	^ filterEnabled ifNil: [false]! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 8/1/2016 14:33'!
filterEnabled: aBoolean

	filterEnabled := aBoolean.
	self updateFilter.! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 8/1/2016 14:10'!
filter: aString
	filter := aString.
	self updateFilter.! !

!DialogWindow methodsFor: 'running' stamp: 'mt 7/19/2016 13:50'!
flash
	"Flash me"
	Beeper beepPrimitive.

	self class useWiggleAnimation ifTrue: [
		#(-2 4 -6 8 -4) do: [:i |
			self left: self left + i.
			self refreshWorld.
		] separatedBy: [(Delay forMilliseconds: 50) wait]
	] ifFalse: [
		1 to: 2 do:[:i|
			self color: Color black.
			self world doOneCycleNow.
			(Delay forMilliseconds: 50) wait.
			self color: Color white.
			self world doOneCycleNow.
			(Delay forMilliseconds: 50) wait] ]! !

!DialogWindow methodsFor: 'running' stamp: 'pre 8/4/2016 12:15'!
getUserResponse

	| hand world |
	(ProvideAnswerNotification signal: self message asString) 
		ifNotNil: [:answer| 
			^ answer = #default
				ifTrue: [result]
				ifFalse: [answer]].

	self message ifEmpty: [messageMorph delete]. "Do not waste space."
	self paneMorph submorphs ifEmpty: [self paneMorph delete]. "Do not waste space."

	hand := self currentHand.
	world := self currentWorld.
	
	self fullBounds.
	self center: preferredPosition.
	self bounds: (self bounds translatedToBeWithin: world bounds).
	self openInWorld: world.
	
	hand keyboardFocus in: [:priorKeyboardFocus |
		hand mouseFocus in: [:priorMouseFocus |
			self exclusive ifTrue: [hand newMouseFocus: self].
			hand newKeyboardFocus: self.

			[self isInWorld] whileTrue:[world doOneSubCycle].

			hand newKeyboardFocus: priorKeyboardFocus.
			self exclusive ifTrue: [
				hand newMouseFocus: priorMouseFocus]]].

	^ result! !

!DialogWindow methodsFor: 'running' stamp: 'mt 7/21/2016 09:16'!
getUserResponseAfter: seconds

	timeout := seconds + 1.
	
	self ensureSelectedButton.	
	self step.
	self updateButtonExtent.
	
	^ self getUserResponse! !

!DialogWindow methodsFor: 'running' stamp: 'mt 7/13/2016 20:02'!
getUserResponseAtHand

	^ self getUserResponseAtHand: ActiveHand! !

!DialogWindow methodsFor: 'running' stamp: 'mt 7/13/2016 21:00'!
getUserResponseAtHand: aHand

	self message ifEmpty: [messageMorph delete]. "Do not waste space."
	self paneMorph submorphs ifEmpty: [self paneMorph delete]. "Do not waste space."

	self moveSelectedButtonToHand: aHand.
	^ self getUserResponse! !

!DialogWindow methodsFor: 'events' stamp: 'mt 7/21/2016 09:57'!
handleMouseUp: event

	super handleMouseUp: event.
	self exclusive ifTrue: [event hand newMouseFocus: self].! !

!DialogWindow methodsFor: 'events' stamp: 'mt 7/13/2016 12:21'!
handlesKeyboard: evt

	^true! !

!DialogWindow methodsFor: 'events' stamp: 'mt 7/13/2016 17:27'!
handlesMouseDown: evt 

	^ true! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 7/14/2016 08:41'!
initialExtent

	^ 200@150! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 8/5/2016 09:39'!
initialize

	super initialize.
	
	self
		changeTableLayout;
		listDirection: #topToBottom;
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap;
		setProperty: #indicateKeyboardFocus toValue: #never.
	
	self createTitle: 'Dialog'.
	self createBody.
	
	self setDefaultParameters.
	
	keyMap := Dictionary new.
	exclusive := true.
	autoCancel := false.
	preferredPosition := ActiveWorld center.! !

!DialogWindow methodsFor: 'dropping/grabbing' stamp: 'mt 7/25/2016 14:14'!
justDroppedInto: aMorph event: event

	"Restore drop shadow if necessary."
	self hasDropShadow: Preferences menuAppearance3d.

	self exclusive ifTrue: [
		"aggressively preserve focus"
		event hand newMouseFocus: self].! !

!DialogWindow methodsFor: 'events' stamp: 'mt 8/1/2016 14:13'!
keyStroke: evt
	| char |
	self stopAutoTrigger.
	char := evt keyCharacter.

	char = Character escape ifTrue: [ ^ self cancelDialog ].
	(char = Character cr or: [char = Character enter]) ifTrue: [ ^ self closeDialog ].

	((char = Character arrowLeft or: [char = Character arrowUp])
		or: [ evt shiftPressed and: [ char = Character tab ] ])
			ifTrue: [ ^ self selectPreviousButton ].
	((char = Character arrowRight or: [char = Character arrowDown])
		or: [ char = Character tab ])
			ifTrue: [ ^ self selectNextButton ].
	
	self filterEnabled ifTrue: [
		char = Character backspace ifTrue: [self filter: (self filter ifEmpty: [''] ifNotEmpty: [:f | f allButLast])].
		(char isAlphaNumeric or: [char = Character space])
			ifTrue: [self filter: self filter, char asString]].
	
	keyMap
		at: char asLowercase
		ifPresent: [ : foundButton | foundButton performAction ]
		ifAbsent: [ "do nothing" ].! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/13/2016 17:15'!
message
	^messageMorph contents! !

!DialogWindow methodsFor: 'accessing - ui' stamp: 'mt 7/13/2016 18:24'!
messageMorph
	^ messageMorph! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/22/2016 10:53'!
message: aStringOrText

	messageMorph contents: aStringOrText.
	self setMessageParameters.! !

!DialogWindow methodsFor: 'events' stamp: 'mt 8/5/2016 09:41'!
mouseDown: event

	self stopAutoTrigger.

	"Always bring me to the front since I am modal"
	self comeToFront.
	
	(self containsPoint: event position) ifFalse: [
		^ self autoCancel
			ifTrue: [self cancelDialog]
			ifFalse: [self flash]].
	
	event hand 
		waitForClicksOrDrag: self 
		event: event 
		selectors: { nil. nil. nil. #startDrag: }
		threshold: HandMorph dragThreshold.! !

!DialogWindow methodsFor: 'events' stamp: 'mt 7/19/2016 13:02'!
mouseUp: event
	self stopAutoTrigger.
! !

!DialogWindow methodsFor: 'position' stamp: 'mt 7/13/2016 20:03'!
moveSelectedButtonToHand: aHand
	"Just let the user confirm the selected button without having to reposition the mouse."
	
	self ensureSelectedButton.
	self moveTo: self fullBounds center + (aHand position - self selectedButton center).! !

!DialogWindow methodsFor: 'position' stamp: 'mt 7/13/2016 17:06'!
moveToHand

	self moveToHand: self activeHand.! !

!DialogWindow methodsFor: 'position' stamp: 'mt 7/13/2016 20:01'!
moveToHand: aHand

	self moveTo: aHand position.! !

!DialogWindow methodsFor: 'position' stamp: 'mt 7/13/2016 19:46'!
moveTo: position

	preferredPosition := position.! !

!DialogWindow methodsFor: 'running' stamp: 'mt 7/25/2016 09:55'!
offerDialogMenu

	| menu |
	menu := MenuMorph new defaultTarget: self.
	menu
		add: (exclusive == true ifTrue: ['<yes>'] ifFalse: ['<no>']), 'be modally exclusive' translated
		action: #toggleExclusive;
		addLine;
		add: 'explore dialog invocation' translated
		action: #exploreInvocation.
		
	menu popUpEvent: self currentEvent in: self world.
	
	[menu isInWorld] whileTrue: [self world doOneSubCycle].
	self exclusive ifTrue: [self activeHand newMouseFocus: self].! !

!DialogWindow methodsFor: 'accessing - ui' stamp: 'mt 7/13/2016 18:24'!
paneMorph
	^ paneMorph! !

!DialogWindow methodsFor: 'events' stamp: 'mt 7/13/2016 12:21'!
processFocusEvent: evt using: dispatcher

	^ dispatcher dispatchFocusEventFully: evt with: self! !

!DialogWindow methodsFor: 'constructing' stamp: 'mt 8/5/2016 09:01'!
registerKeyboardShortcutFor: button 
	"Take the first alpha-numeric character that is not already used as a shortcut, and use it as a shortcut."
	
	(button valueOfProperty: #normalLabel) asString in: [:normalLabel | normalLabel do: [:char |
		char isAlphaNumeric ifTrue: [ keyMap
				at: char asLowercase
				ifPresent: [:ea | "Ignore"]
				ifAbsent: [
					button setProperty: #normalLabel toValue: ('{1} ({2})' format: {normalLabel. char asLowercase}).
					button label: (button valueOfProperty: #normalLabel).
					^ keyMap at: char asLowercase put: button ] ] ] ]! !

!DialogWindow methodsFor: 'constructing' stamp: 'mt 7/13/2016 18:47'!
registerKeyboardShortcuts
	
	self buttons do: [:ea | self registerKeyboardShortcutFor: ea].! !

!DialogWindow methodsFor: 'selection' stamp: 'mt 7/22/2016 15:44'!
selectButton: aButton

	| buttonColor |
	buttonColor :=  ((self userInterfaceTheme selectionModifier ifNil: [ [:c | c muchLighter] ]) value: (aButton valueOfProperty: #normalColor)).
	self deselectButton: selectedButton.
	aButton offColor: buttonColor.
	selectedButton := aButton.! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/13/2016 12:21'!
selectedButton
	^ selectedButton! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/13/2016 18:30'!
selectedButtonIndex
	
	^ self selectedButton
		ifNil: [0]
		ifNotNil: [:btn | self buttons indexOf: btn]! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/13/2016 18:33'!
selectedButtonIndex: anInteger

	anInteger = 0 ifTrue: [^ self selectedButton: nil].
	self selectedButton: (self buttons at: anInteger).! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/13/2016 17:24'!
selectedButton: aButton

	aButton
		ifNil: [self deselectButton: self selectedButton]
		ifNotNil: [self selectButton: aButton].! !

!DialogWindow methodsFor: 'selection' stamp: 'mt 8/1/2016 14:35'!
selectNextButton

	self selectedButton ifNil: [^ self].
	(self buttons anySatisfy: [:ea | ea enabled]) ifFalse: [^ self].
	
	self selectedButtonIndex: self selectedButtonIndex \\ self buttons size + 1.
	self selectedButton enabled ifFalse: [self selectNextButton].! !

!DialogWindow methodsFor: 'selection' stamp: 'mt 8/1/2016 14:35'!
selectPreviousButton

	self selectedButton ifNil: [^ self].
	(self buttons anySatisfy: [:ea | ea enabled]) ifFalse: [^ self].
	
	self selectedButtonIndex: self selectedButtonIndex - 2 \\ self buttons size + 1.
	self selectedButton enabled ifFalse: [self selectPreviousButton].! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 7/25/2016 14:36'!
setDefaultParameters
	"change the receiver's appareance parameters"

	self
		color: (self userInterfaceTheme color ifNil: [Color white]);
		borderStyle: (self userInterfaceTheme borderStyle ifNil: [BorderStyle default]);
		borderColor: (self userInterfaceTheme borderColor ifNil: [Color gray]);
		borderWidth: (self userInterfaceTheme borderWidth ifNil: [1]);
		layoutInset: ((self class roundedDialogCorners and: [self class gradientDialog])
			"This check compensates a bug in balloon."
			ifTrue: [0] ifFalse: [self borderWidth negated asPoint]).

	Preferences menuAppearance3d ifTrue: [self addDropShadow].! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 8/1/2016 14:40'!
setMessageParameters
	
	messageMorph ifNotNil: [
		| fontToUse colorToUse |
		fontToUse := self userInterfaceTheme font ifNil: [TextStyle defaultFont].
		colorToUse := self userInterfaceTheme textColor ifNil: [Color black].
	
		messageMorph contents
			addAttribute: (TextFontReference toFont: fontToUse);
			addAttribute: (TextColor color: colorToUse).
		messageMorph textColor: colorToUse.
		messageMorph releaseParagraph; changed].! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 7/25/2016 14:40'!
setTitleParameters

	(self submorphNamed: #title) ifNotNil: [:title |
		title
			fillStyle: (self class gradientDialog
				ifFalse: [SolidFillStyle color: (self userInterfaceTheme titleColor ifNil: [Color r: 0.658 g: 0.678 b: 0.78])]
				ifTrue: [self titleGradientFor: title from: (self userInterfaceTheme titleColor ifNil: [Color r: 0.658 g: 0.678 b: 0.78])]);
			borderStyle: (self userInterfaceTheme titleBorderStyle ifNil: [BorderStyle default]);
			borderColor: (self userInterfaceTheme titleBorderColor ifNil: [Color r: 0.6 g: 0.7 b: 1]);
			borderWidth: (self userInterfaceTheme titleBorderWidth ifNil: [0]);
			cornerStyle: (self wantsRoundedCorners ifTrue: [#rounded] ifFalse: [#square]);
			vResizing: #shrinkWrap;
			hResizing: #spaceFill;
			wrapCentering: #center;
			cellPositioning: #center;
			cellInset: 0;
			layoutInset: (5@3 corner: 5@ (2+(self wantsRoundedCorners ifFalse: [0] ifTrue: [self cornerRadius])))].
	
	titleMorph ifNotNil: [
		| fontToUse colorToUse |
		fontToUse := self userInterfaceTheme titleFont ifNil: [TextStyle defaultFont].
		colorToUse := self userInterfaceTheme titleTextColor ifNil: [Color black].
	
		titleMorph contents
			addAttribute: (TextFontReference toFont: fontToUse);
			addAttribute: (TextColor color: colorToUse).
		titleMorph releaseParagraph; changed].! !

!DialogWindow methodsFor: 'dropping/grabbing' stamp: 'mt 7/13/2016 17:32'!
startDrag: event

	self hasDropShadow: false.
	event hand grabMorph: self.! !

!DialogWindow methodsFor: 'stepping and presenter' stamp: 'mt 8/2/2016 09:20'!
step
	timeout ifNil: [^self].
	timeout = 0
		ifTrue: [
			self stopStepping.
			selectedButton performAction]
		ifFalse: [
			selectedButton label: ('{1} [{2}]' format: {
				selectedButton valueOfProperty: #normalLabel.
				timeout}).
			timeout := timeout - 1]! !

!DialogWindow methodsFor: 'stepping and presenter' stamp: 'mt 7/13/2016 12:21'!
stepTime
	^1000! !

!DialogWindow methodsFor: 'stepping and presenter' stamp: 'mt 7/13/2016 18:47'!
stopAutoTrigger
	timeout ifNil: [^self].
	timeout := nil.
	self stopStepping.
	selectedButton label: (selectedButton valueOfProperty: #normalLabel). ! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/13/2016 12:21'!
title
	^titleMorph contents! !

!DialogWindow methodsFor: 'initialization' stamp: 'mt 7/25/2016 14:39'!
titleGradientFor: morph from: aColor

	| cc gradient |
	cc := aColor.
	gradient := GradientFillStyle ramp: {
		0.0 -> Color white. 
		0.15 ->(cc mixed: 0.5 with: Color white). 
		0.7 -> cc.
	}.
	gradient origin: morph topLeft.
	gradient direction: 0 @ (TextStyle defaultFont height + 10).
	^ gradient! !

!DialogWindow methodsFor: 'accessing' stamp: 'mt 7/22/2016 10:45'!
title: aString

	titleMorph contents: aString asText.
	self setTitleParameters.! !

!DialogWindow methodsFor: 'running' stamp: 'mt 7/25/2016 09:43'!
toggleExclusive

	self exclusive: self exclusive not.! !

!DialogWindow methodsFor: 'updating' stamp: 'mt 7/21/2016 09:16'!
updateButtonExtent
	
	self updateButtonExtent: 20@10.! !

!DialogWindow methodsFor: 'updating' stamp: 'mt 7/22/2016 11:07'!
updateButtonExtent: margin

	"Update all button extents."
	(buttonRow submorphs collect: [:ea | ea minimumExtent]) max + margin in: [:preferredExtent |
		buttonRow submorphsDo: [:ea | ea extent: preferredExtent]].
	
	"See if horizontal button layout would be more appropriate."
	self flag: #magicNumber. "mt: Remove number with computation, maybe choose button font and 20 characters"
	(buttonRow submorphs collect: [:ea | ea fullBounds width]) sum > 400
		ifTrue: [buttonRow
					hResizing: #shrinkWrap;
					listDirection: #topToBottom;
					layoutInset: (buttonRow owner fullBounds width - (buttonRow owner layoutInset left*2) - buttonRow submorphs first fullBounds width // 2@0)]
		ifFalse: [buttonRow
					hResizing: #spaceFill;
					listDirection: #leftToRight;
					layoutInset: 0].! !

!DialogWindow methodsFor: 'updating' stamp: 'mt 8/1/2016 14:37'!
updateFilter

	self buttons do: [:ea |
		ea enabled: (self filter isEmpty or: [ea label asString includesSubstring: self filter caseSensitive: false])].
	
	filterMorph
		visible: self filter notEmpty;
		disableTableLayout: self filter isEmpty;
		contents: '<', self filter, '>';
		textColor: self messageMorph textColor.
		
	self ensureSelectedButton.! !

!DialogWindow methodsFor: 'updating' stamp: 'mt 7/21/2016 09:18'!
update: aspect

	aspect == #buttons
		ifTrue: [self updateButtonExtent].
		
	^ super update: aspect! !

!DialogWindow methodsFor: 'rounding' stamp: 'mt 7/25/2016 14:28'!
wantsRoundedCorners

	^ self class roundedDialogCorners or: [super wantsRoundedCorners]! !

!DialogWindow methodsFor: 'dropping/grabbing' stamp: 'mt 7/13/2016 12:21'!
wantsToBeDroppedInto: aMorph
	"Return true if it's okay to drop the receiver into aMorph"
	^aMorph isWorldMorph "only into worlds"! !

DialogWindow subclass: #FillInTheBlankMorph
	instanceVariableNames: 'response done textPane responseUponCancel'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!FillInTheBlankMorph class methodsFor: 'default constants' stamp: 'mt 4/14/2015 18:23'!
defaultAnswerExtent
	^  (200@ (3 * Preferences standardDefaultTextFont height))! !

!FillInTheBlankMorph class methodsFor: 'instance creation' stamp: 'bolot 5/18/2000 13:57'!
requestPassword: queryString
	"Create an instance of me whose question is queryString. Invoke it centered at the cursor, and answer the string the user accepts. Answer the empty string if the user cancels."
	"use password font"
	"FillInTheBlankMorph requestPassword: 'Password?'"

	^ self
		requestPassword: queryString
		initialAnswer: ''
		centerAt: Sensor cursorPoint
		inWorld: World
		onCancelReturn: ''
		acceptOnCR: true
! !

!FillInTheBlankMorph class methodsFor: 'instance creation' stamp: 'mt 7/13/2016 20:49'!
requestPassword: queryString initialAnswer: defaultAnswer centerAt: aPoint inWorld: aWorld onCancelReturn: returnOnCancel acceptOnCR: acceptBoolean
	"Create an instance of me whose question is queryString with the given initial answer. Invoke it centered at the given point, and answer the string the user accepts.   If the user cancels, answer returnOnCancel."
	"FillInTheBlankMorph
		request: 'Type something, then type CR.'
		initialAnswer: 'yo ho ho!!'
		centerAt: Display center"

	| aFillInTheBlankMorph |
	aFillInTheBlankMorph := self new
		setPasswordQuery: queryString
		initialAnswer: defaultAnswer
		answerHeight: 50
		acceptOnCR: acceptBoolean.

	aFillInTheBlankMorph createAcceptButton
		action: [aFillInTheBlankMorph textPane accept].
	aFillInTheBlankMorph createCancelButton
		action: [aFillInTheBlankMorph closeDialog: returnOnCancel].
		
	aFillInTheBlankMorph moveTo: aPoint.
	^ aFillInTheBlankMorph getUserResponse! !

!FillInTheBlankMorph class methodsFor: 'instance creation' stamp: 'ar 3/18/2001 00:54'!
request: queryString
	"Create an instance of me whose question is queryString. Invoke it centered at the cursor, and answer the string the user accepts. Answer the empty string if the user cancels."
	"FillInTheBlankMorph request: 'What is your favorite color?'"

	^ self
		request: queryString
		initialAnswer: ''
		centerAt: (ActiveHand ifNil:[Sensor]) cursorPoint! !

!FillInTheBlankMorph class methodsFor: 'instance creation' stamp: 'ar 3/18/2001 00:54'!
request: queryString initialAnswer: defaultAnswer 
	"Create an instance of me whose question is queryString with the given initial answer. Invoke it centered at the given point, and answer the string the user accepts. Answer the empty string if the user cancels."
	"FillInTheBlankMorph
		request: 'What is your favorite color?'
		initialAnswer: 'red, no blue. Ahhh!!'"

	^ self
		request: queryString
		initialAnswer: defaultAnswer
		centerAt: ActiveHand cursorPoint! !

!FillInTheBlankMorph class methodsFor: 'instance creation' stamp: 'ar 3/17/2001 23:43'!
request: queryString initialAnswer: defaultAnswer centerAt: aPoint
	"Create an instance of me whose question is queryString with the given initial answer. Invoke it centered at the given point, and answer the string the user accepts. Answer the empty string if the user cancels.
	This variant is only for calling from within a Morphic project."
	"FillInTheBlankMorph
		request: 'Type something, then type CR.'
		initialAnswer: 'yo ho ho!!'
		centerAt: Display center"

	 ^ self 
		request: queryString 
		initialAnswer: defaultAnswer 
		centerAt: aPoint 
		inWorld: ActiveWorld
! !

!FillInTheBlankMorph class methodsFor: 'instance creation' stamp: 'sw 1/31/2000 11:03'!
request: queryString initialAnswer: defaultAnswer centerAt: aPoint inWorld: aWorld
	"Create an instance of me whose question is queryString with the given initial answer. Invoke it centered at the given point, and answer the string the user accepts.  Answer the empty string if the user cancels."
	"FillInTheBlankMorph
		request: 'Type something, then type CR.'
		initialAnswer: 'yo ho ho!!'
		centerAt: Display center"

	^ self request: queryString initialAnswer: defaultAnswer centerAt: aPoint inWorld: aWorld onCancelReturn: ''! !

!FillInTheBlankMorph class methodsFor: 'instance creation' stamp: 'sw 2/2/2000 22:43'!
request: queryString initialAnswer: defaultAnswer centerAt: aPoint inWorld: aWorld onCancelReturn: returnOnCancel
	"Create an instance of me whose question is queryString with the given initial answer. Invoke it centered at the given point, and answer the string the user accepts.   If the user cancels, answer returnOnCancel.  If user hits cr, treat it as a normal accept."

	"FillInTheBlankMorph
		request: 'Type something, then type CR.'
		initialAnswer: 'yo ho ho!!'
		centerAt: Display center"

	^ self request: queryString initialAnswer: defaultAnswer centerAt: aPoint inWorld: aWorld onCancelReturn: returnOnCancel acceptOnCR: true! !

!FillInTheBlankMorph class methodsFor: 'instance creation' stamp: 'NS 8/1/2000 11:44'!
request: queryString initialAnswer: defaultAnswer centerAt: aPoint inWorld: aWorld onCancelReturn: returnOnCancel acceptOnCR: acceptBoolean
	"Create an instance of me whose question is queryString with the given initial answer. Invoke it centered at the given point, and answer the string the user accepts.   If the user cancels, answer returnOnCancel."

	^ self request: queryString initialAnswer: defaultAnswer centerAt: aPoint 
		inWorld: aWorld onCancelReturn: returnOnCancel 
		acceptOnCR: acceptBoolean answerExtent: self defaultAnswerExtent! !

!FillInTheBlankMorph class methodsFor: 'instance creation' stamp: 'mt 7/13/2016 20:48'!
request: queryString initialAnswer: defaultAnswer centerAt: aPoint inWorld: aWorld onCancelReturn: returnOnCancel acceptOnCR: acceptBoolean answerExtent: answerExtent
	"Create an instance of me whose question is queryString with the given initial answer. Invoke it centered at the given point, and answer the string the user accepts.   If the user cancels, answer returnOnCancel."
	"FillInTheBlankMorph
		request: 'Type something, then type CR.'
		initialAnswer: 'yo ho ho!!'
		centerAt: Display center"

	| aFillInTheBlankMorph |
	aFillInTheBlankMorph := self new
		setQuery: queryString
		initialAnswer: defaultAnswer
		answerExtent: answerExtent
		acceptOnCR: acceptBoolean.

	aFillInTheBlankMorph createAcceptButton
		action: [aFillInTheBlankMorph textPane accept].
	aFillInTheBlankMorph createCancelButton
		action: [aFillInTheBlankMorph closeDialog: returnOnCancel].
		
	aFillInTheBlankMorph moveTo: aPoint.
	^ aFillInTheBlankMorph getUserResponse
! !

!FillInTheBlankMorph class methodsFor: 'instance creation' stamp: 'NS 8/1/2000 11:43'!
request: queryString initialAnswer: defaultAnswer centerAt: aPoint inWorld: aWorld onCancelReturn: returnOnCancel acceptOnCR: acceptBoolean answerHeight: answerHeight
	"Create an instance of me whose question is queryString with the given initial answer. Invoke it centered at the given point, and answer the string the user accepts.   If the user cancels, answer returnOnCancel."
	^ self request: queryString initialAnswer: defaultAnswer centerAt: aPoint 
		inWorld: aWorld onCancelReturn: returnOnCancel acceptOnCR: acceptBoolean 
		answerExtent: self defaultAnswerExtent x @ answerHeight! !

!FillInTheBlankMorph methodsFor: 'menu' stamp: 'ar 10/5/2009 23:24'!
codePaneMenu: aMenu shifted: shifted

	^ StringHolder codePaneMenu: aMenu shifted: shifted.
! !

!FillInTheBlankMorph methodsFor: 'initialization' stamp: 'mt 7/26/2016 11:15'!
createTextPaneAcceptOnCR: acceptBoolean 

	textPane := PluggableTextMorph
				on: self
				text: #response
				accept: #response:
				readSelection: #selectionInterval
				menu: #codePaneMenu:shifted:.
	textPane 
		showScrollBarsOnlyWhenNeeded;
		wantsFrameAdornments: false;
		hasUnacceptedEdits: true;
		acceptOnCR: acceptBoolean;
		setNameTo: 'textPane';
		layoutFrame: (LayoutFrame fractions: (0@0 corner: 1@1));
		hResizing: #spaceFill;
		vResizing: #spaceFill.

	^ textPane! !

!FillInTheBlankMorph methodsFor: 'events' stamp: 'mt 7/19/2016 13:09'!
filterEvent: event for: morph

	(event isKeystroke and: [event keyCharacter = Character escape])
		ifTrue: [event ignore. self cancelDialog].
		
	^ event! !

!FillInTheBlankMorph methodsFor: 'initialization' stamp: 'mt 7/19/2016 13:08'!
initialize

	super initialize.
	self addKeyboardCaptureFilter: self.! !

!FillInTheBlankMorph methodsFor: 'accessing' stamp: 'mt 7/13/2016 20:26'!
response

	^ result
! !

!FillInTheBlankMorph methodsFor: 'accessing' stamp: 'mt 7/13/2016 20:41'!
response: aText
	"Sent when text pane accepts."

	result := aText asString.
	self delete.
! !

!FillInTheBlankMorph methodsFor: 'accessing' stamp: 'mt 7/13/2016 20:38'!
selectionInterval
	^ 1 to: result size
! !

!FillInTheBlankMorph methodsFor: 'initialization' stamp: 'mt 7/14/2016 08:01'!
setDefaultParameters

	super setDefaultParameters.
	textPane ifNotNil: [:tp | tp borderColor: self borderColor].! !

!FillInTheBlankMorph methodsFor: 'initialization' stamp: 'mt 7/10/2016 15:22'!
setPasswordQuery: queryString initialAnswer: initialAnswer answerHeight: answerHeight acceptOnCR: acceptBoolean

	self setQuery: queryString 
		initialAnswer: initialAnswer 
		answerHeight: answerHeight 
		acceptOnCR: acceptBoolean.
	textPane font: (StrikeFont passwordFontSize: 12).! !

!FillInTheBlankMorph methodsFor: 'initialization' stamp: 'mt 7/14/2016 08:00'!
setQuery: queryString initialAnswer: initialAnswer answerExtent: answerExtent acceptOnCR: acceptBoolean 
	
	| text |
	
	result := initialAnswer.
	done := false.

	self paneMorph removeAllMorphs.

	self title: 'Input Requested'.
	self message: queryString.
	
	text := self createTextPaneAcceptOnCR: acceptBoolean.
	self paneMorph addMorphBack: text.

	self paneMorph extent: ((initialAnswer asText asMorph extent + (20@10) max: answerExtent) min: 500@500).	
	self setDefaultParameters.! !

!FillInTheBlankMorph methodsFor: 'initialization' stamp: 'NS 8/1/2000 11:44'!
setQuery: queryString initialAnswer: initialAnswer answerHeight: answerHeight acceptOnCR: acceptBoolean
	self setQuery: queryString initialAnswer: initialAnswer 
		answerExtent: (self class defaultAnswerExtent x @ answerHeight) 
		acceptOnCR: acceptBoolean
! !

!FillInTheBlankMorph methodsFor: 'accessing' stamp: 'mt 7/13/2016 20:47'!
textPane
	^ textPane! !

DialogWindow subclass: #UserDialogBoxMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!UserDialogBoxMorph commentStamp: 'ar 12/11/2009 22:33' prior: 0!
A DialogBoxMorph is Morph used in simple yes/no/confirm dialogs. Strongly modal.!

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'mt 7/20/2016 13:52'!
confirm: aString
	"UserDialogBoxMorph confirm: 'Do you like chocolate?'"
	^self confirm: aString title: 'Please Confirm'! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'ar 12/12/2009 18:42'!
confirm: aString orCancel: cancelBlock
	"UserDialogBoxMorph confirm: 'Do you like chocolate?'"
	^self confirm: aString orCancel: cancelBlock at: nil! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'mt 7/20/2016 13:53'!
confirm: aString orCancel: cancelBlock at: aPointOrNil
	^self
		confirm: aString
		orCancel: cancelBlock
		title: 'Please Confirm'
		at: aPointOrNil! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'mt 8/2/2016 09:16'!
confirm: aString orCancel: cancelBlock title: titleString at: aPointOrNil
	
	(self new
		title: titleString;
		message: aString;
		createButton: 'Yes' translated value: true;
		createButton: 'No' translated  value: false;
		createButton: 'Cancel' translated value: nil;
		selectedButtonIndex: 1; "YES"
		registerKeyboardShortcuts;
		yourself) in: [:dialog |
			^ (aPointOrNil
				ifNil: [dialog getUserResponseAtHand]
				ifNotNil: [
					dialog moveTo: aPointOrNil.
					dialog getUserResponse])
						ifNil: [ cancelBlock value ]]! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'ar 12/12/2009 18:42'!
confirm: aString title: titleString
	"UserDialogBoxMorph confirm: 'Make your choice carefully' withCRs title: 'Do you like chocolate?'"
	^self confirm: aString title: titleString at: nil! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'mt 8/2/2016 09:17'!
confirm: aString title: titleString at: aPointOrNil
	"UserDialogBoxMorph confirm: 'Make your choice carefully' withCRs title: 'Do you like chocolate?'"
	^self new
		title: titleString;
		message: aString;
		createButton: 'Yes' translated value: true;
		createCancelButton: 'No' translated  value: false;
		selectedButtonIndex: 1; "YES"
		registerKeyboardShortcuts;
		getUserResponseAtHand! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'ar 1/23/2010 14:37'!
confirm: aString title: titleString trueChoice: trueChoice falseChoice: falseChoice
	"UserDialogBoxMorph confirm: 'Make your choice carefully' withCRs title: 'Do you like chocolate?' trueChoice: 'Oh yessir!!' falseChoice: 'Not so much...'  "
	^self confirm: aString title: titleString trueChoice: trueChoice falseChoice: falseChoice at: nil! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'mt 8/2/2016 09:17'!
confirm: aString title: titleString trueChoice: trueChoice falseChoice: falseChoice at: aPointOrNil
	"UserDialogBoxMorph confirm: 'Make your choice carefully' withCRs title: 'Do you like chocolate?' trueChoice: 'Oh yessir!!' falseChoice: 'Not so much...'"
	^self new
		title: titleString;
		message: aString;
		createButton: trueChoice translated value: true;
		createCancelButton: falseChoice translated value: false;
		selectedButtonIndex: 1;
		registerKeyboardShortcuts;
		moveTo: (aPointOrNil ifNil: [ActiveWorld center]);
		getUserResponse! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'mt 8/2/2016 09:16'!
confirm: aString title: titleString trueChoice: trueChoice falseChoice: falseChoice default: default triggerAfter: seconds at: aPointOrNil
	"UserDialogBoxMorph confirm: 'I like hot java' title: 'What do you say?' trueChoice: 'You bet!!' falseChoice: 'Nope' default: false triggerAfter: 12 at: 121@212"
	^self new
		title: titleString;
		message: aString;
		createButton: trueChoice translated value: true;
		createCancelButton: falseChoice translated value: false;
		selectedButtonIndex: (default ifTrue: [1] ifFalse: [2]);
		registerKeyboardShortcuts;
		moveTo: (aPointOrNil ifNil: [ActiveWorld center]);
		getUserResponseAfter: seconds! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'ar 1/23/2010 14:37'!
confirm: aString trueChoice: trueChoice falseChoice: falseChoice
	"UserDialogBoxMorph confirm: 'Do you like chocolate?' trueChoice: 'Oh yessir!!' falseChoice: 'Not so much...'  "
	^self confirm: aString title: 'Please confirm:' trueChoice: trueChoice falseChoice: falseChoice at: nil! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'mt 7/20/2016 13:53'!
inform: aString
	"UserDialogBoxMorph inform: 'Squeak is great!!'"
	^self inform: aString title: 'Note' translated! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'ar 12/12/2009 18:42'!
inform: aString title: titleString
	"UserDialogBoxMorph inform: 'Squeak is great!!' title: 'Will you look at this:'"
	^self inform: aString title: titleString at: nil! !

!UserDialogBoxMorph class methodsFor: 'utilities' stamp: 'mt 7/13/2016 21:02'!
inform: aString title: titleString at: aPointOrNil
	"UserDialogBoxMorph inform: 'Squeak is great!!' title: 'Will you look at this:'"
	
	^self new
		title: titleString;
		message: aString;
		createButton: 'OK' translated value: nil;
		getUserResponseAtHand! !

!UserDialogBoxMorph methodsFor: 'accessing' stamp: 'mt 7/13/2016 20:12'!
label
	^ self message! !

!UserDialogBoxMorph methodsFor: 'accessing' stamp: 'mt 7/13/2016 20:12'!
label: aString
	self message: aString.! !

Morph subclass: #DockingBarMorph
	instanceVariableNames: 'originalColor gradientRamp fillsOwner avoidVisibleBordersAtEdge autoGradient selectedItem activeSubMenu oldKeyboardFocus oldMouseFocus'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Menus-DockingBar'!

!DockingBarMorph class methodsFor: 'scripting' stamp: 'dgd 8/31/2004 14:26'!
defaultNameStemForInstances
	^ 'DockingBar'! !

!DockingBarMorph class methodsFor: 'samples' stamp: 'dgd 9/1/2004 19:52'!
example1
	" 
	DockingBarMorph example1.  
	 
	(Color lightBlue wheel: 4) do:[:c | DockingBarMorph example1 
	color: c; borderColor: c twiceDarker]. 
	 
	World deleteDockingBars.
	"
	| instance |
	instance := DockingBarMorph new.
	""
	instance addSpace: 10.
	instance
		addMorphBack: (ClockMorph new show24hr: true).
	instance addSpacer.
	instance
		addMorphBack: (ClockMorph new show24hr: true).
	instance addSpace: 10.
	""
	instance adhereToTop.
	""
	instance autoGradient: true.
	instance layoutInset: 10.
	""
	^ instance openInWorld! !

!DockingBarMorph class methodsFor: 'samples' stamp: 'dgd 9/22/2004 18:53'!
example2
	" 
	DockingBarMorph example2.  
	World deleteDockingBars.  
	"
	| menu |
	menu := DockingBarMorph new.
	""
	menu addSpace: 10.
	menu
		add: 'Squeak'
		icon: MenuIcons smallConfigurationIcon
		subMenu: self squeakMenu.
	menu
		add: 'Configuration'
		icon: MenuIcons smallWindowIcon
		subMenu: self squeakMenu.
	menu addSpace: 10.
	menu addLine.
	menu addSpace: 10.
	menu
		addMorphBack: (ImageMorph new image: MenuIcons smallBackIcon).
	menu addSpace: 10.
	menu
		addMorphBack: (ImageMorph new image: MenuIcons smallForwardIcon).
	menu addSpace: 10.
	menu addLine.
	menu addSpacer.
	""
	menu addMorphBack: ProjectNavigationMorph new speakerIcon.
	""
	menu addSpace: 10.
	menu
		addMorphBack: (ClockMorph new show24hr: true).
	menu addSpace: 10.
	""
	menu adhereToTop.
	""
	menu autoGradient: true.
	""
	^ menu openInWorld! !

!DockingBarMorph class methodsFor: 'samples' stamp: 'dgd 9/1/2004 19:55'!
example3
	" 
	DockingBarMorph example3. 
	World deleteDockingBars.
	"
	(Color lightBlue wheel: 4)
		with: #(#top #bottom #left #right )
		do: [:col :edge | 
			| instance | 
			instance := DockingBarMorph example1.
			instance adhereTo: edge.
			instance color: col.
			instance borderColor: col twiceDarker]! !

!DockingBarMorph class methodsFor: 'samples' stamp: 'tpr 4/9/2014 15:09'!
squeakMenu
	| menu |
	menu := DockingBarMenuMorph new defaultTarget: self.
	menu
		add: 'Hello'
		target: self
		selector: #inform:
		argument: 'Hello World!!'.
	menu
		add: 'Long Hello'
		target: self
		selector: #inform:
		argument: 'Helloooo World!!'.
	menu
		add: 'A very long Hello'
		target: self
		selector: #inform:
		argument: 'Hellooooooooooooooo World!!'.
	menu
		add: 'An incredible long Hello'
		target: self
		selector: #inform:
		argument: 'Hellooooooooooooooooooooooo World!!'.
	^ menu! !

!DockingBarMorph class methodsFor: 'preferences' stamp: 'mt 7/10/2016 12:31'!
themeProperties

	^ super themeProperties, {
		{ #borderColor. 'Colors'. 'Color of the menu''s border.' }.
		{ #borderWidth. 'Geometry'. 'Width of the menu''s border.' }.
		{ #borderStyle. 'Borders'. 'Whether to use a plain border, inset, or outset.' }.
		{ #color. 'Colors'. 'Background color of the menu.' }.

		{ #lineColor. 'Colors'. 'Color of the separators between menu items.' }.
		{ #lineStyle. 'Colors'. 'Use border-style to change appearance.' }.
		{ #lineWidth. 'Geometry'. 'How big the separators should be.' }.
	}! !

!DockingBarMorph methodsFor: 'dropping/grabbing' stamp: 'dgd 9/7/2004 14:47'!
aboutToBeGrabbedBy: aHand 
	"The morph is about to be grabbed, make it float"
	self beFloating.
	self updateBounds.
	self updateColor.
	(self bounds containsPoint: aHand position)
		ifFalse: [self center: aHand position].
self owner restoreFlapsDisplay! !

!DockingBarMorph methodsFor: 'events' stamp: 'kb 2/22/2010 15:09'!
activate: evt 
	"Receiver should be activated; e.g., so that control passes  
	correctly."
	
	oldKeyboardFocus := evt hand keyboardFocus.
	self oldMouseFocus: evt hand mouseFocus.
	evt hand 
		newKeyboardFocus: self;
		newMouseFocus: self.
	self ensureSelectedItem: evt! !

!DockingBarMorph methodsFor: 'control' stamp: 'mt 2/27/2015 13:25'!
activeSubmenu: aSubmenu 
	activeSubMenu isNil
		ifFalse: [activeSubMenu delete].
	activeSubMenu := aSubmenu.
	aSubmenu isNil
		ifTrue: [^ self].
	activeSubMenu updateMenu.
	activeSubMenu selectItem: nil event: nil.
	MenuIcons decorateMenu: activeSubMenu.
	activeSubMenu 
		activatedFromDockingBar: self;
		beSticky;
		resistsRemoval: true;
		removeMatchString! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 9/13/2004 19:59'!
addBlankIconsIfNecessary: anIcon 
	"If any of my items have an icon, ensure that all do by using 
	anIcon for those that don't"
	self items
		reject: [:each | each hasIconOrMarker]
		thenDo: [:each | each icon: anIcon]! !

!DockingBarMorph methodsFor: 'menu' stamp: 'dgd 9/29/2004 20:31'!
addCustomMenuItems: aMenu hand: aHandMorph 
	"Populate aMenu with appropriate menu items for a  
	yellow-button (context menu) click."
	super addCustomMenuItems: aMenu hand: aHandMorph.
	""
	aMenu addLine.
	aMenu addUpdating: #autoGradientString action: #toggleAutoGradient.
	self isFloating
		ifFalse: [""
			aMenu addUpdating: #fillsOwnerString action: #toggleFillsOwner.
			aMenu addUpdating: #avoidVisibleBordersAtEdgeString action: #toggleAvoidVisibleBordersAtEdge]! !

!DockingBarMorph methodsFor: 'construction' stamp: 'dgd 7/28/2005 12:08'!
addDefaultSpace
	"Add a new space of the given size to the receiver."
	^ self addSpace: (Preferences tinyDisplay ifFalse:[10] ifTrue:[3])! !

!DockingBarMorph methodsFor: 'construction' stamp: 'kb 11/19/2009 23:57'!
addItem: aBlock
	| item |
	item := DockingBarItemMorph new.
	aBlock value: item.
	self addMorphBack: item! !

!DockingBarMorph methodsFor: 'construction' stamp: 'mt 7/9/2016 13:05'!
addLine
	"Append a divider line to this menu. Suppress duplicate lines."

	| colorToUse |
	submorphs isEmpty ifTrue: [^ self].
	self lastSubmorph knownName = #line ifTrue: [^ self].
	
	colorToUse := self userInterfaceTheme lineColor ifNil: [Color gray: 0.9].
	self addMorphBack: (Morph new
		color: colorToUse;
		extent: (self userInterfaceTheme lineWidth ifNil: [2]) asPoint;
		borderStyle: (self userInterfaceTheme lineStyle ifNil: [BorderStyle inset]);
		borderColor: colorToUse;
		borderWidth: 1;
		name: #line; "see above"
		yourself).! !

!DockingBarMorph methodsFor: 'construction' stamp: 'dgd 8/31/2004 11:34'!
addSpacer
	"Add a new spacer to the receiver. 
	 
	Spacer are objects that try to use as much space as they can"
	self
		addMorphBack: (AlignmentMorph newSpacer: Color transparent)! !

!DockingBarMorph methodsFor: 'construction' stamp: 'dgd 8/31/2004 11:34'!
addSpace: sizePointOrNumber 
	"Add a new space of the given size to the receiver."
	| space |
	space := RectangleMorph new.
	space extent: sizePointOrNumber asPoint.
	space color: Color transparent.
	space borderWidth: 0.
	self addMorphBack: space! !

!DockingBarMorph methodsFor: 'construction' stamp: 'kb 11/19/2009 17:31'!
add: wordingString icon: aForm help: helpString subMenu: aMenuMorph 
	"Append the given submenu with the given label."
	| item |
	item := DockingBarItemMorph new.
	item contents: wordingString.
	item subMenu: aMenuMorph.
	item icon: aForm.
	helpString isNil
		ifFalse: [item setBalloonText: helpString].
	self addMorphBack: item! !

!DockingBarMorph methodsFor: 'construction' stamp: 'kb 11/19/2009 18:20'!
add: wordingString icon: aForm selectedIcon: anotherForm help: helpString subMenu: aMenuMorph 
	"Append the given submenu with the given label."
	| item |
	item := DockingBarItemMorph new
		contents: wordingString;
		subMenu: aMenuMorph;
		icon: aForm;
		selectedIcon: anotherForm.
	helpString isNil ifFalse: [
		item setBalloonText: helpString ].
	self addMorphBack: item! !

!DockingBarMorph methodsFor: 'construction' stamp: 'dgd 9/10/2004 16:48'!
add: wordingString icon: aForm subMenu: aMenuMorph 
	"Append the given submenu with the given label."
^ self add: wordingString icon: aForm help: nil subMenu: aMenuMorph ! !

!DockingBarMorph methodsFor: 'construction' stamp: 'dgd 9/1/2004 19:08'!
add: aString subMenu: aMenuMorph 
	"Append the given submenu with the given label."
	self add: aString icon: nil subMenu: aMenuMorph ! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 12:01'!
adhereToBottom
	"Instract the receiver to adhere to bottom"
	 self adhereTo:#bottom! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 12:01'!
adhereToLeft
	"Instract the receiver to adhere to left"
	self adhereTo: #left! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 12:01'!
adhereToRight
	"Instract the receiver to adhere to right"
	self adhereTo: #right! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 12:01'!
adhereToTop
	"Instract the receiver to adhere to top"
	self adhereTo: #top! !

!DockingBarMorph methodsFor: 'private - accessing' stamp: 'dgd 8/31/2004 14:35'!
adhereTo: edgeSymbol 
	"Private - Instruct the receiver to adhere to the given edge.  
	 
	Options: #left #top #right #bottom or #none"
	""
	(#(#left #top #right #bottom #none ) includes: edgeSymbol)
		ifFalse: [^ self error: 'invalid option'].
	""
	self setToAdhereToEdge: edgeSymbol.
	self updateLayoutProperties.
	self updateColor! !

!DockingBarMorph methodsFor: 'update' stamp: 'mt 7/26/2016 12:00'!
applyUserInterfaceTheme

	| colorToUse |
	gradientRamp := nil.
	
	super applyUserInterfaceTheme.
	
	self setDefaultParameters.
	
	"Update properties of separating lines."
	colorToUse := self userInterfaceTheme lineColor ifNil: [Color gray: 0.9].
	self submorphs
		select: [:ea | ea knownName = #line]
		thenDo: [:line |
			line
				color: colorToUse;
				extent: (self userInterfaceTheme lineWidth ifNil: [2]) asPoint;
				borderStyle: (self userInterfaceTheme lineStyle ifNil: [BorderStyle inset]);
				borderColor: colorToUse].! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 14:28'!
autoGradient
	"Answer if the receiver is in autoGradient mode"
	^ autoGradient! !

!DockingBarMorph methodsFor: 'menu' stamp: 'dgd 9/29/2004 20:25'!
autoGradientString
	"Answer the string to be shown in a menu to represent the  
	'resistsRemoval' status"
	^ (self autoGradient
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'auto gradient' translated! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 14:28'!
autoGradient: aBoolean 
	"Instruct the receiver to fill the owner or not"
	autoGradient := aBoolean.
	self updateColor! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 13:38'!
avoidVisibleBordersAtEdge
"Answer if the receiver is in avoidVisibleBordersAtEdge mode"
	^ avoidVisibleBordersAtEdge! !

!DockingBarMorph methodsFor: 'menu' stamp: 'dgd 9/29/2004 20:25'!
avoidVisibleBordersAtEdgeString
	"Answer the string to be shown in a menu to represent the  
	'resistsRemoval' status"
	^ (self avoidVisibleBordersAtEdge
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'avoid visible borders at edge' translated! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 13:57'!
avoidVisibleBordersAtEdge: aBoolean 
	"Instruct the receiver to avoid showing the borders at edge"
	avoidVisibleBordersAtEdge := aBoolean.
self updateLayoutProperties.! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 12:02'!
beFloating
	"Instract the receiver to be floating"
	self adhereTo: #none! !

!DockingBarMorph methodsFor: 'meta-actions' stamp: 'dgd 9/1/2004 22:11'!
blueButtonDown: anEvent 
	"Special gestures (cmd-mouse on the Macintosh; Alt-mouse on  
	Windows and Unix) allow a mouse-sensitive morph to be  
	moved or bring up a halo for the morph."
	"In NoviceMode we don't want halos"
	
	Preferences noviceMode 
	ifFalse: [super blueButtonDown: anEvent]
	! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'mt 7/26/2016 12:51'!
color: aColor 
	"Set the receiver's color."

	super color: aColor.
	originalColor := aColor asColor.
	gradientRamp := nil.
""
self updateColor! !

!DockingBarMorph methodsFor: 'events' stamp: 'kb 5/10/2016 21:27'!
deactivate: evt 

	| newKeyboardFocus newMouseFocus |
	self selectItem: nil event: evt.
	newKeyboardFocus := self oldKeyboardFocus.
	newMouseFocus := self oldMouseFocus.
	oldKeyboardFocus := oldMouseFocus := nil.
	evt hand
		newKeyboardFocus: newKeyboardFocus;
		newMouseFocus: newMouseFocus! !

!DockingBarMorph methodsFor: 'submorphs-add/remove' stamp: 'kb 2/22/2010 10:42'!
delete

	ActiveHand removeKeyboardListener: self.
	activeSubMenu
		ifNotNil: [activeSubMenu delete].
	^ super delete! !

!DockingBarMorph methodsFor: 'control' stamp: 'kb 2/22/2010 15:54'!
deleteIfPopUp: evt 
	
	evt ifNotNil: [
		evt hand releaseMouseFocus: self ]! !

!DockingBarMorph methodsFor: 'private - accessing' stamp: 'dgd 8/31/2004 13:56'!
edgeToAdhereTo
	"private - answer the edge where the receiver is adhering to"
	^ self
		valueOfProperty: #edgeToAdhereTo
		ifAbsent: [#none]! !

!DockingBarMorph methodsFor: 'events' stamp: 'kb 2/22/2010 15:09'!
ensureSelectedItem: evt
	
	self selectedItem ifNil: [
		self 
			selectItem: (
				self submorphs 
					detect: [ :each | each isKindOf: DockingBarItemMorph ] 
					ifNone: [ ^self ]) 
			event: evt ]! !

!DockingBarMorph methodsFor: 'geometry' stamp: 'ms 2/8/2013 13:40:43.628'!
extent: aPoint 
	"change the receiver's extent"
	(bounds extent closeTo: aPoint) ifTrue: [^ self].
	super extent: aPoint.
	self updateColor! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 13:35'!
fillsOwner
	"Answer if the receiver is in fillOwner mode"
	^ fillsOwner! !

!DockingBarMorph methodsFor: 'menu' stamp: 'dgd 9/29/2004 20:23'!
fillsOwnerString
	"Answer the string to be shown in a menu to represent the  
	'resistsRemoval' status"
	^ (self fillsOwner
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'fills owner' translated
! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 13:57'!
fillsOwner: aBoolean 
	"Instruct the receiver to fill the owner or not"
	fillsOwner := aBoolean.
self updateLayoutProperties! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'mt 6/9/2016 15:08'!
filterEvent: aKeyboardEvent for: anObject
	"Provide keyboard shortcuts."
	
	| index itemToSelect |

	aKeyboardEvent controlKeyPressed
		ifFalse: [^ aKeyboardEvent].

	(aKeyboardEvent isKeyDown or: [aKeyboardEvent isKeystroke]) "We also need #keyDown for Windows platforms because CTRL+X does not trigger key strokes there..."
		ifFalse: [^ aKeyboardEvent].
			
	"Search field."
	aKeyboardEvent keyCharacter = $0
		ifTrue: [
			self searchBarMorph ifNotNil: [ :morph |
				morph model activate: aKeyboardEvent in: morph ].
			^ aKeyboardEvent ignore "hit!!"].
	
	"Select menu items."
	(aKeyboardEvent keyValue 
		between: $1 asciiValue 
		and: $9 asciiValue)
			ifFalse: [^ aKeyboardEvent].	
			
	index := aKeyboardEvent keyValue - $1 asciiValue + 1.
	itemToSelect := (self submorphs select: [ :each | 
		each isKindOf: DockingBarItemMorph ]) 
			at: index 
			ifAbsent: [^ aKeyboardEvent].
			
	self activate: aKeyboardEvent.
	self 
		selectItem: itemToSelect
		event: aKeyboardEvent.

	^ aKeyboardEvent ignore "hit!!"! !

!DockingBarMorph methodsFor: 'private - layout' stamp: 'mt 7/26/2016 12:51'!
gradientRamp
	| cc |
	cc := originalColor adjustSaturation: -0.08 brightness: 0.4.
	^ gradientRamp ifNil:[gradientRamp := {
			0.0 -> cc. 
			0.25 -> (originalColor mixed: 0.5 with: cc). 
			0.9 -> originalColor.
			1.0 -> originalColor darker.}]! !

!DockingBarMorph methodsFor: 'private - layout' stamp: 'dgd 11/2/2004 12:00'!
gradientRamp: colorRamp 
	gradientRamp := colorRamp.
""
self updateColor! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'kb 2/4/2010 11:30'!
handlesKeyboard: evt

	^true! !

!DockingBarMorph methodsFor: 'testing' stamp: 'mt 6/21/2016 13:57'!
indicateKeyboardFocus
	^ false! !

!DockingBarMorph methodsFor: 'initialize-release' stamp: 'mt 7/9/2016 13:42'!
initialize
	"initialize the receiver"
	super initialize.
	self changeTableLayout.
	
	selectedItem := nil.
	activeSubMenu := nil.
	fillsOwner := true.
	avoidVisibleBordersAtEdge := true.
	autoGradient := MenuMorph gradientMenu.
	self 
		setDefaultParameters ; 
		beFloating ; 
		beSticky ;
		layoutInset: 0 ;
		dropEnabled: true! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 11:57'!
isAdheringToBottom
	"Answer true if the receiver is adhering to bottom"
	^ self edgeToAdhereTo == #bottom! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 11:57'!
isAdheringToLeft
	"Answer true if the receiver is adhering to left"
	^ self edgeToAdhereTo == #left! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 11:57'!
isAdheringToRight
	"Answer true if the receiver is adhering to right"
	^ self edgeToAdhereTo == #right! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 11:57'!
isAdheringToTop
	"Answer true if the receiver is adhering to top"
	^ self edgeToAdhereTo == #top! !

!DockingBarMorph methodsFor: 'testing' stamp: 'dgd 8/31/2004 15:00'!
isDockingBar
	"Return true if the receiver is a docking bar"
	^ true! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/30/2004 23:13'!
isFloating
	"Answer true if the receiver has a float layout"
	^ self isHorizontal not
		and: [self isVertical not]! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 11:58'!
isHorizontal
	"Answer true if the receiver has a horizontal layout"
	^ self isAdheringToTop
		or: [self isAdheringToBottom]! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 9/1/2004 21:23'!
isSticky
	"answer whether the receiver is Sticky"
	^ Preferences noviceMode
		or: [super isSticky] ! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 11:58'!
isVertical
	"Answer true if the receiver has a vertical layout"
	^ self isAdheringToLeft
		or: [self isAdheringToRight]
! !

!DockingBarMorph methodsFor: 'dropping/grabbing' stamp: 'dgd 8/31/2004 14:37'!
justDroppedInto: aMorph event: anEvent 
	| ownerBounds leftRegion droppedPosition rightRegion topRegion bottomRegion |
	super justDroppedInto: aMorph event: anEvent.
	""
	self owner isNil
		ifTrue: [^ self].
	""
	ownerBounds := aMorph bounds.
	topRegion := ownerBounds bottom: ownerBounds top + (ownerBounds height // 5).
	bottomRegion := ownerBounds top: ownerBounds bottom - (ownerBounds height // 5).
	""
	leftRegion := ownerBounds right: ownerBounds left + (ownerBounds width // 5).
	leftRegion := leftRegion top: topRegion bottom.
	leftRegion := leftRegion bottom: bottomRegion top.
	""
	rightRegion := ownerBounds left: ownerBounds right - (ownerBounds width // 5).
	rightRegion := rightRegion top: topRegion bottom.
	rightRegion := rightRegion bottom: bottomRegion top.
	""
	droppedPosition := anEvent position.
	(topRegion containsPoint: droppedPosition)
		ifTrue: [
			^ self adhereToTop].
	(bottomRegion containsPoint: droppedPosition)
		ifTrue: [
			^ self adhereToBottom].
	(leftRegion containsPoint: droppedPosition)
		ifTrue: [
			^ self adhereToLeft].
	(rightRegion containsPoint: droppedPosition)
		ifTrue: [
			^ self adhereToRight].
	""
	self beFloating! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'kb 2/26/2010 14:02'!
keyStroke: evt 

	| asc |
	asc := evt keyCharacter asciiValue.
	asc = 27 ifTrue: [ "escape key" 
		^self deactivate: evt ].
	asc = self selectSubmenuKey ifTrue: [
		self ensureSelectedItem: evt.
		self selectedItem subMenu ifNotNil: [ :subMenu |
			subMenu items ifNotEmpty: [
				subMenu activate: evt.
				^subMenu moveSelectionDown: 1 event: evt ] ] ].
	asc = self previousKey ifTrue: [ ^self moveSelectionDown: -1 event: evt ].
	asc = self nextKey ifTrue: [ ^self moveSelectionDown: 1 event: evt ].
	selectedItem ifNotNil: [ 
		selectedItem subMenu ifNotNil: [ :subMenu |
			" If we didn't handle the keystroke, pass the keyboard focus 
			to the open submenu. "
			evt hand newKeyboardFocus: subMenu.
			subMenu keyStroke: evt ] ]! !

!DockingBarMorph methodsFor: 'WiW support' stamp: 'dgd 9/7/2004 19:25'!
morphicLayerNumber
	"helpful for insuring some morphs always appear in front of or 
	behind others. smaller numbers are in front"
	^ 11! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'mt 6/21/2016 13:27'!
mouseDown: evt

	(self fullContainsPoint: evt position) ifFalse: [
		self selectItem: nil event: evt.
		self deleteIfPopUp: evt ]! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'mt 6/20/2016 16:12'!
mouseMove: evt
	"We might transfer control to *either* the currently active submenu or the pop up owner, if any. Since the active sub menu is always displayed upfront check it first."
	
	selectedItem ifNotNil:[selectedItem activateSubmenu: evt].! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'mt 6/21/2016 13:26'!
mouseUp: evt

	evt hand newMouseFocus: self.! !

!DockingBarMorph methodsFor: 'control' stamp: 'kb 2/4/2010 11:37'!
moveSelectionDown: direction event: evt
	"Move the current selection up or down by one, presumably under keyboard control.
	direction = +/-1"

	| index |
	index := (submorphs indexOf: selectedItem ifAbsent: [1-direction]) + direction.
	submorphs do: "Ensure finite"
		[:unused | | m |
		m := submorphs atWrap: index.
		((m isKindOf: DockingBarItemMorph) and: [m isEnabled]) ifTrue:
			[^ self selectItem: m event: evt].
		"Keep looking for an enabled item"
		index := index + direction sign].
	^ self selectItem: nil event: evt! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'kb 2/4/2010 12:33'!
nextKey

	self isHorizontal ifTrue: [ ^29 " right arrow" ].
	self isVertical ifTrue: [ ^31 " down arrow " ]! !

!DockingBarMorph methodsFor: 'submorphs-accessing' stamp: 'dgd 9/1/2004 18:41'!
noteNewOwner: aMorph 
	"I have just been added as a submorph of aMorph"
	super noteNewOwner: aMorph.

	self submorphs
		do: [:each | each adjustLayoutBounds].
! !

!DockingBarMorph methodsFor: 'events' stamp: 'kb 2/4/2010 12:48'!
oldKeyboardFocus
	
	oldKeyboardFocus = self
		ifTrue: [ ^nil ]
		ifFalse: [ ^oldKeyboardFocus ]! !

!DockingBarMorph methodsFor: 'events' stamp: 'kb 2/4/2010 12:48'!
oldMouseFocus
	
	oldMouseFocus = self
		ifTrue: [ ^nil ]
		ifFalse: [ ^oldMouseFocus ]! !

!DockingBarMorph methodsFor: 'events' stamp: 'kb 2/4/2010 14:49'!
oldMouseFocus: aMorph
	
	(self submorphs includes: aMorph) 
		ifFalse: [ oldMouseFocus := aMorph ]
		ifTrue: [ oldMouseFocus := nil ]
	! !

!DockingBarMorph methodsFor: 'change reporting' stamp: 'dgd 9/1/2004 15:29'!
ownerChanged
"The receiver's owner has changed its layout. "
	self updateBounds.
	^ super ownerChanged! !

!DockingBarMorph methodsFor: 'private - accessing' stamp: 'dgd 9/1/2004 15:19'!
predominantDockingBarsOfChastes: predominantChastes 
	"Private - Answer a collection of the docking bar of my owner  
	that are predominant to the receiver.  
	 
	By 'predominant' we mean docking bar that have the right to  
	get a position before the receiver.  
	 
	The predominance of individual living in the same chaste is  
	determinated by the arrival order.  
	"
	| allDockingBars byChaste byArrival |
	(self owner isNil
			or: [self owner isHandMorph])
		ifTrue: [^ #()].
	""
	allDockingBars := self owner dockingBars.
	""
	byChaste := allDockingBars
				select: [:each | predominantChastes includes: each edgeToAdhereTo].
	""
	(predominantChastes includes: self edgeToAdhereTo)
		ifFalse: [^ byChaste].
	""
	byChaste := byChaste
				reject: [:each | each edgeToAdhereTo = self edgeToAdhereTo].
	""
	byArrival := allDockingBars
				select: [:each | each edgeToAdhereTo = self edgeToAdhereTo].

	byArrival := byArrival copyAfter: self.
	""
	^ byChaste , byArrival! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'kb 2/4/2010 12:33'!
previousKey

	self isHorizontal ifTrue: [ ^28 "left arrow" ].
	self isVertical ifTrue: [ ^30 "up arrow " ]! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'mt 6/20/2016 15:54'!
processFocusEvent: evt using: dispatcher

	^ dispatcher dispatchFocusEventFully: evt with: self! !

!DockingBarMorph methodsFor: 'initialize-release' stamp: 'cmm 5/3/2011 19:47'!
release
	activeSubMenu := selectedItem := oldKeyboardFocus := oldMouseFocus := nil! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 9/2/2004 11:01'!
resistsRemoval
"Answer whether the receiver is marked as resisting removal"
	^ Preferences noviceMode
		or: [super resistsRemoval]! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 9/9/2004 19:45'!
rootMenu
	^ self! !

!DockingBarMorph methodsFor: 'rounding' stamp: 'dgd 8/31/2004 14:16'!
roundedCorners
	"Return a list of those corners to round"
	self isAdheringToTop
		ifTrue: [^ #(2 3 )].
	self isAdheringToBottom
		ifTrue: [^ #(1 4 )].
	self isAdheringToLeft 
		ifTrue: [^ #(3 4 )].
	self isAdheringToRight
		ifTrue: [^ #(1 2 )].
	^ #(1 2 3 4 )! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'mt 4/2/2015 16:14'!
searchBarMorph

	^self submorphs detect: [ :each | each knownName = #searchBar ] ifNone: [ nil ]! !

!DockingBarMorph methodsFor: 'private' stamp: 'kb 2/22/2010 15:06'!
selectedItem

	(selectedItem notNil and: [ 
		selectedItem isSelected ]) ifTrue: [ 
			^selectedItem ].
	^ nil! !

!DockingBarMorph methodsFor: 'control' stamp: 'dgd 9/1/2004 16:40'!
selectItem: aMenuItem event: anEvent 
	selectedItem
		ifNotNil: [selectedItem deselect: anEvent].
	selectedItem := aMenuItem.
	selectedItem
		ifNotNil: [selectedItem select: anEvent]! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'kb 2/4/2010 13:03'!
selectSubmenuKey

	self isAdheringToTop ifTrue: [ ^31 ].
	self isAdheringToRight ifTrue: [ ^28 ].
	self isAdheringToLeft ifTrue: [ ^29 ].
	self isAdheringToBottom ifTrue: [ 30 ].
	^31! !

!DockingBarMorph methodsFor: 'initialize-release' stamp: 'mt 7/26/2016 12:04'!
setDefaultParameters
	"private - set the default parameter using Preferences as the inspiration source"

	self
		color: (self userInterfaceTheme color ifNil: [Color r: 0.9 g: 0.9 b: 0.9]);
		borderStyle: (self userInterfaceTheme borderStyle ifNil: [BorderStyle default]);
		borderColor: (self userInterfaceTheme borderColor ifNil: [Color gray]);
		borderWidth: (self userInterfaceTheme borderWidth ifNil: [0]).
		
	self extent: (Preferences standardMenuFont height asPoint).! !

!DockingBarMorph methodsFor: 'menus' stamp: 'dgd 9/1/2004 15:29'!
snapToEdgeIfAppropriate
	(self owner isNil
			or: [self owner isHandMorph])
		ifTrue: [^ self].
	""
	self updateBounds! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 9/1/2004 16:39'!
stayUp
	^ false! !

!DockingBarMorph methodsFor: 'menu' stamp: 'dgd 9/29/2004 20:27'!
toggleAutoGradient
	self autoGradient: self autoGradient not! !

!DockingBarMorph methodsFor: 'menu' stamp: 'dgd 9/29/2004 20:26'!
toggleAvoidVisibleBordersAtEdge
	self avoidVisibleBordersAtEdge: self avoidVisibleBordersAtEdge not! !

!DockingBarMorph methodsFor: 'menu' stamp: 'dgd 9/29/2004 20:24'!
toggleFillsOwner
	self fillsOwner: self fillsOwner not! !

!DockingBarMorph methodsFor: 'private - layout' stamp: 'dgd 9/1/2004 15:29'!
updateBounds
	"private - update the receiver's bounds"
	self updateExtent.
	self isFloating
		ifFalse: [self updatePosition]! !

!DockingBarMorph methodsFor: 'private - layout' stamp: 'mt 7/25/2016 14:01'!
updateColor
	"private - update the receiver's color"
	| fill |
	MenuMorph gradientMenu ifFalse: [
		self color ~= originalColor ifTrue: [self color: originalColor].
		^ self].
	""
	fill := GradientFillStyle ramp: self gradientRamp.
	""
	fill origin: self topLeft.
	self isVertical
		ifTrue: [fill direction: self width @ 0]
		ifFalse: [fill direction: 0 @ self height].
	""
	self fillStyle: fill! !

!DockingBarMorph methodsFor: 'private - layout' stamp: 'dgd 9/1/2004 15:20'!
updateExtent
	"private - update the receiver's extent"
	| margin |
	self fullBounds.
	self fillsOwner
		ifFalse: [^ self].
	""
	margin := self avoidVisibleBordersAtEdge
				ifTrue: [self borderWidth * 2]
				ifFalse: [0].""
	self isHorizontal
		ifTrue: [self width: self owner width + margin].""
	self isVertical
		ifTrue: [| usedHeight | 
			usedHeight := self usedHeightByPredominantDockingBarsOfChastes: #(#top #bottom ).
			self height: self owner height + margin - usedHeight]! !

!DockingBarMorph methodsFor: 'private - layout' stamp: 'mt 7/9/2016 13:10'!
updateLayoutProperties
	"private - update the layout properties based on adhering,  
	fillsOwner and avoidVisibleBordersAtEdge preferencs"
	""
	(self isHorizontal
			or: [self isFloating])
		ifTrue: [self listDirection: #leftToRight]
		ifFalse: [self listDirection: #topToBottom].
	""
	self hResizing: #shrinkWrap.
	self vResizing: #shrinkWrap.
	self fillsOwner
		ifTrue: [""
			self isHorizontal
				ifTrue: [self hResizing: #spaceFill].
			self isVertical
				ifTrue: [self vResizing: #spaceFill]].
	""
	self isHorizontal ifTrue: [
		self submorphsDo: [:m | m hResizing: #shrinkWrap; vResizing: #spaceFill]].
	self isVertical ifTrue: [
		self submorphsDo: [:m | m vResizing: #shrinkWrap; hResizing: #spaceFill]].
! !

!DockingBarMorph methodsFor: 'private - layout' stamp: 'dgd 9/1/2004 13:06'!
updatePosition
	"private - update the receiver's position"
	| edgeSymbol margin |
	edgeSymbol := self edgeToAdhereTo.
	self
		perform: (edgeSymbol , ':') asSymbol
		with: (self owner perform: edgeSymbol).
	""
	margin := self avoidVisibleBordersAtEdge
				ifTrue: [self borderWidth asPoint]
				ifFalse: [0 asPoint].
	""
	self isAdheringToTop
		ifTrue: [| usedHeight | 
			usedHeight := self usedHeightByPredominantDockingBarsOfChastes: #(#top ).
			self topLeft: self owner topLeft - margin + (0 @ usedHeight)].
	self isAdheringToBottom
		ifTrue: [| usedHeight | 
			usedHeight := self usedHeightByPredominantDockingBarsOfChastes: #(#bottom ).
			self bottomLeft: self owner bottomLeft + (-1 @ 1 * margin) - (0 @ usedHeight)].
	""
	self isAdheringToLeft
		ifTrue: [| usedHeight usedWidth | 
			usedHeight := self usedHeightByPredominantDockingBarsOfChastes: #(#top ).
			usedWidth := self usedWidthByPredominantDockingBarsOfChastes: #(#left ).
			self topLeft: self owner topLeft - margin + (usedWidth @ usedHeight)].
	self isAdheringToRight
		ifTrue: [| usedHeight usedWidth | 
			usedHeight := self usedHeightByPredominantDockingBarsOfChastes: #(#top ).
			usedWidth := self usedWidthByPredominantDockingBarsOfChastes: #(#right ).
			self topRight: self owner topRight + (1 @ -1 * margin) + (usedWidth negated @ usedHeight)]! !

!DockingBarMorph methodsFor: 'private - accessing' stamp: 'dgd 9/1/2004 19:39'!
usedHeightByPredominantDockingBarsOfChastes: predominantChastes 
	"Private - convenience"
	| predominants |
	predominants := self predominantDockingBarsOfChastes: predominantChastes.
	^ predominants isEmpty
		ifTrue: [0]
		ifFalse: [(predominants
				collect: [:each | each height]) sum]! !

!DockingBarMorph methodsFor: 'private - accessing' stamp: 'dgd 9/1/2004 19:38'!
usedWidthByPredominantDockingBarsOfChastes: predominantChastes 
	"Private - convenience"
	| predominants |
	predominants := self predominantDockingBarsOfChastes: predominantChastes.
	^ predominants isEmpty
		ifTrue: [0]
		ifFalse: [(predominants
				collect: [:each | each width]) sum]! !

!DockingBarMorph methodsFor: 'events-processing' stamp: 'mt 6/20/2016 16:07'!
wantsEveryMouseMove
	^ true! !

!DockingBarMorph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 16:33'!
wantsToBeTopmost
	"Answer if the receiver want to be one of the topmost objects in 
	its owner"
	^ true! !

!DockingBarMorph methodsFor: 'menu' stamp: 'dgd 9/18/2004 18:28'!
wantsYellowButtonMenu
	"Answer true if the receiver wants a yellow button menu"
	^ Preferences noviceMode not! !

Morph subclass: #HandMorph
	instanceVariableNames: 'mouseFocus keyboardFocus eventListeners mouseListeners keyboardListeners eventCaptureFilters mouseCaptureFilters keyboardCaptureFilters mouseClickState mouseOverHandler mouseWheelState lastMouseEvent targetOffset damageRecorder cacheCanvas cachedCanvasHasHoles temporaryCursor temporaryCursorOffset hardwareCursor hasChanged savedPatch userInitials lastEventBuffer genieGestureProcessor keyboardInterpreter'
	classVariableNames: 'CompositionWindowManager DoubleClickTime DragThreshold EventStats NewEventRules NormalCursor PasteBuffer SendMouseWheelToKeyboardFocus ShowEvents SynthesizeMouseWheelEvents'
	poolDictionaries: 'EventSensorConstants'
	category: 'Morphic-Kernel'!

!HandMorph commentStamp: '<historical>' prior: 0!
The cursor may be thought of as the HandMorph.  The hand's submorphs hold anything being carried by dragging.  

There is some minimal support for multiple hands in the same world.!

!HandMorph class methodsFor: 'utilities' stamp: 'sma 4/30/2000 10:34'!
attach: aMorph
	"Attach aMorph the current world's primary hand."

	self currentWorld primaryHand attachMorph: aMorph! !

!HandMorph class methodsFor: 'initialization' stamp: 'yo 8/13/2003 15:49'!
clearCompositionWindowManager

	CompositionWindowManager := nil.
! !

!HandMorph class methodsFor: 'initialization' stamp: 'yo 8/13/2003 15:45'!
clearInterpreters

	self allInstances do: [:each | each clearKeyboardInterpreter].
! !

!HandMorph class methodsFor: 'accessing' stamp: 'nice 3/6/2010 22:10'!
compositionWindowManager
	CompositionWindowManager ifNotNil: [^CompositionWindowManager].
	Smalltalk platformName = 'Win32' 
		ifTrue: [^CompositionWindowManager := ImmWin32 new].
	(Smalltalk platformName = 'unix' 
		and: [(Smalltalk windowSystemName) = 'X11']) 
			ifTrue: [^CompositionWindowManager := ImmX11 new].
	^CompositionWindowManager := ImmAbstractPlatform new! !

!HandMorph class methodsFor: 'accessing' stamp: ''!
doubleClickTime

	^ DoubleClickTime
! !

!HandMorph class methodsFor: 'accessing' stamp: ''!
doubleClickTime: milliseconds

	DoubleClickTime := milliseconds.
! !

!HandMorph class methodsFor: 'accessing' stamp: 'mtf 2/20/2011 20:39'!
dragThreshold

	^ DragThreshold
! !

!HandMorph class methodsFor: 'accessing' stamp: 'mtf 2/20/2011 20:39'!
dragThreshold: pixels

	DragThreshold := pixels! !

!HandMorph class methodsFor: 'new-morph participation' stamp: 'di 6/22/97 09:07'!
includeInNewMorphMenu
	"Not to be instantiated from the menu"
	^ false! !

!HandMorph class methodsFor: 'class initialization' stamp: 'cmm 3/5/2011 13:54'!
initialize
	"HandMorph initialize"
	PasteBuffer := nil.
	DoubleClickTime := 350 "milliseconds".
	DragThreshold := 10 "pixels".
	NormalCursor := CursorWithMask normal asCursorForm! !

!HandMorph class methodsFor: 'class initialization' stamp: 'fbs 7/15/2013 20:40'!
localeChanged
	self startUp.! !

!HandMorph class methodsFor: 'utilities' stamp: 'ar 9/17/2000 17:02'!
newEventRules: aBool
	NewEventRules := aBool.! !

!HandMorph class methodsFor: 'preferences' stamp: 'mt 6/6/2016 17:38'!
sendMouseWheelToKeyboardFocus
	<preference: 'Send Mouse Wheel Events to Keyboard Focus'
		categoryList: #(Morphic keyboard mouse)
		description: 'If enabled, follow the behavior known from Microsoft Windows, where the mouse wheel works for the widget that has the keyboard focus. If disabled, follow the Mac OS style, where the mouse wheel is send to the widget under the mouse position'
		type: #Boolean>
	^ SendMouseWheelToKeyboardFocus ifNil: [true]! !

!HandMorph class methodsFor: 'preferences' stamp: 'mt 6/6/2016 17:38'!
sendMouseWheelToKeyboardFocus: aBoolean
	SendMouseWheelToKeyboardFocus := aBoolean.! !

!HandMorph class methodsFor: 'utilities' stamp: 'nk 7/20/2003 10:03'!
showEvents: aBool
	"HandMorph showEvents: true"
	"HandMorph showEvents: false"
	ShowEvents := aBool.
	aBool ifFalse: [ ActiveWorld invalidRect: (0@0 extent: 250@120) ].! !

!HandMorph class methodsFor: 'initialization' stamp: 'yo 8/13/2003 15:49'!
startUp

	self clearCompositionWindowManager.
	self clearInterpreters.
! !

!HandMorph class methodsFor: 'preferences' stamp: 'mt 6/6/2016 17:40'!
synthesizeMouseWheelEvents
	<preference: 'Synthesize Mouse Wheel Events from Keyboard Events'
		categoryList: #(Morphic keyboard mouse)
		description: 'If enabled, convert CTRL+Up/Down (keystroke with modifier key) to mouse wheel up/down. Disable this according to the current VM behavior.'
		type: #Boolean>
	^ SynthesizeMouseWheelEvents ifNil: [true]! !

!HandMorph class methodsFor: 'preferences' stamp: 'mt 6/6/2016 17:39'!
synthesizeMouseWheelEvents: aBoolean
	SynthesizeMouseWheelEvents := aBoolean.! !

!HandMorph methodsFor: 'balloon help' stamp: 'mt 4/5/2016 11:24'!
addBalloonHelp: aBalloonMorph
	"Associate a new balloon morph with this hand."

	self removeAlarm: #spawnBalloonFor:.
	aBalloonMorph ifNotNil: [:m | self balloonHelpList add: m].! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/24/2000 20:43'!
addEventListener: anObject
	"Make anObject a listener for all events. All events will be reported to the object."
	self eventListeners: (self addListener: anObject to: self eventListeners)! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/24/2000 20:42'!
addKeyboardListener: anObject
	"Make anObject a listener for keyboard events. All keyboard events will be reported to the object."
	self keyboardListeners: (self addListener: anObject to: self keyboardListeners)! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/24/2000 20:40'!
addListener: anObject to: aListenerGroup
	"Add anObject to the given listener group. Return the new group."
	| listeners |
	listeners := aListenerGroup.
	(listeners notNil and:[listeners includes: anObject]) ifFalse:[
		listeners
			ifNil:[listeners := WeakArray with: anObject]
			ifNotNil:[listeners := listeners copyWith: anObject]].
	listeners := listeners copyWithout: nil. "obsolete entries"
	^listeners! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/24/2000 20:42'!
addMouseListener: anObject
	"Make anObject a listener for mouse events. All mouse events will be reported to the object."
	self mouseListeners: (self addListener: anObject to: self mouseListeners)! !

!HandMorph methodsFor: 'accessing' stamp: 'tk 10/20/2004 15:54'!
anyButtonPressed
	^lastMouseEvent anyButtonPressed! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'ar 10/8/2000 23:42'!
attachMorph: m
	"Position the center of the given morph under this hand, then grab it.
	This method is used to grab far away or newly created morphs."
	| delta |
	self releaseMouseFocus. "Break focus"
	delta := m bounds extent // 2.
	m position: (self position - delta).
	m formerPosition: m position.
	targetOffset := m position - self position.
	self addMorphBack: m.! !

!HandMorph methodsFor: 'genie-stubs' stamp: 'nk 3/10/2004 10:06'!
autoFocusRectangleBoundsFor: aMorph
	^aMorph bounds! !

!HandMorph methodsFor: 'balloon help' stamp: 'mt 3/31/2016 10:31'!
balloonHelpList
	"Return all balloon morphs associated with this hand"

	^ self
		valueOfProperty: #balloonHelpMorphs
		ifAbsentPut: [OrderedCollection new]! !

!HandMorph methodsFor: 'updating' stamp: 'jm 2/20/98 19:54'!
changed

	hasChanged := true.
! !

!HandMorph methodsFor: 'event handling' stamp: 'tpr 1/5/2005 17:34'!
checkForMoreKeyboard
	"Quick check for more keyboard activity -- Allows, eg, many characters
	to be accumulated into a single replacement during type-in."

	| evtBuf |
	self flag: #arNote.	"Will not work if we don't examine event queue in Sensor"
	evtBuf := Sensor peekKeyboardEvent.
	evtBuf ifNil: [^nil].
	^self generateKeyboardEvent: evtBuf! !

!HandMorph methodsFor: 'multilingual' stamp: 'yo 9/26/2003 22:11'!
clearKeyboardInterpreter

	keyboardInterpreter := nil.
! !

!HandMorph methodsFor: 'accessing' stamp: 'sw 2/11/98 18:00'!
colorForInsets
	"Morphs being dragged by the hand use the world's color"
	^ owner colorForInsets! !

!HandMorph methodsFor: 'focus handling' stamp: 'yo 11/7/2002 19:10'!
compositionWindowManager

	^ self class compositionWindowManager.
! !

!HandMorph methodsFor: 'meta-actions' stamp: 'ar 11/6/2000 13:07'!
copyToPasteBuffer: aMorph
	"Save this morph in the paste buffer. This is mostly useful for copying morphs between projects."
	aMorph ifNil:[^PasteBuffer := nil].
	Cursor wait showWhile:[
		PasteBuffer := aMorph topRendererOrSelf veryDeepCopy.
		PasteBuffer privateOwner: nil].

! !

!HandMorph methodsFor: 'cursor' stamp: 'jwh 6/5/2000 07:38'!
cursorBounds

	^temporaryCursor 
		ifNil: [self position extent: NormalCursor extent]
		ifNotNil: [self position + temporaryCursorOffset extent: temporaryCursor extent]! !

!HandMorph methodsFor: 'event handling' stamp: 'dgd 2/21/2003 22:43'!
cursorPoint
	"Implemented for allowing embedded worlds in an event cycle to query a hand's position and get it in its coordinates. The same can be achieved by #point:from: but this is simply much more convenient since it will look as if the hand is in the lower world."

	| pos |
	pos := self position.
	(ActiveWorld isNil or: [ActiveWorld == owner]) ifTrue: [^pos].
	^ActiveWorld point: pos from: owner! !

!HandMorph methodsFor: 'balloon help' stamp: 'mt 3/31/2016 10:33'!
deleteBalloonTarget: aMorph
	"Delete any existing balloon help."
	
	self balloonHelpList
		do: [:ea | ea delete];
		removeAll.! !

!HandMorph methodsFor: 'genie-stubs' stamp: 'nk 3/10/2004 10:15'!
disableGenieFocus
! !

!HandMorph methodsFor: 'drawing' stamp: 'dgd 2/21/2003 22:43'!
drawOn: aCanvas 
	"Draw the hand itself (i.e., the cursor)."

	| userPic |
	temporaryCursor isNil 
		ifTrue: [aCanvas paintImage: NormalCursor at: bounds topLeft]
		ifFalse: [aCanvas paintImage: temporaryCursor at: bounds topLeft].
	self hasUserInformation 
		ifTrue: 
			[aCanvas 
				drawString: userInitials
				at: self cursorBounds topRight + (0 @ 4)
				font: nil
				color: color.
			(userPic := self userPicture) ifNotNil: 
					[aCanvas paintImage: userPic at: self cursorBounds topRight + (0 @ 24)]]! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'ar 10/5/2000 16:23'!
dropMorphs
	"Drop the morphs at the hands position"
	self dropMorphs: lastMouseEvent.! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'ar 9/14/2000 11:22'!
dropMorphs: anEvent
	"Drop the morphs at the hands position"
	self submorphsReverseDo:[:m|
		"Drop back to front to maintain z-order"
		self dropMorph: m event: anEvent.
	].! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'mt 5/21/2016 09:58'!
dropMorph: aMorph event: anEvent
	"Drop the given morph which was carried by the hand"
	| event dropped |
	(anEvent isMouseUp and:[aMorph shouldDropOnMouseUp not]) ifTrue:[^self].

	"Note: For robustness in drag and drop handling we remove the morph BEFORE we drop him, but we keep his owner set to the hand. This prevents system lockups when there is a problem in drop handling (for example if there's an error in #wantsToBeDroppedInto:). THIS TECHNIQUE IS NOT RECOMMENDED FOR CASUAL USE."
	self privateRemove: aMorph.
	aMorph privateOwner: self.

	dropped := aMorph.
	(dropped hasProperty: #addedFlexAtGrab) 
		ifTrue:[dropped := aMorph removeFlexShell].
	event := DropEvent new setPosition: self position contents: dropped hand: self.
	
	[ "In case of an error, ensure that the morph-to-be-dropped will be disposed. Otherwise it may confuse garbage handler. See the sends of #privateRemove: and #privateOwner: above."
		event := self sendEvent: event focus: nil. "event filters can apply and filtered events will be returned"
		event wasHandled ifFalse: [aMorph rejectDropMorphEvent: event] ]
			ensure: [ aMorph owner == self ifTrue: [aMorph delete] ].
	
	self mouseOverHandler processMouseOver: anEvent.! !

!HandMorph methodsFor: 'genie-stubs' stamp: 'nk 3/11/2004 17:47'!
enableGenie
	self error: 'Genie is not available for this hand'.! !

!HandMorph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:14'!
eventBubbleFilters
	Error signal: 'Hand morphs do only have capture filters. Install top-most bubble filters in the world.'! !

!HandMorph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:11'!
eventCaptureFilters
	^ eventCaptureFilters! !

!HandMorph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:11'!
eventCaptureFilters: anArrayOrNil
	eventCaptureFilters := anArrayOrNil! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/26/2000 01:27'!
eventListeners
	^eventListeners! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/26/2000 01:27'!
eventListeners: anArrayOrNil
	eventListeners := anArrayOrNil! !

!HandMorph methodsFor: 'events-filtering' stamp: 'mt 6/10/2016 14:39'!
filterEvent: aKeyboardEvent for: aMorphOrNil
	"Fixes VM behavior. Usually, there are no mouse wheel events generated by the VM but CTRL+UP/DOWN. Convert these into mouse wheel events.
	
	We installed ourself as keyboard filter only!! No need to check whether this is a keyboard event or not!! See HandMorph >> #initForEvents.
	
	Might be removed in the future if this mapping gets obsolete."
	
	HandMorph synthesizeMouseWheelEvents ifFalse: [^ aKeyboardEvent].
	
	(aKeyboardEvent isKeystroke and: [aKeyboardEvent controlKeyPressed]) ifTrue: [
		aKeyboardEvent keyCharacter caseOf: {
			[Character arrowUp] -> [^ self generateMouseWheelEvent: aKeyboardEvent direction: 2r1000].
			[Character arrowDown] -> [^ self generateMouseWheelEvent: aKeyboardEvent direction: 2r0100].
		} otherwise: [^ aKeyboardEvent]].
	
	^ aKeyboardEvent! !

!HandMorph methodsFor: 'event handling' stamp: 'ar 10/5/2000 23:17'!
flushEvents
	"Flush any events that may be pending"
	self flag: #arNote. "Remove it and fix senders"
	Sensor flushEvents.! !

!HandMorph methodsFor: 'genie-stubs' stamp: 'nk 3/11/2004 17:44'!
focusStartEvent
	^nil! !

!HandMorph methodsFor: 'layout' stamp: 'jm 2/20/98 18:55'!
fullBounds
	"Extend my bounds by the shadow offset when carrying morphs."

	| bnds |
	bnds := super fullBounds.
	submorphs isEmpty
		ifTrue: [^ bnds ]
		ifFalse: [^ bnds topLeft corner: bnds bottomRight + self shadowOffset].
! !

!HandMorph methodsFor: 'drawing' stamp: 'tpr 6/22/2015 15:56'!
fullDrawOn: aCanvas 
	"A HandMorph has unusual drawing requirements:
		1. the hand itself (i.e., the cursor) appears in front of its submorphs
		2. morphs being held by the hand cast a shadow on the world/morphs below
	The illusion is that the hand plucks up morphs and carries them above the world."

	"Note: This version caches an image of the morphs being held by the hand for
	 better performance. This cache is invalidated if one of those morphs changes."

	| disableCaching subBnds |
	self visible ifFalse: [^self].
	(aCanvas isVisible: self fullBounds) ifFalse: [^self].
	(self hasProperty: #errorOnDraw) ifTrue:[^self drawErrorOn: aCanvas].
	disableCaching := false.
	disableCaching 
		ifTrue: 
			[self nonCachingFullDrawOn: aCanvas.
			^self].
	submorphs isEmpty 
		ifTrue: 
			[cacheCanvas := nil.
			^self drawOn: aCanvas].	"just draw the hand itself"

	"special handling of a single submorph that wants to do its own thing
	when being dragged"
	(submorphs size = 1
			and: [submorphs first handledOwnDraggingBy: self on: aCanvas])
		ifTrue: [^ self drawOn: aCanvas].

	subBnds := Rectangle merging: (submorphs collect: [:m | m fullBounds]).
	self updateCacheCanvas: aCanvas.
	(cacheCanvas isNil 
		or: [cachedCanvasHasHoles and: [cacheCanvas depth = 1]]) 
			ifTrue: 
				["could not use caching due to translucency; do full draw"

				self nonCachingFullDrawOn: aCanvas.
				^self].

	"draw the shadow"
	aCanvas asShadowDrawingCanvas translateBy: self shadowOffset
		during: 
			[:shadowCanvas | 
			cachedCanvasHasHoles 
				ifTrue: 
					["Have to draw the real shadow of the form"

					shadowCanvas paintImage: cacheCanvas form at: subBnds origin]
				ifFalse: 
					["Much faster if only have to shade the edge of a solid rectangle"

					(subBnds areasOutside: (subBnds translateBy: self shadowOffset negated)) 
						do: [:r | shadowCanvas fillRectangle: r color: Color black]]].

	"draw morphs in front of the shadow using the cached Form"
	cachedCanvasHasHoles 
		ifTrue: [aCanvas paintImage: cacheCanvas form at: subBnds origin]
		ifFalse: 
			[aCanvas 
				drawImage: cacheCanvas form
				at: subBnds origin
				sourceRect: cacheCanvas form boundingBox].
	self drawOn: aCanvas	"draw the hand itself in front of morphs"! !

!HandMorph methodsFor: 'private events' stamp: 'tpr 1/11/2016 15:33'!
generateDropFilesEvent: evtBuf 
	"Generate the appropriate mouse event for the given raw event buffer"

	"Note: This is still in an experimental phase and will need more work"

	| position buttons modifiers stamp numFiles dragType |
	stamp := evtBuf second.
	stamp = 0 ifTrue: [stamp := Time eventMillisecondClock].
	dragType := evtBuf third.
	position := evtBuf fourth @ evtBuf fifth.
	buttons := 0.
	modifiers := evtBuf sixth.
	buttons := buttons bitOr: (modifiers bitShift: 3).
	numFiles := evtBuf seventh.
	dragType = 4 
		ifTrue: 
			["e.g., drop"

			owner borderWidth: 0.
			^DropFilesEvent new 
				setPosition: position
				contents: numFiles
				hand: self].
	"the others are currently not handled by morphs themselves"
	dragType = 1 
		ifTrue: 
			["experimental drag enter"

			owner
				borderWidth: 4;
				borderColor: owner color asColor negated].
	dragType = 2 
		ifTrue: 
			["experimental drag move"

			].
	dragType = 3 
		ifTrue: 
			["experimental drag leave"

			owner borderWidth: 0].
	^nil! !

!HandMorph methodsFor: 'private events' stamp: 'tpr 1/11/2016 15:33'!
generateKeyboardEvent: evtBuf
	"Generate the appropriate mouse event for the given raw event buffer"

	| buttons modifiers type pressType stamp keyValue |
	stamp := evtBuf second.
	stamp = 0 ifTrue: [stamp := Time eventMillisecondClock].
	pressType := evtBuf fourth.
	pressType = EventKeyDown ifTrue: [type := #keyDown].
	pressType = EventKeyUp ifTrue: [type := #keyUp].
	pressType = EventKeyChar ifTrue: [type := #keystroke].
	modifiers := evtBuf fifth.
	buttons := (modifiers bitShift: 3) bitOr: (lastMouseEvent buttons bitAnd: 7).
	type = #keystroke
		ifTrue: [keyValue := (self keyboardInterpreter nextCharFrom: Sensor firstEvt: evtBuf) asInteger]
		ifFalse: [keyValue := evtBuf third].
	^ KeyboardEvent new
		setType: type
		buttons: buttons
		position: self position
		keyValue: keyValue
		hand: self
		stamp: stamp.
! !

!HandMorph methodsFor: 'private events' stamp: 'tpr 1/11/2016 15:34'!
generateMouseEvent: evtBuf 
	"Generate the appropriate mouse event for the given raw event buffer"

	| position buttons modifiers type trail stamp oldButtons evtChanged |
	evtBuf first = lastEventBuffer first 
		ifTrue: 
			["Workaround for Mac VM bug, *always* generating 3 events on clicks"

			evtChanged := false.
			3 to: evtBuf size
				do: [:i | (lastEventBuffer at: i) = (evtBuf at: i) ifFalse: [evtChanged := true]].
			evtChanged ifFalse: [^nil]].
	stamp := evtBuf second.
	stamp = 0 ifTrue: [stamp := Time eventMillisecondClock].
	position := evtBuf third @ evtBuf fourth.
	buttons := evtBuf fifth.
	modifiers := evtBuf sixth.

	type := buttons = 0 
		ifTrue:[
				lastEventBuffer fifth = 0 		
					ifTrue: [#mouseMove] 	"this time no button and previously no button .. just mouse move"
					ifFalse: [#mouseUp]		"this time no button but previously some button ... therefore button was released"
		]
		ifFalse:[
				buttons = lastEventBuffer fifth
						ifTrue: [#mouseMove]		"button states are the same .. now and past .. therfore a mouse movement"
						ifFalse: [					"button states are different .. button was pressed or released"
							buttons > lastEventBuffer fifth
								ifTrue: [#mouseDown]
								ifFalse:[#mouseUp].
						].
		].
	buttons := buttons bitOr: (modifiers bitShift: 3).
	oldButtons := lastEventBuffer fifth 
				bitOr: (lastEventBuffer sixth bitShift: 3).
	lastEventBuffer := evtBuf.
	type == #mouseMove 
		ifTrue: 
			[trail := self mouseTrailFrom: evtBuf.
			^MouseMoveEvent new 
				setType: type
				startPoint: (self position)
				endPoint: trail last
				trail: trail
				buttons: buttons
				hand: self
				stamp: stamp].
	^MouseButtonEvent new 
		setType: type
		position: position
		which: (oldButtons bitXor: buttons)
		buttons: buttons
		hand: self
		stamp: stamp! !

!HandMorph methodsFor: 'private events' stamp: 'mt 6/10/2016 14:57'!
generateMouseWheelEvent: evtBuf
	"Generate the appropriate mouse wheel event for the given raw event buffer"

	| buttons modifiers deltaX deltaY stamp |
	stamp := evtBuf second.
	stamp = 0 ifTrue: [stamp := Time eventMillisecondClock].
	deltaX := evtBuf third.
	deltaY := evtBuf fourth.
	modifiers := evtBuf fifth.
	buttons := (modifiers bitShift: 3) bitOr: (lastMouseEvent buttons bitAnd: 7).
	^ MouseWheelEvent new
		setType: #mouseWheel
		position: self position
		delta: deltaX@deltaY
		direction: 2r0000
		buttons: buttons	
		hand: self
		stamp: stamp! !

!HandMorph methodsFor: 'private events' stamp: 'mt 6/10/2016 15:19'!
generateMouseWheelEvent: keystrokeEvent direction: direction
	"Generate the appropriate mouse wheel event from the keystrokeEvent. Before calling this, ensure that the control key is pressed.
	
	This method can be discarded once the VM produces real mouse wheel events."
	
	^ MouseWheelEvent new
		setType: #mouseWheel
		position: keystrokeEvent position
		delta: 0 @ ((direction anyMask: 2r1000 "wheel up") ifTrue: [120] ifFalse: [-120])
		direction: direction
		buttons: (keystrokeEvent buttons bitAnd: 2r01111) "drop control key pressed for this conversion"
		hand: keystrokeEvent hand
		stamp: keystrokeEvent timeStamp! !

!HandMorph methodsFor: 'private events' stamp: 'tpr 1/11/2016 15:30'!
generateWindowEvent: evtBuf 
	"Generate the appropriate window event for the given raw event buffer"

	| evt |
	evt := WindowEvent new.
	evt setTimeStamp: evtBuf second.
	evt timeStamp = 0 ifTrue: [evt setTimeStamp: Time eventMillisecondClock].
	evt action: evtBuf third.
	evt rectangle: (Rectangle origin: evtBuf fourth @ evtBuf fifth corner: evtBuf sixth @ evtBuf seventh ).
	
	^evt! !

!HandMorph methodsFor: 'genie-stubs' stamp: 'nk 3/10/2004 10:06'!
genieGestureProcessor
	^nil! !

!HandMorph methodsFor: 'meta-actions' stamp: 'ar 11/21/2000 17:42'!
grabMorph: aMorph
	"Grab the given morph (i.e., add it to this hand and remove it from its current owner) without changing its position. This is used to pick up a morph under the hand's current position, versus attachMorph: which is used to pick up a morph that may not be near this hand."
	| grabbed |
	self releaseMouseFocus. "Break focus"
	grabbed := aMorph aboutToBeGrabbedBy: self.
	grabbed ifNil:[^self].
	grabbed := grabbed topRendererOrSelf.
	^self grabMorph: grabbed from: grabbed owner! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'ar 4/23/2001 15:17'!
grabMorph: aMorph from: formerOwner
	"Grab the given morph (i.e., add it to this hand and remove it from its current owner) without changing its position. This is used to pick up a morph under the hand's current position, versus attachMorph: which is used to pick up a morph that may not be near this hand."

	| grabbed offset targetPoint grabTransform fullTransform |
	self releaseMouseFocus. "Break focus"
	grabbed := aMorph.
	aMorph keepsTransform ifTrue:[
		grabTransform := fullTransform := IdentityTransform new.
	] ifFalse:[
		"Compute the transform to apply to the grabbed morph"
		grabTransform := formerOwner 
			ifNil:		[IdentityTransform new] 
			ifNotNil:	[formerOwner grabTransform].
		"Compute the full transform for the grabbed morph"
		fullTransform := formerOwner 
			ifNil:		[IdentityTransform new] 
			ifNotNil:	[formerOwner transformFrom: owner].
	].
	"targetPoint is point in aMorphs reference frame"
	targetPoint := fullTransform globalPointToLocal: self position.
	"but current position will be determined by grabTransform, so compute offset"
	offset := targetPoint - (grabTransform globalPointToLocal: self position).
	"apply the transform that should be used after grabbing"
	grabbed := grabbed transformedBy: grabTransform.
	grabbed == aMorph 
		ifFalse:	[grabbed setProperty: #addedFlexAtGrab toValue: true].
	"offset target to compensate for differences in transforms"
	grabbed position: grabbed position - offset asIntegerPoint.
	"And compute distance from hand's position"
	targetOffset := grabbed position - self position.
	self addMorphBack: grabbed.
	grabbed justGrabbedFrom: formerOwner.! !

!HandMorph methodsFor: 'gridded cursor' stamp: 'tpr 12/3/2015 17:19'!
griddedPoint: aPoint
	"return the equivalent point snapped to the grid, if indeed any gridding is set"
	self valueOfProperty: #gridStep ifPresentDo: [:grid| |offset|
		offset := self valueOfProperty: #gridOffset ifAbsent: [0@0].
		^ offset + (aPoint + (grid //2) - offset truncateTo: grid)].
	^aPoint! !

!HandMorph methodsFor: 'gridded cursor' stamp: 'tpr 12/3/2015 18:26'!
gridPointRaw
	"return my latest position gridded"
	^self griddedPoint: self currentEvent position! !

!HandMorph methodsFor: 'gridded cursor' stamp: 'tpr 12/3/2015 18:27'!
gridTo: grid origin: offset
	"set a couple of properties to specify gridding for the temporaryCursor; instvars would be nicer"
	self setProperty: #gridStep toValue: grid;
		setProperty: #gridOffset toValue: offset! !

!HandMorph methodsFor: 'halos and balloon help' stamp: 'ar 10/4/2000 13:40'!
halo
	"Return the halo associated with this hand, if any"
	^self valueOfProperty: #halo! !

!HandMorph methodsFor: 'halo handling' stamp: 'ar 10/4/2000 13:40'!
halo: newHalo
	"Set halo associated with this hand"
	| oldHalo |
	oldHalo := self halo.
	(oldHalo isNil or:[oldHalo == newHalo]) ifFalse:[oldHalo delete].
	newHalo
		ifNil:[self removeProperty: #halo]
		ifNotNil:[self setProperty: #halo toValue: newHalo]! !

!HandMorph methodsFor: 'events-processing' stamp: 'mt 6/10/2016 15:08'!
handleEvent: unfilteredEvent

	| filteredEvent |
	owner ifNil: [^ unfilteredEvent  "not necessary but good style -- see Morph >> #handleEvent:"].
	
	self logEvent: unfilteredEvent.

	"Mouse-over events occur really, really, really often. They are kind of the heart beat of the Morphic UI process."
	unfilteredEvent isMouseOver ifTrue: [^ self sendMouseEvent: unfilteredEvent].

	self showEvent: unfilteredEvent.
	self sendListenEvents: unfilteredEvent.
	
	filteredEvent := self sendFilterEventCapture: unfilteredEvent for: nil.
	"filteredEvent := unfilteredEvent" " <-- use this to disable global capture filters"
	
	filteredEvent wasIgnored ifTrue: [
		self mouseOverHandler processMouseOver: lastMouseEvent.
		^ filteredEvent].
	
	filteredEvent isWindowEvent ifTrue: [
		self sendEvent: filteredEvent focus: nil.
		self mouseOverHandler processMouseOver: lastMouseEvent.
		^ filteredEvent].

	filteredEvent isKeyboard ifTrue:[
		self sendKeyboardEvent: filteredEvent.
		self mouseOverHandler processMouseOver: lastMouseEvent.
		^ filteredEvent].
			
	filteredEvent isDropEvent ifTrue:[
		self sendEvent: filteredEvent focus: nil.
		self mouseOverHandler processMouseOver: lastMouseEvent.
		^ filteredEvent].

	filteredEvent isMouse ifFalse: [
		self mouseOverHandler processMouseOver: lastMouseEvent.
		^ filteredEvent].

	" ********** MOUSE EVENT *********** "

	lastMouseEvent := filteredEvent.

	"Check for pending drag or double click operations."
	mouseClickState ifNotNil:[
		(mouseClickState handleEvent: filteredEvent from: self) ifFalse:[
			"Possibly dispatched #click: or something and will not re-establish otherwise"
			self mouseOverHandler processMouseOver: lastMouseEvent.
			^ filteredEvent]].

	filteredEvent isMouseWheel ifTrue: [
		mouseWheelState ifNil: [mouseWheelState := MouseWheelState new].
		mouseWheelState handleEvent: filteredEvent from: self.
		self mouseOverHandler processMouseOver: lastMouseEvent.
		^ filteredEvent].

	filteredEvent isMove ifTrue:[
		self position: filteredEvent position.
		self sendMouseEvent: filteredEvent.
		self mouseOverHandler processMouseOver: lastMouseEvent.
		^ filteredEvent].

	"Issue a synthetic move event if we're not at the position of the event"
	filteredEvent position = self position
		ifFalse: [self moveToEvent: filteredEvent].
	
	"Drop submorphs on button events"
	self hasSubmorphs
		ifTrue:[self dropMorphs: filteredEvent]
		ifFalse:[self sendMouseEvent: filteredEvent].

	self mouseOverHandler processMouseOver: lastMouseEvent.
	^ filteredEvent "not necessary but good style -- see Morph >> #handleEvent:"	! !

!HandMorph methodsFor: 'drawing' stamp: 'ls 4/3/2000 20:30'!
hasChanged
	"Return true if this hand has changed, either because it has moved or because some morph it is holding has changed."

	^ hasChanged ifNil: [ true ]
! !

!HandMorph methodsFor: 'drawing' stamp: 'dgd 2/21/2003 22:46'!
hasUserInformation
	^self userInitials notEmpty or: [self userPicture notNil]! !

!HandMorph methodsFor: 'initialization' stamp: 'mt 6/6/2016 17:20'!
initForEvents
	mouseOverHandler := nil.
	lastMouseEvent := MouseEvent new setType: #mouseMove position: 0@0 buttons: 0 hand: self.
	lastEventBuffer := {1. 0. 0. 0. 0. 0. nil. nil}.
	self resetClickState.
	self addKeyboardCaptureFilter: self. "to convert unusual VM events"! !

!HandMorph methodsFor: 'initialization' stamp: 'ar 10/26/2000 14:58'!
initialize
	super initialize.
	self initForEvents.
	keyboardFocus := nil.
	mouseFocus := nil.
	bounds := 0@0 extent: Cursor normal extent.
	userInitials := ''.
	damageRecorder := DamageRecorder new.
	cachedCanvasHasHoles := false.
	temporaryCursor := temporaryCursorOffset := nil.
	self initForEvents.! !

!HandMorph methodsFor: 'initialization' stamp: 'nk 2/14/2004 18:28'!
interrupted
	"Something went wrong - we're about to bring up a debugger. 
	Release some stuff that could be problematic."
	self releaseAllFoci. "or else debugger might not handle clicks"
! !

!HandMorph methodsFor: 'change reporting' stamp: 'ar 12/30/2001 17:32'!
invalidRect: damageRect from: aMorph
	"Note that a change has occurred and record the given damage rectangle relative to the origin this hand's cache."
	hasChanged := true.
	aMorph == self ifTrue:[^self].
	damageRecorder recordInvalidRect: damageRect.
! !

!HandMorph methodsFor: 'events-processing' stamp: 'nk 2/15/2004 09:01'!
isCapturingGesturePoints
	^false! !

!HandMorph methodsFor: 'genie-stubs' stamp: 'nk 3/11/2004 17:45'!
isGenieAvailable
	"Answer whether the Genie gesture recognizer is available for this hand"
	^false! !

!HandMorph methodsFor: 'genie-stubs' stamp: 'nk 3/11/2004 17:46'!
isGenieEnabled
	"Answer whether the Genie gesture recognizer is enabled for this hand"
	^false! !

!HandMorph methodsFor: 'genie-stubs' stamp: 'nk 3/11/2004 17:46'!
isGenieFocused
	"Answer whether the Genie gesture recognizer is auto-focused for this hand"
	^false! !

!HandMorph methodsFor: 'classification' stamp: ''!
isHandMorph

	^ true! !

!HandMorph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:14'!
keyboardBubbleFilters
	Error signal: 'Hand morphs do only have capture filters. Install top-most bubble filters in the world.'! !

!HandMorph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:11'!
keyboardCaptureFilters
	^ keyboardCaptureFilters! !

!HandMorph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:12'!
keyboardCaptureFilters: anArrayOrNil
	keyboardCaptureFilters := anArrayOrNil! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:09'!
keyboardFocus 
	^ keyboardFocus! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/26/2000 01:30'!
keyboardFocus: aMorphOrNil
	keyboardFocus := aMorphOrNil! !

!HandMorph methodsFor: 'multilingual' stamp: 'yo 7/28/2004 21:35'!
keyboardInterpreter

	^keyboardInterpreter ifNil: [keyboardInterpreter := LanguageEnvironment currentPlatform class defaultInputInterpreter]! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/26/2000 01:28'!
keyboardListeners
	^keyboardListeners! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/26/2000 01:27'!
keyboardListeners: anArrayOrNil
	keyboardListeners := anArrayOrNil! !

!HandMorph methodsFor: 'accessing' stamp: 'ar 10/5/2000 23:17'!
lastEvent
	^ lastMouseEvent! !

!HandMorph methodsFor: 'events-debugging' stamp: 'mt 5/20/2016 13:44'!
logEvent: anEvent
	"Update statistics for processed events."
	
	EventStats ifNil:[EventStats := IdentityDictionary new].
	EventStats at: #count put: (EventStats at: #count ifAbsent:[0]) + 1.
	EventStats at: anEvent type put: (EventStats at: anEvent type ifAbsent:[0]) + 1.! !

!HandMorph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:14'!
mouseBubbleFilters
	Error signal: 'Hand morphs do only have capture filters. Install top-most bubble filters in the world.'! !

!HandMorph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:12'!
mouseCaptureFilters
	^ mouseCaptureFilters! !

!HandMorph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:12'!
mouseCaptureFilters: anArrayOrNil
	mouseCaptureFilters := anArrayOrNil! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:10'!
mouseFocus
	^mouseFocus! !

!HandMorph methodsFor: 'focus handling' stamp: 'nk 2/14/2004 18:44'!
mouseFocus: aMorphOrNil
	mouseFocus := aMorphOrNil! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/26/2000 01:28'!
mouseListeners
	^mouseListeners! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/26/2000 01:27'!
mouseListeners: anArrayOrNil
	mouseListeners := anArrayOrNil! !

!HandMorph methodsFor: 'accessing' stamp: 'ar 9/25/2000 14:24'!
mouseOverHandler
	^mouseOverHandler ifNil:[mouseOverHandler := MouseOverHandler new].! !

!HandMorph methodsFor: 'private events' stamp: 'dgd 2/22/2003 14:58'!
mouseTrailFrom: currentBuf 
	"Current event, a mouse event buffer, is about to be processed.  If there are other similar mouse events queued up, then drop them from the queue, and report the positions inbetween."

	| nextEvent trail |
	trail := WriteStream on: (Array new: 1).
	trail nextPut: currentBuf third @ currentBuf fourth.
	[(nextEvent := Sensor peekEvent) isNil] whileFalse: 
			[nextEvent first = currentBuf first 
				ifFalse: [^trail contents	"different event type"].
			nextEvent fifth = currentBuf fifth 
				ifFalse: [^trail contents	"buttons changed"].
			nextEvent sixth = currentBuf sixth 
				ifFalse: [^trail contents	"modifiers changed"].
			"nextEvent is similar.  Remove it from the queue, and check the next."
			nextEvent := Sensor nextEvent.
			trail nextPut: nextEvent third @ nextEvent fourth].
	^trail contents! !

!HandMorph methodsFor: 'private events' stamp: 'ar 10/6/2000 00:08'!
moveToEvent: anEvent
	"Issue a mouse move event to make the receiver appear at the given position"
	self handleEvent: (MouseMoveEvent new
		setType: #mouseMove 
		startPoint: self position 
		endPoint: anEvent position 
		trail: (Array with: self position with: anEvent position)
		buttons: anEvent buttons
		hand: self
		stamp: anEvent timeStamp)! !

!HandMorph methodsFor: 'drawing' stamp: 'ar 6/17/2010 17:00'!
needsToBeDrawn
	"Return true if this hand must be drawn explicitely instead of being drawn via the hardware cursor. This is the case if it (a) it is a remote hand, (b) it is showing a temporary cursor, or (c) it is not empty and there are any visible submorphs. If using the software cursor, ensure that the hardware cursor is hidden."
	"Details:  Return true if this hand has a saved patch to ensure that is is processed by the world. This saved patch will be deleted after one final display pass when it becomes possible to start using the hardware cursor again. This trick gives us one last display cycle to allow us to remove the software cursor and shadow from the display."
	(savedPatch notNil
		or: [ (submorphs anySatisfy: [ :ea | ea visible ])
			or: [ (temporaryCursor notNil and: [hardwareCursor isNil])
				or: [ self hasUserInformation ]]])
		ifTrue: [
			"using the software cursor; hide the hardware one"
			self showHardwareCursor: false.
			^ true].
	^ false
! !

!HandMorph methodsFor: 'focus handling' stamp: 'mt 4/10/2015 13:03'!
newKeyboardFocus: aMorphOrNil
	"Make the given morph the new keyboard focus, canceling the previous keyboard focus if any. If the argument is nil, the current keyboard focus is cancelled."
	
	| oldFocus newFocus |
	oldFocus := self keyboardFocus.
	newFocus := aMorphOrNil ifNotNil: [:m | m keyboardFocusDelegate].
	
	self keyboardFocus: newFocus.
	
	oldFocus == newFocus ifFalse: [
		oldFocus ifNotNil: [:m | m keyboardFocusChange: false].
		newFocus ifNotNil: [:m | m keyboardFocusChange: true]].
	
	newFocus ifNotNil: [:m |
		self compositionWindowManager keyboardFocusForAMorph: m].

	^ newFocus
! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/26/2000 01:32'!
newMouseFocus: aMorphOrNil
	"Make the given morph the new mouse focus, canceling the previous mouse focus if any. If the argument is nil, the current mouse focus is cancelled."
	self mouseFocus: aMorphOrNil.
! !

!HandMorph methodsFor: 'focus handling' stamp: 'dgd 2/21/2003 22:48'!
newMouseFocus: aMorph event: event 
	aMorph isNil 
		ifFalse: [targetOffset := event cursorPoint - aMorph position].
	^self newMouseFocus: aMorph! !

!HandMorph methodsFor: 'accessing' stamp: 'tk 10/20/2004 15:54'!
noButtonPressed
	"Answer whether any mouse button is not being pressed."

	^self anyButtonPressed not! !

!HandMorph methodsFor: 'drawing' stamp: 'nice 12/27/2009 03:11'!
nonCachingFullDrawOn: aCanvas
	
	"A HandMorph has unusual drawing requirements:
		1. the hand itself (i.e., the cursor) appears in front of its submorphs
		2. morphs being held by the hand cast a shadow on the world/morphs below
	The illusion is that the hand plucks up morphs and carries them above the world."
	"Note: This version does not cache an image of the morphs being held by the hand.
	 Thus, it is slower for complex morphs, but consumes less space."

	submorphs isEmpty ifTrue: [^ self drawOn: aCanvas].  "just draw the hand itself"
	aCanvas asShadowDrawingCanvas
		translateBy: self shadowOffset during:[:shadowCanvas| | shadowForm |
		"Note: We use a shadow form here to prevent drawing
		overlapping morphs multiple times using the transparent
		shadow color."
		shadowForm := self shadowForm.
"
shadowForm displayAt: shadowForm offset negated. Display forceToScreen: (0@0 extent: shadowForm extent).
"
		shadowCanvas paintImage: shadowForm at: shadowForm offset.  "draw shadows"
	].
	"draw morphs in front of shadows"
	self drawSubmorphsOn: aCanvas.
	self drawOn: aCanvas.  "draw the hand itself in front of morphs"
! !

!HandMorph methodsFor: 'event handling' stamp: 'ar 9/25/2000 14:27'!
noticeMouseOver: aMorph event: anEvent
	mouseOverHandler ifNil:[^self].
	mouseOverHandler noticeMouseOver: aMorph event: anEvent.! !

!HandMorph methodsFor: 'objects from disk' stamp: 'ar 10/5/2000 19:48'!
objectForDataStream: refStrm
	| dp |
	"I am about to be written on an object file.  Write a path to me in the other system instead."

	(refStrm project world hands includes: self) ifTrue: [
		^ self].	"owned by the project"
	dp := DiskProxy global: #World selector: #primaryHand args: #().
	refStrm replace: self with: dp.
	^ dp
	"Note, when this file is loaded in an MVC project, this will return nil.  The MenuItemMorph that has this in a field will have that item not work.  Maybe warn the user at load time?"! !

!HandMorph methodsFor: 'paste buffer' stamp: 'ar 10/5/2000 19:10'!
objectToPaste
	"It may need to be sent #startRunning by the client"
	^ Cursor wait showWhile: [PasteBuffer veryDeepCopy]

	"PasteBuffer usableDuplicateIn: self world"
! !

!HandMorph methodsFor: 'halo handling' stamp: 'ar 10/24/2000 18:40'!
obtainHalo: aHalo
	"Used for transfering halos between hands"
	| formerOwner |
	self halo == aHalo ifTrue:[^self].
	"Find former owner"
	formerOwner := self world hands detect:[:h| h halo == aHalo] ifNone:[nil].
	formerOwner ifNotNil:[formerOwner releaseHalo: aHalo].
	self halo: aHalo! !

!HandMorph methodsFor: 'paste buffer' stamp: 'ar 10/5/2000 19:10'!
pasteBuffer
	"Return the paste buffer associated with this hand"
	^ PasteBuffer! !

!HandMorph methodsFor: 'paste buffer' stamp: 'ar 10/5/2000 19:11'!
pasteBuffer: aMorphOrNil
	"Set the contents of the paste buffer."
	PasteBuffer := aMorphOrNil.

! !

!HandMorph methodsFor: 'paste buffer' stamp: 'dgd 9/19/2004 13:49'!
pasteMorph

	| aPastee |
	PasteBuffer ifNil: [^ self inform: 'Nothing to paste.' translated].
	self attachMorph: (aPastee := self objectToPaste).
	aPastee align: aPastee center with: self position.
	aPastee player ifNotNil: [aPastee player startRunning]
! !

!HandMorph methodsFor: 'geometry' stamp: 'ar 3/20/2001 20:34'!
position

	^temporaryCursor
		ifNil: [bounds topLeft]
		ifNotNil: [bounds topLeft - temporaryCursorOffset]! !

!HandMorph methodsFor: 'geometry' stamp: 'tpr 12/3/2015 17:26'!
position: aPoint
	"Overridden to align submorph origins to the grid if gridding is on."
	| adjustedPosition delta box |
	adjustedPosition := aPoint.
	temporaryCursor ifNotNil: [adjustedPosition := (self griddedPoint: adjustedPosition) + temporaryCursorOffset].

	"Copied from Morph to avoid owner layoutChanged"
	"Change the position of this morph and and all of its submorphs."
	delta := adjustedPosition - bounds topLeft.
	delta isZero ifTrue: [^ self].  "Null change"
	box := self fullBounds.
	(delta dotProduct: delta) > 100 ifTrue:[
		"e.g., more than 10 pixels moved"
		self invalidRect: box.
		self invalidRect: (box translateBy: delta).
	] ifFalse:[
		self invalidRect: (box merge: (box translateBy: delta)).
	].
	self privateFullMoveBy: delta.
! !

!HandMorph methodsFor: 'event handling' stamp: 'mt 6/10/2016 14:55'!
processEvents
	"Process user input events from the local input devices."

	| evt evtBuf type hadAny |
	ActiveEvent ifNotNil: 
			["Meaning that we were invoked from within an event response.
		Make sure z-order is up to date"

			self mouseOverHandler processMouseOver: lastMouseEvent].
	hadAny := false.
	[(evtBuf := Sensor nextEvent) isNil] whileFalse: 
			[evt := nil.	"for unknown event types"
			type := evtBuf first.
			type = EventTypeMouse
				ifTrue: [evt := self generateMouseEvent: evtBuf].
			type = EventTypeMouseWheel
				ifTrue: [evt := self generateMouseWheelEvent: evtBuf].
			type = EventTypeKeyboard 
				ifTrue: [evt := self generateKeyboardEvent: evtBuf].
			type = EventTypeDragDropFiles 
				ifTrue: [evt := self generateDropFilesEvent: evtBuf].
			type = EventTypeWindow
				ifTrue:[evt := self generateWindowEvent: evtBuf].
			"All other events are ignored"
			(type ~= EventTypeDragDropFiles and: [evt isNil]) ifTrue: [^self].
			evt isNil 
				ifFalse: 
					["Finally, handle it"

					self handleEvent: evt.
					hadAny := true.

					"For better user feedback, return immediately after a mouse event has been processed."
					evt isMouse ifTrue: [^self]]].
	"note: if we come here we didn't have any mouse events"
	mouseClickState notNil 
		ifTrue: 
			["No mouse events during this cycle. Make sure click states time out accordingly"

			mouseClickState handleEvent: lastMouseEvent asMouseMove from: self].
	hadAny 
		ifFalse: 
			["No pending events. Make sure z-order is up to date"

			self mouseOverHandler processMouseOver: lastMouseEvent]! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:09'!
releaseAllFoci
	mouseFocus := nil.
	keyboardFocus := nil.
! !

!HandMorph methodsFor: 'caching' stamp: 'ar 10/26/2000 15:28'!
releaseCachedState
	| oo ui |
	ui := userInitials.
	super releaseCachedState.
	cacheCanvas := nil.
	oo := owner.
	self removeAllMorphs.
	self initialize.	"nuke everything"
	self privateOwner: oo.
	self releaseAllFoci.
	self userInitials: ui andPicture: (self userPicture).! !

!HandMorph methodsFor: 'halo handling' stamp: 'ar 10/24/2000 18:40'!
releaseHalo: aHalo
	"Used for transfering halos between hands"
	self removeProperty: #halo! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:09'!
releaseKeyboardFocus
	"Release the current keyboard focus unconditionally"
	self newKeyboardFocus: nil.
! !

!HandMorph methodsFor: 'focus handling' stamp: 'mt 4/15/2015 16:20'!
releaseKeyboardFocus: aMorph
	"If the given morph had the keyboard focus before, release it"
	self keyboardFocus == aMorph keyboardFocusDelegate
		ifTrue:[self releaseKeyboardFocus].! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:10'!
releaseMouseFocus
	"Release the current mouse focus unconditionally."
	self newMouseFocus: nil.! !

!HandMorph methodsFor: 'focus handling' stamp: 'ar 10/6/2000 00:10'!
releaseMouseFocus: aMorph
	"If the given morph had the mouse focus before, release it"
	self mouseFocus == aMorph ifTrue:[self releaseMouseFocus].! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/24/2000 20:41'!
removeEventListener: anObject
	"Remove anObject from the current event listeners."
	self eventListeners: (self removeListener: anObject from: self eventListeners).! !

!HandMorph methodsFor: 'halo handling' stamp: 'dgd 4/4/2006 16:14'!
removeHalo
	"remove the receiver's halo (if any)"
	| halo |
	halo := self halo.
	halo
		ifNil: [^ self].
	halo delete.
	self removeProperty: #halo! !

!HandMorph methodsFor: 'halo handling' stamp: 'dgd 9/9/2004 22:44'!
removeHaloFromClick: anEvent on: aMorph 
	| halo |
	halo := self halo
				ifNil: [^ self].
	(halo target hasOwner: self)
		ifTrue: [^ self].
	(halo staysUpWhenMouseIsDownIn: aMorph)
		ifFalse: [self removeHalo]! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/24/2000 20:41'!
removeKeyboardListener: anObject
	"Remove anObject from the current keyboard listeners."
	self keyboardListeners: (self removeListener: anObject from: self keyboardListeners).! !

!HandMorph methodsFor: 'events-listening' stamp: 'dgd 2/21/2003 22:48'!
removeListener: anObject from: aListenerGroup 
	"Remove anObject from the given listener group. Return the new group."

	| listeners |
	aListenerGroup ifNil: [^nil].
	listeners := aListenerGroup.
	listeners := listeners copyWithout: anObject.
	listeners := listeners copyWithout: nil.	"obsolete entries"
	listeners isEmpty ifTrue: [listeners := nil].
	^listeners! !

!HandMorph methodsFor: 'events-listening' stamp: 'ar 10/24/2000 20:41'!
removeMouseListener: anObject
	"Remove anObject from the current mouse listeners."
	self mouseListeners: (self removeListener: anObject from: self mouseListeners).! !

!HandMorph methodsFor: 'balloon help' stamp: 'ar 10/6/2000 00:14'!
removePendingBalloonFor: aMorph
	"Get rid of pending balloon help."
	self removeAlarm: #spawnBalloonFor:.
	self deleteBalloonTarget: aMorph.! !

!HandMorph methodsFor: 'halo handling' stamp: 'ar 8/8/2001 14:49'!
removePendingHaloFor: aMorph
	"Get rid of pending balloon help or halo actions."
	self removeAlarm: #spawnMagicHaloFor:.! !

!HandMorph methodsFor: 'balloon help' stamp: 'mt 3/31/2016 10:34'!
resetBalloonHelp: aBalloonMorph
	"Associate a new balloon morph with this hand. Remove all other ones."

	self
		deleteBalloonTarget: nil;
		addBalloonHelp: aBalloonMorph.! !

!HandMorph methodsFor: 'double click support' stamp: 'ar 9/18/2000 17:16'!
resetClickState
	"Reset the double-click detection state to normal (i.e., not waiting for a double-click)."
	mouseClickState := nil.! !

!HandMorph methodsFor: 'initialization' stamp: 'ar 3/3/2001 15:27'!
resourceJustLoaded
	"In case resource relates to me"
	cacheCanvas := nil.! !

!HandMorph methodsFor: 'drawing' stamp: 'ar 6/17/2010 16:28'!
restoreSavedPatchOn: aCanvas 
	"Clear the changed flag and restore the part of the given canvas under this hand from the previously saved patch. If necessary, handle the transition to using the hardware cursor."

	hasChanged := false.
	savedPatch ifNotNil: 
			[aCanvas drawImage: savedPatch at: savedPatch offset.
			self hasUserInformation ifTrue: [^self].	"cannot use hw cursor if so"
			submorphs notEmpty ifTrue: [^self].
			(temporaryCursor notNil and: [hardwareCursor isNil]) ifTrue: [^self].

			"Make the transition to using hardware cursor. Clear savedPatch and
		 report one final damage rectangle to erase the image of the software cursor."
			super invalidRect: (savedPatch offset 
						extent: savedPatch extent + self shadowOffset)
				from: self.
			self showHardwareCursor: true.
			savedPatch := nil]! !

!HandMorph methodsFor: 'drawing' stamp: 'dgd 2/21/2003 22:49'!
savePatchFrom: aCanvas 
	"Save the part of the given canvas under this hand as a Form and return its bounding rectangle."

	"Details: The previously used patch Form is recycled when possible to reduce the burden on storage management."

	| damageRect myBnds |
	damageRect := myBnds := self fullBounds.
	savedPatch ifNotNil: 
			[damageRect := myBnds merge: (savedPatch offset extent: savedPatch extent)].
	(savedPatch isNil or: [savedPatch extent ~= myBnds extent]) 
		ifTrue: 
			["allocate new patch form if needed"

			savedPatch := aCanvas form allocateForm: myBnds extent].
	aCanvas contentsOfArea: (myBnds translateBy: aCanvas origin)
		into: savedPatch.
	savedPatch offset: myBnds topLeft.
	^damageRect! !

!HandMorph methodsFor: 'selected object' stamp: 'dgd 8/28/2004 16:30'!
selectedObject
	"answer the selected object for the hand or nil is none"
	| halo |
	halo := self halo.
	halo isNil
		ifTrue: [^ nil].
	^ halo target renderedMorph! !

!HandMorph methodsFor: 'private events' stamp: 'ar 3/18/2001 01:43'!
sendEvent: anEvent focus: focusHolder
	"Send the event to the morph currently holding the focus, or if none to the owner of the hand."
	^self sendEvent: anEvent focus: focusHolder clear:[nil]! !

!HandMorph methodsFor: 'private events' stamp: 'mt 5/20/2016 13:21'!
sendEvent: anEvent focus: focusHolder clear: aBlock
	"Send the event to the morph currently holding the focus, or if none to the owner of the hand."
	| result |
	focusHolder ifNotNil:[^self sendFocusEvent: anEvent to: focusHolder clear: aBlock].
	ActiveEvent := anEvent.
	[result := owner processEvent: anEvent]
		ensure: [ActiveEvent := nil].
	^ result == #rejected ifTrue: [anEvent] ifFalse: [result "filtered event"]! !

!HandMorph methodsFor: 'events-processing' stamp: 'mt 5/21/2016 09:47'!
sendFilterEventCaptureAgain: anEvent for: anObjectOrNil
	"Apply keyboard-specific and mouse-specific capturing filters. If a filter changes the event type, filter again. WARNING: This is a powerful mechanism. Filters can create endless loops, which are difficult to debug.
	
	Overwritten to provide keyboard and mouse focus holders to event filters. Note that hence the event target in such global filters can be nil."

	| filteredEvent |
	filteredEvent := anEvent.
	
	filteredEvent isKeyboard ifTrue: [
		filteredEvent := self sendFilterEvent: filteredEvent for: self keyboardFocus to: self keyboardCaptureFilters.
		filteredEvent isKeyboard not ifTrue: [^ self sendFilterEventCaptureAgain: filteredEvent for: anObjectOrNil]].
		
	filteredEvent isMouse ifTrue: [
		filteredEvent := self sendFilterEvent: filteredEvent for: self mouseFocus to: self mouseCaptureFilters.
		filteredEvent isMouse not ifTrue: [^ self sendFilterEventCaptureAgain: filteredEvent for: anObjectOrNil]].

	^ filteredEvent! !

!HandMorph methodsFor: 'private events' stamp: 'mt 6/20/2016 13:56'!
sendFocusEvent: anEvent to: focusHolder clear: aBlock
	"Send the event to the morph currently holding the focus"

	| result w |
	w := focusHolder world ifNil:[aBlock value. ^ anEvent].
	w becomeActiveDuring:[
		ActiveHand := self.
		ActiveEvent := anEvent.
		result := focusHolder processFocusEvent: anEvent.
	].
	^ result == #rejected ifTrue: [anEvent] ifFalse: [result "filtered event"]! !

!HandMorph methodsFor: 'private events' stamp: 'ar 1/2/2010 15:22'!
sendKeyboardEvent: anEvent 
	"Send the event to the morph currently holding the focus, or if none to
	the owner of the hand."
	^ self
		sendEvent: anEvent
		focus: self keyboardFocus
		clear: [self keyboardFocus: nil]! !

!HandMorph methodsFor: 'private events' stamp: 'mt 5/20/2016 14:01'!
sendListenEvents: anEvent
	"Send the given event to all registered event listeners."
	
	self sendListenEvent: anEvent to: self eventListeners.
	
	anEvent isKeyboard
		ifTrue: [self sendListenEvent: anEvent to: self keyboardListeners].
		
	anEvent isMouse
		ifTrue: [self sendListenEvent: anEvent to: self mouseListeners].! !

!HandMorph methodsFor: 'private events' stamp: 'ar 10/26/2000 01:43'!
sendListenEvent: anEvent to: listenerGroup
	"Send the event to the given group of listeners"
	listenerGroup ifNil:[^self].
	listenerGroup do:[:listener| 
		listener ifNotNil:[listener handleListenEvent: anEvent copy]].! !

!HandMorph methodsFor: 'private events' stamp: 'ar 3/18/2001 01:45'!
sendMouseEvent: anEvent
	"Send the event to the morph currently holding the focus, or if none to the owner of the hand."
	^self sendEvent: anEvent focus: self mouseFocus clear:[self mouseFocus: nil]! !

!HandMorph methodsFor: 'drawing' stamp: 'JW 7/12/2005 20:13'!
shadowForm
	"Return a 1-bit shadow of my submorphs.  Assumes submorphs is not empty"
	| bnds canvas |
	bnds := Rectangle merging: (submorphs collect: [:m | m fullBounds]).
	canvas := (Display defaultCanvasClass extent: bnds extent depth: 1) 
		asShadowDrawingCanvas: Color black.
	canvas translateBy: bnds topLeft negated
		during:[:tempCanvas| self drawSubmorphsOn: tempCanvas].
	^ canvas form offset: bnds topLeft! !

!HandMorph methodsFor: 'drop shadows' stamp: ''!
shadowOffset

	^ 6@8! !

!HandMorph methodsFor: 'events-debugging' stamp: 'mt 6/10/2016 10:21'!
showEvent: anEvent
	"Show details about the event on the display form. Useful for debugging."
	
	| message borderWidth |
	ShowEvents == true ifFalse: [^ self].
	
	borderWidth := 5.
	message := String streamContents: [:strm |
		strm
			nextPutAll: '[HandMorph >> #showEvent:]'; cr;
			nextPutAll: 'event'; tab; tab; tab; tab; nextPutAll: anEvent printString; cr;
			nextPutAll: 'keyboard focus'; tab; nextPutAll: self keyboardFocus printString; cr;
			nextPutAll: 'mouse focus'; tab; tab; nextPutAll: self mouseFocus printString].
		
	message := message asDisplayText
		foregroundColor: Color black
		backgroundColor: Color white.
	
	"Offset to support multiple hands debugging."
	Display fill: (0 @ 0 extent: message form extent + (borderWidth asPoint * 2)) rule: Form over fillColor: Color white.
	message displayOn: Display at: borderWidth asPoint + (0 @  ((owner hands indexOf: self) - 1 * message form height)).! !

!HandMorph methodsFor: 'drawing' stamp: 'fbs 7/8/2013 22:48'!
showHardwareCursor: aBool
	"Show/hide the current hardware cursor as indicated."
	| cursor |
	cursor :=  hardwareCursor ifNil:[aBool ifTrue:[Cursor normal] ifFalse:[Cursor blank]].
	Cursor currentCursor == cursor ifFalse: [cursor show].
! !

!HandMorph methodsFor: 'cursor' stamp: 'di 3/6/1999 23:52'!
showTemporaryCursor: cursorOrNil
	"Set the temporary cursor to the given Form. If the argument is nil, revert to the normal cursor."

	self showTemporaryCursor: cursorOrNil hotSpotOffset: 0@0
! !

!HandMorph methodsFor: 'cursor' stamp: 'ar 6/17/2010 17:01'!
showTemporaryCursor: cursorOrNil hotSpotOffset: hotSpotOffset 
	"Set the temporary cursor to the given Form.
	If the argument is nil, revert to the normal hardware cursor."

	self changed.
	temporaryCursorOffset 
		ifNotNil: [bounds := bounds translateBy: temporaryCursorOffset negated].
	cursorOrNil isNil 
		ifTrue: [temporaryCursor := temporaryCursorOffset := hardwareCursor := nil]
		ifFalse: 
			[temporaryCursor := cursorOrNil asCursorForm.
			temporaryCursorOffset := temporaryCursor offset - hotSpotOffset.
			(cursorOrNil isKindOf: Cursor) ifTrue: [hardwareCursor := cursorOrNil]].
	bounds := self cursorBounds.
	self
		userInitials: userInitials andPicture: self userPicture;
		layoutChanged;
		changed;
		showHardwareCursor: (temporaryCursor isNil).! !

!HandMorph methodsFor: 'balloon help' stamp: 'mt 4/11/2015 07:48'!
spawnBalloonFor: aMorph

	aMorph wantsBalloon ifFalse: [^ self].
	aMorph showBalloon: aMorph balloonText hand: self.! !

!HandMorph methodsFor: 'halo handling' stamp: 'ar 8/8/2001 14:50'!
spawnMagicHaloFor: aMorph
	(self halo notNil and:[self halo target == aMorph]) ifTrue:[^self].
	aMorph addMagicHaloFor: self.! !

!HandMorph methodsFor: 'accessing' stamp: ''!
targetOffset
	"Return the offset of the last mouseDown location relative to the origin of the recipient morph. During menu interactions, this is the absolute location of the mouse down event that invoked the menu."

	^ targetOffset
! !

!HandMorph methodsFor: 'grabbing/dropping' stamp: 'di 10/16/2000 10:46'!
targetOffset: offsetPoint
	"Set the offset at which we clicked down in the target morph"

	targetOffset := offsetPoint! !

!HandMorph methodsFor: 'accessing' stamp: 'wiz 12/4/2006 00:16'!
targetPoint
	"Return the new position of the target.
	I.E. return the position of the hand less 
	the original distance between hand and target position"

	^ self position - targetOffset
! !

!HandMorph methodsFor: 'cursor' stamp: 'NS 2/17/2001 11:01'!
temporaryCursor
	^ temporaryCursor! !

!HandMorph methodsFor: 'balloon help' stamp: 'ar 10/3/2000 17:14'!
triggerBalloonFor: aMorph after: timeOut
	"Trigger balloon help after the given time out for some morph"
	self addAlarm: #spawnBalloonFor: with: aMorph after: timeOut.! !

!HandMorph methodsFor: 'halo handling' stamp: 'ar 8/8/2001 14:51'!
triggerHaloFor: aMorph after: timeOut
	"Trigger automatic halo after the given time out for some morph"
	self addAlarm: #spawnMagicHaloFor: with: aMorph after: timeOut! !

!HandMorph methodsFor: 'gridded cursor' stamp: 'tpr 12/3/2015 18:27'!
turnOffGridding
	"remove the gridding properties to stop gridding"
	self removeProperty: #gridStep; removeProperty: #gridOffset! !

!HandMorph methodsFor: 'drawing' stamp: 'dgd 2/21/2003 22:49'!
updateCacheCanvas: aCanvas 
	"Update the cached image of the morphs being held by this hand."

	"Note: The following is an attempt to quickly get out if there's no change"

	| subBnds rectList nPix |
	subBnds := Rectangle merging: (submorphs collect: [:m | m fullBounds]).
	rectList := damageRecorder invalidRectsFullBounds: subBnds.
	damageRecorder reset.
	(rectList isEmpty 
		and: [cacheCanvas notNil and: [cacheCanvas extent = subBnds extent]]) 
			ifTrue: [^self].

	"Always check for real translucency -- can't be cached in a form"
	self submorphsDo: 
			[:m | 
			m wantsToBeCachedByHand 
				ifFalse: 
					[cacheCanvas := nil.
					cachedCanvasHasHoles := true.
					^self]].
	(cacheCanvas isNil or: [cacheCanvas extent ~= subBnds extent]) 
		ifTrue: 
			[cacheCanvas := (aCanvas allocateForm: subBnds extent) getCanvas.
			cacheCanvas translateBy: subBnds origin negated
				during: [:tempCanvas | self drawSubmorphsOn: tempCanvas].
			self submorphsDo: 
					[:m | 
					(m areasRemainingToFill: subBnds) isEmpty 
						ifTrue: [^cachedCanvasHasHoles := false]].
			nPix := cacheCanvas form tallyPixelValues first.
			"--> begin rounded corners hack <---"
			cachedCanvasHasHoles := (nPix = 48 
						and: [submorphs size = 1 and: [submorphs first wantsRoundedCorners]]) 
							ifTrue: [false]
							ifFalse: [nPix > 0].
			"--> end rounded corners hack <---"
			^self].

	"incrementally update the cache canvas"
	cacheCanvas translateBy: subBnds origin negated
		during: 
			[:cc | 
			rectList do: 
					[:r | 
					cc clipBy: r
						during: 
							[:c | 
							c fillColor: Color transparent.
							self drawSubmorphsOn: c]]]! !

!HandMorph methodsFor: 'accessing' stamp: ''!
userInitials

	^ userInitials! !

!HandMorph methodsFor: 'geometry' stamp: 'ar 12/30/2001 20:44'!
userInitials: aString andPicture: aForm

	| cb pictRect initRect f |

	userInitials := aString.
	pictRect := initRect := cb := self cursorBounds.
	userInitials isEmpty ifFalse: [
		f := TextStyle defaultFont.
		initRect := cb topRight + (0@4) extent: (f widthOfString: userInitials)@(f height).
	].
	self userPicture: aForm.
	aForm ifNotNil: [
		pictRect := (self cursorBounds topRight + (0@24)) extent: aForm extent.
	].
	self bounds: ((cb merge: initRect) merge: pictRect).


! !

!HandMorph methodsFor: 'accessing' stamp: 'ar 10/26/2000 15:18'!
userPicture
	^self valueOfProperty: #remoteUserPicture

! !

!HandMorph methodsFor: 'accessing' stamp: 'ar 10/26/2000 15:34'!
userPicture: aFormOrNil
	^self setProperty: #remoteUserPicture toValue: aFormOrNil
! !

!HandMorph methodsFor: 'copying' stamp: 'ar 10/6/2000 00:11'!
veryDeepCopyWith: deepCopier
	"Return self.  Do not copy hands this way."
	^ self! !

!HandMorph methodsFor: 'drawing' stamp: 'nk 10/24/2003 22:12'!
visible: aBoolean
	self needsToBeDrawn ifFalse: [ ^self ].
	super visible: aBoolean! !

!HandMorph methodsFor: 'double click support' stamp: 'mtf 2/20/2011 20:41'!
waitForClicksOrDrag: aMorph event: evt
	"Wait for mouse button and movement events, informing aMorph about events interesting to it via callbacks.
	This message is typically sent to the Hand by aMorph when it first receives a mouse-down event.
	The callback methods invoked on aMorph (which are passed a copy of evt) are:
		#click:	sent when the mouse button goes up within doubleClickTime.
		#doubleClick:	sent when the mouse goes up, down, and up again all within DoubleClickTime.
		#doubleClickTimeout:  sent when the mouse does not have a doubleClick within DoubleClickTime.
		#startDrag:	sent when the mouse moves more than 10 pixels from evt's position within DoubleClickTime.
	Note that mouseMove: and mouseUp: events are not sent to aMorph until it becomes the mouse focus,
	which is typically done by aMorph in its click:, doubleClick:, or drag: methods."
	
	^self waitForClicksOrDrag: aMorph event: evt selectors: #( #click: #doubleClick: #doubleClickTimeout: #startDrag:) threshold: HandMorph dragThreshold! !

!HandMorph methodsFor: 'double click support' stamp: 'nk 7/26/2004 10:32'!
waitForClicksOrDrag: aMorph event: evt selectors: clickAndDragSelectors threshold: threshold

	"Wait for mouse button and movement events, informing aMorph about events interesting to it via callbacks.
	This message is typically sent to the Hand by aMorph when it first receives a mouse-down event.
	The callback methods, named in clickAndDragSelectors and passed a copy of evt, are:
		1 	(click) sent when the mouse button goes up within doubleClickTime.
		2	(doubleClick) sent when the mouse goes up, down, and up again all within DoubleClickTime.
		3	(doubleClickTimeout) sent when the mouse does not have a doubleClick within DoubleClickTime.
		4	(startDrag) sent when the mouse moves more than threshold pixels from evt's position within DoubleClickTime.
	Note that mouseMove: and mouseUp: events are not sent to aMorph until it becomes the mouse focus,
	which is typically done by aMorph in its click:, doubleClick:, or drag: methods."
	
	mouseClickState := 
		MouseClickState new
			client: aMorph 
			click: clickAndDragSelectors first 
			dblClick: clickAndDragSelectors second 
			dblClickTime: DoubleClickTime 
			dblClickTimeout: clickAndDragSelectors third
			drag: clickAndDragSelectors fourth 
			threshold: threshold 
			event: evt.
! !

Morph subclass: #HSVAColorSelectorMorph
	instanceVariableNames: 'hsvMorph aMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!HSVAColorSelectorMorph commentStamp: 'gvc 5/18/2007 12:55' prior: 0!
Colour selector featuring a saturation/volume area, hue selection strip and alpha selection strip.!

!HSVAColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/22/2006 09:58'!
alphaSelected: aFloat
	"The alpha has changed."

	self triggerSelectedColor! !

!HSVAColorSelectorMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 13:43'!
aMorph
	"The alpha-selector morph."
	^ aMorph! !

!HSVAColorSelectorMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 13:43'!
aMorph: anAColorSelectorMorph
	"The alpha-selector morph."
	aMorph := anAColorSelectorMorph! !

!HSVAColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/22/2006 09:58'!
colorSelected: aColor
	"A color has been selected. Set the base color for the alpha channel."

	self aMorph color: aColor.
	self triggerSelectedColor! !

!HSVAColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 13:44'!
defaultColor
	"Answer the default color/fill style for the receiver."
	
	^Color transparent
! !

!HSVAColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/20/2006 14:05'!
hsvMorph
	"Answer the value of hsvMorph"

	^ hsvMorph! !

!HSVAColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/20/2006 14:05'!
hsvMorph: anObject
	"Set the value of hsvMorph"

	hsvMorph := anObject! !

!HSVAColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 16:47'!
initialize
	"Initialize the receiver."

	super initialize.
	self
		extent: 180@168;
		changeTableLayout;
		cellInset: 4;
		aMorph: self newAColorMorph;
		hsvMorph: self newHSVColorMorph;
		addMorphBack: self hsvMorph;
		addMorphBack: self aMorph.
	self aMorph color: self hsvMorph selectedColor! !

!HSVAColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'mt 4/17/2015 14:08'!
newAColorMorph
	"Answer a new alpha color morph."

	^AColorSelectorMorph new
		model: self;
		hResizing: #spaceFill;
		vResizing: #rigid;
		setValueSelector: #alphaSelected:;
		extent: 50@24! !

!HSVAColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 14:00'!
newHSVColorMorph
	"Answer a new hue/saturation/volume color morph."

	^HSVColorSelectorMorph new
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		when: #colorSelected send: #colorSelected: to: self! !

!HSVAColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 14:18'!
selectedColor
	"Answer the selected color."

	^self hsvMorph selectedColor alpha: self aMorph value! !

!HSVAColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/22/2006 11:12'!
selectedColor: aColor
	"Set the hue and sv components."

	self aMorph value: aColor alpha.
	self hsvMorph selectedColor: aColor asNontranslucentColor! !

!HSVAColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'cmm 12/3/2010 14:24'!
triggerSelectedColor
	"Trigger the event for the selected colour"
	self
		triggerEvent: #selectedColor
		with: self selectedColor.
	self changed: #selectedColor! !

Morph subclass: #HSVColorSelectorMorph
	instanceVariableNames: 'svMorph hMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!HSVColorSelectorMorph commentStamp: 'gvc 5/18/2007 12:55' prior: 0!
Colour selector featuring a saturation/volume area and a hue selection strip.!

!HSVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 16:28'!
colorSelected: aColor
	"A color has been selected. Make the hue match."

	"self hMorph value: aColor hue / 360.
	self svMorph basicColor: (Color h: aColor hue s: 1.0 v: 1.0)."
	self triggerEvent: #colorSelected with: aColor! !

!HSVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 13:44'!
defaultColor
	"Answer the default color/fill style for the receiver."
	
	^Color transparent
! !

!HSVColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/21/2006 13:30'!
hMorph
	"Answer the value of hMorph"

	^ hMorph! !

!HSVColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/21/2006 13:30'!
hMorph: anObject
	"Set the value of hMorph"

	hMorph := anObject! !

!HSVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 16:23'!
hue: aFloat
	"Set the hue in the range 0.0 - 1.0. Update the SV morph and hMorph."

	self hMorph value: aFloat.
	self svMorph color: (Color h: aFloat * 359.9 s: 1.0 v: 1.0)! !

!HSVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/26/2006 12:25'!
initialize
	"Initialize the receiver."

	super initialize.
	self
		borderWidth: 0;
		changeTableLayout;
		cellInset: 4;
		listDirection: #leftToRight;
		cellPositioning: #topLeft;
		svMorph: self newSVColorMorph;
		hMorph: self newHColorMorph;
		addMorphBack: self svMorph;
		addMorphBack: self hMorph;
		extent: 192@152;
		hue: 0.5! !

!HSVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/19/2006 12:33'!
newHColorMorph
	"Answer a new hue color morph."

	^HColorSelectorMorph new
		model: self;
		setValueSelector: #hue:;
		hResizing: #rigid;
		vResizing: #spaceFill;
		extent: 36@36! !

!HSVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 5/18/2007 12:54'!
newSVColorMorph
	"Answer a new saturation/volume color morph."

	^SVColorSelectorMorph new
		extent: 152@152;
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		when: #colorSelected send: #colorSelected: to: self! !

!HSVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 13:41'!
selectedColor
	"Answer the selected color."

	^self svMorph selectedColor! !

!HSVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 13:38'!
selectedColor: aColor
	"Set the hue and sv components."

	self hue: aColor hue / 360.
	self svMorph selectedColor: aColor! !

!HSVColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/19/2006 12:26'!
svMorph
	"Answer the value of svMorph"

	^ svMorph! !

!HSVColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/19/2006 12:26'!
svMorph: anObject
	"Set the value of svMorph"

	svMorph := anObject! !

Morph subclass: #ImageMorph
	instanceVariableNames: 'image'
	classVariableNames: 'DefaultForm'
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!ImageMorph commentStamp: 'efc 3/7/2003 17:48' prior: 0!
ImageMorph is a morph that displays a picture (Form). My extent is determined by the extent of my form.

Use #image: to set my picture.

Structure:
 instance var		Type 		Description
 image				Form		The Form to use when drawing

Code examples:
	ImageMorph new openInWorld; grabFromScreen

	(Form fromFileNamed: 'myGraphicsFileName') asMorph openInWorld

Relationship to SketchMorph: ImageMorph should be favored over SketchMorph, a parallel, legacy class -- see the Swiki FAQ for details ( http://minnow.cc.gatech.edu/squeak/1372 ). 
!

!ImageMorph class methodsFor: 'scripting' stamp: 'sw 5/19/1998 18:30'!
authoringPrototype
	| aMorph aForm |
	aMorph := super authoringPrototype.
	aForm := ScriptingSystem formAtKey: 'Image'.
	aForm ifNil: [aForm := aMorph image rotateBy: 90].
	aMorph image: aForm.
	^ aMorph! !

!ImageMorph class methodsFor: 'accessing' stamp: 'ar 6/25/1999 11:59'!
defaultForm
	^DefaultForm! !

!ImageMorph class methodsFor: 'parts bin' stamp: 'tak 7/25/2007 00:06'!
descriptionForPartsBin
	^ self partName:	'Image' translatedNoop
		categories:		#()
		documentation:	'A non-editable picture.  If you use the Paint palette to make a picture, you can edit it afterwards.' translatedNoop! !

!ImageMorph class methodsFor: 'instance creation' stamp: 'sw 10/23/2000 18:21'!
fromString: aString 
	"Create a new ImageMorph which displays the input string in the standard button font"

	^ self fromString: aString font: Preferences standardButtonFont! !

!ImageMorph class methodsFor: 'instance creation' stamp: 'sw 10/23/2000 18:21'!
fromString: aString font: aFont
	"Create a new ImageMorph showing the given string in the given font"

	^ self new image: (StringMorph contents: aString font: aFont) imageForm! !

!ImageMorph class methodsFor: 'class initialization' stamp: 'asm 4/10/2003 13:09'!
initialize
	"ImageMorph initialize"

	| h p d |
	DefaultForm := (Form extent: 80@40 depth: 16).
	h := DefaultForm height // 2.
	0 to: h - 1 do: [:i |
		p := (i * 2)@i.
		d := i asFloat / h asFloat.
		DefaultForm fill:
			(p corner: DefaultForm extent - p)
			fillColor: (Color r: d g: 0.5 b: 1.0 - d)].

	self registerInFlapsRegistry.! !

!ImageMorph class methodsFor: 'class initialization' stamp: 'KR 8/12/2007 10:53'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | cl registerQuad: {#ImageMorph. #authoringPrototype. 'Picture' translatedNoop. 'A non-editable picture of something' translatedNoop} 
						forFlapNamed: 'Supplies']! !

!ImageMorph class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 12:36'!
unload
	"Unload the receiver from global registries"

	self environment at: #Flaps ifPresent: [:cl |
	cl unregisterQuadsWithReceiver: self] ! !

!ImageMorph methodsFor: 'menus' stamp: 'ar 11/7/2000 14:55'!
addCustomMenuItems: aMenu hand: aHand
	super addCustomMenuItems: aMenu hand: aHand.
	aMenu addUpdating: #opacityString action: #changeOpacity! !

!ImageMorph methodsFor: 'accessing' stamp: 'ar 12/12/2001 01:08'!
borderStyle: newStyle
	| newExtent |
	newExtent := 2 * newStyle width + image extent.
	bounds extent = newExtent ifFalse:[super extent: newExtent].
	super borderStyle: newStyle.! !

!ImageMorph methodsFor: 'accessing' stamp: 'ar 12/12/2001 01:11'!
borderWidth: bw
	| newExtent |
	newExtent := 2 * bw + image extent.
	bounds extent = newExtent ifFalse:[super extent: newExtent].
	super borderWidth: bw! !

!ImageMorph methodsFor: 'menu' stamp: 'ar 11/7/2000 14:57'!
changeOpacity
	self isOpaque: self isOpaque not! !

!ImageMorph methodsFor: 'accessing' stamp: 'ul 11/15/2010 11:40'!
color: aColor
        super color: aColor.
        (image depth = 1 and: [aColor isColor]) ifTrue: [
                image colors: {Color transparent. aColor}.
                self changed]! !

!ImageMorph methodsFor: 'drawing' stamp: 'dgd 9/7/2004 17:24'!
drawOn: aCanvas
	| style |
	(style := self borderStyle) ifNotNil:[
		style frameRectangle: bounds on: aCanvas.
	].
	self isOpaque
		ifTrue:[aCanvas drawImage: image at: self innerBounds origin]
		ifFalse:[aCanvas translucentImage: image at: self innerBounds origin]! !

!ImageMorph methodsFor: 'geometry' stamp: 'mt 8/1/2016 16:46'!
extent: aPoint

	super extent: self preferredExtent.! !

!ImageMorph methodsFor: 'accessing' stamp: ''!
form
	"For compatability with SketchMorph."

	^ image
! !

!ImageMorph methodsFor: 'menu commands' stamp: ''!
grabFromScreen

	self image: Form fromUser.
! !

!ImageMorph methodsFor: 'accessing' stamp: 'jm 9/27/97 20:16'!
image

	^ image
! !

!ImageMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:34'!
image: anImage 
	self changed.
	image := anImage depth = 1 
				ifTrue: [ColorForm mappingWhiteToTransparentFrom: anImage]
				ifFalse: [anImage]. 
	super extent: self preferredExtent! !

!ImageMorph methodsFor: 'initialization' stamp: ''!
initialize

	super initialize.
	self image: DefaultForm.
! !

!ImageMorph methodsFor: 'parts bin' stamp: 'sw 6/28/2001 11:32'!
initializeToStandAlone

	super initializeToStandAlone.
	self image: DefaultForm.
! !

!ImageMorph methodsFor: 'testing' stamp: 'eem 6/12/2013 15:29'!
isImageMorph
	^true! !

!ImageMorph methodsFor: 'accessing' stamp: 'dgd 2/16/2003 20:01'!
isOpaque
	"Return true if the receiver is marked as being completely opaque"
	^ self
		valueOfProperty: #isOpaque
		ifAbsent: [false]! !

!ImageMorph methodsFor: 'accessing' stamp: 'ar 11/7/2000 14:57'!
isOpaque: aBool
	"Mark the receiver as being completely opaque or not"
	aBool == false
		ifTrue:[self removeProperty: #isOpaque]
		ifFalse:[self setProperty: #isOpaque toValue: aBool].
	self changed! !

!ImageMorph methodsFor: 'other' stamp: 'sw 12/17/1998 12:11'!
newForm: aForm
	self image: aForm! !

!ImageMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 21:46'!
opacityString
	^ (self isOpaque
		ifTrue: ['<on>']
		ifFalse: ['<off>']), 'opaque' translated! !

!ImageMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:34'!
preferredExtent
	^ image ifNil: [0@0] ifNotNil: [image extent]! !

!ImageMorph methodsFor: 'menu commands' stamp: 'ar 8/6/2009 20:40'!
readFromFile
	| fileName |
	fileName := UIManager default
		request: 'Please enter the image file name'
		initialAnswer: 'fileName'.
	fileName isEmpty ifTrue: [^ self].
	self image: (Form fromFileNamed: fileName).
! !

!ImageMorph methodsFor: 'caching' stamp: ''!
releaseCachedState

	super releaseCachedState.
	image hibernate.
! !

!ImageMorph methodsFor: 'accessing' stamp: 'sw 10/24/2000 05:53'!
setNewImageFrom: formOrNil
	"Change the receiver's image to be one derived from the supplied form.  If nil is supplied, clobber any existing image in the receiver, and in its place put a default graphic, either the one known to the receiver as its default value, else a squeaky mouse"

	|  defaultImage |
	formOrNil ifNotNil: [^ self image: formOrNil].
	defaultImage := self defaultValueOrNil ifNil: [ScriptingSystem squeakyMouseForm].
	self image: defaultImage
! !

!ImageMorph methodsFor: 'accessing' stamp: 'eem 6/12/2013 15:30'!
wantsRecolorHandle
	^ image notNil and: [image depth = 1]! !

!ImageMorph methodsFor: 'accessing' stamp: 'wiz 4/7/2004 15:10'!
withSnapshotBorder
	self borderStyle: ((ComplexBorder style: #complexFramed)
			color: (Color
					r: 0.613
					g: 1.0
					b: 0.516);
			 width: 1;
			
			 yourself)! !

ImageMorph subclass: #ProjectViewMorph
	instanceVariableNames: 'project lastProjectThumbnail'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!ProjectViewMorph commentStamp: '<historical>' prior: 0!
I am a Morphic view of a project. I display a scaled version of the project's thumbnail, which itself is a scaled-down snapshot of the screen taken when the project was last exited. When I am displayed, I check to see if the project thumbnail has changed and, if so, I update my own view of that thumbnail.
!

!ProjectViewMorph class methodsFor: 'scripting' stamp: 'KR 5/23/2008 01:30'!
defaultNameStemForInstances
	^ 'ProjectView' translatedNoop! !

!ProjectViewMorph class methodsFor: 'fileIn/Out' stamp: 'nk 7/16/2003 15:55'!
fileReaderServicesForFile: fullName suffix: suffix

	^({ 'extseg'. 'project'. 'pr'. 'morph'. 'morphs'. 'sp'. '*' } includes: suffix)
		ifTrue: [ self services]
		ifFalse: [#()]! !

!ProjectViewMorph class methodsFor: 'class initialization' stamp: 'hg 8/3/2000 16:54'!
initialize

	FileList registerFileReader: self! !

!ProjectViewMorph class methodsFor: 'project window creation' stamp: 'sw 4/24/2001 12:00'!
newMVCProject
	"Create an instance of me on a new MVC project (in a SystemWindow)."

	| proj window |
	proj := Project new.
	window := (SystemWindow labelled: proj name) model: proj.
	window
		addMorph: (self on: proj)
		frame: (0@0 corner: 1.0@1.0).
	^ window
! !

!ProjectViewMorph class methodsFor: 'project window creation' stamp: 'ar 8/31/2004 20:53'!
newProjectViewInAWindowFor: aProject
	"Return an instance of me on a new Morphic project (in a SystemWindow)."

	| window proj |
	proj := self on: aProject.
	window := (SystemWindow labelled: aProject name) model: aProject.
	window
		addMorph: proj
		frame: (0@0 corner: 1.0@1.0).
	proj borderWidth: 0.
	^ window
! !

!ProjectViewMorph class methodsFor: 'instance creation' stamp: 'jm 5/14/1998 16:19'!
on: aProject

	^ self new on: aProject
! !

!ProjectViewMorph class methodsFor: 'project window creation' stamp: 'RAA 2/2/2002 08:30'!
openFromDirectoryAndFileName: anArray
	
	Project canWeLoadAProjectNow ifFalse: [^ self].
	^ProjectLoading 
		openFromDirectory: anArray first 
		andFileName: anArray second! !

!ProjectViewMorph class methodsFor: 'project window creation' stamp: 'RAA 2/2/2002 08:11'!
openFromDirectory: aDirectory andFileName: aFileName
	
	Project canWeLoadAProjectNow ifFalse: [^ self].
	^ProjectLoading openFromDirectory: aDirectory andFileName: aFileName! !

!ProjectViewMorph class methodsFor: 'project window creation' stamp: 'RAA 2/22/2002 06:19'!
openFromFileList: fullName
	
	self flag: #bob.		"not sent??"

	self halt.

	^self openFromFile:  fullName! !

!ProjectViewMorph class methodsFor: 'project window creation' stamp: 'RAA 2/22/2002 06:19'!
openFromFile: fileName
	
	self flag: #bob.		"better not to use this one. nil directories are not nice.
						see #openFromDirectoryAndFileName: or 
						#openFromDirectory:andFileName: instead"

	self halt.

	Project canWeLoadAProjectNow ifFalse: [^ self].
	^ProjectLoading openFromDirectory: nil andFileName: fileName! !

!ProjectViewMorph class methodsFor: 'instance creation' stamp: 'ar 9/18/2009 21:50'!
openOn: aProject
	"Open a ProjectViewMorph for the project in question"
	ProjectViewOpenNotification signal ifTrue: [
		Preferences projectViewsInWindows ifTrue: [
			(self newProjectViewInAWindowFor: aProject) openInWorld
		] ifFalse: [
			(self on: aProject) openInWorld		"but where??"
		].
	].
! !

!ProjectViewMorph class methodsFor: 'project window creation' stamp: 'KR 3/21/2008 20:53'!
serviceOpenProjectFromFile
	"Answer a service for opening a .pr project file"

	^ (SimpleServiceEntry 
		provider: self 
		label: 'load as project' translatedNoop
		selector: #openFromDirectoryAndFileName:
		description: 'open project from file' translatedNoop
		buttonLabel: 'load' translatedNoop
	)
		argumentGetter: [ :fileList | fileList dirAndFileName]! !

!ProjectViewMorph class methodsFor: 'fileIn/Out' stamp: 'sd 2/1/2002 22:01'!
services

	^ Array with: self serviceOpenProjectFromFile

	! !

!ProjectViewMorph class methodsFor: 'initialize-release' stamp: 'SD 11/15/2001 22:22'!
unload

	FileList unregisterFileReader: self ! !

!ProjectViewMorph methodsFor: 'submorphs-add/remove' stamp: 'tk 12/9/1999 08:03'!
abandon
	"Home ViewMorph of project is going away."

	project := nil.
	super abandon.

! !

!ProjectViewMorph methodsFor: 'layout' stamp: 'sw 8/16/2007 11:40'!
acceptDroppingMorph: morphToDrop event: evt
	"Accept -- in a custom sense here -- a morph dropped on the receiver."

	| myCopy smallR |

	(self isTheRealProjectPresent) ifFalse: [
		^morphToDrop rejectDropMorphEvent: evt.		"can't handle it right now"
	].
	(morphToDrop isKindOf: NewHandleMorph) ifTrue: [	"don't send these"
		^morphToDrop rejectDropMorphEvent: evt.
	].
	self dropEnabled ifFalse:
		[^ morphToDrop rejectDropMorphEvent: evt].

	self eToyRejectDropMorph: morphToDrop event: evt.		"we will send a copy"
	myCopy _ morphToDrop veryDeepCopy.	"gradient fills require doing this second"
	smallR _ (morphToDrop bounds scaleBy: image height / Display height) rounded.
	smallR _ smallR squishedWithin: image boundingBox.
	image getCanvas
		paintImage: (morphToDrop imageForm scaledToSize: smallR extent)
		at: smallR topLeft.
	myCopy openInWorld: project world

! !

!ProjectViewMorph methodsFor: 'as yet unclassified' stamp: 'RAA 11/2/2000 09:56'!
addProjectNameMorph

	| m |

	self removeAllMorphs.
	m := UpdatingStringMorph contents: self safeProjectName font: self fontForName.
	m target: self; getSelector: #safeProjectName; putSelector: #safeProjectName:.
	m useStringFormat; fitContents.
	self addMorphBack: m.
	self updateNamePosition.
	^m

! !

!ProjectViewMorph methodsFor: 'as yet unclassified' stamp: 'RAA 11/2/2000 10:12'!
addProjectNameMorphFiller

	| m |

	self removeAllMorphs.
	m := AlignmentMorph newRow color: Color transparent.
	self addMorphBack: m.
	m
		on: #mouseDown send: #editTheName: to: self;
		on: #mouseUp send: #yourself to: self.
	self updateNamePosition.

! !

!ProjectViewMorph methodsFor: 'events' stamp: 'dgd 8/31/2003 18:40'!
balloonText
	^ 'Click here to enter the
project named
"{1}"' translated format: {project name}! !

!ProjectViewMorph methodsFor: 'accessing' stamp: 'RAA 2/12/2001 14:47'!
borderWidthForRounding

	^1! !

!ProjectViewMorph methodsFor: 'events' stamp: 'RAA 5/17/2000 11:39'!
checkForNewerVersionAndLoad

	self withProgressDo: [
		project loadFromServer
	] 

! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'RAA 8/30/2000 19:10'!
colorAroundName

	^Color gray: 0.8! !

!ProjectViewMorph methodsFor: 'initialization' stamp: 'Richo 2/25/2012 19:35'!
dismissViaHalo
	"The user clicked on the dismiss icon on the halo."

	| choice |
	project ifNil: [^ self delete]. "no current project"
	choice := (PopUpMenu labelArray:{
		'yes - delete icon and remove the project' translated.
		'no - delete icon but keep the project' translated.
		'cancel - do not delete anything' translated.  
	}) startUpWithCaption: ('Do you really want to delete the
project named {1}
and all its contents?' translated format: {project name printString}).
	choice = 1 ifTrue: [^ self expungeProject].
	choice = 2 ifTrue: [^ self delete]! !

!ProjectViewMorph methodsFor: 'events' stamp: 'sw 3/17/2005 23:59'!
doButtonAction
	"My inherent button action consists of entering the project I represent"

	self enter! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'jmv 9/7/2009 00:35'!
drawOn: aCanvas

	| font projectName nameForm rectForName |

	self ensureImageReady.
	super drawOn: aCanvas.
	self isEditingName ifTrue: [^self].

	font := self fontForName.
	projectName := self safeProjectName.
	nameForm := (StringMorph contents: projectName font: font) imageForm.
	nameForm := nameForm scaledToSize: (self extent - (4@2) min: nameForm extent).
	rectForName := self bottomLeft + 
			(self width - nameForm width // 2 @ (nameForm height + 2) negated)
				extent: nameForm extent.
	rectForName topLeft eightNeighbors do: [ :pt |
		aCanvas
			stencil: nameForm 
			at: pt
			color: self colorAroundName.
	].
	aCanvas
		drawImage: nameForm 
		at: rectForName topLeft
! !

!ProjectViewMorph methodsFor: 'as yet unclassified' stamp: 'KR 5/1/2008 01:59'!
editTheName: evt

	self isTheRealProjectPresent ifFalse: [
		^self inform: 'The project is not present and may not be renamed now' translated
	].
	self addProjectNameMorph launchMiniEditor: evt.! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'mt 4/12/2016 19:57'!
ensureImageReady

	self isTheRealProjectPresent ifFalse: [^self].
	project thumbnail ifNil: [
		image fill: image boundingBox rule: Form over 
			fillColor: project color.
		^self
	].
	project thumbnail ~~ lastProjectThumbnail ifTrue: ["scale thumbnail to fit my bounds"
		lastProjectThumbnail := project thumbnail.
		self updateImageFrom: lastProjectThumbnail.
		project thumbnail ifNotNil: [project thumbnail hibernate].
		image borderWidth: 1
	].


! !

!ProjectViewMorph methodsFor: 'events' stamp: 'KR 5/1/2008 02:00'!
enter
	"Enter my project."

	self world == self outermostWorldMorph ifFalse: [^Beeper beep].	"can't do this at the moment"
	project class == DiskProxy 
		ifFalse: 
			[(project world notNil and: 
					[project world isMorph 
						and: [project world hasOwner: self outermostWorldMorph]]) 
				ifTrue: [^Beeper beep	"project is open in a window already"]].
	project class == DiskProxy 
		ifTrue: 
			["When target is not in yet"

			self enterWhenNotPresent.	"will bring it in"
			project class == DiskProxy ifTrue: [^self inform: 'Project not found' translated]].
	(owner isSystemWindow) ifTrue: [project setViewSize: self extent].
	self showMouseState: 3.
	project 
		enter: false
		revert: false
		saveForRevert: false! !

!ProjectViewMorph methodsFor: 'events' stamp: 'RAA 5/17/2000 11:38'!
enterWhenNotPresent

	self withProgressDo: [
		project enter: false revert: false saveForRevert: false.	"will bring it in"
	] 

! !

!ProjectViewMorph methodsFor: 'as yet unclassified' stamp: 'mt 5/1/2016 14:27'!
expungeProject

	(self confirm: ('Do you really want to delete {1}
and all its content?' translated format: {project name}))
		ifFalse: [^ self].
	owner isSystemWindow
		ifTrue: [owner model: nil;
				 delete].
	
	project delete.! !

!ProjectViewMorph methodsFor: 'geometry' stamp: 'RAA 10/27/2000 11:08'!
extent: aPoint
	"Set my image form to the given extent."

	| newExtent scaleP scale |

	((bounds extent = aPoint) and: [image depth = Display depth]) ifFalse: [
		lastProjectThumbnail ifNil: [ lastProjectThumbnail := image ].
		scaleP := aPoint / lastProjectThumbnail extent.
		scale := scaleP "scaleP x asFloat max: scaleP y asFloat".
		newExtent := (lastProjectThumbnail extent * scale) rounded.
		self image: (Form extent: newExtent depth: Display depth).
		self updateImageFrom: lastProjectThumbnail.
	].
	self updateNamePosition.! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'yo 6/9/2008 21:19'!
fontForName

	^(TextStyle default fontOfSize: 15) emphasized: 1
! !

!ProjectViewMorph methodsFor: 'event handling' stamp: 'jm 5/4/1998 22:13'!
handlesMouseDown: evt

	^ true
! !

!ProjectViewMorph methodsFor: 'event handling' stamp: 'RAA 1/26/2001 09:03'!
handlesMouseOverDragging: evt

	^ true! !

!ProjectViewMorph methodsFor: 'event handling' stamp: 'sw 9/22/1999 11:41'!
handlesMouseOver: evt
	^ true! !

!ProjectViewMorph methodsFor: 'initialization' stamp: 'sw 8/16/2007 11:28'!
initialize
	"Initialize the receiver."

	super initialize.
	"currentBorderColor _ Color gray."
	self addProjectNameMorphFiller.
	self enableDragNDrop: true.
	self isOpaque: true.
! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'RAA 11/1/2000 22:45'!
isEditingName

	| nameMorph |
	nameMorph := self findA: UpdatingStringMorph.
	nameMorph ifNil: [^false].

	^nameMorph hasFocus
! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'RAA 11/2/2000 10:38'!
isTheRealProjectPresent

	project ifNil: [^ false].
	project isInMemory ifFalse: [^ false].
	project class == DiskProxy ifTrue: [^ false].
	^true
! !

!ProjectViewMorph methodsFor: 'events' stamp: 'RAA 7/12/2000 07:44'!
lastProjectThumbnail: aForm
	
	lastProjectThumbnail := aForm! !

!ProjectViewMorph methodsFor: 'event handling' stamp: 'fc 12/23/2004 19:38'!
mouseDown: evt

	evt yellowButtonPressed ifTrue:
		[self showMenuForProjectView].
	evt hand newMouseFocus: self.
	self removeProperty: #wasOpenedAsSubproject.
	self showMouseState: 2.! !

!ProjectViewMorph methodsFor: 'event handling' stamp: 'RAA 7/16/2000 14:14'!
mouseEnter: evt

	self showMouseState: 1! !

!ProjectViewMorph methodsFor: 'event handling' stamp: 'RAA 1/26/2001 09:03'!
mouseLeaveDragging: evt

	self mouseLeave: evt

! !

!ProjectViewMorph methodsFor: 'event handling' stamp: 'RAA 7/16/2000 14:13'!
mouseLeave: evt

	self showMouseState: 3.
! !

!ProjectViewMorph methodsFor: 'event handling' stamp: 'RAA 7/16/2000 14:13'!
mouseUp: evt

	((self containsPoint: evt cursorPoint) and: 
				[(self hasProperty: #wasOpenedAsSubproject) not]) ifTrue:
		[^ self enter].
	self showMouseState: 3.
! !

!ProjectViewMorph methodsFor: 'objects from disk' stamp: 'RAA 10/27/2000 11:08'!
objectForDataStream: refStrm
	
	| copy |

	1 = 1 ifTrue: [^self].		"this didn't really work"

	copy := self copy lastProjectThumbnail: nil.
	"refStrm replace: self with: copy."
	^copy
! !

!ProjectViewMorph methodsFor: 'events' stamp: 'RAA 11/2/2000 10:06'!
on: aProject

	project := aProject.
	self addProjectNameMorphFiller.
	lastProjectThumbnail := nil.
	project thumbnail
		ifNil: [self extent: 100@80]		"more like screen dimensions?"
		ifNotNil: [self extent: project thumbnail extent].
! !

!ProjectViewMorph methodsFor: 'accessing' stamp: 'tk 8/30/1999 11:48'!
project
	^project! !

!ProjectViewMorph methodsFor: 'events' stamp: 'RAA 11/2/2000 10:06'!
project: aProject

	project := aProject.
	self addProjectNameMorphFiller.! !

!ProjectViewMorph methodsFor: 'caching' stamp: 'RAA 10/27/2000 10:55'!
releaseCachedState

	"see if we can reduce size of published file, but there may be problems"
	super releaseCachedState.
	lastProjectThumbnail := image.
! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'gm 2/22/2003 13:14'!
safeProjectName
	| projectName args |
	projectName := self valueOfProperty: #SafeProjectName ifAbsent: ['???'].
	self isTheRealProjectPresent 
		ifFalse: 
			[project class == DiskProxy 
				ifTrue: 
					[args := project constructorArgs.
					((args isKindOf: Array) 
						and: [args size = 1 and: [args first isString]]) 
							ifTrue: [^args first]]
				ifFalse: [^projectName]].
	self setProperty: #SafeProjectName toValue: project name.
	^project name! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'gm 2/16/2003 20:34'!
safeProjectName: aString 
	self addProjectNameMorphFiller.
	self isTheRealProjectPresent ifFalse: [^self].
	project renameTo: aString.
	self setProperty: #SafeProjectName toValue: project name.
	self updateNamePosition.
	(owner isSystemWindow) ifTrue: [owner setLabel: aString]! !

!ProjectViewMorph methodsFor: 'events' stamp: 'raa 11/2/2000 10:29'!
seeIfNameChanged

	| nameBefore nameNow |

	nameBefore := self valueOfProperty: #SafeProjectName ifAbsent: ['???'].
	nameNow := self safeProjectName.
	(submorphs notEmpty and: [nameBefore = nameNow]) ifTrue: [^self].
	self addProjectNameMorphFiller.
! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'RAA 1/26/2001 09:14'!
showBorderAs: aColor

	"image border: image boundingBox width: 1 fillColor: aColor.
	currentBorderColor := aColor.
	self changed"
! !

!ProjectViewMorph methodsFor: 'events' stamp: 'ar 9/19/2009 20:16'!
showMenuForProjectView
	| menu |
	(menu := MenuMorph new)
		add: 'enter this project' translated
		action: [^ self enter];
		
		add: 'PUBLISH (also saves a local copy)' translated
		action: [^ project storeOnServerShowProgressOn: self forgetURL: false];
		
		add: 'PUBLISH to a different server' translated
		action: [project forgetExistingURL.
			^ project storeOnServerShowProgressOn: self forgetURL: true];
		
		add: 'see if server version is more recent' translated
		action: [^ self checkForNewerVersionAndLoad];

		addLine;
		add: 'expunge this project' translated
		action: [^ self expungeProject].

	menu title: ('Project Named \"{1}"' translated withCRs format: {project name}).
	menu invokeModal.! !

!ProjectViewMorph methodsFor: 'events' stamp: 'kfr 10/9/2004 10:36'!
showMouseState: anInteger 
	| aMorph |
	(owner isSystemWindow)
		ifTrue: [aMorph := owner]
		ifFalse: [aMorph := self].
	anInteger = 1
		ifTrue: ["enter"
			aMorph
				addMouseActionIndicatorsWidth: 10
				color: (Color blue alpha: 0.3)].
	anInteger = 2
		ifTrue: ["down"
			aMorph
				addMouseActionIndicatorsWidth: 15
				color: (Color blue alpha: 0.7)].
	anInteger = 3
		ifTrue: ["leave"
			aMorph deleteAnyMouseActionIndicators]! !

!ProjectViewMorph methodsFor: 'stepping and presenter' stamp: 'raa 11/2/2000 10:20'!
step
	| cmd |
	"Check for a command that could not be executed in my subproject.  Once it is done, remove the trigger.  If this is too slow, make armsLengthCmd an inst var."

	self seeIfNameChanged.
	cmd := self valueOfProperty: #armsLengthCmd.
	cmd ifNil: [^ super step].
	self removeProperty: #armsLengthCmd.
	project perform: cmd.
	project enter.! !

!ProjectViewMorph methodsFor: 'testing' stamp: 'raa 11/2/2000 10:21'!
stepTime

	^1000! !

!ProjectViewMorph methodsFor: 'fileIn/out' stamp: 'tk 9/8/1999 17:51'!
storeSegment
	"Store my project out on the disk as an ImageSegment.  Keep the outPointers in memory.  Name it <project name>.seg"

	project storeSegment
! !

!ProjectViewMorph methodsFor: 'accessing' stamp: 'di 6/6/2001 21:34'!
thumbnail
	^ project ifNotNil: [project thumbnail]! !

!ProjectViewMorph methodsFor: '*Morphic-Sound-piano rolls' stamp: 'RAA 12/12/2000 10:08'!
triggerActionFromPianoRoll

	WorldState addDeferredUIMessage: [
		project world setProperty: #letTheMusicPlay toValue: true.
		self enter.
	]! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'tpr 3/15/2013 12:42'!
updateImageFrom: sourceForm

	(WarpBlt toForm: image)
		sourceForm: sourceForm;
		cellSize: 2;  "installs a colormap"
		combinationRule: Form over;
		copyQuad: (sourceForm boundingBox) innerCorners
		toRect: image boundingBox.
! !

!ProjectViewMorph methodsFor: 'drawing' stamp: 'RAA 11/2/2000 10:11'!
updateNamePosition

	| nameMorph shadowMorph nameFillerMorph |

	(nameMorph := self findA: UpdatingStringMorph) ifNotNil: [
		nameMorph position:
			(self left + (self width - nameMorph width // 2)) @
			(self bottom - nameMorph height - 2).
	].
	(nameFillerMorph := self findA: AlignmentMorph) ifNotNil: [
		nameFillerMorph
			position: self bottomLeft - (0@20);
			extent: self width@20.
	].
	(shadowMorph := self findA: ImageMorph) ifNotNil: [
		shadowMorph delete	"no longer used"
	].

! !

!ProjectViewMorph methodsFor: 'copying' stamp: 'tk 1/8/1999 08:56'!
veryDeepFixupWith: deepCopier
	"If target and arguments fields were weakly copied, fix them here.  If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

super veryDeepFixupWith: deepCopier.
project := deepCopier references at: project ifAbsent: [project].
lastProjectThumbnail := deepCopier references at: lastProjectThumbnail 
				ifAbsent: [lastProjectThumbnail].
! !

!ProjectViewMorph methodsFor: 'copying' stamp: 'fc 12/23/2004 16:24'!
veryDeepInner: deepCopier 
	"Copy all of my instance variables.  Some need to be not copied at all, but shared.  See DeepCopier class comment."

	super veryDeepInner: deepCopier.
	project := project.	"Weakly copied"
	lastProjectThumbnail := lastProjectThumbnail veryDeepCopyWith: deepCopier.
! !

!ProjectViewMorph methodsFor: 'dropping/grabbing' stamp: 'sw 8/16/2007 11:37'!
wantsDroppedMorph: aMorph event: evt
	"Answer if the receiver would accept a drop of a given morph."

	"If drop-enabled not set, answer false"
	(super wantsDroppedMorph: aMorph event: evt) ifFalse: [^ false].

	"If project not present, not morphic, or not initialized, answer false"
	self isTheRealProjectPresent ifFalse: [^ false].
	project isMorphic ifFalse: [^ false].
	project world viewBox ifNil: [^ false].

	^ true! !

!ProjectViewMorph methodsFor: 'event handling' stamp: 'RAA 10/20/2000 10:12'!
wantsKeyboardFocusFor: aSubmorph

	^true! !

!ProjectViewMorph methodsFor: 'rounding' stamp: 'mt 7/28/2016 10:39'!
wantsRoundedCorners
	^SystemWindow roundedWindowCorners 
		and: [(owner isSystemWindow) not]! !

!ProjectViewMorph methodsFor: 'events' stamp: 'RAA 5/23/2000 10:08'!
withProgressDo: aBlock

	ComplexProgressIndicator new 
		targetMorph: self;
		historyCategory: 'project loading';
		withProgressDo: aBlock
! !

ImageMorph subclass: #SelectedObjectThumbnail
	instanceVariableNames: 'noSelectedThumbnail noSelectedBalloonText'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!SelectedObjectThumbnail class methodsFor: 'instance creation' stamp: 'dgd 9/10/2004 19:31'!
extent: aPoint noSelectedThumbnail: aForm noSelectedBalloonText: aString 
	^ self new
		initializeExtent: aPoint
		noSelectedThumbnail: aForm
		noSelectedBalloonText: aString ! !

!SelectedObjectThumbnail methodsFor: 'initialization' stamp: 'dgd 9/10/2004 19:00'!
initialize
	"Initialize the receiver"
	super initialize.
	""
	self
		image: (Form extent:32@32).
	self color: Color transparent! !

!SelectedObjectThumbnail methodsFor: 'initialization' stamp: 'dgd 9/10/2004 19:29'!
initializeExtent: aPoint noSelectedThumbnail: aForm noSelectedBalloonText: aString 
	self
		image: (Form extent: aPoint).
""
	noSelectedThumbnail := aForm.
	noSelectedBalloonText := aString! !

!SelectedObjectThumbnail methodsFor: 'private' stamp: 'dgd 9/10/2004 18:43'!
makeEmptyThumbnail

^ self makeThumbnailOfColor: Color veryLightGray.
! !

!SelectedObjectThumbnail methodsFor: 'private' stamp: 'dgd 9/10/2004 18:43'!
makeErrorThumbnail
	^ self makeThumbnailOfColor: Color red! !

!SelectedObjectThumbnail methodsFor: 'private' stamp: 'dgd 9/13/2004 12:57'!
makeThumbnailFrom: aMorphOrNil 
	| thumbnail |
	thumbnail := aMorphOrNil isNil
				ifTrue: [noSelectedThumbnail
						ifNil: [self makeEmptyThumbnail]]
				ifFalse: [aMorphOrNil iconOrThumbnail]. 
	""
	self
		image: (thumbnail scaledIntoFormOfSize: self extent)! !

!SelectedObjectThumbnail methodsFor: 'private' stamp: 'dgd 9/10/2004 18:43'!
makeThumbnailOfColor: aColor 
	| form |
	form := Form extent: self extent depth: 32.
	form getCanvas fillColor: aColor.
	self image: form! !

!SelectedObjectThumbnail methodsFor: 'accessing' stamp: 'dgd 9/10/2004 19:28'!
noSelectedBalloonText: aString 
	"Set the balloon text to be used when no object is selected"
	noSelectedBalloonText := aString! !

!SelectedObjectThumbnail methodsFor: 'accessing' stamp: 'dgd 9/10/2004 19:27'!
noSelectedThumbnail: aForm 
	"Set the form to be used when no object is selected"
	noSelectedThumbnail := aForm! !

!SelectedObjectThumbnail methodsFor: 'stepping and presenter' stamp: 'dgd 9/10/2004 19:29'!
step
	| current |
	current := self selectedObject.

	self setBalloonText: (current isNil
				ifTrue: [noSelectedBalloonText]
				ifFalse: [current externalName]).
	""
	self makeThumbnailFrom: current! !

!SelectedObjectThumbnail methodsFor: 'stepping and presenter' stamp: 'dgd 9/10/2004 18:36'!
stepTime
	^ 125! !

ImageMorph subclass: #Sonogram
	instanceVariableNames: 'lastX scrollDelta columnForm minVal maxVal pixValMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Sound-Synthesis'!

!Sonogram commentStamp: '<historical>' prior: 0!
Sonograms are imageMorphs that will repeatedly plot arrays of values as black on white columns moving to the right in time and scrolling left as necessary.!

!Sonogram methodsFor: 'geometry' stamp: 'ar 5/14/2001 23:48'!
extent: newExtent
	super image: (Form extent: newExtent depth: Display depth).
	lastX := -1.
	columnForm := Form extent: (32//image depth)@(image height) depth: image depth.
	pixValMap := ((1 to: 256) collect:
			[:i | columnForm pixelValueFor: (Color gray: (256-i)/255.0)])
		as: Bitmap.
! !

!Sonogram methodsFor: 'all' stamp: 'di 8/26/1999 09:01'!
extent: extent minVal: min maxVal: max scrollDelta: d
	minVal := min.
	maxVal := max.
	scrollDelta := d.
	self extent: extent.

" try following with scrolldelta = 1, 20, 200
	| s data |
	s := Sonogram new extent: 200@50
				minVal: 0.0 maxVal: 1.0 scrollDelta: 20.
	World addMorph: s.
	data := (1 to: 133) collect: [:i | 0.0].
	1 to: 300 do:
		[:i | data at: (i\\133)+1 put: 1.0.
		s plotColumn: data.
		data at: (i\\133)+1 put: 0.0.
		World doOneCycleNow].
	s delete	
"! !

!Sonogram methodsFor: 'all' stamp: 'jdl 3/28/2003 09:30'!
plotColumn: dataArray 
	| chm1 i normVal r |
	columnForm unhibernate.
	chm1 := columnForm height - 1.
	0 to: chm1
		do: 
			[:y | 
			i := y * (dataArray size - 1) // chm1 + 1.
			normVal := ((dataArray at: i) - minVal) / (maxVal - minVal).
			normVal := normVal max: 0.0.
			normVal := normVal min: 1.0.
			columnForm bits at: chm1 - y + 1
				put: (pixValMap at: (normVal * 255.0) truncated + 1)].
	(lastX := lastX + 1) > (image width - 1) ifTrue: [self scroll].
	image 
		copy: (r := lastX @ 0 extent: 1 @ image height)
		from: (32 // image depth - 1) @ 0
		in: columnForm
		rule: Form over.
	"self changed."
	self invalidRect: (r translateBy: self position)! !

!Sonogram methodsFor: 'all' stamp: 'di 8/26/1999 00:40'!
scroll
	image copy: (scrollDelta@0 extent: (image width-scrollDelta)@image height)
			from: image to: 0@0 rule: Form over.
	lastX := lastX - scrollDelta.
	self changed! !

ImageMorph subclass: #ThreePhaseButtonMorph
	instanceVariableNames: 'offImage pressedImage state target actionSelector arguments actWhen label font textColor'
	classVariableNames: 'AuthorModeOwner'
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!ThreePhaseButtonMorph commentStamp: '<historical>' prior: 0!
A button morph with separate images for on, off, and pressed with the mouse. 

When the event actWhen occurs, send actionSelector with 'arguments' to target.  For other events, default to my eventHandler.  The current event is not supplied in the arguments to the actionSelector.  

image (a.k.a. onImage) may not be nil.  offImage and pressedImage may be nil.  nil there means be transparent and show the underlying object.  

Tools for debugging:
Display the images momentarily under program control (for positioning) (self is an instance).
	self state: #on.  self state: #off.
	self state: #pressed.  self state: #off.
Display a rectangle where the button is.
	Display fillWithColor: bounds + (self world viewBox origin).
	self invalidRect: bounds.!

!ThreePhaseButtonMorph class methodsFor: 'instance creation' stamp: 'mt 8/1/2016 16:06'!
checkBox
	"Answer a button pre-initialized with checkbox images."
	| f |
	^self new
		onImage: (f := ScriptingSystem formAtKey: 'CheckBoxOn');
		pressedImage: (ScriptingSystem formAtKey: 'CheckBoxPressed');
		offImage: (ScriptingSystem formAtKey: 'CheckBoxOff');
		extent: f extent + (2@0);
		setDefaultParameters;
		yourself
! !

!ThreePhaseButtonMorph class methodsFor: 'class initialization' stamp: 'ar 5/25/2000 18:01'!
initialize
	"ThreePhaseButtonMorph initialize"
	| extent inset |
	extent := 12@12.
	inset := 3.

	#('CheckBoxOff' 'CheckBoxOn' 'CheckBoxPressed') do: [:button |
		| f r |
		f := ColorForm extent: extent depth: 1.
		f colors: {Color transparent. Color black}.
		f borderWidth: 1.
		r := f boundingBox insetBy: inset.
		button = 'CheckBoxPressed' ifTrue: [f border: r width: 1].
		button = 'CheckBoxOn' ifTrue: [f fillBlack: r].
		ScriptingSystem saveForm: f atKey: button].

	#('RadioButtonOff' 'RadioButtonOn' 'RadioButtonPressed') do: [:button |
		| f r c |
		f := ColorForm extent: extent depth: 1.
		f colors: {Color transparent. Color black}.
		r := f boundingBox.
		c := f getCanvas.
		c frameOval: r color: Color black.
		r := r insetBy: inset.
		button = 'RadioButtonPressed' ifTrue:
			[c frameOval: r color: Color black].
		button = 'RadioButtonOn' ifTrue:
			[c fillOval: r color: Color black].
		ScriptingSystem saveForm: f atKey: button]! !

!ThreePhaseButtonMorph class methodsFor: 'instance creation' stamp: 'mt 8/1/2016 16:06'!
radioButton
	"Answer a button pre-initialized with radiobutton images."
	| f |
	^self new
		onImage: (f := ScriptingSystem formAtKey: 'RadioButtonOn');
		pressedImage: (ScriptingSystem formAtKey: 'RadioButtonPressed');
		offImage: (ScriptingSystem formAtKey: 'RadioButtonOff');
		extent: f extent + (2@0);
		setDefaultParameters;
		yourself
! !

!ThreePhaseButtonMorph class methodsFor: 'preferences' stamp: 'mt 8/1/2016 16:03'!
themeProperties

	^ super themeProperties, {
		{ #color. 'Colors'. 'Color of the button.' }.
		{ #font. 'Fonts'. 'Font for button title.' }.
		{ #textColor. 'Colors'. 'Color for the button title label.' }.
	}! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 6/30/97 10:49'!
actionSelector

	^ actionSelector
! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 6/30/97 10:49'!
actionSelector: aSymbolOrString

	(nil = aSymbolOrString or:
	 ['nil' = aSymbolOrString or:
	 [aSymbolOrString isEmpty]])
		ifTrue: [^ actionSelector := nil].

	actionSelector := aSymbolOrString asSymbol.
! !

!ThreePhaseButtonMorph methodsFor: 'submorphs-add/remove' stamp: 'tk 6/30/97 10:49'!
actWhen: condition
	"Accepts symbols:  #buttonDown, #buttonUp, and #whilePressed"
	actWhen := condition! !

!ThreePhaseButtonMorph methodsFor: 'updating' stamp: 'mt 8/1/2016 16:49'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	self setDefaultParameters.! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 7/1/97 12:39'!
arguments
	^ arguments! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 7/1/97 08:39'!
arguments: aCollection

	arguments := aCollection asArray copy.
! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:31'!
color: c
	
	self onImage ifNotNil: [:form |
		self onImage: ((form asFormOfDepth: 32) collectColors: [:col | c alpha: col alpha])].
	self offImage ifNotNil: [:form |
		self offImage: ((form asFormOfDepth: 32) collectColors: [:col | c alpha: col alpha])].
	self pressedImage ifNotNil: [:form |
		self pressedImage: ((form asFormOfDepth: 32) collectColors: [:col | c alpha: col alpha])].
	
	^ super color: c! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:36'!
currentImage

	state == #off ifTrue: [^ offImage].
	state == #pressed ifTrue: [^ pressedImage].
	state == #on ifTrue: [^ image].
	^ image! !

!ThreePhaseButtonMorph methodsFor: 'button' stamp: 'sw 1/13/2012 01:22'!
doButtonAction
	"Perform the action of this button. Subclasses may override this method. The default behavior is to send the button's actionSelector to its target object with its arguments."

	| args |
	(target notNil and: [actionSelector notNil]) 
		ifTrue: 
			[args := actionSelector numArgs > arguments size
				ifTrue:
					[arguments copyWith: ActiveEvent]
				ifFalse:
					[arguments].
			Cursor normal 
				showWhile: [target perform: actionSelector withArguments: args].
			target isMorph ifTrue: [target changed]]! !

!ThreePhaseButtonMorph methodsFor: 'event handling' stamp: 'nice 12/27/2009 03:10'!
doButtonAction: evt
	
	"Perform the action of this button. Subclasses may override this method. The default behavior is to send the button's actionSelector to its target object with its arguments."

	target ifNil: [^self].
	actionSelector ifNil: [^self].
	Cursor normal showWhile: [ | moreArgs |
		moreArgs := actionSelector numArgs > arguments size ifTrue: [
			arguments copyWith: evt
		] ifFalse: [
			arguments
		].
		target perform: actionSelector withArguments: moreArgs
	]! !

!ThreePhaseButtonMorph methodsFor: 'drawing' stamp: 'mt 8/1/2016 16:39'!
drawOn: aCanvas

	| imageToUse |
	
	imageToUse := self currentImage.
	
	imageToUse ifNotNil: [aCanvas translucentImage: imageToUse at: bounds origin].	
	
	self label ifNotNil: [:lbl |
		aCanvas
			drawString: lbl
			at: bounds origin + (imageToUse ifNil: [0@0] ifNotNil: [:form | (form width @ 0) + (3@0)])
			font: self font
			color: self textColor].! !

!ThreePhaseButtonMorph methodsFor: 'geometry' stamp: 'tk 7/1/97 09:14'!
extent: aPoint
	"Do it normally"
	
	self changed.
	bounds := bounds topLeft extent: aPoint.
	self layoutChanged.
	self changed.
! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:32'!
font
	^ font ifNil: [TextStyle defaultFont]! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:34'!
font: aFont
	font := aFont.
	super extent: self preferredExtent.! !

!ThreePhaseButtonMorph methodsFor: 'event handling' stamp: 'tk 6/30/97 10:52'!
handlesMouseDown: evt

	^ true
! !

!ThreePhaseButtonMorph methodsFor: 'event handling' stamp: 'ar 10/25/2000 18:18'!
handlesMouseStillDown: evt
	^actWhen == #whilePressed! !

!ThreePhaseButtonMorph methodsFor: 'initialization' stamp: 'cmm 5/8/2014 11:01'!
initialize
	super initialize.
	state := #off.
	target := nil.
	actionSelector := #flash.
	arguments := Array empty.
	actWhen := #buttonUp

	"self on: #mouseStillDown send: #dragIfAuthoring: to: self."
		"real move should include a call on dragIfAuthoring: "! !

!ThreePhaseButtonMorph methodsFor: 'testing' stamp: 'sw 3/8/1999 13:56'!
isOn
	^ state == #on! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:18'!
label
	^ label! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:34'!
label: aString
	label := aString.
	super extent: self preferredExtent.! !

!ThreePhaseButtonMorph methodsFor: 'event handling' stamp: 'tak 3/28/2007 14:24'!
mouseDown: evt
	| now dt |
	actWhen == #buttonDown
		ifTrue: [self doButtonAction].
	actWhen == #buttonUp
		ifTrue: [self state: #pressed].
	actWhen == #whilePressed
		ifTrue: 
			[self state: #pressed.
			now _ Time millisecondClockValue.
			"Allow on:send:to: to set the response to events other than actWhen"
			dt _ Time millisecondClockValue - now max: 0.  "Time it took to do"
			"NOTE: this delay is temporary disabled because it makes event reaction delay,
				e.g. the action is not stopped even if you release the button... - Takashi" 
			[dt < 200 ifTrue: [(Delay forMilliseconds: 200-dt) wait]].
			self mouseStillDown: evt].
	super mouseDown: evt! !

!ThreePhaseButtonMorph methodsFor: 'event handling' stamp: 'tak 3/28/2007 14:21'!
mouseMove: evt 
	(#(#buttonUp #whilePressed ) includes: actWhen)
		ifTrue: [(self containsPoint: evt cursorPoint)
				ifTrue: [self state: #pressed]
				ifFalse: [self state: #off]].
	super mouseMove: evt! !

!ThreePhaseButtonMorph methodsFor: 'event handling' stamp: 'ar 10/25/2000 18:17'!
mouseStillDown: evt
	actWhen == #whilePressed ifFalse:[^self].
	(self containsPoint: evt cursorPoint) ifTrue:[self doButtonAction].! !

!ThreePhaseButtonMorph methodsFor: 'event handling' stamp: 'tak 3/28/2007 14:05'!
mouseUp: evt 
	"Allow on:send:to: to set the response to events other than actWhen"
	actWhen == #buttonDown
		ifTrue: [super mouseUp: evt].
	actWhen == #buttonUp
		ifTrue: [(self containsPoint: evt cursorPoint)
				ifTrue: [self state: #on.
					self doButtonAction: evt.
					super mouseUp: evt]
				ifFalse: [self state: #off.
					target
						ifNotNil: ["Allow owner to keep it selected for radio
							buttons"
							target mouseUpBalk: evt]]].
	actWhen == #whilePressed
		ifTrue: [self state: #off.
			super mouseUp: evt]! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 10/19/97 15:02'!
offImage
	^ offImage! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 6/30/97 10:08'!
offImage: aForm
	offImage := aForm.
	self invalidRect: self bounds.! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 10/19/97 15:02'!
onImage
	^ image! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 6/30/97 10:08'!
onImage: aForm
	image := aForm.
	self invalidRect: self bounds.! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:38'!
preferredExtent

	| iw ih lw lh |
	iw := self currentImage width.
	ih := self currentImage height.
	lw := self font widthOfString: (self label ifNil: ['']).
	lh := self font height.
	^ (iw + 3 + lw) @ (ih max: lh)! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 10/19/97 15:02'!
pressedImage
	^ pressedImage! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 6/30/97 10:09'!
pressedImage: aForm
	pressedImage := aForm.
	self invalidRect: self bounds.! !

!ThreePhaseButtonMorph methodsFor: 'printing' stamp: 'dgd 2/22/2003 18:45'!
printOn: aStream 
	| string |
	aStream nextPutAll: '3PButton'.
	arguments notEmpty 
		ifTrue: [string := arguments at: (2 min: arguments size)].
	aStream nextPutAll: '('.
	(string notNil and: [string ~~ self]) 
		ifTrue: 
			[aStream
				print: string;
				space]
		ifFalse: 
			[aStream
				print: actionSelector;
				space].
	aStream
		print: self identityHash;
		nextPutAll: ')'! !

!ThreePhaseButtonMorph methodsFor: 'initialization' stamp: 'mt 8/1/2016 16:33'!
setDefaultParameters

	self
		color: (self userInterfaceTheme color ifNil: [Color black]);
		font: (self userInterfaceTheme font ifNil: [TextStyle defaultFont]);
		textColor: (self userInterfaceTheme textColor ifNil: [Color black]).! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 6/30/97 11:01'!
state: newState
	"Change the image and invalidate the rect."

	newState == state ifTrue: [^ self].
	state := newState.
	self invalidRect: bounds.	"All three images must be the same size"! !

!ThreePhaseButtonMorph methodsFor: 'stepping and presenter' stamp: 'ar 10/11/2000 14:05'!
step
	(self hasProperty: #doesButtonAction) ifTrue:[
		self doButtonAction.
		self setProperty: #didButtonAction toValue: true.
	].! !

!ThreePhaseButtonMorph methodsFor: 'testing' stamp: 'ar 10/11/2000 14:05'!
stepTime
	(self hasProperty: #doesButtonAction) ifTrue:[^1].
	^super stepTime! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 6/30/97 10:50'!
target

	^ target
! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'tk 6/30/97 10:50'!
target: anObject

	target := anObject
! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:32'!
textColor
	^ textColor ifNil: [Color black]! !

!ThreePhaseButtonMorph methodsFor: 'accessing' stamp: 'mt 8/1/2016 16:32'!
textColor: aColor
	textColor := aColor.! !

!ThreePhaseButtonMorph methodsFor: 'copying' stamp: 'tk 1/8/1999 09:02'!
veryDeepFixupWith: deepCopier
	"If target and arguments fields were weakly copied, fix them here.  If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

super veryDeepFixupWith: deepCopier.
target := deepCopier references at: target ifAbsent: [target].
arguments := arguments collect: [:each |
	deepCopier references at: each ifAbsent: [each]].
! !

!ThreePhaseButtonMorph methodsFor: 'copying' stamp: 'tk 1/8/1999 09:01'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all, but shared.  	Warning!!!!  Every instance variable defined in this class must be handled.  We must also implement veryDeepFixupWith:.  See DeepCopier class comment."

super veryDeepInner: deepCopier.
offImage := offImage veryDeepCopyWith: deepCopier.
pressedImage := pressedImage veryDeepCopyWith: deepCopier.
state := state veryDeepCopyWith: deepCopier.
"target := target.		Weakly copied"
"actionSelector := actionSelector.		Symbol"
"arguments := arguments.		Weakly copied"
actWhen := actWhen.		"Symbol"! !

!ThreePhaseButtonMorph methodsFor: 'testing' stamp: 'ar 10/11/2000 14:06'!
wantsSteps
	^(self hasProperty: #doesButtonAction) or:[super wantsSteps]! !

ThreePhaseButtonMorph subclass: #UpdatingThreePhaseButtonMorph
	instanceVariableNames: 'getSelector getArgument'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!UpdatingThreePhaseButtonMorph methodsFor: 'button' stamp: 'bf 10/8/1999 15:08'!
doButtonAction
	"Since the action likely changes our state, do a step so we're updated immediately"
	super doButtonAction.
	self step
! !

!UpdatingThreePhaseButtonMorph methodsFor: 'as yet unclassified' stamp: 'sw 3/8/1999 13:50'!
getSelector: sel
	getSelector := sel! !

!UpdatingThreePhaseButtonMorph methodsFor: 'event handling' stamp: 'bf 10/14/1999 21:11'!
mouseUp: evt
	"Since mouseUp likely changes our state, do a step so we're updated immediately"
	super mouseUp: evt.
	self step! !

!UpdatingThreePhaseButtonMorph methodsFor: 'stepping and presenter' stamp: 'tk 7/14/2000 15:27'!
step
	| newBoolean |
	super step.
	state == #pressed ifTrue: [^ self].
	newBoolean := target perform: getSelector.
	newBoolean == self isOn
		ifFalse:
			[self state: (newBoolean == true ifTrue: [#on] ifFalse: [#off])]! !

!UpdatingThreePhaseButtonMorph methodsFor: 'testing' stamp: 'sw 3/8/1999 13:50'!
wantsSteps
	^ true! !

ImageMorph subclass: #ThumbnailImageMorph
	instanceVariableNames: 'imagePopupMorph desiredExtent isPopup'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!ThumbnailImageMorph commentStamp: 'wiz 2/19/2006 18:10' prior: 0!
A ThumbnailImageMorph is variant of Lex Spoon's CDScreenShotMorph. It displays a thumbnail of the image stored in imagePopupMorph. As a super class of PopupThumbnail morph it is meant to be a thumbnail w/o the popup action. Basicly it provides a scalable thumbnail with the usual morph event behaviors.

A menu item allows for "photographing" any morph on the screen to become the subject of our images.

Instance Variables
	imagePopupMorph:		<anImageMorph>
	image: 					<aForm>
	desiredExtent			<aPoint>
	isPopup					<aBool>
imagePopupMorph
	- an ImageMorph containing the full sized image.
	- it can be set from a morph image via the sight target menu item.
image
	- holds the scaled thumbnail form of our imagePopupMorph image.
desiredExtent
	- holds the desired extent that the thumbnail is expected to fit within.
	- it is guarded to be positive and non-zero.
	- it can be set by extent: so that the yellow halo handle works.
isPopup
	- true when popup feature is on.
	- toggled from red halo menu
		
Setting the size of the thumbnail works somewhat excentrically because the extent of the thumbnail depends both on the desiredExtent and the aspect ratio of the current popup image.
 
With the popup feature off this morph can be picked up and dropped with the mouse.
When the feature is on, a full sized snapshot will be seen when the mouse is pressed.
Since the mouse can't be used for two things at once, moving the morph must be done with the grab halo or brown move halo.!

!ThumbnailImageMorph class methodsFor: 'examples' stamp: 'wiz 9/21/2005 22:31'!
ofDisplay
"ThumbnailImageMorph ofDisplay openInHand"
^self new initializeWithDisplay .! !

!ThumbnailImageMorph methodsFor: 'menu commands' stamp: 'wiz 2/19/2006 14:38'!
addCustomMenuItems: aCustomMenu hand: aHandMorph 
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	self addWorldTargetSightingItems: aCustomMenu hand: aHandMorph .
	self addPopupMenuItems: aCustomMenu hand: aHandMorph! !

!ThumbnailImageMorph methodsFor: 'menu commands' stamp: 'wiz 2/19/2006 14:41'!
addPopupMenuItems: aCustomMenu hand: aHandMorph
" Show and toggle the popUp boolean menu item."

	aCustomMenu addLine. 
	
	aCustomMenu addUpdating: #popupFeatureString  target: self 
	selector: #togglePopupFeature argumentList: #() .

	! !

!ThumbnailImageMorph methodsFor: 'accessing' stamp: 'wiz 9/14/2005 23:53'!
extent: anExtentPoint
"Set the desired extetnt for the thumbnail. It is guarenteed to fit within the desired extent.
The desitedExtent is guarded to prevent deviant forms from being attempted."

self changed . "We might be bigger before we change."
desiredExtent := anExtentPoint guarded.
self newThumbnail: imagePopupMorph image .
! !

!ThumbnailImageMorph methodsFor: 'event handling' stamp: 'wiz 2/19/2006 12:51'!
handlesMouseDown: evt
	^isPopup == true! !

!ThumbnailImageMorph methodsFor: 'initialization' stamp: 'wiz 9/15/2005 17:38'!
initialize
super initialize .
desiredExtent := 90 asPoint.
self newImage: DefaultForm! !

!ThumbnailImageMorph methodsFor: 'initialization' stamp: 'wiz 9/21/2005 22:26'!
initializeWithDisplay
super initialize .
desiredExtent := 90 asPoint.
self newImage: Display! !

!ThumbnailImageMorph methodsFor: 'event handling' stamp: 'wiz 2/19/2006 12:51'!
mouseDown: evt
	
	
	imagePopupMorph center: (self localPointToGlobal: evt position).
	imagePopupMorph bounds: (imagePopupMorph bounds translatedAndSquishedToBeWithin: World bounds).
	imagePopupMorph openInWorld
! !

!ThumbnailImageMorph methodsFor: 'event handling' stamp: 'wiz 2/19/2006 12:50'!
mouseUp: evt
	imagePopupMorph ifNotNil: [
		imagePopupMorph delete	.
		 ]! !

!ThumbnailImageMorph methodsFor: 'accessing' stamp: 'wiz 9/14/2005 23:55'!
newImage: aForm
	"Use aForm as the new popupImage and update the thumbnail image."
	
	imagePopupMorph 
		ifNil: [ imagePopupMorph :=   aForm asMorph]
		ifNotNil: [ imagePopupMorph image: aForm ] .
		
		self newThumbnail: aForm
		
		! !

!ThumbnailImageMorph methodsFor: 'accessing' stamp: 'wiz 2/19/2006 14:12'!
newThumbnail: aForm
	"Use aForm as the new popupImage and update the thumbnail image."
	| scale  thumbForm border smoothing |

		scale := aForm extent scaleTo: desiredExtent .
		smoothing := (scale x < 1.0 or: [ scale y < 1.0 ]) ifTrue: [ 2 ] ifFalse: [ 1 ] .
		
		thumbForm := aForm magnify: aForm boundingBox by: scale smoothing: smoothing .
		
		self image: thumbForm . "heres where we put in a thumbnail"
		"We need the following to keep the border the right size. Otherwise it will shrink."
		(border := self borderStyle) == BorderStyle default ifFalse: [ self borderStyle: border ] .
		
		"We have changed clear the old and show the new"
		self invalidRect: self bounds . ! !

!ThumbnailImageMorph methodsFor: 'accessing' stamp: 'wiz 2/19/2006 14:48'!
popupFeatureString
	"Answer the string to be shown in a menu to represent the  
	stickiness status"
	^ (self yesNoStringFor: (isPopup == true ) )
		, 'Popup feature' translated! !

!ThumbnailImageMorph methodsFor: 'accessing' stamp: 'wiz 11/3/2005 01:17'!
target: aMorph
	"Snap aMorphs current image and show its thumbnail"
	
	self newImage: aMorph imageForm fixAlpha .
		! !

!ThumbnailImageMorph methodsFor: 'accessing' stamp: 'wiz 2/19/2006 14:49'!
togglePopupFeature
	"Change the popup behaviour. Return the new boolean value."
	
	^isPopup := isPopup ~~ true .
		! !

!ThumbnailImageMorph methodsFor: 'accessing' stamp: 'wiz 2/19/2006 14:21'!
yesNoStringFor: aBool
	"Answer the string to be shown in a menu to represent the  
	yes/no status"
	^ (aBool
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		! !

Morph subclass: #LazyListMorph
	instanceVariableNames: 'listItems listIcons listFilterOffsets font selectedRow selectedRows preSelectedRow listSource maxWidth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!LazyListMorph commentStamp: 'efc 8/6/2005 11:34' prior: 0!
The morph that displays the list in a PluggableListMorph.  It is "lazy" because it will only request the list items that it actually needs to display.

I will cache the maximum width of my items in maxWidth to avoid this potentially expensive and frequent computation.!

!LazyListMorph methodsFor: 'drawing' stamp: 'ls 5/15/2001 22:13'!
adjustHeight
	"private.  Adjust our height to match the length of the underlying list"
	self height: (listItems size max: 1) * font height
! !

!LazyListMorph methodsFor: 'drawing' stamp: 'sps 3/9/2004 17:06'!
adjustWidth
	"private.  Adjust our height to match the length of the underlying list"
	self width: ((listSource width max: self hUnadjustedScrollRange) + 20). 
! !

!LazyListMorph methodsFor: 'drawing' stamp: 'ls 5/17/2001 21:57'!
bottomVisibleRowForCanvas: aCanvas
        "return the bottom visible row in aCanvas's clip rectangle"
        ^self rowAtLocation: aCanvas clipRect bottomLeft.
! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 7/23/2016 17:35'!
colorForRow: row
	
	^(selectedRow notNil and: [ row = selectedRow])
		ifTrue: [ self selectionTextColor ]
		ifFalse: [ self color ].! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 11:28'!
displayFilterOn: canvas for: row in: drawBounds font: font
	"Draw filter matches if any."
	
	| fill |
	fill := self filterColor isColor
		ifTrue: [SolidFillStyle color: self filterColor]
		ifFalse: [self filterColor].
	fill isGradientFill ifTrue: [
		fill origin: drawBounds topLeft.
		fill direction: 0@ drawBounds height].
	
	(self filterOffsets: row) do: [:offset | | r |
		r := ((drawBounds left + offset first first) @ drawBounds top corner: (drawBounds left + offset first last) @ drawBounds bottom).
		canvas
			frameAndFillRoundRect: (r outsetBy: 1@0)
			radius: 3
			fillStyle: fill
			borderWidth: 1
			borderColor: fill asColor twiceDarker.
		canvas
			drawString: offset second
			in: r
			font: font
			color: self filterTextColor].! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 11:03'!
display: item atRow: row on: canvas
	"display the given item at row row"

	| drawBounds emphasized rowColor itemAsText |
	itemAsText := item asStringOrText.
	
	"If it is a text, we will only use the first character's emphasis."
	emphasized := itemAsText isText 
		ifTrue: [font emphasized: (itemAsText emphasisAt: 1)] 
		ifFalse: [font].
	
	rowColor := itemAsText isText
		ifTrue: [itemAsText colorAt: 1 ifNone: [self colorForRow: row]]
		ifFalse: [self colorForRow: row].
	
	drawBounds := (self drawBoundsForRow: row) translateBy: (self hMargin @ 0).
	drawBounds := drawBounds intersect: self bounds.

	"Draw icon if existing. Adjust draw bounds in that case."
	(self icon: row) ifNotNil: [ :icon || top |
		top := drawBounds top + ((drawBounds height - icon height) // 2).
		canvas translucentImage: icon at: drawBounds left @ top.
		drawBounds := drawBounds left: drawBounds left + icon width + 2 ].
		
	"We will only draw strings here."
	canvas
		drawString: itemAsText asString
		in: drawBounds
		font: emphasized
		color: rowColor.

	"Draw filter matches if any."
	self
		displayFilterOn: canvas
		for: row
		in: drawBounds
		font: emphasized.! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 7/30/2016 08:26'!
drawBackgroundForMulti: row on: aCanvas 
	"shade the background paler, if this row is selected, but not the current selected row"
	| selectionDrawBounds |
	selectedRow = row ifTrue: [^ self].
	selectionDrawBounds := self drawBoundsForRow: row.
	selectionDrawBounds := selectionDrawBounds intersect: self bounds.
	aCanvas
		fillRectangle: selectionDrawBounds
		color: self multiSelectionColor! !

!LazyListMorph methodsFor: 'list management' stamp: 'ls 7/5/2000 18:21'!
drawBoundsForRow: row
	"calculate the bounds that row should be drawn at.  This might be outside our bounds!!"
	| topLeft drawBounds |
	topLeft := self topLeft x @ (self topLeft y + ((row - 1) * (font height))).
	drawBounds := topLeft extent: self width @ font height.
	^drawBounds! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 7/30/2016 08:26'!
drawOn: aCanvas

	| topRow bottomRow |
	listItems ifEmpty: [ ^self ].
	
	self drawPreSelectionOn: aCanvas.
	
	topRow := self topVisibleRowForCanvas: aCanvas.
	bottomRow := self bottomVisibleRowForCanvas: aCanvas.

	"Draw multi-selection."
	topRow to: bottomRow do: [ :row |
		(listSource itemSelectedAmongMultiple: row) ifTrue: [
			self drawBackgroundForMulti: row on: aCanvas ] ].
	self drawSelectionOn: aCanvas.

	"Draw hovered row if preference enabled."
	PluggableListMorph highlightHoveredRow ifTrue: [
		listSource hoverRow > 0 ifTrue: [
			self highlightHoverRow: listSource hoverRow on: aCanvas ] ].

	"Draw all visible rows."
	topRow to: bottomRow do: [ :row |
		self display: (self item: row) atRow: row on: aCanvas ].

	"Finally, highlight drop row for drag/drop operations.."
	listSource potentialDropRow > 0 ifTrue: [
		self highlightPotentialDropRow: listSource potentialDropRow on: aCanvas ].! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 7/23/2016 17:35'!
drawPreSelectionOn: aCanvas
	
	self 
		drawSelectionFor: preSelectedRow
		withColor: self preSelectionColor 
		on: aCanvas! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 11:32'!
drawSelectionFor: index withColor: color on: aCanvas
	
	| selectionDrawBounds fill |
	index ifNil: [ ^self ].
	index = 0 ifTrue: [ ^self ].
	selectionDrawBounds := self drawBoundsForRow: index.
	selectionDrawBounds := selectionDrawBounds intersect: self bounds.
	
	fill := color isColor
		ifTrue: [SolidFillStyle color: color]
		ifFalse: [color].
	fill isGradientFill ifTrue: [
		fill origin: selectionDrawBounds topLeft.
		fill direction: 0@ selectionDrawBounds height].
	
	aCanvas fillRectangle: selectionDrawBounds fillStyle: fill.! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 7/23/2016 17:35'!
drawSelectionOn: aCanvas
	
	self 
		drawSelectionFor: selectedRow 
		withColor: self selectionColor 
		on: aCanvas! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 7/23/2016 17:45'!
filterColor
	^ self valueOfProperty: #filterColor ifAbsent: [Color yellow]! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 11:29'!
filterColor: aColor

	| cc fill |
	cc := aColor.

	MenuMorph gradientMenu
		ifFalse: [fill := SolidFillStyle color: cc]
		ifTrue: [
			fill := GradientFillStyle ramp: { 
				0.0 -> cc twiceLighter. 
				1 -> cc twiceDarker }].

	self setProperty: #filterColor toValue: fill! !

!LazyListMorph methodsFor: 'list access' stamp: 'mt 4/12/2015 14:22'!
filterOffsets: row
	"Get the character offsets for the matching filter term."
	
	| indexes |
	"Migrate old instances if necessary."
	listFilterOffsets ifNil: [listFilterOffsets := Array new: listItems size].
	
	row <= listFilterOffsets size ifFalse: [
		^ self getFilterOffsets: row].
	
	(indexes := listFilterOffsets at: row) ifNil: [
		indexes := self getFilterOffsets: row.
		listFilterOffsets at: row put: indexes ].
	
	^ indexes! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 7/23/2016 17:45'!
filterTextColor
	^ self valueOfProperty: #filterTextColor ifAbsent: [Color black]! !

!LazyListMorph methodsFor: 'drawing' stamp: 'ls 7/5/2000 18:37'!
font
	"return the font used for drawing.  The response is never nil"
	^font! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 4/12/2015 14:47'!
font: newFont
	font := (newFont ifNil: [ TextStyle default defaultFont ]).
	self adjustHeight.
	listFilterOffsets := Array new: self getListSize withAll: nil.
	self changed.! !

!LazyListMorph methodsFor: 'list access' stamp: 'mt 7/23/2016 17:54'!
getFilterOffsets: row
	"Calculate matching character indexes for the current filter term."
	
	| item filter offsets currentIndex sub |
	filter := listSource filterTerm.
	filter ifEmpty: [^ Array empty].

	item := (self item: row) asStringOrText asString. "See row drawing. Strings only."
	offsets := OrderedCollection new.
	
	currentIndex := 1.
	[currentIndex > 0] whileTrue: [
		currentIndex := item findString: filter startingAt: currentIndex caseSensitive: false.
		currentIndex > 0 ifTrue: [ | left width |
			left := font widthOfString: item from: 1 to: currentIndex-1.
			sub := item copyFrom: currentIndex to: currentIndex + filter size - 1.
			width := font widthOfString: sub.
			offsets addLast: {(left to: left + width). sub}.
			currentIndex := currentIndex + 1] ].
	^ offsets! !

!LazyListMorph methodsFor: 'list access' stamp: 'mt 4/12/2015 13:56'!
getListIcon: row
	"Grab icon directly from the model."
	
	^ listSource iconAt: row
! !

!LazyListMorph methodsFor: 'list access' stamp: 'ls 8/19/2001 14:07'!
getListItem: index
	"grab a list item directly from the model"
	^listSource getListItem: index! !

!LazyListMorph methodsFor: 'list access' stamp: 'ls 5/15/2001 22:11'!
getListSize
	"return the number of items in the list"
	listSource ifNil: [ ^0 ].
	^listSource getListSize! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 7/23/2016 17:38'!
highlightHoverRow: row on: aCanvas
	| drawBounds  |
	drawBounds := self drawBoundsForRow: row.
	drawBounds := drawBounds intersect: self bounds.
	aCanvas fillRectangle: drawBounds color: self hoverColor.! !

!LazyListMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 11:34'!
highlightPotentialDropRow: row  on: aCanvas
	| drawBounds  |
	drawBounds := self drawBoundsForRow: row.
	drawBounds := drawBounds intersect: self bounds.
	aCanvas frameRectangle: drawBounds color: self selectionColor asColor! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 4/1/2015 17:09'!
hMargin

	^ 3! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 7/23/2016 17:38'!
hoverColor
	^ self valueOfProperty: #hoverColor ifAbsent: [Color veryVeryLightGray]! !

!LazyListMorph methodsFor: 'scroll range' stamp: 'cmm 2/27/2011 15:21'!
hUnadjustedScrollRange
"Ok, this is a bit messed up. We need to return the width of the widest item in the list. If we grab every item in the list, it defeats the purpose of LazyListMorph. If we don't, then we don't know the size. 

This is a compromise -- find the widest of the first 30 items, then double it, This width will be updated as new items are installed, so it will always be correct for the visible items. If you know a better way, please chime in."

	| itemsToCheck item index |
	"Check for a cached value"
	maxWidth ifNotNil:[^maxWidth].

	"Compute from scratch"
	itemsToCheck := 30 min: (listItems size).
	maxWidth := 0. 

	"Check the first few items to get a representative sample of the rest of the list."
	index := 1.
	[index < itemsToCheck] whileTrue:
		[ item := self getListItem: index. "Be careful not to actually install this item"
		maxWidth := maxWidth max: (self widthToDisplayItem: item asStringOrText contents).
		index:= index + 1.
		].

	"Add some initial fudge if we didn't check all the items."
	(itemsToCheck < listItems size) ifTrue:[maxWidth := maxWidth*2].

	^maxWidth
! !

!LazyListMorph methodsFor: 'list access' stamp: 'mt 4/12/2015 13:58'!
icon: row
	
	| icon |
	"Migrate old instances if necessary."
	listIcons ifNil: [listIcons := Array new: listItems size].
	
	row <= listIcons size ifFalse: [
		^ self getListIcon: row].
	
	(icon := listIcons at: row) ifNil: [
		icon := self getListIcon: row.
		listIcons at: row put: icon ].
	
	^ icon! !

!LazyListMorph methodsFor: 'initialization' stamp: 'mt 4/12/2015 14:38'!
initialize
	super initialize.
	self color: Color black.
	font := Preferences standardListFont.
	listItems := #().
	listIcons := #().
	listFilterOffsets := #().
	selectedRow := nil.
	selectedRows := PluggableSet integerSet.
	preSelectedRow := nil.
	self adjustHeight.! !

!LazyListMorph methodsFor: 'list access' stamp: 'cmm 2/27/2011 15:26'!
item: index
	"return the index-th item, using the 'listItems' cache"
	| newItem itemWidth |
	(index between: 1 and: listItems size)
		ifFalse: [ "there should have been an update, but there wasn't!!"  ^self getListItem: index].
	(listItems at: index) ifNil: [ 
		newItem := self getListItem: index.
		"Update the width cache."
		maxWidth ifNotNil:[
			itemWidth := self widthToDisplayItem: newItem asStringOrText contents.
			itemWidth > maxWidth ifTrue:[
				maxWidth := itemWidth.
				self adjustWidth.
			]].
		listItems at: index put: newItem ].
	^listItems at: index! !

!LazyListMorph methodsFor: 'list management' stamp: 'mt 4/12/2015 14:38'!
listChanged
	"set newList to be the list of strings to display"
	| size |
	size := self getListSize.
	listItems := Array new: size withAll: nil.
	listIcons := Array new: size withAll: nil.
	listFilterOffsets := Array new: size withAll: nil.
	maxWidth := nil.
	selectedRow := nil.
	selectedRows := PluggableSet integerSet.
	preSelectedRow := nil.
	self adjustHeight.
	self adjustWidth.
	self changed.
! !

!LazyListMorph methodsFor: 'initialization' stamp: 'ls 5/15/2001 22:12'!
listSource: aListSource
	"set the source of list items -- typically a PluggableListMorph"
	listSource := aListSource.
	self listChanged! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 7/30/2016 08:25'!
multiSelectionColor

	^ self valueOfProperty: #multiSelectionColor ifAbsent: [self selectionColor asColor]! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 7/30/2016 08:24'!
multiSelectionColor: aColor

	self setProperty: #multiSelectionColor toValue: aColor.! !

!LazyListMorph methodsFor: 'list management' stamp: 'kb 5/22/2013 16:41'!
preSelectedRow: index
	" Show the user which row is about to become selected, thus providing feedback if there is a delay between the selection gesture and the actual selection of the row. "
	
	preSelectedRow := index.
	self changed.! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 7/23/2016 17:34'!
preSelectionColor
	^ self valueOfProperty: #preSelectionColor ifAbsent: [Color gray]! !

!LazyListMorph methodsFor: 'list access' stamp: 'mt 7/23/2016 17:52'!
resetFilterOffsets
	
	listFilterOffsets := nil.! !

!LazyListMorph methodsFor: 'list management' stamp: 'mt 4/1/2015 10:32'!
rowAtLocation: aPoint
	"return the number of the row at aPoint"
	| y |
	y := aPoint y.
	y < self top ifTrue: [ ^ 1 min: listItems size ].
	^((y - self top // (font height)) + 1) min: listItems size max: 0! !

!LazyListMorph methodsFor: 'updating' stamp: 'mt 3/31/2015 16:20'!
rowChanged: anInteger

	self invalidRect: (self drawBoundsForRow: anInteger).! !

!LazyListMorph methodsFor: 'list management' stamp: 'ls 7/7/2000 10:38'!
selectedRow
	"return the currently selected row, or nil if none is selected"
	^selectedRow! !

!LazyListMorph methodsFor: 'list management' stamp: 'kb 5/22/2013 16:42'!
selectedRow: index
	" Select the index-th row. Clear the pre selection highlight. If nil, remove the current selection. "
	selectedRow := index.
	preSelectedRow := nil.
	self changed.! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 7/23/2016 17:34'!
selectionColor
	^ self valueOfProperty: #selectionColor ifAbsent: [Color blue]! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 11:31'!
selectionColor: aColor

	| cc fill |
	cc := aColor.

	MenuMorph gradientMenu
		ifFalse: [fill := SolidFillStyle color: cc]
		ifTrue: [
			fill := GradientFillStyle ramp: { 
				0.0 -> cc twiceLighter. 
				1 -> cc twiceDarker }].

	self setProperty: #selectionColor toValue: fill! !

!LazyListMorph methodsFor: 'accessing' stamp: 'mt 7/23/2016 17:35'!
selectionTextColor
	^ self valueOfProperty: #selectionTextColor ifAbsent: [Color white]! !

!LazyListMorph methodsFor: 'list management' stamp: 'kb 5/22/2013 16:41'!
selectRow: index
	" Select the index-th row. Clear the pre selection highlight. "
	selectedRows add: index.
	preSelectedRow := nil.
	self changed.! !

!LazyListMorph methodsFor: 'drawing' stamp: 'ls 5/17/2001 21:57'!
topVisibleRowForCanvas: aCanvas
        "return the top visible row in aCanvas's clip rectangle"
        ^self rowAtLocation: aCanvas clipRect topLeft.
! !

!LazyListMorph methodsFor: 'list management' stamp: 'kb 5/22/2013 16:42'!
unselectRow: index
	"unselect the index-th row"
	selectedRows remove: index ifAbsent: [].
	preSelectedRow := nil.
	self changed.! !

!LazyListMorph methodsFor: 'accessing' stamp: 'bf 6/25/2011 13:36'!
userString
	"Do I have a text string to be searched on?"

	^ String streamContents: [:strm |
		1 to: self getListSize do: [:i |
			"must use asStringOrText because that's what the drawing uses, too"
			strm nextPutAll: (self getListItem: i) asStringOrText; cr]]! !

!LazyListMorph methodsFor: 'scroll range' stamp: 'cmm 2/27/2011 15:19'!
widthToDisplayItem: item 
	^ self font widthOfStringOrText: item asStringOrText! !

LazyListMorph subclass: #MulticolumnLazyListMorph
	instanceVariableNames: 'columnWidths'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!MulticolumnLazyListMorph commentStamp: '<historical>' prior: 0!
A variant of LazyListMorph that can display multi-column lists.!

!MulticolumnLazyListMorph methodsFor: 'drawing' stamp: 'mt 4/1/2015 17:14'!
display: items atRow: row on: canvas 
	"display the specified item, which is on the specified row; for Multicolumn 
	lists, items will be a list of strings"
	| drawBounds |
	drawBounds :=  (self drawBoundsForRow: row) translateBy: (self hMargin @ 0).
	drawBounds := drawBounds intersect: self bounds.
	items
		with: (1 to: items size)
		do: [:item :index | 
			"move the bounds to the right at each step"
			index > 1
				ifTrue: [drawBounds := drawBounds left: drawBounds left + 6
									+ (columnWidths at: index - 1)].
			item isText
				ifTrue: [canvas
						drawString: item
						in: drawBounds
						font: (font
								emphasized: (item emphasisAt: 1))
						color: (self colorForRow: row)]
				ifFalse: [canvas
						drawString: item
						in: drawBounds
						font: font
						color: (self colorForRow: row)]]! !

!MulticolumnLazyListMorph methodsFor: 'drawing' stamp: 'ls 5/17/2001 21:58'!
drawOn: aCanvas
        self getListSize = 0 ifTrue:[ ^self ].

        self setColumnWidthsFor: aCanvas.

        super drawOn: aCanvas! !

!MulticolumnLazyListMorph methodsFor: 'as yet unclassified' stamp: 'ls 5/17/2001 21:23'!
getListItem: index
	^listSource getListRow: index! !

!MulticolumnLazyListMorph methodsFor: 'scroll range' stamp: 'sps 4/2/2004 12:16'!
hUnadjustedScrollRange
"multi column list morphs don't use hScrollbars"

	^0

! !

!MulticolumnLazyListMorph methodsFor: 'as yet unclassified' stamp: 'ls 5/18/2001 16:43'!
listChanged
	columnWidths := nil.
	super listChanged! !

!MulticolumnLazyListMorph methodsFor: 'drawing' stamp: 'sps 3/23/2004 15:51'!
setColumnWidthsFor: aCanvas
        | row topRow bottomRow |
        "set columnWidths for drawing on the specified canvas"
		columnWidths ifNil: [
		columnWidths := (self item: 1) collect: [ :ignored | 0 ]. ].
	topRow := (self topVisibleRowForCanvas: aCanvas) max: 1.
	bottomRow :=  (self bottomVisibleRowForCanvas: aCanvas) max: 1.
	topRow > bottomRow ifTrue: [ ^ self ].
	topRow to: bottomRow do: [ :rowIndex |
                row := self item: rowIndex.
                columnWidths := columnWidths with: row collect: [ :currentWidth :item |
				| widthOfItem |
				widthOfItem := (font widthOfStringOrText: item).
				widthOfItem > currentWidth
					ifTrue: [ self changed.  widthOfItem ]
					ifFalse: [ currentWidth ] ] ]! !

!MulticolumnLazyListMorph methodsFor: 'scroll range' stamp: 'ls 4/17/2004 12:21'!
widthToDisplayItem: item
	| widths |
	widths := item collect: [ :each | super widthToDisplayItem: each ].
	^widths sum + (10 * (widths size - 1))   "add in space between the columns"
! !

Morph subclass: #MatrixTransformMorph
	instanceVariableNames: 'transform'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Balloon'!

!MatrixTransformMorph commentStamp: '<historical>' prior: 0!
MatrixTransformMorph is similar to TransformMorph but uses a MatrixTransform2x3 instead of a MorphicTransform. It is used by clients who want use the BalloonEngine for vector-based scaling instead of the standard WarpBlt pixel-based mechanism.!

!MatrixTransformMorph methodsFor: 'rotate scale and flex' stamp: 'ar 11/15/1998 21:55'!
addFlexShell
	"No flex shell necessary"
	self lastRotationDegrees: 0.0.! !

!MatrixTransformMorph methodsFor: 'initialize' stamp: 'wiz 11/6/2005 16:58'!
asFlexOf: aMorph
	"Initialize me with position and bounds of aMorph,
	and with an offset that provides centered rotation."
	self addMorph: aMorph.
	self setRotationCenterFrom: aMorph center .
	self lastRotationDegrees: 0.0. 
	self computeBounds! !

!MatrixTransformMorph methodsFor: 'halos and balloon help' stamp: 'ar 11/15/1998 21:50'!
balloonHelpTextForHandle: aHandle
	aHandle eventHandler firstMouseSelector == #changeRotationCenter:with:
		ifTrue:[^'set center of rotation'].
	^super balloonHelpTextForHandle: aHandle! !

!MatrixTransformMorph methodsFor: 'geometry' stamp: 'mdr 12/19/2001 10:48'!
boundsChangedFrom: oldBounds to: newBounds
	oldBounds extent = newBounds extent ifFalse:[
		transform := transform composedWithGlobal:
			(MatrixTransform2x3 withOffset: oldBounds origin negated).
		transform := transform composedWithGlobal:
			(MatrixTransform2x3 withScale: newBounds extent / oldBounds extent).
		transform := transform composedWithGlobal:
			(MatrixTransform2x3 withOffset: newBounds origin).
	].
	transform offset: transform offset + (newBounds origin - oldBounds origin)! !

!MatrixTransformMorph methodsFor: 'updating' stamp: 'ar 11/12/2000 18:51'!
changed
	^self invalidRect: (self fullBounds insetBy: -1)! !

!MatrixTransformMorph methodsFor: 'flexing' stamp: 'ar 11/16/1998 01:19'!
changeRotationCenter: evt with: rotHandle
	| pos |
	pos := evt cursorPoint.
	rotHandle referencePosition: pos.
	self referencePosition: pos.! !

!MatrixTransformMorph methodsFor: 'geometry' stamp: 'nice 12/27/2009 03:10'!
computeBounds
	| box |
	(submorphs isNil or:[submorphs isEmpty]) ifTrue:[^self].
	box := nil.
	submorphs do:[:m| | subBounds |
		subBounds := self transform localBoundsToGlobal: m bounds.
		box 
			ifNil:[box := subBounds]
			ifNotNil:[box := box quickMerge: subBounds].
	].
	box ifNil:[box := 0@0 corner: 20@20].
	fullBounds := bounds := box! !

!MatrixTransformMorph methodsFor: 'geometry testing' stamp: 'ar 1/15/1999 16:34'!
containsPoint: aPoint
	self visible ifFalse:[^false].
	(bounds containsPoint: aPoint) ifFalse: [^ false].
	self hasSubmorphs
		ifTrue: [self submorphsDo: 
					[:m | (m fullContainsPoint: (self transform globalPointToLocal: aPoint))
							ifTrue: [^ true]].
				^ false]
		ifFalse: [^ true]! !

!MatrixTransformMorph methodsFor: 'drawing' stamp: 'ar 11/15/1998 22:20'!
drawOn: aCanvas! !

!MatrixTransformMorph methodsFor: 'drawing' stamp: 'ar 5/29/1999 09:01'!
drawSubmorphsOn: aCanvas
	aCanvas asBalloonCanvas transformBy: self transform
		during:[:myCanvas| super drawSubmorphsOn: myCanvas].! !

!MatrixTransformMorph methodsFor: 'geometry' stamp: 'ar 11/15/1998 21:52'!
extent: extent
	self handleBoundsChange:[super extent: extent]! !

!MatrixTransformMorph methodsFor: 'layout' stamp: 'nice 12/27/2009 03:10'!
fullBounds
	
	fullBounds ifNil:[
		fullBounds := self bounds.
		submorphs do:[:m| | subBounds |
			subBounds := (self transform localBoundsToGlobal: m fullBounds).
			fullBounds := fullBounds quickMerge: subBounds.
		].
	].
	^fullBounds! !

!MatrixTransformMorph methodsFor: 'geometry testing' stamp: 'ar 11/15/1998 21:52'!
fullContainsPoint: aPoint
	| p |
	self visible ifFalse:[^false].
	(self fullBounds containsPoint: aPoint) ifFalse:[^false].
	(self containsPoint: aPoint) ifTrue:[^true].
	p := self transform globalPointToLocal: aPoint.
	submorphs do:[:m|
		(m fullContainsPoint: p) ifTrue:[^true].
	].
	^false! !

!MatrixTransformMorph methodsFor: 'geometry' stamp: 'ar 11/15/1998 21:52'!
handleBoundsChange: aBlock
	| oldBounds newBounds |
	oldBounds := bounds.
	aBlock value.
	newBounds := bounds.
	self boundsChangedFrom: oldBounds to: newBounds.! !

!MatrixTransformMorph methodsFor: 'flexing' stamp: 'ar 11/15/1998 21:55'!
hasNoScaleOrRotation
	^true! !

!MatrixTransformMorph methodsFor: 'geometry eToy' stamp: 'ar 6/12/2001 06:03'!
heading
	"Return the receiver's heading (in eToy terms)"
	^ self forwardDirection + self innerAngle! !

!MatrixTransformMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:38'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	transform := MatrixTransform2x3 identity! !

!MatrixTransformMorph methodsFor: 'flexing' stamp: 'fbs 11/26/2004 10:59'!
innerAngle
	^ (self transform a11 @ self transform a21) degrees! !

!MatrixTransformMorph methodsFor: 'change reporting' stamp: 'ar 11/12/2000 18:44'!
invalidRect: rect from: aMorph
	aMorph == self
		ifTrue:[super invalidRect: rect from: self]
		ifFalse:[super invalidRect: (self transform localBoundsToGlobal: rect) from: aMorph].! !

!MatrixTransformMorph methodsFor: 'flexing' stamp: 'ar 11/15/1998 21:56'!
lastRotationDegrees
	^(self valueOfProperty: #lastRotationDegrees) ifNil:[0.0].! !

!MatrixTransformMorph methodsFor: 'flexing' stamp: 'ar 11/15/1998 21:56'!
lastRotationDegrees: deg
	deg = 0.0 
		ifTrue:[self removeProperty: #lastRotationDegrees]
		ifFalse:[self setProperty: #lastRotationDegrees toValue: deg]! !

!MatrixTransformMorph methodsFor: 'private' stamp: 'ar 6/12/2001 06:38'!
privateFullMoveBy: delta
	self privateMoveBy: delta.
	transform offset: transform offset + delta.! !

!MatrixTransformMorph methodsFor: 'flexing' stamp: 'ar 11/15/1998 21:56'!
removeFlexShell
	"Do nothing"! !

!MatrixTransformMorph methodsFor: 'flexing' stamp: 'mdr 12/19/2001 10:49'!
rotateBy: delta
	| pt m |
	delta = 0.0 ifTrue:[^self].
	self changed.
	pt := self transformFromWorld globalPointToLocal: self referencePosition.
	m := MatrixTransform2x3 withOffset: pt.
	m := m composedWithLocal: (MatrixTransform2x3 withAngle: delta).
	m := m composedWithLocal: (MatrixTransform2x3 withOffset: pt negated).
	self transform: (transform composedWithLocal: m).
	self changed.! !

!MatrixTransformMorph methodsFor: 'geometry eToy' stamp: 'ar 6/12/2001 05:11'!
rotationCenter
	| pt |
	pt := self transform localPointToGlobal: super rotationCenter.
	^pt - bounds origin / bounds extent asFloatPoint! !

!MatrixTransformMorph methodsFor: 'geometry eToy' stamp: 'ar 6/12/2001 05:07'!
rotationCenter: aPoint
	super rotationCenter: (self transform globalPointToLocal: bounds origin + (bounds extent * aPoint))! !

!MatrixTransformMorph methodsFor: 'flexing' stamp: 'ar 11/15/1998 21:56'!
rotationDegrees: degrees
	| last delta |
	last := self lastRotationDegrees.
	delta := degrees - last.
	self rotateBy: delta.
	self lastRotationDegrees: degrees.! !

!MatrixTransformMorph methodsFor: 'geometry eToy' stamp: 'ar 6/12/2001 05:50'!
setDirectionFrom: aPoint
	| delta degrees |
	delta := (self transformFromWorld globalPointToLocal: aPoint) - super rotationCenter.
	degrees := delta degrees + 90.0.
	self forwardDirection: (degrees \\ 360) rounded.
! !

!MatrixTransformMorph methodsFor: 'menus' stamp: 'jcg 11/1/2001 13:03'!
setRotationCenterFrom: aPoint

	super setRotationCenterFrom: (self transformFromWorld localPointToGlobal: aPoint)
! !

!MatrixTransformMorph methodsFor: 'flexing' stamp: 'ar 9/11/2000 21:16'!
transform
	^ transform ifNil: [MatrixTransform2x3 identity]! !

!MatrixTransformMorph methodsFor: 'geometry' stamp: 'ar 10/6/2000 15:37'!
transformedBy: aTransform
	self transform: (self transform composedWithGlobal: aTransform).! !

!MatrixTransformMorph methodsFor: 'event handling' stamp: 'ar 9/12/2000 01:22'!
transformFrom: uberMorph
	(owner isNil or:[self == uberMorph]) ifTrue:[^self transform].
	^(owner transformFrom: uberMorph) asMatrixTransform2x3 composedWithLocal: self transform! !

!MatrixTransformMorph methodsFor: 'accessing' stamp: 'ar 11/15/1998 21:51'!
transform: aMatrixTransform
	transform := aMatrixTransform.
	self computeBounds.! !

!MatrixTransformMorph methodsFor: 'drawing' stamp: 'md 2/27/2006 09:51'!
visible: aBoolean 
	"set the 'visible' attribute of the receiver to aBoolean"
	extension ifNil: [aBoolean ifTrue: [^ self]].
	self assureExtension visible: aBoolean! !

Morph subclass: #MenuLineMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Menus'!

!MenuLineMorph methodsFor: 'drawing' stamp: 'mt 4/21/2015 17:38'!
drawOn: aCanvas 
	| baseColor |
	baseColor := Preferences menuColorFromWorld
				ifTrue: [owner color twiceDarker]
				ifFalse: [Preferences menuAppearance3d
						ifTrue: [owner color]
						ifFalse: [MenuMorph menuLineColor]].
	Preferences menuAppearance3d
		ifTrue: [
			aCanvas
				fillRectangle: (bounds topLeft corner: bounds rightCenter)
				color: baseColor twiceDarker.
			
			aCanvas
				fillRectangle: (bounds leftCenter corner: bounds bottomRight)
				color: baseColor twiceLighter]
		ifFalse: [
			aCanvas
				fillRectangle: ((bounds topLeft corner: bounds bottomRight) insetBy: (0@0 corner: 0@ 1))
				color: baseColor]! !

!MenuLineMorph methodsFor: 'initialization' stamp: 'ar 11/8/2000 23:09'!
initialize
	super initialize.
	self hResizing: #spaceFill; vResizing: #spaceFill.! !

!MenuLineMorph methodsFor: 'layout' stamp: 'dgd 9/1/2004 19:20'!
minHeight
	"answer the receiver's minHeight"
	^ self isInDockingBar
		ifTrue: [owner isVertical
				ifTrue: [2]
				ifFalse: [10]]
		ifFalse: [2]! !

!MenuLineMorph methodsFor: 'layout' stamp: 'dgd 9/1/2004 19:21'!
minWidth
	"answer the receiver's minWidth"
	^ self isInDockingBar
		ifTrue: [owner isVertical
				ifTrue: [10]
				ifFalse: [2]]
		ifFalse: [10]! !

!MenuLineMorph methodsFor: 'submorphs-accessing' stamp: 'dgd 9/1/2004 19:12'!
noteNewOwner: aMorph 
	"I have just been added as a submorph of aMorph"
	super noteNewOwner: aMorph.
	self updateLayoutInDockingBar! !

!MenuLineMorph methodsFor: 'change reporting' stamp: 'dgd 9/1/2004 19:12'!
ownerChanged
	"The receiver's owner, some kind of a pasteup, has changed its 
	layout."
	super ownerChanged.
	self updateLayoutInDockingBar! !

!MenuLineMorph methodsFor: 'private' stamp: 'dgd 9/1/2004 19:20'!
updateLayoutInDockingBar
	self isInDockingBar
		ifFalse: [^ self].
	""
	owner isVertical
		ifFalse: [""
			self hResizing: #shrinkWrap.
			self vResizing: #spaceFill]
		ifTrue: [""
			self hResizing: #spaceFill.
			self vResizing: #shrinkWrap].
	self extent: self minWidth @ self minHeight! !

Morph subclass: #MenuMorph
	instanceVariableNames: 'defaultTarget selectedItem stayUp popUpOwner activeSubMenu'
	classVariableNames: 'CloseBoxImage CloseBoxImageFlat CloseBoxImageGradient GradientMenu MenuBorderColor MenuBorderWidth MenuColor MenuLineColor MenuSelectionColor MenuTitleBorderColor MenuTitleBorderWidth MenuTitleColor PushPinImage RoundedMenuCorners'
	poolDictionaries: ''
	category: 'Morphic-Menus'!

!MenuMorph commentStamp: '<historical>' prior: 0!
Instance variables:
	defaultTarget 	<Object>				The default target for creating menu items
	selectedItem		<MenuItemMorph> 	The currently selected item in the receiver
	stayUp 			<Boolean>			True if the receiver should stay up after clicks
	popUpOwner 	<MenuItemMorph>	The menu item that automatically invoked the receiver, if any.
	activeSubMenu 	<MenuMorph>		The currently active submenu.!

MenuMorph subclass: #DockingBarMenuMorph
	instanceVariableNames: 'activatorDockingBar'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Menus-DockingBar'!

!DockingBarMenuMorph methodsFor: 'as yet unclassified' stamp: 'kb 2/22/2010 10:21'!
activatedFromDockingBar: aDockingBar 

	activatorDockingBar := aDockingBar! !

!DockingBarMenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/22/2010 16:41'!
handleCRStroke: evt

	evt keyValue = 13 ifFalse: [ ^false ].
	selectedItem ifNotNil: [ selectedItem invokeWithEvent: evt ].
	^true! !

!DockingBarMenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/24/2010 13:16'!
handleLeftStroke: evt

	28 = evt keyValue ifFalse: [ ^false ].
	(self stepIntoSubmenu: evt) ifFalse: [ 
		self deactivate: evt.
		activatorDockingBar moveSelectionDown: -1 event: evt ].
	^true! !

!DockingBarMenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/24/2010 13:16'!
handleRightStroke: evt

	29 = evt keyValue ifFalse: [ ^false ].
	(self stepIntoSubmenu: evt) ifFalse: [
		self deactivate: evt.
		activatorDockingBar moveSelectionDown: 1 event: evt ].
	^true! !

!DockingBarMenuMorph methodsFor: 'rounding' stamp: 'kb 2/22/2010 10:13'!
roundedCorners
	"Return a list of those corners to round"
	activatorDockingBar isFloating
		ifTrue: [^ #(2 3 )].
	activatorDockingBar isAdheringToTop
		ifTrue: [^ #(2 3 )].
	activatorDockingBar isAdheringToBottom
		ifTrue: [^ #(1 4 )].
	activatorDockingBar isAdheringToLeft
		ifTrue: [^ #(3 4 )].
	activatorDockingBar isAdheringToRight
		ifTrue: [^ #(1 2 )]! !

!DockingBarMenuMorph methodsFor: 'copying' stamp: 'kb 2/22/2010 10:16'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all, but shared.  	Warning!!!!  Every instance variable defined in this class must be handled.  We must also implement veryDeepFixupWith:.  See DeepCopier class comment."

	super veryDeepInner: deepCopier.
	activatorDockingBar := activatorDockingBar.  "Weakly copied"
! !

DockingBarMenuMorph subclass: #DockingBarUpdatingMenuMorph
	instanceVariableNames: 'arguments updater updateSelector menuUpdater'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Menus-DockingBar'!

!DockingBarUpdatingMenuMorph methodsFor: 'as yet unclassified' stamp: 'kb 5/22/2012 16:57'!
delete

	owner ifNotNil: [ 
		" When deleted remove my menu items, so I can avoid holding unwanted references to other objects. They will be updated anyway when I become visible again. "
		" The owner notNil condition is necessary because MenuItemMorph >> select: sends delete before I become visible, but after the menu items are updated. "
		self removeAllMorphs ].
	super delete! !

!DockingBarUpdatingMenuMorph methodsFor: 'as yet unclassified' stamp: 'kb 2/22/2010 14:59'!
initialize

	super initialize.
	menuUpdater := MenuUpdater new! !

!DockingBarUpdatingMenuMorph methodsFor: 'as yet unclassified' stamp: 'kb 2/22/2010 15:00'!
updateMenu

	menuUpdater update: self! !

!DockingBarUpdatingMenuMorph methodsFor: 'as yet unclassified' stamp: 'kb 2/22/2010 14:59'!
updater: anObject updateSelector: aSelector

	menuUpdater updater: anObject updateSelector: aSelector! !

!DockingBarUpdatingMenuMorph methodsFor: 'as yet unclassified' stamp: 'kb 2/22/2010 14:59'!
updater: anObject updateSelector: aSelector arguments: anArray

	menuUpdater updater: anObject updateSelector: aSelector arguments: anArray! !

MenuMorph subclass: #DumberMenuMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Explorer'!

!DumberMenuMorph commentStamp: '<historical>' prior: 0!
Contributed by Bob Arning as part of the ObjectExplorer package.
!

!DumberMenuMorph methodsFor: 'menu' stamp: 'RAA 6/21/1999 15:40'!
setInvokingView: invokingView
	"I'd rather not, if that's OK"! !

!MenuMorph class methodsFor: 'utilities' stamp: 'ul 12/12/2009 14:09'!
chooseFrom: aList lines: linesArray title: queryString
	"Choose an item from the given list. Answer the index of the selected item."
	"MenuMorph
		chooseFrom: #('Hello' 'World' 'Here' 'We' 'Go')
		lines: #(2 4)
		title: 'What''s up?'"
	| menu aBlock result |
	(ProvideAnswerNotification signal: queryString) ifNotNil:[:answer | 
		1 to: aList size do:[:i| (aList at: i) = answer ifTrue:[^i]].
		^0].
	aBlock := [:v| result := v].
	menu := self new.
	menu addTitle: queryString.
	1 to: aList size do:[:i| 
		menu add: (aList at: i) asString target: aBlock selector: #value: argument: i.
		(linesArray includes: i) ifTrue:[menu addLine]
	].
	MenuIcons decorateMenu: menu.
	result := 0.
	menu invokeAt: ActiveHand position in: ActiveWorld allowKeyboard: true.
	^result! !

!MenuMorph class methodsFor: 'utilities' stamp: 'ul 12/12/2009 14:09'!
chooseFrom: aList values: valueList lines: linesArray title: queryString
	"Choose an item from the given list. Answer the index of the selected item."
	"MenuMorph
		chooseFrom: #('Hello' 'World' 'Here' 'We' 'Go')
		values: #('Hello' 'World' 'Here' 'We' 'Go')
		lines: #(2 4)
		title: 'What''s up?'"
	| menu aBlock result |
	(ProvideAnswerNotification signal: queryString) ifNotNil:[:answer | 
		1 to: aList size do:[:i| (aList at: i) = answer ifTrue:[^answer]].
		^nil].
	aBlock := [:v| result := v].
	menu := self new.
	menu addTitle: queryString.
	1 to: aList size do:[:i| 
		menu add: (aList at: i) asString target: aBlock selector: #value: argument: (valueList at: i).
		(linesArray includes: i) ifTrue:[menu addLine]
	].
	MenuIcons decorateMenu: menu.
	result := nil.
	menu invokeAt: ActiveHand position in: ActiveWorld allowKeyboard: true.
	^result! !

!MenuMorph class methodsFor: 'images' stamp: 'topa 3/9/2015 00:10'!
closeBoxImage

	^ self gradientMenu
		ifTrue: [self closeBoxImageGradient]
		ifFalse: [self closeBoxImageFlat]! !

!MenuMorph class methodsFor: 'images' stamp: 'topa 3/9/2015 00:11'!
closeBoxImageFlat

	^ CloseBoxImageFlat ifNil: [CloseBoxImageFlat := SystemWindow closeBoxImageFlat]! !

!MenuMorph class methodsFor: 'images' stamp: 'topa 3/9/2015 00:09'!
closeBoxImageGradient

	^ CloseBoxImageGradient ifNil: [CloseBoxImageGradient := SystemWindow closeBoxImageGradient]! !

!MenuMorph class methodsFor: 'utilities' stamp: 'ar 8/4/2009 22:40'!
confirm: queryString
	"Put up a yes/no menu with caption queryString. Answer true if the 
	response is yes, false if no. This is a modal question--the user must 
	respond yes or no."

	"MenuMorph confirm: 'Are you hungry?'"

	^ self confirm: queryString trueChoice: 'Yes' translated falseChoice: 'No' translated! !

!MenuMorph class methodsFor: 'utilities' stamp: 'ar 8/5/2009 01:10'!
confirm: queryString orCancel: cancelBlock 
	"Put up a yes/no/cancel menu with caption aString. Answer true if  the response is yes, false if no. If cancel is chosen, evaluate cancelBlock. This is a modal question--the user must respond yes or no."
	"MenuMorph confirm: 'Reboot universe' orCancel:[^'Nevermind'] "
	| choice |
	choice := self chooseFrom: {'Yes' translated. 'No' translated. 'Cancel' translated}
		lines: #()
		title: queryString.
	choice = 1 ifTrue: [^ true].
	choice = 2 ifTrue: [^ false].
	^ cancelBlock value! !

!MenuMorph class methodsFor: 'utilities' stamp: 'ul 12/12/2009 14:09'!
confirm: queryString trueChoice: trueChoice falseChoice: falseChoice 
	"Put up a yes/no menu with caption queryString. The actual wording for the two choices will be as provided in the trueChoice and falseChoice parameters. Answer true if the response is the true-choice,  false if it's the false-choice. This is a modal question -- the user must respond one way or the other."
	"MenuMorph 
		confirm: 'Are you hungry?'  
		trueChoice: 'yes, I''m famished'  
		falseChoice: 'no, I just ate'"
	| menu aBlock result |
	(ProvideAnswerNotification signal: queryString) 
		ifNotNil:[:answer | ^ trueChoice = answer].
	aBlock := [:v| result := v].
	menu := self new.
	menu addTitle: queryString icon: MenuIcons confirmIcon.
	menu add: trueChoice target: aBlock selector: #value: argument: true.
	menu add: falseChoice target: aBlock selector: #value: argument: false.
	MenuIcons decorateMenu: menu.
	[menu invokeAt: ActiveHand position in: ActiveWorld allowKeyboard: true.
	result == nil] whileTrue.
	^result! !

!MenuMorph class methodsFor: 'instance creation' stamp: 'jm 5/14/1998 17:21'!
entitled: aString
	"Answer a new instance of me with the given title."

	^ self new addTitle: aString
! !

!MenuMorph class methodsFor: 'example' stamp: 'kfr 7/1/2014 12:29'!
example
	"MenuMorph example popUpInWorld"

	| menu |
	menu := MenuMorph new.
	menu addTitle: 'Fruit' translated.
	menu addStayUpItem.
	menu add: 'apples' action: #apples.
	menu add: 'oranges' action: #oranges.
	menu addLine.
	menu addLine.  "extra lines ignored"
	menu add: 'peaches' action: #peaches.
	menu addLine.
	menu add: 'pears' action: #pears.
	menu addLine.
	^ menu
! !

!MenuMorph class methodsFor: 'instance creation' stamp: 'ul 11/15/2010 11:53'!
fromArray: anArray 
	"Construct a menu from anArray. The elements of anArray  
	must be either:  
	* A pair of the form: <label> <selector>  
	or	* The 'dash' (or 'minus sign') symbol  
	 
	Refer to the example at the bottom of the method"
	| menu |

	menu := self new.

	anArray
		do: [:anElement |
			anElement size = 1
				ifTrue: [
					anElement == #- ifFalse: [^ self error: 'badly-formed menu constructor'].
					menu addLine.
				]
				ifFalse: [
					anElement size = 2 ifFalse: [^ self error: 'badly-formed menu constructor'].
					menu add: anElement first action: anElement second.
				]
		].

	^ menu! !

!MenuMorph class methodsFor: 'preferences' stamp: 'mt 2/27/2015 11:31'!
gradientMenu

	<preference: 'gradientMenu'
		category: #menus
		description: 'If true, the menus will have a gradient look.'
		type: #Boolean>
	^ GradientMenu ifNil: [true]! !

!MenuMorph class methodsFor: 'preferences' stamp: 'mt 7/29/2016 15:18'!
gradientMenu: aBoolean

	GradientMenu = aBoolean ifTrue: [^ self].
	GradientMenu := aBoolean.

	self flag: #refactor. "mt: Heavy use of the gradient-menu property..."
	
	SystemProgressMorph reset.
	PluggableTextMorph flushAdornmentCache.

	MenuMorph allSubInstancesDo: [:ea | ea setDefaultParameters; setTitleParameters; updateColor].
	DockingBarMorph allSubInstancesDo: [:ea | ea setDefaultParameters; updateColor].

	TheWorldMainDockingBar updateInstances.
	PluggableListMorph allSubInstancesDo: [:ea | ea setListParameters].
	SimpleHierarchicalListMorph allSubInstancesDo: [:ea | ea setDefaultParameters].
	! !

!MenuMorph class methodsFor: 'utilities' stamp: 'ul 12/12/2009 14:09'!
inform: queryString
	"MenuMorph inform: 'I like Squeak'"
	| menu |
	(ProvideAnswerNotification signal: queryString) 
		ifNotNil:[:answer | ^ self].
	menu := self new.
	menu addTitle: queryString icon: MenuIcons confirmIcon.
	menu add: 'OK' target: self selector: #yourself.
	MenuIcons decorateMenu: menu.
	menu invokeAt: ActiveHand position in: ActiveWorld allowKeyboard: true.! !

!MenuMorph class methodsFor: 'instance creation' stamp: 'tfel 3/26/2010 11:47'!
initialize

	"MenuMorph initialize"
	CloseBoxImage := nil.
	PushPinImage := nil.

! !

!MenuMorph class methodsFor: 'images' stamp: 'MAD 1/2/2011 00:27'!
pushPinImage
	"Answer the push-pin image, creating and caching it at this time if it is absent"

	^ MenuIcons smallPinIcon! !

!MenuMorph class methodsFor: 'preferences' stamp: 'mt 2/27/2015 13:50'!
roundedMenuCorners

	<preference: 'roundedMenuCorners'
		category: #menus
		description: 'Whether morphic menus should have rounded corners.'
		type: #Boolean>
	^ RoundedMenuCorners ifNil: [true]! !

!MenuMorph class methodsFor: 'preferences' stamp: 'mt 2/27/2015 14:01'!
roundedMenuCorners: aBoolean

	RoundedMenuCorners := aBoolean.
	SystemProgressMorph reset.
	TheWorldMainDockingBar updateInstances.! !

!MenuMorph class methodsFor: 'preferences' stamp: 'mt 7/10/2016 12:31'!
themeProperties

	^ super themeProperties, {
		{ #borderColor. 'Colors'. 'Color of the menu''s border.' }.
		{ #borderWidth. 'Borders'. 'Width of the menu''s border.' }.
		{ #borderStyle. 'Borders'. 'Whether to use a plain border, inset, or outset.' }.
		{ #color. 'Colors'. 'Background color of the menu.' }.

		{ #titleBorderColor. 'Colors'. 'Color of the menu title border.' }.
		{ #titleBorderWidth. 'Geometry'. 'Width of the menu title border.' }.
		{ #titleBorderStyle. 'Borders'. 'Whether to use a plain border, inset, or outset for the title.' }.
		{ #titleColor. 'Colors'. 'Background color of the menu'' title.' }.
		{ #titleFont. 'Fonts'. 'Font for menu title.' }.
		{ #titleTextColor. 'Colors'. 'Color for the menu title label.' }.

		{ #lineColor. 'Colors'. 'Color of the separators between menu items.' }.
		{ #lineStyle. 'Colors'. 'Use border-style to change appearance.' }.
		{ #lineWidth. 'Geometry'. 'How big the separators should be.' }.
	}! !

!MenuMorph methodsFor: 'events' stamp: 'kb 2/4/2010 13:06'!
activate: evt
	"Receiver should be activated; e.g., so that control passes correctly."
	evt hand 
		newMouseFocus: self;
		newKeyboardFocus: self! !

!MenuMorph methodsFor: 'control' stamp: 'kb 2/22/2010 10:20'!
activeSubmenu: aSubmenu 
	activeSubMenu ifNotNil: [
		activeSubMenu delete ].
	activeSubMenu := aSubmenu.
	activeSubMenu ifNotNil: [
		activeSubMenu updateMenu ]! !

!MenuMorph methodsFor: 'construction' stamp: 'dgd 10/1/2004 13:48'!
addAllFrom: aMenuMorph 
	aMenuMorph submorphs
		do: [:each | 
			(each isKindOf: MenuItemMorph)
				ifTrue: [self
						add: each contents
						target: each target
						selector: each selector
						argumentList: each arguments].
			(each isKindOf: MenuLineMorph)
				ifTrue: [self addLine]] ! !

!MenuMorph methodsFor: 'accessing' stamp: 'dgd 9/13/2004 19:59'!
addBlankIconsIfNecessary: anIcon 
	"If any of my items have an icon, ensure that all do by using  
	anIcon for those that don't"
	self items
		reject: [:each | each hasIconOrMarker]
		thenDo: [:each | each icon: anIcon]! !

!MenuMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 21:52'!
addCustomMenuItems: aCustomMenu hand: aHandMorph

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu addLine.
	aCustomMenu add: 'add title...' translated action: #addTitle.
	aCustomMenu add: 'set target...' translated action: #setTarget:.
	defaultTarget ifNotNil: [
		aCustomMenu add: 'add item...' translated action: #addItem].
	aCustomMenu add: 'add line' translated action: #addLine.
	(self items count:[:any| any hasSubMenu]) > 0
		ifTrue:[aCustomMenu add: 'detach submenu' translated action: #detachSubMenu:].! !

!MenuMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:41'!
addItem

	| string sel |
	string := UIManager default request: 'Label for new item?'.
	string isEmpty ifTrue: [^ self].
	sel := UIManager default request: 'Selector?'.
	sel isEmpty ifFalse: [sel := sel asSymbol].
	self add: string action: sel.
! !

!MenuMorph methodsFor: 'construction' stamp: 'cmm 3/9/2010 09:40'!
addItem: aBlock
	| item |
	item := MenuItemMorph new.
	aBlock value: item.
	self addMenuItem: item! !

!MenuMorph methodsFor: 'construction' stamp: 'mt 7/9/2016 13:06'!
addLine
	"Append a divider line to this menu. Suppress duplicate lines."

	| colorToUse |
	self hasItems ifFalse: [^ self].
	self lastSubmorph knownName = #line ifTrue: [^ self].
	
	colorToUse := self userInterfaceTheme lineColor ifNil: [Color gray: 0.9].
	self addMorphBack: (Morph new
		color: colorToUse;
		hResizing: #spaceFill;
		height: (self userInterfaceTheme lineWidth ifNil: [2]);
		borderStyle: (self userInterfaceTheme lineStyle ifNil: [BorderStyle inset]);
		borderColor: colorToUse;
		borderWidth: 1;
		name: #line; "see above"
		yourself).! !

!MenuMorph methodsFor: 'construction' stamp: 'sw 2/27/2001 07:50'!
addList: aList
	"Add the given items to this menu, where each item is a pair (<string> <actionSelector>)..  If an element of the list is simply the symobl $-, add a line to the receiver.  The optional third element of each entry, if present, provides balloon help."

	aList do: [:tuple |
		(tuple == #-)
			ifTrue: [self addLine]
			ifFalse:
				[self add: tuple first action: tuple second.
				tuple size > 2 ifTrue:
					[self balloonTextForLastItem: tuple third]]]! !

!MenuMorph methodsFor: 'construction' stamp: 'cmm 3/9/2010 09:40'!
addMenuItem: aMenuItemMorph
	self addMorphBack: aMenuItemMorph! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 2/15/2004 16:11'!
addServices2: services for: served extraLines: linesArray

	services withIndexDo: [:service :i |
		service addServiceFor: served toMenu: self.
		self lastItem setBalloonText: service description.
		(linesArray includes: i)  ifTrue: [self addLine] ]
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 11/26/2002 13:53'!
addServices: services for: served extraLines: linesArray

	services withIndexDo: [:service :i |
		self addService: service for: served.
		submorphs last setBalloonText: service description.
		(linesArray includes: i) | service useLineAfter 
			ifTrue: [self addLine]].
! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 2/15/2004 16:19'!
addService: aService for: serviceUser
	"Append a menu item with the given service. If the item is selected, it will perform the given service."

	aService addServiceFor: serviceUser toMenu: self.! !

!MenuMorph methodsFor: 'construction' stamp: 'mt 7/9/2016 14:21'!
addStayUpIcons
	| title closeBox pinBox |
	title := submorphs
				detect: [:ea | ea hasProperty: #titleString]
				ifNone: [
					"Called too soon. Will add stay-up icons when title is added."
					self setProperty: #needsTitlebarWidgets toValue: true.
					^ self].
	closeBox := SystemWindowButton new target: self;
				 actionSelector: #delete;
				 labelGraphic: self class closeBoxImage;
				 color: Color transparent;
				 extent: self class closeBoxImage extent;
				 borderWidth: 0.
	pinBox := SystemWindowButton new target: self;
				 actionSelector: #stayUp:;
				 arguments: {true};
				 labelGraphic: self class pushPinImage;
				 color: Color transparent;
				 extent: self class pushPinImage extent;
				 borderWidth: 0.
	Preferences noviceMode ifTrue: [
		closeBox setBalloonText: 'close this menu'.
		pinBox setBalloonText: 'keep this menu up'].
		
	title
		addMorphFront: closeBox;
		addMorphBack: pinBox.
		
	self setProperty: #hasTitlebarWidgets toValue: true.
	self removeProperty: #needsTitlebarWidgets.
	self removeStayUpItems! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 4/6/2002 22:41'!
addStayUpItem
	"Append a menu item that can be used to toggle this menu's persistence."

	(self valueOfProperty: #hasTitlebarWidgets ifAbsent: [ false ])
		ifTrue: [ ^self ].
	self addStayUpIcons.! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 4/6/2002 22:41'!
addStayUpItemSpecial
	"Append a menu item that can be used to toggle this menu's persistent."

	"This variant is resistant to the MVC compatibility in #setInvokingView:"

	(self valueOfProperty: #hasTitlebarWidgets ifAbsent: [ false ])
		ifTrue: [ ^self ].
	self addStayUpIcons.! !

!MenuMorph methodsFor: 'menu' stamp: 'ar 8/4/2009 22:20'!
addTitle

	| string |
	string := UIManager default request: 'Title for this menu?'.
	string isEmpty ifTrue: [^ self].
	self addTitle: string.
! !

!MenuMorph methodsFor: 'construction' stamp: 'sw 6/19/1999 23:09'!
addTitle: aString
	"Add a title line at the top of this menu."

	self addTitle: aString updatingSelector: nil updateTarget: nil! !

!MenuMorph methodsFor: 'construction' stamp: 'dgd 9/13/2004 13:35'!
addTitle: aString icon: aForm 
	"Add a title line at the top of this menu."
	self
		addTitle: aString
		icon: aForm
		updatingSelector: nil
		updateTarget: nil ! !

!MenuMorph methodsFor: 'construction' stamp: 'mt 7/9/2016 14:37'!
addTitle: aString icon: aForm updatingSelector: aSelector updateTarget: aTarget 
	"Add a title line at the top of this menu Make aString its initial contents. If aSelector is not nil, then periodically obtain fresh values for its contents by sending aSelector to aTarget."
	
	| container fontToUse colorToUse labelArea |
	
	container := AlignmentMorph newRow.
	self setTitleParametersFor: container.
	
	aForm ifNotNil: [:form | container addMorphBack: form asMorph].

	fontToUse := self userInterfaceTheme titleFont ifNil: [TextStyle defaultFont].
	colorToUse := self userInterfaceTheme titleTextColor ifNil: [Color black].
	
	"Build the label."
	aSelector
		ifNil: [
			labelArea := AlignmentMorph newColumn
				color: Color transparent;
				vResizing: #shrinkWrap;
				wrapCentering: #center;
				cellPositioning: #topCenter.
			aString asString linesDo: [:line | labelArea 
					addMorphBack: ((StringMorph contents: line 
										font: fontToUse)
										color: colorToUse;
										yourself)]]
		ifNotNil: [
			labelArea := UpdatingStringMorph on: aTarget selector: aSelector.
			labelArea font: fontToUse.
			labelArea color: colorToUse.
			labelArea useStringFormat.
			labelArea lock].
		
	container addMorphBack: labelArea.
	""
	container setProperty: #titleString toValue: aString.
	self addMorphFront: container.
	""
	(self hasProperty: #needsTitlebarWidgets)
		ifTrue: [self addStayUpIcons]! !

!MenuMorph methodsFor: 'construction' stamp: 'dgd 4/3/2006 13:01'!
addTitle: aString updatingSelector: aSelector updateTarget: aTarget 
	"Add a title line at the top of this menu Make aString its initial  
	contents.  
	If aSelector is not nil, then periodically obtain fresh values for  
	its contents by sending aSelector to aTarget.."
	^ self
		addTitle: aString
		icon: nil
		updatingSelector: aSelector
		updateTarget: aTarget! !

!MenuMorph methodsFor: 'construction' stamp: 'nk 11/25/2003 09:59'!
addTranslatedList: aList
	"Add the given items to this menu, where each item is a pair (<string> <actionSelector>)..  If an element of the list is simply the symobl $-, add a line to the receiver.  The optional third element of each entry, if present, provides balloon help.
	The first and third items will be translated."

	aList do: [:tuple |
		(tuple == #-)
			ifTrue: [self addLine]
			ifFalse:
				[self add: tuple first translated action: tuple second.
				tuple size > 2 ifTrue:
					[self balloonTextForLastItem: tuple third translated ]]]! !

!MenuMorph methodsFor: 'construction' stamp: 'cmm 5/8/2014 10:56'!
addUpdating: aWordingSelector action: aSymbol 
	self
		addUpdating: aWordingSelector
		target: defaultTarget
		selector: aSymbol
		argumentList: Array empty! !

!MenuMorph methodsFor: 'construction' stamp: 'sw 11/6/2000 13:39'!
addUpdating: wordingSelector enablementSelector: enablementSelector target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  In this variant, the wording of the menu item is obtained by sending the wordingSelector to the target, and the optional enablementSelector determines whether or not the item should be enabled.  Answer the item itself."

	| item |
	item := UpdatingMenuItemMorph new
		target: target;
		selector: aSymbol;
		wordingProvider: target wordingSelector: wordingSelector;
		enablementSelector: enablementSelector;
		arguments: argList asArray.
	self addMorphBack: item.
	^ item
! !

!MenuMorph methodsFor: 'construction' stamp: 'cmm 5/8/2014 10:57'!
addUpdating: aWordingSelector enablement: anEnablementSelector action: aSymbol 
	self
		addUpdating: aWordingSelector
		enablementSelector: anEnablementSelector
		target: defaultTarget
		selector: aSymbol
		argumentList: Array empty! !

!MenuMorph methodsFor: 'construction' stamp: 'cmm 5/8/2014 10:57'!
addUpdating: aWordingSelector target: aTarget action: aSymbol 
	self
		addUpdating: aWordingSelector
		target: aTarget
		selector: aSymbol
		argumentList: Array empty! !

!MenuMorph methodsFor: 'construction' stamp: 'sw 11/6/2000 13:43'!
addUpdating: wordingSelector target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  In this variant, the wording of the menu item is obtained by sending the wordingSelector to the target,  Answer the item added."

	| item |
	item := UpdatingMenuItemMorph new
		target: target;
		selector: aSymbol;
		wordingProvider: target wordingSelector: wordingSelector;
		arguments: argList asArray.
	self addMorphBack: item.
	^ item
! !

!MenuMorph methodsFor: 'construction' stamp: 'sw 8/28/2000 18:01'!
addWithLabel: aLabel enablementSelector: enablementSelector target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument.  In this variant, the wording of the menu item is constant, and the optional enablementSelector determines whether or not the item should be enabled."

	| item |
	item := UpdatingMenuItemMorph new
		target: target;
		selector: aSymbol;
		contents: aLabel;
		wordingProvider: target wordingSelector: nil;
		enablementSelector: enablementSelector;
		arguments: argList asArray.
	self addMorphBack: item.
! !

!MenuMorph methodsFor: 'construction' stamp: 'cmm 5/8/2014 10:57'!
addWithLabel: aLabel enablement: anEnablementSelector action: aSymbol 
	self
		addWithLabel: aLabel
		enablementSelector: anEnablementSelector
		target: defaultTarget
		selector: aSymbol
		argumentList: Array empty! !

!MenuMorph methodsFor: 'construction' stamp: 'cmm 5/8/2014 10:57'!
add: aString action: aSymbolOrValuable 
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the default target object."
	"Details: Note that the menu item added captures the default target object at the time the item is added; the default target can later be changed before added additional items without affecting the targets of previously added entries. The model is that each entry is like a button that knows everything it needs to perform its action."
	aSymbolOrValuable isSymbol
		ifTrue:
			[ self
				add: aString
				target: defaultTarget
				selector: aSymbolOrValuable
				argumentList: Array empty ]
		ifFalse:
			[ self
				add: aString
				target: aSymbolOrValuable
				selector: #value
				argumentList: Array empty ]! !

!MenuMorph methodsFor: 'construction' stamp: 'cmm 5/8/2014 10:56'!
add: wordingString help: helpString action: aSymbol 
	"Append a menu item with the given label. If the item is  
	selected, it will send the given selector to the default target  
	object."
	"Details: Note that the menu item added captures the default  
	target object at the time the item is added; the default target  
	can later be changed before added additional items without  
	affecting the targets of previously added entries. The model is 
	that each entry is like a button that knows everything it needs 
	to perform its action."
	self
		add: wordingString
		target: defaultTarget
		selector: aSymbol
		argumentList: Array empty.
	self balloonTextForLastItem:helpString! !

!MenuMorph methodsFor: 'construction' stamp: 'dgd 9/13/2004 18:58'!
add: wordingString icon: aForm help: helpString subMenu: aMenuMorph 
	"Append the given submenu with the given label."
	| item |
	item := MenuItemMorph new.

	item contents: wordingString.
	item subMenu: aMenuMorph.
	item icon: aForm.
	helpString isNil
		ifFalse: [item setBalloonText: helpString].
	self addMorphBack: item! !

!MenuMorph methodsFor: 'construction' stamp: 'dgd 9/13/2004 14:01'!
add: wordingString icon: aForm subMenu: aMenuMorph 
	"Append the given submenu with the given label."
	^ self
		add: wordingString
		icon: aForm
		help: nil
		subMenu: aMenuMorph! !

!MenuMorph methodsFor: 'construction' stamp: 'sw 5/1/1998 00:48'!
add: aString selector: aSymbol argument: arg

	self add: aString
		target: defaultTarget
		selector: aSymbol
		argumentList: (Array with: arg)
! !

!MenuMorph methodsFor: 'construction' stamp: 'dgd 9/13/2004 14:01'!
add: aString subMenu: aMenuMorph 
	"Append the given submenu with the given label."
	self
		add: aString
		icon: nil
		subMenu: aMenuMorph! !

!MenuMorph methodsFor: 'construction' stamp: 'ar 12/16/2001 16:53'!
add: aString subMenu: aMenuMorph target: target selector: aSymbol argumentList: argList
	"Append the given submenu with the given label."

	| item |
	item := MenuItemMorph new.
	item 
		contents: aString;
		target: target;
		selector: aSymbol;
		arguments: argList asArray;
		subMenu: aMenuMorph.
	self addMorphBack: item.
	^item! !

!MenuMorph methodsFor: 'construction' stamp: 'cmm 5/8/2014 10:57'!
add: aString target: aTarget action: aSymbol 
	self
		add: aString
		target: aTarget
		selector: aSymbol
		argumentList: Array empty! !

!MenuMorph methodsFor: 'construction' stamp: 'cmm 5/8/2014 10:57'!
add: aString target: anObject selector: aSymbol 
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object."
	self
		add: aString
		target: anObject
		selector: aSymbol
		argumentList: Array empty! !

!MenuMorph methodsFor: 'construction' stamp: 'hg 8/3/2000 15:22'!
add: aString target: target selector: aSymbol argumentList: argList
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given arguments. If the selector takes one more argument than the number of arguments in the given list, then the triggering event is supplied as as the last argument."

	| item |
	item := MenuItemMorph new
		contents: aString;
		target: target;
		selector: aSymbol;
		arguments: argList asArray.
	self addMorphBack: item.
! !

!MenuMorph methodsFor: 'construction' stamp: 'jm 11/4/97 07:46'!
add: aString target: target selector: aSymbol argument: arg
	"Append a menu item with the given label. If the item is selected, it will send the given selector to the target object with the given argument."

	self add: aString
		target: target
		selector: aSymbol
		argumentList: (Array with: arg)
! !

!MenuMorph methodsFor: 'update' stamp: 'mt 7/9/2016 15:01'!
applyUserInterfaceTheme

	| colorToUse |
	super applyUserInterfaceTheme.
	
	self setDefaultParameters.
	
	"Update properties of separating lines."
	colorToUse := self userInterfaceTheme lineColor ifNil: [Color gray: 0.9].
	self submorphs
		select: [:ea | ea knownName = #line]
		thenDo: [:line |
			line
				color: colorToUse;
				height: (self userInterfaceTheme lineWidth ifNil: [2]);
				borderStyle: (self userInterfaceTheme lineStyle ifNil: [BorderStyle inset]);
				borderColor: colorToUse].! !

!MenuMorph methodsFor: 'construction' stamp: 'sw 11/5/1998 21:13'!
balloonTextForLastItem: balloonText
	submorphs last setBalloonText: balloonText! !

!MenuMorph methodsFor: 'accessing' stamp: 'sw 12/4/2001 21:22'!
commandKeyHandler
	"Answer the receiver's commandKeyHandler"

	^ self valueOfProperty: #commandKeyHandler ifAbsent: [nil]! !

!MenuMorph methodsFor: 'accessing' stamp: 'sw 12/4/2001 21:23'!
commandKeyHandler: anObject
	"Set the receiver's commandKeyHandler.  Whatever you set here needs to be prepared to respond to the message #commandKeyTypedIntoMenu: "

	self setProperty: #commandKeyHandler toValue: anObject! !

!MenuMorph methodsFor: 'events' stamp: 'kb 2/24/2010 13:14'!
deactivate: evt

	"If a stand-alone menu, just delete it"
	popUpOwner ifNil: [ 
		self delete.
		^true ].
	"If a sub-menu, then deselect, and return focus to outer menu"
	self selectItem: nil event: evt.
	evt hand newMouseFocus: popUpOwner owner.
	evt hand newKeyboardFocus: popUpOwner owner! !

!MenuMorph methodsFor: 'accessing' stamp: 'ar 9/18/2000 13:19'!
defaultTarget
	^defaultTarget! !

!MenuMorph methodsFor: 'construction' stamp: 'jm 11/4/97 07:46'!
defaultTarget: anObject
	"Set the default target for adding menu items."

	defaultTarget := anObject.
! !

!MenuMorph methodsFor: 'initialization' stamp: 'kfr 4/6/2008 21:16'!
delete
	"Delete the receiver."

	activeSubMenu ifNotNil: [activeSubMenu stayUp ifFalse: [activeSubMenu delete]].
	self isFlexed ifTrue: [^ owner delete].
	^ super delete! !

!MenuMorph methodsFor: 'control' stamp: 'hg 8/3/2000 15:28'!
deleteIfPopUp
	"Remove this menu from the screen if stayUp is not true. If it is a submenu, also remove its owning menu."

	stayUp ifFalse: [self topRendererOrSelf delete].
	(popUpOwner notNil and: [popUpOwner isKindOf: MenuItemMorph]) ifTrue: [
		popUpOwner isSelected: false.
		(popUpOwner owner isKindOf: MenuMorph)
			ifTrue: [popUpOwner owner deleteIfPopUp]].
! !

!MenuMorph methodsFor: 'control' stamp: 'ar 10/10/2000 00:54'!
deleteIfPopUp: evt
	"Remove this menu from the screen if stayUp is not true. If it is a submenu, also remove its owning menu."

	stayUp ifFalse: [self topRendererOrSelf delete].
	(popUpOwner notNil) ifTrue: [
		popUpOwner isSelected: false.
		popUpOwner deleteIfPopUp: evt].
	evt ifNotNil:[evt hand releaseMouseFocus: self].! !

!MenuMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 18:26'!
detachSubMenu: evt
	| possibleTargets item subMenu index |
	possibleTargets := self items select:[:any| any hasSubMenu].
	possibleTargets size > 0 ifTrue:[
		index := UIManager default 
				chooseFrom: (possibleTargets collect:[:t| t contents asString])
				title: 'Which menu?'.
		index = 0 ifTrue:[^self]].
	item := possibleTargets at: index.
	subMenu := item subMenu.
	subMenu ifNotNil: [
		item subMenu: nil.
		item delete.
		subMenu stayUp: true.
		subMenu popUpOwner: nil.
		subMenu addTitle: item contents.
		evt hand attachMorph: subMenu].
! !

!MenuMorph methodsFor: 'keyboard control' stamp: 'kb 2/4/2010 00:32'!
displayFiltered: evt
	| matchStr allItems matches feedbackMorph |
	matchStr := self valueOfProperty: #matchString.
	allItems := self submorphs select: [:m | m isKindOf: MenuItemMorph].
	matches :=  allItems select: [:m | | isMatch | 
		isMatch := 
			matchStr isEmpty or: [
				m contents includesSubstring: matchStr caseSensitive: false].
		m isEnabled: isMatch.
		isMatch].
	feedbackMorph := self valueOfProperty: #feedbackMorph.
	feedbackMorph ifNil: [
		feedbackMorph := 
			TextMorph new 
				autoFit: true;
				color: Color darkGray.
		self
			addLine;
			addMorphBack: feedbackMorph lock.
		self setProperty: #feedbackMorph toValue: feedbackMorph.
		self fullBounds.  "Lay out for submorph adjacency"].
	feedbackMorph contents: '<', matchStr, '>'.
	matchStr isEmpty ifTrue: [
		feedbackMorph delete.
		self submorphs last delete.
		self removeProperty: #feedbackMorph].
	" This method is invoked with evt = nil from MenuMorph >> removeMatchString. 
	The current implementation can't select an item without an event. "
	(evt notNil and: [ matches size >= 1 ]) ifTrue: [
		self selectItem: matches first event: evt]! !

!MenuMorph methodsFor: 'menu' stamp: 'ul 12/12/2009 14:09'!
doButtonAction
	"Do the receiver's inherent button action.  Makes sense for the kind of MenuMorph that is a wrapper for a single menu-item -- pass it on the the item"

	(self findA: MenuItemMorph) ifNotNil: [:aMenuItem | aMenuItem doButtonAction]! !

!MenuMorph methodsFor: 'drawing' stamp: 'mt 4/10/2015 17:05'!
drawKeyboardFocusIndicationOn: aCanvas 
	"Draw the menu. Add keyboard-focus feedback if appropriate"

	(self rootMenu hasProperty: #hasUsedKeyboard)
		ifTrue: [ 
			aCanvas
				frameRectangle: self innerBounds
				width: Preferences menuBorderWidth
				color: Preferences keyboardFocusColor].! !

!MenuMorph methodsFor: 'keyboard control' stamp: 'rr 3/24/2004 13:45'!
filterListWith: char 
	| matchString |
	matchString := self valueOfProperty: #matchString ifAbsentPut: [String new].
	matchString := char = Character backspace 
				ifTrue: 
					[matchString isEmpty ifTrue: [matchString] ifFalse: [matchString allButLast]]
				ifFalse: [matchString copyWith: char].
	self setProperty: #matchString toValue: matchString! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/22/2010 15:54'!
handleCommandKeyPress: evt

	(evt commandKeyPressed and: [
		self commandKeyHandler notNil ]) ifTrue: [
			self commandKeyHandler commandKeyTypedIntoMenu: evt.
			self deleteIfPopUp: evt.
			^true ].
	^false! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/22/2010 16:05'!
handleCRStroke: evt

	| selectable |
	evt keyValue = 13 ifFalse: [ ^false ].
	selectedItem ifNotNil: [
		selectedItem hasSubMenu 
			ifTrue: [
				evt hand 
					newMouseFocus: selectedItem subMenu;
					newKeyboardFocus: selectedItem subMenu ]
			ifFalse:  [
				selectedItem invokeWithEvent: evt ].
		^true ].
	(selectable := self items) size = 1 ifTrue: [ 
		selectable first invokeWithEvent: evt ].
	^true! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/22/2010 16:31'!
handleDownStroke: evt

	evt keyValue = 31 ifFalse: [ ^false ].
	self moveSelectionDown: 1 event: evt.
	^true! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/24/2010 13:15'!
handleEscStroke: evt

	evt keyValue = 27 ifFalse: [ ^false ].
	self 
		valueOfProperty: #matchString
		ifPresentDo: [ :str | 
			str isEmpty ifFalse: [ "If filtered, first ESC removes filter"
				self setProperty: #matchString toValue: String new.
				self selectItem: nil event: evt.
				self displayFiltered: evt.
				^true ] ].
	self deactivate: evt.
	^true! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'cmm 1/11/2011 21:08'!
handleFiltering: evt

	| matchString |
	matchString := self valueOfProperty: #matchString ifAbsentPut: [ String new ].
	matchString := evt keyValue = 8 " Character backspace asciiValue "
		ifTrue: [
			matchString isEmpty 
				ifTrue: [ matchString ] 
				ifFalse: [ matchString allButLast ] ]
		ifFalse: [
			matchString copyWith: evt keyCharacter ].
	self setProperty: #matchString toValue: matchString.
	self displayFiltered: evt! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/24/2010 10:41'!
handleLeftStroke: evt

	28 = evt keyValue ifFalse: [ ^false ].
	self stepIntoSubmenu: evt.
	^true! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/22/2010 16:30'!
handlePageDownStorke: evt

	evt keyValue = 12 ifFalse: [ ^false ].
	self moveSelectionDown: 5 event: evt.
	^true! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/22/2010 16:35'!
handlePageDownStroke: evt

	evt keyValue = 12 ifFalse: [ ^false ].
	self moveSelectionDown: 5 event: evt.
	^true! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/22/2010 16:34'!
handlePageUpStroke: evt

	evt keyValue = 11 ifFalse: [ ^false ].
	self moveSelectionDown: -5 event: evt.
	^true! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/24/2010 10:48'!
handleRightStroke: evt

	29 = evt keyValue ifFalse: [ ^false ].
	self stepIntoSubmenu: evt.
	^true! !

!MenuMorph methodsFor: 'keyboard control' stamp: 'sw 12/4/2001 20:13'!
handlesKeyboard: evt
	"Answer whether the receiver handles the keystroke represented by the event"

	^ evt anyModifierKeyPressed not or: [evt commandKeyPressed and: [self commandKeyHandler notNil]]! !

!MenuMorph methodsFor: 'events' stamp: 'ar 9/18/2000 10:13'!
handlesMouseDown: evt
	^true! !

!MenuMorph methodsFor: 'events' stamp: 'mt 6/25/2016 08:37'!
handlesMouseOver: evt
	^ true! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/22/2010 16:12'!
handleUpStorke: evt

	evt keyValue = 30 ifFalse: [ ^false ].
	self moveSelectionDown: -1 event: evt.
	^true! !

!MenuMorph methodsFor: 'accessing' stamp: 'dgd 9/13/2004 13:36'!
hasItems
	"Answer if the receiver has menu items"
	^ submorphs
		anySatisfy: [:each | each isKindOf: MenuItemMorph] ! !

!MenuMorph methodsFor: 'accessing' stamp: 'ar 9/18/2000 11:18'!
hasSubMenu: aMenuMorph
	self items do: [:each | (each hasSubMenu: aMenuMorph) ifTrue:[^true]].
	^ false
! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'cmm 1/11/2011 21:06'!
hideKeyboardHelp
	self deleteBalloon! !

!MenuMorph methodsFor: 'testing' stamp: 'mt 4/10/2015 16:49'!
indicateKeyboardFocus

	^ true! !

!MenuMorph methodsFor: 'modal control' stamp: 'ar 8/6/2009 18:59'!
informUserAt: aPoint during: aBlock
	"Add this menu to the Morphic world during the execution of the given block."
	| title w |
	title := self allMorphs detect: [ :ea | ea hasProperty: #titleString ].
	title := title submorphs first.
	self visible: false.
	w := ActiveWorld.
	aBlock value:[:string|
		self visible ifFalse:[
			w addMorph: self centeredNear: aPoint.
			self visible: true].
		title contents: string.
		self setConstrainedPosition: Sensor cursorPoint hangOut: false.
		self changed.
		w displayWorld		 "show myself"
	]. 
	self delete.
	w displayWorld! !

!MenuMorph methodsFor: 'initialization' stamp: 'mt 7/9/2016 13:38'!
initialize
	super initialize.

	bounds := 0 @ 0 corner: 40 @ 10.

	self setDefaultParameters.

	self changeTableLayout.
	self listDirection: #topToBottom.
	self hResizing: #shrinkWrap.
	self vResizing: #shrinkWrap.
	defaultTarget := nil.
	selectedItem := nil.
	stayUp := false.
	popUpOwner := nil.! !

!MenuMorph methodsFor: 'modal control' stamp: 'ar 8/4/2009 22:39'!
invokeAt: aPoint in: aWorld allowKeyboard: aBoolean
	"Add this menu to the given world centered at the given point. Wait for the user to make a selection and answer it. The selection value returned is an integer in keeping with PopUpMenu, if the menu is converted from an MVC-style menu."
	"Details: This is invoked synchronously from the caller. In order to keep processing inputs and updating the screen while waiting for the user to respond, this method has its own version of the World's event loop." 
	| w originalFocusHolder |
	originalFocusHolder := aWorld primaryHand keyboardFocus.
	self popUpAt: aPoint forHand: aWorld primaryHand in: aWorld allowKeyboard: aBoolean.
	w := aWorld outermostWorldMorph. "containing hand"
	[self isInWorld] whileTrue: [w doOneSubCycle].
	self delete.
	originalFocusHolder ifNotNil: [aWorld primaryHand newKeyboardFocus: originalFocusHolder].
! !

!MenuMorph methodsFor: 'private' stamp: 'ar 10/7/2000 21:08'!
invokeMetaMenu: evt
	stayUp ifFalse:[^self]. "Don't allow this"
	^super invokeMetaMenu: evt! !

!MenuMorph methodsFor: 'modal control' stamp: 'sw 2/3/2002 14:26'!
invokeModal
	"Invoke this menu and don't return until the user has chosen a value.
	See example below on how to use modal menu morphs."
	^ self invokeModal: Preferences menuKeyboardControl

	"Example:
	| menu sub entry |
	menu := MenuMorph new.
	1 to: 3 do: [:i |
		entry := 'Line', i printString.
		sub := MenuMorph new.
		menu add: entry subMenu: sub.
		#('Item A' 'Item B' 'Item C')  do:[:subEntry|
			sub add: subEntry target: menu 
				selector: #modalSelection: argument: {entry. subEntry}]].
	menu invokeModal.
"

! !

!MenuMorph methodsFor: 'modal control' stamp: 'ar 8/6/2009 20:32'!
invokeModalAt: aPoint in: aWorld
	"Add this menu to the given world centered at the given point. Wait for the user to make a selection and answer it. The selection value returned is an integer in keeping with PopUpMenu, if the menu is converted from an MVC-style menu."
	"Details: This is invoked synchronously from the caller. In order to keep processing inputs and updating the screen while waiting for the user to respond, this method has its own version of the World's event loop."

	^ self invokeModalAt: aPoint in: aWorld allowKeyboard: Preferences menuKeyboardControl! !

!MenuMorph methodsFor: 'modal control' stamp: 'KLC 4/11/2004 09:06'!
invokeModalAt: aPoint in: aWorld allowKeyboard: aBoolean
	"Invoke this menu and don't return until the user has chosen a value.
	See senders of this method for finding out how to use modal menu morphs."
	| w originalFocusHolder |
	originalFocusHolder := aWorld primaryHand keyboardFocus.
	self popUpAt: aPoint forHand: aWorld primaryHand in: aWorld allowKeyboard: aBoolean.
	self isModalInvokationDone: false.
	w := aWorld outermostWorldMorph. "containing hand"
	[self isInWorld & self isModalInvokationDone not] whileTrue: [w doOneSubCycle].
	self delete.
	originalFocusHolder ifNotNil: [aWorld primaryHand newKeyboardFocus: originalFocusHolder].
	^ self modalSelection! !

!MenuMorph methodsFor: 'modal control' stamp: 'sw 2/3/2002 14:26'!
invokeModal: allowKeyboardControl
	"Invoke this menu and don't return until the user has chosen a value.  If the allowKeyboarControl boolean is true, permit keyboard control of the menu"

	^ self invokeModalAt: ActiveHand position in: ActiveWorld allowKeyboard: allowKeyboardControl! !

!MenuMorph methodsFor: 'modal control' stamp: 'ar 1/5/2002 21:33'!
isModalInvokationDone
	^self valueOfProperty: #isModalInvokationDone ifAbsent:[false]! !

!MenuMorph methodsFor: 'modal control' stamp: 'ar 1/5/2002 21:34'!
isModalInvokationDone: aBool
	self setProperty: #isModalInvokationDone toValue: aBool
! !

!MenuMorph methodsFor: 'accessing' stamp: 'hg 8/3/2000 15:29'!
items

	^ submorphs select: [:m | m isKindOf: MenuItemMorph]
! !

!MenuMorph methodsFor: 'accessing' stamp: 'nice 12/27/2009 03:11'!
itemWithWording: wording
	"If any of the receiver's items or submenu items have the given wording (case-blind comparison done), then return it, else return nil."
	
	self items do:[:anItem | | found |
		found := anItem itemWithWording: wording.
		found ifNotNil:[^found]].
	^ nil! !

!MenuMorph methodsFor: 'dropping/grabbing' stamp: 'mt 5/8/2015 16:09'!
justDroppedInto: aMorph event: evt
	| halo |
	super justDroppedInto: aMorph event: evt.
	halo := evt hand halo.
	(halo notNil and:[halo target hasOwner: self]) ifTrue:[
		"Grabbed single menu item"
		self addHalo: evt.
	].
	stayUp ifFalse:[evt hand newMouseFocus: self].
	self hasDropShadow: Preferences menuAppearance3d.! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/24/2010 10:49'!
keyStrokeHandlers

	^#(
		handleCommandKeyPress:
		handleCRStroke:
		handleEscStroke:
		handleLeftStroke:
		handleRightStroke:
		handleUpStorke:
		handleDownStroke:
		handlePageUpStroke:
		handlePageDownStroke:)! !

!MenuMorph methodsFor: 'keyboard control' stamp: 'cmm 1/11/2011 21:07'!
keyStroke: evt 
	self hideKeyboardHelp; noteRootMenuHasUsedKeyboard.
	self keyStrokeHandlers
		detect: [:each | self perform: each with: evt]
		ifNone: [self handleFiltering: evt]! !

!MenuMorph methodsFor: 'construction' stamp: 'nice 6/11/2010 21:52'!
labels: labelList lines: linesArray selections: selectionsArray 
	"This method allows the receiver to accept old-style SelectionMenu creation messages. It should be used only for backward compatibility during the MVC-to-Morphic transition. New code should be written using the other menu construction protocol such as addList:."

	"Labels can be either a sting with embedded crs, or a collection of strings."

	| labelArray |
	labelArray := (labelList isString) 
				ifTrue: [labelList lines]
				ifFalse: [labelList]. 
	1 to: labelArray size
		do: 
			[:i | 
			self add: (labelArray at: i) action: (selectionsArray at: i).
			(linesArray includes: i) ifTrue: [self addLine]]! !

!MenuMorph methodsFor: 'accessing' stamp: 'ul 3/21/2010 17:48'!
lastItem
	
	submorphs reverseDo: [ :each |
		(each isKindOf: MenuItemMorph) ifTrue: [ ^each ] ].
	^submorphs last! !

!MenuMorph methodsFor: 'accessing' stamp: 'nice 8/27/2010 22:00'!
lastSelection
	"Return the label of the last selected item or nil."

	^selectedItem ifNotNil: [selectedItem selector]! !

!MenuMorph methodsFor: 'modal control' stamp: 'ar 1/5/2002 21:34'!
modalSelection
	^self valueOfProperty: #modalSelection ifAbsent:[nil]! !

!MenuMorph methodsFor: 'modal control' stamp: 'ar 1/5/2002 21:34'!
modalSelection: anObject
	self setProperty: #modalSelection toValue: anObject.
	self isModalInvokationDone: true! !

!MenuMorph methodsFor: 'private' stamp: 'ar 2/10/2001 00:37'!
morphicLayerNumber

	"helpful for insuring some morphs always appear in front of or behind others.
	smaller numbers are in front"
	^self valueOfProperty: #morphicLayerNumber  ifAbsent: [
		stayUp ifTrue:[100] ifFalse:[10]
	]! !

!MenuMorph methodsFor: 'events' stamp: 'mt 5/8/2015 16:08'!
mouseDown: evt
	"Handle a mouse down event."
	(stayUp or:[self fullContainsPoint: evt position]) 
		ifFalse:[^self deleteIfPopUp: evt]. "click outside"
	self isSticky ifTrue: [^self].
	
	"Grab the menu and drag it to some other place"
	self hasDropShadow: false.
	evt hand grabMorph: self.! !

!MenuMorph methodsFor: 'events' stamp: 'mt 6/25/2016 08:38'!
mouseLeave: evt
	self selectItem: nil event: evt.! !

!MenuMorph methodsFor: 'events' stamp: 'mt 6/21/2016 14:08'!
mouseMove: evt
	" If the mouse moves over an item not selected, we try to set it as selected.
	If this happens depends on that the current selected item wants to release
	its selection. "

	self selectedItem ifNil: [
		self 
			selectItem: (
				self items 
					detect: [ :each | each containsPoint: evt position ] 
					ifNone: [ nil ])
			event: evt].

	"Transfer control to *either* the currently active submenu or the pop up owner, if any. Since the active sub menu is always displayed upfront check it first."	
	self selectedItem ifNotNil: [
		(self selectedItem activateSubmenu: evt) 
			ifTrue: [ ^self ]
			ifFalse: [ 
				(self containsPoint: evt position) ifFalse: [ 
					self selectItem: nil event: evt ] ] ].

	"Note: The following does not traverse upwards but it's the best I can do for now"
	popUpOwner ifNotNil:[(popUpOwner activateOwnerMenu: evt) ifTrue: [^ self]].
! !

!MenuMorph methodsFor: 'events' stamp: 'mt 6/21/2016 13:19'!
mouseUp: evt
	"Handle a mouse up event.
	Note: This might be sent from a modal shell."
	(self fullContainsPoint: evt position) ifFalse:[
		^ self deleteIfPopUp: evt].
	stayUp ifFalse:[
		"Still in pop-up transition; keep focus"
		evt hand newMouseFocus: self].! !

!MenuMorph methodsFor: 'keyboard control' stamp: 'nice 12/27/2009 20:09'!
moveSelectionDown: direction event: evt
	"Move the current selection up or down by one, presumably under keyboard control.
	direction = +/-1"

	| index |
	index := (submorphs indexOf: selectedItem ifAbsent: [1-direction]) + direction.
	submorphs do: "Ensure finite"
		[:unused | | m |
		m := submorphs atWrap: index.
		((m isKindOf: MenuItemMorph) and: [m isEnabled]) ifTrue:
			[^ self selectItem: m event: evt].
		"Keep looking for an enabled item"
		index := index + direction sign].
	^ self selectItem: nil event: evt! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/22/2010 15:53'!
noteRootMenuHasUsedKeyboard

	(self rootMenu hasProperty: #hasUsedKeyboard) ifFalse: [
		self setProperty: #hasUsedKeyboard toValue: true.
		self changed ].! !

!MenuMorph methodsFor: 'control' stamp: 'tpr 12/18/2015 16:10'!
popUpAdjacentTo: rightOrLeftPoint forHand: hand from: sourceItem 
	"Present this menu at the given point under control of the given hand."
	
	| tryToPlace selectedOffset rightPoint leftPoint |
	hand world startSteppingSubmorphsOf: self.
	popUpOwner := sourceItem.
	
	self fullBounds.
	self updateColor.

	"ensure layout is current"
	selectedOffset := (selectedItem
				ifNil: [self items first]) position - self position.
	tryToPlace := [:where :mustFit | | delta | 
			self position: where - selectedOffset.
			delta := self boundsInWorld amountToTranslateWithin: sourceItem worldBounds.
			(delta x = 0
					or: [mustFit])
				ifTrue: [delta = (0 @ 0)
						ifFalse: [self position: self position + delta].
					sourceItem world addMorphFront: self.
					^ self]].
	rightPoint := rightOrLeftPoint first + ((self layoutInset + self borderWidth) @ 0).
	leftPoint := rightOrLeftPoint last + ((self layoutInset + self borderWidth - self width) @ 0).
	tryToPlace
		value: rightPoint value: false;
		 value: leftPoint value: false;
		 value: rightPoint value: true.! !

!MenuMorph methodsFor: 'control' stamp: 'ar 12/27/2001 22:46'!
popUpAt: aPoint forHand: hand in: aWorld
	"Present this menu at the given point under control of the given hand.  Allow keyboard input into the menu."

	^ self popUpAt: aPoint forHand: hand in: aWorld allowKeyboard: Preferences menuKeyboardControl! !

!MenuMorph methodsFor: 'control' stamp: 'cmm 1/11/2011 21:04'!
popUpAt: aPoint forHand: hand in: aWorld allowKeyboard: aBoolean 
	"Present this menu at the given point under control of the given 
	hand."
	| evt |
	aWorld submorphs
		select: [:each | (each isKindOf: MenuMorph)
				and: [each stayUp not]]
		thenCollect: [:menu | menu delete].
	self items isEmpty
		ifTrue: [^ self].
	MenuIcons decorateMenu: self.
	(self submorphs
		select: [:m | m isKindOf: UpdatingMenuItemMorph])
		do: [:m | m updateContents].
	"precompute width"
	self
		positionAt: aPoint
		relativeTo: (selectedItem
				ifNil: [self items first])
		inWorld: aWorld.
	aWorld addMorphFront: self.
	"Acquire focus for valid pop up behavior"
	hand newMouseFocus: self.
	aBoolean
		ifTrue: [hand newKeyboardFocus: self.
			self showKeyboardHelp].
	evt := hand lastEvent.
	(evt isKeyboard
			or: [evt isMouse
					and: [evt anyButtonPressed not]])
		ifTrue: ["Select first item if button not down"
			self moveSelectionDown: 1 event: evt
			"Select first item if button not down"].
	self updateColor.
	self changed! !

!MenuMorph methodsFor: 'control' stamp: 'sw 4/24/2001 11:11'!
popUpEvent: evt in: aWorld
	"Present this menu in response to the given event."

	| aHand aPosition |
	aHand := evt ifNotNil: [evt hand] ifNil: [ActiveHand].
	aPosition := aHand position truncated.
	^ self popUpAt: aPosition forHand: aHand in: aWorld
! !

!MenuMorph methodsFor: 'control' stamp: 'ar 3/18/2001 00:33'!
popUpForHand: hand in: aWorld
	| p |
	"Present this menu under control of the given hand."

	p := hand position truncated.
	^self popUpAt: p forHand: hand in: aWorld
! !

!MenuMorph methodsFor: 'control' stamp: 'sw 2/18/2001 00:52'!
popUpInWorld
	"Present this menu in the current World"

	^ self popUpInWorld: self currentWorld! !

!MenuMorph methodsFor: 'control' stamp: 'ar 10/5/2000 19:31'!
popUpInWorld: aWorld
	"Present this menu under control of the given hand."
	^self popUpAt: aWorld primaryHand position forHand: aWorld primaryHand in: aWorld
! !

!MenuMorph methodsFor: 'control' stamp: 'sw 12/17/2001 16:43'!
popUpNoKeyboard
	"Present this menu in the current World, *not* allowing keyboard input into the menu"

	^ self popUpAt: ActiveHand position forHand: ActiveHand in: ActiveWorld allowKeyboard: false! !

!MenuMorph methodsFor: 'accessing' stamp: 'ar 9/18/2000 10:06'!
popUpOwner
	"Return the current pop-up owner that is the menu item that automatically initiated the receiver."
	^ popUpOwner
! !

!MenuMorph methodsFor: 'accessing' stamp: 'ar 9/18/2000 10:07'!
popUpOwner: aMenuItemMorph
	"Set the current pop-up owner"
	popUpOwner := aMenuItemMorph.
! !

!MenuMorph methodsFor: 'private' stamp: 'sw 5/1/2002 01:39'!
positionAt: aPoint relativeTo: aMenuItem inWorld: aWorld
	"Note: items may not be laid out yet (I found them all to be at 0@0),  
	so we have to add up heights of items above the selected item."

	| i yOffset sub delta |	
	self fullBounds. "force layout"
	i := 0.
	yOffset := 0.
	[(sub := self submorphs at: (i := i + 1)) == aMenuItem]
		whileFalse: [yOffset := yOffset + sub height].

	self position: aPoint - (2 @ (yOffset + 8)).

	"If it doesn't fit, show it to the left, not to the right of the hand."
	self right > aWorld worldBounds right
		ifTrue:
			[self right: aPoint x + 1].

	"Make sure that the menu fits in the world."
	delta := self bounds amountToTranslateWithin:
		(aWorld worldBounds withHeight: ((aWorld worldBounds height - 18) max: (ActiveHand position y) + 1)).
	delta = (0 @ 0) ifFalse: [self position: self position + delta]! !

!MenuMorph methodsFor: 'events' stamp: 'mt 6/20/2016 15:51'!
processFocusEvent: evt using: dispatcher

	^ dispatcher dispatchFocusEventFully: evt with: self! !

!MenuMorph methodsFor: 'keyboard control' stamp: 'dgd 9/9/2004 21:48'!
removeMatchString
	"Remove the matchString, if any."
	self setProperty: #matchString toValue: String new.
	self displayFiltered: nil! !

!MenuMorph methodsFor: 'menu' stamp: 'gm 2/22/2003 13:10'!
removeStayUpBox
	| box |
	submorphs isEmpty ifTrue: [^self].
	(submorphs first isAlignmentMorph) ifFalse: [^self].
	box := submorphs first submorphs last.
	(box isKindOf: IconicButton) 
		ifTrue: 
			[box
				labelGraphic: (Form extent: box extent depth: 8);
				shedSelvedge;
				borderWidth: 0;
				lock]! !

!MenuMorph methodsFor: 'menu' stamp: 'nk 3/31/2002 18:36'!
removeStayUpItems
	| stayUpItems |
	stayUpItems := self items select: [ :item | item isStayUpItem ].
	stayUpItems do: [ :ea | ea delete ].
! !

!MenuMorph methodsFor: 'accessing' stamp: 'di 12/10/2001 22:11'!
rootMenu
	popUpOwner ifNil: [^ self].
	popUpOwner owner ifNil: [^ self].
	^ popUpOwner owner rootMenu! !

!MenuMorph methodsFor: 'keyboard control' stamp: 'rr 3/25/2004 12:14'!
selectCurrentItem: evt 
	| selectable |
	selectedItem ifNotNil: 
			[selectedItem hasSubMenu 
				ifTrue: [self selectSubMenu: evt]
				ifFalse: [selectedItem invokeWithEvent: evt]].
	(selectable := self items) size = 1 
		ifTrue: [selectable first invokeWithEvent: evt]! !

!MenuMorph methodsFor: 'private' stamp: 'ar 9/18/2000 12:12'!
selectedItem
	^selectedItem! !

!MenuMorph methodsFor: 'control' stamp: 'mt 6/25/2016 08:27'!
selectItem: aMenuItem event: anEvent
	"Change the selected item." 
	
	selectedItem ifNotNil:[selectedItem deselect: anEvent].
	selectedItem := aMenuItem.
	selectedItem ifNotNil:[selectedItem select: anEvent].! !

!MenuMorph methodsFor: 'keyboard control' stamp: 'rr 9/15/2005 16:24'!
selectMoreItem: evt
	| allItems more |
	allItems := self submorphs select: [:m | m isKindOf: MenuItemMorph].
	more := allItems detect: [:m | (m contents size >= 4) and: [(m contents first: 4) asString = 'more'.]] ifNone: [^ self flash].
	self selectItem: more event: evt.
	selectedItem invokeWithEvent: evt! !

!MenuMorph methodsFor: 'initialization' stamp: 'mt 7/9/2016 13:40'!
setDefaultParameters
	"change the receiver's appareance parameters"

	self
		color: (self userInterfaceTheme color ifNil: [Color r: 0.9 g: 0.9 b: 0.9]);
		borderStyle: (self userInterfaceTheme borderStyle ifNil: [BorderStyle default]);
		borderColor: (self userInterfaceTheme borderColor ifNil: [Color gray]);
		borderWidth: (self userInterfaceTheme borderWidth ifNil: [1]).

	Preferences menuAppearance3d ifTrue: [self addDropShadow].
	
	self layoutInset: 3.
! !

!MenuMorph methodsFor: 'menu' stamp: 'efo 3/27/2003 23:32'!
setInvokingView: invokingView
	"Re-work every menu item of the form
		<target> perform: <selector>
	to the form
		<target> perform: <selector> orSendTo: <invokingView>.
	This supports MVC's vectoring of non-model messages to the editPane."
	self items do:
		[:item |
		item hasSubMenu 
			ifTrue: [ item subMenu setInvokingView: invokingView]
			ifFalse: [ item arguments isEmpty ifTrue:  "only the simple messages"
						[item arguments: (Array with: item selector with: invokingView).
						item selector: #perform:orSendTo:]]]! !

!MenuMorph methodsFor: 'menu' stamp: 'wiz 1/16/2006 21:40'!
setTarget: evt 
	"Set the default target object to be used for add item commands, and re-target all existing items to the new target or the the invoking hand."

	| oldDefaultTarget |
	oldDefaultTarget := defaultTarget .
	self sightTargets: evt. 
	oldDefaultTarget ~~ defaultTarget 
		ifTrue: [self updateItemsWithTarget: defaultTarget orWithHand: evt hand ].
	! !

!MenuMorph methodsFor: 'initialization' stamp: 'mt 7/25/2016 14:08'!
setTitleParameters
		
	 self setTitleParametersFor: (self allMorphs
			detect: [:each | each hasProperty: #titleString]
			ifNone: [^ self]).! !

!MenuMorph methodsFor: 'initialization' stamp: 'mt 7/9/2016 14:44'!
setTitleParametersFor: aMenuTitle 

	aMenuTitle
		color: (self userInterfaceTheme titleColor ifNil: [Color transparent]);
		borderStyle: (self userInterfaceTheme titleBorderStyle ifNil: [BorderStyle default]);
		borderColor: (self userInterfaceTheme titleBorderColor ifNil: [Color r: 0.6 g: 0.7 b: 1]);
		borderWidth: (self userInterfaceTheme titleBorderWidth ifNil: [0]);
		cornerStyle: (self wantsRoundedCorners ifTrue: [#rounded] ifFalse: [#square]);
		vResizing: #shrinkWrap;
		wrapCentering: #center;
		cellPositioning: #center;
		cellInset: 5;
		layoutInset: (5@0 corner: 5@0).! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'mt 3/31/2015 11:31'!
showKeyboardHelp

	| help |
	help := self balloonMorphClass 
		string: 'Enter text to narrow selection\down to matching items ' withCRs
		for: self 
		corner: #topLeft.
	help popUpAt: self topCenter forHand: self activeHand! !

!MenuMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
stayUp

	^ stayUp
! !

!MenuMorph methodsFor: 'accessing' stamp: 'nk 3/31/2002 15:13'!
stayUp: aBoolean

	stayUp := aBoolean.
	aBoolean ifTrue: [ self removeStayUpBox ].! !

!MenuMorph methodsFor: 'keystroke helpers' stamp: 'kb 2/24/2010 10:39'!
stepIntoSubmenu: evt

	(selectedItem notNil and: [ selectedItem hasSubMenu ]) ifTrue: [
		evt hand newMouseFocus: selectedItem subMenu.
		evt hand newKeyboardFocus: selectedItem subMenu.
		selectedItem subMenu moveSelectionDown: 1 event: evt.
		^true ].
	^false! !

!MenuMorph methodsFor: 'menu' stamp: 'wiz 1/16/2006 21:26'!
target: aMorph
"Set defaultTarget since thats what we got.
For the sake of targetSighting which assumes #target is a word we know."

defaultTarget := aMorph! !

!MenuMorph methodsFor: 'construction' stamp: 'sw 7/1/1999 22:21'!
title: aString
	"Add a title line at the top of this menu."

	self addTitle: aString! !

!MenuMorph methodsFor: 'menu' stamp: 'RAA 1/19/2001 15:10'!
toggleStayUpIgnore: ignored evt: evt

	"This variant is resistant to the MVC compatibility in #setInvokingView:"

	self toggleStayUp: evt.
! !

!MenuMorph methodsFor: 'menu' stamp: 'RAA 1/18/2001 18:21'!
toggleStayUp: evt
	"Toggle my 'stayUp' flag and adjust the menu item to reflect its new state."

	self items do: [:item |
		item isStayUpItem ifTrue:
			[self stayUp: stayUp not.	
			 stayUp
				ifTrue: [item contents: 'dismiss this menu']
				ifFalse: [item contents: 'keep this menu up']]].
	evt hand releaseMouseFocus: self.
	stayUp ifFalse: [self topRendererOrSelf delete].
! !

!MenuMorph methodsFor: 'dropping/grabbing' stamp: 'ar 10/5/2000 18:16'!
undoGrabCommand
	^nil! !

!MenuMorph methodsFor: 'keyboard control' stamp: 'md 1/19/2006 18:02'!
unfilterOrEscape: evt 
	self valueOfProperty: #matchString
		ifPresentDo: 
			[:str | 
			
			str isEmpty 
				ifFalse: 
					["If filtered, first ESC removes filter"

					self setProperty: #matchString toValue: String new.
					self selectItem: nil event: evt.
					self displayFiltered: evt]].
	"If a stand-alone menu, just delete it"
	popUpOwner ifNil: [^self delete].
	"If a sub-menu, then deselect, and return focus to outer menu"
	self selectSuperMenu: evt! !

!MenuMorph methodsFor: 'control' stamp: 'mt 7/26/2016 11:49'!
updateColor
	| fill title cc |
	self class gradientMenu
		ifFalse: [^ self].
	(self fillStyle == self color) not 
		ifTrue: [^ self]. "Don't apply the gradient more than once"
	""
	title := self allMorphs
				detect: [:each | each hasProperty: #titleString]
				ifNone: [].
	
	cc := self color adjustSaturation: -0.08 brightness: 0.4.
	fill := GradientFillStyle ramp: {
		0.0 -> (title ifNil: [cc] ifNotNil: [cc muchLighter]). 
		0.25 -> (self color mixed: 0.5 with: cc). 
		1.0 -> self color}.
	""
	fill
		radial: false;
		origin: self topLeft;
		direction: 0 @ self height.
	""
	" 
	update the title color"

	title ifNotNil: [fill direction: 0@ title height].
	self fillStyle: fill.
	title ifNil: [^ self].
	""
	fill := GradientFillStyle ramp: {
			0.0 -> title color twiceLighter.
			1 -> title color twiceDarker}.
	""
	fill
		origin: title topLeft;
		direction: title width @ 0.
	""
	title fillStyle: fill! !

!MenuMorph methodsFor: 'menu' stamp: 'wiz 3/14/2006 23:40'!
updateItemsWithTarget: aTarget orWithHand: aHand
	"re-target all existing items"
	self items do: 
			[:item | item target ifNotNil: [
			item target isHandMorph 
				ifTrue: [item target: aHand]
				ifFalse: [item target: aTarget] ] ]! !

!MenuMorph methodsFor: 'update' stamp: 'kb 11/20/2009 13:38'!
updateMenu
	" Do nothing "! !

!MenuMorph methodsFor: 'copying' stamp: 'ar 9/18/2000 09:34'!
veryDeepFixupWith: deepCopier
	"If fields were weakly copied, fix them here.  If they were in the tree being copied, fix them up, otherwise point to the originals."

super veryDeepFixupWith: deepCopier.
defaultTarget := deepCopier references at: defaultTarget ifAbsent: [defaultTarget].
popUpOwner := deepCopier references at: popUpOwner ifAbsent: [popUpOwner].
activeSubMenu := deepCopier references at: activeSubMenu ifAbsent:[activeSubMenu].! !

!MenuMorph methodsFor: 'copying' stamp: 'kb 2/22/2010 10:15'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all, but shared.  	Warning!!!!  Every instance variable defined in this class must be handled.  We must also implement veryDeepFixupWith:.  See DeepCopier class comment."

	super veryDeepInner: deepCopier.
	"defaultTarget := defaultTarget.		Weakly copied"
	selectedItem := selectedItem veryDeepCopyWith: deepCopier.
	stayUp := stayUp veryDeepCopyWith: deepCopier.
	popUpOwner := popUpOwner.		"Weakly copied"
	activeSubMenu := activeSubMenu. "Weakly copied"
! !

!MenuMorph methodsFor: 'events' stamp: 'mt 6/20/2016 16:02'!
wantsEveryMouseMove
	^ true! !

!MenuMorph methodsFor: 'rounding' stamp: 'mt 2/27/2015 13:51'!
wantsRoundedCorners

	^ self class roundedMenuCorners or: [super wantsRoundedCorners]! !

!MenuMorph methodsFor: 'control' stamp: 'ul 11/15/2010 11:53'!
wantsToBeDroppedInto: aMorph
	"Return true if it's okay to drop the receiver into aMorph.  A single-item MenuMorph is in effect a button rather than a menu, and as such should not be reluctant to be dropped into another object."

	^ (aMorph isWorldMorph or: [submorphs size = 1]) or:
		[Preferences systemWindowEmbedOK]! !

MenuMorph subclass: #MVCMenuMorph
	instanceVariableNames: 'done mvcSelection'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Menus'!

!MVCMenuMorph commentStamp: '<historical>' prior: 0!
I simulate the MVC menu classes PopUpMenu, SelectionMenu, and CustomMenu when running in a Morphic world. I am also used to implement Utilities>informUser:during:.!

!MVCMenuMorph class methodsFor: 'instance creation' stamp: 'nice 6/11/2010 21:52'!
from: aPopupMenu title: titleStringOrNil
	"Answer a MenuMorph constructed from the given PopUpMenu. Used to simulate MVC-style menus in a Morphic-only world."

	| menu items lines selections labelString j |
	menu := self new.
	titleStringOrNil ifNotNil: [
		titleStringOrNil isEmpty ifFalse: [menu addTitle: titleStringOrNil]].
	labelString := aPopupMenu labelString.
	items := labelString asString lines.
	labelString isText ifTrue:
		["Pass along text emphasis if present"
		j := 1.
		items := items collect:
			[:item | | emphasis |
			j := labelString asString findString: item startingAt: j.
			emphasis := TextEmphasis new emphasisCode: (labelString emphasisAt: j).
			item asText addAttribute: emphasis]].
	lines := aPopupMenu lineArray.
	lines ifNil: [lines := #()].
	menu cancelValue: 0.
	menu defaultTarget: menu.
	selections := (1 to: items size) asArray.
	1 to: items size do: [:i |
		menu add: (items at: i) selector: #selectMVCItem: argument: (selections at: i).
		(lines includes: i) ifTrue: [menu addLine]].
	^ menu
! !

!MVCMenuMorph methodsFor: 'private' stamp: 'ar 9/18/2000 12:10'!
cancelValue: selectionOrNil
	"Set the value to be returned if the user cancels without making a selection."

	mvcSelection := selectionOrNil.
! !

!MVCMenuMorph methodsFor: 'invoking' stamp: 'bf 6/19/2008 17:37'!
displayAt: aPoint during: aBlock
	"Add this menu to the Morphic world during the execution of the given block."

	Smalltalk isMorphic ifFalse: [^ self].

	[ActiveWorld addMorph: self centeredNear: aPoint.
	self world displayWorld.  "show myself"
	aBlock value]
		ensure: [self delete]! !

!MVCMenuMorph methodsFor: 'invoking' stamp: 'nk 4/6/2002 22:33'!
informUserAt: aPoint during: aBlock
	"Add this menu to the Morphic world during the execution of the given block."

	| title w |
	Smalltalk isMorphic ifFalse: [^ self].

	title := self allMorphs detect: [ :ea | ea hasProperty: #titleString ].
	title := title submorphs first.
	self visible: false.
	w := ActiveWorld.
	aBlock value:[:string|
		self visible ifFalse:[
			w addMorph: self centeredNear: aPoint.
			self visible: true].
		title contents: string.
		self setConstrainedPosition: Sensor cursorPoint hangOut: false.
		self changed.
		w displayWorld		 "show myself"
	]. 
	self delete.
	w displayWorld! !

!MVCMenuMorph methodsFor: 'initializing' stamp: 'laza 4/20/2004 10:41'!
initialize
	super initialize.
	self setProperty: #morphicLayerNumber toValue: self morphicLayerNumber
! !

!MVCMenuMorph methodsFor: 'invoking' stamp: 'ar 12/27/2001 22:46'!
invokeAt: aPoint in: aWorld
	"Add this menu to the given world centered at the given point. Wait for the user to make a selection and answer it. The selection value returned is an integer in keeping with PopUpMenu, if the menu is converted from an MVC-style menu."
	"Details: This is invoked synchronously from the caller. In order to keep processing inputs and updating the screen while waiting for the user to respond, this method has its own version of the World's event loop."

	^ self invokeAt: aPoint in: aWorld allowKeyboard: Preferences menuKeyboardControl! !

!MVCMenuMorph methodsFor: 'invoking' stamp: 'KLC 4/11/2004 10:56'!
invokeAt: aPoint in: aWorld allowKeyboard: aBoolean
	"Add this menu to the given world centered at the given point. Wait for the user to make a selection and answer it. The selection value returned is an integer in keeping with PopUpMenu, if the menu is converted from an MVC-style menu."
	"Details: This is invoked synchronously from the caller. In order to keep processing inputs and updating the screen while waiting for the user to respond, this method has its own version of the World's event loop." 
	| w originalFocusHolder |
	self flag: #bob.		"is <aPoint> global or local?"
	self flag: #arNote.	"<aPoint> is local to aWorld"
	originalFocusHolder := aWorld primaryHand keyboardFocus.
	self popUpAt: aPoint forHand: aWorld primaryHand in: aWorld allowKeyboard: aBoolean.
	done := false.
	w := aWorld outermostWorldMorph. "containing hand"
	[self isInWorld & done not] whileTrue: [w doOneSubCycle].
	self delete.
	originalFocusHolder ifNotNil: [aWorld primaryHand newKeyboardFocus: originalFocusHolder].
	^ mvcSelection
! !

!MVCMenuMorph methodsFor: 'WiW support' stamp: 'laza 4/20/2004 10:38'!
morphicLayerNumber
	^self valueOfProperty: #morphicLayerNumber ifAbsent: [10].
! !

!MVCMenuMorph methodsFor: 'private' stamp: 'ar 9/18/2000 12:16'!
selectMVCItem: item
	"Called by the MenuItemMorph that the user selects.
	Record the selection and set the done flag to end this interaction."

	mvcSelection := item.
	done := true.
! !

MenuMorph subclass: #UpdatingMenuMorph
	instanceVariableNames: 'menuUpdater'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Menus'!

!UpdatingMenuMorph methodsFor: 'as yet unclassified' stamp: 'kb 2/22/2010 15:00'!
activate: evt
	"Receiver should be activated; e.g., so that control passes correctly."
	
	self updateMenu.
	super activate: evt! !

!UpdatingMenuMorph methodsFor: 'initialization' stamp: 'kb 2/22/2010 14:57'!
initialize

	super initialize.
	menuUpdater := MenuUpdater new! !

!UpdatingMenuMorph methodsFor: 'update' stamp: 'kb 2/22/2010 15:00'!
updateMenu

	menuUpdater update: self! !

!UpdatingMenuMorph methodsFor: 'initialization' stamp: 'kb 2/22/2010 14:58'!
updater: anObject updateSelector: aSelector

	menuUpdater updater: anObject updateSelector: aSelector! !

!UpdatingMenuMorph methodsFor: 'initialization' stamp: 'kb 2/22/2010 14:58'!
updater: anObject updateSelector: aSelector arguments: anArray

	menuUpdater updater: anObject updateSelector: aSelector arguments: anArray! !

!Morph class methodsFor: 'testing' stamp: 'nk 6/12/2004 09:20'!
allSketchMorphClasses
	"Morph allSketchMorphClasses"
	^ Array
		streamContents: [:s | self
				withAllSubclassesDo: [:cls | cls isSketchMorphClass
						ifTrue: [s nextPut: cls ]]]
! !

!Morph class methodsFor: 'testing' stamp: 'nice 12/27/2009 20:10'!
allSketchMorphForms
	"Answer a Set of forms of SketchMorph (sub) instances, except those 
	used as button images, ones being edited, and those with 0 extent."

	| reasonableForms |
	reasonableForms := Set new.
	Morph allSketchMorphClasses do:
		[:cls | cls allInstances do:
			[:m | | form |
			(m owner isKindOf: SketchEditorMorph orOf: IconicButton)
				ifFalse:
					[form := m form.
					((form width > 0) and: [form height > 0]) ifTrue: [reasonableForms add: form]]]].
	^ reasonableForms! !

!Morph class methodsFor: 'scripting' stamp: 'sw 8/11/97 13:17'!
authoringPrototype
	"Answer an instance of the receiver suitable for placing in a parts bin for authors"
	
	^ self new markAsPartsDonor! !

!Morph class methodsFor: 'fileIn/Out' stamp: 'nk 7/16/2003 15:54'!
fileReaderServicesForFile: fullName suffix: suffix

	^({ 'morph'. 'morphs'. 'sp'. '*' } includes: suffix)
		ifTrue: [
			{SimpleServiceEntry 
				provider: self 
				label: 'load as morph'
				selector: #fromFileName:
				description: 'load as morph'}]
		ifFalse: [#()]! !

!Morph class methodsFor: 'fileIn/Out' stamp: 'yo 8/7/2003 11:02'!
fromFileName: fullName
	"Reconstitute a Morph from the file, presumed to be represent a Morph saved
	via the SmartRefStream mechanism, and open it in an appropriate Morphic world"

 	| aFileStream morphOrList |
	aFileStream := (MultiByteBinaryOrTextStream with: ((FileStream readOnlyFileNamed: fullName) binary contentsOfEntireFile)) binary reset.
	morphOrList := aFileStream fileInObjectAndCode.
	(morphOrList isKindOf: SqueakPage) ifTrue: [morphOrList := morphOrList contentsMorph].
	Smalltalk isMorphic
		ifTrue: [ActiveWorld addMorphsAndModel: morphOrList]
		ifFalse:
			[morphOrList isMorph ifFalse: [self inform: 'Can only load a single morph
into an mvc project via this mechanism.'].
			morphOrList openInWorld]! !

!Morph class methodsFor: 'new-morph participation' stamp: 'di 6/22/97 09:07'!
includeInNewMorphMenu
	"Return true for all classes that can be instantiated from the menu"
	^ true! !

!Morph class methodsFor: 'preferences' stamp: 'mt 4/10/2015 17:22'!
indicateKeyboardFocus

	<preference: 'Indicate Keyboard Focus'
		categoryList: #(keyboard Morphic general)
		description: 'If enabled, there will be a visual highlight drawn onto the morph to help the user find out about the current keyboard focus. This is especially useful when the keyboard focus does not match the mouse position.'
		type: #Boolean>
	^ IndicateKeyboardFocus ifNil: [false]! !

!Morph class methodsFor: 'preferences' stamp: 'mt 4/10/2015 15:27'!
indicateKeyboardFocus: aBoolean

	IndicateKeyboardFocus := aBoolean.! !

!Morph class methodsFor: 'class initialization' stamp: 'cmm 5/8/2014 10:58'!
initialize
	"Morph initialize"
	FileList registerFileReader: self! !

!Morph class methodsFor: 'instance creation' stamp: 'efo 5/3/2002 14:59'!
initializedInstance
	"Answer an instance of the receiver which in some sense is initialized.  In the case of Morphs, this will yield an instance that can be attached to the Hand after having received the same kind of basic initialization that would be obtained from an instance chosen from the 'new morph' menu.
	Return nil if the receiver is reluctant for some reason to return such a thing"

	^ (self class includesSelector: #descriptionForPartsBin)
		ifTrue:
			[self newStandAlone]
		ifFalse:
			[self new]! !

!Morph class methodsFor: 'testing' stamp: 'nk 6/12/2004 09:17'!
isSketchMorphClass
	^false! !

!Morph class methodsFor: 'misc' stamp: 'nice 12/27/2009 20:10'!
morphsUnknownToTheirOwners
	"Return a list of all morphs (other than HandMorphs) whose owners do not contain them in their submorph lists"
	"Morph morphsUnknownToTheirOwners"
	| problemMorphs |
	problemMorphs := OrderedCollection new.
	self allSubInstances do:
		[:m | | itsOwner |
		(m isHandMorph not and: [((itsOwner := m owner) ~~ nil and: [(itsOwner submorphs includes: m) not])])
			ifTrue:
				[problemMorphs add: m]].
	^ problemMorphs! !

!Morph class methodsFor: 'instance creation' stamp: ''!
newBounds: bounds

	^ self new privateBounds: bounds! !

!Morph class methodsFor: 'instance creation' stamp: 'jm 5/29/1998 21:28'!
newBounds: bounds color: color

	^ (self new privateBounds: bounds) privateColor: color
! !

!Morph class methodsFor: 'new-morph participation' stamp: 'sw 6/28/2001 11:33'!
newStandAlone
	"Answer an instance capable of standing by itself as a usable morph."

	^ self basicNew initializeToStandAlone! !

!Morph class methodsFor: 'instance creation' stamp: 'sw 8/4/97 12:05'!
newSticky

	^ self new beSticky! !

!Morph class methodsFor: 'new-morph participation' stamp: 'sw 8/2/2001 12:01'!
partName: aName categories: aList documentation: aDoc
	"Answer a DescriptionForPartsBin which will represent a launch of a new instance of my class via the #newStandAlone protocol sent to my class. Use the category-list and documentation provided"


	^ DescriptionForPartsBin new
		formalName: aName
		categoryList: aList
		documentation: aDoc
		globalReceiverSymbol: self name
		nativitySelector: #newStandAlone! !

!Morph class methodsFor: 'new-morph participation' stamp: 'ar 7/21/2010 20:03'!
partName: aName categories: aList documentation: aDoc sampleImageForm: aForm
	"Answer a DescriptionForPartsBin which will represent a launch of a new instance of my class via the #newStandAlone protocol sent to my class. Use the category-list and documentation provided.  This variant allows an overriding image form to be provided, useful in cases where we don't want to launch a sample instance just to get the form"

	| descr |
	descr := DescriptionForPartsBin new
		formalName: aName
		categoryList: aList
		documentation: aDoc
		globalReceiverSymbol: self name
		nativitySelector: #newStandAlone.
	descr sampleImageForm: aForm.
	^ descr
! !

!Morph class methodsFor: 'preferences' stamp: 'mt 2/27/2015 11:47'!
preferredCornerRadius

	<preference: 'Preferred Corner Radius'
		categoryList: #(Morphic windows menus)
		description: 'If a morph wants rounded corners, use this radius. May be overwritten in subclasses.'
		type: #Number>
	^ PreferredCornerRadius ifNil: [6]! !

!Morph class methodsFor: 'preferences' stamp: 'mt 2/25/2015 10:36'!
preferredCornerRadius: anInteger

	PreferredCornerRadius := anInteger.! !

!Morph class methodsFor: 'fileIn/Out' stamp: 'KR 3/21/2008 20:57'!
serviceLoadMorphFromFile
	"Answer a service for loading a .morph file"

	^ SimpleServiceEntry 
		provider: self 
		label: 'load as morph' translatedNoop
		selector: #fromFileName:
		description: 'load as morph' translatedNoop
		buttonLabel: 'load' translatedNoop! !

!Morph class methodsFor: 'fileIn/Out' stamp: 'sd 2/1/2002 21:45'!
services

	^ Array with: self serviceLoadMorphFromFile! !

!Morph class methodsFor: 'preferences' stamp: 'mt 7/30/2016 07:50'!
themeProperties
	
	| catName |
	catName := 'Other'.
	
	^{ 
		{ #hardShadowColor. catName.  'The color of shadows. An alpha about 0.5 is recommended.' }.
		{ #hardShadowOffset. catName. 'The offset of the shadows behind the morph.' } .
		{ #softShadowColor. catName.  'The color of soft shadows. An alpha less than 0.05 is recommended.' }.
		{ #softShadowOffset. catName. 'The offset of soft shadows behind the morph.' } .
		{ #keyboardFocusColor. catName. 'The color of the keyboard focus indication.' }.
		{ #keyboardFocusWidth. catName. 'The width of the keyboard focus indication.' }.
	}! !

!Morph class methodsFor: 'initialize-release' stamp: 'SD 11/15/2001 22:22'!
unload

	FileList unregisterFileReader: self ! !

!Morph class methodsFor: 'preferences' stamp: 'mt 4/6/2015 13:12'!
useSoftDropShadow

	<preference: 'Use Soft Drop Shadow'
		categoryList: #(Morphic windows menus performance)
		description: 'If drop shadows are enabled, this preference will switch between hard and soft shadows. Soft shadows are more expensive.'
		type: #Boolean>
	^ UseSoftDropShadow ifNil: [ false ]! !

!Morph class methodsFor: 'preferences' stamp: 'mt 7/30/2016 07:54'!
useSoftDropShadow: aBoolean

	UseSoftDropShadow = aBoolean ifTrue: [^ self].
	UseSoftDropShadow := aBoolean.
	
	SystemWindow refreshAllWindows.
	DialogWindow refreshAllDialogs.
	SystemProgressMorph reset.
	TheWorldMainDockingBar updateInstances.! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'tk 12/15/1998 14:23'!
abandon
	"Like delete, but we really intend not to use this morph again.  Clean up a few things."

	self delete! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/5/2000 20:00'!
aboutToBeGrabbedBy: aHand
	"The receiver is being grabbed by a hand.
	Perform necessary adjustments (if any) and return the actual morph
	that should be added to the hand."
	| extentToHandToHand cmd |
	self formerOwner: owner.
	self formerPosition: self position.
	cmd := self undoGrabCommand.
	cmd ifNotNil:[self setProperty: #undoGrabCommand toValue: cmd].
	(extentToHandToHand := self valueOfProperty: #expandedExtent)
			ifNotNil:
				[self removeProperty: #expandedExtent.
				self extent: extentToHandToHand].
	^self "Grab me"! !

!Morph methodsFor: 'menus' stamp: 'sw 10/6/2004 11:38'!
absorbStateFromRenderer: aRenderer 
	"Transfer knownName, actorState, visible, and player info over from aRenderer, which was formerly imposed above me as a transformation shell but is now going away."

	| current |
	(current := aRenderer actorStateOrNil) ifNotNil:
		[self actorState: current.
		aRenderer actorState: nil].

	(current := aRenderer knownName) ifNotNil:
		[self setNameTo: current.
		aRenderer setNameTo: nil].

	(current := aRenderer player) ifNotNil:
		[self player: current.
		current rawCostume: self.
		aRenderer player: nil].

	self visible: aRenderer visible! !

!Morph methodsFor: 'layout' stamp: 'ar 11/12/2000 17:33'!
acceptDroppingMorph: aMorph event: evt
	"This message is sent when a morph is dropped onto a morph that has agreed to accept the dropped morph by responding 'true' to the wantsDroppedMorph:Event: message. This default implementation just adds the given morph to the receiver."
	| layout |
	layout := self layoutPolicy.
	layout ifNil:[^self addMorph: aMorph].
	self privateAddMorph: aMorph 
		atIndex: (layout indexForInserting: aMorph at: evt position in: self).! !

!Morph methodsFor: 'events-accessing' stamp: 'rw 4/25/2002 07:18'!
actionMap
	"Answer an action map"

	| actionMap |
	actionMap := self valueOfProperty: #actionMap.
	actionMap ifNil:
		[actionMap := self createActionMap].
	^ actionMap! !

!Morph methodsFor: 'structure' stamp: 'mt 5/1/2016 18:03'!
activeHand
	
	^ ActiveHand ifNil: [
		self isInWorld
			ifTrue: [self world activeHand]
			ifFalse: [nil]]! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:51'!
actorStateOrNil
	"answer the redeiver's actorState"
	^ extension ifNotNil: [extension actorState]! !

!Morph methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:52'!
actorState: anActorState 
	"change the receiver's actorState"
	self assureExtension actorState: anActorState! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 9/28/2001 08:39'!
actWhen
	"Answer when the receiver, probably being used as a button, should have its action triggered"

	^ self valueOfProperty: #actWhen ifAbsentPut: [#buttonDown]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 9/25/2001 10:23'!
actWhen: aButtonPhase
	"Set the receiver's actWhen trait"

	self setProperty: #actWhen toValue: aButtonPhase! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 2/9/1999 17:43'!
adaptToWorld: aWorld
	"The receiver finds itself operating in a possibly-different new world.  If any of the receiver's parts are world-dependent (such as a target of a SimpleButtonMorph, etc.), then have them adapt accordingly"
	submorphs do: [:m | m adaptToWorld: aWorld].
	self eventHandler ifNotNil:
		[self eventHandler adaptToWorld: aWorld]! !

!Morph methodsFor: 'menus' stamp: 'sw 11/27/2001 14:36'!
addAddHandMenuItemsForHalo: aMenu hand: aHandMorph
	"The former charter of this method was to add halo menu items that pertained specifically to the hand.  Over time this charter has withered, and most morphs reimplement this method simply to add their morph-specific menu items.  So in the latest round, all other implementors in the standard image have been removed.  However, this is left here as a hook for the benefit of existing code in client uses."

! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector after: delayTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: #() after: delayTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector at: scheduledTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: #() at: scheduledTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector withArguments: args after: delayTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: args at: Time millisecondClockValue + delayTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/14/2000 12:15'!
addAlarm: aSelector withArguments: args at: scheduledTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	| scheduler |
	scheduler := self alarmScheduler.
	scheduler ifNotNil:[scheduler addAlarm: aSelector withArguments: args for: self at: scheduledTime].! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector with: arg1 after: delayTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: (Array with: arg1) after: delayTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector with: arg1 at: scheduledTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: (Array with: arg1) at: scheduledTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector with: arg1 with: arg2 after: delayTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: (Array with: arg1 with: arg2) after: delayTime! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:35'!
addAlarm: aSelector with: arg1 with: arg2 at: scheduledTime
	"Add an alarm (that is an action to be executed once) with the given set of parameters"
	^self addAlarm: aSelector withArguments: (Array with: arg1 with: arg2) at: scheduledTime! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'mt 2/10/2014 11:56:55.76'!
addAllMorphsBack: aCollection
	^self privateAddAllMorphs: aCollection atIndex: submorphs size + 1! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'mt 2/10/2014 11:56:36.22'!
addAllMorphsFront: aCollection
	^self privateAddAllMorphs: aCollection atIndex: 1! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'mt 2/10/2014 11:57:47.655'!
addAllMorphs: aCollection
	^self addAllMorphsBack: aCollection! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'mt 2/10/2014 11:28:24.837'!
addAllMorphs: aCollection after: anotherMorph
	^self addAllMorphs: aCollection behind: anotherMorph! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'mt 2/10/2014 11:29:18.817'!
addAllMorphs: aCollection behind: anotherMorph
	^self privateAddAllMorphs: aCollection 
			atIndex: (submorphs indexOf: anotherMorph) + 1! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'mt 2/10/2014 11:29:58.664'!
addAllMorphs: aCollection inFrontOf: anotherMorph
	^self privateAddAllMorphs: aCollection
			atIndex: ((submorphs indexOf: anotherMorph) max: 1)! !

!Morph methodsFor: 'menu' stamp: 'sw 11/27/2001 15:21'!
addBorderStyleMenuItems: aMenu hand: aHandMorph
	"Probably one could offer border-style items even if it's not a borderedMorph, so this remains a loose end for the moment"
! !

!Morph methodsFor: 'layout-menu' stamp: 'dgd 8/30/2003 16:57'!
addCellLayoutMenuItems: aMenu hand: aHand
	"Cell (e.g., child) related items"
	| menu sub |
	menu := MenuMorph new defaultTarget: self.
		menu addUpdating: #hasDisableTableLayoutString action: #changeDisableTableLayout.
		menu addLine.

		sub := MenuMorph new defaultTarget: self.
		#(rigid shrinkWrap spaceFill) do:[:sym|
			sub addUpdating: #hResizingString: target: self selector: #hResizing: argumentList: (Array with: sym)].
		menu add:'horizontal resizing' translated subMenu: sub.

		sub := MenuMorph new defaultTarget: self.
		#(rigid shrinkWrap spaceFill) do:[:sym|
			sub addUpdating: #vResizingString: target: self selector: #vResizing: argumentList: (Array with: sym)].
		menu add:'vertical resizing' translated subMenu: sub.

	aMenu ifNotNil:[aMenu add: 'child layout' translated subMenu: menu].
	^menu! !

!Morph methodsFor: 'menus' stamp: 'dgd 11/15/2003 19:25'!
addCopyItemsTo: aMenu
	"Add copy-like items to the halo menu"

	| subMenu |
	subMenu := MenuMorph new defaultTarget: self.
	subMenu add: 'copy to paste buffer' translated action: #copyToPasteBuffer:.
	subMenu add: 'copy text' translated action: #clipText.
	subMenu add: 'copy Postscript' translated action: #clipPostscript.
	subMenu add: 'print Postscript to file...' translated target: self selector: #printPSToFile.
	aMenu add: 'copy & print...' translated subMenu: subMenu! !

!Morph methodsFor: 'menus' stamp: 'sw 4/27/1998 03:44'!
addCustomHaloMenuItems: aMenu hand: aHandMorph
	"Add morph-specific items to the given menu which was invoked by the given hand from the halo.  To get started, we defer to the counterpart method used with the option-menu, but in time we can have separate menu choices for halo-menus and for option-menus"

	self addCustomMenuItems: aMenu hand: aHandMorph! !

!Morph methodsFor: 'menus' stamp: 'sw 11/27/2001 07:17'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	"Add morph-specific items to the given menu which was invoked by the given hand.  This method provides is invoked both from the halo-menu and from the control-menu regimes."
! !

!Morph methodsFor: 'debug and other' stamp: 'dgd 8/30/2003 20:36'!
addDebuggingItemsTo: aMenu hand: aHandMorph
	aMenu add: 'debug...' translated subMenu:  (self buildDebugMenu: aHandMorph)! !

!Morph methodsFor: 'drop shadows' stamp: 'mt 7/30/2016 08:02'!
addDropShadow

	self
		hasDropShadow: true;
		shadowOffset: (self useSoftDropShadow
			ifTrue: [self userInterfaceTheme softShadowOffset ifNil: [10@8 corner: 10@12]]
			ifFalse: [self userInterfaceTheme hardShadowOffset ifNil: [1@1]]);
		shadowColor: (self useSoftDropShadow
			ifTrue: [self userInterfaceTheme softShadowColor ifNil: [Color black alpha: 0.01]]
			ifFalse: [self userInterfaceTheme hardShadowColor ifNil: [Color black alpha: 0.5]]).! !

!Morph methodsFor: 'drop shadows' stamp: 'dgd 8/30/2003 16:48'!
addDropShadowMenuItems: aMenu hand: aHand
	| menu |
	menu := MenuMorph new defaultTarget: self.
	menu
		addUpdating: #hasDropShadowString
		action: #toggleDropShadow.
	menu addLine.
	menu add: 'shadow color...' translated target: self selector: #changeShadowColor.
	menu add: 'shadow offset...' translated target: self selector: #setShadowOffset:.
	aMenu add: 'drop shadow' translated subMenu: menu.! !

!Morph methodsFor: 'meta-actions' stamp: 'sw 9/29/2007 01:01'!
addEmbeddingMenuItemsTo: aMenu hand: aHandMorph
	"Construct a menu offerring embed targets for the receiver.  If the incoming menu is is not degenerate, add the constructed menu as a submenu; in any case, answer the embed-target menu"

	| menu w |
	menu _ MenuMorph new defaultTarget: self.
	w _ self world.
	self potentialEmbeddingTargets reverseDo: [:m |
		menu add: (m == w ifTrue: ['desktop' translated] ifFalse: [m knownName ifNil:[m class name asString]]) target: m selector: #addMorphFrontFromWorldPosition: argumentList: {self topRendererOrSelf}.
		m == self topRendererOrSelf owner ifTrue:
			[menu lastItem color: Color red]].
	aMenu ifNotNil:
		[menu submorphCount > 0 
			ifTrue:[aMenu add:'embed into' translated subMenu: menu]].
	^ menu! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:08'!
addEventBubbleFilter: anObject
	self eventBubbleFilters: (self addFilter: anObject to: self eventBubbleFilters).! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:07'!
addEventCaptureFilter: anObject
	self eventCaptureFilters: (self addFilter: anObject to: self eventCaptureFilters).! !

!Morph methodsFor: 'menus' stamp: 'nk 2/16/2004 13:29'!
addExportMenuItems: aMenu hand: aHandMorph
	"Add export items to the menu"

	aMenu ifNotNil:
		[ | aSubMenu |
		aSubMenu := MenuMorph new defaultTarget: self.
		aSubMenu add: 'BMP file' translated action: #exportAsBMP.
		aSubMenu add: 'GIF file' translated action: #exportAsGIF.
		aSubMenu add: 'JPEG file' translated action: #exportAsJPEG.
		aSubMenu add: 'PNG file' translated action: #exportAsPNG.
		aMenu add: 'export...' translated subMenu: aSubMenu]
! !

!Morph methodsFor: 'change reporting' stamp: 'ar 8/12/2003 21:50'!
addedMorph: aMorph
	"Notify the receiver that the given morph was just added."
! !

!Morph methodsFor: 'menus' stamp: 'dgd 8/30/2003 16:44'!
addFillStyleMenuItems: aMenu hand: aHand
	"Add the items for changing the current fill style of the Morph"
	| menu |
	self canHaveFillStyles ifFalse:[^aMenu add: 'change color...' translated target: self action: #changeColor].
	menu := MenuMorph new defaultTarget: self.
	self fillStyle addFillStyleMenuItems: menu hand: aHand from: self.
	menu addLine.
	menu add: 'solid fill' translated action: #useSolidFill.
	menu add: 'gradient fill' translated action: #useGradientFill.
	menu add: 'bitmap fill' translated action: #useBitmapFill.
	menu add: 'default fill' translated action: #useDefaultFill.
	aMenu add: 'fill style' translated subMenu: menu.
	"aMenu add: 'change color...' translated action: #changeColor"! !

!Morph methodsFor: 'events-filtering' stamp: 'mt 5/17/2016 22:20'!
addFilter: anObject to: aFilterGroup
	"Add anObject to the given listener group. Return the new group."
	| filters |
	filters := aFilterGroup.
	(filters notNil and:[filters includes: anObject]) ifFalse:[
		filters
			ifNil:[filters := WeakArray with: anObject]
			ifNotNil:[filters := filters copyWith: anObject]].
	filters := filters copyWithout: nil. "obsolete entries"
	^filters! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'kfr 5/22/2011 12:05'!
addFlexShell
	"Wrap a rotating and scaling shell around this morph."

	| oldHalo flexMorph myWorld anIndex morphOwner |

	myWorld := self world.
	oldHalo:= self halo.
	self owner ifNotNil:[ morphOwner := self owner]
					ifNil:[morphOwner := self currentWorld].
	
	anIndex := morphOwner  submorphIndexOf: self.
	morphOwner  addMorph: (flexMorph := self newTransformationMorph asFlexOf: self)
		asElementNumber: anIndex.
	self transferStateToRenderer: flexMorph.
	oldHalo ifNotNil: [oldHalo setTarget: flexMorph].
	myWorld ifNotNil: [myWorld startSteppingSubmorphsOf: flexMorph].

	^ flexMorph! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'di 11/28/2001 18:22'!
addFlexShellIfNecessary
	"If this morph requires a flex shell to scale or rotate,
		then wrap it in one and return it.
	Polygons, eg, may override to return themselves."

	^ self addFlexShell! !

!Morph methodsFor: 'menu' stamp: 'nk 2/15/2004 09:08'!
addGestureMenuItems: aMenu hand: aHandMorph
	"If the receiver wishes the Genie menu items, add a line to the menu and then those Genie items, else do nothing"! !

!Morph methodsFor: 'menu' stamp: 'nk 6/15/2004 07:11'!
addGraphModelYellowButtonItemsTo: aCustomMenu event: evt
	^aCustomMenu! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 11/4/2015 20:10'!
addHalo
	"Invoke a halo programatically (e.g., not from a meta gesture)"
	
	^ self createHalo
		popUpFor: self;
		yourself! !

!Morph methodsFor: 'menus' stamp: 'sw 8/6/2009 18:35'!
addHaloActionsTo: aMenu
	"Add items to aMenu representing actions requestable via halo"

	| subMenu |
	subMenu _ MenuMorph new defaultTarget: self.
	subMenu addTitle: self externalName.
	subMenu addStayUpItemSpecial.
	subMenu addLine.
	subMenu add: 'delete' translated action: #dismissViaHalo.
	subMenu balloonTextForLastItem: 'Delete this object -- warning -- can be destructive!!' translated.

	self maybeAddCollapseItemTo: subMenu.
	subMenu add: 'grab' translated action: #openInHand.
	subMenu balloonTextForLastItem: 'Pick this object up -- warning, since this removes it from its container, it can have adverse effects.' translated.

	subMenu addLine.

	subMenu add: 'resize' translated action: #resizeFromMenu.
	subMenu balloonTextForLastItem: 'Change the size of this object' translated.

	subMenu add: 'duplicate' translated action: #maybeDuplicateMorph.
	subMenu balloonTextForLastItem: 'Hand me a copy of this object' translated.
	"Note that this allows access to the non-instancing duplicate even when this is a uniclass instance"

	self couldMakeSibling ifTrue:
		[subMenu add: 'make a sibling' translated action: #handUserASibling.
		subMenu balloonTextForLastItem: 'Make a new sibling of this object and hand it to me' translated].

	subMenu addLine.
	subMenu add: 'property sheet' translated target: self renderedMorph action: #openAPropertySheet.
	subMenu balloonTextForLastItem: 'Open a property sheet for me. Allows changing lots of stuff at once.' translated.

	subMenu add: 'set color' translated target: self renderedMorph action: #changeColor.
	subMenu balloonTextForLastItem: 'Change the color of this object' translated.

	subMenu add: 'viewer' translated target: self action: #beViewed.
	subMenu balloonTextForLastItem: 'Open a Viewer that will allow everything about this object to be seen and controlled.' translated.

	subMenu add: 'tile browser' translated target: self action: #openInstanceBrowserWithTiles.
	subMenu balloonTextForLastItem: 'Open a tool that will facilitate tile scripting of this object.' translated.

	subMenu add: 'tile representing this object' translated target: self action: #tearOffTile.
	subMenu balloonTextForLastItem: 'Hand me a tile represting this object' translated.
	subMenu addLine.

	subMenu add: 'inspect' translated target: self action: #inspect.
	subMenu balloonTextForLastItem: 'Open an Inspector on this object' translated.

	aMenu add: 'halo actions...' translated subMenu: subMenu
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 11/4/2015 20:10'!
addHalo: evt
	
	^ self createHalo
		popUpFor: self event: evt;
		yourself! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 11/7/1999 21:55'!
addHalo: evt from: formerHaloOwner
	"Transfer a halo from the former halo owner to the receiver"
	^self addHalo: evt! !

!Morph methodsFor: 'halos and balloon help' stamp: 'nice 12/27/2009 03:10'!
addHandlesTo: aHaloMorph box: box
	"Add halo handles to the halo.  Apply the halo filter if appropriate"

	
	aHaloMorph haloBox: box.
	Preferences haloSpecifications  do:
		[:aSpec | | wantsIt aSelector | 
			aSelector :=  aSpec addHandleSelector.
			wantsIt := Preferences selectiveHalos
				ifTrue:
					[self wantsHaloHandleWithSelector: aSelector inHalo: aHaloMorph]
				ifFalse:
					[true].
			wantsIt ifTrue:
				[(#(addMakeSiblingHandle: addDupHandle:) includes: aSelector) ifTrue:
					[wantsIt := self preferredDuplicationHandleSelector = aSelector].
			wantsIt ifTrue:
				[aHaloMorph perform: aSelector with: aSpec]]].

	aHaloMorph innerTarget addOptionalHandlesTo: aHaloMorph box: box! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:08'!
addKeyboardBubbleFilter: anObject
	self keyboardBubbleFilters: (self addFilter: anObject to: self keyboardBubbleFilters).! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:08'!
addKeyboardCaptureFilter: anObject
	self keyboardCaptureFilters: (self addFilter: anObject to: self keyboardCaptureFilters).! !

!Morph methodsFor: 'layout-menu' stamp: 'kfr 4/24/2015 16:32'!
addLayoutMenuItems: topMenu hand: aHand
	| aMenu |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu addUpdating: #hasNoLayoutString action: #changeNoLayout.
	aMenu addUpdating: #hasProportionalLayoutString action: #changeProportionalLayout.
	aMenu addUpdating: #hasTableLayoutString action: #changeTableLayout.
	aMenu addLine.
	aMenu add: 'change layout inset...' translated action: #changeLayoutInset:.
	aMenu addLine.
	self addCellLayoutMenuItems: aMenu hand: aHand.
	self addTableLayoutMenuItems: aMenu hand: aHand.
	aMenu addLine.
	aMenu add: 'copy layout properties' translated action: #copyLayoutProperties.
	topMenu ifNotNil:[topMenu add: 'layout' translated subMenu: aMenu].
	^aMenu! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 11/4/2015 20:18'!
addMagicHaloFor: aHand

	aHand halo ifNotNil: [:halo |
		halo target == self ifTrue:[^self].
		halo isMagicHalo ifFalse:[^self]].
	
	self createHalo
		popUpMagicallyFor: self hand: aHand! !

!Morph methodsFor: 'menus' stamp: 'sw 3/2/2004 22:11'!
addMiscExtrasTo: aMenu
	"Add a submenu of miscellaneous extra items to the menu."

	| realOwner realMorph subMenu |
	subMenu := MenuMorph new defaultTarget: self.
	(self isWorldMorph not and: [(self renderedMorph isSystemWindow) not])
		ifTrue: [subMenu add: 'put in a window' translated action: #embedInWindow].

	self isWorldMorph ifFalse:
		[subMenu add: 'adhere to edge...' translated action: #adhereToEdge.
		subMenu addLine].

	realOwner := (realMorph := self topRendererOrSelf) owner.
	(realOwner isKindOf: TextPlusPasteUpMorph) ifTrue:
		[subMenu add: 'GeeMail stuff...' translated subMenu: (realOwner textPlusMenuFor: realMorph)].

	subMenu
		add: 'add mouse up action' translated action: #addMouseUpAction;
		add: 'remove mouse up action' translated action: #removeMouseUpAction;
		add: 'hand me tiles to fire this button' translated action: #handMeTilesToFire.
	subMenu addLine.
	subMenu add: 'arrowheads on pen trails...' translated action: #setArrowheads.
	subMenu addLine.

	subMenu defaultTarget: self topRendererOrSelf.
	subMenu add: 'draw new path' translated action: #definePath.
	subMenu add: 'follow existing path' translated action: #followPath.
	subMenu add: 'delete existing path' translated action: #deletePath.
	subMenu addLine.

	self addGestureMenuItems: subMenu hand: ActiveHand.

	aMenu add: 'extras...' translated subMenu: subMenu! !

!Morph methodsFor: 'menu' stamp: 'ul 12/12/2009 14:09'!
addModelYellowButtonItemsTo: aCustomMenu event: evt 
	"Give my models a chance to add their context-menu items to  
	aCustomMenu."
	self model
		ifNotNil: [:mod |
			mod
				addModelYellowButtonMenuItemsTo: aCustomMenu
				forMorph: self
				hand: evt hand]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 1/31/2001 12:55'!
addMorphBack: aMorph
	^self privateAddMorph: aMorph atIndex: submorphs size+1! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'RAA 12/15/2000 19:34'!
addMorphCentered: aMorph

	aMorph position: bounds center - (aMorph extent // 2).
	self addMorphFront: aMorph.
! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 12/16/2001 21:08'!
addMorphFrontFromWorldPosition: aMorph
	^self addMorphFront: aMorph fromWorldPosition: aMorph positionInWorld.! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 1/31/2001 12:54'!
addMorphFront: aMorph
	^self privateAddMorph: aMorph atIndex: 1! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 11/15/1998 23:42'!
addMorphFront: aMorph fromWorldPosition: wp

	self addMorphFront: aMorph.
	aMorph position: (self transformFromWorld globalPointToLocal: wp)! !

!Morph methodsFor: 'WiW support' stamp: 'nice 12/27/2009 03:10'!
addMorphInFrontOfLayer: aMorph

	| targetLayer |

	targetLayer := aMorph morphicLayerNumberWithin: self.
	submorphs do: [ :each | | layerHere |
		each == aMorph ifTrue: [^self].
		layerHere := each morphicLayerNumberWithin: self.
		"the <= is the difference - it insures we go to the front of our layer"
		targetLayer <= layerHere ifTrue: [
			^self addMorph: aMorph inFrontOf: each
		].
	].
	self addMorphBack: aMorph.
! !

!Morph methodsFor: 'WiW support' stamp: 'RAA 6/29/2000 10:49'!
addMorphInLayer: aMorph

	submorphs do: [ :each |
		each == aMorph ifTrue: [^self].
		aMorph morphicLayerNumber < each morphicLayerNumber ifTrue: [
			^self addMorph: aMorph inFrontOf: each
		].
	].
	self addMorphBack: aMorph
! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'dgd 2/22/2003 14:26'!
addMorphNearBack: aMorph 
	| bg |
	(submorphs notEmpty and: [submorphs last mustBeBackmost]) 
		ifTrue: 
			[bg := submorphs last.
			bg privateDelete].
	self addMorphBack: aMorph.
	bg ifNotNil: [self addMorphBack: bg]! !

!Morph methodsFor: '*Morphic-Sound-piano rolls' stamp: 'RAA 12/11/2000 17:29'!
addMorphsTo: morphList pianoRoll: pianoRoll eventTime: t betweenTime: leftTime and: rightTime

	"a hack to allow for abitrary morphs to be dropped into piano roll"
	t > rightTime ifTrue: [^ self].  
	t < leftTime ifTrue: [^ self].
	morphList add: (self left: (pianoRoll xForTime: t)).
! !

!Morph methodsFor: 'submorphs-add/remove' stamp: ''!
addMorph: aMorph

	self addMorphFront: aMorph.! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'mt 2/10/2014 11:27:52.161'!
addMorph: newMorph after: aMorph
	^self addMorph: newMorph behind: aMorph! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 4/14/2006 18:11'!
addMorph: aMorph asElementNumber: aNumber
	"Add the given morph so that it becomes the aNumber'th element of my submorph list.  If aMorph is already one of my submorphs, reposition it"

	(submorphs includes: aMorph) ifTrue:
		[aMorph privateDelete].
	(aNumber notNil and: [aNumber <= submorphs size])
		ifTrue:
			[self addMorph: aMorph inFrontOf: (submorphs at: aNumber)]
		ifFalse:
			[self addMorphBack: aMorph]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 1/31/2001 12:44'!
addMorph: newMorph behind: aMorph
	"Add a morph to the list of submorphs behind the specified morph"
	^self privateAddMorph: newMorph atIndex: (submorphs indexOf: aMorph) + 1.
! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'JW 2/1/2001 12:52'!
addMorph: aMorph fullFrame: aLayoutFrame

	aMorph layoutFrame: aLayoutFrame.
	aMorph hResizing: #spaceFill; vResizing: #spaceFill.
	self addMorph: aMorph.

! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 1/31/2001 12:45'!
addMorph: newMorph inFrontOf: aMorph
	"Add a morph to the list of submorphs in front of the specified morph"
	^self privateAddMorph: newMorph atIndex: ((submorphs indexOf: aMorph) max: 1).! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 1/19/2001 07:51'!
addMouseActionIndicatorsWidth: anInteger color: aColor

	self deleteAnyMouseActionIndicators.

	self changed.
	self hasRolloverBorder: true.
	self setProperty: #rolloverWidth toValue: anInteger@anInteger.
	self setProperty: #rolloverColor toValue: aColor.
	self layoutChanged.
	self changed.

! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:08'!
addMouseBubbleFilter: anObject
	self mouseBubbleFilters: (self addFilter: anObject to: self mouseBubbleFilters).! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:08'!
addMouseCaptureFilter: anObject
	self mouseCaptureFilters: (self addFilter: anObject to: self mouseCaptureFilters).! !

!Morph methodsFor: 'debug and other' stamp: 'ar 8/6/2009 20:41'!
addMouseUpAction
	| codeToRun oldCode |
	oldCode := self
				valueOfProperty: #mouseUpCodeToRun
				ifAbsent: [''].
	codeToRun := UIManager default request: 'MouseUp expression:' translated initialAnswer: oldCode.
	self addMouseUpActionWith: codeToRun! !

!Morph methodsFor: 'debug and other' stamp: 'gm 2/22/2003 13:41'!
addMouseUpActionWith: codeToRun 
	((codeToRun isMessageSend) not and: [codeToRun isEmptyOrNil]) 
		ifTrue: [^self].
	self setProperty: #mouseUpCodeToRun toValue: codeToRun.
	self 
		on: #mouseUp
		send: #programmedMouseUp:for:
		to: self.
	self 
		on: #mouseDown
		send: #programmedMouseDown:for:
		to: self.
	self 
		on: #mouseEnter
		send: #programmedMouseEnter:for:
		to: self.
	self 
		on: #mouseLeave
		send: #programmedMouseLeave:for:
		to: self! !

!Morph methodsFor: 'menu' stamp: 'nk 3/10/2004 19:49'!
addMyYellowButtonMenuItemsToSubmorphMenus
	"Answer true if I have items to add to the context menus of my submorphs"

	^true! !

!Morph methodsFor: 'menu' stamp: 'dgd 9/13/2004 19:23'!
addNestedYellowButtonItemsTo: aMenu event: evt 
	"Add items to aMenu starting with me and proceeding down 
	through my submorph chain, 
	letting any submorphs that include the event position 
	contribute their items to the bottom of the menu, separated by 
	a line."
	| underMouse |

	self addYellowButtonMenuItemsTo: aMenu event: evt.

	underMouse := self
				submorphThat: [:each | each containsPoint: evt position]
				ifNone: [^ self].

	(underMouse addMyYellowButtonMenuItemsToSubmorphMenus
			and: [underMouse hasYellowButtonMenu])
		ifTrue: [| submenu |
			aMenu addLine.
			submenu := MenuMorph new defaultTarget: underMouse.
			underMouse addNestedYellowButtonItemsTo: submenu event: evt.
			aMenu
				add: underMouse externalName
				icon: (underMouse iconOrThumbnailOfSize: 16)
				subMenu: submenu
		]
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 9/22/2000 20:41'!
addOptionalHandlesTo: aHalo box: box
	aHalo addDirectionHandles! !

!Morph methodsFor: 'menus' stamp: 'tfel 4/17/2015 14:36'!
addPaintingItemsTo: aMenu hand: aHandMorph 
	| subMenu movies |
	subMenu := MenuMorph new defaultTarget: self.
	subMenu
		add: 'repaint' translated action: #editDrawing;
		add: 'set rotation center' translated action: #setRotationCenter;
		add: 'reset forward-direction' translated action: #resetForwardDirection;
		add: 'set rotation style' translated action: #setRotationStyle;
		add: 'erase pixels of color' translated action: #erasePixelsUsing:;
		add: 'recolor pixels of color' translated action: #recolorPixelsUsing:;
		add: 'reduce color palette' translated action: #reduceColorPalette:;
		add: 'detect edges' translated action: #edgeDetect;
		add: 'sharpen' translated action: #sharpen;
		add: 'blur' translated action: #blur;
		add: 'emboss' translated action: #emboss;
		add: 'add a border around this shape...' translated action: #addBorderToShape:.
	movies := (self world rootMorphsAt: aHandMorph targetPoint) 
				select: [:m | (m isKindOf: MovieMorph) or: [m isSketchMorph]].
	movies size > 1 
		ifTrue: 
			[subMenu add: 'insert into movie' translated action: #insertIntoMovie:].
	aMenu add: 'painting...' translated subMenu: subMenu! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 12/21/1999 17:52'!
addSimpleHandlesTo: aHaloMorph box: aBox
	^ aHaloMorph addSimpleHandlesTo: aHaloMorph box: aBox! !

!Morph methodsFor: 'menus' stamp: 'ul 11/15/2010 11:54'!
addStandardHaloMenuItemsTo: aMenu hand: aHandMorph
	"Add standard halo items to the menu"

	| unlockables |

	self isWorldMorph ifTrue:
		[^ self addWorldHaloMenuItemsTo: aMenu hand: aHandMorph].

	self mustBeBackmost ifFalse:
		[aMenu add: 'send to back' translated action: #goBehind.
		aMenu add: 'bring to front' translated action: #comeToFront.
		self addEmbeddingMenuItemsTo: aMenu hand: aHandMorph.
		aMenu addLine].

	self addFillStyleMenuItems: aMenu hand: aHandMorph.
	self addBorderStyleMenuItems: aMenu hand: aHandMorph.
	self addDropShadowMenuItems: aMenu hand: aHandMorph.
	self addLayoutMenuItems: aMenu hand: aHandMorph.
	self addHaloActionsTo: aMenu.
	owner isTextMorph ifTrue:[self addTextAnchorMenuItems: aMenu hand: aHandMorph].
	aMenu addLine.
	self addToggleItemsToHaloMenu: aMenu.
	aMenu addLine.
	self addCopyItemsTo: aMenu.
	self addPlayerItemsTo: aMenu.
	self addExportMenuItems: aMenu hand: aHandMorph.
	self addStackItemsTo: aMenu.
	self addMiscExtrasTo: aMenu.
	Preferences noviceMode ifFalse:
		[self addDebuggingItemsTo: aMenu hand: aHandMorph].

	aMenu addLine.
	aMenu defaultTarget: self.

	aMenu addLine.

	unlockables := self submorphs select:
		[:m | m isLocked].
	unlockables size = 1 ifTrue:
		[aMenu
			add: ('unlock "{1}"' translated format: unlockables first externalName)
			action: #unlockContents].
	unlockables size > 1 ifTrue:
		[aMenu add: 'unlock all contents' translated action: #unlockContents.
		aMenu add: 'unlock...' translated action: #unlockOneSubpart].

	aMenu defaultTarget: aHandMorph.
! !

!Morph methodsFor: 'layout-menu' stamp: 'dgd 8/30/2003 20:07'!
addTableLayoutMenuItems: aMenu hand: aHand
	| menu sub |
	menu := MenuMorph new defaultTarget: self.
	menu addUpdating: #hasReverseCellsString action: #changeReverseCells.
	menu addUpdating: #hasClipLayoutCellsString action: #changeClipLayoutCells.
	menu addUpdating: #hasRubberBandCellsString action: #changeRubberBandCells.
	menu addLine.
	menu add: 'change cell inset...' translated action: #changeCellInset:.
	menu add: 'change min cell size...' translated action: #changeMinCellSize:.
	menu add: 'change max cell size...' translated action: #changeMaxCellSize:.
	menu addLine.

	sub := MenuMorph new defaultTarget: self.
	#(leftToRight rightToLeft topToBottom bottomToTop) do:[:sym|
		sub addUpdating: #listDirectionString: target: self selector: #changeListDirection: argumentList: (Array with: sym)].
	menu add: 'list direction' translated subMenu: sub.

	sub := MenuMorph new defaultTarget: self.
	#(none leftToRight rightToLeft topToBottom bottomToTop) do:[:sym|
		sub addUpdating: #wrapDirectionString: target: self selector: #wrapDirection: argumentList: (Array with: sym)].
	menu add: 'wrap direction' translated subMenu: sub.

	sub := MenuMorph new defaultTarget: self.
	#(center topLeft topRight bottomLeft bottomRight topCenter leftCenter rightCenter bottomCenter) do:[:sym|
		sub addUpdating: #cellPositioningString: target: self selector: #cellPositioning: argumentList: (Array with: sym)].
	menu add: 'cell positioning' translated subMenu: sub.

	sub := MenuMorph new defaultTarget: self.
	#(topLeft bottomRight center justified) do:[:sym|
		sub addUpdating: #listCenteringString: target: self selector: #listCentering: argumentList: (Array with: sym)].
	menu add: 'list centering' translated subMenu: sub.

	sub := MenuMorph new defaultTarget: self.
	#(topLeft bottomRight center justified) do:[:sym|
		sub addUpdating: #wrapCenteringString: target: self selector: #wrapCentering: argumentList: (Array with: sym)].
	menu add: 'wrap centering' translated subMenu: sub.

	sub := MenuMorph new defaultTarget: self.
	#(none equal) do:[:sym|
		sub addUpdating: #listSpacingString: target: self selector: #listSpacing: argumentList: (Array with: sym)].
	menu add: 'list spacing' translated subMenu: sub.

	sub := MenuMorph new defaultTarget: self.
	#(none localRect localSquare globalRect globalSquare) do:[:sym|
		sub addUpdating: #cellSpacingString: target: self selector: #cellSpacing: argumentList: (Array with: sym)].
	menu add: 'cell spacing' translated subMenu: sub.

	aMenu ifNotNil:[aMenu add: 'table layout' translated subMenu: menu].
	^menu! !

!Morph methodsFor: 'text-anchor' stamp: 'ar 12/17/2001 12:45'!
addTextAnchorMenuItems: topMenu hand: aHand
	| aMenu |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu addUpdating: #hasInlineAnchorString action: #changeInlineAnchor.
	aMenu addUpdating: #hasParagraphAnchorString action: #changeParagraphAnchor.
	aMenu addUpdating: #hasDocumentAnchorString action: #changeDocumentAnchor.
	topMenu ifNotNil:[topMenu add: 'text anchor' subMenu: aMenu].
	^aMenu! !

!Morph methodsFor: 'menu' stamp: 'dgd 7/28/2005 13:02'!
addTitleForHaloMenu: aMenu 
	aMenu
		addTitle: self externalName
		icon: (self iconOrThumbnailOfSize: (Preferences tinyDisplay ifFalse:[28] ifTrue:[16]))! !

!Morph methodsFor: 'menus' stamp: 'dgd 4/3/2006 14:11'!
addToggleItemsToHaloMenu: aMenu
	"Add standard true/false-checkbox items to the memu"

	#(
		(resistsRemovalString toggleResistsRemoval 'whether I should be reistant to easy deletion via the pink X handle' true)
		(stickinessString toggleStickiness 'whether I should be resistant to a drag done by mousing down on me' true)
		(lockedString lockUnlockMorph 'when "locked", I am inert to all user interactions' true)
		(hasClipSubmorphsString changeClipSubmorphs 'whether the parts of objects within me that are outside my bounds should be masked.' false)
		(hasDirectionHandlesString changeDirectionHandles 'whether direction handles are shown with the halo' false)
		(hasDragAndDropEnabledString changeDragAndDrop 'whether I am open to having objects dropped into me' false)
	)
		select:[:each | Preferences noviceMode not or:[each fourth]]
		thenDo:
		[:each |
			aMenu addUpdating: each first action: each second.
			aMenu balloonTextForLastItem: each third translated].

	self couldHaveRoundedCorners ifTrue:
		[aMenu addUpdating: #roundedCornersString action: #toggleCornerRounding.
		aMenu balloonTextForLastItem: 'whether my corners should be rounded' translated]! !

!Morph methodsFor: 'debug and other' stamp: 'sw 1/3/2001 06:42'!
addViewingItemsTo: aMenu
	"Add viewing-related items to the given menu.  If any are added, this method is also responsible for adding a line after them"! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 1/26/2000 19:37'!
addWorldHandlesTo: aHaloMorph box: box
	aHaloMorph haloBox: box.
	Preferences haloSpecificationsForWorld do:
		[:aSpec | 
			aHaloMorph perform: aSpec addHandleSelector with: aSpec].
	aHaloMorph innerTarget addOptionalHandlesTo: aHaloMorph box: box! !

!Morph methodsFor: 'menus' stamp: 'wiz 2/14/2006 18:54'!
addWorldTargetSightingItems: aCustomMenu hand: aHandMorph
"Use cursor to select a point on screen.
Set target from all possible morphs under cursor sight." 
	
	aCustomMenu addLine.
	
	aCustomMenu add: 'sight target' translated action: #sightWorldTargets:.
	! !

!Morph methodsFor: 'menu' stamp: 'nice 1/1/2010 22:05'!
addYellowButtonMenuItemsTo: aMenu event: evt 
	"Populate aMenu with appropriate menu items for a  
	yellow-button (context menu) click."
	aMenu defaultTarget: self.
	""
	Preferences noviceMode
		ifFalse: [aMenu addStayUpItem].
	""
	self addModelYellowButtonItemsTo: aMenu event: evt.
	""
	Preferences generalizedYellowButtonMenu
		ifFalse: [^ self].
	""
	Preferences cmdGesturesEnabled
		ifTrue: [""
			aMenu addLine.
			aMenu add: 'inspect' translated action: #inspect].
	""
	aMenu addLine.
	self world selectedObject == self
		ifTrue: [aMenu add: 'deselect' translated action: #removeHalo]
		ifFalse: [aMenu add: 'select' translated action: #addHalo].
	""
	(self isWorldMorph
			or: [self mustBeBackmost
			or: [self wantsToBeTopmost]])
		ifFalse: [""
			aMenu addLine.
			aMenu add: 'send to back' translated action: #goBehind.
			aMenu add: 'bring to front' translated action: #comeToFront.
			self addEmbeddingMenuItemsTo: aMenu hand: evt hand].
	""
	self isWorldMorph
		ifFalse: [""
	Smalltalk
		at: #NCAAConnectorMorph
		ifPresent: [:connectorClass | 
			aMenu addLine.
			aMenu add: 'connect to' translated action: #startWiring.
			aMenu addLine].
	""

			self isFullOnScreen
				ifFalse: [aMenu add: 'move onscreen' translated action: #goHome]].
	""
	Preferences noviceMode
		ifFalse: [""
			self addLayoutMenuItems: aMenu hand: evt hand.
			(owner notNil
					and: [owner isTextMorph])
				ifTrue: [self addTextAnchorMenuItems: aMenu hand: evt hand]].
	""
	self isWorldMorph
		ifFalse: [""
			aMenu addLine.
			self addToggleItemsToHaloMenu: aMenu].
	""
	aMenu addLine.
	self isWorldMorph
		ifFalse: [aMenu add: 'copy to paste buffer' translated action: #copyToPasteBuffer:].
	(self allStringsAfter: nil) isEmpty
		ifFalse: [aMenu add: 'copy text' translated action: #clipText].
	""
	self addExportMenuItems: aMenu hand: evt hand.
	""
	(Preferences noviceMode not
			and: [self isWorldMorph not])
		ifTrue: [""
			aMenu addLine.
			aMenu add: 'adhere to edge...' translated action: #adhereToEdge].
	""
	self addCustomMenuItems: aMenu hand: evt hand! !

!Morph methodsFor: 'menus' stamp: 'dgd 10/17/2003 22:51'!
adhereToEdge
	| menu |
	menu := MenuMorph new defaultTarget: self.
	#(top right bottom left - center - topLeft topRight bottomRight bottomLeft - none)
		do: [:each |
			each == #-
				ifTrue: [menu addLine]
				ifFalse: [menu add: each asString translated selector: #setToAdhereToEdge: argument: each]].
	menu popUpEvent: self currentEvent in: self world! !

!Morph methodsFor: 'menus' stamp: 'wiz 10/19/2006 00:35'!
adhereToEdge: edgeSymbol 
	| edgeMessage |
	(owner isNil or: [owner isHandMorph]) ifTrue: [^self].
	(owner class canUnderstand:  edgeSymbol) ifFalse:  [^self].
	(self class canUnderstand: ( edgeMessage := (edgeSymbol , ':') asSymbol ))
		 ifFalse:  [^self].
	
	self perform: edgeMessage
		withArguments: (Array with: (owner perform: edgeSymbol))! !

!Morph methodsFor: 'menus' stamp: 'sw 2/3/2000 00:14'!
adjustedCenter
	"Provides a hook for objects to provide a reference point other than the receiver's center,for the purpose of centering a submorph under special circumstances, such as BalloonMorph"

	^ self center! !

!Morph methodsFor: 'menus' stamp: 'sw 2/3/2000 00:12'!
adjustedCenter: c
	"Set the receiver's position based on the #adjustedCenter protocol for adhereToEdge.  By default this simply sets the receiver's center.   Though there are (at its inception anyway) no other implementors of this method, it is required in use with the #adhereToEdge when the centering of a submorph is to be with reference to a rectangle  other than the receiver's center."

	self center: c! !

!Morph methodsFor: 'layout' stamp: 'ar 11/12/2000 17:34'!
adjustLayoutBounds
	"Adjust the receivers bounds depending on the resizing strategy imposed"
	| hFit vFit box myExtent extent |
	hFit := self hResizing.
	vFit := self vResizing.
	(hFit == #shrinkWrap or:[vFit == #shrinkWrap]) ifFalse:[^self]. "not needed"
	box := self layoutBounds.
	myExtent := box extent.
	extent := self submorphBounds corner - box origin.
	hFit == #shrinkWrap ifTrue:[myExtent := extent x @ myExtent y].
	vFit == #shrinkWrap ifTrue:[myExtent := myExtent x @ extent y].
	"Make sure we don't get smaller than minWidth/minHeight"
	myExtent x < self minWidth ifTrue:[
		myExtent := (myExtent x max: 
			(self minWidth - self bounds width + self layoutBounds width)) @ myExtent y].
	myExtent y < self minHeight ifTrue:[
		myExtent := myExtent x @ (myExtent y max:
			(self minHeight - self bounds height + self layoutBounds height))].
	self layoutBounds: (box origin extent: myExtent).! !

!Morph methodsFor: 'accessing' stamp: 'ar 12/18/2001 20:09'!
adoptPaneColor: paneColor
	self submorphsDo:[:m| m adoptPaneColor: paneColor].! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/11/2000 16:34'!
alarmScheduler
	"Return the scheduler being responsible for triggering alarms"
	^self world! !

!Morph methodsFor: 'geometry' stamp: 'di 7/24/97 11:55'!
align: aPoint1 with: aPoint2
	"Translate by aPoint2 - aPoint1."

	^ self position: self position + (aPoint2 - aPoint1)! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'di 11/14/2001 12:50'!
allKnownNames
	"Return a list of all known names based on the scope of the receiver.  Does not include the name of the receiver itself.  Items in parts bins are excluded.  Reimplementors (q.v.) can extend the list"

	^ Array streamContents:
		[:s | self allSubmorphNamesDo: [:n | s nextPut: n]]
! !

!Morph methodsFor: 'menus' stamp: 'ar 10/5/2000 17:20'!
allMenuWordings
	| tempMenu |
	tempMenu := self buildHandleMenu: self currentHand.
	tempMenu allMorphsDo: [:m | m step].  "Get wordings current"
	^ tempMenu allWordings! !

!Morph methodsFor: 'submorphs-accessing' stamp: ''!
allMorphs
	"Return a collection containing all morphs in this composite morph (including the receiver)."

	| all |
	all := OrderedCollection new: 100.
	self allMorphsDo: [: m | all add: m].
	^ all! !

!Morph methodsFor: 'e-toy support' stamp: 'yo 1/9/2004 16:10'!
allMorphsAndBookPagesInto: aSet
	"Return a set of all submorphs.  Don't forget the hidden ones like BookMorph pages that are not showing.  Consider only objects that are in memory (see allNonSubmorphMorphs)." 

	submorphs do: [:m | m allMorphsAndBookPagesInto: aSet].
	self allNonSubmorphMorphs do: [:m | 
			(aSet includes: m) ifFalse: ["Stop infinite recursion"
				m allMorphsAndBookPagesInto: aSet]].
	aSet add: self.
	self player ifNotNil:
		[self player allScriptEditors do: [:e | e allMorphsAndBookPagesInto: aSet]].
	^ aSet! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:27'!
allMorphsDo: aBlock 
	"Evaluate the given block for all morphs in this composite morph (including the receiver)."

	submorphs do: [:m | m allMorphsDo: aBlock].
	aBlock value: self! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ul 12/12/2009 14:09'!
allMorphsWithPlayersDo: aTwoArgumentBlock 
	"Evaluate the given block for all morphs in this composite morph that have non-nil players.
	Also evaluate the block for the receiver if it has a player."

	submorphs do: [:m | m allMorphsWithPlayersDo: aTwoArgumentBlock ].
	self playerRepresented ifNotNil: [ :p | aTwoArgumentBlock value: self value: p ].
! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'sw 10/31/97 20:05'!
allNonSubmorphMorphs
	"Return a collection containing all morphs in this morph which are not currently in the submorph containment hierarchy (put in primarily for bookmorphs)"

	^ OrderedCollection new! !

!Morph methodsFor: 'structure' stamp: 'di 11/13/2000 01:00'!
allOwners
	"Return the owners of the reciever"

	^ Array streamContents: [:strm | self allOwnersDo: [:m | strm nextPut: m]]! !

!Morph methodsFor: 'structure' stamp: 'ar 9/14/2000 16:47'!
allOwnersDo: aBlock
	"Evaluate aBlock with all owners of the receiver"
	owner ifNotNil:[^owner withAllOwnersDo: aBlock].! !

!Morph methodsFor: 'geniestubs' stamp: 'nk 3/10/2004 09:58'!
allowsGestureStart: evt
	^false! !

!Morph methodsFor: 'debug and other' stamp: 'nice 12/27/2009 03:10'!
allStringsAfter: aSubmorph 
	"return an OrderedCollection of strings of text in my submorphs.  If aSubmorph is non-nil, begin with that container."

	| list ok |
	list := OrderedCollection new.
	ok := aSubmorph isNil.
	self allMorphsDo: 
			[:sub | | string | 
			ok ifFalse: [ok := sub == aSubmorph].	"and do this one too"
			ok 
				ifTrue: 
					[(string := sub userString) ifNotNil: 
							[string isString ifTrue: [list add: string] ifFalse: [list addAll: string]]]].
	^list! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ul 12/12/2009 14:09'!
allSubmorphNamesDo: nameBlock
	"Return a list of all known names of submorphs and nested submorphs of the receiver, based on the scope of the receiver.  Items in parts bins are excluded"

	self isPartsBin ifTrue: [^ self]. "Don't report names from parts bins"
	self submorphsDo: 
		[:m | m knownName ifNotNil: [:n | nameBlock value: n].
		m allSubmorphNamesDo: nameBlock].
! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/7/2000 16:27'!
altSpecialCursor0
	"an arrow"
	^(Form
	extent: 16@16
	depth: 8
	fromArray: #( 0 0 0 0 14869218 3806520034 3806520034 3791650816 14848144 2425393296 2425393378 0 14848144 2425393296 2425414144 0 14848144 2425393296 2430730240 0 14848144 2425393296 3791650816 0 14848144 2425393378 3791650816 0 14848144 2425414370 3806461952 0 14848144 2430788322 3806519808 0 14848144 3791651042 3806520034 0 14848226 0 3806520034 3791650816 14868992 0 14869218 3806461952 14811136 0 58082 3806519808 0 0 226 3806520034 0 0 0 3806520034 0 0 0 14869218)
	offset: 0@0)
! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/7/2000 16:28'!
altSpecialCursor1
	"a star and an arrow"
	^(Form
	extent: 31@26
	depth: 8
	fromArray: #( 14417920 0 0 0 0 0 0 0 3705461980 3705461980 3705405440 0 0 0 0 0 3705461980 3705461980 3705461760 0 0 0 0 0 14474460 3705461980 3705405440 0 0 0 0 0 56540 3705461980 3690987520 0 0 3690987520 0 0 220 3705461980 3705461760 0 0 3690987520 0 0 220 3705405440 3705461980 0 0 3705405440 0 0 0 3705461760 56540 3690987520 220 3705405440 0 0 0 3705405440 220 3705461760 220 3705405440 0 0 0 0 0 14474460 220 3705461760 0 0 0 0 0 56540 3691044060 3705461760 0 0 0 0 0 220 3705461980 3705461760 0 0 0 0 56540 3705461980 3705461980 3705461980 3705461980 3705461760 0 0 220 3705461980 3705461980 3705461980 3705461980 3705461760 0 0 0 3705461980 3705461980 3705461980 3705461980 3705405440 0 0 0 14474460 3705461980 3705461980 3705461980 3690987520 0 0 0 56540 3705461980 3705461980 3705461760 0 0 0 0 220 3705461980 3705461980 3705405440 0 0 0 0 0 3705461980 3705461980 3690987520 0 0 0 0 0 3705461980 3705461980 3705405440 0 0 0 0 220 3705461980 3705461980 3705405440 0 0 0 0 220 3705461980 3705461980 3705405440 0 0 0 0 220 3705461980 14474460 3705405440 0 0 0 0 220 3705405440 220 3705461760 0 0 0 0 56540 3690987520 0 3705461760 0 0 0 0 56540 0 0 14474240 0)
	offset: 0@0)! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/7/2000 16:41'!
altSpecialCursor2
	| f |
	"a blue box with transparent center"
	f := Form extent: 32@32 depth: 32.
	f offset: (f extent // 2) negated.
	f fill: f boundingBox rule: Form over fillColor: (Color blue alpha: 0.5).
	f fill: (f boundingBox insetBy: 4) rule: Form over fillColor: Color transparent.
	^f
! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/7/2000 16:42'!
altSpecialCursor3
	
	^self altSpecialCursor3: Color blue! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/7/2000 16:41'!
altSpecialCursor3: aColor
	| f box |
	"a bulls-eye pattern in this color"
	f := Form extent: 32@32 depth: 32.
	f offset: (f extent // 2) negated.
	box := f boundingBox.
	[ box width > 0] whileTrue: [
		f fill: box rule: Form over fillColor: aColor.
		f fill: (box insetBy: 2) rule: Form over fillColor: Color transparent.
		box := box insetBy: 4.
	].
	^f
! !

!Morph methodsFor: 'meta-actions' stamp: 'sw 1/29/2001 02:50'!
applyStatusToAllSiblings: evt
	"Apply the statuses of all my scripts to the script status of all my siblings"

	| aPlayer |
	(aPlayer := self topRendererOrSelf player) belongsToUniClass ifFalse: [self error: 'not uniclass'].
	aPlayer instantiatedUserScriptsDo: 
		[:aScriptInstantiation | aScriptInstantiation assignStatusToAllSiblings]! !

!Morph methodsFor: 'updating' stamp: 'mt 7/1/2016 17:39'!
applyUserInterfaceTheme

	"Re-initialize the shadow. Ensure shadow cache invalidation."
	self hasDropShadow ifTrue: [
		self addDropShadow.
		self removeProperty: #dropShadow.
		self changed].! !

!Morph methodsFor: 'drawing' stamp: 'di 6/24/1998 14:10'!
areasRemainingToFill: aRectangle
	"May be overridden by any subclasses with opaque regions"

	^ Array with: aRectangle! !

!Morph methodsFor: 'stepping and presenter' stamp: 'sw 3/22/2000 14:27'!
arrangeToStartStepping
	"Arrange to start getting sent the 'step' message, but don't do that initial #step call that startStepping does"

	self arrangeToStartSteppingIn: self world! !

!Morph methodsFor: 'stepping and presenter' stamp: 'sw 3/22/2000 14:26'!
arrangeToStartSteppingIn: aWorld
	"Start getting sent the 'step' message in aWorld.  Like startSteppingIn:, but without the initial one to get started'"
	aWorld ifNotNil:
		[aWorld startStepping: self.
		self changed]! !

!Morph methodsFor: 'converting' stamp: 'mt 2/17/2016 15:21'!
asDraggableMorph
	"Huge or complex morphs have a serious impact on performance while being dragged. Use the thumbnail version."

	^ self thumbnail asMorph! !

!Morph methodsFor: 'creation' stamp: 'tk 2/6/1999 22:43'!
asMorph
	^ self! !

!Morph methodsFor: 'e-toy support' stamp: ''!
asNumber: aPointOrNumber
	"Support for e-toy demo."

	aPointOrNumber class = Point
		ifTrue: [^ aPointOrNumber r]
		ifFalse: [^ aPointOrNumber].
! !

!Morph methodsFor: '*morphic-Postscript Canvases' stamp: ''!
asPostscript
	^self asEPS.
! !

!Morph methodsFor: 'converting' stamp: 'wiz 2/19/2006 19:01'!
asSnapshotThumbnail
	^(ThumbnailImageMorph new  newImage: self imageForm ) extent: 90 asPoint .! !

!Morph methodsFor: 'accessing - extension' stamp: 'md 2/27/2006 08:46'!
assureExtension
	"creates an extension for the receiver if needed"
	extension ifNil: [self initializeExtension].
	^ extension! !

!Morph methodsFor: 'player' stamp: 'sw 8/10/2000 00:06'!
assureExternalName
	| aName |
	^ (aName := self knownName) ifNil:
		[self setNameTo: (aName := self externalName).
		^ aName]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:56'!
assureLayoutProperties
	| props |
	props := self layoutProperties.
	props == self ifTrue:[props := nil].
	props ifNil:[
		props := LayoutProperties new initializeFrom: self.
		self layoutProperties: props].
	^props! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:57'!
assureTableProperties
	| props |
	props := self layoutProperties.
	props == self ifTrue:[props := nil].
	props ifNil:[
		props := TableLayoutProperties new initializeFrom: self.
		self layoutProperties: props].
	props includesTableProperties 
		ifFalse:[self layoutProperties: (props := props asTableLayoutProperties)].
	^props! !

!Morph methodsFor: 'fileIn/out' stamp: 'di 11/18/1999 08:35'!
attachToResource
	"Produce a morph from a file -- either a saved .morph file or a graphics file"

	| pathName |
	pathName := Utilities chooseFileWithSuffixFromList: (#('.morph'), Utilities graphicsFileSuffixes)
			withCaption: 'Choose a file
to load'.
	pathName ifNil: [^ self].  "User made no choice"
	pathName == #none ifTrue: [^ self inform: 
'Sorry, no suitable files found
(names should end with .morph, .gif,
.bmp, .jpeg, .jpe, .jp, or .form)'].

	self setProperty: #resourceFilePath toValue: pathName! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 12/20/1999 17:36'!
automaticViewing
	"Backstop, in case this message gets sent to an owner that is not a playfield"
	^ false! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sma 11/11/2000 14:54'!
balloonColor
	^ self
		valueOfProperty: #balloonColor
		ifAbsent: [self defaultBalloonColor]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sma 11/11/2000 14:55'!
balloonColor: aColor
	^ self
		setProperty: #balloonColor
		toValue: aColor! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sd 12/5/2001 20:29'!
balloonFont
	^ self
		valueOfProperty: #balloonFont
		ifAbsent: [self defaultBalloonFont]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sd 12/5/2001 20:30'!
balloonFont: aFont 
	^ self setProperty: #balloonFont toValue: aFont! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 2/7/2000 11:27'!
balloonHelpAligner
	"Answer the morph to which the receiver's balloon help should point"
	^ (self valueOfProperty: #balloonTarget) ifNil: [self]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'dgd 9/7/2004 18:35'!
balloonHelpDelayTime
	"Return the number of milliseconds before a balloon help should be put up on the receiver. The balloon help will only be put up if the receiver responds to #wantsBalloon by returning true."
	^ Preferences balloonHelpDelayTime! !

!Morph methodsFor: 'halos and balloon help' stamp: 'fbs 12/31/2013 12:09'!
balloonHelpTextForHandle: aHandle 
	"Answer a string providing balloon help for the
	given halo handle"
	| itsSelector |
	itsSelector := aHandle eventHandler firstMouseSelector.
	itsSelector == #doRecolor:with:
		ifTrue: [^ Preferences propertySheetFromHalo
				ifTrue: ['Open a property sheet.']
				ifFalse: ['Change color']].
	itsSelector == #mouseDownInDimissHandle:with:
		ifTrue: [^ TrashCanMorph preserveTrash
				ifTrue: ['Move to trash']
				ifFalse: ['Remove from screen']].
	#(#(#addFullHandles 'More halo handles') #(#addSimpleHandles 'Fewer halo handles') #(#chooseEmphasisOrAlignment 'Emphasis & alignment') #(#chooseFont 'Change font') #(#chooseNewGraphicFromHalo 'Choose a new graphic') #(#chooseStyle 'Change style') #(#dismiss 'Remove') #(#doDebug:with: 'Debug') #(#doDirection:with: 'Choose forward direction') #(#doDup:with: 'Duplicate') #(#doMakeSibling:with: 'Make a sibling') #(#doMenu:with: 'Menu') #(#doGrab:with: 'Pick up') #(#editButtonsScript 'See the script for this button') #(#editDrawing 'Repaint') #(#doDupOrMakeSibling:with: 'Duplicate (press shift to make a sibling)') #(#doMakeSiblingOrDup:with: 'Make a sibling (press shift to make simple duplicate)') #(#makeNascentScript 'Make a scratch script') #(#makeNewDrawingWithin 'Paint new object') #(#mouseDownInCollapseHandle:with: 'Collapse') #(#mouseDownOnHelpHandle: 'Help') #(#openViewerForArgument 'Open a Viewer for me. Press shift for a snapshot.') #(#openViewerForTarget:with: 'Open a Viewer for me. Press shift for a snapshot.') #(#paintBackground 'Paint background') #(#prepareToTrackCenterOfRotation:with: 'Move object or set center of rotation') #(#presentViewMenu 'Present the Viewing menu') #(#startDrag:with: 'Move') #(#startGrow:with: 'Change size') #(#startRot:with: 'Rotate') #(#startScale:with: 'Change scale') #(#tearOffTile 'Make a tile representing this object') #(#tearOffTileForTarget:with: 'Make a tile representing this object') #(#trackCenterOfRotation:with: 'Set center of rotation') )
		do: [:pair | itsSelector == pair first
				ifTrue: [^ pair last]].
	^ 'unknown halo handle'translated! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 3/30/2015 14:20'!
balloonMorphClass

	^ NewBalloonMorph useNewBalloonMorph
		ifTrue: [NewBalloonMorph]
		ifFalse: [BalloonMorph]! !

!Morph methodsFor: 'accessing' stamp: 'mt 3/30/2015 11:10'!
balloonText
	"Answer balloon help text or nil, if no help is available.  
	NB: subclasses may override such that they programatically  
	construct the text, for economy's sake, such as model phrases in 
	a Viewer"

	| result |
	extension ifNil: [^nil].
	
	extension balloonText
		ifNotNil: [:balloonText | result := balloonText]
		ifNil: [extension balloonTextSelector
			ifNotNil: [:balloonSelector |
				result := ScriptingSystem helpStringOrNilFor: balloonSelector.
				(result isNil and: [balloonSelector == #methodComment]) 
					ifTrue: [result := self methodCommentAsBalloonHelp].
				((result isNil and: [balloonSelector numArgs = 0]) 
					and: [self respondsTo: balloonSelector]) 
						ifTrue: [result := self perform: balloonSelector]]].
	^ result! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:52'!
balloonTextSelector
	"Answer balloon text selector item in the extension, nil if none"
	^ extension ifNotNil: [extension balloonTextSelector]! !

!Morph methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:50'!
balloonTextSelector: aSelector 
	"change the receiver's balloonTextSelector"
	self assureExtension balloonTextSelector: aSelector! !

!Morph methodsFor: 'accessing' stamp: 'mt 3/30/2015 08:58'!
balloonText: aString

	self assureExtension balloonText: aString.! !

!Morph methodsFor: 'initialization' stamp: 'cmm 5/8/2014 10:57'!
basicInitialize
	"Do basic generic initialization of the instance variables:  
	Set up the receiver, created by a #basicNew and now ready to  
	be initialized, by placing initial values in the instance variables  
	as appropriate"
	owner := nil.
	submorphs := Array empty.
	bounds := self defaultBounds.
	color := self defaultColor! !

!Morph methodsFor: 'user interface' stamp: 'tak 3/15/2005 17:36'!
becomeModal
	self currentWorld
		ifNotNil: [self currentWorld modalWindow: self]! !

!Morph methodsFor: 'accessing' stamp: 'sw 10/31/2001 21:06'!
beFlap: aBool
	"Mark the receiver with the #flap property, or unmark it"

	aBool
		ifTrue:
			[self setProperty: #flap toValue: true.
			self hResizing: #rigid.
			self vResizing: #rigid]
		ifFalse:
			[self removeProperty: #flap]! !

!Morph methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:57'!
beSticky
	"make the receiver sticky"
	self assureExtension sticky: true! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 10/5/2000 16:51'!
beThisWorldsModel

	self world setModel: self.
	self model: nil slotName: nil.	"A world's model cannot have another model"! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:52'!
beUnsticky
	"If the receiver is marked as sticky, make it now be unsticky"
	extension ifNotNil: [extension sticky: false]! !

!Morph methodsFor: 'meta-actions' stamp: 'mt 11/4/2015 21:00'!
blueButtonDown: anEvent
	"Special gestures (cmd-mouse on the Macintosh; Alt-mouse on Windows and Unix) allow a mouse-sensitive morph to be moved or bring up a halo for the morph."
	| h tfm doNotDrag |
	h := anEvent hand halo.
	"Prevent wrap around halo transfers originating from throwing the event back in"
	doNotDrag := false.
	h ifNotNil:[
		(h innerTarget == self) ifTrue:[doNotDrag := true].
		(h innerTarget hasOwner: self) ifTrue:[doNotDrag := true].
		(self hasOwner: h target) ifTrue:[doNotDrag := true]].

	tfm := (self transformedFrom: nil) inverseTransformation.

	"cmd-drag on flexed morphs works better this way"
	h := self addHalo: (anEvent transformedBy: tfm).
	h ifNil: [^ self].
	doNotDrag ifTrue:[^self].
	"Initiate drag transition if requested"
	anEvent hand 
		waitForClicksOrDrag: h
		event: (anEvent transformedBy: tfm)
		selectors: { nil. nil. nil. #startDragTarget:. }
		threshold: HandMorph dragThreshold.
	"Pass focus explicitly here"
	anEvent hand newMouseFocus: h.! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 9/15/2000 20:25'!
blueButtonUp: anEvent
	"Ignored. Theoretically we should never get here since control is transferred to the halo on #blueButtonDown: but subclasses may implement this differently."! !

!Morph methodsFor: 'accessing' stamp: 'ar 8/25/2001 18:28'!
borderColor
	^self borderStyle color! !

!Morph methodsFor: 'accessing' stamp: 'nk 4/15/2004 10:55'!
borderColor: aColorOrSymbolOrNil 
	"Unfortunately, the argument to borderColor could be more than 	just a color. 
	It could also be a symbol, in which case it is to be interpreted as a style identifier.
	But I might not be able to draw that kind of border, so it may have to be ignored.
	Or it could be nil, in which case I should revert to the default border."

	| style newStyle |
	style := self borderStyle.
	style baseColor = aColorOrSymbolOrNil
		ifTrue: [^ self].

	aColorOrSymbolOrNil isColor
		ifTrue: [style style = #none "default border?"
				ifTrue: [self borderStyle: (SimpleBorder width: 0 color: aColorOrSymbolOrNil)]
				ifFalse: [style baseColor: aColorOrSymbolOrNil.
					self changed].
			^ self].

	self
		borderStyle: ( ({ nil. #none } includes: aColorOrSymbolOrNil)
				ifTrue: [BorderStyle default]
				ifFalse: [ "a symbol"
					self doesBevels ifFalse: [ ^self ].
					newStyle := (BorderStyle perform: aColorOrSymbolOrNil)
								color: style color;
								width: style width;
								yourself.
					(self canDrawBorder: newStyle)
						ifTrue: [newStyle]
						ifFalse: [style]])! !

!Morph methodsFor: 'accessing' stamp: 'ar 11/26/2001 14:53'!
borderStyle
	^(self valueOfProperty: #borderStyle ifAbsent:[BorderStyle default]) trackColorFrom: self! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/26/2001 16:18'!
borderStyleForSymbol: aStyleSymbol
	"Answer a suitable BorderStyle for me of the type represented by a given symbol"

	| aStyle existing |
	aStyle := BorderStyle borderStyleForSymbol: aStyleSymbol asSymbol.
	aStyle ifNil: [self error: 'bad style'].
	existing := self borderStyle.
	aStyle width: existing width;
		baseColor: existing baseColor.
	^ (self canDrawBorder: aStyle)
		ifTrue:
			[aStyle]
		ifFalse:
			[nil]! !

!Morph methodsFor: 'accessing' stamp: 'ar 12/11/2001 22:14'!
borderStyle: newStyle
	newStyle = self borderStyle ifFalse:[
		(self canDrawBorder: newStyle) ifFalse:[
			"Replace the suggested border with a simple one"
			^self borderStyle: (BorderStyle width: newStyle width color: (newStyle trackColorFrom: self) color)].
		self setProperty: #borderStyle toValue: newStyle.
		self changed].! !

!Morph methodsFor: 'accessing' stamp: 'ar 8/25/2001 18:28'!
borderWidth
	^self borderStyle width! !

!Morph methodsFor: 'accessing' stamp: 'di 2/6/2001 14:02'!
borderWidthForRounding

	^ self borderWidth! !

!Morph methodsFor: 'accessing' stamp: 'nk 4/14/2004 17:48'!
borderWidth: aNumber
	| style |
	style := self borderStyle.
	style width = aNumber ifTrue: [ ^self ].

	style style = #none
		ifTrue: [ self borderStyle: (SimpleBorder width: aNumber color: Color transparent) ]
		ifFalse: [ style width: aNumber. self changed ].
! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:17'!
bottom
	" Return the y-coordinate of my bottom side "

	^ bounds bottom! !

!Morph methodsFor: 'geometry' stamp: 'di 3/6/2002 13:06'!
bottomCenter

	^ bounds bottomCenter! !

!Morph methodsFor: 'geometry' stamp: 'tk 9/8/97 10:44'!
bottomLeft

	^ bounds bottomLeft! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:08'!
bottomLeft: aPoint
	" Move me so that my bottom left corner is at aPoint. My extent (width & height) are unchanged "

	self position: ((aPoint x) @ (aPoint y - self height)).
! !

!Morph methodsFor: 'geometry' stamp: 'di 6/12/97 11:17'!
bottomRight

	^ bounds bottomRight! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:09'!
bottomRight: aPoint
	" Move me so that my bottom right corner is at aPoint. My extent (width & height) are unchanged "

	self position: ((aPoint x - bounds width) @ (aPoint y - self height))
! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:14'!
bottom: aNumber
	" Move me so that my bottom is at the y-coordinate aNumber. My extent (width & height) are unchanged "

	self position: (bounds left @ (aNumber - self height))! !

!Morph methodsFor: 'drawing' stamp: 'sw 6/4/2000 22:02'!
boundingBoxOfSubmorphs
	| aBox |
	aBox := bounds origin extent: self minimumExtent.  "so won't end up with something empty"
	submorphs do:
		[:m | m visible ifTrue: [aBox := aBox quickMerge: m fullBounds]].
	^ aBox
! !

!Morph methodsFor: 'geometry' stamp: 'jm 8/3/97 15:50'!
bounds
	"Return the bounds of this morph."
	"Note: It is best not to override this method because many methods in Morph and its subclasses use the instance variable directly rather than 'self bounds'. Instead, subclasses should be sure that the bounds instance variable is correct."

	^ bounds
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'RAA 7/21/2000 11:10'!
boundsForBalloon

	"some morphs have bounds that are way too big"
	^self boundsInWorld! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:05'!
boundsInWorld
	^self bounds: self bounds in: self world! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:04'!
boundsIn: referenceMorph
	"Return the receiver's bounds as seen by aMorphs coordinate frame"
	^self bounds: self bounds in: referenceMorph! !

!Morph methodsFor: 'drawing' stamp: 'mt 10/6/2015 13:45'!
boundsWithinCorners
	"Return a single sub-rectangle that lies entirely inside corners
	that are made by me.
	Used to identify large regions of window that do not need to be redrawn."

	^ self wantsRoundedCorners
		ifTrue: [self bounds insetBy: 0@self cornerRadius]
		ifFalse: [self bounds]
! !

!Morph methodsFor: 'geometry' stamp: 'ar 12/14/2000 13:48'!
bounds: newBounds
	| oldExtent newExtent |
	oldExtent := self extent.
	newExtent := newBounds extent.
	(oldExtent dotProduct: oldExtent) <= (newExtent dotProduct: newExtent) ifTrue:[
		"We're growing. First move then resize."
		self position: newBounds topLeft; extent: newExtent.
	] ifFalse:[
		"We're shrinking. First resize then move."
		self extent: newExtent; position: newBounds topLeft.
	].! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:04'!
bounds: aRectangle from: referenceMorph
	"Return the receiver's bounds as seen by aMorphs coordinate frame"
	owner ifNil: [^ aRectangle].
	^(owner transformFrom: referenceMorph) globalBoundsToLocal: aRectangle
! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:04'!
bounds: aRectangle in: referenceMorph
	"Return the receiver's bounds as seen by aMorphs coordinate frame"
	owner ifNil: [^ aRectangle].
	^(owner transformFrom: referenceMorph) localBoundsToGlobal: aRectangle
! !

!Morph methodsFor: 'meta-actions' stamp: 'sw 1/19/2001 18:10'!
bringAllSiblingsToMe: evt
	"bring all siblings of the receiver's player found in the same container to the receiver's location."

	| aPlayer aPosition aContainer |
	(aPlayer := self topRendererOrSelf player) belongsToUniClass ifFalse: [self error: 'not uniclass'].
	aPosition := self topRendererOrSelf position.
	aContainer := self topRendererOrSelf owner.
	(aPlayer class allInstances copyWithout: aPlayer) do:
		[:each |
			(aContainer submorphs includes: each costume) ifTrue:
				[each costume  position: aPosition]]! !

!Morph methodsFor: 'debug and other' stamp: 'mt 4/2/2015 10:47'!
buildDebugMenu: aHand
	"Answer a debugging menu for the receiver.  The hand argument is seemingly historical and plays no role presently"

	| aMenu aPlayer |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu addStayUpItem.
	(self hasProperty: #errorOnDraw) ifTrue:
		[aMenu add: 'start drawing again' translated action: #resumeAfterDrawError.
		aMenu addLine].
	(self hasProperty: #errorOnStep) ifTrue:
		[aMenu add: 'start stepping again' translated action: #resumeAfterStepError.
		aMenu addLine].

	aMenu add: 'inspect morph' translated action: #inspectInMorphic:.
	aMenu add: 'inspect owner chain' translated action: #inspectOwnerChain.
	Smalltalk isMorphic ifFalse:
		[aMenu add: 'inspect morph (in MVC)' translated action: #inspect].

	self isMorphicModel ifTrue:
		[aMenu add: 'inspect model' translated target: self model action: #inspect].
	(aPlayer := self player) ifNotNil:
		[aMenu add: 'inspect player' translated target: aPlayer action: #inspect].

     aMenu add: 'explore morph' translated target: self selector: #exploreInMorphic:.

	aMenu addLine.
	aPlayer ifNotNil:
		[ aMenu add: 'viewer for Player' translated target: self player action: #beViewed.
	aMenu balloonTextForLastItem: 'Opens a viewer on my Player -- this is the same thing you get if you click on the cyan "View" halo handle' translated ].

	aMenu add: 'viewer for Morph' translated target: self action: #viewMorphDirectly.
	aMenu balloonTextForLastItem: 'Opens a Viewer on this Morph, rather than on its Player' translated.
	aMenu addLine.

	aPlayer ifNotNil:
		[aPlayer class isUniClass ifTrue: [
			aMenu add: 'browse player class' translated target: aPlayer action: #browseHierarchy]].
	aMenu add: 'browse morph class' translated target: self selector: #browseHierarchy.
	(self isMorphicModel)
		ifTrue: [aMenu
				add: 'browse model class'
				target: self model
				selector: #browseHierarchy].
	aMenu addLine.

	self addViewingItemsTo: aMenu.
	aMenu 
		add: 'make own subclass' translated action: #subclassMorph;
		add: 'save morph in file' translated  action: #saveOnFile;
		addLine;
		add: 'call #tempCommand' translated action: #tempCommand;
		add: 'define #tempCommand' translated action: #defineTempCommand;
		addLine;

		add: 'control-menu...' translated target: self selector: #invokeMetaMenu:;
		add: 'edit balloon help' translated action: #editBalloonHelpText.

	^ aMenu! !

!Morph methodsFor: 'meta-actions' stamp: 'dgd 9/18/2004 17:23'!
buildHandleMenu: aHand
	"Build the morph menu for the given morph's halo's menu handle. This menu has two sections. The first section contains commands that are interpreted by the hand; the second contains commands provided by the target morph. This method allows the morph to decide which items should be included in the hand's section of the menu."

	| menu |

	(Preferences generalizedYellowButtonMenu
			and: [Preferences noviceMode])
		ifTrue: [^ self buildYellowButtonMenu: aHand].

	menu := MenuMorph new defaultTarget: self.
	menu addStayUpItem.
	menu addLine.
	self addStandardHaloMenuItemsTo: menu hand: aHand.
	menu defaultTarget: aHand.
	self addAddHandMenuItemsForHalo: menu  hand: aHand.
	menu defaultTarget: self.
	self addCustomHaloMenuItems: menu hand: aHand.
	menu defaultTarget: aHand.
	^ menu
! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 1/3/2010 17:01'!
buildMetaMenu: evt
	"Build the morph menu. This menu has two sections. The first section contains commands that are handled by the hand; the second contains commands handled by the argument morph."
	| menu |
	menu := MenuMorph new defaultTarget: self.
	menu addStayUpItem.
	menu add: 'grab' translated action: #grabMorph:.
	menu add: 'copy to paste buffer' translated action: #copyToPasteBuffer:.
	self maybeAddCollapseItemTo: menu.
	menu add: 'delete' translated action: #dismissMorph:.
	menu addLine.
	menu add: 'copy text' translated action: #clipText.
	menu add: 'copy Postscript' translated action: #clipPostscript.
	menu add: 'print Postscript to file...' translated action: #printPSToFile.
	menu addLine.
	menu add: 'go behind' translated action: #goBehind.
	menu add: 'add halo' translated action: #addHalo:.
	menu add: 'duplicate' translated action: #maybeDuplicateMorph:.

	self addEmbeddingMenuItemsTo: menu hand: evt hand.

	menu add: 'resize' translated action: #resizeMorph:.
	"Give the argument control over what should be done about fill styles"
	self addFillStyleMenuItems: menu hand: evt hand.
	self addDropShadowMenuItems: menu hand: evt hand.
	self addLayoutMenuItems: menu hand: evt hand.
	menu addUpdating: #hasClipSubmorphsString target: self selector: #changeClipSubmorphs argumentList: #().
	menu addLine.

	(self morphsAt: evt position) size > 1 ifTrue:
		[menu add: 'submorphs...' translated
			target: self
			selector: #invokeMetaMenuAt:event:
			argument: evt position].
	menu addLine.
	menu add: 'inspect' translated selector: #inspectAt:event: argument: evt position.
	menu add: 'explore' translated action: #explore.
	menu add: 'browse hierarchy' translated action: #browseHierarchy.
	menu add: 'make own subclass' translated action: #subclassMorph.
	menu addLine.
	(self isMorphicModel) ifTrue:
		[menu add: 'save morph as prototype' translated action: #saveAsPrototype.
		(self ~~ self world modelOrNil) ifTrue:
			 [menu add: 'become this world''s model' translated action: #beThisWorldsModel]].
	menu add: 'save morph in file' translated action: #saveOnFile.
	(self hasProperty: #resourceFilePath)
		ifTrue: [((self valueOfProperty: #resourceFilePath) endsWith: '.morph')
				ifTrue: [menu add: 'save as resource' translated action: #saveAsResource].
				menu add: 'update from resource' translated action: #updateFromResource]
		ifFalse: [menu add: 'attach to resource' translated action: #attachToResource].
	menu add: 'show actions' translated action: #showActions.
	menu addLine.
	self addDebuggingItemsTo: menu hand: evt hand.

	self addCustomMenuItems: menu hand: evt hand.
	^ menu
! !

!Morph methodsFor: 'menu' stamp: 'dgd 9/18/2004 17:23'!
buildYellowButtonMenu: aHand 
	"build the morph menu for the yellow button"
	| menu |
	menu := MenuMorph new defaultTarget: self.
	self addNestedYellowButtonItemsTo: menu event: ActiveEvent.
	MenuIcons decorateMenu: menu.
	^ menu! !

!Morph methodsFor: 'visual properties' stamp: 'mt 8/5/2016 21:30'!
canApplyUserInterfaceTheme
	^ self isInWorld and: [(self hasProperty: #noUserInterfaceTheme) not]! !

!Morph methodsFor: 'private' stamp: 'cmm 8/25/2013 13:09'!
canBeEncroached
	"Support for the #smartHorizontalSplitters preference."
	^ true! !

!Morph methodsFor: 'testing' stamp: 'RAA 12/4/2000 10:44'!
canDrawAtHigherResolution

	^false! !

!Morph methodsFor: 'testing' stamp: 'ar 8/25/2001 19:14'!
canDrawBorder: aBorderStyle
	"Return true if the receiver can be drawn with the given border style."
	^true! !

!Morph methodsFor: 'visual properties' stamp: 'ar 6/25/1999 11:11'!
canHaveFillStyles
	"Return true if the receiver can have general fill styles; not just colors.
	This method is for gradually converting old morphs."
	^self class == Morph "no subclasses"! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:54'!
cellInset
	"Layout specific. This property specifies an extra inset for each cell in the layout."
	| props |
	props := self layoutProperties.
	^props ifNil:[0] ifNotNil:[props cellInset].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:54'!
cellInset: aNumber
	"Layout specific. This property specifies an extra inset for each cell in the layout."
	self assureTableProperties cellInset: aNumber.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:54'!
cellPositioning
	"Layout specific. This property describes how the receiver should be layed out in its owner when the bounds of the cell assigned to the receiver do not exactly match its bounds. Possible values are:
		#topLeft, #topRight, #bottomLeft, #bottomRight, #topCenter, #leftCenter, #rightCenter, #bottomCenter, #center 
	which align the receiver's bounds with the cell at the given point."
	| props |
	props := self layoutProperties.
	^props ifNil:[#center] ifNotNil:[props cellPositioning].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 02:48'!
cellPositioningString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self cellPositioning! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:39'!
cellPositioning: aSymbol
	"Layout specific. This property describes how the receiver should be layed out in its owner when the bounds of the cell assigned to the receiver do not exactly match its bounds. Possible values are:
		#topLeft, #topRight, #bottomLeft, #bottomRight, #topCenter, #leftCenter, #rightCenter, #bottomCenter, #center 
	which align the receiver's bounds with the cell at the given point."
	self assureTableProperties cellPositioning: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:55'!
cellSpacing
	"Layout specific. This property describes how the cell size for each element in a list should be computed.
		#globalRect - globally equal rectangular cells
		#globalSquare - globally equal square cells
		#localRect - locally (e.g., per row/column) equal rectangular cells
		#localSquare - locally (e.g., per row/column) equal square cells
		#none - cells are sized based on available row/column constraints
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#none] ifNotNil:[props cellSpacing].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 02:47'!
cellSpacingString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self cellSpacing! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:56'!
cellSpacing: aSymbol
	"Layout specific. This property describes how the cell size for each element in a list should be computed.
		#globalRect - globally equal rectangular cells
		#globalSquare - globally equal square cells
		#localRect - locally (e.g., per row/column) equal rectangular cells
		#localSquare - locally (e.g., per row/column) equal square cells
		#none - cells are sized based on available row/column constraints
	"
	self assureTableProperties cellSpacing: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'geometry' stamp: ''!
center

	^ bounds center! !

!Morph methodsFor: 'geometry' stamp: 'sw 6/11/1999 18:48'!
center: aPoint
	self position: (aPoint - (self extent // 2))! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:08'!
changeCellInset: evt
	| handle |
	handle := HandleMorph new forEachPointDo:[:newPoint |
		self cellInset: (newPoint - evt cursorPoint) asIntegerPoint // 5].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 18:54'!
changeClipLayoutCells
	self invalidRect: self fullBounds.
	self clipLayoutCells: self clipLayoutCells not.
	self invalidRect: self fullBounds.! !

!Morph methodsFor: 'drawing' stamp: 'ar 11/4/2000 23:39'!
changeClipSubmorphs
	self clipSubmorphs: self clipSubmorphs not.! !

!Morph methodsFor: 'menus' stamp: 'cmm 12/4/2010 15:10'!
changeColor
	"Change the color of the receiver -- triggered, e.g. from a menu"
	NewColorPickerMorph useIt
		ifTrue: [ (NewColorPickerMorph on: self) openNear: self fullBoundsInWorld ]
		ifFalse:
			[ ColorPickerMorph new
				 choseModalityFromPreference ;
				 sourceHand: self activeHand ;
				 target: self ;
				 selector: #fillStyle: ;
				 originalColor: self color ;
				
				putUpFor: self
				near: self fullBoundsInWorld ]! !

!Morph methodsFor: 'meta-actions' stamp: 'cmm 12/4/2010 15:11'!
changeColorTarget: anObject selector: aSymbol originalColor: aColor hand: aHand 
	"Put up a color picker for changing some kind of color.  May be modal or modeless, depending on #modalColorPickers setting"
	| desiredLoc |
	self flag: #arNote.
	"Simplify this due to anObject == self for almost all cases"
	desiredLoc := anObject isMorph
		ifTrue:
			[ Rectangle
				center: self position
				extent: 20 ]
		ifFalse:
			[ anObject == self world
				ifTrue: [ anObject viewBox bottomLeft + (20 @ -20) extent: 200 ]
				ifFalse: [ anObject fullBoundsInWorld ] ].
	^ NewColorPickerMorph useIt
		ifTrue:
			[ (NewColorPickerMorph
				on: anObject
				originalColor: aColor
				setColorSelector: aSymbol) openNear: desiredLoc ]
		ifFalse:
			[ ColorPickerMorph new
				 choseModalityFromPreference ;
				 sourceHand: aHand ;
				 target: anObject ;
				 selector: aSymbol ;
				 originalColor: aColor ;
				
				putUpFor: anObject
				near: desiredLoc ;
				 yourself ]! !

!Morph methodsFor: 'menus' stamp: 'ar 11/29/2001 19:57'!
changeDirectionHandles
	^self wantsDirectionHandles: self wantsDirectionHandles not! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 10/31/2000 19:19'!
changeDisableTableLayout
	self disableTableLayout: self disableTableLayout not.
	self layoutChanged.! !

!Morph methodsFor: 'text-anchor' stamp: 'aoy 2/15/2003 21:47'!
changeDocumentAnchor
	"Change the anchor from/to document anchoring"

	| newType |
	newType := self textAnchorType == #document 
		ifTrue: [#paragraph]
		ifFalse: [ #document].
	owner isTextMorph 
		ifTrue: 
			[owner 
				anchorMorph: self
				at: self position
				type: newType]! !

!Morph methodsFor: 'menus' stamp: 'ar 11/2/2000 15:04'!
changeDragAndDrop
	^self enableDragNDrop: self dragNDropEnabled not! !

!Morph methodsFor: 'updating' stamp: 'ar 6/25/2001 19:46'!
changed
	"Report that the area occupied by this morph should be redrawn."
	^fullBounds 
		ifNil:[self invalidRect: self outerBounds]
		ifNotNil:[self invalidRect: fullBounds]! !

!Morph methodsFor: 'text-anchor' stamp: 'aoy 2/15/2003 21:48'!
changeInlineAnchor
	"Change the anchor from/to line anchoring"

	| newType |
	newType := self textAnchorType == #inline 
				ifTrue: [#paragraph]
				ifFalse: [#inline]. 
	owner isTextMorph 
		ifTrue: 
			[owner 
				anchorMorph: self
				at: self position
				type: newType]! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:09'!
changeLayoutInset: evt
	| handle |
	handle := HandleMorph new forEachPointDo:[:newPoint |
		self layoutInset: (newPoint - evt cursorPoint) asIntegerPoint // 5].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:10'!
changeListDirection: aSymbol
	| listDir wrapDir |
	self listDirection: aSymbol.
	(self wrapDirection == #none) ifTrue:[^self].
	"otherwise automatically keep a valid table layout"
	listDir := self listDirection.
	wrapDir := self wrapDirection.
	(listDir == #leftToRight or:[listDir == #rightToLeft]) ifTrue:[
		wrapDir == #leftToRight ifTrue:[^self wrapDirection: #topToBottom].
		wrapDir == #rightToLeft ifTrue:[^self wrapDirection: #bottomToTop].
	] ifFalse:[
		wrapDir == #topToBottom ifTrue:[^self wrapDirection: #leftToRight].
		wrapDir == #bottomToTop ifTrue:[^self wrapDirection: #rightToLeft].
	].
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:10'!
changeMaxCellSize: evt
	| handle |
	handle := HandleMorph new forEachPointDo:[:newPoint |
		self maxCellSize: (newPoint - evt cursorPoint) asIntegerPoint].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:10'!
changeMinCellSize: evt
	| handle |
	handle := HandleMorph new forEachPointDo:[:newPoint |
		self minCellSize: (newPoint - evt cursorPoint) asIntegerPoint].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 10/31/2000 19:19'!
changeNoLayout
	self layoutPolicy ifNil:[^self]. "already no layout"
	self layoutPolicy: nil.
	self layoutChanged.! !

!Morph methodsFor: 'text-anchor' stamp: 'aoy 2/15/2003 21:48'!
changeParagraphAnchor
	"Change the anchor from/to paragraph anchoring"

	| newType |
	newType := self textAnchorType == #paragraph 
		ifTrue: [#document]
		ifFalse: [#paragraph].
	owner isTextMorph 
		ifTrue: 
			[owner 
				anchorMorph: self
				at: self position
				type: newType]! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 10/31/2000 19:19'!
changeProportionalLayout
	| layout |
	((layout := self layoutPolicy) notNil and:[layout isProportionalLayout])
		ifTrue:[^self]. "already proportional layout"
	self layoutPolicy: ProportionalLayout new.
	self layoutChanged.! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:10'!
changeReverseCells
	self reverseTableCells: self reverseTableCells not.! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 11/13/2000 19:10'!
changeRubberBandCells
	self rubberBandCells: self rubberBandCells not.! !

!Morph methodsFor: 'drop shadows' stamp: 'cmm 12/4/2010 15:11'!
changeShadowColor
	"Change the shadow color of the receiver -- triggered, e.g. from a menu"
	NewColorPickerMorph useIt
		ifTrue:
			[ (NewColorPickerMorph
				on: self
				originalColor: self shadowColor
				setColorSelector: #shadowColor:) openNearMorph: self ]
		ifFalse:
			[ ColorPickerMorph new
				 choseModalityFromPreference ;
				 sourceHand: self activeHand ;
				 target: self ;
				 selector: #shadowColor: ;
				 originalColor: self shadowColor ;
				
				putUpFor: self
				near: self fullBoundsInWorld ]! !

!Morph methodsFor: 'layout-menu' stamp: 'ar 10/31/2000 19:20'!
changeTableLayout
	| layout |
	((layout := self layoutPolicy) notNil and:[layout isTableLayout])
		ifTrue:[^self]. "already table layout"
	self layoutPolicy: TableLayout new.
	self layoutChanged.! !

!Morph methodsFor: 'menus' stamp: 'sw 12/17/1998 12:09'!
chooseNewGraphic
	"Used by any morph that can be represented by a graphic"
	self chooseNewGraphicCoexisting: false
! !

!Morph methodsFor: 'menus' stamp: 'sw 9/2/2007 19:04'!
chooseNewGraphicCoexisting: aBoolean 
	"Allow the user to choose a different form for her form-based morph"

	| replacee aGraphicalMenu |
	self isInWorld ifFalse: "menu must have persisted for a not-in-world object."
		[aGraphicalMenu := ActiveWorld submorphThat:
				[:m | (m isKindOf: GraphicalMenu) and: [m target == self]]
			 ifNone:
				[^ self].
		^ aGraphicalMenu show; flashBounds].
	aGraphicalMenu := GraphicalMenu new
				initializeFor: self
				withForms: self reasonableForms
				coexist: aBoolean.
	aBoolean
		ifTrue: [self primaryHand attachMorph: aGraphicalMenu]
		ifFalse: [replacee := self topRendererOrSelf.
			replacee owner replaceSubmorph: replacee by: aGraphicalMenu]! !

!Morph methodsFor: 'menus' stamp: 'sw 10/27/2000 17:34'!
chooseNewGraphicFromHalo
	"Allow the user to select a changed graphic to replace the one in the receiver"

	self currentWorld abandonAllHalos.
	self chooseNewGraphicCoexisting: true
! !

!Morph methodsFor: 'accessing' stamp: 'dgd 9/1/2004 16:10'!
clearArea
	"Answer the clear area of the receiver. It means the area free  
	of docking bars."
	| visTop visBottom visLeft visRight |

	visTop := self top.
	visBottom := self bottom.
	visLeft := self left.
	visRight := self right.

	self dockingBars
		do: [:each | 
			(each isAdheringToTop and: [each bottom > visTop])
				ifTrue: [visTop := each bottom].

			(each isAdheringToBottom and: [each top < visBottom])
				ifTrue: [visBottom := each top].

			(each isAdheringToLeft and: [each right > visLeft])
				ifTrue: [visLeft := each right].

			(each isAdheringToRight and: [each left < visRight])
				ifTrue: [visRight := each left]
		].

	^ Rectangle
		left: visLeft
		right: visRight
		top: visTop
		bottom: visBottom
! !

!Morph methodsFor: 'event handling' stamp: 'tk 9/6/2000 12:42'!
click
	"Pretend the user clicked on me."

	(self handlesMouseDown: nil) ifTrue: [
		self mouseDown: nil.
		self mouseUp: nil].! !

!Morph methodsFor: 'event handling' stamp: 'LC 5/18/2000 09:54'!
click: evt
	"Handle a single-click event. This message is only sent to clients that request it by sending #waitForClicksOrDrag:event: to the initiating hand in their mouseDown: method. This default implementation does nothing.
	LC 2/14/2000 08:32 - added: EventHandler notification"

	self eventHandler ifNotNil:
		[self eventHandler click: evt fromMorph: self].! !

!Morph methodsFor: 'drawing' stamp: 'dgd 2/16/2003 20:02'!
clipLayoutCells
	"Drawing/layout specific. If this property is set, clip the  
	submorphs of the receiver by its cell bounds."
	^ self
		valueOfProperty: #clipLayoutCells
		ifAbsent: [false]! !

!Morph methodsFor: 'drawing' stamp: 'ar 10/29/2000 19:22'!
clipLayoutCells: aBool
	"Drawing/layout specific. If this property is set, clip the submorphs of the receiver by its cell bounds."
	aBool == false
		ifTrue:[self removeProperty: #clipLayoutCells]
		ifFalse:[self setProperty: #clipLayoutCells toValue: aBool].
	self changed.! !

!Morph methodsFor: '*morphic-Postscript Canvases' stamp: 'ar 1/16/2001 17:06'!
clipPostscript
	^Clipboard clipboardText: self asPostscript.

! !

!Morph methodsFor: 'drawing' stamp: 'ar 10/29/2000 19:16'!
clippingBounds
	"Return the bounds to which any submorphs should be clipped if the property is set"
	^self innerBounds! !

!Morph methodsFor: 'drawing' stamp: 'md 4/3/2006 11:53'!
clipSubmorphs
	"Drawing specific. If this property is set, clip the receiver's  
	submorphs to the receiver's clipping bounds."
	
	extension ifNil: [^false].
	^ self
		valueOfProperty: #clipSubmorphs
		ifAbsent: [false]! !

!Morph methodsFor: 'drawing' stamp: 'ar 11/12/2000 18:47'!
clipSubmorphs: aBool
	"Drawing specific. If this property is set, clip the receiver's submorphs to the receiver's clipping bounds."
	self invalidRect: self fullBounds.
	aBool == false
		ifTrue:[self removeProperty: #clipSubmorphs]
		ifFalse:[self setProperty: #clipSubmorphs toValue: aBool].
	self invalidRect: self fullBounds.! !

!Morph methodsFor: 'printing' stamp: 'bf 7/17/2003 12:53'!
clipText
	"Copy the text in the receiver or in its submorphs to the clipboard"
	| content |
	"My own text"
	content := self userString.
	"Or in my submorphs"
	content ifNil: [
		| list |
		list := self allStringsAfter: nil.
		list notEmpty ifTrue: [
			content := String streamContents: [:stream |
				list do: [:each | stream nextPutAll: each; cr]]]].
	"Did we find something?"
	content
		ifNil: [self flash "provide feedback"]
		ifNotNil: [Clipboard clipboardText: content].! !

!Morph methodsFor: 'menus' stamp: 'sw 2/21/2000 15:21'!
collapse
	CollapsedMorph new beReplacementFor: self! !

!Morph methodsFor: 'accessing' stamp: 'tk 2/15/2001 15:55'!
color

	^ color 	"has already been set to ((self valueOfProperty: #fillStyle) asColor)"! !

!Morph methodsFor: 'change reporting' stamp: 'sw 9/10/1998 08:18'!
colorChangedForSubmorph: aSubmorph
	"The color associated with aSubmorph was changed through the UI; react if needed"! !

!Morph methodsFor: 'accessing' stamp: 'ar 8/15/2001 22:40'!
colorForInsets
	"Return the color to be used for shading inset borders.  The default is my own color, but it might want to be, eg, my owner's color.  Whoever's color ends up prevailing, the color itself gets the last chance to determine, so that when, for example, an InfiniteForm serves as the color, callers won't choke on some non-Color object being returned"
	(color isColor and:[color isTransparent and:[owner notNil]]) ifTrue:[^owner colorForInsets].
	^ color colorForInsets
! !

!Morph methodsFor: 'printing' stamp: 'ul 12/9/2010 03:01'!
colorString: aColor 

	aColor ifNil: [ ^'nil' ].
	^aColor name
		ifNil: [ aColor storeString ]
		ifNotNil: [ :colorName | 'Color ', colorName ]! !

!Morph methodsFor: 'accessing' stamp: 'ar 8/6/2001 09:03'!
color: aColor
	"Set the receiver's color.  Directly set the color if appropriate, else go by way of fillStyle"

	(aColor isColor or: [aColor isKindOf: InfiniteForm]) ifFalse:[^ self fillStyle: aColor].
	color = aColor ifFalse:
		[self removeProperty: #fillStyle.
		color := aColor.
		self changed]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'dgd 2/22/2003 14:30'!
comeToFront
	| outerMorph |
	outerMorph := self topRendererOrSelf.
	(outerMorph owner isNil or: [outerMorph owner hasSubmorphs not]) 
		ifTrue: [^self].
	outerMorph owner firstSubmorph == outerMorph 
		ifFalse: [outerMorph owner addMorphFront: outerMorph]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 3/1/2000 11:39'!
comeToFrontAndAddHalo
	self comeToFront.
	self addHalo! !

!Morph methodsFor: 'undo' stamp: 'ar 8/31/2000 23:15'!
commandHistory
	"Return the command history for the receiver"
	| w |
	(w := self world) ifNotNil:[^w commandHistory].
	(w := self currentWorld) ifNotNil:[^w commandHistory].
	^CommandHistory new. "won't really record anything but prevent breaking things"! !

!Morph methodsFor: 'testing' stamp: 'nice 12/27/2009 03:10'!
completeModificationHash

"World completeModificationHash"

	| resultSize result |
	resultSize := 10.
	result := ByteArray new: resultSize.
	self allMorphsDo: [ :each | | here | 
		here := each modificationHash.
		here withIndexDo: [ :ch :index | | i |
			i := index \\ resultSize + 1.
			result at: i put: ((result at: i) bitXor: ch asciiValue)
		].
	].
	^result! !

!Morph methodsFor: 'printing' stamp: ''!
constructorString

	^ String streamContents: [:s | self printConstructorOn: s indent: 0].
! !

!Morph methodsFor: 'structure' stamp: 'mt 5/14/2016 12:37'!
containingWindow
	"Answer a window that contains the receiver. Try to use the model to find the right window. If I do not have a model, use the model of one of my owners. We could also just use #isSystemWindow. This, however, gives system windows the chance to refrain from taking ownership of this morph."

	| component |
	component := self.
	component model isNil ifTrue: [component := self firstOwnerSuchThat: [:m| m model notNil]].
	^(component isNil or: [component isWindowForModel: component model])
		ifTrue: [component]
		ifFalse: [component firstOwnerSuchThat:[:m| m isWindowForModel: component model]]! !

!Morph methodsFor: 'geometry testing' stamp: ''!
containsPoint: aPoint

	^ self bounds containsPoint: aPoint! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/13/2000 17:58'!
containsPoint: aPoint event: anEvent
	"Return true if aPoint is considered to be inside the receiver for the given event.
	The default implementation treats locked children as integral part of their owners."
	(self fullBounds containsPoint: aPoint) ifFalse:[^false].
	(self containsPoint: aPoint) ifTrue:[^true].
	self submorphsDo:[:m|
		(m isLocked and:[m fullContainsPoint: 
			((m transformedFrom: self) globalPointToLocal: aPoint)]) ifTrue:[^true]].
	^false! !

!Morph methodsFor: 'copying' stamp: 'tk 2/19/2001 18:21'!
copy

	^ self veryDeepCopy! !

!Morph methodsFor: 'layout-properties' stamp: 'kfr 4/24/2015 18:16'!
copyLayoutProperties
	^[Clipboard clipboardText: self layoutProperties stringWithLayout] ifError:[ nil ] ! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 10/5/2000 16:44'!
copyToPasteBuffer: evt
	self okayToDuplicate ifTrue:[evt hand copyToPasteBuffer: self].! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'di 10/27/97 23:26'!
copyWithoutSubmorph: sub
	"Needed to get a morph to draw without one of its submorphs.
	NOTE:  This must be thrown away immediately after use."
	^ self clone privateSubmorphs: (submorphs copyWithout: sub)! !

!Morph methodsFor: 'rounding' stamp: 'mt 10/6/2015 13:38'!
cornerRadius

	^ self
		valueOfProperty: #cornerRadius
		ifAbsent: [self class preferredCornerRadius]! !

!Morph methodsFor: 'rounding' stamp: 'mt 10/6/2015 13:38'!
cornerRadius: radius

	self
		setProperty: #cornerRadius
		toValue: radius.

	self changed.! !

!Morph methodsFor: 'rounding' stamp: 'mk 8/7/2005 10:02'!
cornerStyle
	"Returns one of the following symbols:
		#square
		#rounded
	according to the current corner style."

	^ self valueOfProperty: #cornerStyle ifAbsent: [#square]! !

!Morph methodsFor: 'rounding' stamp: 'mk 8/7/2005 10:03'!
cornerStyle: aSymbol
	"This method makes it possible to set up desired corner style. aSymbol has to be one of:
		#square
		#rounded"

	aSymbol == #square
		ifTrue:[self removeProperty: #cornerStyle]
		ifFalse:[self setProperty: #cornerStyle toValue: aSymbol].
	self changed! !

!Morph methodsFor: 'accessing' stamp: 'ar 12/27/2001 17:56'!
couldHaveRoundedCorners
	^ true! !

!Morph methodsFor: 'testing' stamp: 'sw 3/30/2007 04:52'!
couldMakeSibling
	"Answer whether it is appropriate to ask the receiver to make a sibling"

	^ self isWorldMorph not! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 11/4/2015 20:09'!
createHalo

	^ (Smalltalk at: self haloClass ifAbsent: [HaloMorph]) new
		bounds: self worldBoundsForHalo
		yourself! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 10/22/1998 20:28'!
currentPlayerDo: aBlock
	"If the receiver is a viewer/scriptor associated with a current Player object, evaluate the given block against that object"! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 9/8/2000 16:34'!
cursor
	"vacuous backstop in case it gets sent to a morph that doesn't know what to do with it"

	^ 1! !

!Morph methodsFor: 'event handling' stamp: 'sw 3/8/1999 00:17'!
cursorPoint
	^ self currentHand lastEvent cursorPoint! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 9/7/2000 09:28'!
cursor: aNumber
	"vacuous backstop in case it gets sent to a morph that doesn't know what to do with it"
! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 9/13/2002 17:44'!
decimalPlacesForGetter: aGetter
	"Answer the decimal places I prefer for showing a slot with the given getter, or nil if none"

	| decimalPrefs |
	decimalPrefs := self renderedMorph valueOfProperty: #decimalPlacePreferences ifAbsent: [^ nil].
	^ decimalPrefs at: aGetter ifAbsent: [nil]! !

!Morph methodsFor: 'copying' stamp: 'tk 2/14/2001 12:47'!
deepCopy

	self error: 'Please use veryDeepCopy'.
! !

!Morph methodsFor: 'menus' stamp: ''!
defaultArrowheadSize
	
	^ self class defaultArrowheadSize! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sma 11/11/2000 16:15'!
defaultBalloonColor
	^ Display depth <= 2
		ifTrue: [Color white]
		ifFalse: [BalloonMorph balloonColor]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sd 12/5/2001 20:23'!
defaultBalloonFont
	^ BalloonMorph balloonFont! !

!Morph methodsFor: 'visual properties' stamp: 'nk 8/28/2003 15:56'!
defaultBitmapFillForm
	^ImageMorph defaultForm.
! !

!Morph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 15:06'!
defaultBounds
"answer the default bounds for the receiver"
	^ 0 @ 0 corner: 50 @ 40! !

!Morph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:28'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color blue! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/13/2000 14:51'!
defaultEventDispatcher
	"Return the default event dispatcher to use with events that are directly sent to the receiver"
	^MorphicEventDispatcher new! !

!Morph methodsFor: 'user interface' stamp: 'mt 11/13/2015 10:35'!
defaultLabelForInspector
	"Answer the default label to be used for an Inspector window on the receiver."
	^ self printStringLimitedTo: 40! !

!Morph methodsFor: 'accessing' stamp: 'ar 1/3/2010 18:05'!
defaultNameStemForInstances
	^self class name! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 10/24/2000 05:52'!
defaultValueOrNil
	"If the receiver has a property named #defaultValue, return that property's value, else return nil"

	^ self valueOfProperty: #defaultValue ifAbsent: [nil]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 1/11/2000 18:24'!
defersHaloOnClickTo: aSubMorph
	"If a cmd-click on aSubMorph would make it a preferred recipient of the halo, answer true"
	"May want to add a way (via a property) for morphs to assert true here -- this would let certain kinds of morphs that are unusually reluctant to take the halo on initial click"

	^ false
	! !

!Morph methodsFor: 'debug and other' stamp: 'ar 9/27/2005 20:29'!
defineTempCommand
	"To use this, comment out what's below here, and substitute your own code.
You will then be able to invoke it from the standard debugging menus.  If invoked from the world menu, you'll always get it invoked on behalf of the world, but if invoked from an individual morph's meta-menu, it will be invoked on behalf of that individual morph.

Note that you can indeed reimplement tempCommand in an individual morph's class if you wish"

	ToolSet browse: Morph
		selector: #tempCommand! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'mt 5/1/2016 18:16'!
delete
	"Remove the receiver as a submorph of its owner and make its 
	new owner be nil."

	| aWorld |
	self removeHalo.

	self isInWorld ifTrue: [
		self disableSubmorphFocusForHand: self activeHand.
		self activeHand
	  		releaseKeyboardFocus: self;
			releaseMouseFocus: self].

	"Preserve world reference for player notificaiton. See below."
	aWorld := self world ifNil: [World].
	
	owner ifNotNil:[
		self privateDelete.
		self player ifNotNil: [:player |
			player noteDeletionOf: self fromWorld: aWorld]].! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 1/19/2001 07:51'!
deleteAnyMouseActionIndicators

	self changed.
	(self valueOfProperty: #mouseActionIndicatorMorphs ifAbsent: [#()]) do: [ :each |
		each deleteWithSiblings		"one is probably enough, but be safe"
	].
	self removeProperty: #mouseActionIndicatorMorphs.
	self hasRolloverBorder: false.
	self removeProperty: #rolloverWidth.
	self removeProperty: #rolloverColor.
	self layoutChanged.
	self changed.

! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 10/3/2000 17:03'!
deleteBalloon
	"If I am showing a balloon, delete it."
	| w |
	w := self world ifNil:[^self].
	w deleteBalloonTarget: self.! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'dgd 9/1/2004 16:26'!
deleteDockingBars
	"Delete the receiver's docking bars"
	self dockingBars
		do: [:each | each delete]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 7/3/1998 11:02'!
deleteSubmorphsWithProperty: aSymbol
	submorphs copy do:
		[:m | (m hasProperty: aSymbol) ifTrue: [m delete]]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'cmm 5/10/2015 21:16'!
deleteUnlessHasFocus
	"Runs on a step timer because we cannot be guaranteed to get focus change events."
	(ActiveHand keyboardFocus ~= self and: [ self isInWorld ]) ifTrue:
		[ self
			 stopSteppingSelector: #deleteUnlessHasFocus ;
			 delete ]! !

!Morph methodsFor: 'classification' stamp: 'sw 2/26/2002 23:29'!
demandsBoolean
	"Answer whether the receiver will only accept a drop if it is boolean-valued.  Particular to tile-scripting."

	^ self hasProperty: #demandsBoolean! !

!Morph methodsFor: 'thumbnail' stamp: 'sw 10/26/2000 08:32'!
demandsThumbnailing
	"Answer whether the receiver, if in a thumbnailable parts bin, wants to be thumbnailed whether or not size requires it"

	^ false! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:41'!
disableDragNDrop
	self enableDragNDrop: false! !

!Morph methodsFor: 'dispatching' stamp: 'nk 2/15/2004 09:16'!
disableSubmorphFocusForHand: aHandMorph
	"Check whether this morph or any of its submorph has the Genie focus.
	If yes, disable it."
! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:56'!
disableTableLayout
	"Layout specific. Disable laying out the receiver in table layout"
	| props |
	props := self layoutProperties.
	^props ifNil:[false] ifNotNil:[props disableTableLayout].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:06'!
disableTableLayout: aBool
	"Layout specific. Disable laying out the receiver in table layout"
	self assureLayoutProperties disableTableLayout: aBool.
	self layoutChanged.! !

!Morph methodsFor: 'meta-actions' stamp: 'bf 3/1/2012 18:38'!
dismissMorph
	"This is called from an explicit halo destroy/delete action.
	So first disconnect all constraints to keep the graph up to date."

	(self respondsTo: #disconnectAllConstraints) "Connectors package"
		ifTrue: [ self perform: #disconnectAllConstraints ].
	(self respondsTo: #releaseGraphModels) "CGPrereqs package (Connectors)"
		ifTrue: [ self perform: #releaseGraphModels ].
	self world ifNotNilDo: 
			[:w |  w abandonAllHalos; stopStepping: self].
	self delete! !

!Morph methodsFor: 'meta-actions' stamp: 'mir 3/17/2006 18:01'!
dismissMorph: evt
	self dismissMorph! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'fbs 12/31/2013 12:08'!
dismissViaHalo
	"The user has clicked in the delete halo-handle.  This provides a hook in case some concomitant action should be taken, or if the particular morph is not one which should be put in the trash can, for example."

	| cmd |
	self setProperty: #lastPosition toValue: self positionInWorld.
	self dismissMorph.
	TrashCanMorph preserveTrash ifTrue: [ 
		TrashCanMorph slideDismissalsToTrash
			ifTrue:[self slideToTrash: nil]
			ifFalse:[TrashCanMorph moveToTrash: self].
	].

	cmd := Command new cmdWording: 'dismiss ' translated, self externalName.
	cmd undoTarget: ActiveWorld selector: #reintroduceIntoWorld: argument: self.
	cmd redoTarget: ActiveWorld selector: #onceAgainDismiss: argument: self.
	ActiveWorld rememberCommand: cmd! !

!Morph methodsFor: 'button' stamp: 'sw 2/6/2001 23:09'!
doButtonAction
	"If the receiver has a button-action defined, do it now.  The default button action of any morph is, well, to do nothing.  Note that there are several ways -- too many ways -- for morphs to have button-like actions.  This one refers not to the #mouseUpCodeToRun feature, nor does it refer to the Player-scripting mechanism.  Instead it is intended for morph classes whose very nature is to be buttons -- this method provides glue so that arbitrary buttons on the UI can be 'fired' programatticaly from user scripts"! !

!Morph methodsFor: 'user interface' stamp: 'tak 3/15/2005 17:10'!
doCancel
	self delete! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 8/31/2004 16:53'!
dockingBars
	"Answer the receiver's dockingBars"
	^ self submorphs
		select: [:each | each isDockingBar]
! !

!Morph methodsFor: 'accessing' stamp: 'nk 4/15/2004 07:50'!
doesBevels
	"To return true means that this object can show bevelled borders, and
	therefore can accept, eg, #raised or #inset as valid borderColors.
	Must be overridden by subclasses that do not support bevelled borders."

	^ false! !

!Morph methodsFor: 'drawing' stamp: 'tk 8/2/1998 14:33'!
doesOwnRotation
	"Some morphs don't want to TransformMorph to rotate their images, but we do"
	^ false! !

!Morph methodsFor: 'layout' stamp: 'dgd 2/22/2003 14:31'!
doLayoutIn: layoutBounds 
	"Compute a new layout based on the given layout bounds."

	"Note: Testing for #bounds or #layoutBounds would be sufficient to
	figure out if we need an invalidation afterwards but #outerBounds
	is what we need for all leaf nodes so we use that."

	| layout box priorBounds |
	priorBounds := self outerBounds.
	submorphs isEmpty ifTrue: [^fullBounds := priorBounds].
	"Send #ownerChanged to our children"
	submorphs do: [:m | m ownerChanged].
	layout := self layoutPolicy.
	layout ifNotNil: [layout layout: self in: layoutBounds].
	self adjustLayoutBounds.
	fullBounds := self privateFullBounds.
	box := self outerBounds.
	box = priorBounds 
		ifFalse: [self invalidRect: (priorBounds quickMerge: box)]! !

!Morph methodsFor: 'menus' stamp: 'ar 10/25/2000 23:17'!
doMenuItem: menuString
	| aMenu anItem aNominalEvent aHand |
	aMenu := self buildHandleMenu: (aHand := self currentHand).
	aMenu allMorphsDo: [:m | m step].  "Get wordings current"
	anItem := aMenu itemWithWording: menuString.
	anItem ifNil:
		[^ self player scriptingError: 'Menu item not found: ', menuString].
	aNominalEvent :=  MouseButtonEvent new
		setType: #mouseDown
		position: anItem bounds center
		which: 4 "red"
		buttons: 4 "red"
		hand: aHand
		stamp: nil.
	anItem invokeWithEvent: aNominalEvent! !

!Morph methodsFor: 'event handling' stamp: 'jcg 10/2/2001 09:26'!
doubleClickTimeout: evt
	"Handle a double-click timeout event. This message is only sent to clients that request it by sending #waitForClicksOrDrag:event: to the initiating hand in their mouseDown: method. This default implementation does nothing."

	self eventHandler ifNotNil:
		[self eventHandler doubleClickTimeout: evt fromMorph: self].! !

!Morph methodsFor: 'event handling' stamp: 'LC 5/18/2000 09:54'!
doubleClick: evt
	"Handle a double-click event. This message is only sent to clients that request it by sending #waitForClicksOrDrag:event: to the initiating hand in their mouseDown: method. This default implementation does nothing.
	LC 2/14/2000 08:32 - added: EventHandler notification"

	self eventHandler ifNotNil:
		[self eventHandler doubleClick: evt fromMorph: self].! !

!Morph methodsFor: 'naming' stamp: 'sw 10/31/2000 09:28'!
downshiftedNameOfObjectRepresented
	"Answer the downshiped version of the external name of the object represented"

	^ self nameOfObjectRepresented asLowercase! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:50'!
dragEnabled
	"Get this morph's ability to add and remove morphs via drag-n-drop."
	^(self valueOfProperty: #dragEnabled) == true
! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:18'!
dragEnabled: aBool
	^self enableDrag: aBool! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:20'!
dragNDropEnabled
	"Note: This method is only useful for dragEnabled == dropEnabled at all times"
	self separateDragAndDrop.
	^self dragEnabled and:[self dropEnabled]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 18:36'!
dragSelectionColor
	^ Color magenta! !

!Morph methodsFor: 'drawing' stamp: 'mt 10/6/2015 13:45'!
drawDropHighlightOn: aCanvas

	self highlightedForDrop ifTrue: [
		self wantsRoundedCorners
			ifTrue: [aCanvas frameRoundRect: self fullBounds radius: self cornerRadius width: 1 color: self dropHighlightColor]
			ifFalse: [aCanvas frameRectangle: self fullBounds color: self dropHighlightColor]].! !

!Morph methodsFor: 'drawing' stamp: 'mt 10/6/2015 13:45'!
drawDropShadowOn: aCanvas
	"Rectangular shadow with support for rounded corners."
	
	| shadowBounds |
	shadowBounds := self shadowOffset isRectangle
		ifTrue: [self bounds outsetBy: self shadowOffset]
		ifFalse: [self bounds translateBy: (self shadowOffset negated max: 0@0)].
	
	"Only redraw the shadow if the shadow area is affected."
	((aCanvas clipRect intersects: shadowBounds) and: [((self bounds insetBy: (self wantsRoundedCorners ifFalse: [0] ifTrue: [self cornerRadius])) containsRect: aCanvas clipRect) not])
		ifTrue: [
			(self hasProperty: #dropShadow)
				ifFalse: [self updateDropShadowCache].
			aCanvas
				translucentImage: (self valueOfProperty: #dropShadow)
				at: shadowBounds topLeft].! !

!Morph methodsFor: 'drawing' stamp: 'bf 5/23/2011 11:51'!
drawErrorOn: aCanvas
	"The morph (or one of its submorphs) had an error in its drawing method."
	| saneBounds |
	saneBounds := bounds rounded.
	aCanvas
		frameAndFillRectangle: saneBounds
		fillColor: Color red
		borderWidth: 1
		borderColor: Color yellow.
	aCanvas line: saneBounds topLeft to: saneBounds bottomRight width: 1 color: Color yellow.
	aCanvas line: saneBounds topRight to: saneBounds bottomLeft width: 1 color: Color yellow.! !

!Morph methodsFor: 'drawing' stamp: 'mt 7/25/2016 16:02'!
drawKeyboardFocusIndicationOn: aCanvas

	self wantsRoundedCorners
		ifTrue: [aCanvas frameRoundRect: self bounds radius: self cornerRadius width: self keyboardFocusWidth color: self keyboardFocusColor]
		ifFalse: [aCanvas frameRectangle: self bounds width: self keyboardFocusWidth color: self keyboardFocusColor].! !

!Morph methodsFor: 'drawing' stamp: 'mt 10/6/2015 13:45'!
drawMouseDownHighlightOn: aCanvas

	self highlightedForMouseDown ifTrue: [
		self wantsRoundedCorners
			ifTrue: [aCanvas frameRoundRect: self fullBounds radius: self cornerRadius width: 1 color: self color darker darker]
			ifFalse: [aCanvas frameRectangle: self fullBounds color: self color darker darker]].! !

!Morph methodsFor: 'filter streaming' stamp: 'ar 10/26/2000 19:55'!
drawOnCanvas: aCanvas
	^aCanvas fullDraw: self.
! !

!Morph methodsFor: 'drawing' stamp: 'mt 10/6/2015 13:41'!
drawOn: aCanvas

	self wantsRoundedCorners
		ifTrue: [aCanvas frameAndFillRoundRect: self bounds radius: self cornerRadius fillStyle: self fillStyle borderWidth: self borderStyle width borderColor: self borderStyle color]
		ifFalse: [aCanvas fillRectangle: self bounds fillStyle: self fillStyle borderStyle: self borderStyle].
	
! !

!Morph methodsFor: 'drawing' stamp: 'mt 4/10/2015 17:07'!
drawOverlayOn: aCanvas
	"Draw something over all my submorphs."
	
	self drawDropHighlightOn: aCanvas.
	self drawMouseDownHighlightOn: aCanvas.

	(self indicateKeyboardFocus and: [self hasKeyboardFocus])
		ifTrue: [self drawKeyboardFocusIndicationOn: aCanvas].! !

!Morph methodsFor: '*morphic-Postscript Canvases' stamp: 'mpw 8/10/1930 05:21'!
drawPostscriptOn: aCanvas

	self drawOn:aCanvas.
! !

!Morph methodsFor: 'drawing' stamp: 'mt 2/26/2015 18:39'!
drawRolloverBorderOn: aCanvas 
	| colorToUse offsetToUse myShadow newForm f |
	colorToUse := self
				valueOfProperty: #rolloverColor
				ifAbsent: [Color blue alpha: 0.5].
	offsetToUse := self
				valueOfProperty: #rolloverWidth
				ifAbsent: [10 @ 10].
	self hasRolloverBorder: false.
	myShadow := self shadowForm.
	self hasRolloverBorder: true.
	myShadow offset: 0 @ 0.
	f := ColorForm extent: myShadow extent depth: 1.
	myShadow displayOn: f.
	f colors: {Color transparent. colorToUse}.
	newForm := Form extent: offsetToUse * 2 + myShadow extent depth: 32.
	(WarpBlt toForm: newForm) sourceForm: f;
		 cellSize: 1;
		 combinationRule: 3;
		 copyQuad: f boundingBox innerCorners toRect: newForm boundingBox.
		
	self flag: #roundedCorners. "mt: Check for #wantsRoundedCorners and call appropriate things in canvas."
	aCanvas
		translateBy: offsetToUse negated
		during: [:shadowCanvas | 
			shadowCanvas shadowColor: colorToUse.
			shadowCanvas paintImage: newForm at: self position]! !

!Morph methodsFor: 'drawing' stamp: 'dgd 2/22/2003 14:31'!
drawSubmorphsOn: aCanvas 
	"Display submorphs back to front"

	| drawBlock |
	submorphs isEmpty ifTrue: [^self].
	drawBlock := [:canvas | submorphs reverseDo: [:m | canvas fullDrawMorph: m]].
	self clipSubmorphs 
		ifTrue: [aCanvas clipBy: self clippingBounds during: drawBlock]
		ifFalse: [drawBlock value: aCanvas]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:51'!
dropEnabled
	"Get this morph's ability to add and remove morphs via drag-n-drop."
	^(self valueOfProperty: #dropEnabled) == true
! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:18'!
dropEnabled: aBool
	^self enableDrop: aBool! !

!Morph methodsFor: 'event handling' stamp: 'ar 1/10/2001 21:28'!
dropFiles: anEvent
	"Handle a number of files dropped from the OS"
! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/28/2000 10:52'!
dropHighlightColor
	^ Color blue! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 18:08'!
dropSuccessColor
	^ Color blue! !

!Morph methodsFor: 'copying' stamp: 'sw 4/19/2005 17:02'!
duplicate
	"Make and return a duplicate of the receiver"

	| newMorph aName w aPlayer topRend |
	((topRend := self topRendererOrSelf) ~~ self) ifTrue: [^ topRend duplicate].

	self okayToDuplicate ifFalse: [^ self].
	aName := (w := self world) ifNotNil:
		[w nameForCopyIfAlreadyNamed: self].
	newMorph := self veryDeepCopy.
	aName ifNotNil: [newMorph setNameTo: aName].

	newMorph arrangeToStartStepping.
	newMorph privateOwner: nil. "no longer in world"
	newMorph isPartsDonor: false. "no longer parts donor"
	(aPlayer := newMorph player) belongsToUniClass ifTrue:
		[aPlayer class bringScriptsUpToDate].
	aPlayer ifNotNil: [ActiveWorld presenter flushPlayerListCache].
	^ newMorph! !

!Morph methodsFor: 'copying' stamp: 'nk 3/12/2001 17:07'!
duplicateMorphCollection: aCollection
	"Make and return a duplicate of the receiver"

	| newCollection names |

	names := aCollection collect: [ :ea | | newMorph w |
		(w := ea world) ifNotNil:
			[w nameForCopyIfAlreadyNamed: ea].
	].

	newCollection := aCollection veryDeepCopy.

	newCollection with: names do: [ :newMorph :name |
		name ifNotNil: [ newMorph setNameTo: name ].
		newMorph arrangeToStartStepping.
		newMorph privateOwner: nil. "no longer in world"
		newMorph isPartsDonor: false. "no longer parts donor"
	].

	^newCollection! !

!Morph methodsFor: 'meta-actions' stamp: 'wiz 2/19/2006 12:41'!
duplicateMorphImage: evt 
	"Make and return a imageMorph of the receiver's argument imageForm"
	| dup |
	dup := self asSnapshotThumbnail withSnapshotBorder.
	dup bounds: self bounds.
	evt hand grabMorph: dup from: owner.
	"duplicate was ownerless so use #grabMorph:from: here"
	^ dup! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 11/4/2000 17:56'!
duplicateMorph: evt
	"Make and return a duplicate of the receiver's argument"
	| dup |
	dup := self duplicate.
	evt hand grabMorph: dup from: owner. "duplicate was ownerless so use #grabMorph:from: here"
	^dup! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 8/6/2009 20:42'!
editBalloonHelpContent: aString
	| reply |
	reply := UIManager default
		multiLineRequest: 'Edit the balloon help text for ' translated, self externalName
		centerAt: Sensor cursorPoint
		initialAnswer: (aString ifNil: [self noHelpString] ifNotNil: [aString])
		answerHeight: 200.
	reply ifNil: [^ self].  "User cancelled out of the dialog"
	(reply isEmpty or: [reply asString = self noHelpString])
		ifTrue: [self setBalloonText: nil]
		ifFalse: [self setBalloonText: reply]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sma 12/23/1999 13:24'!
editBalloonHelpText
	"Modify the receiver's balloon help text."

	self editBalloonHelpContent: self balloonText! !

!Morph methodsFor: 'e-toy support' stamp: 'mt 7/2/2016 11:48'!
embeddedInMorphicWindowLabeled: labelString
	| window |
	window := (SystemWindow labelled: labelString) model: nil.
	window setStripeColorsFrom: Color white.
	window addMorph: self frame: (0@0 extent: 1@1).
	^ window! !

!Morph methodsFor: 'meta-actions' stamp: 'cmm 12/4/2010 19:27'!
embedInto: evt
	"Embed the receiver into some other morph"
	|  target morphs |
	morphs := self potentialEmbeddingTargets.
	target := UIManager default 
		chooseFrom: (morphs collect:[:m| m knownName ifNil:[m class name asString]])
		values: self potentialEmbeddingTargets
		title: ('Place ', self externalName, ' in...').
	target ifNil:[^self].
	target addMorphFrontFromWorldPosition: self! !

!Morph methodsFor: 'e-toy support' stamp: 'mt 5/17/2016 15:59'!
embedInWindow

	| window worldToUse |

	worldToUse := self world.		"I'm assuming we are already in a world"
	window := (SystemWindow labelled: self defaultLabelForInspector) model: nil.
	window bounds: ((self position - ((0@window labelHeight) + window borderWidth))
						corner: self bottomRight + window borderWidth).
	window addMorph: self frame: (0@0 extent: 1@1).
	window updatePaneColors.
	worldToUse addMorph: window.
	window beKeyWindow.! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:41'!
enableDragNDrop
	self enableDragNDrop: true! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:21'!
enableDragNDrop: aBoolean
	"Set both properties at once"
	self separateDragAndDrop.
	self enableDrag: aBoolean.
	self enableDrop: aBoolean.! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:50'!
enableDrag: aBoolean
	self setProperty: #dragEnabled toValue: aBoolean! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:51'!
enableDrop: aBoolean
	self setProperty: #dropEnabled toValue: aBoolean! !

!Morph methodsFor: '*Morphic-Sound-piano rolls' stamp: 'RAA 12/11/2000 15:48'!
encounteredAtTime: ticks inScorePlayer: scorePlayer atIndex: index inEventTrack: track secsPerTick: secsPerTick

	"a hack to allow for abitrary morphs to be dropped into piano roll"
	self triggerActionFromPianoRoll.! !

!Morph methodsFor: 'WiW support' stamp: 'gk 5/24/2004 15:43'!
eToyRejectDropMorph: morphToDrop event: evt

	| tm am |

	tm := TextMorph new 
		beAllFont: ((TextStyle named: Preferences standardEToysFont familyName) fontOfSize: 24);
		contents: 'GOT IT!!'.
	(am := AlignmentMorph new)
		color: Color yellow;
		layoutInset: 10;
		useRoundedCorners;
		vResizing: #shrinkWrap;
		hResizing: #shrinkWrap;
		addMorph: tm;
		fullBounds;
		position: (self bounds center - (am extent // 2));
		openInWorld: self world.
	SoundService default playSoundNamed: 'yum' ifAbsentReadFrom: 'yum.aif'.
	morphToDrop rejectDropMorphEvent: evt.		"send it back where it came from"
	am delete
! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:05'!
eventBubbleFilters
	^ self valueOfProperty: #eventBubbleFilters! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:05'!
eventBubbleFilters: anArrayOrNil
	^ self setProperty: #eventBubbleFilters toValue: anArrayOrNil! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:05'!
eventCaptureFilters
	^ self valueOfProperty: #eventCaptureFilters! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:05'!
eventCaptureFilters: anArrayOrNil
	^ self setProperty: #eventCaptureFilters toValue: anArrayOrNil! !

!Morph methodsFor: 'events-filtering' stamp: 'mt 6/4/2016 18:21'!
eventFilterDocumentation
	"
	Event filters represent a mechanism to intercept events *before* morphs get the chance to handle them. So, there are still event handlers, which are usually the morphs themselves. However, morphs can also be their own event filters. One application of event filters is keyboard shortcuts. Such filters should invoke the shortcut but ignore the original event to *be sure* that no other morph ever handles that. System windows have window shortcuts. The world has world shortcuts.
	
	When user input events are dispatched, there is a capturing phase and a bubbling phase. This dispatching behavior is described in MorphicEventDispatcher. Capturing means that we are looking for the handler (resp. morph) to handle the event, traversing submorph hierarchy. Bubbling means that once a morph handled the event, all owners get also the chance to handle that event *if* they honor the #wasHandled flag.. Event filters can be attached to any morph for the capturing or the bubbling phase. See it as code that is execute before/after the event handlers.
	
	Event filters are objects that implement #filterEvent:for:. All objects understand that message but do nothing. There is a PluggableEventFilter for a more dynamic programming approach.
	
	Morphic hands have specific focus holders. There is a keyboard focus and a mouse focus. For example, text morphs are good candidates for holders of the keyboard focus. Both foci bypass the traditional event capturing phase, which usually start at the world morph. However, capture filters are also apply before handling the focus event. See Morph >> #handleFocusEvent: for that. Note that after handling the focus event, events do bubble along their owners, which usually do nothing if they honor the #wasHandled flag, and hence all bubble event filters are executed unless events get ignored. Thus, we mimick the behavior of MorphicEventDispatcher in #handleFocusEvent:.
	
	There are event filters for all kinds of events, filters for keyboard events, and filters for mouse events. You can always install generic event filters and then check for the particular event type in that filter. However, it is good style to not install keyboard event filters as generic event filters. It makes code more readable.
	
	You can install global event capture filters in instances of HandMorph. For example,  Project current world firstHand provides access to the project's world's hand. Avoid using globals such as ActiveHand or ActiveWorld. BEWARE that global event filters can have unexpected effects and may be difficult to debug. Note that you cannot install global event bubble filters in the HandMorph. You have to install such a filter in the world because hands are not part of the event bubbling phase.
	
	For an example of keyboard capturing filters see SystemWindow >> #addKeyboardShortcuts and SystemWindow >> #filterEvent:for:. You can also take a look at: PasteUpMorph >> #addKeyboardShortcuts and PasteUpMorph >> #filterEvent:for: and DockingBarMorph >> #filterEvent:for: for a more elaborate example.
	
	Event filters include all the behavior of event listeners. An event listener converted to an event filter would be a global event capture filter that sents a copy of the event to itself:
	
	| listenerFilter someMorph |
	someMorph := Morph new.
	listenerFilter := PluggableEventFilter on: [:event :target | event copy sentTo: someMorph. event].
	Project current world activeHand addEventCaptureFilter: listenerFilter.
	
	Instead of:

	| someMorph |
	someMorph := Morph new.
	Project current world activeHand addEventListener: someMorph.
	
	The filter version might look more complicated than the traditional listener version but note that it is quite rare to install global event filters. The example for keyboard shortcuts reflects the simplicity of event filters."! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:53'!
eventHandler
	"answer the receiver's eventHandler"
	^ extension ifNotNil: [extension eventHandler] ! !

!Morph methodsFor: 'accessing' stamp: 'dgd 2/16/2003 19:25'!
eventHandler: anEventHandler 
	"Note that morphs can share eventHandlers and all is OK. "
	self assureExtension eventHandler: anEventHandler! !

!Morph methodsFor: 'drawing' stamp: 'mt 4/4/2015 14:14'!
expandFullBoundsForDropShadow: aRectangle
	"Return an expanded rectangle for an eventual drop shadow"
	| delta box |

	self shadowOffset isRectangle
		ifTrue: [^ aRectangle outsetBy: self shadowOffset].

	box := aRectangle.
	delta := self shadowOffset.
	box := delta x >= 0 
		ifTrue:[box right: aRectangle right + delta x]
		ifFalse:[box left: aRectangle left + delta x].
	box := delta y >= 0
		ifTrue:[box bottom: aRectangle bottom + delta y]
		ifFalse:[box top: aRectangle top + delta y].
	^box! !

!Morph methodsFor: 'drawing' stamp: 'ar 11/8/2000 19:29'!
expandFullBoundsForRolloverBorder: aRectangle
	| delta |
	delta := self valueOfProperty: #rolloverWidth ifAbsent: [10@10].
	^aRectangle expandBy: delta.

! !

!Morph methodsFor: 'menus' stamp: 'mt 4/11/2015 08:09'!
exploreInMorphic

	ToolSet explore: self.! !

!Morph methodsFor: 'menus' stamp: 'mt 4/11/2015 08:09'!
exploreInMorphic: evt
	
	ToolSet explore: self.! !

!Morph methodsFor: 'menus' stamp: 'cmm 3/29/2012 20:33'!
exportAsBMP
	| fName |
	fName := UIManager default request:'Please enter the name' translated initialAnswer: self externalName,'.bmp'.
	fName isEmpty ifTrue:[^self].
	self exportAsBMPNamed: fName! !

!Morph methodsFor: 'menus' stamp: 'cmm 3/29/2012 20:32'!
exportAsBMPNamed: aString 
	self imageForm writeBMPfileNamed: aString! !

!Morph methodsFor: 'menus' stamp: 'cmm 3/29/2012 20:32'!
exportAsGIF
	| fName |
	fName := UIManager default request:'Please enter the name' translated initialAnswer: self externalName,'.gif'.
	fName isEmpty ifTrue:[^self].
	self exportAsGIFNamed: fName! !

!Morph methodsFor: 'menus' stamp: 'cmm 3/29/2012 20:32'!
exportAsGIFNamed: aString 
	GIFReadWriter
		putForm: self imageForm
		onFileNamed: aString! !

!Morph methodsFor: 'menus' stamp: 'cmm 3/29/2012 20:32'!
exportAsJPEG
	| fName |
	fName := UIManager default request: 'Please enter the name' translated initialAnswer: self externalName,'.jpeg'.
	fName isEmpty ifTrue: [^ self].
	self exportAsJPEGNamed: fName! !

!Morph methodsFor: 'menus' stamp: 'cmm 3/29/2012 20:31'!
exportAsJPEGNamed: aString 
	self imageForm writeJPEGfileNamed: aString! !

!Morph methodsFor: 'menus' stamp: 'cmm 3/29/2012 20:27'!
exportAsPNG
	| fName |
	fName := UIManager default request:'Please enter the name' translated initialAnswer: self externalName,'.png'.
	fName isEmpty ifTrue:[^self].
	self exportAsPNGNamed: fName! !

!Morph methodsFor: 'menus' stamp: 'cmm 3/29/2012 20:27'!
exportAsPNGNamed: aString 
	PNGReadWriter
		putForm: self imageForm
		onFileNamed: aString! !

!Morph methodsFor: 'accessing - extension' stamp: 'dgd 2/16/2003 19:22'!
extension
	"answer the recevier's extension"
	^ extension! !

!Morph methodsFor: 'geometry' stamp: ''!
extent

	^ bounds extent! !

!Morph methodsFor: 'geometry' stamp: 'mt 4/6/2015 11:56'!
extent: aPoint

	(bounds extent closeTo: aPoint) ifTrue: [^ self].
	self changed.
	bounds := (bounds topLeft extent: aPoint) rounded.
	self removeProperty: #dropShadow.
	self layoutChanged.
	self changed.
! !

!Morph methodsFor: 'viewer' stamp: 'sw 10/30/1998 14:46'!
externalName
	^ self knownName ifNil: [self innocuousName]! !

!Morph methodsFor: 'visual properties' stamp: 'dgd 2/16/2003 20:02'!
fillStyle
	"Return the current fillStyle of the receiver."
	^ self
		valueOfProperty: #fillStyle
		ifAbsent: ["Workaround already converted morphs"
			color
				ifNil: [self defaultColor]]! !

!Morph methodsFor: 'visual properties' stamp: 'ar 6/18/1999 07:05'!
fillStyle: aFillStyle
	"Set the current fillStyle of the receiver."
	self setProperty: #fillStyle toValue: aFillStyle.
	"Workaround for Morphs not yet converted"
	color := aFillStyle asColor.
	self changed.! !

!Morph methodsFor: 'visual properties' stamp: 'dgd 1/7/2005 19:31'!
fillWithRamp: rampSpecsOrColor oriented: aRatio 
	rampSpecsOrColor isColor
		ifTrue: [self color: rampSpecsOrColor".
			self borderColor: rampSpecsOrColor muchDarker"]
		ifFalse: [| fill | 
			fill := GradientFillStyle ramp: rampSpecsOrColor.
			fill origin: self bounds topLeft.
			fill direction: (self bounds extent * aRatio) truncated.
			fill radial: false.
			self fillStyle: fill.
			self borderColor: (rampSpecsOrColor first value mixed: 0.5 with: rampSpecsOrColor last value) muchDarker]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'rhi 9/10/2000 12:12'!
findA: aClass
	"Return the first submorph of the receiver that is descended from the given class. Return nil if there is no such submorph. Clients of this code should always check for a nil return value so that the code will be robust if the user takes the morph apart."

	^self submorphs
		detect: [:p | p isKindOf: aClass]
		ifNone: [nil]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'sw 1/9/2001 12:30'!
findDeeplyA: aClass
	"Return a morph in the submorph tree of the receiver that is descended from the given class. Return nil if there is no such morph. Clients of this code should always check for a nil return value so that the code will be robust if the user takes the morph apart."

	^ (self allMorphs copyWithout: self)
		detect: [:p | p isKindOf: aClass]
		ifNone: [nil]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'LC 9/28/1999 19:12'!
findDeepSubmorphThat: block1 ifAbsent: block2 
	self
		allMorphsDo: [:m | (block1 value: m)
				== true ifTrue: [^ m]].
	^ block2 value! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 3/17/2001 15:32'!
findSubmorphBinary: aBlock
	"Use binary search for finding a specific submorph of the receiver. Caller must be certain that the ordering holds for the submorphs."
	^submorphs findBinary: aBlock ifNone:[nil].! !

!Morph methodsFor: 'event handling' stamp: 'RAA 2/12/2001 15:26'!
firstClickTimedOut: evt
	"Useful for double-click candidates who want to know whether or not the click is a single or double. In this case, ignore the #click: and wait for either this or #doubleClick:"

! !

!Morph methodsFor: 'structure' stamp: 'di 11/13/2000 00:48'!
firstOwnerSuchThat: conditionBlock

	self allOwnersDo: [:m | (conditionBlock value: m) ifTrue: [^ m]].
	^ nil
! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:31'!
firstSubmorph
	^submorphs first! !

!Morph methodsFor: 'macpal' stamp: 'mt 4/19/2016 12:03'!
flash
	| originalColor |
	originalColor := self color.
	[ self color:
		(originalColor
			ifNil: [ Color black ]
			ifNotNil: [( (originalColor alpha: 1) adjustSaturation: 0.8 brightness: 0) negated ]) ]
		ensure:
			[ self refreshWorld.
			self color: originalColor ]! !

!Morph methodsFor: 'drawing' stamp: 'sw 11/26/2003 17:43'!
flashBounds
	"Flash the receiver's bounds  -- does not use the receiver's color, thus works with StringMorphs and SketchMorphs, etc., for which #flash is useless.  No senders initially, but useful to send this from a debugger or inspector"

	5 timesRepeat:
		[Display flash: self boundsInWorld  andWait: 120]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/5/2000 18:13'!
formerOwner
	^self valueOfProperty: #formerOwner! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'jmv 3/2/2010 16:16'!
formerOwner: aMorphOrNil 
	aMorphOrNil 
		ifNil: [self removeProperty: #formerOwner]
		ifNotNil: [self setProperty: #formerOwner toValue: aMorphOrNil]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/5/2000 18:13'!
formerPosition
	^self valueOfProperty: #formerPosition! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'jmv 3/2/2010 16:16'!
formerPosition: formerPosition 
	formerPosition 
		ifNil: [self removeProperty: #formerPosition]
		ifNotNil: [self setProperty: #formerPosition toValue: formerPosition]! !

!Morph methodsFor: 'accessing' stamp: 'ar 9/22/2000 13:36'!
forwardDirection
	"Return the receiver's forward direction (in eToy terms)"
	^self valueOfProperty: #forwardDirection ifAbsent:[0.0]! !

!Morph methodsFor: 'layout' stamp: 'ar 1/1/2002 20:00'!
fullBounds
	"Return the bounding box of the receiver and all its children. Recompute the layout if necessary."
	fullBounds ifNotNil:[^fullBounds].
	"Errors at this point can be critical so make sure we catch 'em all right"
	[self doLayoutIn: self layoutBounds] on: Error do:[:ex|
		"This should do it unless you don't screw up the bounds"
		fullBounds := bounds.
		ex pass].
	^fullBounds! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:06'!
fullBoundsInWorld
	^self bounds: self fullBounds in: self world! !

!Morph methodsFor: 'geometry testing' stamp: 'di 5/3/2000 19:05'!
fullContainsPoint: aPoint

	(self fullBounds containsPoint: aPoint) ifFalse: [^ false].  "quick elimination"
	(self containsPoint: aPoint) ifTrue: [^ true].  "quick acceptance"
	submorphs do: [:m | (m fullContainsPoint: aPoint) ifTrue: [^ true]].
	^ false
! !

!Morph methodsFor: 'copying' stamp: 'sw 2/16/2001 16:30'!
fullCopy
	"Deprecated, but maintained for backward compatibility with existing code (no senders in the base 3.0 image).   Calls are revectored to #veryDeepCopy, but note that #veryDeepCopy does not do exactly the same thing that the original #fullCopy did, so beware!!"

	^ self veryDeepCopy! !

!Morph methodsFor: 'drawing' stamp: 'mt 4/10/2015 15:13'!
fullDrawOn: aCanvas
	"Draw the full Morphic structure on the given Canvas"

	self visible ifFalse: [^ self].
	(aCanvas isVisible: self fullBounds) ifFalse:[^self].
	(self hasProperty: #errorOnDraw) ifTrue:[^self drawErrorOn: aCanvas].
	"Note: At some point we should generalize this into some sort of 
	multi-canvas so that we can cross-optimize some drawing operations."

	"Pass 1: Draw eventual drop-shadow"
	self hasDropShadow ifTrue: [self drawDropShadowOn: aCanvas].
	(self hasRolloverBorder and: [(aCanvas seesNothingOutside: self bounds) not])
		ifTrue: [self drawRolloverBorderOn: aCanvas].

	"Pass 2: Draw receiver itself"
	(aCanvas isVisible: self bounds) ifTrue:[aCanvas drawMorph: self].
	self drawSubmorphsOn: aCanvas.
	self drawOverlayOn: aCanvas.! !

!Morph methodsFor: '*morphic-Postscript Canvases' stamp: 'mpw 8/10/1930 05:25'!
fullDrawPostscriptOn: aCanvas

	self fullDrawOn:aCanvas.
! !

!Morph methodsFor: 'caching' stamp: 'jm 11/13/97 16:35'!
fullLoadCachedState
	"Load the cached state of the receiver and its full submorph tree."

	self allMorphsDo: [:m | m loadCachedState].
! !

!Morph methodsFor: 'printing' stamp: ''!
fullPrintOn: aStream

	aStream nextPutAll: self class name , ' newBounds: (';
		print: bounds;
		nextPutAll: ') color: ' , (self colorString: color)! !

!Morph methodsFor: 'caching' stamp: 'jm 11/13/97 16:34'!
fullReleaseCachedState
	"Release the cached state of the receiver and its full submorph tree."

	self allMorphsDo: [:m | m releaseCachedState].
! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 9/1/2000 10:15'!
getNumericValue
	"Only certain kinds of morphs know how to deal with this frontally; here we provide support for a numeric property of any morph"

	^ self valueOfProperty: #numericValue ifAbsent: [0]! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:06'!
globalPointToLocal: aPoint
	^self point: aPoint from: nil! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 9/3/2007 12:41'!
goBehind
	"Move the receiver to bottom z-order."

	| topRend |
	topRend := self topRendererOrSelf.
	topRend owner ifNotNilDo:
		[:own | own addMorphNearBack: topRend]
! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 10/6/2000 16:37'!
grabMorph: evt

	evt hand grabMorph: self! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/6/2000 15:13'!
grabTransform
	"Return the transform for the receiver which should be applied during grabbing"
	^owner ifNil:[IdentityTransform new] ifNotNil:[owner grabTransform]! !

!Morph methodsFor: 'geometry' stamp: 'ar 9/15/2000 14:21'!
griddedPoint: ungriddedPoint

	| griddingContext |
	self flag: #arNote. "Used by event handling - should transform to pasteUp for gridding"
	(griddingContext := self pasteUpMorph) ifNil: [^ ungriddedPoint].
	^ griddingContext gridPoint: ungriddedPoint! !

!Morph methodsFor: 'e-toy support' stamp: 'kfr 9/4/2004 15:22'!
gridFormOrigin: origin grid: smallGrid background: backColor line: lineColor

	| bigGrid gridForm gridOrigin |
	gridOrigin := origin \\ smallGrid.
	bigGrid := (smallGrid asPoint x) @ (smallGrid asPoint y).
	gridForm := Form extent: bigGrid depth: Display depth.
	backColor ifNotNil: [gridForm fillWithColor: backColor].
	gridOrigin x to: gridForm width by: smallGrid x do:
		[:x | gridForm fill: (x@0 extent: 1@gridForm height) fillColor: lineColor].
	gridOrigin y to: gridForm height by: smallGrid y do:
		[:y | gridForm fill: (0@y extent: gridForm width@1) fillColor: lineColor].
	^ InfiniteForm with: gridForm
! !

!Morph methodsFor: 'geometry' stamp: 'di 8/25/2000 00:35'!
gridPoint: ungriddedPoint

	^ ungriddedPoint! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 3/17/2001 13:19'!
halo

	(self outermostWorldMorph ifNil: [^nil]) haloMorphs do: [:h | h target == self ifTrue: [^ h]].
	^ nil! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 9/15/2000 16:13'!
haloClass
	"Answer the name of the desired kind of HaloMorph to launch on behalf of the receiver"

	^ #HaloMorph
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 8/8/2001 15:40'!
haloDelayTime
	"Return the number of milliseconds before a halo should be put up on the receiver. The halo will only be put up if the receiver responds to #wantsHalo by returning true."
	^800! !

!Morph methodsFor: 'events-processing' stamp: 'ar 1/10/2001 21:35'!
handleDropFiles: anEvent
	"Handle a drop from the OS."
	anEvent wasHandled ifTrue:[^self]. "not interested"
	(self wantsDropFiles: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	self dropFiles: anEvent.
! !

!Morph methodsFor: 'events-processing' stamp: 'mt 6/11/2016 11:07'!
handleDropMorph: anEvent
	"Handle a dropping morph."
	| aMorph localPt |
	anEvent wasHandled ifTrue: [^ self]. "Not interested"
	aMorph := anEvent contents.
	"Do a symmetric check if both morphs like each other"
	((self wantsDroppedMorph: aMorph event: anEvent)	"I want her"
		and: [aMorph wantsToBeDroppedInto: self])		"she wants me"
		ifFalse: [aMorph removeProperty: #undoGrabCommand.
				^ self].
	anEvent wasHandled: true.
	"Transform the morph into the receiver's coordinate frame. This is currently incomplete since it only takes the offset into account where it really should take the entire transform."
	localPt := (self transformedFrom: anEvent hand world) "full transform down"
				globalPointToLocal: aMorph referencePosition.
	aMorph referencePosition: localPt.
	self acceptDroppingMorph: aMorph event: anEvent.
	aMorph justDroppedInto: self event: anEvent.
! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'tpr 4/9/2014 15:29'!
handledOwnDraggingBy: aHandMorph on: aCanvas
"this is my chance to do something differrent to the normal dragging work. return true if I did what I wanted, false if not"
	^false! !

!Morph methodsFor: 'events-processing' stamp: 'mt 5/21/2016 09:53'!
handleEvent: anEvent
	"Apply event filters and then handle the resulting event. We have to return the event to chain filters."

	| filteredEvent |
	filteredEvent := self sendFilterEventBubble: anEvent for: self.
	filteredEvent wasIgnored ifFalse: [filteredEvent sentTo: self].
	^ filteredEvent! !

!Morph methodsFor: 'events-processing' stamp: 'mt 6/20/2016 16:13'!
handleFocusEvent: anEvent
	"Handle the given event. This message is sent if the receiver currently has the focus and is therefore receiving events directly from some hand. However, it might already have been handled due to overrides in #processFocusEvent:using:. We might want to get rid of this call-back in the future..."
	
	^ anEvent wasHandled
		ifTrue: [anEvent]
		ifFalse: [self handleEvent: anEvent]! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/15/2000 23:01'!
handleKeyDown: anEvent
	"System level event handling."
	anEvent wasHandled ifTrue:[^self].
	(self handlesKeyboard: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	^self keyDown: anEvent! !

!Morph methodsFor: 'events-processing' stamp: 'mt 6/4/2016 16:39'!
handleKeystroke: anEvent 
	"System level event handling. Has support for automatically grabbing the keyboard focus considering the keyboard focus delegate. See #newKeyboardFocus:"
	
	| handler |
	anEvent wasHandled ifTrue: [^ self].
	(self handlesKeyboard: anEvent) ifFalse: [^ self].
	
	handler := self wantsKeyboardFocus
		ifFalse: [self]
		ifTrue: [(anEvent hand newKeyboardFocus: self) ifNil: [self]].
	anEvent handler: handler.
	
	anEvent wasHandled: true.
	^ handler keyStroke: anEvent! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/15/2000 23:01'!
handleKeyUp: anEvent
	"System level event handling."
	anEvent wasHandled ifTrue:[^self].
	(self handlesKeyboard: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	^self keyUp: anEvent! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/16/2000 14:22'!
handleListenEvent: anEvent
	"Handle the given event. This message is sent if the receiver is a registered listener for the given event."
	^anEvent sentTo: self.! !

!Morph methodsFor: 'events-processing' stamp: 'mt 5/14/2016 11:36'!
handleMouseDown: anEvent
	"System level event handling."
	anEvent wasHandled ifTrue:[^self]. "not interested"
	anEvent hand removePendingBalloonFor: self.
	anEvent hand removePendingHaloFor: self.
	anEvent wasHandled: true.

	(anEvent controlKeyPressed
			and: [anEvent blueButtonChanged not
				and: [Preferences cmdGesturesEnabled]])
		ifTrue: [^ self invokeMetaMenu: anEvent].

	"Make me modal during mouse transitions"
	anEvent hand newMouseFocus: self event: anEvent.
	(anEvent blueButtonChanged and:[Preferences cmdGesturesEnabled]) ifTrue:[^self blueButtonDown: anEvent].
	
	"this mouse down could be the start of a gesture, or the end of a gesture focus"
	(self isGestureStart: anEvent)
		ifTrue: [^ self gestureStart: anEvent].

	self mouseDown: anEvent.

	Preferences maintainHalos
		ifFalse:[ anEvent hand removeHaloFromClick: anEvent on: self ].

	(self handlesMouseStillDown: anEvent) ifTrue:[
		self startStepping: #handleMouseStillDown: 
			at: Time millisecondClockValue + self mouseStillDownThreshold
			arguments: {anEvent copy resetHandlerFields}
			stepTime: self mouseStillDownStepRate ].
! !

!Morph methodsFor: 'events-processing' stamp: 'ar 8/8/2001 15:29'!
handleMouseEnter: anEvent
	"System level event handling."
	(anEvent isDraggingEvent) ifTrue:[
		(self handlesMouseOverDragging: anEvent) ifTrue:[
			anEvent wasHandled: true.
			self mouseEnterDragging: anEvent].
		^self].
	self wantsHalo "If receiver wants halo and balloon, trigger balloon after halo"
		ifTrue:[anEvent hand triggerHaloFor: self after: self haloDelayTime]
		ifFalse:[self wantsBalloon
			ifTrue:[anEvent hand triggerBalloonFor: self after: self balloonHelpDelayTime]].
	(self handlesMouseOver: anEvent) ifTrue:[
		anEvent wasHandled: true.
		self mouseEnter: anEvent.
	].! !

!Morph methodsFor: 'events-processing' stamp: 'ar 10/6/2000 00:15'!
handleMouseLeave: anEvent
	"System level event handling."
	anEvent hand removePendingBalloonFor: self.
	anEvent hand removePendingHaloFor: self.
	anEvent isDraggingEvent ifTrue:[
		(self handlesMouseOverDragging: anEvent) ifTrue:[
			anEvent wasHandled: true.
			self mouseLeaveDragging: anEvent].
		^self].
	(self handlesMouseOver: anEvent) ifTrue:[
		anEvent wasHandled: true.
		self mouseLeave: anEvent.
	].
! !

!Morph methodsFor: 'events-processing' stamp: 'cmm 6/3/2012 21:49'!
handleMouseMove: anEvent 
	"System level event handling."
	anEvent wasHandled ifTrue: [ ^ self ].
	"not interested"
	(self handlesMouseMove: anEvent) ifFalse: [ ^ self ].
	anEvent wasHandled: true.
	self mouseMove: anEvent.
	(self handlesMouseStillDown: anEvent) ifTrue:
		[ "Step at the new location"
		self
			startStepping: #handleMouseStillDown:
			at: Time millisecondClockValue
			arguments: {anEvent copy resetHandlerFields}
			stepTime: self mouseStillDownStepRate ]! !

!Morph methodsFor: 'events-processing' stamp: 'ar 4/23/2001 17:24'!
handleMouseOver: anEvent
	"System level event handling."
	anEvent hand mouseFocus == self ifTrue:[
		"Got this directly through #handleFocusEvent: so check explicitly"
		(self containsPoint: anEvent position event: anEvent) ifFalse:[^self]].
	anEvent hand noticeMouseOver: self event: anEvent! !

!Morph methodsFor: 'events-processing' stamp: 'ar 10/22/2000 17:11'!
handleMouseStillDown: anEvent
	"Called from the stepping mechanism for morphs wanting continuously repeated 'yes the mouse is still down, yes it is still down, yes it has not changed yet, no the mouse is still not up, yes the button is down' etc messages"
	(anEvent hand mouseFocus == self) 
		ifFalse:[^self stopSteppingSelector: #handleMouseStillDown:].
	self mouseStillDown: anEvent.
! !

!Morph methodsFor: 'events-processing' stamp: 'ar 10/22/2000 17:09'!
handleMouseUp: anEvent
	"System level event handling."
	anEvent wasHandled ifTrue:[^self]. "not interested"
	anEvent hand mouseFocus == self ifFalse:[^self]. "Not interested in other parties"
	anEvent hand releaseMouseFocus: self.
	anEvent wasHandled: true.
	anEvent blueButtonChanged
		ifTrue:[self blueButtonUp: anEvent]
		ifFalse:[self mouseUp: anEvent.
				self stopSteppingSelector: #handleMouseStillDown:].! !

!Morph methodsFor: 'events-processing' stamp: 'mt 6/6/2016 16:59'!
handleMouseWheel: anEvent
	"System level event handling."

	anEvent wasHandled ifTrue: [^self].
	(self handlesMouseWheel: anEvent) ifFalse: [^ self].
	anEvent wasHandled: true.
	^ self mouseWheel: anEvent! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 10/7/2000 18:44'!
handlerForBlueButtonDown: anEvent
	"Return the (prospective) handler for a mouse down event. The handler is temporarily installed and can be used for morphs further down the hierarchy to negotiate whether the inner or the outer morph should finally handle the event.
	Note: Halos handle blue button events themselves so we will only be asked if there is currently no halo on top of us."
	self wantsHaloFromClick ifFalse:[^nil].
	anEvent handler ifNil:[^self].
	anEvent handler isPlayfieldLike ifTrue:[^self]. "by default exclude playfields"
	(anEvent shiftPressed)
		ifFalse:[^nil] "let outer guy have it"
		ifTrue:[^self] "let me have it"
! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 10/12/2000 17:07'!
handlerForMetaMenu: evt
	"Return the prospective handler for invoking the meta menu. By default, the top-most morph in the innermost world gets this menu"
	self isWorldMorph ifTrue:[^self].
	evt handler ifNotNil:[evt handler isWorldMorph ifTrue:[^self]].
	^nil! !

!Morph methodsFor: 'event handling' stamp: 'nk 3/10/2004 19:47'!
handlerForMouseDown: anEvent 
	"Return the (prospective) handler for a mouse down event. The handler is temporarily 
	installed and can be used for morphs further down the hierarchy to negotiate whether 
	the inner or the outer morph should finally handle the event."

	anEvent blueButtonPressed
		ifTrue: [^ self handlerForBlueButtonDown: anEvent].
	anEvent yellowButtonPressed
		ifTrue: [^ self handlerForYellowButtonDown: anEvent].
	anEvent controlKeyPressed
		ifTrue: [^ self handlerForMetaMenu: anEvent].
	(self handlesMouseDown: anEvent)
		ifFalse: [^ nil].	"not interested"

	anEvent handler
		ifNil: [^ self ].	"Same priority but I am innermost"

	"Nobody else was interested"
	^self mouseDownPriority >= anEvent handler mouseDownPriority
		ifTrue: [ self]
		ifFalse: [ nil]! !

!Morph methodsFor: 'event handling' stamp: 'nk 3/10/2004 19:48'!
handlerForYellowButtonDown: anEvent 
	"Return the (prospective) handler for a mouse down event with the yellow button pressed.
	The 	handler is temporarily installed and can be used for morphs further 
	down the hierarchy to negotiate whether the inner or the outer 
	morph should finally handle the event."

	(self hasYellowButtonMenu or: [ self handlesMouseDown: anEvent ])
		ifFalse: [ ^ nil].	"Not interested."

	anEvent handler
		ifNil: [^ self].	"Nobody else was interested"

	"Same priority but I am innermost."
	^ self mouseDownPriority >= anEvent handler mouseDownPriority
		ifFalse: [nil ]
		ifTrue: [self]! !

!Morph methodsFor: 'event handling' stamp: 'ar 10/28/2000 22:18'!
handlesKeyboard: evt
	"Return true if the receiver wishes to handle the given keyboard event"
	self eventHandler ifNotNil: [^ self eventHandler handlesKeyboard: evt].
	^ false
! !

!Morph methodsFor: 'event handling' stamp: 'nk 2/14/2004 18:42'!
handlesMouseDown: evt
	"Do I want to receive mouseDown events (mouseDown:, mouseMove:, mouseUp:)?"
	"NOTE: The default response is false, except if you have added sensitivity to mouseDown events using the on:send:to: mechanism.  Subclasses that implement these messages directly should override this one to return true." 

	self eventHandler ifNotNil: [^ self eventHandler handlesMouseDown: evt].
	^ false! !

!Morph methodsFor: 'event handling' stamp: 'mt 6/2/2015 09:49'!
handlesMouseMove: anEvent 
	"Do I want to receive mouseMove: when the hand passes over the receiver?  Rules say that by default a morph gets #mouseMove iff
		* the hand is not dragging anything,
			+ and some button is down,
			+ and the receiver is the current mouse focus."
	self eventHandler ifNotNil: [^ self eventHandler handlesMouseMove: anEvent].
	self wantsEveryMouseMove ifTrue: [^ true].
	anEvent hand hasSubmorphs ifTrue: [ ^ false ].
	(anEvent anyButtonPressed and: [ anEvent hand mouseFocus == self ]) ifFalse: [ ^ false ].
	^ true! !

!Morph methodsFor: 'event handling' stamp: 'di 9/14/1998 07:31'!
handlesMouseOverDragging: evt
	"Return true if I want to receive mouseEnterDragging: and mouseLeaveDragging: when the hand drags something over me (button up or button down), or when the mouse button is down but there is no mouseDown recipient.  The default response is false, except if you have added sensitivity to mouseEnterLaden: or mouseLeaveLaden:, using the on:send:to: mechanism."
	"NOTE:  If the hand state matters in these cases, it may be tested by constructs such as
		event anyButtonPressed
		event hand hasSubmorphs"

	self eventHandler ifNotNil: [^ self eventHandler handlesMouseOverDragging: evt].
	^ false! !

!Morph methodsFor: 'event handling' stamp: 'di 9/14/1998 07:31'!
handlesMouseOver: evt
	"Do I want to receive mouseEnter: and mouseLeave: when the button is up and the hand is empty?  The default response is false, except if you have added sensitivity to mouseEnter: or mouseLeave:, using the on:send:to: mechanism." 

	self eventHandler ifNotNil: [^ self eventHandler handlesMouseOver: evt].
	^ false! !

!Morph methodsFor: 'event handling' stamp: 'ar 10/22/2000 17:06'!
handlesMouseStillDown: evt
	"Return true if the receiver wants to get repeated #mouseStillDown: messages between #mouseDown: and #mouseUp"
	self eventHandler ifNotNil: [^ self eventHandler handlesMouseStillDown: evt].
	^ false
! !

!Morph methodsFor: 'event handling' stamp: 'mt 6/6/2016 17:00'!
handlesMouseWheel: evt
	"Do I want to receive mouseWheel events? The default response is false, except if you have added sensitivity to mouseWheel events using the on:send:to: mechanism.  Subclasses that implement these messages directly should override this one to return true." 

	self eventHandler ifNotNil: [^ self eventHandler handlesMouseWheel: evt].
	^ false! !

!Morph methodsFor: 'events-processing' stamp: 'md 10/22/2003 15:55'!
handleUnknownEvent: anEvent
	"An event of an unknown type was sent to the receiver. What shall we do?!!"
	Beeper beep. 
	anEvent printString displayAt: 0@0.
	anEvent wasHandled: true.! !

!Morph methodsFor: 'events-processing' stamp: 'bf 8/30/2007 20:18'!
handleWindowEvent: anEvent
	"Handle an event concerning our host window"
	anEvent wasHandled ifTrue:[^self]. "not interested"
	(self wantsWindowEvent: anEvent) ifFalse:[^self].
	anEvent wasHandled: true.
	self windowEvent: anEvent.
! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 4/12/2005 23:07'!
handUserASibling
	"Make and hand the user a sibling instance.  Force the creation of a uniclass at this point if one does not already exist for the receiver."

	| topRend |
	topRend := self topRendererOrSelf.
	topRend couldMakeSibling ifFalse: [^ Beeper beep].

	topRend assuredPlayer assureUniClass.
	(topRend makeSiblings: 1) first openInHand! !

!Morph methodsFor: 'layout-menu' stamp: 'dgd 8/30/2003 20:09'!
hasClipLayoutCellsString
	^ (self clipLayoutCells
		ifTrue: ['<on>']
		ifFalse: ['<off>']), 'clip to cell size' translated! !

!Morph methodsFor: 'drawing' stamp: 'dgd 8/30/2003 20:20'!
hasClipSubmorphsString
	"Answer a string that represents the clip-submophs checkbox"
	^ (self clipSubmorphs
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'provide clipping' translated! !

!Morph methodsFor: 'menus' stamp: 'dgd 8/30/2003 20:23'!
hasDirectionHandlesString
	^ (self wantsDirectionHandles
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'direction handles' translated! !

!Morph methodsFor: 'layout-menu' stamp: 'dgd 8/30/2003 16:58'!
hasDisableTableLayoutString
	^ (self disableTableLayout
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'disable layout in tables' translated! !

!Morph methodsFor: 'text-anchor' stamp: 'dgd 9/6/2003 18:14'!
hasDocumentAnchorString
	^ (self textAnchorType == #document
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'Document' translated! !

!Morph methodsFor: 'menus' stamp: 'dgd 8/30/2003 20:24'!
hasDragAndDropEnabledString
	"Answer a string to characterize the drag & drop status of the  
	receiver"
	^ (self dragNDropEnabled
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'accept drops' translated! !

!Morph methodsFor: 'drop shadows' stamp: 'dgd 2/16/2003 21:42'!
hasDropShadow
	"answer whether the receiver has DropShadow"
	^ self
		valueOfProperty: #hasDropShadow
		ifAbsent: [false]! !

!Morph methodsFor: 'drop shadows' stamp: 'dgd 8/30/2003 16:49'!
hasDropShadowString
	^ (self hasDropShadow
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'show shadow' translated! !

!Morph methodsFor: 'drop shadows' stamp: 'mt 4/6/2015 13:39'!
hasDropShadow: aBool

	self hasDropShadow = aBool ifTrue: [^ self].
	self changed.
	aBool
		ifTrue:[self setProperty: #hasDropShadow toValue: true]
		ifFalse:[self removeProperty: #hasDropShadow].
		
	self layoutChanged.
	self changed.! !

!Morph methodsFor: 'accessing - extension' stamp: 'md 2/27/2006 08:31'!
hasExtension
	"answer whether the receiver has extention"
	^ extension notNil! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/10/2015 11:44'!
hasFocus

	self flag: #obsolete. "mt: Use #hasKeyboardFocus instead, which reads live hand information and no cache."
	^ false! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 9/15/2000 16:16'!
hasHalo
	^self hasProperty: #hasHalo.! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 9/28/2000 17:54'!
hasHalo: aBool
	aBool
		ifTrue:[self setProperty: #hasHalo toValue: true]
		ifFalse:[self removeProperty: #hasHalo]! !

!Morph methodsFor: 'text-anchor' stamp: 'dgd 9/6/2003 18:14'!
hasInlineAnchorString
	^ (self textAnchorType == #inline
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'Inline' translated! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/11/2015 16:56'!
hasKeyboardFocus

	^ self activeHand
		ifNil: [false]
		ifNotNil: [:hand | self hasKeyboardFocus: hand]! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/10/2015 17:07'!
hasKeyboardFocus: aHand

	^ aHand keyboardFocus == self keyboardFocusDelegate! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/11/2015 16:56'!
hasMouseFocus

	^ self activeHand
		ifNil: [false]
		ifNotNil: [:hand | self hasMouseFocus: hand]! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/10/2015 11:40'!
hasMouseFocus: aHand

	^ aHand mouseFocus == self! !

!Morph methodsFor: 'layout-menu' stamp: 'dgd 10/8/2003 19:23'!
hasNoLayoutString
	^ (self layoutPolicy isNil
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'no layout' translated! !

!Morph methodsFor: 'structure' stamp: 'ar 10/3/2000 15:36'!
hasOwner: aMorph
	"Return true if the receiver has aMorph in its owner chain"
	aMorph ifNil:[^true].
	self allOwnersDo:[:m| m = aMorph ifTrue:[^true]].
	^false! !

!Morph methodsFor: 'text-anchor' stamp: 'dgd 9/6/2003 18:14'!
hasParagraphAnchorString
	^ (self textAnchorType == #paragraph
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'Paragraph' translated! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 09:47'!
hasProperty: aSymbol 
	"Answer whether the receiver has the property named aSymbol"
	extension ifNil: [^ false].
	^extension hasProperty: aSymbol! !

!Morph methodsFor: 'layout-menu' stamp: 'dgd 8/30/2003 16:55'!
hasProportionalLayoutString
	| layout |
	^ (((layout := self layoutPolicy) notNil
			and: [layout isProportionalLayout])
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'proportional layout' translated! !

!Morph methodsFor: 'layout-menu' stamp: 'dgd 8/30/2003 20:08'!
hasReverseCellsString
	^ (self reverseTableCells
		ifTrue: ['<on>']
		ifFalse: ['<off>']), 'reverse table cells' translated! !

!Morph methodsFor: 'drop shadows' stamp: 'dgd 2/16/2003 21:58'!
hasRolloverBorder
	"answer whether the receiver has RolloverBorder"
	^ self
		valueOfProperty: #hasRolloverBorder
		ifAbsent: [false]! !

!Morph methodsFor: 'drop shadows' stamp: 'RAA 11/7/2000 15:54'!
hasRolloverBorder: aBool
	aBool
		ifTrue:[self setProperty: #hasRolloverBorder toValue: true]
		ifFalse:[self removeProperty: #hasRolloverBorder]! !

!Morph methodsFor: 'layout-menu' stamp: 'dgd 8/30/2003 20:09'!
hasRubberBandCellsString
	^ (self rubberBandCells
		ifTrue: ['<on>']
		ifFalse: ['<off>']), 'rubber band cells' translated! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:32'!
hasSubmorphs
	^submorphs notEmpty! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'nice 5/8/2011 10:50'!
hasSubmorphWithProperty: aSymbol
	^submorphs anySatisfy: [:m | m hasProperty: aSymbol]! !

!Morph methodsFor: 'layout-menu' stamp: 'dgd 8/30/2003 16:59'!
hasTableLayoutString
	| layout |
	^ (((layout := self layoutPolicy) notNil
			and: [layout isTableLayout])
		ifTrue: ['<on>']
		ifFalse: ['<off>'])
		, 'table layout' translated! !

!Morph methodsFor: 'accessing' stamp: 'di 1/3/1999 12:25'!
hasTranslucentColor
	"Answer true if this any of this morph is translucent but not transparent."

	^ color isColor and: [color isTranslucentColor]
! !

!Morph methodsFor: 'menu' stamp: 'dgd 4/4/2006 14:43'!
hasYellowButtonMenu
	"Answer true if I have any items at all for a context (yellow  
	button) menu."
	^ self wantsYellowButtonMenu
			or: [self models anySatisfy: [:each | each hasModelYellowButtonMenuItems]]! !

!Morph methodsFor: 'geometry' stamp: ''!
height

	^ bounds height! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:22'!
height: aNumber
	" Set my height; my position (top-left corner) and width will remain the same "

	self extent: self width@aNumber asInteger.
! !

!Morph methodsFor: 'menus' stamp: 'dgd 4/3/2006 14:09'!
helpButton
	"Answer a button whose action would be to put up help concerning the receiver"

	| aButton |
	aButton := SimpleButtonMorph new.
	aButton
		target: self;
		color: ColorTheme current helpColor;
		borderColor: ColorTheme current helpColor muchDarker;
		borderWidth: 1;
		label: '?' translated font: Preferences standardButtonFont;
		actionSelector: #presentHelp;
		setBalloonText: 'click here for help' translated.
	^ aButton! !

!Morph methodsFor: 'drawing' stamp: 'sw 10/30/1998 18:27'!
hide
	owner ifNil: [^ self].
	self visible ifTrue: [self visible: false.  self changed]! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/30/1998 12:44'!
highlight
	"The receiver is being asked to appear in a highlighted state.  Mostly used for textual morphs"
	self color: self highlightColor! !

!Morph methodsFor: 'accessing' stamp: 'sw 3/6/1999 02:09'!
highlightColor
	
	| val |
	^ (val := self valueOfProperty: #highlightColor)
		ifNotNil:
			[val ifNil: [self error: 'nil highlightColor']]
		ifNil:
			[owner ifNil: [self color] ifNotNil: [owner highlightColor]]! !

!Morph methodsFor: 'accessing' stamp: 'sw 7/2/1998 13:51'!
highlightColor: aColor
	self setProperty: #highlightColor toValue: aColor! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/28/2000 10:53'!
highlightedForDrop
	^(self valueOfProperty: #highlightedForDrop) == true! !

!Morph methodsFor: 'drawing' stamp: 'LC 5/18/2000 08:48'!
highlightedForMouseDown
	^(self valueOfProperty: #highlightedForMouseDown) == true! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/28/2000 11:51'!
highlightForDrop
	self highlightForDrop: true! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/28/2000 12:01'!
highlightForDrop: aBoolean
	self setProperty: #highlightedForDrop toValue: aBoolean.
	self changed! !

!Morph methodsFor: 'drawing' stamp: 'LC 5/18/2000 08:51'!
highlightForMouseDown
	self highlightForMouseDown: true! !

!Morph methodsFor: 'drawing' stamp: 'ar 3/17/2001 15:56'!
highlightForMouseDown: aBoolean
	aBoolean 
		ifTrue:[self setProperty: #highlightedForMouseDown toValue: aBoolean]
		ifFalse:[self removeProperty: #highlightedForMouseDown. self resetExtension].
	self changed! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:57'!
hResizing
	"Layout specific. This property describes how the receiver should be resized with respect to its owner and its children. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit children
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#rigid] ifNotNil:[props hResizing].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/31/2000 20:45'!
hResizingString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self hResizing! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:06'!
hResizing: aSymbol
	"Layout specific. This property describes how the receiver should be resized with respect to its owner and its children. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit children
	"
	self assureLayoutProperties hResizing: aSymbol.
	self layoutChanged.
! !

!Morph methodsFor: 'thumbnail' stamp: 'dgd 9/12/2004 21:12'!
icon
	"Answer a form with an icon to represent the receiver"
	^ self valueOfProperty: #icon! !

!Morph methodsFor: 'thumbnail' stamp: 'mt 2/17/2016 15:20'!
iconOrThumbnail

	^ self icon ifNil: [self thumbnail]! !

!Morph methodsFor: 'thumbnail' stamp: 'dgd 9/13/2004 12:43'!
iconOrThumbnailOfSize: aNumberOrPoint 
	"Answer an appropiate form to represent the receiver"

	^ self iconOrThumbnail scaledIntoFormOfSize: aNumberOrPoint
! !

!Morph methodsFor: 'drawing' stamp: 'jm 6/11/97 17:21'!
imageForm

	^ self imageFormForRectangle: self fullBounds
! !

!Morph methodsFor: 'drawing' stamp: 'di 7/8/1998 12:42'!
imageFormDepth: depth

	^ self imageForm: depth forRectangle: self fullBounds
! !

!Morph methodsFor: 'drawing' stamp: 'di 9/9/1998 22:25'!
imageFormForRectangle: rect

	^ self imageForm: Display depth forRectangle: rect
! !

!Morph methodsFor: 'drawing' stamp: 'ar 9/1/2000 14:23'!
imageFormWithout: stopMorph andStopThere: stopThere
	"Like imageForm, except it does not display stopMorph,
	and it will not display anything above it if stopThere is true.
	Returns a pair of the imageForm and a boolean that is true
		if it has hit stopMorph, and display should stop."
	| canvas rect |
	rect := self fullBounds.
	canvas := ColorPatchCanvas extent: rect extent depth: Display depth.
	canvas stopMorph: stopMorph.
	canvas doStop: stopThere.
	canvas translateBy: rect topLeft negated during:[:tempCanvas| tempCanvas fullDrawMorph: self].
	^ Array with: (canvas form offset: rect topLeft)
			with: canvas foundMorph! !

!Morph methodsFor: 'drawing' stamp: 'nk 9/1/2004 15:08'!
imageForm: depth backgroundColor: aColor forRectangle: rect
	| canvas |
	canvas := Display defaultCanvasClass extent: rect extent depth: depth.
	canvas translateBy: rect topLeft negated
		during:[:tempCanvas| 
			tempCanvas fillRectangle: rect color: aColor.
			tempCanvas fullDrawMorph: self].
	^ canvas form offset: rect topLeft! !

!Morph methodsFor: 'drawing' stamp: 'ar 9/1/2000 14:23'!
imageForm: depth forRectangle: rect
	| canvas |
	canvas := Display defaultCanvasClass extent: rect extent depth: depth.
	canvas translateBy: rect topLeft negated
		during:[:tempCanvas| tempCanvas fullDrawMorph: self].
	^ canvas form offset: rect topLeft! !

!Morph methodsFor: 'initialization' stamp: 'fbs 1/5/2005 09:17'!
inAScrollPane
	"Answer a scroll pane that allows the user to scroll the receiver in either direction.  It will have permanent scroll bars unless you take some special action."

	| widget |
	widget := ScrollPane new.
	widget extent: ((self width min: 300 max: 100) @ (self height min: 150 max: 100));
		borderWidth: 0.
	widget scroller addMorph: self.
	widget setScrollDeltas.
	widget color: self color darker darker.
	^ widget! !

!Morph methodsFor: 'initialization' stamp: 'sw 6/26/2001 10:56'!
inATwoWayScrollPane
	"Answer a two-way scroll pane that allows the user to scroll the receiver in either direction.  It will have permanent scroll bars unless you take some special action."

	| widget |
	widget := TwoWayScrollPane new.
	widget extent: ((self width min: 300 max: 100) @ (self height min: 150 max: 100));
		borderWidth: 0.
	widget scroller addMorph: self.
	widget setScrollDeltas.
	widget color: self color darker darker.
	^ widget! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'tk 10/31/2000 11:04'!
indexOfMorphAbove: aPoint
	"Return index of lowest morph whose bottom is above aPoint.
	Will return 0 if the first morph is not above aPoint."

	submorphs withIndexDo: [:mm :ii | 
		mm fullBounds bottom >= aPoint y ifTrue: [^ ii - 1]].
	^ submorphs size! !

!Morph methodsFor: 'meta-actions' stamp: 'yo 2/12/2005 15:25'!
indicateAllSiblings
	"Indicate all the receiver and all its siblings by flashing momentarily."

	| aPlayer allBoxes |
	(aPlayer := self topRendererOrSelf player) belongsToUniClass ifFalse: [^ self "error: 'not uniclass'"].
	allBoxes := aPlayer class allInstances
		select: [:m | m costume world == ActiveWorld]
		thenCollect: [:m | m costume boundsInWorld].

	5 timesRepeat:
		[Display flashAll: allBoxes andWait: 120]! !

!Morph methodsFor: 'testing' stamp: 'mt 4/10/2015 17:15'!
indicateKeyboardFocus

	^ (self hasProperty: #indicateKeyboardFocus)
		ifTrue: [(self valueOfProperty: #indicateKeyboardFocus) ~~ #never]
		ifFalse: [self class indicateKeyboardFocus]! !

!Morph methodsFor: 'user interface' stamp: 'sw 10/2/97 23:08'!
initialExtent
	| ext |
	(ext := self valueOfProperty: #initialExtent)
		ifNotNil:
			[^ ext].
	^ super initialExtent! !

!Morph methodsFor: 'initialization' stamp: 'cmm 5/8/2014 10:57'!
initialize
	"initialize the state of the receiver"
	owner := nil.
	submorphs := Array empty.
	bounds := self defaultBounds.
	color := self defaultColor! !

!Morph methodsFor: 'accessing - extension' stamp: 'md 2/27/2006 08:35'!
initializeExtension
	"private - initializes the receiver's extension"
	extension := MorphExtension new! !

!Morph methodsFor: 'parts bin' stamp: 'sw 8/12/2001 02:07'!
initializeToStandAlone
	"Set up the receiver, created by a #basicNew and now ready to be initialized, as a fully-formed morph suitable for providing a graphic for a parts bin surrogate, and, when such a parts-bin surrogate is clicked on, for attaching to the hand as a viable stand-alone morph.  Because of historical precedent, #initialize has been expected to handle this burden, though a great number of morphs actually cannot stand alone.  In any case, by default we call the historical #initialize, though unhappily, so that all existing morphs will work no worse than before when using this protocol."

	self initialize! !

!Morph methodsFor: 'printing' stamp: ''!
initString

	^ String streamContents: [:s | self fullPrintOn: s]! !

!Morph methodsFor: 'geometry' stamp: 'DS 4/22/2015 11:12:19.13'!
innerBounds
	"Return the inner rectangle enclosed by the bounds of this morph excluding the space taken by its borders. For an unbordered morph, this is just its bounds."

	^ self bounds insetBy: self borderStyle inset! !

!Morph methodsFor: 'naming' stamp: 'dgd 8/30/2003 15:52'!
innocuousName
	"Choose an innocuous name for the receiver -- one that does not end in the word Morph"

	| className allKnownNames |
	className := self defaultNameStemForInstances.
	(className size > 5 and: [className endsWith: 'Morph'])
		ifTrue: [className := className copyFrom: 1 to: className size - 5].
	className := className asString translated.
	allKnownNames := self world ifNil: [OrderedCollection new] ifNotNil: [self world allKnownNames].
	^ Utilities keyLike: className asString satisfying:
		[:aName | (allKnownNames includes: aName) not]! !

!Morph methodsFor: 'parts bin' stamp: 'di 11/13/2000 00:49'!
inPartsBin

	self isPartsDonor ifTrue: [^ true].
	self allOwnersDo: [:m | m isPartsBin ifTrue: [^ true]].
	^ false
! !

!Morph methodsFor: 'accessing' stamp: 'tk 1/31/2002 10:25'!
insetColor
	owner ifNil:[^self color].
	^ self colorForInsets! !

!Morph methodsFor: 'debug and other' stamp: 'ar 8/7/2009 23:42'!
inspectArgumentsPlayerInMorphic: evt
	evt hand attachMorph: ((Inspector openOn: self player) extent: 300@200)! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 8/6/2009 20:14'!
inspectAt: aPoint event: evt
	| morphs target |
	morphs := self morphsAt: aPoint.
	(morphs includes: self) ifFalse:[morphs := morphs copyWith: self].
	target := UIManager default
		chooseFrom: (morphs collect: [:m | m knownName ifNil:[m class name asString]])
		values: morphs
		title:  ('inspect whom?
(deepest at top)').
	target ifNil:[^self].
	target inspectInMorphic: evt! !

!Morph methodsFor: 'menus' stamp: 'mt 4/11/2015 08:09'!
inspectInMorphic

	ToolSet inspect: self.! !

!Morph methodsFor: 'menus' stamp: 'mt 4/11/2015 08:09'!
inspectInMorphic: evt

	ToolSet inspect: self.
! !

!Morph methodsFor: 'debug and other' stamp: 'sw 11/5/1998 20:31'!
inspectOwnerChain
	self ownerChain inspectWithLabel: 'Owner chain for ', self printString! !

!Morph methodsFor: 'debug and other' stamp: ''!
installModelIn: ignored
	"Simple morphs have no model"
	"See MorphicApp for other behavior"! !

!Morph methodsFor: 'geometry' stamp: 'nk 4/27/2003 16:16'!
intersects: aRectangle
	"Answer whether aRectangle, which is in World coordinates, intersects me."

	^self fullBoundsInWorld intersects: aRectangle! !

!Morph methodsFor: 'initialization' stamp: 'ar 1/31/2001 13:57'!
intoWorld: aWorld
	"The receiver has just appeared in a new world. Note:
		* aWorld can be nil (due to optimizations in other places)
		* owner is already set
		* owner's submorphs may not include receiver yet.
	Important: Keep this method fast - it is run whenever morphs are added."
	aWorld ifNil:[^self].
	self wantsSteps ifTrue:[aWorld startStepping: self].
	self submorphsDo:[:m| m intoWorld: aWorld].
! !

!Morph methodsFor: 'change reporting' stamp: 'ar 11/12/2000 18:50'!
invalidRect: damageRect
	^self invalidRect: damageRect from: self! !

!Morph methodsFor: 'change reporting' stamp: 'md 4/3/2006 11:52'!
invalidRect: aRectangle from: aMorph
	| damageRect |
	aRectangle hasPositiveExtent ifFalse: [ ^self ].
	damageRect := aRectangle.
	aMorph == self ifFalse:[
		"Clip to receiver's clipping bounds if the damage came from a child"
		self clipSubmorphs 
			ifTrue:[damageRect := aRectangle intersect: self clippingBounds]].
	owner ifNotNil: [owner invalidRect: damageRect from: self].! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 8/6/2009 20:15'!
invokeMetaMenuAt: aPoint event: evt
	| morphs target |
	morphs := self morphsAt: aPoint.
	(morphs includes: self) ifFalse:[morphs := morphs copyWith: self].
	morphs size = 1 ifTrue:[morphs first invokeMetaMenu: evt].
	target := UIManager default
		chooseFrom: (morphs collect: [:m | m knownName ifNil:[m class name asString]])
		values: morphs.
	target ifNil:[^self].
	target invokeMetaMenu: evt! !

!Morph methodsFor: 'meta-actions' stamp: 'sw 9/18/2006 12:21'!
invokeMetaMenu: evt
	"Put up the 'meta' menu, invoked via control-click, unless eToyFriendly is true."

	| menu |
	Preferences eToyFriendly ifTrue: [^ self].

	menu _ self buildMetaMenu: evt.
	menu addTitle: self externalName.
	menu popUpEvent: evt in: self world! !

!Morph methodsFor: 'classification' stamp: 'di 5/7/1998 01:21'!
isAlignmentMorph

	^ false! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 10/21/1998 15:54'!
isAViewer
	^ false! !

!Morph methodsFor: 'classification' stamp: 'ar 9/15/2000 17:56'!
isBalloonHelp
	^false! !

!Morph methodsFor: 'classification' stamp: 'ar 1/3/2010 16:11'!
isCompoundTileMorph
	^false! !

!Morph methodsFor: 'testing' stamp: 'dgd 8/31/2004 15:00'!
isDockingBar
	"Return true if the receiver is a docking bar"
	^ false! !

!Morph methodsFor: 'accessing' stamp: 'sw 6/13/2001 01:04'!
isFlap
	"Answer whether the receiver claims to be a flap"

	^ self hasProperty: #flap! !

!Morph methodsFor: 'classification' stamp: 'ar 9/28/2000 13:54'!
isFlapOrTab
	^self isFlap or:[self isFlapTab]! !

!Morph methodsFor: 'classification' stamp: 'ar 9/28/2000 13:53'!
isFlapTab
	^false! !

!Morph methodsFor: 'testing' stamp: 'ar 9/22/2000 13:44'!
isFlexed
	"Return true if the receiver is currently flexed"
	owner ifNil:[^false].
	^owner isFlexMorph! !

!Morph methodsFor: 'classification' stamp: 'jm 4/17/1998 00:44'!
isFlexMorph

	^ false
! !

!Morph methodsFor: 'testing' stamp: 'dgd 9/20/2004 14:31'!
isFullOnScreen
	"Answer if the receiver is full contained in the owner visible  
	area."
	owner isInMemory
		ifFalse: [^ true].
	owner isNil
		ifTrue: [^ true].
	self visible
		ifFalse: [^ true].
	^ owner clearArea containsRect: self fullBounds! !

!Morph methodsFor: 'geniestubs' stamp: 'nk 3/11/2004 17:45'!
isGestureStart: anEvent
	"This mouse down could be the start of a gesture, or the end of a gesture focus"

	anEvent hand isGenieEnabled
		ifFalse: [ ^false ].

	(self allowsGestureStart: anEvent)
		ifTrue: [^ true ].		"could be the start of a gesture"

	"otherwise, check for whether it's time to disable the Genie auto-focus"
	(anEvent hand isGenieFocused
		and: [anEvent whichButton ~= anEvent hand focusStartEvent whichButton])
			ifTrue: [anEvent hand disableGenieFocus].

	^false! !

!Morph methodsFor: 'classification' stamp: ''!
isHandMorph

	^ false! !

!Morph methodsFor: 'testing' stamp: 'eem 6/12/2013 15:29'!
isImageMorph
	^false! !

!Morph methodsFor: 'structure' stamp: 'dgd 9/1/2004 17:17'!
isInDockingBar
	"answer if the receiver is in a menu bar"
	^ (owner notNil) and: [owner isDockingBar]! !

!Morph methodsFor: 'structure' stamp: 'dgd 9/18/2004 15:56'!
isInSystemWindow
	"answer if the receiver is in a system window"
	^ owner isMorph and:[owner isSystemWindow or:[owner isInSystemWindow]]! !

!Morph methodsFor: 'structure' stamp: 'dgd 2/22/2003 19:05'!
isInWorld
	"Return true if this morph is in a world."

	^self world notNil! !

!Morph methodsFor: 'classification' stamp: 'ar 1/3/2010 18:53'!
isKedamaMorph
	^false! !

!Morph methodsFor: 'halos and balloon help' stamp: 'dgd 2/22/2003 19:05'!
isLikelyRecipientForMouseOverHalos
	^self player notNil! !

!Morph methodsFor: 'testing' stamp: 'nk 10/13/2003 18:36'!
isLineMorph
	^false! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:53'!
isLocked
	"answer whether the receiver is Locked"
	extension ifNil: [^ false].
	^ extension locked! !

!Morph methodsFor: 'testing' stamp: 'eem 8/27/2015 16:42'!
isMenuItemMorph
	^false! !

!Morph methodsFor: 'classification' stamp: 'ar 10/3/2000 18:11'!
isModalShell
	^false! !

!Morph methodsFor: 'testing' stamp: ''!
isMorph

	^ true! !

!Morph methodsFor: 'classification' stamp: 'ar 1/3/2010 16:08'!
isNumericReadoutTile
	^false! !

!Morph methodsFor: 'parts bin' stamp: 'sw 8/12/97 14:16'!
isPartsBin
	^ false! !

!Morph methodsFor: 'parts bin' stamp: 'md 2/27/2006 09:53'!
isPartsDonor
	"answer whether the receiver is PartsDonor"
	extension ifNil: [^ false].
	^ extension isPartsDonor! !

!Morph methodsFor: 'parts bin' stamp: 'md 2/27/2006 09:59'!
isPartsDonor: aBoolean 
	"change the receiver's isPartDonor property"
	(extension isNil and: [aBoolean not]) ifTrue: [^ self].
	self assureExtension isPartsDonor: aBoolean! !

!Morph methodsFor: 'classification' stamp: 'ar 1/3/2010 16:10'!
isPhraseTileMorph
	^false! !

!Morph methodsFor: 'classification' stamp: 'sw 1/29/98 21:51'!
isPlayfieldLike
	^ false! !

!Morph methodsFor: 'classification' stamp: 'jm 5/7/1998 13:45'!
isRenderer
	"A *renderer* morph transforms the appearance of its submorph in some manner. For example, it might supply a drop shadow or scale and rotate the morph it encases. Answer true if this morph acts as a renderer. This default implementation returns false."
	"Details: A renderer is assumed to have a single submorph. Renderers may be nested to concatenate their transformations. It is useful to be able to find the outer-most renderer. This can be done by ascending the owner chain from the rendered morph. To find the morph being rendered, one can descend through the (singleton) submorph lists of the renderer chain until a non-renderer is encountered."

	^ false
! !

!Morph methodsFor: 'testing' stamp: 'ar 10/31/2009 13:01'!
isSafeToServe
	"Return true if it is safe to serve this Morph using Nebraska." 
	^true! !

!Morph methodsFor: 'testing' stamp: 'nk 9/4/2004 17:38'!
isSelectionMorph
	^false! !

!Morph methodsFor: 'accessing' stamp: 'sw 10/27/2000 17:42'!
isShared
	"Answer whether the receiver has the #shared property.  This property allows it to be treated as a 'background' item"

	^ self hasProperty: #shared! !

!Morph methodsFor: 'testing' stamp: 'nk 6/12/2004 09:17'!
isSketchMorph
	^self class isSketchMorphClass! !

!Morph methodsFor: 'classification' stamp: 'ar 1/3/2010 16:17'!
isSoundTile
	^false! !

!Morph methodsFor: 'classification' stamp: 'ar 6/30/2001 13:13'!
isStandardViewer
	^false! !

!Morph methodsFor: 'stepping and presenter' stamp: 'sw 3/22/2000 14:28'!
isStepping
	"Return true if the receiver is currently stepping in its world"
	| aWorld |
	^ (aWorld := self world)
		ifNil:		[false]
		ifNotNil:	[aWorld isStepping: self]! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 10/22/2000 16:43'!
isSteppingSelector: aSelector
	"Return true if the receiver is currently stepping in its world"
	| aWorld |
	^ (aWorld := self world)
		ifNil:		[false]
		ifNotNil:	[aWorld isStepping: self selector: aSelector]! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:57'!
isSticky
	"answer whether the receiver is Sticky"
	extension ifNil: [^ false].
	^ extension sticky! !

!Morph methodsFor: 'classification' stamp: 'ar 1/3/2010 16:24'!
isStickySketchMorph
	^false! !

!Morph methodsFor: 'classification' stamp: 'di 11/2/2000 13:24'!
isSyntaxMorph
	^false! !

!Morph methodsFor: 'classification' stamp: 'ar 12/16/2001 18:28'!
isTextMorph
	^false! !

!Morph methodsFor: 'e-toy support' stamp: 'ar 2/7/2001 17:58'!
isTileEditor
	"No, I'm not"
	^false! !

!Morph methodsFor: 'classification' stamp: 'ar 1/3/2010 16:11'!
isTileMorph
	^false! !

!Morph methodsFor: 'classification' stamp: 'ar 1/3/2010 16:08'!
isTilePadMorph
	^false! !

!Morph methodsFor: 'classification' stamp: 'ar 1/3/2010 16:39'!
isViewer
	^false! !

!Morph methodsFor: 'classification' stamp: ''!
isWorldMorph

	^ false! !

!Morph methodsFor: 'classification' stamp: ''!
isWorldOrHandMorph

	^ self isWorldMorph or: [self isHandMorph]! !

!Morph methodsFor: '*Morphic-Sound-piano rolls' stamp: 'RAA 12/11/2000 17:21'!
justDroppedIntoPianoRoll: pianoRoll event: evt
	
	| ambientEvent startTimeInScore |
	startTimeInScore := pianoRoll timeForX: self left.

	ambientEvent := AmbientEvent new 
		morph: self;
		time: startTimeInScore.

	pianoRoll score addAmbientEvent: ambientEvent.

	"self endTime > pianoRoll scorePlayer durationInTicks ifTrue:
		[pianoRoll scorePlayer updateDuration]"
! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'cmm 10/22/2015 14:30:56'!
justDroppedInto: aMorph event: anEvent
	"This message is sent to a dropped morph after it has been dropped on -- and been accepted by -- a drop-sensitive morph"

	| partsBinCase cmd |
	(self formerOwner notNil and: [self formerOwner ~~ aMorph])
		ifTrue: [self removeHalo].
	self formerOwner: nil.
	self formerPosition: nil.
	cmd := self valueOfProperty: #undoGrabCommand.
	cmd ifNotNil:[aMorph rememberCommand: cmd.
				self removeProperty: #undoGrabCommand].
	(partsBinCase := aMorph isPartsBin) ifFalse:
		[self isPartsDonor: false].
	(self isInWorld and: [partsBinCase not]) ifTrue:
		[self world startSteppingSubmorphsOf: self].
	"Note an unhappy inefficiency here:  the startStepping... call will often have already been called in the sequence leading up to entry to this method, but unfortunately the isPartsDonor: call often will not have already happened, with the result that the startStepping... call will not have resulted in the startage of the steppage."

	"An object launched by certain parts-launcher mechanisms should end up fully visible..."
	(self hasProperty: #beFullyVisibleAfterDrop) ifTrue:
		[aMorph == ActiveWorld ifTrue:
			[self goHome].
		self removeProperty: #beFullyVisibleAfterDrop].
! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 2/6/2001 22:12'!
justGrabbedFrom: formerOwner
	"The receiver was just grabbed from its former owner and is now attached to the hand. By default, we pass this message on if we're a renderer."
	(self isRenderer and:[self hasSubmorphs]) 
		ifTrue:[self firstSubmorph justGrabbedFrom: formerOwner].! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'ar 11/24/1998 14:19'!
keepsTransform
	"Return true if the receiver will keep it's transform while being grabbed by a hand."
	^false! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:06'!
keyboardBubbleFilters
	^ self valueOfProperty: #keyboardBubbleFilters! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:06'!
keyboardBubbleFilters: anArrayOrNil
	^ self setProperty: #keyboardBubbleFilters toValue: anArrayOrNil! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:06'!
keyboardCaptureFilters
	^ self valueOfProperty: #keyboardCaptureFilters! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:06'!
keyboardCaptureFilters: anArrayOrNil
	^ self setProperty: #keyboardCaptureFilters toValue: anArrayOrNil! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/19/2015 12:02'!
keyboardFocusChange: aBoolean
	"The message is sent to a morph when its keyboard focus change. The given argument indicates that the receiver is gaining keyboard focus (versus losing) the keyboard focus. Morphs that accept keystrokes should change their appearance in some way when they are the current keyboard focus."

	self eventHandler
		ifNotNil: [:h | h keyboardFocusChange: aBoolean fromMorph: self].

	self indicateKeyboardFocus
		ifTrue: [self changed].! !

!Morph methodsFor: 'drawing' stamp: 'mt 7/25/2016 16:01'!
keyboardFocusColor

	^ self userInterfaceTheme keyboardFocusColor ifNil: [TranslucentColor r: 0.3 g: 0.5 b: 0.5 alpha: 0.5]! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/10/2015 12:18'!
keyboardFocusDelegate
	"If this morph ever tries to acquire the keyboard focus, this can be delegate to, for example, a submorph."
	
	^ self! !

!Morph methodsFor: 'drawing' stamp: 'mt 7/25/2016 16:01'!
keyboardFocusWidth

	^ self userInterfaceTheme keyboardFocusWidth ifNil: [3]! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/19/2015 11:53'!
keyDown: anEvent
	"Handle a key down event. The default response is to let my eventHandler, if any, handle it."

	self eventHandler ifNotNil:
		[self eventHandler keyDown: anEvent fromMorph: self].
! !

!Morph methodsFor: 'event handling' stamp: 'tk 8/10/1998 16:05'!
keyStroke: anEvent
	"Handle a keystroke event.  The default response is to let my eventHandler, if any, handle it."

	self eventHandler ifNotNil:
		[self eventHandler keyStroke: anEvent fromMorph: self].
! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/19/2015 11:53'!
keyUp: anEvent
	"Handle a key up event. The default response is to let my eventHandler, if any, handle it."

	self eventHandler ifNotNil:
		[self eventHandler keyUp: anEvent fromMorph: self].
! !

!Morph methodsFor: 'testing' stamp: 'md 2/27/2006 09:59'!
knownName
	"answer a name by which the receiver is known, or nil if none"
	^ extension ifNotNil: [extension externalName]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:32'!
lastSubmorph
	^submorphs last! !

!Morph methodsFor: 'layout' stamp: 'ar 11/12/2000 23:10'!
layoutBounds
	"Return the bounds for laying out children of the receiver"
	| inset box |
	inset := self layoutInset.
	box := self innerBounds.
	inset isZero ifTrue:[^box].
	^box insetBy: inset.! !

!Morph methodsFor: 'layout' stamp: 'mt 4/25/2015 19:37'!
layoutBounds: aRectangle
	"Set the bounds for laying out children of the receiver.
	Note: written so that #layoutBounds can be changed without touching this method"
	| outer inner |
	outer := self bounds.
	inner := self layoutBounds.
	self bounds: (aRectangle origin + (outer origin - inner origin) corner:
				aRectangle corner + (outer corner - inner corner)).! !

!Morph methodsFor: 'layout' stamp: 'ar 1/27/2001 14:41'!
layoutChanged
	| layout |
	fullBounds ifNil:[^self]. "layout will be recomputed so don't bother"
	fullBounds := nil.
	layout := self layoutPolicy.
	layout ifNotNil:[layout flushLayoutCache].
	owner ifNotNil: [owner layoutChanged].
	"note: does not send #ownerChanged here - we'll do this when computing the new layout"! !

!Morph methodsFor: 'layout-properties' stamp: 'md 2/27/2006 09:59'!
layoutFrame
	"Layout specific. Return the layout frame describing where the  
	receiver should appear in a proportional layout"
	^ extension ifNotNil: [extension layoutFrame]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:20'!
layoutFrame: aLayoutFrame
	"Layout specific. Return the layout frame describing where the receiver should appear in a proportional layout"
	self layoutFrame == aLayoutFrame ifTrue:[^self].
	self assureExtension layoutFrame: aLayoutFrame.
	self layoutChanged.! !

!Morph methodsFor: 'layout' stamp: 'ar 8/6/2001 09:55'!
layoutInBounds: cellBounds
	"Layout specific. Apply the given bounds to the receiver after being layed out in its owner."
	| box aSymbol delta |
	fullBounds ifNil:["We are getting new bounds here but we haven't computed the receiver's layout yet. Although the receiver has reported its minimal size before the actual size it has may differ from what would be after the layout. Normally, this isn't a real problem, but if we have #shrinkWrap constraints then the receiver's bounds may be larger than the cellBounds. THAT is a problem because the centering may not work correctly if the receiver shrinks after the owner layout has been computed. To avoid this problem, we compute the receiver's layout now. Note that the layout computation is based on the new cell bounds rather than the receiver's current bounds."
		cellBounds origin = self bounds origin ifFalse:[
			box := self outerBounds.
			delta := cellBounds origin - self bounds origin.
			self invalidRect: (box merge: (box translateBy: delta)).
			self privateFullMoveBy: delta]. "sigh..."
		box := cellBounds origin extent: "adjust for #rigid receiver"
			(self hResizing == #rigid ifTrue:[self bounds extent x] ifFalse:[cellBounds extent x]) @
			(self vResizing == #rigid ifTrue:[self bounds extent y] ifFalse:[cellBounds extent y]).
		"Compute inset of layout bounds"
		box := box origin - (self bounds origin - self layoutBounds origin) corner:
					box corner - (self bounds corner - self layoutBounds corner).
		"And do the layout within the new bounds"
		self layoutBounds: box.
		self doLayoutIn: box].
	cellBounds = self fullBounds ifTrue:[^self]. "already up to date"
	cellBounds extent = self fullBounds extent "nice fit"
		ifTrue:[^self position: cellBounds origin].
	box := bounds.
	"match #spaceFill constraints"
	self hResizing == #spaceFill 
		ifTrue:[box := box origin extent: cellBounds width @ box height].
	self vResizing == #spaceFill
		ifTrue:[box := box origin extent: box width @ cellBounds height].
	"align accordingly"
	aSymbol := (owner ifNil:[self]) cellPositioning.
	box := box align: (box perform: aSymbol) with: (cellBounds perform: aSymbol).
	"and install new bounds"
	self bounds: box.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 16:38'!
layoutInset
	"Return the extra inset for layouts"
	| props |
	props := self layoutProperties.
	^props ifNil:[0] ifNotNil:[props layoutInset].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 16:38'!
layoutInset: aNumber
	"Return the extra inset for layouts"
	self assureTableProperties layoutInset: aNumber.
	self layoutChanged.! !

!Morph methodsFor: 'layout-menu' stamp: 'dgd 10/19/2003 11:23'!
layoutMenuPropertyString: aSymbol from: currentSetting 
	| onOff wording |
	onOff := aSymbol == currentSetting
				ifTrue: ['<on>']
				ifFalse: ['<off>'].
	""
	wording := String
				streamContents: [:stream | 
					| index | 
					index := 1.
					aSymbol
						keysAndValuesDo: [:idx :ch | ch isUppercase
								ifTrue: [""stream nextPutAll: (aSymbol copyFrom: index to: idx - 1) asLowercase.
									stream nextPutAll: ' '.
									index := idx]].
					index < aSymbol size
						ifTrue: [stream nextPutAll: (aSymbol copyFrom: index to: aSymbol size) asLowercase]].
	""
	^ onOff , wording translated! !

!Morph methodsFor: 'layout-properties' stamp: 'md 2/27/2006 10:00'!
layoutPolicy
	"Layout specific. Return the layout policy describing how children 
	of the receiver should appear."
	^ extension ifNotNil: [ extension layoutPolicy]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:21'!
layoutPolicy: aLayoutPolicy
	"Layout specific. Return the layout policy describing how children of the receiver should appear."
	self layoutPolicy == aLayoutPolicy ifTrue:[^self].
	self assureExtension layoutPolicy: aLayoutPolicy.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'md 2/27/2006 09:58'!
layoutProperties
	"Return the current layout properties associated with the  
	receiver"
	^ extension ifNotNil: [ extension layoutProperties]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:21'!
layoutProperties: newProperties
	"Return the current layout properties associated with the receiver"
	self layoutProperties == newProperties ifTrue:[^self].
	self assureExtension layoutProperties: newProperties.
! !

!Morph methodsFor: 'layout' stamp: 'mt 4/24/2015 10:11'!
layoutProportionallyIn: newBounds
	"Layout specific. Apply the given bounds to the receiver."
	| box frame |
	frame := self layoutFrame ifNil:[^self].
	"compute the cell size the receiver has given its layout frame"
	
	box := frame layout: self bounds in: newBounds.
	box := box topLeft extent: (box extent max: self minExtent).
	
	(box = self bounds) ifTrue:[^self]. "no change"
	^self layoutInBounds: box.! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:16'!
left
	" Return the x-coordinate of my left side "

	^ bounds left! !

!Morph methodsFor: 'geometry' stamp: 'di 3/6/2002 13:06'!
leftCenter

	^ bounds leftCenter! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:15'!
left: aNumber
	" Move me so that my left side is at the x-coordinate aNumber. My extent (width & height) are unchanged "

	self position: (aNumber @ bounds top)! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:58'!
listCentering
	"Layout specific. This property describes how the rows/columns in a list-like layout should be centered.
		#topLeft - center at start of primary direction
		#bottomRight - center at end of primary direction
		#center - center in the middle of primary direction
		#justified - insert extra space inbetween rows/columns
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#topLeft] ifNotNil:[props listCentering].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 02:47'!
listCenteringString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self listCentering! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:05'!
listCentering: aSymbol
	"Layout specific. This property describes how the rows/columns in a list-like layout should be centered.
		#topLeft - center at start of primary direction
		#bottomRight - center at end of primary direction
		#center - center in the middle of primary direction
		#justified - insert extra space inbetween rows/columns
	"
	self assureTableProperties listCentering: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:59'!
listDirection
	"Layout specific. This property describes the direction in which a list-like layout should be applied. Possible values are:
		#leftToRight
		#rightToLeft
		#topToBottom
		#bottomToTop
	indicating the direction in which any layout should take place"
	| props |
	props := self layoutProperties.
	^props ifNil:[#topToBottom] ifNotNil:[props listDirection].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 02:47'!
listDirectionString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self listDirection! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
listDirection: aSymbol
	"Layout specific. This property describes the direction in which a list-like layout should be applied. Possible values are:
		#leftToRight
		#rightToLeft
		#topToBottom
		#bottomToTop
	indicating the direction in which any layout should take place"
	self assureTableProperties listDirection: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:59'!
listSpacing
	"Layout specific. This property describes how the heights for different rows in a table layout should be handled.
		#equal - all rows have the same height
		#none - all rows may have different heights
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#none] ifNotNil:[props listSpacing].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 02:47'!
listSpacingString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self listSpacing! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
listSpacing: aSymbol
	"Layout specific. This property describes how the heights for different rows in a table layout should be handled.
		#equal - all rows have the same height
		#none - all rows may have different heights
	"
	self assureTableProperties listSpacing: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'caching' stamp: 'jm 11/13/97 16:37'!
loadCachedState
	"Load the cached state of this morph. This method may be called to pre-load the cached state of a morph to avoid delays when it is first used. (Cached state can always be recompued on demand, so a morph should not rely on this method being called.) Implementations of this method should do 'super loadCachedState'. This default implementation does nothing."
! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:07'!
localPointToGlobal: aPoint
	^self point: aPoint in: nil! !

!Morph methodsFor: 'accessing' stamp: 'sw 8/4/97 12:05'!
lock
	self lock: true! !

!Morph methodsFor: 'menus' stamp: 'dgd 8/30/2003 20:20'!
lockedString
	"Answer the string to be shown in a menu to represent the 
	'locked' status"
	^ (self isLocked
		ifTrue: ['<on>']
		ifFalse: ['<off>']), 'be locked' translated! !

!Morph methodsFor: 'menus' stamp: 'sw 9/21/2000 22:50'!
lockUnlockMorph
	"If the receiver is locked, unlock it; if unlocked, lock it"

	self isLocked ifTrue: [self unlock] ifFalse: [self lock]! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:58'!
lock: aBoolean 
	"change the receiver's lock property"
	(extension isNil and: [aBoolean not]) ifTrue: [^ self].
	self assureExtension locked: aBoolean! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 9/6/2004 14:17'!
mainDockingBars
	"Answer the receiver's main dockingBars"
	^ self dockingBars
		select: [:each | each hasProperty: #mainDockingBarTimeStamp]! !

!Morph methodsFor: 'e-toy support' stamp: 'fbs 7/12/2013 20:50'!
makeGraphPaper
	| smallGrid backColor lineColor |
	smallGrid := Compiler evaluate: (UIManager default request: 'Enter grid size' translated initialAnswer: '16').
	smallGrid ifNil: [^ self].
	UIManager default informUser: 'Choose a background color' translated during: [backColor := Color fromUser].
	UIManager default informUser: 'Choose a line color' translated during: [lineColor := Color fromUser].
	self makeGraphPaperGrid: smallGrid background: backColor line: lineColor.! !

!Morph methodsFor: 'e-toy support' stamp: 'di 9/7/2000 20:44'!
makeGraphPaperGrid: smallGrid background: backColor line: lineColor

	| gridForm |
	gridForm := self gridFormOrigin: 0@0 grid: smallGrid asPoint background: backColor line: lineColor.
	self color: gridForm.
	self world ifNotNil: [self world fullRepaintNeeded].
	self changed: #newColor.  "propagate to view"
! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 8/6/2009 20:42'!
makeMultipleSiblings: evt
	"Make multiple siblings, first prompting the user for how many"

	| result |
	self topRendererOrSelf couldMakeSibling ifFalse: [^ Beeper beep].
	result := UIManager default request: 'how many siblings do you want?' translated initialAnswer: '2'.
	result isEmptyOrNil ifTrue: [^ self].
	result first isDigit ifFalse: [^ Beeper beep].
	self topRendererOrSelf makeSiblings: result asInteger.! !

!Morph methodsFor: 'menus' stamp: 'sw 10/29/1999 15:34'!
makeNascentScript
	^ self notYetImplemented! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 10/5/2000 19:20'!
makeNewPlayerInstance: evt
	"Make a duplicate of the receiver's argument.  This is called only where the argument has an associated Player as its costumee, and the intent here is to make another instance of the same uniclass as the donor Player itself.  Much works, but there are flaws so this shouldn't be used without recognizing the risks"

	evt hand attachMorph: self usableSiblingInstance! !

!Morph methodsFor: 'meta-actions' stamp: 'sw 1/12/2001 22:47'!
makeSiblingsLookLikeMe: evt
	"Make all my siblings wear the same costume that I am wearing."

	| aPlayer |
	(aPlayer := self topRendererOrSelf player) belongsToUniClass ifFalse: [self error: 'not uniclass'].
	aPlayer class allInstancesDo:
		[:anInstance | anInstance == aPlayer ifFalse:
			[anInstance wearCostumeOf: aPlayer]]! !

!Morph methodsFor: 'meta-actions' stamp: 'nice 12/27/2009 03:10'!
makeSiblings: count
	"Make multiple sibling, and return the list"

	| listOfNewborns aPosition |
	aPosition := self position.
	listOfNewborns := (1 to: count asInteger) asArray collect: 
		[:anIndex | | anInstance |
			anInstance := self usableSiblingInstance.
			owner addMorphFront: anInstance.
			aPosition := aPosition + (10@10).
			anInstance position: aPosition.
			anInstance].
	self currentWorld startSteppingSubmorphsOf: self topRendererOrSelf owner.
	^ listOfNewborns! !

!Morph methodsFor: 'parts bin' stamp: 'di 8/11/1998 13:02'!
markAsPartsDonor
	"Mark the receiver specially so that mouse actions on it are interpreted as 'tearing off a copy'"

	self isPartsDonor: true! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 19:59'!
maxCellSize
	"Layout specific. This property specifies the maximum size of a table cell."
	| props |
	props := self layoutProperties.
	^props ifNil:[SmallInteger maxVal] ifNotNil:[props maxCellSize].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
maxCellSize: aPoint
	"Layout specific. This property specifies the maximum size of a table cell."
	self assureTableProperties maxCellSize: aPoint.
	self layoutChanged.! !

!Morph methodsFor: 'menus' stamp: 'dgd 8/30/2003 20:15'!
maybeAddCollapseItemTo: aMenu
	"If appropriate, add a collapse item to the given menu"

	| anOwner |
	(anOwner := self topRendererOrSelf owner) ifNotNil:
			[anOwner isWorldMorph ifTrue:
				[aMenu add: 'collapse' translated target: self action: #collapse]]! !

!Morph methodsFor: 'meta-actions' stamp: 'sw 4/19/2005 14:55'!
maybeDuplicateMorph
	"Maybe duplicate the morph"

	self okayToDuplicate ifTrue:
		[self topRendererOrSelf duplicate openInHand]! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 10/5/2000 17:32'!
maybeDuplicateMorph: evt
	self okayToDuplicate ifTrue:[^self duplicateMorph: evt]! !

!Morph methodsFor: 'other events' stamp: 'sw 8/1/2001 14:08'!
menuButtonMouseEnter: event
	"The mouse entered a menu-button area; show the menu cursor temporarily"

	event hand showTemporaryCursor: Cursor menu! !

!Morph methodsFor: 'other events' stamp: 'sw 8/1/2001 14:09'!
menuButtonMouseLeave: event
	"The mouse left a menu-button area; restore standard cursor"

	event hand showTemporaryCursor: nil! !

!Morph methodsFor: 'menus' stamp: 'sw 11/22/1999 12:13'!
menuItemAfter: menuString
	| allWordings |
	allWordings := self allMenuWordings.
	^ allWordings atWrap: ((allWordings indexOf: menuString) + 1)! !

!Morph methodsFor: 'menus' stamp: 'sw 11/22/1999 12:14'!
menuItemBefore: menuString
	| allWordings |
	allWordings := self allMenuWordings.
	^ allWordings atWrap: ((allWordings indexOf: menuString) - 1)! !

!Morph methodsFor: 'accessing' stamp: 'ar 1/3/2010 16:10'!
methodCommentAsBalloonHelp
	"Given that I am a morph that is associated with an object and a method, answer a suitable method comment relating to that object & method if possible"

	| inherentSelector actual |
	(inherentSelector := self valueOfProperty: #inherentSelector)
		ifNotNil:
			[(actual := (self firstOwnerSuchThat:[:m| m isPhraseTileMorph or:[m isSyntaxMorph]]) actualObject) ifNotNil:
				[^ actual class precodeCommentOrInheritedCommentFor: inherentSelector]].
	^ nil! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:00'!
minCellSize
	"Layout specific. This property specifies the minimal size of a table cell."
	| props |
	props := self layoutProperties.
	^props ifNil:[0] ifNotNil:[props minCellSize].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
minCellSize: aPoint
	"Layout specific. This property specifies the minimal size of a table cell."
	self assureTableProperties minCellSize: aPoint.
	self layoutChanged.! !

!Morph methodsFor: 'layout' stamp: 'mt 4/24/2015 10:07'!
minExtent
	"Layout specific. Return the minimum size the receiver can be represented in.
	Implementation note: When this message is sent from an owner trying to lay out its children it will traverse down the morph tree and recompute the minimal arrangement of the morphs based on which the minimal extent is returned. When a morph with some layout strategy is encountered, the morph will ask its strategy to compute the new arrangement. However, since the final size given to the receiver is unknown at the point of the query, the assumption is made that the current bounds of the receiver are the base on which the layout should be computed. This scheme prevents strange layout changes when for instance, a table is contained in another table. Unless the inner table has been resized manually (which means its bounds are already enlarged) the arrangement of the inner table will not change here. Thus the entire layout computation is basically an iterative process which may have different results depending on the incremental changes applied."

	| layout minExtent extra hFit vFit |
	hFit := self hResizing.
	vFit := self vResizing.
	(hFit == #spaceFill or: [vFit == #spaceFill]) 
		ifFalse: 
			["The receiver will not adjust to parents layout by growing or shrinking,
		which means that an accurate layout defines the minimum size."

			^self fullBounds extent].

	"An exception -- a receiver with #shrinkWrap constraints but no children is being treated #rigid (the equivalent to a #spaceFill receiver in a non-layouting owner)"
	self hasSubmorphs 
		ifFalse: 
			[hFit == #shrinkWrap ifTrue: [hFit := #rigid].
			vFit == #shrinkWrap ifTrue: [vFit := #rigid]].
	layout := self layoutPolicy.
	layout isNil 
		ifTrue: [minExtent := 0@0]
		ifFalse: [minExtent := layout minExtentOf: self in: self layoutBounds].

	hFit == #rigid 
		ifTrue: [minExtent := self fullBounds extent x @ minExtent y]
		ifFalse: 
			[extra := self bounds width - self layoutBounds width.
			minExtent := (minExtent x + extra) @ minExtent y].
	vFit == #rigid 
		ifTrue: [minExtent := minExtent x @ self fullBounds extent y]
		ifFalse: 
			[extra := self bounds height - self layoutBounds height.
			minExtent := minExtent x @ (minExtent y + extra)].

	^ minExtent max: self minWidth @ self minHeight! !

!Morph methodsFor: 'layout' stamp: 'mt 4/24/2015 10:14'!
minExtent: aPoint

	self error: 'Use #minimumExtent: if you want to set a minimum extent. #minExtent is a layout policy-specific callback.'! !

!Morph methodsFor: 'layout' stamp: 'mt 4/24/2015 09:48'!
minHeight
	"answer the receiver's minHeight"
	^ self
		valueOfProperty: #minHeight
		ifAbsent: [1]! !

!Morph methodsFor: 'layout' stamp: 'dgd 2/22/2003 14:32'!
minHeight: aNumber 
	aNumber isNil 
		ifTrue: [self removeProperty: #minHeight]
		ifFalse: [self setProperty: #minHeight toValue: aNumber].
	self layoutChanged! !

!Morph methodsFor: 'geometry' stamp: 'mt 4/24/2015 09:38'!
minimumExtent
	
	^ self minWidth @ self minHeight! !

!Morph methodsFor: 'geometry' stamp: 'mt 4/24/2015 09:47'!
minimumExtent: aPoint
	"Do not shrink below this extent."

	self
		minWidth: aPoint x;
		minHeight: aPoint y.! !

!Morph methodsFor: 'geometry' stamp: 'mt 4/24/2015 10:18'!
minimumHeight
	"Wrapper for layout-specific function to avoid confusion."
	
	^ self minHeight! !

!Morph methodsFor: 'geometry' stamp: 'mt 4/24/2015 10:18'!
minimumHeight: aNumber
	"Wrapper for layout-specific function to avoid confusion."
	
	self minHeight: aNumber.! !

!Morph methodsFor: 'geometry' stamp: 'mt 4/24/2015 10:18'!
minimumWidth
	"Wrapper for layout-specific function to avoid confusion."
	
	^ self minWidth! !

!Morph methodsFor: 'geometry' stamp: 'mt 4/24/2015 10:18'!
minimumWidth: aNumber
	"Wrapper for layout-specific function to avoid confusion."
	
	self minWidth: aNumber.! !

!Morph methodsFor: 'layout' stamp: 'mt 4/24/2015 09:48'!
minWidth
	"answer the receiver's minWidth"
	^ self
		valueOfProperty: #minWidth
		ifAbsent: [1]! !

!Morph methodsFor: 'layout' stamp: 'dgd 2/22/2003 14:32'!
minWidth: aNumber 
	aNumber isNil 
		ifTrue: [self removeProperty: #minWidth]
		ifFalse: [self setProperty: #minWidth toValue: aNumber].
	self layoutChanged! !

!Morph methodsFor: 'polymorph' stamp: 'gvc 9/11/2006 09:37'!
modalLockTo: aSystemWindow
	"Lock the receiver as a modal owner of the given window."

	self lock! !

!Morph methodsFor: 'polymorph' stamp: 'gvc 9/11/2006 09:38'!
modalUnlockFrom: aSystemWindow
	"Unlock the receiver as a modal owner of the given window."

	self unlock! !

!Morph methodsFor: 'menus' stamp: 'dgd 9/22/2004 20:30'!
model
	^ nil ! !

!Morph methodsFor: 'accessing' stamp: 'sw 10/23/1999 22:35'!
modelOrNil
	^ nil! !

!Morph methodsFor: 'model access' stamp: 'cmm 5/8/2014 10:58'!
models
	"Answer a collection of whatever models I may have."
	^ self modelOrNil
		ifNil: [ Array empty ]
		ifNotNil: [ Array with: self modelOrNil ]! !

!Morph methodsFor: 'testing' stamp: 'RAA 10/20/2000 14:47'!
modificationHash

	^String 
		streamContents: [ :strm |
			self longPrintOn: strm
		]
		limitedTo: 25
! !

!Morph methodsFor: 'WiW support' stamp: 'RAA 7/19/2000 20:44'!
morphicLayerNumber

	"helpful for insuring some morphs always appear in front of or behind others.
	smaller numbers are in front"

	^(owner isNil or: [owner isWorldMorph]) ifTrue: [
		self valueOfProperty: #morphicLayerNumber ifAbsent: [100]
	] ifFalse: [
		owner morphicLayerNumber
	].

	"leave lots of room for special things"! !

!Morph methodsFor: 'WiW support' stamp: 'RAA 2/16/2001 13:54'!
morphicLayerNumberWithin: anOwner

	"helpful for insuring some morphs always appear in front of or behind others.
	smaller numbers are in front"

	^(owner isNil or: [owner isWorldMorph or: [anOwner == owner]]) ifTrue: [
		self valueOfProperty: #morphicLayerNumber ifAbsent: [100]
	] ifFalse: [
		owner morphicLayerNumber
	].

	"leave lots of room for special things"! !

!Morph methodsFor: 'structure' stamp: 'sw 8/29/2000 14:55'!
morphPreceding: aSubmorph
	"Answer the morph immediately preceding aSubmorph, or nil if none"

	| anIndex |
	anIndex := submorphs indexOf: aSubmorph ifAbsent: [^ nil].
	^ anIndex > 1
		ifTrue:
			[submorphs at: (anIndex - 1)]
		ifFalse:
			[nil]! !

!Morph methodsFor: 'printing' stamp: 'RAA 2/26/2001 07:22'!
morphReport

	^self morphReportFor: #(hResizing vResizing bounds)! !

!Morph methodsFor: 'printing' stamp: 'RAA 2/25/2001 17:47'!
morphReportFor: attributeList

	| s |

	s := WriteStream on: String new.
	self
		morphReportFor: attributeList 
		on: s 
		indent: 0.
	StringHolder new contents: s contents; openLabel: 'morph report'! !

!Morph methodsFor: 'printing' stamp: 'RAA 2/25/2001 17:48'!
morphReportFor: attributeList on: aStream indent: anInteger

	anInteger timesRepeat: [aStream tab].
	aStream print: self; space.
	attributeList do: [ :a | aStream print: (self perform: a); space].
	aStream cr.
	submorphs do: [ :sub |
		sub morphReportFor: attributeList on: aStream indent: anInteger + 1
	].! !

!Morph methodsFor: 'thumbnail' stamp: 'sw 8/16/2000 17:40'!
morphRepresented
	"If the receiver is an alias, answer the morph it represents; else answer self"

	^ self! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 10/8/2000 15:40'!
morphsAt: aPoint
	"Return a collection of all morphs in this morph structure that contain the given point, possibly including the receiver itself.  The order is deepest embedding first."
	^self morphsAt: aPoint unlocked: false! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'nice 12/27/2009 03:10'!
morphsAt: aPoint behind: aMorph unlocked: aBool 
	"Return all morphs at aPoint that are behind frontMorph; if aBool is true return only unlocked, visible morphs."

	| isBack all tfm |
	all := (aMorph isNil or: [owner isNil]) 
				ifTrue: 
					["Traverse down"

					(self fullBounds containsPoint: aPoint) ifFalse: [^#()].
					(aBool and: [self isLocked or: [self visible not]]) ifTrue: [^#()].
					nil]
				ifFalse: 
					["Traverse up"

					tfm := self transformedFrom: owner.
					all := owner 
								morphsAt: (tfm localPointToGlobal: aPoint)
								behind: self
								unlocked: aBool.
					WriteStream with: all].
	isBack := aMorph isNil.
	self submorphsDo: 
			[:m | | found | 
			isBack 
				ifTrue: 
					[tfm := m transformedFrom: self.
					found := m 
								morphsAt: (tfm globalPointToLocal: aPoint)
								behind: nil
								unlocked: aBool.
					found notEmpty 
						ifTrue: 
							[all ifNil: [all := WriteStream on: #()].
							all nextPutAll: found]].
			m == aMorph ifTrue: [isBack := true]].
	(isBack and: [self containsPoint: aPoint]) 
		ifTrue: 
			[all ifNil: [^Array with: self].
			all nextPut: self].
	^all ifNil: [#()] ifNotNil: [all contents]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 10/8/2000 15:40'!
morphsAt: aPoint unlocked: aBool
	"Return a collection of all morphs in this morph structure that contain the given point, possibly including the receiver itself.  The order is deepest embedding first."
	| mList |
	mList := WriteStream on: #().
	self morphsAt: aPoint unlocked: aBool do:[:m| mList nextPut: m].
	^mList contents! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'nice 12/27/2009 03:10'!
morphsAt: aPoint unlocked: aBool do: aBlock
	"Evaluate aBlock with all the morphs starting at the receiver which appear at aPoint. If aBool is true take only visible, unlocked morphs into account."
	
	(self fullBounds containsPoint: aPoint) ifFalse:[^self].
	(aBool and:[self isLocked or:[self visible not]]) ifTrue:[^self].
	self submorphsDo:[:m| | tfm |
		tfm := m transformedFrom: self.
		m morphsAt: (tfm globalPointToLocal: aPoint) unlocked: aBool do: aBlock].
	(self containsPoint: aPoint) ifTrue:[aBlock value: self].! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 9/9/2000 17:31'!
morphsInFrontOf: someMorph overlapping: aRectangle do: aBlock
	"Evaluate aBlock with all top-level morphs in front of someMorph that overlap with the given rectangle. someMorph is either an immediate child of the receiver or nil (in which case all submorphs of the receiver are enumerated)."
	self submorphsDo:[:m|
		m == someMorph ifTrue:["Try getting out quickly"
			owner ifNil:[^self].
			^owner morphsInFrontOf: self overlapping: aRectangle do: aBlock].
		(m fullBoundsInWorld intersects: aRectangle)
			ifTrue:[aBlock value: m]].
	owner ifNil:[^self].
	^owner morphsInFrontOf: self overlapping: aRectangle do: aBlock.! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 9/9/2000 17:31'!
morphsInFrontOverlapping: aRectangle
	"Return all top-level morphs in front of someMorph that overlap with the given rectangle."
	| morphList |
	morphList := WriteStream on: Array new.
	self morphsInFrontOf: nil overlapping: aRectangle do:[:m | morphList nextPut: m].
	^morphList contents! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 9/9/2000 17:31'!
morphsInFrontOverlapping: aRectangle do: aBlock
	"Evaluate aBlock with all top-level morphs in front of someMorph that overlap with the given rectangle."
	^self morphsInFrontOf: nil overlapping: aRectangle do: aBlock! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'cmm 3/11/2010 21:40'!
morphToDropInPasteUp: aPasteUp
	^ self! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:06'!
mouseBubbleFilters
	^ self valueOfProperty: #mouseBubbleFilters! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:07'!
mouseBubbleFilters: anArrayOrNil
	^ self setProperty: #mouseBubbleFilters toValue: anArrayOrNil! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:06'!
mouseCaptureFilters
	^ self valueOfProperty: #mouseCaptureFilters! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:07'!
mouseCaptureFilters: anArrayOrNil
	^ self setProperty: #mouseCaptureFilters toValue: anArrayOrNil! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 10/3/2000 17:05'!
mouseDownOnHelpHandle: anEvent
	"The mouse went down in the show-balloon handle"
	
	| str |
	anEvent shiftPressed ifTrue: [^ self editBalloonHelpText].
	str := self balloonText.
	str ifNil: [str := self noHelpString].
	self showBalloon: str hand: anEvent hand.
! !

!Morph methodsFor: 'events-processing' stamp: 'mt 4/19/2015 12:23'!
mouseDownPriority
	"Return the default mouse down priority for the receiver. This supports a morph to get mouse down before one of its children."

	^ self
		valueOfProperty: #mouseDownPriority
		ifAbsent: [(self isPartsDonor or: [self isPartsBin])
			ifTrue:	[50]
			ifFalse:	[0]].
		
	"The above is a workaround for the complete confusion between parts donors and parts bins. Morphs residing in a parts bin may or may not have the parts donor property set; if they have they may or may not actually handle events. To work around this, parts bins get an equal priority to parts donors so that when a morph in the parts bin does have the property set but does not handle the event we still get a copy from picking it up through the parts bin. Argh. This just *cries* for a cleanup."
	"And the above comment is Andreas's from 10/2000, which was formerly retrievable by a #flag: call which however caused a problem when trying to recompile the method from decompiled source."! !

!Morph methodsFor: 'events-processing' stamp: 'mt 4/19/2015 12:24'!
mouseDownPriority: anInteger
	
	anInteger = 0 ifTrue: [^ self removeProperty: #mouseDownPriority].
	
	self
		setProperty: #mouseDownPriority
		toValue: anInteger.! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/19/2015 12:15'!
mouseDown: evt 
	"Handle a mouse down event. The default response is to let my 
	eventHandler, if any, handle it."

	self eventHandler
		ifNotNil: [self eventHandler mouseDown: evt fromMorph: self].

	"Check for option (menu) click"
	evt yellowButtonPressed
		ifTrue: [^ self yellowButtonActivity: evt shiftPressed].
! !

!Morph methodsFor: 'event handling' stamp: 'di 9/14/1998 07:33'!
mouseEnterDragging: evt
	"Handle a mouseEnterDragging event, meaning the mouse just entered my bounds with a button pressed or laden with submorphs.  The default response is to let my eventHandler, if any, handle it, or else to do nothing."

	self eventHandler ifNotNil:
		[^ self eventHandler mouseEnterDragging: evt fromMorph: self].
! !

!Morph methodsFor: 'event handling' stamp: 'tk 8/10/1998 16:00'!
mouseEnter: evt
	"Handle a mouseEnter event, meaning the mouse just entered my bounds with no button pressed. The default response is to let my eventHandler, if any, handle it."

	self eventHandler ifNotNil:
		[self eventHandler mouseEnter: evt fromMorph: self].
! !

!Morph methodsFor: 'event handling' stamp: 'di 9/14/1998 07:38'!
mouseLeaveDragging: evt
	"Handle a mouseLeaveLaden event, meaning the mouse just left my bounds with a button pressed or laden with submorphs. The default response is to let my eventHandler, if any, handle it; else to do nothing."

	self eventHandler ifNotNil:
		[self eventHandler mouseLeaveDragging: evt fromMorph: self]! !

!Morph methodsFor: 'event handling' stamp: 'tk 8/10/1998 16:01'!
mouseLeave: evt
	"Handle a mouseLeave event, meaning the mouse just left my bounds with no button pressed. The default response is to let my eventHandler, if any, handle it."

	self eventHandler ifNotNil:
		[self eventHandler mouseLeave: evt fromMorph: self].
! !

!Morph methodsFor: 'event handling' stamp: 'ar 10/25/2000 18:04'!
mouseMove: evt
	"Handle a mouse move event. The default response is to let my eventHandler, if any, handle it."
	self eventHandler ifNotNil:
		[self eventHandler mouseMove: evt fromMorph: self].
! !

!Morph methodsFor: 'geniestubs' stamp: 'nk 3/11/2004 17:30'!
mouseStillDownStepRate
	"At what rate do I want to receive #mouseStillDown: notifications?"
	^1! !

!Morph methodsFor: 'event handling' stamp: 'ar 10/25/2000 18:02'!
mouseStillDownThreshold
	"Return the number of milliseconds after which mouseStillDown: should be sent"
	^200! !

!Morph methodsFor: 'event handling' stamp: 'ar 10/22/2000 17:08'!
mouseStillDown: evt
	"Handle a mouse move event. The default response is to let my eventHandler, if any, handle it."

	self eventHandler ifNotNil:
		[self eventHandler mouseStillDown: evt fromMorph: self].
! !

!Morph methodsFor: 'debug and other' stamp: 'sw 2/6/2001 22:35'!
mouseUpCodeOrNil
	"If the receiver has a mouseUpCodeToRun, return it, else return nil"

	^ self valueOfProperty: #mouseUpCodeToRun ifAbsent: [nil]! !

!Morph methodsFor: 'event handling' stamp: 'tk 8/10/1998 16:05'!
mouseUp: evt
	"Handle a mouse up event. The default response is to let my eventHandler, if any, handle it."

	self eventHandler ifNotNil:
		[self eventHandler mouseUp: evt fromMorph: self].
! !

!Morph methodsFor: 'event handling' stamp: 'mt 6/6/2016 17:02'!
mouseWheel: anEvent 
	"Handle a mouse wheel event. The default response is to let my eventHandler, if any, handle it."

	self eventHandler ifNotNil: [:handler |
		handler mouseWheel: anEvent fromMorph: self].! !

!Morph methodsFor: 'event handling' stamp: 'dgd 8/28/2004 18:20'!
moveOrResizeFromKeystroke: anEvent 
	"move or resize the receiver based on a keystroke"
	| dir | 

	anEvent keyValue = 28 ifTrue: [dir := -1 @ 0].
	anEvent keyValue = 29 ifTrue: [dir := 1 @ 0].
	anEvent keyValue = 30 ifTrue: [dir := 0 @ -1].
	anEvent keyValue = 31 ifTrue: [dir := 0 @ 1].

	dir notNil
		ifTrue:[
			anEvent controlKeyPressed ifTrue: [dir := dir * 10].

			anEvent shiftPressed
				ifTrue: [self extent: self extent + dir]
				ifFalse: [self position: self position + dir].

			"anEvent wasHandled: true."
	]
! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 4/16/1998 13:46'!
mustBeBackmost
	"Answer whether the receiver needs to be the backmost morph in its owner's submorph list"

	^ false! !

!Morph methodsFor: 'naming' stamp: 'sw 9/21/2000 13:18'!
nameForFindWindowFeature
	"Answer the name to show in a list of windows-and-morphs to represent the receiver"

	^ self knownName ifNil: [self class name]! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'sw 3/27/2001 11:52'!
nameForUndoWording
	"Return wording appropriate to the receiver for use in an undo-related menu item (and perhaps elsewhere)"

	| aName |
	aName := self knownName ifNil: [self renderedMorph class name].
	^ aName truncateTo: 24! !

!Morph methodsFor: 'naming' stamp: 'dgd 2/22/2003 14:33'!
nameInModel
	"Return the name for this morph in the underlying model or nil."

	| w |
	w := self world.
	w isNil ifTrue: [^nil] ifFalse: [^w model nameFor: self]! !

!Morph methodsFor: 'naming' stamp: 'sw 10/31/2000 09:24'!
nameOfObjectRepresented
	"Answer the external name of the object represented"

	^ self externalName! !

!Morph methodsFor: 'naming' stamp: 'gm 2/22/2003 13:16'!
name: aName 
	(aName isString) ifTrue: [self setNameTo: aName]! !

!Morph methodsFor: 'structure' stamp: 'di 11/12/2000 16:13'!
nearestOwnerThat: conditionBlock
	"Return the first enclosing morph for which aBlock evaluates to true, or nil if none"

	^ self firstOwnerSuchThat: conditionBlock
! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'ar 2/16/1999 18:59'!
newTransformationMorph
	^TransformationMorph new! !

!Morph methodsFor: 'halos and balloon help' stamp: 'em 3/24/2005 10:05'!
noHelpString
	^ 'Help not yet supplied' translated! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 9/13/2002 17:45'!
noteDecimalPlaces: aNumber forGetter: aGetter
	"Make a mental note of the user's preference for a particular number of decimal places to be associated with the slot with the given getter"

	(self renderedMorph valueOfProperty: #decimalPlacePreferences ifAbsentPut: [IdentityDictionary new])
		at: aGetter put: aNumber! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 8/13/2003 11:32'!
noteNewOwner: aMorph
	"I have just been added as a submorph of aMorph"! !

!Morph methodsFor: 'objects from disk' stamp: 'dgd 2/22/2003 14:33'!
objectForDataStream: refStrm 
	"I am being written out on an object file"

	| dp |
	self sqkPage ifNotNil: 
			[refStrm rootObject == self | (refStrm rootObject == self sqkPage) 
				ifFalse: 
					[self url notEmpty 
						ifTrue: 
							[dp := self sqkPage copyForSaving.	"be careful touching this object!!"
							refStrm replace: self with: dp.
							^dp]]].
	self prepareToBeSaved.	"Amen"
	^self! !

!Morph methodsFor: 'e-toy support' stamp: 'ar 1/3/2010 16:39'!
objectViewed
	"Answer the morph associated with the player that the structure the receiver currently finds itself within represents."

	^ (self outermostMorphThat: [:o | o isViewer or:[ o isScriptEditorMorph]]) objectViewed
! !

!Morph methodsFor: 'geometry testing' stamp: 'sw 8/12/2008 18:10'!
obtrudesBeyondContainer
	"Answer whether the receiver obtrudes beyond the bounds of its container"

	| top formerOwner |
	top := self topRendererOrSelf.
	top owner ifNil: [^ false].
	^ top owner isHandMorph
		ifTrue:
			[((formerOwner := top formerOwner) notNil and: [formerOwner isInWorld])
				ifFalse:
					[false]
				ifTrue:
					[(formerOwner boundsInWorld containsRect: top boundsInWorld) not]]
		ifFalse:
			[(top owner bounds containsRect: top bounds) not]! !

!Morph methodsFor: 'menu' stamp: 'tak 7/17/2004 18:20'!
offerCostumeViewerMenu: aMenu
	"do nothing"! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 1/3/2010 19:51'!
okayToAddDismissHandle
	"Answer whether a halo on the receiver should offer a dismiss handle.  This provides a hook for making it harder to disassemble some strucures even momentarily"

	^ self resistsRemoval not! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 1/3/2010 19:51'!
okayToAddGrabHandle
	"Answer whether a halo on the receiver should offer a grab handle.  This provides a hook for making it harder to deconstruct some strucures even momentarily"

	^ true! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 11/27/2001 14:50'!
okayToBrownDragEasily
	"Answer whether it it okay for the receiver to be brown-dragged easily -- i.e. repositioned within its container without extracting it.  At present this is just a hook -- nobody declines."

	^ true



"
	^ (self topRendererOrSelf owner isKindOf: PasteUpMorph) and:
		[self layoutPolicy isNil]"! !

!Morph methodsFor: 'player' stamp: 'sw 1/22/2001 14:25'!
okayToDuplicate
	"Formerly this protocol was used to guard against awkward situations when there were anonymous scripts in the etoy system.  Nowadays we just always allow duplication"

	^ true! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 11/27/2001 15:02'!
okayToExtractEasily
	"Answer whether it it okay for the receiver to be extracted easily.  Not yet hooked up to the halo-permissions mechanism."

	^ self topRendererOrSelf owner dragNDropEnabled! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 11/29/2001 06:29'!
okayToResizeEasily
	"Answer whether it is appropriate to have the receiver be easily resized by the user from the halo"

	^ true

	"This one was too jarring, not that it didn't most of the time do the right  thing but because some of the time it didn't, such as in a holder.  If we pursue this path, the test needs to be airtight, obviously...
	^ (self topRendererOrSelf owner isKindOf: PasteUpMorph) and:
		[self layoutPolicy isNil]"! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 11/27/2001 14:44'!
okayToRotateEasily
	"Answer whether it is appropriate for a rotation handle to be shown for the receiver.  This is a hook -- at present nobody declines."

	^ true! !

!Morph methodsFor: 'event handling' stamp: 'sw 9/18/2006 13:59'!
on: eventName send: selector to: recipient
	"When the given event occurs, send the given selector to the given recipient.  If the given selector is nil, rescind any earlier handling for the given event type,"

	self eventHandler ifNil:
		[selector ifNil: [^ self].  "Don't pointlessly create an event handler!!"
		self eventHandler: EventHandler new].
	self eventHandler on: eventName send: selector to: recipient! !

!Morph methodsFor: 'event handling' stamp: 'ar 3/18/2001 17:21'!
on: eventName send: selector to: recipient withValue: value
	"NOTE: selector must take 3 arguments, of which value will be the *** FIRST ***"

	self eventHandler ifNil: [self eventHandler: EventHandler new].
	self eventHandler on: eventName send: selector to: recipient withValue: value
! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 1/3/2010 16:09'!
openAPropertySheet

	Smalltalk at: #ObjectPropertiesMorph ifPresent:[:aClass|
		^aClass basicNew
			targetMorph: self;
			initialize;
			openNearTarget
	].
	Beeper beep.! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 1/3/2010 16:26'!
openATextPropertySheet

	"should only be sent to morphs that are actually supportive"

	Smalltalk at: #TextPropertiesMorph ifPresent:[:aClass|
		^aClass basicNew
			targetMorph: self;
			initialize;
			openNearTarget
	].
	Beeper beep.! !

!Morph methodsFor: 'initialization' stamp: 'RAA 10/18/2000 12:33'!
openCenteredInWorld

	self 
		fullBounds;
		position: Display extent - self extent // 2;
		openInWorld.! !

!Morph methodsFor: 'initialization' stamp: 'sw 3/21/2000 14:46'!
openInHand
	"Attach the receiver to the current hand in the current morphic world"

	self currentHand attachMorph: self! !

!Morph methodsFor: 'initialization' stamp: 'djp 10/24/1999 17:13'!
openInWindow

	^self openInWindowLabeled: self defaultLabelForInspector
! !

!Morph methodsFor: 'initialization' stamp: 'sma 4/22/2000 20:28'!
openInWindowLabeled: aString

	^self openInWindowLabeled: aString inWorld: self currentWorld! !

!Morph methodsFor: 'initialization' stamp: 'mt 5/17/2016 15:59'!
openInWindowLabeled: aString inWorld: aWorld

	| window extent |

	window := (SystemWindow labelled: aString) model: nil.
	window 
		" guess at initial extent"
		bounds:  (RealEstateAgent initialFrameFor: window initialExtent: self fullBounds extent world: aWorld);
		addMorph: self frame: (0@0 extent: 1@1);
		updatePaneColors.
	" calculate extent after adding in case any size related attributes were changed.  Use
	fullBounds in order to trigger re-layout of layout morphs"
	extent := self fullBounds extent + 
			(window borderWidth@window labelHeight) + window borderWidth.
	window extent: extent.
	aWorld addMorph: window.
	window beKeyWindow.
	aWorld startSteppingSubmorphsOf: window.
	^window
! !

!Morph methodsFor: 'initialization' stamp: 'ar 8/15/2009 14:54'!
openInWorld
        "Add this morph to the world."

      self openInWorld: self currentWorld.! !

!Morph methodsFor: 'initialization' stamp: 'dgd 9/1/2004 16:12'!
openInWorld: aWorld
	"Add this morph to the requested World."
	(aWorld visibleClearArea origin ~= (0@0) and: [self position = (0@0)]) ifTrue:
		[self position: aWorld visibleClearArea origin].
	aWorld addMorph: self.
	aWorld startSteppingSubmorphsOf: self! !

!Morph methodsFor: 'polymorph' stamp: 'nice 12/15/2013 22:33'!
openModal: aSystemWindow
	"Open the given window locking the receiver until it is dismissed.
	Answer the system window.
	Restore the original keyboard focus when closed."

	|area mySysWin keyboardFocus|
	keyboardFocus := self activeHand keyboardFocus.
	mySysWin := self isSystemWindow ifTrue: [self] ifFalse: [self ownerThatIsA: SystemWindow].
	mySysWin ifNil: [mySysWin := self].
	mySysWin modalLockTo: aSystemWindow.
	area := RealEstateAgent maximumUsableArea.
	aSystemWindow extent: aSystemWindow initialExtent.
	aSystemWindow position = (0@0)
		ifTrue: [aSystemWindow
				position: self activeHand position - (aSystemWindow extent // 2)].
	aSystemWindow
		bounds: (aSystemWindow bounds translatedToBeWithin: area).
	[ToolBuilder default runModal: aSystemWindow openAsIs]
		ensure: [mySysWin modalUnlockFrom: aSystemWindow.
				self activeHand newKeyboardFocus: keyboardFocus].
	^aSystemWindow! !

!Morph methodsFor: 'initialization' stamp: 'cmm 12/4/2010 14:24'!
openNearMorph: aMorph 
	self
		openNear: aMorph boundsInWorld
		in: (aMorph world ifNil: [ World ])! !

!Morph methodsFor: 'initialization' stamp: 'cmm 12/4/2010 14:25'!
openNear: aRectangle 
	self
		openNear: aRectangle
		in: World! !

!Morph methodsFor: 'initialization' stamp: 'cmm 12/4/2010 14:21'!
openNear: aRectangle in: aWorld
	| wb leftOverlap rightOverlap topOverlap bottomOverlap best |
	wb := aWorld bounds.
	self fullBounds.
	leftOverlap := self width - (aRectangle left - wb left).
	rightOverlap := self width - (wb right - aRectangle right).
	topOverlap := self height - (aRectangle top - wb top).
	bottomOverlap := self height - (wb bottom - aRectangle bottom).
	best := nil.
	{
		{leftOverlap. #topRight:. #topLeft}.
		{rightOverlap. #topLeft:. #topRight}.
		{topOverlap. #bottomLeft:. #topLeft}.
		{bottomOverlap. #topLeft:. #bottomLeft}.
	} do: [ :tuple |
		(best isNil or: [tuple first < best first]) ifTrue: [best := tuple].
	].
	self perform: best second with: (aRectangle perform: best third).
	self bottom: (self bottom min: wb bottom) rounded.
	self right: (self right min: wb right) rounded.
	self top: (self top max: wb top) rounded.
	self left: (self left max: wb left) rounded.
	self openInWorld: aWorld.! !

!Morph methodsFor: 'player viewer' stamp: 'tak 9/2/2006 11:57'!
openViewerForArgument
	Cursor wait
		showWhile: [self presenter viewMorph: self]! !

!Morph methodsFor: 'structure' stamp: 'di 11/13/2000 00:49'!
orOwnerSuchThat: conditionBlock

	(conditionBlock value: self) ifTrue: [^ self].
	self allOwnersDo: [:m | (conditionBlock value: m) ifTrue: [^ m]].
	^ nil

! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 09:47'!
otherProperties
	"answer the receiver's otherProperties"
	^ extension ifNotNil: [extension otherProperties]! !

!Morph methodsFor: 'geometry' stamp: 'ar 11/12/2000 22:06'!
outerBounds
	"Return the 'outer' bounds of the receiver, e.g., the bounds that need to be invalidated when the receiver changes."
	| box |
	box := self bounds.
	self hasDropShadow ifTrue:[box := self expandFullBoundsForDropShadow: box].
	self hasRolloverBorder ifTrue:[box := self expandFullBoundsForRolloverBorder: box].
	^box! !

!Morph methodsFor: 'structure' stamp: 'di 11/13/2000 00:50'!
outermostMorphThat: conditionBlock
	"Return the outermost containing morph for which aBlock is true, or nil if none"

	| outermost |
	self allOwnersDo: [:m | (conditionBlock value: m) ifTrue: [outermost := m]].
	^ outermost! !

!Morph methodsFor: 'menu' stamp: 'nk 3/10/2004 19:51'!
outermostOwnerWithYellowButtonMenu
	"Answer me or my outermost owner that is willing to contribute menu items to a context menu.
	Don't include the world."

	| outermost |
	outermost := self outermostMorphThat: [ :ea |
		ea isWorldMorph not and: [ ea hasYellowButtonMenu ]].
	^outermost ifNil: [ self hasYellowButtonMenu ifTrue: [ self ] ifFalse: []] ! !

!Morph methodsFor: 'structure' stamp: 'ar 3/18/2001 00:12'!
outermostWorldMorph

	| outer |
	World ifNotNil:[^World].
	self flag: #arNote. "stuff below is really only for MVC"
	outer := self outermostMorphThat: [ :x | x isWorldMorph].
	outer ifNotNil: [^outer].
	self isWorldMorph ifTrue: [^self].
	^nil! !

!Morph methodsFor: 'initialization' stamp: 'ar 1/31/2001 13:58'!
outOfWorld: aWorld
	"The receiver has just appeared in a new world. Notes:
		* aWorld can be nil (due to optimizations in other places)
		* owner is still valid
	Important: Keep this method fast - it is run whenever morphs are removed."
	aWorld ifNil:[^self].
	"ar 1/31/2001: We could explicitly stop stepping the receiver here but for the sake of speed I'm for now relying on the lazy machinery in the world itself."
	"aWorld stopStepping: self."
	self submorphsDo:[:m| m outOfWorld: aWorld].
! !

!Morph methodsFor: 'geometry' stamp: 'kfr 11/28/2009 17:14'!
overlapsShadowForm: itsShadow bounds: itsBounds
	"Answer true if itsShadow and my shadow overlap at all"
	| overlapExtent overlap myRect myShadow goalRect goalShadow bb |
	overlap _ self fullBounds intersect: itsBounds.
	overlapExtent _ overlap extent.
	overlapExtent > (0 @ 0)
		ifFalse: [^ false].
	myRect := overlap translateBy: 0 @ 0 - self topLeft.
	myShadow := (self imageForm contentsOfArea: myRect) stencil.
	goalRect := overlap translateBy: 0 @ 0 - itsBounds topLeft.
	goalShadow := (itsShadow contentsOfArea: goalRect) stencil.

			"compute a pixel-by-pixel AND of the two stencils.  Result will be black 
			(pixel value = 1) where black parts of the stencils overlap"
			bb := BitBlt toForm: myShadow.
			bb 
				copyForm: goalShadow
				to: 0 @ 0
				rule: Form and.
	
	^(bb destForm tallyPixelValues second) > 0 ! !

!Morph methodsFor: 'structure' stamp: ''!
owner
	"Returns the owner of this morph, which may be nil."

	^ owner! !

!Morph methodsFor: 'debug and other' stamp: 'dgd 2/22/2003 19:05'!
ownerChain
	"Answer a list of objects representing the receiver and all of its owners.   The first element is the receiver, and the last one is typically the world in which the receiver resides"

	| c next |
	c := OrderedCollection with: self.
	next := self.
	[(next := next owner) notNil] whileTrue: [c add: next].
	^c asArray! !

!Morph methodsFor: 'change reporting' stamp: 'sw 7/8/1998 13:21'!
ownerChanged
	"The receiver's owner, some kind of a pasteup, has changed its layout."

	self snapToEdgeIfAppropriate! !

!Morph methodsFor: 'structure' stamp: 'di 11/12/2000 16:18'!
ownerThatIsA: aClass
	"Return the first enclosing morph that is a kind of aClass, or nil if none"

	^ self firstOwnerSuchThat: [:m | m isKindOf: aClass]! !

!Morph methodsFor: 'structure' stamp: 'di 11/12/2000 16:20'!
ownerThatIsA: firstClass orA: secondClass
	"Return the first enclosing morph that is a kind of one of the two classes given, or nil if none"

	^ self firstOwnerSuchThat: [:m | (m isKindOf: firstClass) or: [m isKindOf: secondClass]]! !

!Morph methodsFor: 'printing' stamp: 'RAA 2/1/2001 17:42'!
pagesHandledAutomatically

	^false! !

!Morph methodsFor: 'parts bin' stamp: 'ar 10/6/2000 22:45'!
partRepresented
	^self! !

!Morph methodsFor: 'structure' stamp: 'sw 7/1/1998 18:02'!
pasteUpMorph
	"Answer the closest containing morph that is a PasteUp morph"
	^ self ownerThatIsA: PasteUpMorph! !

!Morph methodsFor: 'structure' stamp: 'dgd 8/28/2004 18:43'!
pasteUpMorphHandlingTabAmongFields
	"Answer the nearest PasteUpMorph in my owner chain that has the tabAmongFields property, or nil if none"

	| aPasteUp |
	aPasteUp := self owner.
	[aPasteUp notNil] whileTrue:
		[aPasteUp tabAmongFields ifTrue:
			[^ aPasteUp].
		aPasteUp := aPasteUp owner].
	^ nil! !

!Morph methodsFor: '*Morphic-Sound-piano rolls' stamp: 'RAA 12/11/2000 23:21'!
pauseFrom: scorePlayer

	"subclasses should take five"! !

!Morph methodsFor: 'thumbnail' stamp: 'sw 6/16/1999 11:29'!
permitsThumbnailing
	^ true! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 09:54'!
player
	"answer the receiver's player"
	^ extension ifNotNil: [extension player]! !

!Morph methodsFor: 'accessing' stamp: 'dgd 4/4/2006 17:18'!
playerRepresented
	"Answer the player represented by the receiver.  Morphs that serve as references to other morphs reimplement this; be default a morph represents its own player."

	^ self player! !

!Morph methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:53'!
player: anObject 
	"change the receiver's player"
	self assureExtension player: anObject! !

!Morph methodsFor: 'player commands' stamp: 'gk 2/23/2004 21:08'!
playSoundNamed: soundName
	"Play the sound with the given name.
	Does nothing if this image lacks sound playing facilities."

	SoundService default playSoundNamed: soundName asString! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:02'!
pointFromWorld: aPoint
	^self point: aPoint from: self world! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:03'!
pointInWorld: aPoint
	^self point: aPoint in: self world! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:01'!
point: aPoint from: aReferenceMorph

	owner ifNil: [^ aPoint].
	^ (owner transformFrom: aReferenceMorph) globalPointToLocal: aPoint.
! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/25/2000 15:01'!
point: aPoint in: aReferenceMorph

	owner ifNil: [^ aPoint].
	^ (owner transformFrom: aReferenceMorph) localPointToGlobal: aPoint.
! !

!Morph methodsFor: 'geometry' stamp: ''!
position

	^ bounds topLeft! !

!Morph methodsFor: 'geometry' stamp: 'di 9/30/1998 12:11'!
positionInWorld

	^ self pointInWorld: self position.
! !

!Morph methodsFor: 'geometry' stamp: 'sw 10/9/1998 08:56'!
positionSubmorphs
	self submorphsDo:
		[:aMorph | aMorph snapToEdgeIfAppropriate]! !

!Morph methodsFor: 'geometry' stamp: 'laza 4/10/2010 09:51'!
position: aPoint 
	"Change the position of this morph and and all of its
	submorphs. "
	| delta box |
	delta := (aPoint - bounds topLeft) rounded.
	(delta x = 0
			and: [delta y = 0])
		ifTrue: [^ self].
	"Null change"
	box := self fullBounds.
	(delta dotProduct: delta)
			> 100
		ifTrue: ["e.g., more than 10 pixels moved"
			self invalidRect: box.
			self
				invalidRect: (box translateBy: delta)]
		ifFalse: [self
				invalidRect: (box
						merge: (box translateBy: delta))].
	self privateFullMoveBy: delta.
	owner
		ifNotNil: [owner layoutChanged]! !

!Morph methodsFor: 'meta-actions' stamp: 'wiz 1/2/2005 01:06'!
potentialEmbeddingTargets
	"Return the potential targets for embedding the receiver"

	| oneUp topRend |
	(oneUp := (topRend := self topRendererOrSelf) owner) ifNil:[^#()].
	^ (oneUp morphsAt: topRend referencePosition behind: topRend unlocked: true) select:
		[:m | m  isFlexMorph not]! !

!Morph methodsFor: 'meta-actions' stamp: 'wiz 7/17/2004 22:17'!
potentialTargets
	"Return the potential targets for the receiver.
	This is derived from Morph>>potentialEmbeddingTargets."
	owner ifNil:[^#()].
	^owner morphsAt: self referencePosition behind: self unlocked: true not! !

!Morph methodsFor: 'meta-actions' stamp: 'wiz 11/9/2006 23:04'!
potentialTargetsAt: aPoint 
	"Return the potential targets for the receiver.  
	This is derived from Morph>>potentialEmbeddingTargets."
	| realOwner |
	realOwner := self topRendererOrSelf
	owner
		ifNil: [^ #()].
	^ realOwner
		morphsAt: aPoint
		! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 12/31/2004 03:30'!
preferredDuplicationHandleSelector
	"Answer the selector, either #addMakeSiblingHandle: or addDupHandle:, to be offered as the default in a halo open on me"

	Preferences oliveHandleForScriptedObjects ifFalse:
		[^ #addDupHandle:].
	^ self renderedMorph valueOfProperty: #preferredDuplicationHandleSelector ifAbsent:
		[self player class isUniClass
			ifTrue:
				[#addMakeSiblingHandle:]
			ifFalse:
				[#addDupHandle:]]! !

!Morph methodsFor: 'event handling' stamp: 'fbs 1/7/2005 15:43'!
preferredKeyboardBounds

	^ self bounds: self bounds in: World.
! !

!Morph methodsFor: 'event handling' stamp: 'fbs 1/7/2005 15:42'!
preferredKeyboardPosition

	^ (self bounds: self bounds in: World) topLeft.
! !

!Morph methodsFor: 'fileIn/out' stamp: 'tak 2/10/2006 02:24'!
prepareToBeSaved
	"Prepare this morph to be saved to disk. Subclasses should nil out any instance variables that holds state that should not be saved, such as cached Forms. Note that this operation may take more drastic measures than releaseCachedState; for example, it might discard the transcript of an interactive chat session."

	self releaseCachedState.
	self formerOwner: nil.
	self formerPosition: nil.
	self removeProperty: #undoGrabCommand.
	fullBounds := nil! !

!Morph methodsFor: 'accessing' stamp: 'sw 3/3/1999 13:08'!
presenter
	^ owner ifNotNil: [owner presenter] ifNil: [self currentWorld presenter]! !

!Morph methodsFor: 'menus' stamp: 'sw 6/12/2001 21:08'!
presentHelp
	"Present a help message if there is one available"

	self inform: 'Sorry, no help has been
provided here yet.'! !

!Morph methodsFor: 'structure' stamp: 'RAA 6/13/2000 15:01'!
primaryHand

        | outer |
        outer := self outermostWorldMorph ifNil: [^ nil].
        ^ outer activeHand ifNil: [outer firstHand]! !

!Morph methodsFor: 'printing' stamp: ''!
printConstructorOn: aStream indent: level

	^ self printConstructorOn: aStream indent: level nodeDict: IdentityDictionary new
! !

!Morph methodsFor: 'printing' stamp: ''!
printConstructorOn: aStream indent: level nodeDict: nodeDict
	| nodeString |
	(nodeString := nodeDict at: self ifAbsent: [nil])
		ifNotNil: [^ aStream nextPutAll: nodeString].
	submorphs isEmpty ifFalse: [aStream nextPutAll: '('].
	aStream nextPutAll: '('.
	self fullPrintOn: aStream.
	aStream nextPutAll: ')'.
	submorphs isEmpty ifTrue: [^ self].
	submorphs size <= 4
	ifTrue:
		[aStream crtab: level+1;
			nextPutAll: 'addAllMorphs: (Array'.
		1 to: submorphs size do:
			[:i | aStream crtab: level+1; nextPutAll: 'with: '.
			(submorphs at: i) printConstructorOn: aStream indent: level+1 nodeDict: nodeDict].
		aStream nextPutAll: '))']
	ifFalse:
		[aStream crtab: level+1;
			nextPutAll: 'addAllMorphs: ((Array new: ', submorphs size printString, ')'.
		1 to: submorphs size do:
			[:i |
			aStream crtab: level+1; nextPutAll: 'at: ', i printString, ' put: '.
			(submorphs at: i) printConstructorOn: aStream indent: level+1 nodeDict: nodeDict.
			aStream nextPutAll: ';'].
		aStream crtab: level+1; nextPutAll: 'yourself))']! !

!Morph methodsFor: 'printing' stamp: 'dgd 2/22/2003 19:05'!
printOn: aStream 
	| aName |
	super printOn: aStream.
	(aName := self knownName) notNil 
		ifTrue: [aStream nextPutAll: '<' , aName , '>'].
	aStream nextPutAll: '('.
	aStream
		print: self identityHash;
		nextPutAll: ')'! !

!Morph methodsFor: '*morphic-Postscript Canvases' stamp: 'nk 12/29/2003 10:55'!
printPSToFile
	
	self printPSToFileNamed: self externalName! !

!Morph methodsFor: 'printing' stamp: 'RAA 9/18/2000 10:22'!
printSpecs

	| printSpecs |

	printSpecs := self valueOfProperty: #PrintSpecifications.
	printSpecs ifNil: [
		printSpecs := PrintSpecifications defaultSpecs.
		self printSpecs: printSpecs.
	].
	^printSpecs! !

!Morph methodsFor: 'printing' stamp: 'RAA 9/18/2000 10:21'!
printSpecs: aPrintSecification

	self setProperty: #PrintSpecifications toValue: aPrintSecification.
! !

!Morph methodsFor: 'printing' stamp: 'jm 5/28/1998 18:00'!
printStructureOn: aStream indent: tabCount

	tabCount timesRepeat: [aStream tab].
	self printOn: aStream.
	aStream cr.
	self submorphsDo: [:m | m printStructureOn: aStream indent: tabCount + 1].
! !

!Morph methodsFor: 'private' stamp: 'mt 2/10/2014 11:55:34.319'!
privateAddAllMorphs: aCollection atIndex: index
	"Private. Add aCollection of morphs to the receiver"
	| myWorld otherSubmorphs offset |
	(index between: 1 and: submorphs size+1)
		ifFalse: [^ self error: 'index out of range'].
	myWorld := self world.
	otherSubmorphs := submorphs copyWithoutAll: aCollection.
	offset := aCollection count: [:m | (submorphs indexOf: m) between: 1 and: index - 1].
	submorphs := otherSubmorphs copyReplaceFrom: index-offset to: index-offset-1 with: aCollection.
	aCollection do: [:m | | itsOwner itsWorld |
		itsOwner := m owner.
		itsOwner ifNotNil: [
			itsWorld := m world.
			(itsWorld == myWorld) ifFalse: [
				itsWorld ifNotNil: [self privateInvalidateMorph: m].
				m outOfWorld: itsWorld].
			(itsOwner ~~ self) ifTrue: [
				m owner privateRemove: m.
				m owner removedMorph: m ]].
		m privateOwner: self.
		myWorld ifNotNil: [self privateInvalidateMorph: m].
		(myWorld == itsWorld) ifFalse: [m intoWorld: myWorld].
		itsOwner == self ifFalse: [
			self addedMorph: m.
			m noteNewOwner: self ].
	].
	self layoutChanged.
! !

!Morph methodsFor: 'private' stamp: 'nk 10/11/2003 16:08'!
privateAddMorph: aMorph atIndex: index

	| oldIndex myWorld itsWorld oldOwner |
	((index >= 1) and: [index <= (submorphs size + 1)])
		ifFalse: [^ self error: 'index out of range'].
	myWorld := self world.
	oldOwner := aMorph owner.
	(oldOwner == self and: [(oldIndex := submorphs indexOf: aMorph) > 0]) ifTrue:[
		"aMorph's position changes within in the submorph chain"
		oldIndex < index ifTrue:[
			"moving aMorph to back"
			submorphs replaceFrom: oldIndex to: index-2 with: submorphs startingAt: oldIndex+1.
			submorphs at: index-1 put: aMorph.
		] ifFalse:[
			"moving aMorph to front"
			oldIndex-1 to: index by: -1 do:[:i|
				submorphs at: i+1 put: (submorphs at: i)].
			submorphs at: index put: aMorph.
		].
	] ifFalse:[
		"adding a new morph"
		oldOwner ifNotNil:[
			itsWorld := aMorph world.
			itsWorld ifNotNil: [self privateInvalidateMorph: aMorph].
			(itsWorld == myWorld) ifFalse: [aMorph outOfWorld: itsWorld].
			oldOwner privateRemove: aMorph.
			oldOwner removedMorph: aMorph.
		].
		aMorph privateOwner: self.
		submorphs := submorphs copyReplaceFrom: index to: index-1 with: (Array with: aMorph).
		(itsWorld == myWorld) ifFalse: [aMorph intoWorld: myWorld].
	].
	myWorld ifNotNil:[self privateInvalidateMorph: aMorph].
	self layoutChanged.
	oldOwner == self ifFalse: [
		self addedMorph: aMorph.
		aMorph noteNewOwner: self ].
! !

!Morph methodsFor: 'private' stamp: ''!
privateBounds: boundsRect
	"Private!! Use position: and/or extent: instead."

	fullBounds := nil.
	bounds := boundsRect.! !

!Morph methodsFor: 'private' stamp: 'jm 5/29/1998 21:28'!
privateColor: aColor

	color := aColor.
! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 8/10/2003 18:31'!
privateDelete
	"Remove the receiver as a submorph of its owner"
	owner ifNotNil:[owner removeMorph: self].! !

!Morph methodsFor: 'private' stamp: 'RAA 5/23/2000 11:31'!
privateDeleteWithAbsolutelyNoSideEffects
	"Private!! Should only be used by methods that maintain the ower/submorph invariant."
	"used to delete a morph from an inactive world"

	owner ifNil: [^self].
	owner privateRemoveMorphWithAbsolutelyNoSideEffects: self.
	owner := nil.

! !

!Morph methodsFor: 'accessing - extension' stamp: 'dgd 2/16/2003 19:57'!
privateExtension: aMorphExtension
	"private - change the receiver's extension"
	extension := aMorphExtension! !

!Morph methodsFor: 'layout' stamp: 'dgd 2/22/2003 14:33'!
privateFullBounds
	"Private. Compute the actual full bounds of the receiver"

	| box |
	submorphs isEmpty ifTrue: [^self outerBounds].
	box := self outerBounds copy.
	box := box quickMerge: (self clipSubmorphs 
						ifTrue: [self submorphBounds intersect: self clippingBounds]
						ifFalse: [self submorphBounds]).
	^box origin asIntegerPoint corner: box corner asIntegerPoint! !

!Morph methodsFor: 'private' stamp: 'tk 8/30/1998 09:58'!
privateFullBounds: boundsRect
	"Private!! Computed automatically."

	fullBounds := boundsRect.! !

!Morph methodsFor: 'private' stamp: 'ar 12/16/2001 21:47'!
privateFullMoveBy: delta
	"Private!! Relocate me and all of my subMorphs by recursion. Subclasses that implement different coordinate systems may override this method."

	self privateMoveBy: delta.
	1 to: submorphs size do: [:i |
		(submorphs at: i) privateFullMoveBy: delta].
	owner ifNotNil:[
		owner isTextMorph ifTrue:[owner adjustTextAnchor: self]].! !

!Morph methodsFor: 'change reporting' stamp: 'ar 8/12/2003 22:26'!
privateInvalidateMorph: aMorph
	"Private. Invalidate the given morph after adding or removing.
	This method is private because a) we're invalidating the morph 'remotely'
	and b) it forces a fullBounds computation which should not be necessary
	for a general morph c) the morph may or may not actually invalidate
	anything (if it's not in the world nothing will happen) and d) the entire
	mechanism should be rewritten."
	aMorph fullBounds.
	aMorph changed! !

!Morph methodsFor: 'private' stamp: 'ar 1/3/2010 15:52'!
privateMoveBy: delta 
	"Private!! Use 'position:' instead."
	| fill |
	self player ifNotNil: ["Most cases eliminated fast by above test"
		self getPenDown ifTrue: [
			"If this is a costume for a player with its 
			pen down, draw a line."
			self moveWithPenDownBy: delta]].
	bounds := bounds translateBy: delta.
	fullBounds ifNotNil: [fullBounds := fullBounds translateBy: delta].
	fill := self fillStyle.
	fill isOrientedFill ifTrue: [fill origin: fill origin + delta]! !

!Morph methodsFor: 'private' stamp: ''!
privateOwner: aMorph
	"Private!! Should only be used by methods that maintain the ower/submorph invariant."

	owner := aMorph.! !

!Morph methodsFor: 'private' stamp: 'RAA 5/23/2000 11:30'!
privateRemoveMorphWithAbsolutelyNoSideEffects: aMorph
	"Private!! Should only be used by methods that maintain the ower/submorph invariant."
	"used to delete a morph from an inactive world"

	submorphs := submorphs copyWithout: aMorph.

! !

!Morph methodsFor: 'private' stamp: 'di 10/18/2004 21:49'!
privateRemove: aMorph
	"Private!! Should only be used by methods that maintain the ower/submorph invariant."

	submorphs := submorphs copyWithout: aMorph.
	self layoutChanged.! !

!Morph methodsFor: 'private' stamp: ''!
privateSubmorphs
	"Private!! Use 'submorphs' instead."

	^ submorphs! !

!Morph methodsFor: 'private' stamp: ''!
privateSubmorphs: aCollection
	"Private!! Should only be used by methods that maintain the ower/submorph invariant."

	submorphs := aCollection.! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/13/2000 17:14'!
processEvent: anEvent
	"Process the given event using the default event dispatcher."
	^self processEvent: anEvent using: self defaultEventDispatcher! !

!Morph methodsFor: 'events-processing' stamp: 'mt 5/21/2016 09:57'!
processEvent: anEvent using: defaultDispatcher
	"This is the central entry for dispatching events in morphic. Given some event and a default dispatch strategy, find the right receiver and let him handle it.
	WARNING: This is a powerful hook. If you want to use a different event dispatcher from the default, here is the place to hook it in. Depending on how the dispatcher is written (e.g., whether it calls simply #processEvent: or #processEvent:using:) you can change the dispatch strategy for entire trees of morphs. Similarly, you can disable entire trees of morphs from receiving any events whatsoever. Read the documentation in class MorphicEventDispatcher before playing with it.
	Event filters for this capture phase can both influence and overrule the way the receiver can reject the event. If the filter ignores the event, the whole event procecssing will stop. If the filtered event is still not ignored, the receiver can decide to still reject it end regular event dispatch will go on."
	
	| filteredEvent |
	filteredEvent := self sendFilterEventCapture: anEvent for: self.
	filteredEvent wasIgnored ifTrue: [^ filteredEvent].
	
	(self rejectsEvent: filteredEvent) ifTrue:[^#rejected].
	^defaultDispatcher dispatchEvent: filteredEvent with: self! !

!Morph methodsFor: 'events-processing' stamp: 'mt 6/20/2016 11:42'!
processFocusEvent: anEvent

	^self processFocusEvent: anEvent using: self defaultEventDispatcher! !

!Morph methodsFor: 'events-processing' stamp: 'mt 6/20/2016 14:14'!
processFocusEvent: anEvent using: defaultDispatcher
	"Event dispatching shortcut."

	^ defaultDispatcher dispatchFocusEvent: anEvent with: self! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/12/2000 11:16'!
programmedMouseDown: anEvent for: aMorph

	aMorph addMouseActionIndicatorsWidth: 15 color: (Color blue alpha: 0.7).

! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/12/2000 11:16'!
programmedMouseEnter: anEvent for: aMorph

	aMorph addMouseActionIndicatorsWidth: 10 color: (Color blue alpha: 0.3).

! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/12/2000 11:10'!
programmedMouseLeave: anEvent for: aMorph

	self deleteAnyMouseActionIndicators.
! !

!Morph methodsFor: 'debug and other' stamp: 'nice 12/16/2013 15:42'!
programmedMouseUp: anEvent for: aMorph 
	| aCodeString |
	self deleteAnyMouseActionIndicators.
	aCodeString := self valueOfProperty: #mouseUpCodeToRun ifAbsent: [^self].
	(self fullBounds containsPoint: anEvent cursorPoint) ifFalse: [^self].
	
	[(aCodeString isMessageSend) 
		ifTrue: [aCodeString value]
		ifFalse: 
			[Compiler 
				evaluate: aCodeString
				for: self
				notifying: nil]] 
			on: ProgressTargetRequestNotification
			do: [:ex | ex resume: self]	"in case a save/load progress display needs a home"! !

!Morph methodsFor: 'accessing' stamp: 'dgd 3/7/2003 15:24'!
raisedColor
	"Return the color to be used for shading raised borders. The 
	default is my own color, but it might want to be, eg, my 
	owner's color. Whoever's color ends up prevailing, the color 
	itself gets the last chance to determine, so that when, for 
	example, an InfiniteForm serves as the color, callers won't choke 
	on some non-Color object being returned"
	(color isColor
			and: [color isTransparent
					and: [owner notNil]])
		ifTrue: [^ owner raisedColor].
	^ color asColor raisedColor! !

!Morph methodsFor: 'WiW support' stamp: 'RAA 7/16/2000 13:54'!
randomBoundsFor: aMorph

	| trialRect |
	trialRect := (
		self topLeft + 
			((self width * (15 + 75 atRandom/100)) rounded @
			(self height * (15 + 75 atRandom/100)) rounded)
	) extent: aMorph extent.
	^trialRect translateBy: (trialRect amountToTranslateWithin: self bounds)
! !

!Morph methodsFor: 'thumbnail' stamp: 'ar 11/9/2000 20:42'!
readoutForField: fieldSym
	"Provide a readout that will show the value of the slot/pseudoslot of the receiver generated by sending fieldSym to the receiver"

	| aContainer |
	"still need to get this right"
	aContainer := AlignmentMorph newColumn.
	aContainer layoutInset: 0; hResizing: #rigid; vResizing: #shrinkWrap.
	aContainer addMorphBack: (StringMorph new contents: (self perform: fieldSym) asString).
	^ aContainer! !

!Morph methodsFor: 'menus' stamp: 'nk 6/12/2004 22:42'!
reasonableBitmapFillForms
	"Answer an OrderedCollection of forms that could be used to replace my bitmap fill, with my current form first."
	| reasonableForms myGraphic |
	reasonableForms := self class allSketchMorphForms.
	reasonableForms addAll: Imports default images.
	reasonableForms addAll: (BitmapFillStyle allSubInstances collect:[:f| f form]).
	reasonableForms
		remove: (myGraphic := self fillStyle form)
		ifAbsent: [].
	reasonableForms := reasonableForms asOrderedCollection.
	reasonableForms addFirst: myGraphic.
	^reasonableForms! !

!Morph methodsFor: 'menus' stamp: 'nk 6/12/2004 09:55'!
reasonableForms
	"Answer an OrderedCollection of forms that could be used to replace my form, with my current form first."
	| reasonableForms myGraphic |
	reasonableForms := self class allSketchMorphForms.
	reasonableForms addAll: Imports default images.
	reasonableForms
		remove: (myGraphic := self form)
		ifAbsent: [].
	reasonableForms := reasonableForms asOrderedCollection.
	reasonableForms addFirst: myGraphic.
	^reasonableForms! !

!Morph methodsFor: 'geniestubs' stamp: 'nk 3/10/2004 06:38'!
redButtonGestureDictionaryOrName: aSymbolOrDictionary! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 9/11/2004 16:23'!
referencePlayfield
	"Answer the PasteUpMorph to be used for cartesian-coordinate reference"

	| former |
	owner ifNotNil:
		[(self topRendererOrSelf owner isHandMorph and: [(former := self formerOwner) notNil])
			ifTrue:
				[former := former renderedMorph.
				^ former isPlayfieldLike 
					ifTrue: [former]
					ifFalse: [former referencePlayfield]]].

	self allOwnersDo: [:o | o isPlayfieldLike ifTrue: [^ o]].
	^ ActiveWorld! !

!Morph methodsFor: 'drawing' stamp: 'mt 4/19/2016 11:05'!
refreshWorld
	"If this morph is in a world, refresh the world. This can be used to immediately redraw a morph after it changed. Note that you should use this call rather sparsely because it bypasses Morphics damage recording mechanism."
	
	self world ifNotNil: [:w | w displayWorldSafely].! !

!Morph methodsFor: 'accessing' stamp: 'sw 3/6/1999 02:09'!
regularColor
	
	| val |
	^ (val := self valueOfProperty: #regularColor)
		ifNotNil:
			[val ifNil: [self error: 'nil regularColor']]
		ifNil:
			[owner ifNil: [self color] ifNotNil: [owner regularColor]]! !

!Morph methodsFor: 'accessing' stamp: 'sw 7/2/1998 13:51'!
regularColor: aColor
	self setProperty: #regularColor toValue: aColor! !

!Morph methodsFor: 'events-processing' stamp: 'ar 10/5/2000 19:25'!
rejectDropEvent: anEvent
	"This hook allows the receiver to repel a drop operation currently executed. The method is called prior to checking children so the receiver must validate that the event was really designated for it.
	Note that the ordering of the tests below is designed to avoid a (possibly expensive) #fullContainsPoint: test. If the receiver doesn't want to repel the morph anyways we don't need to check after all."
	(self repelsMorph: anEvent contents event: anEvent) ifFalse:[^self]. "not repelled"
	(self fullContainsPoint: anEvent position) ifFalse:[^self]. "not for me"
	"Throw it away"
	anEvent wasHandled: true.
	anEvent contents rejectDropMorphEvent: anEvent.! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'di 12/12/2000 14:35'!
rejectDropMorphEvent: evt
	"The receiver has been rejected, and must be put back somewhere.  There are three cases:
	(1)  It remembers its former owner and position, and goes right back there
	(2)  It remembers its former position only, in which case it was torn off from a parts bin, and the UI is that it floats back to its donor position and then vanishes.
	(3)  Neither former owner nor position is remembered, in which case it is whisked to the Trash"

	self removeProperty: #undoGrabCommand.
	(self formerOwner notNil and: [self formerOwner isPartsBin not]) ifTrue:
		[^ self slideBackToFormerSituation: evt].

	self formerPosition ifNotNil:  "Position but no owner -- can just make it vanish"
		[^ self vanishAfterSlidingTo: self formerPosition event: evt].
		
	self slideToTrash: evt! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/12/2000 23:40'!
rejectsEvent: anEvent
	"Return true to reject the given event. Rejecting an event means neither the receiver nor any of it's submorphs will be given any chance to handle it."
	^self isLocked or:[self visible not]! !

!Morph methodsFor: 'text-anchor' stamp: 'ar 12/16/2001 19:47'!
relativeTextAnchorPosition
	^self valueOfProperty: #relativeTextAnchorPosition! !

!Morph methodsFor: 'text-anchor' stamp: 'ar 12/16/2001 19:22'!
relativeTextAnchorPosition: aPoint
	^self setProperty: #relativeTextAnchorPosition toValue: aPoint! !

!Morph methodsFor: 'events-removing' stamp: 'rw 4/25/2002 07:18'!
releaseActionMap
	"Release the action map"
	
 	self removeProperty: #actionMap! !

!Morph methodsFor: 'caching' stamp: 'md 4/3/2006 12:02'!
releaseCachedState
	"Release any state that can be recomputed on demand, such as the pixel values for a color gradient or the editor state for a TextMorph. This method may be called to save space when a morph becomes inaccessible. Implementations of this method should do 'super releaseCachedState'."
	self borderStyle releaseCachedState. 
! !

!Morph methodsFor: 'accessing' stamp: 'sw 8/29/2000 14:56'!
rememberedColor
	"Answer a rememberedColor, or nil if none"

	^ self valueOfProperty: #rememberedColor ifAbsent: [nil]! !

!Morph methodsFor: 'accessing' stamp: 'sw 8/29/2000 15:47'!
rememberedColor: aColor
	"Place aColor in a property so I can retrieve it later.  A tortuous but expedient flow of data"

	^ self setProperty: #rememberedColor toValue: aColor! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/14/2000 12:14'!
removeAlarm: aSelector
	"Remove the given alarm"
	| scheduler |
	scheduler := self alarmScheduler.
	scheduler ifNotNil:[scheduler removeAlarm: aSelector for: self].! !

!Morph methodsFor: 'events-alarms' stamp: 'ar 9/14/2000 12:15'!
removeAlarm: aSelector at: scheduledTime
	"Remove the given alarm"
	| scheduler |
	scheduler := self alarmScheduler.
	scheduler ifNotNil:[scheduler removeAlarm: aSelector at: scheduledTime for: self].! !

!Morph methodsFor: 'other' stamp: 'sw 10/30/2001 13:12'!
removeAllButFirstSubmorph
	"Remove all of the receiver's submorphs other than the first one."

	self submorphs allButFirst do: [:m | m delete]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'cmm 5/8/2014 10:58'!
removeAllMorphs
	| oldMorphs myWorld |
	myWorld := self world.
	(fullBounds notNil or: [ myWorld notNil ]) ifTrue: [ self invalidRect: self fullBounds ].
	submorphs do:
		[ : m | myWorld ifNotNil: [ m outOfWorld: myWorld ].
		m privateOwner: nil ].
	oldMorphs := submorphs.
	submorphs := Array empty.
	oldMorphs do: [ : m | self removedMorph: m ].
	self layoutChanged! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'nk 10/16/2003 14:02'!
removeAllMorphsIn: aCollection
	"greatly speeds up the removal of *lots* of submorphs"
	| set myWorld |
	set := IdentitySet new: aCollection size * 4 // 3.
	aCollection do: [:each | each owner == self ifTrue: [ set add: each]].
	myWorld := self world.
	(fullBounds notNil or:[myWorld notNil]) ifTrue:[self invalidRect: self fullBounds].
	set do: [:m | myWorld ifNotNil: [ m outOfWorld: myWorld ]. m privateOwner: nil].
	submorphs := submorphs reject: [ :each | set includes: each].
	set do: [ :m | self removedMorph: m ].
	self layoutChanged.
! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 11/12/2000 18:57'!
removeDropShadow
	self hasDropShadow ifFalse:[^self].
	self changed.
	self hasDropShadow: false.
	fullBounds ifNotNil:[fullBounds := self privateFullBounds].
	self changed.! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'ar 8/12/2003 22:01'!
removedMorph: aMorph
	"Notify the receiver that aMorph was just removed from its children"
! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:09'!
removeEventBubbleFilter: anObject
	self eventBubbleFilters: (self removeFilter: anObject from: self eventBubbleFilters).! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:09'!
removeEventCaptureFilter: anObject
	self eventCaptureFilters: (self removeFilter: anObject from: self eventCaptureFilters).! !

!Morph methodsFor: 'events-filtering' stamp: 'mt 5/17/2016 22:21'!
removeFilter: anObject from: aFilterGroup 
	"Remove anObject from the given listener group. Return the new group."

	| filters |
	aFilterGroup ifNil: [^nil].
	filters := aFilterGroup.
	filters := filters copyWithout: anObject.
	filters := filters copyWithout: nil.	"obsolete entries"
	filters isEmpty ifTrue: [filters := nil].
	^filters! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'mu 3/29/2004 17:33'!
removeFlexShell
	self isFlexed
		ifTrue: [self owner removeFlexShell]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'dgd 9/9/2004 22:43'!
removeHalo
	"remove the surrounding halo (if any)"
	self halo isNil
		ifFalse: [self primaryHand removeHalo]! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:09'!
removeKeyboardBubbleFilter: anObject
	self keyboardBubbleFilters: (self removeFilter: anObject from: self keyboardBubbleFilters).! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:09'!
removeKeyboardCaptureFilter: anObject
	self keyboardCaptureFilters: (self removeFilter: anObject from: self keyboardCaptureFilters).! !

!Morph methodsFor: 'event handling' stamp: 'tk 8/10/1998 16:02'!
removeLink: actionCode
	self eventHandler ifNotNil:
		[self eventHandler on: actionCode send: nil to: nil]! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'di 10/18/2004 21:50'!
removeMorph: aMorph
	"Remove the given morph from my submorphs"
	| aWorld |
	aMorph owner == self ifFalse:[^self].
	aWorld := self world.
	aWorld ifNotNil:[
		aMorph outOfWorld: aWorld.
		self privateInvalidateMorph: aMorph.
	].
	self privateRemove: aMorph.
	aMorph privateOwner: nil.
	self removedMorph: aMorph.
! !

!Morph methodsFor: 'events-filtering-bubbling' stamp: 'mt 5/21/2016 09:10'!
removeMouseBubbleFilter: anObject
	self mouseBubbleFilters: (self removeFilter: anObject from: self mouseBubbleFilters).! !

!Morph methodsFor: 'events-filtering-capturing' stamp: 'mt 5/21/2016 09:10'!
removeMouseCaptureFilter: anObject
	self mouseCaptureFilters: (self removeFilter: anObject from: self mouseCaptureFilters).! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 7/7/2000 16:43'!
removeMouseUpAction

	self primaryHand showTemporaryCursor: nil.
	self removeProperty: #mouseUpCodeToRun.
	#(mouseUp mouseEnter mouseLeave mouseDown) do: [ :sym |
		self
			on: sym 
			send: #yourself 
			to: nil.
	]

! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 09:48'!
removeProperty: aSymbol 
	"removes the property named aSymbol if it exists"
	extension ifNil:  [^ self].
	extension removeProperty: aSymbol! !

!Morph methodsFor: 'testing' stamp: 'tk 7/28/2005 04:46'!
renameInternal: aName 
	"Change the internal name (because of a conflict) but leave the external name unchanged.  Change Player class name, but do not change the names that appear in tiles.  When coming in from disk, and have name conflict, References will already have the new name. "

	self knownName = aName ifTrue: [^ aName].
	self topRendererOrSelf setNameTo: aName.
	
	"References dictionary already has key aName"

	"If this player has a viewer flap, it will remain present"

	"Tiles in scripts all stay the same"

	"Compiled methods for scripts have been fixed up because the same association was reused"
	
	^ aName! !

!Morph methodsFor: 'testing' stamp: 'cmm 12/4/2010 20:24'!
renameTo: aName 
	"Set Player name in costume. Update Viewers. Fix all tiles (old style). fix 
	References. New tiles: recompile, and recreate open scripts. If coming in 
	from disk, and have name conflict, References will already have new 
	name."
	| aPresenter putInViewer aPasteUp renderer oldKey assoc classes oldName |
	oldName := self knownName.
	oldName=aName ifTrue: [ ^aName ].
	(renderer := self topRendererOrSelf) setNameTo: aName.
	putInViewer := false.
	((aPresenter := self presenter) isNil or: [renderer player isNil]) 
		ifFalse: 
			[putInViewer := aPresenter currentlyViewing: renderer player.
			putInViewer ifTrue: [renderer player viewerFlapTab hibernate]].
	"empty it temporarily"
	(aPasteUp := self topPasteUp) 
		ifNotNil: [aPasteUp allTileScriptingElements do: [:m | m bringUpToDate]].
	"Fix References dictionary. See restoreReferences to know why oldKey is  
	already aName, but oldName is the old name."
	oldKey := References keyAtIdentityValue: renderer player ifAbsent: [].
	oldKey ifNotNil: 
			[assoc := References associationAt: oldKey.
			oldKey = aName 
				ifFalse: 
					["normal rename"

					assoc key: (renderer player uniqueNameForReferenceFrom: aName).
					References rehash]].
	putInViewer ifTrue: [aPresenter viewMorph: self].
	"recreate my viewer"
	oldKey ifNil: [^aName].
	"Force strings in tiles to be remade with new name. New tiles only."
	Preferences universalTiles ifFalse: [^aName].
	classes := (self systemNavigation allCallsOn: assoc) 
				collect: [:each | each classSymbol].
	classes asSet 
		do: [:clsName | (Smalltalk at: clsName) replaceSilently: oldName to: aName].
	"replace in text body of all methods. Can be wrong!!"
	"Redo the tiles that are showing. This is also done in caller in 
	unhibernate. "
	aPasteUp ifNotNil: 
			[aPasteUp allTileScriptingElements do: 
					[:mm | 
					"just ScriptEditorMorphs"

					nil.
					(mm isScriptEditorMorph) 
						ifTrue: 
							[((mm playerScripted class compiledMethodAt: mm scriptName) 
								hasLiteral: assoc) 
									ifTrue: 
										[mm
											hibernate;
											unhibernate]]]].
	^aName! !

!Morph methodsFor: 'structure' stamp: 'wiz 12/7/2006 15:12'!
renderedMorph
	"This now  gets overridden by rendering morphs."

	^self! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'sw 1/11/1999 20:07'!
repelsMorph: aMorph event: ev
	^ false! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 10/25/1999 23:34'!
replaceSubmorph: oldMorph by: newMorph
	| index itsPosition w |
	oldMorph stopStepping.
	itsPosition := oldMorph referencePositionInWorld.
	index := submorphs indexOf: oldMorph.
	oldMorph privateDelete.
	self privateAddMorph: newMorph atIndex: index.
	newMorph referencePositionInWorld: itsPosition.
	(w := newMorph world) ifNotNil:
		[w startSteppingSubmorphsOf: newMorph]! !

!Morph methodsFor: 'printing' stamp: 'sw 10/27/2000 17:45'!
reportableSize
	"Answer a size worth reporting as the receiver's size in a list view"

	| total |
	total := super reportableSize.
	submorphs do:
		[:m | total := total + m reportableSize].
	^ total! !

!Morph methodsFor: 'thumbnail' stamp: 'sw 12/6/2000 21:28'!
representativeNoTallerThan: maxHeight norWiderThan: maxWidth thumbnailHeight: thumbnailHeight
	"Return a morph representing the receiver but which is no taller than aHeight.  If the receiver is already small enough, just return it, else return a MorphThumbnail companioned to the receiver, enforcing the maxWidth.  If the receiver personally *demands* thumbnailing, do it even if there is no size-related reason to do it."

	self demandsThumbnailing ifFalse:
		[self permitsThumbnailing ifFalse: [^ self].
		(self fullBounds height <= maxHeight and: [self fullBounds width <= maxWidth]) ifTrue: [^ self]].

	^ MorphThumbnail new extent: maxWidth @ (thumbnailHeight min: self fullBounds height); morphRepresented: self! !

!Morph methodsFor: 'fileIn/out' stamp: 'tk 2/17/1999 17:50'!
reserveUrl: urlString
	"Write a dummy object to the server to hold a name and place for this object."

	| dummy ext str |
	dummy := PasteUpMorph new.
	dummy borderWidth: 2.
	dummy setProperty: #initialExtent toValue: (ext := 300@100).
	dummy topLeft: 50@50; extent: ext.	"reset when comes in"
	str := (TextMorph new) topLeft: dummy topLeft + (10@10); 
		extent: dummy width - 15 @ 30.
	dummy addMorph: str.
	str contents: 'This is a place holder only.  Please \find the original page and choose \"send this page to server"' withCRs.
	str extent: dummy width - 15 @ 30.
	dummy saveOnURL: urlString.

	"Claim that url myself"
	self setProperty: #SqueakPage toValue: dummy sqkPage.
	(dummy sqkPage) contentsMorph: self; dirty: true.
	^ self url! !

!Morph methodsFor: 'accessing - extension' stamp: 'md 2/27/2006 08:47'!
resetExtension
	"reset the extension slot if it is not needed"
	(extension notNil and: [extension isDefault]) ifTrue: [extension := nil] ! !

!Morph methodsFor: 'menus' stamp: 'ar 9/22/2000 20:36'!
resetForwardDirection
	self forwardDirection: 0.! !

!Morph methodsFor: '*Morphic-Sound-piano rolls' stamp: 'RAA 12/11/2000 23:22'!
resetFrom: scorePlayer

	"subclasses should revert to their initial state"! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/28/2000 12:02'!
resetHighlightForDrop
	self highlightForDrop: false! !

!Morph methodsFor: 'parts bin' stamp: 'sw 4/22/1998 14:45'!
residesInPartsBin
	"Answer true if the receiver is, or has some ancestor owner who is, a parts bin"
	^ owner ifNotNil: [owner residesInPartsBin] ifNil: [false]! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/15/2001 16:33'!
resistsRemoval
	"Answer whether the receiver is marked as resisting removal"

	^ self hasProperty: #resistsRemoval! !

!Morph methodsFor: 'menus' stamp: 'dgd 8/30/2003 20:18'!
resistsRemovalString
	"Answer the string to be shown in a menu to represent the 
	'resistsRemoval' status"
	^ (self resistsRemoval
		ifTrue: ['<on>']
		ifFalse: ['<off>']), 'resist being deleted' translated! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/15/2001 16:33'!
resistsRemoval: aBoolean
	"Set the receiver's resistsRemoval property as indicated"

	aBoolean
		ifTrue:
			[self setProperty: #resistsRemoval toValue: true]
		ifFalse:
			[self removeProperty: #resistsRemoval]! !

!Morph methodsFor: 'meta-actions' stamp: 'sw 11/27/2001 14:59'!
resizeFromMenu
	"Commence an interaction that will resize the receiver"

	self resizeMorph: ActiveEvent! !

!Morph methodsFor: 'meta-actions' stamp: 'st 9/14/2004 12:30'!
resizeMorph: evt
	| handle |
	handle := HandleMorph new forEachPointDo: [:newPoint | 
		self extent: (self griddedPoint: newPoint) - self bounds topLeft].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'initialization' stamp: 'ar 3/3/2001 15:28'!
resourceJustLoaded
	"In case resource relates to me"
	self releaseCachedState.! !

!Morph methodsFor: 'event handling' stamp: 'sw 11/16/1998 08:06'!
restoreSuspendedEventHandler
	| savedHandler |
	(savedHandler := self valueOfProperty: #suspendedEventHandler) ifNotNil:
		[self eventHandler: savedHandler].
	submorphs do: [:m | m restoreSuspendedEventHandler]
! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 5/24/2000 18:20'!
resumeAfterDrawError

	self changed.
	self removeProperty:#errorOnDraw.
	self changed.! !

!Morph methodsFor: 'debug and other' stamp: 'RAA 5/24/2000 18:20'!
resumeAfterStepError
	"Resume stepping after an error has occured."

	self startStepping. "Will #step"
	self removeProperty:#errorOnStep. "Will remove prop only if #step was okay"
! !

!Morph methodsFor: '*Morphic-Sound-piano rolls' stamp: 'RAA 12/11/2000 23:21'!
resumeFrom: scorePlayer

	"subclasses should continue from their current position"
	"a hack to allow for abitrary morphs to be dropped into piano roll"! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:01'!
reverseTableCells
	"Layout specific. This property describes if the cells should be treated in reverse order of submorphs."
	| props |
	props := self layoutProperties.
	^props ifNil:[false] ifNotNil:[props reverseTableCells].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
reverseTableCells: aBool
	"Layout specific. This property describes if the cells should be treated in reverse order of submorphs."
	self assureTableProperties reverseTableCells: aBool.
	self layoutChanged.! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:16'!
right
	" Return the x-coordinate of my right side "
	^ bounds right! !

!Morph methodsFor: 'geometry' stamp: 'di 3/6/2002 13:06'!
rightCenter

	^ bounds rightCenter! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:15'!
right: aNumber
	" Move me so that my right side is at the x-coordinate aNumber. My extent (width & height) are unchanged "

	self position: ((aNumber - bounds width) @ bounds top)! !

!Morph methodsFor: 'structure' stamp: 'dgd 2/22/2003 14:34'!
root
	"Return the root of the composite morph containing the receiver. The owner of the root is either nil, a WorldMorph, or a HandMorph. If the receiver's owner is nil, the root is the receiver itself. This method always returns a morph."

	(owner isNil or: [owner isWorldOrHandMorph]) ifTrue: [^self].
	^owner root! !

!Morph methodsFor: 'structure' stamp: 'di 8/4/1999 15:41'!
rootAt: location
	"Just return myself, unless I am a WorldWindow.
	If so, then return the appropriate root in that world"

	^ self! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'RAA 6/11/2000 20:41'!
rootMorphsAtGlobal: aPoint
	"Return the list of root morphs containing the given point, excluding the receiver.
	ar 11/8/1999: Moved into morph for an incredibly ugly hack in 3D worlds"

	^ self rootMorphsAt: (self pointFromWorld: aPoint)! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ar 10/8/2000 15:44'!
rootMorphsAt: aPoint
	"Return the list of root morphs containing the given point, excluding the receiver.
	ar 11/8/1999: Moved into morph for an incredibly ugly hack in 3D worlds"
self flag: #arNote. "check this at some point"
	^ self submorphs select:
		[:m | (m fullContainsPoint: aPoint) and: [m isLocked not]]! !

!Morph methodsFor: 'rotate scale and flex' stamp: 'jm 4/25/1998 05:19'!
rotationDegrees
	"Default implementation."

	^ 0.0
! !

!Morph methodsFor: 'e-toy support' stamp: 'ar 9/23/2000 22:40'!
rotationStyle
	"Return the 'rotation style' of the receiver"
	^#normal! !

!Morph methodsFor: 'e-toy support' stamp: 'ar 9/23/2000 22:40'!
rotationStyle: aSymbol
	"Set the 'rotation style' of the receiver; this is ignored for non-sketches"! !

!Morph methodsFor: 'rounding' stamp: 'mk 8/14/2005 13:31'!
roundedCorners
	"Return a list of those corners to round.

		1-4
		|  |
		2-3

	Returned array contains `codes' of those corners, which should be rounded.

	1 denotes top-left corner
	2 denotes bottom-left corner
	3 denotes bottom-right corner
	4 denotes top-right corner.

	Thus, if this method returned #(2 3) that would mean that bottom (left and right)
	corners would be rounded whereas top (left and right) corners wouldn't be rounded.

	This method returns #(1 2 3 4) and that means that all the corners should be rounded."

	^ #(1 2 3 4)! !

!Morph methodsFor: 'rounding' stamp: 'dgd 9/6/2003 18:27'!
roundedCornersString
	"Answer the string to put in a menu that will invite the user to 
	switch to the opposite corner-rounding mode"
	^ (self wantsRoundedCorners
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'round corners' translated! !

!Morph methodsFor: 'miscellaneous' stamp: 'sw 10/4/2006 00:52'!
roundUpStrays
	"Bring submorphs of playfieldlike structures in the receiver's interior back within view."

	self submorphsDo:
		[:m | m isPlayfieldLike ifTrue: [m roundUpStrays]]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:01'!
rubberBandCells
	"Layout specific. This property describes if a parent that is #shrinkWrapped around its children should ignore any #spaceFill children. E.g., when #rubberBandCells is true, the compound layout will always stay at the smallest available size, even though some child may be able to grow."
	| props |
	props := self layoutProperties.
	^props ifNil:[false] ifNotNil:[props rubberBandCells].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:04'!
rubberBandCells: aBool
	"Layout specific. This property describes if a parent that is #shrinkWrapped around its children should ignore any #spaceFill children. E.g., when #rubberBandCells is true, the compound layout will always stay at the smallest available size, even though some child may be able to grow."
	self assureTableProperties rubberBandCells: aBool.
	self layoutChanged.! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 8/6/2009 18:55'!
saveAsPrototype
	(UIManager default confirm: 'Make this morph the prototype for ', self class printString, '?')
		ifFalse: [^ self].
	self class prototype: self.
! !

!Morph methodsFor: 'fileIn/out' stamp: 'di 11/18/1999 08:52'!
saveAsResource

	| pathName |
	(self hasProperty: #resourceFilePath) ifFalse: [^ self].
	pathName := self valueOfProperty: #resourceFilePath.
	(pathName asLowercase endsWith: '.morph') ifFalse:
		[^ self error: 'Can only update morphic resources'].
	(FileStream newFileNamed: pathName) fileOutClass: nil andObject: self.! !

!Morph methodsFor: 'fileIn/out' stamp: 'ar 9/27/2005 21:02'!
saveDocPane

	Smalltalk at: #DocLibrary ifPresent:[:dl| dl external saveDocCheck: self]! !

!Morph methodsFor: 'fileIn/out' stamp: 'cmm 2/24/2011 20:46'!
saveOnFile
	"Ask the user for a filename and save myself on a SmartReferenceStream file.  Writes out the version and class structure.  The file is fileIn-able.  UniClasses will be filed out."
	| aFileName ok |
	aFileName := ('my {1}' translated format: {self class name}) asFileName.	"do better?"
	aFileName := UIManager default request: 'File name? (".morph" will be added to end)' translated 
			initialAnswer: aFileName.
	aFileName isEmpty ifTrue: [^ Beeper beep].
	self allMorphsDo: [:m | m prepareToBeSaved].

	ok := aFileName endsWith: '.morph'.	"don't double them"
	ok := ok | (aFileName endsWith: '.sp').
	ok ifFalse: [aFileName := aFileName,'.morph'].
	self saveOnFileNamed: aFileName! !

!Morph methodsFor: 'fileIn/out' stamp: 'tk 7/16/1999 13:03'!
saveOnURL
	"Ask the user for a url and save myself on a SmartReferenceStream file.  Writes out the version and class structure.  The file is fileIn-able.  UniClasses will be filed out."

	| um pg |
	(pg := self saveOnURLbasic) == #cancel ifTrue: [^ self].
	um := URLMorph newForURL: pg url.
	um setURL: pg url page: pg.
	pg isContentsInMemory ifTrue: [pg computeThumbnail].
	um isBookmark: true.
	um removeAllMorphs.
	um color: Color transparent.
	self primaryHand attachMorph: um.! !

!Morph methodsFor: 'fileIn/out' stamp: 'cmm 3/9/2010 15:13'!
saveOnURLbasic
	"Ask the user for a url and save myself on a SmartReferenceStream file.  Writes out the version and class structure.  The file is fileIn-able.  UniClasses will be filed out."

	| url pg stamp pol |
	(pg := self valueOfProperty: #SqueakPage) ifNil: [pg := SqueakPage new]
		ifNotNil: 
			[pg contentsMorph ~~ self 
				ifTrue: 
					[self inform: 'morph''s SqueakPage property is out of date'.
					pg := SqueakPage new]].
	(url := pg url) ifNil: 
			[url := ServerDirectory defaultStemUrl , '1.sp'.	"A new legal place"
			url := UIManager default 
						request: 'url of a place to store this object.
Must begin with file:// or ftp://'
						initialAnswer: url.
			url isEmpty ifTrue: [^#cancel]].
	stamp := Utilities authorInitialsPerSe.
	stamp isEmptyOrNil ifTrue: [ stamp := '*' ].
	pg saveMorph: self author: stamp.
	SqueakPageCache atURL: url put: pg.	"setProperty: #SqueakPage"
	(pol := pg policy) ifNil: [pol := #neverWrite].
	pg
		policy: #now;
		dirty: true.
	pg write.	"force the write"
	pg policy: pol.
	^pg! !

!Morph methodsFor: 'fileIn/out' stamp: 'cmm 3/9/2010 15:12'!
saveOnURL: suggestedUrlString 
	"Save myself on a SmartReferenceStream file.  If I don't already have a url, use the suggested one.  Writes out the version and class structure.  The file is fileIn-able.  UniClasses will be filed out."
	| url pg stamp pol |
	(pg := self valueOfProperty: #SqueakPage)
		ifNil: [ pg := SqueakPage new ]
		ifNotNil:
			[ pg contentsMorph ~~ self ifTrue:
				[ self inform: 'morph''s SqueakPage property is out of date'.
				pg := SqueakPage new ] ].
	(url := pg url) ifNil: [ url := pg urlNoOverwrite: suggestedUrlString ].
	stamp := Utilities authorInitialsPerSe.
	stamp isEmptyOrNil ifTrue: [ stamp := '*' ].
	pg
		saveMorph: self
		author: stamp.
	SqueakPageCache
		atURL: url
		put: pg.
	"setProperty: #SqueakPage"
	(pol := pg policy) ifNil: [ pol := #neverWrite ].
	pg
		 policy: #now ;
		 dirty: true.
	pg write.
	"force the write"
	pg policy: pol.
	^pg! !

!Morph methodsFor: 'accessing' stamp: 'nk 9/4/2004 10:49'!
scaleFactor
	^self valueOfProperty: #scaleFactor ifAbsent: [ 1.0 ]
! !

!Morph methodsFor: 'geometry' stamp: 'bf 1/5/2000 19:08'!
screenLocation
	"For compatibility only"

	^ self fullBounds origin! !

!Morph methodsFor: 'geometry' stamp: 'sma 2/5/2000 13:58'!
screenRectangle
	"For compatibility only"

	^ self fullBounds! !

!Morph methodsFor: 'selected object' stamp: 'dgd 8/28/2004 16:30'!
selectedObject
	"answer the selected object for the hand or nil is none"
	^ self primaryHand selectedObject! !

!Morph methodsFor: 'events-processing' stamp: 'mt 5/21/2016 09:46'!
sendFilterEventBubbleAgain: anEvent for: anObject
	"Apply keyboard-specific and mouse-specific bubbling filters. If a filter changes the event type, filter again. WARNING: This is a powerful mechanism. Filters can create endless loops, which are difficult to debug."

	| filteredEvent |
	filteredEvent := anEvent.
	
	filteredEvent isKeyboard ifTrue: [
		filteredEvent := self sendFilterEvent: filteredEvent for: anObject to: self keyboardBubbleFilters.
		filteredEvent isKeyboard not ifTrue: [^ self sendFilterEventBubbleAgain: filteredEvent for: anObject]].
		
	filteredEvent isMouse ifTrue: [
		filteredEvent := self sendFilterEvent: filteredEvent for: anObject to: self mouseBubbleFilters.
		filteredEvent isMouse not ifTrue: [^ self sendFilterEventBubbleAgain: filteredEvent for: anObject]].

	^ filteredEvent! !

!Morph methodsFor: 'events-processing' stamp: 'mt 5/21/2016 09:46'!
sendFilterEventBubble: anEvent for: anObject
	"Apply event bubbling filters."

	^ self
		sendFilterEventBubbleAgain: (self sendFilterEvent: anEvent for: anObject to: self eventBubbleFilters)
		for: anObject! !

!Morph methodsFor: 'events-processing' stamp: 'mt 5/21/2016 09:46'!
sendFilterEventCaptureAgain: anEvent for: anObject
	"Apply keyboard-specific and mouse-specific capturing filters. If a filter changes the event type, filter again. WARNING: This is a powerful mechanism. Filters can create endless loops, which are difficult to debug."

	| filteredEvent |
	filteredEvent := anEvent.
	
	filteredEvent isKeyboard ifTrue: [
		filteredEvent := self sendFilterEvent: filteredEvent for: anObject to: self keyboardCaptureFilters.
		filteredEvent isKeyboard not ifTrue: [^ self sendFilterEventCaptureAgain: filteredEvent for: anObject]].
		
	filteredEvent isMouse ifTrue: [
		filteredEvent := self sendFilterEvent: filteredEvent for: anObject to: self mouseCaptureFilters.
		filteredEvent isMouse not ifTrue: [^ self sendFilterEventCaptureAgain: filteredEvent for: anObject]].

	^ filteredEvent! !

!Morph methodsFor: 'events-processing' stamp: 'mt 5/21/2016 09:46'!
sendFilterEventCapture: anEvent for: anObject
	"Apply event capturing filters."

	^ self
		sendFilterEventCaptureAgain: (self sendFilterEvent: anEvent for: anObject to: self eventCaptureFilters)
		for: anObject! !

!Morph methodsFor: 'events-processing' stamp: 'mt 5/21/2016 09:46'!
sendFilterEvent: anEvent for: anObject to: filterGroup
	"An event gets filtered if some event filter wants to filter it."
	
	| filteredEvent |
	filterGroup ifNil: [^anEvent].

	filteredEvent := anEvent.
	filterGroup do: [:filterOrNil | filterOrNil ifNotNil: [:filter |
		filteredEvent := filter filterEvent: filteredEvent for: anObject]].
	
	^ filteredEvent! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:24'!
separateDragAndDrop
	"Conversion only. Separate the old #dragNDropEnabled into #dragEnabled and #dropEnabled and remove the old property."
	| dnd |
	(self hasProperty: #dragNDropEnabled) ifFalse:[^self].
	dnd := (self valueOfProperty: #dragNDropEnabled) == true.
	self dragEnabled: dnd.
	self dropEnabled: dnd.
	self removeProperty: #dragNDropEnabled.
! !

!Morph methodsFor: 'menus' stamp: 'yo 2/17/2005 16:58'!
setArrowheads
	"Let the user edit the size of arrowheads for this object"

	| aParameter result  |
	aParameter := self renderedMorph valueOfProperty:  #arrowSpec ifAbsent:
		[Preferences parameterAt: #arrowSpec ifAbsent: [5 @ 4]].
	result := Morph obtainArrowheadFor: 'Head size for arrowheads: ' translated defaultValue: aParameter asString.
	result ifNotNil:
			[self renderedMorph  setProperty: #arrowSpec toValue: result]
		ifNil:
			[Beeper beep]! !

!Morph methodsFor: 'e-toy support' stamp: 'RAA 3/9/2001 14:37'!
setAsActionInButtonProperties: buttonProperties

	^false	"means I don't know how to be set as a button action"! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 3/30/2015 09:59'!
setBalloonText: stringOrText
	"Set receiver's balloon help text. Pass nil to remove the help."

	self flag: #deprecated. "mt: Use #balloonText:."
	self balloonText: stringOrText.! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 3/30/2015 09:13'!
setBalloonText: stringOrText maxLineLength: aLength 
	"Set receiver's balloon help text. Pass nil to remove the help."
	
	self flag: #deprecated. "mt: Use #balloonText:."
	(extension isNil and: [stringOrText isNil]) ifTrue: [^ self].
	self assureExtension balloonText: 
		(stringOrText ifNotNil: [stringOrText asString withNoLineLongerThan: aLength])! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/26/2001 16:16'!
setBorderStyle: aSymbol
	"Set the border style of my costume"

	| aStyle |
	aStyle := self borderStyleForSymbol: aSymbol.
	aStyle ifNil: [^ self].
	(self canDrawBorder: aStyle)
		ifTrue:
			[self borderStyle: aStyle]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 10/29/1999 17:38'!
setCenteredBalloonText: aString
	self setBalloonText: aString.
	self setProperty: #helpAtCenter toValue: true! !

!Morph methodsFor: 'geometry' stamp: 'mt 4/12/2015 11:47'!
setConstrainedPosition: aPoint hangOut: partiallyOutside
	"Change the position of this morph and and all of its submorphs to aPoint, but don't let me go outside my owner's bounds.  Let me go within two pixels of completely outside if partiallyOutside is true."

	| trialRect delta boundingMorph bRect |
	self flag: #smelly. "mt: The consideration of partiallyOutside is strange... The explicit #laoyutChanged, too."
	owner ifNil:[^self].
	trialRect := aPoint extent: self bounds extent.
	boundingMorph := self topRendererOrSelf owner.
	delta := boundingMorph
			ifNil:    [0@0]
			ifNotNil: [
				bRect := partiallyOutside 
					ifTrue: [boundingMorph bounds insetBy: 
								self extent negated + boundingMorph borderWidth + (2@2)]
					ifFalse: [boundingMorph bounds].
				trialRect amountToTranslateWithin: bRect].
	self position: aPoint + delta.
	self layoutChanged  "So that, eg, surrounding text will readjust"
! !

!Morph methodsFor: 'miscellaneous' stamp: 'sw 7/20/2001 00:15'!
setExtentFromHalo: anExtent
	"The user has dragged the grow box such that the receiver's extent would be anExtent.  Do what's needed"

	self extent: anExtent! !

!Morph methodsFor: 'miscellaneous' stamp: 'sw 2/2/2006 00:43'!
setFlexExtentFromHalo: anExtent
	"The user has dragged the grow box such that the receiver's extent would be anExtent.  Do what's needed.  Set the extent of the top renderer as indicated."

	self addFlexShellIfNecessary.
	self topRendererOrSelf extent: anExtent! !

!Morph methodsFor: 'naming' stamp: 'dgd 2/16/2003 21:57'!
setNamePropertyTo: aName 
	"change the receiver's externalName"
	self assureExtension externalName: aName! !

!Morph methodsFor: 'naming' stamp: 'yo 12/3/2004 17:02'!
setNameTo: aName 
	| nameToUse nameString |
	nameToUse := aName ifNotNil: 
					[(nameString := aName asString) notEmpty ifTrue: [nameString] ifFalse: ['*']].
	self setNamePropertyTo: nameToUse	"no Texts here!!"! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 9/15/2000 06:26'!
setNumericValue: aValue
	"Set the receiver's contents to reflect the given numeric value.  Only certain kinds of morphs know what to do with this, the rest, for now, stash the number in a property, where it may not be visible but at least it won't be lost, and can be retrieved by the companion getter.  This code is never reached under normal circumstances, because the #numericValue slot is not shown in Viewers for most kinds of morphs, and those kinds of morphs that do show it also reimplement this method.  However, this code *could* be reached via a user script which sends #setNumericValue: but whose receiver has been changed, via tile-scripting drag and drop for example, to one that doesn't directly handle numbers"

	ScriptingSystem informScriptingUser: 'an unusual setNumericValue: call was made'.
	self renderedMorph setProperty: #numericValue toValue: aValue
! !

!Morph methodsFor: 'accessing - properties' stamp: 'tk 10/9/2002 08:30'!
setProperties: aList
	"Set many properties at once from a list of prop, value, prop, value"

	1 to: aList size by: 2 do: [:ii |
		self setProperty: (aList at: ii) toValue: (aList at: ii+1)].! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 09:48'!
setProperty: aSymbol toValue: anObject 
	"change the receiver's property named aSymbol to anObject"
	anObject ifNil: [^ self removeProperty: aSymbol].
	self assureExtension setProperty: aSymbol toValue: anObject! !

!Morph methodsFor: 'menus' stamp: 'nice 1/19/2010 01:56'!
setRotationCenter
	| p |
	self world displayWorld.
	p := Cursor crossHair showWhile:
		[Sensor waitButton].
	Sensor waitNoButton.
	self setRotationCenterFrom: (self transformFromWorld globalPointToLocal: p).

! !

!Morph methodsFor: 'menus' stamp: 'ar 9/22/2000 20:14'!
setRotationCenterFrom: aPoint
	self rotationCenter: (aPoint - self bounds origin) / self bounds extent asFloatPoint.! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 10/26/2000 18:58'!
setShadowOffset: evt
	| handle |
	handle := HandleMorph new forEachPointDo:
		[:newPoint | self shadowPoint: newPoint].
	evt hand attachMorph: handle.
	handle startStepping.
! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 7/21/1998 21:18'!
setStandardTexture
	| parms |
	parms := self textureParameters.
	self makeGraphPaperGrid: parms first
		background: parms second
		line: parms third! !

!Morph methodsFor: 'menus' stamp: 'di 12/21/2000 17:18'!
setToAdhereToEdge: anEdge
	anEdge ifNil: [^ self].
	anEdge == #none ifTrue: [^ self removeProperty: #edgeToAdhereTo].
	self setProperty: #edgeToAdhereTo toValue: anEdge.
! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 10/26/2000 18:59'!
shadowColor
	^self valueOfProperty: #shadowColor ifAbsent:[Color black]! !

!Morph methodsFor: 'drop shadows' stamp: 'mt 4/6/2015 12:24'!
shadowColor: aColor
	self shadowColor = aColor ifTrue: [^ self].
	self setProperty: #shadowColor toValue: aColor.
	self removeProperty: #dropShadow.
	self changed.! !

!Morph methodsFor: 'drawing' stamp: 'JW 7/12/2005 20:12'!
shadowForm
	"Return a form representing the 'shadow' of the receiver - e.g., all pixels that are occupied by the receiver are one, all others are zero."
	| canvas |
	canvas := (Display defaultCanvasClass extent: self fullBounds extent depth: 1)
				asShadowDrawingCanvas: Color black. "Color black represents one for 1bpp"
	canvas translateBy: bounds topLeft negated
		during:[:tempCanvas| tempCanvas fullDrawMorph: self].
	^ canvas form offset: bounds topLeft
! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 10/26/2000 18:57'!
shadowOffset
	"Return the current shadow offset"
	^self valueOfProperty: #shadowOffset ifAbsent:[0@0]! !

!Morph methodsFor: 'drop shadows' stamp: 'mt 7/30/2016 08:08'!
shadowOffset: aPoint
	"Set the current shadow offset"

	self shadowOffset = aPoint ifTrue: [^ self].
	self changed.

	(aPoint isNil or: [ aPoint isZero ])
		ifTrue:[self removeProperty: #shadowOffset]
		ifFalse:[self setProperty: #shadowOffset toValue: aPoint].

	self removeProperty: #dropShadow.
	self layoutChanged.
	self changed.! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 11/12/2000 18:58'!
shadowPoint: newPoint
	self changed.
	self shadowOffset: newPoint - self center // 5.
	fullBounds ifNotNil:[fullBounds := self privateFullBounds].
	self changed.! !

!Morph methodsFor: 'geometry' stamp: 'sw 2/16/1999 22:05'!
shiftSubmorphsOtherThan: listNotToShift by: delta
	| rejectList |
	rejectList := listNotToShift ifNil: [OrderedCollection new].
	(submorphs copyWithoutAll: rejectList) do:
		[:m | m position: (m position + delta)]! !

!Morph methodsFor: 'testing' stamp: 'ar 12/3/2001 12:33'!
shouldDropOnMouseUp
	| former |
	former := self formerPosition ifNil:[^false].
	^(former dist: self position) > 10! !

!Morph methodsFor: 'WiW support' stamp: 'ar 3/18/2001 00:14'!
shouldGetStepsFrom: aWorld
	^self world == aWorld! !

!Morph methodsFor: 'player' stamp: 'mir 6/13/2001 14:45'!
shouldRememberCostumes
	^true! !

!Morph methodsFor: 'drawing' stamp: 'sw 10/22/1998 20:29'!
show
	"Make sure this morph is on-stage."
	self visible ifFalse: [self visible: true.  self changed]! !

!Morph methodsFor: 'meta-actions' stamp: 'fbs 5/21/2013 22:33'!
showActions
	"Put up a message list browser of all the code that this morph  
	would run for mouseUp, mouseDown, mouseMove, mouseEnter,  
	mouseLeave, and  
	mouseLinger. tk 9/13/97"
	| list cls selector adder |
	list := SortedCollection new.
	adder := [:mrClass :mrSel | list
				add: (MethodReference class: mrClass selector: mrSel)].
	"the eventHandler"
	self eventHandler
		ifNotNil: [list := self eventHandler methodRefList.
			(self eventHandler handlesMouseDown: nil)
				ifFalse: [adder value: HandMorph value: #grabMorph:]].
	"If not those, then non-default raw events"
	#(#keyStroke: #mouseDown: #mouseEnter: #mouseLeave: #mouseMove: #mouseUp: #doButtonAction )
		do: [:sel | 
			cls := self class whichClassIncludesSelector: sel.
			cls
				ifNotNil: ["want more than default behavior"
					cls == Morph
						ifFalse: [adder value: cls value: sel]]].
	"The mechanism on a Button"
	(self respondsTo: #actionSelector)
		ifTrue: ["A button"
			selector := self actionSelector.
			cls := self target class whichClassIncludesSelector: selector.
			cls
				ifNotNil: ["want more than default behavior"
					cls == Morph
						ifFalse: [adder value: cls value: selector]]].
	MessageSet openMessageList: list name: 'Actions
of ' , self printString autoSelect: nil! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 3/31/2016 10:02'!
showBalloon

	self showBalloon: self balloonText.! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 3/31/2016 11:49'!
showBalloon: msgString
	"Pop up a balloon containing the given string."

	self isInWorld ifFalse: [^ self].
	(msgString isNil or: [msgString isEmpty]) ifTrue: [^ self].
	
	(self balloonMorphClass
		string: msgString
		for: self balloonHelpAligner)
			popUp.! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 4/1/2016 11:16'!
showBalloon: msgString at: point
	"Pop up a balloon containing the given string."

	self isInWorld ifFalse: [^ self].
	(msgString isNil or: [msgString isEmpty]) ifTrue: [^ self].
	
	(self balloonMorphClass
		string: msgString
		for: self balloonHelpAligner)
			popUpAt: point.! !

!Morph methodsFor: 'halos and balloon help' stamp: 'mt 3/31/2016 09:48'!
showBalloon: msgString hand: aHand
	"Pop up a balloon containing the given string. Note that any existing BalloonMorphs will be removed from the world."

	| balloon |
	self isInWorld ifFalse: [^ self].
	
	balloon := (self balloonMorphClass
		string: msgString
		for: self balloonHelpAligner).
		
	balloon
		popUpFor: self
		hand: (aHand ifNil: [self world activeHand]).! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 10/5/2000 19:21'!
showHiders
	self allMorphsDo:[:m | m show]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:35'!
shuffleSubmorphs
	"Randomly shuffle the order of my submorphs.  Don't call this method lightly!!"

	| bg |
	self invalidRect: self fullBounds.
	(submorphs notEmpty and: [submorphs last mustBeBackmost]) 
		ifTrue: 
			[bg := submorphs last.
			bg privateDelete].
	submorphs := submorphs shuffled.
	bg ifNotNil: [self addMorphBack: bg].
	self layoutChanged! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 9/7/2009 14:28'!
sightTargets: event 
	"Return the potential targets for the receiver.  
	This is derived from Morph>>potentialEmbeddingTargets."
	| bullseye candidates choice |
	owner ifNil: [^ #()].
	bullseye := Point fromUserWithCursor: Cursor target.
	candidates := self potentialTargetsAt: bullseye.
	choice := UIManager default 
		chooseFrom: (candidates collect:[:m| m knownName ifNil:[m class name]])
		values: candidates.
	choice ifNotNil:[self target: choice].! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 9/7/2009 14:28'!
sightWorldTargets: event 
	"Return the potential targets for the receiver.  
	This is derived from Morph>>potentialEmbeddingTargets."
	| bullseye myWorld candidates choice |
	myWorld := self world ifNil: [^ #()].
	bullseye := Point fromUserWithCursor: Cursor target.
	candidates := myWorld morphsAt: bullseye.
	choice := UIManager default 
		chooseFrom: (candidates collect:[:m| m knownName ifNil:[m class name]])
		values: candidates.
	choice ifNotNil:[self target: choice].! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'yo 6/19/2007 17:00'!
slideBackToFormerSituation: evt 
	"A drop of the receiver having been rejected, slide it back to where it came from, if possible."

	| slideForm formerOwner formerPosition aWorld startPoint endPoint trans |
	formerOwner := self formerOwner.
	formerPosition := self formerPosition.
	(aWorld := evt hand world) ifNil: [^ self delete]. "Likely a moribund hand from an EventRecorder playback."

	trans := formerOwner transformFromWorld.
	slideForm := trans isPureTranslation 
				ifTrue: [self imageForm offset: 0 @ 0]
				ifFalse: 
					[((TransformationMorph new asFlexOf: self) transform: trans) imageForm 
						offset: 0 @ 0]. 
	startPoint := evt hand fullBounds origin.
	endPoint := trans localPointToGlobal: formerPosition.
	owner removeMorph: self.
	aWorld displayWorld.
	slideForm 
		slideFrom: startPoint
		to: endPoint
		nSteps: 12
		delay: 15.
	"The OLPC Virtual Screen wouldn't notice the last update here."
	Display forceToScreen: (endPoint extent: slideForm extent).
	formerOwner addMorph: self.
	self position: formerPosition.
	self justDroppedInto: formerOwner event: evt! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'fbs 12/31/2013 12:08'!
slideToTrash: evt
	"Perhaps slide the receiver across the screen to a trash can and make it disappear into it.  In any case, remove the receiver from the screen."

	| aForm trash startPoint endPoint morphToSlide |
	((self renderedMorph == ScrapBook default scrapBook) or: [self renderedMorph isKindOf: TrashCanMorph]) ifTrue:
		[self dismissMorph.  ^ self].
	TrashCanMorph slideDismissalsToTrash ifTrue:
		[morphToSlide := self representativeNoTallerThan: 200 norWiderThan: 200 thumbnailHeight: 100.
		aForm := morphToSlide imageForm offset: (0@0).
		trash := ActiveWorld
			findDeepSubmorphThat:
				[:aMorph | (aMorph isKindOf: TrashCanMorph) and:
					[aMorph topRendererOrSelf owner == ActiveWorld]]
			ifAbsent:
				[trash := TrashCanMorph new.
				trash position: ActiveWorld bottomLeft - (0 @ (trash extent y + 26)).
				trash openInWorld.
				trash].
		endPoint := trash fullBoundsInWorld center.
		startPoint := self topRendererOrSelf fullBoundsInWorld center - (aForm extent // 2)].
	self dismissMorph.
	ActiveWorld displayWorld.
	TrashCanMorph slideDismissalsToTrash ifTrue:
		[aForm slideFrom: startPoint to: endPoint nSteps: 12 delay: 15].
	ScrapBook default addToTrash: self! !

!Morph methodsFor: 'menus' stamp: 'sw 8/30/1998 09:42'!
snapToEdgeIfAppropriate
	| edgeSymbol oldBounds aWorld |
	(edgeSymbol := self valueOfProperty: #edgeToAdhereTo) ifNotNil:
		[oldBounds := bounds.
		self adhereToEdge: edgeSymbol.
		bounds ~= oldBounds ifTrue: [(aWorld := self world) ifNotNil: [aWorld viewBox ifNotNil:
			[aWorld displayWorld]]]]! !

!Morph methodsFor: 'layout-properties' stamp: 'dgd 2/16/2003 20:02'!
spaceFillWeight
	"Layout specific. This property describes the relative weight that 
	should be given to the receiver when extra space is distributed 
	between different #spaceFill cells."

	^ self
		valueOfProperty: #spaceFillWeight
		ifAbsent: [1]! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/15/2000 14:16'!
spaceFillWeight: aNumber
	"Layout specific. This property describes the relative weight that should be given to the receiver when extra space is distributed between different #spaceFill cells."
	aNumber = 1
		ifTrue:[self removeProperty: #spaceFillWeight]
		ifFalse:[self setProperty: #spaceFillWeight toValue: aNumber].
	self layoutChanged.! !

!Morph methodsFor: 'naming' stamp: 'gm 2/22/2003 13:16'!
specialNameInModel
	"Return the name for this morph in the underlying model or nil."

	"Not an easy problem.  For now, take the first part of the mouseDownSelector symbol in my eventHandler (fillBrushMouseUp:morph: gives 'fillBrush').  5/26/97 tk"

	| hh |
	(self isMorphicModel) 
		ifTrue: [^self slotName]
		ifFalse: 
			[self eventHandler ifNotNil: 
					[self eventHandler mouseDownSelector ifNotNil: 
							[hh := self eventHandler mouseDownSelector indexOfSubCollection: 'Mouse'
										startingAt: 1.
							hh > 0 
								ifTrue: [^self eventHandler mouseDownSelector copyFrom: 1 to: hh - 1]].
					self eventHandler mouseUpSelector ifNotNil: 
							[hh := self eventHandler mouseUpSelector indexOfSubCollection: 'Mouse'
										startingAt: 1.
							hh > 0 ifTrue: [^self eventHandler mouseUpSelector copyFrom: 1 to: hh - 1]]]].

	"	(self eventHandler mouseDownRecipient respondsTo: #nameFor:) ifTrue: [
					^ self eventHandler mouseDownRecipient nameFor: self]]].	"
	"myModel := self findA: MorphicModel.
			myModel ifNotNil: [^ myModel slotName]"
	^self world specialNameInModelFor: self! !

!Morph methodsFor: 'accessing' stamp: 'tk 12/4/1998 13:06'!
sqkPage
	^ self valueOfProperty: #SqueakPage! !

!Morph methodsFor: 'initialization' stamp: 'sw 9/11/1998 11:13'!
standardPalette
	"Answer a standard palette forced by some level of enclosing presenter, or nil if none"
	| pal aPresenter itsOwner |
	(aPresenter := self presenter) ifNil: [^ nil].
	^ (pal := aPresenter ownStandardPalette)
		ifNotNil: [pal]
		ifNil:	[(itsOwner := aPresenter associatedMorph owner)
					ifNotNil:
						[itsOwner standardPalette]
					ifNil:
						[nil]]! !

!Morph methodsFor: 'stepping and presenter' stamp: ''!
start
	"Start running my script. For ordinary morphs, this means start stepping."

	self startStepping.
! !

!Morph methodsFor: 'event handling' stamp: 'mir 5/23/2000 17:43'!
startDrag: evt
	"Handle a double-click event. This message is only sent to clients that request it by sending #waitForClicksOrDrag:event: to the initiating hand in their mouseDown: method. This default implementation does nothing."

	self eventHandler ifNotNil:
		[self eventHandler startDrag: evt fromMorph: self].! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'mir 1/4/2001 11:02'!
startDrag: anItem with: anObject
	self currentHand attachMorph: anObject! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 1/31/2001 13:07'!
startStepping
	"Start getting sent the 'step' message."
	self startStepping: #stepAt: at: Time millisecondClockValue arguments: nil stepTime: nil.! !

!Morph methodsFor: 'stepping and presenter' stamp: 'sw 7/19/1998 11:51'!
startSteppingIn: aWorld
	"Start getting sent the 'step' message in aWorld"

	self step.  "one to get started!!"
	aWorld ifNotNil: [aWorld startStepping: self].
	self changed! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 10/22/2000 16:42'!
startSteppingSelector: aSelector
	"Start getting sent the 'step' message."
	self startStepping: aSelector at: Time millisecondClockValue arguments: nil stepTime: nil.! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 10/22/2000 16:36'!
startStepping: aSelector at: scheduledTime arguments: args stepTime: stepTime
	"Start stepping the receiver"
	| w |
	w := self world.
	w ifNotNil: [
		w startStepping: self at: scheduledTime selector: aSelector arguments: args stepTime: stepTime.
		self changed].! !

!Morph methodsFor: 'menu' stamp: 'dgd 9/29/2004 13:26'!
startWiring
	Smalltalk
		at: #NCAAConnectorMorph
		ifPresent: [:connectorClass | connectorClass newCurvyArrow startWiringFrom: self] ! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 2/12/2001 17:04'!
step
	"Do some periodic activity. Use startStepping/stopStepping to start and stop getting sent this message. The time between steps is specified by this morph's answer to the stepTime message.  The generic version dispatches control to the player, if any.  The nasty circumlocation about owner's transformation is necessitated by the flexing problem that the player remains in the properties dictionary both of the flex and the real morph.  In the current architecture, only the top renderer's pointer to the player should actually be honored for the purpose of firing."
! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ul 12/12/2009 14:09'!
stepAt: millisecondClockValue
	"Do some periodic activity. Use startStepping/stopStepping to start and stop getting sent this message. The time between steps is specified by this morph's answer to the stepTime message.
	The millisecondClockValue parameter gives the value of the millisecond clock at the moment of dispatch.
	Default is to dispatch to the parameterless step method for the morph, but this protocol makes it possible for some morphs to do differing things depending on the clock value"
	self player ifNotNil:[:p| p stepAt: millisecondClockValue].
	self step
! !

!Morph methodsFor: 'testing' stamp: 'RAA 1/16/2001 17:20'!
stepTime
	"Answer the desired time between steps in milliseconds. This default implementation requests that the 'step' method be called once every second."

	^ self topRendererOrSelf player ifNotNil: [10] ifNil: [1000]! !

!Morph methodsFor: 'menus' stamp: 'dgd 8/30/2003 20:19'!
stickinessString
	"Answer the string to be shown in a menu to represent the  
	stickiness status"
	^ (self isSticky
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'resist being picked up' translated! !

!Morph methodsFor: 'accessing' stamp: 'nice 8/28/2010 01:26'!
sticky: aBoolean 
	"change the receiver's sticky property"
	(extension isNil and: [aBoolean not]) ifTrue: [^ self].
	self assureExtension sticky: aBoolean! !

!Morph methodsFor: 'stepping and presenter' stamp: ''!
stop
	"Stop running my script. For ordinary morphs, this means stop stepping."

	self stopStepping.
! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 12/15/2000 00:00'!
stopStepping
	"Stop getting sent the 'step' message."

	| w |
	w := self world.
	w ifNotNil: [w stopStepping: self].
! !

!Morph methodsFor: 'stepping and presenter' stamp: 'ar 12/15/2000 00:00'!
stopSteppingSelector: aSelector
	"Stop getting sent the given message."
	| w |
	w := self world.
	w ifNotNil: [w stopStepping: self selector: aSelector].
! !

!Morph methodsFor: 'stepping and presenter' stamp: 'sw 10/11/1999 12:59'!
stopSteppingSelfAndSubmorphs
	self allMorphsDo: [:m | m stopStepping]
! !

!Morph methodsFor: 'objects from disk' stamp: 'tk 7/11/1998 18:53'!
storeDataOn: aDataStream
	"Let all Morphs be written out.  All owners are weak references.  They only go out if the owner is in the tree being written."
	| cntInstVars cntIndexedVars ti localInstVars |

	"block my owner unless he is written out by someone else"
	cntInstVars := self class instSize.
	cntIndexedVars := self basicSize.
	localInstVars := Morph instVarNames.
	ti := 2.  
	((localInstVars at: ti) = 'owner') & (Morph superclass == Object) ifFalse:
			[self error: 'this method is out of date'].
	aDataStream
		beginInstance: self class
		size: cntInstVars + cntIndexedVars.
	1 to: ti-1 do:
		[:i | aDataStream nextPut: (self instVarAt: i)].
	aDataStream nextPutWeak: owner.	"owner only written if in our tree"
	ti+1 to: cntInstVars do:
		[:i | aDataStream nextPut: (self instVarAt: i)].
	1 to: cntIndexedVars do:
		[:i | aDataStream nextPut: (self basicAt: i)]! !

!Morph methodsFor: 'printing' stamp: 'jm 5/28/1998 17:58'!
structureString
	"Return a string that showing this morph and all its submorphs in an indented list that reflects its structure."

	| s |
	s := WriteStream on: (String new: 1000).
	self printStructureOn: s indent: 0.
	^ s contents
! !

!Morph methodsFor: 'meta-actions' stamp: 'ul 9/28/2011 14:07'!
subclassMorph
	"Create a new subclass of this morph's class and make this morph be an instance of it."

	| oldClass newClassName newClass |
	oldClass := self class.
	newClassName := UIManager default
		request: 'Please give this new class a name'
		initialAnswer: oldClass name.
	newClassName = '' ifTrue: [^ self].
	(Smalltalk includesKey: newClassName)
		ifTrue: [^ self inform: 'Sorry, there is already a class of that name'].

	newClass := oldClass subclass: newClassName asSymbol
		instanceVariableNames: ''
		classVariableNames: ''
		poolDictionaries: ''
		category: oldClass category asString.
	self becomeForward: (self as: newClass)! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'tk 10/20/2000 13:12'!
submorphAfter
	"Return the submorph after (behind) me, or nil"
	| ii |
	owner ifNil: [^ nil].
	^ (ii := owner submorphIndexOf: self) = owner submorphs size 
		ifTrue: [nil]
		ifFalse: [owner submorphs at: ii+1].
	
! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'tk 10/20/2000 13:13'!
submorphBefore
	"Return the submorph after (behind) me, or nil"
	| ii |
	owner ifNil: [^ nil].
	^ (ii := owner submorphIndexOf: self) = 1 
		ifTrue: [nil]
		ifFalse: [owner submorphs at: ii-1].
	
! !

!Morph methodsFor: 'layout' stamp: 'nice 12/27/2009 03:10'!
submorphBounds
	"Private. Compute the actual full bounds of the receiver"
	| box |
	submorphs do: [:m | | subBox | 
		(m visible) ifTrue: [
			subBox := m fullBounds.
			box 
				ifNil:[box := subBox copy]
				ifNotNil:[box := box quickMerge: subBox]]].
	box ifNil:[^self bounds]. "e.g., having submorphs but not visible"
	^ box origin asIntegerPoint corner: box corner asIntegerPoint
! !

!Morph methodsFor: 'submorphs-accessing' stamp: ''!
submorphCount

	^ submorphs size! !

!Morph methodsFor: 'submorphs-add/remove' stamp: 'sw 9/1/2000 10:16'!
submorphIndexOf: aMorph
	"Assuming aMorph to be one of my submorphs, answer where it occurs in my submorph list"

	^ submorphs indexOf: aMorph ifAbsent: [nil]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'sw 4/9/98 14:26'!
submorphNamed: aName
	^ self submorphNamed: aName ifNone: [nil]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'nice 12/27/2009 03:10'!
submorphNamed: aName ifNone: aBlock 
	"Find the first submorph with this name, or a button with an action selector of that name"

	
	self submorphs do: [:p | p knownName = aName ifTrue: [^p]].
	self submorphs do: 
			[:button | | sub args | 
			(button respondsTo: #actionSelector) 
				ifTrue: [button actionSelector == aName ifTrue: [^button]].
			((button respondsTo: #arguments) and: [(args := button arguments) notNil]) 
				ifTrue: [(args at: 2 ifAbsent: [nil]) == aName ifTrue: [^button]].
			(button isAlignmentMorph) 
				ifTrue: [(sub := button submorphNamed: aName ifNone: [nil]) ifNotNil: [^sub]]].
	^aBlock value! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'rhi 9/10/2000 12:12'!
submorphOfClass: aClass

	^self findA: aClass! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'efc 8/6/2005 11:35'!
submorphs
	"This method returns my actual submorphs collection. Modifying the collection directly could be dangerous; make a copy if you need to alter it."
	^ submorphs ! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'di 11/4/97 14:30'!
submorphsBehind: aMorph do: aBlock
	| behind |
	behind := false.
	submorphs do:
		[:m | m == aMorph ifTrue: [behind := true]
						ifFalse: [behind ifTrue: [aBlock value: m]]].
! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'dgd 2/22/2003 14:35'!
submorphsDo: aBlock 
	submorphs do: aBlock! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'di 11/4/97 14:29'!
submorphsInFrontOf: aMorph do: aBlock
	| behind |
	behind := false.
	submorphs do:
		[:m | m == aMorph ifTrue: [behind := true]
						ifFalse: [behind ifFalse: [aBlock value: m]]].
! !

!Morph methodsFor: 'submorphs-accessing' stamp: ''!
submorphsReverseDo: aBlock

	submorphs reverseDo: aBlock.! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'sw 8/15/97 22:03'!
submorphsSatisfying: aBlock
	^ submorphs select: [:m | (aBlock value: m) == true]! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'ul 8/2/2011 20:05'!
submorphThat: block1 ifNone: block2

	^submorphs detect: block1 ifNone: block2
	! !

!Morph methodsFor: 'submorphs-accessing' stamp: 'sw 7/3/1998 18:47'!
submorphWithProperty: aSymbol
	^ submorphs detect: [:aMorph | aMorph hasProperty: aSymbol] ifNone: [nil]! !

!Morph methodsFor: 'event handling' stamp: 'sw 11/16/1998 08:07'!
suspendEventHandler
	self eventHandler ifNotNil:
		[self setProperty: #suspendedEventHandler toValue: self eventHandler.
		self eventHandler: nil].
	submorphs do: [:m | m suspendEventHandler].  "All those rectangles"! !

!Morph methodsFor: 'event handling' stamp: 'dgd 8/28/2004 18:42'!
tabAmongFields
	^ Preferences tabAmongFields
		or: [self hasProperty: #tabAmongFields] ! !

!Morph methodsFor: 'meta-actions' stamp: 'wiz 7/20/2004 01:25'!
targetFromMenu: aMenu 
	"Some other morph become target of the receiver"
	| newTarget |
	
	newTarget := aMenu startUpWithCaption: self externalName , ' targets...'.
	newTarget
		ifNil: [^ self].
	self target: newTarget! !

!Morph methodsFor: 'meta-actions' stamp: 'ar 8/6/2009 20:16'!
targetWith: evt
	"Some other morph become target of the receiver"
	|  morphs newTarget |
	morphs := self potentialTargets.
	newTarget := UIManager default
		chooseFrom: (morphs collect: [:m | m knownName ifNil:[m class name asString]])
		values: morphs
		title:  self externalName, ' targets...'.
	newTarget ifNil:[^self].
	self target: newTarget.! !

!Morph methodsFor: 'accessing-backstop' stamp: 'wiz 2/14/2006 19:02'!
target: aMorph
"Morphs with targets will override. This backstop does nothing."
"This is here because targeting meta-actions are taken at morph level. 
Do not remove."! !

!Morph methodsFor: 'debug and other' stamp: 'dgd 8/30/2003 20:43'!
tempCommand
	"Generic backstop.  If you care to, you can comment out what's below here, and substitute your own code, though the intention of design of the feature is that you leave this method as it is, and instead reimplement tempCommand in the class of whatever individual morph you care to.  In any case, once you have your own #tempCommand in place, you will then be able to invoke it from the standard debugging menus."

	self inform: 'Before calling tempCommand, you
should first give it a definition.  To
do this, choose "define tempCommand"
from the debug menu.' translated! !

!Morph methodsFor: 'text-anchor' stamp: 'ar 12/16/2001 18:36'!
textAnchorType
	^self valueOfProperty: #textAnchorType ifAbsent:[#document]! !

!Morph methodsFor: 'text-anchor' stamp: 'ar 12/16/2001 18:37'!
textAnchorType: aSymbol
	aSymbol == #document
		ifTrue:[^self removeProperty: #textAnchorType]
		ifFalse:[^self setProperty: #textAnchorType toValue: aSymbol].! !

!Morph methodsFor: 'printing' stamp: 'sw 10/27/2000 17:47'!
textToPaste
	"If the receiver has text to offer pasting, answer it, else answer nil"

	^ nil! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 7/21/1998 21:17'!
textureParameters
	"Answer a triplet giving the preferred grid size, background color, and line color.  The choices here are as suggested by Alan, 9/13/97"

	^ Array with: 16 with: Color lightYellow with: Color lightGreen lighter lighter! !

!Morph methodsFor: 'thumbnail' stamp: 'mt 2/17/2016 15:19'!
thumbnail
	"Answer an appropiate form to represent the receiver"

	| maxExtent fb |
	
	maxExtent := 320 @ 240.
	fb := self fullBounds.
	^ fb area <= (maxExtent x * maxExtent y)
		ifTrue: [self imageForm]
		ifFalse: [self imageFormForRectangle: (fb topLeft extent: maxExtent)]! !

!Morph methodsFor: 'rounding' stamp: 'ar 12/25/2001 19:44'!
toggleCornerRounding
	self cornerStyle == #rounded
		ifTrue: [self cornerStyle: #square]
		ifFalse: [self cornerStyle: #rounded].
	self changed! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'panda 4/25/2000 15:46'!
toggleDragNDrop
	"Toggle this morph's ability to add and remove morphs via drag-n-drop."

		self enableDragNDrop: self dragNDropEnabled not.
! !

!Morph methodsFor: 'drop shadows' stamp: 'ar 10/26/2000 20:16'!
toggleDropShadow
	self hasDropShadow
		ifTrue:[self removeDropShadow]
		ifFalse:[self addDropShadow].! !

!Morph methodsFor: 'accessing' stamp: 'RAA 2/19/2001 17:38'!
toggleLocked
	
	self lock: self isLocked not! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/15/2001 12:21'!
toggleResistsRemoval
	"Toggle the resistsRemoval property"

	self resistsRemoval
		ifTrue:
			[self removeProperty: #resistsRemoval]
		ifFalse:
			[self setProperty: #resistsRemoval toValue: true]! !

!Morph methodsFor: 'accessing' stamp: 'md 2/27/2006 08:53'!
toggleStickiness
	"togle the receiver's Stickiness"
	extension ifNil: [^ self beSticky].
	extension sticky: extension sticky not! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:17'!
top
	" Return the y-coordinate of my top side "

	^ bounds top! !

!Morph methodsFor: 'geometry' stamp: 'di 3/6/2002 13:06'!
topCenter

	^ bounds topCenter! !

!Morph methodsFor: 'geometry' stamp: 'di 6/12/97 11:07'!
topLeft

	^ bounds topLeft! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:10'!
topLeft: aPoint
	" Move me so that my top left corner is at aPoint. My extent (width & height) are unchanged "

	self position: aPoint
! !

!Morph methodsFor: 'structure' stamp: 'sw 8/30/1998 09:47'!
topPasteUp
	"If the receiver is in a world, return that; otherwise return the outermost pasteup morph"
	^ self outermostMorphThat: [:m | m isKindOf: PasteUpMorph]! !

!Morph methodsFor: 'structure' stamp: 'dgd 2/22/2003 19:06'!
topRendererOrSelf
	"Answer the topmost renderer for this morph, or this morph itself if it has no renderer. See the comment in Morph>isRenderer."

	| top topsOwner |
	owner ifNil: [^self].
	self isWorldMorph ifTrue: [^self].	"ignore scaling of this world"
	top := self.
	topsOwner := top owner.
	[topsOwner notNil and: [topsOwner isRenderer]] whileTrue: 
			[top := topsOwner.
			topsOwner := top owner].
	^top! !

!Morph methodsFor: 'geometry' stamp: 'sw 8/20/97 23:04'!
topRight

	^ bounds topRight! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:12'!
topRight: aPoint
	" Move me so that my top right corner is at aPoint. My extent (width & height) are unchanged "

	self position: ((aPoint x - bounds width) @ (aPoint y))
! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:14'!
top: aNumber
	" Move me so that my top is at the y-coordinate aNumber. My extent (width & height) are unchanged "

	self position: (bounds left @ aNumber)! !

!Morph methodsFor: 'halos and balloon help' stamp: 'dgd 9/10/2004 13:59'!
transferHalo: event from: formerHaloOwner
	"Progressively transfer the halo to the next likely recipient"
	| localEvt w target |

	self flag: #workAround. "For halo's distinction between 'target' and 'innerTarget' we need to bypass any renderers."
	(formerHaloOwner == self and:[self isRenderer and:[self wantsHaloFromClick not]]) ifTrue:[
		event shiftPressed ifTrue:[
			target := owner.
			localEvt := event transformedBy: (self transformedFrom: owner).
		] ifFalse:[
			target := self renderedMorph.
			localEvt := event transformedBy: (target transformedFrom: self).
		].
		^target transferHalo: localEvt from: target].

"	formerHaloOwner == self ifTrue:[^ self removeHalo]."

	"Never transfer halo to top-most world"
	(self isWorldMorph and:[owner isNil]) ifFalse:[
		(self wantsHaloFromClick and:[formerHaloOwner ~~ self]) 
			ifTrue:[^self addHalo: event from: formerHaloOwner]].

	event shiftPressed ifTrue:[
		"Pass it outwards"
		owner ifNotNil:[^owner transferHalo: event from: formerHaloOwner].
		"We're at the top level; throw the event back in to find recipient"
		formerHaloOwner removeHalo.
		^self processEvent: event copy resetHandlerFields.
	].
	self submorphsDo:[:m|
		localEvt := event transformedBy: (m transformedFrom: self).
		(m fullContainsPoint: localEvt position) 
			ifTrue:[^m transferHalo: event from: formerHaloOwner].
	].
	"We're at the bottom most level; throw the event back up to the root to find recipient"
	formerHaloOwner removeHalo.

	Preferences maintainHalos ifFalse:[
		(w := self world) ifNil: [ ^self ].
		localEvt := event transformedBy: (self transformedFrom: w) inverseTransformation.
		^w processEvent: localEvt resetHandlerFields.
	].
! !

!Morph methodsFor: 'menus' stamp: 'sw 10/6/2004 12:16'!
transferStateToRenderer: aRenderer
	"Transfer knownName, actorState, visible, and player info over to aRenderer, which is being imposed above me as a transformation shell"

	| current |
	(current := self actorStateOrNil) ifNotNil:
		[aRenderer actorState: current.
		self actorState: nil].

	(current := self knownName) ifNotNil:
		[aRenderer setNameTo: current.
		self setNameTo: nil].

	(current := self player) ifNotNil:
		[aRenderer player: current.
		self player rawCostume: aRenderer.
		"NB player is redundantly pointed to in the extension of both the renderer and the rendee; this is regrettable but many years ago occasionally people tried to make that clean but always ran into problems iirc"
		"self player: nil"].

	aRenderer simplySetVisible: self visible



 

		! !

!Morph methodsFor: 'geometry' stamp: 'ar 10/22/2000 18:03'!
transformedBy: aTransform
	aTransform isIdentity ifTrue:[^self].
	aTransform isPureTranslation ifTrue:[
		^self position: (aTransform localPointToGlobal: self position).
	].
	^self addFlexShell transformedBy: aTransform! !

!Morph methodsFor: 'events-processing' stamp: 'ar 9/15/2000 21:09'!
transformedFrom: uberMorph
	"Return a transform to map coordinates of uberMorph, a morph above me in my owner chain, into the coordinates of MYSELF not any of my children."
	self flag: #arNote. "rename this method"
	owner ifNil:[^IdentityTransform new].
	^ (owner transformFrom: uberMorph)! !

!Morph methodsFor: 'event handling' stamp: 'RAA 6/19/2000 07:13'!
transformFromOutermostWorld
	"Return a transform to map world coordinates into my local coordinates"

	"self isWorldMorph ifTrue: [^ MorphicTransform identity]."
	^ self transformFrom: self outermostWorldMorph! !

!Morph methodsFor: 'event handling' stamp: ''!
transformFromWorld
	"Return a transform to map world coordinates into my local coordinates"

	^ self transformFrom: nil! !

!Morph methodsFor: 'event handling' stamp: 'dgd 2/22/2003 14:36'!
transformFrom: uberMorph 
	"Return a transform to be used to map coordinates in a morph above me into my childrens coordinates, or vice-versa. This is used to support scrolling, scaling, and/or rotation. This default implementation just returns my owner's transform or the identity transform if my owner is nil. 
	Note:  This method cannot be used to map into the receiver's coordinate system!!"

	(self == uberMorph or: [owner isNil]) ifTrue: [^IdentityTransform new].
	^owner transformFrom: uberMorph! !

!Morph methodsFor: 'dropping/grabbing' stamp: ''!
transportedMorph
	^self! !

!Morph methodsFor: '*Morphic-Sound-piano rolls' stamp: 'RAA 12/11/2000 23:05'!
triggerActionFromPianoRoll

	| evt |
	"a hack to allow for abitrary morphs to be dropped into piano roll"
	self world ifNil: [^self].
	evt := MouseEvent new setType: nil position: self center buttons: 0 hand: self world activeHand.
	self programmedMouseUp: evt for: self.

! !

!Morph methodsFor: 'naming' stamp: 'ar 1/3/2010 19:52'!
tryToRenameTo: aName
	"A new name has been submited; make sure it's appropriate, and react accordingly.  This circumlocution provides the hook by which the simple renaming of a field can result in a change to variable names in a stack, etc.  There are some problems to worry about here."

	self renameTo: aName.! !

!Morph methodsFor: 'menus' stamp: 'RAA 11/14/2000 13:46'!
uncollapseSketch

	| uncollapsedVersion w whomToDelete |

	(w := self world) ifNil: [^self].
	uncollapsedVersion := self valueOfProperty: #uncollapsedMorph.
	uncollapsedVersion ifNil: [^self].
	whomToDelete := self valueOfProperty: #collapsedMorphCarrier.
	uncollapsedVersion setProperty: #collapsedPosition toValue: whomToDelete position.

	whomToDelete delete.
	w addMorphFront: uncollapsedVersion.

! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'dgd 8/26/2003 21:44'!
undoGrabCommand
	"Return an undo command for grabbing the receiver"

	| cmd |
	owner ifNil:
		[^ nil]. "no owner - no undo"
	^ (cmd := Command new)
		cmdWording: 'move ' translated, self nameForUndoWording;
		undoTarget: self
		selector: #undoMove:redo:owner:bounds:predecessor:
		arguments: {cmd. false. owner. self bounds. (owner morphPreceding: self)};
		yourself! !

!Morph methodsFor: 'undo' stamp: 'mt 5/17/2016 15:59'!
undoMove: cmd redo: redo owner: formerOwner bounds: formerBounds predecessor: formerPredecessor 
	"Handle undo and redo of move commands in morphic"

	self owner ifNil: [^Beeper beep].
	redo 
		ifFalse: 
			["undo sets up the redo state first"

			cmd 
				redoTarget: self
				selector: #undoMove:redo:owner:bounds:predecessor:
				arguments: { 
						cmd.
						true.
						owner.
						bounds.
						owner morphPreceding: self}].
	formerOwner ifNotNil: 
			[formerPredecessor ifNil: [formerOwner addMorphFront: self]
				ifNotNil: [formerOwner addMorph: self after: formerPredecessor]].
	self bounds: formerBounds.
	(self isSystemWindow) ifTrue: [self beKeyWindow]! !

!Morph methodsFor: 'accessing' stamp: 'sw 11/30/1998 12:44'!
unHighlight
	self color: self regularColor! !

!Morph methodsFor: 'accessing' stamp: 'di 8/11/1998 12:33'!
unlock
	self lock: false! !

!Morph methodsFor: 'accessing' stamp: 'sw 8/15/97 23:59'!
unlockContents
	self submorphsDo:
		[:m | m unlock]! !

!Morph methodsFor: 'e-toy support' stamp: 'ar 8/6/2009 18:55'!
unlockOneSubpart
	| unlockables reply |
	unlockables := self submorphs select:
		[:m | m isLocked].
	unlockables size <= 1 ifTrue: [^ self unlockContents].
	reply := UIManager default
		chooseFrom: (unlockables collect: [:m | m externalName]) 
		values: unlockables
		title:  'Who should be be unlocked?' translated.
	reply isNil ifTrue: [^ self].
	reply unlock! !

!Morph methodsFor: 'fileIn/out' stamp: 'di 11/18/1999 09:15'!
updateAllFromResources

	self allMorphsDo: [:m | m updateFromResource]! !

!Morph methodsFor: 'naming' stamp: 'sw 1/29/2001 02:49'!
updateAllScriptingElements
	"A sledge-hammer sweep from the world down to make sure that all live scripting elements are up to date.  Presently in eclipse, not sent at the moment."

	| aPasteUp |
	(aPasteUp := self topPasteUp) ifNotNil:
		[aPasteUp allTileScriptingElements do: [:m | m bringUpToDate]]! !

!Morph methodsFor: 'events-accessing' stamp: 'rw 4/25/2002 07:17'!
updateableActionMap
	"Answer an updateable action map, saving it in my #actionMap property"
	
	| actionMap |
	actionMap := self valueOfProperty: #actionMap.
	actionMap ifNil:
		[actionMap := self createActionMap.
		self setProperty: #actionMap toValue: actionMap].
	^ actionMap! !

!Morph methodsFor: 'e-toy support' stamp: 'tk 10/19/1999 07:16'!
updateCachedThumbnail
	"If I have a cached thumbnail, then update it.  Copied up from Dan's original version in PasteUpMorph so it can be used by all morphs."
	| cachedThumbnail |

	(cachedThumbnail := self valueOfProperty: #cachedThumbnail) ifNotNil:
		[(cachedThumbnail respondsTo: #computeThumbnail) 
			ifTrue: [cachedThumbnail computeThumbnail]
			ifFalse: [self removeProperty: #computeThumbnail]].
		"Test and removal are because the thumbnail is being replaced by another Morph.  We don't know why.  Need to fix that at the source."! !

!Morph methodsFor: 'drawing' stamp: 'mt 10/6/2015 13:45'!
updateDropShadowCache

	| shadowBounds offset form canvas drawBlock localBounds mask maskCanvas |
	shadowBounds := self shadowOffset isRectangle
		ifTrue: [0@0 corner: (self bounds outsetBy: self shadowOffset) extent]
		ifFalse: [0@0 corner: self extent + self shadowOffset abs].
	offset := self shadowOffset isRectangle
		ifTrue: [0@0]
		ifFalse: [self shadowOffset max: 0@0].
	localBounds := self shadowOffset isRectangle
		ifTrue: [self shadowOffset topLeft extent: self extent]
		ifFalse: [(self shadowOffset negated max: 0@0) extent: self extent].
		
	form := Form extent: shadowBounds extent depth: Display depth.
	canvas := form getCanvas.

	drawBlock := self useSoftDropShadow
		ifFalse: [
			[:c | self wantsRoundedCorners
					ifTrue: [c fillRoundRect: localBounds radius: self cornerRadius fillStyle: self shadowColor]
					ifFalse: [c fillRectangle: localBounds fillStyle: self shadowColor]]]
		ifTrue: [
			[:c | self wantsRoundedCorners
					ifTrue: [0 to: 9 do: [:i |
						c
							fillRoundRect: (shadowBounds insetBy: i)
							radius: (self cornerRadius max: 20) -i
							fillStyle: (self shadowColor alpha: self shadowColor alpha * (i+1))]]
					ifFalse: [0 to: 9 do: [:i | 
						c
							fillRoundRect: (shadowBounds insetBy: i) radius: 20-i
							fillStyle: (self shadowColor alpha: self shadowColor alpha * (i+1))]]]].
			
	canvas 
		translateBy: offset
		during: [ :shadowCanvas | drawBlock value: shadowCanvas].

	"Support transparent morph colors without having the shadow to shine through.."
	mask := Form extent: shadowBounds extent depth: Display depth.
	maskCanvas := mask getCanvas.
	self wantsRoundedCorners
		ifTrue: [maskCanvas fillRoundRect: (localBounds insetBy: self borderWidth) radius: self cornerRadius fillStyle: Color black]
		ifFalse: [maskCanvas fillRectangle: (localBounds insetBy: self borderWidth) fillStyle: Color black].
	mask
		displayOn: form
		at: 0@0
		rule: Form erase.
	
	self setProperty: #dropShadow toValue: form.! !

!Morph methodsFor: 'fileIn/out' stamp: 'nk 1/6/2004 12:38'!
updateFromResource
	| pathName newMorph f |
	(pathName := self valueOfProperty: #resourceFilePath) ifNil: [^self].
	(pathName asLowercase endsWith: '.morph') 
		ifTrue: 
			[newMorph := (FileStream readOnlyFileNamed: pathName) fileInObjectAndCode.
			(newMorph isMorph) 
				ifFalse: [^self error: 'Resource not a single morph']]
		ifFalse: 
			[f := Form fromFileNamed: pathName.
			f ifNil: [^self error: 'unrecognized image file format'].
			newMorph := World drawingClass withForm: f].
	newMorph setProperty: #resourceFilePath toValue: pathName.
	self owner replaceSubmorph: self by: newMorph! !

!Morph methodsFor: 'copying' stamp: 'md 2/27/2006 08:53'!
updateReferencesUsing: aDictionary 
	"Update intra-morph references within a composite morph that 
	has been copied. For example, if a button refers to morph X in 
	the orginal 
	composite then the copy of that button in the new composite 
	should refer to 
	the copy of X in new composite, not the original X. This default 
	implementation updates the contents of any morph-bearing slot. 
	It may be 
	overridden to avoid this behavior if so desired."
	| old |
	Morph instSize + 1
		to: self class instSize
		do: [:i | 
			old := self instVarAt: i.
			old isMorph
				ifTrue: [self
						instVarAt: i
						put: (aDictionary
								at: old
								ifAbsent: [old])]].
	extension ifNotNil: [extension updateReferencesUsing: aDictionary]! !

!Morph methodsFor: 'thumbnail' stamp: 'tk 3/28/2000 11:08'!
updateThumbnailUrl
	"If I have a cached thumbnail, then update it's urls."
	| cachedThumbnail |

	(cachedThumbnail := self valueOfProperty: #cachedThumbnail) ifNotNil:
		[(cachedThumbnail respondsTo: #computeThumbnail) 
			ifTrue: [cachedThumbnail pageMorph: self url inBook: owner url]
			ifFalse: [self removeProperty: #computeThumbnail]].
			"Test and removal are because the thumbnail is being replaced 
			by another Morph.  We don't know why.  Need to fix that at 
			the source."! !

!Morph methodsFor: 'thumbnail' stamp: 'tk 3/28/2000 21:55'!
updateThumbnailUrlInBook: bookUrl
	"If I have a cached thumbnail, then update it's urls."
	| cachedThumbnail |

	(cachedThumbnail := self valueOfProperty: #cachedThumbnail) ifNotNil:
		[(cachedThumbnail respondsTo: #computeThumbnail) 
			ifTrue: [cachedThumbnail pageMorph: self url inBook: bookUrl]
			ifFalse: [self removeProperty: #computeThumbnail]].
			"Test and removal are because the thumbnail is being replaced 
			by another Morph.  We don't know why.  Need to fix that at 
			the source."! !

!Morph methodsFor: 'accessing' stamp: 'tk 2/17/1999 11:45'!
url
	"If I have been assigned a url, return it.  For PasteUpMorphs mostly."
	| sq |
	(sq := self sqkPage) ifNotNil: [^ sq url].
	^ self valueOfProperty: #url
		! !

!Morph methodsFor: 'copying' stamp: 'nk 10/11/2003 16:59'!
usableSiblingInstance
	"Return another similar morph whose Player is of the same class as mine.
	Do not open it in the world."

	| aName usedNames newPlayer newMorph topRenderer |
	(topRenderer := self topRendererOrSelf) == self 
		ifFalse: [^topRenderer usableSiblingInstance].
	self assuredPlayer assureUniClass.
	newMorph := self veryDeepCopySibling.
	newPlayer := newMorph player.
	newPlayer resetCostumeList.
	(aName := self knownName) isNil 
		ifTrue: [self player notNil ifTrue: [aName := newMorph innocuousName]].
	"Force a difference here"
	aName notNil 
		ifTrue: 
			[usedNames := (self world ifNil: [OrderedCollection new]
						ifNotNil: [self world allKnownNames]) copyWith: aName.
			newMorph setNameTo: (Utilities keyLike: aName
						satisfying: [:f | (usedNames includes: f) not])].
	newMorph privateOwner: nil.
	newPlayer assureEventHandlerRepresentsStatus.
	self presenter flushPlayerListCache.
	^newMorph! !

!Morph methodsFor: 'visual properties' stamp: 'nk 8/28/2003 15:57'!
useBitmapFill
	"Make receiver use a solid fill style (e.g., a simple color)"
	| fill |
	self fillStyle isBitmapFill ifTrue:[^self]. "Already done"
	fill := BitmapFillStyle fromForm: self defaultBitmapFillForm.
	"Note: Must fix the origin due to global coordinates"
	fill origin: self bounds origin.
	self fillStyle: fill.! !

!Morph methodsFor: 'visual properties' stamp: 'ar 6/25/1999 11:11'!
useDefaultFill
	"Make receiver use a solid fill style (e.g., a simple color)"
	self fillStyle: self defaultColor.! !

!Morph methodsFor: 'visual properties' stamp: 'sw 3/25/2008 03:17'!
useGradientFill
	"Make receiver use a solid fill style (e.g., a simple color)"

	| fill color1 color2 fil |
	((fil := self fillStyle) notNil and: [fil isSymbol not] and: [fil isGradientFill])  ifTrue:[^self]. "Already done"
	color1 _ self color asColor.
	color2 _ color1 negated.
	fill _ GradientFillStyle ramp: {0.0 -> color1. 1.0 -> color2}.
	fill origin: self topLeft.
	fill direction: 0 @ self bounds extent y.
	fill normal: self bounds extent x @ 0.
	fill radial: false.
	self fillStyle: fill! !

!Morph methodsFor: 'change reporting' stamp: 'tk 8/24/2001 22:07'!
userSelectedColor: aColor
	"The user, via the UI, chose aColor to be the color for the receiver; set it, and tell my owner in case he wishes to react"
	self color: aColor.
	self world ifNotNil: [owner colorChangedForSubmorph: self]! !

!Morph methodsFor: 'accessing' stamp: 'tk 12/16/1998 11:54'!
userString
	"Do I have a text string to be searched on?"

	^ nil! !

!Morph methodsFor: 'drop shadows' stamp: 'mt 4/6/2015 13:54'!
useSoftDropShadow

	^ self
		valueOfProperty: #useSoftDropShadow
		ifAbsent: [self class useSoftDropShadow]! !

!Morph methodsFor: 'drop shadows' stamp: 'mt 4/6/2015 13:57'!
useSoftDropShadow: aBooleanOrNil

	aBooleanOrNil
		ifNil: [self removeProperty: #useSoftDropShadow]
		ifNotNil: [self setProperty: #useSoftDropShadow toValue: aBooleanOrNil].! !

!Morph methodsFor: 'visual properties' stamp: 'ar 6/18/1999 06:57'!
useSolidFill
	"Make receiver use a solid fill style (e.g., a simple color)"
	self fillStyle isSolidFill ifTrue:[^self]. "Already done"
	self fillStyle: self fillStyle asColor. "Try minimizing changes"! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 08:53'!
valueOfProperty: aSymbol 
	"answer the value of the receiver's property named aSymbol"
	^ extension ifNotNil: [extension valueOfProperty: aSymbol]! !

!Morph methodsFor: 'accessing - properties' stamp: 'dgd 2/16/2003 20:55'!
valueOfProperty: aSymbol ifAbsentPut: aBlock 
	"If the receiver possesses a property of the given name, answer  
	its value. If not, then create a property of the given name, give 
	it the value obtained by evaluating aBlock, then answer that  
	value"
	^ self assureExtension valueOfProperty: aSymbol ifAbsentPut: aBlock! !

!Morph methodsFor: 'accessing - properties' stamp: 'ul 9/20/2014 20:32'!
valueOfProperty: aSymbol ifAbsent: aBlock 
	"if the receiver possesses a property of the given name, answer  
	its value. If not then evaluate aBlock and answer the result of  
	this block evaluation"
	
	extension ifNil: [ ^aBlock value ].
	^extension	valueOfProperty: aSymbol ifAbsent: aBlock! !

!Morph methodsFor: 'accessing - properties' stamp: 'md 2/27/2006 08:49'!
valueOfProperty: aSymbol ifPresentDo: aBlock 
	"If the receiver has a property of the given name, evaluate  
	aBlock on behalf of the value of that property"
	extension ifNil:  [^ self].
	^ aBlock value: (extension valueOfProperty: aSymbol ifAbsent: [^ self])! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'HenrikSperreJohansen 1/21/2011 18:34'!
vanishAfterSlidingTo: aPosition event: evt

	| aForm aWorld startPoint endPoint |
	aForm := self imageForm offset: 0@0.
	aWorld := self world.
	startPoint := evt hand fullBounds origin.
	self delete.
	aWorld displayWorld.
	endPoint := aPosition.
	aForm slideFrom: startPoint  to: endPoint nSteps: 12 delay: 15.
	SoundService soundEnabled ifTrue: [TrashCanMorph playDeleteSound].
! !

!Morph methodsFor: 'copying' stamp: 'tk 1/6/1999 17:27'!
veryDeepCopyWith: deepCopier
	"Copy me and the entire tree of objects I point to.  An object in the tree twice is copied once, and both references point to him.  deepCopier holds a dictionary of objects we have seen.  See veryDeepInner:, veryDeepFixupWith:"

	self prepareToBeSaved.
	^ super veryDeepCopyWith: deepCopier! !

!Morph methodsFor: 'copying' stamp: 'tk 2/3/2001 14:29'!
veryDeepFixupWith: deepCopier
	"If some fields were weakly copied, fix new copy here."

	"super veryDeepFixupWith: deepCopier.	Object has no fixups, so don't call it"

	"If my owner is being duplicated too, then store his duplicate.
	 If I am owned outside the duplicated tree, then I am no longer owned!!"
	owner := deepCopier references at: owner ifAbsent: [nil].

! !

!Morph methodsFor: 'copying' stamp: 'md 2/27/2006 08:47'!
veryDeepInner: deepCopier 
	"The inner loop, so it can be overridden when a field should not  
	be traced."
	"super veryDeepInner: deepCopier.	know Object has no inst vars"
	bounds := bounds clone.
	"Points are shared with original"
	"owner := owner.	special, see veryDeepFixupWith:"
	submorphs := submorphs veryDeepCopyWith: deepCopier.
	"each submorph's fixup will install me as the owner"
	"fullBounds := fullBounds.	fullBounds is shared with original!!"
	color := color veryDeepCopyWith: deepCopier.
	"color, if simple, will return self. may be complex"
	extension := (extension veryDeepCopyWith: deepCopier)! !

!Morph methodsFor: 'accessing' stamp: 'dgd 9/27/2004 11:45'!
viewBox
	^ self pasteUpMorph viewBox! !

!Morph methodsFor: 'debug and other' stamp: 'sw 8/4/2001 00:33'!
viewMorphDirectly
	"Open a Viewer directly on the Receiver, i.e. no Player involved"

	self presenter viewObjectDirectly: self renderedMorph

	! !

!Morph methodsFor: 'drawing' stamp: 'md 2/27/2006 08:49'!
visible
	"answer whether the receiver is visible"
	extension ifNil: [^ true].
	^ extension visible! !

!Morph methodsFor: 'accessing' stamp: 'dgd 9/1/2004 16:14'!
visibleClearArea
	"Answer the receiver visible clear area. The intersection 
	between the clear area and the viewbox."
	^ self viewBox intersect: self clearArea! !

!Morph methodsFor: 'drawing' stamp: 'md 2/27/2006 08:49'!
visible: aBoolean 
	"set the 'visible' attribute of the receiver to aBoolean"
	(extension isNil and:[aBoolean]) ifTrue: [^ self].
	self visible == aBoolean ifTrue: [^ self].
	self assureExtension visible: aBoolean.
	self changed! !

!Morph methodsFor: 'layout-properties' stamp: 'tk 10/30/2001 18:39'!
vResizeToFit: aBoolean
	aBoolean ifTrue:[
		self vResizing: #shrinkWrap.
	] ifFalse:[
		self vResizing: #rigid.
	].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:02'!
vResizing
	"Layout specific. This property describes how the receiver should be resized with respect to its owner and its children. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit children
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#rigid] ifNotNil:[props vResizing].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/31/2000 20:45'!
vResizingString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self vResizing! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:03'!
vResizing: aSymbol
	"Layout specific. This property describes how the receiver should be resized with respect to its owner and its children. Possible values are:
		#rigid			-	do not resize the receiver
		#spaceFill		-	resize to fill owner's available space
		#shrinkWrap	- resize to fit children
	"
	self assureLayoutProperties vResizing: aSymbol.
	self layoutChanged.
! !

!Morph methodsFor: 'halos and balloon help' stamp: 'rhi 10/5/2001 20:49'!
wantsBalloon
	"Answer true if receiver wants to show a balloon help text is a few moments."

	^ (self balloonText notNil) and: [Preferences balloonHelpEnabled]! !

!Morph methodsFor: 'connectors-scripting' stamp: 'nk 9/10/2004 11:37'!
wantsConnectorVocabulary
	"Answer true if I want to show a 'connector' vocabulary"
	^false! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 11/29/2001 19:50'!
wantsDirectionHandles
	^self valueOfProperty: #wantsDirectionHandles ifAbsent:[Preferences showDirectionHandles]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'ar 11/29/2001 19:52'!
wantsDirectionHandles: aBool
	aBool == Preferences showDirectionHandles
		ifTrue:[self removeProperty: #wantsDirectionHandles]
		ifFalse:[self setProperty: #wantsDirectionHandles toValue: aBool].
! !

!Morph methodsFor: 'event handling' stamp: 'ar 1/10/2001 21:28'!
wantsDropFiles: anEvent
	"Return true if the receiver wants files dropped from the OS."
	^false! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 10/11/2000 18:24'!
wantsDroppedMorph: aMorph event: evt
	"Return true if the receiver wishes to accept the given morph, which is being dropped by a hand in response to the given event. Note that for a successful drop operation both parties need to agree. The symmetric check is done automatically via aMorph wantsToBeDroppedInto: self."

	^self dropEnabled! !

!Morph methodsFor: 'event handling' stamp: 'di 9/14/2000 11:46'!
wantsEveryMouseMove
	"Unless overridden, this method allows processing to skip mouse move events
	when processing is lagging.  No 'significant' event (down/up, etc) will be skipped."

	^ false! !

!Morph methodsFor: 'halos and balloon help' stamp: 'dgd 2/22/2003 19:06'!
wantsHalo
	| topOwner |
	^(topOwner := self topRendererOrSelf owner) notNil 
		and: [topOwner wantsHaloFor: self]! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 4/8/98 13:26'!
wantsHaloFor: aSubMorph
	^ false! !

!Morph methodsFor: 'halos and balloon help' stamp: 'yo 9/12/2007 13:13'!
wantsHaloFromClick

	^ self valueOfProperty: #wantsHaloFromClick ifAbsent: [^true].! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 11/27/2001 14:49'!
wantsHaloHandleWithSelector: aSelector inHalo: aHaloMorph
	"Answer whether the receiver would like to offer the halo handle with the given selector (e.g. #addCollapseHandle:)"

	(#(addDismissHandle:) includes: aSelector) ifTrue:
		[^ self resistsRemoval not].

	(#( addDragHandle: ) includes: aSelector) ifTrue:
		[^ self okayToBrownDragEasily].

	(#(addGrowHandle: addScaleHandle:) includes: aSelector) ifTrue:
		[^ self okayToResizeEasily].

	(#( addRotateHandle: ) includes: aSelector) ifTrue:
		[^ self okayToRotateEasily].

	(#(addRecolorHandle:) includes: aSelector) ifTrue:
		[^ self renderedMorph wantsRecolorHandle].

	true ifTrue: [^ true]
	! !

!Morph methodsFor: 'event handling' stamp: 'mt 4/10/2015 11:52'!
wantsKeyboardFocus
	"Whether this morph should get the current keyboard focus when handling a keyboard event to speed up further event processing."

	^ false! !

!Morph methodsFor: 'event handling' stamp: 'sw 11/3/97 02:11'!
wantsKeyboardFocusFor: aSubmorph
	"Answer whether a plain mouse click on aSubmorph, a text-edit-capable thing, should result in a text selection there"
	^ false! !

!Morph methodsFor: 'e-toy support' stamp: 'sw 11/27/2001 14:52'!
wantsRecolorHandle
	"Answer whether the receiver would like a recoloring halo handle to be put up.  Since this handle also presently affords access to the property-sheet, it is presently always allowed, even though SketchMorphs don't like regular recoloring"

	^ true
	
! !

!Morph methodsFor: 'rounding' stamp: 'ar 12/22/2001 22:45'!
wantsRoundedCorners
	"Return true if the receiver wants its corners rounded"
	^ self cornerStyle == #rounded! !

!Morph methodsFor: 'halos and balloon help' stamp: 'sw 10/9/2000 16:56'!
wantsScriptorHaloHandle
	"Answer whether the receiver would like to have a Scriptor halo handle put up on its behalf.  Initially, only the ScriptableButton says yes"

	^ false! !

!Morph methodsFor: 'halos and balloon help' stamp: 'nk 6/12/2004 09:32'!
wantsSimpleSketchMorphHandles
	"Answer true if my halo's simple handles should include the simple sketch morph handles."
	^false! !

!Morph methodsFor: 'testing' stamp: 'sw 10/24/2004 15:28'!
wantsSteps
	"Return true if the receiver overrides the default Morph step method."
	"Details: Find first class in superclass chain that implements #step and return true if it isn't class Morph."

	| c |
	self isPartsDonor ifTrue: [^ false].
	(self == self topRendererOrSelf) ifTrue: [self player wantsSteps ifTrue: [^ true]].
	c := self class.
	[c includesSelector: #step] whileFalse: [c := c superclass].
	^ c ~= Morph! !

!Morph methodsFor: 'accessing' stamp: 'ar 6/23/2001 16:06'!
wantsToBeCachedByHand
	"Return true if the receiver wants to be cached by the hand when it is dragged around.
	Note: The default implementation queries all submorphs since subclasses may have shapes that do not fill the receiver's bounds completely."
	self hasTranslucentColor ifTrue:[^false].
	self submorphsDo:[:m|
		m wantsToBeCachedByHand ifFalse:[^false].
	].
	^true! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 9/18/2000 18:34'!
wantsToBeDroppedInto: aMorph
	"Return true if it's okay to drop the receiver into aMorph. This check is symmetric to #wantsDroppedMorph:event: to give both parties a chance of figuring out whether they like each other."
	^true! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'ar 2/10/1999 05:44'!
wantsToBeOpenedInWorld
	"Return true if the receiver wants to be put into the World directly,
	rather than allowing the user to place it (e.g., prevent attaching me
	to the hand after choosing 'new morph' in the world menu)"
	^false! !

!Morph methodsFor: 'accessing' stamp: 'dgd 8/31/2004 16:21'!
wantsToBeTopmost
	"Answer if the receiver want to be one of the topmost objects in its owner"
	^ self isFlapOrTab! !

!Morph methodsFor: 'event handling' stamp: 'bf 8/30/2007 20:28'!
wantsWindowEvents: anEvent
	"Return true if the receiver wants to process host window events. These are only dispatched to the World anyway, but one could have an eventListener in the Hand or a windowEventHandler in the World"
	^false! !

!Morph methodsFor: 'menu' stamp: 'dgd 9/18/2004 18:20'!
wantsYellowButtonMenu
	"Answer true if the receiver wants a yellow button menu"
	self
		valueOfProperty: #wantsYellowButtonMenu
		ifPresentDo: [:value | ^ value].
	""
	self isInSystemWindow
		ifTrue: [^ false].""
	(Preferences noviceMode
			and: [self isInDockingBar])
		ifTrue: [^ false].""
	^ Preferences generalizedYellowButtonMenu! !

!Morph methodsFor: 'menu' stamp: 'dgd 9/18/2004 18:35'!
wantsYellowButtonMenu: aBoolean 
	"Change the receiver to wants or not a yellow button menu"
	self setProperty: #wantsYellowButtonMenu toValue: aBoolean! !

!Morph methodsFor: 'geometry' stamp: ''!
width

	^ bounds width! !

!Morph methodsFor: 'geometry' stamp: 'efc 2/13/2003 18:22'!
width: aNumber
	" Set my width; my position (top-left corner) and height will remain the same "

	self extent: aNumber asInteger@self height.
! !

!Morph methodsFor: 'dropping/grabbing' stamp: 'sw 8/15/2000 16:58'!
willingToBeDiscarded
	^ true! !

!Morph methodsFor: 'event handling' stamp: 'bf 8/30/2007 20:32'!
windowEvent: anEvent
	"Host window event"! !

!Morph methodsFor: 'structure' stamp: 'di 11/13/2000 00:59'!
withAllOwners
	"Return the receiver and all its owners"

	^ Array streamContents: [:strm | self withAllOwnersDo: [:m | strm nextPut: m]]! !

!Morph methodsFor: 'structure' stamp: 'ar 9/14/2000 16:48'!
withAllOwnersDo: aBlock
	"Evaluate aBlock with the receiver and all of its owners"
	aBlock value: self.
	owner ifNotNil:[^owner withAllOwnersDo: aBlock].! !

!Morph methodsFor: 'structure' stamp: 'dgd 2/22/2003 14:36'!
world
	^owner isNil ifTrue: [nil] ifFalse: [owner world]! !

!Morph methodsFor: 'geometry' stamp: 'di 2/23/98 11:36'!
worldBounds
	^ self world bounds! !

!Morph methodsFor: 'geometry' stamp: 'dgd 9/10/2004 12:37'!
worldBoundsForHalo
	"Answer the rectangle to be used as the inner dimension of my halos.
	Allow for showing either bounds or fullBounds, and compensate for the optional bounds rectangle."

	| r |
	r := (Preferences haloEnclosesFullBounds)
		ifFalse: [ self boundsIn: nil ]
		ifTrue: [ self fullBoundsInWorld ].
	Preferences showBoundsInHalo ifTrue: [ ^r outsetBy: 2 ].
	^r! !

!Morph methodsFor: 'event handling' stamp: 'sw 5/6/1998 12:54'!
wouldAcceptKeyboardFocus
	"Answer whether a plain mouse click on the receiver should result in a text selection there"
	^ false! !

!Morph methodsFor: 'event handling' stamp: 'sw 8/29/2000 14:57'!
wouldAcceptKeyboardFocusUponTab
	"Answer whether the receiver is in the running as the new keyboard focus if the tab key were hit at a meta level.  This provides the leverage for tabbing among fields of a card, for example."

	^ false! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:02'!
wrapCentering
	"Layout specific. This property describes how the rows/columns in a list-like layout should be centered.
		#topLeft - center at start of secondary direction
		#bottomRight - center at end of secondary direction
		#center - center in the middle of secondary direction
		#justified - insert extra space inbetween rows/columns
	"
	| props |
	props := self layoutProperties.
	^props ifNil:[#topLeft] ifNotNil:[props wrapCentering].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 03:00'!
wrapCenteringString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self wrapCentering! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:03'!
wrapCentering: aSymbol
	"Layout specific. This property describes how the rows/columns in a list-like layout should be centered.
		#topLeft - center at start of secondary direction
		#bottomRight - center at end of secondary direction
		#center - center in the middle of secondary direction
		#justified - insert extra space inbetween rows/columns
	"
	self assureTableProperties wrapCentering: aSymbol.
	self layoutChanged.! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:03'!
wrapDirection
	"Layout specific. This property describes the direction along which a list-like layout should be wrapped. Possible values are:
		#leftToRight
		#rightToLeft
		#topToBottom
		#bottomToTop
		#none
	indicating in which direction wrapping should occur. This direction must be orthogonal to the list direction, that is if listDirection is #leftToRight or #rightToLeft then wrapDirection must be #topToBottom or #bottomToTop and vice versa."
	| props |
	props := self layoutProperties.
	^props ifNil:[#none] ifNotNil:[props wrapDirection].! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 10/29/2000 03:00'!
wrapDirectionString: aSymbol
	^self layoutMenuPropertyString: aSymbol from: self wrapDirection ! !

!Morph methodsFor: 'layout-properties' stamp: 'ar 11/13/2000 20:03'!
wrapDirection: aSymbol
	"Layout specific. This property describes the direction along which a list-like layout should be wrapped. Possible values are:
		#leftToRight
		#rightToLeft
		#topToBottom
		#bottomToTop
		#none
	indicating in which direction wrapping should occur. This direction must be orthogonal to the list direction, that is if listDirection is #leftToRight or #rightToLeft then wrapDirection must be #topToBottom or #bottomToTop and vice versa."
	self assureTableProperties wrapDirection: aSymbol.
	self layoutChanged.
! !

!Morph methodsFor: 'e-toy support' stamp: 'RAA 2/5/2001 15:35'!
wrappedInWindowWithTitle: aTitle
	| aWindow w2 |
	aWindow := (SystemWindow labelled: aTitle) model: Model new.
	aWindow addMorph: self frame: (0@0 extent: 1@1).
	w2 := aWindow borderWidth * 2.
	w2 := 3.		"oh, well"
	aWindow extent: self fullBounds extent + (0 @ aWindow labelHeight) + (w2 @ w2).
	^ aWindow! !

!Morph methodsFor: 'e-toy support' stamp: 'tk 9/3/1999 11:46'!
wrappedInWindow: aSystemWindow
	| aWindow |
	aWindow := aSystemWindow model: Model new.
	aWindow addMorph: self frame: (0@0 extent: 1@1).
	aWindow extent: self extent.
	^ aWindow! !

!Morph methodsFor: 'event handling' stamp: 'dgd 7/28/2005 13:02'!
yellowButtonActivity: shiftState 
	"Find me or my outermost owner that has items to add to a  
	yellow button menu.  
	shiftState is true if the shift was pressed.  
	Otherwise, build a menu that contains the contributions from  
	myself and my interested submorphs,  
	and present it to the user."
	| menu |
	self isWorldMorph
		ifFalse: [| outerOwner | 
			outerOwner := self outermostOwnerWithYellowButtonMenu.
			outerOwner
				ifNil: [^ self].
			outerOwner == self
				ifFalse: [^ outerOwner yellowButtonActivity: shiftState]].
	menu := self buildYellowButtonMenu: ActiveHand.
	menu
		addTitle: self externalName
		icon: (self iconOrThumbnailOfSize: (Preferences tinyDisplay ifTrue: [16] ifFalse: [28])).
	menu popUpInWorld: self currentWorld! !

!Morph methodsFor: 'geniestubs' stamp: 'nk 3/10/2004 06:38'!
yellowButtonGestureDictionaryOrName: aSymbolOrDictionary! !

Morph subclass: #MovieMorph
	instanceVariableNames: 'playMode msecsPerFrame rotationDegrees scalePoint frameList currentFrameIndex dwellCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!MovieMorph methodsFor: 'menu' stamp: 'wiz 12/4/2006 00:25'!
addCustomMenuItems: aCustomMenu hand: aHandMorph

	| movies subMenu |
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu addLine.
	subMenu := MenuMorph new defaultTarget: self.
	frameList size > 1 ifTrue: [
		subMenu add: 'repaint' translated action: #editDrawing.
		subMenu add: 'set rotation center' translated action: #setRotationCenter.
		subMenu add: 'play once' translated action: #playOnce.
		subMenu add: 'play loop' translated action: #playLoop.
		subMenu add: 'stop playing' translated action: #stopPlaying.
		currentFrameIndex > 1 ifTrue: [
			subMenu add: 'previous frame' translated action: #previousFrame].
		currentFrameIndex < frameList size ifTrue: [
			subMenu add: 'next frame' translated action: #nextFrame]].
	subMenu add: 'extract this frame' translated action: #extractFrame:.
	movies :=
		(self world rootMorphsAt: aHandMorph targetPoint)
			select: [:m | (m isKindOf: MovieMorph) or:
						[m isSketchMorph]].
	(movies size > 1) ifTrue:
		[subMenu add: 'insert into movie' translated action: #insertIntoMovie:].
	aCustomMenu add: 'movie...' translated subMenu: subMenu
! !

!MovieMorph methodsFor: 'menu' stamp: ''!
advanceFrame

	currentFrameIndex < frameList size
		ifTrue: [self setFrame: currentFrameIndex + 1]
		ifFalse: [self setFrame: 1].
! !

!MovieMorph methodsFor: 'geometry testing' stamp: 'dgd 2/22/2003 18:48'!
containsPoint: p 
	| frame |
	frame := self currentFrame.
	^ (frame notNil and: [playMode = #stop]) 
		ifTrue: [frame containsPoint: p]
		ifFalse: [super containsPoint: p]! !

!MovieMorph methodsFor: 'private' stamp: 'jdl 3/28/2003 08:03'!
currentFrame
	frameList isEmpty ifTrue: [^nil].
     currentFrameIndex := currentFrameIndex min: (frameList size).
     currentFrameIndex := currentFrameIndex max: 1.
	^frameList at: currentFrameIndex! !

!MovieMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:28'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color
		r: 1
		g: 0
		b: 1! !

!MovieMorph methodsFor: 'drawing' stamp: 'dgd 2/22/2003 18:47'!
drawOn: aCanvas 
	| frame |
	frame := self currentFrame.
	frame notNil 
		ifTrue: [^frame drawOn: aCanvas]
		ifFalse: [^super drawOn: aCanvas]! !

!MovieMorph methodsFor: 'menu' stamp: 'dgd 2/22/2003 18:47'!
editDrawing
	| frame |
	frame := self currentFrame.
	frame notNil 
		ifTrue: [frame editDrawingIn: self pasteUpMorph forBackground: false]! !

!MovieMorph methodsFor: 'menu' stamp: ''!
extractFrame: evt

	| f |
	f := self currentFrame.
	f ifNil: [^ self].
	frameList := frameList copyWithout: f.
	frameList isEmpty
		ifTrue: [self position: f position]
		ifFalse: [self setFrame: currentFrameIndex].
	evt hand attachMorph: f.
! !

!MovieMorph methodsFor: 'accessing' stamp: ''!
form

	^ self currentFrame form
! !

!MovieMorph methodsFor: 'initialization' stamp: 'cmm 5/8/2014 10:59'!
initialize
	"initialize the state of the receiver"
	super initialize.
	"#stop, #playOnce, or #loop"
	playMode := #stop.
	msecsPerFrame := 200.
	rotationDegrees := 0.
	scalePoint := 1.0 @ 1.0.
	frameList := Array empty.
	currentFrameIndex := 1.
	dwellCount := 0! !

!MovieMorph methodsFor: 'private' stamp: ''!
insertFrames: newFrames
	"Insert the given collection of frames into this movie just after the currentrame."

	frameList isEmpty ifTrue: [
		frameList := newFrames asArray copy.
		self setFrame: 1.
		^ self].

	frameList :=
		frameList
			copyReplaceFrom: currentFrameIndex + 1  "insert before"
			to: currentFrameIndex
			with: newFrames.
! !

!MovieMorph methodsFor: 'menu' stamp: 'wiz 12/4/2006 00:25'!
insertIntoMovie: evt

	| movies aTarget |
	movies :=
		(self world rootMorphsAt: evt hand targetPoint)
			select: [:m | ((m isKindOf: MovieMorph) or:
						 [m isSketchMorph]) and: [m ~= self]].
	movies isEmpty ifTrue: [^ self].
	aTarget := movies first.
	(aTarget isSketchMorph) ifTrue:
		[aTarget := aTarget replaceSelfWithMovie].
	movies first insertFrames: frameList.
	self delete.
! !

!MovieMorph methodsFor: 'menu' stamp: ''!
nextFrame

	currentFrameIndex < frameList size
		ifTrue: [self setFrame: currentFrameIndex + 1].
! !

!MovieMorph methodsFor: 'menu' stamp: ''!
playLoop

	playMode := #loop.
! !

!MovieMorph methodsFor: 'menu' stamp: ''!
playOnce

	self setFrame: 1.
	playMode := #playOnce.
! !

!MovieMorph methodsFor: 'menu' stamp: ''!
previousFrame

	currentFrameIndex > 1
		ifTrue: [self setFrame: currentFrameIndex - 1].
! !

!MovieMorph methodsFor: 'rotate scale and flex' stamp: ''!
rotationDegrees

	^ rotationDegrees
! !

!MovieMorph methodsFor: 'accessing' stamp: 'jm 7/24/97 15:05'!
scalePoint

	^ scalePoint
! !

!MovieMorph methodsFor: 'accessing' stamp: 'jm 7/24/97 15:05'!
scalePoint: newScalePoint

	| frame |
	newScalePoint ~= scalePoint ifTrue: [
		self changed.
		scalePoint := newScalePoint.
		frame := self currentFrame.
		frame ifNotNil: [frame scalePoint: newScalePoint].
		self layoutChanged.
		self changed].
! !

!MovieMorph methodsFor: 'private' stamp: 'jdl 3/28/2003 08:08'!
setFrame: newFrameIndex 
	| oldFrame p newFrame |
	oldFrame := self currentFrame.
	oldFrame ifNil: [^self].
	self changed.
	p := oldFrame referencePosition.
	currentFrameIndex := newFrameIndex.
     currentFrameIndex :=  currentFrameIndex min: (frameList size). 
	currentFrameIndex := currentFrameIndex max: 1.
	newFrame := frameList at: currentFrameIndex.
	newFrame referencePosition: p.
	oldFrame delete.
	self addMorph: newFrame.
	dwellCount := newFrame framesToDwell.
	self layoutChanged.
	self changed! !

!MovieMorph methodsFor: 'stepping and presenter' stamp: ''!
step

	playMode = #stop ifTrue: [^ self].

	dwellCount > 0 ifTrue: [
		dwellCount := dwellCount - 1.
		^ self].

	currentFrameIndex < frameList size
		ifTrue: [^ self setFrame: currentFrameIndex + 1].

	playMode = #loop
		ifTrue: [self setFrame: 1]
		ifFalse: [playMode := #stop].
! !

!MovieMorph methodsFor: 'testing' stamp: ''!
stepTime

	^ msecsPerFrame
! !

!MovieMorph methodsFor: 'menu' stamp: ''!
stopPlaying

	playMode := #stop.
	self setFrame: 1.
! !

Morph subclass: #NewBalloonMorph
	instanceVariableNames: 'balloonOwner textMorph maximumWidth orientation hasTail'
	classVariableNames: 'UseNewBalloonMorph'
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!NewBalloonMorph commentStamp: 'mt 3/31/2015 10:15' prior: 0!
A balloon is a bubble with an optional tail. It contains rich text, which describes something about its balloon-owner.!

!NewBalloonMorph class methodsFor: 'instance creation' stamp: 'mt 3/30/2015 14:17'!
string: str for: morph

	^ self string: str for: morph corner: #bottomLeft! !

!NewBalloonMorph class methodsFor: 'instance creation' stamp: 'mt 7/25/2016 10:18'!
string: message for: morph corner: symbol

	^ self new
		balloonOwner: morph;
		setText: message;
		orientation: symbol;
		yourself! !

!NewBalloonMorph class methodsFor: 'preferences' stamp: 'mt 7/25/2016 10:09'!
themeProperties

	^ super themeProperties, {
		{ #borderColor. 'Colors'. 'Color of the balloon''s border.' }.
		{ #borderWidth. 'Borders'. 'Width of the balloon''s border.' }.
		{ #color. 'Colors', 'Color for the balloon background.' }.
		{ #font. 'Fonts'. 'Font for balloon text if not overridden by text attributes.' }.
		{ #textColor. 'Colors'. 'Color for the balloon text if not overridden by text attributes.' }.
	}! !

!NewBalloonMorph class methodsFor: 'preferences' stamp: 'mt 3/30/2015 11:52'!
useNewBalloonMorph

	<preference: 'Use new-style balloon morphs'
		category: #Morphic
		description: 'The new-style balloon morphs are improved for better reading quality and support rich text.'
		type: #Boolean>
	^ UseNewBalloonMorph ifNil: [true]! !

!NewBalloonMorph class methodsFor: 'preferences' stamp: 'mt 3/30/2015 11:50'!
useNewBalloonMorph: aBoolean

	UseNewBalloonMorph := aBoolean.! !

!NewBalloonMorph methodsFor: 'updating' stamp: 'mt 7/25/2016 10:12'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	self setDefaultParameters.! !

!NewBalloonMorph methodsFor: 'accessing' stamp: 'mt 3/30/2015 11:51'!
balloonOwner

	^ balloonOwner! !

!NewBalloonMorph methodsFor: 'accessing' stamp: 'mt 3/30/2015 11:52'!
balloonOwner: aMorph

	balloonOwner := aMorph.! !

!NewBalloonMorph methodsFor: 'geometry' stamp: 'mt 3/31/2015 10:57'!
bubbleBounds

	^ self bounds insetBy: (0 @ self tailHeight corner: 0 @ self tailHeight)! !

!NewBalloonMorph methodsFor: 'geometry' stamp: 'mt 3/31/2015 10:38'!
bubbleInset

	^ 5@2! !

!NewBalloonMorph methodsFor: 'drawing' stamp: 'mt 10/6/2015 13:44'!
drawDropShadowOn: aCanvas

	aCanvas 
		translateBy: self shadowOffset 
		during: [ :shadowCanvas |
			(shadowCanvas isVisible: self bubbleBounds) ifTrue: [
				self wantsRoundedCorners
					ifTrue: [shadowCanvas fillRoundRect: self bubbleBounds radius: self cornerRadius fillStyle: self shadowColor]
					ifFalse: [shadowCanvas fillRectangle: self bubbleBounds fillStyle: self shadowColor]].
				
				self hasTail ifTrue: [
					shadowCanvas
						drawPolygon: self verticesForTail
						fillStyle: self shadowColor]].

! !

!NewBalloonMorph methodsFor: 'drawing' stamp: 'mt 7/25/2016 10:58'!
drawOn: aCanvas

	self fillStyle isColor
		ifFalse: [self fillStyle isGradientFill
				ifTrue: [self fillStyle direction: 0 @ self height]].
	"Bubble."
	self wantsRoundedCorners
		ifTrue: [aCanvas
			frameAndFillRoundRect: self bubbleBounds
			radius: self cornerRadius fillStyle: self fillStyle borderWidth: self borderStyle width borderColor: self borderStyle color]
		ifFalse: [aCanvas
			fillRectangle: self bubbleBounds
			fillStyle: self fillStyle borderStyle: self borderStyle].

	"Tail."
	self hasTail ifTrue: [
		self verticesForTail in: [:points |
			| pixelOffset |
			pixelOffset := points first y < points second y
				ifFalse: [points first x < points second x
					ifTrue: [self borderStyle width negated @ self borderStyle width] "bottomLeft"
					ifFalse: [self borderStyle width @ self borderStyle width]] "bottomRight"
				ifTrue: [points first x < points second x
					ifTrue: [self borderStyle width negated @ self borderStyle width negated] "topLeft"
					ifFalse: [self borderStyle width @ self borderStyle width negated]]. "topRight"

			aCanvas
				drawPolygon: points
				fillStyle: self fillStyle.
			aCanvas
				line: points first
				to: points second + pixelOffset
				width: self borderStyle width
				color: self borderStyle color.
			aCanvas
				line: points first
				to: points third + pixelOffset
				width: self borderStyle width
				color: self borderStyle color]]! !

!NewBalloonMorph methodsFor: 'accessing' stamp: 'mt 3/31/2015 10:08'!
hasTail

	^ hasTail ifNil: [true]! !

!NewBalloonMorph methodsFor: 'accessing' stamp: 'mt 3/31/2015 10:09'!
hasTail: aBoolean

	hasTail := aBoolean.! !

!NewBalloonMorph methodsFor: 'initialization' stamp: 'mt 7/25/2016 10:13'!
initialize

	super initialize.
	
	self setDefaultParameters.
	
	textMorph := TextMorph new
		wrapFlag: false;
		lock;
		yourself.
	
	self addMorph: textMorph.! !

!NewBalloonMorph methodsFor: 'accessing' stamp: 'mt 3/30/2015 14:50'!
maximumWidth

	^ maximumWidth ifNil: [
		maximumWidth := (self balloonOwner balloonFont widthOf: $m) * Preferences maxBalloonHelpLineLength]! !

!NewBalloonMorph methodsFor: 'accessing' stamp: 'mt 3/30/2015 14:40'!
maximumWidth: anInteger

	maximumWidth := anInteger.! !

!NewBalloonMorph methodsFor: 'WiW support' stamp: 'mt 3/30/2015 11:51'!
morphicLayerNumber

	"helpful for insuring some morphs always appear in front of or behind others.
	smaller numbers are in front"

	^5		"Balloons are very front-like things"! !

!NewBalloonMorph methodsFor: 'geometry' stamp: 'mt 4/15/2015 17:51'!
move: targetPoint

	self perform: self orientation asMutator with: targetPoint.! !

!NewBalloonMorph methodsFor: 'accessing' stamp: 'mt 3/31/2015 10:04'!
orientation
	"Encodes the position of the tail. #topLeft, #topRight, #bottomLeft, #bottomRight"
	
	^ orientation! !

!NewBalloonMorph methodsFor: 'accessing' stamp: 'mt 3/31/2015 10:40'!
orientation: aSymbol

	orientation := aSymbol.
	self changed.! !

!NewBalloonMorph methodsFor: 'initialization' stamp: 'mt 3/31/2016 10:37'!
popUp
	"No tail. Centered in the balloon owner. Do not remove other balloon helps."
	
	| w |	
	self balloonOwner ifNil: [^ self].
	self balloonOwner isInWorld ifFalse: [^ self].
	
	w := self balloonOwner world.
	
	self hasTail: false.
	self center: self balloonOwner center.
	self bounds: (self bounds translatedToBeWithin: w bounds).
	
	w activeHand addBalloonHelp: self.
	self openInWorld.! !

!NewBalloonMorph methodsFor: 'initialization' stamp: 'mt 4/5/2016 14:30'!
popUpAt: point
	"Do not remove other balloon helps."
	
	| w h |	
	self balloonOwner ifNil: [^ self].
	self balloonOwner isInWorld ifFalse: [^ self].
	
	w := self balloonOwner world.
	h := w activeHand.
	
	self hasTail: true.
		
	#(bottomLeft topLeft bottomRight topRight) detect: [:nextOrientation |
		| pointWithOffset |
		self orientation: nextOrientation.
		pointWithOffset := point + self tailOffset.
		self move: pointWithOffset.
		self bounds: (self bounds translatedToBeWithin: w bounds).
		(self bounds perform: self orientation) = pointWithOffset] ifNone: ["Keep last try."].	
	
	w activeHand addBalloonHelp: self.
	self openInWorld.! !

!NewBalloonMorph methodsFor: 'initialization' stamp: 'mt 3/31/2016 10:34'!
popUpAt: point forHand: aHand
	"Pop up the receiver as balloon help for the given hand"

	#(bottomLeft topLeft bottomRight topRight) detect: [:nextOrientation |
		| pointWithOffset |
		self orientation: nextOrientation.
		pointWithOffset := point + self tailOffset.
		self move: pointWithOffset.
		self bounds: (self bounds translatedToBeWithin: aHand world bounds).
		(self bounds perform: self orientation) = pointWithOffset] ifNone: ["Keep last try."].
		
	aHand world addMorphFront: self.
	aHand resetBalloonHelp: self.! !

!NewBalloonMorph methodsFor: 'initialization' stamp: 'mt 3/31/2015 11:31'!
popUpForHand: aHand

	self popUpAt: aHand position forHand: aHand.! !

!NewBalloonMorph methodsFor: 'initialization' stamp: 'mt 4/5/2016 14:29'!
popUpFor: aMorph hand: aHand
	"Pop up the receiver as balloon help for the given hand. Ignore if the hand does not touch the morph."

	(aMorph boundsInWorld containsPoint: aHand position)
		ifFalse: [^ self "Ignore"].
		
	self balloonOwner: aMorph.
	self popUpForHand: aHand.! !

!NewBalloonMorph methodsFor: 'initialization' stamp: 'mt 7/25/2016 11:03'!
setDefaultParameters

	self
		borderWidth: (self userInterfaceTheme borderWidth ifNil: [1]);
		borderColor: (self userInterfaceTheme borderColor ifNil: [Color r: 0.46 g: 0.46 b: 0.353]);
		color: (self userInterfaceTheme color ifNil: [Color r: 0.92 g: 0.92 b: 0.706]);
		hasDropShadow: (Preferences menuAppearance3d and: [self color isTranslucent not]);
		shadowOffset: 1@1;
		shadowColor: (self color muchDarker muchDarker alpha: 0.333);
		orientation: #bottomLeft.
		
	MenuMorph roundedMenuCorners
		ifTrue: [self cornerStyle: #rounded].
		
	"Gradients?"
	MenuMorph gradientMenu ifTrue: [
		| cc fill |
		cc := self color.
		fill := GradientFillStyle ramp: {
			0.0 -> Color white.
			0.15 -> (cc mixed: 0.5 with: Color white).
			0.5 -> cc.
			0.8 -> cc twiceDarker}.
		fill
			origin: self topLeft;
			direction: 0@self height.
		self fillStyle: fill].! !

!NewBalloonMorph methodsFor: 'initialization' stamp: 'mt 7/25/2016 10:15'!
setText: stringOrText

	| text |
	text := stringOrText asText.

	text unembellished ifTrue: [
		text addAttribute: (TextColor color: (self userInterfaceTheme textColor ifNil: [Color black]))].

	text addAttribute: (TextFontReference toFont: (self userInterfaceTheme font ifNil: [TextStyle defaultFont])).
	
	self textMorph wrapFlag: false.
	self textMorph newContents: text.
	self textMorph fullBounds.
	
	(self maximumWidth > 0 and: [self textMorph width > self maximumWidth])
		ifTrue: [
			self textMorph
				wrapFlag: true;
				width: self maximumWidth].
		
	self updateLayout.! !

!NewBalloonMorph methodsFor: 'geometry' stamp: 'mt 3/31/2015 10:10'!
tailHeight
	
	^ 8! !

!NewBalloonMorph methodsFor: 'geometry' stamp: 'mt 3/31/2015 18:25'!
tailOffset

	^ (Dictionary newFrom: {
		#topLeft -> (5@0).
		#topRight -> (-3@0).
		#bottomLeft -> (1@ -1).
		#bottomRight -> (-3 @ -3)}) at: self orientation! !

!NewBalloonMorph methodsFor: 'geometry' stamp: 'mt 3/31/2015 10:58'!
tailPosition

	^ self innerBounds perform: self orientation! !

!NewBalloonMorph methodsFor: 'geometry' stamp: 'mt 3/31/2015 11:25'!
tailWidth
	
	^ 15! !

!NewBalloonMorph methodsFor: 'accessing - ui' stamp: 'mt 3/30/2015 11:57'!
textMorph

	^ textMorph! !

!NewBalloonMorph methodsFor: 'layout' stamp: 'mt 3/31/2015 10:39'!
updateLayout

	self textMorph fullBounds.
	self extent: self textMorph extent + (2* self bubbleInset) +  (0 @ (2*self tailHeight)).
	self textMorph center: self center.! !

!NewBalloonMorph methodsFor: 'drawing' stamp: 'mt 10/6/2015 13:44'!
verticesForTail

	| offset factorX factorY tpos bpos |
	offset := 5 + (self wantsRoundedCorners
		ifTrue: [self cornerRadius]
		ifFalse: [0]).
	tpos := self tailPosition.
	factorX := tpos x < self center x ifTrue: [1] ifFalse: [-1].
	factorY := tpos y > self center y ifTrue: [1] ifFalse: [-1].
	bpos := self bubbleBounds perform: self orientation.
		
	^ {
		tpos.
		bpos + (((offset + self tailWidth) * factorX) @ (self borderStyle width negated * factorY)).
		bpos + ((offset * factorX) @ (self borderStyle width negated * factorY)).}! !

Morph subclass: #NewColorPickerMorph
	instanceVariableNames: 'target setColorSelector hsvaMorph colorPresenter'
	classVariableNames: 'UseIt'
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!NewColorPickerMorph commentStamp: 'cmm 12/3/2010 13:36' prior: 0!
A NewColorPickerMorph is a new widget for choosing colors in Morphic.  Instantiate a NewColorPickerMorph:

	(NewColorPickerMorph
		on: objectToHaveItsColorSet
		getColorSelector: itsColorGetterSymbol
		setColorSelector: itsColorSetterSymbol) openInWorld

!

!NewColorPickerMorph class methodsFor: 'create' stamp: 'cmm 12/4/2010 13:48'!
on: anObject 
	^ self
		on: anObject
		originalColor: anObject color
		setColorSelector: #color:! !

!NewColorPickerMorph class methodsFor: 'create' stamp: 'kfr 5/3/2015 16:20'!
on: objectToHaveItsColorSet originalColor: originalColor setColorSelector: colorSetterSymbol 
	^ self 
		on: objectToHaveItsColorSet
		originalColor: originalColor
		setColorSelector: colorSetterSymbol
		forPropertiesPanel: false! !

!NewColorPickerMorph class methodsFor: 'create' stamp: 'kfr 5/3/2015 16:00'!
on: objectToHaveItsColorSet originalColor: originalColor setColorSelector: colorSetterSymbol forPropertiesPanel: aBool
	^ self new
		setTarget: objectToHaveItsColorSet
		originalColor: originalColor
		setColorSelector: colorSetterSymbol
		forPropertiesPanel: aBool! !

!NewColorPickerMorph class methodsFor: 'accessing' stamp: 'cmm 12/4/2010 15:05'!
useIt
	<preference: 'Use the new color-picker'
		category: 'colors'
		description: 'When true, a newly-enhanced color-picker is used.'
		type: #Boolean>
	^ UseIt ifNil: [ false ]! !

!NewColorPickerMorph class methodsFor: 'accessing' stamp: 'cmm 12/4/2010 14:55'!
useIt: aBoolean
	UseIt := aBoolean! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'cmm 12/4/2010 12:38'!
closeButtonLabel
	^ 'Close' translated! !

!NewColorPickerMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 14:55'!
colorExpression
	"A Smalltalk which can create this color."
	^ self selectedColor printString! !

!NewColorPickerMorph methodsFor: 'accessing' stamp: 'cmm 12/4/2010 12:27'!
colorExpression: aString 
	"Set my color by evaluating aString, a Smalltalk expression which results in a Color instance."
	| col |
	{aString. 
	'Color ' , aString}
		detect:
			[ : each | ([ col := Compiler evaluate: each ]
				on: Error
				do:
					[ : err | nil ]) notNil ]
		ifNone: [ nil ].
	col ifNotNil: [ self selectedColor: col ]! !

!NewColorPickerMorph methodsFor: 'model' stamp: 'cmm 12/3/2010 15:11'!
colorSelected: aColor
	self targetColor: aColor.
	self changed: #colorExpression! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'cmm 12/3/2010 15:00'!
initialize
	super initialize.
	self initializeHsvaMorph! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'cmm 12/4/2010 11:29'!
initializeHsvaMorph
	hsvaMorph := HSVAColorSelectorMorph new
		 hResizing: #spaceFill ;
		 vResizing: #spaceFill ;
		 yourself.
	hsvaMorph
		when: #selectedColor
		send: #colorSelected:
		to: self! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'cmm 12/4/2010 12:38'!
newBottomRow
	^ Morph new
		 color: Color transparent ;
		 changeTableLayout ;
		 listDirection: #leftToRight ;
		 hResizing: #spaceFill; vResizing: #shrinkWrap ;
		 height: 20 ;
		 cellInset: 4 ;
		 addMorph: (StringMorph contents: 'Current selection:' translated) ;
		 addMorphBack: self newColorPresenterMorph ;
		 addMorphBack: self newCloseButton! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'kfr 5/2/2015 19:13'!
newCloseButton
	^ (PluggableButtonMorph
		on: self
		getState: nil
		action: #delete
		label: #closeButtonLabel)
		 vResizing: #spaceFill ;
		hResizing: #spaceFill;
		 yourself! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'mt 4/14/2015 19:01'!
newColorExpressionMorph
	| inputField builder |
	builder := ToolBuilder default.
	inputField := (builder build: (builder pluggableInputFieldSpec new
		model: self;
		getText: #colorExpression;
		setText: #colorExpression:)).
	inputField
		 hResizing: #spaceFill ;
		 vResizing: #rigid ;
		 height: (Preferences standardDefaultTextFont height * 3/2).
	^ inputField! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'cmm 12/3/2010 15:34'!
newColorPresenterMorph
	^ (ColorPresenterMorph
		on: hsvaMorph
		color: #selectedColor)
		 vResizing: #rigid ; height: 20 ;
		 hResizing: #spaceFill ;
		 yourself! !

!NewColorPickerMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 13:53'!
selectedColor
	"The color selected."
	^ hsvaMorph selectedColor! !

!NewColorPickerMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 13:53'!
selectedColor: aColor
	"The color selected."
	hsvaMorph selectedColor: aColor! !

!NewColorPickerMorph methodsFor: 'model' stamp: 'cmm 12/3/2010 13:39'!
setColorSelector
	"Answer the value of setColorSelector"

	^ setColorSelector! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'cmm 12/4/2010 13:46'!
setTarget: objectToHaveItsColorSet originalColor: aColor setColorSelector: colorSetterSymbol 
	target := objectToHaveItsColorSet.
	setColorSelector := colorSetterSymbol.
	hsvaMorph selectedColor: aColor.
	self setup! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'kfr 5/3/2015 16:04'!
setTarget: objectToHaveItsColorSet originalColor: aColor setColorSelector: colorSetterSymbol forPropertiesPanel: aBool 
	target := objectToHaveItsColorSet.
	setColorSelector := colorSetterSymbol.
	hsvaMorph selectedColor: aColor.
	aBool
		ifTrue: [self setupForProperties]
		ifFalse: [self setup]! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'kfr 5/2/2015 22:04'!
setup
	self
		 color: (Color white darker) ;
		 cornerStyle: #rounded ;
		 changeTableLayout ;
		 hResizing: #shrinkWrap ;
		 vResizing: #shrinkWrap ;
		 extent: 240@240 ;
		 addMorphBack: hsvaMorph ;
		 addMorphBack: self newColorExpressionMorph ;
		 addMorphBack: self newBottomRow ;
		 layoutInset: 4 ;
		 cellInset: 2.
		
		Preferences menuAppearance3d
		ifTrue: [self addDropShadow].
	! !

!NewColorPickerMorph methodsFor: 'initialize-release' stamp: 'kfr 5/3/2015 16:06'!
setupForProperties
	self
		 color: (Color white darker) ;
		 changeTableLayout ;
		 hResizing: #shrinkWrap ;
		 vResizing: #shrinkWrap ;
		 extent: 240@240 ;
		 addMorphBack: hsvaMorph ;
		 layoutInset: 4 ;
		 cellInset: 2.! !

!NewColorPickerMorph methodsFor: 'model' stamp: 'cmm 12/3/2010 13:39'!
target
	"Answer the object whose color will be controlled."
	^ target! !

!NewColorPickerMorph methodsFor: 'accessing' stamp: 'cmm 12/3/2010 13:52'!
targetColor: aColor 
	"The color of my target."
	target ifNotNil:
		[ target
			perform: setColorSelector
			with: aColor ]! !

Morph subclass: #PianoRollNoteMorph
	instanceVariableNames: 'trackIndex indexInTrack hitLoc editMode selected notePlaying'
	classVariableNames: 'SoundPlaying'
	poolDictionaries: ''
	category: 'Morphic-Sound'!

!PianoRollNoteMorph commentStamp: '<historical>' prior: 0!
A PianoRollNoteMorph is drawn as a simple mroph, but it carries the necessary state to locate its source sound event via its owner (a PianorRollScoreMorph) and the score therein.  Simple editing of pitch and time placement is provided here.!

!PianoRollNoteMorph methodsFor: 'selecting' stamp: 'di 6/17/1999 12:16'!
deselect

	selected ifFalse: [^ self].
	self changed.
	selected := false.
! !

!PianoRollNoteMorph methodsFor: 'drawing' stamp: 'di 6/17/1999 10:56'!
drawOn: aCanvas

	selected
		ifTrue: [aCanvas frameAndFillRectangle: self fullBounds fillColor: color borderWidth: 1 borderColor: Color black]
		ifFalse: [aCanvas fillRectangle: self bounds color: color].
! !

!PianoRollNoteMorph methodsFor: 'editing' stamp: 'di 6/19/1999 08:14'!
editPitch: evt

	| mk note |
	mk := owner midiKeyForY: evt cursorPoint y.
	note := (owner score tracks at: trackIndex) at: indexInTrack.
	note midiKey = mk ifTrue: [^ self].
	note midiKey: mk.
	self playSound: (self soundOfDuration: 999.0).
	self position: self position x @ ((owner yForMidiKey: mk) - 1)
! !

!PianoRollNoteMorph methodsFor: 'layout' stamp: 'di 6/17/1999 10:55'!
fullBounds

	selected
		ifTrue: [^ bounds expandBy: 1]
		ifFalse: [^ bounds]! !

!PianoRollNoteMorph methodsFor: 'editing' stamp: 'di 6/17/1999 16:08'!
gridToNextQuarter

	owner score gridTrack: trackIndex toQuarter: 1 at: indexInTrack.
	owner rebuildFromScore! !

!PianoRollNoteMorph methodsFor: 'editing' stamp: 'di 6/17/1999 16:08'!
gridToPrevQuarter

	owner score gridTrack: trackIndex toQuarter: -1 at: indexInTrack.
	owner rebuildFromScore! !

!PianoRollNoteMorph methodsFor: 'event handling' stamp: 'di 6/15/1999 14:55'!
handlesMouseDown: evt

	^ owner scorePlayer isPlaying not! !

!PianoRollNoteMorph methodsFor: 'accessing' stamp: 'di 6/17/1999 11:17'!
indexInTrack

	^ indexInTrack! !

!PianoRollNoteMorph methodsFor: 'menu' stamp: 'RAA 6/12/2000 09:03'!
invokeNoteMenu: evt
	"Invoke the note's edit menu."

	| menu |
	menu := MenuMorph new defaultTarget: self.
	menu addList:
		#(('grid to next quarter'		gridToNextQuarter)
		('grid to prev quarter'		gridToPrevQuarter)).

	menu popUpEvent: evt in: self world.
! !

!PianoRollNoteMorph methodsFor: 'event handling' stamp: 'di 6/19/1999 10:26'!
mouseDown: evt

	hitLoc := evt cursorPoint.
	editMode := nil.
	owner submorphsDo:
		[:m | (m isKindOf: PianoRollNoteMorph) ifTrue: [m deselect]].
	selected := true.
	self changed.
	owner selection: (Array with: trackIndex with: indexInTrack with: indexInTrack).
	self playSound! !

!PianoRollNoteMorph methodsFor: 'event handling' stamp: 'dgd 2/22/2003 14:49'!
mouseMove: evt 
	| delta offsetEvt |
	editMode isNil 
		ifTrue: 
			["First movement determines edit mode"

			((delta := evt cursorPoint - hitLoc) dist: 0 @ 0) <= 2 
				ifTrue: [^self	"No significant movement yet."].
			delta x abs > delta y abs 
				ifTrue: 
					[delta x > 0 
						ifTrue: 
							["Horizontal drag"

							editMode := #selectNotes]
						ifFalse: 
							[self playSound: nil.
							offsetEvt := evt copy setCursorPoint: evt cursorPoint + (20 @ 0).
							self invokeNoteMenu: offsetEvt]]
				ifFalse: [editMode := #editPitch	"Vertical drag"]].
	editMode == #editPitch ifTrue: [self editPitch: evt].
	editMode == #selectNotes ifTrue: [self selectNotes: evt]! !

!PianoRollNoteMorph methodsFor: 'event handling' stamp: 'di 6/19/1999 08:29'!
mouseUp: evt

	self playSound: nil! !

!PianoRollNoteMorph methodsFor: 'note playing' stamp: 'di 6/17/1999 11:48'!
noteInScore

	^ (owner score tracks at: trackIndex) at: indexInTrack
! !

!PianoRollNoteMorph methodsFor: 'note playing' stamp: 'di 6/17/1999 11:48'!
noteOfDuration: duration

	| note |
	note := self noteInScore.
	^ (owner scorePlayer instrumentForTrack: trackIndex)
			soundForMidiKey: note midiKey
			dur: duration
			loudness: (note velocity asFloat / 127.0)
! !

!PianoRollNoteMorph methodsFor: 'note playing' stamp: 'di 6/19/1999 08:19'!
playSound
	"This STARTS a single long sound.  It must be stopped by playing another or nil."

	^ self playSound: (self soundOfDuration: 999.0)! !

!PianoRollNoteMorph methodsFor: 'note playing' stamp: 'di 6/19/1999 08:13'!
playSound: aSoundOrNil

	SoundPlaying ifNotNil: [SoundPlaying stopGracefully].
	SoundPlaying := aSoundOrNil.
	SoundPlaying ifNotNil: [SoundPlaying play].! !

!PianoRollNoteMorph methodsFor: 'selecting' stamp: 'di 6/17/1999 12:16'!
select

	selected ifTrue: [^ self].
	selected := true.
	self changed! !

!PianoRollNoteMorph methodsFor: 'selecting' stamp: 'di 6/20/1999 11:44'!
selected

	^ selected! !

!PianoRollNoteMorph methodsFor: 'selecting' stamp: 'dgd 2/22/2003 14:50'!
selectFrom: selection 
	(trackIndex = selection first and: 
			[indexInTrack >= (selection second) and: [indexInTrack <= (selection third)]]) 
		ifTrue: [selected ifFalse: [self select]]
		ifFalse: [selected ifTrue: [self deselect]]! !

!PianoRollNoteMorph methodsFor: 'selecting' stamp: 'di 6/19/1999 10:24'!
selectNotes: evt

	| lastMorph oldEnd saveOwner |
	saveOwner := owner.
	(owner autoScrollForX: evt cursorPoint x) ifTrue:
		["If scroll talkes place I will be deleted and my x-pos will become invalid."
		owner := saveOwner.
		bounds := bounds withLeft: (owner xForTime: self noteInScore time)].
	oldEnd := owner selection last.
	(owner notesInRect: (evt cursorPoint x @ owner top corner: owner bottomRight))
		do: [:m | m trackIndex = trackIndex ifTrue: [m deselect]].
	self select.  lastMorph := self.
	(owner notesInRect: (self left @ owner top corner: evt cursorPoint x @ owner bottom))
		do: [:m | m trackIndex = trackIndex ifTrue: [m select.  lastMorph := m]].
	owner selection: (Array with: trackIndex with: indexInTrack with: lastMorph indexInTrack).
	lastMorph indexInTrack ~= oldEnd ifTrue:
		["Play last note as selection grows or shrinks"
		owner ifNotNil: [lastMorph playSound]]
! !

!PianoRollNoteMorph methodsFor: 'note playing' stamp: 'jm 6/1/2001 01:29'!
soundOfDuration: duration

	| sound |
	sound := MixedSound new.
	sound add: (self noteOfDuration: duration)
		pan: (owner scorePlayer panForTrack: trackIndex)
		volume: owner scorePlayer overallVolume *
				(owner scorePlayer volumeForTrack: trackIndex).
	^ sound reset
! !

!PianoRollNoteMorph methodsFor: 'accessing' stamp: 'di 6/17/1999 11:16'!
trackIndex

	^ trackIndex! !

!PianoRollNoteMorph methodsFor: 'initialization' stamp: 'di 6/17/1999 10:46'!
trackIndex: ti indexInTrack: i

	trackIndex := ti.
	indexInTrack := i.
	selected := false! !

Morph subclass: #PluggableButtonMorph
	instanceVariableNames: 'model label font getStateSelector actionSelector getLabelSelector getMenuSelector shortcutCharacter askBeforeChanging triggerOnMouseDown offColor onColor feedbackColor showSelectionFeedback allButtons arguments argumentsProvider argumentsSelector style hoverColor borderColor textColor labelOffset'
	classVariableNames: 'GradientButton RoundedButtonCorners'
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!PluggableButtonMorph commentStamp: '<historical>' prior: 0!
A PluggableButtonMorph is a combination of an indicator for a boolean value stored in its model and an action button. The action of a button is often, but not always, to toggle the boolean value that it shows. Its pluggable selectors are:

		getStateSelector		fetch a boolean value from the model
		actionSelector		invoke this button's action on the model
		getLabelSelector		fetch this button's lable from the model
		getMenuSelector		fetch a pop-up menu for this button from the model

Any of the above selectors can be nil, meaning that the model does not supply behavior for the given action, and the default behavior should be used. For example, if getStateSelector is nil, then this button shows the state of a read-only boolean that is always false.

The model informs its view(s) of changes by sending #changed: to itself with getStateSelector as a parameter. The view tells the model when the button is pressed by sending actionSelector.

If the actionSelector takes one or more arguments, then the following are relevant:
		arguments			A list of arguments to provide when the actionSelector is called.
		argumentsProvider	The object that is sent the argumentSelector to obtain arguments, if dynamic
		argumentsSelector	The message sent to the argumentProvider to obtain the arguments.

Options:
	askBeforeChanging		have model ask user before allowing a change that could lose edits
	triggerOnMouseDown	do this button's action on mouse down (vs. up) transition
	shortcutCharacter		a place to record an optional shortcut key
!

PluggableButtonMorph subclass: #MultiWindowLabelButtonMorph
	instanceVariableNames: 'savedMultiWindowState'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!MultiWindowLabelButtonMorph commentStamp: 'eem 6/30/2010 16:13' prior: 0!
A MultiWindowLabelButtonMorph is  a means of implementing tabbed windows.  It cooperates with PluggableSystemWindowWithButton to provide a drop-down menu of windows when clicking in the window label.  It requires the model to create a suitable sub-instance of SavedMultiWindowState for inactive windows.  To specify multi-windows use the multiWindowStyle: setter to customize a PluggableWindowSpec.

Instance Variables
	savedMultiWindowState:		<SavedMultiWindowState>

savedMultiWindowState
	- the sequence of models in a multi-window.!

!MultiWindowLabelButtonMorph methodsFor: 'drawing' stamp: 'eem 5/6/2008 13:24'!
drawOn: aCanvas 
	"Be invisible."! !

!MultiWindowLabelButtonMorph methodsFor: 'accessing' stamp: 'eem 6/30/2010 17:03'!
performAction
	"Override to interpret the actionSelector as a menu accessor and to activate that menu."
	actionSelector ifNotNil:
		[(model perform: actionSelector) ifNotNil:
			[:menu|
			menu
				invokeModalAt: self position - (0@5)
				in: ActiveWorld
				allowKeyboard: Preferences menuKeyboardControl]]! !

!MultiWindowLabelButtonMorph methodsFor: 'accessing' stamp: 'eem 6/30/2010 16:09'!
savedMultiWindowState
	"Answer the value of savedMultiWindowState"

	^ savedMultiWindowState! !

!MultiWindowLabelButtonMorph methodsFor: 'accessing' stamp: 'eem 6/30/2010 16:09'!
savedMultiWindowState: anObject
	"Set the value of savedMultiWindowState"

	savedMultiWindowState := anObject! !

!PluggableButtonMorph class methodsFor: 'example' stamp: 'dtl 2/12/2010 19:17'!
example
	"PluggableButtonMorph example openInWorld"

	| s1 s2 s3 b1 b2 b3 row switchClass |
	switchClass := Smalltalk at: #Switch ifAbsent: [^self inform: 'MVC class Switch not present'].
	s1 := switchClass new.
	s2 := switchClass new turnOn.
	s3 := switchClass new.
	s2 onAction: [s3 turnOff].
	s3 onAction: [s2 turnOff].
	b1 := (PluggableButtonMorph on: s1 getState: #isOn action: #switch) label: 'S1'.
	b2 := (PluggableButtonMorph on: s2 getState: #isOn action: #turnOn) label: 'S2'.
	b3 := (PluggableButtonMorph on: s3 getState: #isOn action: #turnOn) label: 'S3'.
	b1
		hResizing: #spaceFill;
		vResizing: #spaceFill.
	b2
		hResizing: #spaceFill;
		vResizing: #spaceFill.
	b3
		hResizing: #spaceFill;
		vResizing: #spaceFill.

	row := AlignmentMorph newRow
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		addAllMorphs: (Array with: b1 with: b2 with: b3);
		extent: 120@35.
	^ row
! !

!PluggableButtonMorph class methodsFor: 'preferences' stamp: 'mt 2/24/2015 08:58'!
gradientButton
	<preference: 'gradientButton'
		category: 'windows'
		description: 'If true, buttons will have a gradient look.'
		type: #Boolean>
	^ GradientButton ifNil: [ true ]! !

!PluggableButtonMorph class methodsFor: 'preferences' stamp: 'mt 7/25/2016 13:43'!
gradientButton: aBoolean

	GradientButton = aBoolean ifTrue: [^ self].
	GradientButton := aBoolean.
	
	PluggableButtonMorph allSubInstancesDo: [:ea | ea updateFillStyle].! !

!PluggableButtonMorph class methodsFor: 'instance creation' stamp: ''!
on: anObject

	^ self on: anObject getState: #isOn action: #switch
! !

!PluggableButtonMorph class methodsFor: 'instance creation' stamp: 'jm 5/4/1998 15:28'!
on: anObject getState: getStateSel action: actionSel

	^ self new
		on: anObject
		getState: getStateSel
		action: actionSel
		label: nil
		menu: nil
! !

!PluggableButtonMorph class methodsFor: 'instance creation' stamp: 'jm 5/4/1998 15:28'!
on: anObject getState: getStateSel action: actionSel label: labelSel

	^ self new
		on: anObject
		getState: getStateSel
		action: actionSel
		label: labelSel
		menu: nil
! !

!PluggableButtonMorph class methodsFor: 'instance creation' stamp: 'jm 5/4/1998 15:29'!
on: anObject getState: getStateSel action: actionSel label: labelSel menu: menuSel

	^ self new
		on: anObject
		getState: getStateSel
		action: actionSel
		label: labelSel
		menu: menuSel
! !

!PluggableButtonMorph class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 13:35'!
roundedButtonCorners
	<preference: 'Rounded Button Corners'
		category: 'windows'
		description: 'Governs whether pluggable buttons in system windows should be rounded'
		type: #Boolean>
	^ RoundedButtonCorners ifNil: [ true ]! !

!PluggableButtonMorph class methodsFor: 'preferences' stamp: 'tfel 2/28/2010 14:06'!
roundedButtonCorners: aBoolean

	RoundedButtonCorners := aBoolean.
	World invalidRect: World bounds from: World.! !

!PluggableButtonMorph class methodsFor: 'preferences' stamp: 'mt 7/22/2016 15:24'!
themeProperties

	^ super themeProperties, {
		{ #borderColor. 'Colors'. 'Color of the button''s border.' }.
		{ #borderWidth. 'Borders'. 'Width of the button''s border.' }.
		{ #borderStyle. 'Borders'. 'Whether to use a plain border, inset, or outset.' }.
		{ #color. 'Colors'. 'Background color of the button.' }.

		{ #font. 'Fonts'. 'Font for button title.' }.
		{ #textColor. 'Colors'. 'Color for the button title label.' }.
	}! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'sw 12/28/2000 16:17'!
actionSelector
	"Answer the receiver's actionSelector"

	^ actionSelector! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: ''!
action: aSymbol 
	"Set actionSelector to be the action defined by aSymbol."

	actionSelector := aSymbol.
! !

!PluggableButtonMorph methodsFor: 'updating' stamp: 'mt 7/22/2016 14:33'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	self setDefaultParameters.! !

!PluggableButtonMorph methodsFor: 'arguments' stamp: 'sw 2/17/2002 05:29'!
argumentsProvider: anObject argumentsSelector: aSelector
	"Set the argument provider and selector"

	argumentsProvider := anObject.
	argumentsSelector := aSelector! !

!PluggableButtonMorph methodsFor: 'arguments' stamp: 'sw 2/17/2002 01:03'!
arguments: args
	"If the receiver takes argument(s) that are static, they can be filled by calling this.  If its argument(s) are to be dynamically determined, then use an argumentProvider and argumentSelector instead"

	arguments := args! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'jm 4/2/98 18:53'!
askBeforeChanging

	^ askBeforeChanging
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'jm 4/7/98 19:17'!
askBeforeChanging: aBoolean
	"If this preference is turned on, then give the model an opportunity to ask the user before accepting a change that might cause unaccepted edits to be lost."

	askBeforeChanging := aBoolean.
! !

!PluggableButtonMorph methodsFor: 'browse' stamp: 'bp 10/19/2009 23:05'!
browseAllImplementorsOfActionSelector
	SystemNavigation default browseAllImplementorsOf: actionSelector localTo: model class! !

!PluggableButtonMorph methodsFor: 'browse' stamp: 'bp 10/19/2009 22:53'!
buildDebugMenu: aHandMorph
	| aMenu |
	aMenu := super buildDebugMenu: aHandMorph.
	aMenu addLine.
	aMenu add: 'implementors of' translated target: self action: #browseAllImplementorsOfActionSelector.
	^ aMenu! !

!PluggableButtonMorph methodsFor: 'drawing' stamp: 'mt 8/1/2016 15:55'!
drawLabelOn: aCanvas

	| fontToUse labelToUse colorToUse labelWidth layoutBounds drawBlock |
	self label ifNil: [^ self].

	layoutBounds := self layoutBounds.
	labelToUse := self label asString.
	fontToUse := self font.
	colorToUse := self textColorToUse.
	
	"Support very narrow buttons. Shrink text to monogram then."
	((layoutBounds width < self labelShrinkThreshold
		and: [self hResizing ~~ #shrinkWrap])
		and: [labelToUse size > 3]) ifTrue: [
			labelToUse := labelToUse first asString. "Show first character only."
			fontToUse := fontToUse emphasized: (TextEmphasis bold) emphasisCode].
	
	labelWidth := fontToUse widthOfString: labelToUse.

	drawBlock := [:c | c
			drawString: labelToUse
			at: (layoutBounds center x - (labelWidth // 2) max: (layoutBounds left)) 
				@ (layoutBounds center y - (fontToUse height // 2))
			font: fontToUse
			color: colorToUse].
		
	self clipSubmorphs
		ifTrue: [aCanvas clipBy: layoutBounds during: drawBlock]
		ifFalse: [drawBlock value: aCanvas]! !

!PluggableButtonMorph methodsFor: 'drawing' stamp: 'mt 4/29/2015 16:48'!
drawMorphLabelOn: aCanvas

	| layoutBounds |
	layoutBounds := self layoutBounds.

	self label privateFullMoveBy: (layoutBounds center - self label center).
	
	self clipSubmorphs
		ifTrue: [aCanvas
			clipBy: layoutBounds
			during: [:c | c fullDrawMorph: self label]]
		ifFalse: [aCanvas fullDrawMorph: self label].! !

!PluggableButtonMorph methodsFor: 'drawing' stamp: 'mt 7/27/2016 10:02'!
drawOn: aCanvas 

	(self fillStyle isColor not and: [self fillStyle isGradientFill])
		ifTrue: [self fillStyle origin: self topLeft; direction: 0@ self height].

	super drawOn: aCanvas.
	
	aCanvas
		translateBy: self labelOffset
		during: [:c |
			self label isMorph
				ifTrue: [self drawMorphLabelOn: c]
				ifFalse: [self drawLabelOn: c]].! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'ar 9/16/2009 21:42'!
enabled
	^true! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 15:32'!
feedbackColor
	^ feedbackColor! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'jm 5/4/1998 17:47'!
feedbackColor: aColor
	"Set the color of this button's selection feedback border."

	feedbackColor := aColor.
	self changed.
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 4/17/2015 12:04'!
font

	^ font ifNil: [Preferences standardButtonFont]! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 9/30/2015 14:07'!
font: aFont

	font = aFont ifTrue: [^ self].
	font := aFont.
	
	self updateMinimumExtent.
	self changed.! !

!PluggableButtonMorph methodsFor: 'private' stamp: 'dgd 2/21/2003 22:40'!
getMenu: shiftPressed 
	"Answer the menu for this button, supplying an empty menu to be filled in. If the menu selector takes an extra argument, pass in the current state of the shift key."

	| menu |
	getMenuSelector isNil ifTrue: [^nil].
	menu := MenuMorph new defaultTarget: model.
	getMenuSelector numArgs = 1 
		ifTrue: [^model perform: getMenuSelector with: menu].
	getMenuSelector numArgs = 2 
		ifTrue: 
			[^model 
				perform: getMenuSelector
				with: menu
				with: shiftPressed].
	^self error: 'The getMenuSelector must be a 1- or 2-keyword symbol'! !

!PluggableButtonMorph methodsFor: 'private' stamp: 'jmv 3/2/2010 16:21'!
getModelState
	"Answer the result of sending the receiver's model the getStateSelector message."

	^ getStateSelector 
		ifNil: [false]
		ifNotNil: [model perform: getStateSelector]! !

!PluggableButtonMorph methodsFor: 'event handling' stamp: 'jm 5/4/1998 16:57'!
handlesMouseDown: evt

	^ true
! !

!PluggableButtonMorph methodsFor: 'event handling' stamp: 'jrp 7/3/2005 18:13'!
handlesMouseOverDragging: evt

	^ true! !

!PluggableButtonMorph methodsFor: 'event handling' stamp: 'jrp 7/2/2005 10:44'!
handlesMouseOver: evt

	^ true! !

!PluggableButtonMorph methodsFor: 'layout-properties' stamp: 'mt 4/26/2015 19:51'!
hResizing: aSymbol
	"We adapt our minimum extent according to our resize behavior."
	
	self hResizing == aSymbol ifTrue: [^ self].
	super hResizing: aSymbol.
	self updateMinimumExtent.! !

!PluggableButtonMorph methodsFor: 'initialization' stamp: 'mt 7/23/2016 10:31'!
initialize

	super initialize.

	"Layout properties."
	self
		extent: 20 @ 15;
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap;
		layoutInset: (4@0 corner: 4@0);
		clipSubmorphs: true;
		wrapCentering: #center;
		cellPositioning: #topCenter.
	
	"Initialize instance variables."
	model := nil.
	label := nil.
	getStateSelector := nil.
	actionSelector := nil.
	getLabelSelector := nil.
	getMenuSelector := nil.
	shortcutCharacter := nil.
	askBeforeChanging := false.
	triggerOnMouseDown := false.
	allButtons := nil.
	argumentsProvider := nil.
	argumentsSelector := nil.
	
	self setDefaultParameters.
! !

!PluggableButtonMorph methodsFor: 'private' stamp: 'RAA 6/12/2000 09:04'!
invokeMenu: evt
	"Invoke my menu in response to the given event."
	| menu |
	menu := self getMenu: evt shiftPressed.
	menu ifNotNil: [menu popUpEvent: evt in: self world]! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 4/17/2015 12:13'!
label
	"Answer the DisplayObject used as this button's label."

	^ label ifNil: ['']
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 7/27/2016 10:01'!
labelOffset
	^ labelOffset ifNil: [0@0]! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 7/27/2016 10:01'!
labelOffset: aPoint
	labelOffset := aPoint.! !

!PluggableButtonMorph methodsFor: 'drawing' stamp: 'mt 4/29/2015 17:05'!
labelShrinkThreshold
	"Determines the minimum width for labels not to be shrunk down to their first character."
	
	^ (self font widthOf: $m)*3! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 7/23/2016 12:34'!
label: aStringOrTextOrMorph

	label = aStringOrTextOrMorph ifTrue: [^ self].
	label := aStringOrTextOrMorph isString
		ifFalse: [aStringOrTextOrMorph asMorph]
		ifTrue: [aStringOrTextOrMorph].
	
	self updateMinimumExtent.
	self changed.! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 4/17/2015 12:06'!
label: aStringOrTextOrMorph font: aFont

	self label: aStringOrTextOrMorph.
	self font: aFont.	! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'jm 5/4/1998 16:52'!
model: anObject
	"Set my model and make me me a dependent of the given object."

	model ifNotNil: [model removeDependent: self].
	anObject ifNotNil: [anObject addDependent: self].
	model := anObject.
! !

!PluggableButtonMorph methodsFor: 'event handling' stamp: 'mt 7/22/2016 14:33'!
mouseDown: evt
	"Details: If this button is triggered on mouse down or the event is the menu gesture, handle it immediately. Otherwise, make a list of buttons (including the receiver) for mouseMove feedback. This allows a simple radio-button effect among the button submorphs of a given morph."

	allButtons := nil.
	evt yellowButtonPressed ifTrue: [^ self invokeMenu: evt].
	triggerOnMouseDown
		ifTrue: [self performAction]
		ifFalse: [
			allButtons := owner submorphs select: [:m | m class = self class].
			self updateFillStyle: evt].
! !

!PluggableButtonMorph methodsFor: 'event handling' stamp: 'mt 7/22/2016 14:34'!
mouseEnter: evt

	self updateFillStyle: evt.! !

!PluggableButtonMorph methodsFor: 'event handling' stamp: 'jrp 7/3/2005 18:13'!
mouseLeaveDragging: evt

	self mouseLeave: evt! !

!PluggableButtonMorph methodsFor: 'event handling' stamp: 'mt 7/22/2016 14:34'!
mouseLeave: evt

	self updateFillStyle: evt.! !

!PluggableButtonMorph methodsFor: 'event handling' stamp: 'mt 7/22/2016 14:34'!
mouseMove: evt

	allButtons ifNil: [^ self].
	allButtons do: [:m | m updateFillStyle: evt].! !

!PluggableButtonMorph methodsFor: 'event handling' stamp: 'mt 7/22/2016 14:35'!
mouseUp: evt

	self updateFillStyle: evt.
	
	allButtons ifNil: [^ self].
	allButtons do: [:m |
		(m containsPoint: evt cursorPoint) ifTrue: [m performAction]].
	allButtons := nil.
	self changed.
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'sw 10/25/1999 14:36'!
offColor
	^ offColor
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 15:14'!
offColor: colorWhenOff
	"Set the fill colors to be used when this button is off."

	| cc |
	cc := colorWhenOff isTransparent ifTrue: [(Color gray: 0.9) alpha: 0.5] ifFalse: [colorWhenOff].
	self
		onColor: ((self userInterfaceTheme selectionModifier ifNil: [ [:c | c adjustBrightness: -0.2] ]) value: cc)
		offColor: cc
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 15:07'!
onColor: colorWhenOn offColor: colorWhenOff
	"Set the fill colors to be used when this button is on/off."

	onColor := colorWhenOn.
	offColor := colorWhenOff.
	
	hoverColor := (self userInterfaceTheme hoverModifier ifNil: [ [:c | c adjustBrightness: -0.1] ]) value: offColor.
	feedbackColor := (self userInterfaceTheme feedbackModifier ifNil: [ [:c | c adjustBrightness: -0.3] ]) value: offColor.
	
	self updateFillStyle.! !

!PluggableButtonMorph methodsFor: 'initialization' stamp: 'mt 7/23/2016 10:41'!
on: anObject getState: getStateSel action: actionSel label: labelSel menu: menuSel

	self model: anObject.
	getStateSelector := getStateSel.
	actionSelector := actionSel.
	getLabelSelector := labelSel.
	getMenuSelector := menuSel.
	
	self update: labelSel.
	self update: getStateSel.
	self updateFillStyle.! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'ul 11/15/2010 11:58'!
performAction
	"Inform the model that this button has been pressed. Sent by the controller when this button is pressed. If the button's actionSelector takes any arguments, they are obtained dynamically by sending the argumentSelector to the argumentsProvider"

	askBeforeChanging ifTrue: [model okToChange ifFalse: [^ self]].
	actionSelector ifNotNil:
		[actionSelector numArgs = 0
			ifTrue:
				[model perform: actionSelector]
			ifFalse:
				[argumentsProvider ifNotNil:
					[arguments := argumentsProvider perform: argumentsSelector].
					model perform: actionSelector withArguments: arguments]]! !

!PluggableButtonMorph methodsFor: 'initialization' stamp: 'mt 7/22/2016 15:34'!
setDefaultParameters
	"change the receiver's appareance parameters"

	self
		color: (self userInterfaceTheme color ifNil: [Color gray: 0.91]);
		borderStyle: (self userInterfaceTheme borderStyle ifNil: [BorderStyle default]);
		borderColor: (self userInterfaceTheme borderColor ifNil: [Color gray]);
		borderWidth: (self userInterfaceTheme borderWidth ifNil: [1]);
		font: (self userInterfaceTheme font ifNil: [TextStyle defaultFont]);
		textColor: (self userInterfaceTheme textColor ifNil: [Color black]).

	borderColor := self borderColor.
	self	offColor: self color.! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'jm 4/2/98 17:43'!
shortcutCharacter
	"Return the Character to be used as a shortcut to turn on this switch, or nil if this switch doesn't have a keyboard shortcut."

	^ shortcutCharacter
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'jm 4/2/98 17:43'!
shortcutCharacter: aCharacter 
	"Set the character to be used as a keyboard shortcut for turning on this switch."

	shortcutCharacter := aCharacter.
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'dtl 9/24/2011 09:33'!
style
	"Treat aSymbol as a hint to modify the button appearance."
	^style
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'dtl 9/24/2011 09:32'!
style: aSymbol
	"Use aSymbol as a hint to modify the button appearance."
	style := aSymbol
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 15:34'!
textColor
	^ textColor ifNil: [Color black "old instances"]! !

!PluggableButtonMorph methodsFor: 'drawing' stamp: 'mt 7/26/2016 10:53'!
textColorToUse

	^ self textColor! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 15:32'!
textColor: aColor
	textColor := aColor.
	self changed.! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'jm 4/2/98 19:26'!
triggerOnMouseDown

	^ triggerOnMouseDown
! !

!PluggableButtonMorph methodsFor: 'accessing' stamp: 'jm 4/7/98 19:16'!
triggerOnMouseDown: aBoolean
	"If this preference is turned on, then trigger my action immediately when the mouse goes down."

	triggerOnMouseDown := aBoolean.
! !

!PluggableButtonMorph methodsFor: 'updating' stamp: 'mt 7/22/2016 14:32'!
updateFillStyle

	self
		updateFillStylePressing: false
		hovering: false.! !

!PluggableButtonMorph methodsFor: 'updating' stamp: 'mt 7/31/2016 11:46'!
updateFillStylePressing: isPressing hovering: isHovering

	| gradient cc |
	"Migrate old instances."
	hoverColor ifNil: [hoverColor := onColor darker].
	
	self labelOffset: (isPressing ifTrue: [1@1] ifFalse: [0@0]).
	
	self getModelState
		ifTrue: [self color: onColor]
		ifFalse: [self color: offColor].
	self borderStyle color: borderColor.
	
	self class gradientButton ifFalse: [
		isPressing ifTrue: [
			self color: feedbackColor.
			self borderStyle color: feedbackColor muchDarker].
		isHovering ifTrue: [
			self color: hoverColor.
			self borderStyle color: borderColor].
		^ self].
	
	isPressing ifTrue: [
		cc := feedbackColor.
		self borderColor: feedbackColor muchDarker.
		gradient := GradientFillStyle ramp: {
			0.0 -> cc muchDarker.
			0.1-> (cc adjustBrightness: -0.2).
			0.5 -> cc.
			0.9-> (cc adjustBrightness: -0.1).
			1 -> cc muchDarker}].
	isHovering ifTrue: [
		cc := hoverColor.
		gradient := GradientFillStyle ramp: {
			0.0 -> Color white.
			0.1-> (cc adjustBrightness: 0.05).
			0.6 -> (cc darker)}].
	gradient ifNil: [
		cc := self color.
		gradient := GradientFillStyle ramp: {
			0.0 -> Color white.
			0.1-> (cc adjustBrightness: 0.05).
			0.6 -> (cc darker)}].
		
	gradient origin: bounds topLeft.
	gradient direction: 0@self height.
	
	self fillStyle: gradient.! !

!PluggableButtonMorph methodsFor: 'updating' stamp: 'mt 7/22/2016 15:08'!
updateFillStyle: evt

	self
		updateFillStylePressing: (evt redButtonPressed and: [self containsPoint: evt position])
		hovering: (evt redButtonPressed not and: [self containsPoint: evt position]).! !

!PluggableButtonMorph methodsFor: 'layout' stamp: 'mt 7/13/2016 18:59'!
updateMinimumExtent

	| hMin vMin |
	self label isMorph
		ifTrue: [^ self minimumExtent: self label minExtent].

	hMin := vMin := 0.
	self hResizing ~~ #spaceFill
		ifTrue: [hMin := (self font widthOfString: self label)].
	self vResizing ~~ #spaceFill
		ifTrue: [vMin := self font height].

	hMin := hMin + (2* self borderStyle width).
	vMin := vMin + (2* self borderStyle width).
			
	self layoutInset isRectangle
		ifTrue: [
			hMin := hMin + self layoutInset left + self layoutInset right.
			vMin := vMin + self layoutInset top + self layoutInset bottom]
		ifFalse: [self layoutInset isPoint
			ifTrue: [
				hMin := hMin + (2* self layoutInset x).
				vMin := vMin + (2* self layoutInset y)]
			ifFalse: [
				hMin := hMin + (2* self layoutInset).
				vMin := vMin + (2* self layoutInset)]].
	
	self minimumExtent: hMin @ vMin.

	"Since we have no submorphs, we have to resize here if we want to shrink wrap."
	self hResizing == #shrinkWrap ifTrue: [self width: hMin].
	self vResizing == #shrinkWrap ifTrue: [self height: vMin].! !

!PluggableButtonMorph methodsFor: 'updating' stamp: 'mt 7/23/2016 10:34'!
update: aParameter 

	getLabelSelector ifNotNil: [:sel |
		aParameter == sel ifTrue: [self label: (model perform: sel)]].
	getStateSelector ifNotNil: [:sel |
		aParameter == sel ifTrue: [self updateFillStyle]].! !

!PluggableButtonMorph methodsFor: 'copying' stamp: 'tk 1/7/1999 16:53'!
veryDeepFixupWith: deepCopier
	"If fields were weakly copied, fix them here.  If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

super veryDeepFixupWith: deepCopier.
model := deepCopier references at: model ifAbsent: [model].
! !

!PluggableButtonMorph methodsFor: 'copying' stamp: 'mt 7/28/2016 12:05'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all, but shared.  	Warning!!!!  Every instance variable defined in this class must be handled.  We must also implement veryDeepFixupWith:.  See DeepCopier class comment."

super veryDeepInner: deepCopier.
"model := model.		Weakly copied"
label := label veryDeepCopyWith: deepCopier.
"getStateSelector := getStateSelector.		a Symbol"
"actionSelector := actionSelector.		a Symbol"
"getLabelSelector := getLabelSelector.		a Symbol"
"getMenuSelector := getMenuSelector.		a Symbol"
shortcutCharacter := shortcutCharacter veryDeepCopyWith: deepCopier.
askBeforeChanging := askBeforeChanging veryDeepCopyWith: deepCopier.
triggerOnMouseDown := triggerOnMouseDown veryDeepCopyWith: deepCopier.
offColor := offColor veryDeepCopyWith: deepCopier.
onColor := onColor veryDeepCopyWith: deepCopier.
feedbackColor := feedbackColor veryDeepCopyWith: deepCopier.
hoverColor := hoverColor veryDeepCopyWith: deepCopier.
borderColor := borderColor veryDeepCopyWith: deepCopier.
textColor := textColor veryDeepCopyWith: deepCopier.
labelOffset := labelOffset veryDeepCopyWith: deepCopier.
allButtons := nil.		"a cache"
arguments := arguments veryDeepCopyWith: deepCopier.
argumentsProvider := argumentsProvider veryDeepCopyWith: deepCopier.
"argumentsSelector := argumentsSelector.   a Symbol" 
style := style.  "a Symbol"! !

!PluggableButtonMorph methodsFor: 'layout-properties' stamp: 'mt 4/26/2015 19:51'!
vResizing: aSymbol
	"We adapt our minimum extent according to our resize behavior."
	
	self vResizing == aSymbol ifTrue: [^ self].
	super vResizing: aSymbol.
	self updateMinimumExtent.! !

!PluggableButtonMorph methodsFor: 'rounding' stamp: 'mt 2/26/2015 18:26'!
wantsRoundedCorners
	"If the button is intended to invoke a menu for selection, provide a visual
	distinction by inverting the rounded corners attribute."

	^ (self class roundedButtonCorners or: [super wantsRoundedCorners])
		xor: style == #menuButton! !

Morph subclass: #PluggableTabButtonMorph
	instanceVariableNames: 'active model textSelector arcLengths subMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Pluggable Widgets'!

!PluggableTabButtonMorph commentStamp: 'KLC 9/17/2004 11:27' prior: 0!
This is a specialized pluggable button morph that is meant to represent a tab in a set of tabs arranged horizontally.  Each tab will overlap slightly when drawn.  All but one tab will be drawn in left to right order in the specified color, but lighter.  The active tab will be drawn last in the full color and slightly taller to indicate that it is selected.  Clicking the active tab has no effect but clicking any other tab will change the active tab to the clicked tab.

This morph does not itself accept any events.  The parent tab set will grab the mouse clicks and handle notifying the appropriate tabs that they have been activated or deactivated.

There is a single selector which provides the text for the button label and affects the width of the tab.  When the width changes the tab will inform its parent that it has changed and that the layout needs to be updated.  The model for the text selector of course should be the client for the tab set.

The button label can be a String, Text, or Morph.  Texts work better than plain Strings.!

!PluggableTabButtonMorph class methodsFor: 'instance creation' stamp: 'KLC 1/22/2004 14:46'!
on: anObject label: getTextSelector
	| instance |
	instance := super new.
	instance model: anObject.
	instance textSelector: getTextSelector.
	^ instance ! !

!PluggableTabButtonMorph methodsFor: 'access' stamp: 'KLC 1/22/2004 14:25'!
active
	active ifNil: [ active := false ].
	^ active! !

!PluggableTabButtonMorph methodsFor: 'access' stamp: 'KLC 1/22/2004 14:26'!
active: aBoolean
	active := aBoolean.
	self changed.! !

!PluggableTabButtonMorph methodsFor: 'private - access' stamp: 'KLC 1/23/2004 14:36'!
arcLengths
	arcLengths ifNil: [ self calculateArcLengths ].
	^ arcLengths! !

!PluggableTabButtonMorph methodsFor: 'private - access' stamp: 'KLC 1/23/2004 14:37'!
arcLengths: anArrayOfIntegers
	arcLengths := anArrayOfIntegers
! !

!PluggableTabButtonMorph methodsFor: 'precalculations' stamp: 'KLC 1/23/2004 14:46'!
calculateArcLengths
	| array radius |
	radius := self cornerRadius.
	array := Array new: radius.
	
	1 to: radius do: [ :i | | x |
		x := i - 0.5.
		array at: i
		 	put: (radius - ((2 * x * radius) - (x * x)) sqrt) asInteger].
		
	self arcLengths: array! !

!PluggableTabButtonMorph methodsFor: 'private - access' stamp: 'KLC 1/23/2004 11:30'!
cornerRadius
	^ 5
! !

!PluggableTabButtonMorph methodsFor: 'drawing' stamp: 'KLC 1/23/2004 15:49'!
drawOn: aCanvas
	self drawTabOn: aCanvas.
	self drawSubMorphOn: aCanvas! !

!PluggableTabButtonMorph methodsFor: 'drawing' stamp: 'KLC 9/17/2004 11:24'!
drawSubMorphOn: aCanvas
	| morphBounds |
	morphBounds := self bounds insetBy: (self cornerRadius + 3) @ (self topInactiveGap // 2 + 2).
	morphBounds := morphBounds translateBy: 0@(self topInactiveGap // 2 + 1).
	self active ifTrue: [
		morphBounds := morphBounds translateBy: 0@((self topInactiveGap // 2 + 1) negated)].
	self subMorph bounds height < (morphBounds height)
		ifTrue: [
			morphBounds := morphBounds
				insetBy: 0@((morphBounds height - self subMorph bounds height) // 2)].
	self subMorph bounds width < (morphBounds width)
		ifTrue: [
			morphBounds := morphBounds
				insetBy: ((morphBounds width - self subMorph bounds width) // 2)@0].

	self subMorph bounds: morphBounds.			
	aCanvas drawMorph: self subMorph! !

!PluggableTabButtonMorph methodsFor: 'drawing' stamp: 'KLC 2/2/2004 15:07'!
drawTabOn: aCanvas
	| top myColor cornerRadius myArcLengths myBounds |
	cornerRadius := self cornerRadius.
	myBounds := self bounds.
	self active
		ifTrue: [ top := myBounds top.
			myColor := self color ]
		ifFalse: [ top := myBounds top + self topInactiveGap.
			myColor := self color whiter whiter ].
	aCanvas fillRectangle:
		((myBounds left + cornerRadius)
				@ (top + cornerRadius)
			corner: (myBounds right - cornerRadius)
						@ self bottom)
		color: myColor.
	aCanvas fillRectangle:
		((myBounds left + (cornerRadius * 2)) @ top
			corner: (myBounds right - (cornerRadius * 2))
				@ (top + cornerRadius))
		color: myColor.
	aCanvas fillOval:
		((myBounds left + self cornerRadius) @ top
			corner: (myBounds left + (self cornerRadius * 3))
				@ (top + (self cornerRadius * 2)))
		color: myColor.
	aCanvas fillOval:
		((myBounds right - (self cornerRadius * 3)) @ top
			corner: (myBounds right - self cornerRadius)
				@ (top + (self cornerRadius * 2)))
		color: myColor.

	myArcLengths := self arcLengths.
	1 to: myArcLengths size do: [ :i | | length |
		length := myArcLengths at: i.
		aCanvas line: (myBounds left + cornerRadius - i) @ (myBounds bottom - 1 )
			to: (myBounds left + cornerRadius - i) @ (myBounds bottom - length - 1)
			color: myColor.
		aCanvas line: (myBounds right - cornerRadius + i - 1) @ (myBounds bottom - 1)
			to: (myBounds right - cornerRadius + i - 1) @ (myBounds bottom - length - 1)
			color: myColor]
	
! !

!PluggableTabButtonMorph methodsFor: 'initialization' stamp: 'KLC 1/22/2004 16:45'!
initialize
	^ super initialize
! !

!PluggableTabButtonMorph methodsFor: 'access' stamp: 'KLC 2/2/2004 14:05'!
innerExtent: aPoint
	"Set the extent based on the primary visible part of the tab.  In other words add twice the cornerRadius to this extent"
	self extent: (aPoint x + (self cornerRadius * 2)) @ (aPoint y)! !

!PluggableTabButtonMorph methodsFor: 'access' stamp: 'KLC 1/22/2004 14:39'!
model
	^ model
! !

!PluggableTabButtonMorph methodsFor: 'access' stamp: 'KLC 1/22/2004 14:39'!
model: anObject
	model := anObject! !

!PluggableTabButtonMorph methodsFor: 'access' stamp: 'KLC 2/2/2004 14:07'!
outerGap
	"The horizontal distance of the outer left and right edges of the tab excluding the inner visible part"
	^ self cornerRadius * 2! !

!PluggableTabButtonMorph methodsFor: 'stepping' stamp: 'KLC 2/2/2004 10:15'!
step
	self subMorph step.
	self changed.
! !

!PluggableTabButtonMorph methodsFor: 'stepping' stamp: 'KLC 1/23/2004 17:31'!
stepTime
	^ self subMorph stepTime
! !

!PluggableTabButtonMorph methodsFor: 'private - access' stamp: 'KLC 1/23/2004 16:40'!
subMorph
	subMorph ifNil: [ self update: self textSelector ].
	^ subMorph! !

!PluggableTabButtonMorph methodsFor: 'private - access' stamp: 'KLC 1/23/2004 16:40'!
subMorph: aMorph
	subMorph := aMorph
! !

!PluggableTabButtonMorph methodsFor: 'access' stamp: 'KLC 1/22/2004 14:39'!
textSelector
	^ textSelector
! !

!PluggableTabButtonMorph methodsFor: 'access' stamp: 'KLC 1/22/2004 14:39'!
textSelector: aSymbol
	textSelector := aSymbol! !

!PluggableTabButtonMorph methodsFor: 'actions' stamp: 'KLC 1/23/2004 15:38'!
toggle
	self active: self active not! !

!PluggableTabButtonMorph methodsFor: 'private - access' stamp: 'KLC 1/23/2004 11:30'!
topInactiveGap
	^ 5! !

!PluggableTabButtonMorph methodsFor: 'updating' stamp: 'KLC 1/23/2004 17:02'!
update: aSelector
	self textSelector ifNotNil: [
		aSelector = self textSelector
			ifTrue: [ | morph |
				(aSelector isSymbol and: [model notNil])
					ifTrue: [
						morph :=
							(self model perform: aSelector) asMorph]
					ifFalse: [ morph := aSelector value asMorph].
				self subMorph: morph]].
	self changed! !

!PluggableTabButtonMorph methodsFor: 'stepping' stamp: 'KLC 1/23/2004 17:31'!
wantsSteps
	^ self subMorph wantsSteps! !

Morph subclass: #SimpleHaloMorph
	instanceVariableNames: 'target positionOffset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!SimpleHaloMorph commentStamp: 'mt 11/6/2015 09:59' prior: 0!
This is a simple base class for halos in the system. It represents the minimal interface used to implement custom halo morphs. 

It provides:

- event handling code to invoke and transfer a halo when clicking the meta-button (blue)
- move the halo's target (morph) when click-hold-drag the meta-button
- one close button as a minimal handle (see #addHandles)

In general, the halo concept consists of one dedicated user interaction (meta-button click) to invoke an additional, interactive view (the halo) for any morph. This interactive view is itself a morph that can have submorphs (e.g. buttons or text fields) to enrich the target morph. Besides button-based interactions (e.g. resize, move, duplicate, etc.), this could also be used to show other, even domain-specific, information.

Use the halo concept to provide means to explore and modify interactive, graphical elements in Squeak and your application. You can benefit from this concept without wasting additional screen space. In non-Squeak applications, the meta-key (typically the mouse-wheel button) is often without real functionality for the user. There, it makes scrolling more convenient---at best. In Squeak, you can easily take advantage of this button click. 

Notice that direct user input is very limited. Many keyboard shortcuts (such as [ctrl]+[c]) are already pre-defined and should not be remapped for your domain-specific applications to avoid user confusion. Key chords (such as [ctrl]+[alt]+[v], [a] from Visual Studio) have to be learned with great effort. 

The left mouse click (red) selects something.
The right mouse click (yellow) invokes a context menu.
Only the middle click, the meta-key, the blue button, is unused in many environments.

This is where the halo concept comes in.

[For two- or single-button mice, the meta-key can be simulated.]!

SimpleHaloMorph subclass: #HaloMorph
	instanceVariableNames: 'innerTarget angleOffset minExtent growingOrRotating directionArrowAnchor haloBox simpleMode originalExtent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!HaloMorph commentStamp: '<historical>' prior: 0!
This morph provides a halo of handles for its target morph. Dragging, duplicating, rotating, and resizing to be done by mousing down on the appropriate handle. There are also handles for help and for a menu of infrequently used operations.!

!HaloMorph methodsFor: 'initialization' stamp: 'sw 10/2/2001 21:20'!
acceptNameEdit
	"If the name is currently under edit, accept the changes"

	| label |
	(label := self findA: NameStringInHalo) ifNotNil:
		[label hasFocus ifTrue:
			[label lostFocusWithoutAccepting]]! !

!HaloMorph methodsFor: 'handles' stamp: 'nk 6/12/2004 09:24'!
addChooseGraphicHandle: haloSpec
	"If the target is a sketch morph, and if the governing preference is set, add a halo handle allowing the user to select a new graphic"

	(Preferences showChooseGraphicHaloHandle and: [innerTarget isSketchMorph]) ifTrue:
		[self addHandle: haloSpec
				on: #mouseDown send: #chooseNewGraphicFromHalo to: innerTarget]
! !

!HaloMorph methodsFor: 'private' stamp: 'sw 4/27/2000 13:39'!
addCircleHandles
	| box |
	simpleMode := false.
	target isWorldMorph ifTrue: [^ self addHandlesForWorldHalos].

	self removeAllMorphs.  "remove old handles, if any"
	self bounds: target renderedMorph worldBoundsForHalo.  "update my size"
	box := self basicBox.

	target addHandlesTo: self box: box.

	self addName.
	growingOrRotating := false.
	self layoutChanged.
	self changed.
! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 12/13/2001 14:07'!
addCollapseHandle: handleSpec
	"Add the collapse handle, with all of its event handlers set up, unless the target's owner is not the world or the hand."

	| collapseHandle |
	(target owner notNil "nil happens, amazingly"
			and: [target owner isWorldOrHandMorph])
		ifFalse: [^ self].
	collapseHandle := self addHandle: handleSpec
		on: #mouseDown send: #mouseDownInCollapseHandle:with: to: self.
	collapseHandle on: #mouseUp send: #maybeCollapse:with: to: self.
	collapseHandle on: #mouseMove send: #setDismissColor:with: to: self
! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 1/26/2000 15:51'!
addDebugHandle: handleSpec
	Preferences debugHaloHandle ifTrue:
		[self addHandle: handleSpec
			on: #mouseDown send: #doDebug:with: to: self]
! !

!HaloMorph methodsFor: 'private' stamp: 'dgd 4/4/2006 16:56'!
addDirectionHandles

	| centerHandle d w directionShaft patch patchColor crossHairColor |
	self showingDirectionHandles ifFalse: [^ self].

	directionArrowAnchor := (target point: target referencePosition in: self world) rounded.
	patch := target imageFormForRectangle: (Rectangle center: directionArrowAnchor extent: 3@3).
	patchColor := patch colorAt: 1@1.

	(directionShaft := LineMorph newSticky makeForwardArrow)
		borderWidth: 2; borderColor: (Color green orColorUnlike: patchColor).
	self positionDirectionShaft: directionShaft.
	self addMorphFront: directionShaft.
	directionShaft setCenteredBalloonText: 'Set forward direction' translated;
		on: #mouseDown send: #doDirection:with: to: self;
		on: #mouseMove send: #trackDirectionArrow:with: to: self;
		on: #mouseUp send: #setDirection:with: to: self.

	d := 15.  "diameter"  w := 3.  "borderWidth"
	crossHairColor := Color red orColorUnlike: patchColor.
	(centerHandle := EllipseMorph newBounds: (0@0 extent: d@d) color: Color transparent)
			borderWidth: w; borderColor: (Color blue orColorUnlike: patchColor);
			addMorph: (LineMorph from: (d//2)@w to: (d//2)@(d-w-1) color: crossHairColor width: 1) lock;
			addMorph: (LineMorph from: w@(d//2) to: (d-w-1)@(d//2) color: crossHairColor width: 1) lock;
			align: centerHandle bounds center with: directionArrowAnchor.
	centerHandle wantsYellowButtonMenu: false.
	self addMorph: centerHandle.
	centerHandle setCenteredBalloonText: 'Rotation center (hold down the shift key and drag from here to change it)' translated;
			on: #mouseDown send: #prepareToTrackCenterOfRotation:with: to: self;
			on: #mouseMove send: #trackCenterOfRotation:with: to: self;
			on: #mouseUp send: #setCenterOfRotation:with: to: self
! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 11/27/2001 11:18'!
addDismissHandle: handleSpec
	"Add the dismiss handle according to the spec, unless selectiveHalos is on and my target resists dismissal"

	| dismissHandle |
	(target okayToAddDismissHandle or: [Preferences selectiveHalos not]) ifTrue:
		[dismissHandle := self addHandle: handleSpec
			on: #mouseDown send: #mouseDownInDimissHandle:with: to: self.
		dismissHandle on: #mouseUp send: #maybeDismiss:with: to: self.
		dismissHandle on: #mouseDown send: #setDismissColor:with: to: self.
		dismissHandle on: #mouseMove send: #setDismissColor:with: to: self]
! !

!HaloMorph methodsFor: 'handles' stamp: 'ar 10/25/2000 17:48'!
addDragHandle: haloSpec
	(self addHandle: haloSpec on: #mouseDown send: #startDrag:with: to: self)
		on: #mouseMove send: #doDrag:with: to: self


! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 8/31/2007 18:17'!
addDupHandle: haloSpec
	"Add the halo that offers duplication, or, when shift is down, make-sibling"

 	| aSelector |
	aSelector := innerTarget couldMakeSibling
		ifTrue:
			[#doDupOrMakeSibling:with:]
		ifFalse:
			[#doDup:with:].

	self addHandle: haloSpec on: #mouseDown send: aSelector to: self

! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 1/28/2000 09:59'!
addFewerHandlesHandle: haloSpec
	self addHandle: haloSpec on: #mouseDown send: #addSimpleHandles to: self
! !

!HaloMorph methodsFor: 'handles' stamp: 'gm 2/22/2003 13:13'!
addFontEmphHandle: haloSpec 
	(innerTarget isTextMorph) 
		ifTrue: 
			[self 
				addHandle: haloSpec
				on: #mouseDown
				send: #chooseEmphasisOrAlignment
				to: innerTarget]! !

!HaloMorph methodsFor: 'handles' stamp: 'gm 2/22/2003 13:13'!
addFontSizeHandle: haloSpec 
	(innerTarget isTextMorph) 
		ifTrue: 
			[self 
				addHandle: haloSpec
				on: #mouseDown
				send: #chooseFont
				to: innerTarget]! !

!HaloMorph methodsFor: 'handles' stamp: 'gm 2/22/2003 13:13'!
addFontStyleHandle: haloSpec 
	(innerTarget isTextMorph) 
		ifTrue: 
			[self 
				addHandle: haloSpec
				on: #mouseDown
				send: #chooseStyle
				to: innerTarget]! !

!HaloMorph methodsFor: 'private' stamp: 'sw 10/29/1999 15:31'!
addFullHandles
	"Later, obey a preference to choose between circle-iconic and solid-circles"
	self addCircleHandles! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 10/27/2000 17:22'!
addGrabHandle: haloSpec
	"If appropriate, add the black halo handle for picking up the target"

	innerTarget okayToAddGrabHandle ifTrue:
		[self addHandle: haloSpec on: #mouseDown send: #doGrab:with: to: self]

! !

!HaloMorph methodsFor: 'private' stamp: 'dgd 9/18/2004 18:23'!
addGraphicalHandleFrom: formKey at: aPoint
	"Add the supplied form as a graphical handle centered at the given point.  Return the handle."
	| handle aForm |
	aForm := (ScriptingSystem formAtKey: formKey) ifNil: [ScriptingSystem formAtKey: #SolidMenu].
	handle := ImageMorph new image: aForm; bounds: (Rectangle center: aPoint extent: aForm extent).
	handle wantsYellowButtonMenu: false.
	self addMorph: handle.
	handle on: #mouseUp send: #endInteraction to: self.
	^ handle
! !

!HaloMorph methodsFor: 'private' stamp: 'dgd 8/28/2003 15:15'!
addGraphicalHandle: formKey at: aPoint on: eventName send: selector to: recipient
	"Add the supplied form as a graphical handle centered at the given point, and set it up to respond to the given event by sending the given selector to the given recipient.  Return the handle."
	| handle |
	handle := self addGraphicalHandleFrom: formKey at: aPoint.
	handle on: eventName send: selector to: recipient.
	handle setBalloonText: (target balloonHelpTextForHandle: handle) translated.
	^ handle
! !

!HaloMorph methodsFor: 'handles' stamp: 'dgd 8/28/2004 18:13'!
addGrowHandle: haloSpec
	target isFlexMorph ifFalse: 
		[(self addHandle: haloSpec
				on: #mouseDown send: #startGrow:with: to: self)
				on: #mouseMove send: #doGrow:with: to: self]
	"This or addScaleHandle:, but not both, will prevail at any one time"
! !

!HaloMorph methodsFor: 'private' stamp: 'cmm 1/10/2011 14:59'!
addHandleAt: aPoint color: aColor icon: iconName on: eventName send: selector to: recipient 
	"Add a handle centered at the given point with the given color, 
	and set it up to respond to the given event by sending the 
	given selector to the given recipient. Return the handle."
	| handle |
	handle := self createHandleAt: aPoint color: (aColor alpha: 0.8) iconName: iconName.
	self addMorph: handle.

	handle on: #mouseUp send: #endInteraction to: self.
	handle on: eventName send: selector to: recipient.
	handle setBalloonText: (target balloonHelpTextForHandle: handle) translated.

	^ handle ! !

!HaloMorph methodsFor: 'private' stamp: 'sw 1/29/2000 18:36'!
addHandleAt: aPoint color: aColor on: eventName send: selector to: recipient
	^ self addHandleAt: aPoint color: aColor icon: nil on: eventName send: selector to: recipient
! !

!HaloMorph methodsFor: 'private' stamp: 'sw 1/28/2000 09:57'!
addHandles
	simpleMode == true
		ifTrue:
			[self addSimpleHandles]
		ifFalse:
			[self addCircleHandles]
! !

!HaloMorph methodsFor: 'private' stamp: 'tak 1/10/2007 14:51'!
addHandlesForWorldHalos
	"Add handles for world halos, like the man said"

	| box w |
	w _ self world ifNil:[target world].
	self removeAllMorphs.  "remove old handles, if any"
	self bounds: target bounds.
	box _ w bounds insetBy: self handleSize // 2.
	target addWorldHandlesTo: self box: box.

	Preferences uniqueNamesInHalos ifTrue:
		[innerTarget assureExternalName].
	self addNameBeneath: (box insetBy: (0@0 corner: 0@10)) string: innerTarget externalName.
	growingOrRotating _ false.
	self layoutChanged.
	self changed.
! !

!HaloMorph methodsFor: 'private' stamp: 'dgd 9/26/2004 19:37'!
addHandle: handleSpec on: eventName send: selector to: recipient 
	"Add a handle within the halo box as per the haloSpec, and set 
	it up to respond to the given event by sending the given 
	selector to the given recipient. Return the handle."
	| handle aPoint |

	aPoint := self
				positionIn: haloBox
				horizontalPlacement: handleSpec horizontalPlacement
				verticalPlacement: handleSpec verticalPlacement.

	handle := self
				addHandleAt: aPoint
				color: (Color colorFrom: handleSpec color)
				icon: handleSpec iconSymbol
				on: eventName
				send: selector
				to: recipient.

	self isMagicHalo
		ifTrue: [
			handle on: #mouseEnter send: #handleEntered to: self.
			handle on: #mouseLeave send: #handleLeft to: self].

	^ handle! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 1/26/2000 16:16'!
addHelpHandle: haloSpec
	target balloonText ifNotNil:
		[(self addHandle: haloSpec on: #mouseDown send: #mouseDownOnHelpHandle: to: innerTarget)
			on: #mouseUp send: #deleteBalloon to: innerTarget]
! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 12/29/2004 22:18'!
addMakeSiblingHandle: haloSpec
	"Add the halo handle that allows a sibling instance to be torn off, or, if the shift key is down, for a deep-copy duplicate to be made."

	self addHandle: haloSpec on: #mouseDown send: #doMakeSiblingOrDup:with: to: self

! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 1/26/2000 16:05'!
addMenuHandle: haloSpec
	self addHandle: haloSpec on: #mouseDown send: #doMenu:with: to: self! !

!HaloMorph methodsFor: 'private' stamp: 'sw 8/16/2000 17:41'!
addName
	"Add a name readout at the bottom of the halo."

	Preferences uniqueNamesInHalos ifTrue:
		[target assureExternalName].

	self addNameBeneath: self basicBox string: target externalName
! !

!HaloMorph methodsFor: 'private' stamp: 'dgd 4/4/2006 16:58'!
addNameBeneath: outerRectangle string: aString
	"Add a name display centered beneath the bottom of the outer rectangle. Return the handle."

	| nameMorph namePosition w |
	w := self world ifNil:[target world].
	nameMorph := NameStringInHalo contents: aString font: Preferences standardHaloLabelFont.
	nameMorph wantsYellowButtonMenu: false.
	nameMorph color: Color black.
	nameMorph useStringFormat; target: innerTarget; putSelector: #tryToRenameTo:.
	namePosition := outerRectangle bottomCenter -
		((nameMorph width // 2) @ (self handleSize negated // 2 - 1)).
	nameMorph position: (namePosition min: w viewBox bottomRight - nameMorph extent y + 2).
	nameMorph balloonTextSelector: #objectNameInHalo.
	self addMorph: nameMorph.
	^ nameMorph! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 1/27/2000 16:23'!
addPaintBgdHandle: haloSpec
	(innerTarget isKindOf: PasteUpMorph) ifTrue:
		[self addHandle: haloSpec
				on: #mouseDown send: #paintBackground to: innerTarget].
! !

!HaloMorph methodsFor: 'handles' stamp: 'ar 11/16/2002 19:24'!
addPoohHandle: handleSpec
	(innerTarget isKindOf: (Smalltalk at: #WonderlandCameraMorph ifAbsent:[nil])) ifTrue:
		[self addHandle: handleSpec on: #mouseDown send: #strokeMode to: innerTarget]
! !

!HaloMorph methodsFor: 'handles' stamp: 'RAA 3/15/2001 11:24'!
addRecolorHandle: haloSpec
	"Add a recolor handle to the receiver, if appropriate"

	| recolorHandle |

	"since this halo now opens a more general properties panel, allow it in all cases"
	"innerTarget canSetColor ifTrue:"

	recolorHandle := self addHandle: haloSpec on: #mouseUp send: #doRecolor:with: to: self.
	recolorHandle on: #mouseUp send: #doRecolor:with: to: self

! !

!HaloMorph methodsFor: 'handles' stamp: 'nk 6/12/2004 09:24'!
addRepaintHandle: haloSpec
	(innerTarget isSketchMorph) ifTrue:
		[self addHandle: haloSpec
				on: #mouseDown send: #editDrawing to: innerTarget]
! !

!HaloMorph methodsFor: 'handles' stamp: 'ar 10/25/2000 17:49'!
addRotateHandle: haloSpec
	(self addHandle: haloSpec on: #mouseDown send: #startRot:with: to: self)
		on: #mouseMove send: #doRot:with: to: self

! !

!HaloMorph methodsFor: 'handles' stamp: 'ar 10/25/2000 17:49'!
addScaleHandle: haloSpec
	target isFlexMorph ifTrue: 
		[(self addHandle: haloSpec
				on: #mouseDown send: #startScale:with: to: self)
				on: #mouseMove send: #doScale:with: to: self].
	"This or addGrowHandle:, but not both, will prevail at any one time"
! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 10/9/2000 16:57'!
addScriptHandle: haloSpec
	"If the halo's innerTarget claims it wants a Script handle, add one to the receiver, forming it as per haloSpec"

	innerTarget wantsScriptorHaloHandle ifTrue:
		[self addHandle: haloSpec
				on: #mouseUp send: #editButtonsScript to: innerTarget]
! !

!HaloMorph methodsFor: 'private' stamp: 'sw 4/27/2000 13:40'!
addSimpleHandles
	target isWorldMorph ifTrue: [^ self addHandlesForWorldHalos].
	self removeAllMorphs.  "remove old handles, if any"
	self bounds: target renderedMorph worldBoundsForHalo.  "update my size"
	self innerTarget addSimpleHandlesTo: self box: self basicBoxForSimpleHalos

! !

!HaloMorph methodsFor: 'private' stamp: 'sw 10/28/1999 15:39'!
addSimpleHandlesForWorldHalos
	"Nothing special at present here -- just use the regular handles.  Cannot rotate or resize world"

	self addHandlesForWorldHalos
! !

!HaloMorph methodsFor: 'halos and balloon help' stamp: 'nk 6/12/2004 09:34'!
addSimpleHandlesTo: aHaloMorph box: aBox
	| aHandle |
	simpleMode := true.

	target isWorldMorph ifTrue: [^ self addSimpleHandlesForWorldHalos].

	self removeAllMorphs.  "remove old handles, if any"
	
	self bounds: target renderedMorph worldBoundsForHalo.  "update my size"
	
	self addHandleAt: (((aBox topLeft + aBox leftCenter) // 2) + self simpleFudgeOffset) color: Color paleBuff icon: 'Halo-MoreHandles'
		on: #mouseDown send: #addFullHandles to: self.

	aHandle := self addGraphicalHandle: #Rotate at: aBox bottomLeft on: #mouseDown send: #startRot:with: to: self.
	aHandle on: #mouseMove send: #doRot:with: to: self.

	target isFlexMorph
		ifTrue: [(self addGraphicalHandle: #Scale at: aBox bottomRight  on: #mouseDown send: #startScale:with: to: self)
				on: #mouseMove send: #doScale:with: to: self]
		ifFalse: [(self addGraphicalHandle: #Scale at: aBox bottomRight on: #mouseDown send: #startGrow:with: to: self)
				on: #mouseMove send: #doGrow:with: to: self].

	innerTarget wantsSimpleSketchMorphHandles ifTrue:
		[self addSimpleSketchMorphHandlesInBox: aBox].

	growingOrRotating := false.
	self layoutChanged.
	self changed.
! !

!HaloMorph methodsFor: 'private' stamp: 'sw 1/27/2000 17:37'!
addSimpleSketchMorphHandlesInBox: box

	self addGraphicalHandle: #PaintTab at: box bottomCenter on: #mouseDown send: #editDrawing to: self innerTarget.

	self addDirectionHandles! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 10/2/2001 22:17'!
addTileHandle: haloSpec
	"Add the 'tear-off-tile' handle from the spec"

	self addHandle: haloSpec on: #mouseDown send: #tearOffTileForTarget:with: to: self
! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 10/2/2001 22:17'!
addViewHandle: haloSpec
	"Add the 'open viewer' handle from the halo spec"

	self addHandle: haloSpec
		on: #mouseDown send: #openViewerForTarget:with: to: self


! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 9/26/2007 11:21'!
addViewingHandle: haloSpec
	"If appropriate, add a special Viewing halo handle to the receiver.  On 26 Sept 07, we decided to eliminate this item from the UI, so the code of is method is now commented out...

	(innerTarget isKindOf: PasteUpMorph) ifTrue:
		[self addHandle: haloSpec
				on: #mouseDown send: #presentViewMenu to: innerTarget].
"
! !

!HaloMorph methodsFor: 'private' stamp: 'tak 1/10/2007 14:52'!
basicBox
	| aBox minSide anExtent w |
	minSide _ 4 * self handleSize.
	anExtent _ ((self width + self handleSize + 8) max: minSide) @
				((self height + self handleSize + 8) max: minSide).
	aBox _ Rectangle center: self center extent: anExtent.
	w _ self world ifNil:[target outermostWorldMorph].
	^ w
		ifNil:
			[aBox]
		ifNotNil:
			[aBox intersect: (w viewBox insetBy: self handleSize // 2)]
! !

!HaloMorph methodsFor: 'private' stamp: 'ar 10/7/2000 23:36'!
basicBoxForSimpleHalos
	| w |
	w := self world ifNil:[target outermostWorldMorph].
	^ (target topRendererOrSelf worldBoundsForHalo expandBy: self handleAllowanceForIconicHalos)
			intersect: (w bounds insetBy: 8@8)
! !

!HaloMorph methodsFor: 'events' stamp: 'mt 11/5/2015 10:03'!
blueButtonDown: event

	self isMagicHalo
		ifFalse: [super blueButtonDown: event]
		ifTrue: [
			self isMagicHalo: false.
			self magicAlpha: 1.0].! !

!HaloMorph methodsFor: 'updating' stamp: 'di 11/17/2001 10:56'!
changed
	"Quicker to invalidate handles individually if target is large (especially the world)"

	self extent > (200@200)
		ifTrue: [(target notNil and: [target ~~ self world]) ifTrue:
					["Invalidate 4 outer strips first, thus subsuming separate damage."
					(self fullBounds areasOutside: target bounds) do:
						[:r | self invalidRect: r]].
				self submorphsDo: [:m | m changed]]
		ifFalse: [super changed].
! !

!HaloMorph methodsFor: 'geometry testing' stamp: 'jmv 3/2/2010 16:02'!
containsPoint: aPoint 
	"This method is overridden so that, once up, the handles will stay up as long as the mouse is within the box that encloses all the handles even if it is not over any handle or over its owner."

	^target
		ifNil: [super containsPoint: aPoint] 
		ifNotNil: [false]! !

!HaloMorph methodsFor: 'events' stamp: 'ar 9/15/2000 16:54'!
containsPoint: aPoint event: anEvent
	"Blue buttons are handled by the halo"
	(anEvent isMouse and:[anEvent isMouseDown and:[anEvent blueButtonPressed]])
		ifFalse:[^super containsPoint: aPoint event: anEvent].
	^bounds containsPoint: anEvent position! !

!HaloMorph methodsFor: 'objects from disk' stamp: 'RAA 12/20/2000 17:46'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	simpleMode ifNil: [simpleMode := false].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!HaloMorph methodsFor: 'private' stamp: 'cmm 1/10/2011 15:00'!
createHandleAt: aPoint color: aColor iconName: iconName 
	| bou handle |
	bou := Rectangle center: aPoint extent: self handleSize asPoint.
	Preferences alternateHandlesLook
		ifTrue: [handle := RectangleMorph newBounds: bou color: aColor.
			handle useRoundedCorners.
			self setColor: aColor toHandle: handle]
		ifFalse: [handle := EllipseMorph newBounds: bou color: aColor].
	handle borderWidth: 0;
		 wantsYellowButtonMenu: false.
	""
	iconName isNil
		ifFalse: [| form |
			form := ScriptingSystem formAtKey: iconName.
			form isNil
				ifFalse: [| image |
					image := ImageMorph new.
					image image: form.
					image color: aColor makeForegroundColor.
					image lock.
					handle addMorphCentered: image]].
	""
	^ handle! !

!HaloMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:28'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color
		r: 0.6
		g: 0.8
		b: 1.0! !

!HaloMorph methodsFor: 'submorphs-add/remove' stamp: 'mt 11/5/2015 10:07'!
delete
	"Delete the halo.  Tell the target that it no longer has the halo; accept any pending edits to the name; and then either actually delete myself or start to fade out"

	self acceptNameEdit.
	self isMagicHalo: false.
	
	Preferences haloTransitions
		ifFalse: [super delete]
		ifTrue: [
			self
				stopStepping;
				startStepping;
				startSteppingSelector: #fadeOutFinally].
		! !

!HaloMorph methodsFor: 'private' stamp: 'di 9/26/2000 15:16'!
directionArrowLength
	^ 25! !

!HaloMorph methodsFor: 'private' stamp: 'tpr 12/31/2013 15:56'!
dismiss
	"Remove my target from the world Seems to be EToy specific."

	| w |
	w := self world.
	w ifNotNil: [w stopStepping: target].
	self delete.
	target dismissViaHalo! !

!HaloMorph methodsFor: 'private' stamp: 'sw 10/27/2002 09:27'!
doDebug: evt with: menuHandle
	"Ask hand to invoke the a debugging menu for my inner target.  If shift key is down, immediately put up an inspector on the inner target"

	| menu |
	self obtainHaloForEvent: evt andRemoveAllHandlesBut: nil.
	self world displayWorld.
	evt shiftPressed ifTrue: 
		[self delete.
		^ innerTarget inspectInMorphic: evt].

	menu := innerTarget buildDebugMenu: evt hand.
	menu addTitle: (innerTarget externalName truncateWithElipsisTo: 40).
	menu popUpEvent: evt in: self world! !

!HaloMorph methodsFor: 'private' stamp: 'sw 8/27/2008 13:51'!
doDirection: anEvent with: directionHandle
	"The mouse went down on the forward-direction halo handle; respond appropriately."

	anEvent hand obtainHalo: self.
	anEvent shiftPressed
		ifTrue:
			[directionArrowAnchor _ (target point: target referencePosition in: self world) rounded.
			self positionDirectionShaft: directionHandle.
			self removeAllHandlesBut: directionHandle.
			directionHandle setProperty: #trackDirectionArrow toValue: true]
		 ifFalse:
			[ActiveHand spawnBalloonFor: directionHandle]! !

!HaloMorph methodsFor: 'private' stamp: 'tk 7/14/2001 11:04'!
doDrag: evt with: dragHandle
	| thePoint |
	evt hand obtainHalo: self.
	thePoint := target point: evt position - positionOffset from: owner.
	target setConstrainedPosition:(target griddedPoint: thePoint) hangOut: true.
! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 3/30/2005 04:26'!
doDupOrMakeSibling: evt with: dupHandle
	"Ask hand to duplicate my target, if shift key *not* pressed, or make a sibling if shift key *is* pressed"

	^ (evt shiftPressed and: [target couldMakeSibling])
		ifTrue:
			[dupHandle color: Color green muchDarker.
			self doMakeSibling: evt with: dupHandle]
		ifFalse:
			[self doDup: evt with: dupHandle]! !

!HaloMorph methodsFor: 'private' stamp: 'nk 9/4/2004 17:38'!
doDup: evt with: dupHandle
	"Ask hand to duplicate my target."

	(target isSelectionMorph) ifTrue:
		[^ target doDup: evt fromHalo: self handle: dupHandle].

	self obtainHaloForEvent: evt andRemoveAllHandlesBut: dupHandle.
	self setTarget: (target duplicateMorph: evt).
	evt hand grabMorph: target.
	self step. "update position if necessary"
	evt hand addMouseListener: self. "Listen for the drop"! !

!HaloMorph methodsFor: 'private' stamp: 'sw 10/2/2001 22:35'!
doGrab: evt with: grabHandle
	"Ask hand to grab my target."

	self obtainHaloForEvent: evt andRemoveAllHandlesBut: grabHandle.
	evt hand grabMorph: target.
	self step. "update position if necessary"
	evt hand addMouseListener: self. "Listen for the drop"! !

!HaloMorph methodsFor: 'private' stamp: 'ul 12/12/2009 14:11'!
doGrow: evt with: growHandle
	"Called while the mouse is down in the grow handle"

	| newExtent extentToUse scale |
	evt hand obtainHalo: self.
	newExtent := (target pointFromWorld: (target griddedPoint: evt cursorPoint - positionOffset))
								- target topLeft.
	evt shiftPressed ifTrue: [
		scale := (newExtent x / (originalExtent x max: 1)) min:
					(newExtent y / (originalExtent y max: 1)).
		newExtent := (originalExtent x * scale) asInteger @ (originalExtent y * scale) asInteger
	].
	(newExtent x < 1 or: [newExtent y < 1 ]) ifTrue: [^ self].
	target renderedMorph setExtentFromHalo: (extentToUse := newExtent).
	growHandle position: evt cursorPoint - (growHandle extent // 2).
	self layoutChanged.
	(self valueOfProperty: #commandInProgress) ifNotNil:  
		[:cmd | "Update the final extent"
			cmd redoTarget: target renderedMorph selector: #setFlexExtentFromHalo: argument: extentToUse]
! !

!HaloMorph methodsFor: 'handles' stamp: 'sw 3/30/2005 04:28'!
doMakeSiblingOrDup: evt with: dupHandle
	"Ask hand to duplicate my target, if shift key *is* pressed, or make a sibling if shift key *not* pressed"

	^ (evt shiftPressed or: [target couldMakeSibling not])
		ifFalse:
			[self doMakeSibling: evt with: dupHandle]
		ifTrue:
			[dupHandle color: Color green.
			self doDup: evt with: dupHandle]! !

!HaloMorph methodsFor: 'private' stamp: 'sw 3/30/2005 04:31'!
doMakeSibling: evt with: dupHandle
	"Ask hand to make a sibling of my target.  Only reachable if target is of a uniclass"

	target couldMakeSibling ifFalse: [^ self].

	target assuredPlayer assureUniClass.
	self obtainHaloForEvent: evt andRemoveAllHandlesBut: dupHandle.
	self setTarget: (target makeSiblings: 1) first.
	evt hand grabMorph: target.
	self step. "update position if necessary"
	evt hand addMouseListener: self. "Listen for the drop"! !

!HaloMorph methodsFor: 'private' stamp: 'ar 11/29/2001 20:01'!
doMenu: evt with: menuHandle
	"Ask hand to invoke the halo menu for my inner target."

	| menu |
	self obtainHaloForEvent: evt andRemoveAllHandlesBut: nil.
	self world displayWorld.
	menu := innerTarget buildHandleMenu: evt hand.
	innerTarget addTitleForHaloMenu: menu.
	menu popUpEvent: evt in: self world.
! !

!HaloMorph methodsFor: 'private' stamp: 'sw 9/20/2001 00:16'!
doRecolor: evt with: aHandle
	"The mouse went down in the 'recolor' halo handle.  Allow the user to change the color of the innerTarget"

	evt hand obtainHalo: self.
	(aHandle containsPoint: evt cursorPoint)
		ifFalse:  "only do it if mouse still in handle on mouse up"
			[self delete.
			target addHalo: evt]
		ifTrue:
			[(Preferences propertySheetFromHalo == evt shiftPressed)
				ifFalse:	[innerTarget openAPropertySheet]
				ifTrue:	[innerTarget changeColor].
			self showingDirectionHandles ifTrue: [self addHandles]]! !

!HaloMorph methodsFor: 'dragging or resizing' stamp: 'mt 11/5/2015 10:10'!
doResizeTarget: evt

	| oldExtent newExtent newPosition |
	newExtent := originalExtent + (evt position - positionOffset * 2).
	
	(newExtent x > 1 and: [newExtent y > 1])
		ifTrue: [
			oldExtent := target extent.
			target setExtentFromHalo: (newExtent min: owner extent).
			newPosition := target position - (target extent - oldExtent // 2).
			newPosition := (newPosition x min: owner extent x - newExtent x max: 0) @ (newPosition y min: owner extent y - newExtent y max: 0).
			target setConstrainedPosition: newPosition hangOut: true].

	self bounds: self target worldBoundsForHalo.! !

!HaloMorph methodsFor: 'private' stamp: 'ul 12/12/2009 14:11'!
doRot: evt with: rotHandle
	"Update the rotation of my target if it is rotatable.  Keep the relevant command object up to date."

	| degrees |
	evt hand obtainHalo: self.
	degrees := (evt cursorPoint - (target pointInWorld: target referencePosition)) degrees.
	degrees := degrees - angleOffset degrees.
	degrees := degrees detentBy: 10.0 atMultiplesOf: 90.0 snap: false.
	degrees = 0.0
		ifTrue: [self setColor: Color lightBlue toHandle: rotHandle]
		ifFalse: [self setColor: Color blue toHandle: rotHandle].
	rotHandle submorphsDo:
		[:m | m color: rotHandle color makeForegroundColor].
	self removeAllHandlesBut: rotHandle.
	self showingDirectionHandles ifFalse:
		[self showDirectionHandles: true addHandles: false].
	self addDirectionHandles.

	target rotationDegrees: degrees.

	rotHandle position: evt cursorPoint - (rotHandle extent // 2).
	(self valueOfProperty: #commandInProgress) ifNotNil:
		[:cmd | "Update the final rotation"
		cmd redoTarget: target renderedMorph selector: #heading: argument: degrees].
	self layoutChanged! !

!HaloMorph methodsFor: 'private' stamp: 'ul 12/12/2009 14:11'!
doScale: evt with: scaleHandle
	"Update the scale of my target if it is scalable."
	| newHandlePos colorToUse |
	evt hand obtainHalo: self.
	newHandlePos := evt cursorPoint - (scaleHandle extent // 2).
	target scaleToMatch: newHandlePos.
	colorToUse := target scale = 1.0
						ifTrue: [Color yellow]
						ifFalse: [Color orange].
	self setColor: colorToUse toHandle: scaleHandle.
	scaleHandle
		submorphsDo: [:m | m color: colorToUse makeForegroundColor].
	scaleHandle position: newHandlePos.
	self layoutChanged.

	(self valueOfProperty: #commandInProgress) ifNotNil:[:cmd |
		"Update the final extent"
		cmd redoTarget: target renderedMorph selector: #setFlexExtentFromHalo: argument: target extent
	].
! !

!HaloMorph methodsFor: 'drawing' stamp: 'ar 8/8/2001 15:13'!
drawSubmorphsOn: aCanvas
	| alpha |
	((alpha := self magicAlpha) = 1.0)
		ifTrue:[^super drawSubmorphsOn: aCanvas].
	^super drawSubmorphsOn: (aCanvas asAlphaBlendingCanvas: alpha)! !

!HaloMorph methodsFor: 'private' stamp: 'ul 12/12/2009 14:10'!
endInteraction
	"Clean up after a user interaction with the a halo control"

	| m |
	self isMagicHalo: false.	"no longer"
	self magicAlpha: 1.0.
	(target isInWorld not or: [owner isNil]) ifTrue: [^self].
	[target isFlexMorph and: [target hasNoScaleOrRotation]] whileTrue: 
			[m := target firstSubmorph.
			target removeFlexShell.
			target := m].
	self isInWorld 
		ifTrue: 
			["make sure handles show in front, even if flex shell added"

			self comeToFront.
			self addHandles].
	(self valueOfProperty: #commandInProgress) ifNotNil: 
			[:cmd | 
			self rememberCommand: cmd.
			self removeProperty: #commandInProgress]! !

!HaloMorph methodsFor: 'stepping' stamp: 'ar 8/8/2001 14:56'!
fadeIn
	self magicAlpha >= 1.0 ifTrue:[self stopSteppingSelector: #fadeIn].
	self magicAlpha: ((self magicAlpha + 0.1) min: 1.0)
! !

!HaloMorph methodsFor: 'stepping' stamp: 'ar 8/8/2001 15:44'!
fadeInInitially
	| max |
	max := self isMagicHalo ifTrue:[0.3] ifFalse:[1.0].
	self magicAlpha >= max ifTrue:[self stopSteppingSelector: #fadeInInitially].
	self magicAlpha: ((self magicAlpha + (max * 0.1)) min: max)
! !

!HaloMorph methodsFor: 'stepping' stamp: 'ar 8/8/2001 14:57'!
fadeOut
	self magicAlpha <= 0.3 ifTrue:[self stopSteppingSelector: #fadeOut].
	self magicAlpha: ((self magicAlpha - 0.1) max: 0.3)
! !

!HaloMorph methodsFor: 'stepping' stamp: 'ar 8/8/2001 15:46'!
fadeOutFinally
	self magicAlpha <= 0.05 ifTrue:[^super delete].
	self magicAlpha <= 0.3 ifTrue:[
		^self magicAlpha: (self magicAlpha - 0.03 max: 0.0)].
	self magicAlpha: ((self magicAlpha * 0.5) max: 0.0)
! !

!HaloMorph methodsFor: 'accessing' stamp: 'sw 1/26/2000 15:36'!
haloBox: aBox
	haloBox := aBox! !

!HaloMorph methodsFor: 'private' stamp: 'sw 1/27/2000 18:42'!
handleAllowanceForIconicHalos
	^ 12! !

!HaloMorph methodsFor: 'stepping' stamp: 'ar 8/8/2001 15:38'!
handleEntered
	self isMagicHalo ifFalse:[^self].
	self stopStepping; startStepping.
	self startSteppingSelector: #fadeIn.
! !

!HaloMorph methodsFor: 'stepping' stamp: 'ar 8/8/2001 15:38'!
handleLeft
	self isMagicHalo ifFalse:[^self].
	self stopStepping; startStepping.
	self startSteppingSelector: #fadeOut.! !

!HaloMorph methodsFor: 'events' stamp: 'nk 6/26/2002 07:19'!
handleListenEvent: anEvent
	"We listen for possible drop events here to add back those handles after a dup/grab operation"

	(anEvent isMouse and:[anEvent isMove not]) ifFalse:[^ self]. "not interested"
	anEvent hand removeMouseListener: self. "done listening"
	(self world ifNil: [target world]) ifNil: [^ self].
	self addHandles  "and get those handles back"! !

!HaloMorph methodsFor: 'private' stamp: 'tak 1/8/2007 19:42'!
handleSize
	^ Preferences biggerHandles
		ifTrue: [30]
		ifFalse: [16]! !

!HaloMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 14:29'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	growingOrRotating := false.
	simpleMode := Preferences simpleHalosInForce ! !

!HaloMorph methodsFor: 'accessing' stamp: 'jm 5/22/1998 16:28'!
innerTarget

	^ innerTarget
! !

!HaloMorph methodsFor: 'accessing' stamp: 'ar 8/8/2001 14:35'!
isMagicHalo
	^self valueOfProperty: #isMagicHalo ifAbsent:[false].! !

!HaloMorph methodsFor: 'accessing' stamp: 'ar 8/8/2001 15:37'!
isMagicHalo: aBool
	self setProperty: #isMagicHalo toValue: aBool.
	aBool ifFalse:[
		"Reset everything"
		self stopStepping. "get rid of all"
		self startStepping. "only those of interest"
	].! !

!HaloMorph methodsFor: 'stepping' stamp: 'nk 6/27/2003 12:28'!
localHaloBoundsFor: aMorph

	"aMorph may be in the hand and perhaps not in our world"

	| r |

	r := aMorph worldBoundsForHalo truncated.
	aMorph world = self world ifFalse: [^r].
	^((self transformFromOutermostWorld) globalBoundsToLocal: r) truncated! !

!HaloMorph methodsFor: 'accessing' stamp: 'ar 8/8/2001 14:28'!
magicAlpha
	^self valueOfProperty: #magicAlpha ifAbsent:[1.0]! !

!HaloMorph methodsFor: 'accessing' stamp: 'ar 8/8/2001 14:42'!
magicAlpha: alpha
	self setProperty: #magicAlpha toValue: alpha.
	self changed.! !

!HaloMorph methodsFor: 'private' stamp: 'aoy 2/15/2003 21:10'!
maybeCollapse: evt with: collapseHandle 
	"Ask hand to collapse my target if mouse comes up in it."

	evt hand obtainHalo: self.
	self delete.
	(collapseHandle containsPoint: evt cursorPoint) 
		ifFalse: 
			[
			target addHalo: evt]
		ifTrue: 
			[
			target collapse]! !

!HaloMorph methodsFor: 'private' stamp: 'ar 8/6/2009 18:23'!
maybeDismiss: evt with: dismissHandle
	"Ask hand to dismiss my target if mouse comes up in it."

	evt hand obtainHalo: self.
	(dismissHandle containsPoint: evt cursorPoint)
		ifFalse:
			[self delete.
			target addHalo: evt]
		ifTrue:
			[target resistsRemoval ifTrue:
				[(UIManager default chooseFrom: {
					'Yes' translated.
					'Um, no, let me reconsider' translated.
				} title: 'Really throw this away' translated) = 1 ifFalse: [^ self]].
			evt hand removeHalo.
			self delete.
			target dismissViaHalo.
			ActiveWorld presenter flushPlayerListCache]! !

!HaloMorph methodsFor: 'private' stamp: 'ar 10/24/2000 18:42'!
maybeDoDup: evt with: dupHandle
	evt hand obtainHalo: self.
	^ target okayToDuplicate ifTrue:
		[self doDup: evt with: dupHandle]! !

!HaloMorph methodsFor: 'WiW support' stamp: 'RAA 6/27/2000 19:12'!
morphicLayerNumber

	"helpful for insuring some morphs always appear in front of or behind others.
	smaller numbers are in front"

	^7		"Halos are very front-like things"! !

!HaloMorph methodsFor: 'private' stamp: 'sw 10/3/2001 00:21'!
mouseDownInCollapseHandle: evt with: collapseHandle
	"The mouse went down in the collapse handle; collapse the morph"

	self obtainHaloForEvent: evt andRemoveAllHandlesBut: collapseHandle.
	self setDismissColor: evt with: collapseHandle! !

!HaloMorph methodsFor: 'private' stamp: 'HenrikSperreJohansen 1/21/2011 18:34'!
mouseDownInDimissHandle: evt with: dismissHandle
	evt hand obtainHalo: self.
	SoundService soundEnabled ifTrue: [TrashCanMorph playMouseEnterSound].
	self removeAllHandlesBut: dismissHandle.
	self setColor: Color darkGray toHandle: dismissHandle.
! !

!HaloMorph methodsFor: 'events' stamp: 'mt 11/5/2015 10:10'!
mouseMove: evt
	
	growingOrRotating
		ifTrue: [self doResizeTarget: evt]
		ifFalse: [self doDragTarget: evt].! !

!HaloMorph methodsFor: 'private' stamp: 'sw 10/2/2001 22:16'!
obtainHaloForEvent: evt andRemoveAllHandlesBut: aHandle
	"Make sure the event's hand correlates with the receiver, and remove all handles except the given one.  If nil is provided as the handles argument, the result is that all handles are removed.  Note that any pending edits to the name-string in the halo are accepted at this time."

	evt hand obtainHalo: self.
	self acceptNameEdit.
	self removeAllHandlesBut: aHandle! !

!HaloMorph methodsFor: 'handles' stamp: 'wiz 3/13/2005 13:53'!
openViewerForTarget: evt with: aHandle 
	"Open a viewer for my inner target or if shift pressed make a snapshot of morph."
	self obtainHaloForEvent: evt andRemoveAllHandlesBut: nil.
	evt shiftPressed
		ifTrue: [target duplicateMorphImage: evt]
		ifFalse: [innerTarget openViewerForArgument]! !

!HaloMorph methodsFor: 'pop up' stamp: 'mt 11/5/2015 10:20'!
popUpFor: morph at: position hand: hand
 
	super popUpFor: morph at: position hand: hand.
	
	self startStepping.
	(Preferences haloTransitions or: [self isMagicHalo]) 
		ifTrue: [
			self magicAlpha: 0.0.
			self startSteppingSelector: #fadeInInitially].! !

!HaloMorph methodsFor: 'pop up' stamp: 'mt 11/5/2015 10:25'!
popUpMagicallyFor: aMorph hand: aHand
	"Programatically pop up a halo for a given hand."

	super
		popUpMagicallyFor: aMorph
		hand: aHand.
		
	Preferences magicHalos
		ifTrue: [self isMagicHalo: true].
	(Preferences haloTransitions not and: [self isMagicHalo])
		ifTrue: [self magicAlpha: 0.2].
! !

!HaloMorph methodsFor: 'private' stamp: 'di 9/26/2000 15:12'!
positionDirectionShaft: shaft
	"Position the shaft."
	| alphaRadians unitVector |
	"Pretty crude and slow approach at present, but a stake in the ground"
	alphaRadians := target heading degreesToRadians.
	unitVector := alphaRadians sin  @ alphaRadians cos negated.
	shaft setVertices: {unitVector * 6 + directionArrowAnchor.  "6 = radius of deadeye circle"
					unitVector * self directionArrowLength + directionArrowAnchor}
! !

!HaloMorph methodsFor: 'handles' stamp: 'ar 1/30/2001 23:32'!
positionIn: aBox horizontalPlacement: horiz verticalPlacement: vert
	| xCoord yCoord |

	horiz == #left
		ifTrue:	[xCoord := aBox left].
	horiz == #leftCenter
		ifTrue:	[xCoord := aBox left + (aBox width // 4)].
	horiz == #center
		ifTrue:	[xCoord := (aBox left + aBox right) // 2].
	horiz == #rightCenter
		ifTrue:	[xCoord := aBox left + ((3 * aBox width) // 4)].
	horiz == #right
		ifTrue:	[xCoord := aBox right].

	vert == #top
		ifTrue:	[yCoord := aBox top].
	vert == #topCenter
		ifTrue:	[yCoord := aBox top + (aBox height // 4)].
	vert == #center
		ifTrue:	[yCoord := (aBox top + aBox bottom) // 2].
	vert == #bottomCenter
		ifTrue:	[yCoord := aBox top + ((3 * aBox height) // 4)].
	vert == #bottom
		ifTrue:	[yCoord := aBox bottom].

	^ xCoord asInteger @ yCoord asInteger! !

!HaloMorph methodsFor: 'geometry' stamp: 'di 9/26/2000 21:03'!
position: pos
	"Halos display imprefectly if their coordinates are non-integral
		-- especially the direction handles."

	^ super position: pos asIntegerPoint! !

!HaloMorph methodsFor: 'private' stamp: 'sw 8/27/2008 02:17'!
prepareToTrackCenterOfRotation: evt with: rotationHandle
	"The mouse went down on the center of rotation."

	evt hand obtainHalo: self.
	evt shiftPressed
		ifTrue:
			[self removeAllHandlesBut: rotationHandle.
			rotationHandle setProperty: #trackCenterOfRotation toValue: true.
			evt hand showTemporaryCursor: Cursor blank]
		ifFalse:
			[ActiveHand spawnBalloonFor: rotationHandle]! !

!HaloMorph methodsFor: 'private' stamp: 'dgd 9/10/2004 13:38'!
removeAllHandlesBut: h
	"Remove all handles except h."
	(Preferences maintainHalos and:[h isNil])
		ifTrue:[self removeHalo]
		ifFalse:[
			submorphs copy do:
				[:m | m == h ifFalse: [m delete]]
		].
! !

!HaloMorph methodsFor: 'private' stamp: 'sw 8/27/2008 02:24'!
setCenterOfRotation: evt with: rotationHandle
	| localPt |
	evt hand obtainHalo: self.
	evt hand showTemporaryCursor: nil.
	(rotationHandle hasProperty: #trackCenterOfRotation) ifTrue:
		[localPt  :=  innerTarget transformFromWorld globalPointToLocal: rotationHandle center.
		innerTarget setRotationCenterFrom: localPt].
	
	rotationHandle removeProperty: #trackCenterOfRotation.
	self endInteraction! !

!HaloMorph methodsFor: 'private' stamp: 'dgd 9/29/2004 19:51'!
setColor: aColor toHandle: aHandle 
	"private - change the color to the given handle, applying the 
	alternate look if corresponds"
	aHandle color: aColor.
	Preferences alternateHandlesLook
		ifTrue: [| fill | 
			fill := GradientFillStyle ramp: {0.0 -> aColor muchLighter. 1.0 -> aColor darker}.
			fill origin: aHandle topLeft.
			fill direction: aHandle extent.
			aHandle fillStyle: fill] ! !

!HaloMorph methodsFor: 'private' stamp: 'sw 8/27/2008 13:31'!
setDirection: anEvent with: directionHandle
	"The user has let up after having dragged the direction arrow; now set the forward direction of the actual SketchMorph accordingly"
	(directionHandle hasProperty: #trackDirectionArrow) ifTrue:
		[anEvent hand obtainHalo: self.
		target setDirectionFrom: directionHandle center.
		directionHandle removeProperty: #trackDirectionArrow.
		self endInteraction]! !

!HaloMorph methodsFor: 'private' stamp: 'dgd 9/29/2004 19:56'!
setDismissColor: evt with: dismissHandle
	"Called on mouseStillDown in the dismiss handle; set the color appropriately."

	| colorToUse |
	evt hand obtainHalo: self.
	colorToUse :=  (dismissHandle containsPoint: evt cursorPoint)
		ifFalse:
			[Color red muchLighter]
		ifTrue:
			[Color lightGray].
	self setColor: colorToUse toHandle: dismissHandle.
! !

!HaloMorph methodsFor: 'accessing' stamp: 'nk 6/12/2004 21:56'!
setTarget: aMorph
	"Private!! Set the target without adding handles."

	target := aMorph topRendererOrSelf.
	innerTarget := target renderedMorph.
	innerTarget wantsDirectionHandles
		ifTrue: [self showDirectionHandles: true addHandles: false].
	target hasHalo: true.

! !

!HaloMorph methodsFor: 'private' stamp: 'di 9/26/2000 15:25'!
showDirectionHandles: wantToShow

	self showDirectionHandles: wantToShow addHandles: true  "called from menu"
! !

!HaloMorph methodsFor: 'private' stamp: 'aoy 2/17/2003 01:27'!
showDirectionHandles: wantToShow addHandles: needHandles 
	directionArrowAnchor := wantToShow 
				ifTrue: [target referencePositionInWorld	"not nil means show"]
				ifFalse: [nil].
	needHandles ifTrue: [self addHandles] ! !

!HaloMorph methodsFor: 'private' stamp: 'dgd 2/22/2003 19:04'!
showingDirectionHandles
	^directionArrowAnchor notNil! !

!HaloMorph methodsFor: 'private' stamp: 'sw 1/27/2000 18:43'!
simpleFudgeOffset
	"account for the difference in basicBoxes between regular and simple handles"

	^ 0@0
! !

!HaloMorph methodsFor: 'dragging or resizing' stamp: 'mt 11/5/2015 10:12'!
startDragTarget: event

	event controlKeyPressed
		ifTrue: [self startResizeTarget: event]
		ifFalse: [
			growingOrRotating := false.
			super startDragTarget: event].! !

!HaloMorph methodsFor: 'private' stamp: 'sw 10/2/2001 22:09'!
startDrag: evt with: dragHandle
	"Drag my target without removing it from its owner."

	| itsOwner |
	self obtainHaloForEvent: evt andRemoveAllHandlesBut: dragHandle.
	positionOffset := dragHandle center - (target point: target position in: owner).

	 ((itsOwner := target topRendererOrSelf owner) notNil and:
			[itsOwner automaticViewing]) ifTrue:
				[target openViewerForArgument]! !

!HaloMorph methodsFor: 'private' stamp: 'sw 2/2/2006 02:46'!
startGrow: evt with: growHandle
	"Initialize resizing of my target.  Launch a command representing it, to support Undo"

	| botRt |
	self obtainHaloForEvent: evt andRemoveAllHandlesBut: growHandle.
	botRt := target point: target bottomRight in: owner.
	positionOffset := (self world viewBox containsPoint: botRt)
		ifTrue: [evt cursorPoint - botRt]
		ifFalse: [0@0].

	self setProperty: #commandInProgress toValue:
		(Command new
			cmdWording: ('resize ' translated, target nameForUndoWording);
			undoTarget: target renderedMorph selector: #setFlexExtentFromHalo: argument: target extent).

	originalExtent := target extent! !

!HaloMorph methodsFor: 'dragging or resizing' stamp: 'mt 11/5/2015 10:12'!
startResizeTarget: event
	"Begin resizing the target"
	growingOrRotating := true.
	positionOffset := event position.
	originalExtent := target extent.
	self removeAllHandlesBut: nil.
	event hand newMouseFocus: self.
	event hand addMouseListener: self. "add handles back on mouse-up"! !

!HaloMorph methodsFor: 'private' stamp: 'sw 2/2/2006 01:31'!
startRot: evt with: rotHandle
	"Initialize rotation of my target if it is rotatable.  Launch a command object to represent the action"

	self obtainHaloForEvent: evt andRemoveAllHandlesBut: rotHandle.
	target isFlexMorph ifFalse: 
		[target isInWorld ifFalse: [self setTarget: target player costume].
		target addFlexShellIfNecessary].
	growingOrRotating := true.

	self removeAllHandlesBut: rotHandle.  "remove all other handles"
	angleOffset := evt cursorPoint - (target pointInWorld: target referencePosition).
	angleOffset := Point
			r: angleOffset r
			degrees: angleOffset degrees - target rotationDegrees.
	self setProperty: #commandInProgress toValue:
		(Command new
			cmdWording: ('rotate ' translated, target nameForUndoWording);
			undoTarget: target renderedMorph selector: #heading: argument: target rotationDegrees)

! !

!HaloMorph methodsFor: 'private' stamp: 'sw 2/2/2006 00:28'!
startScale: evt with: scaleHandle
	"Initialize scaling of my target."

	self obtainHaloForEvent: evt andRemoveAllHandlesBut: scaleHandle.
	target isFlexMorph ifFalse: [target addFlexShellIfNecessary].
	growingOrRotating := true.
	positionOffset := 0@0.

	self setProperty: #commandInProgress toValue:
		(Command new
			cmdWording: ('resize ' translated, target nameForUndoWording);
			undoTarget: target renderedMorph selector: #setFlexExtentFromHalo: argument: target extent).
	originalExtent := target extent
! !

!HaloMorph methodsFor: 'stepping' stamp: 'nk 6/27/2003 12:32'!
step
	| newBounds |
	target
		ifNil: [^ self].
	newBounds := target isWorldMorph
				ifTrue: [target bounds]
				ifFalse: [self localHaloBoundsFor: target renderedMorph].
	newBounds = self bounds
		ifTrue: [^ self].
	newBounds extent = self bounds extent
		ifTrue: [^ self position: newBounds origin].
	growingOrRotating
		ifFalse: [submorphs size > 1
				ifTrue: [self addHandles]].
	"adjust halo bounds if appropriate"
	self bounds: newBounds! !

!HaloMorph methodsFor: 'stepping' stamp: 'jm 7/16/97 06:54'!
stepTime

	^ 0  "every cycle"
! !

!HaloMorph methodsFor: 'accessing' stamp: 'jm 5/7/1998 15:42'!
target: aMorph

	self setTarget: aMorph.
	target ifNotNil: [self addHandles].
! !

!HaloMorph methodsFor: 'handles' stamp: 'ar 1/4/2010 00:28'!
tearOffTileForTarget: evt with: aHandle
	"Tear off a tile representing my inner target.  If shift key is down, open up an instance browser on the morph itself, not the player, with tiles showing, instead"

	self obtainHaloForEvent: evt andRemoveAllHandlesBut: nil.
	innerTarget tearOffTile! !

!HaloMorph methodsFor: 'private' stamp: 'sw 8/27/2008 02:16'!
trackCenterOfRotation: anEvent with: rotationHandle
	(rotationHandle hasProperty: #dragByCenterOfRotation) 
		ifTrue:[^self doDrag: anEvent with: rotationHandle].
	(rotationHandle hasProperty: #trackCenterOfRotation)
		ifTrue:
			[anEvent hand obtainHalo: self.
			rotationHandle center: anEvent cursorPoint]! !

!HaloMorph methodsFor: 'private' stamp: 'sw 8/27/2008 13:29'!
trackDirectionArrow: anEvent with: shaft
	(shaft hasProperty: #trackDirectionArrow) ifTrue:
		[anEvent hand obtainHalo: self.
		shaft setVertices: {directionArrowAnchor. anEvent cursorPoint}.
		self layoutChanged]! !

!HaloMorph methodsFor: 'copying' stamp: 'tk 1/7/1999 17:37'!
veryDeepFixupWith: deepCopier
	"If target and arguments fields were weakly copied, fix them here.  If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

super veryDeepFixupWith: deepCopier.
target := deepCopier references at: target ifAbsent: [target].
innerTarget := deepCopier references at: innerTarget ifAbsent: [innerTarget].
! !

!HaloMorph methodsFor: 'copying' stamp: 'st 9/14/2004 13:03'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all, but shared.  	Warning!!!!  Every instance variable defined in this class must be handled.  We must also implement veryDeepFixupWith:.  See DeepCopier class comment."

	super veryDeepInner: deepCopier.
	"target := target.		Weakly copied"
	"innerTarget := innerTarget.		Weakly copied"
	positionOffset := positionOffset veryDeepCopyWith: deepCopier.
	angleOffset := angleOffset veryDeepCopyWith: deepCopier.
	growingOrRotating := growingOrRotating veryDeepCopyWith: deepCopier.
	directionArrowAnchor := directionArrowAnchor.
	simpleMode := simpleMode.
	haloBox := haloBox.
	originalExtent := originalExtent
! !

!HaloMorph methodsFor: 'events' stamp: 'sw 5/21/1998 15:41'!
wantsKeyboardFocusFor: aSubmorph
	"to allow the name to be edited in the halo in the old tty way; when we morphic-text-ize the name editing, presumably this method should be removed"
	^ true! !

!HaloMorph methodsFor: 'accessing' stamp: 'dgd 9/9/2004 22:55'!
wantsToBeTopmost
	"Answer if the receiver want to be one of the topmost objects in 
	its owner"
	^ true! !

!HaloMorph methodsFor: 'menu' stamp: 'dgd 9/20/2004 19:35'!
wantsYellowButtonMenu
	"Answer true if the receiver wants a yellow button menu"
	^ false! !

!SimpleHaloMorph methodsFor: 'construction' stamp: 'mt 11/5/2015 09:50'!
addHandles
	"This is an example for handles."
	
	self addMorphFront: (IconicButton new
		color: Color red muchLighter;
		borderColor: Color red;
		labelGraphic: MenuIcons smallCancelIcon;
		target: self target;
		actionSelector: #delete;
		bottomRight: self topLeft;
		yourself).! !

!SimpleHaloMorph methodsFor: 'events' stamp: 'mt 11/4/2015 20:58'!
blueButtonDown: event
	"Transfer the halo to the next likely recipient"
	
	self target ifNil: [^self delete].
	event hand obtainHalo: self.

	self positionOffset: (event position - (self target point: self target position in: self owner)).

	"wait for drags or transfer"
	event hand 
		waitForClicksOrDrag: self 
		event: event
		selectors: { #transferHalo:. nil. nil. #startDragTarget:. }
		threshold: HandMorph dragThreshold.! !

!SimpleHaloMorph methodsFor: 'submorphs-add/remove' stamp: 'mt 11/4/2015 20:52'!
delete

	self target hasHalo: false.
	super delete.! !

!SimpleHaloMorph methodsFor: 'dragging' stamp: 'mt 11/4/2015 20:58'!
doDragTarget: event

	self target
		setConstrainedPosition: (self target point: (event position - self positionOffset) from: self owner)
		hangOut: true.
		
	self bounds: self target worldBoundsForHalo.! !

!SimpleHaloMorph methodsFor: 'drawing' stamp: 'mt 11/4/2015 20:42'!
drawOn: aCanvas 
	"Draw this morph only if it has no target."

	(Preferences showBoundsInHalo and: [self target isWorldMorph not])
		ifTrue: [
			| boundsColor |
			boundsColor := MenuMorph menuSelectionColor
						ifNil: [Color blue].
			aCanvas
				frameAndFillRectangle: self bounds
				fillColor: Color transparent
				borderWidth: 2
				borderColor: 
					(boundsColor isTranslucent
						ifTrue: [boundsColor]
						ifFalse: [boundsColor alpha: 0.8])]! !

!SimpleHaloMorph methodsFor: 'events' stamp: 'mt 11/4/2015 20:17'!
handlerForBlueButtonDown: anEvent
	"Blue button was clicked within the receiver"
	^self! !

!SimpleHaloMorph methodsFor: 'accessing' stamp: 'mt 11/5/2015 09:58'!
innerTarget
	"If the target is merely a decorator for another morph, the inner target can be distiguished. Scroll panes, for example, could have their scrolled content as an inner target."

	^ self target! !

!SimpleHaloMorph methodsFor: 'testing' stamp: 'mt 11/4/2015 20:13'!
isMagicHalo

	^ false! !

!SimpleHaloMorph methodsFor: 'events' stamp: 'mt 11/4/2015 20:59'!
mouseMove: event

	self doDragTarget: event.! !

!SimpleHaloMorph methodsFor: 'pop up' stamp: 'mt 11/4/2015 20:05'!
popUpFor: morph

	self
		popUpFor: morph
		hand: (morph world activeHand ifNil: [morph world primaryHand]).! !

!SimpleHaloMorph methodsFor: 'pop up' stamp: 'mt 11/5/2015 10:20'!
popUpFor: morph at: position hand: hand

	self target: morph.
	
	hand halo: self.
	hand world addMorphFront: self.
	
	self positionOffset: position - (morph point: morph position in: self owner).! !

!SimpleHaloMorph methodsFor: 'pop up' stamp: 'mt 11/4/2015 20:22'!
popUpFor: morph event: event

	self
		popUpFor: morph
		at: event position
		hand: event hand.! !

!SimpleHaloMorph methodsFor: 'pop up' stamp: 'jmg 11/27/2015 23:23'!
popUpFor: morph hand: hand

	self
		popUpFor: morph
		at: (hand lastEvent transformedBy: (morph transformedFrom: nil)) position
		hand: hand! !

!SimpleHaloMorph methodsFor: 'pop up' stamp: 'mt 11/4/2015 20:23'!
popUpMagicallyFor: morph hand: hand

	self
		popUpFor: morph
		hand: hand.! !

!SimpleHaloMorph methodsFor: 'accessing' stamp: 'mt 11/4/2015 20:20'!
positionOffset

	^ positionOffset! !

!SimpleHaloMorph methodsFor: 'accessing' stamp: 'mt 11/4/2015 20:20'!
positionOffset: aPoint

	positionOffset := aPoint.! !

!SimpleHaloMorph methodsFor: 'events' stamp: 'mt 11/4/2015 20:45'!
rejectsEvent: anEvent
	"Return true to reject the given event. Rejecting an event means neither the receiver nor any of it's submorphs will be given any chance to handle it."
	(super rejectsEvent: anEvent) ifTrue:[^true].
	anEvent isDropEvent ifTrue:[^true]. "never attempt to drop on halos"
	^false! !

!SimpleHaloMorph methodsFor: 'dragging' stamp: 'mt 11/4/2015 20:58'!
startDragTarget: event

	self positionOffset: (event position - (self target point: self target position in: self owner)).		
	event hand newMouseFocus: self.! !

!SimpleHaloMorph methodsFor: 'testing' stamp: 'mt 11/4/2015 20:41'!
staysUpWhenMouseIsDownIn: aMorph
	^ ((aMorph == self target) or: [aMorph hasOwner: self])! !

!SimpleHaloMorph methodsFor: 'accessing' stamp: 'mt 11/5/2015 10:14'!
target

	^ target ifNil: [target := Morph new]! !

!SimpleHaloMorph methodsFor: 'accessing' stamp: 'mt 11/4/2015 20:29'!
target: morph

	target := morph.
	morph hasHalo: true.
	self addHandles.! !

!SimpleHaloMorph methodsFor: 'pop up' stamp: 'mt 11/5/2015 10:26'!
transferHalo: event
	"Transfer the halo to the next likely recipient"

	self target
		transferHalo: (event transformedBy: (self target transformedFrom: self))
		from: self target.! !

Morph subclass: #SketchMorph
	instanceVariableNames: 'originalForm rotationStyle scalePoint framesToDwell rotatedForm keepAspectRatio'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!SketchMorph commentStamp: '<historical>' prior: 0!
The morph that results when the user draws a color bitmap using the PaintBox (SketchEditorMorph and PaintBoxMorph).  

forwardDirection is the angle at which the object will go forward.  When the rotationStyle is not #normal, then forwardDirection is any angle, while the rotation is highly restricted.  If flexed, this is remembered by the Transform morph.  For non-normal rotationStyle, it is rotationDegrees.

setupAngle (a property) is where the user put the green arrow to indicate which direction on the picture is forward.  When #normal, draw the morph initially at (0.0 - setupAngle).  The enclosing TransformationMorph then rotates it to the true angle.
 
rotationDegrees  In a #normal object, rotationDegrees is constant an equal to setupAngle.
	For non-normal, it is the direction the object is going.

When repainting, set it back to its original state. The green arrow is set to setupAngle, and the sketch is shown as drawn originally (rotationDegrees = 0). 

rotationStyle = normal (turns), leftRight, upDown, fixed.  
When leftRight upDown or fixed, bit map has severe restrictions.
!

SketchMorph subclass: #ColorPickerMorph
	instanceVariableNames: 'selectedColor sourceHand deleteOnMouseUp updateContinuously target selector argument originalColor theSelectorDisplayMorph command isModal clickedTranslucency noChart'
	classVariableNames: 'ColorChart DragBox FeedbackBox RevertBox TransparentBox TransText'
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!ColorPickerMorph commentStamp: 'kfr 10/27/2003 16:16' prior: 0!
A gui for setting color and transparency. Behaviour can be changed with the Preference modalColorPickers.!

!ColorPickerMorph class methodsFor: 'class initialization' stamp: 'nice 12/27/2009 03:11'!
colorPaletteForDepth: depth extent: chartExtent
	"Display a palette of colors sorted horizontally by hue and vertically by lightness. Useful for eyeballing the color gamut of the display, or for choosing a color interactively."
	"Note: It is slow to build this palette, so it should be cached for quick access."
	"(Color colorPaletteForDepth: 16 extent: 190@60) display"

	| startHue palette transHt vSteps transCaption grayWidth hSteps y c x |
	palette := Form extent: chartExtent depth: depth.
	transCaption := "(DisplayText text: 'no color' asText textStyle: (TextConstants at: #ComicPlain)) form storeString"
		(Form extent: 34@9 depth: 1
			fromArray: #(0 0 256 0 256 0 3808663859 2147483648 2491688266 2147483648 2491688266 0 2491688266 0 2466486578 0 0 0)
			offset: 0@0).
	transHt := transCaption height.
	palette fillWhite: (0@0 extent: palette width@transHt).
	palette fillBlack: (0@transHt extent: palette width@1).
	transCaption displayOn: palette at: palette boundingBox topCenter - ((transCaption width // 2)@0).
	grayWidth := 10.
	startHue := 338.0.
	vSteps := palette height - transHt // 2.
	hSteps := palette width - grayWidth.
	x := 0.
	startHue to: startHue + 360.0 by: 360.0/hSteps do: [:h | | basicHue |
		basicHue := Color h: h asFloat s: 1.0 v: 1.0.
		y := transHt+1.
		0 to: vSteps do: [:n |
 			c := basicHue mixed: (n asFloat / vSteps asFloat) with: Color white.
			palette fill: (x@y extent: 1@1) fillColor: c.
			y := y + 1].
		1 to: vSteps do: [:n |
 			c := Color black mixed: (n asFloat / vSteps asFloat) with: basicHue.
			palette fill: (x@y extent: 1@1) fillColor: c.
			y := y + 1].
		x := x + 1].
	y := transHt + 1.
	1 to: vSteps * 2 do: [:n |
 		c := Color black mixed: (n asFloat / (vSteps*2) asFloat) with: Color white.
		palette fill: (x@y extent: 10@1) fillColor: c.
		y := y + 1].
	^ palette
! !

!ColorPickerMorph class methodsFor: 'class initialization' stamp: 'ar 7/8/2006 20:33'!
initialize
	"ColorPickerMorph initialize"

	ColorChart := ColorPickerMorph colorPaletteForDepth: 16 extent: 190@60.
	DragBox :=  (11@0) extent: 9@8.
	RevertBox := (ColorChart width - 20)@1 extent: 9@8.
	FeedbackBox := (ColorChart width - 10)@1 extent: 9@8.
	TransparentBox := DragBox topRight corner: RevertBox bottomLeft.

		ColorChart fillBlack: ((DragBox left - 1)@0 extent: 1@9).
		ColorChart fillBlack: ((TransparentBox left)@0 extent: 1@9).
		ColorChart fillBlack: ((FeedbackBox left - 1)@0 extent: 1@9).
		ColorChart fillBlack: ((RevertBox left - 1)@0 extent: 1@9).
		(Form dotOfSize: 5) displayOn: ColorChart at: DragBox center + (0@1).

	self localeChanged.! !

!ColorPickerMorph class methodsFor: 'class initialization' stamp: 'tak 8/4/2005 14:26'!
localeChanged
	| formTranslator |
	formTranslator := NaturalLanguageFormTranslator localeID: Locale current localeID.
	TransText := formTranslator translate: 'translucent'.
	TransText
		ifNil: [TransText := Form
						extent: 63 @ 8
						depth: 1
						fromArray: #(4194306 1024 4194306 1024 15628058 2476592640 4887714 2485462016 1883804850 2486772764 4756618 2485462016 4748474 1939416064 0 0 )
						offset: 0 @ 0].
	TransText := ColorForm mappingWhiteToTransparentFrom: TransText! !

!ColorPickerMorph class methodsFor: 'class initialization' stamp: 'ar 7/8/2006 20:33'!
noColorCaption
	| formTranslator |
	formTranslator := NaturalLanguageFormTranslator localeID: Locale current localeID.
	^ (formTranslator translate: 'no color')
		ifNil: [Form
				extent: 34 @ 9
				depth: 1
				fromArray: #(0 0 256 0 256 0 3808663859 2147483648 2491688266 2147483648 2491688266 0 2491688266 0 2466486578 0 0 0 )
				offset: 0 @ 0]
! !

!ColorPickerMorph class methodsFor: 'as yet unclassified' stamp: 'sw 10/27/1999 11:40'!
perniciousBorderColor
	"Answer the color of the border lines of a color picker; this color gets reported as you drag the mouse through from the translucent box to the true color area, for example, and can cause some difficulties in some special cases, so it is faithfully reported here in this hard-coded fashion in order that energetic clients wishing to handle it as special-case it can do so."

	^ Color r: 0.0 g: 0.0 b: 0.032! !

!ColorPickerMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 21:17'!
addCustomMenuItems: aCustomMenu hand: aHandMorph

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	deleteOnMouseUp
		ifTrue: [aCustomMenu add: 'stay up' translated action: #toggleDeleteOnMouseUp]
		ifFalse: [aCustomMenu add: 'do not stay up' translated action: #toggleDeleteOnMouseUp].
	updateContinuously
		ifTrue: [aCustomMenu add: 'update only at end' translated action: #toggleUpdateContinuously]
		ifFalse: [aCustomMenu add: 'update continuously' translated action: #toggleUpdateContinuously].
! !

!ColorPickerMorph methodsFor: 'other' stamp: 'di 11/27/1999 09:12'!
addToWorld: world near: box
	| goodLocation |
	goodLocation := self bestPositionNear: box inWorld: world.
	world allMorphsDo:
		[:p | (p isMemberOf: ColorPickerMorph) ifTrue:
		[(p ~~ self and: [p owner notNil and: [p target == target]]) ifTrue:
			[(p selector == selector and: [p argument == argument])
				ifTrue: [^ p comeToFront  "uncover existing picker"]
				ifFalse: ["place second picker relative to first"
						goodLocation := self bestPositionNear: p bounds inWorld: world]]]].
	self position: goodLocation.
	world addMorphFront: self.
	self changed
! !

!ColorPickerMorph methodsFor: 'private' stamp: 'di 9/27/2000 10:36'!
anchorAndRunModeless: aHand
	"If user clicks on the drag-dot of a modal picker,
	anchor it, and change to modeless operation."

	aHand showTemporaryCursor: nil.  "revert to normal cursor"
	self initializeModal: false; originalColor: originalColor.  "reset as modeless"
	aHand flushEvents.  "Drop any events gathered during modal loop"
	aHand position: Sensor cursorPoint; grabMorph: self.  "Slip into drag operation"
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'ar 6/25/1999 11:33'!
argument
	^argument! !

!ColorPickerMorph methodsFor: 'private' stamp: 'ar 7/19/2003 20:40'!
argumentsWith: aColor
	"Return an argument array appropriate to this action selector"

	| nArgs |
	nArgs := selector ifNil:[0] ifNotNil:[selector numArgs].
	nArgs = 0 ifTrue:[^#()].
	nArgs = 1 ifTrue:[^ {aColor}].
	nArgs = 2 ifTrue:[^ {aColor. sourceHand}].
	nArgs = 3 ifTrue:[^ {aColor. argument. sourceHand}].
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'ar 6/25/1999 11:33'!
argument: anObject
	argument := anObject! !

!ColorPickerMorph methodsFor: 'other' stamp: 'di 11/27/1999 08:51'!
bestPositionNear: box inWorld: world
	| points b |
	points := #(topCenter rightCenter bottomCenter leftCenter).  "possible anchors"
	1 to: 4 do:
		[:i |  "Try the four obvious anchor points"
		b := self bounds align: (self bounds perform: (points at: i))
					with: (box perform: (points atWrap: i + 2)).
		(world viewBox containsRect: b) ifTrue:
			[^ b topLeft"  Yes, it fits"]].

	^ 20@20  "when all else fails"
! !

!ColorPickerMorph methodsFor: 'initialization' stamp: 'di 9/28/2000 12:05'!
buildChartForm
	| chartForm |
	chartForm := ColorChart deepCopy asFormOfDepth: Display depth.
	chartForm fill: ((TransparentBox left + 9)@0 extent: 1@9) fillColor: Color lightGray.
	chartForm fill: ((TransparentBox right - 10)@0 extent: 1@9) fillColor: Color lightGray.
	TransText displayOn: chartForm at: 62@0.
	Display depth = 32 ifTrue:
		["Set opaque bits for 32-bit display"
		chartForm fill: chartForm boundingBox rule: Form under
				fillColor: (Color r: 0.0 g: 0.0 b: 0.0 alpha: 1.0)].
	chartForm borderWidth: 1.
	self form: chartForm.
	selectedColor ifNotNil: [self updateAlpha: selectedColor alpha].
	self updateSelectorDisplay.

! !

!ColorPickerMorph methodsFor: 'initialization' stamp: 'sw 9/8/2000 18:14'!
choseModalityFromPreference
	"Decide whether to be modal or not by consulting the prevailing preference"

	self initializeModal: Preferences modalColorPickers! !

!ColorPickerMorph methodsFor: 'geometry testing' stamp: 'LC 2/2/2000 04:28'!
containsPoint: aPoint 
	^ (super containsPoint: aPoint)
		or: [RevertBox containsPoint: aPoint - self topLeft]! !

!ColorPickerMorph methodsFor: 'submorphs-add/remove' stamp: 'nk 4/17/2004 19:34'!
delete
	"The moment of departure has come.
	If the receiver has an affiliated command, finalize it and have the system remember it.
	In any case, delete the receiver"

	(selector isNil or: [ target isNil ]) ifFalse: [
		self rememberCommand: 
			(Command new
				cmdWording: 'color change' translated;
				undoTarget: target selector: selector arguments: (self argumentsWith: originalColor);
				redoTarget: target selector: selector arguments: (self argumentsWith: selectedColor)).
	].
	super delete! !

!ColorPickerMorph methodsFor: 'private' stamp: 'di 9/27/2000 12:55'!
deleteAllBalloons

	self submorphsDo: [:m | m deleteBalloon].
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:15'!
deleteOnMouseUp

	^ deleteOnMouseUp
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:15'!
deleteOnMouseUp: aBoolean

	deleteOnMouseUp := aBoolean.
! !

!ColorPickerMorph methodsFor: 'drawing' stamp: 'di 9/3/1999 13:34'!
drawOn: aCanvas
	aCanvas depth = 1 ifTrue: [aCanvas fillRectangle: self bounds color: Color white].
	Display depth = originalForm depth ifFalse: [self buildChartForm].
	super drawOn: aCanvas! !

!ColorPickerMorph methodsFor: 'kedama' stamp: 'ar 1/3/2010 18:52'!
getColorFromKedamaWorldIfPossible: aGlobalPoint

	self world submorphs do: [:sub |
		 (sub isKedamaMorph) ifTrue: [
			sub morphsAt: aGlobalPoint unlocked: false do: [:e |
				^ e colorAt: (aGlobalPoint - e topLeft).
			].
		].
	].
	^ nil.
! !

!ColorPickerMorph methodsFor: 'event handling' stamp: 'jm 11/4/97 07:15'!
handlesMouseDown: evt

	^ true
! !

!ColorPickerMorph methodsFor: 'other' stamp: 'di 9/25/2000 15:38'!
indicateColorUnderMouse
	"Track the mouse with the special eyedropper cursor, and accept whatever color is under the mouse as the currently-chosen color; reflect that choice in the feedback box, and return that color."

	| pt |
	self pickColorAt: (pt := Sensor cursorPoint).
	isModal ifTrue:
		[self activeHand position: pt.
		self world displayWorldSafely; runStepMethods].
	^ selectedColor	! !

!ColorPickerMorph methodsFor: 'event handling' stamp: 'RAA 2/19/2001 13:16'!
inhibitDragging

	^self hasProperty: #noDraggingThisPicker! !

!ColorPickerMorph methodsFor: 'initialization' stamp: 'ar 9/4/2001 13:26'!
initialize
	"Initialize the receiver.  Obey the modalColorPickers preference when deciding how to configure myself.  This is not quite satisfactory -- we'd like to have explicit calls tell us things like whether whether to be modal, whether to allow transparency, but for the moment, in grand Morphic fashion, this is rather inflexibly all housed right here"

	super initialize.
	self clipSubmorphs: true.
	self buildChartForm.
	
	selectedColor := Color white.
	sourceHand := nil.
	deleteOnMouseUp := false.
	clickedTranslucency := false.
	updateContinuously := true.
	selector := nil.
	target := nil! !

!ColorPickerMorph methodsFor: 'initialization' stamp: 'yo 2/23/2005 17:17'!
initializeForPropertiesPanel
	"Initialize the receiver.  If beModal is true, it will be a modal color picker, else not"

	isModal := false.
	self removeAllMorphs.
	self setProperty: #noDraggingThisPicker toValue: true.

	self addMorph: ((Morph newBounds: (RevertBox translateBy: self topLeft))
			color: Color transparent; setCenteredBalloonText: 'restore original color' translated).
	self addMorph: ((Morph newBounds: (FeedbackBox translateBy: self topLeft))
			color: Color transparent; setCenteredBalloonText: 'shows selected color' translated).
	self addMorph: ((Morph newBounds: (TransparentBox translateBy: self topLeft))
			color: Color transparent; setCenteredBalloonText: 'adjust translucency' translated).

	self buildChartForm.
	
	selectedColor ifNil: [selectedColor := Color white].
	sourceHand := nil.
	deleteOnMouseUp := false.
	updateContinuously := true.
! !

!ColorPickerMorph methodsFor: 'initialization' stamp: 'yo 2/23/2005 17:13'!
initializeModal: beModal
	"Initialize the receiver.  If beModal is true, it will be a modal color picker, else not"

	isModal := beModal.
	self removeAllMorphs.
	isModal ifFalse:
		[theSelectorDisplayMorph := AlignmentMorph newRow
			color: Color white;
			borderWidth: 1;
			borderColor: Color red;
			hResizing: #shrinkWrap;
			vResizing: #shrinkWrap;
			addMorph: (StringMorph contents: 'theSelector' translated).
		self addMorph: theSelectorDisplayMorph.

		self addMorph: (SimpleButtonMorph new borderWidth: 0;
			label: 'x' font: nil; color: Color transparent;
			actionSelector: #delete; target: self; useSquareCorners;
			position: self topLeft - (0@3); extent: 10@12;
			setCenteredBalloonText: 'dismiss color picker' translated)].

	self addMorph: ((Morph newBounds: (DragBox translateBy: self topLeft))
			color: Color transparent; setCenteredBalloonText: 'put me somewhere' translated).
	self addMorph: ((Morph newBounds: (RevertBox translateBy: self topLeft))
			color: Color transparent; setCenteredBalloonText: 'restore original color' translated).
	self addMorph: ((Morph newBounds: (FeedbackBox translateBy: self topLeft))
			color: Color transparent; setCenteredBalloonText: 'shows selected color' translated).
	self addMorph: ((Morph newBounds: (TransparentBox translateBy: self topLeft))
			color: Color transparent; setCenteredBalloonText: 'adjust translucency' translated).

	self buildChartForm.
	
	selectedColor ifNil: [selectedColor := Color white].
	sourceHand := nil.
	deleteOnMouseUp := false.
	updateContinuously := true.
! !

!ColorPickerMorph methodsFor: 'halos and balloon help' stamp: 'sw 7/6/1999 09:07'!
isLikelyRecipientForMouseOverHalos
	^ false! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'nice 12/27/2009 03:11'!
locationIndicator
	
	^self valueOfProperty: #locationIndicator ifAbsent:[ | loc |
		loc := EllipseMorph new.
		loc color: Color transparent; 
			borderWidth: 1; 
			borderColor: Color red; 
			extent: 6@6.
		self setProperty: #locationIndicator toValue: loc.
		self addMorphFront: loc.
		loc]! !

!ColorPickerMorph methodsFor: 'private' stamp: 'dgd 2/21/2003 22:59'!
modalBalloonHelpAtPoint: cursorPoint 
	self flag: #arNote.	"Throw this away. There needs to be another way."
	self submorphsDo: 
			[:m | 
			m wantsBalloon 
				ifTrue: 
					[(m valueOfProperty: #balloon) isNil
						ifTrue: 
							[(m containsPoint: cursorPoint) ifTrue: [m showBalloon: m balloonText]]
						ifFalse: [(m containsPoint: cursorPoint) ifFalse: [m deleteBalloon]]]]! !

!ColorPickerMorph methodsFor: 'event handling' stamp: 'RAA 2/19/2001 13:17'!
mouseDown: evt
	| localPt |
	localPt := evt cursorPoint - self topLeft.
	self deleteAllBalloons.
	clickedTranslucency := TransparentBox containsPoint: localPt.
	self inhibitDragging ifFalse: [
		(DragBox containsPoint: localPt)
			ifTrue: [^ evt hand grabMorph: self].
	].
	(RevertBox containsPoint: localPt)
		ifTrue: [^ self updateColor: originalColor feedbackColor: originalColor].
	self inhibitDragging ifFalse: [self comeToFront].
	sourceHand := evt hand.
	self startStepping.
! !

!ColorPickerMorph methodsFor: 'event handling' stamp: 'yo 9/29/2004 10:24'!
mouseUp: evt

	| c |
	self stopStepping.
	sourceHand := nil.
	deleteOnMouseUp ifTrue: [self delete].
	c := self getColorFromKedamaWorldIfPossible: evt cursorPoint.
	c ifNotNil: [selectedColor := c].
	self updateTargetColor.
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'KR 12/9/2005 22:51'!
originalColor: colorOrSymbol 
	"Set the receiver's original color.  It is at this point that a command is launched to represent the action of the picker, in support of Undo."

	originalColor := (colorOrSymbol isColor) 
				ifTrue: [colorOrSymbol]
				ifFalse: [Color lightGreen].
	originalForm fill: RevertBox fillColor: originalColor.
	selectedColor := originalColor.
	self updateAlpha: originalColor alpha.
	self locationIndicator 
		center: self topLeft + (self positionOfColor: originalColor)! !

!ColorPickerMorph methodsFor: 'private' stamp: 'yo 9/29/2004 10:57'!
pickColorAt: aGlobalPoint 

	| alpha selfRelativePoint pickedColor c |
	clickedTranslucency ifNil: [clickedTranslucency := false].
	selfRelativePoint := (self globalPointToLocal: aGlobalPoint) - self topLeft.
	(FeedbackBox containsPoint: selfRelativePoint) ifTrue: [^ self].
	(RevertBox containsPoint: selfRelativePoint)
		ifTrue: [^ self updateColor: originalColor feedbackColor: originalColor].

	"check for transparent color and update using appropriate feedback color "
	(TransparentBox containsPoint: selfRelativePoint) ifTrue:
		[clickedTranslucency ifFalse: [^ self].  "Can't wander into translucency control"
		alpha := (selfRelativePoint x - TransparentBox left - 10) asFloat /
							(TransparentBox width - 20)
							min: 1.0 max: 0.0.
					"(alpha roundTo: 0.01) printString , '   ' displayAt: 0@0." " -- debug"
		self 
			updateColor: (selectedColor alpha: alpha)
			feedbackColor: (selectedColor alpha: alpha).
		^ self].

	"pick up color, either inside or outside this world"
	clickedTranslucency ifTrue: [^ self].  "Can't wander out of translucency control"
	self locationIndicator visible: false. self refreshWorld.
	pickedColor := Display colorAt: aGlobalPoint.
	c := self getColorFromKedamaWorldIfPossible: aGlobalPoint.
	c ifNotNil: [pickedColor := c].
	self locationIndicator visible: true. self refreshWorld.
	self 
		updateColor: (
			(selectedColor isColor and: [selectedColor isTranslucentColor])
						ifTrue: [pickedColor alpha: selectedColor alpha]
						ifFalse: [pickedColor]
		)
		feedbackColor: pickedColor! !

!ColorPickerMorph methodsFor: 'menu' stamp: 'tfel 8/8/2016 18:00'!
pickUpColorFor: aMorph
	"Show the eyedropper cursor, and modally track the mouse through a mouse-down and mouse-up cycle"

      | aHand localPt c |
	aHand := aMorph ifNil: [self activeHand] ifNotNil: [aMorph activeHand].
	aHand ifNil: [aHand := self currentHand].
	self addToWorld: aHand world near: (aMorph ifNil: [aHand world]) fullBounds.
	self owner ifNil: [^ self].

	aHand showTemporaryCursor: (ScriptingSystem formAtKey: #Eyedropper) 
			hotSpotOffset: 6 @ 31.    "<<<< the form was changed a bit??"

	self updateContinuously: false.
	[Sensor anyButtonPressed]
		whileFalse: 
			 [self trackColorUnderMouse].
	self deleteAllBalloons.

	localPt := Sensor cursorPoint - self topLeft.
	self inhibitDragging ifFalse: [
		(DragBox containsPoint: localPt) ifTrue:
			["Click or drag the drag-dot means to anchor as a modeless picker"
			^ self anchorAndRunModeless: aHand].
	].
	(clickedTranslucency := TransparentBox containsPoint: localPt)
		ifTrue: [selectedColor := originalColor].

	self updateContinuously: true.
	[Sensor anyButtonPressed]
		whileTrue:
			 [self updateTargetColorWith: self indicateColorUnderMouse].
	c := self getColorFromKedamaWorldIfPossible: Sensor cursorPoint.
	c ifNotNil: [selectedColor := c].
	aHand newMouseFocus: nil;
		showTemporaryCursor: nil;
		flushEvents.
	self delete.
		 
 ! !

!ColorPickerMorph methodsFor: 'private' stamp: 'ar 9/4/2001 13:27'!
positionOfColor: aColor
	"Compute the position of the given color in the color chart form"
	| rgbRect x y h s v |
	rgbRect := (0@0 extent: originalForm boundingBox extent) insetBy: (1@10 corner: 11@1).
	h := aColor hue.
	s := aColor saturation.
	v := aColor brightness.
	h = 0.0 ifTrue:["gray"
		^(rgbRect right + 6) @ (rgbRect height * (1.0 - v) + rgbRect top)].
	x := (h + 22 \\ 360 / 360.0 * rgbRect width) rounded.
	y := 0.5.
	s < 1.0 ifTrue:[y := y - (1.0 - s * 0.5)].
	v < 1.0 ifTrue:[y := y + (1.0 - v * 0.5)].
	y := (y * rgbRect height) rounded.
	^x@y + (1@10)! !

!ColorPickerMorph methodsFor: 'other' stamp: 'ar 12/8/2000 15:32'!
putUpFor: aMorph near: aRectangle
	"Put the receiver up on the screen.   Note highly variant behavior depending on the setting of the #modalColorPickers preference"
	| layerNumber |
	aMorph isMorph ifTrue: [
		layerNumber := aMorph morphicLayerNumber.
		aMorph allOwnersDo:[:m|
			layerNumber := layerNumber min: m morphicLayerNumber].
		self setProperty: #morphicLayerNumber toValue: layerNumber - 0.1
	].

	isModal == true "backward compatibility"
		ifTrue:
			[self pickUpColorFor: aMorph]
		ifFalse:
			[self addToWorld:
				((aMorph notNil and: [aMorph world notNil])
					ifTrue:
						[aMorph world]
					ifFalse:
						[self currentWorld])
		  		near:
					(aRectangle ifNil:
						[aMorph ifNil: [100@100 extent: 1@1] ifNotNil: [aMorph fullBoundsInWorld]])]! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:15'!
selectedColor

	^ selectedColor
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:15'!
selector

	^ selector
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'di 8/30/2000 13:40'!
selector: aSymbol
	"Set the selector to be associated with the receiver.  Store it in the receiver's command, if appropriate"

	selector := aSymbol.
	self updateSelectorDisplay! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:15'!
sourceHand

	^ sourceHand
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:15'!
sourceHand: aHand

	sourceHand := aHand.
! !

!ColorPickerMorph methodsFor: 'stepping and presenter' stamp: 'jm 11/4/97 07:15'!
step

	sourceHand ifNotNil:
		[self pickColorAt: sourceHand position].
! !

!ColorPickerMorph methodsFor: 'testing' stamp: 'jm 11/4/97 07:15'!
stepTime

	^ 50
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:15'!
target

	^ target
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'aoy 2/15/2003 21:24'!
target: anObject 
	target := anObject.
	selectedColor := (target respondsTo: #color)  
				ifTrue: [target color]
				ifFalse: [Color white]! !

!ColorPickerMorph methodsFor: 'menu' stamp: 'jm 11/4/97 07:46'!
toggleDeleteOnMouseUp

	deleteOnMouseUp := deleteOnMouseUp not.
! !

!ColorPickerMorph methodsFor: 'menu' stamp: 'jm 11/4/97 07:46'!
toggleUpdateContinuously

	updateContinuously := updateContinuously not.
! !

!ColorPickerMorph methodsFor: 'private' stamp: 'di 9/30/2000 10:07'!
trackColorAt: aGlobalPoint 
	"Before the mouse comes down in a modal color picker, track the color under the cursor, and show it in the feedback box, but do not make transparency changes"

	| selfRelativePoint pickedColor |
	selfRelativePoint := (self globalPointToLocal: aGlobalPoint) - self topLeft.
	(FeedbackBox containsPoint: selfRelativePoint) ifTrue: [^ self].
	(RevertBox containsPoint: selfRelativePoint)
		ifTrue: [^ self updateColor: originalColor feedbackColor: originalColor].

	"check for transparent color and update using appropriate feedback color "
	(TransparentBox containsPoint: selfRelativePoint) ifTrue: [^ self].

	"pick up color, either inside or outside this world"
	pickedColor := Display colorAt: aGlobalPoint.
	self updateColor: (pickedColor alpha: originalColor alpha)
		feedbackColor: pickedColor! !

!ColorPickerMorph methodsFor: 'other' stamp: 'di 9/27/2000 11:48'!
trackColorUnderMouse
	"Track the mouse with the special eyedropper cursor, and accept whatever color is under the mouse as the currently-chosen color; reflect that choice in the feedback box, and return that color."

	| pt |
	selectedColor := originalColor.
	self trackColorAt: (pt := Sensor cursorPoint).
	isModal ifTrue:
		[self activeHand position: pt.
		self world displayWorldSafely; runStepMethods.
		self modalBalloonHelpAtPoint: pt].
	^ selectedColor	! !

!ColorPickerMorph methodsFor: 'private' stamp: 'di 9/28/2000 11:10'!
updateAlpha: alpha
	| sliderRect |
	sliderRect := (TransparentBox left + 10)@1 corner: (TransparentBox right - 9)@9.
	originalForm fill: (sliderRect withRight: sliderRect left + (alpha*sliderRect width))
				fillColor: Color lightGray.
	originalForm fillWhite: (sliderRect withLeft: sliderRect left + (alpha*sliderRect width)).
	originalForm fill: ((TransparentBox right - 9)@1 extent: 8@8)
				fillColor: (alpha < 1.0 ifTrue: [Color white] ifFalse: [Color lightGray]).
	TransText displayOn: originalForm at: 62@1 rule: Form paint.
! !

!ColorPickerMorph methodsFor: 'private' stamp: 'tak 9/12/2007 20:12'!
updateColor: aColor feedbackColor: feedbackColor
	"Set my selected color to the given color if it is different. Give user feedback. Inform the target of the change if the target and selector are not nil." 

	selectedColor = aColor ifTrue: [^ self].  "do nothing if color doesn't change"

	self updateAlpha: aColor alpha.
	originalForm fill: (FeedbackBox insetBy: 2) fillColor: feedbackColor.
	self form: originalForm.
	selectedColor _ aColor.
	updateContinuously ifTrue: [self updateTargetColor].
	self locationIndicator center: self topLeft + (self positionOfColor: feedbackColor).! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
updateContinuously

	^ updateContinuously
! !

!ColorPickerMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
updateContinuously: aBoolean

	updateContinuously := aBoolean.
! !

!ColorPickerMorph methodsFor: 'initialization' stamp: 'sma 4/22/2000 19:39'!
updateSelectorDisplay
	theSelectorDisplayMorph ifNil: [^self].
	theSelectorDisplayMorph position: self bottomLeft.
	theSelectorDisplayMorph firstSubmorph contents: selector asString , ' ' , selectedColor printString! !

!ColorPickerMorph methodsFor: 'private' stamp: 'dgd 2/22/2003 18:41'!
updateTargetColor
	| nArgs |
	(target notNil and: [selector notNil]) 
		ifTrue: 
			[self updateSelectorDisplay.
			nArgs := selector numArgs.
			nArgs = 1 ifTrue: [^target perform: selector with: selectedColor].
			nArgs = 2 
				ifTrue: 
					[^target 
						perform: selector
						with: selectedColor
						with: sourceHand].
			nArgs = 3 
				ifTrue: 
					[^target 
						perform: selector
						with: selectedColor
						with: argument
						with: sourceHand]]! !

!ColorPickerMorph methodsFor: 'private' stamp: 'dgd 2/22/2003 18:41'!
updateTargetColorWith: aColor 
	"Update the target so that it reflects aColor as the color choice"

	(target notNil and: [selector notNil]) 
		ifTrue: 
			[self updateSelectorDisplay.
			^target perform: selector withArguments: (self argumentsWith: aColor)]! !

SketchMorph subclass: #JoystickMorph
	instanceVariableNames: 'handleMorph xScale yScale radiusScale lastAngle autoCenter realJoystickIndex lastRealJoystickValue button1 button2'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!JoystickMorph commentStamp: 'kfr 10/27/2003 16:25' prior: 0!
A widget that simulates a joystick. Mosly used in etoy scripting.!

!JoystickMorph class methodsFor: 'scripting' stamp: 'sw 11/5/97 12:04'!
authoringPrototype
	^ self new markAsPartsDonor! !

!JoystickMorph class methodsFor: 'parts bin' stamp: 'tak 7/25/2007 00:06'!
descriptionForPartsBin
	^ self partName: 	'Joystick' translatedNoop
		categories:		{'Basic' translatedNoop}
		documentation:	'A joystick-like control' translatedNoop! !

!JoystickMorph class methodsFor: 'new-morph participation' stamp: 'sw 4/26/1999 16:02'!
includeInNewMorphMenu
	^ true! !

!JoystickMorph class methodsFor: 'class initialization' stamp: 'asm 4/10/2003 13:10'!
initialize

	self registerInFlapsRegistry.! !

!JoystickMorph class methodsFor: 'class initialization' stamp: 'KR 8/12/2007 10:53'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | cl registerQuad: {#JoystickMorph. #authoringPrototype. 'Joystick' translatedNoop. 'A joystick-like control' translatedNoop} 
						forFlapNamed: 'PlugIn Supplies'.
						cl registerQuad: {#JoystickMorph	. #authoringPrototype. 'Joystick' translatedNoop. 'A joystick-like control' translatedNoop} 
						forFlapNamed: 'Scripting'.
						cl registerQuad: {#JoystickMorph	. #authoringPrototype. 'Joystick' translatedNoop. 'A joystick-like control' translatedNoop}
						forFlapNamed: 'Supplies']! !

!JoystickMorph class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 12:36'!
unload
	"Unload the receiver from global registries"

	self environment at: #Flaps ifPresent: [:cl |
	cl unregisterQuadsWithReceiver: self] ! !

!JoystickMorph methodsFor: 'menu' stamp: 'sw 4/29/2004 20:00'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	"Add custom items to the menu"

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu addLine.
	aCustomMenu add: 'set X range' translated action: #setXRange.
	aCustomMenu add: 'set Y range' translated action: #setYRange.
	aCustomMenu addLine.
	aCustomMenu addUpdating: #autoCenterString target: self action: #toggleAutoCenter.
	aCustomMenu balloonTextForLastItem: 'When auto-center is on, every time you let go of the Joystick knob, it springs back to the neutral position at the center of the device' translated. 
	aCustomMenu addUpdating: #realJoystickString target: self action: #toggleRealJoystick.
	aCustomMenu balloonTextForLastItem: 'Governs whether this joystick should track the motions of a real, physical joystick attached to the computer.' translated. 
	aCustomMenu addUpdating: #joystickNumberString enablementSelector: #realJoystickInUse target: self selector: #chooseJoystickNumber argumentList: #().
	aCustomMenu balloonTextForLastItem: 'Choose which physical device is associated with the joystick.' translated! !

!JoystickMorph methodsFor: 'accessing' stamp: ''!
amount

	^ (handleMorph center - self center) r * radiusScale! !

!JoystickMorph methodsFor: 'accessing' stamp: ''!
angle

	self center = handleMorph center ifTrue: [^ lastAngle].
	^ 360.0 - (handleMorph center - self center) theta radiansToDegrees! !

!JoystickMorph methodsFor: 'menu' stamp: 'sw 4/29/2004 18:30'!
autoCenterString
	"Answer a string characterizing whether or not I have auto-center on"

	^ (autoCenter == true	ifTrue: ['<yes>'] ifFalse: ['<no>']), ('auto-center' translated)! !

!JoystickMorph methodsFor: 'accessing' stamp: 'mir 9/4/2005 16:46'!
button1
	^button1 == true! !

!JoystickMorph methodsFor: 'accessing' stamp: 'mir 9/4/2005 16:46'!
button2
	^button2 == true! !

!JoystickMorph methodsFor: 'menu' stamp: 'nice 1/18/2010 18:58'!
chooseJoystickNumber
	"Allow the user to select a joystick number"

	| result aNumber str |
	str := self lastRealJoystickIndex asString.
	result := UIManager default 
				request: ('Joystick device number (currently {1})' translated format: {str})
				initialAnswer: str.
	aNumber := [result asNumber] on: Error do: [:err | ^Beeper beep].
	(aNumber > 0 and: [aNumber <= 32]) 
		ifFalse: 
			["???"

			^Beeper beep].
	realJoystickIndex := aNumber.
	self setProperty: #lastRealJoystickIndex toValue: aNumber.
	self startStepping! !

!JoystickMorph methodsFor: 'event handling' stamp: 'sw 5/11/1998 13:51'!
handlesMouseDown: evt

	self inPartsBin ifTrue: [^ false].

	true ifTrue: [^ true].  "5/7/98 jhm temporary fix to allow use when rotated"

	(handleMorph fullContainsPoint: evt cursorPoint)
		ifTrue: [^ true]
		ifFalse: [^ super handlesMouseDown: evt].
! !

!JoystickMorph methodsFor: 'initialization' stamp: 'bf 9/5/2005 11:23'!
initialize

	super initialize.
	xScale := 0.25.
	yScale := 0.25.
	radiusScale := 1.0.
	lastAngle := 0.0.
	autoCenter := true.
	self form: ((Form extent: 55@55 depth: 8) fillColor: (Color r: 0.3 g: 0.2 b: 0.2)).
	handleMorph := EllipseMorph new.
	handleMorph color: Color red; extent: 15@15.
	self addMorph: handleMorph.
	self moveHandleToCenter.
	button1 := false.
	button2 := false.
! !

!JoystickMorph methodsFor: 'parts bin' stamp: 'sw 8/12/2001 17:26'!
initializeToStandAlone
	"Circumvent SketchMorph's implementation here"

	self initialize! !

!JoystickMorph methodsFor: 'halos and balloon help' stamp: 'sw 8/11/2004 18:10'!
isLikelyRecipientForMouseOverHalos
	"The automatic mouseover halos interere with the proper functioning of the joystick's knob"

	^ false! !

!JoystickMorph methodsFor: 'menu' stamp: 'yo 2/11/2005 09:19'!
joystickNumberString
	"Answer a string characterizing the joystick number"

	^ 'set real joystick number (now {1})' translated format: {self lastRealJoystickIndex asString}.
! !

!JoystickMorph methodsFor: 'menu' stamp: 'sw 4/29/2004 20:08'!
lastRealJoystickIndex
	"Answer the last remembered real joystick index.  Initialize it to 1 if need be"

	^ self valueOfProperty: #lastRealJoystickIndex ifAbsentPut: [1] ! !

!JoystickMorph methodsFor: 'accessing' stamp: ''!
leftRight

	^ (handleMorph center x - self center x) * xScale
! !

!JoystickMorph methodsFor: 'event handling' stamp: ''!
mouseMove: evt
	"Make handle track the cursor within my bounds."

	| m r center |
	m := handleMorph.
	center := m center.
	r := m owner innerBounds insetBy:
		((center - m fullBounds origin) corner: (m fullBounds corner - center)).
	m position: (evt cursorPoint adhereTo: r) - (m extent // 2).
! !

!JoystickMorph methodsFor: 'event handling' stamp: 'jm 11/18/97 14:26'!
mouseUp: evt

	lastAngle := self angle.
	autoCenter ifTrue: [self moveHandleToCenter].
! !

!JoystickMorph methodsFor: 'other' stamp: ''!
moveHandleToCenter

	handleMorph position: self center - (handleMorph extent // 2).
! !

!JoystickMorph methodsFor: 'menu' stamp: 'sw 4/29/2004 19:57'!
realJoystickInUse
	"Answer whether a real joystick is in use"

	^ realJoystickIndex notNil! !

!JoystickMorph methodsFor: 'menu' stamp: 'sw 4/29/2004 18:29'!
realJoystickString
	"Answer a string characterizing whether or not I am currenty tracking a real joystick"

	^ (realJoystickIndex ifNil: ['<no>'] ifNotNil: ['<yes>']), ('track real joystick' translated)! !

!JoystickMorph methodsFor: 'menu' stamp: 'nice 12/25/2013 22:07'!
setXRange

	| range |
	range := UIManager default
		request:
'Type the maximum value for the X axis' translated
		initialAnswer: ((xScale * (self width - handleMorph width) / 2.0) printShowingMaxDecimalPlaces: 2).
	range isEmpty ifFalse: [
		xScale := (2.0 * range asNumber asFloat) / (self width - handleMorph width)].
! !

!JoystickMorph methodsFor: 'menu' stamp: 'nice 12/25/2013 22:06'!
setYRange

	| range |
	range := UIManager default
		request:
'Type the maximum value for the Y axis'  translated
		initialAnswer: ((yScale * (self width - handleMorph width) / 2.0) printShowingMaxDecimalPlaces: 2).
	range isEmpty ifFalse: [
		yScale := (2.0 * range asNumber asFloat) / (self width - handleMorph width)].
! !

!JoystickMorph methodsFor: 'stepping and presenter' stamp: 'bf 9/5/2005 11:19'!
step
	"Track the real joystick whose index is realJoystickIndex."
	"Details:
	  a. if realJoystickIndex is nil we're not tracking a joystick
	  b. [-joyMax..joyMax] is nominal range of joystick in both X and Y
	  c. [-threshold..threshold] is considered 0 to compensate for poor joystick centering"

	| threshold joyMax joyPt joyBtn m mCenter r scaledPt  |
	super step.  "Run ticking user-written scripts if any"
	realJoystickIndex ifNil: [^ self].
	threshold := 30.
	joyMax := 350.
	joyPt := Sensor joystickXY: realJoystickIndex.
	joyBtn := Sensor joystickButtons: realJoystickIndex.

	button1 := (joyBtn bitAnd: 1) > 0.
	button2 := (joyBtn bitAnd: 2) > 0.
	
	joyPt x abs < threshold ifTrue: [joyPt := 0@joyPt y].
	joyPt y abs < threshold ifTrue: [joyPt := joyPt x@0].
	lastRealJoystickValue = joyPt ifTrue: [^ self].
	lastRealJoystickValue := joyPt.
	m := handleMorph.
	mCenter := m center.
	r := m owner innerBounds insetBy:
		((mCenter - m fullBounds origin) corner: (m fullBounds corner - mCenter)).
	scaledPt := r center + ((r extent * joyPt) / (joyMax * 2)) truncated.
	m position: (scaledPt adhereTo: r) - (m extent // 2).
! !

!JoystickMorph methodsFor: 'stepping and presenter' stamp: 'laza 6/8/2003 11:53'!
stepTime
	"Provide for as-fast-as-possible stepping in the case of a real joystick"

	^ realJoystickIndex
		ifNotNil:
			[0]  "fast as we can to track actual joystick"
		ifNil:
			[super stepTime]! !

!JoystickMorph methodsFor: 'menu' stamp: 'jm 6/22/1998 17:24'!
stopTrackingJoystick

	realJoystickIndex := nil.
	self stopStepping.
! !

!JoystickMorph methodsFor: 'menu' stamp: 'jm 11/18/97 14:33'!
toggleAutoCenter

	autoCenter := autoCenter not.
	autoCenter ifTrue: [self moveHandleToCenter].
! !

!JoystickMorph methodsFor: 'menu' stamp: 'sw 8/11/2004 18:15'!
toggleRealJoystick
	"Toggle whether or not one is using a real joystick"

	realJoystickIndex
		ifNil:
			[realJoystickIndex := self valueOfProperty: #lastRealJoystickIndex ifAbsentPut: [1].
			self startStepping]
		ifNotNil:
			[self stopTrackingJoystick]! !

!JoystickMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:41'!
trackRealJoystick

	| s |
	s := UIManager default
		request: 'Number of joystick to track?'
		initialAnswer: '1'.
	s isEmpty ifTrue: [^ self].
	realJoystickIndex := Number readFromString: s.
	self startStepping.
! !

!JoystickMorph methodsFor: 'accessing' stamp: 'jm 1/19/98 20:48'!
upDown

	^ (self center y - handleMorph center y) * yScale
! !

!SketchMorph class methodsFor: 'scripting' stamp: 'KR 5/23/2008 01:27'!
defaultNameStemForInstances
	^ 'Sketch' translatedNoop! !

!SketchMorph class methodsFor: 'instance creation' stamp: 'ar 3/4/2001 20:59'!
fromFile: aFileName
	^self fromStream: (FileStream readOnlyFileNamed: aFileName)! !

!SketchMorph class methodsFor: 'instance creation' stamp: 'ar 3/4/2001 21:00'!
fromStream: aStream
	^self withForm: (ImageReadWriter formFromStream: aStream)! !

!SketchMorph class methodsFor: 'new-morph participation' stamp: 'sw 9/28/1998 17:15'!
includeInNewMorphMenu
	"Not to be instantiated from the menu"
	^ false! !

!SketchMorph class methodsFor: 'testing' stamp: 'nk 6/12/2004 09:16'!
isSketchMorphClass
	^true! !

!SketchMorph class methodsFor: 'instance creation' stamp: 'nk 11/9/2003 08:11'!
openEditor
	"Create a new SketchMorph and open a SketchMorphEditor on it. 
	Answers the painted SketchMorph."
	"SketchMorph openEditor"
	| newSketch |
	newSketch := (self
				withForm: (Form extent: 100 @ 100 depth: Display depth)) center: self currentWorld center;
				 openInWorld;
				 editDrawing.
	^ newSketch! !

!SketchMorph class methodsFor: 'instance creation' stamp: 'di 2/17/2000 20:21'!
withForm: aForm
	"Note: 'SketchMorph withForm: zz' is MUCH faster
	than 'SketchMorph new form: zz'."

	^ self basicNew initializeWith: aForm! !

!SketchMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:45'!
addBorderToShape: evt
	| str borderWidth borderedForm r |
	str := UIManager default
		request: 'Please enter the desired border width' translated
		initialAnswer: '0'.
	borderWidth := Integer readFrom: (ReadStream on: str).
	(borderWidth between: 1 and: 10) ifFalse: [^ self].

	"Take care of growing appropriately.  Does this lose the reg point?"
	borderedForm := originalForm shapeBorder: Color black width: borderWidth.
	r := borderedForm rectangleEnclosingPixelsNotOfColor: Color transparent.
	self form: (borderedForm copy: r).
! !

!SketchMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 22:16'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	"Add custom menu items"

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu add: 'restore base graphic' translated target: self action: #restoreBaseGraphicFromMenu.
	aCustomMenu add: 'call this my base graphic' translated target: self action: #callThisBaseGraphic.
	aCustomMenu add: 'choose new graphic...' translated target: self action: #chooseNewGraphic.
	aCustomMenu addLine.
	aCustomMenu add: 'set as background' translated target: rotatedForm action: #setAsBackground.
	self addPaintingItemsTo: aCustomMenu hand: aHandMorph! !

!SketchMorph methodsFor: 'menus' stamp: 'ar 6/18/1999 06:40'!
addFillStyleMenuItems: aMenu hand: aHand
	"Do nothing here - we do not allow changing the fill style of a SketchMorph yet."! !

!SketchMorph methodsFor: 'menus' stamp: 'yo 9/14/2006 15:42'!
addToggleItemsToHaloMenu: aCustomMenu 
	"Add  toggle-items to the halo menu"

	super addToggleItemsToHaloMenu: aCustomMenu.
	(Smalltalk includesKey: #B3DRenderEngine) ifTrue: [
		aCustomMenu addUpdating: #useInterpolationString target: self action: #toggleInterpolation.
	].
! !

!SketchMorph methodsFor: 'e-toy support' stamp: 'sw 12/12/2001 13:10'!
baseGraphic
	"Answer my base graphic"

	^ self valueOfProperty: #baseGraphic ifAbsent:
		[self setProperty: #baseGraphic toValue: originalForm.
		^ originalForm]! !

!SketchMorph methodsFor: 'e-toy support' stamp: 'sw 12/12/2001 13:15'!
baseGraphic: aForm
	"Remember the given form as the receiver's base graphic"

	^ self setProperty: #baseGraphic toValue: aForm! !

!SketchMorph methodsFor: 'menu' stamp: 'kfr 4/19/2015 15:55'!
blur

	Cursor wait showWhile:[^(SketchMorph withForm: (self form approxGaussianBlur))  openInHand]! !

!SketchMorph methodsFor: 'menu' stamp: 'yo 3/15/2005 12:04'!
callThisBaseGraphic
	"Set my baseGraphic to be the current form"

	| aGraphic |
	self isInWorld ifFalse: [^ self inform: 

'oops, this menu is a for a morph that
has been replaced, probably because a
"look like" script was run.  Please dismiss
the menu and get a new one!!.  Sorry!!' translated].

	((aGraphic := self valueOfProperty: #baseGraphic)
				notNil and: [aGraphic ~= originalForm])
		ifTrue:
			[self setProperty: #baseGraphic toValue: originalForm]
		ifFalse:
			[self inform: 'this already *was* your baseGraphic' translated]! !

!SketchMorph methodsFor: 'drawing' stamp: 'nice 12/26/2009 21:35'!
canBeEnlargedWithB3D
	^self 
		valueOfProperty: #canBeEnlargedWithB3D
		ifAbsent: [
			| answer |
			answer := self rotatedForm colorsUsed allSatisfy: [ :c | c isTranslucent not].
			self setProperty: #canBeEnlargedWithB3D toValue: answer.
			answer
		]! !

!SketchMorph methodsFor: 'menus' stamp: 'nk 3/27/2001 17:57'!
changePixelsOfColor: c toColor: newColor

	| r |
	originalForm mapColor: c to: newColor.
	r := originalForm rectangleEnclosingPixelsNotOfColor: Color transparent.
	self form: (originalForm copy: r).

! !

!SketchMorph methodsFor: 'pen support' stamp: 'jm 4/22/1998 17:14'!
clearExtent: aPoint fillColor: aColor
	"Make this sketch have the given pixel dimensions and fill it with given color. Its previous contents are replaced."

	self form:
		((Form extent: aPoint depth: Display depth) fillColor: aColor).
! !

!SketchMorph methodsFor: 'menus' stamp: 'sw 8/18/2009 18:52'!
collapse
	"Replace the receiver with a collapsed rendition of itself."

	|  w collapsedVersion a ht tab |

	(w _ self world) ifNil: [^self].
	collapsedVersion _ (self imageForm scaledToSize: 50@50) asMorph.
	collapsedVersion setProperty: #uncollapsedMorph toValue: self.
	collapsedVersion on: #mouseUp send: #uncollapseSketch to: collapsedVersion.
	
	collapsedVersion setBalloonText: ('A collapsed version of {1}.  Click to open it back up.' translated format: {self externalName}).

	self delete.
	w addMorphFront: (
		a _ AlignmentMorph newRow
			hResizing: #shrinkWrap;
			vResizing: #shrinkWrap;
			borderWidth: 4;
			borderColor: Color white;
			addMorph: collapsedVersion;
			yourself).
	a setNameTo: self externalName.
	ht := (tab := ActiveWorld findA: SugarNavTab)
		ifNotNil:
			[tab height]
		ifNil:
			[80].
	a position: 0@ht.

	collapsedVersion setProperty: #collapsedMorphCarrier toValue: a.

	(self valueOfProperty: #collapsedPosition) ifNotNilDo:
		[:priorPosition |
			a position: priorPosition]! !

!SketchMorph methodsFor: 'geometry testing' stamp: ''!
containsPoint: aPoint

	^ (self bounds containsPoint: aPoint) and:
	  [(self rotatedForm isTransparentAt: aPoint - bounds origin) not]
! !

!SketchMorph methodsFor: 'objects from disk' stamp: 'RAA 12/20/2000 17:49'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	scalePoint ifNil: [scalePoint := 1.0@1.0].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!SketchMorph methodsFor: 'drawing' stamp: 'RAA 12/17/2000 14:24'!
drawHighResolutionOn: aCanvas in: aRectangle

	| r finalClipRect scale sourceOrigin sourceExtent sourceRect biggerSource biggerDestExtent interForm offsetInBigger |

	r := aRectangle translateBy: aCanvas origin.
	finalClipRect := r intersect: (aCanvas clipRect translateBy: aCanvas origin).
	self canBeEnlargedWithB3D ifTrue: [
		(WarpBlt toForm: aCanvas form)
			clipRect: finalClipRect;
			sourceForm: originalForm;
			cellSize: 2;  "installs a colormap"
			combinationRule: Form paint;

			copyQuad: originalForm boundingBox innerCorners 
			toRect: r.
		^self
	].
	scale := aRectangle extent / originalForm extent.
	sourceOrigin := originalForm offset + (aCanvas clipRect origin - aRectangle origin / scale).
	sourceExtent := aCanvas clipRect extent / scale.
	sourceRect := sourceOrigin rounded extent: sourceExtent rounded.
	biggerSource := sourceRect expandBy: 1.
	biggerDestExtent := (biggerSource extent * scale) rounded.
	offsetInBigger := (sourceOrigin - biggerSource origin * scale) rounded.

	interForm := Form extent: biggerDestExtent depth: aCanvas depth.
	(originalForm copy: biggerSource)
		displayInterpolatedIn: interForm boundingBox
		on: interForm.
	aCanvas 
		drawImage: interForm 
		at: aCanvas clipRect origin 
		sourceRect: (offsetInBigger extent: aCanvas clipRect extent).


! !

!SketchMorph methodsFor: 'drawing' stamp: 'ar 7/28/2005 16:51'!
drawInterpolatedImage: aForm on: aCanvas
	"Draw the given form onto the canvas using the Balloon 3D engine"
	| engine |
	engine := Smalltalk at: #B3DRenderEngine 
		ifPresent:[:b3d | b3d defaultForPlatformOn: aCanvas form].
	engine == nil ifTrue:[
		self useInterpolation: false.
		^self generateRotatedForm].
	"Setup the engine"
	engine viewport: aCanvas form boundingBox.
	"Install the material to be used (using a plain white emission color)"
	engine material: ((Smalltalk at: #B3DMaterial) new emission: Color white).
	"Install the texture"
	engine texture: aForm.
	"Draw the mesh"
	engine render: ((Smalltalk at: #B3DIndexedQuadMesh) new plainTextureRect).
	"and finish"
	engine finish.! !

!SketchMorph methodsFor: 'drawing' stamp: 'tk 10/9/2002 09:15'!
drawOn: aCanvas
	aCanvas translucentImage: self rotatedForm at: bounds origin
! !

!SketchMorph methodsFor: 'menu' stamp: 'kfr 4/19/2015 15:58'!
edgeDetect
	
	Cursor wait showWhile:[^(SketchMorph withForm: (self form edgeDetect))  openInHand]
	! !

!SketchMorph methodsFor: 'menu' stamp: 'tk 7/6/1998 15:31'!
editDrawing
	self flag: #deferred.  "Don't allow this if the user is already in paint mode, because it creates a very strange situation."
	"costumee ifNotNil: [self forwardDirection: costumee direction]."  "how say this?"
	self editDrawingIn: self pasteUpMorph forBackground: false
! !

!SketchMorph methodsFor: 'menu' stamp: 'nice 12/27/2009 03:11'!
editDrawingIn: aPasteUpMorph forBackground: forBackground
	"Edit an existing sketch."

	| w bnds sketchEditor rotCenter aPaintTab aWorld aPaintBox |
	self world assureNotPaintingElse: [^self].
	w := aPasteUpMorph world.
	w prepareToPaint.
	w displayWorld.
	self visible: false.
	bnds := forBackground 
				ifTrue: [aPasteUpMorph boundsInWorld]
				ifFalse: 
					[bnds := self boundsInWorld expandBy: 60 @ 60.
					(aPasteUpMorph paintingBoundsAround: bnds center) merge: bnds]. 
	sketchEditor := SketchEditorMorph new.
	forBackground 
		ifTrue: [sketchEditor setProperty: #background toValue: true].
	w addMorphFront: sketchEditor.
	sketchEditor 
		initializeFor: self
		inBounds: bnds
		pasteUpMorph: aPasteUpMorph.
	rotCenter := self rotationCenter.

	sketchEditor afterNewPicDo: 
			[:aForm :aRect | | tfx | 
			self visible: true.
			self form: aForm.
			tfx := aPasteUpMorph transformFrom: aPasteUpMorph world.
			self topRendererOrSelf position: (tfx globalPointToLocal: aRect origin).
			self rotationStyle: sketchEditor rotationStyle.
			self forwardDirection: sketchEditor forwardDirection.
			(rotCenter notNil and: [(rotCenter = (0.5 @ 0.5)) not]) ifTrue:
				[self rotationCenter: rotCenter].
			(aPaintTab := (aWorld := self world) paintingFlapTab) 
				ifNotNil: [aPaintTab hideFlap]
				ifNil: [(aPaintBox := aWorld paintBox) ifNotNil: [aPaintBox delete]].
			self presenter drawingJustCompleted: self.

			forBackground ifTrue: [self goBehind	"shouldn't be necessary"]]
		ifNoBits: 
			[ | pal |"If no bits drawn.  Must keep old pic.  Can't have no picture"

			self visible: true.
			aWorld := self currentWorld.
			"sometimes by now I'm no longer in a world myself, but we still need
				 to get ahold of the world so that we can deal with the palette"
			((pal := aPasteUpMorph standardPalette) notNil and: [pal isInWorld]) 
				ifTrue: 
					[(aPaintBox := aWorld paintBox) ifNotNil: [aPaintBox delete].
					pal viewMorph: self]
				ifFalse: 
					[(aPaintTab := (aWorld := self world) paintingFlapTab) 
						ifNotNil: [aPaintTab hideFlap]
						ifNil: [(aPaintBox := aWorld paintBox) ifNotNil: [aPaintBox delete]]]]! !

!SketchMorph methodsFor: 'menu' stamp: 'kfr 4/19/2015 15:58'!
emboss
 	
	Cursor wait showWhile:[^(SketchMorph withForm: (self form emboss))  openInHand]
	! !

!SketchMorph methodsFor: 'menu' stamp: 'cmm 8/2/2010 18:09'!
erasePixelsOfColor: aColor 
	"Let the user specifiy a color such that all pixels of that color should be erased; then do the erasure"
	| newBounds |
	originalForm
		mapColor: aColor
		to: Color transparent.
	newBounds := originalForm rectangleEnclosingPixelsNotOfColor: Color transparent.
	self form: (originalForm copy: newBounds)! !

!SketchMorph methodsFor: 'menu' stamp: 'cmm 8/2/2010 18:19'!
erasePixelsUsing: evt 
	"Let the user specifiy a color such that all pixels of that color should be erased; then do the erasure"
	self
		changeColorTarget: self
		selector: #rememberedColor:
		originalColor: nil
		hand: evt hand.
	self rememberedColor "color to erase"
		ifNil: [ ^ self ]
		ifNotNilDo:
			[ : chosenColor | self erasePixelsOfColor: chosenColor ]! !

!SketchMorph methodsFor: 'geometry' stamp: 'tfel 8/2/2016 11:21'!
extent: newExtent
	"Change my scale to fit myself into the given extent.
	Avoid extents where X or Y is zero."
	
	(newExtent x = 0 or: [newExtent y = 0]) ifTrue: [ ^self ].
	self extent = newExtent ifTrue:[^self].
	self scalePoint: newExtent asFloatPoint / (originalForm extent max: 1@1).
	self layoutChanged.
! !

!SketchMorph methodsFor: 'geometry' stamp: 'nk 6/21/2003 14:00'!
firstIntersectionWithLineFrom: start to: end
	| intersections last |
	intersections := self fullBounds extrapolatedIntersectionsWithLineFrom: start to: end.
	intersections size = 1 ifTrue: [ ^intersections anyOne ].
	intersections isEmpty ifTrue: [ ^nil ].
	intersections := intersections asSortedCollection: [ :a :b | (start dist: a) < (start dist: b) ].
	last := intersections first rounded.
	last pointsTo: intersections last rounded do: [ :pt |
		(self rotatedForm isTransparentAt: (pt - bounds origin)) ifFalse: [ ^last ].
		last := pt.
	].
	^intersections first rounded! !

!SketchMorph methodsFor: 'e-toy support' stamp: 'kks 11/15/2007 00:28'!
flipHorizontal

	|  r |
	r _ self rotationCenter.
	self left:  self left - (1.0 - (2 * r x) * self width).
	self form: (self form flipBy: #horizontal centerAt: self form center).
	self rotationCenter: (1 - r x) @ (r y).! !

!SketchMorph methodsFor: 'e-toy support' stamp: 'kks 11/15/2007 00:32'!
flipVertical

	|  r |
	r _ self rotationCenter.
	self top:  self top - (1.0 - (2 * r y) * self height).
	self form: (self form flipBy: #vertical centerAt: self form center).
	self rotationCenter:  r x @ (1 - r y).! !

!SketchMorph methodsFor: 'accessing' stamp: ''!
form

	^ originalForm
! !

!SketchMorph methodsFor: 'accessing' stamp: 'sw 12/13/2001 12:10'!
form: aForm
	"Set the receiver's form"

	| oldForm topRenderer |
	oldForm := originalForm.
	(self hasProperty: #baseGraphic) ifFalse: [self setProperty: #baseGraphic toValue: aForm].
	originalForm := aForm.
	self rotationCenter: 0.5@0.5.
	self layoutChanged.
	topRenderer := self topRendererOrSelf.

	oldForm ifNotNil: [topRenderer position: topRenderer position + (oldForm extent - aForm extent // 2)].
! !

!SketchMorph methodsFor: 'geometry eToy' stamp: 'ar 9/22/2000 22:31'!
forwardDirection: degrees
	"If not rotating normally, update my rotatedForm"
	super forwardDirection: degrees.
	rotationStyle == #normal ifFalse:[self layoutChanged].! !

!SketchMorph methodsFor: 'accessing' stamp: ''!
framesToDwell

	^ framesToDwell
! !

!SketchMorph methodsFor: 'accessing' stamp: 'jm 7/24/97 15:06'!
framesToDwell: anInteger

	framesToDwell := anInteger.
! !

!SketchMorph methodsFor: 'drawing' stamp: 'nk 8/18/2004 19:07'!
generateInterpolatedForm
	"Draw the given form onto the canvas using the Balloon 3D engine"
	| aCanvas extent |
	extent := (originalForm extent * scalePoint) asIntegerPoint.
	rotatedForm := Form extent: extent asIntegerPoint depth: originalForm depth.
	aCanvas := rotatedForm getCanvas.
	^self drawInterpolatedImage: originalForm on: aCanvas! !

!SketchMorph methodsFor: 'drawing' stamp: 'tpr 3/15/2013 12:43'!
generateRotatedForm
	"Compute my rotatedForm and offsetWhenRotated."

	| scalePt smoothPix pair |
	scalePoint ifNil: [scalePoint := 1 @ 1].
	scalePt := scalePoint x abs @ scalePoint y abs.
	rotationStyle == #none ifTrue: [scalePt := 1 @ 1].
	smoothPix := (scalePt x < 1.0 or: [scalePt y < 1.0]) 
		ifTrue: [2]
		ifFalse: [1].
	rotationStyle = #leftRight 
		ifTrue: 
			[self heading asSmallAngleDegrees < 0.0 
				ifTrue: [scalePt := scalePt x negated @ scalePt y]].
	rotationStyle = #upDown 
		ifTrue: 
			[self heading asSmallAngleDegrees abs > 90.0 
				ifTrue: [scalePt := scalePt x @ scalePt y negated]].
	rotatedForm := scalePt = (1 @ 1) 
				ifTrue: [originalForm]
				ifFalse: 
					["ar 11/19/2001: I am uncertain what happens in the case of rotationStyle ~~ normal"

					(rotationStyle == #normal and: [self useInterpolation]) 
						ifTrue: [^self generateInterpolatedForm].
					pair := WarpBlt 
								rotate: originalForm
								degrees: 0
								center: originalForm boundingBox center
								scaleBy: scalePt
								smoothing: smoothPix.
					pair first]! !

!SketchMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 21:51'!
initialize
"initialize the state of the receiver"
	^ self initializeWith: (ScriptingSystem formAtKey: 'Painting') deepCopy! !

!SketchMorph methodsFor: 'parts bin' stamp: 'sw 6/28/2001 11:33'!
initializeToStandAlone
	super initializeToStandAlone.
	self initializeWith: (ScriptingSystem formAtKey: 'Painting') deepCopy

! !

!SketchMorph methodsFor: 'initialization' stamp: 'yo 9/14/2006 18:54'!
initializeWith: aForm

	super initialize.
	originalForm _ aForm.
	rotationStyle _ #normal.		"styles: #normal, #leftRight, #upDown, or #none"
	scalePoint _ 1.0@1.0.
	framesToDwell _ 1.
	rotatedForm _ originalForm.	"cached rotation of originalForm"
	self extent: originalForm extent.
! !

!SketchMorph methodsFor: 'menu' stamp: 'wiz 12/4/2006 00:32'!
insertIntoMovie: evt

	| movies aTarget |
	movies :=
		(self world rootMorphsAt: evt hand targetPoint)
			select: [:m | ((m isKindOf: MovieMorph) or:
						 [m isSketchMorph]) and: [m ~= self]].
	movies isEmpty ifTrue: [^ self].
	aTarget := movies first.
	(aTarget isSketchMorph) ifTrue: [
		aTarget := aTarget replaceSelfWithMovie].
	aTarget insertFrames: (Array with: self).
	self delete.
! !

!SketchMorph methodsFor: 'halos and balloon help' stamp: 'sw 7/3/1999 20:06'!
isLikelyRecipientForMouseOverHalos
	^ true! !

!SketchMorph methodsFor: 'accessing' stamp: 'mt 4/2/2015 15:17'!
keepAspectRatio

	^ keepAspectRatio ifNil: [false]! !

!SketchMorph methodsFor: 'accessing' stamp: 'mt 4/2/2015 15:17'!
keepAspectRatio: aBoolean

	keepAspectRatio := aBoolean.! !

!SketchMorph methodsFor: 'layout' stamp: 'ar 9/22/2000 14:00'!
layoutChanged
	"Update rotatedForm and compute new bounds."
	self changed.
	self generateRotatedForm.
	bounds := bounds origin extent: rotatedForm extent.
	super layoutChanged.
	self changed.
! !

!SketchMorph methodsFor: 'other' stamp: 'sw 12/1/1998 18:16'!
newForm: aForm
	self originalForm: aForm.
	self layoutChanged! !

!SketchMorph methodsFor: 'accessing' stamp: 'sw 12/12/2001 10:49'!
nominalForm: aForm
	"Ascribe the blank nominal form"

	originalForm := aForm.
	self rotationCenter: 0.5@0.5.
	self layoutChanged
! !

!SketchMorph methodsFor: 'accessing' stamp: 'sw 9/9/1998 13:15'!
originalForm: aForm
	originalForm := aForm! !

!SketchMorph methodsFor: 'pen support' stamp: 'jm 4/22/1998 09:26'!
penOnMyForm
	"Support for experiments with drawing under program control. To get started, make a new SketchMorph in a morphic world. In an inspector, give it the desired pixel dimensions with clearExtent:fillColor:. Then use this method to get a pen to which you can send normal pen commands. Reveal the resulting drawing with revealPenStrokes."

	^ Pen newOnForm: originalForm
! !

!SketchMorph methodsFor: 'menu' stamp: 'cmm 8/2/2010 18:15'!
recolorPixelsOfColor: originalColor with: newColor
	"Let the user select a color to be remapped, and then a color to map that color to, then carry it out."
	| d newForm map |
	d := originalForm depth.
	newForm := Form extent: originalForm extent depth: d.
	map := (Color cachedColormapFrom: d to: d) copy.
	map at: (originalColor indexInMap: map) put: (newColor pixelValueForDepth: d).
	newForm copyBits: newForm boundingBox
		from: originalForm at: 0@0
		colorMap: map.
	self form: newForm.
! !

!SketchMorph methodsFor: 'menu' stamp: 'cmm 8/2/2010 18:18'!
recolorPixelsUsing: evt 
	"Let the user select a color to be remapped, and then a color to map that color to, then carry it out."
	| originalColor newColor |
	self inform: 'choose the color you want to replace' translated.
	self
		changeColorTarget: self
		selector: #rememberedColor:
		originalColor: nil
		hand: evt hand.
	"color to replace"
	originalColor := self rememberedColor ifNil: [ ^ self ].
	self inform: 'now choose the color you want to replace it with' translated.
	self
		changeColorTarget: self
		selector: #rememberedColor:
		originalColor: originalColor
		hand: evt hand.
	"new color"
	newColor := self rememberedColor ifNil: [ ^ self ].
	self
		recolorPixelsOfColor: originalColor
		with: newColor! !

!SketchMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 20:45'!
reduceColorPalette: evt
	"Let the user ask for a reduced number of colors in this sketch"

	| str nColors |
	str := UIManager default
		request: 'Please enter a number greater than one.
(note: this cannot be undone, so answer zero
to abort if you need to make a backup first)' translated
		initialAnswer: '256'.
	nColors := Integer readFrom: (ReadStream on: str).
	(nColors between: 2 and: 256) ifFalse: [^ self].

	originalForm := originalForm copyWithColorsReducedTo: nColors.
	rotatedForm := nil.
	self changed! !

!SketchMorph methodsFor: 'caching' stamp: 'di 3/2/98 14:14'!
releaseCachedState
	"Clear cache of rotated, scaled Form."

	super releaseCachedState.
	rotatedForm := nil.
	originalForm hibernate! !

!SketchMorph methodsFor: 'other' stamp: 'jm 7/24/97 15:06'!
replaceSelfWithMovie
	"Replace this SketchMorph in its owner with a MovieMorph containing this sketch as its only frame. This allows a SketchMorph to be turned into a MovieMorph by just insering additional frames."

	| o movie |
	self changed.
	o := self owner.
	movie := MovieMorph new position: self referencePosition.
	movie insertFrames: (Array with: self).
	o ifNil: [^ movie].
	o addMorphFront: movie.
	^ movie
! !

!SketchMorph methodsFor: 'menu' stamp: 'sw 12/12/2001 13:14'!
restoreBaseGraphic
	"Restore the receiver's base graphic"

	| aGraphic |
	((aGraphic := self baseGraphic) notNil and:
				[aGraphic ~= originalForm])
		ifTrue:
			[self form: aGraphic]! !

!SketchMorph methodsFor: 'menu' stamp: 'yo 3/15/2005 13:57'!
restoreBaseGraphicFromMenu
	"Restore the base graphic -- invoked from a menu, so give interactive feedback if appropriate"

	self isInWorld ifFalse: [^ self inform: 

'oops, this menu is a for a morph that
has been replaced, probably because a
"look like" script was run.  Please dismiss
the menu and get a new one!!.  Sorry!!' translated].

	 self baseGraphic = originalForm ifTrue: [^ self inform: 'This object is *already* showing its baseGraphic' translated].
	self restoreBaseGraphic! !

!SketchMorph methodsFor: 'pen support' stamp: 'jm 4/22/1998 09:08'!
revealPenStrokes
	"This message must be sent after a sequence of pen strokes to make the resulting changes visible."

	rotatedForm := nil.
	self changed.
! !

!SketchMorph methodsFor: 'accessing' stamp: ''!
rotatedForm

	rotatedForm ifNil: [self layoutChanged].
	^ rotatedForm
! !

!SketchMorph methodsFor: 'e-toy support' stamp: ''!
rotationStyle

	^ rotationStyle
! !

!SketchMorph methodsFor: 'e-toy support' stamp: 'bf 6/23/2012 20:35'!
rotationStyle: aSymbol
	"Set my rotation style to #normal, #leftRight, #upDown, or #none. Styles mean:
		#normal		-- continuous 360 degree rotation
		#leftRight		-- quantize angle to left or right facing
		#upDown		-- quantize angle to up or down facing
		#none			-- do not rotate
	Because my rendering code flips the form (see generateRotatedForm) we 'pre-flip' it here to preserve the same visual appearance.
	"

	| wasFlippedX wasFlippedY isFlippedX isFlippedY |
	wasFlippedX := rotationStyle == #leftRight
		and: [ self heading asSmallAngleDegrees < 0.0 ].
	wasFlippedY := rotationStyle == #upDown
		and: [ self heading asSmallAngleDegrees abs > 90.0 ].

	rotationStyle _ aSymbol.

	isFlippedX := rotationStyle == #leftRight
		and: [ self heading asSmallAngleDegrees < 0.0 ].
	isFlippedY := rotationStyle == #upDown
		and: [ self heading asSmallAngleDegrees abs > 90.0 ].

	wasFlippedX == isFlippedX
		ifFalse: [self form: (self form flipBy: #horizontal centerAt: self form center)].
	wasFlippedY == isFlippedY
		ifFalse: [self form: (self form flipBy: #vertical centerAt: self form center)].

	self layoutChanged.
! !

!SketchMorph methodsFor: 'accessing' stamp: 'gm 2/22/2003 13:14'!
scaleFactor
	"Answer the number representing my scaleFactor, assuming the receiver to be unflexed (if flexed, the renderer's scaleFactor is called instead"

	| qty |
	((qty := self scalePoint) isPoint) ifTrue: [^1.0].
	^qty! !

!SketchMorph methodsFor: 'accessing' stamp: 'jm 7/24/97 15:06'!
scalePoint

	scalePoint ifNil: [scalePoint := 1.0@1.0].
	^ scalePoint
! !

!SketchMorph methodsFor: 'accessing' stamp: 'mt 4/2/2015 15:23'!
scalePoint: aPoint

	scalePoint := self keepAspectRatio
		ifTrue: [aPoint max: aPoint transposed]
		ifFalse: [aPoint].
	self layoutChanged.! !

!SketchMorph methodsFor: 'accessing' stamp: 'sw 10/24/2000 05:55'!
setNewFormFrom: formOrNil
	"Set the receiver's form as indicated.   If nil is provided, then a default form will be used, possibly retrieved from the receiver's defaultValue property"

	| defaultImage |
	formOrNil ifNotNil: [^ self form: formOrNil].
	defaultImage := self defaultValueOrNil ifNil: [ScriptingSystem squeakyMouseForm].
	self form: defaultImage
! !

!SketchMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 19:09'!
setRotationStyle
	| selections labels sel reply |
	selections := #(normal leftRight upDown none).
	labels := #('rotate smoothly' 'left-right flip only' 'top-down flip only' 'don''t rotate').
	sel := labels at: (selections indexOf: self rotationStyle ifAbsent:[1]).
	labels := labels collect:[:lbl| sel = lbl ifTrue:['<on>', lbl translated] ifFalse:['<off>', lbl translated]].
	reply := UIManager default chooseFrom: labels values: selections.
	reply ifNotNil: [self rotationStyle: reply].
! !

!SketchMorph methodsFor: 'menu' stamp: 'kfr 4/19/2015 15:59'!
sharpen

	Cursor wait showWhile:[^(SketchMorph withForm: (self form sharpen))  openInHand]
	! !

!SketchMorph methodsFor: 'menu' stamp: 'ar 11/19/2001 22:38'!
toggleInterpolation
	^self useInterpolation: self useInterpolation not! !

!SketchMorph methodsFor: 'accessing' stamp: 'ar 11/16/2002 19:22'!
useInterpolation
	^(self valueOfProperty: #useInterpolation ifAbsent:[false]) 
		and:[Smalltalk includesKey: #B3DRenderEngine]! !

!SketchMorph methodsFor: 'menu' stamp: 'dgd 9/6/2003 18:27'!
useInterpolationString
	^ (self useInterpolation
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'smooth image' translated! !

!SketchMorph methodsFor: 'accessing' stamp: 'nk 1/24/2004 23:46'!
useInterpolation: aBool
	(aBool == true and: [ Smalltalk includesKey: #B3DRenderEngine ])
		ifTrue:[self setProperty: #useInterpolation toValue: aBool]
		ifFalse:[self removeProperty: #useInterpolation].
	self layoutChanged. "to regenerate the form"
! !

!SketchMorph methodsFor: 'halos and balloon help' stamp: 'ar 11/29/2001 19:51'!
wantsDirectionHandles
	^self valueOfProperty: #wantsDirectionHandles ifAbsent:[
		Preferences showDirectionHandles or:[Preferences showDirectionForSketches]]! !

!SketchMorph methodsFor: 'halos and balloon help' stamp: 'ar 11/29/2001 19:52'!
wantsDirectionHandles: aBool
	aBool == (Preferences showDirectionHandles or:[Preferences showDirectionForSketches])
		ifTrue:[self removeProperty: #wantsDirectionHandles]
		ifFalse:[self setProperty: #wantsDirectionHandles toValue: aBool].! !

!SketchMorph methodsFor: 'e-toy support' stamp: 'dgd 8/29/2004 11:05'!
wantsRecolorHandle
	"Answer whether the receiver would like a recolor handle to be  
	put up for it. We'd want to disable this but for the moment  
	that would cut off access to the button part of the properties  
	sheet. So this remains a loose end."
	^ false! !

!SketchMorph methodsFor: 'accessing' stamp: 'nk 6/12/2004 09:32'!
wantsSimpleSketchMorphHandles
	"Answer true if my halo's simple handles should include the simple sketch morph handles."
	^self isMemberOf: SketchMorph! !

Morph subclass: #StringMorph
	instanceVariableNames: 'font emphasis contents hasFocus'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!StringMorph commentStamp: 'efc 3/7/2003 17:34' prior: 0!
StringMorph is a "lightweight" Morph to display a String. It supports only a single font, color, and emphasis combination. For multiple text styles, use TextMorph.

Structure:
instance var    	Type              Description 
font 			StrikeFont 		(normally nil; then the accessor #font gives back TextStyle 
				or nil			defaultFont) 
emphasis 		SmallInteger	bitmask determining character attributes (underline, bold, 								italics, narrow, struckout) 
contents 		String 			The text that will be displayed. 
hasFocus 		Boolean 		Do I have the keyboard focus or not? 

If you shift-click on a StringMorph you can edit its string. This is accomplished the following way: StringMorph can launch a StringMorphEditor if it receives a #mouseDown event.

A StringMorph may also be used like a SimpleButtonMorph to do an action when clicked. Use the menu 'extras' / 'add mouseUpAction'.

The following propery will be defined:
aStringMorph valueOfProperty: #mouseUpCodeToRun!

StringMorph subclass: #IndentingListItemMorph
	instanceVariableNames: 'indentLevel isExpanded complexContents firstChild container nextSibling icon backgroundColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Explorer'!

!IndentingListItemMorph commentStamp: '<historical>' prior: 0!
An IndentingListItemMorph is a StringMorph that draws itself with an optional toggle at its left, as part of the display of the SimpleHierarchicalListMorph.

It will also display lines around the toggle if the #showLinesInHierarchyViews Preference is set.

Instance variables:

indentLevel <SmallInteger> 	the indent level, from 0 at the root and increasing by 1 at each level of the hierarchy.

isExpanded <Boolean>		true if this item is expanded (showing its children)

complexContents <ListItemWrapper>	an adapter wrapping my represented item that can answer its children, etc.
	
firstChild <IndentingListItemMorph|nil>	my first child, or nil if none
	
container <SimpleHierarchicalListMorph>	my container
	
nextSibling <IndentingListItemMorph|nil>	the next item in the linked list of siblings, or nil if none.

Contributed by Bob Arning as part of the ObjectExplorer package.
Don't blame him if it's not perfect.  We wanted to get it out for people to play with.!

!IndentingListItemMorph class methodsFor: 'defaults' stamp: 'mt 3/6/2015 11:36'!
iconColumnIndex
	"Hack. For now, say who gets the icon here. We need a generic solution for icons in multi-column trees. PluggableTreeMorph does something in that direction."
	^ 2! !

!IndentingListItemMorph methodsFor: 'drag and drop' stamp: 'nk 6/12/2004 16:49'!
acceptDroppingMorph: toDrop event: evt
	complexContents acceptDroppingObject: toDrop complexContents.
	toDrop delete.
	self highlightForDrop: false.! !

!IndentingListItemMorph methodsFor: 'container protocol - private' stamp: 'RAA 7/30/2000 19:49'!
addChildrenForList: hostList addingTo: morphList withExpandedItems: expandedItems

	firstChild ifNotNil: [
		firstChild withSiblingsDo: [ :aNode | aNode delete].
	].
	firstChild := nil.
	complexContents hasContents ifFalse: [^self].
	firstChild := hostList 
		addMorphsTo: morphList
		from: complexContents contents 
		allowSorting: true
		withExpandedItems: expandedItems
		atLevel: indentLevel + 1.
	! !

!IndentingListItemMorph methodsFor: 'filtering' stamp: 'mt 3/13/2015 10:10'!
applyFilter: filter

	self
		applyFilter: filter
		depthOffset: self indentLevel.! !

!IndentingListItemMorph methodsFor: 'filtering' stamp: 'mt 3/13/2015 10:27'!
applyFilter: filter depthOffset: offset

	| selfMatch childMatch |
	self isExpanded ifTrue: [self toggleExpandedState].
	
	selfMatch := self matches: filter.
	childMatch := self matchesAnyChild: filter depthOffset: offset.
	
	selfMatch | childMatch ifFalse: [^ self hide].
	
	selfMatch ifTrue: [
		self backgroundColor: ((Color gray: 0.85) alpha: 0.5)].
	childMatch ifTrue: [
		self toggleExpandedState.
		self childrenDo: [:child | child applyFilter: filter depthOffset: offset]].! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 3/8/2015 17:05'!
backgroundColor
	^ backgroundColor! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 3/8/2015 17:06'!
backgroundColor: aColor
	backgroundColor := aColor.
	self changed.! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'RAA 7/21/2000 11:00'!
balloonText

	^complexContents balloonText ifNil: [super balloonText]! !

!IndentingListItemMorph methodsFor: 'halos and balloon help' stamp: 'RAA 7/21/2000 11:13'!
boundsForBalloon

	"some morphs have bounds that are way too big"
	container ifNil: [^super boundsForBalloon].
	^self boundsInWorld intersect: container boundsInWorld! !

!IndentingListItemMorph methodsFor: 'testing' stamp: 'RAA 3/31/1999 17:44'!
canExpand

	^complexContents hasContents! !

!IndentingListItemMorph methodsFor: 'private' stamp: 'mt 4/8/2015 19:36'!
charactersOccluded
	"Answer the number of characters occluded in my #visibleList by my right edge."
	| listIndex leftEdgeOfRightmostColumn eachString indexOfLastVisible iconWidth totalWidth |
	listIndex := 0.
	leftEdgeOfRightmostColumn := container columns
		ifNil: [ 0 ]
		ifNotNil:
			[ : cols | (1 to: cols size - 1)
				inject: 0
				into:
					[ : sum : each | sum + (self widthOfColumn: each) ] ].
	eachString := container columns
		ifNil: [ self complexContents asString ]
		ifNotNil:
			[ : cols | self contentsAtColumn: container columns size ].
	iconWidth := self icon
		ifNil: [ 0 ]
		ifNotNil:
			[ : icon | icon width + 2 ].
	totalWidth := self toggleBounds right.
	indexOfLastVisible := ((1 to: eachString size)
		detect:
			[ : stringIndex | (totalWidth:=totalWidth+(self fontToUse widthOf: (eachString at: stringIndex))) >
				(container width -
					(container vIsScrollbarShowing
						ifTrue: [ container vScrollBar width ]
						ifFalse: [ 0 ]) - iconWidth - leftEdgeOfRightmostColumn) ]
		ifNone: [ eachString size + 1 ]) - 1.
	^ eachString size - indexOfLastVisible! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'panda 4/28/2000 15:30'!
children
	| children |
	children := OrderedCollection new.
	self childrenDo: [:each | children add: each].
	^children! !

!IndentingListItemMorph methodsFor: 'enumeration' stamp: 'ul 8/2/2011 20:02'!
childrenDo: aBlock

	firstChild ifNotNil: [
		firstChild withSiblingsDo: aBlock ]! !

!IndentingListItemMorph methodsFor: 'container protocol' stamp: 'mt 3/29/2015 14:20'!
collapse

	self isExpanded ifFalse: [^ self].
	
	self isExpanded: false.
	
	firstChild ifNotNil: [:collapsingNode |
	 	| toDelete |
		toDelete := OrderedCollection new.
		collapsingNode withSiblingsDo: [:aNode | aNode recursiveAddTo: toDelete].
		container noteRemovalOfAll: toDelete].
	
	self changed.! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'mt 7/26/2016 13:05'!
colorToUse

	^ (self valueOfProperty: #wasRefreshed ifAbsent: [false])
		ifTrue: [complexContents highlightColor ifNil: [self highlightTextColor]]
		ifFalse: [
			self isSelected ifTrue: [^ self selectionTextColor].			
			complexContents preferredColor ifNil: [self color]]! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'RAA 7/11/1998 14:34'!
complexContents

	^complexContents! !

!IndentingListItemMorph methodsFor: 'accessing - columns' stamp: 'mt 3/6/2015 10:48'!
contentsAtColumn: index
	"Split string contents at <tab> character."
	
	| column scanner cell |
	column := 0.
	scanner := ReadStream on: contents asString.
	[(cell := scanner upTo: Character tab) notEmpty]
		whileTrue: [column := column + 1. column = index ifTrue: [^ cell]].
	^ ''! !

!IndentingListItemMorph methodsFor: 'accessing - columns' stamp: 'mt 3/6/2015 10:48'!
contentsSplitByColumns
	"Split string contents at <tab> character."
	
	| result scanner cell |
	result := OrderedCollection new.
	scanner := ReadStream on: contents asString.
	[(cell := scanner upTo: Character tab) notEmpty]
		whileTrue: [result add: cell].
	^ result! !

!IndentingListItemMorph methodsFor: 'initialization' stamp: 'ul 9/3/2012 14:05'!
defaultColor

	^complexContents
		ifNil: [ super defaultColor ]
		ifNotNil: [ complexContents preferredColor ]! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 10:36'!
drawHoverOn: aCanvas 

	aCanvas
		fillRectangle: self bounds
		color: self hoverColor.! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'topa 3/2/2016 00:03'!
drawLinesOn: aCanvas lineColor: lineColor indentThreshold: aNumber
	self indentLevel > aNumber ifTrue: [
		| hasToggle |
		hasToggle := self hasToggle.
		"Draw line from toggle to text"
		self drawLineToggleToTextOn: aCanvas lineColor: lineColor hasToggle: hasToggle.

		"Draw the line from my toggle to the nextSibling's toggle"
		self nextVisibleSibling ifNotNil: [ self drawLinesToNextSiblingOn: aCanvas lineColor: lineColor hasToggle: hasToggle ]].

	"If I have children and am expanded, draw a line to my first child"
	((self firstVisibleChild notNil
		and: [self isExpanded])
		and: [self firstVisibleChild indentLevel > aNumber])
			ifTrue: [ self drawLinesToFirstChildOn: aCanvas lineColor: lineColor ]! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'mt 3/8/2015 18:24'!
drawLinesToFirstChildOn: aCanvas lineColor: lineColor 
	"Draw line from me to next sibling"

	| child vLineX vLineTop vLineBottom childBounds childCenter |
	child := self firstVisibleChild.
	childBounds := child toggleBounds.
	childCenter := childBounds center.
	vLineX := childCenter x - 1.
	vLineTop := bounds bottom.
	child hasToggle
		ifTrue: [vLineBottom := childCenter y - 7]
		ifFalse: [vLineBottom := childCenter y].
	aCanvas
		line: vLineX @ vLineTop
		to: vLineX @ vLineBottom
		width: 1
		color: lineColor! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'mt 3/8/2015 18:24'!
drawLinesToNextSiblingOn: aCanvas lineColor: lineColor hasToggle: hasToggle
	| myBounds nextSibBounds vLineX myCenter vLineTop vLineBottom |
	myBounds := self toggleBounds.
	nextSibBounds := self nextVisibleSibling toggleBounds.
	myCenter := myBounds center.
	vLineX := myCenter x - 1.
	vLineTop := myCenter y.
	vLineBottom := nextSibBounds center y.
	"Draw line from me to next sibling"
	aCanvas
		line: vLineX @ vLineTop
		to: vLineX @ vLineBottom
		width: 1
		color: lineColor! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'dgd 9/25/2004 22:47'!
drawLineToggleToTextOn: aCanvas lineColor: lineColor hasToggle: hasToggle
	"If I am not the only item in my container, draw the line between:
		- my toggle (if any) or my left edge (if no toggle)
		- and my text left edge"

	| myBounds myCenter hLineY hLineLeft |
	self isSoleItem ifTrue: [ ^self ].
	myBounds := self toggleBounds.
	myCenter := myBounds center.
	hLineY := myCenter y.
	hLineLeft := myCenter x - 1.
	"Draw line from toggle to text"
	aCanvas
		line: hLineLeft @ hLineY
		to: myBounds right + 0 @ hLineY
		width: 1
		color: lineColor! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 10:05'!
drawOn: aCanvas
	
	| tRect sRect columnScanner columnLeft |
	self backgroundColor ifNotNil: [:c |
		aCanvas fillRectangle: self innerBounds color: c].

	tRect := self toggleRectangle.	
	self drawToggleOn: aCanvas in: tRect.

	sRect := bounds withLeft: tRect right + self hMargin.
	sRect := sRect top: sRect top + sRect bottom - self fontToUse height // 2.	
	
	(container columns isNil or: [(contents asString indexOf: Character tab) = 0]) ifTrue: [
		icon ifNotNil: [
			aCanvas
				translucentImage: icon
				at: sRect left @ (self top + (self height - icon height // 2)).
			sRect := sRect left: sRect left + icon width + 2.
		].
		
		aCanvas drawString: contents asString in: sRect font: self fontToUse color: self colorToUse.
	
	] ifFalse: [
		columnLeft := sRect left.
		columnScanner := ReadStream on: contents asString.
		container columns withIndexDo: [ :widthSpec :column | | columnRect columnData columnWidth |
			"Draw icon."
			column = self class iconColumnIndex ifTrue: [
				icon ifNotNil: [
					aCanvas
						translucentImage: icon
						at: columnLeft @ (self top + (self height - icon height // 2)).
					columnLeft := columnLeft + icon width + 2]].

			columnWidth := self widthOfColumn: column.
			columnRect := columnLeft @ sRect top extent: columnWidth @ sRect height.
			columnData := columnScanner upTo: Character tab.
			
			"Draw string."
			columnData ifNotEmpty: [
				aCanvas drawString: columnData in: columnRect font: self fontToUse color: self colorToUse].

			"Compute next column offset."			
			columnLeft := columnRect right + 5.
			column = 1 ifTrue: [columnLeft := columnLeft - tRect right + self left].
			
		].
	]! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 11:43'!
drawSelectionOn: aCanvas 

	| fill |
	fill := self selectionColor isColor
		ifTrue: [SolidFillStyle color: self selectionColor]
		ifFalse: [self selectionColor].
	fill isGradientFill ifTrue: [
		fill origin: self topLeft.
		fill direction: 0@ self height].

	aCanvas
		fillRectangle: self bounds
		fillStyle: fill.! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'nk 2/19/2004 18:19'!
drawToggleOn: aCanvas in: aRectangle

	| aForm centeringOffset |
	complexContents hasContents ifFalse: [^self].
	aForm := isExpanded 
		ifTrue: [container expandedForm]
		ifFalse: [container notExpandedForm].
	centeringOffset := ((aRectangle height - aForm extent y) / 2.0) rounded.
	^aCanvas 
		paintImage: aForm 
		at: (aRectangle topLeft translateBy: 0 @ centeringOffset).
! !

!IndentingListItemMorph methodsFor: 'container protocol' stamp: 'mt 3/29/2015 14:20'!
expand

 	| newChildren c |

	(self isExpanded or: [self canExpand not])
		ifTrue: [^ self].
	
	(c := self getChildren) ifEmpty: [
		"Due to the guessing in #canExpand, it may still fail here."
		^ self].

	self isExpanded: true.

	newChildren := container 
		addSubmorphsAfter: self 
		fromCollection: c 
		allowSorting: true.

	firstChild := newChildren first.! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:24'!
filterColor

	^ self valueOfProperty: #filterColor ifAbsent: [Color yellow]! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 11:41'!
filterColor: aColor

	| cc fill |
	cc := aColor.

	MenuMorph gradientMenu
		ifFalse: [fill := SolidFillStyle color: cc]
		ifTrue: [
			fill := GradientFillStyle ramp: { 
				0.0 -> cc twiceLighter. 
				1 -> cc twiceDarker }].
			
	self setProperty: #filterColor toValue: fill.! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:25'!
filterTextColor

	^ self valueOfProperty: #filterTextColor ifAbsent: [Color black]! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:25'!
filterTextColor: aColor

	self setProperty: #filterTextColor toValue: aColor.! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'RAA 8/2/1999 16:48'!
firstChild

	^firstChild! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 3/8/2015 18:23'!
firstVisibleChild

	^ self firstChild ifNotNil: [:c |
		c visible ifTrue: [c] ifFalse: [c nextVisibleSibling]]! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/26/2016 12:58'!
fitContents

	super fitContents.
	self width: container preferredSubmorphWidth.! !

!IndentingListItemMorph methodsFor: 'model access' stamp: 'mt 3/8/2015 17:36'!
getChildren

	^ self getChildrenFor: complexContents! !

!IndentingListItemMorph methodsFor: 'model access' stamp: 'mt 3/8/2015 17:36'!
getChildrenFor: model

	^ model contents! !

!IndentingListItemMorph methodsFor: 'model access' stamp: 'mt 3/8/2015 17:17'!
getIcon

	^ complexContents icon! !

!IndentingListItemMorph methodsFor: 'model access' stamp: 'mt 3/8/2015 17:23'!
getLabel

	^ self getLabelFor: complexContents! !

!IndentingListItemMorph methodsFor: 'model access' stamp: 'mt 3/8/2015 17:23'!
getLabelFor: model

	^ model asString! !

!IndentingListItemMorph methodsFor: 'testing' stamp: 'dgd 9/25/2004 22:25'!
hasIcon
	"Answer whether the receiver has an icon."
	^ icon notNil! !

!IndentingListItemMorph methodsFor: 'private' stamp: 'nk 2/19/2004 18:29'!
hasToggle
	^ complexContents hasContents! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 10:04'!
highlight

	self changed.
	
! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:46'!
highlightTextColor

	^ self valueOfProperty: #highlightTextColor ifAbsent: [Color red]! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:46'!
highlightTextColor: aColor

	self setProperty: #highlightTextColor toValue: aColor.! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'topa 3/2/2016 00:07'!
hMargin

	^ 3! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:38'!
hoverColor

	^ self valueOfProperty: #hoverColor ifAbsent: [Color veryLightGray]! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:33'!
hoverColor: aColor

	self setProperty: #hoverColor toValue: aColor.! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'dgd 9/25/2004 22:27'!
icon
	"answer the receiver's icon"
	^ icon! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'RAA 7/30/2000 19:15'!
indentLevel

	^indentLevel! !

!IndentingListItemMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:42'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	indentLevel := 0.
	isExpanded := false! !

!IndentingListItemMorph methodsFor: 'initialization' stamp: 'mt 3/29/2015 14:25'!
initWithContents: anObject prior: priorMorph forList: hostList indentLevel: newLevel

	container := hostList.
	
	complexContents := anObject.
	anObject addDependent: self.
	
	self initWithContents: self getLabel font: Preferences standardListFont emphasis: nil.
	indentLevel := 0.
	isExpanded := false.
 	nextSibling := firstChild := nil.
	priorMorph ifNotNil: [
		priorMorph nextSibling: self.
	].
	indentLevel := newLevel.
	icon := self getIcon.
	self extent: self minWidth @ self minHeight! !

!IndentingListItemMorph methodsFor: 'mouse events' stamp: 'ar 3/17/2001 17:32'!
inToggleArea: aPoint

	^self toggleRectangle containsPoint: aPoint! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'RAA 7/31/1998 00:30'!
isExpanded

	^isExpanded! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'RAA 7/31/1998 00:48'!
isExpanded: aBoolean

	isExpanded := aBoolean! !

!IndentingListItemMorph methodsFor: 'testing' stamp: 'nk 3/8/2004 09:14'!
isFirstItem
	^owner submorphs first == self! !

!IndentingListItemMorph methodsFor: 'testing' stamp: 'mt 7/26/2016 10:35'!
isSelected

	^ container ifNil: [false] ifNotNil: [container selectedMorph == self]! !

!IndentingListItemMorph methodsFor: 'testing' stamp: 'nk 3/8/2004 09:15'!
isSoleItem
	^self isFirstItem and: [ owner submorphs size = 1 ]! !

!IndentingListItemMorph methodsFor: 'filtering' stamp: 'mt 3/13/2015 10:11'!
matchesAnyChild: pattern depthOffset: offset

	| maxDepth next current |
	maxDepth := PluggableTreeMorph maximumSearchDepth - self indentLevel + offset.
	maxDepth <= 0 ifTrue: [^ false].
	
	next := (self getChildren collect: [:obj | 1 -> obj]) asOrderedCollection.
	[next notEmpty] whileTrue: [
		current := next removeFirst.
		
		(self matches: pattern in: current value)
			ifTrue: [^ true].
		
		current key < maxDepth ifTrue: [
			next addAll: ((self getChildrenFor: current value) collect: [:obj | (current key + 1) -> obj])].
		].
	
	^ false! !

!IndentingListItemMorph methodsFor: 'filtering' stamp: 'mt 3/8/2015 17:37'!
matches: pattern

	^ self matches: pattern in: complexContents! !

!IndentingListItemMorph methodsFor: 'filtering' stamp: 'cmm 3/10/2015 16:37'!
matches: pattern in: model 
	^ ((PluggableTreeMorph filterByLabelsOnly
		ifTrue: [ model itemName ]
		ifFalse: [ self getLabelFor: model ])
			findString: pattern
			startingAt: 1
			caseSensitive: false) > 0! !

!IndentingListItemMorph methodsFor: 'layout' stamp: 'dgd 9/25/2004 22:26'!
minHeight
	| iconHeight |
	iconHeight := self hasIcon
				ifTrue: [self icon height + 2]
				ifFalse: [0].
	^ self fontToUse height max: iconHeight ! !

!IndentingListItemMorph methodsFor: 'layout' stamp: 'dgd 9/25/2004 22:26'!
minWidth
	| iconWidth |
	iconWidth := self hasIcon
				ifTrue: [self icon width + 2]
				ifFalse: [0].
	^ (self fontToUse widthOfString: contents)
		+ iconWidth ! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'RAA 7/11/1998 12:15'!
nextSibling

	^nextSibling! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'RAA 8/1/1998 01:05'!
nextSibling: anotherMorph

	nextSibling := anotherMorph! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 3/8/2015 18:25'!
nextVisibleSibling

	| m |
	m := self nextSibling.
	[m isNil or: [m visible]] whileFalse: [
		m := m nextSibling].
	^ m! !

!IndentingListItemMorph methodsFor: 'container protocol - private' stamp: 'nk 10/14/2003 10:36'!
openPath: anArray 
	| found |
	anArray isEmpty
		ifTrue: [^ container setSelectedMorph: nil].
	found := nil.
	self
		withSiblingsDo: [:each | found
				ifNil: [(each complexContents asString = anArray first
							or: [anArray first isNil])
						ifTrue: [found := each]]].
	found
		ifNil: ["try again with no case sensitivity"
			self
				withSiblingsDo: [:each | found
						ifNil: [(each complexContents asString sameAs: anArray first)
								ifTrue: [found := each]]]].
	found
		ifNotNil: [found isExpanded
				ifFalse: [found toggleExpandedState.
					container adjustSubmorphPositions].
			found changed.
			anArray size = 1
				ifTrue: [^ container setSelectedMorph: found].
			^ found firstChild
				ifNil: [container setSelectedMorph: nil]
				ifNotNil: [found firstChild openPath: anArray allButFirst]].
	^ container setSelectedMorph: nil! !

!IndentingListItemMorph methodsFor: 'accessing - columns' stamp: 'mt 3/6/2015 10:48'!
preferredColumnCount

	^ self contentsSplitByColumns size! !

!IndentingListItemMorph methodsFor: 'accessing - columns' stamp: 'mt 3/6/2015 11:34'!
preferredWidthOfColumn: index

	^ (self fontToUse widthOfString: (self contentsAtColumn: index)) + 
		(index = 1 ifTrue: [self toggleRectangle right - self left] ifFalse: [0])! !

!IndentingListItemMorph methodsFor: 'container protocol - private' stamp: 'RAA 6/21/1999 14:54'!
recursiveAddTo: aCollection

	firstChild ifNotNil: [
		firstChild withSiblingsDo: [ :aNode | aNode recursiveAddTo: aCollection].
	].
	aCollection add: self
	! !

!IndentingListItemMorph methodsFor: 'container protocol - private' stamp: 'RAA 4/2/1999 18:02'!
recursiveDelete

	firstChild ifNotNil: [
		firstChild withSiblingsDo: [ :aNode | aNode recursiveDelete].
	].
	self delete
	! !

!IndentingListItemMorph methodsFor: 'initialization' stamp: 'mt 7/26/2016 12:59'!
refresh

	self contents: self getLabel.
	icon := self getIcon.
	
	(self valueOfProperty: #wasRefreshed ifAbsent: [false]) ifFalse: [
		self setProperty: #wasRefreshed toValue: true].! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:26'!
selectionColor

	^ self valueOfProperty: #selectionColor ifAbsent: [Color blue]! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 11:42'!
selectionColor: aColor

	| cc fill |
	cc := aColor.

	MenuMorph gradientMenu
		ifFalse: [fill := SolidFillStyle color: cc]
		ifTrue: [
			fill := GradientFillStyle ramp: { 
				0.0 -> cc twiceLighter. 
				1 -> cc twiceDarker }].

	self setProperty: #selectionColor toValue: fill.! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:26'!
selectionTextColor

	^ self valueOfProperty: #selectionTextColor ifAbsent: [Color white]! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'mt 7/24/2016 10:26'!
selectionTextColor: aColor

	self setProperty: #selectionTextColor toValue: aColor.! !

!IndentingListItemMorph methodsFor: 'private' stamp: 'nk 12/5/2002 15:16'!
toggleBounds
	^self toggleRectangle! !

!IndentingListItemMorph methodsFor: 'container protocol' stamp: 'mt 3/29/2015 14:21'!
toggleExpandedState

	self isExpanded
		ifTrue: [self collapse]
		ifFalse: [self expand].! !

!IndentingListItemMorph methodsFor: 'private' stamp: 'mt 4/1/2015 17:21'!
toggleRectangle

	| h |
	h := bounds height.
	^(bounds left + self hMargin + (12 * indentLevel)) @ bounds top extent: 12@h! !

!IndentingListItemMorph methodsFor: 'drawing' stamp: 'mt 7/24/2016 10:04'!
unhighlight

	self changed.
	
	
! !

!IndentingListItemMorph methodsFor: 'updating' stamp: 'mt 5/14/2015 18:38'!
update: aspect
	"See ListItemWrapper and subclasses for possible change aspects."
	
	aspect = #contents ifTrue: [
		self isExpanded ifTrue: [self toggleExpandedState].
		self canExpand ifTrue: [self toggleExpandedState].
		container adjustSubmorphPositions].
		
	super update: aspect.! !

!IndentingListItemMorph methodsFor: 'accessing' stamp: 'bf 2/9/2004 10:55'!
userString
	"Add leading tabs to my userString"
	^ (String new: indentLevel withAll: Character tab), super userString
! !

!IndentingListItemMorph methodsFor: 'accessing - columns' stamp: 'mt 5/1/2016 11:24'!
widthOfColumn: columnIndex 

	^ (container widthOfColumn: columnIndex) ifNil: [self width]! !

!IndentingListItemMorph methodsFor: 'converting' stamp: 'RAA 3/31/1999 12:13'!
withoutListWrapper

	^complexContents withoutListWrapper! !

!IndentingListItemMorph methodsFor: 'private' stamp: 'RAA 7/11/1998 14:25'!
withSiblingsDo: aBlock

	| node |
	node := self.
	[node isNil] whileFalse: [
		aBlock value: node.
		node := node nextSibling
	].! !

StringMorph subclass: #MenuItemMorph
	instanceVariableNames: 'isEnabled subMenu isSelected target selector arguments icon'
	classVariableNames: 'SubMenuMarker'
	poolDictionaries: ''
	category: 'Morphic-Menus'!

!MenuItemMorph commentStamp: '<historical>' prior: 0!
I represent an item in a menu.

Instance variables:
	isEnabled 	<Boolean>	True if the menu item can be executed.
	subMenu 	<MenuMorph | nil>	The submenu to activate automatically when the user mouses over the item.
	isSelected 	<Boolean>	True if the item is currently selected.
	target 		<Object>		The target of the associated action.
	selector 		<Symbol>	The associated action.
	arguments 	<Array>		The arguments for the associated action.
	icon		<Form | nil>	An optional icon form to be displayed to my left.

If I have a dynamic marker, created by strings like <yes> or <no> in my contents, it will be installed as a submorph.!

MenuItemMorph subclass: #DockingBarItemMorph
	instanceVariableNames: 'selectedIcon'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Menus-DockingBar'!

!DockingBarItemMorph methodsFor: 'selecting' stamp: 'mt 7/26/2016 13:10'!
adjacentTo

	| roundedCornersOffset verticalOffset |
	roundedCornersOffset := MenuMorph roundedMenuCorners
		ifTrue: [Morph preferredCornerRadius negated]
		ifFalse: [0].
	verticalOffset := 2.

	owner isFloating
		ifTrue: [^ {self bounds bottomLeft + (roundedCornersOffset @ verticalOffset)}].
	owner isAdheringToTop
		ifTrue: [^ {self bounds bottomLeft + (roundedCornersOffset @ verticalOffset)}].
	owner isAdheringToLeft
		ifTrue: [^ {self bounds topRight + (roundedCornersOffset @ verticalOffset)}].
	owner isAdheringToBottom
		ifTrue: [^ {self bounds topLeft + (roundedCornersOffset @ verticalOffset)}].
	owner isAdheringToRight
		ifTrue: [^ {self bounds topLeft + (roundedCornersOffset @ verticalOffset negated)}].
	^ {self bounds bottomLeft + (roundedCornersOffset @ 5)}! !

!DockingBarItemMorph methodsFor: 'private' stamp: 'kb 2/4/2010 14:15'!
createSubmenu

	^DockingBarMenuMorph new! !

!DockingBarItemMorph methodsFor: 'private' stamp: 'kb 2/4/2010 14:15'!
createUpdatingSubmenu

	^DockingBarUpdatingMenuMorph new! !

!DockingBarItemMorph methodsFor: 'drawing' stamp: 'mt 7/9/2016 15:39'!
drawIconOn: aCanvas 

	| pos |
	self hasIcon ifTrue: [
		| iconForm | 
		iconForm := self iconForm.

		pos := (contents
			ifEmpty: [self left + (self width - iconForm width // 2)]
			ifNotEmpty: [self left])
				@ (self top + (self height - iconForm height // 2)).

		aCanvas
			translucentImage: iconForm 
			at: pos].! !

!DockingBarItemMorph methodsFor: 'drawing' stamp: 'mt 7/9/2016 15:42'!
drawLabelOn: aCanvas 

	| stringBounds |	
	self contents ifEmpty: [^ self].
	
	stringBounds := bounds.
	
	self hasIcon ifTrue: [
		stringBounds := stringBounds left: stringBounds left + self iconForm width + 2 ].
	
	"Vertical centering."
	stringBounds := stringBounds top: stringBounds top + stringBounds bottom - self fontToUse height // 2.
	"Horizontal centering."
	stringBounds := stringBounds left: stringBounds left + (stringBounds width - (self fontToUse widthOfString: contents) // 2) abs.

	aCanvas
		drawString: contents
		in: stringBounds
		font: self fontToUse
		color: self colorToUse.! !

!DockingBarItemMorph methodsFor: 'drawing' stamp: 'mt 7/9/2016 15:21'!
drawSubMenuMarkerOn: aCanvas 
	"Ignore."! !

!DockingBarItemMorph methodsFor: 'as yet unclassified' stamp: 'kb 11/19/2009 18:22'!
iconForm
	"private - answer the form to be used as the icon"
	^isEnabled
		ifTrue: [
			(isSelected and: [ selectedIcon notNil ])
				ifTrue: [ selectedIcon ]
				ifFalse: [ icon ] ]
		ifFalse: [
			icon asGrayScale ]! !

!DockingBarItemMorph methodsFor: 'events' stamp: 'mt 6/21/2016 13:58'!
mouseDown: evt
	"Handle a mouse down event. Menu items get activated when the mouse is over them."

	(evt shiftPressed and:[self wantsKeyboardFocusOnShiftClick]) ifTrue: [ ^super mouseDown: evt ].  "enable label editing" 
	isSelected
		ifTrue: [
			owner selectItem: nil event: evt. ]
		ifFalse: [
			owner activate: evt. "Redirect to menu for valid transitions"
			owner selectItem: self event: evt. ]
! !

!DockingBarItemMorph methodsFor: 'events' stamp: 'mt 6/25/2016 08:47'!
mouseEnter: evt
	"Do not hover docking bar items directory. Mouse-down required. But if you already see a submenu, support hovering."

	owner selectedItem ifNotNil: [owner selectItem: self event: evt]! !

!DockingBarItemMorph methodsFor: 'events' stamp: 'mt 6/21/2016 13:35'!
mouseUp: evt
	"Handle a mouse up event. Menu items get activated when the mouse is over them. Do nothing if we're not in a 'valid menu transition', meaning that the current hand focus must be aimed at the owning menu."
	
	evt hand mouseFocus == owner ifFalse: [ ^self ].
	self contentString ifNotNil: [
		self contents: self contentString withMarkers: true inverse: true.
		self refreshWorld.
		(Delay forMilliseconds: 200) wait ].! !

!DockingBarItemMorph methodsFor: 'as yet unclassified' stamp: 'kb 11/19/2009 18:20'!
selectedIcon: aForm

	selectedIcon := aForm! !

!DockingBarItemMorph methodsFor: 'private' stamp: 'kb 11/19/2009 17:57'!
selectionFillStyle

	| fill |
	fill := super selectionFillStyle.
	fill isColor ifTrue: [ ^fill ].
	self owner isVertical
		ifFalse: [ fill direction: 0 @ self height ]
		ifTrue: [ fill direction: self width @ 0 ].
	^fill! !

!DockingBarItemMorph methodsFor: 'selecting' stamp: 'kb 2/26/2010 14:02'!
select: evt
	
	super select: evt.
	subMenu ifNotNil: [
		evt hand newKeyboardFocus: subMenu ]! !

!DockingBarItemMorph methodsFor: 'private' stamp: 'kb 11/19/2009 18:03'!
updateLayoutInDockingBar
	
	owner isVertical
		ifTrue: [
			self hResizing: #spaceFill.
			self vResizing: #shrinkWrap ]
		ifFalse: [
			self hResizing: #shrinkWrap.
			self vResizing: #spaceFill ].
	self extent: self minWidth @ self minHeight! !

!DockingBarItemMorph methodsFor: 'events' stamp: 'tpr 11/14/2015 11:27'!
wantsKeyboardFocusOnShiftClick
	"set this preference to false to prevent user editing of docking bar menu items"
	^Preferences valueOfPreference: #allowMenubarItemEditing ifAbsent: [true]! !

!DockingBarItemMorph methodsFor: 'rounding' stamp: 'kb 11/19/2009 17:53'!
wantsRoundedCorners

	^false! !

!MenuItemMorph class methodsFor: 'preferences' stamp: 'mt 7/10/2016 11:49'!
applyUserInterfaceTheme

	SubMenuMarker := (UserInterfaceTheme current get: #subMenuMarker for: self)
		ifNil: [self defaultSubMenuMarker].! !

!MenuItemMorph class methodsFor: 'defaults' stamp: 'mt 7/10/2016 11:49'!
defaultSubMenuMarker

	| f |
	f := Form
		extent: 5@9
		fromArray: #(2147483648 3221225472 3758096384 4026531840 4160749568 4026531840 3758096384 3221225472 2147483648)
		offset: 0@0.
	^ ColorForm mappingWhiteToTransparentFrom: f! !

!MenuItemMorph class methodsFor: 'class initialization' stamp: 'mt 7/10/2016 11:49'!
initialize
	"MenuItemMorph initialize"

	SubMenuMarker := self defaultSubMenuMarker.
! !

!MenuItemMorph class methodsFor: 'preferences' stamp: 'mt 8/5/2016 08:37'!
themeProperties

	^ super themeProperties, {	
		{ #font. 'Fonts'. 'Font for menu items.' }.
		{ #textColor. 'Colors'. 'Color for the menu item''s labels.' }.
		{ #disabledTextColor. 'Colors'. 'Color to use for disabled menu item labels.' }.
		{ #selectionColor. 'Colors'. 'Color used for items when hovering or selecting them.' }.
		{ #selectionTextColor. 'Colors'. 'Color used for label when hovering or selecting them.' }.
		{ #subMenuMarker. 'Forms'. 'The form to be used to indicate a submenu.' }.
	}! !

!MenuItemMorph methodsFor: 'grabbing' stamp: 'spfa 3/13/2004 18:34'!
aboutToBeGrabbedBy: aHand
	"Don't allow the receiver to act outside a Menu"
	| menu box |
	(owner notNil and:[owner submorphs size = 1]) ifTrue:[
		"I am a lonely menuitem already; just grab my owner"
		owner stayUp: true.
		^owner 	aboutToBeGrabbedBy: aHand].
	box := self bounds.
	menu := MenuMorph new defaultTarget: nil.
	menu addMorphFront: self.
	menu bounds: box.
	menu stayUp: true.
	self isSelected: false.
	^menu! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'kb 11/20/2009 00:27'!
action: aBlock

	self 
		target: aBlock;
		selector: #value! !

!MenuItemMorph methodsFor: 'events' stamp: 'ar 10/10/2000 01:38'!
activateOwnerMenu: evt
	"Activate our owner menu; e.g., pass control to it"
	owner ifNil:[^false]. "not applicable"
	(owner fullContainsPoint: evt position) ifFalse:[^false].
	owner activate: evt.
	^true! !

!MenuItemMorph methodsFor: 'events' stamp: 'mt 6/25/2016 07:59'!
activateSubmenu: evt
	"Activate our submenu; e.g., pass control to it"
	subMenu ifNil:[^false]. "not applicable"
	(subMenu fullContainsPoint: evt position) ifFalse:[^false].
	subMenu activate: evt.
	^true! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'ar 3/17/2001 20:16'!
adaptToWorld: aWorld

	super adaptToWorld: aWorld.
	target := target adaptedToWorld: aWorld.! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'kb 2/4/2010 14:14'!
addSubMenu: aBlock

	subMenu := self createSubmenu.
	aBlock value: subMenu.
	self changed.
! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'kb 11/20/2009 13:01'!
addUpdatingSubMenu: aBlock

	subMenu := UpdatingMenuMorph new.
	subMenu updater: aBlock updateSelector: #value:.
	self changed.
! !

!MenuItemMorph methodsFor: 'selecting' stamp: 'mt 2/27/2015 13:33'!
adjacentTo

	^{ self outerBounds topRight. self bounds topLeft }! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'sw 10/3/2002 20:50'!
allWordingsNotInSubMenus: verbotenSubmenuContentsList
	"Answer a collection of the wordings of all items and subitems, but omit the stay-up item, and also any items in any submenu whose tag is in verbotenSubmenuContentsList"

	self isStayUpItem ifTrue:[^ #()].
	subMenu ifNotNil:
		[^ (verbotenSubmenuContentsList includes: self contents asString)
			ifTrue:
				[#()]
			ifFalse:
				[subMenu allWordingsNotInSubMenus: verbotenSubmenuContentsList]].

	^ Array with: self contents asString! !

!MenuItemMorph methodsFor: 'updating' stamp: 'mt 7/9/2016 15:58'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	
	self
		color: (self userInterfaceTheme textColor ifNil: [Color black]);
		font: (self userInterfaceTheme font ifNil: [TextStyle defaultFont]).
	! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
arguments

	^ arguments
! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
arguments: aCollection

	arguments := aCollection.
! !

!MenuItemMorph methodsFor: 'private' stamp: 'dgd 3/31/2006 12:15'!
bottomArrow
	^ ColorForm
		mappingWhiteToTransparentFrom: ((SubMenuMarker rotateBy: 90) asFormOfDepth:8)! !

!MenuItemMorph methodsFor: 'browse' stamp: 'bp 10/19/2009 23:03'!
browseAllImplementorsOfRealSelector
	SystemNavigation default browseAllImplementorsOf: self realSelector localTo: target class! !

!MenuItemMorph methodsFor: 'browse' stamp: 'bp 10/19/2009 22:54'!
buildDebugMenu: aHandMorph
	| aMenu |
	aMenu := super buildDebugMenu: aHandMorph.
	aMenu addLine.
	aMenu add: 'implementors of' translated target: self action: #browseAllImplementorsOfRealSelector.
	^ aMenu! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'mt 7/9/2016 15:59'!
colorToUse

	^ isSelected & isEnabled
		ifTrue: [ self userInterfaceTheme selectionTextColor ifNil: [Color white] ]
		ifFalse: [ color ]! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'ar 1/16/2001 16:58'!
contentString
	^self valueOfProperty: #contentString! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'jmv 3/2/2010 16:11'!
contentString: aString 
	aString 
		ifNil: [self removeProperty: #contentString]
		ifNotNil: [self setProperty: #contentString toValue: aString]! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'ar 9/17/2000 18:32'!
contents: aString
	^self contents: aString withMarkers: true! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'ar 1/16/2001 16:57'!
contents: aString withMarkers: aBool
	^self contents: aString withMarkers: aBool inverse: false! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'nk 3/10/2004 15:55'!
contents: aString withMarkers: aBool inverse: inverse 
	"Set the menu item entry. If aBool is true, parse aString for embedded markers."

	| markerIndex marker |
	self contentString: nil.	"get rid of old"
	aBool ifFalse: [^super contents: aString].
	self removeAllMorphs.	"get rid of old markers if updating"
	self hasIcon ifTrue: [ self icon: nil ].
	(aString notEmpty and: [aString first = $<]) 
		ifFalse: [^super contents: aString].
	markerIndex := aString indexOf: $>.
	markerIndex = 0 ifTrue: [^super contents: aString].
	marker := (aString copyFrom: 1 to: markerIndex) asLowercase.
	(#('<on>' '<off>' '<yes>' '<no>') includes: marker) 
		ifFalse: [^super contents: aString].
	self contentString: aString.	"remember actual string"
	marker := (marker = '<on>' or: [marker = '<yes>']) ~= inverse 
				ifTrue: [self onImage]
				ifFalse: [self offImage].
	super contents:  (aString copyFrom: markerIndex + 1 to: aString size).
	"And set the marker"
	marker := ImageMorph new image: marker.
	marker position: self left @ (self top + 2).
	self addMorphFront: marker! !

!MenuItemMorph methodsFor: 'private' stamp: 'kb 2/4/2010 14:13'!
createSubmenu

	^MenuMorph new! !

!MenuItemMorph methodsFor: 'private' stamp: 'kb 2/4/2010 14:13'!
createUpdatingSubmenu

	^UpdatingMenuMorph new! !

!MenuItemMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 15:07'!
defaultBounds
"answer the default bounds for the receiver"
	^ 0 @ 0 extent: 10 @ 10! !

!MenuItemMorph methodsFor: 'initialization' stamp: 'ar 10/10/2000 02:05'!
deleteIfPopUp: evt
	"Recurse up for nested pop ups"
	owner ifNotNil:[owner deleteIfPopUp: evt].! !

!MenuItemMorph methodsFor: 'private' stamp: 'hg 8/3/2000 15:21'!
deselectItem
	| item |
	self isSelected: false.
	subMenu ifNotNil: [subMenu deleteIfPopUp].
	(owner isKindOf: MenuMorph) ifTrue:
		[item := owner popUpOwner.
		(item isKindOf: MenuItemMorph) ifTrue: [item deselectItem]].
! !

!MenuItemMorph methodsFor: 'selecting' stamp: 'mt 6/25/2016 07:58'!
deselect: evt

	self isSelected: false.
	subMenu ifNotNil: [
		owner ifNotNil: [ owner activeSubmenu: nil ] ].! !

!MenuItemMorph methodsFor: 'events' stamp: 'sw 2/7/2001 00:03'!
doButtonAction
	"Called programattically, this should trigger the action for which the receiver is programmed"

	self invokeWithEvent: nil! !

!MenuItemMorph methodsFor: 'drawing' stamp: 'mt 7/9/2016 15:19'!
drawBackgroundOn: aCanvas 

	isSelected & isEnabled
		ifTrue: [aCanvas fillRectangle: self bounds fillStyle: self selectionFillStyle].! !

!MenuItemMorph methodsFor: 'drawing' stamp: 'mt 7/9/2016 15:16'!
drawIconOn: aCanvas 

	self hasIcon ifTrue: [
		| iconForm | 
		iconForm := self iconForm.
		aCanvas
			translucentImage: iconForm 
			at: self bounds left @ (self top + (self height - iconForm height // 2)).].! !

!MenuItemMorph methodsFor: 'drawing' stamp: 'mt 7/9/2016 15:42'!
drawLabelOn: aCanvas 

	| stringBounds |	
	stringBounds := bounds.
	
	self hasIcon ifTrue: [
		stringBounds := stringBounds left: stringBounds left + self iconForm width + 2 ].
	self hasMarker ifTrue: [
		stringBounds := stringBounds left: stringBounds left + self submorphBounds width + 8 ].
	
	stringBounds := stringBounds top: stringBounds top + stringBounds bottom - self fontToUse height // 2.

	aCanvas
		drawString: contents
		in: stringBounds
		font: self fontToUse
		color: self colorToUse.! !

!MenuItemMorph methodsFor: 'drawing' stamp: 'mt 7/9/2016 15:20'!
drawOn: aCanvas 

	self
		drawBackgroundOn: aCanvas;
		drawIconOn: aCanvas;
		drawLabelOn: aCanvas;
		drawSubMenuMarkerOn: aCanvas.! !

!MenuItemMorph methodsFor: 'drawing' stamp: 'mt 7/9/2016 15:19'!
drawSubMenuMarkerOn: aCanvas 

	self hasSubMenu ifTrue: [
		| subMenuMarker subMenuMarkerPosition | 
		subMenuMarker := self subMenuMarker.
		subMenuMarkerPosition := self right - subMenuMarker width @ (self top + self bottom - subMenuMarker height // 2).
		aCanvas paintImage: subMenuMarker at: subMenuMarkerPosition ]! !

!MenuItemMorph methodsFor: 'grabbing' stamp: 'spfa 3/13/2004 18:32'!
duplicateMorph: evt
	"Make and return a duplicate of the receiver's argument"
	| dup menu |
	dup := self duplicate isSelected: false.
	menu := MenuMorph new defaultTarget: nil.
	menu addMorphFront: dup.
	menu bounds: self bounds.
	menu stayUp: true.
	evt hand grabMorph: menu from: owner. "duplicate was ownerless so use #grabMorph:from: here"
	^menu! !

!MenuItemMorph methodsFor: 'events' stamp: 'ar 10/10/2000 22:45'!
handleMouseUp: anEvent
	"The handling of control between menu item requires them to act on mouse up even if not the current focus. This is different from the default behavior which really only wants to handle mouse ups when they got mouse downs before"
	anEvent wasHandled ifTrue:[^self]. "not interested"
	anEvent hand releaseMouseFocus: self.
	anEvent wasHandled: true.
	anEvent blueButtonChanged
		ifTrue:[self blueButtonUp: anEvent]
		ifFalse:[self mouseUp: anEvent].! !

!MenuItemMorph methodsFor: 'events' stamp: 'jm 11/4/97 07:15'!
handlesMouseDown: evt

	^ true
! !

!MenuItemMorph methodsFor: 'events' stamp: 'ar 9/18/2000 21:46'!
handlesMouseOverDragging: evt
	^true! !

!MenuItemMorph methodsFor: 'events' stamp: 'ar 9/16/2000 14:40'!
handlesMouseOver: anEvent
	^true! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'topa 10/17/2014 19:46'!
hasIcon
	"Answer whether the receiver has an icon. If menues without icons are
	requested, pretend to not bear one unless we have not content."
	^ self icon notNil and: [
		Preferences menuWithIcons or: [self contents isEmptyOrNil]].! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'nk 3/10/2004 15:19'!
hasIconOrMarker
	"Answer whether the receiver has an icon or a marker."
	^ self hasIcon or: [ submorphs isEmpty not ]! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'nk 3/10/2004 15:25'!
hasMarker
	"Answer whether the receiver has a marker morph."
	^ submorphs isEmpty not! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'ar 9/18/2000 11:40'!
hasSubMenu
	"Return true if the receiver has a submenu"
	^subMenu notNil! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'ar 9/18/2000 11:18'!
hasSubMenu: aMenuMorph
	subMenu ifNil:[^false].
	subMenu == aMenuMorph ifTrue:[^true].
	^subMenu hasSubMenu: aMenuMorph! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'kb 11/20/2009 00:29'!
help: aString

	self setBalloonText: aString! !

!MenuItemMorph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:58'!
hResizing
	"Default to #spaceFill"
	| props |
	props := self layoutProperties.
	^props ifNil:[#spaceFill] ifNotNil:[props hResizing].! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'dgd 3/22/2003 14:45'!
icon
	"answer the receiver's icon"
	^ icon! !

!MenuItemMorph methodsFor: 'private' stamp: 'dgd 9/1/2004 17:10'!
iconForm
	"private - answer the form to be used as the icon"
	^ isEnabled
		ifTrue: [self icon]
		ifFalse: [self icon asGrayScale]! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'dgd 9/13/2004 19:00'!
icon: aForm 
	"change the the receiver's icon"
	icon := aForm.
	self height: self minHeight.
self width: self minWidth! !

!MenuItemMorph methodsFor: 'initialization' stamp: 'mt 7/9/2016 15:57'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	
	contents := ''.
	hasFocus := false.
	isEnabled := true.
	subMenu := nil.
	isSelected := false.
	target := nil.
	selector := nil.
	arguments := nil.
	font := self userInterfaceTheme font ifNil: [TextStyle defaultFont].
	self
		color: (self userInterfaceTheme textColor ifNil: [Color black]);
		hResizing: #spaceFill;
		vResizing: #shrinkWrap! !

!MenuItemMorph methodsFor: 'events' stamp: 'nice 12/27/2009 03:10'!
invokeWithEvent: evt
	"Perform the action associated with the given menu item."

	| w |
	self isEnabled ifFalse: [^ self].
	target class == HandMorph ifTrue: [(self notObsolete) ifFalse: [^ self]].
	owner ifNotNil:[self isStayUpItem ifFalse:[
		self flag: #workAround. "The tile system invokes menus straightforwardly so the menu might not be in the world."
		(w := self world) ifNotNil:[
			owner deleteIfPopUp: evt.
			"Repair damage before invoking the action for better feedback"
			w displayWorldSafely]]].
	selector ifNil:[^self].
	Cursor normal showWhile: [ | selArgCount |  "show cursor in case item opens a new MVC window"
		(selArgCount := selector numArgs) = 0
			ifTrue:
				[target perform: selector]
			ifFalse:
				[selArgCount = arguments size
					ifTrue: [target perform: selector withArguments: arguments]
					ifFalse: [target perform: selector withArguments: (arguments copyWith: evt)]]].! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
isEnabled

	^ isEnabled
! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'mt 8/5/2016 08:39'!
isEnabled: aBoolean

	| colorToUse |
	isEnabled = aBoolean ifTrue: [^ self].
	isEnabled := aBoolean.
	
	colorToUse := isEnabled
		ifTrue: [self userInterfaceTheme textColor ifNil: [Color black]]
		ifFalse: [self userInterfaceTheme disabledTextColor ifNil: [Color gray]].
	
	self color: colorToUse.! !

!MenuItemMorph methodsFor: 'testing' stamp: 'eem 8/27/2015 16:42'!
isMenuItemMorph
	^true! !

!MenuItemMorph methodsFor: 'selecting' stamp: 'dgd 9/9/2004 21:26'!
isSelected
	^ isSelected ! !

!MenuItemMorph methodsFor: 'selecting' stamp: 'ar 9/18/2000 11:09'!
isSelected: aBoolean

	isSelected := aBoolean.
	self changed.
! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'RAA 1/18/2001 18:24'!
isStayUpItem

	^selector == #toggleStayUp: or: [selector == #toggleStayUpIgnore:evt:]! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'ar 9/18/2000 11:22'!
itemWithWording: wording
	"If any of the receiver's items or submenu items have the given wording (case-blind comparison done), then return it, else return nil."
	(self contents asString sameAs: wording) ifTrue:[^self].
	subMenu ifNotNil:[^subMenu itemWithWording: wording].
	^nil! !

!MenuItemMorph methodsFor: 'private' stamp: 'dgd 3/31/2006 12:16'!
leftArrow
	^ ColorForm
		mappingWhiteToTransparentFrom: ((SubMenuMarker rotateBy: 180)asFormOfDepth: 8)! !

!MenuItemMorph methodsFor: 'layout' stamp: 'tfel 2/27/2010 19:27'!
minHeight
	| iconHeight |
	iconHeight := self hasIcon
				ifTrue: [self icon height + 2]
				ifFalse: [0].
	^ self fontToUse height + 2 max: iconHeight! !

!MenuItemMorph methodsFor: 'layout' stamp: 'mt 7/9/2016 15:31'!
minWidth

	| subMenuWidth iconWidth markerWidth |
	subMenuWidth := self hasSubMenu
		ifTrue: [ 10 ]
		ifFalse: [ 0 ].
	iconWidth := self hasIcon
		ifTrue: [ self icon width + 2 ]
		ifFalse: [ 0 ].
	markerWidth := self hasMarker
		ifTrue: [ self submorphBounds width + 8 ]
		ifFalse: [ 0 ].
	^(self fontToUse widthOfString: contents)
		+ subMenuWidth + iconWidth + markerWidth + self stringMargin! !

!MenuItemMorph methodsFor: 'events' stamp: 'kb 11/19/2009 17:44'!
mouseDown: evt
	"Handle a mouse down event. Menu items get activated when the mouse is over them."

	evt shiftPressed ifTrue: [ ^super mouseDown: evt ].  "enable label editing" 
	evt hand newMouseFocus: owner. "Redirect to menu for valid transitions"
	owner selectItem: self event: evt! !

!MenuItemMorph methodsFor: 'events' stamp: 'mt 6/25/2016 08:41'!
mouseEnterDragging: evt

	self mouseEnter: evt.! !

!MenuItemMorph methodsFor: 'events' stamp: 'mt 6/25/2016 08:41'!
mouseEnter: evt
	"The mouse entered the receiver"

	owner ifNotNil: [ owner selectItem: self event: evt ]! !

!MenuItemMorph methodsFor: 'events' stamp: 'kb 11/19/2009 17:50'!
mouseUp: evt
	"Handle a mouse up event. Menu items get activated when the mouse is over them. Do nothing if we're not in a 'valid menu transition', meaning that the current hand focus must be aimed at the owning menu."
	
	evt hand mouseFocus == owner ifFalse: [ ^self ].
	self contentString ifNotNil: [
		self contents: self contentString withMarkers: true inverse: true.
		self refreshWorld.
		(Delay forMilliseconds: 200) wait ].
	self deselect: evt.
	self invokeWithEvent: evt.		
! !

!MenuItemMorph methodsFor: 'submorphs-accessing' stamp: 'dgd 9/9/2004 20:25'!
noteNewOwner: aMorph 
	"I have just been added as a submorph of aMorph"
	super noteNewOwner: aMorph.

	self updateLayoutInDockingBar! !

!MenuItemMorph methodsFor: 'private' stamp: 'ar 9/18/2000 10:27'!
notObsolete
	"Provide backward compatibility with messages being sent to the Hand.  Remove this when no projects made prior to 2.9 are likely to be used.  If this method is removed early, the worst that can happen is a notifier when invoking an item in an obsolete menu."

	(HandMorph canUnderstand: (selector)) ifTrue: [^ true]. 	"a modern one"

	self inform: 'This world menu is obsolete.
Please dismiss the menu and open a new one.'.
	^ false
! !

!MenuItemMorph methodsFor: 'private' stamp: 'ar 9/17/2000 18:36'!
offImage
	"Return the form to be used for indicating an '<off>' marker"
	| form |
	form := Form extent: (self fontToUse ascent-2) asPoint depth: 16.
	(form getCanvas)
		frameAndFillRectangle: form boundingBox fillColor: (Color gray: 0.9) 
			borderWidth: 1 borderColor: Color black.
	^form! !

!MenuItemMorph methodsFor: 'private' stamp: 'ar 9/19/2000 09:34'!
onImage
	"Return the form to be used for indicating an '<off>' marker"
	| form |
	form := Form extent: (self fontToUse ascent-2) asPoint depth: 16.
	(form getCanvas)
		frameAndFillRectangle: form boundingBox fillColor: (Color gray: 0.8) 
			borderWidth: 1 borderColor: Color black;
		fillRectangle: (form boundingBox insetBy: 2) fillStyle: Color black.
	^form! !

!MenuItemMorph methodsFor: 'change reporting' stamp: 'dgd 9/1/2004 18:29'!
ownerChanged
	"The receiver's owner, some kind of a pasteup, has changed its 
	layout."
	super ownerChanged.
	self updateLayoutInDockingBar! !

!MenuItemMorph methodsFor: 'browse' stamp: 'bp 10/19/2009 22:22'!
realSelector
	selector == #perform:orSendTo: ifTrue: [^arguments first].
	^selector! !

!MenuItemMorph methodsFor: 'private' stamp: 'dgd 9/1/2004 18:07'!
rightArrow

	^ SubMenuMarker! !

!MenuItemMorph methodsFor: 'private' stamp: 'mt 7/22/2016 11:58'!
selectionFillStyle
	" Answer the fill style to use with the receiver is the selected element "

	| fill baseColor |
	baseColor := self userInterfaceTheme selectionColor ifNil: [Color r: 0.4 g: 0.5 b: 0.7].
	MenuMorph gradientMenu ifFalse: [ ^baseColor ].

	fill := GradientFillStyle ramp: { 
		0.0 -> baseColor twiceLighter. 
		1 -> baseColor twiceDarker }.
	fill origin: self topLeft.
	fill direction: 0 @ self height.
	^ fill! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
selector

	^ selector
! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
selector: aSymbol

	selector := aSymbol.
! !

!MenuItemMorph methodsFor: 'selecting' stamp: 'kb 11/20/2009 13:03'!
select: evt
	self isSelected: true.
	owner activeSubmenu: subMenu.
	subMenu ifNotNil: [
		subMenu delete.
		subMenu
			popUpAdjacentTo: self adjacentTo
			forHand: evt hand
			from: self.
		subMenu selectItem: nil event: evt].! !

!MenuItemMorph methodsFor: 'layout' stamp: 'mt 7/9/2016 15:26'!
stringMargin

	^Preferences tinyDisplay
		ifTrue: [ 1 ]
		ifFalse: [ 6 ]! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
subMenu

	^ subMenu
! !

!MenuItemMorph methodsFor: 'private' stamp: 'kb 11/19/2009 17:59'!
subMenuMarker
	"private - answer the form to be used as submenu marker"
	
	^self rightArrow! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'kb 2/4/2010 14:13'!
subMenuUpdater: updater selector: selector

	subMenu := self createUpdatingSubmenu.
	subMenu updater: updater updateSelector: selector.
	self changed.
! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'kb 2/4/2010 14:13'!
subMenuUpdater: updater selector: selector arguments: arguments

	subMenu := self createUpdatingSubmenu.
	subMenu updater: updater updateSelector: selector arguments: arguments.
	self changed.
! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
subMenu: aMenuMorph

	subMenu := aMenuMorph.
	self changed.
! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
target

	^ target! !

!MenuItemMorph methodsFor: 'accessing' stamp: 'jm 11/4/97 07:46'!
target: anObject

	target := anObject.
! !

!MenuItemMorph methodsFor: 'geometry-testing' stamp: 'kb 11/17/2009 13:06'!
triangle: points containsPoint: p
	" Computes if p is in the triangle defined by points.
	p should be a Point, and points should be an array with three Points.
	I took the algorithm from the bottom of this page: 
		http://www.blackpawn.com/texts/pointinpoly/default.html "

	| a b c v0 v1 v2 dot00 dot01 dot02 dot11 dot12 denom invDenom u v |
	a := points first.
	b := points second.
	c := points third.
	" Compute vectors "
	v0 := c - a.
	v1 := b - a.
	v2 := p - a.
	" Compute dot products "
	dot00 := v0 dotProduct: v0.
	dot01 := v0 dotProduct: v1.
	dot02 := v0 dotProduct: v2.
	dot11 := v1 dotProduct: v1.
	dot12 := v1 dotProduct: v2.
	" Compute barycentric coordinates "
	denom := dot00 * dot11 - (dot01 * dot01).
	denom = 0 ifTrue: [ ^false ].
	invDenom := 1 / denom.
	u := (dot11 * dot02 - (dot01 * dot12)) * invDenom.
	v := (dot00 * dot12 - (dot01 * dot02)) * invDenom.
	" Check if point is in triangle "
	^u >= 0 and: [ v >= 0 and: [ u + v <= 1 ] ]! !

!MenuItemMorph methodsFor: 'private' stamp: 'dgd 3/31/2006 12:16'!
upArrow
	^ ColorForm
		mappingWhiteToTransparentFrom: ((SubMenuMarker rotateBy: 270)asFormOfDepth: 8)! !

!MenuItemMorph methodsFor: 'private' stamp: 'kb 11/19/2009 18:03'!
updateLayoutInDockingBar! !

!MenuItemMorph methodsFor: 'copying' stamp: 'sw 9/25/2002 03:24'!
veryDeepFixupWith: deepCopier
	"If target and arguments fields were weakly copied, fix them here.  If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

	super veryDeepFixupWith: deepCopier.
	target := deepCopier references at: target ifAbsent: [target].
	arguments notNil ifTrue:
	[arguments := arguments collect: [:each |
		deepCopier references at: each ifAbsent: [each]]]! !

!MenuItemMorph methodsFor: 'copying' stamp: 'mt 6/25/2016 07:59'!
veryDeepInner: deepCopier 
	"Copy all of my instance variables. Some need to be not copied  
	at all, but shared. Warning!!!! Every instance variable defined in  
	this class must be handled. We must also implement  
	veryDeepFixupWith:. See DeepCopier class comment."
	super veryDeepInner: deepCopier.
	isEnabled := isEnabled veryDeepCopyWith: deepCopier.
	subMenu := subMenu veryDeepCopyWith: deepCopier.
	isSelected := isSelected veryDeepCopyWith: deepCopier.
	icon := icon veryDeepCopyWith: deepCopier.
	"target := target.		Weakly copied"
	"selector := selector.		a Symbol"
	arguments := arguments.! !

!MenuItemMorph methodsFor: 'layout-properties' stamp: 'ar 11/14/2000 17:59'!
vResizing
	"Default to #shrinkWrap"
	| props |
	props := self layoutProperties.
	^props ifNil:[#shrinkWrap] ifNotNil:[props vResizing].! !

!MenuItemMorph methodsFor: 'meta actions' stamp: 'ar 10/10/2000 02:13'!
wantsHaloFromClick
	"Only if I'm not a lonely submenu"
	^owner notNil and:[owner submorphs size > 1]! !

MenuItemMorph subclass: #UpdatingMenuItemMorph
	instanceVariableNames: 'wordingProvider wordingSelector enablementSelector wordingArgument'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Menus'!

!UpdatingMenuItemMorph commentStamp: '<historical>' prior: 0!
A menu item whose textual label and whose enablement are updatable.  The wordingProvider provides the current wording, upon being being sent the wordingSelector.

The item can also dynamically update whether or not it should be enabled; to do this, give it an enablementSelector, which is also sent to the wordingProvider..!

!UpdatingMenuItemMorph methodsFor: 'stepping and presenter' stamp: 'ar 10/7/2000 15:35'!
arrangeToStartSteppingIn: aWorld
	super arrangeToStartSteppingIn: aWorld.
	self updateContents.! !

!UpdatingMenuItemMorph methodsFor: 'enablement' stamp: 'ajh 1/21/2003 13:17'!
enablement 

	enablementSelector isBlock
		ifTrue: [^ enablementSelector value]
		ifFalse: [enablementSelector numArgs = 0
				ifTrue: [^ wordingProvider perform: enablementSelector]
				ifFalse: [^ wordingProvider perform: enablementSelector
										withArguments: arguments]]! !

!UpdatingMenuItemMorph methodsFor: 'enablement' stamp: 'nice 3/13/2010 21:46'!
enablementSelector: aSelector 
	enablementSelector := aSelector isBlock 
				ifTrue: [aSelector copyForSaving]
				ifFalse: [aSelector] ! !

!UpdatingMenuItemMorph methodsFor: 'stepping and presenter' stamp: 'ar 10/7/2000 15:34'!
step
	super step.
	self updateContents.! !

!UpdatingMenuItemMorph methodsFor: 'testing' stamp: 'sw 6/11/1999 18:31'!
stepTime
	^ 1200! !

!UpdatingMenuItemMorph methodsFor: 'wording' stamp: 'wiz 2/11/2008 16:52'!
target: anObject
"For us if the old target and wording provider are the same update both."
target = wordingProvider 
	ifTrue: [ wordingProvider := target := anObject ]
	ifFalse: [  target := anObject ] .
! !

!UpdatingMenuItemMorph methodsFor: 'world' stamp: 'laza 5/16/2011 12:54'!
updateContents
	"Update the receiver's contents"

	| newString enablement nArgs |
	((wordingProvider isNil) or: [wordingSelector isNil]) ifFalse: [
		nArgs := wordingSelector numArgs.
		newString := nArgs = 0
			ifTrue:
				[wordingProvider perform: wordingSelector]
			ifFalse:
				[(nArgs = 1 and: [wordingArgument notNil])
					ifTrue:
						[wordingProvider perform: wordingSelector with: wordingArgument]
					ifFalse:
						[nArgs == arguments size ifTrue:
							[wordingProvider perform: wordingSelector withArguments: arguments]]].
		newString = (self contentString ifNil: [ contents ])
			ifFalse: [self contents: newString.
				MenuIcons decorateMenu: owner ]].
	enablementSelector ifNotNil:
		[(enablement := self enablement) == isEnabled 
			ifFalse:	[self isEnabled: enablement]]! !

!UpdatingMenuItemMorph methodsFor: 'wording' stamp: 'sw 11/6/2000 09:55'!
wordingArgument: anArgument
	"Set the receiver's wordingArgument as indicated"

	wordingArgument := anArgument! !

!UpdatingMenuItemMorph methodsFor: 'wording' stamp: 'sw 6/11/1999 15:12'!
wordingProvider: aProvider wordingSelector: aSelector
	wordingProvider := aProvider.
	wordingSelector := aSelector! !

StringMorph subclass: #PopUpChoiceMorph
	instanceVariableNames: 'target actionSelector arguments getItemsSelector getItemsArgs choiceSelector choiceArgs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!PopUpChoiceMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/2/98 00:05'!
actionSelector

	^ actionSelector
! !

!PopUpChoiceMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/2/98 00:05'!
actionSelector: aSymbolOrString

	(nil = aSymbolOrString or:
	 ['nil' = aSymbolOrString or:
	 [aSymbolOrString isEmpty]])
		ifTrue: [^ actionSelector := nil].

	actionSelector := aSymbolOrString asSymbol.
! !

!PopUpChoiceMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/2/98 00:05'!
arguments

	^ arguments
! !

!PopUpChoiceMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/2/98 00:05'!
arguments: aCollection

	arguments := aCollection asArray copy.
! !

!PopUpChoiceMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/2/98 00:33'!
getItemsArgs

	^ getItemsArgs
! !

!PopUpChoiceMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/2/98 00:33'!
getItemsArgs: aCollection

	getItemsArgs := aCollection asArray copy.
! !

!PopUpChoiceMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/2/98 00:32'!
getItemsSelector

	^ getItemsSelector
! !

!PopUpChoiceMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/2/98 00:32'!
getItemsSelector: aSymbolOrString

	(nil = aSymbolOrString or:
	 ['nil' = aSymbolOrString or:
	 [aSymbolOrString isEmpty]])
		ifTrue: [^ getItemsSelector := nil].

	getItemsSelector := aSymbolOrString asSymbol.
! !

!PopUpChoiceMorph methodsFor: 'event handling' stamp: 'jm 2/2/98 00:20'!
handlesMouseDown: evt

	^ true
! !

!PopUpChoiceMorph methodsFor: 'initialization' stamp: 'cmm 5/8/2014 10:59'!
initialize
"initialize the state of the receiver"
	super initialize.
	self contents: 'PopUpChoice of Colors'.
	target := Color.
	actionSelector := nil.
	arguments := Array empty.
	getItemsSelector := #colorNames.
	getItemsArgs := Array empty! !

!PopUpChoiceMorph methodsFor: 'event handling' stamp: 'dtl 12/30/2013 17:47'!
mouseDown: evt 
	| items selectedItem |
	(target isNil or: [getItemsSelector isNil]) ifTrue: [^self].
	items := ((target perform: getItemsSelector withArguments: getItemsArgs)
				ifNil: [#()]) asOrderedCollection.
	selectedItem := UIManager default 
		chooseFrom: items
		values: items.
	selectedItem ifNil: [^self].
	self contentsClipped: selectedItem.	"Client can override this if necess"
	actionSelector ifNotNil: 
			[target perform: actionSelector
				withArguments: (arguments copyWith: selectedItem)]! !

!PopUpChoiceMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/2/98 00:05'!
target

	^ target
! !

!PopUpChoiceMorph methodsFor: 'as yet unclassified' stamp: 'jm 2/2/98 00:05'!
target: anObject

	target := anObject
! !

!PopUpChoiceMorph methodsFor: 'copying' stamp: 'di 3/24/1999 09:57'!
veryDeepFixupWith: deepCopier
	"If target and arguments fields were weakly copied, fix them here.  If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

super veryDeepFixupWith: deepCopier.
target := deepCopier references at: target ifAbsent: [target].
arguments := arguments collect: [:each |
	deepCopier references at: each ifAbsent: [each]].
getItemsArgs := getItemsArgs collect: [:each |
	deepCopier references at: each ifAbsent: [each]].
choiceArgs ifNotNil: [choiceArgs := choiceArgs collect: [:each |
	deepCopier references at: each ifAbsent: [each]]].! !

!PopUpChoiceMorph methodsFor: 'copying' stamp: 'tk 1/8/1999 09:43'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all, but shared.  	Warning!!!!  Every instance variable defined in this class must be handled.  We must also implement veryDeepFixupWith:.  See DeepCopier class comment."

super veryDeepInner: deepCopier.
"target := target.		Weakly copied"
"actionSelector := actionSelector.		a Symbol"
"arguments := arguments.		All weakly copied"
"getItemsSelector := getItemsSelector.		a Symbol"
"getItemsArgs := getItemsArgs.		All weakly copied"
"choiceSelector := choiceSelector.		a Symbol"
choiceArgs := choiceArgs.		"All weakly copied"
     ! !

!StringMorph class methodsFor: 'scripting' stamp: 'sw 5/6/1998 14:00'!
authoringPrototype
	^ super authoringPrototype contents: 'String'! !

!StringMorph class methodsFor: 'instance creation' stamp: 'sw 8/22/97 22:19'!
contents: aString
	" 'StringMorph contents: str' is faster than 'StringMorph new contents: str' "
	^ self contents: aString font: nil! !

!StringMorph class methodsFor: 'instance creation' stamp: 'di 4/1/1999 17:15'!
contents: aString font: aFont
	^ self basicNew initWithContents: aString font: aFont emphasis: 0! !

!StringMorph class methodsFor: 'instance creation' stamp: 'di 4/1/1999 17:15'!
contents: aString font: aFont emphasis: emphasisCode
	^ self basicNew initWithContents: aString font: aFont emphasis: emphasisCode! !

!StringMorph class methodsFor: 'testing' stamp: 'di 5/6/1998 21:07'!
test
	"Return a morph with lots of strings for testing display speed."
	| c |
	c := AlignmentMorph newColumn.
	SystemOrganization categories do:
		[:cat | c addMorph: (StringMorph new contents: cat)].
	^ c! !

!StringMorph class methodsFor: 'testing' stamp: 'di 5/6/1998 21:08'!
test2
	"Return a morph with lots of strings for testing display speed."
	| c r |
	c := AlignmentMorph newColumn.
	SystemOrganization categories reverseDo:
		[:cat | c addMorph: (StringMorph new contents: cat)].
	r := RectangleMorph new extent: c fullBounds extent.
	c submorphsDo: [:m | r addMorph: m].
	^ r
! !

!StringMorph methodsFor: 'editing' stamp: ''!
acceptContents
	"The message is sent when the user hits enter or Cmd-S. Accept the current contents and end editing. This default implementation does nothing."
! !

!StringMorph methodsFor: 'editing' stamp: 'sw 9/8/1999 17:04'!
acceptValue: aValue
	| val |
	self contents: (val := aValue asString).
	^ val! !

!StringMorph methodsFor: 'menu' stamp: 'yo 3/24/2008 18:22'!
addCustomMenuItems: aCustomMenu hand: aHandMorph

	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	aCustomMenu add: 'change font' translated action: #changeFont.
	aCustomMenu add: 'change emphasis' translated action: #changeEmphasis.
	aCustomMenu addUpdating: #usePangoString target: self action: #toggleUsePango.
! !

!StringMorph methodsFor: 'halos and balloon help' stamp: 'sw 4/2/2007 15:38'!
addOptionalHandlesTo: aHalo box: box
	"eventually, add more handles for font..."

	self flag: #deferred.
	^ super addOptionalHandlesTo: aHalo box: box

	"Eventually...
	self addFontHandlesTo: aHalo box: box"! !

!StringMorph methodsFor: 'halos and balloon help' stamp: 'sw 6/6/2001 13:34'!
boundsForBalloon
	"Some morphs have bounds that are way too big.  This is a contorted way of making things work okay in PluggableListMorphs, whose list elements historically have huge widths"

	| ownerOwner |
	^ ((owner notNil and: [(ownerOwner := owner owner) notNil]) and:
			[ownerOwner isKindOf: PluggableListMorph])
		ifTrue:
			[self boundsInWorld intersect: ownerOwner boundsInWorld]
		ifFalse:
			[super boundsForBalloon]! !

!StringMorph methodsFor: 'editing' stamp: 'sw 9/17/1999 13:27'!
cancelEdits

	self doneWithEdits! !

!StringMorph methodsFor: 'menu' stamp: 'ar 8/6/2009 19:11'!
changeEmphasis

	| reply aList |
	aList := #(normal bold italic narrow underlined struckOut).
	reply := UIManager default 
		chooseFrom: (aList collect: [:t | t translated]) 
		values: aList..
	reply ifNotNil:[
		self emphasis: (TextEmphasis perform: reply) emphasisCode.
	].
! !

!StringMorph methodsFor: 'menu' stamp: 'ul 8/23/2010 00:49'!
changeFont
	
	| chooser originalFont |
	originalFont := self fontToUse.
	self openModal: (
		Cursor wait showWhile: [
			| window |
			window := UIManager default 
				chooseFont: 'Choose a Font' 
				for: self 
				setSelector: #font: 
				getSelector: originalFont.
			"We have to save the model here, because it will be gone when the window is closed."
			chooser := window model. 
			window ]).
	originalFont = self fontToUse ifFalse: [
		"Ensure that we restore the original font in case it was changed, but Cancel was clicked."
		self font: (chooser result ifNil: [ originalFont ]) ]! !

!StringMorph methodsFor: 'accessing' stamp: ''!
contents

	^ contents! !

!StringMorph methodsFor: 'accessing' stamp: ''!
contentsClipped: aString
	"Change my text, but do not change my size as a result"
	contents = aString ifTrue: [^ self].  "No substantive change"
	contents := aString.
	self changed! !

!StringMorph methodsFor: 'accessing' stamp: 'cao 2/28/2011 18:17'!
contents: newContents 
	| scanner |
	contents := newContents isText
				ifTrue: [scanner := StringMorphAttributeScanner new initializeFromStringMorph: self.
					(newContents attributesAt: 1 forStyle: self font textStyle)
						do: [:attr | attr emphasizeScanner: scanner].
					emphasis := scanner emphasis.
					font := scanner font.
					color := scanner textColor.
					newContents string]
				ifFalse: [contents = newContents
						ifTrue: [^ self].
					"no substantive change"
					newContents].
	self fitContents! !

!StringMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:31'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color black! !

!StringMorph methodsFor: 'editing' stamp: 'di 9/6/1999 22:44'!
doneWithEdits

	hasFocus := false! !

!StringMorph methodsFor: 'drawing' stamp: 'ar 12/31/2001 02:38'!
drawOn: aCanvas

	aCanvas drawString: contents in: bounds font: self fontToUse color: color.! !

!StringMorph methodsFor: 'font' stamp: 'efc 2/22/2003 21:35'!
emphasis: aNumber
	"Set the receiver's emphasis as indicated. aNumber is a bitmask with the following format:

	bit	attribute
	1	bold
	2	italic
	4	underlined
	8	narrow
	16	struckOut"

	"examples: 0 -> plain.  
	1 -> bold.  2 -> italic.  3 -> bold italic.  4 -> underlined  
	5 -> bold underlined.  6 -> italic underlined.   7 -> bold italic underlined   
	etc..."

	emphasis := aNumber.
	^ self font: font emphasis: emphasis! !

!StringMorph methodsFor: 'accessing' stamp: 'ar 12/12/2001 02:44'!
fitContents

	| newBounds boundsChanged |
	newBounds := self measureContents.
	boundsChanged := bounds extent ~= newBounds.
	self extent: newBounds.		"default short-circuits if bounds not changed"
	boundsChanged ifFalse: [self changed]! !

!StringMorph methodsFor: 'objects from disk' stamp: 'yo 1/9/2007 01:07'!
fixUponLoad: aProject seg: anImageSegment
	"We are in an old project that is being loaded from disk.
Fix up conventions that have changed."

	| substituteFont |
	substituteFont _ (aProject projectParameterAt: #substitutedFont).
	(substituteFont notNil and: [self font == substituteFont])
			ifTrue: [ self fitContents ].

	^ super fixUponLoad: aProject seg: anImageSegment! !

!StringMorph methodsFor: 'accessing' stamp: 'ar 1/31/2001 19:33'!
font
	"who came up with #fontToUse rather than font?!!"
	^self fontToUse! !

!StringMorph methodsFor: 'accessing' stamp: 'tk 8/28/2000 13:59'!
fontName: fontName size: fontSize

	^ self font: (StrikeFont familyName: fontName size: fontSize) 
			emphasis: 0! !

!StringMorph methodsFor: 'accessing' stamp: 'dgd 2/21/2003 23:07'!
fontToUse
	| fontToUse |
	fontToUse := font isNil ifTrue: [TextStyle defaultFont] ifFalse: [font].
	(emphasis isNil or: [emphasis = 0]) 
		ifTrue: [^fontToUse]
		ifFalse: [^fontToUse emphasized: emphasis]! !

!StringMorph methodsFor: 'printing' stamp: 'yo 1/5/2007 19:29'!
font: aFont 
	"Set the font my text will use. The emphasis remains unchanged."

	aFont = font ifTrue: [^ self].
	font _ aFont.
	^ self font: font emphasis: emphasis! !

!StringMorph methodsFor: 'accessing' stamp: 'di 4/2/1999 16:11'!
font: aFont emphasis: emphasisCode
	font := aFont.
	emphasis := emphasisCode.
	self fitContents.
"
in inspector say,
	 self font: (TextStyle default fontAt: 2) emphasis: 1
"! !

!StringMorph methodsFor: 'layout' stamp: 'nk 5/11/2001 09:33'!
fullBounds
	self contents ifNil: [ self contents: 'String Morph' ].
	^super fullBounds! !

!StringMorph methodsFor: 'event handling' stamp: 'ar 10/6/2000 00:16'!
handlesMouseDown: evt
	^ (evt shiftPressed and: [self wantsKeyboardFocusOnShiftClick])
		ifTrue: [true]
		ifFalse: [super handlesMouseDown: evt].
! !

!StringMorph methodsFor: 'event handling' stamp: 'sw 9/8/1999 11:26'!
hasFocus
	^ hasFocus! !

!StringMorph methodsFor: 'accessing' stamp: 'jmv 9/7/2009 00:34'!
hasTranslucentColor

	^true! !

!StringMorph methodsFor: 'drawing' stamp: 'mt 1/19/2015 15:03'!
imageForm: depth forRectangle: rect
	| canvas |
	canvas := Display defaultCanvasClass extent: rect extent depth: depth.
	canvas form fillColor: self color negated. 
	canvas translateBy: rect topLeft negated
		during:[:tempCanvas| tempCanvas fullDrawMorph: self].
	^ canvas form offset: rect topLeft! !

!StringMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 20:42'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	font := nil.
	emphasis := 0.
	hasFocus := false! !

!StringMorph methodsFor: 'parts bin' stamp: 'dgd 2/14/2003 21:58'!
initializeToStandAlone
	super initializeToStandAlone.
	
	font := nil.
	emphasis := 0.
	hasFocus := false.
	self contents: 'String: Shift-click on me to edit'! !

!StringMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 21:57'!
initWithContents: aString font: aFont emphasis: emphasisCode 
	super initialize.
	
	font := aFont.
	emphasis := emphasisCode.
	hasFocus := false.
	self contents: aString! !

!StringMorph methodsFor: 'accessing' stamp: 'sw 9/8/1999 11:10'!
interimContents: aString
	"The receiver is under edit and aString represents the string the user sees as she edits, which typically will not have been accepted and indeed may be abandoned"

	self contents: aString! !

!StringMorph methodsFor: 'accessing' stamp: 'tak 9/7/2006 11:19'!
label: aString font: aFont
	"compatible protocol used in ScriptEditorMorph>>bringUpToDate"
	self contents: aString.
	self font: aFont! !

!StringMorph methodsFor: 'editing' stamp: 'nk 2/24/2005 20:11'!
launchMiniEditor: evt

	| textMorph |
	hasFocus := true.  "Really only means edit in progress for this morph"
	textMorph := StringMorphEditor new contentsAsIs: contents.
	textMorph beAllFont: self fontToUse.
	textMorph bounds: (self bounds expandBy: 0@2).
	self addMorphFront: textMorph.
	evt hand newKeyboardFocus: textMorph.
	textMorph editor selectFrom: 1 to: textMorph paragraph text string size! !

!StringMorph methodsFor: 'drawing' stamp: 'tk 8/1/2001 14:15'!
lookTranslucent

	"keep the text the same color (black)"! !

!StringMorph methodsFor: 'editing' stamp: 'sw 9/8/1999 10:42'!
lostFocusWithoutAccepting
	"The message is sent when the user, having been in an editing episode on the receiver, changes the keyboard focus -- typically by clicking on some editable text somewhere else -- without having accepted the current edits."

	self acceptContents! !

!StringMorph methodsFor: 'accessing' stamp: 'ar 12/30/2001 20:45'!
measureContents
	| f |
	f := self fontToUse.
	^(((f widthOfString: contents) max: self minimumWidth)  @ f height).! !

!StringMorph methodsFor: 'connectors-layout' stamp: 'nk 4/29/2004 22:49'!
minHeight
"answer the receiver's minHeight"
	^ super minHeight max: self fontToUse height! !

!StringMorph methodsFor: 'accessing' stamp: 'sw 9/8/1999 13:44'!
minimumWidth
	"Answer the minimum width that the receiver can have.  A nonzero value here keeps the receiver from degenerating into something that cannot ever be seen or touched again!!  Obeyed by fitContents."

	^ 3! !

!StringMorph methodsFor: 'event handling' stamp: 'di 9/5/1999 17:25'!
mouseDown: evt
	"If the shift key is pressed, make this string the keyboard input focus."

	(evt shiftPressed and: [self wantsKeyboardFocusOnShiftClick])
		ifTrue: [self launchMiniEditor: evt]
		ifFalse: [super mouseDown: evt].
! !

!StringMorph methodsFor: 'printing' stamp: 'jm 11/3/97 16:52'!
printOn: aStream

	super printOn: aStream.
	aStream print: contents.
! !

!StringMorph methodsFor: 'accessing' stamp: 'sw 12/6/1999 13:16'!
setWidth: width

	self extent: width @ (font ifNil: [TextStyle defaultFont]) height! !

!StringMorph methodsFor: 'accessing' stamp: 'tk 12/16/1998 11:55'!
userString
	"Do I have a text string to be searched on?"

	^ contents! !

!StringMorph methodsFor: 'accessing' stamp: 'sw 9/16/1999 22:57'!
valueFromContents
	"Return a new value from the current contents string."
	^ contents! !

!StringMorph methodsFor: 'editing' stamp: 'sw 7/21/1999 14:59'!
wantsKeyboardFocusOnShiftClick
	^ owner topRendererOrSelf wantsKeyboardFocusFor: self
! !

!StringMorph methodsFor: 'event handling' stamp: 'sw 5/6/1998 15:45'!
wouldAcceptKeyboardFocus
	^ self isLocked not! !

StringMorph subclass: #UpdatingStringMorph
	instanceVariableNames: 'format target lastValue getSelector putSelector floatPrecision growable stepTime autoAcceptOnFocusLoss minimumWidth maximumWidth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!UpdatingStringMorph commentStamp: '<historical>' prior: 0!
A StringMorph that constantly tries to show the current data from the target object.  When sent #step, it shows what the target objects has (target perform: getSelector).  When edited (with shift-click), it writes back to the target.

floatPrecision = 1. to round to integer.
floatPrecision = .1 to round to 1 decimal place, etc.

Even when ((target == nil) or: [getSelector == nil]), the user would still like to edit the string with shift-click.!

UpdatingStringMorph subclass: #NameStringInHalo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!NameStringInHalo commentStamp: 'kfr 10/27/2003 16:29' prior: 0!
Shows the name of the morph in the halo. !

!NameStringInHalo methodsFor: 'editing' stamp: 'sw 9/17/1999 13:41'!
cancelEdits
	self interimContents: target externalName.
	super cancelEdits! !

!NameStringInHalo methodsFor: 'drawing' stamp: 'sw 9/7/1999 21:27'!
drawOn: aCanvas
	aCanvas fillRectangle: self bounds color: Color white.
	super drawOn: aCanvas.! !

!NameStringInHalo methodsFor: 'accessing' stamp: 'sw 9/17/1999 13:17'!
interimContents: aString
	self contents: aString.
	self placeContents! !

!NameStringInHalo methodsFor: 'as yet unclassified' stamp: 'di 11/25/1999 23:40'!
placeContents
	| namePosition |
	(owner notNil and: [owner isInWorld]) ifTrue:
		[namePosition := owner basicBox bottomCenter -
			((self width // 2) @ (owner handleSize negated // 2 - 1)).
		namePosition := namePosition min: self world viewBox bottomRight - self extent y + 2.
		self bounds: (namePosition extent: self extent)]! !

!UpdatingStringMorph class methodsFor: 'instance creation' stamp: 'sw 3/10/2000 17:27'!
on: targetObject selector: aSymbol

	^ self new
		getSelector: aSymbol;
		target: targetObject

! !

!UpdatingStringMorph methodsFor: 'editing' stamp: ''!
acceptContents

	self informTarget.
! !

!UpdatingStringMorph methodsFor: 'target access' stamp: 'sw 3/10/2005 01:53'!
acceptValueFromTarget: v
	"Accept a value from the target"

	lastValue := v.
	self format == #string ifTrue: [^ v asString].
	self format == #symbol ifTrue: [^ v asString translated].
	(format == #default and: [v isNumber]) ifTrue:
		[^ self stringForNumericValue: v].
	^ v printString translated! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'tk 8/14/2000 23:11'!
acceptValue: aValue

	self updateContentsFrom: (self acceptValueFromTarget: aValue).
! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'dgd 8/30/2003 22:21'!
addCustomMenuItems: menu hand: aHandMorph 
	| prefix |
	super addCustomMenuItems: menu hand: aHandMorph.
	prefix := (self growable
				ifTrue: ['stop being growable']
				ifFalse: ['start being growable']) translated.
	menu add: prefix action: #toggleGrowability.
	menu add: 'decimal places...' translated action: #setPrecision.
	menu add: 'font size...' translated action: #setFontSize.
	menu add: 'font style...' translated action: #setFontStyle! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'sw 9/9/1999 16:47'!
autoAcceptOnFocusLoss
	^ autoAcceptOnFocusLoss ~~ false! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'sw 9/8/1999 10:45'!
autoAcceptOnFocusLoss: aBoolean
	autoAcceptOnFocusLoss := aBoolean! !

!UpdatingStringMorph methodsFor: 'target access' stamp: 'yo 6/7/2004 21:35'!
checkTarget
	""
	getSelector ifNil: [^ true].
	^ getSelector numArgs = 0.
! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'aoy 2/17/2003 01:15'!
contents: newContents 
	"This is the original StringMorph implementation of #contents:, restored down in UpdatingStringMorph because a recent 'optimization' of the StringMorph version of this method broke UpdatingStringMorphs."

	contents := newContents isText 
				ifTrue:  
					[emphasis := newContents emphasisAt: 1.
					newContents string]
				ifFalse: 
					[contents = newContents ifTrue: [^self].	"no substantive change"
					newContents].
	self fitContents.
	self changed! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'sw 9/13/2002 17:55'!
decimalPlaces
	"Answer the number of decimal places to show."

	| places |
	(places := self valueOfProperty: #decimalPlaces) ifNotNil: [^ places].
	self setProperty: #decimalPlaces toValue: (places := Utilities decimalPlacesForFloatPrecision: self floatPrecision).
	^ places! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'ar 1/3/2010 19:57'!
decimalPlaces: aNumber
	"Set the receiver's number of decimal places to be shown.  If my target is a morph or a player, tell it about the change, in case it wants to remember it."

	| constrained |
	self setProperty: #decimalPlaces toValue: (constrained := aNumber min: 11).
	self pvtFloatPrecision: (Utilities floatPrecisionForDecimalPlaces: constrained).
	(target isMorph or:[target isPlayer]) ifTrue:
		[target noteDecimalPlaces: constrained forGetter: getSelector]! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'nk 1/11/2004 15:29'!
doneWithEdits
	"If in a SyntaxMorph, shrink min width after editing"

	| editor |
	super doneWithEdits.
	(owner respondsTo: #parseNode) ifTrue: [minimumWidth := 8].
	editor := (submorphs detect: [ :sm | sm isKindOf: StringMorphEditor ] ifNone: [ ^self ]).
	editor delete.! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'yo 8/28/2008 15:50'!
fitContents

	| newExtent |
	newExtent := self measureContents.
	newExtent := ((newExtent x max: self minimumWidth) min: self maximumWidth) @ newExtent y.
	(self extent = newExtent) ifFalse:
		[self extent: newExtent.
		self changed]
! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'sw 7/27/2001 18:20'!
floatPrecision
	"Answer the floatPrecision to use:
		1.0 ->	show whole number
		0.1	->	show one digit of precision
		.01 ->	show two digits of precision
		etc.
	Initialize the floatPrecision to 1 if it is not already defined"

	floatPrecision isNumber ifFalse:
		[self target: target].  "Fixes up errant cases from earlier bug"
	^ floatPrecision
! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'sw 9/11/2002 14:44'!
format
	"Answer the receiver's format: #default or #string"

	^ format ifNil: [format := #default]! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: ''!
getSelector

	^ getSelector
! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: ''!
getSelector: aSymbol

	getSelector := aSymbol.
! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'jm 5/26/1999 16:21'!
growable

	^ growable ~~ false
! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'jm 5/26/1999 16:22'!
growable: aBoolean

	growable := aBoolean.
! !

!UpdatingStringMorph methodsFor: 'events-processing' stamp: 'sw 11/15/2001 18:17'!
handlerForMouseDown: evt
	"Answer an object to field the mouseDown event provided, or nil if none"

	| aHandler |
	aHandler := super handlerForMouseDown: evt.
	aHandler == self ifTrue:	[^ self]. "I would get it anyways"
	"Note: This is a hack to allow value editing in viewers"
	((owner wantsKeyboardFocusFor: self) and:
		[self userEditsAllowed]) ifTrue: [^ self].
	^ aHandler! !

!UpdatingStringMorph methodsFor: 'event handling' stamp: 'ar 10/6/2000 00:17'!
handlesMouseDown: evt
	(owner wantsKeyboardFocusFor: self)
		ifTrue:[^true].
	^ super handlesMouseDown: evt! !

!UpdatingStringMorph methodsFor: 'target access' stamp: 'dgd 2/22/2003 19:01'!
informTarget
	"Obtain a value from my contents, and tell my target about it.  The putSelector can take one argument (traditional) or two (as used by Croquet)"

	| newValue typeIn |
	(target notNil and: [putSelector notNil]) 
		ifTrue: 
			[typeIn := contents.
			(newValue := self valueFromContents) ifNotNil: 
					[self checkTarget.
					putSelector numArgs = 1 
						ifTrue: [target perform: putSelector with: newValue].
					putSelector numArgs = 2 
						ifTrue: 
							[target 
								perform: putSelector
								with: newValue
								with: self].
					target isMorph ifTrue: [target changed]].
			self fitContents.
			(format == #default and: [newValue isNumber]) 
				ifTrue: [self setDecimalPlacesFromTypeIn: typeIn]]! !

!UpdatingStringMorph methodsFor: 'initialization' stamp: 'yo 6/19/2007 23:40'!
initialize
	"Initialize the receiver to have default values in its instance variables."
	super initialize.
""
	format _ #default.
	"formats: #string, #default"
	target _ getSelector _ putSelector _ nil.
	floatPrecision _ 1.
	growable _ true.
	stepTime _ nil.
	autoAcceptOnFocusLoss _ true.
	minimumWidth _ 8.
	maximumWidth _ 366! !

!UpdatingStringMorph methodsFor: 'target access' stamp: 'sw 1/6/2005 01:27'!
isEtoyReadout
	"Answer whether the receiver can serve as an etoy readout"

	^ true! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'tk 11/29/2000 13:54'!
lostFocusWithoutAccepting
	"The message is sent when the user, having been in an editing episode on the receiver, changes the keyboard focus -- typically by clicking on some editable text somewhere else -- without having accepted the current edits."

	self autoAcceptOnFocusLoss ifTrue: [self doneWithEdits; acceptContents]! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'sw 9/10/1999 10:07'!
maximumWidth
	"Answer the maximum width that the receiver can have.   A nil value means no maximum, and for practical purposes results in a value of 99999 here temporarily, for help in future debugging"

	^ maximumWidth ifNil: [99999]! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'sw 9/10/1999 09:59'!
minimumWidth
	"Answer the minimum width that the receiver can have.  A nonzero value here keeps the receiver from degenerating into something that cannot ever be seen or touched again!!  Obeyed by fitContents."

	^ minimumWidth ifNil: [minimumWidth := 8]! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'tk 12/1/2000 15:08'!
minimumWidth: aWidth
	"Set the minimum width that the receiver can have.  A nonzero value here keeps the receiver from degenerating into something that cannot ever be seen or touched again!!  Obeyed by fitContents."

	minimumWidth := aWidth! !

!UpdatingStringMorph methodsFor: 'event handling' stamp: 'sw 11/15/2001 10:15'!
mouseDown: evt
	"The mouse went down over the receiver.  If appropriate, launch a mini-editor so that the user can commence text-editing here"

	(owner wantsKeyboardFocusFor: self) ifTrue:
		[self userEditsAllowed ifTrue:
			[(owner respondsTo: #parseNode)
					ifTrue: 	"leave space for editing"
						[minimumWidth := (49 max: minimumWidth)].
			self launchMiniEditor: evt]]! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: ''!
putSelector

	^ putSelector
! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: ''!
putSelector: aSymbol

	putSelector := aSymbol.
! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'sw 9/13/2002 17:58'!
pvtFloatPrecision: aNumber
	"Private - Set the floatPrecision instance variable to the given number"

	floatPrecision := aNumber! !

!UpdatingStringMorph methodsFor: 'target access' stamp: 'tfel 8/9/2016 16:40'!
readFromTarget
	"Update my readout from my target"

	| v ret places |
	(target isNil or: [getSelector isNil]) ifTrue: [^contents].
	ret _ self checkTarget.
	ret ifFalse: [^ '0'].
	((target isMorph) or:[target isPlayerLike]) ifTrue:[
	places _ target decimalPlacesForGetter: getSelector.
	(places ~= nil and: [ places ~= (self valueOfProperty: #decimalPlaces ifAbsent: [0])])  ifTrue: [ self decimalPlaces: places ]].
	v := target perform: getSelector.	"scriptPerformer"
	(v isKindOf: Text) ifTrue: [v := v asString].
	^self acceptValueFromTarget: v! !

!UpdatingStringMorph methodsFor: 'target access' stamp: 'sw 3/3/2005 02:55'!
setDecimalPlacesFromTypeIn: typeIn
	"The user has typed in a number as the new value of the receiver.  Glean off decimal-places-preference from the type-in"

	| decimalPointPosition tail places |
	(typeIn includes: $e) ifTrue: [^ self].
	decimalPointPosition := typeIn indexOf: $. ifAbsent: [nil].
	places := 0.
	decimalPointPosition
		ifNotNil:
			[tail := typeIn copyFrom: decimalPointPosition + 1 to: typeIn size.
			[places < tail size and: [(tail at: (places + 1)) isDigit]]
				whileTrue:
					[places := places + 1]].
		
	self decimalPlaces: places! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'sw 9/11/2002 09:57'!
setDecimalPlaces: places
	"Set the number of decimal places, and update the display."

	self decimalPlaces: places.
	self acceptValueFromTarget: lastValue! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'ar 8/6/2009 19:18'!
setFontSize
	| sizes reply family |
	family := font ifNil: [TextStyle default] ifNotNil: [font textStyle].
	family ifNil: [family := TextStyle default].  "safety net -- this line SHOULD be unnecessary now"
	sizes := 	family fontNamesWithPointSizes.
	reply := UIManager default chooseFrom: sizes values: sizes.
	reply ifNotNil:
		[self font: (family fontAt: (sizes indexOf: reply))]! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'nice 10/21/2009 00:16'!
setFontStyle
	| aList reply style |
	aList := (TextConstants select: [:anItem | anItem isKindOf: TextStyle]) 
				keys asArray.
	reply := UIManager default chooseFrom: aList values: aList.
	reply notNil 
		ifTrue: 
			[(style := TextStyle named: reply) ifNil: 
					[Beeper beep.
					^true].
			self font: style defaultFont]! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'sw 1/15/2007 02:27'!
setPrecision
	"Allow the user to specify a number of decimal places.  This UI is invoked from a menu.  Nowadays the precision can be set by simple type-in, making this menu approach mostly obsolete.  However, it's still useful for read-only readouts, where type-in is not allowed."

	| aMenu |
	aMenu _ MenuMorph new.
	aMenu addTitle: ('How many decimal places? (currently {1})' translated format: {self decimalPlaces}).
	0 to: 10 do:
		[:places |
			aMenu add: places asString target: self selector: #setDecimalPlaces: argument: places].
	aMenu popUpInWorld! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'sw 11/15/2001 20:20'!
setToAllowTextEdit
	"Set up the receiver so that it will be receptive to text editing, even if there is no putSelector provided"

	self setProperty: #okToTextEdit toValue: true! !

!UpdatingStringMorph methodsFor: 'stepping and presenter' stamp: 'sw 7/15/1999 07:28'!
step
	| s |
	super step.
	hasFocus ifFalse:
		["update contents, but only if user isn't editing this string"
		s := self readFromTarget.
		s = contents ifFalse:
			[self updateContentsFrom: s]]
! !

!UpdatingStringMorph methodsFor: 'testing' stamp: 'yo 6/19/2007 23:42'!
stepTime

	^ stepTime ifNil: [200]
! !

!UpdatingStringMorph methodsFor: 'stepping' stamp: 'jm 5/26/1999 16:23'!
stepTime: mSecsPerStep

	stepTime := mSecsPerStep truncated.
! !

!UpdatingStringMorph methodsFor: 'target access' stamp: 'sw 3/3/2005 02:53'!
stringForNumericValue: aValue
	"Answer a suitably-formatted string representing the value."

	| barePrintString |
	((barePrintString := aValue printString) includes: $e)  ifTrue: [^ barePrintString].
	^ aValue printShowingDecimalPlaces: self decimalPlaces! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: ''!
target

	^ target
! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'sw 3/11/2000 20:05'!
target: anObject

	target := anObject.
	getSelector ifNotNil: [floatPrecision := anObject defaultFloatPrecisionFor: getSelector]
! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'sw 6/26/1998 07:47'!
toggleGrowability
	growable := self growable not.
	self updateContentsFrom: self readFromTarget.
	growable ifTrue: [self fitContents]! !

!UpdatingStringMorph methodsFor: 'stepping' stamp: 'sw 6/26/1998 07:31'!
updateContentsFrom: aValue
	self growable
		ifTrue:
			[self contents: aValue]
		ifFalse:
			[self contentsClipped: aValue]! !

!UpdatingStringMorph methodsFor: 'formats' stamp: ''!
useDefaultFormat
	"Use the object's own printString format."

	format := #default.
! !

!UpdatingStringMorph methodsFor: 'editing' stamp: 'sw 11/15/2001 09:43'!
userEditsAllowed
	"Answer whether user-edits are allowed to this field"

	^ putSelector notNil or: [self hasProperty: #okToTextEdit]! !

!UpdatingStringMorph methodsFor: 'formats' stamp: ''!
useStringFormat

	format := #string.! !

!UpdatingStringMorph methodsFor: 'formats' stamp: 'yo 1/12/2005 14:28'!
useSymbolFormat

	format := #symbol.! !

!UpdatingStringMorph methodsFor: 'accessing' stamp: 'ar 1/3/2010 16:08'!
valueFromContents
	"Return a new value from the current contents string."

"
	| expression tilePadMorphOrNil asNumberBlock |
	asNumberBlock := [:string | [string asNumber]
				on: Error
				do: []].
	format = #string
		ifTrue: [^ contents].
	(format = #default
			and: [self owner isKindOf: NumericReadoutTile])
		ifTrue: [^ asNumberBlock value: contents].
	tilePadMorphOrNil := self ownerThatIsA: TilePadMorph.
	(tilePadMorphOrNil notNil
			and: [tilePadMorphOrNil type = #Number])
		ifTrue: [^ asNumberBlock value: contents].
	expression := Vocabulary eToyVocabulary translationKeyFor: contents.
	expression isNil
		ifTrue: [expression := contents].
	^ Compiler evaluate: expression
"

	format = #symbol ifTrue: [^ lastValue].
	format = #string ifTrue: [^ contents].
	(owner notNil and: [owner isNumericReadoutTile]) ifTrue: [
		^ Number readFrom: contents
	].
	target ifNotNil: [target owner ifNotNil: [
		((target owner isTilePadMorph) and: [target owner type = #Number])
			ifTrue: [^ Number readFrom: contents]]].
	^ Compiler evaluate: contents
! !

!UpdatingStringMorph methodsFor: 'copying' stamp: 'tk 1/7/1999 15:37'!
veryDeepFixupWith: deepCopier
	"If target field is weakly copied, fix it here.  If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

super veryDeepFixupWith: deepCopier.
target := deepCopier references at: target ifAbsent: [target].
! !

!UpdatingStringMorph methodsFor: 'copying' stamp: 'tfel 8/9/2016 16:40'!
veryDeepInner: deepCopier
	"Copy all of my instance variables.  Some need to be not copied at all, but shared."

	super veryDeepInner: deepCopier.
	format _ format veryDeepCopyWith: deepCopier.
	target _ target.					"Weakly copied"
	lastValue _ lastValue veryDeepCopyWith: deepCopier.
	getSelector _ getSelector.			"Symbol"
	putSelector _ putSelector.		"Symbol"
	floatPrecision _ floatPrecision veryDeepCopyWith: deepCopier.
	growable _ growable veryDeepCopyWith: deepCopier.
	stepTime _ stepTime veryDeepCopyWith: deepCopier.
	autoAcceptOnFocusLoss _ autoAcceptOnFocusLoss veryDeepCopyWith: deepCopier.
	minimumWidth _ minimumWidth veryDeepCopyWith: deepCopier.
	maximumWidth _ maximumWidth veryDeepCopyWith: deepCopier.
	self setProperty: #decimalPlaces toValue: ((self valueOfProperty: #decimalPlaces ifAbsent: [0]) veryDeepCopyWith: deepCopier).
! !

!UpdatingStringMorph methodsFor: 'event handling' stamp: 'sw 5/6/1998 12:59'!
wouldAcceptKeyboardFocus
	^ (self hasProperty: #okToTextEdit) or: [super wouldAcceptKeyboardFocus]! !

Morph subclass: #SVColorSelectorMorph
	instanceVariableNames: 'selectedColor locationMorph'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!SVColorSelectorMorph commentStamp: 'gvc 8/8/2007 14:36' prior: 0!
A colour selector that displays an area with saturation on the x axis and volume on the y axis. Provides interactive selection of colour by mouse. For the moment it is event rather than model based.
Setting the color will specify the hue and setting the selectedColor will specify the saturation and volume (may have a different hue to that displayed if not in sync).!

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 5/23/2007 11:54'!
adoptPaneColor: paneColor
	"Pass on to the border too."
	
	super adoptPaneColor: paneColor.
	self borderStyle baseColor: paneColor twiceDarker! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 14:15'!
basicColor: aColor
	"Set the gradient colors."
	
	super color: aColor asNontranslucentColor.
	self
		fillStyle: self gradient! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 15:13'!
blackGradient
	"Answer the black gradient. Top to bottom, transparent to black."

	^(InterpolatedGradientFillStyle colors: {Color black alpha: 0. Color black})
		origin: self innerBounds topLeft;
		direction: 0@self innerBounds height! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/19/2006 11:36'!
blackGradientMorph
	"Answer the black gradient morph."

	^Morph new
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		fillStyle: self blackGradient! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 15:19'!
borderWidth: anInteger
	"Update the gradients after setting."
	
	super borderWidth: anInteger.
	self updateGradients! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 10:42'!
colorAt: aPoint
	"Answer the color in the world at the given point."
	
	^self isInWorld
		ifTrue: [(Display colorAt: aPoint) asNontranslucentColor ]
		ifFalse: [Color black]! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 14:15'!
color: aColor
	"Set the gradient colors."
	
	self
		basicColor: aColor;
		selectedColor: (Color h: aColor hue s: self selectedColor saturation v: self selectedColor brightness)! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/22/2006 10:33'!
extent: p
	"Update the gradient directions."

	super extent: p.
	self updateGradients! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 10:49'!
fillStyle: fillStyle
	"If it is a color then override with gradient."
	
	fillStyle isColor
		ifTrue: [self color: fillStyle]
		ifFalse: [super fillStyle: fillStyle]! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 7/30/2009 13:52'!
gradient
	"Answer the base gradient."

	|b|
	b := self innerBounds.
	^(GradientFillStyle colors: {Color white. self color})
		origin: b topLeft;
		direction: (b width@0)! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/20/2006 12:50'!
handlesMouseDown: evt
	"Yes for down and move.." 

	^true! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 16:33'!
handlesMouseOverDragging: evt
	"Yes, make the location morph visible when leaving."
	
	^true! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'mt 4/19/2016 12:10'!
hideLocation
	"Hide the location morph and update the display."
	
	self locationMorph visible: false.
	self refreshWorld.! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/22/2006 10:22'!
initialize
	"Initialize the receiver."

	super initialize.
	self locationMorph: self newLocationMorph.
	self
		clipSubmorphs: true;
		color: Color blue;
		borderStyle: (BorderStyle inset width: 1);
		addMorphBack: self locationMorph;
		addMorphBack: self blackGradientMorph! !

!SVColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/20/2006 13:37'!
locationMorph
	"Answer the value of locationMorph"

	^ locationMorph! !

!SVColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/20/2006 13:37'!
locationMorph: anObject
	"Set the value of locationMorph"

	locationMorph := anObject! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 3/19/2010 16:21'!
mouseDown: evt 
	"Handle a mouse down event. Select the color at the mouse position."
	
	evt redButtonPressed
		ifFalse: [^super mouseDown: evt].
	evt hand showTemporaryCursor: (Cursor crossHair copy offset: -9 @ -9).
	self hideLocation.
	self selectColorAt: evt position.
	^super mouseDown: evt! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 3/19/2010 16:21'!
mouseEnterDragging: evt
	"Make the location morph invisible when entering."
	
	self hideLocation.
	evt hand showTemporaryCursor: (Cursor crossHair copy offset: -9 @ -9).! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 5/23/2007 11:56'!
mouseLeaveDragging: evt
	"Make the location morph visible when leaving."
	
	evt hand showTemporaryCursor: nil.
	self showLocation! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/22/2006 12:11'!
mouseMove: evt 
	"Handle a mouse move event. Select the color at the mouse position."
	
	evt redButtonPressed
		ifFalse: [^super mouseMove: evt].
	self selectColorAt: evt position.
	^super mouseMove: evt! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 12:12'!
mouseUp: evt 
	"Handle a up event. Show the location morph again."
	
	evt hand showTemporaryCursor: nil.
	self updateSelectedLocation.
	self locationMorph visible: true! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 12:16'!
newLocationMorph
	"Answer a new morph indicating the location of the selected color."

	^ImageMorph new
		image: Cursor crossHair withMask asCursorForm! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/22/2006 12:12'!
selectColorAt: aPoint
	"Set the color at the given position."
	
	|b p|
	b := self innerBounds.
	p := (b containsPoint: aPoint)
		ifTrue: [aPoint]
		ifFalse: [b pointNearestTo: aPoint].
	p := p - b topLeft / b extent.
	self selectedColor: (Color
		h: self color hue
		s: p x
		v: 1.0 - p y)! !

!SVColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/19/2006 15:08'!
selectedColor
	"Answer the value of selectedColor"

	^selectedColor ifNil: [self color]! !

!SVColorSelectorMorph methodsFor: 'accessing' stamp: 'gvc 9/21/2006 14:01'!
selectedColor: aColor
	"Set the value of selectedColor."

	selectedColor := aColor.
	self locationMorph visible ifTrue: [self updateSelectedLocation].
	self triggerEvent: #colorSelected with: aColor! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 15:25'!
selectedLocation
	"Answer the location within the receiver of the selected colour
	relative to the receiver's top left."

	|b c x y|
	b := self innerBounds.
	c := self selectedColor.
	x := c saturation * (b width - 1).
	y := 1 - c brightness * (b height - 1).
	^(x truncated @ y truncated) + b topLeft! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'mt 4/19/2016 12:11'!
showLocation
	"Show the location morph and update the display."
	
	self locationMorph visible: true.
	self refreshWorld.! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 7/30/2009 13:52'!
updateGradients
	"Update the gradient directions."

	|bgm b|
	b := self innerBounds.
	bgm := self submorphs last.
	bgm bounds: b.
	bgm fillStyle
		origin: b topLeft;
		direction: 0@b height.
	self fillStyle
		origin: b topLeft;
		direction: (b width@0).
	self updateSelectedLocation! !

!SVColorSelectorMorph methodsFor: 'as yet unclassified' stamp: 'gvc 9/21/2006 12:17'!
updateSelectedLocation
	"Position the location morph to indicate the selected colour."
	
	self locationMorph
		position: (self selectedLocation - (self locationMorph extent // 2 + (self locationMorph extent \\ 2)))! !

Morph subclass: #SystemProgressBarMorph
	instanceVariableNames: 'barSize barColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!SystemProgressBarMorph commentStamp: 'laza 4/9/2004 11:47' prior: 0!
Instances of this morph get used by SystemProgressMoprh to quickly display a progress bar.!

!SystemProgressBarMorph class methodsFor: 'preferences' stamp: 'mt 7/22/2016 11:27'!
themeProperties

	^ super themeProperties, {
		{ #borderColor. 'Colors'. 'Color of the progress bar''s border.' }.
		{ #borderWidth. 'Borders'. 'Width of the progress bar''s border.' }.
		{ #borderStyle. 'Borders'. 'Whether to use a plain border, inset, or outset.' }.
		{ #color. 'Colors'. 'Background color of the progress bar.' }.
		{ #barColor. 'Colors'. 'Color of the progress bar''s bar.' }.
	}! !

!SystemProgressBarMorph methodsFor: 'updating' stamp: 'mt 7/22/2016 11:29'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	
	self setDefaultParameters.! !

!SystemProgressBarMorph methodsFor: 'accessing' stamp: 'mt 8/3/2016 15:13'!
barColor

	^ barColor! !

!SystemProgressBarMorph methodsFor: 'accessing' stamp: 'mt 8/3/2016 15:14'!
barColor: aColor

	| cc fill |
	cc := aColor.

	MenuMorph gradientMenu
		ifFalse: [fill := SolidFillStyle color: cc]
		ifTrue: [
			fill := GradientFillStyle ramp: { 
				0.0 -> cc twiceLighter. 
				1 -> cc twiceDarker }].

	barColor := fill.
	self changed.! !

!SystemProgressBarMorph methodsFor: 'accessing' stamp: 'mt 8/3/2016 15:13'!
barSize

	^ barSize! !

!SystemProgressBarMorph methodsFor: 'accessing' stamp: 'laza 4/9/2004 10:37'!
barSize: anInteger
	barSize := anInteger.
	self changed.! !

!SystemProgressBarMorph methodsFor: 'drawing' stamp: 'mt 8/3/2016 15:22'!
drawOn: aCanvas

	| area fill |
	super drawOn: aCanvas.
	
	self barSize > 0 ifTrue: [
		area := self innerBounds.
		area := area origin extent: (self barSize min: area extent x)@area extent y.

		fill := self barColor isColor
			ifTrue: [SolidFillStyle color: self barColor]
			ifFalse: [self barColor].
		fill isGradientFill ifTrue: [
			fill origin: area origin.
			fill direction: 0@ area height].

		aCanvas
			fillRectangle: area
			fillStyle: fill
			borderStyle: (SimpleBorder new width: 1; color: fill asColor muchDarker).
	].
! !

!SystemProgressBarMorph methodsFor: 'initialization' stamp: 'mt 7/22/2016 11:36'!
initialize
	super initialize.

	self setDefaultParameters.
	barSize := 0.
! !

!SystemProgressBarMorph methodsFor: 'initialization' stamp: 'mt 8/3/2016 15:21'!
setDefaultParameters
	"change the receiver's appareance parameters"

	self
		color: (self userInterfaceTheme color ifNil: [Color r: 0.977 g: 0.977 b: 0.977]);
		borderStyle: (self userInterfaceTheme borderStyle ifNil: [BorderStyle default]);
		borderColor: (self userInterfaceTheme borderColor ifNil: [Color transparent]);
		borderWidth: (self userInterfaceTheme borderWidth ifNil: [0]);
		barColor: (self userInterfaceTheme barColor ifNil: [Color r: 0.72 g: 0.72 b: 0.9]).! !

Morph subclass: #SystemProgressMorph
	instanceVariableNames: 'activeSlots bars labels font lock requestedPosition textColor'
	classVariableNames: 'BarHeight BarWidth Inset UniqueInstance'
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!SystemProgressMorph commentStamp: '<historical>' prior: 0!
An single instance of this morph class is used to display progress while the system is busy, eg. while it receives code updates or does a fileIn. To give the user progress information you don't deal directly with SystemProgressMorph. You keep on using the well established way of progress notification, that has been a long time in the system, is widely used and does not depend on the existence of SystemProgressMorph. For more information on this look at the example in this class or look at the comment of the method displayProgressAt:from:to:during: in class String.

SystemProgressMorph is not meant to be used as a component inside other morphs.

You can switch back to the old style of progress display by disabling the morphicProgressStyle setting in the morphic section of the preferences.!

!SystemProgressMorph class methodsFor: 'preferences' stamp: 'mt 7/26/2016 15:40'!
applyUserInterfaceTheme

	self reset.! !

!SystemProgressMorph class methodsFor: 'instance creation' stamp: 'laza 12/8/2010 12:16'!
close: aBlock
	| slot |
	slot := aBlock value: SmallInteger maxVal. "This should prevent a redraw"
	aBlock receiver freeSlot: slot.
	
! !

!SystemProgressMorph class methodsFor: 'examples' stamp: 'ul 6/17/2011 12:31'!
example
	"SystemProgressMorph example"
	'Progress' 
		displayProgressFrom: 0 to: 1000
		during: [:bar | 0 to: 1000 do: [:i | bar value: i. (Delay forMilliseconds: 2) wait]]
! !

!SystemProgressMorph class methodsFor: 'examples' stamp: 'ul 6/17/2011 12:31'!
exampleChangeLabel
	"SystemProgressMorph exampleChangeLabel"
	| classes |
	classes := Smalltalk allClasses copyFrom: 1 to: 100.
	'InitialLabel' 
		displayProgressFrom: 0 to: classes size
		during: [:bar | 1 to: classes size do: [:i |
				bar value: i.
				bar value: i printString, '/', classes size printString, ' ', (classes at: i) printString.
				(Delay forMilliseconds: 100) wait]]
! !

!SystemProgressMorph class methodsFor: 'examples' stamp: 'laza 9/8/2009 22:04'!
exampleLabelOnly
	"SystemProgressMorph exampleLabelOnly"
	| words |
	words := #(zero one two three four five six seven eight nine ten) reversed.
	UIManager default informUserDuring: [:bar |
		words do: [:each|
			bar value: 'Countdown: ', each.
			(Delay forSeconds: 1) wait]].! !

!SystemProgressMorph class methodsFor: 'instance creation' stamp: 'laza 9/7/2009 22:48'!
informUserAt: aPoint during: workBlock
	ProgressInitiationException 
		display: ' '
		at: aPoint 
		from: 0 
		to: 0 
		during: workBlock! !

!SystemProgressMorph class methodsFor: 'class initialization' stamp: 'laza 10/9/2009 11:06'!
initialize
	"SystemProgressMorph initialize; reset"
	BarHeight := 8.
	BarWidth := 300.
	Inset := 30@30! !

!SystemProgressMorph class methodsFor: 'instance creation' stamp: 'cmm 6/4/2011 17:00'!
label: shortDescription min: minValue max: maxValue
"This method is no longer used, but kept for a while longer to ensure no difficulties updating via the trunk."
	^ self 
		position: Display center
		label: shortDescription
		min: minValue
		max: maxValue! !

!SystemProgressMorph class methodsFor: 'instance creation' stamp: 'laza 4/6/2004 21:17'!
new
	^self shouldNotImplement! !

!SystemProgressMorph class methodsFor: 'instance creation' stamp: 'cmm 5/27/2011 14:06'!
position: aPoint label: shortDescription min: minValue max: maxValue 
	UniqueInstance ifNil: [ UniqueInstance := super new ].
	^ UniqueInstance
		position: aPoint
		label: (shortDescription contractTo: 100)
		min: minValue asFloat
		max: maxValue asFloat! !

!SystemProgressMorph class methodsFor: 'instance creation' stamp: 'laza 4/9/2004 10:35'!
reset
	"SystemProgressMorph reset"
	UniqueInstance ifNotNil: [UniqueInstance delete].
	UniqueInstance := nil.! !

!SystemProgressMorph class methodsFor: 'preferences' stamp: 'mt 7/22/2016 11:13'!
themeProperties

	^ super themeProperties, {
		{ #borderColor. 'Colors'. 'Color of the progress'' border.' }.
		{ #borderWidth. 'Borders'. 'Width of the progress'' border.' }.
		{ #borderStyle. 'Borders'. 'Whether to use a plain border, inset, or outset.' }.
		{ #color. 'Colors'. 'Background color of the progress.' }.

		{ #font. 'Fonts'. 'Font for bar labels.' }.
		{ #textColor. 'Colors'. 'Color for bar labels.' }.
	}! !

!SystemProgressMorph class methodsFor: 'instance creation' stamp: 'ar 6/30/2009 23:24'!
uniqueInstance
	^UniqueInstance ifNil:[super new]! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 13:04'!
activeSlots

	^ activeSlots! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 13:04'!
activeSlots: anObject

	activeSlots := anObject! !

!SystemProgressMorph methodsFor: 'updating' stamp: 'mt 7/26/2016 15:39'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	
	self setDefaultParameters.! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 13:04'!
bars

	^ bars! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 13:04'!
bars: anObject

	bars := anObject! !

!SystemProgressMorph methodsFor: 'submorphs-add/remove' stamp: 'laza 4/20/2004 12:01'!
dismissViaHalo
	self class reset! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 13:04'!
font

	^ font! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/26/2016 15:40'!
font: anObject

	font := anObject.
	self labels select: [:ea | ea notNil] thenDo: [:ea | ea font: font].! !

!SystemProgressMorph methodsFor: 'private' stamp: 'mt 7/22/2016 13:06'!
freeSlot: number
	number > 0 ifFalse: [^self].
	lock critical: [| label |
		label := self labels at: number.
		(label isNil or: [label owner isNil]) ifTrue: [^self]. "Has been freed before"
		label delete.
		(self bars at: number) delete.
		self activeSlots: self activeSlots - 1.
		self activeSlots = 0
			ifTrue: [self delete]
			ifFalse: [self reposition]]! !

!SystemProgressMorph methodsFor: 'initialization' stamp: 'mt 7/22/2016 11:20'!
initialize
	super initialize.
	activeSlots := 0.
	bars := Array new: 10.
	labels := Array new: 10.
	lock := Semaphore forMutualExclusion.
	self setDefaultParameters;
		setProperty: #morphicLayerNumber toValue: self morphicLayerNumber;
		layoutPolicy: TableLayout new;
		listDirection: #topToBottom;
		cellPositioning: #leftCenter;
		cellInset: 5;
		listCentering: #center;
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap;
		layoutInset: Inset;
		minWidth: 150! !

!SystemProgressMorph methodsFor: 'labelling' stamp: 'laza 9/9/2009 01:15'!
labelAt: progressBlock put: aString
	"Change the label for the given progressBlock to aString."
	progressBlock value: aString! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 13:04'!
labels

	^ labels! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 13:04'!
labels: anObject

	labels := anObject! !

!SystemProgressMorph methodsFor: 'initialization' stamp: 'laza 4/18/2004 21:31'!
morphicLayerNumber
	"progress morphs are behind menus and balloons, but in front of most other stuff"
	^self valueOfProperty: #morphicLayerNumber ifAbsent: [12].
! !

!SystemProgressMorph methodsFor: 'private' stamp: 'mt 7/26/2016 15:39'!
nextSlotFor: shortDescription
	
	lock critical: [ | label bar slots |
		slots := self labels size.
		self activeSlots = slots ifTrue: [^0].
		self activeSlots: self activeSlots + 1.
		1 to: slots do: [:index |
			label := (self labels at: index).
			label ifNil: [
				bar := self bars at: index put: (SystemProgressBarMorph new extent: BarWidth@BarHeight).
				label := self labels at: index put: ((StringMorph contents: shortDescription font: self font) color: self textColor).
				self
					addMorphBack: label;
					addMorphBack: bar.
				^index].
			label owner ifNil: [
				bar := self bars at: index.
				label := self labels at: index.
				self
					addMorphBack: (label contents: shortDescription);
					addMorphBack: (bar barSize: 0).
				^index]]]
		! !

!SystemProgressMorph methodsFor: 'private' stamp: 'mt 7/22/2016 13:07'!
position: aPoint label: shortDescription min: minValue max: maxValue
	| slot range barSize lastRefresh |
	self requestedPosition: aPoint.
	((range := maxValue - minValue) < 0 or: [(slot := self nextSlotFor: shortDescription) = 0])
		ifTrue: [^[:barVal| 0 ]].
	range <= 0 ifTrue: [self removeMorph: (self bars at: slot)].
	self reposition.
	self openInWorld.
	barSize := -1. "Enforces a inital draw of the morph"
	lastRefresh := 0.
	^[:barVal | | newBarSize |
		barVal isString ifTrue: [
			self setLabel: barVal at: slot.
			self currentWorld displayWorld].
		(barVal isNumber and: [range >= 1 and: [barVal between: minValue and: maxValue]]) ifTrue: [
			newBarSize := (barVal - minValue / range * BarWidth) truncated.
			newBarSize = barSize ifFalse: [
				barSize := newBarSize.
				(self bars at: slot) barSize: barSize.
				Time utcMicrosecondClock - lastRefresh > 25000 ifTrue: [
					self currentWorld displayWorld.
					lastRefresh := Time utcMicrosecondClock]]].
		slot]! !

!SystemProgressMorph methodsFor: 'private' stamp: 'cmm 6/4/2011 16:58'!
recenter
	| position |
"This method is no longer used, but kept for a while longer to ensure no difficulties updating via the trunk."
	"Put ourself in the center of the display"
	self align: self fullBounds center with: Display boundingBox center.
	"Check to see if labels are wider than progress bars. In that case do
	a centered instead of the default left aligned layout."
	position :=	self width > (Inset x * 2 + (self borderWidth * 2) + BarWidth)
					ifTrue: [#topCenter]
					ifFalse: [#leftCenter].
	self cellPositioning: position! !

!SystemProgressMorph methodsFor: 'private' stamp: 'mt 7/22/2016 13:09'!
reposition
	"Put ourself in the requested position on the display, but ensure completely within the bounds of the display"
	| position |
	self bounds:
		((self fullBounds
			align: self fullBounds center
			with: (self requestedPosition ifNil: [ self fullBounds center ])) translatedToBeWithin: Display boundingBox).
	"Check to see if labels are wider than progress bars. In that case do
	a centered instead of the default left aligned layout."
	position := self width > (Inset x * 2 + (self borderWidth * 2) + BarWidth)
		ifTrue: [ #topCenter ]
		ifFalse: [ #leftCenter ].
	self cellPositioning: position! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 13:04'!
requestedPosition

	^ requestedPosition! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/22/2016 13:04'!
requestedPosition: anObject

	requestedPosition := anObject! !

!SystemProgressMorph methodsFor: 'initialization' stamp: 'mt 7/26/2016 15:39'!
setDefaultParameters
	"change the receiver's appareance parameters"

	self
		color: (self userInterfaceTheme color ifNil: [Color r: 0.9 g: 0.9 b: 0.9]);
		borderStyle: (self userInterfaceTheme borderStyle ifNil: [BorderStyle default]);
		borderColor: (self userInterfaceTheme borderColor ifNil: [Color gray]);
		borderWidth: (self userInterfaceTheme borderWidth ifNil: [1]).

	Preferences menuAppearance3d ifTrue: [self addDropShadow].

	self
		font: (self userInterfaceTheme font ifNil: [TextStyle defaultFont]);
		textColor: (self userInterfaceTheme textColor ifNil: [Color black]).

	self
		updateColor: self
		color: self color
		intensity: 1.! !

!SystemProgressMorph methodsFor: 'labelling' stamp: 'mt 7/22/2016 13:09'!
setLabel: shortDescription at: slot
	(self labels at: slot) contents: shortDescription.
	self reposition! !

!SystemProgressMorph methodsFor: 'dropping/grabbing' stamp: 'laza 4/20/2004 11:38'!
slideToTrash: evt
	"If the user needs to dismiss a progress morph by hand, start with a 
	fresh instance next time."
	self dismissViaHalo! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/26/2016 15:38'!
textColor

	^ textColor ifNil: [Color black]! !

!SystemProgressMorph methodsFor: 'accessing' stamp: 'mt 7/26/2016 15:41'!
textColor: aColor

	textColor := aColor.
	self labels select: [:ea | ea notNil] thenDo: [:ea | ea color: textColor].! !

!SystemProgressMorph methodsFor: 'initialization' stamp: 'mt 7/26/2016 15:36'!
updateColor: aMorph color: aColor intensity: anInteger 
	"update the apareance of aMorph"
	| fill cc |
	MenuMorph gradientMenu
		ifFalse: [^ self].

	
	cc := aColor adjustSaturation: -0.08 brightness: 0.4.
	fill := GradientFillStyle ramp: {
		0.0 -> cc. 
		0.25 -> (aColor mixed: 0.5 with: cc). 
		1.0 -> aColor}.
	
	fill radial: false;
		origin: aMorph topLeft;
		direction: 0 @ aMorph height.
	aMorph fillStyle: fill! !

!SystemProgressMorph methodsFor: 'rounding' stamp: 'mt 7/22/2016 11:11'!
wantsRoundedCorners

	^ MenuMorph roundedMenuCorners or: [super wantsRoundedCorners]! !

Morph subclass: #TransferMorph
	instanceVariableNames: 'transferType passenger draggedMorph source dropNotifyRecipient resultRecipient copy'
	classVariableNames: 'CopyPlusIcon'
	poolDictionaries: ''
	category: 'Morphic-Support'!

!TransferMorph commentStamp: 'nk 6/16/2003 16:52' prior: 0!
This is a Morph that is used to visually indicate the progress of a drag operation, and also as a container for various bits of drag state information.

It polls the shift state in its step method to update its copy state (shift pressed = should copy).

And if you hit the Escape key while dragging, it aborts the drag operation.!

!TransferMorph class methodsFor: 'class initialization' stamp: 'mir 5/5/2000 14:49'!
initIcons
	"TransferMorph initIcons"

	CopyPlusIcon := Form
		extent: 16@16
		depth: 8
		fromArray: #( 0 0 65535 0 0 0 16768220 4278190080 0 0 16768220 4278190080 0 255 4294958300 4294967040 0 65500 3705461980 3705462015 0 65500 3705461980 3705462015 0 255 4294958300 4294967295 0 0 16768220 4278190080 0 0 16768220 4278190080 0 0 65535 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0)
		offset: 0@0! !

!TransferMorph class methodsFor: 'class initialization' stamp: 'mir 5/5/2000 14:48'!
initialize
	"TransferMorph initialize"

	self initIcons! !

!TransferMorph class methodsFor: 'instance creation' stamp: 'mt 11/4/2015 18:08'!
withPassenger: anObject 

	^ self
		withPassenger: anObject
		from: nil! !

!TransferMorph class methodsFor: 'instance creation' stamp: 'mt 11/4/2015 18:08'!
withPassenger: anObject from: source 
	
	^ self new
		passenger: anObject;
		source: source;
		yourself! !

!TransferMorph methodsFor: 'dropping/grabbing' stamp: 'mt 11/4/2015 18:31'!
aboutToBeGrabbedBy: aHand 
	"The receiver is being grabbed by a hand.                           
	Perform necessary adjustments (if any) and return the actual morph    
	     that should be added to the hand."
	"Since this morph has been initialized automatically with bounds origin   
	     0@0, we have to move it to aHand position."
	super aboutToBeGrabbedBy: aHand.

	self align: self fullBounds bottomLeft with: aHand position.
	aHand newKeyboardFocus: self.! !

!TransferMorph methodsFor: 'private' stamp: 'sr 6/6/2000 07:19'!
animationForMoveSuccess: success 
	| start stop slideForm |
	success
		ifTrue: [^ self]
		ifFalse: 
			[start := self fullBounds origin.
			stop := self source bounds origin].
	start = stop ifTrue: [^ self].
	slideForm := self imageFormForRectangle: ((self fullBounds origin corner: self fullBounds corner + self activeHand shadowOffset)
					merge: self activeHand bounds).
	slideForm offset: 0 @ 0.
	slideForm
		slideWithFirstFrom: start
		to: stop
		nSteps: 12
		delay: 20! !

!TransferMorph methodsFor: 'initialization' stamp: 'dgd 3/7/2003 15:00'!
defaultColor
"answer the default color/fill style for the receiver"
	^ Color blue alpha: 0.4! !

!TransferMorph methodsFor: 'event handling' stamp: 'mt 11/4/2015 18:17'!
doCopy

	copy := true.
	self updateCopyIcon.! !

!TransferMorph methodsFor: 'event handling' stamp: 'mt 11/4/2015 18:17'!
doMove

	copy := false.
	self updateCopyIcon.! !

!TransferMorph methodsFor: 'accessing' stamp: 'panda 4/28/2000 16:11'!
dragTransferType
	^transferType! !

!TransferMorph methodsFor: 'accessing' stamp: 'panda 4/28/2000 16:11'!
dragTransferType: aSymbol
	transferType := aSymbol! !

!TransferMorph methodsFor: 'initialization' stamp: 'mt 11/4/2015 18:33'!
initialize

	super initialize.

	self
		changeTableLayout;
		listDirection: #leftToRight;
		hResizing: #shrinkWrap;
		vResizing: #shrinkWrap;
		layoutInset: 3;
		cellInset: 3;
		wrapCentering: #center;
		cellPositioning: #leftCenter;
		setProperty: #indicateKeyboardFocus toValue: #never.
	
	self doMove.
	
	self on: #keyStroke send: #keyStroke: to: self.
	self on: #keyUp send: #updateFromUserInputEvent: to: self.
	self on: #keyDown send: #updateFromUserInputEvent: to: self.! !

!TransferMorph methodsFor: 'dropping/grabbing' stamp: 'cmm 3/11/2010 22:59'!
justDroppedInto: targetMorph event: anEvent 
	"If only world wants this TransferMorph, treat it as unaccepted (see also >>delete)."
	super
		justDroppedInto: targetMorph
		event: anEvent.
	self animationForMoveSuccess: true.
	self delete! !

!TransferMorph methodsFor: 'event handling' stamp: 'mt 11/4/2015 18:26'!
keyStroke: evt
	"Abort the drag on an escape"

	evt keyCharacter = Character escape ifTrue: [self delete].! !

!TransferMorph methodsFor: 'dropping/grabbing' stamp: 'cmm 3/11/2010 22:10'!
morphToDropInPasteUp: aPasteUpMorph
	^ aPasteUpMorph morphToDropForTransferMorph: self! !

!TransferMorph methodsFor: 'accessing' stamp: 'sr 4/16/2000 18:52'!
passenger
	^passenger! !

!TransferMorph methodsFor: 'accessing' stamp: 'mt 11/4/2015 18:13'!
passenger: anObject

	passenger := anObject.

	self
		removeAllMorphs;
		addMorph: passenger asDraggableMorph;
		updateCopyIcon.! !

!TransferMorph methodsFor: 'accessing' stamp: 'mir 5/5/2000 14:39'!
shouldCopy
	^copy! !

!TransferMorph methodsFor: 'accessing' stamp: 'mt 11/4/2015 15:28'!
shouldMove
	^ self shouldCopy not! !

!TransferMorph methodsFor: 'accessing' stamp: 'sr 4/16/2000 11:55'!
source
	^source! !

!TransferMorph methodsFor: 'accessing' stamp: 'sr 4/16/2000 18:53'!
source: anObject
	source := anObject! !

!TransferMorph methodsFor: 'dropping/grabbing' stamp: 'ar 10/6/2000 17:30'!
undoGrabCommand
	^nil! !

!TransferMorph methodsFor: 'private' stamp: 'mt 11/4/2015 18:16'!
updateCopyIcon

	(self submorphNamed: #tmCopyIcon)
		ifNil: [self shouldCopy ifTrue: [
			self addMorphFront: (ImageMorph new image: CopyPlusIcon; name: #tmCopyIcon; yourself)]]
		ifNotNil: [:copyIcon | self shouldCopy ifFalse: [
			copyIcon delete]]! !

!TransferMorph methodsFor: 'event handling' stamp: 'mt 11/4/2015 18:18'!
updateFromUserInputEvent: evt

	evt shiftPressed
		ifTrue: [self doCopy]
		ifFalse: [self doMove].! !

!TransferMorph methodsFor: 'dropping/grabbing' stamp: 'cmm 3/11/2010 22:49'!
wantsToBeDroppedInto: aMorph
	^ aMorph isWorldMorph
		ifTrue: [ aMorph hasTransferMorphConverter ]
		ifFalse: [ super wantsToBeDroppedInto: aMorph ]! !

Morph subclass: #TransformMorph
	instanceVariableNames: 'transform smoothing localBounds'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!TransformMorph commentStamp: 'wiz 11/6/2005 15:59' prior: 0!
A TransformMorph introduces a 2-D transformation between its (global) coordinates and the (local) coordinates of its submorphs, while also clipping all display to its bounds.  Specifically, with no offset, angle or scaling, a submorph with coordinates (0@0) will appear exactly at the topLeft of the windowMorph (its position).  Rotation and scaling are relative to the local origin, (0@0).

instance var	type				description
 transform		MorphicTransform	The coordinate transform between my coordinates and the
									local coordinates of my submorphs.
 smoothing		anInteger in 1..3	Perform smoothing of my contents during drawing
										1 No smoothing (#smoothingOff)
										2 Smoothing w/ edge adjacent pixels (#smoothingOn)
										3 Smoothing w/ edge and corner adj pixels
			
 localBounds	Rectangle or nil		caches the value of #localSubmorphBounds for performance

TransformMorphs operate with two different display strategies, depending on whether the transformation is a pure translation or not.  If so, then they simply use a clipping canvas and display their submorphs with the appropriate offset.  If the transformation includes scaling or rotation, then a caching canvas is used, whose active area covers the fullBounds of the submorphs intersected with the source quadrilateral corresponding to the window bounds.!

TransformMorph subclass: #TransformationMorph
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Basic'!

!TransformationMorph commentStamp: 'mk 8/16/2005 11:58' prior: 0!
A TransformationMorph is like a transformMorph, except that it does not clip, and its bounds include its entire submorph.  TransformationMorphs are assumed to have only one submorph -- the idea is that it is a wrapper that enables its submorph to scale and rotate.  A TMorph may come to have more than one submorph if, eg, a menu sprouts a sub menu, using the transformationMorph temporarily as its world, but this ability is only sparsely supported (as in layoutChanged).

See TransformationMorph class example1 method.!

!TransformationMorph class methodsFor: 'example' stamp: 'mk 8/14/2005 16:07'!
example1
	| stringMorph transformationMorph |
	stringMorph := 'vertical text' asMorph.
	transformationMorph := TransformationMorph new asFlexOf: stringMorph.
	transformationMorph angle: Float pi / 2.
	transformationMorph position: 5@5.
	transformationMorph openInWorld.! !

!TransformationMorph methodsFor: 'private' stamp: 'ar 1/3/2010 15:52'!
adjustAfter: changeBlock 
	"Cause this morph to remain cetered where it was before, and
	choose appropriate smoothing, after a change of scale or rotation."
	| oldRefPos |
	oldRefPos := self referencePosition.
	changeBlock value.
	self chooseSmoothing.
	self actorStateOrNil ifNotNil:[
		self penUpWhile: [self position: self position + (oldRefPos - self referencePosition)].
	].
	self layoutChanged.
	owner ifNotNil: [owner invalidRect: bounds]
! !

!TransformationMorph methodsFor: 'initialization' stamp: 'di 2/21/98 14:35'!
asFlexOf: aMorph
	"Initialize me with position and bounds of aMorph,
	and with an offset that provides centered rotation."
	| pos |
	pos := aMorph position.
	self addMorph: aMorph.
	aMorph position: (aMorph extent // 2) negated.
	self position: pos.
	transform := transform withOffset: aMorph position - pos
! !

!TransformationMorph methodsFor: 'private' stamp: 'yo 9/21/2006 12:40'!
chooseSmoothing
	"Choose appropriate smoothing, after a change of scale or rotation."

	smoothing := (self scale < 1.0 or: [self angle ~= (self angle roundTo: Float pi / 2.0)]) 
		ifTrue: [1]
		ifFalse: [1]! !

!TransformationMorph methodsFor: 'geometry' stamp: 'ar 11/21/2000 17:00'!
computeBounds
	self hasSubmorphs ifTrue:
		[bounds := (transform localBoundsToGlobal:
					(Rectangle merging:
						(self submorphs collect: [:m | m fullBounds]))) truncated
				expandBy: 1].
	fullBounds := bounds.! !

!TransformationMorph methodsFor: 'geometry eToy' stamp: 'di 10/1/2000 11:48'!
degreesOfFlex
	"Return any rotation due to flexing"
	^ self rotationDegrees! !

!TransformationMorph methodsFor: 'drawing' stamp: 'di 2/23/98 19:59'!
drawOn: aCanvas
	submorphs isEmpty ifTrue: [super drawOn: aCanvas]! !

!TransformationMorph methodsFor: 'geometry' stamp: 'nice 12/27/2009 20:21'!
extent: newExtent
	
	self adjustAfter:
		[ | scaleFactor |
		scaleFactor := (self scale * newExtent r / self fullBounds extent r) max: 0.1.
		self scale: (scaleFactor detentBy: 0.1 atMultiplesOf: 1.0 snap: false)]! !

!TransformationMorph methodsFor: 'initialization' stamp: 'di 9/30/1998 23:12'!
flexing: aMorph byTransformation: tfm
	"Initialize me with position and bounds of aMorph,
	and with an offset that provides centered rotation."

	(aMorph isKindOf: TransformationMorph)
		ifTrue: [aMorph submorphsDo: [:m | self addMorph: m clone]]
		ifFalse: [self addMorph: aMorph].
	transform := tfm.
	self chooseSmoothing.
	self layoutChanged.! !

!TransformationMorph methodsFor: 'accessing' stamp: 'wiz 12/7/2006 19:10'!
forwardDirection
	"Return the rendee's forward direction. 
	If I have no rendee then return 0.0 degrees "
	| rendee |
	( rendee := self renderedMorph) == self  ifTrue: [ ^ 0.0 ] .
	
	^ rendee forwardDirection! !

!TransformationMorph methodsFor: 'geometry eToy' stamp: 'wiz 12/7/2006 19:10'!
forwardDirection: degrees
 "If we have a rendee set its forward direction. Else do nothing." 

| rendee |
( rendee := self renderedMorph) == self ifTrue: [ ^ self  ] .
	^rendee forwardDirection: degrees! !

!TransformationMorph methodsFor: 'dropping/grabbing' stamp: 'ar 10/22/2000 18:26'!
grabTransform
	"Return the transform for the receiver which should be applied during grabbing"
	self renderedMorph isWorldMorph 
		ifTrue:[^owner ifNil:[IdentityTransform new] ifNotNil:[owner grabTransform]].
	^owner ifNil:[self transform] ifNotNil:[owner grabTransform composedWithLocal: self transform]! !

!TransformationMorph methodsFor: 'accessing' stamp: 'jm 4/25/1998 05:55'!
hasNoScaleOrRotation

	^ transform isPureTranslation
! !

!TransformationMorph methodsFor: 'geometry eToy' stamp: 'wiz 12/7/2006 16:30'!
heading
	"End recusion when necessary."
	| rendee |
	(rendee := self renderedMorph) == self ifTrue: [ ^0.0 ] .
	^ rendee heading! !

!TransformationMorph methodsFor: 'geometry eToy' stamp: 'wiz 4/18/2007 01:16'!
heading: newHeading
 "If we have a rendee set its heading. Else do nothing." 

| rendee |
( rendee := self renderedMorph) == self ifTrue: [ ^ self  ] .
	^rendee heading: newHeading! !

!TransformationMorph methodsFor: 'naming' stamp: 'sw 5/13/1998 10:32'!
innocuousName
	| r |
	^ (r := self renderedMorph) == self
		ifTrue: [super innocuousName] ifFalse: [r innocuousName]! !

!TransformationMorph methodsFor: 'classification' stamp: 'jm 4/17/1998 00:44'!
isFlexMorph

	^ true
! !

!TransformationMorph methodsFor: 'classification' stamp: 'jm 5/7/1998 13:46'!
isRenderer

	^ true
! !

!TransformationMorph methodsFor: 'testing' stamp: 'wiz 11/24/2006 00:09'!
isSticky
submorphs isEmpty ifFalse: [ ^ 	submorphs first isSticky ] .
	
^false! !

!TransformationMorph methodsFor: 'layout' stamp: 'ar 10/25/2000 16:21'!
layoutChanged
	"Recompute bounds as a result of change"
	self computeBounds.
	super layoutChanged! !

!TransformationMorph methodsFor: 'printing' stamp: 'dgd 2/21/2003 22:42'!
printOn: aStream 
	super printOn: aStream.
	submorphs isEmpty 
		ifTrue: [aStream nextPutAll: ' with no transformee!!']
		ifFalse: [aStream nextPutAll: ' on ' , submorphs first printString]! !

!TransformationMorph methodsFor: 'geometry eToy' stamp: 'sw 3/28/2001 14:24'!
referencePosition
	"Answer the  receiver's reference position, bullet-proofed against infinite recursion in the unlikely but occasionally-seen case that I am my own renderee"

	| rendered |
	^ (rendered := self renderedMorph) == self
		ifTrue:
			[super referencePosition]
		ifFalse:
			[transform localPointToGlobal: rendered referencePosition]! !

!TransformationMorph methodsFor: 'menu' stamp: 'nice 12/27/2009 20:21'!
removeFlexShell
	"Remove the shell used to make a morph rotatable and scalable."

	| oldHalo unflexed pensDown myWorld refPos aPosition |
	refPos := self referencePosition.
	myWorld := self world.
	oldHalo := self halo.
	submorphs isEmpty ifTrue: [^ self delete].
	aPosition := (owner submorphIndexOf: self) ifNil: [1].
	unflexed := self firstSubmorph.
	pensDown := OrderedCollection new.
	self allMorphsDo:  "Note any pens down -- must not be down during the move"
		[:m | | player |
		((player := m player) notNil and: [player getPenDown]) ifTrue:
			[m == player costume ifTrue:
				[pensDown add: player.
				player setPenDown: false]]].
	self submorphs do: [:m |
		m position: self center - (m extent // 2).
		owner addMorph: m asElementNumber: aPosition].
	unflexed absorbStateFromRenderer: self.
	pensDown do: [:p | p setPenDown: true].
	oldHalo ifNotNil: [oldHalo setTarget: unflexed].
	myWorld ifNotNil: [myWorld startSteppingSubmorphsOf: unflexed].
	self delete.
	unflexed referencePosition: refPos.
	^ unflexed! !

!TransformationMorph methodsFor: 'classification' stamp: 'wiz 12/7/2006 14:05'!
renderedMorph
"We are a renderer. Answer appropriately."

submorphs isEmpty ifTrue: [^self].
	^self firstSubmorph renderedMorph! !

!TransformationMorph methodsFor: 'submorphs-add/remove' stamp: 'di 11/18/1999 15:44'!
replaceSubmorph: oldMorph by: newMorph
	| t b |
	t := transform.
	b := bounds.
	super replaceSubmorph: oldMorph by: newMorph.
	transform := t.
	bounds := b.
	self layoutChanged! !

!TransformationMorph methodsFor: 'rotate scale and flex' stamp: 'di 2/20/98 14:53'!
rotationDegrees
	^ self angle radiansToDegrees negated! !

!TransformationMorph methodsFor: 'accessing' stamp: 'ar 9/22/2000 14:17'!
rotationDegrees: degrees
	self adjustAfter:[self angle: degrees degreesToRadians negated]! !

!TransformationMorph methodsFor: 'accessing' stamp: 'sw 2/15/2002 02:27'!
scaleFactor
	"Answer the scaleFactor"

	^ transform scale! !

!TransformationMorph methodsFor: 'accessing' stamp: 'jdl 3/28/2003 08:13'!
scaleToMatch: aPoint 
	| scaleFactor tfm originalScale |
	tfm := transform withScale: 1.0.
	originalScale := ((tfm localBoundsToGlobal: self renderedMorph fullBounds) 
				corner - (tfm localPointToGlobal: self renderedMorph referencePosition)) 
				r.
	"Catch cases where the reference point is on fullBounds corner"
	originalScale := originalScale max: 1.0.
	scaleFactor := (aPoint - self referencePosition) r / originalScale.
	scaleFactor := scaleFactor < 1.0 
				ifTrue: 
					[scaleFactor 
						detentBy: 0.05
						atMultiplesOf: 0.25
						snap: false]
				ifFalse: 
					[scaleFactor 
						detentBy: 0.1
						atMultiplesOf: 0.5
						snap: false].
	self adjustAfter: [self scale: ((scaleFactor min: 8.0) max: 0.1)]! !

!TransformationMorph methodsFor: 'geometry eToy' stamp: 'ar 6/12/2001 05:23'!
setDirectionFrom: aPoint
	| delta degrees inner |
	inner := self renderedMorph.
	inner == self ifTrue:[^self].
	delta := (inner transformFromWorld globalPointToLocal: aPoint) - inner referencePosition.
	degrees := delta degrees + 90.0.
	self forwardDirection: (degrees \\ 360) rounded.
! !

!TransformationMorph methodsFor: 'dropping/grabbing' stamp: 'tpr 2/4/2016 20:48'!
shouldDropOnMouseUp
	"check if my submorph should drop "
	^self firstSubmorph shouldDropOnMouseUp! !

!TransformationMorph methodsFor: 'geometry eToy' stamp: 'sw 10/6/2004 12:15'!
simplySetVisible: aBoolean
	"Set the receiver's visibility property.  This mild circumlocution is because my own #visible: method would also set the visibility flag of my flexee, which in this case is pointless because it's the flexee that calls this."

	super visible: aBoolean! !

!TransformationMorph methodsFor: 'testing' stamp: 'mdr 10/3/2000 11:28'!
stepTime
	"Answer the stepTime of my rendered morph if posible"

	| rendered |
	rendered := self renderedMorph.
	rendered = self ifTrue: [^super stepTime].	"Hack to avoid infinite recursion"
	^rendered stepTime.
	! !

!TransformationMorph methodsFor: 'geometry' stamp: 'ar 10/25/2000 16:24'!
transformedBy: aTransform
	self changed.
	self transform: (self transform composedWithGlobal: aTransform).
	self computeBounds.
	self changed.! !

!TransformationMorph methodsFor: 'geometry eToy' stamp: 'sw 10/6/2004 11:33'!
visible: aBoolean
	"Set the receiver's visibility property"

	super visible: aBoolean.
	submorphs isEmptyOrNil ifFalse: [submorphs first visible: aBoolean]! !

!TransformMorph methodsFor: 'menu' stamp: 'dgd 8/30/2003 22:20'!
addCustomMenuItems: aCustomMenu hand: aHandMorph
	super addCustomMenuItems: aCustomMenu hand: aHandMorph.
	smoothing = 1
		ifTrue: [aCustomMenu add: 'turn on smoothing' translated action: #smoothingOn]
		ifFalse: [aCustomMenu add: 'turn off smoothing' translated action: #smoothingOff]! !

!TransformMorph methodsFor: 'accessing' stamp: ''!
angle
	^ transform angle! !

!TransformMorph methodsFor: 'accessing' stamp: 'ar 1/30/2001 23:20'!
angle: newAngle

	self changed.
	transform := transform withAngle: newAngle.
	self layoutChanged.
	self changed! !

!TransformMorph methodsFor: 'accessing' stamp: 'di 2/23/98 14:44'!
colorForInsets
	^ owner ifNil: [color] ifNotNil: [owner color]! !

!TransformMorph methodsFor: 'geometry testing' stamp: 'wiz 11/6/2005 15:53'!
containsPoint: aPoint
	(bounds containsPoint: aPoint) ifFalse: [^ false].
	self hasSubmorphs
		ifTrue: [ | localPoint |  localPoint := (transform globalPointToLocal: aPoint) .
				self submorphsDo: 
					[:m | (m containsPoint: localPoint)
							ifTrue: [^ true]].
				^ false]
		ifFalse: [^ true]! !

!TransformMorph methodsFor: 'initialization' stamp: 'dgd 2/14/2003 22:31'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color lightGreen! !

!TransformMorph methodsFor: 'drawing' stamp: 'ar 8/30/2009 14:03'!
drawSubmorphsOn: aCanvas

	aCanvas transformBy: transform
		clippingTo: self innerBounds
		during: [:myCanvas |
			(self angle ~= 0.0 or: [self scale ~= 1.0])
				ifTrue:[ 
					AbstractFont forceNonSubPixelDuring:[
						submorphs reverseDo:[:m | myCanvas fullDrawMorph: m] ] ]
				ifFalse:[
					submorphs reverseDo:[:m | myCanvas fullDrawMorph: m] ] ] 
		smoothing: smoothing! !

!TransformMorph methodsFor: 'geometry' stamp: 'mt 4/22/2015 09:23'!
extent: aPoint
	"Change my layout only if I change my extent. See my implementation of #layoutChanged."

	(bounds extent closeTo: aPoint) ifTrue: [^ self].
	super extent: aPoint.
	super layoutChanged.! !

!TransformMorph methodsFor: 'dropping/grabbing' stamp: 'ar 10/22/2000 18:00'!
grabTransform
	"Return the transform for the receiver which should be applied during grabbing"
	^owner ifNil:[self transform] ifNotNil:[owner grabTransform composedWithLocal: self transform]! !

!TransformMorph methodsFor: 'initialization' stamp: 'mt 3/1/2015 17:21'!
initialize
	"initialize the state of the receiver"
	super initialize.
	""
	
	smoothing := 1.
	transform := MorphicTransform identity.
	self clipSubmorphs: true.! !

!TransformMorph methodsFor: 'change reporting' stamp: 'tpr 2/4/2016 17:44'!
invalidRect: damageRect from: aMorph
	"Translate damage reports from submorphs by the scrollOffset."
	aMorph == self
		ifTrue:[super invalidRect: damageRect from: self]
		ifFalse:[super invalidRect: (((transform localBoundsToGlobal: damageRect) intersect: bounds) expanded) from: self].! !

!TransformMorph methodsFor: 'layout' stamp: 'mt 3/1/2015 17:34'!
layoutBounds

	^ (transform globalBoundsToLocal: super layoutBounds) truncated
! !

!TransformMorph methodsFor: 'geometry' stamp: 'mt 4/22/2015 09:19'!
layoutChanged

	"A submorph could have moved, thus changing my localBounds. Invalidate the cache."
	localBounds := nil.
	"Only discard my fullBounds. Do not tell my owner."
	fullBounds := nil.! !

!TransformMorph methodsFor: 'geometry' stamp: 'efc 7/24/2003 16:41'!
localSubmorphBounds
	"Answer, in my coordinate system, the bounds of all my submorphs (or nil if no submorphs). We will cache this value for performance. The value is invalidated upon recieving #layoutChanged."

	localBounds ifNil:[
		self submorphsDo:[:m |
			localBounds ifNil: [localBounds := m fullBounds]
						ifNotNil: [localBounds := localBounds quickMerge: m fullBounds]].
	].	

	^ localBounds! !

!TransformMorph methodsFor: 'geometry' stamp: 'nk 4/12/2002 14:02'!
localVisibleSubmorphBounds
	"Answer, in my coordinate system, the bounds of all my visible submorphs (or nil if no visible submorphs)"
	| subBounds |
	subBounds := nil.
	self submorphsDo: [:m |
		(m visible) ifTrue: [
			subBounds
				ifNil: [subBounds := m fullBounds copy]
				ifNotNil: [subBounds := subBounds quickMerge: m fullBounds]]
			].
	^subBounds! !

!TransformMorph methodsFor: 'geometry' stamp: 'nk 3/8/2004 11:05'!
numberOfItemsInView
	"Answer the number of my submorphs whose (transformed) bounds intersect mine.
	This includes items that are only partially visible.
	Ignore visibility of submorphs."

	^(submorphs select: [ :ea | self innerBounds intersects: (transform localBoundsToGlobal: ea bounds) ]) size! !

!TransformMorph methodsFor: 'geometry' stamp: 'nk 3/8/2004 11:06'!
numberOfItemsPotentiallyInView
	"Answer the number of items that could potentially be viewed in full,
	computed as my visible height divided by the average height of my submorphs.
	Ignore visibility of submorphs."

	^self innerBounds height // (self localSubmorphBounds height / submorphs size)! !

!TransformMorph methodsFor: 'geometry' stamp: 'ge 9/6/2006 17:51'!
numberOfItemsPotentiallyInViewWith: submorphCount
	"Answer the number of items that could potentially be viewed in full,
	computed as my visible height divided by the average height of my submorphs.
	Ignore visibility of submorphs."

	^self innerBounds height // (self localSubmorphBounds height / submorphCount)! !

!TransformMorph methodsFor: 'accessing' stamp: ''!
offset
	^ transform offset + self innerBounds topLeft! !

!TransformMorph methodsFor: 'accessing' stamp: ''!
offset: newOffset

	transform := transform withOffset: newOffset - self innerBounds topLeft.
	self changed! !

!TransformMorph methodsFor: 'private' stamp: 'kfr 8/7/2004 18:48'!
privateFullMoveBy: delta
	"Private!! Relocate me, but not my subMorphs."

	self privateMoveBy: delta.
	transform :=  (transform asMorphicTransform) withOffset: (transform offset - delta).
! !

!TransformMorph methodsFor: 'accessing' stamp: ''!
scale
	^ transform scale! !

!TransformMorph methodsFor: 'accessing' stamp: 'jm 4/17/1998 05:23'!
scale: newScale

	self changed.
	transform := transform withScale: newScale.
	self layoutChanged.
	self changed.
! !

!TransformMorph methodsFor: 'accessing' stamp: ''!
setOffset: newOffset angle: newAngle scale: newScale

	transform := MorphicTransform offset: newOffset angle: newAngle scale: newScale.
	self changed! !

!TransformMorph methodsFor: 'accessing' stamp: 'sps 11/29/2002 17:03'!
smoothing
	^smoothing
! !

!TransformMorph methodsFor: 'accessing' stamp: ''!
smoothingOff
	smoothing := 1.
	self changed! !

!TransformMorph methodsFor: 'accessing' stamp: ''!
smoothingOn
	smoothing := 2.
	self changed! !

!TransformMorph methodsFor: 'accessing' stamp: ''!
smoothing: cellSize
	smoothing := cellSize.
	self changed! !

!TransformMorph methodsFor: 'layout' stamp: 'nk 4/12/2002 14:03'!
submorphBounds
	"Answer, in owner coordinates, the bounds of my visible submorphs, or my bounds"
	| box |
	box := self localVisibleSubmorphBounds.
	^(box ifNotNil: [ transform localBoundsToGlobal: box ] ifNil: [ self bounds ]) truncated.
! !

!TransformMorph methodsFor: 'accessing' stamp: 'ar 5/19/1999 18:04'!
transform
	^transform! !

!TransformMorph methodsFor: 'event handling' stamp: 'dgd 2/21/2003 22:52'!
transformFrom: uberMorph 
	"Return a transform to map coorinates of uberMorph, a morph above me in my owner chain, into the coordinates of my submorphs."

	(self == uberMorph or: [owner isNil]) ifTrue: [^transform].
	^(owner transformFrom: uberMorph) composedWithLocal: transform! !

!TransformMorph methodsFor: 'accessing' stamp: 'ar 5/19/1999 18:04'!
transform: aTransform
	transform := aTransform.! !

!TransformMorph methodsFor: 'halos and balloon help' stamp: 'sw 12/29/1999 15:51'!
wantsHaloFromClick
	^ false! !

Morph subclass: #TranslucentProgessMorph
	instanceVariableNames: 'opaqueBackgroundColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!TranslucentProgessMorph methodsFor: 'drawing' stamp: 'nk 7/12/2003 08:59'!
drawOn: aCanvas

	| revealPercentage revealingStyle revealingColor revealingBounds revealToggle x baseColor revealTimes secondsRemaining stringToDraw where fontToUse innerBounds |
	
	innerBounds := bounds.
	opaqueBackgroundColor ifNotNil: [
		aCanvas 
			frameAndFillRectangle: bounds
			fillColor: opaqueBackgroundColor
			borderWidth: 8
			borderColor: Color blue.
		innerBounds := innerBounds insetBy: 8.
	].
	revealTimes := (self valueOfProperty: #revealTimes) ifNil: [^self].
	revealPercentage := (revealTimes first / revealTimes second) asFloat.
	revealingStyle := self revealingStyle.
	x := self valueOfProperty: #progressStageNumber ifAbsent: [1].
	baseColor := Color perform: (#(red blue green magenta cyan yellow) atPin: x).
	revealingColor := baseColor alpha: 0.2.
	revealingStyle = 3 ifTrue: [	"wrap and change color"
		revealPercentage > 1.0 ifTrue: [
			revealingColor := baseColor alpha: (0.2 + (revealingStyle / 10) min: 0.5).
		].
		revealPercentage := revealPercentage fractionPart.
	].
	revealingStyle = 2 ifTrue: [	"peg at 75 and blink"
		revealPercentage > 0.75 ifTrue: [
			revealToggle := self valueOfProperty: #revealToggle ifAbsent: [true].
			self setProperty: #revealToggle toValue: revealToggle not.
			revealToggle ifTrue: [revealingColor := baseColor alpha: 0.8.].
		].
		revealPercentage := revealPercentage min: 0.75.
	].
	revealingBounds := innerBounds withLeft: innerBounds left + (innerBounds width * revealPercentage) truncated.
	aCanvas 
		fillRectangle: revealingBounds
		color: revealingColor.
	secondsRemaining := (revealTimes second - revealTimes first / 1000) rounded.
	secondsRemaining > 0 ifTrue: [
		fontToUse := StrikeFont familyName: Preferences standardEToysFont familyName size: 24.
		stringToDraw := secondsRemaining printString.
		where := innerBounds corner - ((fontToUse widthOfString: stringToDraw) @ fontToUse height).
		aCanvas 
			drawString: stringToDraw 
			in: (where corner: innerBounds corner)
			font: fontToUse
			color: Color black.
		aCanvas
			drawString: stringToDraw 
			in: (where - (1@1) corner: innerBounds corner)
			font: fontToUse
			color: Color white.
	]. 


! !

!TranslucentProgessMorph methodsFor: 'WiW support' stamp: 'RAA 7/19/2000 18:52'!
morphicLayerNumber

	"helpful for insuring some morphs always appear in front of or behind others.
	smaller numbers are in front"

	^self valueOfProperty: #morphicLayerNumber ifAbsent: [12].

	"progress morphs are behind menus and balloons, but in front of most other stuff"! !

!TranslucentProgessMorph methodsFor: 'as yet unclassified' stamp: 'RAA 6/29/2000 11:35'!
opaqueBackgroundColor: aColor

	opaqueBackgroundColor := aColor! !

!TranslucentProgessMorph methodsFor: 'as yet unclassified' stamp: 'RAA 5/23/2000 11:19'!
revealingStyle

">>>>
	1 = original, no change after 100%
	2 = hold at last 25% and blink until done
	3 = wrap around from 100% back to 0 and go again. change color after first
<<<<"
	^3
! !

Object subclass: #MorphExtension
	instanceVariableNames: 'locked visible sticky balloonText balloonTextSelector externalName isPartsDonor actorState player eventHandler otherProperties'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Kernel'!

!MorphExtension commentStamp: '<historical>' prior: 0!
MorphExtension provides access to extra instance state that is not required in most simple morphs.  This allows simple morphs to remain relatively lightweight while still admitting more complex structures as necessary.  The otherProperties field takes this policy to the extreme of allowing any number of additional named attributes, albeit at a certain cost in speed and space.!

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:56'!
actorState
	"answer the redeiver's actorState"
	^ actorState ! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:53'!
actorState: anActorState 
"change the receiver's actorState"
	actorState := anActorState! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:41'!
assureOtherProperties
	"creates an otherProperties for the receiver if needed"
	otherProperties ifNil: [self initializeOtherProperties].
	^ otherProperties! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/10/1998 12:52'!
balloonText
	^ balloonText! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/10/1998 12:52'!
balloonTextSelector
	^ balloonTextSelector! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:51'!
balloonTextSelector: aSymbol 
	"change the receiver's balloonTextSelector"
	balloonTextSelector := aSymbol! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/10/1998 12:55'!
balloonText: newValue
	balloonText := newValue! !

!MorphExtension methodsFor: 'objects from disk' stamp: 'tk 4/8/1999 12:45'!
comeFullyUpOnReload: smartRefStream
	"inst vars have default booplean values."

	locked ifNil: [locked := false].
	visible ifNil: [visible := true].
	sticky ifNil: [sticky := false].
	isPartsDonor ifNil: [isPartsDonor := false].
	^ self! !

!MorphExtension methodsFor: 'connectors-copying' stamp: 'nk 5/1/2004 17:20'!
copyWeakly
	"list of names of properties whose values should be weak-copied when veryDeepCopying a morph.  See DeepCopier."

	^ #(formerOwner newPermanentPlayer logger graphModel gestureDictionaryOrName)
	"add yours to this list" 

	"formerOwner should really be nil at the time of the copy, but this will work just fine."! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:51'!
eventHandler
	"answer the receiver's eventHandler"
	^ eventHandler ! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/10/1998 12:56'!
eventHandler: newValue
	eventHandler := newValue! !

!MorphExtension methodsFor: 'viewer' stamp: 'di 8/10/1998 14:47'!
externalName
	^ externalName! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:57'!
externalName: aString 
	"change the receiver's externalName"
	externalName := aString! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:36'!
hasOtherProperties
	"answer whether the receiver has otherProperties"
	^ otherProperties notNil! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:42'!
hasProperty: aSymbol 
	"Answer whether the receiver has the property named aSymbol"
	| property |
	otherProperties ifNil: [^ false].
	property := otherProperties at: aSymbol ifAbsent: [].
	property isNil ifTrue: [^ false].
	property == false ifTrue: [^ false].
	^ true! !

!MorphExtension methodsFor: 'initialization' stamp: 'di 8/16/1998 12:02'!
initialize
	"Init all booleans to default values"
	locked := false.
	visible := true.
	sticky := false.
	isPartsDonor := false.
! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:36'!
initializeOtherProperties
	"private - initializes the receiver's otherProperties"
	otherProperties :=  IdentityDictionary new! !

!MorphExtension methodsFor: 'other' stamp: 'ar 8/6/2009 18:56'!
inspectElement
	"Create and schedule an Inspector on the otherProperties and the 
	named properties."
	| key obj |
	key := UIManager default chooseFrom: self sortedPropertyNames values: self sortedPropertyNames  title: 'Inspect which property?'.
	key
		ifNil: [^ self].
	obj := otherProperties
				at: key
				ifAbsent: ['nOT a vALuE'].
	obj = 'nOT a vALuE'
		ifTrue: [(self perform: key) inspect
			"named properties"]
		ifFalse: [obj inspect]! !

!MorphExtension methodsFor: 'other' stamp: 'md 2/27/2006 08:42'!
isDefault
	"Return true if the receiver is a default and can be omitted"
	locked == true
		ifTrue: [^ false].
	visible == false
		ifTrue: [^ false].
	sticky == true
		ifTrue: [^ false].
	balloonText isNil
		ifFalse: [^ false].
	balloonTextSelector isNil
		ifFalse: [^ false].
	externalName isNil
		ifFalse: [^ false].
	isPartsDonor == true
		ifTrue: [^ false].
	actorState isNil
		ifFalse: [^ false].
	player isNil
		ifFalse: [^ false].
	eventHandler isNil
		ifFalse: [^ false].
	otherProperties ifNotNil: [otherProperties isEmpty ifFalse: [^ false]].
	^ true! !

!MorphExtension methodsFor: 'parts bin' stamp: 'dgd 2/16/2003 21:37'!
isPartsDonor
	"answer whether the receiver is PartsDonor"
	^ isPartsDonor! !

!MorphExtension methodsFor: 'parts bin' stamp: 'dgd 2/16/2003 21:40'!
isPartsDonor: aBoolean 
	"change the receiver's isPartDonor property"
	isPartsDonor := aBoolean! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'ul 9/20/2014 20:37'!
layoutFrame
	^self valueOfProperty: #layoutFrame! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'dgd 2/22/2003 13:32'!
layoutFrame: aLayoutFrame 
	aLayoutFrame isNil
		ifTrue: [self removeProperty: #layoutFrame]
		ifFalse: [self setProperty: #layoutFrame toValue: aLayoutFrame]! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'ul 9/20/2014 20:36'!
layoutPolicy
	^self valueOfProperty: #layoutPolicy! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'dgd 2/22/2003 13:32'!
layoutPolicy: aLayoutPolicy 
	aLayoutPolicy isNil
		ifTrue: [self removeProperty: #layoutPolicy]
		ifFalse: [self setProperty: #layoutPolicy toValue: aLayoutPolicy]! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'ul 9/20/2014 20:37'!
layoutProperties
	^self valueOfProperty: #layoutProperties! !

!MorphExtension methodsFor: 'accessing - layout properties' stamp: 'dgd 2/22/2003 13:32'!
layoutProperties: newProperties 
	"Return the current layout properties associated with the receiver"

	newProperties isNil
		ifTrue: [self removeProperty: #layoutProperties]
		ifFalse: [self setProperty: #layoutProperties toValue: newProperties]! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:38'!
locked
	"answer whether the receiver is Locked"
	^ locked! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:48'!
locked: aBoolean 
	"change the receiver's locked property"
	locked := aBoolean! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'dgd 2/16/2003 21:04'!
otherProperties
	"answer the receiver's otherProperties"
	^ otherProperties! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:42'!
player
	"answer the receiver's player"
	^ player! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:53'!
player: anObject 
	"change the receiver's player"
	player := anObject ! !

!MorphExtension methodsFor: 'printing' stamp: 'md 2/27/2006 08:45'!
printOn: aStream 
	"Append to the argument, aStream, a sequence of characters that 
	identifies the receiver." 
	super printOn: aStream.
	aStream nextPutAll: ' ' , self identityHashPrintString.
	locked == true
		ifTrue: [aStream nextPutAll: ' [locked] '].
	visible == false
		ifTrue: [aStream nextPutAll: '[not visible] '].
	sticky == true
		ifTrue: [aStream nextPutAll: ' [sticky] '].
	balloonText
		ifNotNil: [aStream nextPutAll: ' [balloonText] '].
	balloonTextSelector
		ifNotNil: [aStream nextPutAll: ' [balloonTextSelector: ' , balloonTextSelector printString , '] '].
	externalName
		ifNotNil: [aStream nextPutAll: ' [externalName = ' , externalName , ' ] '].
	isPartsDonor == true
		ifTrue: [aStream nextPutAll: ' [isPartsDonor] '].
	player
		ifNotNil: [aStream nextPutAll: ' [player = ' , player printString , '] '].
	eventHandler
		ifNotNil: [aStream nextPutAll: ' [eventHandler = ' , eventHandler printString , '] '].
	(otherProperties isNil or: [otherProperties isEmpty ]) ifTrue: [^ self].
	aStream nextPutAll: ' [other: '.
	self otherProperties
		keysDo: [:aKey | aStream nextPutAll: ' (' , aKey , ' -> ' , (self otherProperties at: aKey) printString , ')'].
	aStream nextPut: $]! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'jmv 7/28/2009 18:15'!
privateOtherProperties: anIdentityDictionary 
	"private - change the receiver's otherProperties"
	otherProperties := anIdentityDictionary ! !

!MorphExtension methodsFor: 'connectors-copying' stamp: 'nk 5/1/2004 17:23'!
propertyNamesNotCopied
	"list of names of properties whose values should be deleted when veryDeepCopying a morph.
	See DeepCopier."

	^ #(connectedConstraints connectionHighlights highlightedTargets)
	"add yours to this list" 
! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:37'!
removeOtherProperties
	"Remove the 'other' properties"
	otherProperties := nil! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:43'!
removeProperty: aSymbol 
	"removes the property named aSymbol if it exists"
	otherProperties ifNil: [^ self].
	otherProperties removeKey: aSymbol ifAbsent: [].
	otherProperties isEmpty ifTrue: [self removeOtherProperties]! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'dgd 2/16/2003 21:49'!
setProperty: aSymbol toValue: abObject 
	"change the receiver's property named aSymbol to anObject"
	self assureOtherProperties at: aSymbol put: abObject! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'md 2/27/2006 08:45'!
sortedPropertyNames
	"answer the receiver's property names in a sorted way"

	| props |
	props := WriteStream on: (Array new: 10).
	locked == true ifTrue: [props nextPut: #locked].
	visible == false ifTrue: [props nextPut: #visible].
	sticky == true ifTrue: [props nextPut: #sticky].
	balloonText isNil ifFalse: [props nextPut: #balloonText].
	balloonTextSelector isNil ifFalse: [props nextPut: #balloonTextSelector].
	externalName isNil ifFalse: [props nextPut: #externalName].
	isPartsDonor == true ifTrue: [props nextPut: #isPartsDonor].
	actorState isNil ifFalse: [props nextPut: #actorState].
	player isNil ifFalse: [props nextPut: #player].
	eventHandler isNil ifFalse: [props nextPut: #eventHandler].
	 otherProperties ifNotNil: [otherProperties associationsDo: [:a | props nextPut: a key]].
	^props contents sort: [:s1 :s2 | s1 <= s2]! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/14/1998 13:07'!
sticky
	^ sticky! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:47'!
sticky: aBoolean 
	"change the receiver's sticky property"
	sticky := aBoolean! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'ul 9/20/2014 20:36'!
valueOfProperty: aSymbol 
"answer the value of the receiver's property named aSymbol"

	^otherProperties ifNotNil: [ otherProperties at: aSymbol ifAbsent: nil ]! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'dgd 2/16/2003 21:28'!
valueOfProperty: aSymbol ifAbsentPut: aBlock 
	"If the receiver possesses a property of the given name, answer  
	its value. If not, then create a property of the given name, give 
	it the value obtained by evaluating aBlock, then answer that  
	value"
	^self assureOtherProperties at: aSymbol ifAbsentPut: aBlock! !

!MorphExtension methodsFor: 'accessing - other properties' stamp: 'ul 9/20/2014 20:30'!
valueOfProperty: aSymbol ifAbsent: aBlock 
	"if the receiver possesses a property of the given name, answer  
	its value. If not then evaluate aBlock and answer the result of  
	this block evaluation"
	otherProperties ifNil: [^ aBlock value].
	^otherProperties at: aSymbol ifAbsent: aBlock! !

!MorphExtension methodsFor: 'connectors-copying' stamp: 'nk 5/1/2004 17:39'!
veryDeepFixupWith: deepCopier 
	"If target and arguments fields were weakly copied, fix them here.
	If they were in the tree being copied, fix them up, otherwise point to the originals!!!!"

	super veryDeepFixupWith: deepCopier.
	otherProperties ifNil: [ ^self ].

	"Properties whose values are only copied weakly replace those values if they were copied via another path"
	self copyWeakly do: [ :propertyName |
		otherProperties at: propertyName ifPresent: [ :property |
			otherProperties at: propertyName
				put: (deepCopier references at: property ifAbsent: [ property ])]].
! !

!MorphExtension methodsFor: 'connectors-copying' stamp: 'nk 5/1/2004 17:45'!
veryDeepInner: deepCopier 
	"Copy all of my instance variables.
	Some otherProperties need to be not copied at all, but shared. Their names are given by copyWeakly.
	Some otherProperties should not be copied or shared. Their names are given by propertyNamesNotCopied.
	This is special code for the dictionary. See DeepCopier, and veryDeepFixupWith:."

	| namesOfWeaklyCopiedProperties weaklyCopiedValues |
	super veryDeepInner: deepCopier.
	locked := locked veryDeepCopyWith: deepCopier.
	visible := visible veryDeepCopyWith: deepCopier.
	sticky := sticky veryDeepCopyWith: deepCopier.
	balloonText := balloonText veryDeepCopyWith: deepCopier.
	balloonTextSelector := balloonTextSelector veryDeepCopyWith: deepCopier.
	externalName := externalName veryDeepCopyWith: deepCopier.
	isPartsDonor := isPartsDonor veryDeepCopyWith: deepCopier.
	actorState := actorState veryDeepCopyWith: deepCopier.
	player := player veryDeepCopyWith: deepCopier.		"Do copy the player of this morph"
	eventHandler := eventHandler veryDeepCopyWith: deepCopier. 	"has its own restrictions"

	otherProperties ifNil: [ ^self ].

	otherProperties := otherProperties copy.
	self propertyNamesNotCopied do: [ :propName | otherProperties removeKey: propName ifAbsent: [] ].

	namesOfWeaklyCopiedProperties := self copyWeakly.
	weaklyCopiedValues := namesOfWeaklyCopiedProperties collect: [  :propName | otherProperties removeKey: propName ifAbsent: [] ].

	"Now copy all the others."
	otherProperties := otherProperties veryDeepCopyWith: deepCopier.

	"And replace the weak ones."
	namesOfWeaklyCopiedProperties with: weaklyCopiedValues do: [ :name :value | value ifNotNil: [ otherProperties at: name put: value ]].
! !

!MorphExtension methodsFor: 'accessing' stamp: 'dgd 2/16/2003 21:41'!
visible
	"answer whether the receiver is visible"
	^ visible! !

!MorphExtension methodsFor: 'accessing' stamp: 'di 8/10/1998 12:55'!
visible: newValue
	visible := newValue! !

Object subclass: #MorphHierarchy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Widgets'!

!MorphHierarchy class methodsFor: 'opening' stamp: 'dgd 9/25/2004 21:50'!
openOrDelete
	| oldMorph |
	oldMorph := World submorphs
				detect: [:each | each hasProperty: #morphHierarchy]
				ifNone: [| newMorph | 
					newMorph := self new asMorph.
					newMorph bottomLeft: ActiveHand position.
					newMorph openInWorld.
					newMorph isFullOnScreen
						ifFalse: [newMorph goHome].
					^ self].
	""
	oldMorph delete! !

!MorphHierarchy methodsFor: 'private' stamp: 'dgd 9/26/2004 18:28'!
asMorph
	"Answer the morph version of the receiver"
	| morph |
	morph := MorphHierarchyListMorph
				on: self
				list: #roots
				selected: nil
				changeSelected: #selected:.
	""
	^ morph inAContainer! !

!MorphHierarchy methodsFor: 'accessing' stamp: 'dgd 9/26/2004 18:29'!
roots
	"Answer the roots for the Object Hierarchy, that means answer the World"
	^ {MorphListItemWrapper with: World}! !

!MorphHierarchy methodsFor: 'accessing' stamp: 'dgd 9/26/2004 18:30'!
selected: aMorphListItemWrapper 
	"Change the selected object"
	| newSelection |
	aMorphListItemWrapper isNil
		ifTrue: [^ self].
	newSelection := aMorphListItemWrapper withoutListWrapper.
	newSelection == World selectedObject
		ifTrue: [newSelection removeHalo]
		ifFalse: [newSelection addHalo].
	self changed: #selected! !

Object subclass: #MorphicAlarmQueue
	instanceVariableNames: 'mutex sequenceNumber heap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!MorphicAlarmQueue commentStamp: 'jcg 1/9/2010 13:34' prior: 0!
MorphicAlarmQueue is a specialized Heap.  The main change is to stamp each added MorphicAlarm with a sequence number to ensure that alarms scheduled for the same time are executed in the order that they were added.!

!MorphicAlarmQueue class methodsFor: 'class initialization' stamp: 'jcg 1/9/2010 12:42'!
convertAllAlarms
	"Alarms should be kept in a MorphicAlarmQueue, not a Heap."
	WorldState allSubInstancesDo: [:ws | ws convertAlarms]! !

!MorphicAlarmQueue class methodsFor: 'class initialization' stamp: 'jcg 1/9/2010 12:37'!
initialize
	self convertAllAlarms.! !

!MorphicAlarmQueue methodsFor: 'adding' stamp: 'ul 5/31/2016 01:06'!
add: aMorphicAlarm

	aMorphicAlarm sequenceNumber: self nextSequenceNumber.
	^heap add: aMorphicAlarm! !

!MorphicAlarmQueue methodsFor: 'enumeration' stamp: 'ul 5/31/2016 01:15'!
alarmsDoSafely: aBlock

	heap asArray do: aBlock! !

!MorphicAlarmQueue methodsFor: 'accessing' stamp: 'ul 5/31/2016 01:18'!
incrementScheduledTimesBy: anInteger

	heap do: [ :each | each scheduledTime: each scheduledTime + anInteger ]! !

!MorphicAlarmQueue methodsFor: 'initialize' stamp: 'ul 5/31/2016 01:06'!
initialize

	super initialize.
	mutex := Mutex new.
	sequenceNumber := 0.
	heap := Heap sortBlock: [ :alarmA :alarmB |
		alarmA scheduledTime = alarmB scheduledTime 
			ifFalse: [ alarmA scheduledTime < alarmB scheduledTime ]
			ifTrue: [
				alarmA sequenceNumber = alarmB sequenceNumber
					ifFalse: [ alarmA sequenceNumber < alarmB sequenceNumber ]
					ifTrue: [ self error: 'These alarms run at the same time' ] ] ]! !

!MorphicAlarmQueue methodsFor: 'accessing' stamp: 'ul 5/31/2016 01:07'!
mutex
	
	^mutex! !

!MorphicAlarmQueue methodsFor: 'private' stamp: 'ul 6/2/2016 20:24'!
nextSequenceNumber

	(sequenceNumber := sequenceNumber + 1) = 16r3FFFFFFF ifTrue: [
		"Sequence number overflow... reassign sequence numbers starting at 1."
		heap sort.
		1 to: heap size do: [ :index |
			(heap at: index) sequenceNumber: index ].
		sequenceNumber := heap size + 1 ].
	^sequenceNumber! !

!MorphicAlarmQueue methodsFor: 'copying' stamp: 'ul 5/31/2016 01:33'!
postCopy

	super postCopy.
	mutex := Mutex new.
	heap := heap copy! !

!MorphicAlarmQueue methodsFor: 'removing' stamp: 'ul 5/31/2016 01:33'!
removeAlarmWithReceiver: receiver selector: selector

	^heap
		detect: [ :each | each receiver == receiver and: [ each selector == selector ] ]
		ifFound: [ :found | heap remove: found ]
		ifNone: [ ]! !

!MorphicAlarmQueue methodsFor: 'removing' stamp: 'ul 5/31/2016 01:11'!
removeAllAlarmsBefore: nowTime

	^Array new: 10 streamContents: [ :stream |
		[ heap notEmpty and: [ heap first scheduledTime < nowTime ] ]
			whileTrue: [ stream nextPut: heap removeFirst ] ]! !

Object subclass: #MorphicEvent
	instanceVariableNames: 'timeStamp source'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!MorphicEvent commentStamp: '<historical>' prior: 0!
This class represents the base for all events.

Instance variables:
	stamp	<Integer>	The millisecond clock time stamp (based on Time millisecondClock)
	source	<Hand | nil>	If non-nil the hand that generated the event.!

MorphicEvent subclass: #DropEvent
	instanceVariableNames: 'position contents wasHandled'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 18:33'!
contents
	^contents! !

!DropEvent methodsFor: 'initialize' stamp: 'ar 10/10/2000 01:19'!
copyHandlerState: anEvent
	"Copy the handler state from anEvent. Used for quickly transferring handler information between transformed events."
	wasHandled := anEvent wasHandled.! !

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 19:21'!
cursorPoint
	"For compatibility with mouse events"
	^position! !

!DropEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 18:33'!
isDropEvent
	^true! !

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 18:33'!
position
	^position! !

!DropEvent methodsFor: 'printing' stamp: 'ar 9/14/2000 18:15'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self position printString; space.
	aStream nextPutAll: self type.
	aStream nextPut: $].! !

!DropEvent methodsFor: 'initialize' stamp: 'ar 10/10/2000 01:18'!
resetHandlerFields
	"Reset anything that is used to cross-communicate between two eventual handlers during event dispatch"
	wasHandled := false.! !

!DropEvent methodsFor: 'dispatching' stamp: 'ar 1/10/2001 21:24'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	self type == #dropEvent ifTrue:[^anObject handleDropMorph: self].! !

!DropEvent methodsFor: 'private' stamp: 'ar 9/13/2000 19:23'!
setPosition: pos contents: aMorph hand: aHand
	position := pos.
	contents := aMorph.
	source := aHand.
	wasHandled := false.! !

!DropEvent methodsFor: 'transforming' stamp: 'ar 10/7/2000 18:28'!
transformBy: aMorphicTransform
	"Transform the receiver into a local coordinate system."
	position :=  aMorphicTransform globalPointToLocal: position.! !

!DropEvent methodsFor: 'transforming' stamp: 'ar 10/7/2000 18:28'!
transformedBy: aMorphicTransform
	"Return the receiver transformed by the given transform into a local coordinate system."
	^self shallowCopy transformBy: aMorphicTransform! !

!DropEvent methodsFor: 'transforming' stamp: 'mt 5/20/2016 10:12'!
translateTo: point

	position := point.! !

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 18:33'!
type
	^#dropEvent! !

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 18:44'!
wasHandled
	^wasHandled! !

!DropEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 18:44'!
wasHandled: aBool
	wasHandled := aBool.! !

DropEvent subclass: #DropFilesEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!DropFilesEvent methodsFor: 'dispatching' stamp: 'ar 1/10/2001 21:35'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	self type == #dropFilesEvent ifTrue:[^anObject handleDropFiles: self].! !

!DropFilesEvent methodsFor: 'accessing' stamp: 'ar 1/10/2001 21:35'!
type
	^#dropFilesEvent! !

!MorphicEvent class methodsFor: 'instance creation' stamp: 'ar 10/26/2000 00:44'!
convertObsolete: anEvent
	"ar 10/25/2000: This method is used to convert OLD MorphicEvents into new ones."
	| type cursorPoint buttons keyValue sourceHand |
	type := anEvent type.
	cursorPoint := anEvent cursorPoint.
	buttons := anEvent buttons.
	keyValue := anEvent keyValue.
	sourceHand := anEvent hand.
	type == #mouseMove ifTrue:[
		^MouseMoveEvent new
			setType: #mouseMove 
			startPoint: cursorPoint
			endPoint: cursorPoint
			trail: #() 
			buttons: buttons 
			hand: sourceHand 
			stamp: nil].
	(type == #mouseDown) | (type == #mouseUp) ifTrue:[
			^MouseButtonEvent new
				setType: type
				position: cursorPoint
				which: 0
				buttons: buttons
				hand: sourceHand
				stamp: nil].
	(type == #keystroke) | (type == #keyDown) | (type == #keyUp) ifTrue:[
		^KeyboardEvent new
			setType: type
			buttons: buttons
			position: cursorPoint
			keyValue: keyValue
			hand: sourceHand
			stamp: nil].
	^nil! !

!MorphicEvent class methodsFor: 'instance creation' stamp: 'nice 12/27/2009 03:11'!
readFromObsolete: aStream
	"Read one of those old and now obsolete events from the stream"
	| type x y buttons keyValue typeString |
	typeString := String streamContents:
		[:s | | c |   [(c := aStream next) isLetter] whileTrue: [s nextPut: c]].
	typeString = 'mouseMove'
		ifTrue: [type := #mouseMove  "fast treatment of common case"]
		ifFalse: [type := typeString asSymbol].

	x := Integer readFrom: aStream.
	aStream skip: 1.
	y := Integer readFrom: aStream.
	aStream skip: 1.

	buttons := Integer readFrom: aStream.
	aStream skip: 1.

	keyValue := Integer readFrom: aStream.

	typeString = 'mouseMove' ifTrue:[
		^MouseMoveEvent new
			setType: #mouseMove 
			startPoint: x@y 
			endPoint: x@y 
			trail: #() 
			buttons: buttons 
			hand: nil 
			stamp: nil].
	(typeString = 'mouseDown') | (typeString = 'mouseUp') ifTrue:[
			^MouseButtonEvent new
				setType: type
				position: x@y
				which: 0
				buttons: buttons
				hand: nil
				stamp: nil].
	(typeString = 'keystroke') | (typeString = 'keyDown') | (typeString = 'keyUp') ifTrue:[
		^KeyboardEvent new
			setType: type
			buttons: buttons
			position: x@y
			keyValue: keyValue
			hand: nil
			stamp: nil].

	^nil! !

!MorphicEvent class methodsFor: 'instance creation' stamp: 'mt 6/6/2016 09:48'!
readFrom: aStream
	"Read a MorphicEvent from the given stream."
	| typeString |
	typeString := String streamContents:
		[:s | | c |   [(c := aStream next) isLetter] whileTrue: [s nextPut: c]].
	typeString = 'mouseMove' ifTrue:[^MouseMoveEvent type: #mouseMove readFrom: aStream].
	typeString = 'mouseDown' ifTrue:[^MouseButtonEvent type: #mouseDown readFrom: aStream].
	typeString = 'mouseUp' ifTrue:[^MouseButtonEvent type: #mouseUp readFrom: aStream].
	typeString = 'mouseWheel' ifTrue:[^MouseWheelEvent type: #mouseWheel readFrom: aStream].

	typeString = 'keystroke' ifTrue:[^KeyboardEvent type: #keystroke readFrom: aStream].
	typeString = 'keyDown' ifTrue:[^KeyboardEvent type: #keyDown readFrom: aStream].
	typeString = 'keyUp' ifTrue:[^KeyboardEvent type: #keyUp readFrom: aStream].

	typeString = 'mouseOver' ifTrue:[^MouseEvent type: #mouseOver readFrom: aStream].
	typeString = 'mouseEnter' ifTrue:[^MouseEvent type: #mouseEnter readFrom: aStream].
	typeString = 'mouseLeave' ifTrue:[^MouseEvent type: #mouseLeave readFrom: aStream].

	typeString = 'unknown' ifTrue:[^MorphicUnknownEvent type: #unknown readFrom: aStream].

	^nil
! !

!MorphicEvent class methodsFor: 'instance creation' stamp: 'ar 10/24/2000 16:32'!
type: eventType readFrom: aStream
	^self new type: eventType readFrom: aStream! !

!MorphicEvent methodsFor: 'object fileIn' stamp: 'RAA 12/20/2000 16:05'!
convertOctober2000: varDict using: smartRefStrm
	"ar 10/25/2000: This method is used to convert OLD MorphicEvents into new ones."
	"These are going away #('type' 'cursorPoint' 'buttons' 'keyValue' 'sourceHand').  Possibly store their info in another variable?"
	| type cursorPoint buttons keyValue sourceHand |
	type := varDict at: 'type'.
	cursorPoint := varDict at: 'cursorPoint'.
	buttons := varDict at: 'buttons'.
	keyValue := varDict at: 'keyValue'.
	sourceHand := varDict at: 'sourceHand'.
	type == #mouseMove ifTrue:[
		^MouseMoveEvent new
			setType: #mouseMove 
			startPoint: cursorPoint
			endPoint: cursorPoint
			trail: #() 
			buttons: buttons 
			hand: sourceHand 
			stamp: nil].
	(type == #mouseDown) | (type == #mouseUp) ifTrue:[
			^MouseButtonEvent new
				setType: type
				position: cursorPoint
				which: 0
				buttons: buttons
				hand: sourceHand
				stamp: nil].
	(type == #keystroke) | (type == #keyDown) | (type == #keyUp) ifTrue:[
		^KeyboardEvent new
			setType: type
			buttons: buttons
			position: cursorPoint
			keyValue: keyValue
			hand: sourceHand
			stamp: nil].
	"All others will be handled there"
	^MorphicUnknownEvent new! !

!MorphicEvent methodsFor: 'objects from disk' stamp: 'nice 12/27/2009 03:11'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	

	"ar 10/25/2000: This method is used to convert OLD MorphicEvents into new ones."
	varDict at: 'cursorPoint' ifPresent: [ :x | | answer | 
		answer := self convertOctober2000: varDict using: smartRefStrm.
		varDict removeKey: 'cursorPoint'.	"avoid doing this again"
		^answer
	].
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.


! !

!MorphicEvent methodsFor: 'initialize' stamp: 'ar 10/10/2000 01:18'!
copyHandlerState: anEvent
	"Copy the handler state from anEvent. Used for quickly transferring handler information between transformed events."
! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 10/10/2000 21:28'!
cursorPoint
	"Backward compatibility. Use #position instead"
	^ self position! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 16:48'!
hand
	"Return the source that generated the event"
	^source! !

!MorphicEvent methodsFor: 'comparing' stamp: 'ar 9/13/2000 15:36'!
hash
	^self type hash! !

!MorphicEvent methodsFor: 'dispatching' stamp: 'mt 5/21/2016 10:06'!
ignore

	self wasIgnored: true.! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/22/2000 10:36'!
isDraggingEvent
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 19:17'!
isDropEvent
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 19:19'!
isKeyboard
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 10/10/2000 21:27'!
isKeystroke
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:37'!
isMorphicEvent
	^true! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 19:19'!
isMouse
	^false! !

!MorphicEvent methodsFor: 'testing' stamp: 'ar 9/14/2000 18:21'!
isMouseOver
	^self type == #mouseOver! !

!MorphicEvent methodsFor: 'testing' stamp: 'JMM 10/6/2004 21:35'!
isWindowEvent
	^false! !

!MorphicEvent methodsFor: 'accessing' stamp: 'wiz 12/8/2004 23:13'!
position
	"Since cursorPoint is defined and refers to position it should be defined
	here as well"
	^ self subclassResponsibility! !

!MorphicEvent methodsFor: 'initialize' stamp: 'ar 10/10/2000 01:18'!
resetHandlerFields
	"Reset anything that is used to cross-communicate between two eventual handlers during event dispatch"! !

!MorphicEvent methodsFor: 'dispatching' stamp: 'ar 9/15/2000 21:12'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	^anObject handleUnknownEvent: self! !

!MorphicEvent methodsFor: 'private' stamp: 'ar 10/25/2000 21:26'!
setHand: aHand
	source := aHand! !

!MorphicEvent methodsFor: 'private' stamp: 'ar 10/25/2000 20:53'!
setTimeStamp: stamp
	timeStamp := stamp.! !

!MorphicEvent methodsFor: 'accessing' stamp: 'tpr 1/11/2016 15:30'!
timeStamp
	"Return the millisecond clock value at which the event was generated"
	^timeStamp ifNil:[timeStamp := Time eventMillisecondClock]! !

!MorphicEvent methodsFor: 'transforming' stamp: 'ar 9/13/2000 15:47'!
transformedBy: aMorphicTransform
	"Return the receiver transformed by the given transform into a local coordinate system."
! !

!MorphicEvent methodsFor: 'transforming' stamp: 'mt 5/20/2016 10:12'!
translateTo: point
! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:34'!
type
	"Return a symbol indicating the type this event."
	^self subclassResponsibility! !

!MorphicEvent methodsFor: 'initialize' stamp: 'ar 10/24/2000 16:21'!
type: eventType readFrom: aStream
	"Read a MorphicEvent from the given stream."
! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 10/10/2000 01:19'!
wasHandled
	"Return true if this event was handled. May be ignored for some types of events."
	^false! !

!MorphicEvent methodsFor: 'accessing' stamp: 'ar 10/10/2000 01:20'!
wasHandled: aBool
	"Determine if this event was handled. May be ignored for some types of events."! !

!MorphicEvent methodsFor: 'accessing' stamp: 'mt 5/19/2016 13:12'!
wasIgnored
	^ false! !

!MorphicEvent methodsFor: 'accessing' stamp: 'mt 5/19/2016 13:12'!
wasIgnored: boolean
! !

!MorphicEvent methodsFor: 'comparing' stamp: 'ar 9/13/2000 15:36'!
= anEvent
	anEvent isMorphicEvent ifFalse:[^false].
	^self type = anEvent type! !

MorphicEvent subclass: #MorphicUnknownEvent
	instanceVariableNames: 'type argument'
	classVariableNames: ''
	poolDictionaries: 'EventSensorConstants'
	category: 'Morphic-Events'!

!MorphicUnknownEvent methodsFor: 'accessing' stamp: 'ar 10/25/2000 20:04'!
argument
	^argument! !

!MorphicUnknownEvent methodsFor: 'accessing' stamp: 'ar 10/25/2000 20:04'!
argument: arg
	argument := arg! !

!MorphicUnknownEvent methodsFor: 'objects from disk' stamp: 'RAA 12/20/2000 17:48'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	type ifNil: [type := #startSound].
	source ifNil: [source := varDict at: 'sourceHand'].
	argument ifNil: [argument := varDict at: 'sound' ifAbsent: [nil]].	"???"
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!MorphicUnknownEvent methodsFor: 'accessing' stamp: 'ar 10/25/2000 19:55'!
position
	^0@0! !

!MorphicUnknownEvent methodsFor: 'private' stamp: 'ar 10/25/2000 19:59'!
setType: evtType argument: arg
	type := evtType.
	argument := arg.! !

!MorphicUnknownEvent methodsFor: 'private' stamp: 'ar 10/25/2000 19:58'!
setType: evtType argument: arg hand: evtHand stamp: stamp
	type := evtType.
	argument := arg.
	source := evtHand.
	timeStamp := stamp.! !

!MorphicUnknownEvent methodsFor: 'printing' stamp: 'ar 10/26/2000 01:19'!
storeOn: aStream
	aStream nextPutAll: 'unknown'.
	aStream space.
	self timeStamp storeOn: aStream.
	aStream space.
	{type. argument} storeOn: aStream.! !

!MorphicUnknownEvent methodsFor: 'accessing' stamp: 'ar 10/25/2000 19:55'!
type
	^type! !

!MorphicUnknownEvent methodsFor: 'initialize' stamp: 'ar 10/26/2000 01:20'!
type: eventType readFrom: aStream
	| typeAndArg |
	timeStamp := Integer readFrom: aStream.
	aStream skip: 1.
	typeAndArg := Object readFrom: aStream.
	type := typeAndArg first.
	argument := typeAndArg last.! !

MorphicEvent subclass: #UserInputEvent
	instanceVariableNames: 'type buttons position handler wasHandled wasIgnored'
	classVariableNames: ''
	poolDictionaries: 'EventSensorConstants'
	category: 'Morphic-Events'!

UserInputEvent subclass: #KeyboardEvent
	instanceVariableNames: 'keyValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!KeyboardEvent methodsFor: 'comparing' stamp: 'ar 9/13/2000 15:50'!
hash
	^buttons hash + keyValue hash
! !

!KeyboardEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:49'!
isKeyboard
	^true! !

!KeyboardEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:49'!
isKeyDown
	^self type == #keyDown! !

!KeyboardEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:49'!
isKeystroke
	^self type == #keystroke! !

!KeyboardEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:49'!
isKeyUp
	^self type == #keyUp! !

!KeyboardEvent methodsFor: 'testing' stamp: 'ar 10/9/2000 00:43'!
isMouseMove
	^false! !

!KeyboardEvent methodsFor: 'keyboard' stamp: 'ar 9/13/2000 15:51'!
keyCharacter
	"Answer the character corresponding this keystroke. This is defined only for keystroke events."

	^ keyValue asCharacter! !

!KeyboardEvent methodsFor: 'keyboard' stamp: 'nk 10/13/2004 10:43'!
keyString
	"Answer the string value for this keystroke. This is defined only for keystroke events."

	^ String streamContents: [ :s | self printKeyStringOn: s ]! !

!KeyboardEvent methodsFor: 'keyboard' stamp: 'ar 9/13/2000 15:51'!
keyValue
	"Answer the ascii value for this keystroke. This is defined only for keystroke events."

	^ keyValue! !

!KeyboardEvent methodsFor: 'printing' stamp: 'tk 10/13/2004 15:19'!
printKeyStringOn: aStream
	"Print a readable string representing the receiver on a given stream"

	| kc inBrackets firstBracket keyString |
	kc := self keyCharacter.
	inBrackets := false.
	firstBracket := [ inBrackets ifFalse: [ aStream nextPut: $<. inBrackets := true ]].
	self controlKeyPressed ifTrue: [ 	firstBracket value. aStream nextPutAll: 'Ctrl-' ].
	self commandKeyPressed ifTrue: [ firstBracket value. aStream nextPutAll: 'Cmd-' ].
	(buttons anyMask: 32) ifTrue: [ firstBracket value. aStream nextPutAll: 'Opt-' ].
	(self shiftPressed and: [ keyValue between: 1 and: 31 ])
		ifTrue: [ firstBracket value. aStream nextPutAll: 'Shift-' ].

	(self controlKeyPressed and: [ keyValue <= 26 ])
			ifTrue:
				[aStream nextPut: (keyValue + $a asciiValue - 1) asCharacter]
			ifFalse: 
				[keyString := (kc caseOf: {
					[ Character space ] -> [ ' ' ].
					[ Character tab ] -> [ 'tab' ].
					[ Character cr ] -> [ 'cr' ].
					[ Character lf ] -> [ 'lf' ].
					[ Character enter ] -> [ 'enter' ].

					[ Character backspace ] -> [ 'backspace' ].
					[ Character delete ] -> [ 'delete' ].

					[ Character escape ] -> [ 'escape' ].

					[ Character arrowDown ] -> [ 'down' ].
					[ Character arrowUp ] -> [ 'up' ].
					[ Character arrowLeft ] -> [ 'left' ].
					[ Character arrowRight ] -> [ 'right' ].

					[ Character end ] -> [ 'end' ].
					[ Character home ] -> [ 'home' ].
					[ Character pageDown ] -> [ 'pageDown' ].
					[ Character pageUp ] -> [ 'pageUp' ].

					[ Character euro ] -> [ 'euro' ].
					[ Character insert ] -> [ 'insert' ].

				} otherwise: [ String with: kc ]).
				keyString size > 1 ifTrue: [ firstBracket value ].
				aStream nextPutAll: keyString].

	inBrackets ifTrue: [aStream nextPut: $> ]! !

!KeyboardEvent methodsFor: 'printing' stamp: 'mt 6/6/2016 17:17'!
printOn: aStream
	"Print the receiver on a stream"

	aStream nextPut: $[.
	aStream nextPutAll: self cursorPoint printString; space.
	aStream nextPutAll: type; nextPutAll: ' '''.
	self printKeyStringOn: aStream.
	aStream nextPut: $'; space.
	aStream nextPutAll: timeStamp printString.
	aStream nextPut: $]! !

!KeyboardEvent methodsFor: 'private' stamp: 'djr 3/30/2010 12:38'!
scanCode: ignore
	" OB-Tests expects this "! !

!KeyboardEvent methodsFor: 'dispatching' stamp: 'ar 9/15/2000 21:13'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #keystroke ifTrue:[^anObject handleKeystroke: self].
	type == #keyDown ifTrue:[^anObject handleKeyDown: self].
	type == #keyUp ifTrue:[^anObject handleKeyUp: self].
	^super sentTo: anObject.! !

!KeyboardEvent methodsFor: 'private' stamp: 'ar 10/5/2000 23:54'!
setType: aSymbol buttons: anInteger position: pos keyValue: aValue hand: aHand stamp: stamp
	type := aSymbol.
	buttons := anInteger.
	position := pos.
	keyValue := aValue.
	source := aHand.
	wasHandled := false.
	timeStamp := stamp.! !

!KeyboardEvent methodsFor: 'printing' stamp: 'ar 10/25/2000 22:07'!
storeOn: aStream

	aStream nextPutAll: type.
	aStream space.
	self timeStamp storeOn: aStream.
	aStream space.
	buttons storeOn: aStream.
	aStream space.
	keyValue storeOn: aStream.
! !

!KeyboardEvent methodsFor: 'initialize' stamp: 'ar 10/25/2000 22:08'!
type: eventType readFrom: aStream
	type := eventType.
	timeStamp := Integer readFrom: aStream.
	aStream skip: 1.
	buttons := Integer readFrom: aStream.
	aStream skip: 1.
	keyValue := Integer readFrom: aStream.! !

!KeyboardEvent methodsFor: 'comparing' stamp: 'ar 10/24/2000 17:44'!
= aMorphicEvent
	super = aMorphicEvent ifFalse:[^false].
	buttons = aMorphicEvent buttons ifFalse: [^ false].
	keyValue = aMorphicEvent keyValue ifFalse: [^ false].
	^ true
! !

UserInputEvent subclass: #MouseEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

MouseEvent subclass: #MouseButtonEvent
	instanceVariableNames: 'whichButton'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!MouseButtonEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 19:58'!
blueButtonChanged
	"Answer true if the blue mouse button has changed. This is the third mouse button or cmd+click on the Mac."

	^ whichButton anyMask: 1! !

!MouseButtonEvent methodsFor: 'printing' stamp: 'mt 6/7/2016 10:36'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self cursorPoint printString; space.
	aStream nextPutAll: type; space.
	aStream nextPutAll: self modifierString.
	aStream nextPutAll: self buttonString.
	aStream nextPutAll: self whichButtonString; space.
	aStream nextPutAll: timeStamp printString.
	aStream nextPut: $].! !

!MouseButtonEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 19:58'!
redButtonChanged
	"Answer true if the red mouse button has changed. This is the first mouse button."

	^ whichButton anyMask: 4! !

!MouseButtonEvent methodsFor: 'dispatching' stamp: 'ar 9/16/2000 13:05'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseDown ifTrue:[^anObject handleMouseDown: self].
	type == #mouseUp ifTrue:[^anObject handleMouseUp: self].
	^super sentTo: anObject! !

!MouseButtonEvent methodsFor: 'private' stamp: 'ar 10/5/2000 23:55'!
setType: evtType position: evtPos which: button buttons: evtButtons hand: evtHand stamp: stamp
	type := evtType.
	position := evtPos.
	buttons := evtButtons.
	source := evtHand.
	wasHandled := false.
	whichButton := button.
	timeStamp := stamp.! !

!MouseButtonEvent methodsFor: 'printing' stamp: 'ar 10/24/2000 16:29'!
storeOn: aStream
	super storeOn: aStream.
	aStream space.
	whichButton storeOn: aStream.! !

!MouseButtonEvent methodsFor: 'initialize' stamp: 'ar 10/24/2000 16:29'!
type: eventType readFrom: aStream
	super type: eventType readFrom: aStream.
	aStream skip: 1.
	whichButton := Integer readFrom: aStream.! !

!MouseButtonEvent methodsFor: 'accessing' stamp: 'nk 3/11/2004 17:44'!
whichButton
	^whichButton! !

!MouseButtonEvent methodsFor: 'printing' stamp: 'mt 6/7/2016 10:36'!
whichButtonString
	"Return a string identifying the changed buttons"
	
	^ String streamContents: [:stream |
		stream nextPutAll: '( '.
		self redButtonChanged ifTrue: [stream nextPutAll: 'red '].
		self yellowButtonChanged ifTrue: [stream nextPutAll: 'yellow '].
		self blueButtonChanged ifTrue: [stream nextPutAll: 'blue '].
		stream nextPutAll: ')']! !

!MouseButtonEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 19:59'!
yellowButtonChanged
	"Answer true if the yellow mouse button has changed. This is the second mouse button or option+click on the Mac."

	^ whichButton anyMask: 2! !

!MouseEvent class methodsFor: 'constants' stamp: 'NS 5/19/2003 15:16'!
anyButton
	^ 7! !

!MouseEvent class methodsFor: 'constants' stamp: 'NS 5/19/2003 15:16'!
blueButton
	^ 1! !

!MouseEvent class methodsFor: 'constants' stamp: 'NS 5/19/2003 15:16'!
redButton
	^ 4! !

!MouseEvent class methodsFor: 'constants' stamp: 'NS 5/19/2003 15:16'!
yellowButton
	^ 2! !

!MouseEvent methodsFor: 'button state' stamp: 'NS 5/19/2003 15:17'!
anyButtonPressed
	"Answer true if any mouse button is being pressed."

	^ buttons anyMask: self class anyButton! !

!MouseEvent methodsFor: 'converting' stamp: 'ar 10/10/2000 21:17'!
asMouseEnter
	^self clone setType: #mouseEnter! !

!MouseEvent methodsFor: 'converting' stamp: 'ar 10/10/2000 21:17'!
asMouseLeave
	^self clone setType: #mouseLeave! !

!MouseEvent methodsFor: 'converting' stamp: 'tpr 1/11/2016 14:22'!
asMouseMove
	"Convert the receiver into a mouse move"
	^MouseMoveEvent new setType: #mouseMove startPoint: position endPoint: position trail: {position. position} buttons: buttons hand: source stamp: Time eventMillisecondClock! !

!MouseEvent methodsFor: 'converting' stamp: 'ar 9/25/2000 14:29'!
asMouseOver
	"Convert the receiver into a mouse over event"
	^MouseEvent new setType: #mouseOver position: position buttons: buttons hand: source! !

!MouseEvent methodsFor: 'button state' stamp: 'NS 5/19/2003 15:17'!
blueButtonPressed
	"Answer true if the blue mouse button is being pressed. This is the third mouse button or cmd+click on the Mac."

	^ buttons anyMask: self class blueButton! !

!MouseEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 22:51'!
cursorPoint
	"Answer the location of the cursor's hotspot when this event occured."

	^ position! !

!MouseEvent methodsFor: 'comparing' stamp: 'ar 9/15/2000 22:47'!
hash
	^ position hash + buttons hash! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 10/5/2000 19:43'!
isDraggingEvent
	source ifNil:[^false].
	source hasSubmorphs ifTrue:[^true].
	self anyButtonPressed ifTrue:[^true].
	^false! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:30'!
isMouse
	^true! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:32'!
isMouseDown
	^self type == #mouseDown! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:32'!
isMouseEnter
	^self type == #mouseEnter! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:32'!
isMouseLeave
	^self type == #mouseLeave! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:32'!
isMouseMove
	^self type == #mouseMove! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 15:32'!
isMouseUp
	^self type == #mouseUp! !

!MouseEvent methodsFor: 'testing' stamp: 'mt 6/6/2016 09:39'!
isMouseWheel
	^ false! !

!MouseEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 19:29'!
isMove
	^false! !

!MouseEvent methodsFor: 'printing' stamp: 'ar 10/7/2000 22:01'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self cursorPoint printString; space.
	aStream nextPutAll: type; space.
	aStream nextPutAll: self modifierString.
	aStream nextPutAll: self buttonString.
	aStream nextPutAll: timeStamp printString.
	aStream nextPut: $].! !

!MouseEvent methodsFor: 'button state' stamp: 'NS 5/19/2003 15:17'!
redButtonPressed
	"Answer true if the red mouse button is being pressed. This is the first mouse button."

	^ buttons anyMask: self class redButton! !

!MouseEvent methodsFor: 'dispatching' stamp: 'ar 10/10/2000 21:15'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseOver ifTrue:[^anObject handleMouseOver: self].
	type == #mouseEnter ifTrue:[^anObject handleMouseEnter: self].
	type == #mouseLeave ifTrue:[^anObject handleMouseLeave: self].
	^super sentTo: anObject.! !

!MouseEvent methodsFor: 'private' stamp: 'ar 10/10/2000 21:15'!
setType: aSymbol
	"For quick conversion between event types"
	type := aSymbol.! !

!MouseEvent methodsFor: 'private' stamp: 'ar 9/15/2000 22:53'!
setType: evtType position: evtPos buttons: evtButtons hand: evtHand
	type := evtType.
	position := evtPos.
	buttons := evtButtons.
	source := evtHand.
	wasHandled := false.! !

!MouseEvent methodsFor: 'printing' stamp: 'ar 10/25/2000 22:09'!
storeOn: aStream

	aStream nextPutAll: type.
	aStream space.
	self timeStamp storeOn: aStream.
	aStream space.
	position x storeOn: aStream.
	aStream space.
	position y storeOn: aStream.
	aStream space.
	buttons storeOn: aStream.! !

!MouseEvent methodsFor: 'button state' stamp: 'ar 9/15/2000 22:51'!
targetPoint
	"Answer the location of the cursor's hotspot, adjusted by the offset
	of the last mouseDown relative to the recipient morph."

	^ position - source targetOffset! !

!MouseEvent methodsFor: 'initialize' stamp: 'ar 10/25/2000 22:08'!
type: eventType readFrom: aStream
	| x y |
	type := eventType.
	timeStamp := Integer readFrom: aStream.
	aStream skip: 1.
	x := Integer readFrom: aStream.
	aStream skip: 1.
	y := Integer readFrom: aStream.
	aStream skip: 1.
	buttons := Integer readFrom: aStream.
	position := x@y.
! !

!MouseEvent methodsFor: 'button state' stamp: 'NS 5/19/2003 15:17'!
yellowButtonPressed
	"Answer true if the yellow mouse button is being pressed. This is the second mouse button or option+click on the Mac."

	^ buttons anyMask: self class yellowButton! !

!MouseEvent methodsFor: 'comparing' stamp: 'ar 9/15/2000 22:50'!
= aMorphicEvent
	super = aMorphicEvent ifFalse:[^false].
	position = aMorphicEvent position ifFalse: [^ false].
	buttons = aMorphicEvent buttons ifFalse: [^ false].
	^ true
! !

MouseEvent subclass: #MouseMoveEvent
	instanceVariableNames: 'startPoint trail'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!MouseMoveEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 22:51'!
endPoint
	"Return the point where the movement ended."
	^position! !

!MouseMoveEvent methodsFor: 'comparing' stamp: 'ar 9/15/2000 22:49'!
hash
	^ position hash + startPoint hash + buttons hash! !

!MouseMoveEvent methodsFor: 'testing' stamp: 'ar 9/13/2000 19:29'!
isMove
	^true! !

!MouseMoveEvent methodsFor: 'printing' stamp: 'ar 10/7/2000 22:00'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self startPoint printString; space.
	aStream nextPutAll: self endPoint printString; space.
	aStream nextPutAll: self type; space.
	aStream nextPutAll: self modifierString.
	aStream nextPutAll: self buttonString.
	aStream nextPutAll: timeStamp printString.
	aStream nextPut: $].! !

!MouseMoveEvent methodsFor: 'dispatching' stamp: 'ar 10/10/2000 21:15'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseMove ifTrue:[^anObject handleMouseMove: self].
	^super sentTo: anObject.
! !

!MouseMoveEvent methodsFor: 'private' stamp: 'ar 10/5/2000 23:55'!
setType: evtType startPoint: evtStart endPoint: evtEnd trail: evtTrail buttons: evtButtons hand: evtHand stamp: stamp
	type := evtType.
	startPoint := evtStart.
	position := evtEnd.
	trail := evtTrail.
	buttons := evtButtons.
	source := evtHand.
	wasHandled := false.
	timeStamp := stamp.! !

!MouseMoveEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 16:25'!
startPoint
	"Return the point where the movement started."
	^startPoint! !

!MouseMoveEvent methodsFor: 'printing' stamp: 'ar 10/24/2000 16:30'!
storeOn: aStream
	super storeOn: aStream.
	aStream space.
	self startPoint x storeOn: aStream.
	aStream space.
	self startPoint y storeOn: aStream.
	aStream space.
	"trail storeOn: aStream."! !

!MouseMoveEvent methodsFor: 'accessing' stamp: 'ar 10/24/2000 16:33'!
trail
	"Return any immediate points that have been assembled along the move"
	^trail ifNil:[#()]! !

!MouseMoveEvent methodsFor: 'transforming' stamp: 'ar 9/15/2000 22:52'!
transformBy: aMorphicTransform
	"Transform the receiver into a local coordinate system."
	position :=  aMorphicTransform globalPointToLocal: position.
	startPoint :=  aMorphicTransform globalPointToLocal: startPoint.! !

!MouseMoveEvent methodsFor: 'transforming' stamp: 'ar 9/15/2000 22:52'!
translateBy: delta
	"add delta to cursorPoint, and return the new event"
	position := position + delta.
	startPoint := startPoint + delta.! !

!MouseMoveEvent methodsFor: 'initialize' stamp: 'ar 10/24/2000 16:31'!
type: eventType readFrom: aStream
	| x y |
	super type: eventType readFrom: aStream.
	aStream skip: 1.
	x := Integer readFrom: aStream.
	aStream skip: 1.
	y := Integer readFrom: aStream.
	startPoint := x@y.! !

!MouseMoveEvent methodsFor: 'comparing' stamp: 'ar 9/15/2000 22:49'!
= aMorphicEvent
	super = aMorphicEvent ifFalse:[^false].
	position = aMorphicEvent position ifFalse: [^ false].
	startPoint = aMorphicEvent startPoint ifFalse: [^ false].
	buttons = aMorphicEvent buttons ifFalse: [^ false].
	^ true
! !

MouseEvent subclass: #MouseWheelEvent
	instanceVariableNames: 'delta direction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!MouseWheelEvent methodsFor: 'initialization' stamp: 'mt 6/10/2016 14:24'!
initialize

	super initialize.
	direction := 2r0000.
	delta := 0@0.! !

!MouseWheelEvent methodsFor: 'testing' stamp: 'mt 6/6/2016 09:39'!
isMouseWheel
	^ true! !

!MouseWheelEvent methodsFor: 'testing' stamp: 'mt 6/10/2016 14:21'!
isWheelDown
	^ direction anyMask: 2r0100! !

!MouseWheelEvent methodsFor: 'testing' stamp: 'mt 6/10/2016 14:21'!
isWheelLeft
	^ direction anyMask: 2r0010! !

!MouseWheelEvent methodsFor: 'testing' stamp: 'mt 6/10/2016 14:21'!
isWheelRight
	^ direction anyMask: 2r0001! !

!MouseWheelEvent methodsFor: 'testing' stamp: 'mt 6/10/2016 14:21'!
isWheelUp
	^ direction anyMask: 2r1000! !

!MouseWheelEvent methodsFor: 'printing' stamp: 'mt 6/10/2016 15:47'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self cursorPoint printString; space.
	aStream nextPutAll: type; space.
	aStream nextPutAll: self wheelString.
	aStream nextPutAll: self wheelDelta printString; space.
	aStream nextPutAll: self modifierString.
	aStream nextPutAll: self buttonString.
	aStream nextPutAll: timeStamp printString; space.
	aStream nextPut: $].! !

!MouseWheelEvent methodsFor: 'dispatching' stamp: 'mt 6/6/2016 09:43'!
sentTo: anObject
	"Dispatch the receiver into anObject"
	type == #mouseWheel ifTrue: [^ anObject handleMouseWheel: self].
	^ super sentTo: anObject! !

!MouseWheelEvent methodsFor: 'private' stamp: 'mt 6/10/2016 14:57'!
setType: evtType position: evtPos delta: evtDelta direction: dirFlags buttons: evtButtons hand: evtHand stamp: stamp
	type := evtType.
	position := evtPos.
	buttons := evtButtons.
	source := evtHand.
	wasHandled := false.
	direction := dirFlags.
	delta := evtDelta.
	timeStamp := stamp.! !

!MouseWheelEvent methodsFor: 'initialization' stamp: 'mt 6/10/2016 15:43'!
setWheelDown

	direction := direction bitOr: 2r0100.! !

!MouseWheelEvent methodsFor: 'initialization' stamp: 'mt 6/10/2016 15:43'!
setWheelLeft

	direction := direction bitOr: 2r0010.! !

!MouseWheelEvent methodsFor: 'initialization' stamp: 'mt 6/10/2016 15:44'!
setWheelRight

	direction := direction bitOr: 2r0001.! !

!MouseWheelEvent methodsFor: 'initialization' stamp: 'mt 6/10/2016 15:44'!
setWheelUp

	direction := direction bitOr: 2r1000.! !

!MouseWheelEvent methodsFor: 'printing' stamp: 'mt 6/10/2016 14:35'!
storeOn: aStream
	"Note: We generate the same array as provided by the VM. BUT we are a subclass of MouseEvent now even if we cannot store the position right now. This is awkward. But we still store the position and the wheel delta. *sign*"
	
	super storeOn: aStream.
	aStream space.
	delta x storeOn: aStream.
	aStream space.
	delta y storeOn: aStream.
	aStream space.
	direction storeOn: aStream.
! !

!MouseWheelEvent methodsFor: 'initialization' stamp: 'mt 6/10/2016 15:51'!
type: eventType readFrom: aStream

	| deltaX deltaY |
	super type: eventType readFrom: aStream.
	aStream skip: 1.
	deltaX := Integer readFrom: aStream.
	aStream skip: 1.
	deltaY := Integer readFrom: aStream.
	aStream skip: 1.
	direction := Integer readFrom: aStream.
	delta := deltaX @ deltaY.
! !

!MouseWheelEvent methodsFor: 'accessing' stamp: 'mt 6/10/2016 14:15'!
wheelDelta

	^ delta! !

!MouseWheelEvent methodsFor: 'printing' stamp: 'mt 6/10/2016 15:46'!
wheelString
	"Return a string identifying the wheel state"
	
	^ String streamContents: [:stream |
		self isWheelUp ifTrue: [stream nextPutAll: 'up '].
		self isWheelDown ifTrue: [stream nextPutAll: 'down '].
		self isWheelLeft ifTrue: [stream nextPutAll: 'left '].
		self isWheelRight ifTrue: [stream nextPutAll: 'right ']]! !

!UserInputEvent methodsFor: 'modifier state' stamp: 'ar 9/13/2000 15:43'!
anyModifierKeyPressed
	"ignore, however, the shift keys 'cause that's not REALLY a command key "

	^ self buttons anyMask: 16r70	"cmd | opt | ctrl"! !

!UserInputEvent methodsFor: 'printing' stamp: 'ar 10/7/2000 21:57'!
buttonString
	"Return a string identifying the currently pressed buttons"
	| string |
	string := ''.
	self redButtonPressed ifTrue:[string := string,'red '].
	self yellowButtonPressed ifTrue:[string := string,'yellow '].
	self blueButtonPressed ifTrue:[string := string,'blue '].
	^string! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:45'!
buttons
	"Return the a word encoding the mouse and modifier buttons for this event."

	^ buttons! !

!UserInputEvent methodsFor: 'modifier state' stamp: 'ar 9/13/2000 15:43'!
commandKeyPressed
	"Answer true if the command key on the keyboard was being held down when this event occurred."

	^ buttons anyMask: 64! !

!UserInputEvent methodsFor: 'modifier state' stamp: 'ar 9/13/2000 15:44'!
controlKeyPressed
	"Answer true if the control key on the keyboard was being held down when this event occurred."

	^ buttons anyMask: 16! !

!UserInputEvent methodsFor: 'initialize' stamp: 'ar 9/13/2000 15:54'!
copyHandlerState: anEvent
	"Copy the handler state from anEvent. Used for quickly transferring handler information between transformed events."
	handler := anEvent handler.
	wasHandled := anEvent wasHandled.! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:54'!
handler
	^handler! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:54'!
handler: anObject
	handler := anObject! !

!UserInputEvent methodsFor: 'initialize' stamp: 'mt 5/19/2016 13:15'!
initialize

	super initialize.
	wasIgnored := false.! !

!UserInputEvent methodsFor: 'printing' stamp: 'ar 10/7/2000 21:56'!
modifierString
	"Return a string identifying the currently pressed modifiers"
	| string |
	string := ''.
	self commandKeyPressed ifTrue:[string := string,'CMD '].
	self shiftPressed ifTrue:[string := string,'SHIFT '].
	self controlKeyPressed ifTrue:[string := string,'CTRL '].
	^string! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/15/2000 22:45'!
position
	^position! !

!UserInputEvent methodsFor: 'initialize' stamp: 'ar 9/13/2000 15:54'!
resetHandlerFields
	"Reset anything that is used to cross-communicate between two eventual handlers during event dispatch"
	handler := nil.
	wasHandled := false.! !

!UserInputEvent methodsFor: 'private' stamp: 'ar 10/24/2000 16:33'!
setPosition: aPoint
	position := aPoint! !

!UserInputEvent methodsFor: 'modifier state' stamp: 'ar 9/13/2000 15:44'!
shiftPressed
	"Answer true if the shift key on the keyboard was being held down when this event occurred."

	^ buttons anyMask: 8
! !

!UserInputEvent methodsFor: 'transforming' stamp: 'ar 10/9/2000 00:38'!
transformBy: aMorphicTransform
	"Transform the receiver into a local coordinate system."
	position :=  aMorphicTransform globalPointToLocal: position.! !

!UserInputEvent methodsFor: 'transforming' stamp: 'ar 10/9/2000 00:38'!
transformedBy: aMorphicTransform
	"Return the receiver transformed by the given transform into a local coordinate system."
	^self shallowCopy transformBy: aMorphicTransform! !

!UserInputEvent methodsFor: 'transforming' stamp: 'ar 10/9/2000 00:37'!
translateBy: delta
	"add delta to cursorPoint, and return the new event"
	position := position + delta.! !

!UserInputEvent methodsFor: 'transforming' stamp: 'ar 10/9/2000 00:38'!
translatedBy: delta
	"add delta to cursorPoint, and return the new event"
	^self shallowCopy translateBy: delta! !

!UserInputEvent methodsFor: 'transforming' stamp: 'mt 5/20/2016 10:11'!
translateTo: point

	position := point.! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:45'!
type
	"Return a symbol indicating the type this event."

	^ type! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:54'!
wasHandled
	^wasHandled! !

!UserInputEvent methodsFor: 'accessing' stamp: 'ar 9/13/2000 15:54'!
wasHandled: aBool
	wasHandled := aBool.! !

!UserInputEvent methodsFor: 'accessing' stamp: 'bf 6/8/2016 18:06'!
wasIgnored
	^ wasIgnored == true! !

!UserInputEvent methodsFor: 'accessing' stamp: 'mt 5/19/2016 12:30'!
wasIgnored: aBoolean
	wasIgnored := aBoolean.! !

MorphicEvent subclass: #WindowEvent
	instanceVariableNames: 'action rectangle'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!WindowEvent commentStamp: '<historical>' prior: 0!
I'm an event related to the host window, only dispatched to the World. !

!WindowEvent methodsFor: 'accessing' stamp: 'JMM 10/6/2004 21:11'!
action
	^action! !

!WindowEvent methodsFor: 'accessing' stamp: 'JMM 10/6/2004 21:10'!
action: aValue
	action := aValue.! !

!WindowEvent methodsFor: 'testing' stamp: 'JMM 10/6/2004 21:35'!
isWindowEvent
	^true! !

!WindowEvent methodsFor: 'printing' stamp: 'bf 8/30/2007 20:39'!
printOn: aStream

	aStream nextPut: $[.
	aStream nextPutAll: self type; space.
	aStream nextPut: $(; print: self rectangle; nextPut: $).
	aStream nextPut: $].! !

!WindowEvent methodsFor: 'accessing' stamp: 'JMM 10/6/2004 21:12'!
rectangle
	^rectangle! !

!WindowEvent methodsFor: 'accessing' stamp: 'JMM 10/6/2004 21:12'!
rectangle: aValue
	rectangle := aValue.! !

!WindowEvent methodsFor: 'dispatching' stamp: 'bf 8/30/2007 17:31'!
sentTo:anObject
	"Dispatch the receiver into anObject"
	^anObject handleWindowEvent: self.! !

!WindowEvent methodsFor: 'accessing' stamp: 'bf 8/30/2007 20:42'!
type
	"This should match the definitions in sq.h"
	^#(
		windowMetricChange
		windowClose
		windowIconise
		windowActivated
		windowPaint
	) at: action ifAbsent: [#windowEventUnknown]! !

Object subclass: #MorphicEventDispatcher
	instanceVariableNames: 'lastType lastDispatch'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!MorphicEventDispatcher commentStamp: '<historical>' prior: 0!
The class represents a strategy for dispatching events to some immediate child of a morph. It is used by morphs to delegate the somewhat complex action of dispatching events accurately. !

!MorphicEventDispatcher methodsFor: 'specific' stamp: 'mt 5/20/2016 15:21'!
dispatchDefault: anEvent with: aMorph
	"Dispatch the given event. The event will be passed to the front-most visible submorph that contains the position wrt. to the event."

	"Try to get out quickly"
	anEvent wasIgnored ifTrue: [^anEvent "propagate the ignored event"].
	(aMorph fullBounds containsPoint: anEvent position) ifFalse:[^#rejected].

	^ self
		dispatchEvent: anEvent
		withHandler: aMorph
		withMorph: aMorph! !

!MorphicEventDispatcher methodsFor: 'specific' stamp: 'mt 5/20/2016 15:21'!
dispatchDropEvent: anEvent with: aMorph
	"Find the appropriate receiver for the event and let it handle it. The dispatch is similar to the default dispatch with one difference: Morphs are given the chance to reject an entire drop operation. If the operation is rejected, no drop will be executed."

	"Try to get out quickly"
	anEvent wasIgnored ifTrue: [^anEvent "propagate the ignored event"].
	(aMorph fullBounds containsPoint: anEvent position) ifFalse:[^#rejected].

	"Give aMorph a chance to repel the dropping morph"
	aMorph rejectDropEvent: anEvent.
	anEvent wasHandled ifTrue:[^anEvent].

	^ self
		dispatchEvent: anEvent
		withHandler: aMorph
		withMorph: aMorph! !

!MorphicEventDispatcher methodsFor: 'support' stamp: 'mt 6/20/2016 15:04'!
dispatchEvent: anEvent toSubmorphsOf: aMorph
	"Dispatch the given event to the submorphs of the given morph. For coordinate transformations, work only with copies. Either return the given event or a copy of any filtered event to employ immutability to some extent. --- PRIVATE!!"

	| localEvent filteredEvent |	
	aMorph submorphsDo: [:child |
		localEvent := anEvent transformedBy: (child transformedFrom: aMorph).
		filteredEvent := child
			processEvent: localEvent
			using: self. "use same dispatcher"
		filteredEvent == #rejected ifFalse: [ "some event or #rejected symbol"
			self flag: #overlappingChildren. "mt: We cannot give two overlapping siblings the chance to handle the event!!"	
			^ self nextFromOriginal: anEvent local: localEvent filtered: filteredEvent]].

	^ #rejected! !

!MorphicEventDispatcher methodsFor: 'support' stamp: 'mt 5/20/2016 15:12'!
dispatchEvent: anEvent withHandler: aHandler withMorph: aMorph
	"Perform the actual event dispatch. Use the given object as handler. Ask submorphs first to handle the event. Then bubble up. Stop if ignored. Note that event rejection and event filters are two separete concepts. Filters come from the outside. Rejection is a morph's decision.
	
	* The top-most chain of visible, unlocked morphs containing the event position will get a chance to handle the event.
	* When travelling up, the prospective handler is always executed. The handler needs to check if the event was handled before as well as checking if somebody else's handler has been installed.
	* If another handler has been installed but the event was not handled it means that somebody up in the hierarchy wants to handle the event."
	
	| result filteredEvent |
	
	result := self dispatchEvent: anEvent toSubmorphsOf: aMorph.

	result == #rejected "Anybody?"
		ifFalse: [filteredEvent := result]
		ifTrue: [
			"My submorphs did not want it. Do I want it anyway? It's about locked children..."
			(aMorph containsPoint: anEvent position event: anEvent)
				ifFalse: [^ #rejected].
			filteredEvent := anEvent "there was no filtering, only basic rejects"].
	
	"Receiver is in the top-most unlocked, visible chain."
	(aHandler notNil and: [filteredEvent wasIgnored not])
		ifTrue: [filteredEvent := aHandler handleEvent: filteredEvent].

	^ filteredEvent! !

!MorphicEventDispatcher methodsFor: 'dispatching' stamp: 'bf 8/30/2007 20:30'!
dispatchEvent: anEvent with: aMorph
	"Dispatch the given event for a morph that has chosen the receiver to dispatch its events. The method implements a shortcut for repeated dispatches of events using the same dispatcher."
	anEvent type == lastType ifTrue:[^self perform: lastDispatch with: anEvent with: aMorph].
	"Otherwise classify"
	lastType := anEvent type.
	anEvent isMouse ifTrue:[
		anEvent isMouseDown ifTrue:[
			lastDispatch := #dispatchMouseDown:with:.
			^self dispatchMouseDown: anEvent with: aMorph]].
	anEvent type == #dropEvent ifTrue:[
		lastDispatch := #dispatchDropEvent:with:.
		^self dispatchDropEvent: anEvent with: aMorph].
	anEvent isWindowEvent ifTrue:[
		lastDispatch := #dispatchWindowEvent:with:.
		^self dispatchWindowEvent: anEvent with: aMorph].
	lastDispatch := #dispatchDefault:with:.
	^self dispatchDefault: anEvent with: aMorph! !

!MorphicEventDispatcher methodsFor: 'focus events' stamp: 'mt 6/20/2016 16:57'!
dispatchFocusEventFully: anEventWithGlobalPosition with: focusMorph
	"Dispatch the given event to the given morph. Do capturing, processing in sub-tree, and bubbling."
	
	| currentEvent |
	"1) Capturing phase."
	currentEvent := self doCapturingForFocusEvent: anEventWithGlobalPosition with: focusMorph.
	currentEvent == #rejected ifTrue: [^ #rejected].
	currentEvent wasIgnored ifTrue: [^ currentEvent].
	
	"2) Sub-tree processing."
	currentEvent := self doProcessingForFocusEvent: currentEvent with: focusMorph.
	currentEvent wasIgnored ifTrue: [^ currentEvent].

	"3) Let the focus morph handle the event. Usually no effect because previous sub-tree processing involved the focus morph already -- at least in the bubbling phase. Skip it?"
	currentEvent := self doHandlingForFocusEvent: currentEvent with: focusMorph.
	currentEvent wasIgnored ifTrue: [^ currentEvent].
	
	"4) Bubbling phase."
	^ self doBubblingForFocusEvent: currentEvent with: focusMorph! !

!MorphicEventDispatcher methodsFor: 'focus events' stamp: 'mt 6/20/2016 16:57'!
dispatchFocusEvent: anEventWithGlobalPosition with: focusMorph
	"Dispatch the given event to the given morph. Simulate capturing phase, handle the event, then do bubbling."

	| currentEvent |
	"1) Capturing phase."
	currentEvent := self doCapturingForFocusEvent: anEventWithGlobalPosition with: focusMorph.
	currentEvent == #rejected ifTrue: [^ #rejected].
	currentEvent wasIgnored ifTrue: [^ currentEvent].
	
	"2) No sub-tree processing here. Use #dispatchFocusEventFully:with: if you want that, too."
	
	"3) Let the focus morph handle the event." 
	currentEvent := self doHandlingForFocusEvent: currentEvent with: focusMorph.
	currentEvent wasIgnored ifTrue: [^ currentEvent].
	
	"4) Bubbling phase"
	^ self doBubblingForFocusEvent: currentEvent with: focusMorph! !

!MorphicEventDispatcher methodsFor: 'specific' stamp: 'mt 5/20/2016 15:20'!
dispatchMouseDown: anEvent with: aMorph
	"Find the appropriate receiver for the event and let it handle it. When travelling down the hierarchy a prospective handler for the event is installed. This prospective handler can be used by submorphs wishing to handle the mouse down for negotiating who the receiver is."
	
	| handler lastHandler |
	
	"Try to get out quickly"
	anEvent wasIgnored ifTrue: [^anEvent "propagate the ignored event"].
	(aMorph fullBounds containsPoint: anEvent position) ifFalse:[^#rejected].

	"In case the mouse wasn't even in the receiver"
	lastHandler := anEvent handler. 

	"Negotiate and install the actual handler."
	handler := aMorph handlerForMouseDown: anEvent.
	handler ifNotNil: [anEvent handler: handler].

	[^ self
		dispatchEvent: anEvent
		withHandler: handler
		withMorph: aMorph
	] ensure: [
		anEvent handler: lastHandler. "good style"].! !

!MorphicEventDispatcher methodsFor: 'specific' stamp: 'mt 5/20/2016 15:19'!
dispatchWindowEvent: anEvent with: aMorph
	"Host window events do not have a position and are only dispatched to the World"

	anEvent wasIgnored ifTrue: [^anEvent].
	aMorph isWorldMorph ifFalse: [^#rejected].
	anEvent wasHandled ifTrue:[^anEvent].
	^aMorph handleEvent: anEvent! !

!MorphicEventDispatcher methodsFor: 'support' stamp: 'mt 6/20/2016 16:37'!
doBubblingForFocusEvent: anEvent with: focusMorph
	"Simulate real event bubbling up to the focused morph's outermost owner. Applies event bubble filters via Morph >> #handleEvent:. Watch out for coordinate transformations and some globals (ActiveWorld, ...)."

	| currentEvent filteredEvent localEvent referenceMorph |
	currentEvent := anEvent.
	referenceMorph := anEvent hand.

	focusMorph allOwnersDo: [:ownerMorph |
		ownerMorph == anEvent hand ifFalse: [ "Never bubble up to the hand morph but only up to the world."
			localEvent := currentEvent transformedBy: (ownerMorph transformedFrom: referenceMorph).
			filteredEvent := ownerMorph handleEvent: localEvent.
			currentEvent := self nextFromOriginal: currentEvent local: localEvent filtered: filteredEvent.
			currentEvent wasIgnored ifTrue: [^ currentEvent]]].

	^ currentEvent! !

!MorphicEventDispatcher methodsFor: 'support' stamp: 'mt 6/20/2016 16:45'!
doCapturingForFocusEvent: anEvent with: focusMorph
	"Simulate real event capturing down to the focused morph. Apply event capture filters. Watch out for coordinate transformations. Keep the filter-ignore-reject order like in Morph >> #processEvent:using:."

	| currentEvent filteredEvent localEvent referenceMorph |
	currentEvent := anEvent.
	referenceMorph := anEvent hand.

	"Event capturing. Filters only because the capturing phase was bypassed by using the keyboard/mouse focus."
	focusMorph withAllOwners reverseDo: [:ownerMorph | "reverse order to comply with regular dispatching"
		ownerMorph == anEvent hand ifFalse: [ "Never dispatch the hand morph. It already did so."
			localEvent := currentEvent transformedBy: (ownerMorph transformedFrom: referenceMorph).
			
			filteredEvent := ownerMorph sendFilterEventCapture: localEvent for: ownerMorph.
			
			"Ignoring has higher priority but the reject-check must be with local event coodinates."
			(filteredEvent wasIgnored not and: [ownerMorph rejectsEvent: filteredEvent])
				ifTrue: [^ #rejected].
			
			currentEvent := self nextFromOriginal: currentEvent local: localEvent filtered: filteredEvent.
			currentEvent wasIgnored ifTrue: [^ currentEvent]]].
	
	^ currentEvent! !

!MorphicEventDispatcher methodsFor: 'support' stamp: 'mt 6/20/2016 16:40'!
doHandlingForFocusEvent: currentEvent with: focusMorph

	| localEvent filteredEvent |
	localEvent := currentEvent transformedBy: (focusMorph transformedFrom: currentEvent hand).
	filteredEvent := focusMorph handleFocusEvent: localEvent.
	^ self nextFromOriginal: currentEvent local: localEvent filtered: filteredEvent.! !

!MorphicEventDispatcher methodsFor: 'support' stamp: 'mt 6/20/2016 16:56'!
doProcessingForFocusEvent: currentEvent with: focusMorph
	"Sub-tree processing (including capturing from focus morph down to something and bubbling up back to focus morph). Never reject in the end."
	
	| localEvent filteredEvent |
	localEvent := currentEvent transformedBy: (focusMorph transformedFrom: currentEvent hand).
	filteredEvent := focusMorph processEvent: localEvent using: self.
	^ filteredEvent == #rejected
		ifTrue: [currentEvent] "Can happen if you click, e.g., outside the bounds of the focus morph"
		ifFalse: [self nextFromOriginal: currentEvent local: localEvent filtered: filteredEvent]! !

!MorphicEventDispatcher methodsFor: 'support' stamp: 'mt 6/20/2016 15:26'!
nextFromOriginal: originalEvent local: localEvent filtered: filteredEvent
	"Take the filtered event if different but always keep the original coordinates."
	
	filteredEvent == localEvent
		ifTrue: [ "Use original event but keep track of ignored flag."
			localEvent wasHandled ifTrue: [originalEvent copyHandlerState: localEvent].
			originalEvent wasIgnored: localEvent wasIgnored.
			^ originalEvent]
		ifFalse: [ "There was an event transformation. Copy, revert coordinates, keep handler state."
			| result |
			result := filteredEvent copy. "Never mutate position without copying. MouseClickState etc. will break otherwise."
			result translateTo: originalEvent position. "restore to untransformed coordinates"
			result wasHandled ifFalse: [result copyHandlerState: originalEvent].
			^ result].
! !

Object subclass: #MouseClickState
	instanceVariableNames: 'clickClient clickState firstClickDown firstClickUp firstClickTime clickSelector dblClickSelector dblClickTime dblClickTimeoutSelector dragSelector dragThreshold'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!MouseClickState commentStamp: '<historical>' prior: 0!
MouseClickState is a simple class managing the distinction between clicks, double clicks, and drag operations. It has been factored out of HandMorph due to the many instVars.

Instance variables:
	clickClient 	<Morph>		The client wishing to receive #click:, #dblClick:, or #drag messages
	clickState 	<Symbol>	The internal state of handling the last event (#firstClickDown, #firstClickUp, #firstClickTimedOut)
	firstClickDown 	<MorphicEvent>	The #mouseDown event after which the client wished to receive #click: or similar messages
	firstClickUp 	<MorphicEvent>	The first mouse up event which came in before the double click time out was exceeded (it is sent if there is a timout after the first mouse up event occured)
	firstClickTime 	<Integer>	The millisecond clock value of the first event
	clickSelector 	<Symbol>	The selector to use for sending #click: messages
	dblClickSelector 	<Symbol>	The selector to use for sending #doubleClick: messages
	dblClickTime 	<Integer>	Timout in milliseconds for a double click operation
	dragSelector 	<Symbol>	The selector to use for sending #drag: messages
	dragThreshold 	<Integer>	Threshold used for determining if a #drag: message is sent (pixels!!)
!

!MouseClickState methodsFor: 'event handling' stamp: 'jcg 9/21/2001 11:23'!
click

	clickSelector ifNotNil: [clickClient perform: clickSelector with: firstClickDown]! !

!MouseClickState methodsFor: 'initialize' stamp: 'jcg 9/21/2001 13:08'!
client: aMorph click: aClickSelector dblClick: aDblClickSelector dblClickTime: timeOut dblClickTimeout: aDblClickTimeoutSelector drag: aDragSelector threshold: aNumber event: firstClickEvent
	clickClient := aMorph.
	clickSelector := aClickSelector.
	dblClickSelector := aDblClickSelector.
	dblClickTime := timeOut.
	dblClickTimeoutSelector := aDblClickTimeoutSelector.
	dragSelector := aDragSelector.
	dragThreshold := aNumber.
	firstClickDown := firstClickEvent.
	firstClickTime := firstClickEvent timeStamp.
	clickState := #firstClickDown.! !

!MouseClickState methodsFor: 'event handling' stamp: 'jcg 9/21/2001 11:24'!
doubleClick

	dblClickSelector ifNotNil: [clickClient perform: dblClickSelector with: firstClickDown]! !

!MouseClickState methodsFor: 'event handling' stamp: 'jcg 9/21/2001 13:09'!
doubleClickTimeout

	dblClickTimeoutSelector ifNotNil: [
		clickClient perform: dblClickTimeoutSelector with: firstClickDown]! !

!MouseClickState methodsFor: 'event handling' stamp: 'jcg 9/21/2001 11:27'!
drag: event

	dragSelector ifNotNil: [clickClient perform: dragSelector with: event]! !

!MouseClickState methodsFor: 'event handling' stamp: 'nk 7/26/2004 10:21'!
handleEvent: evt from: aHand
	"Process the given mouse event to detect a click, double-click, or drag.
	Return true if the event should be processed by the sender, false if it shouldn't.
	NOTE: This method heavily relies on getting *all* mouse button events."
	| localEvt timedOut isDrag |
	timedOut := (evt timeStamp - firstClickTime) > dblClickTime.
	localEvt := evt transformedBy: (clickClient transformedFrom: aHand owner).
	isDrag := (localEvt position - firstClickDown position) r > dragThreshold.
	clickState == #firstClickDown ifTrue: [
		"Careful here - if we had a slow cycle we may have a timedOut mouseUp event"
		(timedOut and:[localEvt isMouseUp not]) ifTrue:[
			"timeout before #mouseUp -> keep waiting for drag if requested"
			clickState := #firstClickTimedOut.
			dragSelector ifNil:[
				aHand resetClickState.
				self doubleClickTimeout; click "***"].
			^true].
		localEvt isMouseUp ifTrue:[

			(timedOut or:[dblClickSelector isNil]) ifTrue:[
				self click.
				aHand resetClickState.
				^true].
			"Otherwise transfer to #firstClickUp"
			firstClickUp := evt copy.
			clickState := #firstClickUp.
			"If timedOut or the client's not interested in dbl clicks get outta here"
			self click.
			aHand handleEvent: firstClickUp.
			^false].
		isDrag ifTrue:["drag start"
			self doubleClickTimeout. "***"
			aHand resetClickState.
			dragSelector "If no drag selector send #click instead"
				ifNil: [self click]
				ifNotNil: [self drag: firstClickDown].
			^true].
		^false].

	clickState == #firstClickTimedOut ifTrue:[
		localEvt isMouseUp ifTrue:["neither drag nor double click"
			aHand resetClickState.
			self doubleClickTimeout; click. "***"
			^true].
		isDrag ifTrue:["drag start"
			aHand resetClickState.
			self doubleClickTimeout; drag: firstClickDown. "***"
			^true].
		^false].

	clickState = #firstClickUp ifTrue:[
		(timedOut) ifTrue:[
			"timed out after mouseUp - signal timeout and pass the event"
			aHand resetClickState.
			self doubleClickTimeout. "***"
			^true].
		localEvt isMouseDown ifTrue:["double click"
			clickState := #secondClickDown.
			^false]].

	clickState == #secondClickDown ifTrue: [
		timedOut ifTrue:[
			"timed out after second mouseDown - pass event after signaling timeout"
			aHand resetClickState.
			self doubleClickTimeout. "***"
			^true].
		isDrag ifTrue: ["drag start"
			self doubleClickTimeout. "***"
			aHand resetClickState.
			dragSelector "If no drag selector send #click instead"
				ifNil: [self click]
				ifNotNil: [self drag: firstClickDown].
			^true].
		localEvt isMouseUp ifTrue: ["double click"
			aHand resetClickState.
			self doubleClick.
			^false]
	].

	^true
! !

!MouseClickState methodsFor: 'as yet unclassified' stamp: 'nk 7/26/2004 09:13'!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $[; print: clickState; nextPut: $]
! !

Object subclass: #MouseOverHandler
	instanceVariableNames: 'mouseOverMorphs enteredMorphs overMorphs leftMorphs'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!MouseOverHandler methodsFor: 'initialize-release' stamp: 'ul 11/23/2009 02:28'!
initialize

	mouseOverMorphs := Array new.
	self initializeTrackedMorphs! !

!MouseOverHandler methodsFor: 'initialize-release' stamp: 'ul 11/23/2009 02:28'!
initializeTrackedMorphs

	leftMorphs := OrderedCollection new.
	overMorphs := WriteStream on: #().
	enteredMorphs := WriteStream on: #().! !

!MouseOverHandler methodsFor: 'event handling' stamp: 'ul 11/22/2009 21:53'!
noticeMouseOver: aMorph event: anEvent
	"Remember that the mouse is currently over some morph"

	leftMorphs remove: aMorph ifAbsent: [
		enteredMorphs nextPut: aMorph ].
	overMorphs nextPut: aMorph.
! !

!MouseOverHandler methodsFor: 'event handling' stamp: 'mt 6/4/2016 18:41'!
processMouseOver: anEvent 
	"Re-establish the z-order for all morphs wrt the given event"

	| hand localEvt focus evt |
	anEvent ifNil: [^ self].
	hand := anEvent hand.
	leftMorphs := mouseOverMorphs asIdentitySet.
	"Assume some coherence for the number of objects in over list"
	overMorphs := WriteStream on: (Array new: leftMorphs size).
	enteredMorphs := WriteStream on: #().
	"Now go looking for eventual mouse overs"
	hand handleEvent: anEvent asMouseOver.
	"Get out early if there's no change"
	(leftMorphs isEmpty and: [ enteredMorphs position = 0 ]) 
		ifTrue: [ ^self initializeTrackedMorphs ].
	focus := hand mouseFocus.
	"Send #mouseLeave as appropriate"
	evt := anEvent asMouseLeave.
	"Keep the order of the left morphs by recreating it from the mouseOverMorphs"
	leftMorphs size > 1 
		ifTrue:
			[leftMorphs := mouseOverMorphs select: [:m | leftMorphs includes: m]].
			leftMorphs do: [ :m | 
			(m == focus or: [m hasOwner: focus])
				ifFalse: [ overMorphs nextPut: m ]
				ifTrue: 
					[ localEvt := evt transformedBy: (m transformedFrom: hand).
					m handleEvent: localEvt ] ].
	enteredMorphs ifNil: [ "inform: was called in handleEvent:"
		^self initializeTrackedMorphs ].
	"Send #mouseEnter as appropriate"
	evt := anEvent asMouseEnter.
	enteredMorphs contents reverseDo: [ :m | 
		(m == focus or: [m hasOwner: focus]) ifTrue: [
			localEvt := evt transformedBy: (m transformedFrom: hand).
			m handleEvent: localEvt ] ].
	"And remember the over list"
	overMorphs ifNotNil: [ mouseOverMorphs := overMorphs contents ].
	self initializeTrackedMorphs! !

Object subclass: #MouseWheelState
	instanceVariableNames: 'currentDelta'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!MouseWheelState methodsFor: 'event processing' stamp: 'mt 6/10/2016 15:53'!
handleEvent: aMouseWheelEvent from: aHand
	"Every 120 units, raise the wheel flags for convenient mouse wheel programming. We choose not to send multiple mouse-wheel events for multiples of 120 because applications can always react to the actual delta values if they want to do more scrolling or zooming."
	
	| sign |
	currentDelta := currentDelta + aMouseWheelEvent wheelDelta.

	sign := currentDelta sign.
	currentDelta := currentDelta abs.

	(currentDelta x // 120) > 0 ifTrue: [
		sign x = 1
			ifTrue: [aMouseWheelEvent setWheelRight]
			ifFalse: [aMouseWheelEvent setWheelLeft]].

	(currentDelta y // 120) > 0 ifTrue: [
		sign y = 1
			ifTrue: [aMouseWheelEvent setWheelUp]
			ifFalse: [aMouseWheelEvent setWheelDown]].
		
	currentDelta := currentDelta \\ 120.
	currentDelta := currentDelta * sign.

	"Finally, send the event."
	HandMorph sendMouseWheelToKeyboardFocus
		ifFalse: [aHand sendMouseEvent: aMouseWheelEvent]
		ifTrue: [aHand sendEvent: aMouseWheelEvent focus: aHand keyboardFocus clear: [aHand keyboardFocus: nil]].
! !

!MouseWheelState methodsFor: 'initialize-release' stamp: 'mt 6/10/2016 14:42'!
initialize

	super initialize.
	currentDelta := 0@0.! !

Object subclass: #NewParagraph
	instanceVariableNames: 'text textStyle firstCharacterIndex container lines positionWhenComposed offsetToEnd maxRightX selectionStart selectionStop wantsColumnBreaks focused caretRect showCaret caretColor selectionColor unfocusedSelectionColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!NewParagraph commentStamp: '<historical>' prior: 0!
A Paragraph represents text that has been laid out, or composed, in some container.
	text 		A Text with encoded per-character emphasis.
	textStyle	A TextStyle with font set, line height and horizontal alignment.
	firstCharacterIndex    The starting index in text for this paragraph, allowing
				composition of a long text into a number of containers.
	container	A Rectangle or TextContainer that determines where text can go.
	lines		An Array of TextLines comprising the final layout of the text
				after it has been composed within its container.
	positionWhenComposed   As its name implies.  Allows display at new locations
				without the need to recompose the text.
Lines are ordered vertically.  However, for a given y, there may be several lines in left to right order.  Lines must never be empty, even if text is empty.

Notes on yet another hack - 5 Feb 2001

We really need to clean up #composeLinesFrom:to:delta:into:priorLines:atY:!!!!!!

I added one more habdful of code to correct:

This is an annoying bug that's been around for a couple of years, but I finally figured out how to duplicate the problem, so I figured I'd just report it now.  (It doesn't necessarily have to be fixed for 3.0 if it looks messy, but if it's a simple fix, it would be worth it.)

In Morphic, if you have the following text in a workspace:

This is line 1
This is line 2

**and** you have a return character after line 2, you will normally be able to click the mouse two times below line 2 in order to select all the text.  If you edit line 2 (e.g. so that it reads "line number 2"), you can still select all the text by clicking below the second line.  However, if you edit line 1, you will not be able to select all the text from the bottom in the same way.  Things get messed up such that the last return character seems to be gone.  In this state, if you position the cursor immediately after the 2, and press the right arrow, the cursor jumps to the beginning of line 2... oof. (report by Doug Way)

While I don't have a very deep understanding of the above mentioned method, I was able to determine that text ending in a CR worked better in the editor when the last entry in <lines> had a start of text size + 1 and a stop of text size. I have accordingly added code near the end to ensure this. It seems to have fixed the problem, but we do need to clean this baby up some day. - Bob
!

!NewParagraph methodsFor: 'access' stamp: 'di 11/16/97 09:02'!
adjustedFirstCharacterIndex
	"Return the index in the text where this paragraph WOULD begin if nothing had changed, except the size of the text -- ie if there have only been an insertion of deletion in the preceding morphs"
	offsetToEnd ifNil: [^ -1].
	^ text size - offsetToEnd! !

!NewParagraph methodsFor: 'private' stamp: 'di 11/8/97 15:47'!
adjustLineIndicesBy: delta
	firstCharacterIndex := firstCharacterIndex + delta.
	lines do: [:line | line slide: delta].
! !

!NewParagraph methodsFor: 'private' stamp: 'mt 4/27/2015 15:10'!
adjustRightX
	| shrink |
	shrink := container right - maxRightX.
	lines do: [:line | line paddingWidth: (line paddingWidth - shrink)].
	container := container withRight: maxRightX + self caretWidth.! !

!NewParagraph methodsFor: 'display' stamp: 'di 8/13/2000 12:27'!
asParagraphForPostscript

	^ self! !

!NewParagraph methodsFor: 'editing' stamp: 'ar 12/13/2010 23:26'!
attributesAt: aPoint
	"Answer the attributes at the given point"
	^text attributesAt: (self characterBlockAtPoint: aPoint) stringIndex forStyle: textStyle! !

!NewParagraph methodsFor: 'access' stamp: 'mt 7/24/2016 13:28'!
caretColor
	^ caretColor ifNil: [Color red]! !

!NewParagraph methodsFor: 'access' stamp: 'mt 7/24/2016 13:31'!
caretColor: aColor
	caretColor := aColor.! !

!NewParagraph methodsFor: 'access' stamp: 'ar 12/15/2008 22:41'!
caretRect
	"The rectangle in which the caret was last drawn,
	 or nil if the last drawing drew a range-selection rather than insertion point."
	^ caretRect! !

!NewParagraph methodsFor: 'access' stamp: 'mt 7/24/2016 13:53'!
caretWidth
	^ Editor dumbbellCursor
		ifTrue: [ 3 ]
		ifFalse: [ 2 ]! !

!NewParagraph methodsFor: 'alignment' stamp: 'di 10/25/97 19:26'!
centered 
	textStyle centered! !

!NewParagraph methodsFor: 'selection' stamp: 'nice 9/29/2013 15:30'!
characterBlockAtPoint: aPoint 
	"Answer a CharacterBlock for the character in the text at aPoint."
	| line |
	line := lines at: (self lineIndexForPoint: aPoint).
	^(CharacterBlockScanner new text: text textStyle: textStyle)
		characterBlockAtPoint: aPoint index: nil
		in: line! !

!NewParagraph methodsFor: 'selection' stamp: 'nice 9/29/2013 15:30'!
characterBlockForIndex: index 
	"Answer a CharacterBlock for the character in text at index."
	| line |
	line := lines at: (self lineIndexOfCharacterIndex: index).
	^ (CharacterBlockScanner new text: text textStyle: textStyle)
		characterBlockAtPoint: nil index: ((index max: line first) min: text size+1)
		in: line! !

!NewParagraph methodsFor: 'editing' stamp: 'nice 12/27/2009 20:13'!
clickAt: clickPoint for: model controller: editor
	"Give sensitive text a chance to fire.  Display flash: (100@100 extent: 100@100)."
	| startBlock action |
	action := false.
	startBlock := self characterBlockAtPoint: clickPoint.
	(text attributesAt: startBlock stringIndex forStyle: textStyle) 
		do: [:att | | range target box boxes |
			att mayActOnClick ifTrue:
				[(target := model) ifNil: [target := editor morph].
				range := text rangeOf: att startingAt: startBlock stringIndex.
				boxes := self selectionRectsFrom: (self characterBlockForIndex: range first) 
							to: (self characterBlockForIndex: range last+1).
				box := boxes detect: [:each | each containsPoint: clickPoint] ifNone: [nil].
				box ifNotNil:
					[ box := (editor transformFrom: nil) invertBoundsRect: box.
					editor morph allOwnersDo: [ :m | box := box intersect: (m boundsInWorld) ].
					Utilities awaitMouseUpIn: box
						repeating: []
						ifSucceed: [(att actOnClickFor: target in: self at: clickPoint editor: editor) ifTrue: [action := true]].
					Cursor currentCursor == Cursor webLink ifTrue:[Cursor normal show].
				]]].
	^ action! !

!NewParagraph methodsFor: 'composition' stamp: 'tpr 9/25/2013 13:59'!
composeAll
	self composeLinesFrom: firstCharacterIndex to: text size delta: 0
			into: OrderedCollection new priorLines: Array new atY: container top! !

!NewParagraph methodsFor: 'composition' stamp: 'di 11/15/97 09:21'!
composeAllStartingAt: characterIndex
	firstCharacterIndex := characterIndex.
	offsetToEnd := text size - firstCharacterIndex.
	self composeAll! !

!NewParagraph methodsFor: 'composition' stamp: 'RAA 5/7/2001 10:58'!
composeLinesFrom: start to: stop delta: delta into: lineColl priorLines: priorLines
	atY: startingY
	"While the section from start to stop has changed, composition may ripple all the way to the end of the text.  However in a rectangular container, if we ever find a line beginning with the same character as before (ie corresponding to delta in the old lines), then we can just copy the old lines from there to the end of the container, with adjusted indices and y-values"

	| newResult |

	newResult := TextComposer new
		composeLinesFrom: start 
		to: stop 
		delta: delta 
		into: lineColl 
		priorLines: priorLines
		atY: startingY
		textStyle: textStyle 
		text: text 
		container: container
		wantsColumnBreaks: wantsColumnBreaks == true.
	lines := newResult first asArray.
	maxRightX := newResult second.
	^maxRightX
! !

!NewParagraph methodsFor: 'composition' stamp: 'di 11/8/97 15:31'!
compose: t style: ts from: startingIndex in: textContainer
	text := t.
	textStyle := ts.
	firstCharacterIndex := startingIndex.
	offsetToEnd := text size - firstCharacterIndex.
	container := textContainer.
	self composeAll! !

!NewParagraph methodsFor: 'composition' stamp: 'di 10/22/97 11:13'!
compositionRectangle
	^ container! !

!NewParagraph methodsFor: 'selection' stamp: 'jm 11/19/97 22:56'!
containsPoint: aPoint
	^ (lines at: (self lineIndexForPoint: aPoint)) rectangle
		containsPoint: aPoint! !

!NewParagraph methodsFor: 'copying' stamp: 'di 5/21/1998 21:45'!
deepCopy
	"Don't want to copy the container (etc) or fonts in the TextStyle."
	| new |
	new := self copy.
	new textStyle: textStyle copy
		lines: lines copy
		text: text deepCopy.
	^ new! !

!NewParagraph methodsFor: 'selection' stamp: 'di 10/5/1998 12:59'!
defaultCharacterBlock
	^ (CharacterBlock new stringIndex: firstCharacterIndex text: text
			topLeft: lines first topLeft extent: 0 @ 0)
		textLine: lines first! !

!NewParagraph methodsFor: 'display' stamp: 'mt 7/24/2016 15:19'!
displayDumbbellCursorOn: aCanvas at: leftX in: line

	| w |
	w := 2.
	self focused ifFalse: [^ w].
	
	1 to: w
		do: 
			[:i | 
			"Draw caret triangles at top and bottom"

			aCanvas fillRectangle: ((leftX - w + i - 1) @ (line top + i - 1) 
						extent: ((w - i) * 2 + 3) @ 1)
				color: self caretColor.
			aCanvas fillRectangle: ((leftX - w + i - 1) @ (line bottom - i) 
						extent: ((w - i) * 2 + 3) @ 1)
				color: self caretColor].
	
	aCanvas
		line: leftX @ line top
		to: leftX @ (line bottom-1)
		color: self caretColor.
	
	^ w! !

!NewParagraph methodsFor: 'fonts-display' stamp: 'nk 3/20/2004 11:13'!
displayOn: aCanvas using: displayScanner at: somePosition
	"Send all visible lines to the displayScanner for display"
	| visibleRectangle offset leftInRun line |
	visibleRectangle := aCanvas clipRect.
	offset := (somePosition - positionWhenComposed) truncated.
	leftInRun := 0.
	(self lineIndexForPoint: visibleRectangle topLeft)
		to: (self lineIndexForPoint: visibleRectangle bottomRight)
		do: [:i | line := lines at: i.
			self displaySelectionInLine: line on: aCanvas.
			line first <= line last ifTrue:
				[leftInRun := displayScanner displayLine: line
								offset: offset leftInRun: leftInRun]].
! !

!NewParagraph methodsFor: 'display' stamp: 'mt 7/24/2016 15:15'!
displaySelectionInLine: line on: aCanvas 
	| leftX rightX w |
	selectionStart ifNil: [^self].	"No selection"
	aCanvas isShadowDrawing ifTrue: [ ^self ].	"don't draw selection with shadow"
	selectionStart = selectionStop 
		ifTrue: 
			["Only show caret on line where clicked"

			selectionStart textLine ~= line ifTrue: [^self]]
		ifFalse:
			["Test entire selection before or after here"

			(selectionStop stringIndex < line first 
				or: [selectionStart stringIndex > (line last + 1)]) ifTrue: [^self].	"No selection on this line"
			(selectionStop stringIndex = line first 
				and: [selectionStop textLine ~= line]) ifTrue: [^self].	"Selection ends on line above"
			(selectionStart stringIndex = (line last + 1) 
				and: [selectionStop textLine ~= line]) ifTrue: [^self]].	"Selection begins on line below"
	leftX := (selectionStart stringIndex < line first 
				ifTrue: [line ]
				ifFalse: [selectionStart ])left.
	rightX := (selectionStop stringIndex > (line last + 1) or: 
					[selectionStop stringIndex = (line last + 1) 
						and: [selectionStop textLine ~= line]]) 
				ifTrue: [line right]
				ifFalse: [selectionStop left].
	selectionStart = selectionStop 
		ifTrue: [
			rightX := rightX + 1.
			caretRect := (leftX-2) @ line top corner: (rightX+2)@ line bottom. "sigh..."
			self showCaret ifFalse: [^self].
			w := (Editor dumbbellCursor
				ifTrue: [self displayDumbbellCursorOn: aCanvas at: leftX in: line]
				ifFalse: [self displaySimpleCursorOn: aCanvas at: leftX in: line]).
			caretRect := (leftX-w) @ line top corner: (rightX+w)@ line bottom]
		ifFalse: [
			caretRect := nil.
			aCanvas fillRectangle: (leftX @ line top corner: rightX @ line bottom)
				color: (self focused ifTrue: [self selectionColor] ifFalse: [self unfocusedSelectionColor])]! !

!NewParagraph methodsFor: 'display' stamp: 'mt 7/24/2016 15:19'!
displaySimpleCursorOn: aCanvas at: leftX in: line

	self focused ifFalse: [^ 1].
	
	aCanvas
		line: leftX @ (line top+1)
		to: leftX @ (line bottom-1)
		color: self caretColor.

	aCanvas
		line: leftX+1 @ (line top+1)
		to: leftX+1 @ (line bottom-1)
		color: (self caretColor alpha: 0.3).
	
	^ 1! !

!NewParagraph methodsFor: 'access' stamp: 'di 10/24/97 17:38'!
extent
	^ container width @ (lines last bottom - lines first top)! !

!NewParagraph methodsFor: 'private' stamp: 'ul 5/12/2011 00:30'!
fastFindFirstLineSuchThat: lineBlock
	"Perform a binary search of the lines array and return the index
	of the first element for which lineBlock evaluates as true.
	This assumes the condition is one that goes from false to true for
	increasing line numbers (as, eg, yval > somey or start char > somex).
	If lineBlock is not true for any element, return size+1."
	
	^lines
		findBinaryIndex: [ :each | 
			(lineBlock value: each)
				ifTrue: [ -1 ]
				ifFalse: [ 1 ] ]
		ifNone: [ :lower :upper | upper ]! !

!NewParagraph methodsFor: 'access' stamp: 'di 11/8/97 15:41'!
firstCharacterIndex
	^ firstCharacterIndex! !

!NewParagraph methodsFor: 'composition' stamp: 'RAA 2/25/2001 15:02'!
fixLastWithHeight: lineHeightGuess
"This awful bit is to ensure that if we have scanned all the text and the last character is a CR that there is a null line at the end of lines. Sometimes this was not happening which caused anomalous selections when selecting all the text. This is implemented as a post-composition fixup because I coul;dn't figure out where to put it in the main logic."

	| oldLastLine newRectangle line |

	(text size > 1 and: [text last = Character cr]) ifFalse: [^self].

	oldLastLine := lines last.
	oldLastLine last - oldLastLine first >= 0 ifFalse: [^self].
	oldLastLine last = text size ifFalse: [^self].

	newRectangle := oldLastLine left @ oldLastLine bottom 
				extent: 0@(oldLastLine bottom - oldLastLine top).
	"Even though we may be below the bottom of the container,
	it is still necessary to compose the last line for consistency..."

	line := TextLine start: text size+1 stop: text size internalSpaces: 0 paddingWidth: 0.
	line rectangle: newRectangle.
	line lineHeight: lineHeightGuess baseline: textStyle baseline.
	lines := lines, (Array with: line).
! !

!NewParagraph methodsFor: 'access' stamp: 'rr 3/22/2004 12:42'!
focused
	focused ifNil: [focused := false].
	^ focused! !

!NewParagraph methodsFor: 'access' stamp: 'rr 3/22/2004 12:41'!
focused: aBoolean
	focused := aBoolean! !

!NewParagraph methodsFor: 'private' stamp: 'nice 11/17/2009 00:59'!
indentationOfLineIndex: lineIndex ifBlank: aBlock
	"Answer the number of leading tabs in the line at lineIndex.  If there are
	 no visible characters, pass the number of tabs to aBlock and return its value.
	 If the line is word-wrap overflow, back up a line and recur."

	| arrayIndex first last crlf |
	crlf := CharacterSet crlf.
	arrayIndex := lineIndex.
	[first := (lines at: arrayIndex) first.
	 first > 1 and: [crlf includes: (text string at: first - 1)]] whileTrue: "word wrap"
		[arrayIndex := arrayIndex - 1].
	last := (lines at: arrayIndex) last.
	
	^(text string copyFrom: first to: last) indentationIfBlank: aBlock.
! !

!NewParagraph methodsFor: 'initialize-release' stamp: 'tak 12/21/2004 13:29'!
initialize
	self positionWhenComposed: 0 @ 0! !

!NewParagraph methodsFor: 'alignment' stamp: 'di 10/25/97 19:26'!
justified 
	textStyle justified! !

!NewParagraph methodsFor: 'access' stamp: 'di 10/23/97 21:01'!
lastCharacterIndex
	^ lines last last! !

!NewParagraph methodsFor: 'private' stamp: 'ul 1/3/2010 06:50'!
lastLine

	^lines last! !

!NewParagraph methodsFor: 'alignment' stamp: 'di 10/25/97 19:26'!
leftFlush 
	textStyle leftFlush! !

!NewParagraph methodsFor: 'private' stamp: 'nice 2/8/2010 14:26'!
lineIndexForCharacter: characterIndex
	"Deprecated"
	
	^self lineIndexOfCharacterIndex: characterIndex ! !

!NewParagraph methodsFor: 'private' stamp: 'di 4/14/98 13:13'!
lineIndexForPoint: aPoint
	"Answer the index of the line in which to select the character nearest to aPoint."
	| i py |
	py := aPoint y truncated.

	"Find the first line at this y-value"
	i := (self fastFindFirstLineSuchThat: [:line | line bottom > py]) min: lines size.

	"Now find the first line at this x-value"
	[i < lines size and: [(lines at: i+1) top = (lines at: i) top
				and: [aPoint x >= (lines at: i+1) left]]]
		whileTrue: [i := i + 1].
	^ i! !

!NewParagraph methodsFor: 'private' stamp: 'nice 2/8/2010 14:26'!
lineIndexOfCharacterIndex: index
	"Answer the index of the line in which to select the character at index."
	^ (self fastFindFirstLineSuchThat: [:line | line first > index]) - 1 max: 1! !

!NewParagraph methodsFor: 'private' stamp: 'di 10/24/97 17:40'!
lines
	^ lines! !

!NewParagraph methodsFor: 'private' stamp: 'edc 6/18/2004 09:10'!
moveBy: delta
	lines do: [:line | line moveBy: delta].
	positionWhenComposed ifNotNil:[
	positionWhenComposed := positionWhenComposed + delta].
	container := container translateBy: delta! !

!NewParagraph methodsFor: 'access' stamp: 'sbw 10/13/1999 22:31'!
numberOfLines

	^lines size! !

!NewParagraph methodsFor: 'private' stamp: 'di 10/21/97 21:36'!
positionWhenComposed: pos
	positionWhenComposed := pos! !

!NewParagraph methodsFor: 'composition' stamp: 'tpr 9/25/2013 13:59'!
recomposeFrom: start to: stop delta: delta
	"Recompose this paragraph.  The altered portion is between start and stop.
	Recomposition may continue to the end of the text, due to a ripple effect.
	Delta is the amount by which the current text is longer than it was
	when its current lines were composed."
	| startLine newLines |
	"Have to recompose line above in case a word-break was affected."
	startLine := (self lineIndexOfCharacterIndex: start) - 1 max: 1.
	[startLine > 1 and: [(lines at: startLine-1) top = (lines at: startLine) top]]
		whileTrue: [startLine := startLine - 1].  "Find leftmost of line pieces"
	newLines := OrderedCollection new: lines size + 1.
	1 to: startLine-1 do: [:i | newLines addLast: (lines at: i)].
	self composeLinesFrom: (lines at: startLine) first to: stop delta: delta
			into: newLines priorLines: lines
			atY: (lines at: startLine) top! !

!NewParagraph methodsFor: 'editing' stamp: 'jmv 7/29/2009 22:19'!
replaceFrom: start to: stop with: aText
	"Edit the text, and then recompose the lines." 
	text replaceFrom: start to: stop with: aText.
	self recomposeFrom: start to: start + aText size - 1 delta: aText size - (stop-start+1)! !

!NewParagraph methodsFor: 'editing' stamp: 'di 4/28/1999 10:14'!
replaceFrom: start to: stop with: aText displaying: displayBoolean 
	"Edit the text, and then recompose the lines." 
	text replaceFrom: start to: stop with: aText.
	self recomposeFrom: start to: start + aText size - 1 delta: aText size - (stop-start+1)! !

!NewParagraph methodsFor: 'alignment' stamp: 'di 10/25/97 19:26'!
rightFlush 
	textStyle rightFlush! !

!NewParagraph methodsFor: 'access' stamp: 'mt 7/24/2016 14:19'!
selectionColor
	^ selectionColor ifNil: [Color blue muchLighter]! !

!NewParagraph methodsFor: 'access' stamp: 'mt 7/24/2016 13:31'!
selectionColor: aColor
	selectionColor := aColor.! !

!NewParagraph methodsFor: 'selection' stamp: 'di 11/30/97 12:10'!
selectionRects
	"Return an array of rectangles representing the selection region."
	selectionStart ifNil: [^ Array new].
	^ self selectionRectsFrom: selectionStart to: selectionStop! !

!NewParagraph methodsFor: 'selection' stamp: 'nice 2/8/2010 14:24'!
selectionRectsFrom: characterBlock1 to: characterBlock2 
	"Return an array of rectangles representing the area between the two character blocks given as arguments."
	| line1 line2 rects cb1 cb2 w |
	characterBlock1 <= characterBlock2
		ifTrue: [cb1 := characterBlock1.  cb2 := characterBlock2]
		ifFalse: [cb2 := characterBlock1.  cb1 := characterBlock2].
	cb1 = cb2 ifTrue:
		[w := self caretWidth.
		^ Array with: (cb1 topLeft - (w@0) corner: cb1 bottomLeft + ((w+1)@0))].
	line1 := self lineIndexOfCharacterIndex: cb1 stringIndex.
	line2 := self lineIndexOfCharacterIndex: cb2 stringIndex.
	line1 = line2 ifTrue:
		[^ Array with: (cb1 topLeft corner: cb2 bottomRight)].
	rects := OrderedCollection new.
	rects addLast: (cb1 topLeft corner: (lines at: line1) bottomRight).
	line1+1 to: line2-1 do: [ :i |
		| line |
		line := lines at: i.
		(line left = rects last left and: [ line right = rects last right ])
			ifTrue: [ "new line has same margins as old one -- merge them, so that the caller gets as few rectangles as possible"
					| lastRect |
					lastRect := rects removeLast.
					rects add: (lastRect bottom: line bottom) ]
			ifFalse: [ "differing margins; cannot merge"
					rects add: line rectangle ] ].

	rects addLast: ((lines at: line2) topLeft corner: cb2 bottomLeft).
	^ rects! !

!NewParagraph methodsFor: 'selection' stamp: 'di 12/2/97 19:57'!
selectionStart: startBlock selectionStop: stopBlock
	selectionStart := startBlock.
	selectionStop := stopBlock.! !

!NewParagraph methodsFor: 'access' stamp: 'ar 12/15/2008 22:42'!
showCaret
	^showCaret ifNil:[true]
! !

!NewParagraph methodsFor: 'access' stamp: 'ar 12/15/2008 22:42'!
showCaret: aBool
	showCaret := aBool
! !

!NewParagraph methodsFor: 'access' stamp: 'sw 1/13/98 21:31'!
string
	^ text string! !

!NewParagraph methodsFor: 'access' stamp: 'di 10/21/97 14:39'!
text
	^ text! !

!NewParagraph methodsFor: 'access' stamp: 'jm 11/19/97 20:27'!
textOwner: ignored  "See TextOnCurve"! !

!NewParagraph methodsFor: 'access' stamp: 'di 10/21/97 14:39'!
textStyle
	^ textStyle! !

!NewParagraph methodsFor: 'access' stamp: 'di 10/23/97 19:33'!
textStyle: aTextStyle 
	"Set the style by which the receiver should display its text."
	textStyle := aTextStyle! !

!NewParagraph methodsFor: 'private' stamp: 'di 5/21/1998 21:47'!
textStyle: ts lines: l text: t
	"Private -- just a service for deepCopy"
	textStyle := ts.
	lines := l.
	text := t.! !

!NewParagraph methodsFor: 'access' stamp: 'mt 7/24/2016 13:30'!
unfocusedSelectionColor
	^ unfocusedSelectionColor ifNil: [Color gray: 0.9]! !

!NewParagraph methodsFor: 'access' stamp: 'mt 7/24/2016 13:31'!
unfocusedSelectionColor: aColor
	unfocusedSelectionColor := aColor.! !

!NewParagraph methodsFor: 'access' stamp: 'RAA 5/6/2001 15:04'!
wantsColumnBreaks

	^wantsColumnBreaks! !

!NewParagraph methodsFor: 'access' stamp: 'RAA 5/6/2001 15:03'!
wantsColumnBreaks: aBoolean

	wantsColumnBreaks := aBoolean! !

Object subclass: #NullEncoder
	instanceVariableNames: 'target filterSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

NullEncoder subclass: #FlattenEncoder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!FlattenEncoder commentStamp: '<historical>' prior: 0!
The simplest possible encoding:  leave the objects as is.
!

FlattenEncoder subclass: #Canvas
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!Canvas commentStamp: '<historical>' prior: 0!
A canvas is a two-dimensional medium on which morphs are drawn in a device-independent manner. Canvases keep track of the origin and clipping rectangle, as well as the underlying drawing medium (such as a window, pixmap, or postscript script).

Subclasses must implement (at least) the following methods:
	* Drawing:
		#fillOval:color:borderWidth:borderColor:
		#frameAndFillRectangle:fillColor:borderWidth:borderColor:
		#drawPolygon:color:borderWidth:borderColor:
		#image:at:sourceRect:rule:
		#stencil:at:sourceRect:rule:
		#line:to:width:color:
		#paragraph:bounds:color:
		#text:bounds:font:color:
	* Support
		#clipBy:during:
		#translateBy:during:
		#translateBy:clippingTo:during:
		#transformBy:clippingTo:during:
!

!Canvas class methodsFor: 'configuring' stamp: ''!
filterSelector
	^#drawOnCanvas:.! !

!Canvas methodsFor: 'converting' stamp: 'ar 8/8/2001 14:22'!
asAlphaBlendingCanvas: alpha
	^(AlphaBlendingCanvas on: self) alpha: alpha! !

!Canvas methodsFor: 'converting' stamp: 'ar 6/24/1999 17:46'!
asShadowDrawingCanvas
	^self asShadowDrawingCanvas: (Color black alpha: 0.5).! !

!Canvas methodsFor: 'converting' stamp: 'ar 8/8/2001 14:14'!
asShadowDrawingCanvas: aColor
	^(ShadowDrawingCanvas on: self) shadowColor: aColor! !

!Canvas methodsFor: 'drawing-support' stamp: 'gm 2/22/2003 14:53'!
cache: aRectangle using: aCache during: aBlock 
	"Cache the execution of aBlock by the given cache.
	Note: At some point we may want to actually *create* the cache here;
		for now we're only using it."

	(aCache notNil 
		and: [(aCache isForm) and: [aCache extent = aRectangle extent]]) 
			ifTrue: [^self paintImage: aCache at: aRectangle origin].
	aBlock value: self! !

!Canvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 02:53'!
clipBy: aRectangle during: aBlock
	"Set a clipping rectangle active only during the execution of aBlock.
	Note: In the future we may want to have more general clip shapes - not just rectangles"
	^self subclassResponsibility! !

!Canvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 14:07'!
clipRect
	"Return the currently active clipping rectangle"
	^self subclassResponsibility! !

!Canvas methodsFor: 'accessing' stamp: 'ar 2/12/2000 18:17'!
contentsOfArea: aRectangle
	"Return the contents of the given area"
	^self contentsOfArea: aRectangle into: (Form extent: aRectangle extent depth: self depth)! !

!Canvas methodsFor: 'accessing' stamp: 'ar 2/12/2000 18:17'!
contentsOfArea: aRectangle into: aForm
	"Return the contents of the given area"
	^self subclassResponsibility! !

!Canvas methodsFor: 'copying' stamp: 'jm 8/2/97 13:54'!
copy

	^ self clone
! !

!Canvas methodsFor: 'copying' stamp: 'ls 3/20/2000 21:24'!
copyClipRect: newClipRect

	^ ClippingCanvas canvas: self clipRect: newClipRect
! !

!Canvas methodsFor: 'accessing' stamp: ''!
depth

	^ Display depth
! !

!Canvas methodsFor: 'Nebraska/embeddedWorlds' stamp: 'RAA 11/7/2000 13:54'!
displayIsFullyUpdated! !

!Canvas methodsFor: 'testing' stamp: 'di 8/12/2000 15:04'!
doesRoundedCorners 

	^ true! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 2/16/2000 23:45'!
drawImage: aForm at: aPoint
	"Draw the given Form, which is assumed to be a Form or ColorForm"

	self drawImage: aForm
		at: aPoint
		sourceRect: aForm boundingBox! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 2/17/2000 01:47'!
drawImage: aForm at: aPoint sourceRect: sourceRect
	"Draw the given form."
	self shadowColor ifNotNil:[
		^self fillRectangle: ((aForm boundingBox intersect: sourceRect) translateBy: aPoint)
				color: self shadowColor].
	^self image: aForm
		at: aPoint
		sourceRect: sourceRect
		rule: Form over! !

!Canvas methodsFor: 'drawing-general' stamp: 'ar 12/30/2001 15:23'!
drawMorph: aMorph
	self draw: aMorph! !

!Canvas methodsFor: 'drawing-polygons' stamp: 'ar 6/18/1999 08:56'!
drawPolygon: vertices color: aColor borderWidth: bw borderColor: bc
	"Draw the given polygon."
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-polygons' stamp: 'ar 6/25/1999 12:18'!
drawPolygon: vertices fillStyle: aFillStyle
	"Fill the given polygon."
	self drawPolygon: vertices fillStyle: aFillStyle borderWidth: 0 borderColor: Color transparent! !

!Canvas methodsFor: 'drawing-polygons' stamp: 'ar 6/18/1999 08:58'!
drawPolygon: vertices fillStyle: aFillStyle borderWidth: bw borderColor: bc
	"Fill the given polygon.
	Note: The default implementation does not recognize any enhanced fill styles"
	self drawPolygon: vertices color: aFillStyle asColor borderWidth: bw borderColor: bc! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/30/2001 20:35'!
drawString: s at: pt

	^ self drawString: s from: 1 to: s size at: pt font: nil color: Color black! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:25'!
drawString: s at: pt font: aFont color: aColor

	^ self drawString: s from: 1 to: s size at: pt font: aFont color: aColor! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/30/2001 20:36'!
drawString: s from: firstIndex to: lastIndex at: pt font: font color: aColor
	self drawString: s from: firstIndex to: lastIndex in: (pt extent: 10000@10000) font: font color: aColor! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/30/2001 20:37'!
drawString: s from: firstIndex to: lastIndex in: boundsRect font: fontOrNil color: c
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:39'!
drawString: s in: boundsRect
	^self drawString: s from: 1 to: s size in: boundsRect font: nil color: Color black! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:13'!
drawString: s in: boundsRect font: fontOrNil color: c
	^self drawString: s from: 1 to: s size in: boundsRect font: fontOrNil color: c! !

!Canvas methodsFor: 'drawing-general' stamp: 'ar 5/29/1999 05:14'!
draw: anObject
	^anObject drawOn: self! !

!Canvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 14:15'!
extent
	"Return the physical extent of the output device"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing' stamp: 'ar 6/17/1999 01:18'!
fillColor: aColor
	"Fill the receiver with the given color.
	Note: This method should be named differently since it is intended to fill the background and thus fills even if the color is transparent"
	^self fillRectangle: self clipRect color: (aColor alpha: 1.0).! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:45'!
fillOval: r color: c

	self fillOval: r color: c borderWidth: 0 borderColor: Color transparent.
! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:45'!
fillOval: r color: c borderWidth: borderWidth borderColor: borderColor
	"Fill the given oval."
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:51'!
fillOval: aRectangle fillStyle: aFillStyle
	"Fill the given oval."
	^self fillOval: aRectangle fillStyle: aFillStyle borderWidth: 0 borderColor: Color transparent! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:50'!
fillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc
	"Fill the given oval.
	Note: The default implementation does not recognize any enhanced fill styles"
	self fillOval: aRectangle color: aFillStyle asColor borderWidth: bw borderColor: bc! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/18/1999 07:32'!
fillRectangle: r color: c
	"Fill the rectangle using the given color"
	^self 
		frameAndFillRectangle: r
		fillColor: c
		borderWidth: 0
		borderColor: Color transparent! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/18/1999 07:34'!
fillRectangle: aRectangle fillStyle: aFillStyle
	"Fill the given rectangle.
	Note: The default implementation does not recognize any enhanced fill styles"
	self fillRectangle: aRectangle color: aFillStyle asColor.! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'DS 4/22/2015 11:13:57.415'!
fillRectangle: aRectangle fillStyle: aFillStyle borderStyle: aBorderStyle
	"Fill the given rectangle."
	aFillStyle isTransparent ifFalse:[
		self fillRectangle: (aRectangle insetBy: aBorderStyle inset) fillStyle: aFillStyle].
	aBorderStyle ifNil:[^self].
	aBorderStyle width <= 0 ifTrue:[^self].
	aBorderStyle frameRectangle: aRectangle on: self.! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'mt 3/6/2015 08:55'!
fillRoundRect: aRectangle radius: radius fillStyle: fillStyle

	self
		fillRectangle: aRectangle
		fillStyle: fillStyle.! !

!Canvas methodsFor: 'initialization' stamp: 'ar 5/27/2000 21:50'!
finish
	"If there are any pending operations on the receiver complete them. Do not return before all modifications have taken effect."
	^self flush! !

!Canvas methodsFor: 'initialization' stamp: 'ar 6/17/2010 16:43'!
finish: allDamage
	"If there are any pending operations on the receiver complete them. 
	Do not return before all modifications have taken effect."
	^self finish! !

!Canvas methodsFor: 'initialization' stamp: 'ar 2/9/1999 06:29'!
flush! !

!Canvas methodsFor: 'other' stamp: ''!
flushDisplay
		" Dummy ."! !

!Canvas methodsFor: 'other' stamp: ''!
forceToScreen:rect
	" dummy "
! !

!Canvas methodsFor: 'accessing' stamp: 'jm 6/2/1998 06:39'!
form

	^ Display
! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/18/1999 07:32'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth borderColor: borderColor
	"Draw the rectangle using the given attributes"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'RAA 8/14/2000 14:22'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth topLeftColor: topLeftColor bottomRightColor: bottomRightColor
	"Draw the rectangle using the given attributes.
	Note: This is a *very* simple implementation"
	| bw pt |
	self frameAndFillRectangle: r
		fillColor: fillColor
		borderWidth: borderWidth
		borderColor: bottomRightColor.
	bottomRightColor = topLeftColor ifFalse: [
		bw := borderWidth asPoint.
		pt := r topLeft + (bw // 2).
		self line: pt to: pt + ((r extent x - bw x)@0) width: borderWidth color: topLeftColor.
		self line: pt to: pt + (0@(r extent y - bw y)) width: borderWidth color: topLeftColor.
	].! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'mt 3/6/2015 08:58'!
frameAndFillRoundRect: aRectangle radius: cornerRadius fillStyle: fillStyle borderWidth: bw borderColor: bc

	self
		frameAndFillRectangle: aRectangle
		fillColor: fillStyle asColor
		borderWidth: bw
		borderColor: bc.! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:45'!
frameOval: r color: c

	self fillOval: r color: Color transparent borderWidth: 1 borderColor: c.
! !

!Canvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:45'!
frameOval: r width: w color: c

	self fillOval: r color: Color transparent borderWidth: w borderColor: c.
! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/18/1999 07:33'!
frameRectangle: r color: c

	self frameRectangle: r width: 1 color: c.
! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/18/1999 07:33'!
frameRectangle: r width: w color: c
	^self frameAndFillRectangle: r fillColor: Color transparent borderWidth: w borderColor: c.! !

!Canvas methodsFor: 'drawing-rectangles' stamp: 'mt 3/6/2015 08:59'!
frameRoundRect: aRectangle radius: radius width: borderWidth color: borderColor

	self
		frameRectangle: aRectangle
		width: borderWidth
		color: borderColor.! !

!Canvas methodsFor: 'drawing-general' stamp: 'ar 12/30/2001 15:23'!
fullDrawMorph: aMorph
	self fullDraw: aMorph! !

!Canvas methodsFor: 'drawing-general' stamp: ''!
fullDraw: anObject
	^anObject fullDrawOn: self! !

!Canvas methodsFor: 'drawing-obsolete' stamp: 'ar 2/12/2000 18:11'!
imageWithOpaqueWhite: aForm at: aPoint
	"Note: This protocol is deprecated. Use #drawImage: instead"
	self image: aForm
		at: aPoint
		sourceRect: (0@0 extent: aForm extent)
		rule: Form over.
! !

!Canvas methodsFor: 'drawing-obsolete' stamp: 'ar 2/12/2000 18:10'!
image: aForm at: aPoint
	"Note: This protocol is deprecated. Use #paintImage: instead."
	self image: aForm
		at: aPoint
		sourceRect: aForm boundingBox
		rule: Form paint.
! !

!Canvas methodsFor: 'drawing-obsolete' stamp: 'ar 2/12/2000 18:11'!
image: aForm at: aPoint rule: combinationRule
	"Note: This protocol is deprecated. Use one of the explicit image drawing messages (#paintImage, #drawImage) instead."
	self image: aForm
		at: aPoint
		sourceRect: aForm boundingBox
		rule: combinationRule.
! !

!Canvas methodsFor: 'private' stamp: 'ar 2/12/2000 18:12'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	"Note: The public use of this protocol is deprecated. It will become private. Nobody in the outside world must assume that a thing like a combination rule has any specific effect."
	^self subclassResponsibility! !

!Canvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:21'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule alpha: sourceAlpha
	"Privately used for blending forms w/ constant alpha. Fall back to simpler case by defaul."
	^self image: aForm at: aPoint sourceRect: sourceRect rule: rule! !

!Canvas methodsFor: 'testing' stamp: 'ar 11/13/1998 13:19'!
isBalloonCanvas
	^false! !

!Canvas methodsFor: 'testing' stamp: 'nk 1/1/2004 21:09'!
isPostscriptCanvas
	^false! !

!Canvas methodsFor: 'testing' stamp: 'ar 6/22/1999 19:03'!
isShadowDrawing
	^false! !

!Canvas methodsFor: 'testing' stamp: 'ar 6/22/1999 14:10'!
isVisible: aRectangle
	"Return true if the given rectangle is (partially) visible"
	^self clipRect intersects: aRectangle
! !

!Canvas methodsFor: 'drawing' stamp: 'ar 6/17/1999 01:30'!
line: pt1 to: pt2 brushForm: brush
	"Obsolete - will be removed in the future"! !

!Canvas methodsFor: 'drawing' stamp: 'jm 8/2/97 13:54'!
line: pt1 to: pt2 color: c

	self line: pt1 to: pt2 width: 1 color: c.
! !

!Canvas methodsFor: 'drawing' stamp: 'ar 6/17/1999 01:31'!
line: pt1 to: pt2 width: w color: c
	"Draw a line using the given width and color"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing' stamp: 'aoy 2/15/2003 21:41'!
line: pt1 to: pt2 width: width color: color1 dashLength: s1 secondColor: color2 secondDashLength: s2 startingOffset: startingOffset 
	"Draw a line using the given width, colors and dash lengths.
	Originally written by Stephan Rudlof; tweaked by Dan Ingalls
	to use startingOffset for sliding offset as in 'ants' animations.
	Returns the sum of the starting offset and the length of this line."

	| dist deltaBig colors nextPhase segmentOffset phase segmentLength startPoint distDone endPoint segLens |
	dist := pt1 dist: pt2.
	dist = 0 ifTrue: [^startingOffset].
	s1 = 0 & (s2 = 0) ifTrue: [^startingOffset].
	deltaBig := pt2 - pt1.
	colors := { 
				color1.
				color2}.
	segLens := { 
				s1 asFloat.
				s2 asFloat}.
	nextPhase := { 
				2.
				1}.

	"Figure out what phase we are in and how far, given startingOffset."
	segmentOffset := startingOffset \\ (s1 + s2).
	segmentLength := segmentOffset < s1 
		ifTrue: 
			[phase := 1.
			s1 - segmentOffset]
		ifFalse: 
			[phase := 2.
			 s1 + s2 - segmentOffset].
	startPoint := pt1.
	distDone := 0.0.
	[distDone < dist] whileTrue: 
			[segmentLength := segmentLength min: dist - distDone.
			endPoint := startPoint + (deltaBig * segmentLength / dist).
			self 
				line: startPoint truncated
				to: endPoint truncated
				width: width
				color: (colors at: phase).
			distDone := distDone + segmentLength.
			phase := nextPhase at: phase.
			startPoint := endPoint.
			segmentLength := segLens at: phase].
	^startingOffset + dist! !

!Canvas methodsFor: 'drawing' stamp: 'sr 4/27/2000 03:31'!
line: pt1 to: pt2 width: w1 color: c1 stepWidth: s1 secondWidth: w2 secondColor: c2 secondStepWidth: s2 
	"Draw a line using the given width, colors and steps; both steps can  
	have different stepWidths (firstStep, secondStep), draw widths and  
	colors."
	| bigSteps offsetPoint dist p1p2Vec deltaBig delta1 delta2 lastPoint bigStep |
	s1 = 0 & (s2 = 0) ifTrue: [^ self].
	dist := pt1 dist: pt2.
	dist = 0 ifTrue: [^ self].
	bigStep := s1 + s2.
	bigSteps := dist / bigStep.
	p1p2Vec := pt2 - pt1.
	deltaBig := p1p2Vec / bigSteps.
	delta1 := deltaBig * (s1 / bigStep).
	delta2 := deltaBig * (s2 / bigStep).
	dist <= s1
		ifTrue: 
			[self
				line: pt1 rounded
				to: pt2 rounded
				width: w1
				color: c1.
			^ self].
	0 to: bigSteps truncated - 1 do: 
		[:bigStepIx | 
		self
			line: (pt1 + (offsetPoint := deltaBig * bigStepIx)) rounded
			to: (pt1 + (offsetPoint := offsetPoint + delta1)) rounded
			width: w1
			color: c1.
		self
			line: (pt1 + offsetPoint) rounded
			to: (pt1 + (offsetPoint + delta2)) rounded
			width: w2
			color: c2].
	"if there was no loop, offsetPoint is nil"
	lastPoint := pt1 + ((offsetPoint ifNil: [0 @ 0])
					+ delta2).
	(lastPoint dist: pt2)
		<= s1
		ifTrue: [self
				line: lastPoint rounded
				to: pt2 rounded
				width: w1
				color: c1]
		ifFalse: 
			[self
				line: lastPoint rounded
				to: (lastPoint + delta1) rounded
				width: w1
				color: c1.
			self
				line: (lastPoint + delta1) rounded
				to: pt2
				width: w1
				color: c2]! !

!Canvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 14:11'!
origin
	"Return the current origin for drawing operations"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 2/16/2000 23:48'!
paintImage: aForm at: aPoint
	"Draw the given Form, which is assumed to be a Form or ColorForm following the convention that zero is the transparent pixel value."

	self paintImage: aForm
		at: aPoint
		sourceRect: aForm boundingBox
! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 2/17/2000 01:48'!
paintImage: aForm at: aPoint sourceRect: sourceRect
	"Draw the given Form, which is assumed to be a Form or ColorForm following the convention that zero is the transparent pixel value."
	self shadowColor ifNotNil:[
		^self stencil: aForm at: aPoint sourceRect: sourceRect color: self shadowColor].
	^self image: aForm
		at: aPoint
		sourceRect: sourceRect
		rule: Form paint! !

!Canvas methodsFor: 'drawing' stamp: 'ar 6/17/1999 01:31'!
paragraph: paragraph bounds: bounds color: c
	"Draw the given paragraph"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing' stamp: 'ar 6/17/1999 01:32'!
point: p color: c
	"Obsolete - will be removed in the future"! !

!Canvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 01:43'!
preserveStateDuring: aBlock
	"Preserve the full canvas state during the execution of aBlock"
	^aBlock value: self copy! !

!Canvas methodsFor: 'drawing' stamp: 'ar 2/5/1999 18:28'!
render: anObject
	"Do some 3D operations with the object if possible"! !

!Canvas methodsFor: 'initialization' stamp: 'di 9/22/1999 19:21'!
reset
	"Reset the canvas."

	super initWithTarget:self class defaultTarget.
! !

!Canvas methodsFor: 'testing' stamp: 'di 9/24/2000 16:10'!
seesNothingOutside: aRectangle
	"Return true if this canvas will not touch anything outside aRectangle"
	^ aRectangle containsRect: self clipRect
! !

!Canvas methodsFor: 'accessing' stamp: 'ar 2/17/2000 01:46'!
shadowColor
	"Return the current override color or nil if no such color exists"
	^nil! !

!Canvas methodsFor: 'accessing' stamp: 'ar 2/17/2000 01:46'!
shadowColor: aColor
	"Set a shadow color. If set this color overrides any client-supplied color."! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 6/25/1999 12:17'!
stencil: stencilForm at: aPoint color: aColor
	"Flood this canvas with aColor wherever stencilForm has non-zero pixels"
	^self stencil: stencilForm
		at: aPoint
		sourceRect: stencilForm boundingBox
		color: aColor! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 6/25/1999 12:17'!
stencil: stencilForm at: aPoint sourceRect: sourceRect color: aColor
	"Flood this canvas with aColor wherever stencilForm has non-zero pixels"
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:40'!
text: s at: pt font: fontOrNil color: c
	"OBSOLETE"
	^ self drawString: s at: pt font: fontOrNil color: c! !

!Canvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:40'!
text: s bounds: boundsRect font: fontOrNil color: c
	"OBSOLETE"
	^self drawString: s in: boundsRect font: fontOrNil color: c! !

!Canvas methodsFor: 'Nebraska/embeddedWorlds' stamp: 'ul 8/2/2011 19:57'!
transform2By: aDisplayTransform clippingTo: aClipRect during: aBlock smoothing: cellSize

	"an attempt to use #displayInterpolatedOn: instead of WarpBlt."

	| patchRect subCanvas pureRect biggerPatch biggerClip interForm |

	self flag: #bob.		"added to Canvas in hopes it will work for Nebraska"
	(aDisplayTransform isPureTranslation) ifTrue: [
		^aBlock value: (self copyOffset: aDisplayTransform offset negated truncated
							clipRect: aClipRect)
	].
	"Prepare an appropriate warp from patch to aClipRect"
	pureRect := (aDisplayTransform globalBoundsToLocal: aClipRect).
	patchRect := pureRect rounded.
	patchRect area = 0 ifTrue: [^self]. 	"oh, well!!"
	biggerPatch := patchRect expandBy: 1.
	biggerClip := (aDisplayTransform localBoundsToGlobal: biggerPatch) rounded.

	"Render the submorphs visible in the clipping rectangle, as patchForm"
	subCanvas := FormCanvas extent: biggerPatch extent depth: self depth.
	self isShadowDrawing ifTrue: [
		subCanvas shadowColor: self shadowColor
	].

	"this biggerPatch/biggerClip is an attempt to improve positioning of the final image in high magnification conditions. Since we cannot grab fractional pixels from the source, take one extra and then take just the part we need from the expanded form"

	subCanvas 
		translateBy: biggerPatch topLeft negated rounded
		during: aBlock.
	interForm := Form extent: biggerClip extent depth: self depth.
	subCanvas form 
		displayInterpolatedIn: interForm boundingBox
		on: interForm.
	self 
		drawImage: interForm 
		at: aClipRect origin 
		sourceRect: (aClipRect origin - biggerClip origin extent: aClipRect extent)

! !

!Canvas methodsFor: 'drawing-support' stamp: 'di 10/16/1999 16:02'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock
	"Transform the receiver by the given display transformation during the execution of aBlock. The given clip rectangle defines the *global* (e.g., outer) rectangle against which the receiver should clip (which would be equivalent to 'self clipRect: aClipRect; transformBy: aDisplayTransform')."
	^ self transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock smoothing: 1
! !

!Canvas methodsFor: 'drawing-support' stamp: 'di 10/16/1999 15:56'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock smoothing: cellSize
	"Transform the receiver by the given display transformation during the execution of aBlock. The given clip rectangle defines the *global* (e.g., outer) rectangle against which the receiver should clip (which would be equivalent to 'self clipRect: aClipRect; transformBy: aDisplayTransform')."
	^ self subclassResponsibility! !

!Canvas methodsFor: 'other' stamp: ''!
translateBy:aPoint clippingTo:aRect during:aBlock
	^aBlock value:(self copyOffset:aPoint clipRect:aRect).! !

!Canvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 03:00'!
translateBy: delta during: aBlock
	"Set a translation only during the execution of aBlock."
	^self subclassResponsibility! !

!Canvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 14:08'!
translateTo: newOrigin clippingTo: aRectangle during: aBlock
	"Set a new origin and clipping rectangle only during the execution of aBlock."
	self translateBy: newOrigin - self origin 
		clippingTo: (aRectangle translateBy: self origin negated) 
		during: aBlock! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 2/17/2000 14:05'!
translucentImage: aForm at: aPoint
	"Draw a translucent image using the best available way of representing translucency."
	self translucentImage: aForm
		at: aPoint
		sourceRect: aForm boundingBox! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 2/10/2004 17:19'!
translucentImage: aForm at: aPoint sourceRect: sourceRect
	"Draw a translucent image using the best available way of representing translucency.
	Note: This will be fixed in the future."
	self shadowColor ifNotNil:[
		^self stencil: aForm at: aPoint sourceRect: sourceRect color: self shadowColor].
	(self depth < 32 or:[aForm isTranslucent not]) 
		ifTrue:[^self paintImage: aForm at: aPoint sourceRect: sourceRect].
	self image: aForm
		at: aPoint
		sourceRect: sourceRect
		rule: Form blend! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 12/28/2001 23:44'!
warpImage: aForm transform: aTransform
	"Warp the given form using aTransform"
	^self warpImage: aForm transform: aTransform at: 0@0! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 12/28/2001 23:54'!
warpImage: aForm transform: aTransform at: extraOffset
	"Warp the given form using aTransform.
	TODO: Use transform to figure out appropriate cell size"
	^self warpImage: aForm transform: aTransform at: extraOffset sourceRect: aForm boundingBox cellSize: 1! !

!Canvas methodsFor: 'drawing-images' stamp: 'ar 12/29/2001 00:20'!
warpImage: aForm transform: aTransform at: extraOffset sourceRect: sourceRect cellSize: cellSize
	"Warp the given using the appropriate transform and offset."
	^self subclassResponsibility! !

Canvas subclass: #ColorMappingCanvas
	instanceVariableNames: 'myCanvas'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

ColorMappingCanvas subclass: #AlphaBlendingCanvas
	instanceVariableNames: 'alpha'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!AlphaBlendingCanvas methodsFor: 'accessing' stamp: 'ar 8/8/2001 14:24'!
alpha
	^alpha! !

!AlphaBlendingCanvas methodsFor: 'accessing' stamp: 'ar 8/8/2001 14:24'!
alpha: newAlpha
	alpha := newAlpha.! !

!AlphaBlendingCanvas methodsFor: 'private' stamp: 'bf 10/28/2003 15:46'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	"Draw the given form. For the 'paint' combination rule use stenciling otherwise simply fill the source rectangle."
	rule = Form paint ifTrue:[
		^myCanvas
			image: aForm
			at: aPoint
			sourceRect: sourceRect
			rule: Form paintAlpha
			alpha: alpha.
	].
	rule = Form over ifTrue:[
		^myCanvas
			image: aForm
			at: aPoint
			sourceRect: sourceRect
			rule: Form blendAlpha
			alpha: alpha.
	].! !

!AlphaBlendingCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:23'!
mapColor: aColor
	aColor isColor ifFalse:[^aColor]. "Should not happen but who knows..."
	aColor isTransparent ifTrue:[^aColor].
	aColor isOpaque ifTrue:[^aColor alpha: alpha].
	^aColor alpha: (aColor alpha * alpha)! !

!AlphaBlendingCanvas methodsFor: 'private' stamp: 'topa 3/6/2015 10:33'!
mapFillStyle: aFillStyle

	^ self alpha = 1.0
		ifTrue: [aFillStyle]
		ifFalse: [super mapFillStyle: aFillStyle]
! !

!AlphaBlendingCanvas methodsFor: 'initialization' stamp: 'ar 8/8/2001 14:18'!
on: aCanvas
	myCanvas := aCanvas.
	alpha := 1.0.! !

!ColorMappingCanvas class methodsFor: 'instance creation' stamp: 'ar 6/22/1999 18:23'!
on: aCanvas
	^self new on: aCanvas! !

!ColorMappingCanvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 18:19'!
clipBy: aRectangle during: aBlock
	"Set a clipping rectangle active only during the execution of aBlock.
	Note: In the future we may want to have more general clip shapes - not just rectangles"
	| oldCanvas |
	oldCanvas := myCanvas.
	myCanvas clipBy: aRectangle during:[:newCanvas|
		myCanvas := newCanvas.
		aBlock value: self].
	myCanvas := oldCanvas! !

!ColorMappingCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 17:40'!
clipRect
	^myCanvas clipRect! !

!ColorMappingCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 17:39'!
depth
	^myCanvas depth! !

!ColorMappingCanvas methodsFor: 'drawing-polygons' stamp: 'mir 9/12/2001 14:24'!
drawPolygon: vertices color: aColor borderWidth: bw borderColor: bc
	"Draw the given polygon."
	^myCanvas
		drawPolygon: vertices
		color: aColor
		borderWidth: bw
		borderColor: (self mapColor: bc)! !

!ColorMappingCanvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:28'!
drawString: s from: firstIndex to: lastIndex in: boundsRect font: fontOrNil color: c
	"Draw the given string in the given font and color clipped to the given rectangle. If the font is nil, the default font is used."
	myCanvas
		drawString: s from: firstIndex to: lastIndex 
		in: boundsRect 
		font: fontOrNil 
		color: (self mapColor: c)! !

!ColorMappingCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 17:39'!
extent
	^myCanvas extent! !

!ColorMappingCanvas methodsFor: 'drawing-ovals' stamp: 'ar 6/22/1999 17:59'!
fillOval: r color: c borderWidth: borderWidth borderColor: borderColor
	"Fill the given oval."
	myCanvas
		fillOval: r
		color: (self mapColor: c)
		borderWidth: borderWidth
		borderColor: (self mapColor: borderColor)! !

!ColorMappingCanvas methodsFor: 'initialization' stamp: 'ar 6/22/1999 18:24'!
flush
	myCanvas flush.! !

!ColorMappingCanvas methodsFor: 'accessing' stamp: 'ar 6/24/1999 17:54'!
form
	^myCanvas form! !

!ColorMappingCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/22/1999 17:59'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth borderColor: borderColor
	"Draw the rectangle using the given attributes"
	myCanvas
		frameAndFillRectangle: r
		fillColor: (self mapColor: fillColor)
		borderWidth: borderWidth
		borderColor: (self mapColor: borderColor)! !

!ColorMappingCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/22/1999 18:01'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth topLeftColor: topLeftColor bottomRightColor: bottomRightColor
	"Draw the rectangle using the given attributes"
	myCanvas
		frameAndFillRectangle: r 
		fillColor: (self mapColor: fillColor) 
		borderWidth: borderWidth 
		topLeftColor: (self mapColor: topLeftColor)
		bottomRightColor: (self mapColor: bottomRightColor)! !

!ColorMappingCanvas methodsFor: 'drawing-rectangles' stamp: 'topa 3/6/2015 10:25'!
frameAndFillRoundRect: aRectangle radius: cornerRadius fillStyle: fillStyle borderWidth: bw borderColor: bc
	"Draw the rectangle using the given attributes"
	myCanvas
		frameAndFillRoundRect: aRectangle
		radius: cornerRadius
		fillStyle: (self mapFillStyle: fillStyle)
		borderWidth: bw
		borderColor: (self mapColor: bc)
! !

!ColorMappingCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:15'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	"Draw the given form. For the 'paint' combination rule use stenciling otherwise simply fill the source rectangle."
	^myCanvas image: aForm at: aPoint sourceRect: sourceRect rule: rule.! !

!ColorMappingCanvas methodsFor: 'testing' stamp: 'ar 8/8/2001 14:16'!
isShadowDrawing
	^myCanvas isShadowDrawing! !

!ColorMappingCanvas methodsFor: 'drawing' stamp: 'ar 6/22/1999 18:15'!
line: pt1 to: pt2 width: w color: c
	"Draw a line using the given width and color"
	myCanvas
		line: pt1
		to: pt2
		width: w
		color: (self mapColor: c).! !

!ColorMappingCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:15'!
mapColor: aColor
	^aColor! !

!ColorMappingCanvas methodsFor: 'private' stamp: 'topa 3/6/2015 10:33'!
mapFillStyle: aFillStyle

	^ aFillStyle isGradientFill
		ifFalse: [self mapColor: aFillStyle asColor]
		ifTrue: [aFillStyle copy colorRamp: (aFillStyle colorRamp collect: [:assoc | assoc key -> (self mapColor: assoc value)])].
! !

!ColorMappingCanvas methodsFor: 'initialization' stamp: 'ar 8/8/2001 14:14'!
on: aCanvas
	myCanvas := aCanvas.! !

!ColorMappingCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 17:39'!
origin
	^myCanvas origin! !

!ColorMappingCanvas methodsFor: 'drawing' stamp: 'ar 6/22/1999 18:16'!
paragraph: paragraph bounds: bounds color: c
	"Draw the given paragraph"
	myCanvas
		paragraph: paragraph
		bounds: bounds
		color: (self mapColor: c)! !

!ColorMappingCanvas methodsFor: 'drawing-support' stamp: 'nice 12/26/2009 19:22'!
preserveStateDuring: aBlock
	"Preserve the full canvas state during the execution of aBlock"
	| oldCanvas result |
	oldCanvas := myCanvas.
	result := myCanvas preserveStateDuring:[:newCanvas|
		myCanvas := newCanvas.
		aBlock value: self].
	myCanvas := oldCanvas.
	^result! !

!ColorMappingCanvas methodsFor: 'initialization' stamp: 'ar 6/22/1999 18:23'!
reset
	myCanvas reset.! !

!ColorMappingCanvas methodsFor: 'drawing-images' stamp: 'ar 6/24/1999 18:26'!
stencil: aForm at: aPoint color: aColor
	myCanvas
		stencil: aForm
		at: aPoint
		color: (self mapColor: aColor)! !

!ColorMappingCanvas methodsFor: 'drawing-images' stamp: 'ar 6/24/1999 18:26'!
stencil: aForm at: aPoint sourceRect: aRect color: aColor
	myCanvas
		stencil: aForm
		at: aPoint
		sourceRect: aRect
		color: (self mapColor: aColor)! !

!ColorMappingCanvas methodsFor: 'drawing-support' stamp: 'di 10/16/1999 16:01'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock	 smoothing: cellSize

	"Transform the receiver by the given display transformation during the execution of aBlock. The given clip rectangle defines the *global* (e.g., outer) rectangle against which the receiver should clip (which would be equivalent to 'self clipRect: aClipRect; transformBy: aDisplayTransform')."
	| oldCanvas |
	oldCanvas := myCanvas.
	myCanvas transformBy: aDisplayTransform
		clippingTo: aClipRect
		during: [:newCanvas |
				myCanvas := newCanvas.
				aBlock value: self]
		smoothing: cellSize.
	myCanvas := oldCanvas.! !

!ColorMappingCanvas methodsFor: 'other' stamp: 'ar 6/22/1999 18:21'!
translateBy: delta clippingTo: aRectangle during: aBlock
	"Set a translation and clipping rectangle only during the execution of aBlock."
	| oldCanvas |
	oldCanvas := myCanvas.
	myCanvas translateBy: delta clippingTo: aRectangle during:[:newCanvas|
		myCanvas := newCanvas.
		aBlock value: self].
	myCanvas := oldCanvas.! !

!ColorMappingCanvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 18:22'!
translateBy: delta during: aBlock
	"Set a translation only during the execution of aBlock."
	| oldCanvas |
	oldCanvas := myCanvas.
	myCanvas translateBy: delta during:[:newCanvas|
		myCanvas := newCanvas.
		aBlock value: self].
	myCanvas := oldCanvas.! !

!ColorMappingCanvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 18:22'!
translateTo: newOrigin clippingTo: aRectangle during: aBlock
	"Set a new origin and clipping rectangle only during the execution of aBlock."
	| oldCanvas |
	oldCanvas := myCanvas.
	myCanvas translateTo: newOrigin clippingTo: aRectangle during:[:newCanvas|
		myCanvas := newCanvas.
		aBlock value: self].
	myCanvas := oldCanvas.! !

ColorMappingCanvas subclass: #ShadowDrawingCanvas
	instanceVariableNames: 'shadowColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!ShadowDrawingCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:13'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	"Draw the given form. For the 'paint' combination rule use stenciling otherwise simply fill the source rectangle."
	rule = Form paint ifTrue:[
		^myCanvas
			stencil: aForm
			at: aPoint
			sourceRect: sourceRect
			color: shadowColor
	] ifFalse:[
		^myCanvas
			fillRectangle: (sourceRect translateBy: aPoint)
			color: shadowColor
	].! !

!ShadowDrawingCanvas methodsFor: 'testing' stamp: 'ar 8/8/2001 14:16'!
isShadowDrawing
	^true! !

!ShadowDrawingCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:14'!
mapColor: aColor
	aColor isColor ifFalse:[^aColor]. "Should not happen but who knows..."
	^aColor isTransparent
		ifTrue:[aColor]
		ifFalse:[shadowColor]! !

!ShadowDrawingCanvas methodsFor: 'initialization' stamp: 'ar 8/8/2001 14:14'!
on: aCanvas
	myCanvas := aCanvas.
	shadowColor := Color black.! !

!ShadowDrawingCanvas methodsFor: 'accessing' stamp: 'ar 8/8/2001 14:14'!
shadowColor
	^shadowColor! !

!ShadowDrawingCanvas methodsFor: 'accessing' stamp: 'ar 8/8/2001 14:14'!
shadowColor: aColor
	shadowColor := aColor! !

Canvas subclass: #FormCanvas
	instanceVariableNames: 'origin clipRect form port shadowColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!FormCanvas commentStamp: '<historical>' prior: 0!
Note that when shadowDrawing is true, shadowStipple may be either a color, for a solid shadow of the given color, or it may be a stipple used to simulate gray shading when the display cannot support alpha blending.!

FormCanvas subclass: #BalloonCanvas
	instanceVariableNames: 'transform colorTransform engine aaLevel deferred'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Balloon'!

!BalloonCanvas commentStamp: '<historical>' prior: 0!
BalloonCanvas is a canvas using the BalloonEngine for drawing wherever possible. It has various methods which other canvases do not support due to the extra features of the balloon engine.!

!BalloonCanvas methodsFor: 'accessing' stamp: 'ar 11/13/1998 01:02'!
aaLevel
	^aaLevel! !

!BalloonCanvas methodsFor: 'accessing' stamp: 'ar 12/30/1998 10:53'!
aaLevel: newLevel
	"Only allow changes to aaLevel if we're working on >= 8 bit forms"
	form depth >= 8 ifFalse:[^self].
	aaLevel = newLevel ifTrue:[^self].
	self flush.	"In case there are pending primitives in the engine"
	aaLevel := newLevel.
	engine ifNotNil:[engine aaLevel: aaLevel].! !

!BalloonCanvas methodsFor: 'converting' stamp: 'ar 11/11/1998 22:57'!
asBalloonCanvas
	^self! !

!BalloonCanvas methodsFor: 'transforming' stamp: 'ar 11/24/1998 14:45'!
colorTransformBy: aColorTransform
	aColorTransform ifNil:[^self].
	colorTransform 
		ifNil:[colorTransform := aColorTransform]
		ifNotNil:[colorTransform := colorTransform composedWithLocal: aColorTransform]! !

!BalloonCanvas methodsFor: 'copying' stamp: 'ar 11/24/1998 22:33'!
copy
	self flush.
	^super copy resetEngine! !

!BalloonCanvas methodsFor: 'accessing' stamp: 'ar 12/30/1998 10:54'!
deferred
	^deferred! !

!BalloonCanvas methodsFor: 'accessing' stamp: 'ar 12/30/1998 10:55'!
deferred: aBoolean
	deferred == aBoolean ifTrue:[^self].
	self flush. "Force pending prims on screen"
	deferred := aBoolean.
	engine ifNotNil:[engine deferred: aBoolean].! !

!BalloonCanvas methodsFor: 'balloon drawing' stamp: 'DSM 10/15/1999 15:14'!
drawBezier3Shape: vertices color: c borderWidth: borderWidth borderColor:
borderColor
	self drawBezierShape: (Bezier3Segment convertBezier3ToBezier2:
vertices) color: c borderWidth: borderWidth borderColor: borderColor! !

!BalloonCanvas methodsFor: 'balloon drawing' stamp: 'ar 2/17/2000 00:25'!
drawBezierShape: vertices color: c borderWidth: borderWidth borderColor: borderColor
	"Draw a boundary shape that is defined by a list of vertices.
	Each three subsequent vertices define a quadratic bezier segment.
	For lines, the control point should be set to either the start or the end
	of the bezier curve."
	| fillC borderC |
	fillC := self shadowColor ifNil:[c].
	borderC := self shadowColor ifNil:[borderColor].
	self ensuredEngine
		drawBezierShape: vertices
		fill: fillC
		borderWidth: borderWidth
		borderColor: borderC
		transform: transform.! !

!BalloonCanvas methodsFor: 'balloon drawing' stamp: 'ar 11/24/1998 15:16'!
drawCompressedShape: compressedShape
	"Draw a compressed shape"
	self ensuredEngine
		drawCompressedShape: compressedShape
		transform: transform.! !

!BalloonCanvas methodsFor: 'balloon drawing' stamp: 'DSM 10/15/1999 15:18'!
drawGeneralBezier3Shape: contours color: c borderWidth: borderWidth
borderColor: borderColor
	| b2 |
	b2 := contours collect: [:b3 | Bezier3Segment
convertBezier3ToBezier2: b3 ].
	self drawGeneralBezierShape: b2 color: c borderWidth: borderWidth
borderColor: borderColor! !

!BalloonCanvas methodsFor: 'balloon drawing' stamp: 'ar 2/17/2000 00:24'!
drawGeneralBezierShape: contours color: c borderWidth: borderWidth borderColor: borderColor
	"Draw a general boundary shape (e.g., possibly containing holes)"
	| fillC borderC |
	fillC := self shadowColor ifNil:[c].
	borderC := self shadowColor ifNil:[borderColor].
	self ensuredEngine
		drawGeneralBezierShape: contours
		fill: fillC
		borderWidth: borderWidth
		borderColor: borderC
		transform: transform.! !

!BalloonCanvas methodsFor: 'balloon drawing' stamp: 'ar 2/17/2000 00:24'!
drawGeneralPolygon: contours color: c borderWidth: borderWidth borderColor: borderColor
	"Draw a general polygon (e.g., a polygon that can contain holes)"
	| fillC borderC |
	fillC := self shadowColor ifNil:[c].
	borderC := self shadowColor ifNil:[borderColor].
	self ensuredEngine
		drawGeneralPolygon: contours
		fill: fillC
		borderWidth: borderWidth
		borderColor: borderC
		transform: transform.! !

!BalloonCanvas methodsFor: 'balloon drawing' stamp: 'ar 2/17/2000 00:24'!
drawOval: r color: c borderWidth: borderWidth borderColor: borderColor
	"Draw the oval defined by the given rectangle"
	| fillC borderC |
	fillC := self shadowColor ifNil:[c].
	borderC := self shadowColor ifNil:[borderColor].
	self ensuredEngine
		drawOval: r
		fill: fillC
		borderWidth: borderWidth
		borderColor: borderC
		transform: transform.! !

!BalloonCanvas methodsFor: 'drawing-polygons' stamp: 'ar 8/26/2001 22:14'!
drawPolygon: vertices fillStyle: aFillStyle
	"Fill the given polygon."
	self drawPolygon: vertices fillStyle: aFillStyle borderWidth: 0 borderColor: nil! !

!BalloonCanvas methodsFor: 'drawing-polygons' stamp: 'cmm 2/17/2013 14:33'!
drawPolygon: vertices fillStyle: aFillStyle borderWidth: borderWidth borderColor: borderColor
	"Draw a simple polygon defined by the list of vertices."
	| fillC borderC |
	vertices ifEmpty: [ ^ self ].
	fillC := self shadowColor ifNil:[aFillStyle].
	borderC := self shadowColor ifNil:[borderColor].
	self ensuredEngine
		drawPolygon: (vertices copyWith: vertices first)
		fill: fillC
		borderWidth: borderWidth
		borderColor: borderC
		transform: transform.! !

!BalloonCanvas methodsFor: 'balloon drawing' stamp: 'ar 2/17/2000 00:24'!
drawRectangle: r color: c borderWidth: borderWidth borderColor: borderColor
	"Draw a rectangle"
	| fillC borderC |
	fillC := self shadowColor ifNil:[c].
	borderC := self shadowColor ifNil:[borderColor].
	self ensuredEngine
		drawRectangle: r
		fill: fillC
		borderWidth: borderWidth
		borderColor: borderC
		transform: transform.! !

!BalloonCanvas methodsFor: 'TODO' stamp: 'ar 12/31/2001 02:27'!
drawString: s from: firstIndex to: lastIndex in: boundsRect font: fontOrNil color: c
	(self ifNoTransformWithIn: boundsRect)
		ifTrue:[^super drawString: s from: firstIndex to: lastIndex in: boundsRect font: fontOrNil color: c]! !

!BalloonCanvas methodsFor: 'accessing' stamp: 'ar 2/13/2001 21:07'!
ensuredEngine
	engine ifNil:[
		engine := BalloonEngine new.
		"engine := BalloonDebugEngine new"
		engine aaLevel: aaLevel.
		engine bitBlt: port.
		engine destOffset: origin.
		engine clipRect: clipRect.
		engine deferred: deferred.
		engine].
	engine colorTransform: colorTransform.
	engine edgeTransform: transform.
	^engine! !

!BalloonCanvas methodsFor: 'drawing' stamp: 'ar 2/9/1999 05:40'!
fillColor: c
	"Note: This always fills, even if the color is transparent."
	"Note2: To achieve the above we must make sure that c is NOT transparent"
	self frameAndFillRectangle: form boundingBox 
		fillColor: (c alpha: 1.0)
		borderWidth: 0
		borderColor: nil! !

!BalloonCanvas methodsFor: 'drawing' stamp: 'ar 2/9/1999 05:51'!
fillOval: r color: c borderWidth: borderWidth borderColor: borderColor
	"Draw a filled and outlined oval"
	"Note: The optimization test below should actually read:
		self ifNoTransformWithIn: (r insetBy: borderWidth // 2)
	but since borderWidth is assumed to be very small related to r we don't check it."

	(self ifNoTransformWithIn: r)
		ifTrue:[^super fillOval: r color: c borderWidth: borderWidth borderColor: borderColor].

	^self drawOval: (r insetBy: borderWidth // 2) 
			color: c 
			borderWidth: borderWidth 
			borderColor: borderColor! !

!BalloonCanvas methodsFor: 'drawing-ovals' stamp: 'ar 6/18/1999 08:50'!
fillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc
	"Fill the given rectangle."
	^self drawOval: (aRectangle insetBy: bw // 2)
			color: aFillStyle "@@: Name confusion!!!!!!"
			borderWidth: bw
			borderColor: bc
! !

!BalloonCanvas methodsFor: 'drawing' stamp: 'ar 2/9/1999 05:40'!
fillRectangle: r color: c
	"Fill the rectangle with the given color"
	^self frameAndFillRectangle: r
			fillColor: c
			borderWidth: 0
			borderColor: nil! !

!BalloonCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/18/1999 08:48'!
fillRectangle: aRectangle fillStyle: aFillStyle
	"Fill the given rectangle."
	^self drawRectangle: aRectangle
			color: aFillStyle "@@: Name confusion!!!!!!"
			borderWidth: 0
			borderColor: nil
! !

!BalloonCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 6/8/2003 19:03'!
fillRoundRect: aRectangle radius: radius fillStyle: fillStyle
	| points |
	radius asPoint <= (0@0) 
		ifTrue:[^self fillRectangle: aRectangle fillStyle: fillStyle].
	(radius * 2) asPoint >= aRectangle extent 
		ifTrue:[^self fillOval: aRectangle fillStyle: fillStyle].
	"decompose aRectangle into bezier form"
	points := self makeRoundRectShape: aRectangle radius: radius.
	"blast the bezier shape out"
	self
		drawBezierShape: points
		color: fillStyle
		borderWidth: 0
		borderColor: nil.
! !

!BalloonCanvas methodsFor: 'initialize' stamp: 'ar 11/24/1998 15:28'!
flush
	"Force all pending primitives onscreen"
	engine ifNotNil:[engine flush].! !

!BalloonCanvas methodsFor: 'drawing' stamp: 'ar 2/9/1999 06:26'!
frameAndFillRectangle: r fillColor: c borderWidth: borderWidth borderColor: borderColor
	"Draw a filled and outlined rectangle"
	"Note: The optimization test below should actually read:
		self ifNoTransformWithIn: (r insetBy: borderWidth // 2)
	but since borderWidth is assumed to be very small related to r we don't check it."

	(self ifNoTransformWithIn: r)
		ifTrue:[^super frameAndFillRectangle: r fillColor: c borderWidth: borderWidth borderColor: borderColor].

	^self drawRectangle: (r insetBy: borderWidth // 2) 
			color: c 
			borderWidth: borderWidth 
			borderColor: borderColor! !

!BalloonCanvas methodsFor: 'drawing' stamp: 'ar 2/9/1999 05:52'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth topLeftColor: topLeftColor bottomRightColor: bottomRightColor
	"Draw a beveled or raised rectangle"
	| bw |

	"Note: The optimization test below should actually read:
		self ifNoTransformWithIn: (r insetBy: borderWidth // 2)
	but since borderWidth is assumed to be very small related to r we don't check it."

	(self ifNoTransformWithIn: r)
		ifTrue:[^super frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth topLeftColor: topLeftColor bottomRightColor: bottomRightColor].

	"Fill rectangle and draw top and left border"
	bw := borderWidth // 2.
	self drawRectangle: (r insetBy: bw)
		color: fillColor
		borderWidth: borderWidth
		borderColor: topLeftColor.
	"Now draw bottom right border."
	self drawPolygon: (Array with: r topRight + (bw negated@bw) 
							with: r bottomRight - bw asPoint
							with: r bottomLeft + (bw@bw negated))
		color: nil
		borderWidth: borderWidth
		borderColor: bottomRightColor.! !

!BalloonCanvas methodsFor: 'drawing' stamp: 'nk 5/1/2004 12:25'!
frameRectangle: r width: w color: c
	"Draw a frame around the given rectangle"
	^self frameAndFillRectangle: r
			fillColor: Color transparent
			borderWidth: w
			borderColor: c! !

!BalloonCanvas methodsFor: 'drawing-rectangles' stamp: 'mt 9/4/2015 08:36'!
frameRoundRect: aRectangle radius: radius width: borderWidth color: borderColor
	| outerPoints innerRect innerRadius innerPoints |
	(borderWidth isZero or:[borderColor isTransparent])
		ifTrue:[^self].
	radius asPoint <= (0@0) 
		ifTrue:[^self frameRectangle: aRectangle width: borderWidth color: borderColor].
	"decompose inner rectangle into bezier shape"
	innerRect := aRectangle insetBy: borderWidth.
	innerRect area <= 0 
		ifTrue:[^self fillRoundRect: aRectangle radius: radius fillStyle: borderColor].
	innerRadius := (radius - borderWidth) asPoint.
	innerPoints := self makeRoundRectShape: innerRect radius: innerRadius.
	"decompose outer rectangle into bezier shape"
	outerPoints := self makeRoundRectShape: aRectangle radius: radius.
	self
		drawGeneralBezierShape: (Array with: outerPoints with: innerPoints)
		color: borderColor
		borderWidth: 0
		borderColor: nil.! !

!BalloonCanvas methodsFor: 'private' stamp: 'marcus.denker 9/14/2008 21:01'!
ifNoTransformWithIn: box
	"Return true if the current transformation does not affect the given bounding box"
	| delta |
	transform ifNil: [^true].
	delta := (transform localPointToGlobal: box origin) - box origin.
	^(transform localPointToGlobal: box corner) - box corner = delta! !

!BalloonCanvas methodsFor: 'private' stamp: 'tpr 3/15/2013 12:40'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	| warp dstRect srcQuad dstOffset center |
	(self ifNoTransformWithIn: sourceRect) & false
		ifTrue:[^super image: aForm at: aPoint sourceRect: sourceRect rule: rule].
	dstRect := (transform localBoundsToGlobal: (aForm boundingBox translateBy: aPoint)).
	dstOffset := 0@0. "dstRect origin."
	"dstRect := 0@0 corner: dstRect extent."
	center := 0@0."transform globalPointToLocal: dstRect origin."
	srcQuad := transform globalPointsToLocal: (dstRect innerCorners).
	srcQuad := srcQuad collect:[:pt| pt - aPoint].
	warp := (WarpBlt toForm: form)
			sourceForm: aForm;
			cellSize: 2;  "installs a new colormap if cellSize > 1"
			combinationRule: Form over.
	warp copyQuad: srcQuad toRect: (dstRect translateBy: dstOffset).

	self frameRectangle: (aForm boundingBox translateBy: aPoint) color: Color green.

	"... TODO ... create a bitmap fill style from the form and use it for a simple rectangle."! !

!BalloonCanvas methodsFor: 'initialize' stamp: 'alain.plantec 5/28/2009 09:42'!
initialize
	super initialize.
	aaLevel := 1.
	deferred := false.! !

!BalloonCanvas methodsFor: 'testing' stamp: 'ar 11/13/1998 13:19'!
isBalloonCanvas
	^true! !

!BalloonCanvas methodsFor: 'testing' stamp: 'ar 11/12/1998 01:07'!
isVisible: aRectangle
	^transform 
		ifNil:[super isVisible: aRectangle]
		ifNotNil:[super isVisible: (transform localBoundsToGlobal: aRectangle)]! !

!BalloonCanvas methodsFor: 'TODO' stamp: 'ar 2/9/1999 05:46'!
line: point1 to: point2 brushForm: brush
	"Who's gonna use this?"
	| pt1 pt2 |
	self flush. "Sorry, but necessary..."
	transform 
		ifNil:[pt1 := point1. pt2 := point2]
		ifNotNil:[pt1 := transform localPointToGlobal: point1.
				pt2 := transform localPointToGlobal: point2].
	^super line: pt1 to: pt2 brushForm: brush! !

!BalloonCanvas methodsFor: 'drawing' stamp: 'ul 11/15/2010 11:17'!
line: pt1 to: pt2 width: w color: c
	"Draw a line from pt1 to: pt2"

	(aaLevel = 1 and: [self ifNoTransformWithIn:(pt1 rect: pt2)])
		ifTrue:[^super line: pt1 to: pt2 width: w color: c].
	^self drawPolygon: (Array with: pt1 with: pt2)
		color: c
		borderWidth: w
		borderColor: c! !

!BalloonCanvas methodsFor: 'private' stamp: 'nice 12/29/2010 16:10'!
makeRoundRectShape: aRectangle radius: radius
	"decompose a rounded rectangle into bezier form"
	| ovalDiameter rectExtent segments points endPoint seg idx offset rectOffset |
	ovalDiameter := (radius * 2) asPoint min: aRectangle extent.
	(ovalDiameter x <= 0 or: [ovalDiameter y <= 0]) ifTrue: [
		"degenerates into rectangle - just hand back four lines"
		| topLeft topRight bottomLeft bottomRight |
		topLeft := aRectangle topLeft.
		topRight := aRectangle topRight.
		bottomLeft := aRectangle bottomLeft.
		bottomRight := aRectangle bottomRight.

		points := Array new: 4 * 3.
		points at: 1 put: topLeft.
		points at: 2 put: topLeft.
		points at: 3 put: topRight.

		points at: 4 put: topRight.
		points at: 5 put: topRight.
		points at: 6 put: bottomRight.

		points at: 7 put: bottomRight.
		points at: 8 put: bottomRight.
		points at: 9 put: bottomLeft.

		points at: 10 put: bottomLeft.
		points at: 11 put: bottomLeft.
		points at: 12 put: topLeft.
		^points
	].
	rectExtent := aRectangle extent - ovalDiameter.
	rectOffset := aRectangle origin.
	segments := Bezier2Segment makeEllipseSegments: (0 @ 0 extent: ovalDiameter).
	"patch up the segments to include lines connecting the oval parts.
	we need: 8*3 points for the oval parts + 4*3 points for the connecting lines"
	points := Array new: 12 * 3.
	idx := 0.
	"Tweaked offsets to clean up curves. MAD"
	endPoint := segments last end + rectOffset + (0 @ -1).
	1 to: 8 by: 2 do: [:i |
		i = 1 ifTrue: [offset := rectOffset + (rectExtent x @ 0) + (1 @ -1)]. "top, tr"
		i = 3 ifTrue: [offset := rectOffset + rectExtent + (1 @ 1)]. "right, br"
		i = 5 ifTrue: [offset := rectOffset + (0 @ rectExtent y) + (0 @ 1)]. "bottom, bl"
		i = 7 ifTrue: [offset := rectOffset + (0 @ -1)]."left, tl"
		seg := segments at: i.
		"insert a line segment for the horizontal part of the round rect"
		points at: (idx := idx + 1) put: endPoint.
		points at: (idx := idx + 1) put: endPoint.
		points at: (idx := idx + 1) put: seg start + offset.
		"now the first half-arc"
		points at: (idx := idx + 1) put: seg start + offset.
		points at: (idx := idx + 1) put: seg via + offset.
		points at: (idx := idx + 1) put: seg end + offset.
		"the second half-arc"
		seg := segments at: i + 1.
		points at: (idx := idx + 1) put: seg start + offset.
		points at: (idx := idx + 1) put: seg via + offset.
		points at: (idx := idx + 1) put: seg end + offset.
		endPoint := seg end + offset.
	].
	^points! !

!BalloonCanvas methodsFor: 'TODO' stamp: 'ar 2/9/1999 05:46'!
paragraph: para bounds: bounds color: c
	(self ifNoTransformWithIn: bounds)
		ifTrue:[^super paragraph: para bounds: bounds color: c].! !

!BalloonCanvas methodsFor: 'drawing' stamp: 'ar 11/11/1998 19:39'!
point: pt color: c
	"Is there any use for this?"
	| myPt |
	transform 
		ifNil:[myPt := pt]
		ifNotNil:[myPt := transform localPointToGlobal: pt].
	^super point: myPt color: c! !

!BalloonCanvas methodsFor: 'transforming' stamp: 'ar 12/30/1998 10:47'!
preserveStateDuring: aBlock
	| state result |
	state := BalloonState new.
	state transform: transform.
	state colorTransform: colorTransform.
	state aaLevel: self aaLevel.
	result := aBlock value: self.
	transform := state transform.
	colorTransform := state colorTransform.
	self aaLevel: state aaLevel.
	^result! !

!BalloonCanvas methodsFor: 'initialize' stamp: 'ar 11/11/1998 20:25'!
resetEngine
	engine := nil.! !

!BalloonCanvas methodsFor: 'transforming' stamp: 'ar 11/12/1998 00:32'!
transformBy: aTransform
	aTransform ifNil:[^self].
	transform 
		ifNil:[transform := aTransform]
		ifNotNil:[transform := transform composedWithLocal: aTransform]! !

!BalloonCanvas methodsFor: 'transforming' stamp: 'ar 5/29/1999 08:59'!
transformBy: aDisplayTransform during: aBlock
	| myTransform result |
	myTransform := transform.
	self transformBy: aDisplayTransform.
	result := aBlock value: self.
	transform := myTransform.
	^result! !

FormCanvas subclass: #ColorPatchCanvas
	instanceVariableNames: 'stopMorph foundMorph doStop'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!ColorPatchCanvas commentStamp: '<historical>' prior: 0!
I generate patches of Morphic worlds that views below certain Morphs. This facility is used for the end-user scripting system.!

!ColorPatchCanvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 16:34'!
clipBy: aRectangle during: aBlock
	"Set a clipping rectangle active only during the execution of aBlock.
	Note: In the future we may want to have more general clip shapes - not just rectangles"
	| tempCanvas |
	tempCanvas := (self copyClipRect: aRectangle).
	aBlock value: tempCanvas.
	foundMorph := tempCanvas foundMorph.! !

!ColorPatchCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 16:15'!
doStop
	^doStop! !

!ColorPatchCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 16:15'!
doStop: aBoolean
	doStop := aBoolean! !

!ColorPatchCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 16:15'!
foundMorph
	^foundMorph! !

!ColorPatchCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 16:38'!
foundMorph: aBoolean
	foundMorph := aBoolean! !

!ColorPatchCanvas methodsFor: 'drawing-general' stamp: 'ar 6/22/1999 16:14'!
fullDrawMorph: aMorph
	(foundMorph and:[doStop]) ifTrue:[^self]. "Found it and should stop"
	aMorph == stopMorph ifTrue:[
		"Never draw the stopMorph"
		foundMorph := true.
		^self].
	^super fullDrawMorph: aMorph.! !

!ColorPatchCanvas methodsFor: 'drawing-support' stamp: 'nice 12/26/2009 19:22'!
preserveStateDuring: aBlock
	"Preserve the full canvas state during the execution of aBlock.
	Note: This does *not* include the state in the receiver (e.g., foundMorph)."
	| tempCanvas result |
	tempCanvas := self copy.
	result := aBlock value: tempCanvas.
	foundMorph := tempCanvas foundMorph.
	^result! !

!ColorPatchCanvas methodsFor: 'initialization' stamp: 'ar 6/22/1999 16:18'!
reset
	"Initialize the receiver to act just as a FormCanvas"
	super reset.
	foundMorph := false.
	doStop := false.
	stopMorph := nil.! !

!ColorPatchCanvas methodsFor: 'private' stamp: 'ar 6/22/1999 16:18'!
setForm: aForm
	"Initialize the receiver to act just as a FormCanvas"
	super setForm: aForm.
	stopMorph := nil.
	doStop := false.
	foundMorph := false.! !

!ColorPatchCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 16:14'!
stopMorph
	^stopMorph! !

!ColorPatchCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 16:14'!
stopMorph: aMorph
	stopMorph := aMorph! !

!ColorPatchCanvas methodsFor: 'drawing-support' stamp: 'ul 8/2/2011 19:58'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock smoothing: cellSize
	"Note: This method has been originally copied from TransformationMorph."
	| innerRect patchRect sourceQuad warp start subCanvas |
	(aDisplayTransform isPureTranslation) ifTrue:[
		subCanvas := self copyOffset: aDisplayTransform offset negated truncated
							clipRect: aClipRect.
		aBlock value: subCanvas.
		foundMorph := subCanvas foundMorph.
		^self
	].
	"Prepare an appropriate warp from patch to innerRect"
	innerRect := aClipRect.
	patchRect := aDisplayTransform globalBoundsToLocal:
					(self clipRect intersect: innerRect).
	sourceQuad := (aDisplayTransform sourceQuadFor: innerRect)
					collect: [:p | p - patchRect topLeft].
	warp := self warpFrom: sourceQuad toRect: innerRect.
	warp cellSize: cellSize.

	"Render the submorphs visible in the clipping rectangle, as patchForm"
	start := (self depth = 1 and: [self isShadowDrawing not])
		"If this is true B&W, then we need a first pass for erasure."
		ifTrue: [1] ifFalse: [2].
	start to: 2 do:
		[:i | "If i=1 we first make a shadow and erase it for opaque whites in B&W"
		subCanvas := ColorPatchCanvas extent: patchRect extent depth: self depth.
		subCanvas stopMorph: stopMorph.
		subCanvas foundMorph: foundMorph.
		subCanvas doStop: doStop.
		i=1	ifTrue: [subCanvas shadowColor: Color black.
					warp combinationRule: Form erase]
			ifFalse: [self isShadowDrawing ifTrue:
					[subCanvas shadowColor: self shadowColor].
					warp combinationRule: Form paint].
		subCanvas
			translateBy: patchRect topLeft negated
			during: aBlock.
		i = 2 ifTrue:[foundMorph := subCanvas foundMorph].
		warp sourceForm: subCanvas form; warpBits.
		warp sourceForm: nil.  subCanvas := nil "release space for next loop"]
! !

!ColorPatchCanvas methodsFor: 'other' stamp: 'ar 6/22/1999 16:39'!
translateBy: delta clippingTo: aRectangle during: aBlock
	"Set a translation and clipping rectangle only during the execution of aBlock."
	| tempCanvas |
	tempCanvas := self copyOffset: delta clipRect: aRectangle.
	aBlock value: tempCanvas.
	foundMorph := tempCanvas foundMorph.! !

!ColorPatchCanvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 16:39'!
translateBy: delta during: aBlock
	"Set a translation only during the execution of aBlock."
	| tempCanvas |
	tempCanvas := self copyOffset: delta.
	aBlock value: tempCanvas.
	foundMorph := tempCanvas foundMorph.! !

!ColorPatchCanvas methodsFor: 'drawing-support' stamp: 'ar 6/22/1999 16:40'!
translateTo: newOrigin clippingTo: aRectangle during: aBlock
	"Set a new origin and clipping rectangle only during the execution of aBlock."
	| tempCanvas |
	tempCanvas := self copyOrigin: newOrigin clipRect: aRectangle.
	aBlock value: tempCanvas.
	foundMorph := tempCanvas foundMorph.! !

!FormCanvas class methodsFor: 'instance creation' stamp: ''!
extent: aPoint

	^ self extent: aPoint depth: Display depth
! !

!FormCanvas class methodsFor: 'instance creation' stamp: ''!
extent: extent depth: depth

	^ self new setForm: (Form extent: extent depth: depth)! !

!FormCanvas class methodsFor: 'instance creation' stamp: 'nk 7/4/2003 10:11'!
extent: extent depth: depth origin: aPoint clipRect: aRectangle

	^ self new
		setForm: (Form extent: extent depth: depth);
		setOrigin: aPoint clipRect: aRectangle;
		yourself! !

!FormCanvas class methodsFor: 'instance creation' stamp: 'jm 8/2/97 13:54'!
on: aForm

	^ self new setForm: aForm
! !

!FormCanvas class methodsFor: 'testing' stamp: 'ar 12/31/2001 02:26'!
test1
	"FormCanvas test1"

	| canvas |
	canvas := FormCanvas extent: 200@200.
	canvas fillColor: (Color black).
	canvas line: 10@10 to: 50@30 width: 1 color: (Color red).
	canvas frameRectangle: ((20@20) corner: (120@120)) width: 4 color: (Color green).
	canvas point: 100@100 color: (Color black).
	canvas drawString: 'Hello, World!!' at: 40@40 font: nil color: (Color cyan).
	canvas fillRectangle: ((10@80) corner: (31@121)) color: (Color magenta).
	canvas fillOval: ((10@80) corner: (31@121)) color: (Color cyan).
	canvas frameOval: ((40@80) corner: (61@121)) color: (Color blue).
	canvas frameOval: ((70@80) corner: (91@121)) width: 3 color: (Color red alpha: 0.2).
	canvas fillRectangle: ((130@30) corner: (170@80)) color: (Color lightYellow).
	canvas showAt: 0@0.
! !

!FormCanvas class methodsFor: 'testing' stamp: 'ar 12/31/2001 02:26'!
test2
	"FormCanvas test2"

	| baseCanvas p |
	baseCanvas := FormCanvas extent: 200@200.
	p := Sensor cursorPoint.
	[Sensor anyButtonPressed] whileFalse: [
		baseCanvas translateBy: (Sensor cursorPoint - p) during:[:canvas|
			canvas fillColor: Color white.
			canvas line: 10@10 to: 50@30 width: 1 color: Color red.
			canvas frameRectangle: ((20@20) corner: (120@120)) width: 4 color: Color green.
			canvas point: 100@100 color: Color black.
			canvas drawString: 'Hello, World!!' at: 40@40 font: nil color: Color cyan.
			canvas fillRectangle: ((10@80) corner: (31@121)) color: Color magenta.
			canvas fillOval: ((10@80) corner: (31@121)) color: Color cyan.
			canvas frameOval: ((40@80) corner: (61@121)) color: Color blue.
			canvas frameOval: ((70@80) corner: (91@121)) width: 3 color: Color red.
			canvas fillRectangle: ((130@30) corner: (170@80)) color: Color lightYellow.
			canvas showAt: 0@0]].
! !

!FormCanvas class methodsFor: 'testing' stamp: 'ar 12/31/2001 02:25'!
test3
	"FormCanvas test3"

	| baseCanvas |
	baseCanvas := FormCanvas extent: 200@200.
	baseCanvas fillColor: Color white.
	baseCanvas translateBy: 10@10 during:[:canvas|
		canvas shadowColor: (Color black alpha: 0.5).
		canvas line: 10@10 to: 50@30 width: 1 color: Color red.
		canvas frameRectangle: ((20@20) corner: (120@120)) width: 4 color: Color green.
		canvas point: 100@100 color: Color black.
		canvas drawString: 'Hello, World!!' at: 40@40 font: nil color: Color cyan.
		canvas fillRectangle: ((10@80) corner: (31@121)) color: Color magenta.
		canvas fillOval: ((10@80) corner: (31@121)) color: Color cyan.
		canvas frameOval: ((40@80) corner: (61@121)) color: Color blue.
		canvas frameOval: ((70@80) corner: (91@121)) width: 3 color: Color red.
		canvas fillRectangle: ((130@30) corner: (170@80)) color: Color lightYellow.
		canvas showAt: 0@0.
	].! !

!FormCanvas methodsFor: 'accessing' stamp: 'ar 5/28/2000 17:11'!
allocateForm: extentPoint
	"Allocate a new form which is similar to the receiver"
	^form allocateForm: extentPoint! !

!FormCanvas methodsFor: 'other' stamp: 'ar 11/11/1998 22:57'!
asBalloonCanvas
	^(BalloonCanvas on: form) setOrigin: origin clipRect: clipRect! !

!FormCanvas methodsFor: 'converting' stamp: 'ar 2/17/2000 00:17'!
asShadowDrawingCanvas
	"Note: This is sort of an optimization here since since the logic is all there"
	^self copy shadowColor: (Color black alpha: 0.5)! !

!FormCanvas methodsFor: 'converting' stamp: 'ar 2/17/2000 00:16'!
asShadowDrawingCanvas: aColor
	"Note: This is sort of an optimization here since since the logic is all there"
	^self copy shadowColor: aColor! !

!FormCanvas methodsFor: 'drawing-ovals' stamp: 'RAA 11/6/2000 15:21'!
balloonFillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc

	self asBalloonCanvas
		fillOval: aRectangle
		fillStyle: aFillStyle
		borderWidth: bw
		borderColor: bc! !

!FormCanvas methodsFor: 'private' stamp: 'RAA 7/28/2000 07:39'!
balloonFillRectangle: aRectangle fillStyle: aFillStyle

	self asBalloonCanvas fillRectangle: aRectangle fillStyle: aFillStyle.! !

!FormCanvas methodsFor: 'private' stamp: 'ar 7/22/2009 21:35'!
balloonFillRoundRect: aRectangle radius: radius fillStyle: fillStyle
	^self asBalloonCanvas fillRoundRect: aRectangle radius: radius fillStyle: fillStyle! !

!FormCanvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 03:02'!
clipBy: aRectangle during: aBlock
	"Set a clipping rectangle active only during the execution of aBlock.
	Note: In the future we may want to have more general clip shapes - not just rectangles"
	^aBlock value: (self copyClipRect: aRectangle)! !

!FormCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 14:06'!
clipRect
	"Return the currently active clipping rectangle"
	^ clipRect translateBy: origin negated! !

!FormCanvas methodsFor: 'accessing' stamp: 'ar 12/31/2001 03:26'!
contentsOfArea: aRectangle into: aForm
	| bb |
	self flush.
	bb := BitBlt toForm: aForm.
	bb sourceForm: form; combinationRule: Form over;
		sourceX: (aRectangle left + origin x); sourceY: (aRectangle top + origin y);
		width: aRectangle width; height: aRectangle height;
		copyBits.
	^aForm! !

!FormCanvas methodsFor: 'copying' stamp: 'jm 8/2/97 14:00'!
copy
	"Make a copy the receiver on the same underlying Form but with its own grafPort."

	^ self clone resetGrafPort
! !

!FormCanvas methodsFor: 'copying' stamp: 'ar 6/17/1999 02:51'!
copyClipRect: aRectangle
	^ self copyOrigin: origin clipRect: (aRectangle translateBy: origin)
! !

!FormCanvas methodsFor: 'copying' stamp: 'ar 6/17/1999 02:52'!
copyOffset: aPoint
	^ self copyOrigin: origin + aPoint clipRect: clipRect! !

!FormCanvas methodsFor: 'copying' stamp: 'ar 6/17/1999 02:52'!
copyOffset: aPoint clipRect: sourceClip
	"Make a copy of me offset by aPoint, and further clipped
	by sourceClip, a rectangle in the un-offset coordinates"
	^ self copyOrigin: aPoint + origin
		clipRect: ((sourceClip translateBy: origin) intersect: clipRect)! !

!FormCanvas methodsFor: 'copying' stamp: 'ar 6/17/1999 02:52'!
copyOrigin: aPoint clipRect: aRectangle
	"Return a copy of this canvas with the given origin. The clipping rectangle of this canvas is the intersection of the given rectangle and the receiver's current clipping rectangle. This allows the clipping rectangles of nested clipping morphs to be composed."
	^ self copy
		setOrigin: aPoint
		clipRect: (clipRect intersect: aRectangle)! !

!FormCanvas methodsFor: 'accessing' stamp: ''!
depth

	^ form depth
! !

!FormCanvas methodsFor: 'drawing-polygons' stamp: 'ar 6/18/1999 08:57'!
drawPolygon: vertices color: aColor borderWidth: bw borderColor: bc
	"Generalize for the BalloonCanvas"
	^self drawPolygon: vertices fillStyle: aColor borderWidth: bw borderColor: bc! !

!FormCanvas methodsFor: 'drawing-polygons' stamp: 'ar 12/6/2000 14:59'!
drawPolygon: vertices fillStyle: aFillStyle borderWidth: bw borderColor: bc
	"Use a BalloonCanvas"
	self asBalloonCanvas 
		drawPolygon: vertices asArray
		fillStyle: (self shadowColor ifNil:[aFillStyle])
		borderWidth: bw 
		borderColor: bc! !

!FormCanvas methodsFor: 'drawing-text' stamp: 'ar 2/5/2002 19:03'!
drawString: aString from: firstIndex to: lastIndex at: aPoint font: fontOrNil color: c
	| font |
	port colorMap: nil.
	font := fontOrNil ifNil: [TextStyle defaultFont].
	port combinationRule: Form paint.
	font installOn: port
		foregroundColor: (self shadowColor ifNil:[c]) 
		backgroundColor: Color transparent.
	font displayString: aString on: port 
		from: firstIndex to: lastIndex at: (origin + aPoint) kern: 0.! !

!FormCanvas methodsFor: 'drawing-text' stamp: 'ar 2/5/2002 19:03'!
drawString: aString from: firstIndex to: lastIndex in: bounds font: fontOrNil color: c
	| font portRect |
	port colorMap: nil.
	portRect := port clipRect.
	port clipByX1: bounds left + origin x 
		y1: bounds top + origin y 
		x2: bounds right + origin x 
		y2: bounds bottom + origin y.
	font := fontOrNil ifNil: [TextStyle defaultFont].
	port combinationRule: Form paint.
	font installOn: port
		foregroundColor: (self shadowColor ifNil:[c]) 
		backgroundColor: Color transparent.
	font displayString: aString asString on: port 
		from: firstIndex to: lastIndex at: (bounds topLeft + origin) kern: 0.
	port clipRect: portRect.! !

!FormCanvas methodsFor: 'accessing' stamp: ''!
extent

	^ form extent! !

!FormCanvas methodsFor: 'drawing' stamp: 'ar 5/14/2000 15:50'!
fillColor: c
	"Note: This always fills, even if the color is transparent."
	self setClearColor: c.
	port fillRect: form boundingBox offset: origin.! !

!FormCanvas methodsFor: 'drawing-ovals' stamp: 'di 5/25/2001 01:40'!
fillOval: r color: fillColor borderWidth: borderWidth borderColor: borderColor
	| rect |
	"draw the border of the oval"
	rect := (r translateBy: origin) truncated.
	(borderWidth = 0 or: [borderColor isTransparent]) ifFalse:[
		self setFillColor: borderColor.
		(r area > 10000 or: [fillColor isTranslucent]) 
			ifTrue: [port frameOval: rect borderWidth: borderWidth]
			ifFalse: [port fillOval: rect]]. "faster this way"
	"fill the inside"
	fillColor isTransparent ifFalse:
		[self setFillColor: fillColor.
		port fillOval: (rect insetBy: borderWidth)].
! !

!FormCanvas methodsFor: 'drawing-ovals' stamp: 'RAA 11/6/2000 16:42'!
fillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc
	"Fill the given oval."

	self flag: #bob.		"this and its siblings could be moved up to Canvas with the
						right #balloonFillOval:..."

	self shadowColor ifNotNil:
		[^self fillOval: aRectangle color: aFillStyle asColor borderWidth: bw borderColor: bc].
	(aFillStyle isBitmapFill and:[aFillStyle isKindOf: InfiniteForm]) ifTrue:[
		self flag: #fixThis.
		^self fillOval: aRectangle color: aFillStyle borderWidth: bw borderColor: bc].
	(aFillStyle isSolidFill) ifTrue:[
		^self fillOval: aRectangle color: aFillStyle asColor borderWidth: bw borderColor: bc].
	"Use a BalloonCanvas instead"
	self balloonFillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc! !

!FormCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 10/26/2000 19:26'!
fillRectangle: aRectangle fillStyle: aFillStyle
	"Fill the given rectangle."
	| pattern |
	self shadowColor ifNotNil:
		[^self fillRectangle: aRectangle color: aFillStyle asColor].

	(aFillStyle isKindOf: InfiniteForm) ifTrue: [
		^self infiniteFillRectangle: aRectangle fillStyle: aFillStyle
	].

	(aFillStyle isSolidFill) 
		ifTrue:[^self fillRectangle: aRectangle color: aFillStyle asColor].
	"We have a very special case for filling with infinite forms"
	(aFillStyle isBitmapFill and:[aFillStyle origin = (0@0)]) ifTrue:[
		pattern := aFillStyle form.
		(aFillStyle direction = (pattern width @ 0) 
			and:[aFillStyle normal = (0@pattern height)]) ifTrue:[
				"Can use an InfiniteForm"
				^self fillRectangle: aRectangle color: (InfiniteForm with: pattern)].
	].
	"Use a BalloonCanvas instead"
	self balloonFillRectangle: aRectangle fillStyle: aFillStyle.! !

!FormCanvas methodsFor: 'drawing-rectangles' stamp: 'mt 1/13/2016 09:02'!
fillRoundRect: aRectangle radius: radius fillStyle: fillStyle
	fillStyle isTransparent ifTrue:[^self].
	radius asPoint <= (0@0) 
		ifTrue:[^self fillRectangle: aRectangle fillStyle: fillStyle].
	fillStyle isSolidFill 
		ifFalse:[^self balloonFillRoundRect: aRectangle radius: radius fillStyle: fillStyle].
	self setFillColor: (shadowColor ifNil:[fillStyle asColor]).
	^port fillRoundRect: (aRectangle translateBy: origin) truncated radius: radius.
! !

!FormCanvas methodsFor: 'initialization' stamp: 'ar 5/27/2000 21:51'!
finish
	"If there are any pending operations on the receiver complete them. Do not return before all modifications have taken effect."
	form finish! !

!FormCanvas methodsFor: 'other' stamp: ''!
flushDisplay
		Display deferUpdates: false; forceDisplayUpdate.! !

!FormCanvas methodsFor: 'other' stamp: ''!
forceToScreen:rect
	^Display forceToScreen:rect.
! !

!FormCanvas methodsFor: 'accessing' stamp: ''!
form

	^ form! !

!FormCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 5/14/2000 15:50'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth borderColor: borderColor
	| rect |
	rect := r translateBy: origin.
	"draw the border of the rectangle"
	borderColor isTransparent ifFalse:[
		self setFillColor: borderColor.
		(r area > 10000 or: [fillColor isTranslucent]) ifTrue: [
			port frameRect: rect borderWidth: borderWidth.
		] ifFalse: ["for small rectangles, it's faster to fill the entire outer rectangle
					than to compute and fill the border rects"
					port fillRect: rect offset: origin]].

	"fill the inside"
	fillColor isTransparent ifFalse:
		[self setFillColor: fillColor.
		port fillRect: (rect insetBy: borderWidth) offset: origin].! !

!FormCanvas methodsFor: 'drawing-rectangles' stamp: 'ar 2/16/2000 22:07'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth topLeftColor: topLeftColor bottomRightColor: bottomRightColor

	| w h rect |
	"First use quick code for top and left borders and fill"
	self frameAndFillRectangle: r
		fillColor: fillColor
		borderWidth: borderWidth
		borderColor: topLeftColor.

	"Now use slow code for bevelled bottom and right borders"
	bottomRightColor isTransparent ifFalse: [
		borderWidth isNumber
			ifTrue: [w := h := borderWidth]
			ifFalse: [w := borderWidth x.   h := borderWidth y].
		rect := r translateBy: origin.
		self setFillColor: bottomRightColor.
		port 
			 frameRectRight: rect width: w;
			 frameRectBottom: rect height: h].
! !

!FormCanvas methodsFor: 'drawing-rectangles' stamp: 'mt 1/13/2016 09:02'!
frameAndFillRoundRect: aRectangle radius: cornerRadius fillStyle: fillStyle borderWidth: bw borderColor: bc
	"Draw a rounded rectangle"
	self shadowColor ifNotNil:[
		^self fillRoundRect: aRectangle radius: cornerRadius fillStyle: shadowColor.
	].
	"see if the round rect is degenerate"
	cornerRadius asPoint <= (0@0) 
		ifTrue:[^self frameAndFillRectangle: aRectangle fillColor: fillStyle asColor borderWidth: bw borderColor: bc].
	"Okay it's a rounded rectangle"
	fillStyle isTransparent ifFalse:["fill interior"
		| innerRect radius |
		innerRect := aRectangle.
		radius := cornerRadius.
		bw isZero ifFalse:[
			innerRect := innerRect insetBy: bw.
			radius := radius - bw.
		].
		self fillRoundRect: innerRect radius: radius fillStyle: fillStyle.
	].
	self frameRoundRect: aRectangle radius: cornerRadius width: bw color: bc
! !

!FormCanvas methodsFor: 'drawing-rectangles' stamp: 'mt 9/4/2015 08:35'!
frameRoundRect: aRectangle radius: radius width: borderWidth color: borderColor
	"Frame a rounded rectangle with the given attributes."
	| innerRect |
	(borderWidth isZero or:[borderColor isTransparent])
		ifTrue:[^self].
	radius asPoint <= (0@0) 
		ifTrue:[^self frameRectangle: aRectangle width: borderWidth color: borderColor].
	"decompose inner rectangle into bezier shape"
	innerRect := aRectangle insetBy: borderWidth.
	innerRect area <= 0 
		ifTrue:[^self fillRoundRect: aRectangle radius: radius fillStyle: borderColor].
	self setFillColor: borderColor.
	port 
		frameRoundRect: (aRectangle translateBy: origin) truncated 
		radius: radius truncated 
		borderWidth: borderWidth truncated.
! !

!FormCanvas methodsFor: 'private' stamp: 'bf 5/8/2011 16:40'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule 
	"Draw the portion of the given Form defined by sourceRect at the given point using the given BitBlt combination rule."
	port colorMap: (aForm colormapIfNeededFor: form); fillColor: nil.
	port image: aForm at: aPoint + origin sourceRect: sourceRect rule: rule.
	(form depth = 32 and: [aForm depth = 16])
		ifTrue: [port image: nil at: aPoint + origin sourceRect: sourceRect rule: 40 "fixAlpha:with:"].
! !

!FormCanvas methodsFor: 'private' stamp: 'ar 8/8/2001 14:21'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule alpha: sourceAlpha
	"Draw the portion of the given Form defined by sourceRect at the given point using the given BitBlt combination rule."
	port colorMap: (aForm colormapIfNeededFor: form); fillColor: nil.
	port image: aForm at: aPoint + origin sourceRect: sourceRect rule: rule alpha: sourceAlpha.! !

!FormCanvas methodsFor: 'private' stamp: 'RAA 2/6/2001 14:00'!
infiniteFillRectangle: aRectangle fillStyle: aFillStyle

	| additionalOffset rInPortTerms clippedPort targetTopLeft clipOffset ex |

	"this is a bit of a kludge to get the form to be aligned where I *think* it should be.
	something better is needed, but not now"

	additionalOffset := 0@0.
	ex := aFillStyle form extent.
	rInPortTerms := aRectangle translateBy: origin.
	clippedPort := port clippedBy: rInPortTerms.
	targetTopLeft := clippedPort clipRect topLeft truncateTo: ex.
	clipOffset := rInPortTerms topLeft - targetTopLeft.
	additionalOffset := (clipOffset \\ ex) - ex.
	^aFillStyle
		displayOnPort: clippedPort
		offsetBy: additionalOffset
! !

!FormCanvas methodsFor: 'testing' stamp: 'ar 2/17/2000 00:24'!
isShadowDrawing
	^ self shadowColor notNil! !

!FormCanvas methodsFor: 'testing' stamp: 'ar 6/22/1999 14:08'!
isVisible: aRectangle
	"Optimization"
	(aRectangle right + origin x) < clipRect left	ifTrue: [^ false].
	(aRectangle left + origin x) > clipRect right	ifTrue: [^ false].
	(aRectangle bottom + origin y) < clipRect top	ifTrue: [^ false].
	(aRectangle top + origin y) > clipRect bottom	ifTrue: [^ false].
	^ true
! !

!FormCanvas methodsFor: 'drawing' stamp: 'ar 5/14/2001 23:34'!
line: pt1 to: pt2 brushForm: brush
	| offset |
	offset := origin.
	self setPaintColor: Color black.
	port sourceForm: brush; fillColor: nil;
		sourceRect: brush boundingBox;
		colorMap: (brush colormapIfNeededFor: form);
		drawFrom: (pt1 + offset) to: (pt2 + offset)! !

!FormCanvas methodsFor: 'drawing' stamp: 'ar 2/16/2000 22:07'!
line: pt1 to: pt2 width: w color: c
	| offset |
	offset := origin - (w // 2) asPoint.
	self setFillColor: c.
	port width: w; height: w;
		drawFrom: (pt1 + offset) to: (pt2 + offset)! !

!FormCanvas methodsFor: 'accessing' stamp: 'ar 6/22/1999 14:10'!
origin
	"Return the current origin for drawing operations"
	^ origin! !

!FormCanvas methodsFor: 'drawing' stamp: 'di 9/12/2001 21:38'!
paragraph: para bounds: bounds color: c

	| scanner |
	self setPaintColor: c.
	scanner := (port clippedBy: (bounds translateBy: origin)) displayScannerFor: para
		foreground: (self shadowColor ifNil:[c]) background: Color transparent
		ignoreColorChanges: self shadowColor notNil.
	para displayOn: (self copyClipRect: bounds) using: scanner at: origin+ bounds topLeft.
! !

!FormCanvas methodsFor: 'drawing' stamp: ''!
point: pt color: c

	form colorAt: (pt + origin) put: c.! !

!FormCanvas methodsFor: 'private' stamp: 'tpr 3/15/2013 12:45'!
portClass
	"Return the class used as port"
	^BitBlt asGrafPort! !

!FormCanvas methodsFor: 'printing' stamp: 'ar 5/28/2000 17:07'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll:' on: '; print: form.! !

!FormCanvas methodsFor: 'private' stamp: 'RAA 12/17/2000 13:24'!
privateClipRect

	^clipRect! !

!FormCanvas methodsFor: 'private' stamp: 'RAA 12/17/2000 13:25'!
privatePort

	^port! !

!FormCanvas methodsFor: 'private' stamp: 'tpr 3/15/2013 12:42'!
privateWarp: aForm transform: aTransform at: extraOffset sourceRect: sourceRect cellSize: cellSize
	"Warp the given using the appropriate transform and offset."
	| globalRect sourceQuad warp tfm |
	tfm := aTransform.
	globalRect := tfm localBoundsToGlobal: sourceRect.
	sourceQuad := (tfm sourceQuadFor: globalRect) collect:[:p| p - sourceRect topLeft].
	extraOffset ifNotNil:[globalRect := globalRect translateBy: extraOffset].
     warp := (WarpBlt toForm: port destForm)
                combinationRule: Form paint;
                sourceQuad: sourceQuad destRect: (globalRect origin corner: globalRect corner+(1@1));
                clipRect: port clipRect.
	warp cellSize: cellSize.
	warp sourceForm: aForm.
	warp warpBits! !

!FormCanvas methodsFor: 'drawing' stamp: 'ar 9/9/2000 22:18'!
render: anObject
	"Do some 3D operations with the object if possible"
	^self asBalloonCanvas render: anObject! !

!FormCanvas methodsFor: 'initialization' stamp: 'ar 12/30/2010 18:01'!
reset

	origin := 0@0.							"origin of the top-left corner of this cavas"
	form ifNil:[
		"This code path will never be executed after the changes in setForm:
		are installed, so it can be removed in due time."
		clipRect := (0@0 corner: 10000@10000).
	] ifNotNil:[
		clipRect := (0@0 corner: form extent).	"default clipping rectangle"
	].
	self shadowColor: nil.! !

!FormCanvas methodsFor: 'private' stamp: 'ar 5/25/2000 17:25'!
resetGrafPort
	"Private!! Create a new grafPort for a new copy."

	port := self portClass toForm: form.
	port clipRect: clipRect.
! !

!FormCanvas methodsFor: 'private' stamp: 'tpr 9/15/2004 10:28'!
setClearColor: aColor
	"Install a new clear color - e.g., a color is used for clearing the background"
	| clearColor |
	clearColor := aColor ifNil:[Color transparent].
	clearColor isColor ifFalse:[
		(clearColor isKindOf: InfiniteForm) ifFalse:[^self error:'Cannot install color'].
		^port fillPattern: clearColor; combinationRule: Form over].
	"Okay, so clearColor really *is* a color"
	port sourceForm: nil.
	port combinationRule: Form over.
	port fillPattern: clearColor.
	self depth = 8 ifTrue:[
		"Use a stipple pattern"
		port fillColor: (form balancedPatternFor: clearColor)].
! !

!FormCanvas methodsFor: 'private' stamp: 'tpr 9/15/2004 10:28'!
setFillColor: aColor
	"Install a new color used for filling."
	| screen patternWord fillColor |
	fillColor := self shadowColor ifNil:[aColor].
	fillColor ifNil:[fillColor := Color transparent].
	fillColor isColor ifFalse:[
		(fillColor isKindOf: InfiniteForm) ifFalse:[^self error:'Cannot install color'].
		^port fillPattern: fillColor; combinationRule: Form over].
	"Okay, so fillColor really *is* a color"
	port sourceForm: nil.
	fillColor isTranslucent ifFalse:[
		port combinationRule: Form over.
		port fillPattern: fillColor.
		self depth = 8 ifTrue:[
			"In 8 bit depth it's usually a good idea to use a stipple pattern"
			port fillColor: (form balancedPatternFor: fillColor)].
		^self].
	"fillColor is some translucent color"

	self depth > 8 ifTrue:[
		"BitBlt setup for alpha masked transfer"
		port fillPattern: fillColor.
		self depth = 16
			ifTrue:[port alphaBits: fillColor privateAlpha; combinationRule: 30]
			ifFalse:[port combinationRule: Form blend].
		^self].
	"Can't represent actual transparency -- use stipple pattern"
	screen := Color translucentMaskFor: fillColor alpha depth: self depth.
	patternWord := form pixelWordFor: fillColor.
	port fillPattern: (screen collect: [:maskWord | maskWord bitAnd: patternWord]).
	port combinationRule: Form paint.
! !

!FormCanvas methodsFor: 'private' stamp: 'ar 12/30/2010 17:49'!
setForm: aForm

	form := aForm.
	port := self portClass toForm: form.
	self reset.! !

!FormCanvas methodsFor: 'private' stamp: 'ar 6/22/1999 14:06'!
setOrigin: aPoint clipRect: aRectangle

	origin := aPoint.
	clipRect := aRectangle.
	port clipRect: aRectangle.
! !

!FormCanvas methodsFor: 'private' stamp: 'tpr 9/15/2004 10:28'!
setPaintColor: aColor
	"Install a new color used for filling."
	| paintColor screen patternWord |
	paintColor := self shadowColor ifNil:[aColor].
	paintColor ifNil:[paintColor := Color transparent].
	paintColor isColor ifFalse:[
		(paintColor isKindOf: InfiniteForm) ifFalse:[^self error:'Cannot install color'].
		^port fillPattern: paintColor; combinationRule: Form paint].
	"Okay, so paintColor really *is* a color"
	port sourceForm: nil.
	(paintColor isTranslucent) ifFalse:[
		port fillPattern: paintColor.
		port combinationRule: Form paint.
		self depth = 8 ifTrue:[
			port fillColor: (form balancedPatternFor: paintColor)].
		^self].
	"paintColor is translucent color"

	self depth > 8 ifTrue:[
		"BitBlt setup for alpha mapped transfer"
		port fillPattern: paintColor.
		self depth = 16
			ifTrue:[port alphaBits: paintColor privateAlpha; combinationRule: 31]
			ifFalse:[port combinationRule: Form blend].
		^self].

	"Can't represent actual transparency -- use stipple pattern"
	screen := Color translucentMaskFor: paintColor alpha depth: self depth.
	patternWord := form pixelWordFor: paintColor.
	port fillPattern: (screen collect: [:maskWord | maskWord bitAnd: patternWord]).
	port combinationRule: Form paint
! !

!FormCanvas methodsFor: 'accessing' stamp: 'ar 2/17/2000 00:26'!
shadowColor
	^shadowColor! !

!FormCanvas methodsFor: 'accessing' stamp: 'ar 2/17/2000 00:26'!
shadowColor: aColor
	shadowColor := aColor! !

!FormCanvas methodsFor: 'other' stamp: ''!
showAt: pt

	^ form displayAt: pt! !

!FormCanvas methodsFor: 'other' stamp: 'tpr 3/15/2013 12:30'!
showAt: pt invalidRects: updateRects
	| blt |
	blt := (BitBlt toForm: Display)
		sourceForm: form;
		combinationRule: Form over.
	updateRects do:
		[:rect |
		blt sourceRect: rect;
			destOrigin: rect topLeft + pt;
			copyBits]! !

!FormCanvas methodsFor: 'drawing-images' stamp: 'tpr 9/15/2004 10:27'!
stencil: stencilForm at: aPoint sourceRect: sourceRect color: aColor
	"Flood this canvas with aColor wherever stencilForm has non-zero pixels"
	self setPaintColor: aColor.
	port colorMap: stencilForm maskingMap.
	port stencil: stencilForm
		at: aPoint + origin
		sourceRect: sourceRect.! !

!FormCanvas methodsFor: 'drawing-support' stamp: 'bf 10/9/2012 14:58'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock	 smoothing: cellSize

	"Note: This method has been originally copied from TransformationMorph."
	| innerRect patchRect sourceQuad warp start subCanvas |
	(aDisplayTransform isPureTranslation) ifTrue:[
		^aBlock value: (self copyOffset: aDisplayTransform offset negated truncated
							clipRect: aClipRect)
	].
	"Prepare an appropriate warp from patch to innerRect"
	innerRect := aClipRect.
	patchRect := (aDisplayTransform globalBoundsToLocal: innerRect) truncated.
	sourceQuad := (aDisplayTransform sourceQuadFor: innerRect)
					collect: [:p | p - patchRect topLeft].
	warp := self warpFrom: sourceQuad toRect: innerRect.

	"Render the submorphs visible in the clipping rectangle, as patchForm"
	start := (self depth = 1 and: [self isShadowDrawing not])
		"If this is true B&W, then we need a first pass for erasure."
		ifTrue: [1] ifFalse: [2].
	start to: 2 do:
		[:i | "If i=1 we first make a shadow and erase it for opaque whites in B&W"
		subCanvas := self class extent: patchRect extent depth: self depth.
		i=1	ifTrue: [subCanvas shadowColor: Color black.
					warp combinationRule: Form erase]
			ifFalse: [self isShadowDrawing ifTrue:
					[subCanvas shadowColor: self shadowColor].
				warp combinationRule: (self depth = 32
					ifTrue: [Form blendAlphaScaled]
					ifFalse: [Form paint])].
		subCanvas
			translateBy: patchRect topLeft negated
			during: aBlock.
		warp sourceForm: subCanvas form; cellSize: cellSize; warpBits.
		warp sourceForm: nil.  subCanvas := nil "release space for next loop"]
! !

!FormCanvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 03:02'!
translateBy: delta during: aBlock
	"Set a translation only during the execution of aBlock."
	^aBlock value: (self copyOffset: delta)! !

!FormCanvas methodsFor: 'drawing-support' stamp: 'ar 6/17/1999 02:55'!
translateTo: newOrigin clippingTo: aRectangle during: aBlock
	"Set a new origin and clipping rectangle only during the execution of aBlock."
	aBlock value: (self copyOrigin: newOrigin clipRect: aRectangle)! !

!FormCanvas methodsFor: 'other' stamp: 'tpr 3/15/2013 12:42'!
warpFrom: sourceQuad toRect: destRect
        ^ (WarpBlt toForm: port destForm)
                combinationRule: Form paint;
                sourceQuad: sourceQuad destRect: (destRect translateBy: origin);
                clipRect: clipRect! !

!FormCanvas methodsFor: 'drawing-images' stamp: 'ar 12/30/2001 16:36'!
warpImage: aForm transform: aTransform at: extraOffset sourceRect: sourceRect cellSize: cellSize
	"Warp the given using the appropriate transform and offset."
	| tfm |
	tfm := (MatrixTransform2x3 withOffset: origin) composedWithLocal: aTransform.
	^self privateWarp: aForm transform: tfm at: extraOffset sourceRect: sourceRect cellSize: cellSize! !

Canvas subclass: #PluggableCanvas
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!PluggableCanvas commentStamp: '<historical>' prior: 0!
An abstract canvas which modifies the behavior of an underlying canvas in some way.  Subclasses should implement apply:, which takes a one argument block and an actual canvas to draw on.  See apply: for the specific definition.!

PluggableCanvas subclass: #ClippingCanvas
	instanceVariableNames: 'canvas clipRect'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!ClippingCanvas commentStamp: '<historical>' prior: 0!
A modified canvas which clips all drawing commands.!

!ClippingCanvas class methodsFor: 'instance creation' stamp: 'ls 3/20/2000 20:45'!
canvas: aCanvas  clipRect: aRectangle
	^self new canvas: aCanvas  clipRect: aRectangle! !

!ClippingCanvas methodsFor: 'private' stamp: 'ls 3/20/2000 20:44'!
apply: aBlock
	"apply the given block to the inner canvas with clipRect as the clipping rectangle"
	canvas clipBy: clipRect during: aBlock! !

!ClippingCanvas methodsFor: 'initialization' stamp: 'ls 3/20/2000 20:44'!
canvas: aCanvas  clipRect: aRectangle
	canvas := aCanvas.
	clipRect := aRectangle.! !

!ClippingCanvas methodsFor: 'accessing' stamp: 'ls 3/25/2000 22:56'!
clipRect
	^clipRect! !

!ClippingCanvas methodsFor: 'accessing' stamp: 'ls 3/26/2000 14:22'!
contentsOfArea: aRectangle into: aForm
	self flag: #hack.    "ignore the clipping specification for this command.  This is purely so that CachingCanvas will work properly when clipped.  There *has* to be a clean way to do this...."
	^canvas contentsOfArea: aRectangle into: aForm! !

!ClippingCanvas methodsFor: 'accessing' stamp: 'ls 3/20/2000 21:17'!
form
	^canvas form! !

!ClippingCanvas methodsFor: 'testing' stamp: 'ls 3/20/2000 21:17'!
isBalloonCanvas
	^canvas isBalloonCanvas! !

!ClippingCanvas methodsFor: 'testing' stamp: 'ls 3/20/2000 21:18'!
isShadowDrawing
	^canvas isShadowDrawing! !

!ClippingCanvas methodsFor: 'accessing' stamp: 'ls 3/20/2000 21:15'!
shadowColor
	^canvas shadowColor! !

PluggableCanvas subclass: #MultiCanvas
	instanceVariableNames: 'canvases extent depth'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!MultiCanvas commentStamp: '<historical>' prior: 0!
A canvas which forwards drawing commands to sub-canvases.!

!MultiCanvas methodsFor: 'accessing' stamp: 'ls 3/20/2000 20:48'!
addCanvas: aCanvas
	canvases add: aCanvas! !

!MultiCanvas methodsFor: 'initialization' stamp: 'RAA 8/1/2000 13:50'!
allocateForm: extentPoint
	"Allocate a new form which is similar to the receiver and can be used for accelerated blts"
	^Form extent: extentPoint depth: self depth! !

!MultiCanvas methodsFor: 'private' stamp: 'RAA 11/6/2000 14:17'!
apply: aCommand

	self flag: #roundedRudeness.	
	"This rudeness is to help get rounded corners to work right on RemoteCanvases. Since the RemoteCanvas has no other way to read its bits, we are grabbing them from Display for now. To support this, we need to see that the Display is written before any RemoteCanvases"

	canvases do: [ :canvas | 
		(canvas isKindOf: FormCanvas) ifTrue: [aCommand value: canvas]
	].
	canvases do: [ :canvas | 
		(canvas isKindOf: FormCanvas) ifFalse: [aCommand value: canvas]
	].
! !

!MultiCanvas methodsFor: 'accessing' stamp: 'RAA 11/7/2000 17:46'!
clipRect
	
	^super clipRect ifNil: [
		0@0 extent: 5000@5000
	].! !

!MultiCanvas methodsFor: 'accessing' stamp: 'RAA 8/14/2000 10:27'!
contentsOfArea: aRectangle into: aForm

	self apply: [ :c |
		(c isKindOf: FormCanvas) ifTrue: [
			c contentsOfArea: aRectangle into: aForm.
			^aForm
		].
	].
	self apply: [ :c |
		c contentsOfArea: aRectangle into: aForm.
		^aForm.
	].
	^aForm! !

!MultiCanvas methodsFor: 'accessing' stamp: 'ls 4/8/2000 22:35'!
depth
	^depth! !

!MultiCanvas methodsFor: 'initialization' stamp: 'ls 4/8/2000 22:35'!
depth: newDepth
	"set the extent to be used with this canvas"
	depth := newDepth.! !

!MultiCanvas methodsFor: 'accessing' stamp: 'ls 4/8/2000 22:35'!
extent
	^extent! !

!MultiCanvas methodsFor: 'initialization' stamp: 'ls 4/8/2000 22:34'!
extent: newExtent
	"set the extent to be used with this canvas"
	extent := newExtent.! !

!MultiCanvas methodsFor: 'initialization' stamp: 'ls 4/8/2000 22:34'!
initialize
	canvases := Set new.
	extent := 600@400.
	depth := 32. ! !

!MultiCanvas methodsFor: 'accessing' stamp: 'ls 3/20/2000 20:48'!
removeCanvas: aCanvas
	canvases remove: aCanvas ifAbsent: []! !

PluggableCanvas subclass: #NullCanvas
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!NullCanvas commentStamp: '<historical>' prior: 0!
A canvas which ignores all drawing commands.!

!NullCanvas methodsFor: 'drawing-support' stamp: 'ls 3/27/2000 21:41'!
clipBy: region during: aBlock
	"do this in order that timing runs work better"
	aBlock value: self! !

!NullCanvas methodsFor: 'accessing' stamp: 'ls 3/20/2000 21:11'!
clipRect
	^1@1 extent: 99@99! !

!NullCanvas methodsFor: 'copying' stamp: 'ls 3/20/2000 21:26'!
copyClipRect: clipRect
	"who cares what the clipping rectangle is?"
	^self! !

!NullCanvas methodsFor: 'accessing' stamp: 'ls 3/20/2000 21:11'!
extent
	^100@100! !

!NullCanvas methodsFor: 'accessing' stamp: 'ls 3/20/2000 21:12'!
form
	^Form extent: self extent! !

!NullCanvas methodsFor: 'accessing' stamp: 'ls 3/20/2000 21:13'!
origin
	^0@0! !

!NullCanvas methodsFor: 'drawing-support' stamp: 'ls 3/27/2000 21:39'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock smoothing: cellSize
	"do this in order that timing runs work better"
	aBlock value: self! !

!NullCanvas methodsFor: 'drawing-support' stamp: 'ls 3/27/2000 21:40'!
translateBy: delta during: aBlock
	"do this in order that timing runs work better"
	aBlock value: self! !

!PluggableCanvas methodsFor: 'private' stamp: 'ls 3/20/2000 20:46'!
apply: aBlock
	"evaluate aBlock with a canvas to do a drawing command on.  See implementors for examples"! !

!PluggableCanvas methodsFor: 'canvas methods' stamp: 'RAA 11/6/2000 16:33'!
balloonFillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc

	self apply: [ :c | 
		c balloonFillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc
	]! !

!PluggableCanvas methodsFor: 'canvas methods' stamp: 'RAA 7/28/2000 06:52'!
balloonFillRectangle: aRectangle fillStyle: aFillStyle

	self apply: [ :c | c balloonFillRectangle: aRectangle fillStyle: aFillStyle ]! !

!PluggableCanvas methodsFor: 'drawing-support' stamp: 'ls 3/20/2000 19:59'!
clipBy: newClipRect during: aBlock
	self apply: [ :c |
		c clipBy: newClipRect during: aBlock ]! !

!PluggableCanvas methodsFor: 'accessing' stamp: 'ls 3/20/2000 20:02'!
clipRect
	| innerClipRect |
	self apply: [ :c |
		innerClipRect := c clipRect ].
	^innerClipRect! !

!PluggableCanvas methodsFor: 'accessing' stamp: 'ls 3/26/2000 13:57'!
contentsOfArea: aRectangle into: aForm
	self apply: [ :c |
		c contentsOfArea: aRectangle into: aForm ].
	^aForm! !

!PluggableCanvas methodsFor: 'drawing-polygons' stamp: 'ls 3/20/2000 20:01'!
drawPolygon: vertices color: aColor borderWidth: bw borderColor: bc
	self apply: [ :c |
		c drawPolygon: vertices color: aColor borderWidth: bw borderColor: bc ]! !

!PluggableCanvas methodsFor: 'drawing-text' stamp: 'ar 12/31/2001 02:28'!
drawString: s from: firstIndex to: lastIndex in: boundsRect font: fontOrNil color: c
	self apply: [ :clippedCanvas |
		clippedCanvas drawString: s from: firstIndex to: lastIndex in: boundsRect font: fontOrNil color: c]! !

!PluggableCanvas methodsFor: 'accessing' stamp: 'RAA 8/13/2000 18:56'!
extent

	self apply: [ :c | ^c extent ].
! !

!PluggableCanvas methodsFor: 'drawing-ovals' stamp: 'ls 3/20/2000 20:03'!
fillOval: r color: c borderWidth: borderWidth borderColor: borderColor
	self apply: [ :clippedCanvas |
		clippedCanvas fillOval: r color: c borderWidth: borderWidth borderColor: borderColor ]! !

!PluggableCanvas methodsFor: 'drawing-ovals' stamp: 'RAA 11/6/2000 16:32'!
fillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc
	"Fill the given oval."
	self shadowColor ifNotNil:
		[^self fillOval: aRectangle color: aFillStyle asColor borderWidth: bw borderColor: bc].
	(aFillStyle isBitmapFill and:[aFillStyle isKindOf: InfiniteForm]) ifTrue:[
		self flag: #fixThis.
		^self fillOval: aRectangle color: aFillStyle borderWidth: bw borderColor: bc].
	(aFillStyle isSolidFill) ifTrue:[
		^self fillOval: aRectangle color: aFillStyle asColor borderWidth: bw borderColor: bc].
	"Use a BalloonCanvas instead"
	self balloonFillOval: aRectangle fillStyle: aFillStyle borderWidth: bw borderColor: bc! !

!PluggableCanvas methodsFor: 'drawing-rectangles' stamp: 'RAA 8/25/2000 14:40'!
fillRectangle: aRectangle fillStyle: aFillStyle

	| pattern |

	self shadowColor ifNotNil: [^self fillRectangle: aRectangle color: self shadowColor].

	(aFillStyle isKindOf: InfiniteForm) ifTrue: [
		^self infiniteFillRectangle: aRectangle fillStyle: aFillStyle
	].

	aFillStyle isSolidFill ifTrue:[ ^self fillRectangle: aRectangle color: aFillStyle asColor].

	"We have a very special case for filling with infinite forms"
	(aFillStyle isBitmapFill and:[aFillStyle origin = (0@0)]) ifTrue:[
		pattern := aFillStyle form.
		(aFillStyle direction = (pattern width @ 0) 
			and:[aFillStyle normal = (0@pattern height)]) ifTrue:[
				"Can use an InfiniteForm"
				^self fillRectangle: aRectangle color: (InfiniteForm with: pattern)].
	].
	"Use a BalloonCanvas instead"
	self balloonFillRectangle: aRectangle fillStyle: aFillStyle.
! !

!PluggableCanvas methodsFor: 'initialization' stamp: 'ls 3/20/2000 21:16'!
flush
	self apply: [ :c |
		c flush ]! !

!PluggableCanvas methodsFor: 'other' stamp: 'ls 3/20/2000 21:16'!
flushDisplay
	self apply: [ :c |
		c flushDisplay ]! !

!PluggableCanvas methodsFor: 'other' stamp: 'RAA 7/20/2000 16:49'!
forceToScreen: rect

	self apply: [ :c |
		c forceToScreen: rect ]! !

!PluggableCanvas methodsFor: 'drawing-rectangles' stamp: 'ls 3/20/2000 20:04'!
frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth borderColor: borderColor
	self apply: [ :c |
		c frameAndFillRectangle: r fillColor: fillColor borderWidth: borderWidth borderColor: borderColor ]! !

!PluggableCanvas methodsFor: 'drawing-rectangles' stamp: 'topa 3/6/2015 00:06'!
frameAndFillRoundRect: aRectangle radius: cornerRadius fillStyle: fillStyle borderWidth: bw borderColor: bc
	self apply: [ :c |
		c frameAndFillRoundRect: aRectangle radius: cornerRadius fillStyle: fillStyle borderWidth: bw borderColor: bc ]! !

!PluggableCanvas methodsFor: 'private' stamp: 'ls 3/20/2000 20:30'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	self apply:  [ :c |
		c image: aForm at: aPoint sourceRect: sourceRect rule: rule ]! !

!PluggableCanvas methodsFor: 'canvas methods' stamp: 'RAA 8/25/2000 13:34'!
infiniteFillRectangle: aRectangle fillStyle: aFillStyle

	self apply: [ :c | c infiniteFillRectangle: aRectangle fillStyle: aFillStyle ]! !

!PluggableCanvas methodsFor: 'drawing' stamp: 'ls 3/20/2000 20:31'!
line: pt1 to: pt2 brushForm: brush
	self apply: [ :c |
		c line: pt1 to: pt2 brushForm: brush ]! !

!PluggableCanvas methodsFor: 'drawing' stamp: 'ls 3/20/2000 20:31'!
line: pt1 to: pt2 width: w color: c
	self apply: [ :clippedCanvas |
		clippedCanvas line: pt1 to: pt2 width: w color: c ]! !

!PluggableCanvas methodsFor: 'accessing' stamp: 'RAA 8/13/2000 18:57'!
origin

	self apply: [ :c | ^c origin ].
! !

!PluggableCanvas methodsFor: 'drawing-images' stamp: 'ls 3/20/2000 20:32'!
paintImage: aForm at: aPoint
	self apply: [ :c |
		c paintImage: aForm at: aPoint ]! !

!PluggableCanvas methodsFor: 'drawing-images' stamp: 'ls 3/20/2000 20:32'!
paintImage: aForm at: aPoint sourceRect: sourceRect
	self apply: [ :c |
		c paintImage: aForm at: aPoint sourceRect: sourceRect ]! !

!PluggableCanvas methodsFor: 'drawing' stamp: 'ls 3/20/2000 20:33'!
paragraph: paragraph bounds: bounds color: color
	self apply: [ :c |
		c paragraph: paragraph bounds: bounds color: color ]! !

!PluggableCanvas methodsFor: 'drawing' stamp: 'ls 3/20/2000 20:34'!
render: anObject
	self apply: [ :c |
		c render: anObject ]! !

!PluggableCanvas methodsFor: 'accessing' stamp: 'ls 3/20/2000 21:14'!
shadowColor: color
	self apply: [ :c |
		c shadowColor: color ]! !

!PluggableCanvas methodsFor: 'canvas methods' stamp: 'ls 3/25/2000 15:53'!
showAt: pt invalidRects: updateRects
	self apply: [ :c |
		c showAt: pt invalidRects: updateRects ]! !

!PluggableCanvas methodsFor: 'drawing-images' stamp: 'ls 3/20/2000 20:35'!
stencil: stencilForm at: aPoint sourceRect: sourceRect color: aColor
	self apply: [ :c |
		c stencil: stencilForm at: aPoint sourceRect: sourceRect color: aColor ]! !

!PluggableCanvas methodsFor: 'drawing-support' stamp: 'ls 3/20/2000 20:35'!
transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock smoothing: cellSize

	self apply: [ :clippedCanvas |
		clippedCanvas transformBy: aDisplayTransform clippingTo: aClipRect during: aBlock smoothing: cellSize ]! !

!PluggableCanvas methodsFor: 'other' stamp: 'ls 3/20/2000 20:37'!
translateBy: aPoint clippingTo: aRect during: aBlock
	self apply: [ :clippedCanvas |
		clippedCanvas translateBy: aPoint clippingTo: aRect during: aBlock ]! !

!PluggableCanvas methodsFor: 'drawing-support' stamp: 'ls 3/20/2000 20:37'!
translateBy: delta during: aBlock
	self apply: [ :clippedCanvas |
		 clippedCanvas translateBy: delta during: aBlock ]! !

!FlattenEncoder class methodsFor: 'configuring' stamp: 'MPW 1/1/1901 00:08'!
filterSelector
	^#flattenOnStream:
! !

!FlattenEncoder methodsFor: 'writing' stamp: 'MPW 1/1/1901 21:51'!
cr
	^self print:String cr.

! !

!FlattenEncoder methodsFor: 'filter streaming' stamp: 'MPW 1/1/1901 01:32'!
elementSeparator
	^target elementSeparator.! !

!FlattenEncoder methodsFor: 'writing' stamp: 'MPW 1/1/1901 21:50'!
writeArrayedCollection:anArrayedCollection
	^self writeCollectionContents:anArrayedCollection.

! !

!FlattenEncoder methodsFor: 'writing' stamp: 'MPW 1/4/1901 08:26'!
writeCollectionContents:aCollection
    ^self writeCollectionContents:aCollection separator:self elementSeparator iterationMessage:#do:.

! !

!FlattenEncoder methodsFor: 'writing' stamp: 'MPW 1/4/1901 08:26'!
writeCollectionContents:aCollection separator:separator
	^self writeCollectionContents:aCollection separator:separator iterationMessage:#do:.! !

!FlattenEncoder methodsFor: 'writing' stamp: 'MPW 1/4/1901 08:25'!
writeCollectionContents:aCollection separator:separator iterationMessage:op
	| first |
	first := true.
	aCollection perform:op with: [ :each |  first ifFalse:[ self writeObject:separator ]. self write:each. first:=false.].
! !

!FlattenEncoder methodsFor: 'writing' stamp: 'MPW 1/1/1901 01:03'!
writeCollection:aCollection
	^self writeCollectionContents:aCollection.

! !

!FlattenEncoder methodsFor: 'writing' stamp: 'MPW 1/4/1901 08:29'!
writeDictionaryContents:aCollection separator:separator
	^self writeCollectionContents:aCollection separator:separator iterationMessage:#associationsDo:.! !

!FlattenEncoder methodsFor: 'writing' stamp: 'MPW 1/4/1901 08:30'!
writeDictionary:aCollection
	^self writeDictionaryContents:aCollection separator:nil.

! !

!NullEncoder class methodsFor: 'configuring' stamp: 'MPW 1/1/1901 00:02'!
defaultTarget
	^OrderedCollection new.
! !

!NullEncoder class methodsFor: 'configuring' stamp: 'MPW 1/1/1901 00:02'!
filterSelector
	^#writeOnFilterStream:
! !

!NullEncoder class methodsFor: 'processing' stamp: 'MPW 1/1/1901 01:20'!
process:anObject
	^self stream process:anObject.

! !

!NullEncoder class methodsFor: 'creation' stamp: 'MPW 1/1/1901 00:55'!
stream
	^self streamOn:self defaultTarget. 
! !

!NullEncoder class methodsFor: 'creation' stamp: 'MPW 1/1/1901 02:20'!
streamOnFile:fileName
	^self new initWithTarget:(FileStream newFileNamed: fileName).
! !

!NullEncoder class methodsFor: 'creation' stamp: 'MPW 1/1/1901 01:15'!
streamOn:newTargetCollection
	^self new initWithTarget:newTargetCollection.
! !

!NullEncoder class methodsFor: 'creation' stamp: 'MPW 1/1/1901 00:05'!
stream:newTarget
	^self new initWithTarget:newTarget.
! !

!NullEncoder methodsFor: 'accessing' stamp: 'RAA 9/17/2000 11:53'!
close

	^target close.
! !

!NullEncoder methodsFor: 'accessing' stamp: 'MPW 1/1/1901 00:56'!
contents
	^target contents.
! !

!NullEncoder methodsFor: 'writing' stamp: 'MPW 1/1/1901 01:17'!
forward:anObject
	anObject ~= nil ifTrue:[target write:anObject].
! !

!NullEncoder methodsFor: 'initialization' stamp: 'MPW 1/1/1901 00:04'!
initWithTarget:aTarget
	target := aTarget.
	filterSelector := self class filterSelector.
	^self.
! !

!NullEncoder methodsFor: 'processing' stamp: 'MPW 1/1/1901 01:19'!
process:anObject
	self write:anObject.
	^self contents.! !

!NullEncoder methodsFor: 'accessing' stamp: 'MPW 1/1/1901 00:16'!
target
	^target.! !

!NullEncoder methodsFor: 'writing' stamp: 'MPW 1/1/1901 01:16'!
writeObject:anObject
	^self forward:anObject.
! !

!NullEncoder methodsFor: 'writing' stamp: 'mpw 8/13/1999 10:54'!
write:anObject
	filterSelector  ifNil:[filterSelector:=self class filterSelector].
	anObject ifNotNil: [anObject perform:filterSelector with:self].
! !

!Object methodsFor: '*morphic' stamp: 'ar 8/8/2009 12:01'!
asDraggableMorph
	"Converts the receiver into a Morph suitable for dragging"
	^(StringMorph contents: (
			(self respondsTo: #dragLabel) 
				ifTrue:[self dragLabel] 
				ifFalse:[self printString]))
		color: Color white;
		yourself! !

!Object methodsFor: '*morphic' stamp: 'nk 2/26/2004 13:33'!
asMorph
	"Open a morph, as best one can, on the receiver"

	^ self asStringMorph

	"
234 asMorph
(ScriptingSystem formAtKey: #TinyMenu) asMorph
'fred' asMorph
"

! !

!Object methodsFor: '*morphic' stamp: 'nk 2/26/2004 13:35'!
asStringMorph
	"Open a StringMorph, as best one can, on the receiver"

	^ self asStringOrText asStringMorph
! !

!Object methodsFor: '*morphic' stamp: 'nk 2/26/2004 13:35'!
asTextMorph
	"Open a TextMorph, as best one can, on the receiver"

	^ TextMorph new contentsAsIs: self asStringOrText
! !

!Object methodsFor: '*Morphic-Events-Filtering' stamp: 'mt 5/31/2016 12:41'!
filterEvent: anEvent for: aMorphOrNil
	"Get the chance to intercept the event. Note that global event filters might not be provided with the actual target. Returns the event that should be processed any further. If you want to ignore them, call #ignore on them."
	
	^ anEvent "no change"! !

!Object methodsFor: '*Morphic-Explorer' stamp: 'RAA 6/21/1999 11:27'!
hasContentsInExplorer

	^self basicSize > 0 or: [self class allInstVarNames isEmpty not]
! !

!Object methodsFor: '*morphic' stamp: 'cmm 8/19/2013 22:12'!
isPluggableListMorph
	^ false! !

!Object methodsFor: '*morphic' stamp: 'sw 1/29/2002 21:45'!
openAsMorph
	"Open a morph, as best one can, on the receiver"

	^ self asMorph openInHand

"
234 openAsMorph
(ScriptingSystem formAtKey: #TinyMenu) openAsMorph
'fred' openAsMorph
"! !

Object subclass: #PluggableEventFilter
	instanceVariableNames: 'filterBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Events'!

!PluggableEventFilter class methodsFor: 'instance creation' stamp: 'mt 5/17/2016 22:48'!
on: filterBlock

	^ self new filterBlock: filterBlock! !

!PluggableEventFilter methodsFor: 'accessing' stamp: 'mt 5/17/2016 22:44'!
filterBlock

	^ filterBlock! !

!PluggableEventFilter methodsFor: 'accessing' stamp: 'mt 5/17/2016 22:44'!
filterBlock: anObject

	filterBlock := anObject! !

!PluggableEventFilter methodsFor: 'events-filtering' stamp: 'mt 5/19/2016 11:49'!
filterEvent: event for: target

	^ self filterBlock
		ifNil: [event]
		ifNotNil: [:b | b cull: event cull: target]! !

Object subclass: #Presenter
	instanceVariableNames: ''
	classVariableNames: 'DefaultPresenterClass'
	poolDictionaries: ''
	category: 'Morphic-Worlds'!

!Presenter commentStamp: '<historical>' prior: 0!
Optionally associated with a PasteUpMorph, provides a local scope for the running of scripts.

Once more valuable, may be again, but at present occupies primarily a historical niche.

Maintains a playerList cache.

Holds, optionally three 'standard items' -- standardPlayer standardPlayfield standardPalette -- originally providing idiomatic support of ongoing squeak-team internal work, but now extended to more general applicability.

   !

!Presenter class methodsFor: 'accessing' stamp: 'ar 1/3/2010 18:10'!
defaultPresenterClass
	"The default presenter class to use"
	^DefaultPresenterClass ifNil:[self]! !

!Presenter class methodsFor: 'accessing' stamp: 'ar 1/3/2010 18:10'!
defaultPresenterClass: aPresenterClass
	"The default presenter class to use"
	DefaultPresenterClass := aPresenterClass! !

!Presenter methodsFor: 'stubs' stamp: 'sw 5/1/2005 00:22'!
allCurrentlyTickingScriptInstantiations
	"Answer a list of ScriptInstantiation objects representing all the scripts within the scope of the receiver which are currently ticking."

	^ Array streamContents:
		[:aStream | 
			self allExtantPlayers do:
				[:aPlayer | aPlayer instantiatedUserScriptsDo:
					[:aScriptInstantiation |
						aScriptInstantiation status == #ticking ifTrue:
							[aStream nextPut: aScriptInstantiation]]]]! !

!Presenter methodsFor: 'stubs' stamp: 'ar 1/3/2010 18:15'!
allExtantPlayers
	^#()! !

!Presenter methodsFor: 'access' stamp: 'ar 1/3/2010 19:11'!
associatedMorph: m! !

!Presenter methodsFor: 'stubs' stamp: 'KR 5/1/2008 01:54'!
browseAllScriptsTextually
	"Open a method-list browser on all the scripts in the project"

	| aList aMethodList |
	self flushPlayerListCache.  "Just to be certain we get everything"

	(aList _ self uniclassesAndCounts) size == 0 ifTrue: [^ self inform: 'there are no scripted players' translated].
	aMethodList _ OrderedCollection new.
	aList do:
		[:aPair | aPair first addMethodReferencesTo: aMethodList].
	aMethodList size > 0 ifFalse: [^ self inform: 'there are no scripts in this project!!' translated].
	
	SystemNavigation new 
		browseMessageList: aMethodList 
		name: 'All scripts in this project' 
		autoSelect: nil

"
ActiveWorld presenter browseAllScriptsTextually
"! !

!Presenter methodsFor: 'stubs' stamp: 'ar 1/3/2010 18:13'!
currentlyViewing: aPlayer
	^false! !

!Presenter methodsFor: 'stubs' stamp: 'ar 1/3/2010 18:15'!
drawingJustCompleted: aSketch! !

!Presenter methodsFor: 'stubs' stamp: 'ar 1/3/2010 18:12'!
flushPlayerListCache! !

!Presenter methodsFor: 'stubs' stamp: 'ar 1/4/2010 00:18'!
morph: aMorph droppedIntoPasteUpMorph: aPasteUpMorph! !

!Presenter methodsFor: 'stubs' stamp: 'ar 1/3/2010 18:14'!
ownStandardPalette
	^nil! !

!Presenter methodsFor: 'stubs' stamp: 'ar 1/3/2010 18:14'!
positionStandardPlayer! !

!Presenter methodsFor: 'stubs' stamp: 'ar 1/3/2010 18:13'!
viewMorph: aMorph
	aMorph inspect.
! !

!Presenter methodsFor: 'stubs' stamp: 'ar 1/3/2010 18:14'!
viewObjectDirectly: aMorph
	aMorph inspect.
! !

Object subclass: #StandardScriptingSystem
	instanceVariableNames: ''
	classVariableNames: 'ClassVarNamesInUse FormDictionary HelpStrings StandardPartsBin'
	poolDictionaries: ''
	category: 'Morphic-Worlds'!

!StandardScriptingSystem commentStamp: '<historical>' prior: 0!
An instance of this is installed as the value of the global variable "ScriptingSystem".  Client subclasses are invited, such as one used internally by squeak team for ongoing internal work.!

!StandardScriptingSystem class methodsFor: 'class initialization' stamp: 'ar 2/27/2010 00:19'!
cleanUp: agressive
	"Clean up unreferenced players. If agressive, reinitialize and nuke players"

	self removeUnreferencedPlayers.
	agressive ifTrue:[
		References keys do: [:k | References removeKey: k].
		self initialize.
	].! !

!StandardScriptingSystem class methodsFor: 'class initialization' stamp: 'asm 4/11/2003 19:04'!
initialize
	"Initialize the scripting system.  Sometimes this method is vacuously changed just to get it in a changeset so that its invocation will occur as part of an update"

	(self environment at: #ScriptingSystem ifAbsent: [nil]) ifNil:
		[self environment at: #ScriptingSystem put: self new].

	ScriptingSystem
		initializeHelpStrings.

	self registerInFlapsRegistry.

"StandardScriptingSystem initialize"! !

!StandardScriptingSystem class methodsFor: 'class initialization' stamp: 'cmm 1/18/2014 17:08'!
removeUnreferencedPlayers
	"Remove existing but unreferenced player references"
	"StandardScriptingSystem removeUnreferencedPlayers"
	References keys do:
		[ : key | | ref |
		ref := References at: key.
		((ref respondsTo: #costume) and: [ ref costume pasteUpMorph isNil ]) ifTrue: [ References removeKey: key ] ]! !

!StandardScriptingSystem methodsFor: 'utilities' stamp: 'sw 10/30/2000 16:33'!
allClassVarNamesInSystem
	"Compute and answer a set of all the class variable names known to the sytem from any class"

	| aList |
	aList := OrderedCollection new.
	Object withAllSubclasses do:
		[:c | aList addAll: c allClassVarNames].
	^ aList asSet

	"ScriptingSystem allClassVarNamesInSystem"
! !

!StandardScriptingSystem methodsFor: 'utilities' stamp: 'mt 8/26/2015 17:44'!
customizeForEToyUsers: aBoolean
	"If aBoolean is true, set things up for etoy users.  If it's false, unset some of those things.  Some things are set when switching into etoy mode but not reversed when switching out of etoy mode."
 
	#(	
		(allowEtoyUserCustomEvents	no		reverse)
		(balloonHelpEnabled			yes		dontReverse)
		(debugHaloHandle			no		reverse)
		(modalColorPickers			yes		dontReverse)
		(oliveHandleForScriptedObjects	no	dontReverse)
		(uniqueNamesInHalos		yes		reverse)
		(useUndo					yes		dontReverse)
		(infiniteUndo				no		dontReverse)
		(warnIfNoChangesFile		no		reverse)
		(warnIfNoSourcesFile		no		reverse)) do:
			[:trip |
				(aBoolean or: [trip third == #reverse]) ifTrue:
					[Preferences setFlag: trip first toValue:
						((trip second == #yes) & aBoolean) | ((trip second == #no) & aBoolean not)]]! !

!StandardScriptingSystem methodsFor: 'form dictionary' stamp: 'ar 3/3/2001 19:45'!
deletePrivateGraphics
	"ScriptingSystem deletePrivateGraphics"
	self deletePrivateGraphics: self privateGraphics
		afterStoringToFileNamed: 'disGraphics'! !

!StandardScriptingSystem methodsFor: 'form dictionary' stamp: 'nice 12/26/2009 21:56'!
deletePrivateGraphics: nameList afterStoringToFileNamed: aFileName
	"This method is used to strip private graphics from the FormDictionary and store them on a file of the given name"

	| replacement toRemove aReferenceStream |
	toRemove := Dictionary new.
	replacement := FormDictionary at: #Gets.

	nameList do:
		[:aKey |
			| keySymbol |
			keySymbol := aKey asSymbol.
			(toRemove at: keySymbol put: (self formAtKey: keySymbol)).
			FormDictionary at: keySymbol put: replacement].

	aReferenceStream := ReferenceStream fileNamed: aFileName.
	aReferenceStream nextPut: toRemove.
	aReferenceStream close! !

!StandardScriptingSystem methodsFor: 'form dictionary' stamp: 'tfel 8/1/2016 15:58'!
formAtKey: aString
	"Answer the form saved under the given key"

	Symbol hasInterned: aString ifTrue:
		[:aKey | ^ FormDictionary at: aKey ifAbsent: [FormDictionary at: #Cat]].
	^ FormDictionary at: #Cat! !

!StandardScriptingSystem methodsFor: 'form dictionary' stamp: 'ar 3/3/2001 19:46'!
formAtKey: aKey extent: extent depth: depth
	"ScriptingSystem saveForm: (TileMorph downPicture) atKey: 'downArrow'"
	^ FormDictionary at: aKey asSymbol ifAbsent: [Form extent: extent depth: depth]! !

!StandardScriptingSystem methodsFor: 'form dictionary' stamp: 'ar 3/3/2001 19:49'!
formDictionary
	^FormDictionary! !

!StandardScriptingSystem methodsFor: 'help dictionary' stamp: 'dgd 9/1/2003 14:25'!
helpStringOrNilFor: aSymbol 
	"If my HelpStrings dictionary has an entry at the given symbol, 
	answer that entry's value, else answer nil"
	HelpStrings
		at: aSymbol
		ifPresent:[:string | ^ string translated].
^ nil! !

!StandardScriptingSystem methodsFor: 'help dictionary' stamp: 'nice 10/20/2009 21:03'!
initializeHelpStrings
	"Initialize the data structure that determines, for the etoy system, help messages for various scripting elements.  The structure is built up by letting every Morph subclass contribute elements simply by implementing method #helpContributions.  Consult implementors of #helpContributions for examples of how this goes."

	"ScriptingSystem initializeHelpStrings"

	| aDictionary |
	aDictionary := IdentityDictionary new.  
	"For safety, the new copy is built up in this temp first, so that if an error occurs during the creation of the structure, the old version will remain remain in place"

	Morph withAllSubclasses do:
		[:aClass | (aClass class includesSelector: #helpContributions)
			ifTrue:
				[aClass helpContributions do:
					[:pair | aDictionary at: pair first put: pair second]]].

		HelpStrings := aDictionary! !

!StandardScriptingSystem methodsFor: 'form dictionary' stamp: 'sd 5/11/2003 21:32'!
inspectFormDictionary
	"ScriptingSystem inspectFormDictionary"
	
	GraphicalDictionaryMenu openOn: FormDictionary withLabel: 'Testing One Two Three'! !

!StandardScriptingSystem methodsFor: 'form dictionary' stamp: 'ar 3/3/2001 19:44'!
privateGraphics
	"ScriptingSystem deletePrivateGraphics"
	^#(#BadgeMiniPic #BadgePic #Broom #CedarPic #CollagePic #CoverMain #CoverSpiral #CoverTexture #Fred #ImagiPic #KayaPic #StudioPic)! !

!StandardScriptingSystem methodsFor: 'form dictionary' stamp: 'sw 10/6/1999 20:57'!
saveForm: aForm atKey: aKey
	FormDictionary at: aKey asSymbol put: aForm! !

!StandardScriptingSystem methodsFor: 'utilities' stamp: 'sw 4/6/2005 09:30'!
soundNamesToSuppress
	"Answer a list of sound-names that are not to be offered in sound-choice pop-ups unless they are the current choice"

	^ #('scrape' 'scritch' 'peaks')! !

!StandardScriptingSystem methodsFor: 'form dictionary' stamp: 'sw 10/24/1998 14:12'!
squeakyMouseForm
	^ self formAtKey: 'squeakyMouse'

"
	ScriptingSystem saveForm: (Form
	extent: 30@29
	depth: 16
	fromArray: #( 1811114995 1878286257 2012637171 1811180532 1811180533 1811179508 1811180532 1811179508 1744006133 1878289396 1811180533 1878289396 1744007156 1674736630 1744006132 1811114995 1811181556 1744006131 1811246068 1811180532 1811179508 1811180532 1744071668 1811113972 1811180532 1811180532 1811179507 1878288338 1945529332 1744071668 1743941620 1811112945 1811179506 1811114995 1744006131 1744006130 1744005106 1811048434 1811113969 1743939570 1811179506 1743939571 1676833782 1676765171 1811047410 1744006131 1811048435 1811116020 1811180531 1743939571 1811048435 1743939570 1743939570 1743939570 1743940594 1744005106 1811181556 1811180532 1676766196 1743939570 1878420468 1676963830 1189896082 1811245044 1744137204 1744070644 1811179508 1811113971 1743939571 1811179508 1811246070 1811309524 1811302093 1811310580 1811246068 1674867703 1744049472 1120606594 1118465013 1744137205 1811179508 1811180532 1744071667 1744006132 1811112947 1811247095 1605584589 358761132 289435638 1676830707 1741975543 1462778473 1811312631 702891724 1811310548 1945528308 1811178450 1945528307 1878288372 1878353875 1878421494 1051471335 1809213397 1118524175 1811246068 1945659348 1185698607 1878486005 1672694510 1118531574 1607626741 1878420467 1811180533 1743942645 1744072693 1811301035 1185770487 1878486006 1324239597 1811180533 1811116019 1120623438 1878352818 1945462739 704868339 1878289395 1811049459 1878221808 1878223859 1743876083 1811162563 1945463796 1811181556 1464746666 1811116018 1809019893 1120551562 1945464821 1741844468 1466842760 1878289395 1811048434 1811050483 1811050483 1878223859 1049188174 1741910004 1811181556 1256998634 1811114994 1878289396 1466840647 1744007156 1744006131 1676877216 1743940596 1878222835 1743938545 1878351792 1676833781 358641652 1743940596 1811050484 845566798 1811113970 1811114995 1811163652 1811112913 1878420468 1878282028 1811179506 1607560178 1878289395 1676900342 1878351825 1466853330 1811113971 1811116019 635659217 1811179506 1811245045 1676942754 1744137206 1744201717 1676962806 1676962805 1811310581 1676896245 1744199635 1811376117 1744072695 1744005109 1811244019 499279861 1811310581 1811244020 1811293668 1399943159 1605528567 1744136181 982063522 986342388 1744070645 1744189066 430063308 1744071669 1744070644 1744067504 566519797 1744136181 1744137205 1743999854 912813044 1811311606 1742162607 4195488 283139922 1945531382 1253113857 144710948 1601400791 1811246069 1811167879 1464821747 1744136180 1674799094 1811178482 843473875 1811311606 1878533542 2106790 2080066222 1876193270 696845376 627472380 1185772536 1878355957 1743990309 1744007157 1676898294 1744006132 1811114996 1743941620 1811180533 1809204941 4194368 4217681 1878290421 1252982848 4194336 1670540278 1739811795 1878353906 1744006131 1811179506 1744007157 1744005106 1945462771 1811182582 1811311574 1393641133 1462856629 2012638196 1876382449 1112301394 1742041045 1945596917 1676833781 1811113970 1811179507 1811180532 1672705014 1674735606 1672697648 1945725943 1878551479 1809215479 1811312629 1809216504 1809215479 1809215478 1462853490 1878487029 1744007158 1744005075 1811239726 704979363 495004132 700789287 562372997 631646663 1739998892 4194400 1116497846 698688932 562375109 770124262 633609569 495070758 1257010166 562315916 1809279958 2012894002 1047280171 980237901 910966381 1668677696 4194400 6314867 1047281260 908804749 910968495 1393719290 1809279959 1185750370 1809214455 1878469062 423836236 1532188466 1601592148 1462986647 1672937568 4194368 6319062 1603622706 1601525554 1601522417 1047336194 770206679 1878487031 1878409899 977955830 1809145716 1118586509 980105834 980045584 1811372914 980104778 1605526483 1395605131 910769804 1118651052 1534358520 1809136234 1118596053 1532059506 1878485973 1326456163 1945660374 1742106615 1811311607 1945725942 1742107641 1744072693 1811311605 1744203767 1878551543 564478604 1878553591 1603428242 1811048433 1811049459 1051290611 1744006131 1811049459 1878156273 1743874034 1744007156 1743874033 1811048434 1811113970 1743939571 1743933228 1603301363 1743875059 1811049458 1945461745 1811181556 1811113971 1811049458 1811048434 1811116020 1878287346 1878223857 1743940594 1744006130 1744007157 1945395153 1945400309 1811048434 1743810547 1676765170 1878353906 1811113970 1743874032 1810983921 1743874033 1811113971 1676765169 1743874034 1743940593 1743939569 1811047409 1676765168 1743940595 1810981872 1945397235 1607560179 1743941620 1810982897 1810983921 1811048433 1744007155 1743875059 1811048434 1743875058 1743939568 1676832754 1811116019 1811114994 1811244019 1676962805 1677029367 1811244020 1744005106 1743940594 1811246068 1744070645 1676961781 1744004084 1676897269 1811180533 1878353908 1744004083 1744070645)
	offset: 0@0) atKey: 'squeakyMouse'"! !

!StandardScriptingSystem methodsFor: 'utilities' stamp: 'sw 11/26/1999 15:44'!
stripGraphicsForExternalRelease
	"ScriptingSystem stripGraphicsForExternalRelease"

	|  replacement |
	replacement := FormDictionary at: #Gets.

	#('BadgeMiniPic' 'BadgePic' 'Broom' 'CedarPic' 'CollagePic' 'CoverMain' 'CoverSpiral' 'CoverTexture' 'Fred' 'ImagiPic' 'KayaPic' 'StudioPic')
		do:
			[:aKey | FormDictionary at: aKey asSymbol put: replacement]! !

Object subclass: #StringMorphAttributeScanner
	instanceVariableNames: 'fontNumber textColor emphasis alignment actualFont indent kern'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!StringMorphAttributeScanner commentStamp: '<historical>' prior: 0!
A StringMorphAttributeScanner provides the interface of a CharacterScanner so that text attributes may be collected from a Text and used elsewhere, like in setting the attributes of a StringMorph.
!

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 13:12'!
actualFont
	"Answer the value of actualFont"

	^ actualFont ifNil: [ TextStyle defaultFont ]! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 12:40'!
addEmphasis: anInteger
	"Set the value of emphasis"

	emphasis := emphasis bitOr: anInteger! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 12:41'!
addKern: kernDelta
	"Set the current kern amount."
	kern := kern + kernDelta! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 12:34'!
alignment
	"Answer the value of alignment"

	^ alignment! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 12:34'!
emphasis
	"Answer the value of emphasis"

	^ emphasis! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 13:14'!
font
	"Answer the value of font"

	^self textStyle fontAt: self fontNumber! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 13:11'!
fontNumber
	"Answer the value of font"

	^ fontNumber! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 12:34'!
indent
	"Answer the value of indent"

	^ indent! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 12:37'!
indentationLevel: anInteger
	"Set the value of indent"

	indent := anInteger! !

!StringMorphAttributeScanner methodsFor: 'initialize-release' stamp: 'nk 2/26/2004 13:10'!
initialize
	emphasis := 0.
	indent := 0.
	kern := 0.
	fontNumber := 1.
	actualFont := TextStyle defaultFont! !

!StringMorphAttributeScanner methodsFor: 'string morph' stamp: 'nk 2/26/2004 13:09'!
initializeFromStringMorph: aStringMorph
	| style |
	actualFont := aStringMorph font ifNil: [ TextStyle defaultFont ].
	style := actualFont textStyle.
	emphasis := actualFont emphasis.
	fontNumber := (style fontIndexOf: actualFont) ifNil: [ 1 ].
	textColor := aStringMorph color.
! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 12:34'!
kern
	"Answer the value of kern"

	^ kern! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'ul 12/12/2009 14:06'!
setActualFont: aFont
	"Set the value of actualFont, from a TextFontReference"

	actualFont := aFont.
	aFont textStyle ifNotNil: [ :ts | fontNumber := ts fontIndexOf: aFont ]! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 12:39'!
setAlignment: aSymbol
	"Set the value of alignment"

	alignment := aSymbol! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 13:10'!
setFont: fontNum
	"Set the value of font"

	fontNumber := fontNum! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 12:34'!
textColor
	"Answer the value of textColor"

	^ textColor! !

!StringMorphAttributeScanner methodsFor: 'scanning' stamp: 'nk 2/26/2004 12:34'!
textColor: anObject
	"Set the value of textColor"

	textColor := anObject! !

!StringMorphAttributeScanner methodsFor: 'accessing' stamp: 'nk 2/26/2004 13:12'!
textStyle
	^self actualFont textStyle ifNil: [ TextStyle default ]! !

Object subclass: #TextContainer
	instanceVariableNames: 'textMorph shadowForm vertProfile minWidth rectangleCache fillsOwner avoidsOcclusions'
	classVariableNames: 'OuterMargin'
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextContainer commentStamp: '<historical>' prior: 0!
A TextContainer models the shape of an ownerMorph, possibly occluded by one or more occludingMorphs, and scans this shape to provide a list of rectangles suitable for layout of text.  It does this by displaying the shadow of the ownerMorph in black, and any occludingMorphs in white, on its shadowForm.  It then scans horizontal strips of appropriate height to find unbroken intervals of black, greater than minWidth in extent.  Conputation of the rectangles is done on demand, and results are cached so that text can be redisplayed without having to recompute the rectangles.!

!TextContainer class methodsFor: 'class initialization' stamp: 'di 11/16/97 09:25'!
initialize    "TextContainer initialize"
	OuterMargin := 2! !

!TextContainer methodsFor: 'access' stamp: 'di 11/4/97 14:05'!
avoidsOcclusions
	^ avoidsOcclusions ifNil: [false]! !

!TextContainer methodsFor: 'access' stamp: 'di 11/13/97 14:45'!
avoidsOcclusions: aBoolean
	avoidsOcclusions := aBoolean.
	self releaseCachedState! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 10/27/97 23:09'!
bottom
	"Note we should really check for contiguous pixels here"
	^ (self vertProfile findLast: [:count | count >= minWidth])
		+ shadowForm offset y! !

!TextContainer methodsFor: 'private' stamp: 'nice 12/27/2009 22:39'!
bounds
	| bounds theText |
	self fillsOwner ifFalse: [^ textMorph textBounds].
	theText := textMorph.
	bounds := theText owner innerBounds.
	bounds := bounds insetBy: (textMorph valueOfProperty: #margins ifAbsent: [1@1]).
	theText owner submorphsBehind: theText do:
		[:m | bounds swallow: m fullBounds].
	^ bounds! !

!TextContainer methodsFor: 'private' stamp: 'nice 12/27/2009 03:10'!
computeShadow
	| canvas bounds theText |
	bounds := self bounds.
	theText := textMorph.
	canvas := (Display defaultCanvasClass extent: bounds extent depth: 1)
			shadowColor: Color black.
	canvas translateBy: bounds topLeft negated during:[:tempCanvas| | back |
		self fillsOwner
			ifTrue: [tempCanvas fullDrawMorph: (theText owner copyWithoutSubmorph: theText)]
			ifFalse: [tempCanvas fillRectangle: textMorph bounds color: Color black].
		self avoidsOcclusions ifTrue:
			[back := tempCanvas form deepCopy.
			tempCanvas form fillWhite.
			theText owner submorphsInFrontOf: theText do:
				[:m | (textMorph isLinkedTo: m)
					ifTrue: []
					ifFalse: [tempCanvas fullDrawMorph: m]].
			back displayOn: tempCanvas form at: 0@0 rule: Form reverse].
	].
	shadowForm := canvas form offset: bounds topLeft.
	vertProfile := shadowForm  yTallyPixelValue: 1 orNot: false.
	rectangleCache := Dictionary new.
	^ shadowForm! !

!TextContainer methodsFor: 'access' stamp: 'di 11/4/97 14:05'!
fillsOwner
	^ fillsOwner ifNil: [true]! !

!TextContainer methodsFor: 'access' stamp: 'di 11/13/97 14:45'!
fillsOwner: aBoolean
	fillsOwner := aBoolean.
	self releaseCachedState! !

!TextContainer methodsFor: 'private' stamp: 'di 11/4/97 14:06'!
for: aTextMorph minWidth: wid
	textMorph := aTextMorph.
	minWidth := wid.
	fillsOwner := true.
	avoidsOcclusions := false.! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 10/28/97 18:33'!
left 
	^ textMorph owner left! !

!TextContainer methodsFor: 'access' stamp: 'jm 11/19/97 19:28'!
paragraphClass
	^ NewParagraph! !

!TextContainer methodsFor: 'container protocol' stamp: 'dgd 2/22/2003 19:06'!
rectanglesAt: lineY height: lineHeight 
	"Return a list of rectangles that are at least minWidth wide
	in the specified horizontal strip of the shadowForm.
	Cache the results for later retrieval if the owner does not change."

	| hProfile rects thisWidth thisX count pair outerWidth lineRect lineForm |
	pair := Array with: lineY with: lineHeight.
	rects := rectangleCache at: pair ifAbsent: [nil].
	rects ifNotNil: [^rects].
	outerWidth := minWidth + (2 * OuterMargin).
	self shadowForm.	"Compute the shape"
	lineRect := 0 @ (lineY - shadowForm offset y) 
				extent: shadowForm width @ lineHeight.
	lineForm := shadowForm copy: lineRect.

	"Check for a full line -- frequent case"
	(lineForm tallyPixelValues second) = lineRect area 
		ifTrue: 
			[rects := Array with: (shadowForm offset x @ lineY extent: lineRect extent)]
		ifFalse: 
			["No such luck -- scan the horizontal profile for segments of minWidth"

			hProfile := lineForm xTallyPixelValue: 1 orNot: false.
			rects := OrderedCollection new.
			thisWidth := 0.
			thisX := 0.
			1 to: hProfile size
				do: 
					[:i | 
					count := hProfile at: i.
					count >= lineHeight 
						ifTrue: [thisWidth := thisWidth + 1]
						ifFalse: 
							[thisWidth >= outerWidth 
								ifTrue: 
									[rects addLast: ((thisX + shadowForm offset x) @ lineY 
												extent: thisWidth @ lineHeight)].
							thisWidth := 0.
							thisX := i]].
			thisWidth >= outerWidth 
				ifTrue: 
					[rects addLast: ((thisX + shadowForm offset x) @ lineY 
								extent: thisWidth @ lineHeight)]].
	rects := rects collect: [:r | r insetBy: OuterMargin @ 0].
	rectangleCache at: pair put: rects.
	^rects! !

!TextContainer methodsFor: 'access' stamp: 'di 11/16/97 09:39'!
releaseCachedState

	shadowForm := nil.
	vertProfile := nil.
	rectangleCache := Dictionary new.
! !

!TextContainer methodsFor: 'private' stamp: 'di 10/27/97 23:09'!
shadowForm
	shadowForm ifNil: [self computeShadow].
	^ shadowForm! !

!TextContainer methodsFor: 'access' stamp: 'tk 8/31/2000 14:50'!
textMorph
	^ textMorph! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 11/16/97 09:33'!
top
	"Note we should really check for contiguous pixels here"
	| outerWidth |
	outerWidth := minWidth + (2*OuterMargin).
	^ (self vertProfile findFirst: [:count | count >= outerWidth]) - 1
		+ shadowForm offset y! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 10/28/97 18:33'!
topLeft  "for compatibility"
	^ textMorph owner topLeft! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 11/7/97 12:01'!
translateBy: delta
	self releaseCachedState! !

!TextContainer methodsFor: 'private' stamp: 'di 10/27/97 23:08'!
vertProfile
	vertProfile ifNil: [self computeShadow].
	^ vertProfile! !

!TextContainer methodsFor: 'container protocol' stamp: 'di 10/28/97 18:32'!
width  "for compatibility"
	^ textMorph owner width! !

Object subclass: #TextEditorCommand
	instanceVariableNames: 'interval message paragraph selection contentsBefore contentsAfter intervalBefore intervalAfter valid messageToUndo messageToRedo intervalBetween type isCompositeUndo isCompositeRedo'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:03'!
contentsAfter

	^ contentsAfter! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:03'!
contentsAfter: anObject

	contentsAfter := anObject! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:03'!
contentsBefore

	^ contentsBefore! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:03'!
contentsBefore: anObject

	contentsBefore := anObject! !

!TextEditorCommand methodsFor: 'testing' stamp: 'mt 12/21/2015 13:56'!
hasInsertedSomething

	^ self contentsBefore isEmpty! !

!TextEditorCommand methodsFor: 'testing' stamp: 'mt 11/10/2015 09:58'!
hasReplacedSomething

	^ self contentsBefore size > 0! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:08'!
intervalAfter

	^ intervalAfter! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:08'!
intervalAfter: anObject

	intervalAfter := anObject! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:08'!
intervalBefore

	^ intervalBefore! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:08'!
intervalBefore: anObject

	intervalBefore := anObject! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:51'!
intervalBetween

	^ intervalBetween! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:51'!
intervalBetween: anObject

	intervalBetween := anObject! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/10/2015 10:28'!
isCompositeRedo

	^ isCompositeRedo! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/10/2015 10:29'!
isCompositeRedo: boolean

	isCompositeRedo := boolean.! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/10/2015 10:28'!
isCompositeUndo

	^ isCompositeUndo! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/10/2015 10:29'!
isCompositeUndo: boolean

	isCompositeUndo := boolean.! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:45'!
messageToRedo
	^ messageToRedo! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:45'!
messageToRedo: msg
	messageToRedo := msg.! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 14:45'!
messageToUndo
	^ messageToUndo! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 15:26'!
messageToUndo: msg
	messageToUndo := msg.! !

!TextEditorCommand methodsFor: 'copying' stamp: 'mt 11/8/2015 13:15'!
postCopy

	super postCopy.
	
	contentsAfter := contentsAfter copy.
	contentsBefore := contentsBefore copy.
	intervalAfter := intervalAfter copy.
	intervalBefore := intervalBefore copy.
	intervalBetween := intervalBetween copy.

	messageToUndo := messageToUndo copy.
	messageToRedo := messageToRedo copy.! !

!TextEditorCommand methodsFor: 'undo/redo' stamp: 'mt 11/7/2015 17:04'!
redoIn: editor

	self messageToRedo sendTo: editor.! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 21:41'!
type

	^ type! !

!TextEditorCommand methodsFor: 'accessing' stamp: 'mt 11/7/2015 21:41'!
type: symbol

	type := symbol.! !

!TextEditorCommand methodsFor: 'undo/redo' stamp: 'mt 11/7/2015 17:26'!
undoIn: editor

	self messageToUndo sendTo: editor.! !

Object subclass: #TextEditorCommandHistory
	instanceVariableNames: 'commands currentIndex'
	classVariableNames: 'MaximumTextHistoryDepth'
	poolDictionaries: ''
	category: 'Morphic-Text Support'!

!TextEditorCommandHistory class methodsFor: 'preferences' stamp: 'mt 11/13/2015 09:51'!
maximumTextHistoryDepth
	<preference: 'Maximum text edit (undo/redo) history depth'
		categoryList:  #(editing Morphic)
		description: 'Specifies the number of editor commands that will be stored in a text editor'
		type: #Number>
	^ MaximumTextHistoryDepth ifNil: [500]! !

!TextEditorCommandHistory class methodsFor: 'preferences' stamp: 'mt 11/8/2015 09:39'!
maximumTextHistoryDepth: number

	MaximumTextHistoryDepth := number.! !

!TextEditorCommandHistory methodsFor: 'undo/redo' stamp: 'mt 11/8/2015 10:18'!
beginRemember: command

	commands := commands
		copyFrom: (1 max: (currentIndex + 2 - self class maximumTextHistoryDepth))
		to: (currentIndex min: commands size).
	commands := commands, {command}.
	currentIndex := commands size - 1. "Select the new command."! !

!TextEditorCommandHistory methodsFor: 'accessing' stamp: 'mt 11/8/2015 10:03'!
current

	^ self next! !

!TextEditorCommandHistory methodsFor: 'enumeration' stamp: 'mt 11/8/2015 15:04'!
detect: block

	self do: [:command |
		(block value: command) ifTrue: [^ command]].
	^ nil! !

!TextEditorCommandHistory methodsFor: 'enumeration' stamp: 'mt 11/26/2015 16:50'!
do: block
	"Goes backward in history."
	
	((currentIndex min: commands size) to: 1 by: -1)
		do: [:i | block value: (commands at: i)].! !

!TextEditorCommandHistory methodsFor: 'undo/redo' stamp: 'mt 11/8/2015 10:01'!
finishRemember

	currentIndex := commands size.! !

!TextEditorCommandHistory methodsFor: 'testing' stamp: 'mt 2/5/2016 15:07'!
hasCursorMoved: cursorIndex

	^ self hasPrevious and: [self previous intervalAfter first ~= cursorIndex]! !

!TextEditorCommandHistory methodsFor: 'testing' stamp: 'mt 12/21/2015 13:55'!
hasInsertedSomething

	^ self hasPrevious and: [self previous hasInsertedSomething]! !

!TextEditorCommandHistory methodsFor: 'accessing' stamp: 'mt 11/7/2015 23:47'!
hasNext

	^ currentIndex < commands size! !

!TextEditorCommandHistory methodsFor: 'accessing' stamp: 'mt 11/8/2015 09:56'!
hasPrevious

	^ currentIndex > 0! !

!TextEditorCommandHistory methodsFor: 'testing' stamp: 'mt 11/26/2015 16:46'!
hasReplacedSomething

	^ self hasPrevious and: [self previous hasReplacedSomething]! !

!TextEditorCommandHistory methodsFor: 'initialization' stamp: 'mt 11/7/2015 23:41'!
initialize

	super initialize.
	self reset.! !

!TextEditorCommandHistory methodsFor: 'accessing' stamp: 'mt 11/8/2015 09:57'!
next

	^ self hasNext
		ifTrue: [commands at: currentIndex+1]
		ifFalse: [nil]! !

!TextEditorCommandHistory methodsFor: 'accessing' stamp: 'mt 11/8/2015 09:58'!
previous

	^ self hasPrevious
		ifTrue: [commands at: currentIndex]
		ifFalse: [nil]! !

!TextEditorCommandHistory methodsFor: 'undo/redo' stamp: 'mt 11/10/2015 10:44'!
redoIn: editor

	self hasNext ifFalse: [^ self].
	
	[self current redoIn: editor]
		ensure: [currentIndex := currentIndex + 1].

	self previous isCompositeRedo == true ifTrue: [self redoIn: editor].! !

!TextEditorCommandHistory methodsFor: 'initialization' stamp: 'mt 11/7/2015 23:40'!
reset

	commands := #().
	currentIndex := 0.! !

!TextEditorCommandHistory methodsFor: 'undo/redo' stamp: 'mt 11/10/2015 10:40'!
undoIn: editor

	self hasPrevious ifFalse: [^ self].
	
	currentIndex := currentIndex - 1.
	self current undoIn: editor.

	self current isCompositeUndo == true ifTrue: [self undoIn: editor].! !

Object subclass: #TheWorldMainDockingBar
	instanceVariableNames: ''
	classVariableNames: 'Instance ShowSecondsInClock TS TwentyFourHourClock'
	poolDictionaries: ''
	category: 'Morphic-Kernel'!

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'mt 7/10/2016 12:33'!
applyUserInterfaceTheme

	self updateInstances.! !

!TheWorldMainDockingBar class methodsFor: 'initialize-release' stamp: 'fbs 4/17/2013 17:16'!
cleanUp
	self updateInstances.! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:11'!
disableMenuPreference: aPreferenceSymbol

	self setMenuPreference: aPreferenceSymbol to: false.! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:11'!
enableMenuPreference: aPreferenceSymbol

	self setMenuPreference: aPreferenceSymbol to: true.! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:22'!
getMenuPreference: aPreferenceSymbol

	^ self getMenuPreference: aPreferenceSymbol ifAbsent: [true]! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:22'!
getMenuPreference: aPreferenceSymbol ifAbsent: aBlock

	| project |
	^ (project := Project current) isMorphic
		ifTrue: [project projectPreferenceFlagDictionary at: aPreferenceSymbol ifAbsent: aBlock]
		ifFalse: aBlock
! !

!TheWorldMainDockingBar class methodsFor: 'class initialization' stamp: 'dtl 1/18/2011 21:37'!
initialize
	"self initialize"
	
	Locale addLocalChangedListener: self.
	self updateInstances.! !

!TheWorldMainDockingBar class methodsFor: 'access' stamp: 'cmm 11/16/2010 19:14'!
instance
	"Answer the receiver's instance"
	^ Instance ifNil: [ Instance := self new ]! !

!TheWorldMainDockingBar class methodsFor: 'access' stamp: 'cmm 11/16/2010 19:15'!
instance: aTheWorldMainDockingBar
	Instance := aTheWorldMainDockingBar.
	self updateInstances! !

!TheWorldMainDockingBar class methodsFor: 'events' stamp: 'kb 11/19/2009 18:48'!
localeChanged
	self updateInstances! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:27'!
setMenuPreference: aPreferenceSymbol to: aBoolean
	| project |
	(project := Project current) isMorphic ifTrue: [
		project projectPreferenceFlagDictionary at: aPreferenceSymbol  put: aBoolean.
		(aBoolean ~= (Preferences perform: aPreferenceSymbol))
			ifTrue: [Preferences setPreference: aPreferenceSymbol toValue: aBoolean]].
	self updateInstances.! !

!TheWorldMainDockingBar class methodsFor: 'timestamping' stamp: 'kb 11/19/2009 18:48'!
setTimeStamp
	"Change the receiver's timeStamp"
	TS := UUID new! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:26'!
showAppsMenuInWorldMainDockingBar
	
	<preference: 'Show ''Apps'' menu in world main docking bar'
		category: 'docking bars'
		description: 'Whether world''s main docking bar should show the ''Apps'' menu.'
		type: #Boolean>
	^ self getMenuPreference: #showAppsMenuInWorldMainDockingBar ifAbsent: [true]
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:23'!
showAppsMenuInWorldMainDockingBar: aBoolean
	
	self
		setMenuPreference: #showAppsMenuInWorldMainDockingBar
		to: aBoolean.
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'mt 5/14/2016 09:44'!
showDoMenuInWorldMainDockingBar
	
	<preference: 'Show ''Do'' menu in world main docking bar'
		category: 'docking bars'
		description: 'Whether world''s main docking bar should show the ''Do'' menu.'
		type: #Boolean>
	^ self getMenuPreference: #showDoMenuInWorldMainDockingBar ifAbsent: [true]
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'mt 5/14/2016 09:44'!
showDoMenuInWorldMainDockingBar: aBoolean
	
	self
		setMenuPreference: #showDoMenuInWorldMainDockingBar
		to: aBoolean.
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:26'!
showExtrasMenuInWorldMainDockingBar
	
	<preference: 'Show ''Extras'' menu in world main docking bar'
		category: 'docking bars'
		description: 'Whether world''s main docking bar should show the ''Extras'' menu.'
		type: #Boolean>
	^ self getMenuPreference: #showExtrasMenuInWorldMainDockingBar ifAbsent: [true]
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:23'!
showExtrasMenuInWorldMainDockingBar: aBoolean
	
	self
		setMenuPreference: #showExtrasMenuInWorldMainDockingBar
		to: aBoolean.
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:26'!
showHelpMenuInWorldMainDockingBar
	
	<preference: 'Show ''Help'' menu in world main docking bar'
		category: 'docking bars'
		description: 'Whether world''s main docking bar should show the ''Help'' menu.'
		type: #Boolean>
	^ self getMenuPreference: #showHelpMenuInWorldMainDockingBar ifAbsent: [true]
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:24'!
showHelpMenuInWorldMainDockingBar: aBoolean
	
	self
		setMenuPreference: #showHelpMenuInWorldMainDockingBar
		to: aBoolean.
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:26'!
showMainMenuInWorldMainDockingBar
	
	<preference: 'Show ''Main'' menu in world main docking bar'
		category: 'docking bars'
		description: 'Whether world''s main docking bar should show the ''Main'' menu.'
		type: #Boolean>
	^ self getMenuPreference: #showMainMenuInWorldMainDockingBar ifAbsent: [true]
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:24'!
showMainMenuInWorldMainDockingBar: aBoolean
	
	self
		setMenuPreference: #showMainMenuInWorldMainDockingBar
		to: aBoolean.
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:26'!
showProjectsMenuInWorldMainDockingBar
	
	<preference: 'Show ''Projects'' menu in world main docking bar'
		category: 'docking bars'
		description: 'Whether world''s main docking bar should show the ''Projects'' menu.'
		type: #Boolean>
	^ self getMenuPreference: #showProjectsMenuInWorldMainDockingBar ifAbsent: [true]
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:23'!
showProjectsMenuInWorldMainDockingBar: aBoolean
	
	self
		setMenuPreference: #showProjectsMenuInWorldMainDockingBar
		to: aBoolean.
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'mt 1/18/2016 08:27'!
showSecondsInClock
	<preference: 'Show seconds in clock'
		category: 'docking bars'
		description: 'Whether the clock in the world''s main docking bar should display seconds.'
		type: #Boolean>
	^ ShowSecondsInClock ifNil: [ true ]! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'mt 1/18/2016 08:27'!
showSecondsInClock: boolean

	ShowSecondsInClock := boolean.
	self updateInstances.! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:27'!
showToolsMenuInWorldMainDockingBar
	
	<preference: 'Show ''Tools'' menu in world main docking bar'
		category: 'docking bars'
		description: 'Whether world''s main docking bar should show the ''Tools'' menu.'
		type: #Boolean>
	^ self getMenuPreference: #showToolsMenuInWorldMainDockingBar ifAbsent: [true]
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:23'!
showToolsMenuInWorldMainDockingBar: aBoolean
	
	self
		setMenuPreference: #showToolsMenuInWorldMainDockingBar
		to: aBoolean.
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:26'!
showWindowsMenuInWorldMainDockingBar
	
	<preference: 'Show ''Windows'' menu in world main docking bar'
		category: 'docking bars'
		description: 'Whether world''s main docking bar should show the ''Windows'' menu.'
		type: #Boolean>
	^ self getMenuPreference: #showWindowsMenuInWorldMainDockingBar ifAbsent: [true]
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'topa 10/9/2014 23:23'!
showWindowsMenuInWorldMainDockingBar: aBoolean
	
	self
		setMenuPreference: #showWindowsMenuInWorldMainDockingBar
		to: aBoolean.
! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'kb 11/19/2009 18:48'!
showWorldMainDockingBar
	
	<preference: 'Show world main docking bar'
		category: 'docking bars'
		description: 'Whether world''s main docking bar should be shown or not.'
		type: #Boolean>
	^Project current showWorldMainDockingBar! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'kb 11/19/2009 18:48'!
showWorldMainDockingBar: aBoolean
	
	Project current showWorldMainDockingBar: aBoolean! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'mt 7/10/2016 13:01'!
themeProperties

	^ {
		{ #textColor. 'Colors'. 'The color for the clock and other labels.' }.
		{ #font. 'Fonts'. 'The font for the clock and other labels.' }.
		{ #logoColor. 'Colors'. 'The color of the Squeak logo.' }.
		{ #selectionLogoColor. 'Colors'. 'The color of the Squeak logo when it is selected.' }.
	}! !

!TheWorldMainDockingBar class methodsFor: 'timestamping' stamp: 'kb 11/19/2009 18:48'!
timeStamp
	"Answer the receiver's timeStamp"
	^ TS! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'cmm 5/10/2015 18:03'!
twentyFourHourClock
	<preference: '24-hour Time Clock'
		category: 'docking bars'
		description: 'Whether the clock in the world''s main docking bar should present the time in 24-hour format.'
		type: #Boolean>
	^ TwentyFourHourClock ifNil: [ true ]! !

!TheWorldMainDockingBar class methodsFor: 'preferences' stamp: 'cmm 5/10/2015 18:05'!
twentyFourHourClock: aBoolean
	TwentyFourHourClock := aBoolean.
	self updateInstances! !

!TheWorldMainDockingBar class methodsFor: 'events' stamp: 'topa 10/20/2015 08:46:31'!
updateInstances
	"The class has changed, time to update the instances"

	self setTimeStamp.
	Project current in: [:project |
		project isMorphic ifTrue: [
			project assureMainDockingBarPresenceMatchesPreference]].! !

!TheWorldMainDockingBar class methodsFor: 'events' stamp: 'kb 11/19/2009 18:48'!
updateInstances: anEvent 
	"The class has changed, time to update the instances"
	(anEvent itemClass == self
			or: [anEvent itemClass == self class])
		ifFalse: [^ self].
	""
	self updateInstances! !

!TheWorldMainDockingBar methodsFor: 'submenu - squeak' stamp: 'kb 11/20/2009 17:16'!
aboutMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'About Squeak...' translated;
			target: self;
			selector: #aboutSqueak ]! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'mt 8/2/2016 13:17'!
aboutSqueak
	
	| m |
	 m := SystemReporter open.
	 m label: 'About Squeak' translated.! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'laza 3/3/2010 05:18'!
allOtherWindowsLike: window
	^ self allVisibleWindows reject: [:each |
		each model name ~= window model name or: [each = window]]! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'laza 3/2/2010 11:29'!
allVisibleWindows
	^SystemWindow windowsIn: World satisfying: [ :w | w visible ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'laza 3/3/2010 05:04'!
allWindowsLike: window
	^ self allVisibleWindows reject: [:each | each model ~= window model or: [each = window]]! !

!TheWorldMainDockingBar methodsFor: 'updating' stamp: 'mt 7/10/2016 12:33'!
applyUserInterfaceTheme
	"Do it on the class-side."! !

!TheWorldMainDockingBar methodsFor: 'submenu - apps' stamp: 'ar 8/6/2010 23:40'!
appsMenuOn: aDockingBar
	"Create a menu with the registered apps"

	aDockingBar addItem: [ :item |
		item
			contents: 'Apps' translated;
			subMenuUpdater: self
			selector: #listAppsOn: ]
! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:08'!
browserMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Browser' translated;
			help: 'Open a browser' translated;
			icon: (self colorIcon: Browser basicNew windowColorToUse);
			target: StandardToolSet;
			selector: #openClassBrowser ]! !

!TheWorldMainDockingBar methodsFor: 'right side' stamp: 'mt 7/10/2016 12:44'!
clockOn: aDockingBar 
	aDockingBar 
		addMorphBack: (ClockMorph new
			font: self fontToUse;
			color: self colorToUse;
			show24hr: self class twentyFourHourClock;
			showSeconds: self class showSecondsInClock;
			yourself);
		addDefaultSpace! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'mt 3/6/2015 08:24'!
closeAllWindows
	self allVisibleWindows do: [:each |
		each model canDiscardEdits ifTrue: [each delete]]! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'mt 3/6/2015 08:40'!
closeAllWindowsButWorkspaces

	(UserDialogBoxMorph
		confirm: 'There might be unsaved changes.\Do you really want to close all windows\that are no workspaces?' withCRs
		title: 'Only keep workspaces') ifTrue: [
			self allVisibleWindows
				reject: [:each | each model isKindOf: Workspace]
				thenDo: [:each | [each delete] valueSupplyingAnswer: true]].! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'laza 3/3/2010 05:18'!
closeAllWindowsBut: window
	(self allOtherWindowsLike: window) do: [:each |
		each model canDiscardEdits ifTrue: [each delete]]! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'laza 3/3/2010 05:17'!
closeAllWindowsLike: window
	self closeAllWindowsBut: window.
	window model canDiscardEdits ifTrue: [window delete]! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'mt 3/6/2015 08:39'!
closeAllWindowsUnsafe

	(UserDialogBoxMorph
		confirm: 'There might be unsaved changes.\Do you really want to close all windows?' withCRs
		title: 'Close All Windows') ifTrue: [
			self allVisibleWindows do: [:each | [each delete] valueSupplyingAnswer: true]].! !

!TheWorldMainDockingBar methodsFor: 'private' stamp: 'mt 7/2/2016 12:43'!
colorIcon: aColor

	"Guess if 'uniform window colors' are used and avoid all icons to be just gray"
	(aColor = (UserInterfaceTheme current get: #uniformWindowColor for: Model) or: [Preferences tinyDisplay]) ifTrue: [ ^nil ].
	^(aColor iconOrThumbnailOfSize: 14)
		borderWidth: 3 color: MenuMorph menuColor muchDarker;
		borderWidth: 2 color: Color transparent! !

!TheWorldMainDockingBar methodsFor: 'private' stamp: 'mt 7/10/2016 12:44'!
colorToUse

	^ self userInterfaceTheme textColor ifNil: [Color black]! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 13:25'!
commandKeyHelp
	"Open a window giving command key help."

	self
		openHelp: #SqueakTutorialsCommandKey
		topic: nil
		styled: false! !

!TheWorldMainDockingBar methodsFor: 'construction' stamp: 'mt 7/26/2016 12:53'!
createDockingBar
	"Create a docking bar from the receiver's representation"
	
	| dockingBar |
	dockingBar := DockingBarMorph new
		adhereToTop;
		color: MenuMorph menuColor;
		autoGradient: MenuMorph gradientMenu;
		borderWidth: 0.
	self fillDockingBar: dockingBar.
	self labelIfNeeded: dockingBar.
	^ dockingBar! !

!TheWorldMainDockingBar methodsFor: 'construction' stamp: 'mt 5/14/2016 09:47'!
doMenuOn: aDockingBar 

	aDockingBar addItem: [ :item |
		item
			contents: 'Do' translated;
			subMenuUpdater: self
			selector: #listCommonRequestsOn: ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:08'!
dualChangeSorterMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Dual Change Sorter' translated;
			help: 'Open a Dual Change Sorter' translated;
			icon: (self colorIcon: ChangeSorter basicNew windowColorToUse);
			target: DualChangeSorter;
			selector: #open ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - extras' stamp: 'mt 7/2/2016 11:06'!
editCurrentTheme

	UserInterfaceTheme current explore.! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 13:26'!
extendingTheSystem
	
	self
		openHelp: #SqueakProjectHelp
		topic: #extendingTheSystem
		styled: false.! !

!TheWorldMainDockingBar methodsFor: 'submenu - extras' stamp: 'mt 7/2/2016 10:38'!
extrasMenuOn: aDockingBar 

	aDockingBar addItem: [ :it|
		it 	contents: 'Extras' translated;
			addSubMenu: [:menu|
				menu addItem:[:item|
					item
						contents: 'Recover Changes' translated;
						help: 'Recover changes after a crash' translated;
						icon: MenuIcons smallDocumentClockIcon;
						target: ChangeList;
						selector: #browseRecentLog].
				menu addLine.
				menu addItem:[:item|
					item
						contents: 'Themes & Colors' translated;
						subMenuUpdater: self
						selector: #themesAndWindowColorsOn: ].
				menu addItem:[:item|
					item
						contents: 'Set Author Initials' translated;
						help: 'Sets the author initials' translated;
						icon: MenuIcons smallUserQuestionIcon;
						target: Utilities;
						selector: #setAuthorInitials].
				menu addItem:[:item|
					item
						contents: 'Restore Display (r)' translated;
						help: 'Redraws the entire display' translated;
						target: Project current;
						selector: #restoreDisplay].
				menu addItem:[:item|
					item
						contents: 'Rebuild Menus' translated;
						help: 'Rebuilds the menu bar' translated;
						target: TheWorldMainDockingBar;
						selector: #updateInstances].
				menu addLine.
				menu addItem:[:item|
					item
						contents: 'Start Profiler' translated;
						help: 'Starts the profiler' translated;
						icon: MenuIcons smallTimerIcon;
						target: self;
						selector: #startMessageTally].
				menu addItem:[:item|
					item
						contents: 'Collect Garbage' translated;
						help: 'Run the garbage collector and report space usage' translated;
						target: Utilities;
						selector: #garbageCollectAndReport].
				menu addItem:[:item|
					item
						contents: 'Purge Undo Records' translated;
						help: 'Save space by removing all the undo information remembered in all projects' translated;
						target: CommandHistory;
						selector: #resetAllHistory].
				menu addItem:[:item|
					item
						contents: 'VM statistics' translated;
						help: 'Virtual Machine information' translated;
						target: self;
						selector: #vmStatistics].
				menu addLine.
				menu addItem:[:item|
					item
						contents: 'Graphical Imports' translated;
						help: 'View the global repository called ImageImports; you can easily import external graphics into ImageImports via the FileList' translated;
						target: (Imports default);
						selector: #viewImages].
				menu addItem:[:item|
					item
						contents: 'Standard Graphics Library' translated;
						help: 'Lets you view and change the system''s standard library of graphics' translated;
						target: ScriptingSystem;
						selector: #inspectFormDictionary].
				menu addItem:[:item|
					item
						contents: 'Annotation Setup' translated;
						help: 'Click here to get a little window that will allow you to specify which types of annotations, in which order, you wish to see in the annotation panes of browsers and other tools' translated;
						target: Preferences;
						selector: #editAnnotations].
				menu addItem:[:item|
					item
						contents: 'Browse My Changes' translated;
						help: 'Browse all of my changes since the last time #condenseSources was run.' translated;
						target: SystemNavigation new;
						selector: #browseMyChanges].
			] ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:08'!
fileListMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'File List' translated;
			help: 'Open a file list' translated;
			icon: (self colorIcon: FileList basicNew windowColorToUse);
			target: StandardToolSet;
			selector: #openFileList ]! !

!TheWorldMainDockingBar methodsFor: 'construction' stamp: 'mt 4/2/2015 15:09'!
fillDockingBar: aDockingBar 
	"Private - fill the given docking bar"
	
	self menusOn: aDockingBar.
	aDockingBar addSpacer.
	self projectNameOn: aDockingBar.
	aDockingBar addSpacer.
	self rightSideOn: aDockingBar.
	aDockingBar
		setProperty: #mainDockingBarTimeStamp 
		toValue: self class timeStamp.! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 13:26'!
fontSizeSummary
	
	self
		openHelp: #SqueakToolsHelp
		topic: #fontSizeSummary
		styled: false.! !

!TheWorldMainDockingBar methodsFor: 'private' stamp: 'mt 7/10/2016 12:44'!
fontToUse

	^ self userInterfaceTheme font ifNil: [TextStyle defaultFont]! !

!TheWorldMainDockingBar methodsFor: 'private' stamp: 'kfr 4/19/2015 12:20'!
gradientRamp

	^{ 
		0.0 -> Color white.
		1.0 -> MenuMorph menuColor darker }! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 14:26'!
helpMenuOn: aDockingBar

	aDockingBar addItem: [ :it |
		it	contents: 'Help' translated;
			addSubMenu: [ :menu | 
				menu addItem: [:item |
						item
							contents: 'Squeak Help' translated;
							help: 'Integrated Help System' translated;
							target: self;
							selector: #squeakHelp].
				
				menu addLine.
				
				menu addItem:[:item|
					item
						contents: 'Online Resources' translated;
						help: 'Online resources for Squeak' translated;
						target: self;
						icon: MenuIcons smallHelpIcon;
						selector: #squeakOnlineResources].
				menu addItem:[:item|
					item
						contents: 'Keyboard Shortcuts' translated;
						help: 'Keyboard bindings used in Squeak' translated;
						target: self;
						selector: #commandKeyHelp ].
				menu addItem:[:item|
					item
						contents: 'Font Size Summary' translated;
						help: 'Font size summary.' translated;
						target: self;
						selector: #fontSizeSummary ].
				menu addItem:[:item|
					item
						contents: 'Useful Expressions' translated;
						help: 'Useful expressions' translated;
						target: self;
						selector: #usefulExpressions ].
				menu addLine.
				menu addItem:[:item|
					item
						contents: 'Terse Guide to Squeak' translated;
						help: 'Concise information about language and environment' translated;
						target: self;
						selector: #terseGuideToSqueak].
				menu addItem:[:item|
					item
						contents: 'Extending the system' translated;
						help: 'Includes code snippets to evaluate for extending the system' translated;
						target: self;
						icon: MenuIcons smallHelpIcon;
						selector: #extendingTheSystem].
				menu addLine.
				
			menu addItem:[:item|
				item
					contents: 'Release Notes' translated;
					help: 'Changes in this release' translated ;
					target: self;
					selector: #releaseNotes].
			menu addItem:[:item|
				item
					contents: 'Working With Squeak' translated;
					help: 'Information for new users' ;
					target: self;
					selector: #workingWithSqueak].
			menu addItem:[:item|
				item
					contents: 'The Squeak User Interface' translated;
					help: 'Descriptions of some of the more-unusual UI elements in Squeak' ;
					target: self;
					selector: #squeakUserInterface].
			menu addItem:[:item|
				item
					contents: 'License Information' translated;
					help: String empty ;
					target: self;
					selector: #licenseInformation].				
				
		
			menu addLine.
			menu addItem: [:item |
				item
					contents: 'About Squeak' translated;
					help: 'SystemReporter status of the image and runtime environment' translated;
					target: self;
					selector: #aboutSqueak].
		]]! !

!TheWorldMainDockingBar methodsFor: 'submenu - projects' stamp: 'MAD 1/1/2011 19:18'!
jumpToProjectMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Jump To Project' translated;
			icon: MenuIcons smallProjectJumpIcon;
			subMenuUpdater: self
			selector: #updateJumpToProjectSubMenu: ]! !

!TheWorldMainDockingBar methodsFor: 'private' stamp: 'eem 7/4/2010 19:32'!
labelIfNeeded: aDockingBar 
	"Label the given docking bar with the project name, if needed"
	(aDockingBar submorphWithProperty: #projectNameMorph)
		contents: (Project current isTopProject ifTrue: [''] ifFalse: [Project current name]);
		fitContents! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 13:26'!
licenseInformation
	
	self
		openHelp: #SqueakLicenseHelp
		topic: #officialLicense
		styled: false.! !

!TheWorldMainDockingBar methodsFor: 'submenu - apps' stamp: 'ar 8/6/2010 23:44'!
listAppsOn: menu
	"Update the apps list in the menu"

	| args |
	TheWorldMenu registeredOpenCommands do:[:spec|
		args := spec second.
		menu addItem: [ :item |
			item
				contents: spec first translated;
				target: args first;
				selector: args second].
	].
! !

!TheWorldMainDockingBar methodsFor: 'submenu - do' stamp: 'mt 5/14/2016 09:42'!
listCommonRequestsOn: aMenu

	| strings |
	strings := Utilities commonRequestStrings contents.

	strings asString linesDo: [:aString |
		aString = '-'
			ifTrue: [aMenu addLine]
			ifFalse: [aMenu add: (aString ifEmpty: [' ']) target: Utilities selector: #eval: argument: aString]].

	aMenu addLine.
	aMenu add: 'edit this list' translated target: Utilities action: #editCommonRequestStrings.! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'mt 5/17/2016 16:04'!
listWindowsOn: menu

	| windows |
	windows := SortedCollection sortBlock: [:winA :winB |
		winA model name = winB model name
			ifTrue: [winA label < winB label]
			ifFalse: [winA model name < winB model name]].
	windows addAll: self allVisibleWindows.
	windows ifEmpty: [ 
		menu addItem: [ :item | 
			item
				contents: 'No Windows' translated;
				isEnabled: false ] ].
	windows do: [ :each |
		menu addItem: [ :item |
			item 
				contents: (self windowMenuItemLabelFor: each);
				icon: (self colorIcon: each model defaultBackgroundColor);
				target: each;
				selector: #comeToFront;
				subMenuUpdater: self
				selector: #windowMenuFor:on:
				arguments: { each };
				action: [ each beKeyWindow; expand ] ] ].
	menu
		addLine;
		add: 'Close all windows' target: self selector: #closeAllWindowsUnsafe;
		addItem: [:item | item
			contents: 'Close all windows without changes';
			target: self;
			icon: MenuIcons smallBroomIcon;
			selector: #closeAllWindows];
		add: 'Close all windows but workspaces' target: self selector: #closeAllWindowsButWorkspaces.! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'kb 11/20/2009 17:42'!
loadProject

	World worldMenu loadProject! !

!TheWorldMainDockingBar methodsFor: 'submenu - projects' stamp: 'mt 5/10/2015 13:31'!
loadProjectMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Load Project' translated;
			help: 'Load a project from a file' translated;
			target: self;
			selector: #loadProject ]! !

!TheWorldMainDockingBar methodsFor: 'construction' stamp: 'mt 6/15/2016 08:11'!
menusOn: aDockingBar

	self class showMainMenuInWorldMainDockingBar ifTrue: [
		self squeakMenuOn: aDockingBar].

	self class showProjectsMenuInWorldMainDockingBar ifTrue: [
		self projectsMenuOn: aDockingBar].

	self class showToolsMenuInWorldMainDockingBar ifTrue: [
		self toolsMenuOn: aDockingBar].

	self class showAppsMenuInWorldMainDockingBar ifTrue: [
		self appsMenuOn: aDockingBar].

	self class showDoMenuInWorldMainDockingBar ifTrue: [
		self doMenuOn: aDockingBar].

	self class showExtrasMenuInWorldMainDockingBar ifTrue: [
		self extrasMenuOn: aDockingBar].

	self class showWindowsMenuInWorldMainDockingBar ifTrue: [
		self windowsMenuOn: aDockingBar].

	self class showHelpMenuInWorldMainDockingBar ifTrue: [
		self helpMenuOn: aDockingBar].! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:08'!
methodFinderMenuItemOn: menu 

	menu addItem: [:item | 
		item 
			contents: 'Method Finder' translated;
			help: 'Open the Method Finder' translated;
			icon: (self colorIcon: SelectorBrowser basicNew windowColorToUse);
			target: StandardToolSet;
			selector: #openSelectorBrowser]! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:08'!
monticelloBrowserMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Monticello Browser' translated;
			help: 'Open a Monticello Browser' translated;
			icon: (self colorIcon: MCTool basicNew windowColorToUse);
			target: MCWorkingCopyBrowser;
			selector: #open ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:08'!
monticelloConfigurationsMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Monticello Configurations' translated;
			help: 'Open a Monticello Configurations Editor' translated;
			icon: (self colorIcon: MCConfigurationBrowser basicNew windowColorToUse);
			target: MCConfigurationBrowser;
			selector: #open ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - projects' stamp: 'kb 11/26/2009 20:41'!
newProjectMenuItemOn: menu

	menu addItem: [ :item | 
		item
			contents: 'New Project' translated;
			help: 'Start a new MorphicProject' translated;
			icon: MenuIcons smallProjectIcon;
			target: self;
			selector: #newProject:;
			arguments: { MorphicProject };
			subMenuUpdater:  self
			selector: #updateNewProjectSubMenu: ]! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'ar 11/25/2009 19:48'!
newProject: projectClass
	"Create a new project of the given type"
	| newProject |
	"Allow the project to return nil from #new to indicate that it was canceled."
	newProject := projectClass new ifNil:[^self].
	ProjectViewMorph openOn: newProject.
	newProject enter.! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 14:24'!
openHelp: bookSymbol topic: topicSymbol styled: boolean

	| openSelector |
	openSelector := boolean ifTrue: [#openForCodeOn:] ifFalse: [#openOn:].

	(Smalltalk classNamed: 'HelpBrowser')
		ifNil: [self inform: 'Sorry, there is no help system installed.' translated]
		ifNotNil: [:helpClass |
			(Smalltalk classNamed: bookSymbol)
				ifNil: [self inform: 'Sorry, the help book you requested does not exist.']
				ifNotNil: [:book |
					topicSymbol
						ifNil: [(helpClass perform: openSelector with: book) model showFirstTopic]
						ifNotNil: [helpClass perform: openSelector with: (book perform: topicSymbol)]]]! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:08'!
preferenceBrowserMenuItemOn: menu
	Smalltalk at: #PreferenceBrowser ifPresent:[:pb|
		menu addItem: [ :item |
			item
				contents: 'Preferences' translated;
				help: 'Open a Preferences Browser' translated;
				icon: (self colorIcon: pb basicNew windowColorToUse);
				target: pb;
				selector: #open ]
	].! !

!TheWorldMainDockingBar methodsFor: 'submenu - projects' stamp: 'mt 5/10/2015 13:32'!
previousProjectMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Previous Project' translated;
			help: 'Return to the most-recently-visited project' translated;
			target: World;
			selector: #goBack ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:09'!
processBrowserMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Process Browser' translated;
			help: 'Open a Process Browser' translated;
			icon: (self colorIcon: ProcessBrowser basicNew windowColorToUse);
			target: ProcessBrowser;
			selector: #open ]! !

!TheWorldMainDockingBar methodsFor: 'right side' stamp: 'eem 5/6/2014 14:00'!
projectNameOn: aDockingBar
	| morph |
	morph := StringMorph contents: ''.
	morph setProperty: #projectNameMorph toValue: #projectNameMorph.
	aDockingBar addMorphBack: morph.
	self labelIfNeeded: aDockingBar! !

!TheWorldMainDockingBar methodsFor: 'construction' stamp: 'mt 5/10/2015 13:32'!
projectsMenuOn: aDockingBar

	aDockingBar addItem: [ :item |
		item
			contents: 'Projects' translated;
			addSubMenu: [ :menu | 
				self newProjectMenuItemOn: menu.
				menu addLine.
				self
					saveProjectMenuItemOn: menu;
					loadProjectMenuItemOn: menu.
				menu addLine.
				self
					previousProjectMenuItemOn: menu;
					jumpToProjectMenuItemOn: menu ] ]
! !

!TheWorldMainDockingBar methodsFor: 'submenu - squeak' stamp: 'MAD 12/22/2010 00:11'!
quitMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Quit' translated;
			help: 'Quit out of Squeak' translated;
			icon: MenuIcons smallQuitNoSaveIcon;
			target: self;
			selector: #quitSqueak ]! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'fbs 7/25/2013 07:20'!
quitSqueak

	^Smalltalk
		snapshot: (
			UserDialogBoxMorph 
				confirm: 'Save changes before quitting?' translated 
				orCancel: [ ^self ]
				at: World center)
		andQuit: true

	! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 13:35'!
releaseNotes
	
	self
		openHelp: #SqueakReleaseNotes
		topic: nil
		styled: false.! !

!TheWorldMainDockingBar methodsFor: 'construction' stamp: 'rhi 5/18/2010 21:04'!
rightSideOn: aDockingBar

	self
		searchBarOn: aDockingBar;
		clockOn: aDockingBar;
		toggleFullScreenOn: aDockingBar.! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'fbs 7/25/2013 07:20'!
save

	Smalltalk snapshot: true andQuit: false! !

!TheWorldMainDockingBar methodsFor: 'submenu - squeak' stamp: 'mt 5/10/2015 13:16'!
saveAndQuitMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Save And Quit' translated;
			help: 'Save the current state of Squeak on disk, and quit out of Squeak' translated;
			target: self;
			selector: #saveAndQuitSqueak ]! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'fbs 7/25/2013 07:20'!
saveAndQuitSqueak

	Smalltalk snapshot: true andQuit: true! !

!TheWorldMainDockingBar methodsFor: 'submenu - squeak' stamp: 'mt 5/10/2015 13:30'!
saveAsMenuItemOn: menu

	menu addItem: [ :item |	
		item
			contents: 'Save As...' translated;
			help: 'Save the current state of Squeak on disk under a new name' translated;
			target: self;
			selector: #saveImageAs ]! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'fbs 7/25/2013 07:20'!
saveAsNewVersion

	Smalltalk saveAsNewVersion! !

!TheWorldMainDockingBar methodsFor: 'submenu - squeak' stamp: 'mt 5/10/2015 13:16'!
saveAsNewVersionMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Save As New Version' translated;
			help: 'Save the current state of Squeak on disk under a version-stamped name' translated;
			target: self;
			selector: #saveAsNewVersion ]! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'fbs 7/25/2013 07:21'!
saveImage

	Smalltalk saveSession! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'fbs 7/25/2013 07:21'!
saveImageAs

	Smalltalk saveAs! !

!TheWorldMainDockingBar methodsFor: 'submenu - squeak' stamp: 'mt 5/10/2015 13:31'!
saveMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Save' translated;
			help: 'Save the current state of Squeak on disk' translated;
			icon: MenuIcons smallSaveIcon;
			target: self;
			selector: #saveImage ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - projects' stamp: 'mt 5/10/2015 13:31'!
saveProjectMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Save Project' translated;
			help: 'Save this project on a file' translated;
			target: World;
			selector: #saveOnFile ]! !

!TheWorldMainDockingBar methodsFor: 'right side' stamp: 'mt 4/7/2015 11:53'!
searchBarOn: aDockingBar

	aDockingBar 
		addMorphBack: (SearchBar build vResizing: #spaceFill; width: 200);
		addDefaultSpace! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:09'!
simpleChangeSorterMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Simple Change Sorter' translated;
			help: 'Open a Change Sorter' translated;
			icon: (self colorIcon: ChangeSorter basicNew windowColorToUse);
			target: ChangeSorter;
			selector: #open ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 13:27'!
squeakHelp

	(Smalltalk classNamed: 'HelpBrowser')
		ifNil: [self inform: 'Sorry, there is no help system installed.' translated]
		ifNotNil: [:helpClass | helpClass open].! !

!TheWorldMainDockingBar methodsFor: 'construction' stamp: 'mt 7/10/2016 13:03'!
squeakMenuOn: aDockingBar 
	"Private - fill the given docking bar"
	
	aDockingBar addItem: [ :item |
		item
			contents: '';
			icon: (MenuIcons squeakLogoIconColorized:
				(self userInterfaceTheme logoColor ifNil: [Color black]));
			selectedIcon: (MenuIcons squeakLogoIconColorized:
				(self userInterfaceTheme selectionLogoColor ifNil: [Color white]));
			addSubMenu: [ :menu | 
				self
					aboutMenuItemOn: menu;
					updateMenuItemOn: menu.
				menu addLine.
				self 
					saveMenuItemOn: menu;
					saveAsMenuItemOn: menu;
					saveAsNewVersionMenuItemOn: menu.
				menu addLine.
				self
					saveAndQuitMenuItemOn: menu;
					quitMenuItemOn: menu ] ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 13:27'!
squeakOnlineResources
	
	self
		openHelp: #SqueakProjectHelp
		topic: #squeakResourcesOnline
		styled: false.! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 13:27'!
squeakUserInterface
	
	self
		openHelp: #SqueakProjectHelp
		topic: #squeakUserInterface
		styled: false.! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'mt 2/16/2016 09:46'!
startMessageTally
	(self confirm: 'MessageTally will start now,
and stop when the cursor goes
to the top of the screen') ifTrue:
		[MessageTally spyOn:
			[[Sensor peekPosition y > 0] whileTrue: [World doOneCycle]]]! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 14:25'!
terseGuideToSqueak

	self
		openHelp: #TerseGuideHelp
		topic: nil
		styled: true.! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:09'!
testRunnerMenuItemOn: menu
	Smalltalk at: #TestRunner ifPresent:[:aClass|
		menu addItem: [ :item |
			item
				contents: 'Test Runner' translated;
				help: 'Open the Test Runner' translated;
				icon: (self colorIcon: aClass basicNew windowColorToUse);
				target: aClass;
				selector: #open ]
	].! !

!TheWorldMainDockingBar methodsFor: 'submenu - extras' stamp: 'mt 7/25/2016 14:23'!
themesAndWindowColorsOn: menu

	| themes |
	themes := UserInterfaceTheme allThemes asArray sortBy: [:t1 :t2 |
		t1 name <= t2 name].
	
	menu addItem:[:item|
		item
			contents: (Model useColorfulWindows ifTrue: ['<yes>'] ifFalse: ['<no>']), 'Colorful Windows' translated;
			target: self;
			selector: #toggleColorfulWindows].
	menu addItem:[:item|
		item
			contents: (SystemWindow gradientWindow not ifTrue: ['<yes>'] ifFalse: ['<no>']), 'Flat Widget Look' translated;
			target: self;
			selector: #toggleGradients].
	menu addItem:[:item |
		item
			contents: ((Preferences valueOfFlag: #menuAppearance3d) ifTrue: ['<yes>'] ifFalse: ['<no>']), 'Shadows' translated;
			target: self;
			selector: #toggleShadows].
	menu addItem:[:item |
		item
			contents: (SystemWindow roundedWindowCorners ifTrue: ['<yes>'] ifFalse: ['<no>']), 'Rounded Widget Look' translated;
			target: self;
			selector: #toggleRoundedLook].

	
	menu addLine.
	
	themes ifEmpty: [ 
		menu addItem: [ :item | 
			item
				contents: '(No UI themes found.)' translated;
				isEnabled: false ] ].
	themes do: [ :each |
		menu addItem: [ :item |
			item 
				contents: (UserInterfaceTheme current == each ifTrue: ['<yes>'] ifFalse: ['<no>']), each name;
				target: each;
				selector: #apply ] ].
	menu
		addLine;
		add: 'Edit Current UI Theme...' target: self selector: #editCurrentTheme.! !

!TheWorldMainDockingBar methodsFor: 'submenu - extras' stamp: 'mt 7/24/2016 09:31'!
toggleColorfulWindows

	Model useColorfulWindows: Model useColorfulWindows not.! !

!TheWorldMainDockingBar methodsFor: 'right side' stamp: 'mt 8/2/2016 16:52'!
toggleFullScreenOn: aDockingBar 
	
	| toggleMorph onIcon offIcon box bgColor |
	offIcon := MenuIcons fullscreenWireframeIconColorized:
					(self userInterfaceTheme logoColor ifNil: [Color black]).
	onIcon := MenuIcons fullscreenWireframeIconColorized:
					(self userInterfaceTheme selectionLogoColor ifNil: [Color white]).
	bgColor := (UserInterfaceTheme current get: #selectionColor for: #DockingBarItemMorph) ifNil: [Color blue].
	
	toggleMorph := offIcon asMorph.
	
	box := Morph new
		color: Color transparent;
		hResizing: #shrinkWrap;
		vResizing: #spaceFill;
		width: toggleMorph width;
		changeTableLayout;
		
		borderWidth: 1;
		borderColor: Color transparent;
		balloonText: 'toggle full screen mode' translated;
		addMorph: toggleMorph.
		
	toggleMorph setToAdhereToEdge: #rightCenter. 
		
	box
		on: #mouseUp
		send: #value
		to:
			[ DisplayScreen toggleFullScreen. 
			"toggleMorph image: MenuIcons smallFullscreenOffIcon" ] ;

		on: #mouseEnter
		send: #value
		to: [
			toggleMorph image: onIcon.
			box color: bgColor; borderColor: bgColor];
		
		on: #mouseLeave
		send: #value
		to: [
			toggleMorph image: offIcon.
			box color: Color transparent; borderColor: Color transparent].
				
	aDockingBar addMorphBack: box! !

!TheWorldMainDockingBar methodsFor: 'submenu - extras' stamp: 'mt 7/25/2016 14:41'!
toggleGradients

	| switch |
	switch := SystemWindow gradientWindow not.
	
	SystemWindow gradientWindow: switch.
	DialogWindow gradientDialog: switch.
	MenuMorph gradientMenu: switch.
	ScrollBar gradientScrollBar: switch.
	PluggableButtonMorph gradientButton: switch.! !

!TheWorldMainDockingBar methodsFor: 'submenu - extras' stamp: 'mt 7/25/2016 14:28'!
toggleRoundedLook

	| switch |
	switch := SystemWindow roundedWindowCorners not.
	
	SystemWindow roundedWindowCorners: switch.
	DialogWindow roundedDialogCorners: switch.
	PluggableButtonMorph roundedButtonCorners: switch.
	MenuMorph roundedMenuCorners: switch.
	ScrollBar roundedScrollBarLook: switch.! !

!TheWorldMainDockingBar methodsFor: 'submenu - extras' stamp: 'mt 7/27/2016 09:53'!
toggleShadows

	Preferences toggle: #menuAppearance3d.! !

!TheWorldMainDockingBar methodsFor: 'construction' stamp: 'jmg 8/17/2011 17:43'!
toolsMenuOn: aDockingBar 

	aDockingBar addItem: [ :item |
		item
			contents: 'Tools' translated;
			addSubMenu: [ :menu | 
				self
					browserMenuItemOn: menu;
					workspaceMenuItemOn: menu;
					transcriptMenuItemOn: menu;
					testRunnerMenuItemOn: menu;
					methodFinderMenuItemOn: menu.
				menu addLine.
				self 
					monticelloBrowserMenuItemOn: menu;
					monticelloConfigurationsMenuItemOn: menu;
					simpleChangeSorterMenuItemOn: menu;
					dualChangeSorterMenuItemOn: menu.
				menu addLine.
				self
					processBrowserMenuItemOn: menu;
					preferenceBrowserMenuItemOn: menu;
					fileListMenuItemOn: menu.
			] ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:11'!
transcriptMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Transcript' translated;
			help: 'Open the Transcript' translated;
			icon: (self colorIcon: Transcript windowColorToUse);
			target: Transcript;
			selector: #open ]! !

!TheWorldMainDockingBar methodsFor: 'private' stamp: 'eem 5/6/2014 14:07'!
updateIfNeeded: aDockingBar 
	"Update the given docking bar if needed"
	| timeStamp |
	timeStamp := aDockingBar
					valueOfProperty: #mainDockingBarTimeStamp
					ifAbsent: [].
	timeStamp ~= self class timeStamp
		ifTrue:
			[aDockingBar
				 release;
				 removeAllMorphs.
			 self fillDockingBar: aDockingBar]
		ifFalse:
			[self labelIfNeeded: aDockingBar]! !

!TheWorldMainDockingBar methodsFor: 'submenu - projects' stamp: 'dtl 11/23/2009 21:48'!
updateJumpToProjectSubMenu: subMenu

	subMenu defaultTarget: Project.
	Project current buildJumpToMenu: subMenu! !

!TheWorldMainDockingBar methodsFor: 'submenu - squeak' stamp: 'mt 5/10/2015 13:30'!
updateMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Update Squeak' translated;
			help: 'Load latest code updates via the internet' translated;
			icon: MenuIcons smallChangesIcon;
			target: self;
			selector: #updateSqueak ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - projects' stamp: 'kb 11/26/2009 20:40'!
updateNewProjectSubMenu: menu

	Project allSubclasses do: [ :each |
		menu addItem: [ :item | 
			item
				contents: ('New ', each name) translated;
				help: ('Start a new ', each name) translated;
				target: self;
				selector: #newProject:;
				arguments: { each } ] ]! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'fbs 12/6/2013 21:01'!
updateSqueak
	MCMcmUpdater updateFromServer.! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 14:24'!
usefulExpressions
	
	self
		openHelp: #SqueakTutorials
		topic: #usefulExpressions
		styled: true.! !

!TheWorldMainDockingBar methodsFor: 'menu actions' stamp: 'fbs 7/25/2013 07:21'!
vmStatistics
	"Open a string view on a report of vm statistics"

	(StringHolder new contents: Smalltalk vmStatisticsReportString)
		openLabel: 'VM Statistics'! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'cmm 6/21/2011 20:09'!
windowMenuFor: window on: menu
	menu 
		addItem: [ :item |
			item
				contents: 'Close' translated ;
				target: window;
				selector: #delete ];
		addItem: [ :item |
			item
				contents: 'Close all like this' translated ;
				target: self;
				selector: #closeAllWindowsLike:;
				arguments: { window } ];
		addItem: [ :item |
			item
				contents: 'Close all but this' translated ;
				target: self;
				selector: #closeAllWindowsBut:;
				arguments: { window } ];
		addItem: [ :item |
			item 
				contents: 'Expand or Contract' translated ;
				target: window;
				selector: #expandBoxHit ]! !

!TheWorldMainDockingBar methodsFor: 'submenu - windows' stamp: 'laza 3/3/2010 04:59'!
windowMenuItemLabelFor: window
	| s |
	s := WriteStream on: String new.
	window model canDiscardEdits ifFalse: [ s nextPut: $* ].
	window isCollapsed ifTrue: [ s nextPut: $( ].
	s nextPutAll: window label.
	window isCollapsed ifTrue: [ s nextPut: $) ].
	^s contents contractTo: 50! !

!TheWorldMainDockingBar methodsFor: 'construction' stamp: 'kb 11/20/2009 17:55'!
windowsMenuOn: aDockingBar

	aDockingBar addItem: [ :item |
		item
			contents: 'Windows' translated;
			subMenuUpdater: self
			selector: #listWindowsOn: ]
! !

!TheWorldMainDockingBar methodsFor: 'submenu - help' stamp: 'mt 8/2/2016 13:28'!
workingWithSqueak
	
	self
		openHelp: #SqueakProjectHelp
		topic: #workingWithSqueak
		styled: false.! !

!TheWorldMainDockingBar methodsFor: 'submenu - tools' stamp: 'mt 7/10/2016 12:11'!
workspaceMenuItemOn: menu

	menu addItem: [ :item |
		item
			contents: 'Workspace' translated;
			help: 'Open a Workspace' translated;
			icon: (self colorIcon: Workspace basicNew windowColorToUse);
			target: StandardToolSet;
			selector: #openWorkspace ]! !

Object subclass: #TheWorldMenu
	instanceVariableNames: 'myProject myWorld myHand'
	classVariableNames: 'OpenMenuRegistry'
	poolDictionaries: ''
	category: 'Morphic-Kernel'!

!TheWorldMenu commentStamp: 'sw 10/5/2002 00:44' prior: 0!
Instances of TheWorldMenu serve to present the primary Squeak menu obtained by clicking on open desktop, which is variously spoken of as the "screen menu", the "desktop menu", or the "world menu".

myProject is the Project I pertain to.
myWorld is the world, a PasteUpMorph, that I pertain to.
myHand is the hand that invoked the menu.!

!TheWorldMenu class methodsFor: 'class initialization' stamp: 'ar 2/26/2010 23:06'!
cleanUp
	"Flush out obsolete entries"

	self removeObsolete! !

!TheWorldMenu class methodsFor: 'open-menu registry' stamp: 'nice 12/27/2009 03:11'!
loadSqueakMap
	"Load the externally-maintained SqueakMap package if it is not already loaded.  Based on code by Göran Hultgren"

	| server |
	Socket initializeNetwork.
	server := #('map1.squeakfoundation.org' 'map2.squeakfoundation.org' 'map.squeak.org' 'map.bluefish.se' 'marvin.bluefish.se:8000')
		detect: [:srv | | addr answer |
			addr := NetNameResolver addressForName: (srv upTo: $:) timeout: 5.
			addr notNil and: [
				answer := HTTPSocket httpGet: ('http://', srv, '/sm/ping').
				answer isString not and: [answer contents = 'pong']]]
		ifNone: [^ self inform: 'Sorry, no SqueakMap master server responding.'].
	server ifNotNil: ["Ok, found an SqueakMap server"
		ChangeSet newChangesFromStream:
			((('http://', server, '/sm/packagebyname/squeakmap/downloadurl')
			asUrl retrieveContents content) asUrl retrieveContents content unzipped
			readStream)
		named: 'SqueakMap']! !

!TheWorldMenu class methodsFor: 'open-menu registry' stamp: 'sw 11/11/2002 00:27'!
openPackageLoader
	"If this method is reached, it means that SMLoader has not yet been loaded; after SqueakMap has come into the image, a different receiver/selector will have been installed under 'Package Loader'; if this method is reached when theoretically SqueakMap is already loaded, presumably this is a grandfathered menu item in a still-up menu, so get the message on to its appropriate recipient."

	| loaderClass |
	((loaderClass := Smalltalk at: #SMLoader ifAbsent: [nil]) isKindOf: Class)
		ifTrue:
			[^ loaderClass open].

	(self confirm: 
'This requires that you first install "SqueakMap" into your image.
SqueakMap is a new architecture for finding, installing, and
publishing packages in Squeak.
Would you like to install SqueakMap now?' )
		ifTrue:
			[self loadSqueakMap]! !

!TheWorldMenu class methodsFor: 'open-menu registry' stamp: 'nk 6/29/2003 13:56'!
registeredOpenCommands
	"Answer the list of dynamic open commands, sorted by description"
	
	^self registry asArray sort: [ :a :b | a first asLowercase < b first asLowercase ]! !

!TheWorldMenu class methodsFor: 'open-menu registry' stamp: 'sw 11/10/2002 22:56'!
registerOpenCommand: anArray
	"The array received should be of form {'A Label String'. {TargetObject. #command}  'A Help String'} ; the final element is optional but if present will be used to supply balloon help for the menu item in the Open menu.
	If any previous registration of the same label string is already known, delete the old one."

	self unregisterOpenCommand: anArray first.
	OpenMenuRegistry addLast: anArray! !

!TheWorldMenu class methodsFor: 'open-menu registry' stamp: 'nk 6/29/2003 13:55'!
registry
	"Answer the registry of dynamic open commands"
	
	^OpenMenuRegistry ifNil: [OpenMenuRegistry := OrderedCollection new].
! !

!TheWorldMenu class methodsFor: 'open-menu registry' stamp: 'ar 9/27/2005 21:49'!
removeObsolete
	"Remove all obsolete commands"	
	self registry removeAllSuchThat: [:e | e second first class isObsolete].! !

!TheWorldMenu class methodsFor: 'open-menu registry' stamp: 'nk 6/29/2003 13:55'!
unregisterOpenCommandWithReceiver: aReceiver
	"Remove the open command with the given object as receiver from the registry"
	
	self registry removeAllSuchThat: [:e | e second first == aReceiver]! !

!TheWorldMenu class methodsFor: 'open-menu registry' stamp: 'nk 6/29/2003 13:55'!
unregisterOpenCommand: label
	"Remove the open command with the given label from the registry"
	
	self registry removeAllSuchThat: [:e | e first = label]! !

!TheWorldMenu methodsFor: 'menu' stamp: 'nk 2/15/2004 09:38'!
addGestureHelpItemsTo: aMenuMorph 
! !

!TheWorldMenu methodsFor: 'construction' stamp: 'cbr 3/26/2010 00:46'!
addObjectsAndTools: menu
	self
		fillIn: menu
		from: {
			nil.
			{ 'objects (o)'. { #myWorld. #activateObjectsTool }. 'A tool for finding and obtaining many kinds of objects' }.
			{ 'new morph...'. { self. #newMorph }. 'Offers a variety of ways to create new objects' }.
			nil.
			{ 'authoring tools...'. { self. #scriptingDo }. 'A menu of choices useful for authoring' }.
			{ 'playfield options...'. { self. #playfieldDo }. 'A menu of options pertaining to this object as viewed as a playfield' }.
			{ 'flaps...'. { self. #flapsDo }. 'A menu relating to use of flaps.  For best results, use "keep this menu up"' }.
			{ 'projects...'. { self. #projectDo }. 'A menu of commands relating to use of projects' }.
			{ 'telemorphic...' . {self. #remoteDo}.  'commands for doing multi-machine "telemorphic" experiments'}.
			nil
		}! !

!TheWorldMenu methodsFor: 'construction' stamp: 'ar 8/2/2010 20:51'!
addPrintAndDebug: menu
	Preferences simpleMenus ifFalse: [
		self
			fillIn: menu
			from: {
				{ 'make screenshot'. {self. #saveScreenshot}. 'makes a screenshot and saves it to disk'}.
				"{ 'print PS to file...'. { self. #printWorldOnFile }. 'write the world into a postscript file' }."
				{ 'debug...'. { self. #debugDo }. 'a menu of debugging items' }
			} ]! !

!TheWorldMenu methodsFor: 'construction' stamp: 'mha 2/3/2009 16:33'!
addProjectEntries: menu
	self
		fillIn: menu
		from: {
			nil.
			{ 'previous project'. { #myWorld. #goBack }. 'return to the most-recently-visited project' }.
			{ 'jump to project...'. { #myWorld. #jumpToProject }. 'put up a list of all projects, letting me choose one to go to' }.
			{ 'save project on file...'. { #myWorld. #saveOnFile }. 'save this project on a file' }.
			{'load project from file...'. {self. #loadProject}. 'load a project from a file' }.
			nil
		}! !

!TheWorldMenu methodsFor: 'construction' stamp: 'mt 4/13/2016 10:37'!
addRestoreDisplay: menu
	self
		fillIn: menu
		from: {
			{'restore display (r)'. { Project current. #restoreDisplay }. 'repaint the screen -- useful for removing unwanted display artifacts, lingering cursors, etc.' }.
			nil
		}! !

!TheWorldMenu methodsFor: 'construction' stamp: 'fbs 7/25/2013 07:21'!
addSaveAndQuit: menu
	self
		fillIn: menu
		from: {
			nil.
			{ 'save'. { Smalltalk. #saveSession }. 'save the current version of the image on disk' }.
			{ 'save as...'. { Smalltalk. #saveAs }. 'save the current version of the image on disk under a new name.' }.
			{ 'save as new version'. { Smalltalk. #saveAsNewVersion }. 'give the current image a new version-stamped name and save it under that name on disk.' }.
			{ 'save and quit'. { self. #saveAndQuit }. 'save the current image on disk, and quit out of Squeak.' }.
			{ 'quit'. { self. #quitSession }. 'quit out of Squeak.' }
		}! !

!TheWorldMenu methodsFor: 'construction' stamp: 'mha 2/3/2009 16:43'!
addUtilities: menu
	Preferences simpleMenus ifFalse: [
		self
			fillIn: menu
			from: {
				{ 'open...'. { self. #openWindow } }.
				{ 'windows...'. { self. #windowsDo } }.
				{ 'changes...'. { self. #changesDo } }
			} ].
	self
		fillIn: menu
		from: {
			{ 'help...'. { self. #helpDo }. 'puts up a menu of useful items for updating the system, determining what version you are running, and much else' }.
			{ 'appearance...'. { self. #appearanceDo }. 'put up a menu offering many controls over appearance.' }
		}.
	Preferences simpleMenus ifFalse: [
		self
			fillIn: menu
			from: {
				{ 'do...'. { Utilities. #offerCommonRequests }. 'put up an editible list of convenient expressions, and evaluate the one selected.' }
			} ]! !

!TheWorldMenu methodsFor: 'construction' stamp: 'gm 2/28/2003 01:42'!
alphabeticalMorphMenu
	| list splitLists menu firstChar lastChar subMenu |
	list := Morph withAllSubclasses select: [:m | m includeInNewMorphMenu].
	list := list asArray sortBy: [:c1 :c2 | c1 name < c2 name].
	splitLists := self splitNewMorphList: list depth: 3.
	menu := MenuMorph new defaultTarget: self.
	1 to: splitLists size
		do: 
			[:i | 
			firstChar := i = 1 
				ifTrue: [$A]
				ifFalse: 
					[((splitLists at: i - 1) last name first asInteger + 1) 
								asCharacter].
			lastChar := i = splitLists size 
						ifTrue: [$Z]
						ifFalse: [(splitLists at: i) last name first].
			subMenu := MenuMorph new.
			(splitLists at: i) do: 
					[:cl | 
					subMenu 
						add: cl name
						target: self
						selector: #newMorphOfClass:event:
						argument: cl].
			menu add: firstChar asString , ' - ' , lastChar asString subMenu: subMenu].
	^menu! !

!TheWorldMenu methodsFor: 'popups' stamp: 'RAA 5/26/2000 10:25'!
appearanceDo
	"Build and show the appearance menu for the world."

	self doPopUp: self appearanceMenu! !

!TheWorldMenu methodsFor: 'construction' stamp: 'mt 8/5/2016 10:15'!
appearanceMenu
	"Build the appearance menu for the world."

	^self fillIn: (self menu: 'appearance...') from: {

		{'preferences...' . { self . #openPreferencesBrowser} . 'Opens a "Preferences Browser" which allows you to alter many settings' } .
		{'choose set of preferences...' . { Preferences . #offerThemesMenu} . 'Presents you with a menu of sets or preferences; each item''s balloon-help will tell you about the particular set.  If you choose one, many different preferences that come along are set at the same time; you can subsequently change any settings by using a Preferences Panel'} .
		nil .
		{'system fonts...' . { self . #standardFontDo} . 'Choose the standard fonts to use for code, lists, menus, window titles, etc.'}.
		{'text highlight color...' . { Preferences . #chooseTextHighlightColor} . 'Choose which color should be used for text highlighting in Morphic.'}.
		{'insertion point color...' . { Preferences . #chooseInsertionPointColor} . 'Choose which color to use for the text insertion point in Morphic.'}.
		{'keyboard focus color' . { Preferences . #chooseKeyboardFocusColor} . 'Choose which color to use for highlighting which pane has the keyboard focus'}.
		nil.
		{#menuColorString . { self . #toggleMenuColorPolicy} . 'Governs whether menu colors should be derived from the desktop color.'}.
		{#roundedCornersString . { self . #toggleRoundedCorners} . 'Governs whether morphic windows and menus should have rounded corners.'}.
		nil.
		{'full screen on' . { DisplayScreen . #fullScreenOn} . 'puts you in full-screen mode, if not already there.'}.
		{'full screen off' . { DisplayScreen . #fullScreenOff} . 'if in full-screen mode, takes you out of it.'}.
		nil.
		{'set display depth...' . {self. #setDisplayDepth} . 'choose how many bits per pixel.'}.
		{'set desktop color...' . {self. #changeBackgroundColor} . 'choose a uniform color to use as desktop background.'}.
		{'set gradient color...' . {self. #setGradientColor} . 'choose second color to use as gradient for desktop background.'}.
		{'use texture background' . { #myWorld . #setStandardTexture} . 'apply a graph-paper-like texture background to the desktop.'}.
		nil.
		{'clear turtle trails from desktop' . { #myWorld . #clearTurtleTrails} . 'remove any pigment laid down on the desktop by objects moving with their pens down.'}.
		{'pen-trail arrowhead size...' . { Preferences. #setArrowheads} . 'choose the shape to be used in arrowheads on pen trails.'}.

	}! !

!TheWorldMenu methodsFor: 'construction' stamp: 'cbr 3/26/2010 01:26'!
buildWorldMenu
	"Build the menu that is put up when the screen-desktop is clicked on"
	| menu |
	menu := MenuMorph new defaultTarget: self.
	menu commandKeyHandler: self.
	self colorForDebugging: menu.
	menu addStayUpItem.
	self makeConvenient: menu.
	Smalltalk at: #ServiceGUI ifPresent:[:sgui|
		sgui worldMenu: menu.
		sgui onlyServices ifTrue: [^ menu].
	].
	self addProjectEntries: menu.
	myWorld addUndoItemsTo: menu.
	self addRestoreDisplay: menu.
	self addUtilities: menu.
	self addObjectsAndTools: menu.
	self addPrintAndDebug: menu.
	self addSaveAndQuit: menu.
	^ menu! !

!TheWorldMenu methodsFor: 'commands' stamp: 'cmm 5/12/2011 21:10'!
changeBackgroundColor
	"Let the user select a new background color for the world"
	myWorld
		changeColorTarget: myWorld
		selector: #color:
		originalColor: myWorld color asColor
		hand: myWorld activeHand! !

!TheWorldMenu methodsFor: 'popups' stamp: 'RAA 5/26/2000 10:26'!
changesDo
	"Build the changes menu for the world."

	self doPopUp: self changesMenu! !

!TheWorldMenu methodsFor: 'construction' stamp: 'cbr 8/2/2010 18:34'!
changesMenu
        "Build the changes menu for the world."

        | menu |
        menu := self menu: 'changes...'.
        self fillIn: menu from: {
                { 'file out current change set' . { ChangeSet current . #verboseFileOut}.
                                'Write the current change set out to a file whose name reflects the change set name and the current date & time.'}.
                { 'create new change set...' . { ChangeSet . #newChangeSet}. 'Create a new change set and make it the current one.'}.
                { 'browse changed methods' . { ChangeSet  . #browseChangedMessages}.  'Open a message-list browser showing all methods in the current change set'}.
                { 'check change set for slips' . { self  . #lookForSlips}.
                                'Check the current change set for halts, references to the Transcript, etc., and if any such thing is found, open up a message-list browser detailing all possible slips.'}.

                nil.
                { 'simple change sorter' . {self. #openChangeSorter1}.  'Open a 3-paned changed-set viewing tool'}.
                { 'dual change sorter' . {self. #openChangeSorter2}.
                                'Open a change sorter that shows you two change sets at a time, making it easy to copy and move methods and classes between them.'}.
               { 'find a change sorter (C)' . { #myWorld . #findAChangeSorter: }. 'Brings an open change sorter to the front, creating one if necessary, and makes it the active window'}.
                nil.
                { 'browse recent submissions (R)' . { Utilities . #browseRecentSubmissions}.
                                'Open a new recent-submissions browser.  A recent-submissions browser is a message-list browser that shows the most recent methods that have been submitted.  If you submit changes within that browser, it will keep up-to-date, always showing the most recent submissions.'}.
			{ 'browse my changes' . { SystemNavigation new . #browseMyChanges }.
					'Browse all of my changes since the last time #condenseSources was run.'}.
			nil.
                { 'recently logged changes...' . { self . #browseRecentLog}.'Open a change-list browser on the latter part of the changes log.  You can use this browser to recover logged changes which were not saved in your image, in the event of a crash or other interruption.'}.

                { 'recent log file...' . { Smalltalk . #writeRecentToFile}.
                                'Create a file holding the logged changes (going as far back as you wish), and open a window on that file.'}.

                nil.
                { 'save world as morph file' . {self. #saveWorldInFile}. 'Save a file that, when reloaded, reconstitutes the current World.'}.
                nil.
        }.
        ^ menu! !

!TheWorldMenu methodsFor: 'commands' stamp: 'ar 8/6/2009 19:15'!
cleanUpWorld
	(UIManager default confirm:
'This will remove all windows except those
containing unsubmitted text edits, and will
also remove all non-window morphs (other
than flaps) found on the desktop.  Are you
sure you want to do this?' translated)
		ifFalse: [^ self].

	myWorld allNonFlapRelatedSubmorphs do:
		[:m | m delete].
	(SystemWindow windowsIn: myWorld satisfying: [:w | w model canDiscardEdits])
		do: [:w | w delete]! !

!TheWorldMenu methodsFor: 'construction' stamp: 'RAA 6/16/2000 09:50'!
colorForDebugging: aMenu

        "aMenu color: self myMenuColor"

        "aMenu color: Color lightRed"

! !

!TheWorldMenu methodsFor: 'action' stamp: 'sw 12/4/2001 21:02'!
commandKeyTypedIntoMenu: evt
	"The user typed a command-key into the given menu; dispatch it"

	myWorld keystrokeInWorld: evt ! !

!TheWorldMenu methodsFor: 'popups' stamp: 'RAA 5/26/2000 10:26'!
debugDo

	self doPopUp: self debugMenu! !

!TheWorldMenu methodsFor: 'construction' stamp: 'mt 4/12/2016 17:09'!
debugMenu

        | menu |

        menu := self menu: 'debug...'.
        self fillIn: menu from: { 
                { 'inspect world' . { #myWorld . #inspect } }.
                { 'explore world' . { #myWorld . #explore } }.
                { 'inspect model' . { myWorld model . #inspect } }.
                        " { 'talk to world...' . { self . #typeInMessageToWorld } }."
                { 'start MessageTally' . { self . #startMessageTally } }.
                { 'start/browse MessageTally' . { self . #startThenBrowseMessageTally } }.
                { 'open process browser' . { self . #openProcessBrowser } }.
                nil.
                        "(self hasProperty: #errorOnDraw) ifTrue:  Later make this come up only when needed."
                { 'start drawing again' . { #myWorld . #resumeAfterDrawError } }.
                { 'start stepping again' . { #myWorld . #resumeAfterStepError } }.
                nil.
                { 'call #tempCommand' . { #myWorld . #tempCommand } }.
                { 'define #tempCommand' . { #myWorld . #defineTempCommand } }.
        }.
	self haltOnceEnabled
		ifTrue: [menu
				add: 'disable halt/inspect once' translated
				target: menu
				action: #clearHaltOnce]
		ifFalse: [menu
				add: 'enable halt/inspect once' translated
				target: menu
				action: #setHaltOnce].
	^menu
	! !

!TheWorldMenu methodsFor: 'action' stamp: 'ar 3/17/2001 23:46'!
doMenuItem: aCollection with: event
	| realTarget selector nArgs |
	selector := aCollection second.
	nArgs := selector numArgs.
	realTarget := aCollection first.
	realTarget == #myWorld ifTrue: [realTarget := myWorld].
	realTarget == #myHand ifTrue: [realTarget := myHand].
	realTarget == #myProject ifTrue: [realTarget := self projectForMyWorld].
	^nArgs = 0 
		ifTrue:[realTarget perform: selector]
		ifFalse:[realTarget perform: selector with: event].
! !

!TheWorldMenu methodsFor: 'popups' stamp: 'RAA 6/12/2000 09:13'!
doPopUp: aMenu

	aMenu popUpForHand: myHand in: myWorld.
! !

!TheWorldMenu methodsFor: 'construction' stamp: 'nice 12/27/2009 03:11'!
fillIn: aMenu from: dataForMenu
	"A menu constructor utility by RAA.  dataForMenu is a list of items which mean:
			nil							Indicates to add a line

			first element is symbol		Add updating item with the symbol as the wording selector
			second element is a list		second element has the receiver and selector

			first element is a string		Add menu item with the string as its wording
			second element is a list		second element has the receiver and selector

			a third element exists		Use it as the balloon text
			a fourth element exists		Use it as the enablement selector (updating case only)"
	

	dataForMenu do: [ :itemData | | item |
		itemData ifNil: [aMenu addLine] ifNotNil:
			[item := (itemData first isKindOf: Symbol)
				ifTrue: 
					[aMenu 
						addUpdating: itemData first 
						target: self 
						selector: #doMenuItem:with: 
						argumentList: {itemData second}]
				 ifFalse:
					[aMenu 
						add: itemData first translated
						target: self 
						selector: #doMenuItem:with: 
						argumentList: {itemData second}].
			itemData size >= 3 ifTrue:
				[aMenu balloonTextForLastItem: itemData third translated.
			itemData size >= 4 ifTrue:
				[item enablementSelector: itemData fourth]]]].

	^ aMenu! !

!TheWorldMenu methodsFor: 'commands' stamp: 'nk 2/15/2004 09:37'!
garbageCollect
	"Do a garbage collection, and report results to the user."

	Utilities garbageCollectAndReport! !

!TheWorldMenu methodsFor: 'popups' stamp: 'RAA 5/26/2000 10:26'!
helpDo
	"Build and show the help menu for the world."

	self doPopUp: self helpMenu! !

!TheWorldMenu methodsFor: 'construction' stamp: 'mt 8/2/2016 12:55'!
helpMenu
        "Build the help menu for the world."
        |  menu |

  	menu := self menu: 'help...'.

        self fillIn: menu from:
        {
                {'about this system...'. {Smalltalk. #aboutThisSystem}. 'current version information.'}.
                {'update code from server'. {MCMcmUpdater. #updateFromServer}. 'load latest code updates via the internet'}.
                {'preferences...'. {self. #openPreferencesBrowser}. 'view and change various options.'}.
			 {'set language...' . {Project. #chooseNaturalLanguage}. 'choose the language in which tiles should be displayed.'} .
                nil.
               {'command-key help'. { TheWorldMainDockingBar instance . #commandKeyHelp}. 'summary of keyboard shortcuts.'}
	}.

	self addGestureHelpItemsTo: menu.

	self fillIn: menu from:
	{
                {'world menu help'. { self . #worldMenuHelp}. 'helps find menu items buried in submenus.'}.
                        "{'info about flaps' . { Utilities . #explainFlaps}. 'describes how to enable and use flaps.'}."
                {'font size summary' . { TextStyle . #fontSizeSummary}.  'summary of names and sizes of available fonts.'}.
                {'useful expressions' . { Utilities . #openStandardWorkspace}. 'a window full of useful expressions.'}.
			 {'annotation setup...' . { Preferences . #editAnnotations}. 'Click here to get a little window that will allow you to specify which types of annotations, in which order, you wish to see in the annotation panes of browsers and other tools'}.
			nil.
                {'graphical imports' . { Imports default . #viewImages}.  'view the global repository called ImageImports; you can easily import external graphics into ImageImports via the FileList'}.
                {'standard graphics library' . { ScriptingSystem . #inspectFormDictionary}.  'lets you view and change the system''s standard library of graphics.'}.
                nil.
                {'telemorphic...' . {self. #remoteDo}.  'commands for doing multi-machine "telemorphic" experiments'}.
                {#soundEnablingString . { SoundService . #toggleSoundEnabled}. 'turning sound off will completely disable Squeak''s use of sound.'}.
                nil.

                {'set author initials...' . { Utilities . #setAuthorInitials }. 'supply initials to be used to identify the author of code and other content.'}.
                {'vm statistics' . { self . #vmStatistics}.  'obtain some intriguing data about the vm.'}.
			  nil.
			  {'purge undo records' . { CommandHistory . #resetAllHistory }. 'save space by removing all the undo information remembered in all projects.'}.
                {'space left' . { self . #garbageCollect}. 'perform a full garbage-collection and report how many bytes of space remain in the image.'}.
        }.

	^menu

! !

!TheWorldMenu methodsFor: 'commands' stamp: 'laza 3/5/2010 19:47'!
loadProject

	| stdFileMenuResult path |
	"Put up a Menu and let the user choose a '.project' file to load.  Create a thumbnail and jump into the project."

	Project canWeLoadAProjectNow ifFalse: [^ self].
	path := FileList2 modalFolderSelector.
	path ifNil: [^ nil].
	stdFileMenuResult := ((StandardFileMenu new) pattern: '*.pr'; 
		oldFileFrom: path) 
			startUpWithCaption: 'Select a File:' translated.
	stdFileMenuResult ifNil: [^ nil].
	ProjectLoading 
		openFromDirectory: stdFileMenuResult directory 
		andFileName: stdFileMenuResult name
! !

!TheWorldMenu methodsFor: 'commands' stamp: 'sd 5/23/2003 14:49'!
lookForSlips

	ChangeSet current lookForSlips! !

!TheWorldMenu methodsFor: 'construction' stamp: 'ar 1/4/2010 18:50'!
makeConvenient: menu
	self
		fillIn: menu
		from: {
			{ 'Browser'. { StandardToolSet. #openClassBrowser }. 'open a browser' }.
			{ 'Workspace'. { Workspace. #open }. 'open a workspace' }.
			{ 'Transcript'. { Transcript. #open }. 'open a transcript' }.
			Smalltalk at: #TestRunner ifPresent:[:aClass|
				{ 'Test Runner'. { aClass. #open }. 'open a test runner' }.
			].
			nil
		}! !

!TheWorldMenu methodsFor: 'action' stamp: 'RAA 5/26/2000 10:18'!
menuColorString

	^ Preferences menuColorString! !

!TheWorldMenu methodsFor: 'mechanics' stamp: 'dgd 8/26/2003 21:05'!
menu: titleString
	"Create a menu with the given title, ready for filling"

	| menu |
	(menu := MenuMorph entitled: titleString translated) 
		defaultTarget: self; 
		addStayUpItem;
		commandKeyHandler: self.
	self colorForDebugging: menu.
	^ menu
! !

!TheWorldMenu methodsFor: 'commands' stamp: 'RAA 5/24/2000 17:31'!
mvcProjectsAllowed

	^Preferences mvcProjectsAllowed and: [Smalltalk includesKey: #StandardSystemView]! !

!TheWorldMenu methodsFor: 'construction' stamp: 'md 2/22/2006 21:24'!
newMorph
	"The user requested 'new morph' from the world menu.  Put up a menu that allows many ways of obtaining new morphs."

	| menu |

	menu := self menu: 'Add a new morph'.
	menu 
		add: 'from paste buffer' translated target: myHand action: #pasteMorph;
		add: 'from alphabetical list' translated subMenu: self alphabeticalMorphMenu;
		add: 'from a file...' translated target: self action: #readMorphFromAFile.
	menu addLine.
	menu add: 'grab rectangle from screen' translated target: myWorld action: #grabDrawingFromScreen:;
		add: 'grab with lasso from screen' translated target: myWorld action: #grabLassoFromScreen:;
		add: 'grab rubber band from screen' translated target: myWorld action: #grabRubberBandFromScreen:;
		add: 'grab flood area from screen' translated target: myWorld action: #grabFloodFromScreen:.
	menu addLine.
	menu add: 'make new drawing' translated target: myWorld action: #newDrawingFromMenu:;
		add: 'make link to project...' translated target: self action: #projectThumbnail.

	self doPopUp: menu.
! !

!TheWorldMenu methodsFor: 'commands' stamp: 'RAA 5/25/2000 07:42'!
newMorphOfClass: morphClass event: evt
	"Attach a new morph of the given class to the invoking hand."

	| m |
	m := morphClass new.
	m installModelIn: myWorld.  "a chance to install model pointers"
	m wantsToBeOpenedInWorld
		ifTrue:[myWorld addMorph: m]
		ifFalse:[evt hand attachMorph: m].
	myWorld startSteppingSubmorphsOf: m.
! !

!TheWorldMenu methodsFor: 'commands' stamp: 'ar 9/27/2005 20:33'!
openBrowser 
	"Create and schedule a Browser view for browsing code."
	ToolSet browse: nil selector: nil! !

!TheWorldMenu methodsFor: 'commands' stamp: 'asm 1/17/2006 19:17'!
openFileDirectly

	FileList openFileDirectly! !

!TheWorldMenu methodsFor: 'commands' stamp: 'ar 10/4/2009 22:18'!
openFileList
	FileList open.! !

!TheWorldMenu methodsFor: 'construction' stamp: 'dtl 3/3/2011 21:06'!
openMenu
	"Build the open window menu for the world."

	| menu |
	menu := self menu: 'open...'.
	menu defaultTarget: ToolSet default.
	menu addList: ToolSet menuItems.
	menu defaultTarget: self.
	self fillIn: menu from: {
		nil.
		{'file...' . { self . #openFileDirectly} . 'Lets you open a window on a single file'}.
		{'transcript (t)' . {self . #openTranscript}. 'A window used to report messages sent to Transcript' }.
		"{'inner world' . { WorldWindow . #test1} }."
		nil.
	}.
	self fillIn: menu from: self class registeredOpenCommands.
	menu addLine.

	self mvcProjectsAllowed ifTrue:
		[self fillIn: menu from: { {'mvc project' . {self. #openMVCProject} . 'Creates a new project of the classic "mvc" style'} }].

	self fillIn: menu from: { 
		{'morphic project' . {self. #openMorphicProject} . 'Creates a new morphic project'}.
	}.
	Smalltalk at: #SMxMorphicProject ifPresent: [:p |
		self fillIn: menu from: { 
			{ 'simple morphic project' . { self . #openSMxMorphicProject } . 'Creates a new simple morphic project' }.
		}
	].
	^menu
! !

!TheWorldMenu methodsFor: 'commands' stamp: 'ar 9/18/2009 20:12'!
openMorphicProject
	"Open a morphic project from within a morphic project"
	MorphicProject openViewOn: nil
! !

!TheWorldMenu methodsFor: 'commands' stamp: 'ar 9/18/2009 21:29'!
openMVCProject
	"Open a new MVC project (only if MVC is present)"
	Smalltalk at: #MVCProject ifPresent:[:projClass|
		ProjectViewMorph openOn: projClass new.
	].! !

!TheWorldMenu methodsFor: 'commands' stamp: 'ar 1/2/2010 15:42'!
openPreferencesBrowser
	"Open a preferences browser"
	^Smalltalk at: #PreferenceBrowser ifPresent:[:pb| pb open].
! !

!TheWorldMenu methodsFor: 'commands' stamp: 'ar 8/7/2009 22:28'!
openTranscript

	Transcript openLabel: 'Transcript'! !

!TheWorldMenu methodsFor: 'popups' stamp: 'RAA 5/26/2000 10:26'!
openWindow

	self doPopUp: self openMenu! !

!TheWorldMenu methodsFor: 'commands' stamp: 'ar 9/27/2005 20:49'!
openWorkspace

	UIManager default edit: '' label: 'Workspace'! !

!TheWorldMenu methodsFor: 'popups' stamp: 'RAA 5/26/2000 10:26'!
projectDo
	"Build and show the project menu for the world."

	self doPopUp: self projectMenu! !

!TheWorldMenu methodsFor: 'commands' stamp: 'RAA 6/15/2000 10:17'!
projectForMyWorld

        ^myProject ifNil: [myProject := myWorld project]! !

!TheWorldMenu methodsFor: 'construction' stamp: 'dtl 3/3/2011 20:59'!
projectMenu
	"Build the project menu for the world."
	| menu |

	self flag: #bob0302.

	menu := self menu: 'projects...'.
	self fillIn: menu from: { 
		{ 'save on server (also makes a local copy)' . { #myProject . #storeOnServer } }.
		{ 'save to a different server' . { #myProject . #saveAs } }.
		{ 'save project on local file only' . { #myWorld . #saveOnFile } }.
		{ 'see if server version is more recent...' . { #myProject . #loadFromServer } }.
		{ 'load project from file...' . { self . #loadProject } }.
		nil.
	}.

	self mvcProjectsAllowed ifTrue: [
		self fillIn: menu from: {
			{ 'create new mvc project'. { self . #openMVCProject } }.
		}
	].
	self fillIn: menu from: { 
		{ 'create new morphic project' . { self . #openMorphicProject } }.
	}.
	Smalltalk at: #SMxMorphicProject ifPresent: [:p |
		self fillIn: menu from: { 
			{ 'create new simple morphic project' . { self . #openSMxMorphicProject } }.
		}
	].
	self fillIn: menu from: { 
		nil.
		{ 'go to previous project' . { Project . #returnToPreviousProject } }.
		{ 'go to next project' . { Project . #advanceToNextProject } }.
		{ 'jump to project...' . { #myWorld . #jumpToProject } }.
	}.
	Preferences simpleMenus ifFalse: [
		self fillIn: menu from: { 
			nil.
			{ 'save for future revert' . { #myProject . #saveForRevert } }.
			{ 'revert to saved copy' . { #myProject . #revert } }.
		}.
	].

	^ menu! !

!TheWorldMenu methodsFor: 'action' stamp: 'ar 9/7/2009 14:33'!
projectThumbnail
	"Offer the user a menu of project names. Attach to the hand a thumbnail of the project the user selects."

	| projName pr names values |
	names := OrderedCollection with: Project current name, ' (current)'.
	values := OrderedCollection with: Project current name.
	Project allNames do: [:n | names add: n. values add: n].
	projName := UIManager default 
		chooseFrom: names values: values lines: #(1) title: 'Select a project'.
	projName ifNotNil:
		[(pr := Project named: projName) 
			ifNotNil: [myHand attachMorph: (ProjectViewMorph on: pr)]
			ifNil: [self inform: 'can''t seem to find that project']].! !

!TheWorldMenu methodsFor: 'commands' stamp: 'fbs 7/25/2013 07:21'!
quitSession

	Smalltalk
		snapshot: (UserDialogBoxMorph 
			confirm: 'Save changes before quitting?' translated 
			orCancel: [^ self]
			at: World center)
		andQuit: true! !

!TheWorldMenu methodsFor: 'commands' stamp: 'dgd 9/21/2003 13:20'!
readMorphFromAFile
	"Produce a morph from a file -- either a saved .morph file or a graphics file"

	| morphOrList ff aName f m |
	aName := Utilities chooseFileWithSuffixFromList:
(#('.morph'), Utilities graphicsFileSuffixes) withCaption: 'Choose a file
to load' translated.
	aName ifNil: [^ self].  "User made no choice"
	aName == #none ifTrue: [^ self inform: 
'Sorry, no suitable files found
(names should end with .morph, .gif,
.bmp, .jpeg, .jpe, .jp, or .form)' translated].

	(aName asLowercase endsWith: '.morph')
		ifTrue:
			[ff := FileStream readOnlyFileNamed: aName.
			morphOrList := ff fileInObjectAndCode.		"code filed in is the Model class"
			"the file may contain either a single morph or an array of morphs"
			myWorld addMorphsAndModel: morphOrList]
		ifFalse:
			[f := Form fromFileNamed: aName.
			f ifNil: [^ self error: 'unrecognized image file format' translated].
			m := myWorld drawingClass new form: f.
			myHand attachMorph: m]
! !

!TheWorldMenu methodsFor: 'popups' stamp: 'RAA 5/26/2000 10:26'!
remoteDo

	self doPopUp: self remoteMenu! !

!TheWorldMenu methodsFor: 'construction' stamp: 'Richo 5/18/2010 14:13'!
remoteMenu
        "Build the Telemorphic menu for the world."

        ^self fillIn: (self menu: 'Telemorphic' translatedNoop) from: {
                { 'local host address' translatedNoop. { #myWorld . #reportLocalAddress } }.
                { 'connect remote user' translatedNoop. { #myWorld . #connectRemoteUser } }.
                { 'disconnect remote user' translatedNoop. { #myWorld . #disconnectRemoteUser } }.
                { 'disconnect all remote users' translatedNoop. { #myWorld . #disconnectAllRemoteUsers } }.
        }! !

!TheWorldMenu methodsFor: 'action' stamp: 'mt 7/28/2016 10:43'!
roundedCornersString

	^ (((SystemWindow roundedWindowCorners)
		ifTrue: ['stop']
		ifFalse: ['start']) , ' rounding window corners') translated! !

!TheWorldMenu methodsFor: 'commands' stamp: 'RAA 5/24/2000 17:54'!
saveAndQuit

	Smalltalk snapshot: true andQuit: true! !

!TheWorldMenu methodsFor: 'action' stamp: 'ar 8/2/2010 20:51'!
saveScreenshot
	"Make a screenshot of the world and save it to a file"

	SampledSound playSoundNamed: 'camera'.
	PNGReadWriter putForm: myWorld imageForm onFileNamed:
		(FileDirectory default nextNameFor: 'SqueakScreen' extension:'png').
! !

!TheWorldMenu methodsFor: 'commands' stamp: 'ar 8/6/2009 20:46'!
saveWorldInFile
	"Save the world's submorphs, model, and stepList in a file.  "

	| fileName fileStream aClass |
	fileName := UIManager default request: 'File name for this morph?'.
	fileName isEmpty ifTrue: [^ self].  "abort"

	"Save only model, stepList, submorphs in this world"
	myWorld submorphsDo: [:m |
		m allMorphsDo: [:subM | subM prepareToBeSaved]].	"Amen"

	fileStream := FileStream newFileNamed: fileName, '.morph'.
	aClass := myWorld model ifNil: [nil] ifNotNil: [myWorld model class].
	fileStream fileOutClass: aClass andObject: myWorld.
! !

!TheWorldMenu methodsFor: 'commands' stamp: 'MarcoSchmidt 10/27/2009 16:46'!
setDisplayDepth
	"Let the user choose a new depth for the display. "

	| result oldDepth allDepths allLabels hasBoth |
	oldDepth := Display nativeDepth.
	allDepths := #(1 -1 2 -2 4 -4 8 -8 16 -16 32 -32) select: [:d | Display supportsDisplayDepth: d].
	hasBoth := (allDepths anySatisfy:[:d| d > 0]) and:[allDepths anySatisfy:[:d| d < 0]].
	allLabels := allDepths collect:[:d|
		String streamContents:[:s|
			s nextPutAll: (d = oldDepth ifTrue:['<on>'] ifFalse:['<off>']).
			s print: d abs.
			hasBoth ifTrue:[s nextPutAll: (d > 0 ifTrue:['  (big endian)'] ifFalse:['  (little endian)'])].
		]].
	result := UIManager default
		chooseFrom: allLabels 
		values: allDepths 
		title: 'Choose a display depth' translated.
	result ifNotNil: [Display newDepth: result].
	oldDepth := oldDepth abs.
	(Smalltalk isMorphic and: [(Display depth < 4) ~= (oldDepth < 4)])
		ifTrue:
			["Repaint windows since they look better all white in depth < 4"
			(SystemWindow windowsIn: myWorld satisfying: [:w | true]) do:
				[:w |
				oldDepth < 4
					ifTrue: [w restoreDefaultPaneColor]
					ifFalse: [w updatePaneColors]]]! !

!TheWorldMenu methodsFor: 'action' stamp: 'RAA 6/27/2000 09:04'!
setGradientColor

	myWorld setGradientColor: myHand lastEvent! !

!TheWorldMenu methodsFor: 'action' stamp: 'nice 9/21/2013 23:14'!
soundEnablingString

	^ SoundService soundEnablingString! !

!TheWorldMenu methodsFor: 'commands' stamp: 'aoy 2/15/2003 21:19'!
splitNewMorphList: list depth: d 
	| middle c prev next out |
	d <= 0 ifTrue: [^Array with: list].
	middle := list size // 2 + 1.
	c := (list at: middle) name first.
	prev := middle - 1.
	[prev > 0 and: [(list at: prev) name first = c]] 
		whileTrue: [prev := prev - 1].
	next := middle + 1.
	[next <= list size and: [(list at: next) name first = c]] 
		whileTrue: [next := next + 1].
	"Choose the better cluster"
	middle := middle - prev < (next - middle) 
				ifTrue: [prev + 1]
				ifFalse: [next]. 
	middle = 1 ifTrue: [middle := next].
	middle >= list size ifTrue: [middle := prev + 1].
	(middle = 1 or: [middle >= list size]) ifTrue: [^Array with: list].
	out := WriteStream on: Array new.
	out nextPutAll: (self splitNewMorphList: (list copyFrom: 1 to: middle - 1)
				depth: d - 1).
	out 
		nextPutAll: (self splitNewMorphList: (list copyFrom: middle to: list size)
				depth: d - 1).
	^out contents! !

!TheWorldMenu methodsFor: 'action' stamp: 'RAA 5/26/2000 11:31'!
staggerPolicyString

	^ Preferences staggerPolicyString! !

!TheWorldMenu methodsFor: 'popups' stamp: 'RAA 5/26/2000 10:27'!
standardFontDo
	"Build and show the standard font menu"

	self doPopUp: Preferences fontConfigurationMenu! !

!TheWorldMenu methodsFor: 'commands' stamp: 'mt 2/16/2016 09:46'!
startMessageTally

	(self confirm: 'MessageTally will start now,
and stop when the cursor goes
to the top of the screen') ifTrue:
		[MessageTally spyOn:
			[[Sensor peekPosition y > 0] whileTrue: [World doOneCycle]]]! !

!TheWorldMenu methodsFor: 'windows & flaps menu' stamp: 'dao 10/1/2004 13:12'!
suppressFlapsString
	"Answer the wording of the suppress-flaps item"

	^ Project current suppressFlapsString! !

!TheWorldMenu methodsFor: 'action' stamp: 'mt 8/27/2015 10:14'!
toggleMenuColorPolicy

	Preferences toggle: #menuColorFromWorld.! !

!TheWorldMenu methodsFor: 'action' stamp: 'mt 7/28/2016 10:38'!
toggleRoundedCorners

	SystemWindow roundedWindowCorners: SystemWindow roundedWindowCorners not.! !

!TheWorldMenu methodsFor: 'action' stamp: 'mt 8/27/2015 10:12'!
toggleWindowPolicy

	Preferences toggle: #reverseWindowStagger.! !

!TheWorldMenu methodsFor: 'commands' stamp: 'fbs 7/25/2013 07:21'!
vmStatistics
	"Open a string view on a report of vm statistics"

	(StringHolder new contents: Smalltalk vmStatisticsReportString)
		openLabel: 'VM Statistics'! !

!TheWorldMenu methodsFor: 'windows & flaps menu' stamp: 'RAA 5/26/2000 10:27'!
windowsDo
	"Build the windows menu for the world."

	self doPopUp: self windowsMenu! !

!TheWorldMenu methodsFor: 'windows & flaps menu' stamp: 'kfr 5/18/2011 22:20'!
windowsMenu
        "Build the windows menu for the world."

        ^ self fillIn: (self menu: 'windows' translatedNoop) from: {  
                { 'find window' translatedNoop. { #myWorld . #findWindow: }. 'Presents a list of all windows; if you choose one from the list, it becomes the active window.' translatedNoop}.

                { 'find changed browsers...' translatedNoop. { #myWorld . #findDirtyBrowsers: }. 'Presents a list of browsers that have unsubmitted changes; if you choose one from the list, it becomes the active window.' translatedNoop}.

                { 'find changed windows...' translatedNoop. { #myWorld . #findDirtyWindows: }. 'Presents a list of all windows that have unsubmitted changes; if you choose one from the list, it becomes the active window.' translatedNoop}.
			nil.

                { 'find a transcript (t)' translatedNoop. { #myWorld . #findATranscript: }. 'Brings an open Transcript to the front, creating one if necessary, and makes it the active window' translatedNoop}.

               { 'find a fileList (L)' translatedNoop. { #myWorld . #findAFileList: }. 'Brings an open fileList to the front, creating one if necessary, and makes it the active window' translatedNoop}.

               { 'find a change sorter (C)' translatedNoop. { #myWorld . #findAChangeSorter: }. 'Brings an open change sorter to the front, creating one if necessary, and makes it the active window' translatedNoop}.

			{ 'find message names (W)' translatedNoop. { #myWorld . #findAMessageNamesWindow: }. 'Brings an open MessageNames window to the front, creating one if necessary, and makes it the active window' translatedNoop}.

			 nil.
                { #staggerPolicyString . { self . #toggleWindowPolicy }. 'stagger: new windows positioned so you can see a portion of each one.
                tile: new windows positioned so that they do not overlap others, if possible.' translatedNoop}.

                nil.
                { 'collapse all windows' translatedNoop. { #myWorld . #collapseAllWindows }. 'Reduce all open windows to collapsed forms that only show titles.' translatedNoop}.
                { 'collapse all objects' translatedNoop. { #myWorld . #collapseAllWindowsAndNonWindows }. 'Reduce all open windows and all other objects on the desktop to labeled tabs' translatedNoop}.
                { 'expand all' translatedNoop. { #myWorld . #expandAllCollapsedObjects }. 'Expand all collapsed windows and other collapsed objects back to their expanded forms.' translatedNoop}.
		
                { 'close top window (w)' translatedNoop. { SystemWindow . #closeTopWindow }. 'Close the topmost window if possible.' translatedNoop}.
                { 'send top window to back (\)' translatedNoop. { SystemWindow . #sendTopWindowToBack  }. 'Make the topmost window become the backmost one, and activate the window just beneath it.' translatedNoop}.
			 { 'move windows onscreen' translatedNoop. { #myWorld . #bringWindowsFullOnscreen }. 'Make all windows fully visible on the screen' translatedNoop}.

                nil.
                { 'delete unchanged windows' translatedNoop. { #myWorld . #closeUnchangedWindows }. 'Deletes all windows that do not have unsaved text edits.' translatedNoop}.
                { 'delete non-windows' translatedNoop. { #myWorld . #deleteNonWindows }. 'Deletes all non-window morphs lying on the world.' translatedNoop}.
                { 'delete both of the above' translatedNoop. { self . #cleanUpWorld }. 'deletes all unchanged windows and also all non-window morphs lying on the world, other than flaps.' translatedNoop}.

        }! !

!TheWorldMenu methodsFor: 'commands' stamp: 'nice 12/27/2009 20:20'!
worldMenuHelp
	| explanation aList |
	"self currentWorld primaryHand worldMenuHelp"

	aList := OrderedCollection new.
	#(helpMenu changesMenu openMenu debugMenu projectMenu scriptingMenu windowsMenu playfieldMenu appearanceMenu flapsMenu) 
		with:
	#('help' 'changes' 'open' 'debug' 'projects' 'authoring tools' 'windows' 'playfield options' 'appearance' 'flaps') do:
		[:sel :title | | aMenu |
		aMenu := self perform: sel.
			aMenu items do:
				[:it | | cnts |
				(((cnts := it contents) = 'keep this menu up') or: [cnts isEmpty])
					ifFalse: [aList add: (cnts, ' - ', title translated)]]].
	aList := aList asSortedCollection: [:a :b | a asLowercase < b asLowercase].

	explanation := String streamContents: [:aStream | aList do:
		[:anItem | aStream nextPutAll: anItem; cr]].

	(StringHolder new contents: explanation)
		openLabel: 'Where in the world menu is...' translated! !

!TheWorldMenu methodsFor: 'mechanics' stamp: 'RAA 5/24/2000 19:02'!
world: aWorld project: aProject hand: aHand

	myWorld := aWorld.
	myProject := aProject.
	myHand := aHand.! !

Object subclass: #WorldState
	instanceVariableNames: 'hands activeHand viewBox canvas damageRecorder stepList lastStepTime lastStepMessage lastCycleTime commandHistory alarms lastAlarmTime remoteServer multiCanvas interCycleDelay'
	classVariableNames: 'CanSurrenderToOS DeferredUIMessages DisableDeferredUpdates LastCycleTime MinCycleLapse'
	poolDictionaries: ''
	category: 'Morphic-Worlds'!

!WorldState commentStamp: 'ls 7/10/2003 19:30' prior: 0!
The state of a Morphic world.  (This needs some serious commenting!!!!)


The MinCycleLapse variable holds the minimum amount of time that a morphic cycle is allowed to take.  If a cycle takes less than this, then interCyclePause: will wait until the full time has been used up.!

!WorldState class methodsFor: 'class initialization' stamp: 'RAA 7/15/2000 12:58'!
addDeferredUIMessage: valuableObject

	self deferredUIMessages nextPut: valuableObject.

! !

!WorldState class methodsFor: 'as yet unclassified' stamp: 'RAA 8/14/2000 16:40'!
canSurrenderToOS: aBoolean

	CanSurrenderToOS := aBoolean! !

!WorldState class methodsFor: 'objects from disk' stamp: 'bf 4/11/2013 14:30'!
classVersion

	^2		"force cleanup of alarms and stepList"! !

!WorldState class methodsFor: 'class initialization' stamp: 'ar 2/26/2010 23:06'!
cleanUp
	"Reset command histories"

	self allInstances do: [ :ea | ea clearCommandHistory ].! !

!WorldState class methodsFor: 'accessing' stamp: 'sk 1/26/2003 18:29'!
deferredExecutionTimeLimit
	"Answer the maximum time in milliseconds that should be spent dispatching deferred UI messages in WorldState>>runStepMethodsIn:."

	^ 200! !

!WorldState class methodsFor: 'class initialization' stamp: 'RAA 7/15/2000 12:58'!
deferredUIMessages

	^DeferredUIMessages ifNil: [DeferredUIMessages := SharedQueue new].
! !

!WorldState class methodsFor: 'class initialization' stamp: 'RAA 7/15/2000 12:56'!
initialize
	"WorldState initialize"

	MinCycleLapse := 20.		"allows 50 frames per second..."
	DisableDeferredUpdates := false.
	DeferredUIMessages := SharedQueue new.! !

!WorldState class methodsFor: 'as yet unclassified' stamp: 'RAA 7/23/2000 17:28'!
lastCycleTime

	^LastCycleTime! !

!WorldState class methodsFor: 'objects from disk' stamp: 'bf 4/11/2013 15:15'!
withClassVersion: aVersion
	aVersion <= self classVersion ifTrue: [^self].
	^super withClassVersion: aVersion! !

!WorldState methodsFor: 'hands' stamp: 'ar 1/22/2001 14:26'!
activeHand

	^ ActiveHand! !

!WorldState methodsFor: 'alarms' stamp: 'jcg 1/9/2010 14:51'!
addAlarm: aSelector withArguments: argArray for: aTarget at: scheduledTime
	"Add a new alarm with the given set of parameters"
	self lockAlarmsDuring: [:locked |
		locked add:	(MorphicAlarm 
						scheduledAt: scheduledTime
						receiver: aTarget
						selector: aSelector
						arguments: argArray).
	]! !

!WorldState methodsFor: 'hands' stamp: 'ar 10/26/2000 14:51'!
addHand: aHandMorph
	"Add the given hand to the list of hands for this world."

	hands := (hands copyWithout: aHandMorph) copyWith: aHandMorph.
! !

!WorldState methodsFor: 'Nebraska support' stamp: 'ar 10/26/2000 14:11'!
addRemoteCanvas: c
	self canvas: nil. "force recomputation"! !

!WorldState methodsFor: 'alarms' stamp: 'ul 5/31/2016 01:18'!
adjustAlarmTimes: nowTime
	"Adjust the alarm times after some clock weirdness (such as roll-over, image-startup etc)"
	
	| deltaTime |
	(deltaTime := nowTime - lastAlarmTime) = 0 ifTrue: [ ^self ].
	self lockAlarmsDuring: [ :locked | locked incrementScheduledTimesBy: deltaTime ]! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 15:22'!
adjustWakeupTimesIfNecessary
	"Fix the wakeup times in my step list if necessary. This is needed after a snapshot, after a long pause (say because some other view had control or because the user was selecting from an MVC-style menu) or when the millisecond clock wraps around (a very rare occurence with a 32-bit clock!!)."

	| now |
	now := Time millisecondClockValue.
	((now < lastStepTime) or: [(now - lastStepTime) > 5000])
		 ifTrue: [self adjustWakeupTimes: now].  "clock slipped"
! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:23'!
adjustWakeupTimes: now
	"Fix the wakeup times in my step list. This is necessary when this world has been restarted after a pause, say because some other view had control, after a snapshot, or because the millisecond clock has wrapped around. (The latter is a rare occurence with a 32-bit clock!!)"
	| deltaTime |
	deltaTime := now - lastStepTime.
	stepList do:[:entry| entry scheduledTime: entry scheduledTime + deltaTime].
	lastStepTime := now.
! !

!WorldState methodsFor: 'alarms' stamp: 'jcg 1/9/2010 12:31'!
alarms

	^alarms ifNil: [alarms := MorphicAlarmQueue new]! !

!WorldState methodsFor: 'canvas' stamp: 'ar 10/26/2000 14:10'!
assuredCanvas
	remoteServer ifNotNil:[^self assuredRemoteCanvas].
	(canvas isNil or: [(canvas extent ~= viewBox extent) or: [canvas form depth ~= Display depth]])
		ifTrue:
			["allocate a new offscreen canvas the size of the window"
			self canvas: (Display defaultCanvasClass extent: viewBox extent)].
	^ self canvas! !

!WorldState methodsFor: 'Nebraska support' stamp: 'ar 10/26/2000 14:17'!
assuredRemoteCanvas
	| newCanvas |
	(self canvas notNil) ifTrue: [ ^self canvas ].
	newCanvas := MultiCanvas new.
	newCanvas depth: 32.
	newCanvas extent: viewBox extent.
	self remoteCanvasesDo: [ :c | newCanvas addCanvas: c ].
	newCanvas addCanvas: Display getCanvas.

	"newCanvas := CachingCanvas on: newCanvas."
	self canvas: newCanvas.
	^newCanvas! !

!WorldState methodsFor: 'canvas' stamp: 'di 6/7/1999 17:44'!
canvas

	^ canvas! !

!WorldState methodsFor: 'canvas' stamp: 'dgd 2/22/2003 13:29'!
canvas: x 
	canvas := x.
	damageRecorder isNil 
		ifTrue: [damageRecorder := DamageRecorder new]
		ifFalse: [damageRecorder doFullRepaint]! !

!WorldState methodsFor: 'update cycle' stamp: 'RAA 5/24/2000 13:13'!
checkIfUpdateNeeded

	damageRecorder updateIsNeeded ifTrue: [^true].
	hands do: [:h | (h hasChanged and: [h needsToBeDrawn]) ifTrue: [^true]].
	^false  "display is already up-to-date"
! !

!WorldState methodsFor: 'stepping' stamp: 'ul 5/31/2016 01:27'!
cleanseStepListForWorld: aWorld
	"Remove morphs from the step list that are not in this World.  Often were in a flap that has moved on to another world."

	| deletions |
	deletions := nil.
	stepList do: [:entry |
		entry receiver world == aWorld ifFalse: [
			(deletions ifNil: [ deletions := OrderedCollection new ])
				addLast: entry ] ].

	deletions ifNotNil:[
		deletions do: [:entry|
			self stopStepping: entry receiver]].

	self lockAlarmsDuring: [:locked |
		locked alarmsDoSafely: [ :entry |
			| morphToStep |
			morphToStep := entry receiver.
			(morphToStep isMorph and: [morphToStep world == aWorld]) ifFalse: [
				locked removeAlarmWithReceiver: entry receiver selector: entry selector ] ] ]! !

!WorldState methodsFor: 'undo support' stamp: 'RAA 9/21/2000 20:05'!
clearCommandHistory

	"useful prior to project saves"
	commandHistory := nil! !

!WorldState methodsFor: 'undo' stamp: 'ar 8/31/2000 22:57'!
commandHistory
	^commandHistory ifNil:[commandHistory := CommandHistory new]! !

!WorldState methodsFor: 'object fileIn' stamp: 'jcg 1/9/2010 13:28'!
convertAlarms
	"We now store the alarms in a MorphicAlarmQueue, rather than a Heap."
	alarms ifNotNil: [
		alarms class == MorphicAlarmQueue ifFalse: [
			| oldAlarms |
			oldAlarms := alarms.
			alarms := MorphicAlarmQueue new.
			oldAlarms do: [:alarm | alarms add: alarm]			
		]
	].! !

!WorldState methodsFor: 'object fileIn' stamp: 'nice 12/27/2009 03:11'!
convertStepList
	"Convert the old-style step list (an Array of Arrays) into the new-style StepMessage heap"

	| newList |
	(stepList isKindOf: Heap) 
		ifTrue: 
			[^stepList sortBlock: self stepListSortBlock	"ensure that we have a cleaner block"].
	newList := Heap sortBlock: self stepListSortBlock.
	stepList do: 
			[:entry | | wakeupTime morphToStep | 
			wakeupTime := entry second.
			morphToStep := entry first.
			newList add: (StepMessage 
						scheduledAt: wakeupTime
						stepTime: nil
						receiver: morphToStep
						selector: #stepAt:
						arguments: nil)].
	stepList := newList! !

!WorldState methodsFor: 'objects from disk' stamp: 'RAA 1/5/2001 10:51'!
convertToCurrentVersion: varDict refStream: smartRefStrm
	
	"Convert the old to new step lists"
	self convertStepList.
	self convertAlarms.
	^super convertToCurrentVersion: varDict refStream: smartRefStrm.

! !

!WorldState methodsFor: 'update cycle' stamp: 'dtl 2/27/2010 10:57'!
displayWorldSafely: aWorld
	"Update this world's display and keep track of errors during draw methods."

	[aWorld displayWorld] ifError: [:err :rcvr |
		"Handle a drawing error"
		| errCtx errMorph |
		errCtx := thisContext.
		[
			errCtx := errCtx sender.
			"Search the sender chain to find the morph causing the problem"
			[errCtx notNil and:[(errCtx receiver isMorph) not]] 
				whileTrue:[errCtx := errCtx sender].
			"If we're at the root of the context chain then we have a fatal drawing problem"
			errCtx ifNil:[^Project current handleFatalDrawingError: err].
			errMorph := errCtx receiver.
			"If the morph causing the problem has already the #drawError flag set,
			then search for the next morph above in the caller chain."
			errMorph hasProperty: #errorOnDraw
		] whileTrue.
		errMorph setProperty: #errorOnDraw toValue: true.
		"Install the old error handler, so we can re-raise the error"
		rcvr error: err.
	].! !

!WorldState methodsFor: 'update cycle' stamp: 'mt 2/25/2015 10:16'!
displayWorld: aWorld submorphs: submorphs
	"Update this world's display."

	| deferredUpdateMode handsToDraw allDamage handDamageRects worldDamageRects |

	submorphs do: [:m | m fullBounds].  "force re-layout if needed"
	self checkIfUpdateNeeded ifFalse: [^ self].  "display is already up-to-date"

	deferredUpdateMode := self doDeferredUpdatingFor: aWorld.
	deferredUpdateMode ifFalse: [self assuredCanvas].

	worldDamageRects := self drawWorld: aWorld submorphs: submorphs invalidAreasOn: canvas.  "repair world's damage on canvas"
	"self handsDo:[:h| h noticeDamageRects: worldDamageRects]."
	handsToDraw := self selectHandsToDrawForDamage: worldDamageRects.
	handDamageRects := handsToDraw collect: [:h | h savePatchFrom: canvas].
	allDamage := worldDamageRects, handDamageRects.

	handsToDraw reverseDo: [:h | canvas fullDrawMorph: h].  "draw hands onto world canvas"

	"*make this true to flash damaged areas for testing*"
	Preferences debugShowDamage ifTrue: [aWorld flashRects: allDamage color: Color black].

	canvas finish: allDamage.

	"quickly copy altered rects of canvas to Display:"
	deferredUpdateMode
		ifTrue: [self forceDamageToScreen: allDamage]
		ifFalse: [canvas showAt: aWorld viewBox origin invalidRects: allDamage].
	handsToDraw do: [:h | h restoreSavedPatchOn: canvas].  "restore world canvas under hands"
	Display deferUpdates: false; forceDisplayUpdate.
! !

!WorldState methodsFor: 'update cycle' stamp: 'ar 6/17/2010 16:04'!
doDeferredUpdatingFor: aWorld
        "If this platform supports deferred updates, then make my canvas be the Display (or a rectangular portion of it), set the Display to deferred update mode, and answer true. Otherwise, do nothing and answer false. One can set the class variable DisableDeferredUpdates to true to completely disable the deferred updating feature."
	| properDisplay |
	PasteUpMorph disableDeferredUpdates ifTrue: [^ false].
	(Display deferUpdates: true) ifNil: [^ false].  "deferred updates not supported"
	remoteServer ifNotNil:[
		self assuredCanvas.
		^true].
	properDisplay := canvas notNil and: [canvas form == Display].
	aWorld == World ifTrue: [  "this world fills the entire Display"
		properDisplay ifFalse: [
			aWorld viewBox: Display boundingBox.    "do first since it may clear canvas"
			self canvas: (Display getCanvas copyClipRect: Display boundingBox).
		]
	].
	^ true
! !

!WorldState methodsFor: 'canvas' stamp: 'RAA 5/25/2000 15:12'!
doFullRepaint

	damageRecorder doFullRepaint
! !

!WorldState methodsFor: 'update cycle' stamp: 'ls 5/6/2003 16:51'!
doOneCycleFor: aWorld
	"Do one cycle of the interaction loop. This method is called repeatedly when the world is running.

This is a moderately private method; a better alternative is usually either to wait for events or to check the state of things from #step methods."

	self interCyclePause: (Preferences higherPerformance ifTrue: [1] ifFalse: [MinCycleLapse]).
	self doOneCycleNowFor: aWorld.! !

!WorldState methodsFor: 'update cycle' stamp: 'md 3/19/2006 22:16'!
doOneCycleNowFor: aWorld
	"Immediately do one cycle of the interaction loop.
	This should not be called directly, but only via doOneCycleFor:"

	| capturingGesture |
	DisplayScreen checkForNewScreenSize.
	capturingGesture := false.
	"self flag: #bob.	"	"need to consider remote hands in lower worlds"

	"process user input events"
	LastCycleTime := Time millisecondClockValue.
	self handsDo: [:h |
		ActiveHand := h.
		h processEvents.
		capturingGesture := capturingGesture or: [ h isCapturingGesturePoints ].
		ActiveHand := nil
	].

	"the default is the primary hand"
	ActiveHand := self hands first.

	"The gesture recognizer needs enough points to be accurate.
	Therefore morph stepping is disabled while capturing points for the recognizer"
	capturingGesture ifFalse: 
		[aWorld runStepMethods.		"there are currently some variations here"
		self displayWorldSafely: aWorld].
! !

!WorldState methodsFor: 'update cycle' stamp: 'ar 1/22/2001 14:26'!
doOneSubCycleFor: aWorld
	"Like doOneCycle, but preserves activeHand."

	| currentHand |
	currentHand := ActiveHand.
	self doOneCycleFor: aWorld.
	ActiveHand := currentHand! !

!WorldState methodsFor: 'update cycle' stamp: 'mt 4/16/2015 10:59'!
drawWorld: aWorld submorphs: submorphs invalidAreasOn: aCanvas 
	"Redraw the damaged areas of the given canvas and clear the damage list. Return a collection of the areas that
were redrawn."

	| rectList n morphs rects validList |
	rectList := damageRecorder invalidRectsFullBounds: aWorld viewBox.
	"sort by areas to draw largest portions first"
	rectList := rectList asArray sort: [:r1 :r2 | r1 area > r2 area].
	damageRecorder reset.
	n := submorphs size.
	morphs := OrderedCollection new: n * 2.
	rects := OrderedCollection new: n * 2.
	validList := OrderedCollection new: n * 2.

	"This is added in case we are drawing to a form that is to be used as a texture, and we want the background to be translucent."
	aWorld color isTranslucent ifTrue:
		[rectList do: [:r | aCanvas form fill: r fillColor: aWorld color]].

	rectList do: 
			[:dirtyRect | 
			dirtyRect allAreasOutsideList: validList
				do: 
					[:r | | mm rectToFill remnants c rect i | 
					"Experimental top-down drawing --
			Traverses top to bottom, stopping if the entire area is filled.
			If only a single rectangle remains, then continue with the reduced rectangle."

					rectToFill := r.
					i := 1.
					[rectToFill isNil or: [i > n]] whileFalse: 
							[mm := submorphs at: i.
							((mm fullBounds intersects: r) and: [mm visible]) 
								ifTrue: 
									[morphs addLast: mm.
									rects addLast: rectToFill.
									remnants := mm areasRemainingToFill: rectToFill.
									remnants size = 1 ifTrue: [rectToFill := remnants first].
									remnants isEmpty ifTrue: [rectToFill := nil]].
							i := i + 1].

					"Now paint from bottom to top, but using the reduced rectangles."
					rectToFill 
						ifNotNil: [(c := aCanvas copyClipRect: rectToFill) drawMorph: aWorld].
					[morphs isEmpty] whileFalse: 
							[(rect := rects removeLast) == rectToFill 
								ifFalse: [c := aCanvas copyClipRect: (rectToFill := rect)].
							c fullDrawMorph: morphs removeLast].
					morphs reset.
					rects reset.
					validList add: r]].
	^validList! !

!WorldState methodsFor: 'update cycle' stamp: 'ar 4/19/2001 05:45'!
forceDamageToScreen: allDamage

	"here for the convenience of NebraskaWorldState"
	Display forceDamageToScreen: allDamage.
	self remoteCanvasesDo: [ :each | 
		allDamage do: [:r | each forceToScreen: r].
		each displayIsFullyUpdated.
	].! !

!WorldState methodsFor: 'hands' stamp: 'di 6/7/1999 17:40'!
hands

	^ hands! !

!WorldState methodsFor: 'hands' stamp: 'RAA 5/24/2000 10:13'!
handsDo: aBlock

	^ hands do: aBlock! !

!WorldState methodsFor: 'hands' stamp: 'RAA 5/24/2000 12:09'!
handsReverseDo: aBlock

	^ hands reverseDo: aBlock! !

!WorldState methodsFor: 'initialization' stamp: 'RAA 1/5/2001 10:47'!
initialize

	hands := Array new.
	damageRecorder:= DamageRecorder new.
	stepList := Heap sortBlock: self stepListSortBlock.
	lastStepTime := 0.
	lastAlarmTime := 0.! !

!WorldState methodsFor: 'update cycle' stamp: 'mt 6/29/2016 09:32'!
interCyclePause: milliSecs
	"delay enough that the previous cycle plus the amount of delay will equal milliSecs.  If the cycle is already expensive, then no delay occurs.  However, if the system is idly waiting for interaction from the user, the method will delay for a proportionally long time and cause the overall CPU usage of Squeak to be low.
	If the preference #serverMode is enabled, always do a complete delay of 50ms, independant of my argument. This prevents the freezing problem described in Mantis #6581"

	| millisecondsToWait |
	millisecondsToWait := Preferences serverMode
		ifTrue: [ 50 ]
		ifFalse: [
			(lastCycleTime isNil or: [ CanSurrenderToOS == false ])
				ifTrue: [ 0 ]
				ifFalse: [ milliSecs - (Time millisecondsSince: lastCycleTime) ] ].
	(Preferences serverMode or: [millisecondsToWait > 0 and: [millisecondsToWait <= milliSecs]])
		ifTrue: [
			(interCycleDelay isNil or: [ interCycleDelay beingWaitedOn ])
				ifTrue: [ interCycleDelay := Delay forMilliseconds: millisecondsToWait ]
				ifFalse: [ interCycleDelay delayDuration: millisecondsToWait ].
			interCycleDelay wait ].
	lastCycleTime := Time millisecondClockValue.
	CanSurrenderToOS := true.! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:09'!
isStepping: aMorph
	"Return true if the given morph is in the step list."
	lastStepMessage ifNotNil:[(lastStepMessage receiver == aMorph) ifTrue:[^true]].
	stepList do:[:entry| entry receiver == aMorph ifTrue:[^true]].
	^ false! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:09'!
isStepping: aMorph selector: aSelector
	"Return true if the given morph is in the step list."
	lastStepMessage ifNotNil:[
		(lastStepMessage receiver == aMorph and:[lastStepMessage selector == aSelector])
			ifTrue:[^true]].
	stepList do:[:entry| (entry receiver == aMorph and:[entry selector == aSelector]) ifTrue:[^true]].
	^ false! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:05'!
listOfSteppingMorphs
	^stepList collect:[:entry| entry receiver].
! !

!WorldState methodsFor: 'alarms' stamp: 'jcg 1/9/2010 15:00'!
lockAlarmsDuring: actionBlock
	"All accesses to the alarms queue is synchronized by a mutex.  Answer the result of evaluating the 1-argument 'actionBlock'."
	alarms ifNil: [alarms := MorphicAlarmQueue new].
	^alarms mutex critical: [
		actionBlock value: alarms
	]! !

!WorldState methodsFor: 'canvas' stamp: 'ar 1/30/2001 23:25'!
recordDamagedRect: damageRect

	damageRecorder ifNotNil: [damageRecorder recordInvalidRect: damageRect truncated]
! !

!WorldState methodsFor: 'Nebraska support' stamp: 'ar 10/26/2000 14:44'!
releaseRemoteServer
	"My server has been transferred to some other world. Release pending references"
	remoteServer := nil.
	self canvas: nil.! !

!WorldState methodsFor: 'Nebraska support' stamp: 'ar 10/26/2000 14:15'!
remoteCanvasesDo: aBlock
	remoteServer ifNil:[^self].
	^remoteServer clients do:[:client| aBlock value: client canvas]! !

!WorldState methodsFor: 'Nebraska support' stamp: 'ar 10/26/2000 14:08'!
remoteServer
	^remoteServer! !

!WorldState methodsFor: 'Nebraska support' stamp: 'ar 10/26/2000 14:29'!
remoteServer: aNebraskaServer
	remoteServer ifNotNil:[remoteServer destroy].
	remoteServer := aNebraskaServer.
	self canvas: nil.! !

!WorldState methodsFor: 'alarms' stamp: 'ul 5/31/2016 01:13'!
removeAlarm: aSelector for: aTarget 
	"Remove the alarm with the given selector"
	self lockAlarmsDuring: [:locked |
		locked removeAlarmWithReceiver: aTarget selector: aSelector ]! !

!WorldState methodsFor: 'hands' stamp: 'ar 1/22/2001 14:26'!
removeHand: aHandMorph
	"Remove the given hand from the list of hands for this world."

	(hands includes: aHandMorph) ifFalse: [^self].
	hands := hands copyWithout: aHandMorph.
	ActiveHand == aHandMorph ifTrue: [ActiveHand := nil].
! !

!WorldState methodsFor: 'Nebraska support' stamp: 'ar 10/26/2000 14:17'!
removeRemoteCanvas: c
	self canvas: nil.	"force withdrawal of remote from MultiCanvas"
! !

!WorldState methodsFor: 'canvas' stamp: 'RAA 5/25/2000 15:10'!
resetDamageRecorder

	damageRecorder reset
! !

!WorldState methodsFor: 'stepping' stamp: 'dgd 2/22/2003 13:31'!
runLocalStepMethodsIn: aWorld 
	"Run morph 'step' methods (LOCAL TO THIS WORLD) whose time has come. Purge any morphs that are no longer in this world.
	ar 3/13/1999: Remove buggy morphs from the step list so that they don't raise repeated errors."

	| now morphToStep stepTime priorWorld |
	now := Time millisecondClockValue.
	priorWorld := ActiveWorld.
	ActiveWorld := aWorld.
	self triggerAlarmsBefore: now.
	stepList isEmpty 
		ifTrue: 
			[ActiveWorld := priorWorld.
			^self].
	(now < lastStepTime or: [now - lastStepTime > 5000]) 
		ifTrue: [self adjustWakeupTimes: now].	"clock slipped"
	[stepList isEmpty not and: [stepList first scheduledTime < now]] 
		whileTrue: 
			[lastStepMessage := stepList removeFirst.
			morphToStep := lastStepMessage receiver.
			(morphToStep shouldGetStepsFrom: aWorld) 
				ifTrue: 
					[lastStepMessage value: now.
					lastStepMessage ifNotNil: 
							[stepTime := lastStepMessage stepTime ifNil: [morphToStep stepTime].
							lastStepMessage scheduledTime: now + (stepTime max: 1).
							stepList add: lastStepMessage]].
			lastStepMessage := nil].
	lastStepTime := now.
	ActiveWorld := priorWorld! !

!WorldState methodsFor: 'stepping' stamp: 'ar 2/4/2010 22:37'!
runStepMethodsIn: aWorld
	"Perform periodic activity inbetween event cycles"
	| queue msg limit stamp |
	"Limit processing of deferredUIMessages to a max. amount of time"
	limit := self class deferredExecutionTimeLimit.
	stamp := Time millisecondClockValue.
	queue := self class deferredUIMessages.
	[(Time millisecondsSince: stamp) >= limit 
		or:[(msg := queue nextOrNil) == nil]] 
			whileFalse: [msg value].
	self runLocalStepMethodsIn: aWorld.
! !

!WorldState methodsFor: 'hands' stamp: 'nice 5/8/2011 12:18'!
selectHandsToDrawForDamage: damageList
	"Select the set of hands that must be redrawn because either (a) the hand itself has changed or (b) the hand intersects some damage rectangle."

	| result |
	result := OrderedCollection new.
	hands do: [:h |
		h needsToBeDrawn ifTrue: [
			h hasChanged
				ifTrue: [result add: h]
				ifFalse: [
					| hBnds |
					hBnds := h fullBounds.
					(damageList anySatisfy: [:r | r intersects: hBnds])
						ifTrue: [result add: h]]]].
	^ result
! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:36'!
startStepping: aMorph at: scheduledTime selector: aSelector arguments: args stepTime: stepTime
	"Add the given morph to the step list. Do nothing if it is already being stepped."

	self stopStepping: aMorph selector: aSelector.
	self adjustWakeupTimesIfNecessary.
	stepList add:(
		StepMessage 
			scheduledAt: scheduledTime
			stepTime: stepTime
			receiver: aMorph
			selector: aSelector
			arguments: args)! !

!WorldState methodsFor: 'initialization' stamp: 'sw 9/5/2000 06:39'!
stepListSize
	^ stepList size! !

!WorldState methodsFor: 'initialization' stamp: 'nice 12/27/2009 01:33'!
stepListSortBlock
	^[ :stepMsg1 :stepMsg2 | 
		stepMsg1 scheduledTime <= stepMsg2 scheduledTime.
	]! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:10'!
stopStepping: aMorph
	"Remove the given morph from the step list."
	lastStepMessage ifNotNil:[
		(lastStepMessage receiver == aMorph) ifTrue:[lastStepMessage := nil]].
	stepList removeAll: (stepList select:[:stepMsg| stepMsg receiver == aMorph]).
! !

!WorldState methodsFor: 'stepping' stamp: 'ar 10/22/2000 16:10'!
stopStepping: aMorph selector: aSelector
	"Remove the given morph from the step list."
	lastStepMessage ifNotNil:[
		(lastStepMessage receiver == aMorph and:[lastStepMessage selector == aSelector])
			ifTrue:[lastStepMessage := nil]].
	stepList removeAll: (stepList select:[:stepMsg| stepMsg receiver == aMorph and:[stepMsg selector == aSelector]]).! !

!WorldState methodsFor: 'alarms' stamp: 'ul 5/31/2016 01:12'!
triggerAlarmsBefore: nowTime
	"Trigger all pending alarms that are to be executed before nowTime."
	| triggered |
	lastAlarmTime ifNil:[lastAlarmTime := nowTime].
	(nowTime < lastAlarmTime or:[nowTime - lastAlarmTime > 10000])
		ifTrue:[self adjustAlarmTimes: nowTime].
	triggered := self lockAlarmsDuring: [:pending |
		pending removeAllAlarmsBefore: nowTime ].
	triggered do: [:alarm | alarm value: nowTime].
	lastAlarmTime := nowTime.! !

!WorldState methodsFor: 'canvas' stamp: 'di 6/7/1999 17:44'!
viewBox

	^ viewBox! !

!WorldState methodsFor: 'canvas' stamp: 'di 6/7/1999 17:58'!
viewBox: x

	viewBox := x! !

!Integer methodsFor: '*Morphic-Explorer' stamp: 'ul 11/11/2009 19:47'!
explorerContents

	^#(
		('hexadecimal' 16)
		('octal' 8)
		('binary' 2)) collect: [ :each |
			ObjectExplorerWrapper
				with: each first translated
				name: (self printStringBase: each second)
				model: self ]! !

!Integer methodsFor: '*Morphic-Explorer' stamp: 'laza 3/17/2005 13:38'!
hasContentsInExplorer
	^true! !

!DisplayScreen methodsFor: '*Morphic-blitter defaults' stamp: 'ar 5/28/2000 12:02'!
defaultCanvasClass
	"Return the WarpBlt version to use when I am active"
	^FormCanvas! !

!Form class methodsFor: '*Morphic-examples' stamp: 'tpr 11/13/2015 15:36'!
exampleColorSees
	"Form exampleColorSees"
	"First column as above shows the sneaky red/yellow pirate sneaking up on the blue/peach galleon.
	Second column shows the 1bpp made from the red/yellow/transparent - white -> ignore this, black -> test this
	Third shows the hit area - where red touches blue - superimposed on the original scene.
	Fourth column is the tally of hits via the old algorithm
	Last column shows the tally of hits via the new prim"	
		
	|formA formB maskA  offset tally map intersection left top dCanvas sensitiveColor soughtColor index|
	formA := formB := maskA := offset := tally := map := intersection :=  nil. "just to shut up the compiler when testing"
	ActiveWorld restoreMorphicDisplay; doOneCycle.

	sensitiveColor := Color red.
	soughtColor := Color blue.

	top := 50.
	dCanvas := FormCanvas on: Display.
	-50 to: 80 by: 10 do:[:p|
		offset:= p@0. "vary this to check different states"
		left := 10.

		formA := (Form extent: 100@50 depth: 32) asFormOfDepth: 16 "so we can try original forms of other depths".
		formB := Form extent: 100@50 depth: 32.

		"make a red square in the middle of the form"
		(FormCanvas on: formA) fillRectangle: (25@25 extent: 50@5) fillStyle: sensitiveColor.
		(FormCanvas on: formA) fillRectangle: (25@30 extent: 50@5) fillStyle: Color transparent.
		(FormCanvas on: formA) fillRectangle: (25@35 extent: 50@50) fillStyle: Color yellow.
		"formA displayOn: Display at: left@top rule: Form paint.
		dCanvas frameRectangle: (left@top extent: formA extent) width:2 color: Color green.
		left := left + 150."

		"make a blue block on the right half of the form"
		(FormCanvas on: formB) fillRectangle: (50@0 extent: 50@100) fillStyle: soughtColor.
		(FormCanvas on: formB) fillRectangle: (60@0 extent: 10@100) fillStyle: Color palePeach.
		"formB displayOn: Display at: left@top rule: Form paint.
		dCanvas frameRectangle: (left@top extent: formA extent) width:2 color: Color green.
		left := left + 150."

		intersection := (formA boundingBox translateBy: offset) intersect: (formB boundingBox).

		formB displayOn: Display at: left@top rule: Form paint.
		formA displayOn: Display at: (left@top) + offset rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.
		left := left + 150.
	
		maskA := Form extent: intersection extent depth: 1.

		map := Bitmap new: (1 bitShift: (formA depth min: 15)).
		map at: (index := sensitiveColor indexInMap: map) put: 1.

		maskA copyBits: (intersection translateBy:  offset negated) from: formA at: 0@0 colorMap: map.
		formB displayOn: Display at: left@top rule: Form paint.
		formA displayOn: Display at: (left@top) + offset rule: Form paint.
		maskA displayOn: Display at: (left@top) + intersection origin rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.	left := left + 150.

		"intersect world pixels of the color we're looking for with sensitive pixels mask"
		map at: index put: 0.  "clear map and reuse it"
		map at: (soughtColor indexInMap: map) put: 1.

		maskA
	 		copyBits: intersection
			from: formB at: 0@0 clippingBox: formB boundingBox
			rule: Form and
			fillColor: nil
			map: map.

		formB displayOn: Display at: left@top rule: Form paint.
		formA displayOn: Display at: (left@top) + offset rule: Form paint.
		maskA displayOn: Display at: (left@top) + intersection origin rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.
		left := left + 170.
		
		(maskA tallyPixelValues at: 2) asString asDisplayText displayOn: Display at: left@(top +20).
		left := left + 70.
		
		"now try using the new primitive"
		tally := (BitBlt
			destForm: formB
			sourceForm: formA
			fillColor: nil
			combinationRule: 3 "really ought to work with nil but prim code checks"
			destOrigin: intersection origin
			sourceOrigin: (offset negated max: 0@0)
			extent: intersection extent 
			clipRect: intersection)
				primCompareColor: ((sensitiveColor pixelValueForDepth: formA depth) ) to: ((soughtColor pixelValueForDepth: formB depth) ) test: (Form compareMatchColor bitOr: Form compareTallyFlag).
		tally  asString asDisplayText displayOn: Display at: left@(top +20).
		top:= top + 60]

! !

!Form class methodsFor: '*Morphic-examples' stamp: 'tpr 11/13/2015 15:37'!
exampleTouchingColor
	"Form exampleTouchingColor"
	"Demonstrate the algorithm used in Scratch code to determine if a sprite's non-transparent pixels touch a
	particular color pixel of the background upon which it is displayed.
	First column as above shows the sneaky red/yellow pirate sneaking up on the blue/peach galleon.
	Second column shows the 1bpp made from the red/yellow/transparent - white -> ignore this, black -> test this
	Third shows the hit area (black) superimposed on the original scene
	Fourth column is the tally of hits via the old algorithm
	Last column shows the tally of hits via the new prim"	
	|formA formB maskA  offset tally map intersection left top dCanvas ignoreColor soughtColor|
	formA := formB := maskA := offset := tally := map := intersection :=  nil. "just to shut up the compiler when testing"
	ActiveWorld restoreMorphicDisplay; doOneCycle.

	ignoreColor := Color transparent.
	soughtColor := Color blue.

	top := 50.
	dCanvas := FormCanvas on: Display.
	-50 to: 80 by: 10 do:[:p|
		offset:= p@0. "vary this to check different states"
		left := 10.

		formA := (Form extent: 100@50 depth: 32) asFormOfDepth: 16 "so we can try original forms of other depths".
		formB := Form extent: 100@50 depth: 32.

		"make a red square in the middle of the form"
		(FormCanvas on: formA) fillRectangle: (25@25 extent: 50@5) fillStyle: Color red.
		(FormCanvas on: formA) fillRectangle: (25@30 extent: 50@5) fillStyle: Color transparent.
		(FormCanvas on: formA) fillRectangle: (25@35 extent: 50@50) fillStyle: Color yellow.
		"formA displayOn: Display at: left@top rule: Form paint.
		dCanvas frameRectangle: (left@top extent: formA extent) width:2 color: Color green.
		left := left + 150."

		"make a blue block on the right half of the form"
		(FormCanvas on: formB) fillRectangle: (50@0 extent: 50@100) fillStyle: soughtColor.
		(FormCanvas on: formB) fillRectangle: (60@0 extent: 10@100) fillStyle: Color palePeach.
		"formB displayOn: Display at: left@top rule: Form paint.
		dCanvas frameRectangle: (left@top extent: formA extent) width:2 color: Color green.
		left := left + 150."

		intersection := (formA boundingBox translateBy: offset) intersect: (formB boundingBox).

		formB displayOn: Display at: left@top rule: Form paint.
		formA displayOn: Display at: (left@top) + offset rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.
		left := left + 150.
	
		maskA := Form extent: intersection extent depth: 1.

		map := Bitmap new: (1 bitShift: (formA depth min: 15)).
		map atAllPut: 1.
		map at: ( ignoreColor indexInMap: map) put: 0.

		maskA copyBits: (intersection translateBy:  offset negated) from: formA at: 0@0 colorMap: map.
		formB displayOn: Display at: left@top rule: Form paint.
		formA displayOn: Display at: (left@top) + offset rule: Form paint.
		maskA displayOn: Display at: (left@top) + intersection origin rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.	left := left + 150.

		"intersect world pixels of the color we're looking for with sensitive pixels mask"
		map atAllPut: 0.  "clear map and reuse it"
		map at: (soughtColor indexInMap: map) put: 1.

		maskA
	 		copyBits: intersection
			from: formB at: 0@0 clippingBox: formB boundingBox
			rule: Form and
			fillColor: nil
			map: map.

		formB displayOn: Display at: left@top rule: Form paint.
		formA displayOn: Display at: (left@top) + offset rule: Form paint.
		maskA displayOn: Display at: (left@top) + intersection origin rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.
		left := left + 170.
		
		(maskA tallyPixelValues at: 2) asString asDisplayText displayOn: Display at: left@(top +20).
		left := left + 70.
		
		"now try using the new primitive"
		tally := (BitBlt
			destForm: formB
			sourceForm: formA
			fillColor: nil
			combinationRule: 3 "really ought to work with nil but prim code checks"
			destOrigin: intersection origin
			sourceOrigin: (offset negated max: 0@0)
			extent: intersection extent 
			clipRect: intersection)
				primCompareColor: ((ignoreColor pixelValueForDepth: formA depth) bitAnd: 16rFFFFFF) to: ((soughtColor pixelValueForDepth: formB depth) bitAnd: 16rFFFFFF) test: (Form compareNotColorAMatchColorB bitOr: Form compareTallyFlag).
		tally  asString asDisplayText displayOn: Display at: left@(top +20).
		top:= top + 60]
! !

!Form class methodsFor: '*Morphic-examples' stamp: 'tpr 11/13/2015 15:36'!
exampleTouchTest
	"Form exampleTouchTest"
	"Demonstrate the algorithm used in Scratch code to determine if a sprite's non-transparent pixels touch a 
	non-transparent pixel of the background upon which it is displayed.
	First column shows a form with a red block in the midst of transparent area sneaking up on a form with a transparent LHS and blue RHS. 	The green frame shows the intersection area.
	Second column shows in grey the part of the red that is within the intersection.
	Third column shows in black the blue that is within the intersection.
	Fourth column shows just the A touching B area.
	Fifth column is the tally of hits via the old algorithm
	Last column shows the tally of hits via the new prim"
	|formA formB maskA maskB offset tally map intersection left top dCanvas|
	formA := formB := maskA := maskB := offset := tally := map := intersection :=  nil. "just to shut up the compiler when testing"

	ActiveWorld restoreMorphicDisplay; doOneCycle.

	top := 50.
	dCanvas := FormCanvas on: Display.
	-50 to: 80 by: 10 do:[:p|
		offset:= p@0. "vary this to check different states"
		left := 10.

		formA := Form extent: 100@50 depth: 32.
		formB := Form extent: 100@50 depth: 16.

		"make a red square in the middle of the form"
		(FormCanvas on: formA) fillRectangle: (25@25 extent: 50@5) fillStyle: Color yellow.
		(FormCanvas on: formA) fillRectangle: (25@30 extent: 50@5) fillStyle: Color transparent.
		(FormCanvas on: formA) fillRectangle: (25@35 extent: 50@50) fillStyle: Color red.
		"formA displayOn: Display at: left@top rule: Form paint.
		dCanvas frameRectangle: (left@top extent: formA extent) width:2 color: Color green.
		left := left + 150."

		"make a blue block on the right half of the form"
		(FormCanvas on: formB) fillRectangle: (50@0 extent: 50@100) fillStyle: Color blue.
		(FormCanvas on: formB) fillRectangle: (60@0 extent: 10@100) fillStyle: Color palePeach.
		"formB displayOn: Display at: left@top rule: Form paint.
		dCanvas frameRectangle: (left@top extent: formA extent) width:2 color: Color green.
		left := left + 150."

		intersection := (formA boundingBox translateBy: offset) intersect: (formB boundingBox).

		formB displayOn: Display at: left@top rule: Form paint.
		formA displayOn: Display at: (left@top) + offset rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.
		left := left + 150.

		maskA := Form extent: intersection extent depth: 2.
		formA displayOn: maskA at: offset  - intersection origin rule: Form paint.
		formB displayOn: Display at: left@top rule: Form paint.
		formA displayOn: Display at: (left@top) + offset rule: Form paint.
		maskA displayOn: Display at: (left@top) + intersection origin rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.
		left := left + 150.

		maskB := Form extent: intersection extent depth: 2.
		formB displayOn: maskB at: intersection origin negated rule: Form paint.
		formB displayOn: Display at: left@top rule: Form paint.
		formA displayOn: Display at: (left@top) + offset rule: Form paint.
		maskB displayOn: Display at: (left@top) + intersection origin rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.
		left := left + 150.

		map := Bitmap new: 4 withAll: 1.
		map at: 1 put: 0.  "transparent"

		maskA copyBits: maskA boundingBox from: maskA at: 0@0 colorMap: map.
		"maskA displayOn: Display at: (left@top) + intersection origin rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.
		left := left + 150."

		maskB copyBits: maskB boundingBox from: maskB at: 0@0 colorMap: map.
		"maskB displayOn: Display at: (left@top) + intersection origin rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.
		left := left + 150."

		maskB displayOn: maskA at: 0@0 rule: Form and.
		maskA displayOn: Display at: (left@top) + intersection origin rule: Form paint.
		dCanvas frameRectangle: (intersection translateBy: left@top) width:2 color: Color green.
		left := left + 170.
		
		(maskA boundingBox area -( maskA tallyPixelValues at: 1)) asString asDisplayText displayOn: Display at: left@(top +20).
		left := left + 70.
		
		"now try using the new primitive"
		tally := (BitBlt
			destForm: formB
			sourceForm: formA
			fillColor: nil
			combinationRule: 3 "really ought to work with nil but prim code checks"
			destOrigin: intersection origin
			sourceOrigin: (offset negated max: 0@0)
			extent: intersection extent 
			clipRect: intersection)
				primCompareColor: ((Color transparent pixelValueForDepth: formA depth) bitAnd: 16rFFFFFF) to: ((Color transparent pixelValueForDepth: formB depth) bitAnd: 16rFFFFFF) test: (Form compareNotColorANotColorB bitOr: Form compareTallyFlag).
		tally  asString asDisplayText displayOn: Display at: left@(top +20).
		top:= top + 60]


! !

!Form methodsFor: '*Morphic' stamp: 'ar 11/7/1999 20:29'!
asMorph
	^ImageMorph new image: self! !

!Form methodsFor: '*Morphic' stamp: 'ar 7/8/2006 21:01'!
iconOrThumbnailOfSize: aNumberOrPoint 
	"Answer an appropiate form to represent the receiver"
	^ self scaledIntoFormOfSize: aNumberOrPoint! !

!Form methodsFor: '*Morphic' stamp: 'mt 8/25/2015 10:43'!
scaledIntoFormOfSize: aNumberOrPoint 
	
	^ self scaledIntoFormOfSize: aNumberOrPoint smoothing: 8! !

!Form methodsFor: '*Morphic' stamp: 'mt 12/10/2015 09:43'!
scaledIntoFormOfSize: aNumberOrPoint smoothing: factor
	"Scale and center the receiver into a form of a given size"

	| extent scaledForm result |

	extent := aNumberOrPoint asPoint.
	extent = self extent ifTrue: [^ self copy].

	scaledForm := self scaledToSize: extent smoothing: factor.

	result := self species extent: extent depth: self depth.
	result getCanvas
		translucentImage: scaledForm
		at: extent - scaledForm extent // 2.

	^ result
! !

!Form methodsFor: '*Morphic-Support-image manipulation' stamp: 'LB 8/26/2002 18:08'!
stencil
	"return a 1-bit deep, black-and-white stencil of myself"

	| canvas |
	canvas := FormCanvas extent: self extent depth: 1.
	canvas fillColor: (Color white).

	canvas stencil: self at: 0@0  
				sourceRect: (Rectangle origin: 0@0 corner: self extent) color: Color black.

	^ canvas form
! !

!Debugger class methodsFor: '*Morphic-instance creation' stamp: 'mt 2/15/2016 15:27'!
morphicContext: aContext
	"Answer an instance of me for debugging the active process starting with the given context."

	^ self new
		process: Processor activeProcess
		controller: nil
		context: aContext! !

!Debugger class methodsFor: '*Morphic-opening' stamp: 'mt 2/15/2016 15:29'!
morphicOpenContext: aContext label: aString contents: contentsStringOrNil
	"Open a notifier in response to an error, halt, or notify. A notifier view just shows a short view of the sender stack and provides a menu that lets the user open a full debugger."
	<primitive: 19> "Simulation guard"
	ErrorRecursion not & Preferences logDebuggerStackToFile ifTrue:
		[Smalltalk logSqueakError: aString inContext: aContext].
	ErrorRecursion ifTrue:[
		ErrorRecursion := false.
		self primitiveError: aString].
	ErrorRecursion := true.
	self informExistingDebugger: aContext label: aString.
	(Debugger morphicContext: aContext)
		openNotifierContents: contentsStringOrNil
		label: aString.
	ErrorRecursion := false.
	Processor activeProcess suspend.! !

!Debugger class methodsFor: '*Morphic-opening' stamp: 'mt 2/15/2016 15:25'!
morphicOpenInterrupt: aString onProcess: interruptedProcess
	"Open a notifier in response to an interrupt. An interrupt occurs when the user types the interrupt key (cmd-. on Macs, ctrl-c or alt-. on other systems) or when the low-space watcher detects that memory is low."
	| debugger |
	<primitive: 19> "Simulation guard"
	debugger := self new.
	debugger
		process: interruptedProcess
		controller: nil
		context: interruptedProcess suspendedContext.
	debugger externalInterrupt: true.

	Preferences logDebuggerStackToFile ifTrue:
		[(aString includesSubstring: 'Space') & (aString includesSubstring: 'low')
			ifTrue: [Smalltalk logError: aString inContext: debugger interruptedContext to: 'LowSpaceDebug.log']
			"logging disabled for 4.3 release, see
				http://lists.squeak.org/pipermail/squeak-dev/2011-December/162503.html"
			"ifFalse: [Smalltalk logSqueakError: aString inContext: debugger interruptedContext]"].

	Preferences eToyFriendly ifTrue: [World stopRunningAll].
	^debugger
		openNotifierContents: nil label: aString;
		yourself
! !

!Debugger class methodsFor: '*Morphic-opening' stamp: 'mt 4/17/2016 17:31'!
morphicOpenOn: process context: context label: title contents: contentsStringOrNil fullView: full
	"Open a notifier in response to an error, halt, or notify. A notifier view just shows a short view of the sender stack and provides a menu that lets the user open a full debugger."

	| errorWasInUIProcess debugger |
	ErrorRecursion ifTrue: [
		"self assert: process == Project current uiProcess -- DOCUMENTATION ONLY"
		ErrorRecursion := false.
		^ Project current handleFatalDrawingError: title].

	[ErrorRecursion not & Preferences logDebuggerStackToFile
		ifTrue: [Smalltalk logSqueakError: title inContext: context]] on: Error do: [:ex | ex return: nil].
	
	ErrorRecursion := true.

	errorWasInUIProcess := Project current spawnNewProcessIfThisIsUI: process.

	"Schedule debugging in deferred UI message because
		1) If process is the current UI process, it is already broken.
		2) If process is some other process, it must not execute UI code"
	Project current addDeferredUIMessage: [ 		
		debugger := self new process: process controller: nil context: context.
		full
			ifTrue: [debugger openFullNoSuspendLabel: title]
			ifFalse: [debugger openNotifierContents: contentsStringOrNil label: title].
		debugger errorWasInUIProcess: errorWasInUIProcess.
		
		"Try drawing the debugger tool at least once to avoid freeze."
		ActiveWorld displayWorldSafely.
		
		ErrorRecursion := false.
	].
	process suspend.
! !

!Debugger methodsFor: '*Morphic-opening' stamp: 'mt 3/30/2015 15:53'!
morphicResumeProcess: aTopView 

	| processToResume |
	processToResume := interruptedProcess.
	interruptedProcess := nil. "Before delete, so release doesn't terminate it"
	aTopView delete.
	World displayWorld. "We have to redraw *before* resuming the old process."
	Smalltalk installLowSpaceWatcher. "restart low space handler"

	savedCursor
		ifNotNil: [Cursor currentCursor: savedCursor].
	processToResume isTerminated ifFalse: [
		errorWasInUIProcess
					ifTrue: [Project resumeProcess: processToResume]
					ifFalse: [processToResume resume]].
	"if old process was terminated, just terminate current one"
	errorWasInUIProcess == false
		ifFalse: [Processor terminateActive]! !

!BitBlt class methodsFor: '*Morphic' stamp: 'ar 5/28/2000 12:04'!
asGrafPort
	"Return the GrafPort associated with the receiver"
	^GrafPort! !

BitBlt subclass: #GrafPort
	instanceVariableNames: 'alpha fillPattern lastFont lastFontForegroundColor lastFontBackgroundColor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Support'!

!GrafPort methodsFor: 'accessing' stamp: 'ar 2/17/2000 01:07'!
alphaBits: a
	alpha := a! !

!GrafPort methodsFor: 'copying' stamp: 'ar 12/30/2001 20:32'!
clippedBy: aRectangle
	^ self copy clipBy: aRectangle! !

!GrafPort methodsFor: 'accessing' stamp: 'ar 5/28/2000 14:41'!
contentsOfArea: aRectangle into: aForm
	destForm 
		displayOn: aForm 
		at:  aRectangle origin
		clippingBox: (0@0 extent: aRectangle extent).
	^aForm! !

!GrafPort methodsFor: 'copying' stamp: 'jmv 3/2/2010 16:02'!
copyBits
	"Override copybits to do translucency if desired"

	(combinationRule >= 30 and: [combinationRule <= 31]) 
		ifTrue: [
			self copyBitsTranslucent: (alpha ifNil: [255])]
		ifFalse: [super copyBits]! !

!GrafPort methodsFor: 'accessing' stamp: 'nice 9/29/2013 15:33'!
displayScannerFor: para foreground: foreColor background: backColor ignoreColorChanges: shadowMode
	^ (DisplayScanner new text: para text textStyle: para textStyle
			foreground: foreColor background: backColor fillBlt: self
			ignoreColorChanges: shadowMode)
		setPort: self clone
! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/16/2000 22:32'!
fillOval: rect
	| centerX centerY nextY yBias xBias outer nextOuterX |
	rect area <= 0 ifTrue: [^ self].
	height := 1.
	yBias := rect height odd ifTrue: [0] ifFalse: [-1].
	xBias := rect width odd ifTrue: [1] ifFalse: [0].
	centerX := rect center x.
	centerY := rect center y.
	outer := EllipseMidpointTracer new on: rect.
	nextY := rect height // 2.
	[nextY > 0] whileTrue:[
		nextOuterX := outer stepInY.
		width := (nextOuterX bitShift: 1) + xBias.
		destX := centerX - nextOuterX.
		destY := centerY - nextY.
		self copyBits.
		destY := centerY + nextY + yBias.
		self copyBits.
		nextY := nextY - 1.
	].
	destY := centerY.
	height := 1 + yBias.
	width := rect width.
	destX := rect left.
	self copyBits.
! !

!GrafPort methodsFor: 'accessing' stamp: 'ar 2/17/2000 01:09'!
fillPattern: anObject
	fillPattern := anObject.
	self fillColor: anObject.! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 5/17/2000 21:20'!
fillRect: rect offset: aPoint
	"The offset is really just for stupid InfiniteForms."
	| fc |
	fillPattern class == InfiniteForm ifTrue:[
		fc := halftoneForm.
		self fillColor: nil.
		fillPattern displayOnPort: ((self clippedBy: rect) colorMap: nil) at: aPoint.
		halftoneForm := fc.
		^self].

	destX := rect left.
	destY := rect top.
	sourceX := 0.
	sourceY := 0.
	width := rect width.
	height := rect height.
	self copyBits.! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 7/22/2009 21:34'!
fillRoundRect: aRectangle radius: radius
	| nextY outer nextOuterX ovalDiameter rectExtent rectOffset rectX rectY rectWidth rectHeight ovalRadius ovalRect |
	aRectangle area <= 0 ifTrue: [^ self].
	ovalDiameter := (radius * 2) asPoint min: aRectangle extent.
	(ovalDiameter x <= 0 or:[ovalDiameter y <= 0]) ifTrue:[
		^self fillRect: aRectangle offset: 0@0.
	].
	"force diameter to be even - this simplifies lots of stuff"
	ovalRadius := (ovalDiameter x // 2) @ (ovalDiameter y // 2).
	(ovalRadius x <= 0 or:[ovalRadius y <= 0]) ifTrue:[
		^self fillRect: aRectangle offset: 0@0.
	].
	ovalDiameter := ovalRadius * 2.
	rectExtent := aRectangle extent - ovalDiameter.
	rectWidth := rectExtent x.
	rectHeight := rectExtent y.
	rectOffset := aRectangle origin + ovalRadius.
	rectX := rectOffset x.
	rectY := rectOffset y.

	ovalRect := ovalRadius negated extent: ovalDiameter.

	height := 1.
	outer := EllipseMidpointTracer new on: ovalRect.
	nextY := ovalRadius y.
	"upper and lower portions of round rect"
	[nextY > 0] whileTrue:[
		nextOuterX := outer stepInY.
		width := nextOuterX * 2 + rectWidth.
		destX := rectX - nextOuterX.
		destY := rectY - nextY.
		self copyBits.
		destY := rectY + nextY + rectHeight - 1.
		self copyBits.
		nextY := nextY - 1.
	].
	destX := aRectangle left.
	destY := rectOffset y.
	height := rectHeight.
	width := aRectangle width.
	self copyBits.
! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/16/2000 22:26'!
frameOval: rect borderWidth: borderWidth
	| centerX centerY nextY yBias xBias wp outer inner nextOuterX nextInnerX fillAlpha |
	rect area <= 0 ifTrue: [^ self].
	height := 1.
	wp := borderWidth asPoint.
	yBias := rect height odd ifTrue: [0] ifFalse: [-1].
	xBias := rect width odd ifTrue: [1] ifFalse: [0].
	centerX := rect center x.
	centerY := rect center y.
	outer := EllipseMidpointTracer new on: rect.
	inner := EllipseMidpointTracer new on: (rect insetBy: wp).
	nextY := rect height // 2.
	1 to: (wp y min: nextY) do:[:i|
		nextOuterX := outer stepInY.
		width := (nextOuterX bitShift: 1) + xBias.
		destX := centerX - nextOuterX.
		destY := centerY - nextY.
		self copyBits.
		destY := centerY + nextY + yBias.
		self copyBits.
		nextY := nextY - 1.
	].
	[nextY > 0] whileTrue:[
		nextOuterX := outer stepInY.
		nextInnerX := inner stepInY.
		destX := centerX - nextOuterX.
		destY := centerY - nextY.
		width := nextOuterX - nextInnerX.
		self copyBits.
		destX := centerX + nextInnerX + xBias.
		self copyBits.
		destX := centerX - nextOuterX.
		destY := centerY + nextY + yBias.
		self copyBits.
		destX := centerX + nextInnerX + xBias.
		self copyBits.
		nextY := nextY - 1.
	].
	destY := centerY.
	height := 1 + yBias.
	width := wp x.
	destX := rect left.
	self copyBits.
	destX := rect right - wp x.
	self copyBits.
! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/17/2000 01:08'!
frameRectBottom: rect height: h

	destX := rect left + 1.
	destY := rect bottom - 1.
	width := rect width - 2.
	height := 1.
	1 to: h do: [:i |
		self copyBits.
		destX := destX + 1.
		destY := destY - 1.
		width := width - 2].
! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/17/2000 01:08'!
frameRectRight: rect width: w

	width := 1.
	height := rect height - 1.
	destX := rect right - 1.
	destY := rect top + 1.
	1 to: w do: [:i |
		self copyBits.
		destX := destX - 1.
		destY := destY + 1.
		height := height - 2].
! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/17/2000 14:44'!
frameRect: rect borderWidth: borderWidth
	sourceX := 0.
	sourceY := 0.
	(rect areasOutside: (rect insetBy: borderWidth)) do:
		[:edgeStrip | self destRect: edgeStrip; copyBits].
! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 7/22/2009 21:33'!
frameRoundRect: aRectangle radius: radius borderWidth: borderWidth
	| nextY outer nextOuterX ovalDiameter rectExtent rectOffset rectX rectY rectWidth rectHeight ovalRadius ovalRect innerRadius innerDiameter innerRect inner nextInnerX wp |
	aRectangle area <= 0 ifTrue: [^ self].
	ovalDiameter := (radius * 2) asPoint min: aRectangle extent.
	(ovalDiameter x <= 0 or:[ovalDiameter y <= 0]) ifTrue:[
		^self fillRect: aRectangle offset: 0@0.
	].
	"force diameter to be even - this simplifies lots of stuff"
	ovalRadius := (ovalDiameter x // 2) @ (ovalDiameter y // 2).
	(ovalRadius x <= 0 or:[ovalRadius y <= 0]) ifTrue:[
		^self fillRect: aRectangle offset: 0@0.
	].
	wp := borderWidth asPoint.
	ovalDiameter := ovalRadius * 2.
	innerRadius := ovalRadius - borderWidth max: 0@0.
	innerDiameter := innerRadius * 2.

	rectExtent := aRectangle extent - ovalDiameter.
	rectWidth := rectExtent x.
	rectHeight := rectExtent y.

	rectOffset := aRectangle origin + ovalRadius.
	rectX := rectOffset x.
	rectY := rectOffset y.

	ovalRect := 0@0 extent: ovalDiameter.
	innerRect := 0@0 extent: innerDiameter.

	height := 1.
	outer := EllipseMidpointTracer new on: ovalRect.
	inner := EllipseMidpointTracer new on: innerRect.

	nextY := ovalRadius y.

	1 to: (wp y min: nextY) do:[:i|
		nextOuterX := outer stepInY.
		width := nextOuterX * 2 + rectWidth.
		destX := rectX - nextOuterX.
		destY := rectY - nextY.
		self copyBits.
		destY := rectY + nextY + rectHeight - 1.
		self copyBits.
		nextY := nextY - 1.
	].
	[nextY > 0] whileTrue:[
		nextOuterX := outer stepInY.
		nextInnerX := inner stepInY.
		destX := rectX - nextOuterX.
		destY := rectY - nextY.
		width := nextOuterX - nextInnerX.
		self copyBits.
		destX := rectX + nextInnerX + rectWidth.
		self copyBits.
		destX := rectX - nextOuterX.
		destY := rectY + nextY + rectHeight-1.
		self copyBits.
		destX := rectX + nextInnerX + rectWidth.
		self copyBits.
		nextY := nextY - 1.
	].

	destX := aRectangle left.
	destY := rectOffset y.
	height := rectHeight.
	width := wp x.
	self copyBits.
	destX := aRectangle right - width.
	self copyBits.
	innerRadius y = 0 ifTrue:[
		destX := aRectangle left + wp x.
		destY := rectY.
		width := rectWidth.
		height := wp y - ovalRadius y.
		self copyBits.
		destY := aRectangle bottom - wp y.
		self copyBits.
	].! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/17/2000 01:08'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule
	"Draw the portion of the given Form defined by sourceRect at the given point using the given BitBlt combination rule."

	sourceForm := aForm.
	combinationRule := rule.
	self sourceRect: sourceRect.
	self destOrigin: aPoint.
	self copyBits! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 8/8/2001 14:26'!
image: aForm at: aPoint sourceRect: sourceRect rule: rule alpha: sourceAlpha
	"Draw the portion of the given Form defined by sourceRect at the given point using the given BitBlt combination rule."

	sourceForm := aForm.
	combinationRule := rule.
	self sourceRect: sourceRect.
	self destOrigin: aPoint.
	self copyBitsTranslucent: (alpha := (sourceAlpha * 255) truncated min: 255 max: 0).! !

!GrafPort methodsFor: 'private' stamp: 'yo 1/8/2005 09:12'!
installStrikeFont: aStrikeFont

	^ self installStrikeFont: aStrikeFont foregroundColor: (lastFontForegroundColor ifNil: [Color black]) backgroundColor: (lastFontBackgroundColor ifNil: [Color transparent]).
! !

!GrafPort methodsFor: 'private' stamp: 'tween 6/8/2008 12:29'!
installStrikeFont: aStrikeFont foregroundColor: foregroundColor backgroundColor: backgroundColor
	super installStrikeFont: aStrikeFont foregroundColor: foregroundColor backgroundColor: backgroundColor.
	aStrikeFont glyphs depth = 1 ifTrue: [
		alpha := foregroundColor privateAlpha.
		"dynamically switch between blend modes to support translucent text"
		"To handle the transition from TTCFont to StrikeFont, rule 34 must be taken into account."
		alpha = 255 ifTrue:[
			combinationRule = 30 ifTrue: [combinationRule := Form over].
			combinationRule = 31 ifTrue: [combinationRule := Form paint].
			combinationRule = 34 ifTrue: [combinationRule := Form paint].
			combinationRule = 41 ifTrue: [combinationRule := Form paint]. "41 is  SPRmode"
		] ifFalse:[
			combinationRule = Form over ifTrue: [combinationRule := 30].
			combinationRule = Form paint ifTrue: [combinationRule := 31].
			combinationRule = 34 ifTrue: [combinationRule := 31].
			combinationRule = 41 ifTrue: [combinationRule := 31]. "41 is SPR mode"
		]
	].
	lastFont := aStrikeFont.
	lastFontForegroundColor := foregroundColor.
	lastFontBackgroundColor := backgroundColor.
! !

!GrafPort methodsFor: 'private' stamp: 'yo 1/12/2005 16:39'!
installTTCFont: aTTCFont

	^ self installTTCFont: aTTCFont foregroundColor: (lastFontForegroundColor ifNil: [Color black]) backgroundColor: (lastFontBackgroundColor ifNil: [Color transparent]).
! !

!GrafPort methodsFor: 'private' stamp: 'yo 1/8/2005 09:12'!
installTTCFont: aTTCFont foregroundColor: foregroundColor backgroundColor: backgroundColor

	super installTTCFont: aTTCFont foregroundColor: foregroundColor backgroundColor: backgroundColor.
	lastFont := aTTCFont.
	lastFontForegroundColor := foregroundColor.
	lastFontBackgroundColor := backgroundColor.
! !

!GrafPort methodsFor: 'private' stamp: 'yo 1/8/2005 09:13'!
lastFont

	^ lastFont.
! !

!GrafPort methodsFor: 'accessing' stamp: 'tween 4/5/2007 08:03'!
lastFontForegroundColor
	^lastFontForegroundColor! !

!GrafPort methodsFor: 'drawing support' stamp: 'ar 2/17/2000 00:31'!
stencil: stencilForm at: aPoint sourceRect: aRect
	"Paint using aColor wherever stencilForm has non-zero pixels"
	self sourceForm: stencilForm;
		destOrigin: aPoint;
		sourceRect: aRect.
	self copyBits! !

!Text methodsFor: '*Morphic-converting' stamp: 'nk 2/26/2004 13:32'!
asMorph
	^ self asTextMorph! !

!Text methodsFor: '*Morphic-converting' stamp: 'nk 2/26/2004 13:31'!
asStringMorph
	^ StringMorph
		contents: self string
		font: (self fontAt: 1 withStyle: TextStyle default)
		emphasis: (self emphasisAt: 1)! !

!Text methodsFor: '*Morphic-converting' stamp: 'nk 2/26/2004 13:32'!
asTextMorph
	^ TextMorph new contentsAsIs: self! !

!Set methodsFor: '*Morphic-Explorer' stamp: 'ul 11/11/2009 19:52'!
explorerContents 

	^Array new: self size streamContents: [ :stream |
		self do: [ :each |
			stream nextPut: (
				ObjectExplorerWrapper
					with: each
					name: (stream position + 1) printString
					model: self) ] ]! !

!Set methodsFor: '*Morphic-Explorer' stamp: 'yo 8/27/2008 23:09'!
hasContentsInExplorer

	^self notEmpty! !

Project subclass: #MorphicProject
	instanceVariableNames: 'uiProcess'
	classVariableNames: 'DefaultFill'
	poolDictionaries: ''
	category: 'Morphic-Support'!

!MorphicProject commentStamp: 'dtl 7/13/2013 15:40' prior: 0!
An MorphicProject is a project with a Morphic user interface. Its world is a PasteUpMorph, and its UI manager is a MorphicUIManager. It uses a MorphicToolBuilder to create the views for various tools. It has a single uiProcess for its world, and Morph updates and user interaction occur in the context of that UI process.
!

!MorphicProject class methodsFor: 'utilities' stamp: 'nice 7/20/2014 22:19'!
compileNewDefaultBackgroundFrom: aForm 
	"Compile a new #defaultBackgroundForm method in this class-side which uses aForm as the background."
	| formStream base64 |
	formStream := ByteArray new writeStream.
	PNGReadWriter
		putForm: aForm
		onStream: formStream.
	base64 := formStream base64Encoded.
	self class
		compile:
			(String streamContents:
				[ : stream | stream
					 nextPutAll: 'defaultBackgroundForm
	^ Form fromBinaryStream: ' ;
					 nextPut: $' ;
					 nextPutAll: base64 ;
					 nextPut: $' ;
					nextPutAll: ' base64Decoded asByteArray readStream' ])
		classified: 'class initialization'! !

!MorphicProject class methodsFor: 'class initialization' stamp: 'ar 9/18/2009 20:43'!
convertOldProjects
	"Convert older Morphic projects to be MorphicProjects"
	Project allInstancesDo:[:p|
		p isMorphic ifTrue:[p primitiveChangeClassTo: self basicNew].
	].! !

!MorphicProject class methodsFor: 'utilities' stamp: 'ar 9/18/2009 20:33'!
defaultFill
	"Answer the default fill to use for new Morphic Worlds"
	^DefaultFill ifNil:[Color white].
! !

!MorphicProject class methodsFor: 'utilities' stamp: 'ar 9/18/2009 20:34'!
defaultFill: aFill
	"Answer the default fill to use for new Morphic Worlds"
	DefaultFill := aFill.! !

!MorphicProject class methodsFor: 'class initialization' stamp: 'mt 4/21/2015 18:26'!
initialize	"MorphicProject initialize"
	"Initialize the default Morphic Project background"
	self defaultFill: (SolidFillStyle color: Color white).
	CurrentProject ifNil:
		["This is the Top Project."
		CurrentProject := super new initialProject.
		CurrentProject spawnNewProcessAndTerminateOld: true].
	self convertOldProjects.
	Project current isMorphic ifTrue:[
		"Set the default background in the current world"
		Project current world color: self defaultFill.
	].
! !

!MorphicProject class methodsFor: 'instance creation' stamp: 'dtl 12/6/2009 17:16'!
openViewAndEnter: morphOrList
	"morphOrList has been reconstituted from a file. Open it in an appropriate
	Morphic world."

	| contents |
	contents := morphOrList.
	(morphOrList isKindOf: SqueakPage) ifTrue: [
		contents := morphOrList contentsMorph
	].
	(contents isKindOf: PasteUpMorph) ifFalse:
		[^ self inform: 'This is not a PasteUpMorph or
exported Project.' translated].
	(self openViewOn: contents) enter
! !

!MorphicProject class methodsFor: 'instance creation' stamp: 'ar 9/18/2009 21:31'!
openViewOn: aPasteUpOrNil
	"Create a new Morphic Project and open a view for the project"
	| newProject |
	newProject := self new.
	aPasteUpOrNil ifNotNil: [newProject installPasteUpAsWorld: aPasteUpOrNil].
	ProjectViewMorph openOn: newProject.
	^newProject
! !

!MorphicProject class methodsFor: 'utilities' stamp: 'mt 5/1/2016 16:53'!
releaseProjectReferences: outgoingProject
	"Iterate over all project-view morphs, wherever they may be located. Also consider image segments."
	
	ImageSegment allSubInstancesDo: [:seg |
		seg ifOutPointer: outgoingProject thenAllObjectsDo: [:obj |
			(obj isKindOf: ProjectViewMorph) ifTrue: [
				obj owner isSystemWindow
					ifTrue: [obj owner model: nil; delete].
				obj abandon]]].
	
	ProjectViewMorph allSubInstancesDo: [:p | 
		p owner isSystemWindow ifTrue: [p owner model: nil; delete].
		p project == outgoingProject ifTrue: [p abandon]].! !

!MorphicProject class methodsFor: 'preferences' stamp: 'mt 8/5/2016 10:28'!
themeProperties

	^ super themeProperties, {
		{ #background. 'Colors'. 'Color or fill-style for projects. Use-defined background will not be overwritten.' }.
	}! !

!MorphicProject class methodsFor: 'shrinking' stamp: 'mt 5/9/2016 10:24'!
unloadMorphic
	"MorphicProject unloadMorphic"

	Project current isMorphic ifTrue: [
		^ Error signal: 'You can only unload Morphic from within another kind of project.'].

	MorphicProject removeProjectsFromSystem.
	
	#(ActiveHand ActiveWorld ActiveEvent World) do: [:ea |
		Smalltalk globals removeKey: ea].

	{ 'ToolBuilder-Morphic' . 'MorphicTests' . 'MorphicExtras' . 'Morphic' }
		do: [ :package | (MCPackage named: package) unload ].

! !

!MorphicProject methodsFor: 'scheduling' stamp: 'dtl 9/20/2009 18:52'!
addDeferredUIMessage: valuableObject 
	"Arrange for valuableObject to be evaluated at a time when the user interface
	is in a coherent state."

	WorldState addDeferredUIMessage: valuableObject! !

!MorphicProject methodsFor: 'utilities' stamp: 'dtl 11/25/2009 13:54'!
addItem: item toMenu: menu selection: action color: aColor thumbnail: aForm
	"Add menu item representing the sender to a menu"

	menu
		add: item
		selector: #jumpToSelection:
		argument: action.
	menu lastItem color: aColor.
	aForm isNil
		ifFalse: [menu lastItem
				icon: (aForm
						scaledIntoFormOfSize: (Preferences tinyDisplay
								ifTrue: [16]
								ifFalse: [28]))]! !

!MorphicProject methodsFor: 'subprojects' stamp: 'mt 5/1/2016 17:43'!
addProject: project

	| view |
	super addProject: project.
	
	view := Preferences projectViewsInWindows
		ifTrue: [ProjectViewMorph newProjectViewInAWindowFor: project]
		ifFalse: [ProjectViewMorph on: project].
	
	"Do not use #openInWorld: because SystemWindow does things with real-estate manager, which depends on too much global state."
	self world
		addMorph: view;
		startSteppingSubmorphsOf: view.! !

!MorphicProject methodsFor: 'updating' stamp: 'mt 8/5/2016 10:47'!
applyUserInterfaceTheme

	super applyUserInterfaceTheme.
	self setWorldBackground: false.! !

!MorphicProject methodsFor: 'file in/out' stamp: 'nice 2/9/2010 15:17'!
armsLengthCommand: aCommand withDescription: aString
	| tempProject foolingForm tempCanvas bbox crossHatchColor stride |
	"Set things up so that this aCommand is sent to self as a message
after jumping to the parentProject.  For things that can't be executed
while in this project, such as saveAs, loadFromServer, storeOnServer.  See
ProjectViewMorph step."

	world borderWidth: 0.	"get rid of the silly default border"
	tempProject := MorphicProject new.
	foolingForm := world imageForm.		"make them think they never left"
	tempCanvas := foolingForm getCanvas.
	bbox := foolingForm boundingBox.
	crossHatchColor := Color yellow alpha: 0.3.
	stride := 20.
	10 to: bbox width by: stride do: [ :x |
		tempCanvas fillRectangle: (x@0 extent: 1@bbox height) fillStyle: crossHatchColor.
	].
	10 to: bbox height by: stride do: [ :y |
		tempCanvas fillRectangle: (0@y extent: bbox width@1) fillStyle: crossHatchColor.
	].

	tempProject world color: (InfiniteForm with: foolingForm).
	tempProject projectParameters 
		at: #armsLengthCmd 
		put: (
			DoCommandOnceMorph new
				addText: aString;
				actionBlock: [
					self doArmsLengthCommand: aCommand.
				]
		).
	tempProject projectParameters 
		at: #deleteWhenEnteringNewProject 
		put: true.
	tempProject enter
! !

!MorphicProject methodsFor: 'flaps support' stamp: 'dtl 10/31/2009 15:14'!
assureFlapIntegrity
	"Make certain that the items on the disabled-global-flap list are actually global flaps, and if not, get rid of them.  Also, old (and damaging) parameters that held references to actual disabled flaps are cleansed"

	| disabledFlapIDs currentGlobalIDs oldList |
	disabledFlapIDs := self parameterAt: #disabledGlobalFlapIDs ifAbsent: [Set new].
	currentGlobalIDs := Flaps globalFlapTabsIfAny collect: [:f | f flapID].
	oldList := Project current projectParameterAt: #disabledGlobalFlaps ifAbsent: [nil].
	oldList ifNotNil:
		[disabledFlapIDs := oldList collect: [:aFlap | aFlap flapID].
		disabledFlapIDs addAll: {'Scripting' translated. 'Stack Tools' translated. 'Painting' translated}].
	disabledFlapIDs := disabledFlapIDs select: [:anID | currentGlobalIDs includes: anID].
	self projectParameterAt: #disabledGlobalFlapIDs put: disabledFlapIDs asSet.
	self assureNavigatorPresenceMatchesPreference.

	projectParameters ifNotNil:
		[projectParameters removeKey: #disabledGlobalFlaps ifAbsent: []]! !

!MorphicProject methodsFor: 'docking bars support' stamp: 'dtl 10/31/2009 14:20'!
assureMainDockingBarPresenceMatchesPreference
	"Synchronize the state of the receiver's dockings with the  
	preference"
	(self showWorldMainDockingBar)
		ifTrue: [self createOrUpdateMainDockingBar]
		ifFalse: [self removeMainDockingBar]! !

!MorphicProject methodsFor: 'menu messages' stamp: 'dtl 11/6/2009 21:33'!
assureNavigatorPresenceMatchesPreference
	"Make sure that the current project conforms to the presence/absence of the navigator"

	| navigator navType wantIt |
	wantIt :=  Preferences classicNavigatorEnabled and: [Preferences showProjectNavigator].
	navType := ProjectNavigationMorph preferredNavigator.
	navigator := world findA: navType.
	wantIt
		ifFalse:
			[navigator ifNotNil: [navigator delete]]
		ifTrue:
			[navigator isNil ifTrue: 
				[(navigator := navType new)
					bottomLeft: world bottomLeft;
					openInWorld: world]]! !

!MorphicProject methodsFor: 'editors' stamp: 'dtl 1/23/2010 18:26'!
bitEdit: aForm
	"Create and schedule a view located in an area designated by the user 
	that contains a view of the receiver magnified by 8@8 that can be 
	modified using the Bit Editor. It also contains a view of the original 
	form."

	aForm currentHand attachMorph: (FatBitsPaint new editForm: aForm;
			 magnification: 8;
			 brushColor: Color black;
			 penSize: 1;
			 yourself)
! !

!MorphicProject methodsFor: 'editors' stamp: 'dtl 1/24/2010 16:09'!
bitEdit: aForm at: magnifiedFormLocation scale: scaleFactor
	"Create and schedule a view whose top left corner is magnifiedLocation 
	and that contains a view of aForm magnified by scaleFactor that  can be
	modified using the Bit Editor. It also contains a view of the original form."

	self inform: 'A Morphic editor has not been implemented. Enter an MVC project to edit this form.'! !

!MorphicProject methodsFor: 'updating' stamp: 'mt 8/5/2016 10:47'!
canApplyUserInterfaceTheme

	^ true! !

!MorphicProject methodsFor: 'language' stamp: 'tfel 7/28/2016 16:04'!
chooseNaturalLanguage
	"Put up a menu allowing the user to choose the natural language for the project"

	| aMenu availableLanguages |
	aMenu := MenuMorph new defaultTarget: self.
	aMenu addTitle: 'choose language' translated.
	aMenu lastItem setBalloonText: 'This controls the human language in which tiles should be viewed.  It is potentially extensible to be a true localization mechanism, but initially it only works in the classic tile scripting system.  Each project has its own private language choice' translated.
	Preferences noviceMode
		ifFalse:[aMenu addStayUpItem].

	availableLanguages := NaturalLanguageTranslator availableLanguageLocaleIDs
										asSortedCollection:[:x :y | x displayName < y displayName].

	availableLanguages do:
		[:localeID |
			aMenu addUpdating: #stringForLanguageNameIs: target: Locale selector:  #switchAndInstallFontToID: argumentList: {localeID}].
	aMenu popUpInWorld

"Project current chooseNaturalLanguage"! !

!MorphicProject methodsFor: 'flaps support' stamp: 'cmm 12/6/2010 15:09'!
cleanseDisabledGlobalFlapIDsList
	"Make certain that the items on the disabled-global-flap list are actually global flaps, and if not, get rid of them"

	| disabledFlapIDs currentGlobalIDs oldList |
	disabledFlapIDs := self parameterAt: #disabledGlobalFlapIDs ifAbsent: [Set new].
	currentGlobalIDs := Flaps globalFlapTabsIfAny collect: [:f | f flapID].
	oldList := Project current projectParameterAt: #disabledGlobalFlaps ifAbsent: [nil].
	oldList ifNotNil:
		[disabledFlapIDs := oldList select: [:aFlap | aFlap flapID]].
	disabledFlapIDs := disabledFlapIDs select: [:anID | currentGlobalIDs includes: anID].
	self projectParameterAt: #disabledGlobalFlapIDs put: disabledFlapIDs.

	self removeParameter: #disabledGlobalFlaps.
! !

!MorphicProject methodsFor: 'accessing' stamp: 'mt 4/12/2016 19:02'!
color

	^ self world isInMemory
		ifTrue: [Color black]
		ifFalse: [Color brown]! !

!MorphicProject methodsFor: 'utilities' stamp: 'mt 5/4/2016 15:33'!
composeDisplayTextIntoForm: displayText

	| tmpText canvas |
	tmpText := TextMorph new contentsAsIs: displayText text deepCopy.
	displayText foregroundColor ifNotNil: [:color |
		tmpText text addAttribute: (TextColor color: color)].
	displayText backgroundColor ifNotNil: [:color |
		tmpText backgroundColor: color].
	tmpText setTextStyle: displayText textStyle.
	canvas := FormCanvas on: (Form extent: tmpText extent depth: 32).
	tmpText drawOn: canvas.
	^ canvas form! !

!MorphicProject methodsFor: 'docking bars support' stamp: 'mt 5/31/2016 12:01'!
createOrUpdateMainDockingBar
	"Private - create a new main docking bar or update the current one"
	| w mainDockingBars |
	w := self world.
	mainDockingBars := w mainDockingBars.
	mainDockingBars isEmpty ifTrue:
		[ "no docking bar, just create a new one"
		self dockingBar createDockingBar
			 openInWorld: w .
		^ self ].
	"update if needed"
	mainDockingBars do:
		[ : each | self dockingBar updateIfNeeded: each ]! !

!MorphicProject methodsFor: 'utilities' stamp: 'yo 6/20/2007 01:24'!
createViewIfAppropriate
	"Create a project view for the receiver and place it appropriately on the screen."

	| aMorph requiredWidth existing proposedV proposedH despair |
	ProjectViewOpenNotification signal ifTrue:
		[Preferences projectViewsInWindows
			ifTrue:
				[(ProjectViewMorph newProjectViewInAWindowFor: self) openInWorld]
			ifFalse:
				[aMorph := ProjectViewMorph on: self.
				requiredWidth := aMorph width + 10.
				existing := ActiveWorld submorphs
					select: [:m | m isKindOf: ProjectViewMorph]
					thenCollect: [:m | m fullBoundsInWorld].
				proposedV := 85.
				proposedH := 10.
				despair := false.
				[despair not and: [((proposedH @ proposedV) extent: requiredWidth) intersectsAny: existing]] whileTrue:
					[proposedH := proposedH + requiredWidth.
					proposedH + requiredWidth > ActiveWorld right ifTrue:
						[proposedH := 10.
						proposedV := proposedV + 90.
						proposedV > (ActiveWorld bottom - 90)
							ifTrue:
								[proposedH := ActiveWorld center x - 45.
								proposedV := ActiveWorld center y - 30.
								despair := true]]].
				aMorph position: (proposedH @ proposedV).
				aMorph openInWorld]]! !

!MorphicProject methodsFor: 'protocols' stamp: 'ul 3/15/2010 06:54'!
currentVocabulary

	^ActiveWorld currentVocabulary! !

!MorphicProject methodsFor: 'debugging' stamp: 'mt 4/11/2016 10:41'!
debugMethod: aCompiledMethod forReceiver: anObject inContext: aContextOrNil

	| guineaPig debugger debuggerWindow context |
	guineaPig :=
		[aCompiledMethod
			valueWithReceiver: anObject
			 arguments: (aContextOrNil ifNil: [ #() ] ifNotNil: [ { aContextOrNil } ]).
		 guineaPig := nil "spot the return from aCompiledMethod"] newProcess.
	context := guineaPig suspendedContext.
	debugger := Debugger new
		process: guineaPig
		controller: nil
		context: context.
	debuggerWindow := debugger openFullNoSuspendLabel: 'Debug it'.
	"Now step into the expression.  But if it is quick (is implemented as a primtiive, e.g. `0')
	 it will return immediately back to the block that is sent newProcess above.  Guard
	 against that with the check for home being thisContext."
	[debugger interruptedContext method == aCompiledMethod]
		whileFalse:
			[(guineaPig isNil
			  and: [debugger interruptedContext home == thisContext]) ifTrue:
				[debuggerWindow delete.
				 UIManager default inform: 'Nothing to debug; expression is optimized'.
				 ^self].
			debugger send]! !

!MorphicProject methodsFor: 'release' stamp: 'mt 5/1/2016 17:01'!
deletingProject: outgoingProject

	(self world submorphs
		select: [:m | m isSystemWindow and: [m model == outgoingProject]]
		thenCollect: [:window | window paneMorphs first])
		do: [:projectViewMorph |
			projectViewMorph owner "window" model: nil; delete.
			projectViewMorph abandon].

	super deletingProject: outgoingProject.! !

!MorphicProject methodsFor: 'transcripter' stamp: 'dtl 12/14/2013 16:20'!
displayTranscripter: transcripter
	"A transcripter is a minimal user interface to support an emergency evaluator.
	A Morphic project uses class NewParagraph."
	transcripter morphicDisplayText! !

!MorphicProject methodsFor: 'docking bars support' stamp: 'cmm 12/6/2010 15:05'!
dockingBar
	^ self
		projectParameterAt: #dockingBar
		ifAbsent: [ TheWorldMainDockingBar instance ]! !

!MorphicProject methodsFor: 'docking bars support' stamp: 'cmm 12/6/2010 15:26'!
dockingBar: aTheWorldMainDockingBar 
	self
		projectParameterAt: #dockingBar
		put: aTheWorldMainDockingBar.
	self isCurrentProject ifTrue: [ TheWorldMainDockingBar instance: aTheWorldMainDockingBar ]! !

!MorphicProject methodsFor: 'utilities' stamp: 'dtl 11/30/2009 22:27'!
do: aBlock withProgressInfoOn: aMorphOrNil label: aString
	"Evaluate aBlock with a labeled progress bar"

	ComplexProgressIndicator new 
		targetMorph: aMorphOrNil;
		historyCategory: aString;
		withProgressDo: aBlock
! !

!MorphicProject methodsFor: 'editors' stamp: 'dtl 1/24/2010 12:10'!
editCharacter: character ofFont: strikeFont
	"Open a bit editor on a character in the given strike font."
	"Note that BitEditor only works in MVC currently."

	"(TextStyle default fontAt: 1) edit: $="

	self inform: 'A Morphic editor has not been implemented. Enter an MVC project to edit this font.'! !

!MorphicProject methodsFor: 'flaps support' stamp: 'dtl 10/31/2009 15:16'!
enableDisableGlobalFlap: aFlapTab
	"For the benefit of pre-existing which-global-flap buttons from a design now left behind."

	self flag: #toRemove.
	^ self inform: 
'Sorry, this is an obsolete menu; please
dismiss it and get a fresh menu.  Thanks.'.! !

!MorphicProject methodsFor: 'squeaklet on server' stamp: 'dtl 11/27/2009 13:42'!
enterIfThereOrFind: aProjectName

	| newProject |
	newProject := Project named: aProjectName.
	newProject ifNotNil: [^newProject enter].

	ComplexProgressIndicator new 
		targetMorph: nil;
		historyCategory: 'project loading';
		withProgressDo: [
			[
				newProject := self fromMyServerLoad: aProjectName
			] 
				on: ProjectViewOpenNotification
				do: [ :ex | ex resume: false]		
					"we probably don't want a project view morph in this case"
		].

	newProject ifNotNil: [^newProject enter].
	Beeper beep.! !

!MorphicProject methodsFor: 'file in/out' stamp: 'eem 5/22/2016 17:50'!
exportSegmentWithChangeSet: aChangeSetOrNil fileName: aFileName
directory: aDirectory
	"Store my project out on the disk as an *exported*
ImageSegment.  All outPointers will be in a form that can be resolved
in the target image.  Name it <project name>.extSeg.  Whatdo we do
about subProjects, especially if they are out as local image
segments?  Force them to come in?
	Player classes are included automatically."

	| is str ans revertSeg roots holder collector fd mgr stacks |

	"Files out a changeSet first, so that a project can contain
its own classes"
	world ifNil: [^ false].
	world presenter ifNil: [^ false].

	ScrapBook default emptyScrapBook.
	world currentHand pasteBuffer: nil.	  "don't write the paste buffer."
	world currentHand mouseOverHandler initialize.	  "forget about any
	references here"
		"Display checkCurrentHandForObjectToPaste."
	Command initialize.
	world clearCommandHistory.
	world fullReleaseCachedState; releaseViewers.
	world cleanseStepList.
	world localFlapTabs size = world flapTabs size ifFalse: [
		self error: 'Still holding onto Global flaps'].
	world releaseSqueakPages.
	holder := Project allProjects.	"force them in to outPointers, where
	DiskProxys are made"

	"Just export me, not my previous version"
	revertSeg := self parameterAt: #revertToMe.
	self removeParameter: #revertToMe.

	roots := OrderedCollection new.
	roots add: self; add: world; add: transcript; add: aChangeSetOrNil; add: thumbnail; add: world activeHand.

		"; addAll: classList; addAll: (classList collect: [:cls | cls class])"

	roots := roots reject: [ :x | x isNil].	"early saves may not have
	active hand or thumbnail"

	fd := aDirectory directoryNamed: self resourceDirectoryName.
	fd assureExistence.
	"Clean up resource references before writing out"
	mgr := self resourceManager.
	self resourceManager: nil.
	ResourceCollector current: ResourceCollector new.
	ResourceCollector current localDirectory: fd.
	ResourceCollector current baseUrl: self resourceUrl.
	ResourceCollector current initializeFrom: mgr.
	ProgressNotification signal: '2:findingResources' extra:
'(collecting resources...)' translated.
	"Must activate old world because this is run at #armsLength.
	Otherwise references to ActiveWorld, ActiveHand, or ActiveEvent
	will not be captured correctly if referenced from blocks or user code."
	world becomeActiveDuring:[
		is := ImageSegment new copySmartRootsExport: roots asArray.
		"old way was (is := ImageSegment new
copyFromRootsForExport: roots asArray)"
	].
	self resourceManager: mgr.
	collector := ResourceCollector current.
	ResourceCollector current: nil.
	ProgressNotification signal: '2:foundResources' extra: ''.
	is state = #tooBig ifTrue: [
		collector replaceAll.
		^ false].

	str := ''.
	"considered legal to save a project that has never been entered"
	(is outPointers includes: world) ifTrue: [
		str := str, '\Project''s own world is not in the segment.' translated withCRs].
	str isEmpty ifFalse: [
		ans := UIManager default chooseFrom: {
			'Do not write file' translated.
			'Write file anyway' translated.
			'Debug' translated.
		} title: str.
		ans = 1 ifTrue: [
			revertSeg ifNotNil: [projectParameters at:
	#revertToMe put: revertSeg].
			collector replaceAll.
			^ false].
		ans = 3 ifTrue: [
			collector replaceAll.
			self halt: 'Segment not written' translated]].
		stacks := is findStacks.

		is
			writeForExportWithSources: aFileName
			inDirectory: fd
			changeSet: aChangeSetOrNil.
		SecurityManager default signFile: aFileName directory: fd.
		"Compress all files and update check sums"
		collector forgetObsolete.
		self storeResourceList: collector in: fd.
		self storeHtmlPageIn: fd.
		self storeManifestFileIn: fd.
		self writeStackText: stacks in: fd registerIn: collector.
		"local proj.005.myStack.t"
		self compressFilesIn: fd to: aFileName in: aDirectory
	resources: collector.
				"also deletes the resource directory"
		"Now update everything that we know about"
		mgr updateResourcesFrom: collector.

	revertSeg ifNotNil: [projectParameters at: #revertToMe put: revertSeg].
	holder.

	collector replaceAll.

	world flapTabs do: [:ft |
			(ft respondsTo: #unhibernate) ifTrue: [ft unhibernate]].
	is arrayOfRoots do: [:obj |
		obj isScriptEditorMorph ifTrue: [obj unhibernate]].
	^ true
! !

!MorphicProject methodsFor: 'enter' stamp: 'mt 5/1/2016 13:20'!
finalEnterActions: leavingProject
	"Perform the final actions necessary as the receiver project is entered"

	| navigator armsLengthCmd navType thingsToUnhibernate |
	World := world.  "Signifies Morphic"
	world install.
	world transferRemoteServerFrom: leavingProject world.
	"(revertFlag | saveForRevert | forceRevert) ifFalse: [
		(Preferences valueOfFlag: #projectsSentToDisk) ifTrue: [
			self storeSomeSegment]]."
	
	"Transfer event recorder to me."
	leavingProject isMorphic ifTrue: [
		leavingProject world pauseEventRecorder ifNotNil: [:rec |
			rec resumeIn: world]].

	world triggerOpeningScripts.


	self initializeMenus.
	self projectParameters 
		at: #projectsToBeDeleted 
		ifPresent: [ :projectsToBeDeleted |
			self removeParameter: #projectsToBeDeleted.
			projectsToBeDeleted do: [:each | each delete]].

	Locale switchAndInstallFontToID: self localeID.

	thingsToUnhibernate := world valueOfProperty: #thingsToUnhibernate ifAbsent: [#()].
	thingsToUnhibernate do: [:each | each unhibernate].
	world removeProperty: #thingsToUnhibernate.

	navType := ProjectNavigationMorph preferredNavigator.
	armsLengthCmd := self parameterAt: #armsLengthCmd ifAbsent: [nil].
	navigator := world findA: navType.
	(Preferences classicNavigatorEnabled and: [Preferences showProjectNavigator and: [navigator isNil]]) ifTrue:
		[(navigator := navType new)
			bottomLeft: world bottomLeft;
			openInWorld: world].
	navigator notNil & armsLengthCmd notNil ifTrue:
		[navigator color: Color lightBlue].
	armsLengthCmd notNil ifTrue:
		[Preferences showFlapsWhenPublishing
			ifFalse:
				[self flapsSuppressed: true.
				navigator ifNotNil:	[navigator visible: false]].
		armsLengthCmd openInWorld: world].
	world reformulateUpdatingMenus.
	world presenter positionStandardPlayer.
	self assureMainDockingBarPresenceMatchesPreference.

	world repairEmbeddedWorlds.! !

!MorphicProject methodsFor: 'enter' stamp: 'mt 4/17/2016 18:24'!
finalExitActions: enteringProject

	world triggerClosingScripts.

	"Pause sound players, subject to preference settings"
	(world hasProperty: #letTheMusicPlay)
		ifTrue: [world removeProperty: #letTheMusicPlay]
		ifFalse: [SoundService stop].

	world sleep.
	
	(world findA: ProjectNavigationMorph)
		ifNotNil: [:navigator | navigator retractIfAppropriate].

	"Clean-up global state."
	World := nil.
	ActiveWorld := ActiveHand := ActiveEvent := nil.
	Sensor flushAllButDandDEvents. ! !

!MorphicProject methodsFor: 'utilities' stamp: 'dtl 12/2/2009 20:29'!
findAFolderForProject: aProject label: dialogLabel
	"Find a folder for saving or loading a project"

	^FileList2 modalFolderSelectorForProject: aProject
! !

!MorphicProject methodsFor: 'utilities' stamp: 'nice 12/27/2009 03:10'!
findProjectView: projectDescription
	"In this world, find the morph that holds onto the project described by projectDescription.
	projectDescription can be a project, or the name of a project.  The project may be
	represented by a DiskProxy. The holder morph may be at any depth in the world.."

	| pName |
	pName := (projectDescription isString) 
		ifTrue: [projectDescription]
		ifFalse: [projectDescription name].
	world allMorphsDo: [:pvm | | dpName |
	pvm class == ProjectViewMorph ifTrue: [
		(pvm project class == Project and: 
			[pvm project name = pName]) ifTrue: [^ pvm].
			pvm project class == DiskProxy ifTrue: [ 
			dpName := pvm project constructorArgs first.
			dpName := (dpName findTokens: '/') last.
			dpName := (Project parseProjectFileName: dpName unescapePercents) first.
			dpName = pName ifTrue: [^ pvm]]]].
	^ nil! !

!MorphicProject methodsFor: 'flaps support' stamp: 'dtl 11/13/2009 18:19'!
flapsSuppressed: aBoolean
	"Make the setting of the flag that governs whether global flaps are suppressed in the project be as indicated and add or remove the actual flaps"

	super flapsSuppressed: aBoolean.
	self == Project current
		ifFalse:   "Anomalous case where this project is not the current one."
			[aBoolean
				ifTrue:		
					[Flaps globalFlapTabsIfAny do:
						[:aFlapTab | Flaps removeFlapTab: aFlapTab keepInList: true]]

				ifFalse:
					[Smalltalk isMorphic  ifTrue:
						[self currentWorld addGlobalFlaps]]].
	Project current assureNavigatorPresenceMatchesPreference! !

!MorphicProject methodsFor: 'editors' stamp: 'dtl 2/10/2010 14:20'!
formEdit: aForm
	"Start up an instance of the form editor on a form."
 
	self inform: 'A Morphic editor has not been implemented. Enter an MVC project to edit this form or use #bitEdit'! !

!MorphicProject methodsFor: 'editors' stamp: 'dtl 2/10/2010 13:18'!
formViewClass
	"Answer a class suitable for a view on a form or collection of forms"

	^ GraphicalDictionaryMenu! !

!MorphicProject methodsFor: 'futures' stamp: 'jcg 1/9/2010 13:12'!
future: receiver do: aSelector at: deltaMSecs args: args
	"Send a message deltaSeconds into the future.  No response is expected."
	| msg |
	msg := MessageSend receiver: receiver selector: aSelector arguments: args.
	deltaMSecs = 0 
		ifTrue: [self addDeferredUIMessage: msg]
		ifFalse: [
			world 
				addAlarm: #addDeferredUIMessage: 
				withArguments: {msg}
				for: self
				at: (Time millisecondClockValue + deltaMSecs)
		]..
	^nil! !

!MorphicProject methodsFor: 'futures' stamp: 'jcg 1/9/2010 13:12'!
future: receiver send: aSelector at: deltaMSecs args: args
	"Send a message deltaSeconds into the future.  Answers a Promise that will be resolved at some time in the future."
	| pr closure |
	pr := Promise new.
	closure := [pr resolveWith: (receiver perform: aSelector withArguments: args)].
	deltaMSecs = 0
		ifTrue: [self addDeferredUIMessage: closure]
		ifFalse: [
			world 
				addAlarm: #addDeferredUIMessage: 
				withArguments: {closure}
				for: self
				at: (Time millisecondClockValue + deltaMSecs)
		].
	^pr
		! !

!MorphicProject methodsFor: 'flaps support' stamp: 'dtl 10/31/2009 15:19'!
globalFlapEnabledString: aFlapTab
	"Answer the string to be shown in a menu to represent the status of the given flap regarding whether it it should be shown in this project."

	^ (self isFlapEnabled: aFlapTab)
		ifTrue:
			['<on>', aFlapTab wording]
		ifFalse:
			['<off>', aFlapTab wording]! !

!MorphicProject methodsFor: 'flaps support' stamp: 'dtl 10/31/2009 15:20'!
globalFlapWithIDEnabledString: aFlapID
	"Answer the string to be shown in a menu to represent the status of the given flap regarding whether it it should be shown in this project."

	| aFlapTab |
	aFlapTab := Flaps globalFlapTabWithID: aFlapID.
	^ (self isFlapEnabled: aFlapTab)
		ifTrue:
			['<on>', aFlapTab wording]
		ifFalse:
			['<off>', aFlapTab wording]! !

!MorphicProject methodsFor: 'initialize' stamp: 'mt 8/5/2016 10:34'!
initialize
	"Initialize a new Morphic Project"
	super initialize.
	world := PasteUpMorph newWorldForProject: self.
	self setWorldBackground: true.
	Locale switchToID: CurrentProject localeID.
	Preferences useVectorVocabulary ifTrue: [world installVectorVocabulary]! !

!MorphicProject methodsFor: 'enter' stamp: 'nice 8/5/2011 22:47'!
initializeMenus
	"Menu setting for these classes may have been modified by another
	Morphic project, e.g. SimpleMorphic. Ensure that they are restored to the
	expected menus."

	Editor initialize.! !

!MorphicProject methodsFor: 'transcripter' stamp: 'dtl 12/14/2013 15:06'!
initializeParagraphForTranscripter: transcripter
	"A transcripter is a minimal user interface to support an emergency evaluator.
	A Morphic project uses class NewParagraph."
	transcripter morphicInitializeParagraph: NewParagraph! !

!MorphicProject methodsFor: 'initialize' stamp: 'ar 11/12/2009 01:11'!
initMorphic
	"Written so that Morphic can still be removed.  Note that #initialize is never actually called for a morphic project -- see the senders of this method."

	self flag: #toRemove. "check if this method still used by Etoys"
	Smalltalk verifyMorphicAvailability ifFalse: [^ nil].
	changeSet := ChangeSet new.
	transcript := TranscriptStream new.
	displayDepth := Display depth.
	parentProject := CurrentProject.
	world := PasteUpMorph newWorldForProject: self.
	Locale switchToID: CurrentProject localeID.
	self initializeProjectPreferences. "Do this *after* a world is installed so that the project will be recognized as a morphic one."
	Preferences useVectorVocabulary ifTrue: [world installVectorVocabulary]! !

!MorphicProject methodsFor: 'initialize' stamp: 'mt 8/5/2016 10:34'!
installPasteUpAsWorld: pasteUpMorph
	"(ProjectViewMorph newMorphicProjectOn: aPasteUpMorph) openInWorld."

	world := pasteUpMorph beWorldForProject: self.
	self setWorldBackground: true.! !

!MorphicProject methodsFor: 'scheduling' stamp: 'fm 9/11/2010 00:54'!
interruptName: labelString
	"Create a Notifier on the active scheduling process with the given label."

	^ self interruptName: labelString preemptedProcess: nil! !

!MorphicProject methodsFor: 'utilities' stamp: 'mt 4/12/2016 19:34'!
interruptName: labelString preemptedProcess: theInterruptedProcess
	"Create a Notifier on the active scheduling process with the given label."

	| preemptedProcess projectProcess |
	ActiveHand ifNotNil:[ActiveHand interrupted].
	ActiveWorld := world. "reinstall active globals"
	ActiveHand := world primaryHand.
	ActiveHand interrupted. "make sure this one's interrupted too"
	ActiveEvent := nil.

	projectProcess := self uiProcess.	"we still need the accessor for a while"
	preemptedProcess := theInterruptedProcess ifNil: [Processor preemptedProcess].
	"Only debug preempted process if its priority is >= projectProcess' priority"
	preemptedProcess priority < projectProcess priority 
		ifTrue:[preemptedProcess := projectProcess].
	preemptedProcess suspend.
	ToolSet interrupt: preemptedProcess label: labelString.! !

!MorphicProject methodsFor: 'display' stamp: 'mt 4/13/2016 10:30'!
invalidate
	"Invalidate the entire project so that a redraw will be forced later."
	world restoreMorphicDisplay.! !

!MorphicProject methodsFor: 'flaps support' stamp: 'dtl 10/31/2009 15:20'!
isFlapEnabled:  aFlapTab
	"Answer whether the given flap tab is enabled in this project"

	^ self isFlapIDEnabled: aFlapTab flapID! !

!MorphicProject methodsFor: 'flaps support' stamp: 'dtl 10/31/2009 15:21'!
isFlapIDEnabled:  aFlapID
	"Answer whether a flap of the given ID is enabled in this project"

	| disabledFlapIDs  |
	disabledFlapIDs := self parameterAt: #disabledGlobalFlapIDs ifAbsent: [^ true].
	^ (disabledFlapIDs includes: aFlapID) not! !

!MorphicProject methodsFor: 'enter' stamp: 'dtl 11/7/2009 21:08'!
isIncompletelyLoaded
	"Answer true if project is incomplete and should be loaded from server "

	(world isKindOf: StringMorph)
		ifTrue: [self inform: 'This project is not all here. I will try to load a complete version.' translated.
			^ true].
	^ false! !

!MorphicProject methodsFor: 'testing' stamp: 'ar 9/18/2009 20:15'!
isMorphic
	"Duh."
	^true! !

!MorphicProject methodsFor: 'utilities' stamp: 'dtl 11/23/2009 21:45'!
jumpToProject
	"Present a list of potential projects and enter the one selected."

	"Project current jumpToProject"

	| menu |
menu:=MenuMorph new.
menu defaultTarget: self.
	menu := self buildJumpToMenu: menu.
	menu popUpInWorld! !

!MorphicProject methodsFor: 'file in/out' stamp: 'dtl 11/29/2009 13:20'!
loadFromServer: newerAutomatically
	"If a newer version of me is on the server, load it."
	| pair resp server |
	self assureIntegerVersion.

	self isCurrentProject ifTrue: ["exit, then do the command"
		^ self armsLengthCommand: #loadFromServer withDescription: 'Loading' translated
	].
	server := self tryToFindAServerWithMe ifNil: [^ nil].
	pair := self class mostRecent: self name onServer: server.
	pair first ifNil: [^ self inform: ('can''t find file on server for {1}' translated format: {self name})].
	self currentVersionNumber > pair second ifTrue: [
		^ self inform: ('That server has an older version of the project.' translated)].
	version = (Project parseProjectFileName: pair first) second ifTrue: [
		resp := (UIManager default chooseFrom: 
				(Array with: 'Reload anyway' translated 
						with: 'Cancel' translated withCRs) 
				title:  'The only changes are the ones you made here.' translated).
		resp ~= 1 ifTrue: [^ nil]
	] ifFalse: [
		newerAutomatically ifFalse: [
			resp := (UIManager default 
						chooseFrom: #('Load it' 'Cancel') 
						title:  'A newer version exists on the server.').
			resp ~= 1 ifTrue: [^ nil]
		].
	].

	"let's avoid renaming the loaded change set since it will be replacing ours"
	self projectParameters at: #loadingNewerVersion put: true.

	ComplexProgressIndicator new 
		targetMorph: nil;
		historyCategory: 'project loading';
		withProgressDo: [
			ProjectLoading
				installRemoteNamed: pair first
				from: server
				named: self name
				in: parentProject
		]
! !

!MorphicProject methodsFor: 'menu messages' stamp: 'ar 9/18/2009 20:17'!
makeThumbnail
	"Make a thumbnail image of this project from the Display."
	world displayWorldSafely. "clean pending damage"
	super makeThumbnail.! !

!MorphicProject methodsFor: 'display' stamp: 'mt 7/28/2016 15:05'!
noDisplayDuring: block

	self world noDisplayDuring: block.! !

!MorphicProject methodsFor: 'file in/out' stamp: 'sw 10/3/2009 17:56'!
noteManifestDetailsIn: manifestInfo
	"The receiver is a project being loaded.  From the dictionary provided, absorb and remember whether it's an 'old' (pre-olpc) project, and remember the GUID, user, and prev-GUID associated with the project when these data are available in the incoming manifest."

	| manifestDict oldProject |
	manifestInfo isEmptyOrNil ifTrue: [^ self projectParameterAt: #oldProject put: true].

	manifestDict := (manifestInfo isKindOf: Dictionary) ifTrue: [manifestInfo] ifFalse: [manifestInfo first].

	oldProject := ((manifestDict at: 'Squeak-Version' ifAbsent: ['']) beginsWith: 'etoys') not.
	self projectParameterAt: #oldProject put: oldProject.

	manifestDict at: #URI ifPresent: [:aUri | self projectParameterAt: #URI put: aUri].
	manifestDict at: #user ifPresent: [:aUser | self projectParameterAt: #user put: aUser].
	manifestDict at: #'prev-URI' ifPresent: [:aUri | self projectParameterAt: #'prev-URI' put: aUri]! !

!MorphicProject methodsFor: 'utilities' stamp: 'ar 1/10/2010 10:23'!
offerMenu: menuSelector from: aModel shifted: aBoolean
	"Pop up a menu whose target is aModel and whose contents are provided
	by sending the menuSelector to the model. The menuSelector takes two
	arguments: a menu, and a boolean representing the shift state."

	| aMenu |
	aMenu := MenuMorph new defaultTarget: aModel.
	aModel perform: menuSelector with: aMenu with: aBoolean.
	aMenu popUpInWorld
! !

!MorphicProject methodsFor: 'release' stamp: 'mt 5/1/2016 13:39'!
okToChange
	"If the view is from somewhere else than the current project, just delete it."
	
	^ self parent ~~ Project current
		or: [super okToChange]! !

!MorphicProject methodsFor: 'squeaklet on server' stamp: 'dtl 11/27/2009 13:38'!
openBlankProjectNamed: projName

	| proj projViewer |

	proj := MorphicProject openViewOn: nil.
	proj changeSet name: projName.
	proj world addMorph: (
		TextMorph new 
			beAllFont: ((TextStyle default fontOfSize: 26) emphasized: 1);
			color: Color red;
			contents: 'Welcome to a new project - ',projName
	).
	proj setParent: self.
	projViewer := (self findProjectView: projName) ifNil: [^proj].
	(projViewer owner isSystemWindow) ifTrue: [
			projViewer owner model: proj].
	^ projViewer project: proj! !

!MorphicProject methodsFor: 'editors' stamp: 'mt 4/12/2016 19:34'!
openImage: aForm name: fullName saveResource: aBoolean
	"Open a view on an image. If aBoolean is true, save the image as a project resource."

	aBoolean ifTrue:
		[self resourceManager 
			addResource: aForm 
			url: (FileDirectory urlForFileNamed: fullName) asString].
	(world drawingClass withForm: aForm) openInWorld
! !

!MorphicProject methodsFor: 'initialize' stamp: 'ar 9/18/2009 21:33'!
openProject: aProject
	"Create a new for a new project in the context of the receiver"
	ProjectViewMorph openOn: aProject.! !

!MorphicProject methodsFor: 'utilities' stamp: 'mt 4/12/2016 19:34'!
pointerMoved
	"Pointer has moved during a drag operation. Perform any necessary updates.
	In Morphic, pay the price for reading the sensor directly."

	world activeHand
			newMouseFocus: nil;
			showTemporaryCursor: nil;
			flushEvents! !

!MorphicProject methodsFor: 'release' stamp: 'dtl 11/22/2009 17:53'!
prepareForDelete
	"The window in which the project is housed is about to deleted. Perform
	any necessary actions to prepare for deletion."

	| is list |
	Smalltalk at: #WonderlandCameraMorph ifPresent:[:aClass |
		world submorphs do:   "special release for wonderlands"
					[:m | (m isKindOf: aClass)
							and: [m getWonderland release]]].
	"Remove Player classes and metaclasses owned by project"
	is := ImageSegment new arrayOfRoots: (Array with: self).
	(list := is rootsIncludingPlayers) ifNotNil:
		[list do: [:playerCls | 
			(playerCls respondsTo: #isMeta) ifTrue:
				[playerCls isMeta ifFalse:
					[playerCls removeFromSystemUnlogged]]]]

! !

!MorphicProject methodsFor: 'docking bars support' stamp: 'dtl 10/31/2009 14:21'!
removeMainDockingBar
	"Remove the receiver's main docking bars"
	self world mainDockingBars
		do: [:each | each delete]! !

!MorphicProject methodsFor: 'display' stamp: 'mt 4/17/2016 15:48'!
restore
	"Display world safely. Catch all errors to avoid image freeze. We assume that the world will avoid drawing erroneous morphs twice"

	| finished |
	finished := false.
	
	[finished] whileFalse: [
		[world displayWorldSafely. finished := true]
			on: Error do: [:err | world fullRepaintNeeded]].
! !

!MorphicProject methodsFor: 'enter' stamp: 'mt 4/17/2016 18:17'!
resumeEventRecorder: recorder

	recorder ifNotNil: [:rec | rec resumeIn: world].! !

!MorphicProject methodsFor: 'enter' stamp: 'mt 4/17/2016 18:50'!
scheduleProcessForEnter
	"Complete the enter: by launching a new process"

	self spawnNewProcess.! !

!MorphicProject methodsFor: 'dispatching' stamp: 'dtl 1/30/2010 15:50'!
selectorPrefixForDispatch
	"A string to be prepended to selectors for project specific methods"

	^ 'morphic'! !

!MorphicProject methodsFor: 'utilities' stamp: 'mt 8/5/2016 10:33'!
setAsBackground: aForm
	"Set  aForm as a background image."

	| thisWorld newColor |
	thisWorld := self currentWorld.
	newColor := InfiniteForm with: aForm.
	aForm rememberCommand:
		(Command new cmdWording: 'set background to a picture' translated;
			undoTarget: thisWorld selector: #color: argument: thisWorld color;
			redoTarget: thisWorld selector: #color: argument: newColor).
	thisWorld color: newColor.
	thisWorld setProperty: #hasCustomBackground toValue: true.
! !

!MorphicProject methodsFor: 'flaps support' stamp: 'mt 4/12/2016 19:35'!
setFlaps

	| flapTabs flapIDs sharedFlapTabs navigationMorph |
	self flag: #toRemove. "check if this method still used by Etoys"

	flapTabs := ActiveWorld flapTabs.
	flapIDs := flapTabs collect: [:tab | tab knownName].
	flapTabs
		do: [:tab | (tab isMemberOf: ViewerFlapTab)
				ifFalse: [tab isGlobalFlap
						ifTrue: [Flaps removeFlapTab: tab keepInList: false.
							tab currentWorld reformulateUpdatingMenus]
						ifFalse: [| referent | 
							referent := tab referent.
							referent isInWorld
								ifTrue: [referent delete].
							tab delete]]].
	sharedFlapTabs := Flaps classPool at: #SharedFlapTabs.
	flapIDs
		do: [:id | 
			id = 'Navigator' translated
				ifTrue: [sharedFlapTabs add: Flaps newNavigatorFlap].
			id = 'Widgets' translated
				ifTrue: [sharedFlapTabs add: Flaps newWidgetsFlap].
			id = 'Tools' translated
				ifTrue: [sharedFlapTabs add: Flaps newToolsFlap].
			id = 'Squeak' translated
				ifTrue: [sharedFlapTabs add: Flaps newSqueakFlap].
			id = 'Supplies' translated
				ifTrue: [sharedFlapTabs add: Flaps newSuppliesFlap].
			id = 'Stack Tools' translated
				ifTrue: [sharedFlapTabs add: Flaps newStackToolsFlap].
			id = 'Painting' translated
				ifTrue: [sharedFlapTabs add: Flaps newPaintingFlap].
			id = 'Objects' translated
				ifTrue: [sharedFlapTabs add: Flaps newObjectsFlap ]].
	2 timesRepeat: [flapIDs do: [:id | Flaps enableDisableGlobalFlapWithID: id]].
	ActiveWorld flapTabs
		do: [:flapTab | flapTab isCurrentlyTextual
				ifTrue: [flapTab changeTabText: flapTab knownName]].
	Flaps positionNavigatorAndOtherFlapsAccordingToPreference.
	navigationMorph := ActiveWorld findDeeplyA: ProjectNavigationMorph preferredNavigator.
	navigationMorph isNil
		ifTrue: [^ self].
	navigationMorph allMorphs
		do: [:morph | morph class == SimpleButtonDelayedMenuMorph
				ifTrue: [(morph findA: ImageMorph) isNil
						ifTrue: [| label | 
							label := morph label.
							label isNil
								ifFalse: [| name | 
									name := morph knownName.
									name isNil
										ifTrue: [morph name: label.
											name := label].
									morph label: name translated]]]]! !

!MorphicProject methodsFor: 'initialize' stamp: 'mt 8/5/2016 10:56'!
setWorldBackground: force

	((world hasProperty: #hasCustomBackground) and: [force not])
		ifTrue: [^ self].

	"If the user has a custom background, propagate it into the new project."
	((Project current ~~ self and: [Project current isMorphic]) and: [Project current world hasProperty: #hasCustomBackground])
		ifTrue: [
			world fillStyle: Project current world fillStyle.
			world setProperty: #hasCustomBackground toValue: true]
		ifFalse: [
			world removeProperty: #hasCustomBackground.
			world fillStyle: (self userInterfaceTheme background ifNil: [self class defaultFill])].! !

!MorphicProject methodsFor: 'utilities' stamp: 'mt 4/12/2016 19:36'!
showImage: aForm named: imageName
	"Show an image, possibly attached to the pointer for positioning"

	HandMorph attach: (world drawingClass withForm: aForm)
! !

!MorphicProject methodsFor: 'docking bars support' stamp: 'kb 11/19/2009 12:02'!
showWorldMainDockingBar

	^ self projectPreferenceFlagDictionary
		at: #showWorldMainDockingBar
		ifAbsent: [ false ]! !

!MorphicProject methodsFor: 'docking bars support' stamp: 'dtl 10/31/2009 14:22'!
showWorldMainDockingBarString
	^ (self showWorldMainDockingBar
		ifTrue: ['<yes>']
		ifFalse: ['<no>'])
		, 'show main docking bar (M)' translated! !

!MorphicProject methodsFor: 'docking bars support' stamp: 'mt 9/20/2015 16:11'!
showWorldMainDockingBar: aBoolean 
	"Change the receiver to show the main docking bar"
	self projectPreferenceFlagDictionary at: #showWorldMainDockingBar put: aBoolean.
	self assureMainDockingBarPresenceMatchesPreference! !

!MorphicProject methodsFor: 'active process' stamp: 'mt 4/12/2016 19:36'!
spawnNewProcess

	uiProcess := [
		[world doOneCycle.  Processor yield ] repeat.
	] newProcess priority: Processor userSchedulingPriority.
	uiProcess resume! !

!MorphicProject methodsFor: 'active process' stamp: 'dtl 7/13/2013 17:08'!
spawnNewProcessAndTerminateOld: terminate

	self spawnNewProcess.
	terminate
		ifTrue: [Processor terminateActive]
		ifFalse: [Processor activeProcess suspend]! !

!MorphicProject methodsFor: 'active process' stamp: 'mt 7/28/2016 10:28'!
spawnNewProcessIfThisIsUI: suspendedProcess
	"Initialize a UI process if needed. Answer true if suspendedProcess was interrupted
	from a UI process."
	self uiProcess == suspendedProcess ifTrue: [
		self spawnNewProcess.
		^true
	].

	"Ensure that the UI process is running."
	self uiProcess
		ifNil: [self spawnNewProcess]
		ifNotNil: [:p | p isSuspended ifTrue: [
			self restoreDisplay.
			self uiProcess resume]].

	^false		"no new process was created"
! !

!MorphicProject methodsFor: 'file in/out' stamp: 'dtl 3/30/2014 09:51'!
storeSegment
	"Store my project out on the disk as an ImageSegment.  Keep the outPointers in memory.  Name it <project name>.seg.  *** Caller must be holding (Project alInstances) to keep subprojects from going out. ***"

	| is sizeHint |
	(World == world) ifTrue: [^ false]. 
		"self inform: 'Can''t send the current world out'."
	world isInMemory ifFalse: [^ false].  "already done"
	world ifNil: [^ false].  world presenter ifNil: [^ false].

	ScrapBook default emptyScrapBook.
	World checkCurrentHandForObjectToPaste.
	world releaseSqueakPages.
	sizeHint := self projectParameters at: #segmentSize ifAbsent: [0].

	is := ImageSegment new copyFromRootsLocalFileFor: 
				(Array with: world presenter with: world)	"world, and all Players"
			 sizeHint: sizeHint.

	is state = #tooBig ifTrue: [^ false].
	is segment size < 2000 ifTrue: ["debugging" 
		Transcript show: self name, ' only ', is segment size printString, 
			'bytes in Segment.'; cr].
	self projectParameters at: #segmentSize put: is segment size.
	is extract; writeToFile: self name.
	^ true
! !

!MorphicProject methodsFor: 'file in/out' stamp: 'dtl 3/30/2014 09:51'!
storeSegmentNoFile
	"For testing.  Make an ImageSegment.  Keep the outPointers in memory.  Also useful if you want to enumerate the objects in the segment afterwards (allObjectsDo:)"

	| is |
	(World == world) ifTrue: [^ self].		" inform: 'Can''t send the current world out'."
	world isInMemory ifFalse: [^ self].  "already done"
	world ifNil: [^ self].  world presenter ifNil: [^ self].

	"Do this on project enter"
	World flapTabs do: [:ft | ft referent adaptToWorld: World].
		"Hack to keep the Menu flap from pointing at my project"
	"Preferences setPreference: #useGlobalFlaps toValue: false."
	"Utilities globalFlapTabsIfAny do:
		[:aFlapTab | Utilities removeFlapTab: aFlapTab keepInList: false].
	Utilities clobberFlapTabList.	"
	"project world deleteAllFlapArtifacts."
	"self currentWorld deleteAllFlapArtifacts.	"
	ScrapBook default emptyScrapBook.
	World checkCurrentHandForObjectToPaste2.

	is := ImageSegment new copyFromRootsLocalFileFor: 
			(Array with: world presenter with: world)	"world, and all Players"
		sizeHint: 0.

	is segment size < 800 ifTrue: ["debugging" 
		Transcript show: self name, ' did not get enough objects'; cr.  ^ Beeper beep].

	is extract.
	"is instVarAt: 2 put: is segment clone."		"different memory"
! !

!MorphicProject methodsFor: 'subprojects' stamp: 'mt 4/14/2016 21:26'!
subProjects
	"Answer a list of all the subprojects  of the receiver. "
	
	self flag: #fix. "mt: Collect other projects that have this as parent. See Project >> #allProjects"
	^world submorphs 
		select: [:m | (m isSystemWindow) and: [m model isKindOf: Project]]
		thenCollect: [:m | m model].! !

!MorphicProject methodsFor: 'flaps support' stamp: 'dtl 10/31/2009 15:21'!
suppressFlapsString
	^ (self flapsSuppressed
		ifTrue: ['<no>']
		ifFalse: ['<yes>']), 'show shared tabs (F)' translated! !

!MorphicProject methodsFor: 'enter' stamp: 'mt 4/17/2016 21:54'!
suspendProcessForDebug

	| p |
	self assert: Processor activeProcess == uiProcess.
	
	p := uiProcess.
	uiProcess := nil.
	p suspend.! !

!MorphicProject methodsFor: 'enter' stamp: 'mt 4/12/2016 18:19'!
terminateProcessForLeave
	"There is only one UI process. Kill it."
	
	self assert: Processor activeProcess == uiProcess.
	uiProcess := nil.
	Processor terminateActive.! !

!MorphicProject methodsFor: 'utilities' stamp: 'mt 4/12/2016 19:37'!
textWindows
	"Answer a dictionary of all system windows for text display keyed by window title.
	Generate new window titles as required to ensure unique keys in the dictionary."

	| aDict windows title |
	aDict := Dictionary new.
	windows := world submorphs select: [:m | m isSystemWindow].
	windows do:
		[:w | | assoc |
		assoc := w titleAndPaneText.
		assoc ifNotNil:
			[w holdsTranscript ifFalse:
				[title := assoc key.
				(aDict includesKey: title) ifTrue: [ | newKey | "Ensure unique keys in aDict"
					(1 to: 100) detect: [:e |
							newKey := title, '-', e asString.
							(aDict includesKey: newKey) not].
					title := newKey.
					assoc := newKey -> assoc value].
				aDict add: assoc]]].
	^ aDict
! !

!MorphicProject methodsFor: 'flaps support' stamp: 'dtl 10/31/2009 15:23'!
toggleFlapsSuppressed
	"Project current toggleFlapsSuppressed"

	^self flapsSuppressed: self flapsSuppressed not.! !

!MorphicProject methodsFor: 'docking bars support' stamp: 'dtl 10/31/2009 14:22'!
toggleShowWorldMainDockingBar
	self showWorldMainDockingBar: self showWorldMainDockingBar not! !

!MorphicProject methodsFor: 'active process' stamp: 'dtl 7/13/2013 16:40'!
uiProcess
	^uiProcess! !

!MorphicProject methodsFor: 'active process' stamp: 'dtl 7/11/2013 19:37'!
uiProcess: newUIProcess
	^uiProcess := newUIProcess! !

!MorphicProject methodsFor: 'language' stamp: 'tfel 8/3/2016 13:22'!
updateLocaleDependents
	"Set the project's natural language as indicated"

	ActiveWorld allTileScriptingElements do: [:viewerOrScriptor |
			viewerOrScriptor localeChanged].

	Flaps disableGlobalFlaps: false.
	(Preferences eToyFriendly or: [Smalltalk globals at: #SugarNavigatorBar ifPresent: [:c | c showSugarNavigator] ifAbsent: [false]])
		ifTrue: [
			Flaps addAndEnableEToyFlaps.
			ActiveWorld addGlobalFlaps]
		ifFalse: [Flaps enableGlobalFlaps].

	(Project current isFlapIDEnabled: 'Navigator' translated)
		ifFalse: [Flaps enableDisableGlobalFlapWithID: 'Navigator' translated].

	ScrapBook default emptyScrapBook.
	MenuIcons initializeTranslations.
	
	super updateLocaleDependents.

	"self setFlaps.
	self setPaletteFor: aLanguageSymbol."
! !

!MorphicProject methodsFor: 'display' stamp: 'mt 5/6/2016 14:56'!
viewLocFor: exitedProject 
	"Look for a view of the exitedProject, and return its center"

	world submorphsDo: [:v |
			(v isSystemWindow and: [v model == exitedProject])
				ifTrue: [^ v center]].
	^ super viewLocFor: exitedProject! !

!MorphicProject methodsFor: 'enter' stamp: 'dtl 12/28/2009 22:55'!
wakeUpTopWindow
	"Image has been restarted, and the startUp list has been processed. Perform
	any additional actions needed to restart the user interface."

	SystemWindow wakeUpTopWindowUponStartup! !

!Project class methodsFor: '*Morphic-Support' stamp: 'mt 4/14/2016 21:09'!
allMorphicProjects

	^ self allProjects select: [:p | p isMorphic]! !

!SyntaxError class methodsFor: '*Morphic-Support' stamp: 'di 8/17/1998 10:22'!
buildMorphicViewOn: aSyntaxError
	"Answer an Morphic view on the given SyntaxError."
	| window |
	window := (SystemWindow labelled: 'Syntax Error') model: aSyntaxError.

	window addMorph: (PluggableListMorph on: aSyntaxError list: #list
			selected: #listIndex changeSelected: nil menu: #listMenu:)
		frame: (0@0 corner: 1@0.15).

	window addMorph: (PluggableTextMorph on: aSyntaxError text: #contents
			accept: #contents:notifying: readSelection: #contentsSelection
			menu: #codePaneMenu:shifted:)
		frame: (0@0.15 corner: 1@1).

	^ window openInWorldExtent: 380@220! !

!SyntaxError class methodsFor: '*Morphic-Support' stamp: 'dtl 7/9/2013 08:15'!
morphicOpen: aSyntaxError
	"Answer a view whose model is an instance of me."

	self buildMorphicViewOn: aSyntaxError.
	Project current spawnNewProcessIfThisIsUI: Processor activeProcess.
	^ Processor activeProcess suspend! !

PluggableSystemWindow subclass: #PluggableSystemWindowWithLabelButton
	instanceVariableNames: 'labelButton'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Windows'!

!PluggableSystemWindowWithLabelButton commentStamp: 'eem 6/29/2010 19:02' prior: 0!
A PluggableSystemWindowWithLabelButton provides the pull-down menu of browser panes in the multi-pane browser.

Instance Variables
	labelButton:		<BrowserPaneButtonMorph>

labelButton
	- the labelButton is a hack to hold onto the panes of a multi-paned browser so that an inst var doesn't have to be added to the Browser hierarchy.
!

!PluggableSystemWindowWithLabelButton methodsFor: 'resize/collapse' stamp: 'eem 5/6/2008 12:58'!
adjustExtraButton
	labelButton ifNil: [^self].
	labelButton
		position: self innerBounds topLeft + (20 @ -3);
		extent: self innerBounds extent * (0.4@1)! !

!PluggableSystemWindowWithLabelButton methodsFor: 'geometry' stamp: 'eem 5/6/2008 10:45'!
extent: newExtent
	super extent: newExtent.
	self adjustExtraButton! !

!PluggableSystemWindowWithLabelButton methodsFor: 'accessing' stamp: 'eem 6/30/2010 18:56'!
savedMultiWindowState: aSavedMultiWindowState
	labelButton := MultiWindowLabelButtonMorph
						on: aSavedMultiWindowState
						getState: nil
						action: #selectWindowsMenu.
	labelButton
		triggerOnMouseDown: true;
		savedMultiWindowState: aSavedMultiWindowState.
	labelButton
		layoutFrame: (LayoutFrame "i.e. in the left 40% of the title bar"
						fractions: (0@0 corner: 0.4@0)
						offsets: (20@ -20 corner: 0@0));
		hResizing: #spaceFill;
		vResizing: #spaceFill.
	self addMorphFront: labelButton.
	self adjustExtraButton! !

!CustomMenu methodsFor: '*Morphic-invocation' stamp: 'wiz 7/20/2004 12:20'!
startUpWithCaption: caption at: aPoint 
	"Build and invoke this menu with no initial selection. Answer the  
	selection associated with the menu item chosen by the user or nil if  
	none is chosen; use the provided caption"
	^ self startUp: nil withCaption: caption at: aPoint! !

!CustomMenu methodsFor: '*Morphic-invocation' stamp: 'wiz 7/20/2004 12:18'!
startUp: initialSelection withCaption: caption at: aPoint 
	"Build and invoke this menu with the given initial selection and caption. 
	Answer the selection associated with the menu item chosen by the user 
	or nil if none is chosen."
	self build.
	initialSelection notNil
		ifTrue: [self preSelect: initialSelection].
	^ super startUpWithCaption: caption at: aPoint! !

Array variableSubclass: #Cubic
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Morphic-Collections-Arrayed'!

!Cubic commentStamp: 'wiz 6/17/2004 20:31' prior: 0!
I am a segment between to points. In the form of a cubic polynomial that can be evaluated between 0..1 to obtain the end points and intermediate values.
!

!Cubic methodsFor: 'cubic support' stamp: 'wiz 6/17/2004 22:32'!
bestSegments
	"Return the smallest integer number of segments that give the 
	best curve."
	^ self honeIn: self calcEnoughSegments! !

!Cubic methodsFor: 'cubic support' stamp: 'wiz 6/18/2004 23:12'!
calcEnoughSegments
	"Find the power of two that represents a sufficient number of  
	segments for this cubic.  
	The measure is the sum of distances for the segments.  
	We want this to be close enough not affect the straightness of  
	the drawn lines. Which means within one pixel."
	"^ self 
	enough: 2 
	withMeasure: (self measureFor: 1) 
	withIn: self leeway 
	This ran into a problem when the curve was an s-curve with 
	inflections. Besides honeIn will check to see if 1 is better than 
	two so we lose nothing by starting a little higher."
	^ self
		enough: 4
		withMeasure: (self measureFor: 2)
		withIn: self leeway! !

!Cubic methodsFor: 'cubic support' stamp: 'wiz 7/18/2004 22:50'!
enough: nTry withMeasure: lastMeasure withIn: closeEnough
"See comment in calcEnoughSegments for which I am a helper"
	| measure |
	measure := self measureFor: nTry.
	measure > (lastMeasure + closeEnough)
		ifFalse: [^ nTry // 2].
	^ self
		enough: 2 * nTry
		withMeasure: measure
		withIn: closeEnough! !

!Cubic methodsFor: 'cubic support' stamp: 'ul 7/25/2011 01:03'!
honeIn: enough 
	"Find if there is a smaller n than enough that give the same  
	measure for n."
	
	enough < 2 ifTrue: [ ^enough].
	^ self
		honeIn: enough
		step: enough // 2
		measure: (self measureFor: enough)
		withIn: self leeway! !

!Cubic methodsFor: 'cubic support' stamp: 'wiz 6/17/2004 23:45'!
honeIn: centerN step: step measure: measure withIn: closeEnough 
	"Pick the best n by binary search."
	| nTry |
	step < 1
		ifTrue: [^ centerN].
	nTry := centerN - step.
	^ measure > (closeEnough
				+ (self measureFor: nTry))
		ifTrue: [self
				honeIn: centerN
				step: step // 2
				measure: measure
				withIn: closeEnough]
		ifFalse: [self
				honeIn: nTry
				step: step // 2
				measure: measure
				withIn: closeEnough]! !

!Cubic methodsFor: 'cubic support' stamp: 'wiz 6/19/2004 00:00'!
leeway
	"How close can measure be"
	^ 0.1! !

!Cubic methodsFor: 'cubic support' stamp: 'ul 7/25/2011 00:56'!
measureFor: n 
	"Return a distance measure for cubic curve with n segments. 
	For convienence and accuracy we use the sum of the
	distances. "
	"first point is poly of 0."
	| p1 measure |
	p1 := self at: 1.
	measure := 0.
	1 to: n do: [ :i |
		| x p2 |
		x := i asFloat / n.
		p2 := self polynomialEval: x @ x.
		measure := measure + (p2 dist: p1).
		p1 := p2 ].
	^measure! !

!Dictionary methodsFor: '*Morphic-Explorer' stamp: 'ul 11/11/2009 19:39'!
explorerContents

	^self keysSortedSafely replace: [ :key |
		ObjectExplorerWrapper
			with: (self at: key)
			name: (key printString contractTo: 32)
			model: self ]
! !

!Dictionary methodsFor: '*Morphic-Explorer' stamp: 'bf 8/20/1999 15:07'!
hasContentsInExplorer

	^self isEmpty not! !

!BitmapFillStyle methodsFor: '*Morphic-Balloon' stamp: 'dgd 10/17/2003 22:34'!
addFillStyleMenuItems: aMenu hand: aHand from: aMorph
	"Add the items for changing the current fill style of the receiver"
	aMenu add: 'choose new graphic' translated target: self selector: #chooseNewGraphicIn:event: argument: aMorph.
	aMenu add: 'grab new graphic' translated target: self selector: #grabNewGraphicIn:event: argument: aMorph.
	super addFillStyleMenuItems: aMenu hand: aHand from: aMorph.! !

!BitmapFillStyle methodsFor: '*Morphic-Balloon' stamp: 'nk 6/12/2004 09:59'!
chooseNewGraphicIn: aMorph event: evt 
	"Used by any morph that can be represented by a graphic"
	| aGraphicalMenu |
	aGraphicalMenu := GraphicalMenu new
				initializeFor: self
				withForms: aMorph reasonableBitmapFillForms
				coexist: true.
	aGraphicalMenu selector: #newForm:forMorph:;
		 argument: aMorph.
	evt hand attachMorph: aGraphicalMenu! !

!BitmapFillStyle methodsFor: '*Morphic-Balloon' stamp: 'wiz 8/30/2003 16:54'!
grabNewGraphicIn: aMorph event: evt 
	"Used by any morph that can be represented by a graphic"
	| fill |
	fill := Form fromUser.
	fill boundingBox area = 0
		ifTrue: [^ self].
	self form: fill.
	self direction: fill width @ 0.
	self normal: 0 @ fill height.
	aMorph changed! !

!BitmapFillStyle methodsFor: '*Morphic-Balloon' stamp: 'ar 6/25/1999 11:57'!
newForm: aForm forMorph: aMorph
	self form: aForm.
	self direction: (aForm width @ 0).
	self normal: (0 @ aForm height).
	aMorph changed.! !

!PluggableFileList methodsFor: '*Morphic-FileList' stamp: 'dtl 2/2/2010 21:12'!
morphicOpenLabel: aString in: aWorld
	"Open a view of an instance of me."
	"PluggableFileList new morphicOpenLabel: 'foo' in: World"
	| windowMorph volListMorph templateMorph fileListMorph leftButtonMorph middleButtonMorph rightButtonMorph |
	
	self directory: directory.
	windowMorph := (SystemWindow labelled: aString) model: self.

	volListMorph := PluggableListMorph on: self
		list: #volumeList
		selected: #volumeListIndex
		changeSelected: #volumeListIndex:
		menu: #volumeMenu:.
	volListMorph autoDeselect: false.
	windowMorph addMorph: volListMorph frame: (0@0 corner: 0.4@0.5625).

	templateMorph := PluggableTextMorph on: self
		text: #pattern
		accept: #pattern:.
	templateMorph askBeforeDiscardingEdits: false.
	windowMorph addMorph: templateMorph frame: (0@0.5625 corner: 0.4@0.75).

	fileListMorph := PluggableListMorph on: self
		list: #fileList
		selected: #fileListIndex
		changeSelected: #fileListIndex:
		menu: #fileListMenu:.

	windowMorph addMorph: fileListMorph frame: (0.4@0 corner: 1.0@0.75).

	leftButtonMorph := PluggableButtonMorph 
		on: self
		getState: #leftButtonState
		action: #leftButtonPressed.
	leftButtonMorph
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		label: 'Cancel';
		onColor: Color red offColor: Color red;
		feedbackColor: Color orange;
		borderWidth: 3.

	middleButtonMorph := PluggableButtonMorph
		on: self
		getState: nil
		action: nil.
	middleButtonMorph
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		label: prompt;
		onColor: Color lightYellow offColor: Color lightYellow;
		feedbackColor: Color lightYellow;
		borderWidth: 1.

	rightButtonMorph := PluggableButtonMorph
		on: self
		getState: #rightButtonState
		action: #rightButtonPressed.
	rightButtonMorph
		hResizing: #spaceFill;
		vResizing: #spaceFill;
		label: 'Accept';
		onColor: Color green offColor: Color lightYellow;
		feedbackColor: Color black;
		borderWidth: (self canAccept ifTrue: [3] ifFalse: [1]).
	"self canAccept ifFalse: [rightButtonMorph controller: NoController new]."

	windowMorph
		addMorph: leftButtonMorph frame: (0@0.75 corner: 0.25@1.0);
		addMorph: middleButtonMorph frame: (0.25@0.75 corner: 0.75@1.0);
		addMorph: rightButtonMorph frame: (0.75@0.75 corner: 1.0@1.0).

	self changed: #getSelectionSel.

    windowMorph openInWorld: aWorld.
    [windowMorph model notNil]
       whileTrue: [aWorld doOneCycle].
    ^self result
! !

!TTFontDescription methodsFor: '*Morphic-TrueType' stamp: 'ar 11/14/1998 23:47'!
asMorph
	^TTSampleFontMorph font: self! !

!String methodsFor: '*Morphic' stamp: 'ar 4/10/2005 17:07'!
asMorph 
	"Answer the receiver as a StringMorph"

	^ StringMorph contents: self

"'bugs black blood' asMorph openInHand"! !

!String methodsFor: '*Morphic' stamp: 'nk 2/26/2004 13:38'!
asStringMorph 
	"Answer the receiver as a StringMorph"

	^ StringMorph contents: self

"'bugs black blood' asStringMorph openInHand"! !

!String methodsFor: '*Morphic-Explorer' stamp: 'RAA 6/21/1999 11:28'!
hasContentsInExplorer

	^false! !

!String methodsFor: '*Morphic-converting' stamp: 'yo 11/3/2004 19:24'!
openAsMorph
	"Open the receiver as a morph"

	^ self asMorph openInHand ! !

!CompiledMethod methodsFor: '*Morphic-Explorer' stamp: 'eem 5/15/2008 13:14'!
explorerContents
	"(CompiledMethod compiledMethodAt: #explorerContents) explore"
	
	^Array streamContents:
		[:s| | tokens |
		tokens := Scanner new scanTokens: (self headerDescription readStream skipTo: $"; upTo: $").
		s nextPut: (ObjectExplorerWrapper
						with: ((0 to: tokens size by: 2) collect:
								[:i| i = 0 ifTrue: [self header] ifFalse: [{tokens at: i - 1. tokens at: i}]])
						name: 'header'
						model: self).
		(1 to: self numLiterals) do:
			[:key|
			s nextPut: (ObjectExplorerWrapper
							with: (self literalAt: key)
							name: ('literal', key printString contractTo: 32)
							model: self)].
		self isQuick
			ifTrue: [s nextPut: (ObjectExplorerWrapper
									with: self symbolic
									name: #symbolic
									model: self)]
			ifFalse:
				[self symbolicLinesDo:
					[:pc :line|
					pc <= 1
						ifTrue:
							[s nextPut: (ObjectExplorerWrapper
											with: line
											name: 'pragma'
											model: self)]
						ifFalse:
							[s nextPut: (ObjectExplorerWrapper
											with: line
											name: pc printString
											model: self)]]].
				"should be self numLiterals + 1 * Smalltalk wordSize + 1"
		self endPC + 1
			to: self basicSize
			do: [:key|
				s nextPut: (ObjectExplorerWrapper
								with: (self basicAt: key)
								name: key printString
								model: self)]]! !
FontImporterTool initialize!
Editor initialize!
SmalltalkEditor initialize!
TextEditor initialize!
MenuIcons initialize!
CornerGripMorph initialize!
ScrollBar initialize!
CurvierMorph initialize!
ScorePlayerMorph initialize!
TextFieldMorph initialize!
TextMorph initialize!
HandMorph initialize!
ImageMorph initialize!
ProjectViewMorph initialize!
ThreePhaseButtonMorph initialize!
MenuMorph initialize!
Morph initialize!
ColorPickerMorph initialize!
JoystickMorph initialize!
MenuItemMorph initialize!
SystemProgressMorph initialize!
TransferMorph initialize!
MorphicAlarmQueue initialize!
StandardScriptingSystem initialize!
TextContainer initialize!
TheWorldMainDockingBar initialize!
WorldState initialize!
MorphicProject initialize!
